#-------------------------------------------------------------------------
#  RootNewMacros.cmake
#-------------------------------------------------------------------------
cmake_policy(SET CMP0011 NEW ) # See "cmake --help-policy CMP0011" for more details
if(CMAKE_VERSION VERSION_GREATER 2.8.12)
  cmake_policy(SET CMP0022 NEW) # See "cmake --help-policy CMP0022" for more details
endif()

set(lib lib)
set(bin bin)
if(WIN32)
  set(ssuffix .bat)
  set(scomment rem)
  set(libprefix lib)
  set(ld_library_path PATH)
  set(libsuffix .dll)
  set(runtimedir ${CMAKE_INSTALL_BINDIR})
elseif(APPLE)
  set(ld_library_path DYLD_LIBRARY_PATH)
  set(ssuffix .csh)
  set(scomment \#)
  set(libprefix lib)
  set(libsuffix .so)
  set(runtimedir ${CMAKE_INSTALL_LIBDIR})
else()
  set(ld_library_path LD_LIBRARY_PATH)
  set(ssuffix .csh)
  set(scomment \#)
  set(libprefix lib)
  set(libsuffix .so) 
  set(runtimedir ${CMAKE_INSTALL_LIBDIR})
endif()

if(soversion)
  set(ROOT_LIBRARY_PROPERTIES ${ROOT_LIBRARY_PROPERTIES}
      VERSION ${ROOT_MAJOR_VERSION}.${ROOT_MINOR_VERSION}
      SOVERSION ${ROOT_MAJOR_VERSION}
      SUFFIX ${libsuffix}
      PREFIX ${libprefix} )
else()
  set(ROOT_LIBRARY_PROPERTIES ${ROOT_LIBRARY_PROPERTIES}
      SUFFIX ${libsuffix}
      PREFIX ${libprefix}
      IMPORT_PREFIX ${libprefix} )
endif()


#---Modify the behaviour for local and non-local builds-------------------

if(CMAKE_PROJECT_NAME STREQUAL ROOT)
    set(rootcint_cmd rootcling_tmp)
    set(rlibmap_cmd rlibmap)
    set(genreflex_cmd genreflex)
    set(ROOTCINTDEP rootcling_tmp)
else()
    set(rootcint_cmd rootcling)
    set(rlibmap_cmd rlibmap)
    set(genreflex_cmd genreflex)
    set(ROOTCINTDEP)
endif()

set(CMAKE_INCLUDE_CURRENT_DIR OFF)

include(CMakeMacroParseArguments)

#-------------------------------------------------------------------------
#Add "realclean" target
ADD_CUSTOM_TARGET (realclean COMMAND 
                   "${CMAKE_COMMAND}" -P "${PROJECT_SOURCE_DIR}/CmakePackage/Realclean.cmake"
)

ADD_CUSTOM_TARGET (uninstall COMMAND 
                   "${CMAKE_COMMAND}" -P "${PROJECT_SOURCE_DIR}/CmakePackage/cmake_uninstall.cmake"
)
#-------------------------------------------------------------------------




#-------------------------------------------------------------------------
#---ROOT_GET_SOURCES( <variable> cwd <sources> ...)
#  <variable> : cmake variable name
#  cwd : current sources directory
#  <sources> : sources files and directories
# * put the <sources> files in the entry <variable> variable
# * if one of the element in <sources> is a directory then the function
# go in this directory and puts its files in <variable>
#-------------------------------------------------------------------------
function(ROOT_GET_SOURCES variable cwd )
  set(sources)
  foreach( fp ${ARGN})
    if( IS_ABSOLUTE ${fp})
      file(GLOB files ${fp})
    else()
      file(GLOB files RELATIVE ${CMAKE_CURRENT_SOURCE_DIR} ${cwd}/${fp})
    endif()
    if(files)
      foreach(s ${files})
         # Eliminate G__* files only when using wildcards
        if(fp MATCHES "[*]" AND s MATCHES "(^|/)G__") # Eliminate G__* files
        elseif(s MATCHES "${cwd}/G__")
          set(sources ${fp} ${sources})  
        else()
          set(sources ${sources} ${s})
        endif()
      endforeach()
    else()
      if(fp MATCHES "(^|/)G__")
        set(sources ${fp} ${sources})
      else()
        set(sources ${sources} ${fp})
      endif()
    endif()
  endforeach()
  set(${variable} ${sources} PARENT_SCOPE)
endfunction()



#-------------------------------------------------------------------------
#---URANIE_GENERATE_DICTIONARY( dictionary sourcefiles LINKDEF linkdef 
# OPTIONS opt1 opt2 ...)
# dictionary : variable containing the name (string) of the dictionary file
# sourcefiles : variable containing a list of the source files associated
# to the dictionary
# LINKDEF : option flag "LINKDEF"
# linkdef : variable containing the linkdef file associated to the dictionary
# OPTIONS : option flag "OPTIONS"
# opt1, opt2 : options variable (not used, cf the rootcint_cmd command)
# * function wich generates the dictionary and linkdef files
#-------------------------------------------------------------------------
function(URANIE_GENERATE_DICTIONARY dictionary)
  #PARSE_ARGUMENTS(ARG "MODULE;LINKDEF;OPTIONS" ${ARGN})
  CMAKE_PARSE_ARGUMENTS(ARG "STAGE1;MULTIDICT" "MODULE" "LINKDEF;OPTIONS;DEPENDENCIES" ${ARGN})

  set(ARG_TMP)
  #CHANGE_FILE_EXTENSION(*.cxx *.h ARG_TMP "${ARG_DEFAULT_ARGS}")
  CHANGE_FILE_EXTENSION(*.cxx *.h ARG_TMP "${ARG_UNPARSED_ARGUMENTS}")
  CHANGE_FILE_EXTENSION(*.C *.H ARG_TMP "${ARG_TMP}")
  CHANGE_FILE_EXTENSION(*.c *.h ARG_TMP "${ARG_TMP}")
  set(ALL_HEADERS ${ALL_HEADERS} ${ARG_TMP} PARENT_SCOPE)

  #---roottest compability---------------------------------
  if(CMAKE_ROOTTEST_DICT)
    set(CMAKE_INSTALL_LIBDIR ${CMAKE_CURRENT_BINARY_DIR})
    set(libprefix "")
  endif()

  #---Get the list of header files-------------------------
  set(headerfiles)
  #foreach(fp ${ARG_UNPARSED_ARGUMENTS})
  foreach(fp ${ARG_TMP})
    file(GLOB files ${fp})
    if(files)
      foreach(f ${files})
        if(NOT f MATCHES LinkDef)
          set(headerfiles ${headerfiles} ${fp})
        endif()
      endforeach()
    else()
      set(headerfiles ${headerfiles} ${fp})
    endif()
  endforeach()
  set(rheaderfiles "${headerfiles}" )
  #string(REPLACE "${CMAKE_CURRENT_SOURCE_DIR}/inc/" ""  rheaderfiles "${headerfiles}")
  #---Get the list of include directories------------------
  get_directory_property(incdirs INCLUDE_DIRECTORIES)
  if(CMAKE_PROJECT_NAME STREQUAL ROOT)
    set(includedirs -I${CMAKE_SOURCE_DIR}
                    -I${CMAKE_BINARY_DIR}/include)
  else()
    set(includedirs -I${CMAKE_CURRENT_SOURCE_DIR}) 
  endif() 
  foreach( d ${incdirs})    
    ## FGA::2017/01/17 cf post https://root.cern.ch/phpBB3/viewtopic.php?t=5672 dated Fri Oct 12, 2007 12:22
    if(NOT  ${d} STREQUAL "/usr/include")
      if ( CMAKE_VERBOSE_MAKEFILE STREQUAL "ON")
	message(STATUS " *** add [${d}] to the include directories")
      endif ()
   set(includedirs ${includedirs} -I${d})
    endif()
  endforeach()
  list(REMOVE_DUPLICATES includedirs)
  #---Get the list of definitions---------------------------
  get_directory_property(defs COMPILE_DEFINITIONS)
  foreach( d ${defs})
    if((NOT d MATCHES "=") AND (NOT d MATCHES "^[$]<.*>$")) # avoid generator expressions
      set(definitions ${definitions} -D${d})
    endif()
  endforeach()
  #---Get LinkDef.h file------------------------------------
  foreach( f ${ARG_LINKDEF})
    if( IS_ABSOLUTE ${f})
      set(_linkdef ${_linkdef} ${f})
    else()
      if(EXISTS ${CMAKE_CURRENT_SOURCE_DIR}/inc/${f})
        set(_linkdef ${_linkdef} ${CMAKE_CURRENT_SOURCE_DIR}/inc/${f})
      else()
        set(_linkdef ${_linkdef} ${CMAKE_CURRENT_SOURCE_DIR}/${f})
      endif()
    endif()
  endforeach()

  #---Build the names for library, pcm and rootmap file ----
  get_filename_component(dict_base_name ${dictionary} NAME_WE)
  if(dict_base_name MATCHES "^G__")
    string(SUBSTRING ${dictionary} 3 -1 deduced_arg_module)
  else()
    set(deduced_arg_module ${dict_base_name})
  endif()
  
  #---Set the library output directory-----------------------
  if(DEFINED CMAKE_LIBRARY_OUTPUT_DIRECTORY)
    set(library_output_dir ${CMAKE_LIBRARY_OUTPUT_DIRECTORY})
  else()
    set(library_output_dir ${CMAKE_CURRENT_BINARY_DIR})
  endif()

  if(ARG_MODULE)
    set(library_name ${libprefix}${ARG_MODULE}${libsuffix})
    if(ARG_MULTIDICT)
      set(newargs -s ${library_output_dir}/${library_name} -multiDict)
      set(pcm_name ${library_output_dir}/${libprefix}${ARG_MODULE}_${dictionary}_rdict.pcm)
      set(rootmap_name ${library_output_dir}/${libprefix}${deduced_arg_module}.rootmap)
    else()
      set(newargs -s ${library_output_dir}/${library_name})
      set(pcm_name ${library_output_dir}/${libprefix}${ARG_MODULE}_rdict.pcm)
      set(rootmap_name ${library_output_dir}/${libprefix}${ARG_MODULE}.rootmap)
    endif()
  else()
    set(library_name ${libprefix}${deduced_arg_module}${libsuffix})
    set(pcm_name ${dictionary}_rdict.pcm)
    set(rootmap_name ${library_output_dir}/${libprefix}${deduced_arg_module}.rootmap)
  endif()

  if(CMAKE_ROOTTEST_NOROOTMAP)
    set(rootmapname )
    set(rootmapargs )
  else()
    set(rootmapargs -rml ${library_name} -rmf ${rootmap_name})
  endif()

  #---Get the library and module dependencies-----------------
  if(ARG_DEPENDENCIES)
    foreach(dep ${ARG_DEPENDENCIES})
      set(newargs ${newargs} -m  ${libprefix}${dep}_rdict.pcm)
    endforeach()
  endif()

  #---what rootcling command to use--------------------------
  if(ARG_STAGE1)
    set(command rootcling_tmp)
    set(pcm_name)
  else()
    if(CMAKE_PROJECT_NAME STREQUAL ROOT)
      set(command rootcling -rootbuild)
      set(ROOTCINTDEP rootcling)
    else()
      set(command rootcling)
    endif()
  endif()

  #---call rootcint------------------------------------------
  add_custom_command(OUTPUT ${dictionary}.cxx ${pcm_name} ${rootmap_name}
                     COMMAND ${command} -f  ${dictionary}.cxx ${newargs} ${rootmapargs}
                                        -c ${ARG_OPTIONS} ${definitions} 
                                        -I${CMAKE_INSTALL_PREFIX}/include ${includedirs} ${rheaderfiles} 
                                        ${_linkdef}
                     DEPENDS ${headerfiles} ${_linkdef} ${ROOTCINTDEP})
  get_filename_component(dictname ${dictionary} NAME)

  #---roottest compability
  #if(CMAKE_ROOTTEST_DICT OR (NOT DEFINED CMAKE_LIBRARY_OUTPUT_DIRECTORY))
  if(CMAKE_ROOTTEST_DICT)
    add_custom_target(${dictname} DEPENDS ${dictionary}.cxx)
  else()
    add_custom_target(${dictname} DEPENDS ${dictionary}.cxx)

    set_property(GLOBAL APPEND PROPERTY ROOT_DICTIONARY_TARGETS ${dictname})
    set_property(GLOBAL APPEND PROPERTY ROOT_DICTIONARY_FILES ${CMAKE_CURRENT_BINARY_DIR}/${dictionary}.cxx)

    if(ARG_STAGE1)
      install(FILES ${rootmap_name}
                    DESTINATION ${CMAKE_INSTALL_LIBDIR} COMPONENT libraries)
    else()
      install(FILES ${pcm_name} ${rootmap_name}
                    DESTINATION ${runtimedir} COMPONENT libraries)
    endif()
  endif()
endfunction()



#-------------------------------------------------------------------------
#---URANIE_LINKER_LIBRARY( <name> source1 source2 ...[TYPE STATIC|SHARED] 
# [DLLEXPORT] LIBRARIES lib1 lib2 ... DEPENDENCIES lib3 l1b4)
# <name> : library name
# source1, source2 : list of source files contains also the dictionary file
# TYPE STATIC|SHARED : precises if the library is static or dynamic(shared)
# DLLEXPORT : option flag "DLLEXPORT"
# LIBRARIES : option flag "LIBRARIES"
# lib1, lib2 : list of libraries to link
# DEPENDENCIES : option flag "DEPENDENCIES"
# lib3, lib4 : list of dependencies of the library
# * Generates a library file
#-------------------------------------------------------------------------
function(URANIE_LINKER_LIBRARY library)
  CMAKE_PARSE_ARGUMENTS(ARG "DLLEXPORT;CMAKENOEXPORT;TEST" "TYPE" "LIBRARIES;DEPENDENCIES"  ${ARGN})
  if(COMMAND cmake_policy)
    cmake_policy(SET CMP0003 NEW)
  endif(COMMAND cmake_policy)
  ROOT_GET_SOURCES(lib_srcs src ${ARG_UNPARSED_ARGUMENTS})
  
  if ( CMAKE_VERBOSE_MAKEFILE STREQUAL "ON")
      message(STATUS " ============================ Target ${library} : ROOT_GET_SOURCES")
      message(STATUS " ===== ARG_LIBRARIES  : ${ARG_LIBRARIES} ")
      message(STATUS " =============================================")
      message(STATUS " ===== ARG_DEPENDENCIES  : ${ARG_DEPENDENCIES} ")
      message(STATUS " =============================================")
      message(STATUS " =============================================")
      message(STATUS " ===== lib_srcs  : ${lib_srcs} ")
      message(STATUS " ============================ End Of Target ${library} : ROOT_GET_SOURCES ")
  endif()
  
  if(NOT ARG_TYPE)
    set(ARG_TYPE SHARED)
  endif()
  if(ARG_TEST) #building a test, so add EXCLUDE_FROM_ALL
    set(_all EXCLUDE_FROM_ALL)
  endif()
  include_directories(BEFORE ${CMAKE_CURRENT_SOURCE_DIR} ${CMAKE_BINARY_DIR}/include )
  set(library_name ${library})
  if ( CMAKE_VERBOSE_MAKEFILE STREQUAL "ON")
      message(STATUS " === URANIE_LINKER_LIBRARY for ${library_name} ")
  endif()

  if(TARGET ${library})
    message(STATUS "Target ${library} already exists. Renaming target name to ${library}_new")
    set(library ${library}_new)
  endif()
  if(WIN32 AND ARG_TYPE STREQUAL SHARED AND NOT ARG_DLLEXPORT) 
    # --------------------------------------------------
    # ----------------- Case Of Windows ----------------
    # --------------------------------------------------
    # ----- create a list of all the object files ------
    if(CMAKE_GENERATOR MATCHES "Visual Studio")
      #foreach(src1 ${lib_srcs})
      #  if(NOT src1 MATCHES "[.]h$|[.]icc$|[.]hxx$|[.]hpp$")
      #    string (REPLACE ${CMAKE_CURRENT_SOURCE_DIR} "" src2 ${src1})
      #    string (REPLACE ${CMAKE_CURRENT_BINARY_DIR} "" src3 ${src2})     
      #    string (REPLACE ".." "__" src ${src3})     
      #    get_filename_component(name ${src} NAME_WE)
      #    set(lib_objs ${lib_objs} ${library}.dir/${CMAKE_CFG_INTDIR}/${name}.obj)
      #  endif()
      #endforeach()
      set(lib_objs ${lib_objs} ${library}.dir/${CMAKE_CFG_INTDIR}/*.obj)
    else()
      foreach(src1 ${lib_srcs})
        if(NOT src1 MATCHES "[.]h$|[.]icc$|[.]hxx$|[.]hpp$")
          string (REPLACE ${CMAKE_CURRENT_SOURCE_DIR} "" src2 ${src1})
          string (REPLACE ${CMAKE_CURRENT_BINARY_DIR} "" src3 ${src2})           
          string (REPLACE ".." "__" src ${src3})     
          get_filename_component(name ${src} NAME)
          get_filename_component(path ${src} PATH)
          set(lib_objs ${lib_objs} ${CMAKE_CURRENT_BINARY_DIR}/CMakeFiles/${library}.dir/${path}/${name}.obj)
        endif()
      endforeach()
    endif()

    #-----------------modifies arg_dependencies-------------------------------
    set(tmp_arg_dependencies)
    foreach(dep ${ARG_DEPENDENCIES})
        string(LENGTH ${libprefix} lenpref)
        string(SUBSTRING ${dep} 0 ${lenpref} tmp_str )
        string(COMPARE NOTEQUAL ${tmp_str} ${libprefix} comp)
        string(FIND ${dep} "Uranie" comp1)
        string(FIND ${dep} "pcl" comp2)
        string(FIND ${dep} "pthread" comp3)		
		string(FIND ${dep} "python" comp4)
		string(FIND ${dep} "libnlopt" comp5)
        string(FIND ${dep} "jsoncpp" comp6)
		string(COMPARE EQUAL "-1" ${comp2} comp2)
		string(COMPARE EQUAL "-1" ${comp4} comp4)
		string(COMPARE EQUAL "-1" ${comp5} comp5)
		string(COMPARE EQUAL "-1" ${comp6} comp6)
        if(${comp} AND ${comp1} AND ${comp2} AND ${comp3} AND ${comp4} AND ${comp5} AND ${comp6})
            set(tmp_arg_dependencies ${tmp_arg_dependencies} ${libprefix}${dep})
        else()
            set(tmp_arg_dependencies ${tmp_arg_dependencies} ${dep})
        endif()
    endforeach(dep ${ARG_DEPENDENCIES})
    set(ARG_DEPENDENCIES  ${tmp_arg_dependencies})

    #---create a shared library with the .def file------------------------
    add_library(${library} SHARED ${_all} ${lib_srcs})
    install(FILES $<TARGET_PDB_FILE:${library}> DESTINATION bin OPTIONAL)
    target_link_libraries(${library} ${ARG_LIBRARIES} ${ARG_DEPENDENCIES})
    set_target_properties(${library} PROPERTIES ${ROOT_LIBRARY_PROPERTIES} LINK_FLAGS -DEF:${library}.def)
    #---set the .def file as generated----------------------------------------
    set_source_files_properties(${library}.def PROPERTIES GENERATED 1)
    #---create a custom pre-link command that runs bindexplib-----------------
    add_custom_command(TARGET ${library} PRE_LINK
                       COMMAND bindexplib
                       ARGS -o ${library}.def ${libprefix}${library} ${lib_objs}
                       DEPENDS bindexplib )
  else()
    # --------------------------------------------------
    # ------------------ Case Of Unix ------------------
    # --------------------------------------------------
    #3.9.0 add_library( ${library} ${ARG_TYPE} ${lib_srcs})
    add_library( ${library} ${ARG_TYPE} ${_all} ${lib_srcs})
    if(ARG_TYPE STREQUAL SHARED)
      set_target_properties(${library} PROPERTIES  ${ROOT_LIBRARY_PROPERTIES})

    endif()
    if(explicitlink OR ROOT_explicitlink_FOUND)
      target_link_libraries(${library} ${ARG_LIBRARIES} ${ARG_DEPENDENCIES})
    else()
      target_link_libraries(${library} ${ARG_LIBRARIES})
    endif()
  endif()
  if(TARGET G__${library})
    add_dependencies(${library} G__${library})
  endif()
  if(TARGET move_headers)
    add_dependencies(${library} move_headers)
  endif()  
  set_property(GLOBAL APPEND PROPERTY ROOT_EXPORTED_TARGETS ${library})
  set_target_properties(${library} PROPERTIES OUTPUT_NAME ${library_name})
  set_target_properties(${library} PROPERTIES LINK_INTERFACE_LIBRARIES "${ARG_DEPENDENCIES}")
  #----Installation details-------------------------------------------------
  #if(NOT ARG_TEST AND CMAKE_LIBRARY_OUTPUT_DIRECTORY)
  if(NOT ARG_TEST)
    if(ARG_CMAKENOEXPORT)
      install(TARGETS ${library} RUNTIME DESTINATION ${CMAKE_INSTALL_BINDIR} COMPONENT libraries
                                 LIBRARY DESTINATION ${CMAKE_INSTALL_LIBDIR} COMPONENT libraries
                                 ARCHIVE DESTINATION ${CMAKE_INSTALL_LIBDIR} COMPONENT libraries)
    else()
      install(TARGETS ${library} EXPORT ${CMAKE_PROJECT_NAME}Exports
                                 RUNTIME DESTINATION ${CMAKE_INSTALL_BINDIR} COMPONENT libraries
                                 LIBRARY DESTINATION ${CMAKE_INSTALL_LIBDIR} COMPONENT libraries
                                 ARCHIVE DESTINATION ${CMAKE_INSTALL_LIBDIR} COMPONENT libraries)
    endif()
  endif()
endfunction(URANIE_LINKER_LIBRARY)



#-------------------------------------------------------------------------
#---URANIE_USE_PACKAGE( package )
# package : variable containing the relatie path to the repertory to include,
# this path is relative from the root of sources repertories corresponding
# to the cmake variable CMAKE_SOURCE_DIR
# * if the repertory contains a CMakeLists.txt file, this macro includes 
# the repertory and all its subdirectories containing a CMakeLists.txt
# file
#-------------------------------------------------------------------------
macro( URANIE_USE_PACKAGE package )
  if(IntegratedBuild)
    if( EXISTS ${CMAKE_SOURCE_DIR}/${package}/CMakeLists.txt)
      set(_use_packages ${_use_packages} ${package}) 
      include_directories( ${CMAKE_SOURCE_DIR}/${package} ) 
      set_property(GLOBAL APPEND PROPERTY ROOT_BUILDTREE_PACKAGES ${package})
      file(READ ${CMAKE_SOURCE_DIR}/${package}/CMakeLists.txt file_contents)
      string( REGEX MATCHALL "URANIE_USE_PACKAGE[ ]*[(][ ]*([^ )])+" vars ${file_contents})
      foreach( var ${vars})
        string(REGEX REPLACE "URANIE_USE_PACKAGE[ ]*[(][ ]*([^ )])" "\\1" p ${var})
        #---avoid calling the same one at the same directory level --------------------------
        list(FIND _use_packages ${p} _done)
        if(_done EQUAL -1)
          URANIE_USE_PACKAGE(${p})
        endif()
      endforeach()
    endif()
  endif()

endmacro(URANIE_USE_PACKAGE)

#-------------------------------------------------------------------------
#---URANIE_GENERATE_ROOTMAP( library LINKDEF linkdef LIBRARY lib 
# DEPENDENCIES lib1 lib2 )
# library : name of rootmap file
# LINKDEF : option flag "LINKDEF"
# linkdef : variable containing the names of the linkdef files
# LIBRARY : option flag "LIBRARY"
# lib : list of libraries to link
# DEPENDENCIES : option flag "DEPENDENCIES"
# lib1, lib2 : list of dependencies of the library
# * Creates a target "map${libname}" to generate a rootmap file
#-------------------------------------------------------------------------
function(URANIE_GENERATE_ROOTMAP library)
  return()  #--- Not needed anymore
  PARSE_ARGUMENTS(ARG "LINKDEF;LIBRARY;DEPENDENCIES" "" ${ARGN})
  cmake_policy(SET CMP0002 OLD)
  set(CMAKE_LIBRARY_OUTPUT_DIRECTORY ${CMAKE_BINARY_DIR}/lib)
  file(MAKE_DIRECTORY ${CMAKE_LIBRARY_OUTPUT_DIRECTORY})

  get_filename_component(libname ${library} NAME_WE)
  get_filename_component(path ${library} PATH)
  set(outfile ${CMAKE_LIBRARY_OUTPUT_DIRECTORY}/${libprefix}${libname}.rootmap)
  foreach( f ${ARG_LINKDEF})
    if( IS_ABSOLUTE ${f})
      set(_linkdef ${_linkdef} ${f})
    else() 
      set(_linkdef ${_linkdef} ${CMAKE_CURRENT_SOURCE_DIR}/${f})
    endif()
  endforeach()
  foreach(d ${ARG_DEPENDENCIES})
  
    get_filename_component(_ext ${d} EXT)
    if(IS_ABSOLUTE ${d})
       string(FIND ${d} "/" deb REVERSE)
       string(SUBSTRING 
               ${d}
               ${deb}
               -1
               tmp_d
              )
    string(REGEX REPLACE "/" "" d ${tmp_d})
    endif(IS_ABSOLUTE ${d})
	
    if(_ext)
      set(_dependencies ${_dependencies} ${d})
    else()
      set(_dependencies ${_dependencies} ${libprefix}${d}${CMAKE_SHARED_LIBRARY_SUFFIX})
    endif()
	
  endforeach()
  if(ARG_LIBRARY)
    set(_library ${ARG_LIBRARY})
  else()
    set(_library ${libprefix}${library}${CMAKE_SHARED_LIBRARY_SUFFIX})
  endif()
  
  #---Build the rootmap file------------------------------------------------
  add_custom_target(map${libname} ${rlibmap_cmd} -f -o ${outfile} -l ${_library} -d ${_dependencies} -c ${_linkdef}) 

  #---Add dependencies ------------------------------------------------
  add_dependencies(map${libname} ${libname})
 
  #---Define the installation target  ------------------------------------
  install(FILES ${outfile} DESTINATION ${CMAKE_INSTALL_PREFIX}/lib)
endfunction(URANIE_GENERATE_ROOTMAP)

#-------------------------------------------------------------------------
#---URANIE_INSTALL_HEADERS([dir1 dir2 ...])
# dir1, dir2 : list of header files
# * Add to the target "install" a list of files to install in the
# ${CMAKE_INSTALL_PREFIX}/include repertory 
#-------------------------------------------------------------------------
function(URANIE_INSTALL_HEADERS head)

install(FILES ${head} ${ARGN} DESTINATION ${CMAKE_INSTALL_PREFIX}/include
                PERMISSIONS OWNER_READ GROUP_READ WORLD_READ OWNER_WRITE GROUP_WRITE OWNER_EXECUTE GROUP_EXECUTE        
  )  

endfunction(URANIE_INSTALL_HEADERS)



#-------------------------------------------------------------------------
#---ROOT_EXECUTABLE( <name> source1 source2 ... LIBRARIES library1 
#library2 ...)
# <name> : name of the executable
# source1, source2 : list of source files compiled to generate the 
#executable
# LIBRARIES : option flag "LIBRARIES"
# library1, library2 : list of libraries to link to the executable
# * Generates a executable from the sources given and links it to
# the libraries given
#-------------------------------------------------------------------------
function(URANIE_EXECUTABLE executable)
  if(COMMAND cmake_policy)
      cmake_policy(SET CMP0003 NEW)
  endif(COMMAND cmake_policy)
  PARSE_ARGUMENTS(ARG "LIBRARIES" "CMAKENOEXPORT;NOINSTALL" ${ARGN})
  ROOT_GET_SOURCES(exe_srcs src ${ARG_DEFAULT_ARGS})
  set(executable_name ${executable})
  if(TARGET ${executable})
    message("Target ${executable} already exists. Renaming target name to ${executable}_new")
    set(executable ${executable}_new)
  endif()
  include_directories(BEFORE ${CMAKE_CURRENT_SOURCE_DIR} ${CMAKE_BINARY_DIR}/include )
  add_executable( ${executable} ${exe_srcs})
  if(WIN32)
	set(tmp_arg_libraries)
    foreach(dep ${ARG_LIBRARIES})
        string(LENGTH ${libprefix} lenpref)
        string(SUBSTRING ${dep} 0 ${lenpref} tmp_str )
        string(COMPARE NOTEQUAL ${tmp_str} ${libprefix} comp)

        set(list_exception "Uranie" "pcl" "pthread" "cppunit" "uranie_commands" "code_paths"
          "python_commands" "mpi_commands" "uranie_hooks")
		set(comp1 1)
		foreach(except ${list_exception})
			string(FIND ${dep} ${except} comp2)
			if(${comp1} AND ${comp2})
				set(comp1 1)
			else(${comp1} AND ${comp2})
				set(comp1 0)
			endif(${comp1} AND ${comp2})
		endforeach(except ${list_exception}) 
		
		if(IS_ABSOLUTE ${dep})
			set(comp FALSE)
		endif()
        if(${comp} AND ${comp1})
            set(tmp_arg_libraries ${tmp_arg_libraries} ${libprefix}${dep})
        else()
            set(tmp_arg_libraries ${tmp_arg_libraries} ${dep})
        endif()
    endforeach(dep ${ARG_LIBRARIES})
    set(ARG_LIBRARIES  ${tmp_arg_libraries})
  endif(WIN32)
  target_link_libraries(${executable} ${ARG_LIBRARIES} )
  if(WIN32 AND ${executable} MATCHES .exe)  
    set_target_properties(${executable} PROPERTIES SUFFIX "")
  endif()
  set_property(GLOBAL APPEND PROPERTY ROOT_EXPORTED_TARGETS ${executable})
  set_target_properties(${executable} PROPERTIES OUTPUT_NAME ${executable_name})
#----Installation details-------------------------------------------------
  if(NOT ARG_NOINSTALL)
    if(ARG_CMAKENOEXPORT)
      install(TARGETS ${executable} RUNTIME DESTINATION ${CMAKE_INSTALL_BINDIR} COMPONENT applications)
    else()
      install(TARGETS ${executable} EXPORT ${CMAKE_PROJECT_NAME}Exports RUNTIME DESTINATION ${CMAKE_INSTALL_BINDIR} COMPONENT applications)
    endif()
  endif()
endfunction(URANIE_EXECUTABLE)



#-------------------------------------------------------------------------
#---SET_RUNTIME_PATH( var [LD_LIBRARY_PATH | PATH] )
#-------------------------------------------------------------------------
# function( SET_RUNTIME_PATH var pathname)
#   set( dirs ${CMAKE_LIBRARY_OUTPUT_DIRECTORY}/${CMAKE_CFG_INTDIR})
#   get_property(found_packages GLOBAL PROPERTY PACKAGES_FOUND)
#   get_property(found_projects GLOBAL PROPERTY PROJECTS_FOUND)
#   foreach( package ${found_projects} ${found_packages} )
#      foreach( env ${${package}_environment})
#          if(env MATCHES "^${pathname}[+]=.*")
#             string(REGEX REPLACE "^${pathname}[+]=(.+)" "\\1"  val ${env})
#             set(dirs ${dirs} ${val})
#          endif()
#      endforeach()
#   endforeach()
#   if(WIN32)
#     string(REPLACE ";" "[:]" dirs "${dirs}")
#   else()
#     string(REPLACE ";" ":" dirs "${dirs}")
#   endif()
#   set(${var} "${dirs}" PARENT_SCOPE)
# endfunction()



#----------------------------------------------------------------------------
#------URANIE_ADD_TEST()
# * add test module if the cppunit library is found
#----------------------------------------------------------------------------
macro(URANIE_ADD_TEST)

  if(CPPUNIT_FOUND)
    message(STATUS "  >> Building ${libname} Unitary Tests...")
    add_subdirectory(teST)
    message(STATUS "  >> Building ${libname} Unitary Tests... - Built")
  else(CPPUNIT_FOUND)
    message(STATUS "CPPUNIT is not present in your PATH/LD_LIBRARY_PATH env variable. 'make test' cannot be performed!")
  endif(CPPUNIT_FOUND)

endmacro(URANIE_ADD_TEST)

#----------------------------------------------------------------------------


#----------------------------------------------------------------------------
#------URANIE_ADD_TEST_EXECUTABLE()
# *  Adds the test executable per OS
#----------------------------------------------------------------------------
macro(URANIE_ADD_TEST_EXECUTABLE )

    if(WIN32)
        add_test(${ARGV0} ${CMAKE_COMMAND} -E chdir 
                    ${CMAKE_BINARY_DIR}${ARGV1}  
                ${CMAKE_INSTALL_PREFIX}/${runtimedir}/${ARGV2})
    endif(WIN32)
    if(UNIX)
        add_test(${ARGV0}  
                ${CMAKE_BINARY_DIR}${ARGV1}/${ARGV2})
    endif(UNIX)

    #---Add dependencies ------------------------------------------------
    add_dependencies(map ${ARGV2})

endmacro(URANIE_ADD_TEST_EXECUTABLE)

#----------------------------------------------------------------------------


#----------------------------------------------------------------------------
#------URANIE_ADD_TESTDATA_FILE( module_path )
# module_path : relative path of the repertory where the files to be copied
# are searched (relative path from root source repertory)
# * copy all the files with the extensions: .ref, .dat, .C, .in, .py, .wei,
# .pmml, .csv, .xml, .dtd, .json . The files are copied in the same relative path
# repertory from the build directory
#----------------------------------------------------------------------------
macro(URANIE_ADD_TESTDATA_FILE module_path)

  if(NOT ${URANIE_SOURCE_DIR} STREQUAL ${URANIE_BINARY_DIR})
  # Fichiers pour le module
    file(COPY ${URANIE_SOURCE_DIR}${module_path}/  DESTINATION ${URANIE_BINARY_DIR}${module_path}/  FILES_MATCHING PATTERN "*.ref" PATTERN "CVS" EXCLUDE )
    file(COPY ${URANIE_SOURCE_DIR}${module_path}/  DESTINATION ${URANIE_BINARY_DIR}${module_path}/  FILES_MATCHING PATTERN "*.dat" PATTERN "CVS" EXCLUDE )
    file(COPY ${URANIE_SOURCE_DIR}${module_path}/  DESTINATION ${URANIE_BINARY_DIR}${module_path}/  FILES_MATCHING PATTERN "*.C" PATTERN "CVS" EXCLUDE )
    file(COPY ${URANIE_SOURCE_DIR}${module_path}/  DESTINATION ${URANIE_BINARY_DIR}${module_path}/  FILES_MATCHING PATTERN "*.in" PATTERN "CVS" EXCLUDE )
    file(COPY ${URANIE_SOURCE_DIR}${module_path}/  DESTINATION ${URANIE_BINARY_DIR}${module_path}/  FILES_MATCHING PATTERN "*.py" PATTERN "CVS" EXCLUDE )
    file(COPY ${URANIE_SOURCE_DIR}${module_path}/  DESTINATION ${URANIE_BINARY_DIR}${module_path}/  FILES_MATCHING PATTERN "*.wei" PATTERN "CVS" EXCLUDE )
    file(COPY ${URANIE_SOURCE_DIR}${module_path}/  DESTINATION ${URANIE_BINARY_DIR}${module_path}/  FILES_MATCHING PATTERN "*.pmml" PATTERN "CVS" EXCLUDE )
    file(COPY ${URANIE_SOURCE_DIR}${module_path}/  DESTINATION ${URANIE_BINARY_DIR}${module_path}/  FILES_MATCHING PATTERN "*.csv" PATTERN "CVS" EXCLUDE )
    file(COPY ${URANIE_SOURCE_DIR}${module_path}/  DESTINATION ${URANIE_BINARY_DIR}${module_path}/  FILES_MATCHING PATTERN "*.xml" PATTERN "CVS" EXCLUDE )
    file(COPY ${URANIE_SOURCE_DIR}${module_path}/  DESTINATION ${URANIE_BINARY_DIR}${module_path}/  FILES_MATCHING PATTERN "*.dtd" PATTERN "CVS" EXCLUDE )
    file(COPY ${URANIE_SOURCE_DIR}${module_path}/  DESTINATION ${URANIE_BINARY_DIR}${module_path}/  FILES_MATCHING PATTERN "*.cpp" PATTERN "CVS" EXCLUDE )
    file(COPY ${URANIE_SOURCE_DIR}${module_path}/  DESTINATION ${URANIE_BINARY_DIR}${module_path}/  FILES_MATCHING PATTERN "*.f" PATTERN "CVS" EXCLUDE )
    file(COPY ${URANIE_SOURCE_DIR}${module_path}/  DESTINATION ${URANIE_BINARY_DIR}${module_path}/  FILES_MATCHING PATTERN "*.abs" PATTERN "CVS" EXCLUDE )
    file(COPY ${URANIE_SOURCE_DIR}${module_path}/  DESTINATION ${URANIE_BINARY_DIR}${module_path}/  FILES_MATCHING PATTERN "*.json" PATTERN "CVS" EXCLUDE )
    file(COPY ${URANIE_SOURCE_DIR}${module_path}/  DESTINATION ${URANIE_BINARY_DIR}${module_path}/  FILES_MATCHING PATTERN "*.fmu" PATTERN "CVS" EXCLUDE )

  endif(NOT ${URANIE_SOURCE_DIR} STREQUAL ${URANIE_BINARY_DIR})  

endmacro(URANIE_ADD_TESTDATA_FILE)

#----------------------------------------------------------------------------

#----------------------------------------------------------------------------
#-----URANIE_WRITE_TEST_FILE()
# * copy files needed in the main CMakeLists.txt to use ctest for unitary 
# testing
#----------------------------------------------------------------------------
macro(URANIE_WRITE_TEST_FILE)

    if(NOT ${URANIE_SOURCE_DIR} STREQUAL ${URANIE_BINARY_DIR})
        if(EXISTS ${URANIE_SOURCE_DIR}/Testing.cmake AND EXISTS ${URANIE_SOURCE_DIR}/CTestConfig.cmake)
            if(WIN32)
                file(STRINGS ${URANIE_SOURCE_DIR}/Testing.cmake myfile)
                set(mymodifiedfile "")
                file(WRITE ${URANIE_BINARY_DIR}/Testing.cmake ${mymodifiedfile})
                foreach(str ${myfile})
                    string(COMPARE NOTEQUAL ${str} "" notempty)
                    if(notempty)
                        string(REGEX REPLACE "CTEST_SCRIPT_DIRECTORY" "URANIE_SOURCE_DIR" str_tmp ${str})
                        file(APPEND ${URANIE_BINARY_DIR}/Testing.cmake ${str_tmp}\n)
                    endif(notempty)
                endforeach(str)
                if(EXISTS ${URANIE_SOURCE_DIR}/TestingDI.cmake)
                    file(STRINGS ${URANIE_SOURCE_DIR}/TestingDI.cmake myfile)
                    set(mymodifiedfile "")
                    file(WRITE ${URANIE_BINARY_DIR}/TestingDI.cmake ${mymodifiedfile})
                    foreach(str ${myfile})
                        string(COMPARE NOTEQUAL ${str} "" notempty)
                        if(notempty)
                            string(REGEX REPLACE "CTEST_SCRIPT_DIRECTORY" "URANIE_SOURCE_DIR" str_tmp ${str})
                            file(APPEND ${URANIE_BINARY_DIR}/TestingDI.cmake ${str_tmp}\n)
                        endif(notempty)
                    endforeach(str)
                 endif(EXISTS ${URANIE_SOURCE_DIR}/TestingDI.cmake)
            endif(WIN32)
            if(UNIX)
                EXEC_PROGRAM(sed ARGS 's%\${CTEST_SCRIPT_DIRECTORY}%${URANIE_SOURCE_DIR}%g' ${URANIE_SOURCE_DIR}/Testing.cmake > ${URANIE_BINARY_DIR}/Testing.cmake OUTPUT_VARIABLE BLA)
                 if(EXISTS ${URANIE_SOURCE_DIR}/TestingDI.cmake)
                      EXEC_PROGRAM(sed ARGS 's%\${CTEST_SCRIPT_DIRECTORY}%${URANIE_SOURCE_DIR}%g' ${URANIE_SOURCE_DIR}/TestingDI.cmake > ${URANIE_BINARY_DIR}/TestingDI.cmake OUTPUT_VARIABLE BLA)
                 endif(EXISTS ${URANIE_SOURCE_DIR}/TestingDI.cmake)
            endif(UNIX)
            file(COPY ${URANIE_SOURCE_DIR}/CTestConfig.cmake  DESTINATION ${URANIE_BINARY_DIR})
            if(UNIX)
              file(COPY ${URANIE_SOURCE_DIR}/CTestCustom.cmake DESTINATION ${URANIE_BINARY_DIR})
            endif()
        endif(EXISTS ${URANIE_SOURCE_DIR}/Testing.cmake AND EXISTS ${URANIE_SOURCE_DIR}/CTestConfig.cmake)
        file(COPY ${URANIE_SOURCE_DIR}/testReport.xsl     DESTINATION ${URANIE_BINARY_DIR})
        file(COPY ${URANIE_SOURCE_DIR}/UTReport.xml       DESTINATION ${URANIE_BINARY_DIR})
        file(COPY ${URANIE_SOURCE_DIR}/UTReport.xsl       DESTINATION ${URANIE_BINARY_DIR})
    endif(NOT ${URANIE_SOURCE_DIR} STREQUAL ${URANIE_BINARY_DIR})


endmacro(URANIE_WRITE_TEST_FILE)


#----------------------------------------------------------------------------
#------URANIE_INCLUDE_DIRECTORIES()
# * Replace the lines in CMakelist files : 
# set(INCLUDE_DIRECTORIES ${dir_1} ...)/include_directories(INCLUDE_DIRECTORIES)
# Use only URANIE_INCLUDE_DIRECTORIES({dir_1} {dir_2} ...)
# dir_1 dir_2 : list of the repertory path to include in INCLUDE_DIRECTORIES
#----------------------------------------------------------------------------

macro(URANIE_INCLUDE_DIRECTORIES)

    PARSE_ARGUMENTS(ARG "LIST"  "" ${ARGN})
    include_directories(${ARG_DEFAULT_ARGS})

    if ( CMAKE_VERBOSE_MAKEFILE STREQUAL "ON")
 
        get_property(dirs DIRECTORY ${CMAKE_CURRENT_SOURCE_DIR} PROPERTY INCLUDE_DIRECTORIES)
        message(STATUS "  INCLUDE_DIRECTORIES = '${dirs}'")
    
    endif ()



endmacro(URANIE_INCLUDE_DIRECTORIES)

