#! /usr/bin/python
#------------------------------------------------------------------------------#
# This script is generating aliases for URANIE namespaces in order to simplify 
# there access via the Python language.
# The idea is to use URANIE's LinkDef files to create a list of Python modules 
# which can then be imported in Python scripts. This allows to write:
#
#   from URANIE import DataServer
#
# while before you had to write:
# 
#   import ROOT
#   DataServer = ROOT.URANIE.DataServer
#
# It is even possible to do the highly discouraged:
#
#   from URANIE.DataServer import *
#
# which gives direct access to TDataServer and friends. This was impossible 
# before. 
#
# To be able to do this, the script needs $URANIESYS to be set, or to be given 
# the proper installation directory as parameter (default is current directory).
# If needed, it will create:
#   - $URANIESYS/lib/python/URANIE directory containing the submodules
#   - $URANIESYS/lib/python/loadmodules.py which is a script automatically 
#     loading the standard modules
#   - $URANIESYS/pyuranie which is an executable lauching python's command line 
#     after running the loadmodule.py script.
#
# NOTE: All classes are not loaded via this system. However, they are unlikely 
#       to be used. In case they need an access anyway, they are still available 
#       via the "old" way, i.e. ROOT.URANIE.NameSpace.ClassName.  
#
# Nicolas Gilardi, 2010.02.24
# Modifications
# * Cedric Forgues (2010-10-27)
#   - The script pyuranie is now executable from nowhere if URANIESYS system variable is set.
#------------------------------------------------------------------------------#



import sys, os, stat
import os.path

def readLinkDef(fileName):
    '''
    Reads a ROOT LinkDef file and returns the list of class and struct names.
    The "C++ namespace" format is converted into a "Python module" format 
    (e.g. "namespace::classname" becomes "namespace.classname").
    '''
    lines = [x.strip() for x in open(fileName).readlines()]
    classList = []
    defcount = 0
    listDef = []
    listDefFound = []
    
    for l in lines:
        if "#pragma " in l:
            if "//" in l:
                # commented lines are skipped
                continue
            elif "list<" in l:
                # std::list class are not easy to handle, especially when they 
                # contain C++ pointers.
                continue                
            else:
                if " class " in l:
                    tmpLine = l.split("class")
                elif " struct " in l:
                    tmpLine = l.split("struct")                
                else:
                    # we only care about class and struct
                    continue
                
                if defcount == 0:
                    classList.append(tmpLine[-1].strip())
                    # Clean up the class names
                    classList[-1] = classList[-1].replace(";","")
                    classList[-1] = classList[-1].replace("+","")
                    classList[-1] = classList[-1].replace("-","")
                    classList[-1] = classList[-1].replace("::",".")
                    pass
                pass
                
                
        elif "#ifdef" in l:
            tmpLine = l.split("ifdef")
            listDef.append(tmpLine[-1].strip())
            if tmpLine[-1].strip() not in uranieListDef:
                listDefFound.append(tmpLine[-1].strip())
                defcount = defcount + 1
                pass
            pass
        elif "#ifndef" in l:
            tmpLine = l.split("ifndef")
            listDef.append(tmpLine[-1].strip())
            if tmpLine[-1].strip() in uranieListDef:
                listDefFound.append(tmpLine[-1].strip())
                defcount = defcount + 1
                pass
            pass
        elif "#endif" in l:
            if listDef.pop() in listDefFound:
                defcount = defcount - 1
                pass
            pass
        pass
               
    return classList
            

def findLinkDef(classList, dirname, names):
    '''
    Finds the LinkDef files in "names" and append the new class names found to "classList".
    The directory name "dirname" is checked to verify that it is a valid source 
    repository. 
    This function is to be called by the os.path.walk() function.
    '''
    if "ihm" in dirname:
        # Command line and scripting don't really need IHM classes, do they ?
        pass
    else:    
        exceptions =["headingLinkDef.h","TXMLProblemLinkDef.h","TSumOfSquaresLinkDef.h"]
        dirList = dirname.split(os.sep)
        if "souRCE" in dirList[-1]:
            # We verify that we are in a valid source directory where a usefull 
            # LinkDef can be found.
            for n in names: 
                if "~" not in n and "LinkDef.h" in n:
                    if n not in exceptions:
                        # Verify that the file is a valid LinkDef
                        print(dirname + os.sep + n)
                        for className in readLinkDef(dirname + os.sep + n): 
                            if className not in classList:
                                classList.append(className)
                                pass
                            pass
                        pass
    
                                         
if __name__ == "__main__":   
        
    # Set the uraniePath variable
    if len(sys.argv) > 1:
        uranieSrc = sys.argv[1]
    else:
        uranieSrc = os.getcwd()
        print("*** WARNING: no path given for URANIE source. Using %s"%uranieSrc)

    if len(sys.argv) > 2:
        uranieDest = sys.argv[2]
    elif 'URANIESYS' in os.environ:
        uranieDest = os.environ['URANIESYS']
    else:
        uranieDest = os.getcwd()
        print("*** WARNING: no path given for URANIE installation. Using %s"%uranieDest)
             
    if len(sys.argv) > 3:
        uranieDef = sys.argv[3]
        print(uranieDef)
        pass
    else:
        uranieDef = ""
        pass
    pass

    tmpList = uranieDef.split('-D')
    uranieListDef = []
    for i in range(len(tmpList)):
        uranieListDef.append(tmpList[i].strip())
    
    
    #--------------------------------------#
    # Retrieve the modules and class names #
    #--------------------------------------#
    print("Now looking for LinkDef files in %s"%uranieSrc)
    
    classList = []
    for dpath, dname, files in os.walk(uranieSrc) :
        findLinkDef(classList, dpath, files)
    
    if len(classList) > 0:
        # Extract the modules names
        moduleList = []
        for c in classList:
            if "std" in c:
                pass
            elif "URANIE" in c:
                tmpName = c.split(".")
                if tmpName[1] not in moduleList:
                    moduleList.append(tmpName[1])
        
        #----------------------#
        # Building the package #
        #----------------------#

        modulePath = str.join(os.sep, [uranieDest,"python","URANIE"])
        if not os.path.exists(modulePath):
            os.makedirs(modulePath)
            
        initLines = []
        
        # Building the modules files
        for m in moduleList:
            lines = ["import ROOT\n\n"]
            for c in classList:
                if "." in c and m in c.split(".")[1] and "std" not in c:
                    className = c.split(".")[-1]
                    lines.append("%s = ROOT.%s\n"%(className, c))
            f = open(modulePath + os.sep + "%s.py"%m, "w")
            f.writelines(lines)
            f.close()    
        
        # Building the __init__.py file       
        f = open(modulePath + os.sep + "__init__.py","w")
        f.write("import ROOT\n")
        f.close()

        # if not os.path.exists(str.join(os.sep, [uranieDest, "python","loadmodules.py"])):            
        #     # Create a python script loading the main modules
        #     f = open(str.join(os.sep, [uranieDest, "python", "loadmodules.py"]), "w")
        #     f.write('from array import array\n')
        #     f.write('import URANIE\n')
        #     f.write('print("-----------------------------------")\n')
        #     f.write('from URANIE import ROOT\n')
        #     f.write('print("--- ROOT          module loaded ---")\n')
        #     for m in moduleList:
        #         f.write('from URANIE.%s import *\n'%m)
        #         f.write('print("--- %s    module loaded ---")\n'%m)
        #         pass
        #     f.write('print("-----------------------------------")\n\n')
        #     f.write('import sys\n')
        #     f.write('if len(sys.argv) > 1:\n')
        #     f.write('    arg = sys.argv[1]\n')
        #     f.write('    print("***** launch ",arg," with pyuranie *******")\n')
        #     f.write('    execfile(arg)\n')
        #     f.write('    pass\n')
        #     f.write('\n')
        #     f.write('\n')
        #     f.close()            
        #     # Create a "pyuranie" file loading the main modules
        #     f = open(str.join(os.sep, [uranieDest, "python", "pyuranie"]), "w")
        #     f.write("#! /bin/bash\n")
        #     if 'URANIESYS' not in os.environ :
        #         print("Beware, don't forget to set the URANIESYS variable before executing the pyuranie script")
        #     f.write('option="-i"\n')
        #     f.write('arg=$1\n')
        #     f.write('if [  $# -ge 2  ] \n')
        #     f.write('then\n')
        #     f.write('    if [ $1 == "-q" ]\n')
        #     f.write('    then\n')
        #     f.write('        option=""\n')
        #     f.write('    else\n')
        #     f.write('        echo "option $1 does not exist!"\n')
        #     f.write('    fi\n')
        #     f.write('    arg=$2\n')
        #     f.write('fi\n')
        #     f.write("python $option $URANIESYS/lib/python/loadmodules.py $arg\n")
        #     f.close()
        #     # Make it executable
        #     os.chmod(str.join(os.sep, [uranieDest,"python","pyuranie"]), 0o755)
        #         #stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH ) 
    else:
        print("*** ERROR: no LinkDef file found.")
        
        

