//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TCauchyDistribution.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2009-11-20 10:13:36

 \brief Implementation of the class URANIE::DataServer::TCauchyDistribution

 */
#include "TMath.h"
#include "UExceptions.h"
#include "TCauchyDistribution.h"
#include <iostream>
using std::cout;
using std::endl;

ClassImp (URANIE::DataServer::TCauchyDistribution)
namespace URANIE
{
namespace DataServer
{
// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------

TCauchyDistribution::TCauchyDistribution() :
        TInfiniteDistribution("", TStochasticAttribute::kCauchy), _dparameterGamma(
                1.0), _dparameterMedian(0.0)
{
    computeTheoreticalInformation();
}

TCauchyDistribution::TCauchyDistribution(TString str) :
        TInfiniteDistribution(str, TStochasticAttribute::kCauchy), _dparameterGamma(
                1.0), _dparameterMedian(0.0)
{
    computeTheoreticalInformation();
}

TCauchyDistribution::TCauchyDistribution(TString str, Double_t gamma,
        Double_t median) :
        TInfiniteDistribution(str, TStochasticAttribute::kCauchy)
{
    setParameters(gamma, median);
}

TCauchyDistribution::TCauchyDistribution(URANIE::DataServer::TAttribute *att) :
        TInfiniteDistribution(att, TStochasticAttribute::kCauchy), _dparameterGamma(
                1.0), _dparameterMedian(0.0)
{
    computeTheoreticalInformation();
    //  cout << "Constructeur URANIE::DataServer::TCauchyDistribution att[" << this << "]att->getName()[" << att->getName()<<"]" << endl;
}

// -----------------------------------------------------------------------
//  Destructors
// -----------------------------------------------------------------------        

TCauchyDistribution::~TCauchyDistribution()
{
    //  cout << "detruit TCauchyDistribution" << endl;

}

// -----------------------------------------------------------------------
//  Parameters
// -----------------------------------------------------------------------

void TCauchyDistribution::setParameterMedian(Double_t median, Bool_t recompute)
{
    _dparameterMedian = median;
    if (recompute)
        computeTheoreticalInformation();
}

void TCauchyDistribution::setParameterGamma(Double_t gamma, Bool_t recompute)
{
    try
    {
        if (gamma < 1e-4)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TCauchyDistribution::setParameterGamma: parameter gamma [%f] is too close to zero: lower limit is 1e-4",
                            gamma));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    _dparameterGamma = gamma;

    if (recompute)
        computeTheoreticalInformation();

}

void TCauchyDistribution::setParameters(Double_t gamma, Double_t median,
        Bool_t recompute)
{
    setParameterGamma(gamma, kFALSE);
    setParameterMedian(median, kFALSE);

    if (recompute)
        computeTheoreticalInformation();
}

// -----------------------------------------------------------------------
//  Theoretical Information
// -----------------------------------------------------------------------

void TCauchyDistribution::computeTheoreticalInformation()
{
    _dtheoreticalMode = _dparameterMedian;
}

Double_t TCauchyDistribution::getTheoreticalMean()
{
    try
    {
        throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                Form(
                        "TCauchyDistribution::getTheoreticalMean WARNING : theoretical mean is not defined for Cauchy Distribution."));
    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    return _dtheoreticalMean;
}

Double_t TCauchyDistribution::getTheoreticalStdDev()
{
    try
    {
        throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                Form(
                        "TCauchyDistribution::getTheoreticalStdDev WARNING : theoretical standard deviation is not defined for Cauchy Distribution."));
    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    return _dtheoreticalStdDev;
}

// -----------------------------------------------------------------------
//  Distribution
// -----------------------------------------------------------------------

Double_t TCauchyDistribution::getPDF(Double_t x)
{
    Double_t pdf = 0.0;

    if (x >= lowerBound && x <= upperBound)
    {
        pdf = _dparameterGamma / TMath::Pi();
        pdf /= _dparameterGamma * _dparameterGamma
                + (x - _dparameterMedian) * (x - _dparameterMedian);
        // Take into account the case where at least one of the bounds is finite
        pdf /= _pmax - _pmin;
    }

    return pdf;
}

Double_t TCauchyDistribution::getCDF(Double_t x)
{
    Double_t cdf = 0.0;
    if (x >= lowerBound && x <= upperBound)
    {
        cdf = TMath::ATan((x - _dparameterMedian) / _dparameterGamma)
                / TMath::Pi() + 0.5;
        // Take into account the case where at least one of the bounds is finite
        //cdf /= _pmax - _pmin;
        cdf = (cdf - _pmin) / (_pmax - _pmin);
    }
    else if (x > upperBound)
        cdf = 1.0;

    return cdf;
}

Double_t TCauchyDistribution::getInvCDF(Double_t p)
{
    // Verify that p is a valid probability
    try
    {
        if ((p < 0.0) || (p > 1.0))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TCauchyDistribution::getInvCDF: parameter p [%f] is not a valid probability: p must be in the interval [0,1]",
                            p));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }
    // Take into account the case where at least one of the bounds is finite            
    Double_t realP = p * (_pmax - _pmin) + _pmin;

    Double_t x = _dparameterMedian
            + _dparameterGamma * TMath::Tan(TMath::Pi() * (realP - 0.5));
    return x;
}

// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------

void TCauchyDistribution::printLog(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    TStochasticAttribute::printLog();
    cout << "*******************************" << endl;
    cout << "** TCauchyDistribution::printLog  ******" << endl;
    cout << "**   Law[" << _lawType << "] TCauchyDistribution" << endl;
    cout << "**   Bounds[" << lowerBound << ", " << upperBound << "]" << endl;
    cout << "**   Scale parameter gamma[" << _dparameterGamma << "]" << endl;
    cout << "**   Location parameter x0[" << _dparameterMedian << "]" << endl;
    cout << "**   Theoretical Mean[*** undefined ***]" << endl;
    cout << "**   Theoretical StdDev[*** undefined ***]" << endl;
    cout << "**   Theoretical Mode[" << _dtheoreticalMode << "]" << endl;
    cout << "*******************************" << endl;
}

} // Fin du namespace DataServer
} // Fin du namespace URANIE

