//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TCustomDistribution.cxx
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date   Mon Jan 19 12:00:00 CET 2009

 \brief Implementation of the class URANIE::DataServer::TCustomDistribution

 totot
 */
#include <iostream>
using std::cout;
using std::endl;

#include "TMath.h"

// URANIE
#include "UExceptions.h"
#include "TCustomDistribution.h"

ClassImp (URANIE::DataServer::TCustomDistribution)

namespace URANIE
{
namespace DataServer
{

//===============================//
// Constructors and Destructors //
//===============================//

TCustomDistribution::TCustomDistribution()
{
    iNbValues = 0;
    _dSumFreq = 0.0;
    pValues = NULL;
    pFrequencies = NULL;
    clearVectors();
}

TCustomDistribution::TCustomDistribution(TString str) :
        TDiscreteAttribute(str, TStochasticAttribute::kMultinomial)
{
    iNbValues = 0;
    _dSumFreq = 0.0;
    pValues = NULL;
    pFrequencies = NULL;
}

TCustomDistribution::TCustomDistribution(TString name, Int_t nbValues,
        Double_t* values, Double_t* frequencies) :
        TDiscreteAttribute(name, TStochasticAttribute::kMultinomial)
{
    iNbValues = 0;
    _dSumFreq = 0.0;
    pValues = NULL;
    pFrequencies = NULL;
    setDistribution(nbValues, values, frequencies);
}

TCustomDistribution::~TCustomDistribution()
{
    //     cout << "detruit TCustomDistribution" << endl;
    if (pValues)
    {
        delete[] pValues;
        pValues = NULL;
    }
    if (pFrequencies)
    {
        delete[] pFrequencies;
        pFrequencies = NULL;
    }

}

//=====================//
// Getters and Setters //
//=====================//      

Double_t TCustomDistribution::getValue(Int_t i)
{
    try
    {
        if (pValues == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TCustomDistribution::getValue ERROR: the list of input values is empty.");
        if (i >= iNbValues)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TCustomDistribution::getValue ERROR: index [%d] is greater than the number of values [%d]",
                            i, iNbValues));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    return pValues[i];
}

Double_t TCustomDistribution::getFrequency(Int_t i)
{
    try
    {
        if (pFrequencies == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TCustomDistribution::getFrequency ERROR: the list of frequencies is empty.");
        if (i >= iNbValues)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TCustomDistribution::getFrequency ERROR: index [%d] is greater than the number of frequencies [%d]",
                            i, iNbValues));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    return pFrequencies[i];
}

Double_t TCustomDistribution::getProba(Int_t i)
{
    try
    {
        if (pFrequencies == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TCustomDistribution::getProba ERROR: the list of frequencies is empty.");
        if (i >= iNbValues)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TCustomDistribution::getProba ERROR: index [%d] is greater than the number of frequencies [%d]",
                            i, iNbValues));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    return pFrequencies[i] / _dSumFreq;
}

void TCustomDistribution::setDistribution(Int_t nbValues, Double_t* values,
        Double_t* frequencies)
{
    if (pValues)
    {
        delete[] pValues;
        pValues = NULL;
        delete[] pFrequencies;
        pFrequencies = NULL;
    }

    iNbValues = nbValues;
    pValues = new Double_t[iNbValues];
    pFrequencies = new Double_t[iNbValues];

    for (Int_t i = 0; i < iNbValues; i++)
    {
        pValues[i] = values[i];
        if (frequencies)
        {
            try
            {
                if (frequencies[i] <= 0)
                {
                    delete[] pValues;
                    pValues = NULL;
                    delete[] pFrequencies;
                    pFrequencies = NULL;
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__,
                            __LINE__,
                            Form(
                                    "TCustomDistribution::setDistribution ERROR: invalide frequency: frequency %d is negative or null [%e].",
                                    i, frequencies[i]));
                }
            } catch (URANIE::Exceptions::UErrorExceptions& ue)
            {
                ue.printMessage();
                throw ue;
            }
            pFrequencies[i] = frequencies[i];
        }
        else
        {
            pFrequencies[i] = 1.0 / iNbValues;
        }
        _dSumFreq += pFrequencies[i];
    }
}

void TCustomDistribution::setAllValues(Double_t* values)
{
    for (Int_t i = 0; i < iNbValues; i++)
    {
        pValues[i] = values[i];
    }
}

void TCustomDistribution::setValue(Int_t i, Double_t value)
{
    try
    {
        if (pValues == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TCustomDistribution::setValue ERROR: the list of input values is empty.");
        if (i >= iNbValues)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TCustomDistribution::setValue ERROR: index [%d] is greater than the number of values [%d]",
                            i, iNbValues));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    pValues[i] = value;
}

void TCustomDistribution::setAllFrequencies(Double_t* frequencies)
{
    _dSumFreq = 0.0;
    for (Int_t i = 0; i < iNbValues; i++)
    {
        try
        {
            if (frequencies[i] <= 0)
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        Form(
                                "TCustomDistribution::setAllFrequencies ERROR: invalide frequency: frequency %d is negative or null [%e].",
                                i, frequencies[i]));
        } catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            throw ue;
        }
        pFrequencies[i] = frequencies[i];
        _dSumFreq += pFrequencies[i];
    }
}

void TCustomDistribution::setFrequency(Int_t i, Double_t frequency)
{
    try
    {
        if (frequency <= 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TCustomDistribution::setFrequency ERROR: invalide frequency: frequency %d is negative or null [%e].",
                            i, frequency));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    try
    {
        if (pFrequencies == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TCustomDistribution::setFrequency ERROR: the list of frequencies is empty.");
        if (i >= iNbValues)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TCustomDistribution::setFrequency ERROR: index [%d] is greater than the number of frequencies [%d]",
                            i, iNbValues));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    _dSumFreq -= pFrequencies[i];
    pFrequencies[i] = frequency;
    _dSumFreq += pFrequencies[i];
}

//=======================//
// Statistical Functions //
//=======================//

void TCustomDistribution::computeTheoreticalInformation()
{
    try
    {
        throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                "TCustomDistribution::computeTheoreticalInformation WARNING : Mean and Variance are meaningless for discrete distributions");
    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

Double_t TCustomDistribution::getCDF(Double_t x)
{
    UNUSED(x);
    try
    {
        throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                "TCustomDistribution::getCDF WARNING : Cumulative Distribution Function is meaningless for discrete distributions");
    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    return 0.0;
}

Double_t TCustomDistribution::getPDF(Double_t x, Double_t precision)
{
    // If x is not in the distribution, its probability is 0.0
    Double_t p = 0.0;

    for (Int_t i = 0; i < iNbValues; i++)
    {
        if (TMath::Abs(x - pValues[i]) < precision)
        {
            p = pFrequencies[i];
            break;
        }
    }

    if (_blog)
    {
        cout << "TCustomDistribution::getPDF x[" << x << "]  p[" << p << "]"
                << endl;
    }

    return p / _dSumFreq;
}

Double_t TCustomDistribution::getInvCDF(Double_t p)
{
    Double_t x;
    Double_t cumul = 0.0;

    for (Int_t i = 0; i < iNbValues; i++)
    {
        cumul += pFrequencies[i] / _dSumFreq;
        if (p <= cumul)
        {
            x = pValues[i];
            break;
        }
    }

    if (_blog)
    {
        cout << "TCustomDistribution::getInvCDF p[" << p << "]  x[" << x << "]"
                << endl;
    }
    return x;
}

//=====//
// Log //
//=====//

void TCustomDistribution::printLog(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    //     TInfiniteDistribution::printLog();
    cout << "*******************************" << endl;
    cout << "** TCustomDistribution::printLog  ******" << endl;
    cout << "**   Name    [" << GetName() << "]" << endl;
    cout << "**   label   [" << GetTitle() << "]" << endl;
    cout << "**   unite   [" << getUnity() << "]" << endl;
    cout << "**   law     [" << _lawType << "]" << endl;
    cout << "**   type    [" << _nType << "]" << endl;
    cout << "**   nValues [" << iNbValues << "]" << endl;
    cout << "*******************************" << endl;
}

} // Fin du namespace DataServer
} // Fin du namespace URANIE

