//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TCustomDistribution.h
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date   

 \brief Interface of the class URANIE::DataServer::TCustomDistribution

 */
/**

 * @class URANIE::DataServer::TCustomDistribution
 *
 * @brief Description of the class TCustomDistribution.
 *
 * This class allows to define a discrete distribution of probabilities by associating a list of 
 * "labels" with a list of frequencies. The labels are real numbers, but no order is assumed
 * between them, i.e. one can't say that "label 1.0 is smaller than label 2.0".
 * The consequence is that the notions of Probability Density Function and Cumulative Distribution 
 * Function are irrelevant and the functions getPDF, getCDF and getInvCDF don't work the same as 
 * for TStochasticAttribute objects:
 * - getPDF: returns the probability of the given label;
 * - getCDF: returns nothing;
 * - getInvCDF: returns a label drawn from a "pseudo-cumulative" distribution function.
 *
 */
#ifndef TCUSTOMDISTRIBUTION_H
#define TCUSTOMDISTRIBUTION_H

//Uranie
#include "TDiscreteAttribute.h"

namespace URANIE
{
namespace DataServer
{
class TCustomDistribution: public TDiscreteAttribute
{
private:
protected:
    Int_t iNbValues;
    Double_t _dSumFreq;
    Double_t* pValues;
    Double_t* pFrequencies;
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    TCustomDistribution();

    //! Constructor by name only
    /*! 
     
     \param name (TString) Name of the attribute
     */
    TCustomDistribution(TString name);

    //! Constructor by name and statistics
    /*! 
     This constructor allows to define the distribution. The values 
     are considered as labels, and not numerical values. This means that 
     there is no need to sort them. The associated "frequencies" 
     are given in a separated array. They are not probabilities as 
     their sum does not need to be equal to 1. 
     However, it is assumed that the distribution is complete, i.e. 
     the discrete random variable can only take values from the one 
     given.
     \param name (TString) Name of the attribute
     \param nbValues (Int_t) Number of values taken by the discrete variable 
     \param values (Double_t*) List of values taken by the discrete variable
     \param frequencies (Double_t*) List of the frequencies of each value. 
     If NULL, the values are set to be equiprobable. Default = NULL.
     */
    TCustomDistribution(TString name, Int_t nbValues, Double_t* values,
            Double_t* frequencies = NULL);

    //! Default destructor
    virtual ~TCustomDistribution();
    //@}
    //---------------------------------------------
    /** @name Getters and Setters methods
     Getting and setting attribute values.
     **/
    //@{
    //! Get the number of values
    virtual Int_t getNbValues()
    {
        return iNbValues;
    }

    //! Get the values as a Double_t array
    virtual Double_t* getAllValues()
    {
        return pValues;
    }

    //! Get a value by index
    /*!
     \param i (Int_t): index of the value to retrieve
     \exception UErrorExceptions if no value array exists or index is invalid.
     */
    virtual Double_t getValue(Int_t i);

    //! Get the frequencies as a Double_t array
    virtual Double_t* getAllFrequencies()
    {
        return pFrequencies;
    }

    //! Get a frequency by index
    /*!
     \param i (Int_t): index of the frequency to retrieve
     \exception UErrorExceptions if no frequency array exists or index is invalid.
     */
    virtual Double_t getFrequency(Int_t i);

    //! Return the probability corresponding to the given index
    /*!
     \param i (Int_t): index of the probability to retrieve
     \exception UErrorExceptions if no frequency array exists or index is invalid.
     */
    virtual Double_t getProba(Int_t i);

    //! Set the value and frequency arrays. This function erase any 
    //! previous arrays and reset the "iNbValues" parameter.
    /*!
     The values are considered as labels, and not numbers. Thus, they 
     don't need to be sorted. The associated "frequencies"      
     are given in a separated array. They are not probabilities as 
     their sum does not need to be equal to 1. 
     However, it is assumed that the distribution is complete, i.e. 
     the discrete random variable can only take values from the one 
     given.
     \param nbValues (Int_t) Number of values taken by the discrete variable 
     \param values (Double_t*) List of values taken by the discrete variable
     \param frequencies (Double_t*) List of the frequencies of each value. 
     If NULL, the values are set to be equiprobable (default = NULL).
     \exception UErrorExceptions if a frequency is negative or null.
     */
    virtual void setDistribution(Int_t nbValues, Double_t* values,
            Double_t* frequencies = NULL);

    //! Set the values as a Double_t array.
    /*! The number of the array must match the one of the existing frequency array
     \param values (Double_t*): array of size "iNbValues" containing the new input values.  
     */
    virtual void setAllValues(Double_t* values);

    //! Set a value by index
    /*!
     \param i (Int_t): index of the value to set.
     \param value (Double_t): value to set.
     \exception UErrorExceptions if no value array exists or index is invalid.
     */
    virtual void setValue(Int_t i, Double_t value);

    //! Set the frequencies as a Double_t array. 
    /*! The size of the array must match iNbValues.
     \param frequencies (Double_t*): array of size "iNbValues" containing the new frequencies.
     \exception UErrorExceptions if a frequency is negative or null.
     */
    virtual void setAllFrequencies(Double_t* frequencies);

    //! Set a single frequency by index. 
    /*!
     \param i (Int_t): index of the frequency to set.
     \param frequency (Double_t): frequency to set.
     \exception UErrorExceptions if a frequency is negative or null.
     \exception UErrorExceptions if no frequency array exists or index is invalid.
     */
    virtual void setFrequency(Int_t i, Double_t frequency);
    //@}

    /** @name Statistical methods
     The statistical methods ( pdf, cdf, quantile, ...).
     **/
    //@{                
    //! Mean and Variance are meaningless for non ordered discrete values.
    /*!
     \exception always throws a UWarningExceptions.
     */
    virtual void computeTheoreticalInformation();

    //! CDF is meaningless for non ordered discrete values.
    /*!
     This function returns 0.0.
     \exception always throws a UWarningExceptions.
     */
    virtual Double_t getCDF(Double_t x);

    //! Returns the probability of a value.
    /*!
     If x belongs to the distribution, its probability will be equal 
     to the corresponding value in the pFrequencies array divided by the 
     sum of the pFrequencies values.
     NOTE: only values belonging to the distribution have a non-zero
     probability. 
     \param x (Double_t): the value for which the probability is computed 
     \param precision (Double_t): the precision to determine if x is 
     equal to a value of the distribution.
     */
    virtual Double_t getPDF(Double_t x, Double_t precision);

    //! Returns the probability of a value x, with a precision on x fixed at 1e-9.
    /*!
     If x belongs to the distribution, its probability will be equal 
     to the corresponding value in the pFrequencies array divided by the 
     sum of the pFrequencies values. 
     NOTE: only values belonging to the distribution have a non-zero
     probability. The precision to determine if x is equal to a value of the 
     distribution is set to 1e-9.
     \param x (Double_t): the value for which the probability is computed. 
     */
    virtual Double_t getPDF(Double_t x)
    {
        return getPDF(x, 1e-9);
    }

    //! Returns a value from the distribution.
    /*! 
     The choice of the value is based on a pseudo-cumulative 
     function where the order of the discrete values is the order 
     of appearance in the pValues array.
     \param p (Double_t): a probability value between 0 and 1
     */
    virtual Double_t getInvCDF(Double_t p);
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::DataServer::TCustomDistribution, ID_DATASERVER)
};
} // Fin du namespace DataServer
} // Fin du namespace URANIE
#endif
