//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TDataServer.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   thu apr 26 14:14:11 CEST 2007

 $Revision$
 $Date$

 \brief Interface of the class URANIE::DataServer::TDataServer

 add doc 
 */
/*!
 @class URANIE::DataServer::TDataServer

 @brief Description of the class TDataServer.

 To be written by the developper. 

 */
#ifndef TDATASERVER_H
#define TDATASERVER_H

#ifndef ROOT_TCut
#include "TCut.h"
#endif
#ifndef ROOT_TGraph
#include "TGraph.h"
#endif
#ifndef ROOT_TMatrixD
#include "TMatrixD.h"
#endif
#ifndef ROOT_TNtupleD
#include "TNtupleD.h"
#endif

// Uranie
#include "DataServer.h"
#include "TDSNtupleD.h"
#include "TAttribute.h"
#include "TStochAttribut.h"
#include "TDataSpecification.h"
#include "TPatternsEventList.h"
#include "UMessageLogger.h"

#include <iostream>
#include <fstream>
#include <sstream>
#include <map>

using std::cout;
using std::endl;
using std::string;
using std::map;
using std::vector;
using std::ios;

namespace URANIE
{
namespace DataServer
{
class TDataServer: public TNamed
{

  // Attributes
 public:
  enum EOrigin /// define the origin of the TDS (how it has been constructed)
  {
    kUnknown, kASCIIFile, kTree, kSQL
  };
  enum ENormalisation /// define the kind of normalisation to be chosen in normalize method
  {
    kCR, kCentered, kCenteredSum0fSquareOne, kMinusOneOne, kZeroOne
  };
  enum ECAHMethod /// define the method ued to compute CAH
  {
    kWard, kWardExact, kCentroid, kAverage, kSingle, kComplete
  };

  //! Garbage collector list for drawing
  TObjArray *_drawingGarbageCollector;

  //---------------------------------------------
  /** @name Constructor and Destructor
   **/
  //@{
  //! Default constructor
  TDataServer();
  //! Default constructor with a name and a title
  TDataServer(const char *name, const char* title);
  //! Copy constructor
  TDataServer(const TDataServer &tds);
  //! Default destructor
  virtual ~TDataServer();
  //@}
  
  //---------------------------------------------
  /** @name Loading data
      Three ways to load data:
      \li  From an ASCII file with  at least one line in header to size nb of values.
      \code
      TDataServer *fServer = new TDataServer();
      tds->fileDataRead("_ntupledataread_.dat");
      \endcode
      If vectors or string are in the file, a second line is mandatory to explicit this: here is an example of three variable 
      file containing a double x, a string s and a vector v. Element of vectors are slit using commas. A blank line is also 
      mandatory after the header
      \code
      #COLUMN_NAMES: x|s|v
      #COLUMN_TYPES: D|S[V
      \endcode
    
      \li From a TDSNtupleD contained in a ROOT file with selection of variable, added with others possible one by formula then  
      with deletion of  patterns verifying a criterion.
      \code
      TDataServer *fServer = new TDataServer();
      tds->ntupleDataRead("hsimple.root","ntuple","pxx*py:*:py*px", "px*px+py*py<2.0");
      \endcode
      Thus, we construct a TDataserver \em tds from the \em TNtuple \em ntuple contained in the file \em hsimple.root where all 
      original variables are stored from beginning (character \b *). Two new variables are defined by the following equations 
      "px*py" et "py*px". Points verifying the criterion "px*px+py*py<2.0" are excluded from new data.
     
      \li Load a \b TTree which exists in memory
      \code
      TDataServer *fServer = new TDataServer();
      fServer->loadTree(TTree * tt);
      \endcode
     
  **/

  //@{
  //! Reads a  \b ASCII file with a header
  /*!
    The format of the ASCII datafile \b is similar to  \em the format of \em \salome, except for the line containing titles of variables. 
     
    \param name (const char*) The name of the ASCII file 
    \param bSave (Bool_t) true if we save the data in a binary file, false else.
    \param bDontAddAttribute (Bool_t) Default is false. If true and an attribute is already found in TDataSpecification with a given name, 
    no new TAttribute is created. With this option one can use fileDataRead with a TDS in which one needs to specify TStochasticAttribute for instance
  */
    Bool_t fileDataRead(const char* name, Bool_t bSave = kTRUE, Bool_t bDontAddAttribute = kFALSE, Option_t *option="");

#ifdef WITH_JSONCPP
    Bool_t fileDataReadJSon(const char* name, Bool_t bSave = kTRUE);
#endif
    
  //! Reads data from a TDSNtupleD in a ROOT file
  /*!
   * This methods takes an existing file and read the tree to load it. On the contrary to loadTree, TDataSpecification is assumed 
   to be empty while the TDSNtupleD is rebuilt completly deleting.
   * \param file The name of the root file
   * \param tree The name of the TDSNtupleD
   * \param svar The list of attributes. Formula are possible
   * \param cut  The selection of patterns. Points that verify this criterion are excluded (as explicitely said by the name cut)
   \exception Can't open a file, the tree is not found in the input file or is empty, the TDS is not empty, 
   */
  Bool_t ntupleDataRead(const char* file, const char *tree, const char* svar =
			"*", const char *cut = "");
  //! Load a tree
  /*
    This method takes a tree and investigate its content to rebuild a TDS from it.
    TDataSpecification and TDSNtupleD are rebuilt completly deleting all possible previous information
    \param tt: the tree to be read
   */
  Bool_t loadTree(TTree* tt);
  //@}

  //---------------------------------------------
  /** @name Export data       
   Few methods to export the data in order to re-use them
  **/
  //@{  
  //! Exports the data in a ASCII file
  /*!
    The ASCII File can be loaded with a TDataServer object
     
    \param filename (const char *) File name
    \param varexp (const char *)["*"] The list of attributes to export separated by the character ":"
    \param select (const char *)[""] The selection of patterns to the export
    \param option: can precise the precision with which the double are written by typing "precision=" with a digit from 1 to 9
  */
  void exportData(const char *filename, const char *varexp = "*",
		  const char *select = "", Option_t *option = "");

  //! Exports the data in a header file ( C / C++)
  /*!
    Export the data as a header file so that they can be included in a C++ code.
    Vectors used an initialisation style that is only compliant to C++11, but have not yet been tested
    \param filename (const char *) File name
    \param varexp (const char *)["*"] The list of attributes to export separated by the character ":"
    \param select (const char *)[""] The selection of patterns to the export
    \param option: can precise the precision with which the double are written by typing "precision="
  */
  void exportDataHeader(const char *filename, const char *varexp = "*",
      const char *select = "", Option_t *option = "");

  //! Exports the data in a ASCII file reading by NeMo
  /*!
    The ASCII File will be load with the NeMo tool
     
    The format of this file is as the following next
    \verbatim
    #NombreExemples 53
    #NombreEntrees 2
    #NombreSorties 1
     
    1.800000000e+00 5.400000000e+01 5.400000000e+01 
    1.950000000e+00 5.100000000e+01 5.100000000e+01 
    1.833000000e+00 5.400000000e+01 5.400000000e+01
    ...
    \endverbatim
     
    \par using
     
    \code
    TDataServer *tds = new TDataServer("tdsgeyser", "Data from the geyser");
    tds->fileDataRead("geyser.dat");
    tds->addAttribute("y", "sqrt(x2) * x1");
    tds->exportDataNeMo("geyser.nemo", "x1:x2", "y", "x2<55.0");
    \endcode
     
    \param filename (const char *) File name
    \param varexpinput (const char *) The list of input attributes to export separated by the character ":"
    \param varexpoutput (const char *) The list of output attributes to export separated by the character ":"
    \param select (const char *)[""] The selection of patterns to the export
     
  */
  void exportDataNeMo(const char *filename, const char *varexpinput,
		      const char *varexpoutput, const char *select = "");

  //!Export data in a tex file with a ConTex format
  /*!
    Exports the data in a tex file (not modified to deal with Vectors/Strings)     
    \param filename (const char *) File name
    \param varexp (const char *)["*"] The list of attributes to export separated by the character ":"
    \param select (const char *)[""] The selection of patterns to the export
  */
  void exportDataContext(const char *filename, const char *varexp = "*",
			 const char *select = "");

#ifdef WITH_JSONCPP
  void exportDataJSon(const char *filename, const char *varexp = "*",
            const char *select = "", Option_t *option = "");
#endif

    //! Save the ntupleD in the archive (ROOT File)
  /*!
    The name of the ROOT file is given by the name of 
    \param option (Option_t*) [""] The option is the second argument of TFile::Open method (UPDATE|NEW|RECREATE)
  */
  void saveTuple(Option_t* option);
  //@}


  //---------------------------------------------
  /** @name Manipulate attributs
      Several ways for adding/manipulate attributes
      \li through a formula of other existing attributes
      \li through a TAttribute
      \li merge with an other TDataServer
      \li using normalisation methods (CenterReduct, transform the range on the intervals \f$[-1.0, 1.0]\f$ or  \f$[0.0, 1.0]\f$)
  **/
  //@{

    //! check the attribute content
    /*!
      \param tds: the tds
      \param varexp: list of variable
      \param fromopt: call from option
    */
    static string checkAttributes(TDataServer *tds, string varexp, map<string, vector<int> > &mAttributeElements);
    
  //! Adds an attribute with the name and the range
  /*!
    \param name: the name of the attribute
    \param dmin: its minimum value
    \param dmax: its maximum value
   */
  void addAttribute(TString name, Double_t dmin, Double_t dmax);

  //! Adds an attribute with the name and the content
  /*!
    \param name: the name of the attribute
    \param array: the array of data
    \param size: the size of the array
   */
    void addAttributeUsingData(TString name, double *arr, int size);
    
  //! Adds an attribute with the name and title
  /*!
    \param name: the name of the attribute
    \param title: the title of the attribute
   */
  void addAttributeByNameTitle(TString name, TString title);

  //! Adds an attribute with the name only
  /*!
    \param name: the name of the attribute
    \param title: the title of the attribute
   */
  void addAttribute(const char *name);

  //! Adds a new attribute defined by an analytical function
  /*! 
    This function creates a new TAttributeFormula object and add it to 
    the TDS. If a data tree exist, a new branch is automatically created 
    and filled with the results of the attribute's function.
    \param name (TString):    the name of the new attribute.
    \param formula (TString): the formula of the new attribute.
    \param title (TString):   the label of the new attribute (default = "").
    \param unit (TString):    the unit of the new attribute (default = "").
    \param update (TString):  update the back up file with resulting tree (default = true).
    \exception UErrorException if an attribute with the given name already exist.
  */
    void addAttribute(TString name, TString formula, TString title = "",TString unit = "", bool update=true);

  //! fill TFormula attributes
  void fillOthersAttributes(bool update=kFALSE);

  //! Adds an attribute from an attribute
  void addAttribute(URANIE::DataServer::TAttribute *att);
  
  //! Deletes an attribute given by these name
  /*!
    This methods remove the TAttribute from the list of attribute but the corresponding branch is not destroyed in the tree.
    \param name (TString) The name of the attribute to delete
    \exception the attribute does not exist
  */
  void delAttribute(TString name);

  //! Concatenates the variables of two dataservers
  /*!
    Both dataservers must have the name numbers of entries.
    Variables names are tested before considering their value, so if a variable already exists in both tables 
    the second one is copied as well, changing its name to "tdsname.variablename"
     
    \dot
     
    digraph DataserverMerge {
    node [shape=record];
    tds1 [shape=record,label="<x1> X1|<u1> U1|<x2> x2"];
    tds2 [shape=record,label="<x1> X1|<u2> U2"];
    tds [shape=record,label="<x1> X1|<u1> U1|<x2> x2|<x1b> TDS2Name.X1|<u2> U2"];
    tds1:x1 -> tds:x1;
    tds1:u1 -> tds:u1;
    tds1:x2 -> tds:x2;
    tds2:x1 -> tds:x1b;
    tds2:u2 -> tds:u2;
    }
     
    \enddot
     
    One of the optional argument set the use of ROOT::AddFriend method (previous default). This method does not make the new
    variable visible from a TTree::Scan or when looking at list of Attributes available in the TDS. The new default 
    (bwithFriend==false) is a proper concatenation that allows to close and delete the second tds once the merging is done.

    \param tds (TDataServer *) pointer of the DataServer to be merged into the original tds.     
    \param varexpinput: list of attribute from the second tds to be merged
    \param bwithFriends: use ROOT friend method or copy the information completely in the original tds
     
    \exception  if both dataservers have not the same number of patterns
  */
  void merge(TDataServer *tds, const char* varexpinput="*", bool bwithFriend=kFALSE);
  //! Normalize attributes by several methods of normalisation (CenterReduct, on the intervals  \f$[-1.0, 1.0]\f$ or  \f$[0.0, 1.0]\f$)
  /*!
    Several methods of normalisation are provided:
    \li CenterReduct : \b kCR \em [Default]
    \f[
    \tilde{x} = \frac{x - \mu_{x}}{\sigma_{x}}
    \f]
    \li Centered : \b kCentered
    \f[
    \tilde{x} = x - \mu_{x}
    \f]
    \li transform \f$x\f$ on the interval \f$[-1.0, 1.0]\f$ : \b kMinusOneOne
    \f[
    \tilde{x} = 2.0 * \frac{ x - x_{Min}}{x_{Max} - x_{Min}} -1.0
    \f]
    \li transform \f$x\f$ on the interval \f$[0.0, 1.0]\f$ : \b kZeroOne
    \f[
    \tilde{x} = \frac{ x - x_{Min}}{x_{Max} - x_{Min}}
    \f]
     
    The name of the normalized attribute is given by the name of the original attribute and the suffix string.
     
    \param varexp (const char*) The list of attributes 
    \param suffix (const char*)["CR"] The suffix caracter to add of each attributes to create the normalize attribute
    \param method (ENormalisation)[kCR] the method of normalisation
    \param global (bool)[true]: Use only in the case of vector, this boolean explicitely state whether the min, max, mean and std
    are the global one (only one for all patterns and all vector elements) or not (then specific min, max, mean and std are used for
    every component of the vector)
  */
  void normalize(const char* varexp = "", const char* suffix = "_CR",
		 ENormalisation method = kCR, bool global=true);

  //@}

  //---------------------------------------------
  /** @name Select/Cut/Mix patterns
      Criterion to keep/exclude individuals in a database. There is an example:
      \code
      tds->setSelect("( x1*x1 + x2*x2 < 2.0 ) & ( x1 < 30. )");
      \endcode
      or to exclude individuals
      \code
      tds->setCut("x1 >= 30.");
      \endcode
     
  **/
  //@{
  //! Specifies the criterion by a string to select individuals
  /*!
    The selection is tested and if it differs from the previous one (currently used) the quantiles and statistical caracteristic 
    of the attributes are cleared.
    \param str (TString) A string to define selection criterion.
  */
  void setSelect(TString str)
  {
    if( str.EqualTo(TString(_ccut.GetTitle())) )
      return;
    _ccut = TCut(str);
    cleanStatAndQuantiles();
  }
  
  //! Specifies the criterion to  exclude individuals
  /*
    The cut is tested and if it differs from the previous one (currently used) the quantiles and statistical caracteristic 
    of the attributes are cleared.
    \param cut (TCut): A TCut to reject event
   */
  void setCut(TCut cut)
  {
    string thecut = "!("+string(cut.GetTitle())+")";
    if( string(_ccut.GetTitle())==thecut )
      return;
    _ccut = !cut;
    cleanStatAndQuantiles();
  }
    
  //! Specifies the criterion to  exclude individuals
  /*
    The cut is tested and if it differs from the previous one (currently used) the quantiles and statistical caracteristic 
    of the attributes are cleared.
    \param cut (TCut): A TCut to reject event
   */
  void setCut(const char * str)
  {
    string thecut = "!("+string(str)+")";
    if( string(_ccut.GetTitle())==thecut )
      return;
    _ccut = !TCut(str);
    cleanStatAndQuantiles();
  }

  //! Specifies the criterion to  exclude individuals
  /*! 
    The cut is tested and if it differs from the previous one (currently used) the quantiles and statistical caracteristic 
    of the attributes are cleared.
    \param str (TString)  A string to define rejection criterion.
  */
  void setCut(TString str)
  {
    string thecut = "!("+string(str)+")";
    if( string(_ccut.GetTitle())==thecut )
      return;
    _ccut = !TCut(str);
    cleanStatAndQuantiles();
  }

  //! Clears the selection of patterns
  void clearSelect()
  {
    if( string(_ccut.GetTitle())=="" )
      return;
    _ccut = "";
    cleanStatAndQuantiles();
  }

  //! Clears the rejection of patterns
  void clearCut()
  {
    clearSelect();
  }

  //! Utility function to clean the quantiles and stat of all attribute if needed
  /*
    Not to be calles by users
  */
  void cleanStatAndQuantiles()
  {
    if ( gErrorIgnoreLevel < kWarning-1 )
        (*_fLogger) <<  " Selection is changing ==> clearing the TAttribute computed statistics and quantiles" << Endl;
    TList *listOfAtt = getListOfAttributes("*",false);
    TIter next(listOfAtt);
    TAttribute *att;
    while ((att = (TAttribute*) next()))
        att->clearVectors();

    if(listOfAtt!=NULL)
        delete listOfAtt;
  }
  
  //! Mix the patterns
  /*!
    To mix the patterns
  */
  void mix();
  //@}


  //---------------------------------------------
  /** @name Statistical method
      The functions developed to perform statistical treatments
  **/
  //@{
  //! make a CAH
  /*!
    \param varexp (const char *) attributes to make CAH. If mass exists, put it at the last attribute after a 
    character "," (as "x1:x2:x3,mass")
    \param CAHmethod (ECAHMethod)[ kWard | kCentroid | kAverage | kSingle | kComplete }(kWard) the method of CAH
  */
  TMatrixD computeCAH(const char *varexp = "*", ECAHMethod CAHmethod = kWard,
		      const char *selection = "", Option_t *option = "");

  //! Computes the statistics features (minimum and maximum values, mean and standard-deviation)
  /*!
    This method compute the statistical properties of the attribute precised in the varexp list. The results is then stored in 
    the considered attribute itself and can be access through the TAttribute::getX method (where X is Minimum, Maximum,
    Mean or Std). This method deals with both double branch and vectors when their size is constant through patterns.
    \warning if the vector size is not constant through all considered patterns, no computation is done.
    \param varexp: list of variable to be used
    \param selection: additional selection to be used on top of the global one defined by TDataServer::setSelect()
    \param option: no specific option so far
   */
  void computeStatistic(const char *varexp = "*", const char *selection = "", Option_t *option = "");

  //! Computes the correlation matrix
  /*!
    Compute the correlation matrix of the list of considered variable. This method deals with both double branch and 
    vectors when their size is constant through patterns. If so every elements is considered as an input of the matrix

    \warning if the vector size is not constant through all considered patterns, no computation is done.
    \param varexp: list of variable to be used
    \param selection: additional selection to be used on top of the global one defined by TDataServer::setSelect()
    \param option: can check correlation on rank if "rank" is used in option.
   */
  TMatrixD computeCorrelationMatrix(const char *varexp = "*", const char *selection = "", Option_t *option = "");
  //! Computes the rank
  /*!
    Compute the rank of the chosen attribute through the global list of patterns. This method deals with both double branch and 
    vectors when their size is constant through patterns. If so every elements is considered as an input and  a vector of rank
    is created whose elements ith is the ranl of the ith element of the considered vector through all patterns.

    \warning if the vector size is not constant through all considered patterns, no computation is done.
    \param varexp: list of variable to be used
    \param option: no option implemented so far
   */
  void computeRank(const char *varexp = "*", Option_t *option = "");

  //! Estimates quantiles of attribute "attName" with confidence level (Wilks method)
  /*!
    Returns an estimate of the quantiles of the attribute "attName" corresponding to the given probabilities and confidence level.
    This method calls TDataServer::estimateQuantile(const char *attName, Int_t nProba, Double_t* proba, Double_t* quantile, Double_t confidence = 0.5, Option_t *option = ""). See its explanation for more details
    
    \param  nProba number of quantiles to compute
    \param  proba (Double_t)  Proba for which the quantile is computed.
    \param quantile (Double_t &) Variable storing the result
    
    \warning The results are stored in quantile for double-branch case. In the case of vector, quantile contains only the last element 
    quantile values. In all cases, the quantiles are stored in the attribute itself and can be accessed through getQuantile methods.

  */
  void estimateQuantile(const char *attName, Int_t nProba, Double_t* proba,
			Double_t* quantile, Double_t confidence = 0.5,
			Option_t *option = "");
  //! Estimates quantiles of attribute "attName" with confidence level (Wilks method)
  /*!
    Returns an estimate of the quantile of the attribute "attName" corresponding to the given probability and confidence level.
    
    \param attName (const char*) name of the attribute.
    \warning  if more than one attribute name is given,  only the first one will be used !
    \param  nProba number of quantiles to compute
    \param  proba (Double_t)  Proba for which the quantile is computed.
    \param quantile (Double_t &) Variable storing the result
    \param confidence (Double_t)[0.5]  confidence level
    \param option (Option_t)[""] Options passed to the method
    if option contains "interpolate", use the interpolate method between two values
  
    \warning The result is stored in quantile for double-branch case. In the case of vector, quantile contains only the last element 
    quantile values. In all cases, the quantiles are stored in the attribute itself and can be accessed through getQuantile methods.

  */
  void estimateQuantile(const char *attName, Double_t proba,
			Double_t &quantile, Double_t confidence = 0.5,
			Option_t *option = "");

  //! Compute a quantile of the empirical distribution of an attribute
  /*!          
    Returns the quantile of the empirical distribution of the attribute
    "attName".
    For a given probability p, the corresponding quantile q is given by:
    \f[ q = (1-p) x_k + p x_{k+1} \f]
    where \f$ x_k\f$ is the kth smallest value of the n attName values.
    The parameter "type" determines how k is computed. 
     
    \warning  if more than one attribute name is given, only the first one will be used !

    \warning The result is stored in quantile for double-branch case. In the case of vector, quantile contains only the last element 
    quantile values. In all cases, the quantiles are stored in the attribute itself and can be accessed through getQuantile methods.

    \param attName (const char*) name of the attribute.
    \param proba (Double_t)  Proba for which the quantile is computed.
    \param quantile (Double_t &) Variable storing the result
    \param type (Int_t) quantile type (Default = 7). Possible values are:
    <ul>
    <li>Discontinuous cases:</li>
    <ul>
    <li> 1: k = Int( p * n); if p*n == k, q = x_k; else q = x_{k+1}  </li>
    <li> 2: k = Int( p * n); if p*n == k, q = 0.5(x_k + x_{k+1}); else q = x_{k+1} </li>
    <li> 3: k = Int( p * n - 0.5); if p*n-0.5 == k and k is even, q = x_k; else q = x_{k+1} (Default in SAS) </li>
    </ul>
    <li> Piecewise linear interpolations:</li>
    <ul>
    <li> 4: k = Int( p * n ) </li>
    <li> 5: k = Int( p * n - 0.5 ) </li>     
    <li> 6: k = Int( p * (n+1) )         (Default in Minitab and SPSS) </li>
    <li> 7: k = Int( p * (n-1) + 1)      (Default in ROOT, S and R) </li>
    <li> 8: k = Int( p * (n+1/3) + 1/3 ) (approximately median unbiased) </li>
    <li> 9: k = Int( p * (n+1/4) + 3/8 ) (approximately unbiased if x is normaly distributed) </li>
    </ul>
    </ul>
    NOTE: the notation Int(a) represents the integer part of a >= 0.
     
    \par References:
    <ol>
    <li> Hyndman, R.J and Fan, Y, (1996) "Sample quantiles in statistical packages" American Statistician, 50, 361-365 </li>
    <li> R Project documentation for the function quantile of package {stats}   </li>
    <li> ROOT documentation of the TMath::Quantiles function </li>
    </ol>
  */
  void computeQuantile(const char *attName, Double_t proba,
		       Double_t &quantile, Int_t type = 7);
  //! Compute quantiles of the empirical distribution of an attribute
  /*!          
    Returns the quantile of the empirical distribution of the attribute
    "attName", running nProba times the TDataServer::computeQuantile(const char *attName, Double_t proba, Double_t &quantile, Int_t type = 7) method. See its explanation for more details
    
    \param  nProba number of quantiles to compute
    \param  proba (Double_t)  Proba for which the quantile is computed.
    \param quantile (Double_t &) Variable storing the result
    
    \warning The results are stored in quantile for double-branch case. In the case of vector, quantile contains only the last element 
    quantile values. In all cases, the quantiles are stored in the attribute itself and can be accessed through getQuantile methods.
  */
  void computeQuantile(const char *attName, Int_t nProba, Double_t* proba,
		       Double_t* quantiles, Int_t type = 7);
  //Add by MSA
  //! Compute the quantile alpha by comtrol variate
  /*!
    \brief
    \param yname : name of the attribut
    \param alpha :  Proba for which the quantile is computed
    \param zname : name of he control variate,
    \param zapha :  quantile of the control variate,
    \param quantile : Variable storing the quantile
    \param rho : Variable storing the variance reduction
    \par Description : 
    The estimation of the quantile is determinate in 4  steps.
    <ul>
    <li>compute the wj: \f$ w_j=\frac{\alpha}{N_0} 1\!\!1_{Z_j\le z_\alpha} + \frac{\alpha}{N_1} 1\!\!1_{Z_j>z_\alpha} \f$ </li>
    <li>compute the sum of wj: \f$\sum\limits_{i=1}^j w_i \ \  where \ j=1...nS\f$ </li>
    <li>determinate \f$ K=inf\left\{j, \sum_{i=0}^{j} w_i > \alpha \right\}  \ \  where \  j=1...nS\f$ </li>
    <li>determinate the empirique quantile : \f$ Y_{\alpha,n}=Y_{(K)} \f$</li>
    </ul>
    \brief The coefficient \f$ \rho_l \f$ need to compute the reduction of variance is compute
    thanks to this equation: \f$\rho_l = \frac{P[Y\le y_\alpha ,Z\le z_\alpha]-\alpha^2}{\alpha-\alpha^2} \f$
     
    \par Reminder:
    The reduction of variance is \f$1-{\rho_l}^2 \f$    

    \warning not modified to take into account vectors
  */
  void computeQuantileCV(TString yname, Double_t alpha, TString zname,
			 Double_t zapha, Double_t &quantile, Double_t & rho);

  //! Computes the value of quantile alpha
  /*!
     
    \par Description
     
    This is a rewriting of the ROOT::TMath::Quantiles() function which was 
    found buggy as of February 2010. The present algorithm is derived from 
    http://mathworld.wolfram.com/Quantile.html. The function prototype is the 
    same as ROOT::TMath::Quantiles.

    \warning  Internal method not to be called by users
  */
  void internalQuantilesComputation(Int_t nData, Int_t nProba, Double_t* data,
				    Double_t* quantiles, Double_t* proba, Bool_t isSorted = kTRUE,
				    Int_t* index = 0, Int_t type = 7);

  //! Computes the index of the quantile alpha with a confidence beta (Wilks method)
  /*!
     
    \par Description
     
    Given a sorted array of n realizations of a random variable X, a quantile 
    alpha and a confidence level beta, this function returns the index of the 
    realization Q for which P(X <= Q) = alpha with a confidence level beta.  
     
    For "large" values of n (i.e n > 1000), in order to reduce numerical errors, 
    an approximation of the Wilks formula is used.
     
    (ref. ?)      
    \par Usage
     
    \verbatim
    TDataServer* tds = new TDataServer("tdsQuantiles","test for quantiles");
    tds->computeQuantile("x",0.95,q);
    \endverbatim
     
    \param alpha : probability defining the quantile
    \param beta  : confidence level
    \param n     : sample size
  */
  Int_t computeIndexQuantileWilks(Double_t alpha, Double_t beta, Int_t n);
  
  //! Computes the inverse of the cumulative of a normal distribution
  /*!          
    \par Description
     
    For a given normal distribution N(mean, stddev) and a probability p, this 
    function computes the value X for which the integral of N from -Inf to X 
    is equal to p.
     
    By default, mean = 0.0 and stddev = 1.0.
     
    \par Usage
     
    \verbatim
    TDataServer* tds = new TDataServer("tdsInvCDF","test for inverse of normal CDF");
    tds->computeInverseNormalCDF(0.5, 5.0, 2.5);
    \endverbatim
     
    \param p      : probability
    \param mean   : mean of the normal distribution
    \param stddev : standard deviation of the normal distribution
  */
  Double_t computeInverseNormalCDF(Double_t p, Double_t mean = 0.0, Double_t stddev = 1.0);
  
  //@}
  
  
  //---------------------------------------------
  /** @name Visualisation
      The functions developed to perform visualization of data
  **/
  //@{
  //! Scan content of the TDS
  /*!
    Print out the content of the tree given the parameter requested
    \param varexp: list of attribute to be dumped
    \param selection: selection criterion to be applied on top of the global one
    \param option: option to be passed to ROOT::TTree::Scan method (which is call at the end)
   */  
  Int_t scan(const char *varexp = "", const char *selection = "",
	     Option_t *option = "");
  //! Scan content of the TDS
  /*!
    Calls TDataServer::scan method
   */
  Int_t Scan(const char *varexp, const char *selection = "",
	     Option_t *option = "")
  {
    return scan(varexp, selection, option);
  }

    //!Crate a canvas if needed
    void checkCanvasCreation(Option_t *option="");
  //! Draws histogram for a given index of a variable
  /*! 
    \param varexp (const char*) name of the attribute
    \param selection (const char*) the formula to select pattern ("")
    \param option (Option_t*) drawing options ("")
    The "nclass=" option allows to chose the number of histogram bins. Uranie provides 4 choices:          
    \li nclass=root:    uses the ROOT's default number of bins
    \li nclass=sturges: computes the number of bins via struges method
    \li nclass=fd:      computes the number of bins via fd method
    \li nclass=scott:   computes the number of bins via scott method
    \li nclass=n:       uses n bins, where n is a positive integer value
  */
    void draw(const char* varexp, const char* selection = "", Option_t* option = "", bool internalCall=kFALSE);
  //! Draws histogram for a given index of a variable
  /*! 
    Calls TDataServer::draw method
  */
  void Draw(const char* varexp, const char* selection = "", Option_t* option =
            "")
  {
    draw(varexp, selection, option);
  }

  //! Draw a histogram of the attribute specified by its index in the TDataServer
  /*! 
    The index refers to the position at which the attribute was inserted into the TDataServer
    \param nOrder (Int_t) Index of the attribute
    \param selection (const char*) the formula to select pattern ("")
    \param option (Option_t*) drawing options ("")
  */
  void draw(UInt_t nOrder, const char* selection = "", Option_t* option = "");

  //! Histogram of a variable
  void drawHistogram(const char* sVar, const char* selection = "",
		     Option_t* option = "");
  //! Graph of a variable depanding of his apparition order in the list
  void drawIter(const char* sVar, const char* selection = "",
		Option_t* option = "");
  //! Experimental CDF of a single variable
  /*!
    This function differs from drawCDF on the fact that it doesn't produce a 
    histogram. Instead, each datum is plotted, which produces a curve that is 
    more accurate but maybe not as smooth as the one given by drawCDF.
  */
  void drawECDF(const char* varexp, const char* selection = "",
		Option_t* option = "");

  /*!
    Draw the QQ plot of a variable. With "same" in option, the plot is drawn on top of existing one.
  */
    void drawQQPlot(const char* varexp, const char* laws, int nq, Option_t* option = "");

  /*!
    Draw the PP plot of a variable. With "same" in option, the plot is drawn on top of existing one.
  */
    void drawPPPlot(const char* varexp, const char* laws, int nq, Option_t* option = "");

  //! CDF d'une variable
  void drawCDF(const char* varexp, const char* selection = "",
	       Option_t* option = "");

  void drawBoxPlot(const char* varexp, const char* selection = "",
		   Option_t* option = "");

  void startViewer();
  void StartViewer()
  {
    startViewer();
  }

  //! Draw 2D scatterplots "tufte"
  /*!
    See http://addictedtor.free.fr/graphiques/RGraphGallery.php?graph=81
  */
  void drawTufte(const char* varexp, const char* selection = "",
		 Option_t* option = "");
  //! Draws Scatterplots
  void drawScatterplot(const char* varexp, const char* selection = "",
		       Option_t* option = "");
  //! Draws Profile
  /*!
    Draw a profile (mean and sigma) from a scatterplot \f$ y \f$ versus \f$ x \f$
     
    The profile between (xmin, xmax) value in the first abscissis and (ymin, ymax) ine the second one.
    where
    \li xmin = min(x) + 0.01 [ Max(x) - Min(x) ]
    \li xmax = Max(x) + 0.01 [ Max(x) - Min(x) ]
     
    \exception When the number of attributes \em varexp is different to 2  
  */
  void drawProfile(const char* varexp, const char* selection = "",
		   Option_t* option = "");
  //! draw the scatterplots for each possible pairs \f$(x_j, x_i)\f$
  /*!
    \param varexp (const char*) the list of attributes to draw the pairs ("*").  By default, ("*") all the attributs are used.
    \param selection (const char*) the formula to select pattern ("")
    \param option (Option_t*) drawing options (""). No more ROOT options are treated.
  */
  void drawPairs(const char* varexp = "*", const char* selection = "",
		 Option_t* option = "");
  //! Draw a cobweb
  /*
    This graph is as \b TParallelCoordonate in ROOT but is called \b cobweb in the uncertainty community (book of Saltelli).
  */
  void drawCobWeb(const char* varexp, const char* selection = "",
		  Option_t* option = "");
  //@}  

  
  //---------------------------------------------
  /** @name Setters
      This function are used to change parameters
  **/
  //@{

  //! Change the name
    /*!
      Method added to deal with the registration of the dataserver in gROOT; It overwrite the TObject one.
    */
    void SetName(const char *name);
    //! Check wether the name exists already    
    void checkName();
    
  //! Sets an iterator
  /*!
    Defines the attribute named \em sname as an index.     
    \param sname (TString) The name of the attribute which is defined as an index 
  */
  void setIterator(TString sname);

  //! Set the TDSNtupleD of _datatree variable.
  /*!
    \param ntd (TDSNtupleD *) a pointer on a TDSNtupleD. The pointer is not deleted before the (new) value is set. If _datatree has 
    already been initialized, you might want to use the deleteTuple method before to avoid memory leak.
  */
    void setTuple(TDSNtupleD *ntd, bool newismain=true, const char* att="*")
  {
    _datatree = ((ntd==NULL) ? NULL : new TDSNtupleD(ntd, att, "", newismain));
  }
  //! Set the TDSNtupleD of _datatree variable.
  /*!
    \param ntd (TTree *) a pointer on a TTree. The pointer is not deleted before the (new) value is set. If _datatree has 
    already been initialized, you might want to use the deleteTuple method before to avoid memory leak.
  */
  void setTuple(TTree *ntd, bool newismain=true)
  {
    _datatree = ((ntd==NULL)? NULL :new TDSNtupleD(ntd,"*","",newismain));
  }

  //@}


  //---------------------------------------------
  /** @name Getter functions
   **/
  //@{
  
  //!  Gets the list of attributes from a string
  /*!
    Returns the list of attributes. The character "*" means all attributes.
    \param onlyAttribute: boolean to specify if the list of attribute should contains only kAttribute and kConstant, or all kind of attribute 
    (kInternal and kIterator as well)
  */
  TList * getListOfAttributes(const char* sListOfAttributes = "*", bool onlyAttribute=true);

    //!  Gets the list of attributes as a string
    /*!
      Returns the list of attributes as a string in classical uranie format. 
      \param onlyNumerical: boolean to specify if the list of attribute should contains only kDouble and kVector, or whether it should include kString as well.
      \param onlyAttribute: boolean to specify if the list of attribute should contains only kAttribute and kConstant, or all kind of attribute 
      (kInternal and kIterator as well)
    */
    string getStringListOfAttributes(bool onlyNumerical=true, bool onlyAttribute=true);
    
  //! Returns the index of the attribute
  /*!
    Return the index of the attribute under study in the list of attribute
    
    \param att: attribute under study
    \param onlyAttribute: boolean to specify if the list of attribute should contains only kAttribute and kConstant, or all kind of attribute 
    (kInternal and kIterator as well)
   */
  int getAttributeIndex(TAttribute * att, bool onlyAttribute=false) const;
 
  //! Returns the index of the attribute
  /*!
    Return the index of the attribute under study in the list of attribute
    
    \param att: attribute under study
    \param onlyAttribute: boolean to specify if the list of attribute should contains only real attribute (kAttribute and kConstant), or all 
    kind of attribute (kInternal and kIterator as well)
   */
  int getAttributeIndex(TString name, bool onlyAttribute=false) const;

  //! Returns a pointer of the attribute given by this name
  /*!
    \exception If an attribute with this name does not exist.
  */
  TAttribute *getAttribute(TString str);


  //! Returns a pointer of the attribute given by this name
  /*!
    \param ind: index of the attribute
    \param onlyAttribute: boolean to specify if the list of attribute should contains only real attribute (kAttribute and kConstant), or all 
    kind of attribute (kInternal and kIterator as well)
    \exception If an attribute with this name does not exist.
  */
  TAttribute *getAttribute(Int_t ind, bool onlyAttribute=kTRUE);

  //! Returns the value at index "entry" of the leaf "attName".
  /*!
    If ielement != -1 it returns the ielement-th of the vector, if the concerned attribute is a vector
    \exception if attribute is a string
    \exception if attribute is a string and ielement == -1 or ielement > vector.size()
  */
  Double_t getValue(TString attName, Int_t entry, Int_t ielement=-1);

  //! Returns a vector of double for entry "entry" of the leaf "attName".
  /*!
    \exception if attribute is a string or double
    \exception if attribute is not existing
  */
  vector<double> getVector(TString attName, Int_t entry);
  //! Returns a string for entry "entry" of the leaf "attName".
  /*!
    \exception if attribute is a vector or double
    \exception if attribute is not existing
  */
  string getString(TString attName, Int_t entry);

  //! Returns the matrix of data
  /*!
    Works only
   */
  TMatrixD getMatrix(const char* varexp = "*", const char* selection = "",
		     Option_t* option = "");

  //! Returns the number of patterns in the TDataServer
  /*!
    \exception the TNtuple is NULL
  */
  Int_t getNPatterns();
  
  //! Returns the name of the iterator
  const char *getIteratorName()
  {
    return ((_biteratorNameChanged) ? _siteratorChangedName.Data() : Form("%s__n__iter__", GetName()) );
  }

  void changeIteratorName(const char *name)
  {
    _siteratorChangedName=TString(name);
    _biteratorNameChanged=kTRUE;
  }
  
   //! Returns the name of the tree
  char * getDataTreeName();
  //! Returns the name of the file
  char * getDataFileName();
  //! Returns the name of the archive file
  char * getArchiveFileName();

  //! Return the origin of the considered TDS
  EOrigin getOrigin()
  {
    return _norigin;
  }

  //! Returns the number of attributes in the TDataServer
  /*!
    Returns the number of attributes
    \param onlyAttribute: boolean to specify if the list of attribute should contains only real attribute (kAttribute and kConstant), or all 
    kind of attribute (kInternal and kIterator as well)
  */
  Int_t getNAttributes(bool onlyAttribute=true);

  //! Return the number of input Attributes
  /*!
    Return the number of attributes which are \em Input \em attributes
  */
  Int_t getNInputAttributes();
  
  //! Return the number of output Attributes
  /*!
    Return the number of attributes which are \em Ouput \em attributes
  */
  Int_t getNOutputAttributes();

  //! Return the selection TCut
  TCut getCut()
  {
    return _ccut;
  }

  //! Return a TDataSpecification pointer
  TDataSpecification* getDataSpecification()
  {
    return _dataSpecification;
  }

  //! Returns a pointer to the TDSNtupleD
  /*!
    \param option: useless so far
   */
  TDSNtupleD *GetTuple(Option_t *option = "V")
  {
    return getTuple(option);
  }
    //! Returns a pointer to the TDSNtupleD
  /*!
    \param option: useless so far
   */
  TDSNtupleD *getTuple(Option_t *option = "V");
  
  //@}      

  //---------------------------------------------
  /** @name Bootstrap methods
      The functions developed to perform \b bootstrap analysis
  **/
  //@{
  //! Define a bootstrap database
  /*!
    When the size is :
    \li equal to 0; the TEventList is put to NULL ( original TEventList = \f$ {1, 2, 3, \cdots, nS}\f$
    \li is non-positive; the size is given by the size of the TDataServer (nS = tds->getNPatterns())
    \li is greate than 1; it is the size which must be lesser than nS = tds->getNPatterns().
     
    \param nsize (Int_t) The size of the bootstrap database [-1]
    \param option (Option_t) The option passed [""]("sort|seed").
    If option contains \em "sort", the TEventList is sorted
    If option contains \em "seed=", the seed parameter is passed to the TRandom object
  */
  void bootstrap(Int_t nsize = -1, Option_t *option = "");
  //@}

  
  //---------------------------------------------
  /** @name Miscellaneous
   **/
  //@{

  //! Creates the tuple of the TDataServer
  void createTuple(Option_t *option = "");

  //! Fills a new Branch for an attribute by applying the given formula
  /*!
    This function takes the name of an existing attribute and creates 
    a new branch for it in the data tree. Then, it fills the branch 
    according to the given formula. If the attribute is a 
    TAttributeFormula, its formula is used.
    \param name    (TString): name of the attribute.
    \param formula (TString): formula to apply if the attribute isn't 
    a TAttributeFormula (default = "")
    \param update  (Bool_t):  tells to update the "backup" file of the 
    data server or not. As the operation of 
    writting data on disk can be very slow, 
    it can be useful to set this parameter 
    to kFALSE to disable it (default = kTRUE)
    \exception UranieError if there is no data tree.
    \exception UranieError if the attribute doesn't exist
    \exception UranieError if the attribute is a TAttributeFormula 
    and the formula parameter is not empty.
    \exception UranieError if the attribute isn't a TAttributeFormula 
    and the formula parameter is empty                
    \exception UranieError if the formula is unvalid.
    \exception UranieError if the attribute is already filled.
  */
  void fillAttribute(TString name, TString formula = "", Bool_t update = kTRUE);


    //! Fills a new Branch for an attribute by copying a constant uentry.
  /*!
    Very advanced, this function is mainly used internly to help dealing with 
    constant attribute that could be string or vector and that could be written
    in the output sample.
    \param name    (TString): name of the attribute.
    \param ConstUEntry(UEntry*): UEntry containing the information to be stored.
    \param update  (Bool_t):  tells to update the "backup" file of the 
    data server or not. As the operation of 
    writting data on disk can be very slow, 
    it can be useful to set this parameter 
    to kFALSE to disable it (default = kTRUE)
    \exception UranieError if there is no data tree.
    \exception UranieError if the UEntry doesn't contains data
    \exception UranieError if the attribute is already filled.
  */
  void fillConstantAttribute(TString name, UEntry *ConstUEntry);

  //! Checks if the attribute given by a name exist
  /*!
    \return kTRUE if an attribute given by the name exist, else kFALSE 
  */
  Bool_t isAttribute(TString name) const;

  
  //! Delete the TDSNtupleD of _datatree variable.
  void deleteTuple();  

  //! Check if the data server contains data
  /*!
    This function returns TRUE if the data server contains a 
    TDSNtupleD. It returns FALSE otherwise.
  */
  Bool_t hasTuple()
  {
    return _datatree != NULL;
  }

   void keepFinalTuple(bool val)
  {
      _bsaveTuple = val;
  }
  //@}

  //---------------------------------------------
  /** @name Printing Log
   **/
  //@{
  void setLog()
  {
    _blog = kTRUE;
  }
  void unsetLog()
  {
    _blog = kFALSE;
  }
  void changeLog()
  {
    _blog = _blog ? kFALSE : kTRUE;
  }
  Bool_t getLog()
  {
    return _blog;
  }
  virtual void printLog(Option_t *option = "");

    static TString getUranieVersion(){ return TString("Uranie v")+TString(URANIE_RELEASE);}
    static Int_t doBanner(){
      string bold_on, bold_off;
#ifndef WIN32
		  bold_on = "\033[1m";
		  bold_off = "\033[0m";
#endif
        cout << "\n" << bold_on << "--- Uranie v" << URANIE_RELEASE
             << " --- Developed with ROOT (" << gROOT->GetVersion() << ")" << bold_off << "\n"
             << "                      Copyright (C) 2013-2022 CEA/DES "  << "\n"
             << "                      Contact: support-uranie@cea.fr "   << "\n"
             << "                      Date: " << URANIE_RELEASE_DATE 
             << "\n\n";
        return 0;
    }

        
  //@}
    
    static  void findParametersLaw(TString soption, vector<double> &vec, bool blog);
    
    
  ClassDef(URANIE::DataServer::TDataServer, ID_DATASERVER)

    private:
  TDataSpecification* _dataSpecification;      ///< Specification of data
  TDSNtupleD* _datatree;      ///< Tree with data
  TCanvas *_canvas;   ///< The canvas to work on;
  TTree* _friend;      ///< Tree added by friend method
  Bool_t _blog;          ///< log printing
  Bool_t _bQuantVectPrinted; ///< print the message for the vector quantile calculation only once
  Bool_t _biteratorNameChanged; ///< To know whether this iterator name has been changed
  Bool_t _bsaveTuple;
  TString _sdatafile;     ///< The original data file (ASCII mode)
  TString _sarchivefile;  ///< The archive file (.root)
  TString _shortDataTreeName; ///< The shortened name of the TDataServer;
  TString _siteratorChangedName; ///< The shortened name of the TDataServer;
  TCut _ccut;          ///< Selection of pattern
  
  void drawLegend(const char* varexp, const char* selection = "",
		  const char * histname = "htemp");

    //! Internal method, not to be used standalonely
    /*!
      Draw the QQ plot of a variable. With "same" in option, the plot is drawn on top of existing one.
    */
    void drawQQorPPPlot(const char* varexp, const char* laws, int nq, Option_t* option );
    
  EOrigin _norigin; ///< The origin of the dataserver (ASCII File, TTree, SQL...)

  TPatternsEventList *_pelBootstrap;  ///< The bootstrap list of input pattern

  URANIE::DataServer::UMessageLogger *_fLogger;   ///<  Message logger
  
  //---------------------------------------------
  /** @name Internal methods
      The internal functions that are called only by other methods
  **/
  //@{
  //! Draw a CDF of a TPossibility attribute (RaFu)
  void drawCDFRaFu(const char* varexp, const char* selection = "",
		   Option_t* option = "");
 
  //@}

    
};

} // Fin du namespace DataServer
} // Fin du namespace URANIE
#endif
