//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TDataSpecification.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   thu apr 26 14:14:11 CEST 2007

 $Revision$
 $Date$

 \brief Implementation of the class URANIE::DataServer::TDataSpecification

 It contains informations on \em header of a TDataServer
 \par Modifications
 \li JB Blanchard (CEA) 2016-2-22: Modifications to homogeneity purpose: several methods were defined both in TDS and here
 but not consistently (dangerous) when considering kAttribute and kConstant. All getAttribute/ AttributeIndex/ ListOfAttr 
are defined here and call from TDS after checking  that TDataSpecification is not null. Might impact other module as well.
 \li M. Salvoldi (OCC) (11-07-29 16:30): Modification of the method getAttribute
 -> if the attribut is unknow of the tds, we search if this attribut have a "father" which are known of the tds
 */

#include <memory>
#include <time.h>
#include <algorithm>

// Uranie
#include "TDataSpecification.h"
#include "UExceptions.h"

ClassImp (URANIE::DataServer::TDataSpecification)

namespace URANIE
{
namespace DataServer
{
  // -----------------------------------------------------------------------
  //  Constructors and Destructor
  // -----------------------------------------------------------------------
  TDataSpecification::TDataSpecification() :
    TNamed("_header_", "_title_"), _blog(kFALSE), _relationName("noname")
  {
    _listOfAttributes = new TList();
  }

  TDataSpecification::TDataSpecification(const char *name, const char* title) :
    TNamed(name, title), _blog(kFALSE), _relationName(title) 
  {
    _listOfAttributes = new TList();
  }

  TDataSpecification::TDataSpecification(URANIE::DataServer::TDataSpecification *tdsp)
  {
    _listOfAttributes = new TList();

    TList *inputfromtdsp = tdsp->getListOfAttributes("*",false);
    TIter nextIn(inputfromtdsp);
    TAttribute* att;

    while( (att = (TAttribute*)nextIn() ) )
      this->addAttribute(att);
    
    delete inputfromtdsp;
  }
  
  TDataSpecification::~TDataSpecification()
  {
    TAttribute *att;

    if (_blog)
      cout << "Destructeur::TDataSpecification[" << this << "] taille["
	   << _listOfAttributes->GetEntries() << "]" << endl;

    if(_blog)
        _listOfAttributes->ls();

    if (_listOfAttributes)
      {
          Int_t ncrt = 0, nNotFound=0, nbTotal=_listOfAttributes->GetEntries();
          while ((att = ((TAttribute *) _listOfAttributes->First()+nNotFound)))
	  {
            if (att != NULL)
	      {
                if (_blog)
		  cout << " ** ncrt[" << ncrt << "/" << _listOfAttributes->GetEntries() << "] \t";
                TAttribute *test = (TAttribute *) _listOfAttributes->Remove(att);
                if (att != test)
                {
                    if (_blog)
                        cout
                            << "WARNING :: Destructor::TDataSpecification with TAttribute "
                            << att << "  " << test << "  nNotFound will be increased from " << nNotFound << " to " << nNotFound+1 << std::endl ;
                    nNotFound++;
                }
                else
                {
                    att->delShare();
                    if (_blog)
                        cout << "     ** DelShare with TAttribute["
                             << att->GetName() << "] ... with new share["
                             << att->getShare() << "]" << endl;
                    if (att->getShare() == 0)
                    {
                        if (_blog)
                            att->setLog();
                        delete att;
                        att = NULL;
                    }
                }
	      }
            else
	      {
                if (_blog)
		  cout << " ** Att already NULL" << endl;
	      }
            ncrt++;
            if( ncrt >= nbTotal ) break;
	  }

          delete _listOfAttributes;
          _listOfAttributes = NULL;
      }

  }

  // -----------------------------------------------------------------------
  //  Setting and Getting informations
  // -----------------------------------------------------------------------

TString TDataSpecification::getVariablesNames(TString separator, bool noIterator)
  {
    TString sLabelVariable;
    int written=0;
    
    for (Int_t indx = 0; indx < getNAttributes(); indx++)
      {
	  if(noIterator && getAttribute(indx)->getOrigin() == TAttribute::kIterator )
	      continue;
        if (written > 0)
	  sLabelVariable += separator;
        sLabelVariable += getAttribute(indx)->GetName();
	written++;
      }
    return sLabelVariable;
  }

TString TDataSpecification::getVariablesTypes(TString separator, bool noIterator)
  {
    TString sLabelVariable;
    int written=0;
    
    for (Int_t indx = 0; indx < getNAttributes(); indx++)
      {
	  if(noIterator && getAttribute(indx)->getOrigin() == TAttribute::kIterator )
	      continue;
        if (written > 0)
	  sLabelVariable += separator;
        sLabelVariable += getAttribute(indx)->getDataTypeStr();
	written++;
      }
    return sLabelVariable;
  }

TString TDataSpecification::getVariablesNamesTypes(TString separator, bool noIterator)
{

    TString sLabelVariable;
    int written=0;
    
    for (Int_t indx = 0; indx < getNAttributes(); indx++)
    {
	if(noIterator && getAttribute(indx)->getOrigin() == TAttribute::kIterator )
	    continue;
        if (written > 0)
	    sLabelVariable += separator;
        sLabelVariable += getAttribute(indx)->GetName()+TString("(")+getAttribute(indx)->getDataTypeStr()+TString(")");
	written++;
    }
    return sLabelVariable;
    
}
bool TDataSpecification::isComparableToTDSP(TString otherone, bool isJustContained)
{

    TString local = getVariablesNamesTypes(":", true);

    vector<string> locals, otherones;
    URANIE::DataServer::UEntry::StringParser(string(local),":",locals);
    URANIE::DataServer::UEntry::StringParser(string(otherone),":",otherones);
    if( otherones.size() != locals.size() && !isJustContained )
	return false;
    
    for(unsigned int it=0; it<locals.size(); it++)	
	if( find( otherones.begin(), otherones.end(), locals.at(it) ) == otherones.end() )
	    return false;

    return true;
    
}

bool TDataSpecification::isComparableToTDSP(TDataSpecification *tdsp, bool isJustContained)
{

    return isComparableToTDSP( tdsp->getVariablesNamesTypes(":",true), isJustContained);

}

  ///////////////////////////////////////////////////////////
  //  management of attributes (to delete)
  ///////////////////////////////////////////////////////////

  ///////////////////////////////////////////////////////////
  //  management of attributes by the TList 
  ///////////////////////////////////////////////////////////
void TDataSpecification::addAttribute(TAttribute* miningAttribute, bool dontAddSizeOne)
  {
    try
      {
        // Test if the attribute define by name existe
        if (_listOfAttributes->FindObject(miningAttribute->GetName()) == NULL)
	  {
            miningAttribute->addShare();
            _listOfAttributes->Add(miningAttribute);
	    if( miningAttribute->getDataType() == TAttribute::kVector && !dontAddSizeOne)
	    {
		addAttribute( new TAttribute( (SIZENAME+"__"+string(miningAttribute->GetName()) ).c_str()) );
		getAttribute((SIZENAME+"__"+string(miningAttribute->GetName()) ).c_str() )->setOrigin(TAttribute::kInternal) ;
		getAttribute((SIZENAME+"__"+string(miningAttribute->GetName()) ).c_str() )->delShare() ;
	    }
	  }
        else
	  {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form(
							    "TDataSpecification::addAttribute[%s]  Exist an attribute with the same name",
							    miningAttribute->GetName()));
	  }
      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
        throw ue;
        exit(-1);
      }
  }

  void TDataSpecification::delAttribute(TString name)
  {
    try
      {
        TAttribute *att =
	  dynamic_cast<URANIE::DataServer::TAttribute *>(_listOfAttributes->FindObject(
										       name));

        if (att)
	  {
            att->delShare();
            _listOfAttributes->Remove(att);
	    if (att->getShare() == 0)
	      {
		if (_blog)
		  att->setLog();
		delete att;
		att = NULL;
	      }
	  }
        else
	  {
            //_listOfAttributes->Print();
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form(
							    "TDataSpecification::delAttribute[%s]  attribute with name does not exist",
							    name.Data()));
	  }
      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
        throw ue;
        exit(-1);
      }
  }

  //Method to get the list of attribute
  TList * TDataSpecification::getListOfAttributes(const char* sListOfAttributes, bool onlyAttribute)
  {
    //Imported from TDataServer by JBB 2016-2-22 for consistency between TDS and TDataSpecification 
    TList *listOfAttributes = new TList();

    // Determine data branches
    TString satts(sListOfAttributes);

    // Case of all attributes
    if ((satts == "*") || (satts.Length() == 0))
      {
        // Passage par le TDataSpecification pour courcircuiter la loop couteuse
        Int_t nallpat = getNAttributes();
        for (Int_t i = 0; i < nallpat; i++)
	  {
            TAttribute* att = getAttribute(i);
	    if (!onlyAttribute)
	      listOfAttributes->Add(att);	      
            else if (att->getOrigin() == TAttribute::kAttribute || att->getOrigin() == TAttribute::kConstant) //Mod by JBB
	      listOfAttributes->Add(att);
	  }
      }
    else
      {

        // The list of attributes to compute correlation matrix
        satts += TString(" ");
        satts.ReplaceAll(":", " ");
        if (_blog)
	  cout << " ** satts[" << satts << "]" << endl << endl;

        istringstream inStream((char *) satts.Data());
        string inName;
        while (inStream >> inName)
	  {
            TString svar(inName.c_str());
            if (_blog)
	      cout << " ** svar[" << svar << "]" << endl;

	    try
	      {

		if (!isAttribute(svar))
		  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							     Form(
								  "TDataSpecification::getListOfAttributes[sListOfAttributes=\"%s\"]  The attribute \"%s\" does not exist.",
								  sListOfAttributes, svar.Data()));
	      }catch(URANIE::Exceptions::UErrorExceptions &ue)
	      {
		ue.printMessage();
		delete listOfAttributes;
		listOfAttributes=NULL;
		throw ue;
	      }

            TAttribute *att = getAttribute(svar);
	    if (!onlyAttribute)
	      listOfAttributes->Add(att);
	    else if (att->getOrigin() == TAttribute::kAttribute || att->getOrigin() == TAttribute::kConstant) //Mod by JBB
	      listOfAttributes->Add(att);
	  }
      }

    Int_t natts = listOfAttributes->GetEntries();
    if (_blog)
      {
        cout << endl << " **  Attributes listing  [" << natts << "]" << endl;
        listOfAttributes->Print();
      }

    return listOfAttributes;
    
  }
  
  TAttribute* TDataSpecification::getAttribute(TString name) const
  {
    TAttribute* att = 0;

    try
      {
        att = dynamic_cast<URANIE::DataServer::TAttribute *>(_listOfAttributes->FindObject(name));

        if (att == NULL)
	  {
            Int_t i = 0;
            Int_t find = 0;
            while (find == 0)
	      {
                TAttribute* att_msi = getAttribute(i);
                if (att_msi->isInput() && att_msi->haveSon())
		  {
                    TAttribute* att_son = att_msi->getSonAttribute();
                    if (att_son->GetName() == name)
		      {
                        att_msi->printLog();
                        att = att_msi;
                        if (_blog)
			  cout << " *** Fin de la recherche avec[" << name
			       << "] [" << att_msi->GetName()
			       << "] -- fils [" << att_son->GetName()
			       << "]" << endl;
                        if (att == NULL)
			  cout
			    << " ************************************ att is null "
			    << endl;
                        else
			  att->printLog();
                        find = 1;
		      }

		  }
                i = i + 1;
                if (i == this->getNAttributes())
		  find = 2;
	      }
            if (_blog)
	      _listOfAttributes->Print();
            if (find != 1)
	      throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							 Form(
							      "TDataSpecification::getAttribute[%s]  attribute with name does not exist",
							      name.Data()));
	  }
      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }

    return att;

  }

  TAttribute* TDataSpecification::getAttribute(Int_t index, bool onlyAttribute) const
  {
    try
      {
        if ((index > _listOfAttributes->GetEntries()) || (index < 0))
	  {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form(
							    "TDataSpecification::getAttribute[%d]  The index is out of range (%d).",
							    index, _listOfAttributes->GetEntries()));
	  }
	if(!onlyAttribute)
	  return dynamic_cast<URANIE::DataServer::TAttribute *>(_listOfAttributes->At(index));

	Int_t theIndex = -1;
        Int_t cptvar = 0;
        for (Int_t ivar = 0; ivar < _listOfAttributes->GetEntries() ; ivar++)
	  {
            if ((getAttribute(ivar)->getOrigin() == TAttribute::kAttribute) ||
                (getAttribute(ivar)->getOrigin() == TAttribute::kConstant)  )
	      cptvar++;
            if (cptvar == index + 1)
	      {
                theIndex = ivar;
                break;
	      }
	  }
        if (theIndex > -1)
	  return getAttribute(theIndex);
        else
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form(
							  "TDataServer::getAttribute[%d]  The total number of attributes is %d.",
							  index, cptvar));
	
      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
        throw ue;
        exit(-1);
      }
  }

  Int_t TDataSpecification::getAttributeIndex(TString name, bool onlyAttribute) const
  {
    TAttribute *att =
      dynamic_cast<URANIE::DataServer::TAttribute *>(_listOfAttributes->FindObject(
										   name));

    try
      {
        if (att)
	  {
            return getAttributeIndex(att, onlyAttribute);
	  }
        else
	  {
            //_listOfAttributes->Print();
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form(
							    "TDataSpecification::getAttributeIndex[%s]  attribute with name does not exist",
							    name.Data()));
	  }
      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
        throw ue;
        exit(-1);
      }
  }

  Int_t TDataSpecification::getAttributeIndex(TAttribute* att, bool onlyAttribute) const
  {
    try
      {
        if (!isAttribute(att->GetName()))
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form(
							  "TDataSpecification::getAttributeIndex[%s]  attribute with name does not exist",
							  att->GetName()));
      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
        throw ue;
        exit(-1);
      }
    if(!onlyAttribute)
      return _listOfAttributes->IndexOf(att);
    else
      {
	Int_t natt=0;
	Int_t nvar = getNAttributes();
	for (Int_t ivar = 0; ivar < nvar; ivar++)
	  {
	    if (_blog)
	      cout << " ** ";
	    if ( (getAttribute(ivar)->getOrigin() == TAttribute::kAttribute) ||
		 (getAttribute(ivar)->getOrigin() == TAttribute::kConstant)   )
	      {
		if( getAttribute(ivar) == att )
		  break;
		natt++;
		if (_blog)
		  cout << " + ";
	      }
	    else
	      {
		if (_blog)
		  cout << "   ";
	      }

	    if (_blog)
	      cout << getAttribute(ivar)->GetName()
		   << " : " << natt << "/" << nvar << endl;
	  }
	return natt;
      }
  }

  Bool_t TDataSpecification::isAttribute(TString name) const
  {
    Bool_t bfound = kFALSE;

    TAttribute *att =
      dynamic_cast<URANIE::DataServer::TAttribute *>(_listOfAttributes->FindObject(
										   name));
    if (att)
      bfound = kTRUE;
    return bfound;
  }

  Int_t TDataSpecification::getNAttributes(bool onlyAttribute) const
  {
    Int_t natt = 0;
    if (_blog)
      cout << endl << "****** TDataSpecification::getNAttributes" << endl;

    if(!onlyAttribute)
      return _listOfAttributes->GetEntries();
    
    Int_t nvar = getNAttributes();
    for (Int_t ivar = 0; ivar < nvar; ivar++)
      {
	if (_blog)
	  cout << " ** ";
	if ( (getAttribute(ivar)->getOrigin() == TAttribute::kAttribute) ||
	    (getAttribute(ivar)->getOrigin() == TAttribute::kConstant)   )
	  {
	    natt++;
	    if (_blog)
	      cout << " + ";
	  }
	else
	  {
	    if (_blog)
	      cout << "   ";
	  }

	if (_blog)
	  cout << getAttribute(ivar)->GetName()
	       << " : " << natt << "/" << nvar << endl;
      }
    if (_blog)
      cout << "****** End of TDataSpecification::getNAttributes[" << natt << "]"
	   << endl;
    
    return natt;
  }

  Int_t TDataSpecification::getNOutputAttributes() const
  {
    Int_t natt = 0;
    if (_blog)
      cout << endl << "****** TDataSpecification::getNAttributes" << endl;

    Int_t nvar = getNAttributes();
    for (Int_t ivar = 0; ivar < nvar; ivar++)
      {
	if (_blog)
	  cout << " ** ";
	if ( getAttribute(ivar)->isOutput() )
	  {
	    natt++;
	    if (_blog)
	      cout << " o ";
	  }
	else
	  {
	    if (_blog)
	      cout << " ? ";
	  }
	if (_blog)
	  cout << getAttribute(ivar)->GetName() << " : " << natt << "/" << nvar << endl;
      }
    
    return natt;
  }

  Int_t TDataSpecification::getNInputAttributes() const
  {
    Int_t natt = 0;
    if (_blog)
      cout << endl << " ** TDataSpecification::getNInputAttributes" << endl;

    Int_t nvar = getNAttributes();
    for (Int_t ivar = 0; ivar < nvar; ivar++)
      {
	if (_blog)
	  cout << " ** ";
	if (getAttribute(ivar)->isInput())
	  {
	    natt++;
	    if (_blog)
	      cout << " i ";
	  }
	else
	  {
	    if (_blog)
	      cout << " ? ";
	  }
	if (_blog)
	  cout << getAttribute(ivar)->GetName()
	       << " : " << natt << "/" << nvar << endl;
      }

    if (_blog)
      cout << " ** End of TDataSpecification::getNInputAttributes[" << natt << "]"
	   << endl;
    return natt;
  }
  
  ////////////////////////////////////////////////////////////
  //  PrintLog 
  ///////////////////////////////////////////////////////////
  void TDataSpecification::printLog()
  {
    Int_t natt = getNAttributes();

    cout << "*******************************" << endl;
    cout << "** TDataSpecification::printLog  ******" << endl;
    cout << "**   Name[" << GetName() << "] Title[" << GetTitle() << "]"
	 << endl;
    cout << "**   relationName[" << _relationName << "]" << endl;
    cout << "**   attributs[" << natt << "]" << endl;
    cout << "**** With _listOfAttributes " << endl;
    for (Int_t iatt = 0; iatt < natt; iatt++)
      {
        cout << " Attribute[" << iatt << "/" << natt << "]" << endl;
        getAttribute(iatt)->printLog();
      }
    cout
      << "** TDataSpecification::fin de printLog *******************************"
      << endl;
  }

  ////////////////////////////////////////////////////////////
  //  Protected method 
  ////////////////////////////////////////////////////////////
  void TDataSpecification::clear()
  {
    _listOfAttributes->Clear();
  }

} // Fin du namespace DataServer
} // Fin du namespace URANIE
