//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TDataSpecification.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   jeu avr 26 14:14:11 CEST 2007
 
 $Revision$
 $Date$
 
 \brief Interface of the class URANIE::DataServer::TDataSpecification
 
 */
#ifndef TDataSpecification_H
#define TDataSpecification_H

// STL
#include <string>

#include <iostream>
using std::cout;
using std::endl;

// ROOT
#include "TList.h"

//Uranie
#include "TAttribute.h"
#include "DataServer.h"

/*!
 @class URANIE::DataServer::TDataSpecification
 @brief The collection of attributes specifying how to interpret the input data attributes.

 A description of the attributes.

 */
namespace URANIE
{
namespace DataServer
{
class TDataSpecification: public TNamed
{
protected:
    //---------------------------------------------
    /** @name Protected members
     **/
    //@{
    Bool_t _blog;                          ///< Log printing
    TString _relationName;		            ///< Relation name
    TList *_listOfAttributes;                     ///< The list of attributes
    //! Clear the list of attributes
    void clear();
    //@}

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    TDataSpecification();
    //! Default constructor with a name and a title
    TDataSpecification(const char *name, const char* title);
     //! Copy constructor 
    TDataSpecification(URANIE::DataServer::TDataSpecification *tdsp);
    //! Default Desctructor
    virtual ~TDataSpecification();
    //@}

    //---------------------------------------------
    /** @name Setting and Getting informations
     **/
    //@{
    TString GetClassName() const
    {
        return TString("TDataSpecification");
    }

    //! Defines the realtion name
    TString getRelationName() const
    {
        return _relationName;
    }
    void setRelationName(TString name)
    {
        _relationName = name;
    }

    TString getVariablesNames(TString separator = ":", bool noIterator=false);
    TString getVariablesTypes(TString separator = ":", bool noIterator=false);
    TString getVariablesNamesTypes(TString separator = ":", bool noIterator=false);
    //!Check that content is equivalent (the boolean is for partial)
    bool isComparableToTDSP(TDataSpecification *tdsp, bool isJustContained=false);
    bool isComparableToTDSP(TString namestypes, bool isJustContained=false);
    //@}

    //---------------------------------------------
    /** @name management of attributes by the TList
     **/
    //@{
    //! Gets the number of Attributes
    Int_t getNAttributes(bool onlyAttribute=false) const;
    Int_t getNOutputAttributes() const;
    Int_t getNInputAttributes() const;
    //! Adds an attribute in the TList
    void addAttribute(TAttribute*, bool dontAddSizeOne=false);
    //! Deletes an attribute given by this name
    void delAttribute(TString name);
    //! Returns the attribute given by the name
    /*!
     \exception If the attribute does not exist.
     */
    TAttribute* getAttribute(TString name) const;
    //! Returns the attribute given by the index, bool is for removing internal and iterator.
    TAttribute* getAttribute(Int_t index, bool onlyAttribute=false) const;
    //! Returns the index of an attribute
    /*!
     \return Returns the index of the attributes given by the name in the collection.
     Returns -1 when attribute not found.
     */
    Int_t getAttributeIndex(TAttribute* att, bool onlyAttribute=false) const;

    //! Returns the index of an attribute given by the name
    /*!
     \return Returns the index of the attributes givent ny the name in the collection.
     \exception UWarningExceptions when no attribute with this name exist in the collection.
     */
    Int_t getAttributeIndex(TString name, bool onlyAttribute=false) const;

    //! Checks if the attribute given by a name exists.
    /*!
     \return kTRUE if an attribute given by the name exists, else kFALSE 
     */
    Bool_t isAttribute(TString name) const;

    //!  Gets the list of attributes from a string
    /*!
      Returns the list of attributes. The character "*" means all attributes.
    */
    TList * getListOfAttributes(const char* sListOfAttributes = "*", bool onlyAttribute=false);

    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    void setLog()
    {
        _blog = kTRUE;
    }
    void unsetLog()
    {
        _blog = kFALSE;
    }
    void changeLog()
    {
        _blog = _blog ? kFALSE : kTRUE;
    }
    Bool_t getLog()
    {
        return _blog;
    }
    void printLog();
    //@}

    ClassDef(URANIE::DataServer::TDataSpecification, ID_DATASERVER)
    //Sp�cification des variables contenues dans un serveur de donn�es	  

};
} // Fin du namespace DataServer
} // Fin du namespace URANIE

#endif
