//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TExponentialDistribution.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Thu Jan 18 17:08:58 CET 2007

 \brief Interface of the class URANIE::DataServer::TExponentialDistribution

 */
/**

 * @class URANIE::DataServer::TExponentialDistribution
 *
 * @brief Description of the class TExponentialDistribution.
 *

 This class defines a stochastic attribute following an Exponential distribution.

 This distribution has two parameters: the rate parameter \f$ \lambda > 0 \f$ and the 
 mode \f$ x_{min} \f$. The law is defined over the interval \f$ [x_{min}, \inf) \f$.

 \par Probability Density Function (PDF):
 \li \f$ f(x) = \lambda * \exp\left( - \lambda * (x-x_{min})\right) \f$ for \f$ x \geq x_{min} \f$
 \li \f$ f(x) = 0.0 \f$ otherwise

 \par Cumulative Distribution Function (CDF):
 \li \f$ F(x) = 0.0 \f$ for \f$ x < x_{min} \f$
 \li \f$ F(x) = 1 - \exp\left( - \lambda * (x-x_{min})\right) \f$ otherwise

 \par Inverse CDF:
 \li \f$ F^{-1}(p) = x_{min} - \ln\left( 1 - p \right)/\lambda  \f$ for \f$ p \in [0,1] \f$

 *
 */
#ifndef TEXPONENTIALDISTRIBUTION_H
#define TEXPONENTIALDISTRIBUTION_H

//Uranie
#include "TStochAttribut.h"

namespace URANIE
{
namespace DataServer
{
class TExponentialDistribution: public TInfiniteDistribution
{
    // Attributes
private:
    Double_t _dparameterLambda;   ///< the rate parameter \f$ \lambda \f$
    Double_t _dparameterMode;     ///< the mode parameter \f$ x_{min} \f$

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    /*!
     Default rate is 0.1, default mode is 0.0 
     */
    TExponentialDistribution();

    //! Constructor setting the name
    /*!
     Default rate is 0.1, default mode is 0.0 

     \param name (TString) Name of the attribute
     */
    TExponentialDistribution(TString name);

    //! Constructor setting name, scale and location
    /*!                 
     \param name (TString) Name of the attribute
     \param lambda (Double_t) the rate parameter \f$ \lambda \f$
     \param mode (Double_t) the mode parameter \f$ x_{min} \f$ (Default = 0.0)
     */
    TExponentialDistribution(TString name, Double_t lambda,
            Double_t mode = 0.0);

    //! Construction of a TExponentialDistribution from an existing TAttribute object
    /*!
     \warning this copy constructor is incomplete. Please do not use.
     (Nicolas Gilardi, 2010.06.11)

     \param att (TAttribute *) Pointer to an existing TAttribute object
     \param lambda (Double_t) the rate parameter \f$ \lambda \f$
     \param mode (Double_t) the mode parameter \f$ x_{min} \f$ (Default = 0.0)                
     */
    TExponentialDistribution(URANIE::DataServer::TAttribute *att,
            Double_t lambda, Double_t mode = 0.0);

    //! Default destructor
    virtual ~TExponentialDistribution();

    //@}
    /** @name Law parameters
     This law has two parameters: the rate \f$ \lambda \f$ and the mode 
     \f$ x_{min} \f$. Both are defined here. 
     **/
    //@{            
    //! Set the rate parameter
    /*!
     \param lambda (Double_t) the rate parameter
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)
     
     \exception UErrorExceptions if lambda < 1e-4.            
     */
    virtual void setParameterLambda(Double_t lambda, Bool_t recompute = kTRUE);

    //! Return the rate parameter
    Double_t getParameterLambda()
    {
        return _dparameterLambda;
    }

    //! Set the mode parameter
    /*!
     \param mode (Double_t) the mode parameter
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE) 
     
     \warning in order to avoid border effects, if the lowerBound attribute 
     is set to minus infinity, it is re-initialised to mode.
     */
    virtual void setParameterMode(Double_t mode, Bool_t recompute = kTRUE);

    //! Return the mode parameter
    Double_t getParameterMode()
    {
        return _dparameterMode;
    }

    //! Set all the parameters
    /*!
     \param lambda (Double_t) the rate parameter \f$ \lambda \f$
     \param mode (Double_t) the mode parameter \f$ x_{min} \f$ (Default = 0.0)                
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)
     */
    virtual void setParameters(Double_t lambda, Double_t mode = 0.0,
            Bool_t recompute = kTRUE);
    //@}

    /** @name Distribution specific methods
     PDF, CDF, etc.
     **/
    //@{
    //! Computes the theoretical values of various statistical measures.
    virtual void computeTheoreticalInformation();

    //! Returns the probability density for the value x
    /*!                
     \param x (Double_t) An acceptable value for the attribute                 
     */
    virtual Double_t getPDF(Double_t x);

    //! Returns the cumulative distribution for the value x
    /*!
     \param x (Double_t) An acceptable value for the attribute                
     */
    virtual Double_t getCDF(Double_t x);

    //! Returns the value corresponding to the cumulative density p
    /*!                 
     \param p (Double_t) A valid probability
     
     \exception UErrorExceptions if p is outside of the interval [0,1]                
     */
    virtual Double_t getInvCDF(Double_t p);
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::DataServer::TExponentialDistribution, ID_DATASERVER)
    //Classe de
};

} // Fin du namespace DataServer
} // Fin du namespace URANIE
#endif
