//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TGammaDistribution.cxx
 \author Guillaume Damblin <guillaume.damblin@cea.fr>
 \date   2009-11-20 10:09:30 \brief Interface of the class URANIE::DataServer::TGammaDistribution

 \brief Implementation of the class URANIE::DataServer::TGammaDistribution
 */

#include <iostream>
using std::cout;
using std::endl;

#include "TMath.h"
#include "TGammaDistribution.h"

ClassImp (URANIE::DataServer::TGammaDistribution)

namespace URANIE
{
namespace DataServer
{

// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------

TGammaDistribution::TGammaDistribution() :
  TInfiniteDistribution("", TStochasticAttribute::kGamma), 
	_dparameterAlpha(1.0), _dparameterScale(1.0), _dparameterLocation(0.0)
{
    setLowerBound(_dparameterLocation);
    if (_blog)
        cout << "Constructing TGammaDistribution" << endl;

    computeTheoreticalInformation();
}

TGammaDistribution::TGammaDistribution(TString str) :
        TInfiniteDistribution(str, TStochasticAttribute::kGamma),
	_dparameterAlpha(1.0), _dparameterScale(1.0), _dparameterLocation(0.0)
{
    setLowerBound(_dparameterLocation);

    if (_blog)
        cout
            << "Constructing URANIE::DataServer::TGammaDistribution with: Name["
            << str << "]" << endl;

    computeTheoreticalInformation();
}

  TGammaDistribution::TGammaDistribution(TString str,  Double_t alpha, Double_t scale, Double_t location) :
      TInfiniteDistribution(str, TStochasticAttribute::kGamma),
      _dparameterAlpha(alpha), _dparameterScale(scale), _dparameterLocation(location)
 {
     setLowerBound(_dparameterLocation);
     if (_blog)
         cout
             << "Constructing URANIE::DataServer::TGammaDistribution with: Name["
             << str << "] alpha[" << alpha << "] scale[" << scale << "]"
             << endl;
     
     setParameterAlpha(alpha, kFALSE);
     setParameterScale(scale, kFALSE);
     setParameterLocation(location, kFALSE);
     computeTheoreticalInformation();
 }

TGammaDistribution::TGammaDistribution(URANIE::DataServer::TAttribute *att, Double_t alpha, Double_t scale, Double_t location) :
  TInfiniteDistribution(att, TStochasticAttribute::kGamma),
  _dparameterAlpha(alpha), _dparameterScale(scale), _dparameterLocation(location)
{

  setLowerBound(_dparameterLocation);
  if (_blog)
    cout
      << "Constructing URANIE::DataServer::TGammaDistribution  from URANIE::DataServer::TAttribute object "
      << att->GetName() << endl;

  setParameterAlpha(alpha, kFALSE);
  setParameterScale(scale, kFALSE);
  setParameterLocation(location, kFALSE);
  computeTheoreticalInformation();
}


// -----------------------------------------------------------------------
//  Destructors
// -----------------------------------------------------------------------       

TGammaDistribution::~TGammaDistribution()
{
  if (_blog)
    cout
      << "Destroying URANIE::DataServer::TGammaDistribution object "
      << GetName() << endl;
}

// -----------------------------------------------------------------------
//  Parameters
// -----------------------------------------------------------------------        

void TGammaDistribution::setParameterAlpha(Double_t alpha, Bool_t recompute)
{
  
  _dparameterAlpha = alpha;
  try
  {
      if(alpha <= 0)
          throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                     Form(
                                                         "TGammaDistribution::setParameterAlpha: alpha parameter value [%f] should be greater than 0",
                                                         alpha));
  }catch (URANIE::Exceptions::UErrorExceptions& ue)
  {
      ue.printMessage();
      throw(ue);
  }
  
  if (recompute)
    computeTheoreticalInformation();
}


void TGammaDistribution::setParameterScale(Double_t scale, Bool_t recompute)
{
  
    _dparameterScale = scale;
    try
    {
        if(scale <= 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form(
                                                           "TGammaDistribution::setParameterScale: scale parameter value [%f] should be greater than 0",
                                                           scale));
    }catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }
    
    if (recompute)
        computeTheoreticalInformation();
}


void TGammaDistribution::setParameterLocation(Double_t location, Bool_t recompute)
{
  
  _dparameterLocation = location;
      
  if (recompute)
    computeTheoreticalInformation();
}

void TGammaDistribution::setParameters( Double_t alpha, Double_t scale, Double_t location, Bool_t recompute)
{
  setParameterAlpha(alpha, kFALSE);
  setParameterScale(scale, kFALSE);
  setParameterLocation(location, kFALSE);
  
  if (recompute)
    computeTheoreticalInformation();
}

// -----------------------------------------------------------------------
//  Theoretical Information
// -----------------------------------------------------------------------        

void TGammaDistribution::computeTheoreticalInformation()
{

    if(_blog)
        cout << " ***  TGammaDistribution::computeTheoreticalInformation " << endl;

    _dtheoreticalMean = _dparameterAlpha*_dparameterScale+_dparameterLocation;

    _dtheoreticalStdDev = TMath::Sqrt(_dparameterAlpha*_dparameterScale*_dparameterScale);

    _dtheoreticalMode = ((_dparameterAlpha<1) ? -1 : (_dparameterAlpha-1) * _dparameterScale);
    
    
}


// -----------------------------------------------------------------------
//  Distribution
// -----------------------------------------------------------------------        

Double_t TGammaDistribution::getPDF(Double_t x)
{	
    
    Double_t y=(x-_dparameterLocation)/_dparameterScale;

    if(y <=0)
        return 0;

    Double_t pdf = TMath::Power(y,_dparameterAlpha-1)*TMath::Exp(-y)/(TMath::Gamma(_dparameterAlpha)*_dparameterScale);
    if ((x >= lowerBound) && (x <= upperBound))
        pdf /= _pmax - _pmin;
    else
        pdf=0;
    
    return pdf;
}

Double_t TGammaDistribution::getCDF(Double_t x)
{

    Double_t cdf = 0;
    if ((x >= lowerBound) && (x <= upperBound))
    {
        cdf = TMath::Gamma(_dparameterAlpha,(x-_dparameterLocation)/_dparameterScale);
        cdf = (cdf - _pmin) / (_pmax - _pmin);
    }
    else if (x > upperBound)
        cdf = 1.0;
   
    return cdf;
}

Double_t TGammaDistribution::getInvCDF(Double_t p)
{

    // Verify that p is a valid probability
    try
    {
        if ((p < 0.0) || (p > 1.0))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form(
                                                           "TGammaDistribution::getInvCDF: parameter p [%f] is not a valid probability: p must be in the interval [0,1]",
                                                           p));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    if (_blog)
        cout << " *** Looking for the value x corresponding to probability p["
             << p << "]" << endl;

    // Take into account the case where at least one of the bounds is finite            
    Double_t realP = p;// * (_pmax - _pmin) + _pmin;
    
    Double_t x0 = lowerBound, x1 = upperBound, px;
    Double_t xcrt = (upperBound + lowerBound) / 2.0;
    //Double_t xtol = x0 * 0.000000000000001, ptol = 0.000000000000001;
    Double_t xtol = ((x0!=0)?x0:0.1) * 1e-14, ptol = 1e-15;
    //Double_t xtol = ((x0!=0)?x0:1) * nextafter(0., HUGE_VAL), ptol = 1e-15;

    Int_t ncpt = 0;

    bool hasbroken=false;
    while (TMath::Abs(x1 - x0) > xtol)
    {
        px = getCDF(xcrt);
        if (_blog)
            cout << " ** number of computations [" << ncpt
                 << "], current solution[" << xcrt
                 << "], corresponding probability[" << px
                 << "], lower limit[" << x0 << "], upper limit[" << x1
                 << "]  diff versus tol ["<<setprecision(20)<<TMath::Abs(x1 - x0)<<"/"<<xtol<<"]";
        if (TMath::Abs(px - realP) < ptol)
        {
            if (_blog)
                cout << " ** solution found[" << xcrt
                     << "], corresponding probability[" << px
                     << "], requested probability[" << p << "]" << endl;
            return xcrt;
        }
        else if (px < realP)
            x0 = xcrt;
        else if (px > realP)
            x1 = xcrt;

        if (_blog)
            cout << " --- new lower limit[" << x0 << "], new upper limit[" << x1
                 << "]";

        xcrt = 0.5 * (x0 + x1);
        if (_blog)
            cout << " ************* new proposed solution[" << xcrt << "]"
                 << endl;
        ncpt++;

        if(ncpt>500)
        {
            hasbroken=true;
            break;
        }
    }

     if(hasbroken)
    {
        URANIE::DataServer::UMessageLogger _localmess("TGamma");
        _localmess <<  " TGammaDistribution::getInvCDF the requested invCDF value might be not accurate, the numerical precisions limit is reached"<< Endl;
    }
    
    return xcrt;
}

// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------

void TGammaDistribution::printLog(Option_t *option)
{

    UNUSED(option);
        
    //TInfiniteDistribution::printLog();
    cout << "*******************************" << endl;
    cout << "** TGammaDistribution::printLog  ******" << endl;
    cout << "**   Law[" << _lawType << "] Gamma" << endl;
    cout << "**   Bounds[" << lowerBound << ", " << upperBound << "]" << endl;
    cout << "**   Parameter Alpha[" << _dparameterAlpha << "]" << endl;
    cout << "**   Parameter Scale[" << _dparameterScale << "]" << endl;
    cout << "**   Parameter Location[" << _dparameterLocation << "]" << endl;
    cout << "**   Theoretical Mean[" << _dtheoreticalMean << "]" << endl;
    cout << "**   Theoretical StdDev[" << _dtheoreticalStdDev << "]" << endl;
    cout << "**   Theoretical Mode[" << _dtheoreticalMode << "]" << endl;

 
    cout << "*******************************" << endl;
}

} // Fin du namespace DataServer
} // Fin du namespace URANIE

// End Of File
