//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TGenParetoDistribution.h
 \author Roman NIKOLAEV <roman.nikolaev@opencascade.com>
 \date   Tue Feb 19 12:00:00 CET 2013

 \brief Interface of the class URANIE::DataServer::TGenParetoDistribution

 */
/*!
 
 \class URANIE::DataServer::TGenParetoDistribution

 \brief Description of the class TGenParetoDistribution.
 
 This class defines a stochastic attribute following a Generalized Pareto Distribution (GPD).
 
 It is specified by three parameters: location \f$ \mu \f$ , scale \f$ \sigma \f$ (strictly positive), and shape \f$ \xi \f$ .   
 
 \par Probability Density Function (PDF):
 \li \f$ f(x) =\frac{1}{\sigma}(1 + \xi(\frac{x-\mu}{\sigma}))^{-(1/\xi +1)} \f$
 
 \par Cumulative Distribution Function (CDF):
 \li \f$ F_{(\xi,\mu,\sigma)}(x) = \begin{cases}1 - \left(1+ \frac{\xi(x-\mu)}{\sigma}\right)^{-1/\xi} \,\: \xi \neq 0, \\ 1 - \exp \left(-\frac{x-\mu}{\sigma}\right) \:\: \xi = 0.\end{cases} \f$
 
 \par Inverse CDF:
 \li \f$ f(p) =\frac{((1-p)^{-\xi}-1)\delta}{\xi}+\mu \f$
 
 
 For the details about estimation of the parameters see:
 
 \li J. R. M. Hosking and J. R. Wallis. Parameter and Quantile Estimation for the Generalized Pareto Distribution
 Source: Technometrics, Vol. 29, No. 3 (Aug., 1987), pp. 339-349
 
 \li Scott D. Grimshaw. Computing Maximum Likelihood Estimates for the Generalized Pareto Distribution
 Source: Technometrics, Vol. 35, No. 2 (May, 1993), pp. 185-191
 
 \li V. P. SINGH & H. GUO. Parameter estimation for 3-parameter generalized pareto distribution by the principle of maximum entropy (POME)
 Hydrological Sciences -Journal- des Sciences Hydrologiques Vol. 40, No. 2(April, 1995), pp. 165 - 181.
 
 */

#ifndef TGENPARETODISTRIBUTION_H
#define TGENPARETODISTRIBUTION_H

//Uranie
#include "TStochAttribut.h"

namespace URANIE
{
namespace DataServer
{
class TGenParetoDistribution: public TInfiniteDistribution
{
private:
protected:
    Double_t _dparameterXi;     ///< The shape parameter \f$ \xi\f$  
    Double_t _dparameterMu;     ///< The location parameter \f$ \mu\f$ 
    Double_t _dparameterSigma;  ///< The scale parameter \f$ \sigma\f$

    TList *_listOfInputAttributes;             //< The list of attributes
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    TGenParetoDistribution();

    //! Constructor by name only
    /*! 
     
     \param name (TString) Name of the attribute
     */
    TGenParetoDistribution(TString name);

    //! Constructor by name and statistics
    /*! 
     
     \param name (TString) Name of the attribute
     \param mu (Double_t) location parameter.
     \param sigma (Double_t) scale parameter (sigma > 0).
     \param xi (Double_t) shape parameter.
     */
    TGenParetoDistribution(TString name, Double_t mu, Double_t sigma,
            Double_t xi);

    //! Constructor by attribute and statistics
    /*!                 
     \param att (URANIE::DataServer::TAttribute*) pointer to attribute
     \param mu (Double_t) location parameter.
     \param sigma (Double_t) scale parameter (sigma > 0).
     \param xi (Double_t) shape parameter.
     */
    TGenParetoDistribution(URANIE::DataServer::TAttribute *att, Double_t mu,
            Double_t sigma, Double_t xi);
    //! Default destructor
    virtual ~TGenParetoDistribution();
    //@}
    //---------------------------------------------
    /** @name Getters and Setters methods
     Getting and setting attribute values.
     **/
    //@{
    //! Sets the location parameter
    virtual void setParameterMu(Double_t mu, Bool_t recompute = kTRUE);
    //! Sets the scale parameter
    virtual void setParameterSigma(Double_t sigma, Bool_t recompute = kTRUE);
    //! Sets the shape parameter
    virtual void setParameterXi(Double_t xi, Bool_t recompute = kTRUE);
    //! Sets the scale, shape and location parameters
    virtual void setParameters(Double_t mu, Double_t sigma, Double_t xi,
			       Bool_t recompute = kTRUE);

    //! Gets the location
    virtual Double_t getParameterMu();
    //! Gets the scale
    virtual Double_t getParameterSigma();
    //! Gets the scale
    virtual Double_t getParameterXi();

    /*! Defines equation to estimate the Xi parameter.
     Used by the Method of Moments.
     */
    Double_t getXiEquation(Double_t x, Double_t skewness);

    //! Calculates the \f$ \frac{x-\mu}{\sigma}\f$ value.
    /*!                                                
     \exception UErrorExceptions if case if GPD is not defined for the provided x value.
     */
    Double_t getZ(Double_t x);

    //@}
    /** @name Statistical methods
     The statistical methods ( pdf, cdf, quantile, ...).
     **/
    //@{
    //! Computes the theoretical values of various statistical measures, like mean, standard deviation, mode, etc.
    virtual void computeTheoreticalInformation();

    //! Returns the probability density for the value x
    /*!                
     \param x (Double_t) An acceptable value for the attribute                 
     */
    virtual Double_t getPDF(Double_t x);

    //! Returns the cumulative distribution for the value x
    /*!
     \param x (Double_t) An acceptable value for the attribute                
     */
    virtual Double_t getCDF(Double_t x);

    //! Returns the value corresponding to the cumulative density p
    /*!                                 
     \param p (Double_t) A valid probability
     
     \exception UErrorExceptions if p is outside of the interval [0,1]                
     */
    virtual Double_t getInvCDF(Double_t p);
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}


    ClassDef(URANIE::DataServer::TGenParetoDistribution, ID_DATASERVER)
    //Classe de             
};
} // Fin du namespace DataServer
} // Fin du namespace URANIE
#endif
