//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TGeneralizedNormalV2Distribution.cxx
 \author Rudy Chocat <rudy.chocat@cea.fr>
 \date   28 oct. 2021 11:13:50

 \brief Implementation of the class URANIE::DataServer::TGeneralizedNormalV2Distribution

 */
#include <iostream>
using std::cout;
using std::endl;

#include "TMath.h"
#include "TNormalDistribution.h"
#include "TGeneralizedNormalV2Distribution.h"
 #include "Math/QuantFuncMathCore.h"

ClassImp (URANIE::DataServer::TGeneralizedNormalV2Distribution)

namespace URANIE
{
namespace DataServer
{
// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------

TGeneralizedNormalV2Distribution::TGeneralizedNormalV2Distribution() :
        TInfiniteDistribution("", TStochasticAttribute::kGeneralizedNormalV2), _dparameterXi(
                0.0), _dparameterAlpha(1.0), _dparameterKappa(0.0)
{
    //  cout << "construit TGeneralizedNormalV2Distribution" << endl;
    if(_dparameterKappa<0.0)
    {
        TAttribute::setLowerBound( _dparameterXi + _dparameterAlpha/_dparameterKappa+ 10e-12, true );
    }
    else if(_dparameterKappa>0.0)
    {
        TAttribute::setUpperBound( _dparameterXi + _dparameterAlpha/_dparameterKappa- 10e-12, true );
    }
    computeTheoreticalInformation();
}

TGeneralizedNormalV2Distribution::TGeneralizedNormalV2Distribution(TString str) :
        TInfiniteDistribution(str, TStochasticAttribute::kGeneralizedNormalV2), _dparameterXi(
                0.0), _dparameterAlpha(1.0), _dparameterKappa(0.0)
{
    //  cout << "Constructeur URANIE::DataServer::TGeneralizedNormalV2Distribution[" << this << "]str[" << str<<"]"<<endl;
    if(_dparameterKappa<0.0)
    {
        TAttribute::setLowerBound( _dparameterXi + _dparameterAlpha/_dparameterKappa+ 10e-12, true );
    }
    else if(_dparameterKappa>0.0)
    {
        TAttribute::setUpperBound( _dparameterXi + _dparameterAlpha/_dparameterKappa- 10e-12, true );
    }
    computeTheoreticalInformation();
}

TGeneralizedNormalV2Distribution::TGeneralizedNormalV2Distribution(TString str, Double_t xi,
        Double_t alpha, Double_t kappa) :
        TInfiniteDistribution(str, TStochasticAttribute::kGeneralizedNormalV2)
{
    //  cout << "Constructeur URANIE::DataServer::TGeneralizedNormalV2Distribution[" << this << "]str[" << str<<"] Mean[" << mean <<"] Alpha[" << alpha << "]" Beta[" << kappa << "]" << endl;
     _dparameterXi = xi;
    _dparameterAlpha = alpha;
    _dparameterKappa = kappa;
    computeTheoreticalInformation();
    
    setParameters(xi, alpha, kappa, true);
    
}

TGeneralizedNormalV2Distribution::TGeneralizedNormalV2Distribution(URANIE::DataServer::TAttribute *att,
        Double_t xi, Double_t alpha, Double_t kappa) :
        TInfiniteDistribution(att, TStochasticAttribute::kGeneralizedNormalV2)
{
    //  cout << "Constructeur URANIE::DataServer::TGeneralizedNormalV2Distribution att[" << this << "]att->getName()[" << att->getName()<<"] Mean[" << mean <<"] Alpha[" << alpha << "]" Beta[" << beta << "]" << endl;
    _dparameterXi = xi;
    _dparameterAlpha = alpha;
    _dparameterKappa = kappa;
    computeTheoreticalInformation();
    setParameters(xi, alpha, kappa, true);
}

// -----------------------------------------------------------------------
//  Destructors
// -----------------------------------------------------------------------

TGeneralizedNormalV2Distribution::~TGeneralizedNormalV2Distribution()
{
    //  cout << "detruit TGeneralizedNormalV2Distribution" << endl; 

}

// -----------------------------------------------------------------------
//  Parameters
// -----------------------------------------------------------------------        

void TGeneralizedNormalV2Distribution::setParameterXi(Double_t xi, Bool_t recompute)
{
    _dparameterXi = xi;

    if (recompute)
        computeTheoreticalInformation();
}

void TGeneralizedNormalV2Distribution::setParameterAlpha(Double_t alpha, Bool_t recompute)
{
    try
    {
        Double_t refMean = 0.0; 
        refMean = TMath::Abs(_dtheoreticalMean);
        refMean = 1e-6 * refMean;
        
        if (alpha < 0.0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TGeneralizedNormalV2Distribution::setParameterAlpha: parameter alpha [%f] is negative: it must be strictly positive",
                            alpha));
        if (alpha == 0.0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TGeneralizedNormalV2Distribution::setParameterAlpha: parameter alpha [%f] is null: it must be strictly positive",
                            alpha));            
        if (TMath::Abs(alpha) < refMean)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TGeneralizedNormalV2Distribution::setParameterAlpha: parameter alpha [%f] is too low: lower limit is 1e-6 * mean [%f]",
                            alpha, 1e-6 * TMath::Abs(refMean)));
        if ((refMean==0.0)&&(TMath::Abs(alpha) < 1e-6 ))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TGeneralizedNormalV2Distribution::setParameterAlpha: parameter alpha [%f] is too close to zero: lower limit is 1e-6",
                            alpha));
        
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    _dparameterAlpha = alpha;

    if (recompute)
        computeTheoreticalInformation();
}

void TGeneralizedNormalV2Distribution::setParameterKappa(Double_t kappa, Bool_t recompute)
{
//     try
//     {
//         if (kappa < 0.0)
//             throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
//                     Form(
//                             "TGeneralizedNormalV2Distribution::setParameterKappa: parameter kappa [%f] is negative: it must be positive",
//                             kappa));
//     } catch (URANIE::Exceptions::UErrorExceptions& ue)
//     {
//         ue.printMessage();
//         throw(ue);
//     }

    _dparameterKappa = kappa;

    if (recompute)
        computeTheoreticalInformation();
}

void TGeneralizedNormalV2Distribution::setParameters(Double_t xi, Double_t alpha, Double_t kappa,
        Bool_t recompute)
{
    setParameterXi(xi, kFALSE);
    setParameterAlpha(alpha, kFALSE);
    setParameterKappa(kappa, kFALSE);
    if(_dparameterKappa<0.0)
    {
        TAttribute::setLowerBound( _dparameterXi + _dparameterAlpha/_dparameterKappa+ 10e-12, true );
    }
    else if(_dparameterKappa>0.0)
    {
        TAttribute::setUpperBound( _dparameterXi + _dparameterAlpha/_dparameterKappa- 10e-12, true );
    }
    if (recompute)
        computeTheoreticalInformation();
}

// -----------------------------------------------------------------------
//  Theoretical Information
// -----------------------------------------------------------------------        

void TGeneralizedNormalV2Distribution::computeTheoreticalInformation()
{
//     if(_dparameterKappa<0.0)
//     {
//         lowerBound = _dparameterXi + _dparameterAlpha/_dparameterKappa + 10e-9;
//     }
//     else if(_dparameterKappa>0.0)
//     {
//         upperBound = _dparameterXi + _dparameterAlpha/_dparameterKappa - 10e-9;;
//     }
//     
    
    Double_t theoreticalMean = 0.0; 
    Double_t theoreticalStdDev = 1.0; 
    Double_t theoreticalMode = 0.0; 
    if(_dparameterKappa!=0.0)
    {
        theoreticalMean = _dparameterXi - ( _dparameterAlpha / _dparameterKappa ) * ( TMath::Exp( _dparameterKappa*_dparameterKappa /2.0 ) - 1.0 )    ;
        theoreticalStdDev = TMath::Sqrt( ( (_dparameterAlpha * _dparameterAlpha)/(_dparameterKappa * _dparameterKappa) ) * TMath::Exp( _dparameterKappa*_dparameterKappa ) * ( TMath::Exp( _dparameterKappa*_dparameterKappa ) - 1.0 ));
    }
    else
    {
        theoreticalMean = _dparameterXi ;
        theoreticalStdDev = TMath::Sqrt( _dparameterAlpha * _dparameterAlpha );
    }

    theoreticalMode = _dparameterXi;
    
    _dtheoreticalMean = theoreticalMean;
    _dtheoreticalStdDev = theoreticalStdDev; 
    _dtheoreticalMode = theoreticalMode;
}

// -----------------------------------------------------------------------
//  Distribution
// -----------------------------------------------------------------------        

Double_t TGeneralizedNormalV2Distribution::getPDF(Double_t x)
{
    Double_t pdf = 0;
    Double_t y = 0;
// if(_dparameterKappa<0.0)
//     {
//         TAttribute::setLowerBound( _dparameterXi + _dparameterAlpha/_dparameterKappa+ 10e-12, true );
//     }
//     else if(_dparameterKappa>0.0)
//     {
//         TAttribute::setUpperBound( _dparameterXi + _dparameterAlpha/_dparameterKappa- 10e-12, true );
//     }
    
    if ((x > lowerBound) && (x < upperBound))
    {
        if(_dparameterKappa==0.0)
            y = (x - _dparameterXi) / _dparameterAlpha ;
        else
            y = - (1.0 / _dparameterKappa) * TMath::Log( 1.0 - ( _dparameterKappa * ( x - _dparameterXi )  ) / ( _dparameterAlpha ) ) ;
        
        pdf = TNormalDistribution().getPDF(y) / ( _dparameterAlpha - _dparameterKappa * (x - _dparameterXi) ) ;

        // Take into account the case where at least one of the bounds is finite
        pdf /= _pmax - _pmin;
    }
    return pdf;
}

Double_t TGeneralizedNormalV2Distribution::getCDF(Double_t x)
{
    Double_t cdf = 0;
    Double_t y = 0;
//     cout << "up " << TAttribute::getUpperBound() << endl;
//     if(_dparameterKappa<0.0)
//     {
//         lowerBound = _dparameterXi + _dparameterAlpha/_dparameterKappa + 10e-9;
//     }
//     else if(_dparameterKappa>0.0)
//     {
//         upperBound = _dparameterXi + _dparameterAlpha/_dparameterKappa - 10e-9;;
//     }
//     cout <<" up = " << upperBound <<endl;
    if ((x > lowerBound) && (x < upperBound))
    {
        if(_dparameterKappa==0.0)
            y = (x - _dparameterXi) / _dparameterAlpha ;
        else
            y = - (1.0 / _dparameterKappa) * TMath::Log( 1.0 - ( _dparameterKappa * ( x - _dparameterXi )  ) / ( _dparameterAlpha ) ) ;
        
        cdf = TNormalDistribution().getCDF(y);
        
        // Take into account the case where at least one of the bounds is finite
        //cdf /= _pmax - _pmin;
        cdf = (cdf - _pmin) / (_pmax - _pmin);
    }
    else if (x >= upperBound)
        cdf = 1.0;

    return cdf;
}

Double_t TGeneralizedNormalV2Distribution::getInvCDF(Double_t p)
{
    
        // Verify that p is a valid probability
    try
    {
        if ((p < 0.0) || (p > 1.0))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form(
                                                           "TGeneralizedNormalV2Distribution::getInvCDF: parameter p [%f] is not a valid probability: p must be in the interval [0,1]",
                                                           p));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    if (_blog)
        cout << " *** Looking for the value x corresponding to probability p["
             << p << "]" << endl;

    // Take into account the case where at least one of the bounds is finite            
    Double_t realP = p;// * (_pmax - _pmin) + _pmin;
    
    Double_t x0 = lowerBound, x1 = upperBound, px;
    Double_t xcrt = (upperBound + lowerBound) / 2.0;
    //Double_t xtol = x0 * 0.000000000000001, ptol = 0.000000000000001;
    Double_t xtol = ((x0!=0)?x0:0.1) * 1e-14, ptol = 1e-20;
    //Double_t xtol = ((x0!=0)?x0:1) * nextafter(0., HUGE_VAL), ptol = 1e-15;

    Int_t ncpt = 0;
    bool hasbroken=false;
    while (TMath::Abs(x1 - x0) > xtol)
    {
        px = getCDF(xcrt);
        if (_blog)
            cout << " ** number of computations [" << ncpt
                 << "], current solution[" << xcrt
                 << "], corresponding probability[" << px
                 << "], lower limit[" << x0 << "], upper limit[" << x1
                 << "]  diff versus tol ["<<setprecision(20)<<TMath::Abs(x1 - x0)<<"/"<<xtol<<"]";
        if (TMath::Abs(px - realP) < ptol)
        {
            if (_blog)
                cout << " ** solution found[" << xcrt
                     << "], corresponding probability[" << px
                     << "], requested probability[" << p << "]" << endl;
            return xcrt;
        }
        else if (px < realP)
            x0 = xcrt;
        else if (px > realP)
            x1 = xcrt;

        if (_blog)
            cout << " --- new lower limit[" << x0 << "], new upper limit[" << x1
                 << "]";

        xcrt = 0.5 * (x0 + x1);
        if (_blog)
            cout << " ************* new proposed solution[" << xcrt << "]"
                 << endl;
        ncpt++;

        if(ncpt>500)
        {
            hasbroken=true;
            break;
        }
    }

     if(hasbroken)
    {
        URANIE::DataServer::UMessageLogger _localmess("TGeneralizedNormalV2Distribution");
        _localmess <<  " TGeneralizedNormalV2Distribution::getInvCDF the requested invCDF value might be not accurate, the numerical precisions limit is reached"<< Endl;
    }
    
    return xcrt;
    
}

// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------

void TGeneralizedNormalV2Distribution::printLog(Option_t *option)
{
    //  TInfiniteDistribution::printLog();
    TString soption = option;
    soption.ToLower();
    cout << "*******************************" << endl;
    cout << "** TGeneralizedNormalV2Distribution::printLog  ******" << endl;
    cout << "**   Law[" << _lawType << "] Normal" << endl;
    cout << "**   Bounds[" << lowerBound << ", " << upperBound << "]" << endl;
    cout << "**   Position parameter xi[" << _dparameterXi << "]" << endl;
    cout << "**   Scale parameter alpha [" << _dparameterAlpha << "]" << endl;
    cout << "**   Shape parameter beta [" << _dparameterKappa << "]" << endl;
    cout << "**   Theoretical Mean[" << _dtheoreticalMean << "]" << endl;
    cout << "**   Theoretical StdDev[" << _dtheoreticalStdDev << "]" << endl;
//     cout << "**   Theoretical Mode[" << _dtheoreticalMode << "]" << endl;
    cout << "*******************************" << endl;
}

} // Fin du namespace DataServer
} // Fin du namespace URANIE

