//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TGumbelMaxDistribution.h
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date   Mon Jan 19 12:00:00 CET 2009

 \brief Interface of the class URANIE::DataServer::TGumbelMaxDistribution

 */
/**

 * @class URANIE::DataServer::TGumbelMaxDistribution
 *
 * @brief Description of the class TGumbelMaxDistribution.

 This class defines a stochastic attribute following an Gumbel Max distribution.

 The Gumbel Max law describes the distribution of the maxima of a set of samples, 
 like, for instance, the distribution of the annual maximum level of a river.

 This law has two parameters: the mode \f$ \mu \f$ and the scale \f$ \beta \f$, 
 which is strictly positive.

 \par Probability Density Function (PDF):
 \li \f$ f(x) =  z * \exp(-z)/\beta \f$ where \f$ z = \exp\left(-(x - \mu)/\beta \right) \f$

 \par Cumulative Distribution Function (CDF):
 \li \f$ F(x) = \exp(-z)\f$ where \f$ z = \exp\left(-(x - \mu)/\beta \right) \f$

 \par Inverse CDF:
 \li \f$ F^{-1}(p) = \mu - \beta \ln\left( - \ln(p) \right)\f$ for \f$ p \in [0,1] \f$ 


 */
#ifndef TGUMBELMAXDISTRIBUTION_H
#define TGUMBELMAXDISTRIBUTION_H

//Uranie
#include "TStochAttribut.h"

namespace URANIE
{
namespace DataServer
{
class TGumbelMaxDistribution: public TInfiniteDistribution
{
private:
    Double_t _dparameterMode;  ///< the mode parameter \f$ \mu \f$
    Double_t _dparameterScale; ///< the scale parameter \f$ \beta \f$
protected:
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    /*!
     Default mode is 0.0, default scale is 1.0
     */
    TGumbelMaxDistribution();

    //! Constructor setting the name
    /*!
     Default mode is 0.0, default scale is 1.0
     
     \param name (TString) Name of the attribute
     */
    TGumbelMaxDistribution(TString name);

    //! Constructor setting name, mode and scale
    /*!                 
     \param name (TString) Name of the attribute
     \param mu (Double_t) Mode parameter \f$ \mu \f$
     \param beta (Double_t) Scale parameter \f$ \beta \f$
     */
    TGumbelMaxDistribution(TString name, Double_t mu, Double_t beta);

    //! Construction of a TGumbelMaxDistribution from an existing TAttribute object
    /*!
     \warning this copy constructor is incomplete. Please do not use.
     (Nicolas Gilardi, 2010.06.11)
     
     \param att (TAttribute*) pointer to attribute
     \param mu (Double_t) Mode parameter \f$ \mu \f$
     \param beta (Double_t) Scale parameter \f$ \beta \f$
     */
    TGumbelMaxDistribution(URANIE::DataServer::TAttribute *att, Double_t mu,
            Double_t beta);

    //! Default destructor
    virtual ~TGumbelMaxDistribution();
    //@}
    //---------------------------------------------
    /** @name Law parameters
     This law has two parameters: the mode \f$ \mu \f$ and the scale \f$ \beta \f$.
     Both are defined here.
     **/
    //@{
    //! Set the mode parameter
    /*!
     \param mu (Double_t) Mode parameter \f$ \mu \f$
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE) 
     */
    virtual void setParameterMode(Double_t mu, Bool_t recompute = kTRUE);

    //! Get the mode parameter
    virtual Double_t getParameterMode()
    {
        return _dparameterMode;
    }

    //! Set the scale parameter
    /*!
     \param beta (Double_t) Scale parameter \f$ \beta \f$
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)
     
     \exception UErrorExceptions if beta <  1e-6 * mu                
     */
    virtual void setParameterScale(Double_t beta, Bool_t recompute = kTRUE);

    //! Get the scale parameter
    virtual Double_t getParameterScale()
    {
        return _dparameterScale;
    }

    //! Set all the parameters
    /*!
     \param mu (Double_t) Mode parameter \f$ \mu \f$
     \param beta (Double_t) Scale parameter \f$ \beta \f$
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE) 
     */
    virtual void setParameters(Double_t mu, Double_t beta, Bool_t recompute =
            kTRUE);
    //@}

    /** @name Distribution specific methods
     PDF, CDF, etc.
     **/
    //@{
    //! Computes the theoretical values of various statistical measures.
    virtual void computeTheoreticalInformation();

    //! Returns the probability density for the value x
    /*!
     \param x (Double_t) An acceptable value for the attribute                                  
     */
    virtual Double_t getPDF(Double_t x);

    //! Returns the cumulative distribution for the value x
    /*!
     \param x (Double_t) An acceptable value for the attribute
     */
    virtual Double_t getCDF(Double_t x);

    //! Returns the value corresponding to the cumulative density p
    /*!
     \param p (Double_t) A valid probability
     
     \exception UErrorExceptions if p is outside of the interval [0,1]
     */
    virtual Double_t getInvCDF(Double_t p);
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::DataServer::TGumbelMaxDistribution, ID_DATASERVER)
    //Classe de
};
} // Fin du namespace DataServer
} // Fin du namespace URANIE
#endif
