//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TInvGammaDistribution.h
 \author Guillaume Damblin <guillaume.damblin@cea.fr>
 \date   2009-11-20 10:09:30 \brief Interface of the class URANIE::DataServer::TInvGammaDistribution

 */
/**

 * @class URANIE::DataServer::TInvGammaDistribution
 *
 * @brief Description of the class TInvGammaDistribution.
 *
 This class defines a stochastic attribute following an InvGamma distribution.
 
 The distribution is defined over the interval \f$ ]0,+\infty [ \f$. It only depends on a shape parameter \f$ \alpha>0 \f$ in the standardized form.
 To scale and/or shift the distribution, one can use the scale parameter and the location parameter respectively.
 
 \par Probability Density Function (PDF):
 
 \li \f$ f(x) = x^{-\alpha - 1} / \Gamma(\alpha) \times exp{(-1/x)} \f$ is the standardized form. 
  To shift and/or scale the distribution use the \f$ loc \f$ and \f$ scale \f$ parameters. The density becomes \f$ \frac{f(y)}{scale} \f$ where \f$ y=(x-loc)/scale \f$ .
  
 
 \par Cumulative Distribution Function (CDF):
 
 \li \f$ F(x) = \frac{\gamma(\alpha,scale/(x-loc))}{\Gamma(\alpha)}\f$ where the numerator is the upper incomplete gamma function.
 
 \par Inverse CDF:
 There is no analytical representation of the Inverse CDF of the Invgamma distribution. 
 The value \f$ F^{-1}(p) \f$ is estimated by dichotomy from the CDF.

*/
#ifndef TINVGAMMADISTRIBUTION_H
#define TINVGAMMADISTRIBUTION_H

//Uranie
#include "TStochAttribut.h"

namespace URANIE
{
namespace DataServer
{
class TInvGammaDistribution: public TInfiniteDistribution
{
    // Attributes
private:
  Double_t _dparameterAlpha;      ///< the shape parameter
  Double_t _dparameterScale;      ///< the scale parameter
  Double_t _dparameterLocation;   ///< the location parameter

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    /*!
     Default shape is 1.0, default scale and location are 1.0 and 0.0 respectively. 
     */
    TInvGammaDistribution();

    //! Constructor setting the name
    /*!
      Default shape is 1.0, default scale and location are 1.0 and 0.0 respectively.

     \param name (TString) Name of the attribute
     */
    TInvGammaDistribution(TString name);

   

    //! Constructor setting name, scale and location
    /*!                 
     \param name (TString) Name of the attribute
     \param alpha (Double_t) the shape parameter \f$ \alpha \f$
     \param scale (Double_t) the scale parameter (Default = 0.0)
     \param location (Double_t) the location parameter (Default = 0.0)

     */
  TInvGammaDistribution(TString name, Double_t alpha, Double_t scale = 1.0, Double_t location = 0.0);

    //! Construction of a TInvGammaDistribution from an existing TAttribute object
    /*!
     \warning this copy constructor is incomplete. Please do not use.
     (Nicolas Gilardi, 2010.06.11)

     \param att (TAttribute *) Pointer to an existing TAttribute object
     \param alpha (Double_t) the shape parameter \f$ \alpha \f$
     \param scale (Double_t) the scale parameter (Default = 0.0)
     \param location (Double_t) the location parameter (Default = 0.0)             
     */
  TInvGammaDistribution(URANIE::DataServer::TAttribute *att, Double_t alpha, Double_t scale = 1.0, Double_t location = 0.0);

    //! Default destructor
    virtual ~TInvGammaDistribution();

    //@}
    //@{
    /** @name Getters and Setters parameters
     The inverse-gamma distribution has three parameters: the shape \f$ \alpha \f$, a scale parameter (default equal to 1) and a location parameter (default equal to 0).
     
    **/
    
    //! Return the shape parameter
    Double_t getParameterAlpha()
    {
      return _dparameterAlpha;
    }

    
    //! Return the scale parameter
    Double_t getParameterScale()
    {
      return _dparameterScale;
    }
    
    //! Return the location parameter
    Double_t getParameterLocation()
    {
      return _dparameterLocation;
    }

    //! Set all the parameters
    /*!
     \param lambda (Double_t) the rate parameter \f$ \lambda \f$
     \param mode (Double_t) the mode parameter \f$ x_{min} \f$ (Default = 0.0)                
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)
     */
    virtual void setParameters(Double_t alpha, Double_t scale = 1.0, Double_t location=0,
            Bool_t recompute = kTRUE);
    
    //! Set alpha value
    /*!
      \param  alpha (double) : value of alpha to be used
      \param  recompute (bool, default=true) : recompute the theoretical properties
    */
    void setParameterAlpha(Double_t alpha, Bool_t recompute = kTRUE);
    //! Set scale value
    /*!
      \param  scale (double) : value of scale to be used
      \param  recompute (bool, default=true) : recompute the theoretical properties
    */
    void setParameterScale(Double_t scale, Bool_t recompute = kTRUE);
    //! Set location value
    /*!
      Change the minimum used (default being 0)
      \param  location (double) : value of location to be used
      \param  recompute (bool, default=true) : recompute the theoretical properties
    */
    void setParameterLocation(Double_t location, Bool_t recompute = kTRUE);
    //@}

    /** @name Distribution specific methods
     PDF, CDF, etc.
     **/
    //@{
    //! Computes the theoretical values of various statistical measures.
    virtual void computeTheoreticalInformation();

    //! Returns the probability density for the value x
    /*!                
     \param x (Double_t) An acceptable value for the attribute                 
     */
    virtual Double_t getPDF(Double_t x);

    //! Returns the cumulative distribution for the value x
    /*!
     \param x (Double_t) An acceptable value for the attribute                
     */
    virtual Double_t getCDF(Double_t x);

    //! Returns the value corresponding to the cumulative density p
    /*!                 
     \param p (Double_t) A valid probability
     
     \exception UErrorExceptions if p is outside of the interval [0,1]                
     */
    virtual Double_t getInvCDF(Double_t p);
    //@}
    
    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}


    ClassDef(URANIE::DataServer::TInvGammaDistribution, ID_DATASERVER)
    //Classe de
};

} // Fin du namespace DataServer
} // Fin du namespace URANIE
#endif
