//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TNormalDistribution.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Thu Jan 18 17:08:58 CET 2007

 \brief Interface of the class URANIE::DataServer::TNormalDistribution

 */
/**

 * @class URANIE::DataServer::TNormalDistribution
 *
 * @brief Description of the class TNormalDistribution.

 This class defines a stochastic attribute following an Normal (or Gaussian) distribution.

 This distribution has two parameters: the mean parameter \f$ \mu \f$ and the 
 standard deviation \f$ \sigma > 0 \f$.

 \par Probability Density Function (PDF):
 \li \f$ f(x) =  \exp\left( -( x - \mu)^2 / 2\sigma^2 \right) / \sqrt{2\pi\sigma^2}\f$

 \par Cumulative Distribution Function (CDF):
 \li \f$ F(x) = 0.5 * \left( 1 + erf\left( (x-\mu)/\sqrt{2\sigma^2}\right) \right)\f$

 \par Inverse CDF:
 \li \f$ F^{-1}(p) =  \mu + erf^{-1}\left( 2p-1 \right) * \sqrt{2\sigma^2}\f$ for \f$ p \in [0,1] \f$


 \note In order to reduce numerical errors, the algorithm used to compute \f$ F^{-1}(p) \f$ 
 is actually either Peter John Acklam's algorithm (http://home.online.no/~pjacklam/notes/invnorm) 
 or the ROOT implementation of AS241 algorithm (Michael J. Wichura, Applied Statistics, vol. 37, 1988, pp. 477-484). 
 The latter is used by default.


 */
#ifndef TNORMALDISTRIBUTION_H
#define TNORMALDISTRIBUTION_H

//Uranie
#include "TStochAttribut.h"

namespace URANIE
{
namespace DataServer
{
class TNormalDistribution: public TInfiniteDistribution
{
private:
    Double_t _dparameterMu;    ///< Mean parameter \f$ \mu \f$
    Double_t _dparameterSigma; ///< Standard deviation parameter \f$ \sigma \f$

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    /*!
     Default mean is 0.0, default standard deviation is 1.0 
     */
    TNormalDistribution();

    //! Constructor setting the name
    /*!
     Default mean is 0.0, default standard deviation is 1.0 

     \param name (TString) Name of the attribute
     */
    TNormalDistribution(TString name);

    //! Constructor setting name, scale and location
    /*!                 
     \param name (TString) Name of the attribute
     \param mu (Double_t) Mean parameter \f$ \mu \f$
     \param sigma (Double_t) Standard deviation parameter \f$ \sigma \f$
     */
    TNormalDistribution(TString name, Double_t mu, Double_t sigma);

    //! Construction of a TNormalDistribution from an existing TAttribute object
    /*!
     \warning this copy constructor is incomplete. Please do not use.
     (Nicolas Gilardi, 2010.06.11)

     \param att (TAttribute *) Pointer to an existing TAttribute object
     \param mu (Double_t) Mean parameter \f$ \mu \f$
     \param sigma (Double_t) Standard deviation parameter \f$ \sigma \f$
     */
    TNormalDistribution(URANIE::DataServer::TAttribute *att, Double_t mu,
            Double_t sigma);

    //! Default destructor
    virtual ~TNormalDistribution();
    //@}

    /** @name Law parameters
     This law has two parameters: the mean \f$ \mu \f$ and the standard deviation 
     \f$ \sigma \f$. Both are defined here. 
     **/
    //@{            
    //! Set the mean parameter
    /*!
     \param mu (Double_t) Mean parameter \f$ \mu \f$
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)                            
     */
    virtual void setParameterMu(Double_t mu, Bool_t recompute = kTRUE);

    //! Return the rate parameter
    Double_t getParameterMu()
    {
        return _dparameterMu;
    }

    //! Set the standard deviation parameter
    /*!
     \param sigma (Double_t) Standard deviation parameter \f$ \sigma \f$
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)

     \exception UErrorExceptions if sigma <  1e-6 * mu                
     */
    virtual void setParameterSigma(Double_t sigma, Bool_t recompute = kTRUE);

    //! Return the mode parameter
    Double_t getParameterSigma()
    {
        return _dparameterSigma;
    }

    //! Set all the parameters
    /*!
     \param mu (Double_t) Mean parameter \f$ \mu \f$
     \param sigma (Double_t) Standard deviation parameter \f$ \sigma \f$                
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)
     */
    virtual void setParameters(Double_t mu, Double_t sigma, Bool_t recompute =
            kTRUE);
    //@}

    /** @name Distribution specific methods
     PDF, CDF, etc.
     **/
    //@{
    //! Computes the theoretical values of various statistical measures.
    virtual void computeTheoreticalInformation();

    //! Returns the probability density for the value x
    /*!                
     \param x (Double_t) An acceptable value for the attribute                 
     */
    virtual Double_t getPDF(Double_t x);

    //! Returns the cumulative distribution for the value x
    /*!
     \param x (Double_t) An acceptable value for the attribute                
     */
    virtual Double_t getCDF(Double_t x);

    //! Returns the value corresponding to the cumulative density p
    /*!                 
     The ROOT implementation of AS241 algorithm is used by default.
     
     \param p (Double_t) A valid probability
     
     \exception UErrorExceptions if p is outside of the interval [0,1]                
     */
    virtual Double_t getInvCDF(Double_t p);
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::DataServer::TNormalDistribution, ID_DATASERVER)
    //Classe de
};

} // Fin du namespace DataServer
} // Fin du namespace URANIE
#endif
