//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TPCA.cxx
 \b Created : 2012-09-10 07:41:46
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \brief Interface file for the class URANIE::DataServer::TPCA
 
 Execute the Principal Component Analysis from a TDS.
 
 */
//ROOT
#include "TMath.h"
#include "TEllipse.h"
#include "TLine.h"
#include "TMatrixTSym.h"
#include "TMatrixDSymEigen.h"
#include "TPad.h"
#include "TH2.h"
#include "TLatex.h"

// Uranie
#include "TPCA.h"
#include "UExceptions.h"

ClassImp (URANIE::DataServer::TPCA)

namespace URANIE
{
namespace DataServer
{
TPCA::TPCA(URANIE::DataServer::TDataServer *tds, const char *varexp,
        Option_t * option) :
        _blog(kFALSE), _tds(tds), _ntdResult(NULL), _ntdVarResult(NULL), _ntdLoading(NULL),
        _vecEigenValues(1), _vecSumPctEigenValues(1), _matEigenVectors(1, 1)
{
    TString soption = option;
    soption.ToLower();
    try
    {
        if (_tds == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TPCA::constructor ERROR : the pointer of \"tds\" is empty");

        //  Get the list of Input attributes
        _listOfInputAttributes = _tds->getListOfAttributes(varexp);
        Int_t natt = _listOfInputAttributes->GetEntries();

        if (_blog)
        {
            cout << " ** natt[" << natt << "]" << endl;
            _listOfInputAttributes->Print();
        }

        _matCorr.ResizeTo(natt, natt);
        _matCorr.Zero();

        // compute statistic
        _tds->computeStatistic(varexp);

        // Compute the matrix of correlation
        _matCorr = _tds->computeCorrelationMatrix(varexp);

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

TPCA::~TPCA()
{
    cout << " ** Destroy TPCA" << endl;

    if (_listOfInputAttributes)
        delete _listOfInputAttributes;
    _listOfInputAttributes = NULL;

    if( gROOT->GetList() != 0 &&  !gROOT->GetList()->IsEmpty())
    {
        if(_ntdResult != NULL)
            delete _ntdResult;
        _ntdResult=NULL;

        if(_ntdVarResult != NULL)
            delete _ntdVarResult;
        _ntdVarResult=NULL;

        if(_ntdLoading != NULL)
            delete _ntdLoading;
        _ntdLoading=NULL;
    }
}

// -----------------------------------------------------------------------
//  Computation
// -----------------------------------------------------------------------
void TPCA::compute(Option_t* option)
{
    if (_blog)
    {
        cout << " ** Compute PCA[" << option << "]" << endl;
        _matCorr.Print();
    }

    TMatrixDSym sym;
    sym.Use(_matCorr.GetNrows(), _matCorr.GetMatrixArray());
    TMatrixDSymEigen eigen(sym);

    // Eigen values
    _vecEigenValues.ResizeTo(_matCorr.GetNrows());
    _vecEigenValues = eigen.GetEigenValues();

    // Eigen vectors
    _matEigenVectors.ResizeTo(_matCorr.GetNrows(), _matCorr.GetNrows());
    _matEigenVectors = eigen.GetEigenVectors();

    Double_t dsum = 0.0;
    try
    {
        for (Int_t i = 0; i < _listOfInputAttributes->GetEntries(); i++)
        {
            if (_vecEigenValues[i] < 0.0)
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        Form(
                                "TPCA::compute method\n The %dth eigenvalue is negative[%e]",
                                i, _vecEigenValues[i]));
            dsum += _vecEigenValues[i];
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    _ntdResult = new TNtupleD(Form("PCA_%s", _tds->GetName()),
            Form("PCA for the %s TDS", _tds->GetName()),
            "i:eigen:eigen_pct:sum_eigen_pct");
    _ntdResult->SetMarkerColor(kRed);
    _ntdResult->SetMarkerSize(0.7);
    _ntdResult->SetMarkerStyle(8);

    _vecSumPctEigenValues.ResizeTo(_matCorr.GetNrows());
    Double_t dsumcrt = 0.0;
    for (Int_t i = 0; i < _listOfInputAttributes->GetEntries(); i++)
    {
        dsumcrt += _vecEigenValues[i];
        _vecSumPctEigenValues[i] = 100.0 * dsumcrt / dsum;
        if (_blog)
            cout << " EigenValue i[" << i << "] fEigenValues["
                    << _vecEigenValues[i] << "] Pct["
                    << 100.0 * _vecEigenValues[i] / dsum << "] SumPct["
                    << _vecSumPctEigenValues[i] << "] Sum[" << dsum << "]"
                    << endl;
        _ntdResult->Fill(1.0 + i, _vecEigenValues[i],
                100.0 * _vecEigenValues[i] / dsum, _vecSumPctEigenValues[i]);
    }

    // Fill the Principal Components
    Int_t npat = _tds->getNPatterns();
    Int_t natt = _listOfInputAttributes->GetEntries();
    Double_t dmean = 0., dsigma = 0.;
    TIter next(_listOfInputAttributes);
    Int_t i=0, j = 0;

    // ======================== Initialising stuff
    TMatrixD orig(npat,natt);
    TMatrixD projvar(natt,natt);
    vector<double> totcosevt(npat,0);
    vector<double> totcosvar(natt,0);
    vector<double> totcontr(natt,0);

    TString BranchesName="";TString BranchesType="";
    vector<string> names;
    
    ///for (Int_t i = 0; i < natt; i++)
    while (TAttribute *att = (TAttribute *) next())        
    {
        dmean = _tds->getAttribute(att->GetName())->getMean();
        dsigma = _tds->getAttribute(att->GetName())->getStd();
        if(i!=0)
        {
            BranchesName+=":";
            BranchesType+=":";
        }
        BranchesName += Form("PC_%d:cosca_%d:contr_%d",i+1,i+1,i+1);
        BranchesType += "D:D:D";
        names.push_back( string(Form("PC_%d",i+1)) );
        names.push_back( string(Form("cosca_%d",i+1)) );
        names.push_back( string(Form("contr_%d",i+1)) );
        for (Int_t ipat = 0; ipat < npat; ipat++)
        {
            orig(ipat,i)=(_tds->getValue(att->GetName(), ipat) - dmean )/ dsigma;
            totcosevt[ipat]+=orig(ipat,i)*orig(ipat,i);
        }
        for(int jcp=0; jcp<natt; jcp++)
        {
            projvar(i,jcp)=_matEigenVectors(i,jcp)*_matEigenVectors(i,jcp)*_vecEigenValues(jcp);
            // tot.at(i)+=_matEigenVectors(i,jcp)*_matEigenVectors(i,jcp)*_vecEigenValues(jcp);
        }
        i++;
    }

    // New coordinate of the patterns
    TMatrixD Prod=orig*_matEigenVectors;

    for(unsigned int icontr=0; icontr<totcontr.size(); icontr++)
        for (Int_t ipat = 0; ipat < npat; ipat++)
        {
            //cout<<orig(ipat,icontr)<<" au carre divise par "<<npat<<" fois "<<_vecEigenValues[i]<<" = "<<orig(ipat,i)*orig(ipat,i) / (npat*_vecEigenValues[i])<<endl;
            totcontr[icontr]+=Prod(ipat,icontr)*Prod(ipat,icontr) / (npat*_vecEigenValues[icontr]);
        }
    
    TDSNtupleD *Results = new TDSNtupleD(Form("PCA_proj_%s", _tds->GetName()),
                                         Form("PCA %s TDS", _tds->GetName()), BranchesName.Data());                                         

    vector<double> dpca (3*natt,0);
    for (Int_t ipat = 0; ipat < npat; ipat++)
    {
        for(i=0; i<natt; i++)
        {
            dpca[3*i+0]=Prod(ipat,i);
            dpca[3*i+1]=dpca[3*i]*dpca[3*i]/totcosevt[ipat];
            dpca[3*i+2]=dpca[3*i]*dpca[3*i]/(npat*_vecEigenValues[i]) * 1./totcontr[i];
            
        }
        Results->Fill(&dpca[0]);
    }

    for(unsigned int ibr=0; ibr<names.size(); ibr++)
        _tds->addAttribute( names.at(ibr).c_str() );
    _tds->getTuple()->Merger(Results,BranchesName.Data(),BranchesName.Data());

    // Dealing with variables now

    _ntdVarResult = new TDSNtupleD(Form("PCA_variable_%s", _tds->GetName()), Form("PCA variable %s TDS", _tds->GetName()),
                                   ("Variable:"+BranchesName).Data(),("S:"+BranchesType).Data() );
    
    stringstream sstr;
    int icp=0;
    TIter next2(_listOfInputAttributes);
    while (TAttribute *att = (TAttribute *) next2())
    {
        sstr.str("");
        sstr << " " << att->GetName() << " ";
        for( int jcp=0; jcp<natt; jcp++)
        {
            sstr << _matEigenVectors(icp,jcp) << " " << projvar(icp,jcp) << " " <<  _matEigenVectors(icp,jcp)*_matEigenVectors(icp,jcp) << " " ;
        }
        icp++;
        _ntdVarResult->Fill( sstr.str() );
    }
    
    //clean
    delete Results;
}

// -----------------------------------------------------------------------
//  drawPCA
// -----------------------------------------------------------------------
void TPCA::drawPCA(Int_t indX, Int_t indY, Option_t* option)
{
    if (_blog)
        cout << " ** draw PCA (" << indX << ", " << indY << ") Option["
                << option << "]" << endl;

    Int_t natt = _listOfInputAttributes->GetEntries();
    try
    {
        if (indX > natt)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TPCA::drawPCA method\n The first index \"indx\" %d is greater than the number ofattributes (%d)",
                            indX, natt));
        if (indY > natt)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TPCA::drawPCA method\n The second index \"indY\" %d is greater than the number ofattributes (%d)",
                            indY, natt));
        if (_ntdResult == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TPCA::drawPCA method\n The \"TPCA::compute()\" method must be call before the \"TPCA::drawPCA()\" method");

        // Draw the Graph
        _tds->draw(Form("PC_%d:PC_%d", indY, indX));

        // Label of the Axis with Percentage of variance explained by the component
        Double_t dPctX = _vecSumPctEigenValues[indX-1];
        if (indX > 1)
            dPctX -= _vecSumPctEigenValues[indX - 2];
        Double_t dPctY = _vecSumPctEigenValues[indY-1];
        if (indY > 1)
            dPctY -= _vecSumPctEigenValues[indY - 2];

        if (_blog)
            cout << " ** PCT X(" << dPctX << ") -- Y(" << dPctY << ")" << endl;
        TH2F *htemp = (TH2F*) gPad->GetPrimitive("__tdshisto__0");
        htemp->SetXTitle(Form("Component %d (%5.2f %%)", indX, dPctX));
        htemp->SetYTitle(Form("Component %d (%5.2f %%)", indY, dPctY));
        htemp->GetYaxis()->SetTitleOffset(1.3);
        htemp->SetTitle("");

        
        TString soption = option;
        if(soption != "")
        {            

            if( _tds->isAttribute(soption) )
            {
            
                _tds->getTuple()->Draw(Form("PC_%d:PC_%d", indY, indX),"","goff");
        
                Double_t* valY = _tds->getTuple()->GetV1();
                Double_t* valX = _tds->getTuple()->GetV2();

                int nS=_tds->getNPatterns();
            
                TGraph *gr = new TGraph(nS);
                TLatex *tl = 0;
                TString sname = "";
                for (Int_t i = 0; i < nS; i++)
                {
                    sname = _tds->getString(soption,i);
                    if (_blog)
                        cout << " ** sname[" << sname << "]   x(" << valX[i] << ") Y("
                             << valY[i] << ")" << endl;
                    tl = new TLatex(valX[i] * 1.01, valY[i] * 1.01, sname);
                    tl->SetTextAngle(0);
                    tl->SetTextSize(0.05);
                    tl->SetTextColor(kBlack);
                    gr->GetListOfFunctions()->Add(tl);
                }
                gr->Draw("same");

                gPad->Modified();
                gPad->Update();
            }
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

// -----------------------------------------------------------------------
//  drawLoading
// -----------------------------------------------------------------------
void TPCA::drawLoading(Int_t indX, Int_t indY, Option_t* option)
{
    if (_blog)
        cout << " ** drawLoading PCA  (" << indX << ", " << indY
                << ") - Option[" << option << "]" << endl;

    Int_t natt = _listOfInputAttributes->GetEntries();

    try
    {
        if (indX > natt)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TPCA::drawLoading method\n The first index \"indx\" %d is greater than the number ofattributes (%d)",
                            indX, natt));
        if (indY > natt)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TPCA::drawLoading method\n The second index \"indY\" %d is greater than the number ofattributes (%d)",
                            indY, natt));
        if (_ntdResult == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TPCA::drawLoading method\n The \"TPCA::compute()\" method must be call before the \"TPCA::drawLoading()\" method");

        // Create the TNtupleD of Loading
        if (!_ntdLoading)
        {

            if (_blog)
                cout << " create tjhe TNtupleD of Loading natt[" << natt << "]"
                        << endl;
            TString sLabel = "i";

            for (Int_t i = 0; i < natt; i++)
                sLabel += Form(":PCAL_%d", i + 1);
            if (_blog)
                cout << " ** sLabel[" << sLabel << "]" << endl;

            _ntdLoading = new TNtupleD(Form("Loading_%s", _tds->GetName()),
                    Form("Loading on PCA for the %s TDS", _tds->GetName()),
                    sLabel);
            _ntdLoading->SetMarkerColor(kRed);
            _ntdLoading->SetMarkerSize(0.7);
            _ntdLoading->SetMarkerStyle(8);

            Double_t *dloading = new Double_t[natt + 1];
            for (Int_t i = 0; i < natt; i++)
            {
                dloading[0] = i + 1;

                for (Int_t j = 0; j < natt; j++)
                {
                    dloading[j + 1] = _matEigenVectors[i][j]
                            * TMath::Sqrt(_vecEigenValues[j]); // / (Double_t)(nX) );
                }
                _ntdLoading->Fill(dloading);

            }
            //      _ntdLoading->Scan("*");

        }

        // Draw the Loading
        TString sVarDraw = Form("PCAL_%d:PCAL_%d", indY, indX);
        if (_blog)
            cout << " ** sVarDraw[" << sVarDraw << "]" << endl;

        if(gPad != NULL)
            gPad->Range(0,0,1,1);

        Double_t dPctX = _vecSumPctEigenValues[indX-1];
        if (indX > 1)
            dPctX -= _vecSumPctEigenValues[indX - 2];
        Double_t dPctY = _vecSumPctEigenValues[indY-1];
        if (indY > 1)
            dPctY -= _vecSumPctEigenValues[indY - 2];

        if (_blog)
            cout << " ** PCT X(" << dPctX << ") -- Y(" << dPctY << ")" << endl;

        // Axis nicely done
        TH2F *htemp = new TH2F("AxisTemp",Form(";Component %d (%5.2f %%);Component %d (%5.2f %%)", indX, dPctX, indY, dPctY),100,-1.2,1.2,100,-1.2,1.2);
        htemp->SetStats(false);
        htemp->GetYaxis()->SetTitleOffset(1.3);
        htemp->SetTitle("");
        htemp->Draw();

        // Unitary Circle
        TEllipse *el1 = new TEllipse(0.0,0.0,1);
        el1->SetLineColor(kGray);
        el1->Draw();

        // Diameters of the circle
        TLine *l1 = new TLine(-1.0,0.0,1.0,0.0);
        l1->SetLineColor(kGray);
        l1->SetLineStyle(4);
        l1->Draw();
        l1->DrawLine(0.0,-1.0,0.0,1.0);
        
        // Loading
        _ntdLoading->Draw(sVarDraw,"","same");
        
        Double_t* valY = _ntdLoading->GetV1();
        Double_t* valX = _ntdLoading->GetV2();

        TGraph *gr = new TGraph(natt);
        TLatex *tl = 0;
        TString sname = "";
        for (Int_t i = 0; i < natt; i++)
        {
            sname = _listOfInputAttributes->At(i)->GetName();
            if (_blog)
                cout << " ** sname[" << sname << "]   x(" << valX[i] << ") Y("
                        << valY[i] << ")" << endl;
            tl = new TLatex(valX[i] * 1.01, valY[i] * 1.01, sname);
            tl->SetTextAngle(0);
            tl->SetTextSize(0.05);
            tl->SetTextColor(kBlack);
            gr->GetListOfFunctions()->Add(tl);
        }
        gr->Draw("same");

        gPad->Modified();
        gPad->Update();

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

}

// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------
void TPCA::printLog(Option_t *option)
{
    cout << " ** TPCA::printLog[" << option << "]" << endl;
    cout << " ** Name[" << GetName() << "] Title[" << GetTitle() << "]" << endl;
    cout << " ** TDS Name[" << _tds->GetName() << "] Title[" << _tds->GetTitle()
            << "]" << endl;
    cout << " ** List of Attributes size["
            << _listOfInputAttributes->GetEntries() << "]" << endl;
    _listOfInputAttributes->Print();
    cout << " ** Correlation Matrix" << endl;
    _matCorr.Print();

}

} // End Of namespace DataServer
} // End Of namespace URANIE

