//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TStochAttribut.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Fri Oct  1 15:56:37 2004

 $Revision$
 $Date$

 \brief  Class of attributes following a probability density function


 */
#ifndef TSTOCHASTICATTRIBUTE_H
#define TSTOCHASTICATTRIBUTE_H

#include "TString.h"
#include <iostream>

// Uranie
#include "DataServer.h"
#include "TAttribute.h"
#include "UExceptions.h"

namespace URANIE
{
namespace DataServer
{

/*! \class URANIE::DataServer::TStochasticAttribute
 This is the parent class for all attributes which values can be generated by a 
 TSampler. Child objects are random variables following a specific law.
 
 \warning This class contains the following pure abstract methods
 \li \ref computeTheoreticalInformation
 \li \ref getPDF
 \li \ref getCDF
 \li \ref getInvCDF
 
 Each must be redefined in the daughter classes.
 
 */
class TStochasticAttribute: public TAttribute
{
public:

    //! The list of laws that a TStochasticAttribute can follow
    enum ELawType
    {
        kUniform,
        kLogUniform,
        kNormal,
        kStudent,
        kLogNormal,
        kTrapezium,
        kTriangular,
        kLogTriangular,
        kExponential,
        kBeta,
	kGamma,
	kInvGamma,
        kCauchy,
        kUniformByParts,
        kWeibull,
        kGumbel,
        kMultinomial,
        kGenPareto,
        kGeneralizedNormal,
        kGeneralizedNormalV2,
        kUnknown
    };

    //! The list of orthogonal polynomial families that can represent a given TStochasticAttribute
    enum EStochasticBasis
    {
        kUnknownBasis, kHermite, kLegendre, kLaguerre, kJacobi
    };         ///////// *Hermite, Legendre, Laguerre, Jacobi

protected:

    ELawType _lawType;             ///< The law type ELawType
    Double_t _dtheoreticalMean;    ///< The theoretical mean
    Double_t _dtheoreticalStdDev;  ///< The theoretical standard deviation
    Double_t _dtheoreticalMode;    ///< The theoretical mode
    Double_t _pmin;           ///< The min value of cumulative density function 
    Double_t _pmax;            ///< The max value of cumulative density function
    EStochasticBasis _nStochasticBasis; ///< The stochastic representation of an orthogonal basis

public:
    // -----------------------------------------------------------------------
    //  Constructors
    // -----------------------------------------------------------------------
    //! Default Constructor
    TStochasticAttribute();

    //! Constructor setting the name 
    /*!
     \param str (TString)  Name of the attribute
     */
    TStochasticAttribute(TString str);

    //! Constructor setting the name and the law type 
    /*!
     \param str (TString)    Name of the attribute
     \param value (ELawType) The type of the probability law of the attribute. 
     The available types are defined in the ELawType 
     enumerator
     */
    TStochasticAttribute(TString str, ELawType value);

    //! Constructor setting the name, the law type, and the range (lower and upper bounds)
    /*!
     \param str (TString)    Name of the attribute
     \param value (ELawType) The type of the probability law of the attribute. 
     The available types are defined in the ELawType 
     enumerator
     \param min (Double_t)   Lower bound of the attribute's range
     \param max (Double_t)   Upper bound of the attribute's range
     */
    TStochasticAttribute(TString str, ELawType value, Double_t min,
            Double_t max);

    //! Construction of a TStochasticAttribute from an existing TAttribute object.
    /*!
     \warning this copy constructor is incomplete. Please do not use.
     (Nicolas Gilardi, 2010.05.28)

     \param object (TAttribute *) Pointer to the TAttribute object to copy;
     \param value  (ELawType)     The type of the probability law of the attribute. 
     The available types are defined in the ELawType 
     enum, e.g.: kUniform, kNormal, kBeta...     
     */
    TStochasticAttribute(TAttribute *object, ELawType value);

    //! Default Destructor
    virtual ~TStochasticAttribute();

    // -----------------------------------------------------------------------
    //  Getter and setter methods
    // -----------------------------------------------------------------------

    /** @name Statistical methods
     Statistical methods ( pdf, cdf, theoretical information...).
     **/
    //@{
    //! Computes the theoretical values of various statistical measures, like mean, standard deviation, mode, etc.
    /*!
     \warning if the law is truncated, the theoretical values will 
     not be the one from the "complete" law, but the one 
     from the truncated law.
     As a consequence, this function must be called each 
     time the lower and upper bounds of the attribute are 
     reset.
     */
    virtual void computeTheoreticalInformation() = 0;

    //! Returns the probability density for the value x
    /*!
     \param x (Double_t) An acceptable value for the attribute 
     
     \warning if x is outside of the [lowerBound, upperBound] 
     interval, getPDF(x) must return 0.0
     */
    virtual Double_t getPDF(Double_t x) = 0;
    double getPDFforTF1(double *x) { return getPDF(x[0]); }
    
    //! Returns the cumulative distribution for the value x
    /*!
     \param x (Double_t) An acceptable value for the attribute
     
     \warning the behaviour of this function with respect to 
     boundaries is:
     \li if x <= lowerBound: getCDF(x) == 0.0
     \li if x in [lowerBound, upperBound]: getCDF(x) in [0.0, 1.0]
     \li if x >= upperBound: getCDF(x) == 1.0
     */
    virtual Double_t getCDF(Double_t x) = 0;
    double getCDFforTF1(double *x) { return getCDF(x[0]); }
    
    //! Returns the value corresponding to the cumulative density p
    /*!
     This function is used by the Sampler to generate acceptable values for 
     the attribute

     It can also be used to retrieve the theoretical quantiles 
     of a random variable. For instance, the theoretical value of the 
     median is equal to getInvCDF(0.5).                

     \param p (Double_t) A valid probability
     
     \exception UErrorExceptions if p is outside of the interval [0,1]
     
     \warning This function must always return a value between 
     lowerBound and upperBound. 
     */
    virtual Double_t getInvCDF(Double_t p) = 0;
    double getInvCDFforTF1(double *x) { return getInvCDF(x[0]); }
    
    //! Returns the theoretical value of the mean
    virtual Double_t getTheoreticalMean()
    {
        return _dtheoreticalMean;
    }

    //! Returns the theoretical value of the standard deviation
    virtual Double_t getTheoreticalStdDev()
    {
        return _dtheoreticalStdDev;
    }

    //! Returns the theoretical value of the mode
    virtual Double_t getTheoreticalMode()
    {
        return _dtheoreticalMode;
    }
    //@}

    //---------------------------------------------
    /** @name Statistical Representation
     **/
    //@{
    //! Sets the law type from an integer value
    /*!
     \param value (ELawType) The type of the probability law of the attribute. 
     The available types are defined in the ELawType 
     enumerator.
     */
    void setLawType(ELawType value)
    {
        _lawType = value;
    }

    //! Sets the law type from a law name
    /*!
     The following choices are available for the law name: 
     \li uniform
     \li logUniform
     \li normal
     \li logNormal
     \li triangular
     \li logTriangular
     \li exponential
     \li beta
     \li cauchy
     \li uniformByParts
     \li weibull
     \li gumbel
     \li multinomial
     \li unknown
     
     \param name (TString) Name of the law. It is case insensitive.                 
     */
    void setLawType(TString name);

    //! Returns the law type as an integer
    //const int getLawType() const { return _lawType;}
    int getLawType()
    {
        return _lawType;
    }

    //! Defines an orthogonal polynomial family to represent the stochastic attribute
    /*!                                
     \param n (EStochasticBasis) The orthogonal polynomial family. It is 
     chosen from EStochasticBasis enumerator. 
     
     Example:
     \code
     TTriangularDistribution *ttri = new TTriangularDistribution("triangle", -1.0, 1.0, 0.0);
     tunif->setStochasticBasis(TStochasticAttribute::kJacobi);                
     tunif->printLog();
     \endcode
     */
    void setStochasticBasis(EStochasticBasis n)
    {
        _nStochasticBasis = n;
    }

    //! Sets the default orthogonal polynomial family representing the stochastic attribute
    /*!
     The implicit rule is:
     \li (Log)Uniform law : \b  Legendre
     \li (Log)Normal law  : \b  Hermite
     \li Exponential law  : \b  Laguerre
     \li Beta law (2 parameters) : \b Jacobi
     
     The default for other laws is undefined (kUnknownBasis)
     */
    void setDefaultStochasticBasis();

    //! Returns the orthogonal polynomial which is used to represent the attribute.
    /*!                                
     Example:
     \code
     TUniformDistribution *tunif = new TUniformDistribution("unif", -1.0, 2.0);
     cout << " ** Stochastic polynomial [" << tunif->getStochasticBasis() << "]" <<  endl;
     \endcode
     
     More precisely, for a pointer tatt on an object of the TStochasticAttribute family:
     \code
     switch( tatt->getStochasticBasis() ) {
     case TStochasticAttribute::kHermite: {
     cout << " *** case Hermite" << endl;
     break;
     }
     case TStochasticAttribute::kLegendre: {
     cout << " *** case Legendre" << endl;
     break;
     }
     case TStochasticAttribute::kUnknownBasis: {
     cout << " *** Warning: unknown basis ***" << endl;
     }
     }
     \endcode                
     */
    EStochasticBasis getStochasticBasis() const
    {
        return _nStochasticBasis;
    }
    //@}

    //---------------------------------------------
    /** @name Treat the boundaries of an attribute
     **/
    //@{
    //! Returns the minimum value of the cumulative distribution function
    virtual Double_t getPMin()
    {
        return _pmin;
    }
    //! Returns the maximum value of the cumulative distribution function
    virtual Double_t getPMax()
    {
        return _pmax;
    }
    //@}


    //! Display information specific to this class
    virtual void printLog();

    ClassDef(TStochasticAttribute, ID_DATASERVER)
    //Definition of a stochastic variable
};

/*! \class URANIE::DataServer::TInfiniteDistribution
 This is the parent class for all laws with at least one infinite bound.
 This class takes care of the calculation of new probability limits when 
 an infinite law is truncated.        
 */
class TInfiniteDistribution: public TStochasticAttribute
{

    // Attributes
protected:

public:
    // -----------------------------------------------------------------------
    //  Constructors
    // -----------------------------------------------------------------------
    //! Default Constructor
    TInfiniteDistribution();

    //! Constructor setting the name 
    /*!
     \param str (TString)  Name of the attribute
     */
    TInfiniteDistribution(TString str);

    //! Constructor setting the name and the law type 
    /*!
     \param str (TString)    Name of the attribute
     \param value (ELawType) The type of the probability law of the attribute. 
     The available types are defined in the ELawType 
     enumerator, e.g.: kUniform, kNormal, kBeta...
     */
    TInfiniteDistribution(TString str, ELawType value);

    //! Construction of a TInfiniteDistribution from an existing TAttribute object.
    /*!
     \param object (TAttribute *)  Pointer to the TAttribute object to copy
     \param value  (ELawType)      The type of the probability law of the attribute. 
     The available types are defined in the ELawType 
     enumerator, e.g.: kUniform, kNormal, kBeta...

     \warning this copy constructor is incomplete. Please do not use.
     (Nicolas Gilardi, 2010.05.28)                                          
     */
    TInfiniteDistribution(TAttribute *object, ELawType value);

    //! Default Destructor
    virtual ~TInfiniteDistribution();

    // -----------------------------------------------------------------------
    //  Getter and setter methods
    // -----------------------------------------------------------------------

    /** @name Treat the boundaries of an attribute
     **/
    //@{
    //! Sets the lower bound of the attribute
    /*!
     As we are dealing with infinite distribution, setting a finite 
     lower bound means that the law is truncated. Theoretical 
     information must be recalcutated, as well as the _pmin value.
     
     \param val (Double_t) The value of the new lower bound
     
     \exception UErrorExceptions if the lower bound is superior to the upper bound
     
     \warning when the law is not truncated, its lower bound is supposed 
     to be infinite. For untruncated laws with a finite lower 
     bound, this function must be redefined.
     */
    virtual void setLowerBound(Double_t val);

    //! Sets the upper bound of the attribute
    /*!
     As we are dealing with infinite distribution, setting a finite 
     upper bound means that the law is truncated. Theoretical 
     information must be recalcutated, as well as the _pmax value. 
     
     \param val (Double_t) The value of the new upper bound
     
     \exception UErrorExceptions if the upper bound is inferior to the lower bound

     \warning when the law is not truncated, its upper bound is supposed 
     to be infinite. For untruncated laws with a finite upper 
     bound, this function must be redefined.
     */
    virtual void setUpperBound(Double_t val);

    //! Sets the lower and upper bounds of the attribute
    /*!
     As we are dealing with infinite distribution, setting finite bounds 
     means that the law is truncated. Theoretical information must be 
     recalcutated, as well as the _pmin and _pmax values. 

     \param lower (Double_t) The value of the new lower bound
     \param upper (Double_t) The value of the new upper bound
     
     \exception UErrorExceptions if the upper bound is inferior to the lower bound

     \warning when the law is not truncated, its bounds are supposed 
     to be infinite. For untruncated laws with a finite bound, 
     this function must be redefined.                
     */
    virtual void setBounds(Double_t lower, Double_t upper);
    //@}

    virtual void printLog();ClassDef(TInfiniteDistribution, ID_DATASERVER)
    //Definition of an stochastic variable
};
} // Fin du namespace DataServer
} // Fin du namespace URANIE

#endif
