//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TStudentDistribution.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Thu Jan 18 17:08:58 CET 2007

 \brief Implementation of the class URANIE::DataServer::TStudentDistribution

 */
#include <iostream>
using std::cout;
using std::endl;

#include "TMath.h"
#include "Math/ProbFuncMathCore.h"
#include "Math/QuantFuncMathMore.h"
#include "Math/PdfFuncMathCore.h"
#include "TStudentDistribution.h"

ClassImp (URANIE::DataServer::TStudentDistribution)

namespace URANIE
{
namespace DataServer
{
// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------

TStudentDistribution::TStudentDistribution() :
        TInfiniteDistribution("", TStochasticAttribute::kStudent), _dparameterK(1.0)
{
    //  cout << "construit TStudentDistribution" << endl;
    computeTheoreticalInformation();
}

TStudentDistribution::TStudentDistribution(TString str) :
        TInfiniteDistribution(str, TStochasticAttribute::kStudent), _dparameterK(1.0)
{
    //  cout << "Constructeur URANIE::DataServer::TStudentDistribution[" << this << "]str[" << str<<"]"<<endl;
    computeTheoreticalInformation();
}

TStudentDistribution::TStudentDistribution(TString str, Double_t k) :
        TInfiniteDistribution(str, TStochasticAttribute::kStudent)
{
    //  cout << "Constructeur URANIE::DataServer::TStudentDistribution[" << this << "]str[" << str<<"] Mean[" << mean <<"] Sigma[" << sigma << "]" << endl;
    setParameters(k);
}

TStudentDistribution::TStudentDistribution(URANIE::DataServer::TAttribute *att, Double_t k):
    TInfiniteDistribution(att, TStochasticAttribute::kStudent)
{
    //  cout << "Constructeur URANIE::DataServer::TStudentDistribution att[" << this << "]att->getName()[" << att->getName()<<"] Mean[" << mean <<"] Sigma[" << sigma << "]" << endl;
    setParameters(k);
}

// -----------------------------------------------------------------------
//  Destructors
// -----------------------------------------------------------------------

TStudentDistribution::~TStudentDistribution()
{
    //  cout << "detruit TStudentDistribution" << endl; 

}

// -----------------------------------------------------------------------
//  Parameters
// -----------------------------------------------------------------------        

void TStudentDistribution::setParameterK(Double_t k, Bool_t recompute)
{
    try
    {
        if (k < 1e-6 * _dparameterK)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TStudentDistribution::setParameterK: parameter k [%f] is too close to zero: lower limit is 1e-6 * k [%f]",
                            k, 1e-6 * _dparameterK));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    _dparameterK = k;

    if (recompute)
        computeTheoreticalInformation();
}

void TStudentDistribution::setParameters(Double_t k, Bool_t recompute)
{
    setParameterK(k, kFALSE);

    if (recompute)
        computeTheoreticalInformation();
}

// -----------------------------------------------------------------------
//  Theoretical Information
// -----------------------------------------------------------------------        

void TStudentDistribution::computeTheoreticalInformation()
{
    _dtheoreticalMean = ((_dparameterK>1)? 0 : -99);

    _dtheoreticalStdDev = ((_dparameterK>2)? _dparameterK/(_dparameterK-2) : -99);

    _dtheoreticalMode = 0;
}

// -----------------------------------------------------------------------
//  Distribution
// -----------------------------------------------------------------------        

Double_t TStudentDistribution::getPDF(Double_t x)
{
    Double_t pdf = 0;
    if ((x >= lowerBound) && (x <= upperBound))
    {
        pdf = ROOT::Math::tdistribution_pdf(x,_dparameterK);
        pdf /= _pmax - _pmin;
    }
    return pdf;
}

Double_t TStudentDistribution::getCDF(Double_t x)
{
    Double_t cdf = 0.0;
    if ((x >= lowerBound) && (x <= upperBound))
    {
        cdf = ROOT::Math::tdistribution_cdf(x,_dparameterK);
        // Take into account the case where at least one of the bounds is finite
        //cdf /= _pmax - _pmin;
        cdf = (cdf - _pmin) / (_pmax - _pmin);
    }
    else if (x > upperBound)
        cdf = 1.0;

    return cdf;
}

Double_t TStudentDistribution::getInvCDF(Double_t p)
{
    // Verify that p is a valid probability
    try
    {
        if ((p < 0.0) || (p > 1.0))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TStudentDistribution::getInvCDF: parameter p [%f] is not a valid probability: p must be in the interval [0,1]",
                            p));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }
    // Take into account the case where at least one of the bounds is finite            
    Double_t realP = p * (_pmax - _pmin) + _pmin;

    double  x = ROOT::Math::tdistribution_quantile(realP,_dparameterK);
    return x;
}

// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------

void TStudentDistribution::printLog(Option_t *option)
{
    //  TInfiniteDistribution::printLog();
    TString soption = option;
    soption.ToLower();
    cout << "*******************************" << endl;
    cout << "** TStudentDistribution::printLog  ******" << endl;
    cout << "**   Law[" << _lawType << "] Student" << endl;
    cout << "**   Bounds[" << lowerBound << ", " << upperBound << "]" << endl;
    cout << "**   Mean parameter k[" << _dparameterK << "]" << endl;
    cout << "**   Theoretical Mean[" << _dtheoreticalMean << "]" << endl;
    cout << "**   Theoretical StdDev[" << _dtheoreticalStdDev << "]" << endl;
    cout << "**   Theoretical Mode[" << _dtheoreticalMode << "]" << endl;
    cout << "*******************************" << endl;
}

} // Fin du namespace DataServer
} // Fin du namespace URANIE

