//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TTrapeziumPossibility.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2010-07-07 15:52:43

 \brief  Class of attributes following a probability density function


 */
#ifndef TTRAPEZIUMPOSSIBILITY_H
#define TTRAPEZIUMPOSSIBILITY_H

#include "TString.h"
#include <iostream>

// Uranie
#include "DataServer.h"
#include "TPossibilityAttribute.h"

namespace URANIE
{
namespace DataServer
{

/*! \class URANIE::DataServer::TTrapeziumPossibility
 This is the parent class for all attributes which values can be generated by a 
 possibilty (cf RaFu IRSN). Child objects are possibility attributes following a specific possibility.
 
 Each must be redefined in the daughter classes.
 
 */
class TTrapeziumPossibility: public TPossibilityAttribute
{
protected:
    Double_t _dlowerMode;       ///< The lower mode of the trapezium possibility
    Double_t _dupperMode;       ///< The upper mode of the trapezium possibility
public:
    // -----------------------------------------------------------------------
    //  Constructors
    // -----------------------------------------------------------------------
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Constructor setting the name 
    /*!
     \param str (TString)  Name of the attribute
     */
    TTrapeziumPossibility(TString str);

    //! Constructor setting the name and the law type 
    /*!
     \param str (TString)    Name of the attribute
     \param min (Double_t)   Lower bound of the attribute's range.
     \param max (Double_t)   Upper bound of the attribute's range.
     \param lowMode (Double_t)   Lower mode of the trapezium possibility.
     \param upperMode (Double_t)   Upper mode of the trapezium possibility.  	  
     \param monotonous (EMonotonousRelationship) The relationship of monotonous.
     */
    TTrapeziumPossibility(TString str, Double_t min, Double_t max,
            Double_t lowMode, Double_t upperMode,
            TPossibilityAttribute::EMonotonousRelationship monotonous);

    //! Default Destructor
    virtual ~TTrapeziumPossibility();
    //@}

    //---------------------------------------------
    /** @name The lower and upper modes of the trapezium possibility
     **/
    //@{
    //! Return the lower mode of the trapezium possibility
    Double_t getLowerMode()
    {
        return _dlowerMode;
    }
    //! Set the lower mode of the trapezium possibility
    void setLowerMode(Double_t dval);
    //! Return the  upper mode of the trapezium possibility
    Double_t getUpperMode()
    {
        return _dupperMode;
    }
    //! Set the upper mode of the trapezium possibility
    void setUpperMode(Double_t dval);
    //! Set the lower and uppers modes of the trapezium possibility
    void setModes(Double_t lowVal, Double_t upVal);
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    //! Display information specific to this class
    virtual void printLog();
    //@}

    ClassDef(TTrapeziumPossibility, ID_DATASERVER)
    //Definition of a stochastic variable
};
} // Fin du namespace DataServer
} // Fin du namespace URANIE

#endif
