//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TTriangularDistribution.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Thu Jan 18 17:08:58 CET 2007

 \brief Implementation of the class URANIE::DataServer::TTriangularDistribution

 */
#include <iostream>
using std::cout;
using std::endl;

#include "TMath.h"
#include "TTriangularDistribution.h"

ClassImp (URANIE::DataServer::TTriangularDistribution)

namespace URANIE
{
namespace DataServer
{
// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------

TTriangularDistribution::TTriangularDistribution() :
        TStochasticAttribute("", TStochasticAttribute::kTriangular, 0.0, 1.0)
{
    if (_blog)
        cout << "Constructing URANIE::DataServer::TTriangularDistribution"
                << endl;
    _dparameterMode = 0.5;
    computeTheoreticalInformation();
}

TTriangularDistribution::TTriangularDistribution(TString str) :
        TStochasticAttribute(str, TStochasticAttribute::kTriangular, 0.0, 1.0)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TTriangularDistribution with: Name["
                << str << "]" << endl;
    _dparameterMode = 0.5;
    computeTheoreticalInformation();
}

TTriangularDistribution::TTriangularDistribution(TString str, Double_t min,
        Double_t max) :
        TStochasticAttribute(str, TStochasticAttribute::kTriangular, min, max)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TTriangularDistribution with: Name["
                << str << "] Min[" << min << "] Max[" << max << "]" << endl;
    _dparameterMode = (max + min) / 2.;
    computeTheoreticalInformation();
}

TTriangularDistribution::TTriangularDistribution(TString str, Double_t min,
        Double_t max, Double_t mode) :
        TStochasticAttribute(str, TStochasticAttribute::kTriangular, min, max)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TTriangularDistribution with: Name["
                << str << "] Min[" << min << "] Max[" << max << "] Mode["
                << mode << "]" << endl;
    setParameterMode(mode);
}

TTriangularDistribution::TTriangularDistribution(
        URANIE::DataServer::TAttribute *att, Double_t mode) :
        TStochasticAttribute(att, TStochasticAttribute::kTriangular)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TTriangularDistribution from existing TAttribute ["
                << att->GetName() << "] and Mode[" << mode << "]" << endl;
    setParameterMode(mode);
}

// -----------------------------------------------------------------------
//  Destructors
// -----------------------------------------------------------------------

TTriangularDistribution::~TTriangularDistribution()
{
    if (_blog)
        cout << "Destroying URANIE::DataServer::TTriangularDistribution "
                << GetName() << endl;

}

// -----------------------------------------------------------------------
//  Parameters
// -----------------------------------------------------------------------

void TTriangularDistribution::setParameterMode(Double_t mode)
{
    // Verify that mode is between lowerBound and upperBound
    try
    {
        if ((mode < lowerBound) || (mode > upperBound))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TTriangularDistribution::setParameterMode: parameter mode [%f] is out of bounds [%f,%f]",
                            mode, lowerBound, upperBound));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    _dparameterMode = mode;
    computeTheoreticalInformation();
}

void TTriangularDistribution::setParameters(Double_t lower, Double_t upper,
        Double_t mode)
{
    // Verify that mode is between lower and upper
    try
    {
        if ((mode < lower) || (mode > upper))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TTriangularDistribution::setParameters: the new parameter mode [%f] is out of the new bounds [%f,%f]",
                            mode, lower, upper));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    setBounds(lower, upper);
    _dparameterMode = mode;

    computeTheoreticalInformation();
}

// -----------------------------------------------------------------------
//  Theoretical Information
// -----------------------------------------------------------------------

void TTriangularDistribution::computeTheoreticalInformation()
{
    _dtheoreticalMean = (lowerBound + upperBound + _dparameterMode) / 3.0;

    Double_t tmpvar = lowerBound * lowerBound + upperBound * upperBound
            + _dparameterMode * _dparameterMode;
    tmpvar -= lowerBound * upperBound + lowerBound * _dparameterMode
            + _dparameterMode * upperBound;
    tmpvar /= 18.0;

    _dtheoreticalStdDev = TMath::Sqrt(tmpvar);

    _dtheoreticalMode = _dparameterMode;

    _dleftArea = getCDF(_dparameterMode);
}

// -----------------------------------------------------------------------
//  Distribution
// -----------------------------------------------------------------------        

Double_t TTriangularDistribution::getPDF(Double_t x)
{
    Double_t pdf = 0.;
    if ((x > lowerBound) && (x < upperBound))
    {
        if (x < _dparameterMode)
            pdf = 2.0 * (x - lowerBound)
                    / ((upperBound - lowerBound)
                            * (_dparameterMode - lowerBound));
        else
            pdf = 2.0 * (upperBound - x)
                    / ((upperBound - lowerBound)
                            * (upperBound - _dparameterMode));
    }
    return pdf;
}

Double_t TTriangularDistribution::getCDF(Double_t x)
{
    Double_t cdf;

    if (x < lowerBound)
        cdf = 0.0;
    else if (x > upperBound)
        cdf = 1.0;
    else
    {
        if (x < _dparameterMode)
            cdf = (x - lowerBound) * (x - lowerBound)
                    / ((upperBound - lowerBound)
                            * (_dparameterMode - lowerBound));
        else
            cdf = 1.0
                    - (upperBound - x) * (upperBound - x)
                            / ((upperBound - lowerBound)
                                    * (upperBound - _dparameterMode));
    }
    return cdf;
}

Double_t TTriangularDistribution::getInvCDF(Double_t p)
{
    // Verify that p is a valid probability
    try
    {
        if ((p < 0.0) || (p > 1.0))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TTriangularDistribution::getInvCDF: parameter p [%f] is not a valid probability: p must be in the interval [0,1]",
                            p));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    Double_t x;
    if (p < _dleftArea)
        x = lowerBound
                + TMath::Sqrt(
                        p * (upperBound - lowerBound)
                                * (_dparameterMode - lowerBound));
    else
        x = upperBound
                - TMath::Sqrt(
                        (1.0 - p) * (upperBound - lowerBound)
                                * (upperBound - _dparameterMode));
    return x;
}

// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------      

void TTriangularDistribution::printLog(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    cout << "*******************************" << endl;
    cout << "** TTriangularDistribution::printLog  ******" << endl;
    cout << "**   Law[" << _lawType << "] TTriangularDistribution" << endl;
    cout << "**   Bounds[" << lowerBound << ", " << upperBound << "]" << endl;
    cout << "**   Mode parameter[" << _dparameterMode << "]" << endl;
    cout << "**   Theoretical Mean[" << _dtheoreticalMean << "]" << endl;
    cout << "**   Theoretical StdDev[" << _dtheoreticalStdDev << "]" << endl;
    cout << "**   Theoretical Mode[" << _dtheoreticalMode << "]" << endl;
    cout << "*******************************" << endl;
}

} // Fin du namespace DataServer
} // Fin du namespace URANIE

