//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TTriangularPossibility.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2010-07-07 15:52:43

 \brief  Class of attributes following a probability density function


 */
#ifndef TTRIANGULARPOSSIBITY_H
#define TTRIANGULARPOSSIBITY_H

#include "TString.h"
#include <iostream>

// Uranie
#include "DataServer.h"
#include "TPossibilityAttribute.h"

namespace URANIE
{
namespace DataServer
{

/*! \class URANIE::DataServer::TTriangularPossibility
 This is the parent class for all attributes which values can be generated by a 
 possibilty (cf RaFu IRSN). Child objects are possibility attributes following a specific possibility.
 
 Each must be redefined in the daughter classes.
 
 */
class TTriangularPossibility: public TPossibilityAttribute
{
protected:
    Double_t _dmode;                  ///< The mod of the triangular possibility
public:
    // -----------------------------------------------------------------------
    //  Constructors
    // -----------------------------------------------------------------------
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Constructor setting the name 
    /*!
     \param str (TString)  Name of the attribute
     */
    TTriangularPossibility(TString str);

    //! Constructor setting the name and the law type 
    /*!
     \param str (TString)    Name of the attribute
     \param min (Double_t)   Lower bound of the attribute's range.
     \param max (Double_t)   Upper bound of the attribute's range.	  
     \param mod (Double_t)   The x-value of maximim of possibility.
     \param monotonous (EMonotonousRelationship) The relationship of monotonous.
     */
    TTriangularPossibility(TString str, Double_t min, Double_t max,
            Double_t mod,
            TPossibilityAttribute::EMonotonousRelationship monotonous);

    //! Default Destructor
    virtual ~TTriangularPossibility();
    //@}

    //---------------------------------------------
    /** @name The mode of the triangular possibility
     **/
    //@{
    //! Return the x-value of maximim of possibility
    Double_t getMode()
    {
        return _dmode;
    }
    //! Set the x-value of maximim of possibility
    void setMode(Double_t x);
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    //! Display information specific to this class
    virtual void printLog();
    //@}

    ClassDef(TTriangularPossibility, ID_DATASERVER)
    //Definition of a stochastic variable
};
} // Fin du namespace DataServer
} // Fin du namespace URANIE

#endif
