//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TUniformByPartsDistribution.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2009-12-07 11:11:42

 \brief Interface of the class URANIE::DataServer::TUniformByPartsDistribution


 */
/**
 *   
 * @class URANIE::DataServer::TUniformByPartsDistribution
 *
 * @brief Description of the class TUniformByPartsDistribution.
 * 
 This class defines a stochastic attribute following a Uniform distribution by parts.

 User parameters are the minimum value \f$ x_{min} \f$ , the median \f$ x_{median} \f$ , 
 and the maximum value \f$ x_{max} \f$ .

 The distribution is uniform on the ranges \f$ [x_{min}, x_{median}] \f$ and 
 \f$ ]x_{median}, x_{max}] \f$ , which means that:
 \li by definition, the probability for x to belong either to \f$ [x_{min}, x_{median}] \f$ or to \f$ ]x_{median}, x_{max}] \f$ is 0.5
 \li the density of \f$ x_1 \in [x_{min}, x_{median}] \f$ may be different from the one of \f$ x_2 \in ]x_{median}, x_{max}] \f$  

 \par Probability Density Function (PDF):
 \li \f$ f(x) = 0.5/(x_{median}-x_{min}) \f$ for \f$ x \in [x_{min},x_{median}] \f$
 \li \f$ f(x) = 0.5/(x_{max}-x_{median}) \f$ for \f$ x \in ]x_{median},x_{max}] \f$
 \li \f$ f(x) = 0 \f$ otherwise

 \par Cumulative Distribution Function (CDF):
 \li \f$ F(x) = 0 \f$ for \f$ x < x_{min} \f$ 
 \li \f$ F(x) = 0.5 * (x-x_{min})/(x_{median}-x_{min})\f$ for \f$ x \in [x_{min},x_{median}] \f$
 \li \f$ F(x) = 0.5 + 0.5 * (x-x_{median})/(x_{max}-x_{median})\f$ for \f$ x \in ]x_{median},x_{max}] \f$
 \li \f$ F(x) = 1 \f$ for \f$ x > x_{max} \f$ 

 \par Inverse CDF:
 \li \f$ F^{-1}(p) = x_{min} + 2 * p * ( x_{median} - x_{min} ) \f$ for \f$ p \in [0, 0.5] \f$
 \li \f$ F^{-1}(p) = x_{median} + 2 * (p-0.5) * ( x_{max} - x_{median} )  \f$ for \f$ p \in ]0.5, 1] \f$


 */
#ifndef TUNIFORMBYPARTSDISTRIBUTION_H
#define TUNIFORMBYPARTSDISTRIBUTION_H

//Uranie
#include "TStochAttribut.h"

namespace URANIE
{
namespace DataServer
{
class TUniformByPartsDistribution: public TStochasticAttribute
{
protected:
    Double_t _dparameterMedian; ///< The median parameter

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    /*!
     The default range is [0,1] with median = 0.5
     */
    TUniformByPartsDistribution();

    //! Constructor setting the name
    /*! 
     The default range is [0,1] with median = 0.5
     
     \param name (TString) Name of the attribute
     */
    TUniformByPartsDistribution(TString name);

    //! Constructor setting name and range
    /*!
     The default value for the median is (max-min)/2
     
     \param name (TString)  Name of the attribute
     \param min  (Double_t) Lower bound
     \param max  (Double_t) Upper bound
     */
    TUniformByPartsDistribution(TString name, Double_t min, Double_t max);

    //! Constructor setting name, range and median
    /*!                 
     \param name   (TString)  Name of the attribut
     \param min    (Double_t) Lower bound
     \param max    (Double_t) Upper bound
     \param median (Double_t) Median value
     */
    TUniformByPartsDistribution(TString name, Double_t min, Double_t max,
            Double_t median);

    //! Construction of a TUniformDistribution from an existing TAttribute object
    /*!
     \warning this copy constructor is incomplete. Please do not use.
     (Nicolas Gilardi, 2010.05.28)
     
     \param att (TAttribute *) Pointer to an existing TAttribute object                                
     */
    TUniformByPartsDistribution(URANIE::DataServer::TAttribute *att);

    //! Default destructor
    virtual ~TUniformByPartsDistribution();
    //@}

    /** @name Law parameters
     This law has three parameters: the lower bound, the upper bound and 
     the median. The bounds parameters are defined in the class URANIE::DataServer::TAttribute.
     The median parameter is defined here.
     **/
    //@{
    //! Set the value of the median parameter
    /*!
     \param median (Double_t) Value of the median parameter
     
     \exception UErrorExceptions if median is outside of the interval [lowerBound,upperBound]
     */
    virtual void setParameterMedian(Double_t median);

    //! Returns the value of the median parameter
    virtual Double_t getParameterMedian()
    {
        return _dparameterMedian;
    }

    //! Set all parameters at once
    /*!
     \param lower  (Double_t) value of the lower bound
     \param upper  (Double_t) value of the upper bound
     \param median (Double_t) value of the median
     
     \exception UErrorExceptions if median is outside of the interval [lower,upper]
     */
    virtual void setParameters(Double_t lower, Double_t upper, Double_t median);

    //@}

    /** @name Distribution specific methods
     PDF, CDF, etc.
     **/
    //@{                
    //! Computes the theoretical values of various statistical measures, like mean, standard deviation, mode, etc.
    virtual void computeTheoreticalInformation();

    //! Returns the probability density for the value x
    /*!
     \param x (Double_t) An acceptable value for the attribute                 
     */
    virtual Double_t getPDF(Double_t x);

    //! Returns the cumulative distribution for the value x
    /*!
     \param x (Double_t) An acceptable value for the attribute                
     */
    virtual Double_t getCDF(Double_t x);

    //! Returns the value corresponding to the cumulative density p
    /*!
     \param p (Double_t) A valid probability
     
     \exception UErrorExceptions if p is outside of the interval [0,1]                
     */
    virtual Double_t getInvCDF(Double_t p);
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    //! Display information specific to this class
    /*!
     \param option (Option_t *) character chain containing a list of options (Default = "")
     
     \note definir une liste d'options et un exemple d'utilisation (Nicolas Gilardi 2010.05.31)
     */
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::DataServer::TUniformByPartsDistribution, ID_DATASERVER)
    //Classe de
};

} // Fin du namespace DataServer
} // Fin du namespace URANIE
#endif
