//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TUniformDistribution.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Thu Jan 18 17:08:58 CET 2007

 \brief Implementation of the class URANIE::DataServer::TUniformDistribution

 */
#include "TMath.h"
#include "TUniformDistribution.h"
#include <iostream>
using std::cout;
using std::endl;

ClassImp (URANIE::DataServer::TUniformDistribution)

namespace URANIE
{
namespace DataServer
{

// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------

TUniformDistribution::TUniformDistribution() :
        TStochasticAttribute("", TStochasticAttribute::kUniform)
{
    if (_blog)
        cout << "Constructing URANIE::DataServer::TUniformDistribution" << endl;
    computeTheoreticalInformation();
}

TUniformDistribution::TUniformDistribution(TString str) :
        TStochasticAttribute(str, TStochasticAttribute::kUniform)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TUniformDistribution with: Name ["
                << str << "]" << endl;
    computeTheoreticalInformation();
}

TUniformDistribution::TUniformDistribution(TString str, Double_t min,
        Double_t max) :
        TStochasticAttribute(str, TStochasticAttribute::kUniform, min, max)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TUniformDistribution with: Name ["
                << str << "] Min [" << min << "] Max [" << max << "]" << endl;
    computeTheoreticalInformation();
}

TUniformDistribution::TUniformDistribution(URANIE::DataServer::TAttribute *att) :
        TStochasticAttribute(att, TStochasticAttribute::kUniform)
{
    if (_blog)
        cout
                << "Constructing URANIE::DataServer::TUniformDistribution from existing TAttribute ["
                << att->GetName() << "]" << endl;
    computeTheoreticalInformation();
}

// -----------------------------------------------------------------------
//  Destructors
// -----------------------------------------------------------------------

TUniformDistribution::~TUniformDistribution()
{
    if (_blog)
        cout << "Destroying URANIE::DataServer::TUniformDistribution "
                << GetName() << endl;

}

// -----------------------------------------------------------------------
//  Parameters
// -----------------------------------------------------------------------

void TUniformDistribution::setParameters(Double_t lower, Double_t upper)
{
    // Verify that median is between lower and upper
    try
    {
        if ( lower >= upper)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TUniformDistribution::setParameters: lower bound [%f] is equal to or greater than upper one [%f]",
                                                            lower, upper));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }
    
    setBounds(lower, upper);

    computeTheoreticalInformation();
}


// -----------------------------------------------------------------------
//  Theoretical Information
// -----------------------------------------------------------------------

void TUniformDistribution::computeTheoreticalInformation()
{
    _dtheoreticalMean = (upperBound + lowerBound) / 2.0;
    _dtheoreticalStdDev = (upperBound - lowerBound) / TMath::Sqrt(12);

    // Warning: any value inside the interval is a valid Mode. I chose 
    //          the middle of the range...
    _dtheoreticalMode = (upperBound + lowerBound) / 2.0;
}

// -----------------------------------------------------------------------
//  Distribution
// -----------------------------------------------------------------------

Double_t TUniformDistribution::getPDF(Double_t x)
{
    Double_t pdf = 0.;
    if ((x >= lowerBound) && (x <= upperBound))
        pdf = 1.0 / (upperBound - lowerBound);
    return pdf;
}

Double_t TUniformDistribution::getCDF(Double_t x)
{
    Double_t cdf;
    if (x < lowerBound)
        cdf = 0.0;
    else if (x > upperBound)
        cdf = 1.0;
    else
        cdf = (x - lowerBound) / (upperBound - lowerBound);
    return cdf;
}

Double_t TUniformDistribution::getInvCDF(Double_t p)
{
    // Verify that p is a valid probability
    try
    {
        if ((p < 0.0) || (p > 1.0))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TUniformDistribution::getInvCDF: parameter p [%f] is not a valid probability: p must be in the interval [0,1]",
                            p));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }

    return lowerBound + p * (upperBound - lowerBound);
}

// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------        

void TUniformDistribution::printLog(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    TStochasticAttribute::printLog();
    cout << "*******************************" << endl;
    cout << "** TUniformDistribution::printLog  ******" << endl;
    cout << "**   Law[" << _lawType << "] TUniformDistribution" << endl;
    cout << "**   Bounds[" << lowerBound << ", " << upperBound << "]" << endl;
    cout << "**   Theoretical Mean[" << _dtheoreticalMean << "]" << endl;
    cout << "**   Theoretical StdDev[" << _dtheoreticalStdDev << "]" << endl;
    cout << "**   Theoretical Mode[" << _dtheoreticalMode << "]" << endl;
    cout << "*******************************" << endl;
}

} // Fin du namespace DataServer
} // Fin du namespace URANIE

