//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TWeibullDistribution.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Thu Jan 18 17:08:58 CET 2007

 \brief Interface of the class URANIE::DataServer::TWeibullDistribution

 */
/**

 * @class URANIE::DataServer::TWeibullDistribution
 *
 * @brief Description of the class TWeibullDistribution.
 *

 This class defines a stochastic attribute following an Weibull distribution.

 This law is defined by three parameters: the <b>location</b> \f$ x_{min} \f$, 
 the <b>scale</b> \f$ \lambda > 0 \f$, and the <b>shape</b> \f$ k > 0 \f$. The law is 
 defined over the interval \f$ [x_{min}, \inf) \f$. 

 This law gives the distribution of failures, where the failure rate is proportional 
 to a power of time. The shape parameter, \f$ k \f$, is that power plus one, and so can be 
 interpreted directly as follows:

 \li A value of k < 1 indicates that the failure rate decreases over time. This happens when early failures leaves only the most robust elements and eliminate the others. 
 \li A value of k = 1 indicates that the failure rate is constant over time. This might suggest random external events are causing failure.
 \li A value of k > 1 indicates that the failure rate increases with time. This happens if there is an "aging" process.

 (inspired from http://en.wikipedia.org/wiki/Weibull_distribution)


 \par Probability Density Function (PDF):
 \li \f$ f(x) = k/\lambda * ((x-x_{min})/\lambda)^{k-1} * \exp\left(-((x-x_{min})/\lambda)^{k}\right)\f$ for \f$ x \geq x_{min} \f$
 \li \f$ f(x) = 0 \f$ otherwise

 \par Cumulative Distribution Function (CDF):
 \li \f$ F(x) = 0 \f$ for \f$ x < x_{min} \f$
 \li \f$ F(x) = 1 - \exp\left(-((x-x_{min})/\lambda)^k\right) \f$ otherwise

 \par Inverse CDF:
 \li \f$ F^{-1}(p) = x_{min} + \lambda * (-\ln(1-p))^{1/k} \f$ for \f$ p \in [0,1] \f$ 


 */
#ifndef TWEIBULLDISTRIBUTION_H
#define TWEIBULLDISTRIBUTION_H

//Uranie
#include "TStochAttribut.h"

namespace URANIE
{
namespace DataServer
{
class TWeibullDistribution: public TInfiniteDistribution
{
private:
    Double_t _dparameterLambda;  ///< The scale parameter \f$ \lambda \f$
    Double_t _dparameterK;       ///< The shape parameter \f$ k \f$
    Double_t _dparameterXmin;    ///< The location parameter \f$ x_{min} \f$
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    /*!
     Default scale is 1.0, default shape is 1.0, default location is 0.0
     */
    TWeibullDistribution();

    //! Constructor setting the name
    /*!
     Default scale is 1.0, default shape is 1.0, default location is 0.0
     
     \param name (TString) Name of the attribute
     */
    TWeibullDistribution(TString name);

    //! Constructor setting name, scale, shape and location
    /*!                 
     \param name (TString) Name of the attribute
     \param lambda (Double_t) Scale parameter \f$ \lambda \f$ 
     \param k (Double_t) Shape parameter \f$ k \f$ (Default = 1.0)
     \param xmin (Double_t) Location parameter \f$ x_{min} \f$ (Default = 0.0)
     */
    TWeibullDistribution(TString name, Double_t lambda, Double_t k = 1.0,
            Double_t xmin = 0.0);

    //! Construction of a TWeibullDistribution from an existing TAttribute object
    /*!
     \warning this copy constructor is incomplete. Please do not use.
     (Nicolas Gilardi, 2010.06.11)

     \param att (URANIE::DataServer::TAttribute) the attribute which we want to make a random parameter
     \param lambda (Double_t) Scale parameter \f$ \lambda \f$ 
     \param k (Double_t) Shape parameter \f$ k \f$ (Default = 1.0)
     \param xmin (Double_t) Location parameter \f$ x_{min} \f$ (Default = 0.0)
     */
    TWeibullDistribution(URANIE::DataServer::TAttribute *att, Double_t lambda,
            Double_t k = 1.0, Double_t xmin = 0.0);

    //! Default destructor
    virtual ~TWeibullDistribution();
    //@}

    /** @name Law parameters
     This law has three parameters: the scale \f$ \lambda \f$, the shape 
     \f$ k \f$ and the location \f$ x_{min} \f$. All are defined here.
     **/
    //@{
    //! Set the scale parameter \f$ \lambda \f$
    /*!
     \param lambda (Double_t) Scale parameter \f$ \lambda \f$
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)
     
     \exception UErrorExceptions if lambda <  1e-4
     */
    virtual void setParameterLambda(Double_t lambda, Bool_t recompute = kTRUE);

    //! Get the scale parameter value
    Double_t getParameterLambda()
    {
        return _dparameterLambda;
    }

    //! Set the shape parameter \f$ k \f$
    /*!
     \param k (Double_t) Shape parameter \f$ k \f$
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)
     
     \exception UErrorExceptions if k <  1e-4
     */
    virtual void setParameterK(Double_t k, Bool_t recompute = kTRUE);

    //! Get the shape parameter value
    Double_t getParameterK()
    {
        return _dparameterK;
    }

    //! Set the location parameter \f$ x_{min} \f$
    /*!
     \param xmin (Double_t) Location parameter \f$ x_{min} \f$
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)

     \warning in order to avoid border effects, if the lowerBound attribute 
     is set to minus infinity, it is re-initialised to xmin.                
     */
    virtual void setParameterXmin(Double_t xmin, Bool_t recompute = kTRUE);

    //! Get the location parameter value
    Double_t getParameterXmin()
    {
        return _dparameterXmin;
    }

    //! Set all the parameters
    /*!
     \param lambda (Double_t) Scale parameter \f$ \lambda \f$
     \param k (Double_t) Shape parameter \f$ k \f$    
     \param xmin (Double_t) Location parameter \f$ x_{min} \f$
     \param recompute (Bool_t) should we recompute the theoretical information ? (Default = kTRUE)
     */
    void setParameters(Double_t lambda, Double_t k, Double_t xmin,
            Bool_t recompute = kTRUE);
    //@}

    /** @name Distribution specific methods
     PDF, CDF, etc.
     **/
    //@{
    //! Computes the theoretical values of various statistical measures.
    virtual void computeTheoreticalInformation();

    //! Returns the probability density for the value x
    /*!
     \param x (Double_t) An acceptable value for the attribute                                  
     */
    virtual Double_t getPDF(Double_t x);

    //! Returns the cumulative distribution for the value x
    /*!
     \param x (Double_t) An acceptable value for the attribute
     */
    virtual Double_t getCDF(Double_t x);

    //! Returns the value corresponding to the cumulative density p
    /*!
     \param p (Double_t) A valid probability
     
     \exception UErrorExceptions if p is outside of the interval [0,1]
     */
    virtual Double_t getInvCDF(Double_t p);
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::DataServer::TWeibullDistribution, ID_DATASERVER)
    //Classe de
};

} // Fin du namespace DataServer
} // Fin du namespace URANIE
#endif
