//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   UEntry.cxx
 \author Jean-Baptiste Blanchard <jean-baptiste.banchard@cea.fr>

 $Revision$
 $Date$
*/

#define UENTRY_CXX

#include "UEntry.h"
#include <sstream>
#include "TMath.h"

ClassImp (URANIE::DataServer::UEntry)

namespace URANIE
{
namespace DataServer
{

//========================================
//      Constructors & Destructors
//========================================
UEntry::UEntry():
    TObject(), tag(UEntry::Empty), name("[empty_const]"), bisnew(false)
{
    //val.dou = 0; //new double[size]; val.dou[0]=0;
    val.dou=NULL;
}
    
UEntry::UEntry(UEntry::TAG atag, string aname):
    TObject(), tag(atag), name(aname), bisnew(true)
{
//	XYZ<<" creating entry "<<name<<" tag "<<tag<<" ";
    switch(tag)
    {
    case UEntry::Double: val.dou = new double(); break; 
    case UEntry::String: val.str = new string(""); break; 
    case UEntry::Vector: val.vec = new vector<double>(); break;
    default: break;
    }
}

UEntry::UEntry(UEntry::TAG atag, string aname, void *ptr):
    TObject(), tag(atag), name(aname), bisnew(false)
{
//	XYZ<<" creating entry "<<name<<" tag "<<tag<<" ";
    switch(tag)
    {
    case UEntry::Double: val.dou = (double*) ptr; break; 
    case UEntry::String: val.str = (string*) ptr; break; 
    case UEntry::Vector: val.vec = (vector<double>*)ptr; break;
    default: break;
    }
}
    
UEntry::~UEntry()
{
    if( bisnew )
    {
	//cout<<"deleting entry "<<name<<" tag "<<tag<<endl;
	switch(tag)
	{
	case UEntry::Double: delete val.dou; val.dou = NULL; break;
	case UEntry::String: delete val.str; val.str = NULL; break;
	case UEntry::Vector: delete val.vec; val.vec = NULL; break;
	default: break;
	}
    }
}

//========================================
//         Content modification
//========================================

void UEntry::setAddress( void *ptr )
{
    if( bisnew )
    {
	switch(tag)
	{
	case UEntry::Double: delete val.dou; val.dou = NULL; break;
	case UEntry::String: delete val.str; val.str = NULL; break;
	case UEntry::Vector: delete val.vec; val.vec = NULL; break;
	default: break;
	}
    }
//	XYZ<<" setaddress entry "<<name<<" tag "<<tag<<" ";
    switch(tag)
    {
    case UEntry::Double: val.dou = (double*)(ptr); break; 
    case UEntry::String: val.str = (string*)(ptr); break; 
    case UEntry::Vector: val.vec = (vector<double>*)(ptr); break;
    default: break;
    }
}

string UEntry::getAddress( )
{
    stringstream sstr;
    switch(tag)
    {
    case UEntry::Double: sstr << val.dou; break; 
    case UEntry::String: sstr << val.str; break; 
    case UEntry::Vector: sstr << val.vec; break;
    default: break;
    }
    return sstr.str();
}
    
string UEntry::dumpStr( TString delim, TString beg, TString end, TString format )
{
    stringstream buffer; buffer.precision(SSTR_PREC);
    switch(tag)
    {
    case UEntry::Double:
	if( format=="%e" )
	    buffer << std::scientific <<std::setprecision(SSTR_PREC) << *val.dou;
	else if( format=="%g" )
	    buffer << std::setprecision(SSTR_PREC) << *val.dou;
	else
	{
	    if( format.Contains("%i") || format.Contains("%d") )
	    {
		int nval = TMath::FloorNint(*val.dou);
		buffer << nval ;
	    }
	    else
		buffer << Form( format.Data(), *val.dou);
	}
	break;
    case UEntry::String: buffer << beg << *val.str << end;  break;
    case UEntry::Vector: buffer << beg << URANIE::DataServer::UEntry::dumpVectorStr(val.vec,-1, delim, format) << end;  break;
    default: break;
    }
    return buffer.str();
}

string UEntry::dumpStr( unsigned int nentry, bool &moretowrite, TString beg, TString end, TString format )
{
    stringstream buffer; buffer.precision(SSTR_PREC);
    moretowrite=false;
    switch(tag)
    {
    case UEntry::Double:
	if(nentry==0)
	{	   
	    if( format=="%e" )
		buffer << std::scientific <<std::setprecision(SSTR_PREC) << *val.dou;
	    else
	    {
		if( format.Contains("%i") || format.Contains("%d") )
		{
		    int nval = TMath::FloorNint(*val.dou);
		    buffer << nval ;
		}
		else
		    buffer << Form( format.Data(), *val.dou);
	    }
	}
	break;
    case UEntry::String:
	if(nentry==0) buffer << beg << *val.str << end;
	break;
    case UEntry::Vector:
	if( nentry < val.vec->size() )
	{
	    if( format=="%e" )
		buffer << std::scientific <<std::setprecision(SSTR_PREC) << val.vec->at(nentry);
	    else
	    {
		if( format.Contains("%i") || format.Contains("%d") )
		{
		    int nval = TMath::FloorNint(val.vec->at(nentry));
		    buffer << nval ;
		}
		else
		    buffer << Form( format.Data(), val.vec->at(nentry));
	    }
	    if( nentry < (val.vec->size()-1) )
		moretowrite=true;
	}
	break;
    default: break;
    }
    return buffer.str();
}

void UEntry::clearEntry()
{
    switch(tag)
    {
    case UEntry::Double : *val.dou = 0; break;
    case UEntry::String : *val.str = ""; break;
    case UEntry::Vector : val.vec->clear(); break;
    default: break;
    }
}

int UEntry::getDim()
{
    switch(tag)
    {
    case UEntry::Double : return -1;
    case UEntry::String : return -2;
    case UEntry::Vector : return val.vec->size();
    default: return 0;
    }
}

//========================================
//          Content fillers
//========================================

void UEntry::copyEntry(UEntry *src)
{
    switch(src->tag)
    {
    case UEntry::Double : *val.dou = *src->val.dou; break;
    case UEntry::String : *val.str = *src->val.str; break;
    case UEntry::Vector :
	val.vec->clear();
	for(unsigned int ivec=0; ivec< src->val.vec->size(); ivec++)
	    val.vec->push_back( src->val.vec->at(ivec) );
	break;
    default: break;
    }
}


void UEntry::pushEntry( void *ptr, bool isdouble )
{
    if(tag==UEntry::Double && isdouble) *val.dou = *(double*)(ptr);
    else if(tag==UEntry::Double) *val.dou = *(float*)(ptr);
    if(tag==UEntry::String) *val.str = *(string*)(ptr);
    if(tag==UEntry::Vector && isdouble) val.vec->push_back(*(double*)(ptr));
    else if(tag==UEntry::Vector) val.vec->push_back(*(float*)(ptr));
}

void UEntry::pushEntry( double value )
{
    if(tag==UEntry::Double) *val.dou = value;
    else if(tag==UEntry::Vector) val.vec->push_back(value);
}

void UEntry::pushEntry( TString value, TString del, TString beg, TString end )
{

    if(del!="") value.ReplaceAll(del,"");
    value.ReplaceAll(beg,"");
    value.ReplaceAll(end,"");
    
    switch(tag)
    {
    case UEntry::Double :
	if(value!=" " && value != "")
	    *val.dou = atof(value);
	break;
    case UEntry::String : *val.str = string(value); break;
    case UEntry::Vector :
	if(value!=" " && value != "")
	    val.vec->push_back(atof(value));
	break;
    default: break;
    }
}

void UEntry::assignEntry( void *ptr, bool isdouble )
{
    if(tag==UEntry::Double && isdouble) *val.dou = *(double*)(ptr);
    else if(tag==UEntry::Double) *val.dou = *(float*)(ptr);
    if(tag==UEntry::String) *val.str = *(string*)(ptr);
    if(tag==UEntry::Vector) 
    {    
	val.vec->clear();
	vector<double> buffer = *(vector<double>*)(ptr);
	for(unsigned int ival=0; ival<buffer.size(); ival++)
	    val.vec->push_back( buffer.at(ival) );
    }
//	XYZ<<" for entry "<<name<<" tag "<<tag<<" assign value "<<*val.dou<<" input "<<*(double*)(ptr)<<"  ptr  "<<ptr<<endl;
}
    
void UEntry::fillEntry( TString line, TString delim, TString beg, TString end )
{

    //XYZ<<"beg ["<<beg<<"] end ["<<end<<"]  "<<line<<endl;
    if( beg!="" ) line.ReplaceAll(beg,"");
    if( end!="" ) line.ReplaceAll(end,"");
	    
    switch (tag)
    {
    case UEntry::Double :
	if(line!=" " && line != "")
	    *val.dou = atof(line.Data());
	break;
    case UEntry::String : *val.str = line; break;
    case UEntry::Vector :
    {
	val.vec->clear();
	istringstream tempor(line.Data()); string test;
	if( delim==" ")
	{
	    while(tempor >> test )
		if( test != " " && test != "")
		    val.vec->push_back( atof(test.c_str()) );
	}
	else
	{
	    char mydelim=delim.Data()[0];
	    while(std::getline(tempor, test, mydelim))
		if( test != " " && test != "")
		    val.vec->push_back( atof(test.c_str()) );
	}
    }
    break;
    default: break;
    }
}
    

//========================================
//          Utility functions
//========================================
  
bool UEntry::isWhiteLine(string line)
{
    for (Int_t _i = 0; _i < Int_t(line.length()); _i++)
    {
	if (line[_i] == ' ')
	    continue;
	else if (line[_i] == '\0' || line[_i] == '\n')
	    return true;
	else
	    return false;
    }
    return true;
}

void UEntry::StringParser(string input, string delim, vector<string> &res, bool LeaveEmpty)
{

    res.clear();
    size_t pos = 0;
    string oldinput, oldinput2, token;
    string otherdelim="\t";
    if(delim==" ") input=string(TString(input).ReplaceAll(otherdelim, delim) );

    vector<string> res2;
    while ((pos = input.find("\"")) != std::string::npos)
    {
	token = input.substr(0, pos);
	res2.push_back(token);
	oldinput=input;
	input.erase(0, pos + delim.length());
    }
    res2.push_back(input);
    
    for(unsigned int ires2=0; ires2<res2.size(); ires2++)
    {
	string input2=res2.at(ires2);
	if(ires2%2==1) res.push_back(input2);
	else
	{
	    while ((pos = input2.find(delim)) != std::string::npos)
	    {
		token = input2.substr(0, pos);
		if ( LeaveEmpty && token != delim) res.push_back(token);
		else if(!LeaveEmpty && token!="" && token != delim) res.push_back(token);
		oldinput2=input2;
		input2.erase(0, pos + delim.length());
	    }
	    if ( LeaveEmpty && input2 != delim ) res.push_back(input2);
	    else if(!LeaveEmpty && input2!="" && input2 != delim)  res.push_back(input2);
	}
    }

}

string UEntry::dumpVectorStr(vector<double> *buffer, int precision, string delim, TString format)
{
    stringstream value; value.precision(SSTR_PREC);
    for(unsigned int ival=0; ival<buffer->size(); ival++)
    {
	if(format!="%e" && format!="%g")
	{
	    if( format.Contains("%i") || format.Contains("%i") )
	    {
		int nval = TMath::FloorNint(buffer->at(ival));
		value << Form( format.Data(), nval ) << ((ival!=buffer->size()-1) ? "," : "") ;
	    }
	    else
		value << Form( format.Data(), buffer->at(ival) ) << ((ival!=buffer->size()-1) ? "," : "") ;
	}
	else if(format=="%e")
	{
	    if (precision != -1)
		value << std::scientific <<std::setprecision(precision) << buffer->at(ival) << ((ival!=buffer->size()-1) ? "," : "") ;
	    else
		value << std::scientific << buffer->at(ival) << ((ival!=buffer->size()-1) ? delim : "") ;
	}
       	else if(format=="%g")
	{
	    if (precision != -1)
		value << std::setprecision(precision) << buffer->at(ival) << ((ival!=buffer->size()-1) ? "," : "") ;
	    else
		value <<  buffer->at(ival) << ((ival!=buffer->size()-1) ? delim : "") ;
	}
    }
    return value.str();
}

void UEntry::fillVectorFromStr(TString input, TString delim, vector<double> &buffer)
{
    
        
    buffer.clear();
    istringstream tempor(input.Data()); string test;
    if( delim==" ")
    {
        while(tempor >> test )
            if( test != " " && test != "")
                buffer.push_back( atof(test.c_str()) );
    }
    else
    {
        char mydelim=delim.Data()[0];
        while(std::getline(tempor, test, mydelim))
            if( test != " " && test != "")
                buffer.push_back( atof(test.c_str()) );
    }
}


std::istream& UEntry::safeGetline(std::istream& is, std::string& t)
{
    t.clear();
    std::istream::sentry se(is, true);
    std::streambuf* sb = is.rdbuf();

    for(;;) {
	int c = sb->sbumpc();
	switch (c) {
	case '\n':
	    return is;
	case '\r':
	    if(sb->sgetc() == '\n')
		sb->sbumpc();
	    return is;
	case EOF:
	    // Also handle the case when the last line has no line ending
	    if(t.empty())
		is.setstate(std::ios::eofbit);
	    return is;
	default:
	    t += (char)c;
	}
    }
}




}
}
