//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   UTimer.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   jeu avr 26 14:14:11 CEST 2007
 
 $Revision$
 $Date$
 
 */

#include "UTimer.h"
#include "Riostream.h"


const TString BC_white = "\033[1;37m";
const TString BC_red = "\033[0;31m";
const TString BC_blue = "\033[0;34m";
const TString BC__b0 = "\033[47m";
const TString BC__b1 = "\033[1;42m";
const TString BC__f1 = "\033[33m";
const TString EC__ = "\033[0m";

const TString URANIE::DataServer::UTimer::fgClassName = "UTimer";
const Int_t URANIE::DataServer::UTimer::fgNbins = 24;

ClassImp (URANIE::DataServer::UTimer);

namespace URANIE
{
namespace DataServer
{

//_______________________________________________________________________
UTimer::UTimer(Bool_t colourfulOutput) :
        fNcounts(0), fPrefix(UTimer::fgClassName), fColourfulOutput(
                colourfulOutput)
{
    // constructor
    fLogger = new URANIE::DataServer::UMessageLogger(fPrefix.Data());

    Reset();
}

//_______________________________________________________________________
UTimer::UTimer(Int_t ncounts, TString prefix, Bool_t colourfulOutput) :
        fNcounts(ncounts), fColourfulOutput(colourfulOutput)
{
    // standard constructor: ncounts gives the total number of counts that 
    // the loop will iterate through. At each call of the timer, the current
    // number of counts is provided by the user, so that the timer can obtain
    // the due time from linearly interpolating the spent time.
    if (prefix == "")
        fPrefix = UTimer::fgClassName;
    else
        fPrefix = prefix;

    fLogger = new URANIE::DataServer::UMessageLogger(fPrefix.Data());

    Reset();
}

//_______________________________________________________________________
UTimer::~UTimer(void)
{
    // destructor
    delete fLogger;
    fLogger = NULL;
}

void UTimer::Init(Int_t ncounts)
{
    // timer initialisation
    fNcounts = ncounts;
    Reset();
}

//_______________________________________________________________________
void UTimer::Reset(void)
{
    // resets timer
    TStopwatch::Start (kTRUE);
}

//_______________________________________________________________________
Double_t UTimer::ElapsedSeconds(void)
{
    // computes elapsed tim in seconds
    Double_t rt = TStopwatch::RealTime();
    TStopwatch::Start (kFALSE);
    return rt;
}
//_______________________________________________________________________

TString UTimer::GetElapsedTime(Bool_t Scientific)
{
    // returns pretty string with elaplsed time
    return SecToText(ElapsedSeconds(), Scientific);
}

//_______________________________________________________________________
TString UTimer::GetLeftTime(Int_t icounts)
{
    // returns pretty string with time left
    Double_t leftTime = (
            icounts <= 0 ? -1 :
            icounts > fNcounts ?
                    -1 :
                    Double_t(fNcounts - icounts) / Double_t(icounts)
                            * ElapsedSeconds());

    return SecToText(leftTime, kFALSE);
}

//_______________________________________________________________________
void UTimer::DrawProgressBar(Int_t icounts, Int_t nFailure)
{
    // draws progress bar in color or B&W
    // caution: 

    // sanity check:
    if (icounts > fNcounts - 1)
        icounts = fNcounts - 1;
    if (icounts < 0)
        icounts = 0;
    Int_t ic = Int_t(Float_t(icounts) / Float_t(fNcounts) * fgNbins);

    std::clog << fLogger->GetPrintedSource();
    if (fColourfulOutput)
        std::clog << BC__b1 << BC__f1 << "[" << EC__;
    else
        std::clog << "[";
    for (Int_t i = 0; i < ic; i++)
    {
        if (fColourfulOutput)
            std::clog << BC__b1 << BC__f1 << ">" << EC__;
        else
            std::clog << ">";
    }
    for (Int_t i = ic + 1; i < fgNbins; i++)
    {
        if (fColourfulOutput)
            std::clog << BC__b1 << BC__f1 << "." << EC__;
        else
            std::clog << ".";
    }
    if (fColourfulOutput)
        std::clog << BC__b1 << BC__f1 << "]" << EC__;
    else
        std::clog << "]";

    // timing information
    if (fColourfulOutput)
    {
        std::clog << EC__ << " ";
        std::clog << "(" << BC_red
                << Int_t((100 * (icounts + 1)) / Float_t(fNcounts)) << "%"
                << EC__ << ", " << "time left: " << this->GetLeftTime(icounts)
                << EC__ << " : Failure " << BC_red << nFailure << EC__ << ") ";
    }
    else
    {
        std::clog << "] ";
        std::clog << "(" << Int_t((100 * (icounts + 1)) / Float_t(fNcounts)) << "%"
                << ", " << "time left: " << this->GetLeftTime(icounts)
                << " : Failure " << nFailure << ") ";
    }
    std::clog << "\r" << std::flush;
}

//_______________________________________________________________________
TString UTimer::SecToText(Double_t seconds, Bool_t Scientific)
{
    // pretty string output
    TString out = "";
    if (Scientific)
        out = Form("%.3g sec", seconds);
    else if (seconds < 0)
        out = "unknown";
    else if (seconds <= 300)
        out = Form("%i sec", Int_t(seconds));
    else
    {
        if (seconds > 3600)
        {
            Int_t h = Int_t(seconds / 3600);
            if (h <= 1)
                out = Form("%i hr : ", h);
            else
                out = Form("%i hrs : ", h);

            seconds = Int_t(seconds) % 3600;
        }
        Int_t m = Int_t(seconds / 60);
        if (m <= 1)
            out += Form("%i min", m);
        else
            out += Form("%i mins", m);
    }

    return (fColourfulOutput) ? BC_red + out + EC__ : out;
}

} //  //namespace DataServer
} //namespace URANIE
