//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTDataServer.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   $Date$
 
 \brief  Test file of the class URANIE::DataServer::TDataServer
 
 Created: wen nov 21 07:59:24 2007


 \par Modifications
 \li JB Blanchard (CEA) 2016-2-22: Add and complete few tests to cope with new implementation of string and vectors within the TDSNtupleD
 \li maud Salvoldi 2011/09/28
 - String comparison with strcmp in testGetArchiveFileName()
 - replace Form("%s/%s....) by gSystem->ConcatFileName(....) in testGetArchiveFileName()
 \li Cedric Forgues (C.S.) 2010/08/30
 Changing values of reference for testComputeCorrelationMatrix2, testComputeCorrelationMatrix3, testComputeCorrelationMatrix4 due to new implementation for Sort method between versions 5.18.00 and 5.26.00 of ROOT.
 \li Cedric Forgues (C.S.) 2009/06/16
 Adding tests for DI 49.  
 \li Cedric Forgues (C.S.) 2009/05/10
 Adding of testPatternsCardinalWithWhiteLine()  
 \li Cedric Forgues (C.S.) 2009/04/23
 Adding of testPatternsCardinalWithWhiteLine()  
 \li Fabrice Gaudier 2010-04-13 16:09:58
 add the test testExportData
 \li Cedric Forgues (C.S.) 2009/04/12
 testDrawWithSelection()  
 \li Cedric Forgues (C.S.) 2009/01/28
 testComputeCorrelationMatrix()
 \li FGA 2009/12/21
 Add in the test testFileDataRead when the ASCII contains more than 1000 attributes (limit in ROOT)
 \li Cedric Forgues (C.S.) 2009/12/15
 testLeafIteratorPositionCreateTuple(), testLeafIteratorPosition_FileWithoutIterator(), testLeafIteratorPosition_FileWithIterator(), testLeafIteratorPosition_Merge2Files()
 \li Cedric Forgues (C.S.) 2009/12/08
 testComputeCorrelationMatrix()
 ********************************************************************* 
 */

// CPPUNIT
#define DBL_PRECISION_COMPARE 1.e-10
#define DBL_PRECISION_COMPARE22 1.e-22

// Uranie
#include "TestTDataServer.h"
#include "TAttribute.h"
#include "TAttributeFormula.h"
#include "UDisableTest.h"
#include "uranie_commands.h"

#include "TLeaf.h"
#include "TFile.h"
#include "TRandom3.h"

#include "TSystem.h"
#include "TPad.h"
#include "TCanvas.h"

#define GEYSERFILENAME "geyser.dat"
#define GEYSERFILENAMEWITHOUTCARRIAGERETURN "geyserWithoutCarriageReturn.dat"
#define GEYSERFILENAMEWITHWHITELINESORCARRIAGERETURNATEOF "geyserWithWhiteLinesOrCarriageReturnAtEof.dat"
#define GEYSERWITHCARRIAGERETURNINSIDEDATA "geyserWithCarriageReturnInsideData.dat"
#define GEYSERWITHWHITELINESINSIDEDATA "geyserWithWhiteLinesInsideData.dat"
#define GEYSERWITHWINDOWSEOLFILENAME "geyserWithWindowsEol.dat"
#define RAOTOUTENBURGFILENAME "raotoutenburg.dat"
#define GEYSERFILENAMENOCOLUMNNAMES "geyserWithColumnTitles.dat"
#define GEYSERFILENAMENOLINEHEADER "geyserNoLineHeader.dat"
#define FLOWRATEUNIFORDESIGN "flowrateUniformDesign.dat"
#define FLOWRATEUNIFORDESIGN_IT "flowrateUniformDesign_withIterator.dat"
#define LARGEFILE "largeFile.dat"
#define TMPEXPORTFILENAME "_export_.dat"
#define STRINGANDVECTOR "myDataStrVector.dat"
#define STRINGANDVECTORSAMENAME "myDataStrVectorSameName.dat"
#define STRINGANDVECTORNOTCOMPLETETYPE "myDataStrVectorNotCompleteType.dat"
#define STRINGANDVECTORMISSINGFIELD "myDataStrVectorMissingField.dat"
#define STRINGANDVECTORTOOMUCHFIELD "myDataStrVectorTooMuchField.dat"
#define STRINGANDVECTOR2 "myDataStrVector2.dat"

// Data Header
#include "youngsmodulus.h"

// ROOT
#include "TString.h"
#include "TEventList.h"

#include <iostream>
using namespace std;

TDataServerTest::TDataServerTest() :
        _bLog(kFALSE)
{
    _workingTestPath = gSystem->pwd();
}

TDataServerTest::~TDataServerTest()
{
}

void TDataServerTest::cleanWorkingDirectories(const char *option)
{
    if (_bLog)
        cout << " ** TDataServerTest::cleanWorkingDirectories option[" << option
	     << "] _workingTestPath[" << _workingTestPath << "]" << endl;

    UranieTests::removeFile(_workingTestPath.Data(), "geyser.h");
    UranieTests::removeFile(_workingTestPath.Data(), "geyser.tex");
    UranieTests::removeFile(_workingTestPath.Data(), "geyser.nemo");
    UranieTests::removeFile(_workingTestPath.Data(), "geyser.root");
    UranieTests::removeFile(_workingTestPath.Data(), "flowrateUniformDesign.root");
    UranieTests::removeFile(_workingTestPath.Data(), "_flowrateUniformDesignMerge_.dat");
    UranieTests::removeFile(_workingTestPath.Data(), "_flowrateUniformDesignMerge_.root");
    UranieTests::removeFile(_workingTestPath.Data(), "fakeFile.tex");
    UranieTests::removeFile(_workingTestPath.Data(), "_export_.dat");
    UranieTests::removeFile(_workingTestPath.Data(), "_export_.root");
    UranieTests::removeFile(_workingTestPath.Data(), "_dataserver_.root");
    UranieTests::removeFile(_workingTestPath.Data(), "pouet.root");
    UranieTests::removeFile(_workingTestPath.Data(), "myDataStrVector.root");
    UranieTests::removeFile(_workingTestPath.Data(), "myDataStrVector2.root");
    UranieTests::removeFile(_workingTestPath.Data(), "myDataStrVectorMissingField.root");
    UranieTests::removeFile(_workingTestPath.Data(), "myDataStrVectorSameName.root");
    UranieTests::removeFile(_workingTestPath.Data(), "myDataStrVectorTooMuchField.root");
    UranieTests::removeFile(_workingTestPath.Data(), "geyserWithWindowsEol.root");
    UranieTests::removeFile(_workingTestPath.Data(), "largeFile.root");
    UranieTests::removeFile(_workingTestPath.Data(), "geyserWithoutCarriageReturn.root");
    UranieTests::removeFile(_workingTestPath.Data(), "geyserWithWhiteLinesOrCarriageReturnAtEof.root");
    UranieTests::removeFile(_workingTestPath.Data(), "geyserWithCarriageReturnInsideData.root");
    UranieTests::removeFile(_workingTestPath.Data(), "geyserWithWhiteLinesInsideData.root");
    UranieTests::removeFile(_workingTestPath.Data(), "raotoutenburg.root");
    UranieTests::removeFile(_workingTestPath.Data(), "BirthDeath.root");
    UranieTests::removeFile(_workingTestPath.Data(), "flowrateUniformDesign_withIterator.root");
    UranieTests::removeFile(_workingTestPath.Data(), "geyserWithColumnTitles.root");
    UranieTests::removeFile(_workingTestPath.Data(), "tds1.root");
    UranieTests::removeFile(_workingTestPath.Data(), "tds2.root");
    UranieTests::removeFile(_workingTestPath.Data(), "stringandvector.h");
    UranieTests::removeFile(_workingTestPath.Data(), "stringandvector.nemo");
    UranieTests::removeFile(_workingTestPath.Data(), "DummyNAME.root");
    UranieTests::removeFile(_workingTestPath.Data(), "DummyNAME2.root");
    UranieTests::removeFile(_workingTestPath.Data(), "shortGeyser.root");
    UranieTests::removeFile(_workingTestPath.Data(), "myDataTimeEvol.root");
    UranieTests::removeFile(_workingTestPath.Data(), "myDataSimpleCorrel.root");
    UranieTests::removeFile(_workingTestPath.Data(), "myDataSimpleNorm.root");
    UranieTests::removeFile(_workingTestPath.Data(), "myDataStrVector2.root");
    UranieTests::removeFile(_workingTestPath.Data(), "TDS_Flowrate_100Zj.root");
    UranieTests::removeFile(_workingTestPath.Data(), "TDS_Flowrate_zaZ.root");
    UranieTests::removeFile(_workingTestPath.Data(), "temporaltree.root");
    UranieTests::removeFile(_workingTestPath.Data(), "levelESobol1000-1.root");
    UranieTests::removeFile(_workingTestPath.Data(), "basic.root");
    // UranieTests::removeFile(_workingTestPath.Data(), "bis_example1.json");
    // UranieTests::removeFile(_workingTestPath.Data(), Form("bis_bis_%s", STRINGANDVECTOR2));
    // UranieTests::removeFile(_workingTestPath.Data(), Form("bis_%s", STRINGANDVECTOR2));
}

void TDataServerTest::setUp()
{
    if (_bLog)
    {
        cout << endl;
        cout
                << "*******************************************************************************"
                << endl;
    }
    gSystem->cd(_workingTestPath);
    cleanWorkingDirectories("setUp");
    if (_bLog)
        cout
                << "*******************************************************************************"
                << endl;

    tds1 = new URANIE::DataServer::TDataServer("tds1", "my title 1");
    tds2 = new URANIE::DataServer::TDataServer();
    tds3 = new URANIE::DataServer::TDataServer("tds3", "the title 3");
    tds4 = new URANIE::DataServer::TDataServer();
    tds5 = new URANIE::DataServer::TDataServer();
    tdsgeyser = new URANIE::DataServer::TDataServer();
    tdsJSON=NULL;
}

void TDataServerTest::tearDown()
{
    if (tds1)
    {
        delete tds1;
        tds1 = NULL;
    }
    
    if (tds2)
    {
        delete tds2;
        tds2 = NULL;
    }
    
    if (tds3)
    {
        delete tds3;
        tds3 = NULL;
    }
    if (tds4)
    {
        delete tds4;
        tds4 = NULL;
    }
    if (tds5)
    {
        delete tds5;
        tds5 = NULL;
    }
    if (tdsgeyser)
    {
        delete tdsgeyser;
        tdsgeyser = NULL;
    }
    if (tdsJSON)
    {
        delete tdsJSON;
        tdsJSON = NULL;
    }

    if (_bLog)
    {
        cout << endl;
        cout
                << "*******************************************************************************"
                << endl;
    }
    gSystem->cd(_workingTestPath);
    cleanWorkingDirectories("tearDown");
    if (_bLog)
        cout
                << "*******************************************************************************"
                << endl;
}

////////////////////////////////
///                          ///                            
///         Tests            ///
///                          ///
////////////////////////////////

void TDataServerTest::testNames()
{
    TString sExpected("tds1"), sDefault("_tds_"), sSet("nameSet");

    tds3->SetName(sSet);

    CPPUNIT_ASSERT(!strcmp(tds1->GetName(), sExpected));
    CPPUNIT_ASSERT(!strcmp(tds2->GetName(), sDefault));
    CPPUNIT_ASSERT(!strcmp(tds3->GetName(), sSet));
}

void TDataServerTest::testNameAttributes()
{
    TString sExpected("x1"), sDefault("_tds_"), sSet("nameSet");

    tds1->fileDataRead(GEYSERFILENAME);
    tds2->fileDataRead(GEYSERFILENAME);

    tds2->getAttribute("x1")->SetName(sSet);

    CPPUNIT_ASSERT(!strcmp(tds1->getAttribute("x1")->GetName(), sExpected));
    CPPUNIT_ASSERT(!strcmp(tds2->getAttribute(sSet)->GetName(), sSet));
}

void TDataServerTest::testTitles()
{
    TString sExpected("my title 1"), sDefault("_title_"), sSet("titleSet");

    tds3->SetTitle(sSet);

    CPPUNIT_ASSERT(!strcmp(tds1->GetTitle(), sExpected));
    CPPUNIT_ASSERT(!strcmp(tds2->GetTitle(), sDefault));
    CPPUNIT_ASSERT(!strcmp(tds3->GetTitle(), sSet));
}

//Add few tests on String and Vectors
void TDataServerTest::testFileDataRead() 
{
    tds1->unsetLog();
    tds1->fileDataRead(GEYSERFILENAME);

    CPPUNIT_ASSERT_EQUAL(2, tds1->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(272, tds1->getNPatterns());
    tds1->computeStatistic("x1:x2");

    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getAttribute("x1")->getMean(),
            3.48778308823529, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getAttribute("x1")->getStd(),
            1.14137125110521, DBL_PRECISION_COMPARE);

    tds2->unsetLog();
    tds2->fileDataRead(GEYSERFILENAMENOCOLUMNNAMES);

    CPPUNIT_ASSERT_EQUAL(2, tds2->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(272, tds2->getNPatterns());

    tds3->unsetLog();
    tds3->fileDataRead(GEYSERWITHWINDOWSEOLFILENAME);
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds3->isAttribute("x2"));

    tds4->unsetLog();
    tds4->fileDataRead(LARGEFILE);
    CPPUNIT_ASSERT_EQUAL(3939, tds4->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(9, tds4->getNPatterns());
    CPPUNIT_ASSERT_DOUBLES_EQUAL( tds4->getValue("tdsDecay__n__iter__",0), 0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( tds4->getValue("in_CF252",0), 2558200, DBL_PRECISION_COMPARE);

    tds5->fileDataRead(STRINGANDVECTOR);
    CPPUNIT_ASSERT_EQUAL(11, (Int_t)(tds5->getTuple()->GetEntries()));
    CPPUNIT_ASSERT_DOUBLES_EQUAL(25, tds5->getValue("xfactor",0,2), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT( ("odd"== tds5->getString("parity",0)) );

    delete tds1;
    tds1=NULL;

    tds1 = new URANIE::DataServer::TDataServer("tds1", "my title 1");
    tds1->fileDataRead(STRINGANDVECTORSAMENAME);
    CPPUNIT_ASSERT( tds1->getDataSpecification()->getVariablesNames().EqualTo("xfactor:parity:parity_1:xfactor_1:size__xfactor_1:tds1__n__iter__"));

}
void TDataServerTest::testExportData()
{
    tds1->fileDataRead(GEYSERFILENAME);

    tds1->unsetLog();
    tds1->exportData(TMPEXPORTFILENAME);
    URANIE::DataServer::TDataServer *tdstmp =
            new URANIE::DataServer::TDataServer();
    tdstmp->fileDataRead(TMPEXPORTFILENAME);
    CPPUNIT_ASSERT_EQUAL(2, tdstmp->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(272, tdstmp->getNPatterns());
    delete tdstmp;
    tdstmp = NULL;

    tds1->unsetLog();
    tds1->exportData(TMPEXPORTFILENAME, "x1");
    tdstmp = new URANIE::DataServer::TDataServer();
    tdstmp->fileDataRead(TMPEXPORTFILENAME);

    CPPUNIT_ASSERT_EQUAL(272, tdstmp->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(1, tdstmp->getNAttributes());
    delete tdstmp;
    tdstmp = NULL;

    tds1->unsetLog();
    tds1->exportData(TMPEXPORTFILENAME, "", "x1<3.20");
    tdstmp = new URANIE::DataServer::TDataServer();
    tdstmp->fileDataRead(TMPEXPORTFILENAME);

    CPPUNIT_ASSERT_EQUAL(98, tdstmp->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(2, tdstmp->getNAttributes());
    delete tdstmp;
    tdstmp = NULL;

    tds1->unsetLog();
    tds1->exportData(TMPEXPORTFILENAME, "x2", "x1<3.20");
    tdstmp = new URANIE::DataServer::TDataServer();
    tdstmp->fileDataRead(TMPEXPORTFILENAME);

    CPPUNIT_ASSERT_EQUAL(98, tdstmp->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(1, tdstmp->getNAttributes());
    delete tdstmp;
    tdstmp = NULL;

    tds1->unsetLog();
    tds1->setCut(TString("x1<3.20"));
    tds1->exportData(TMPEXPORTFILENAME);
    tds1->clearSelect();
    tdstmp = new URANIE::DataServer::TDataServer();
    tdstmp->fileDataRead(TMPEXPORTFILENAME);

    CPPUNIT_ASSERT_EQUAL(174, tdstmp->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(2, tdstmp->getNAttributes());
    delete tdstmp;
    tdstmp = NULL;

    tds1->unsetLog();
    tds1->setSelect(TString("x1<3.20"));
    tds1->exportData(TMPEXPORTFILENAME);
    tds1->clearSelect();
    tdstmp = new URANIE::DataServer::TDataServer();
    tdstmp->fileDataRead(TMPEXPORTFILENAME);

    CPPUNIT_ASSERT_EQUAL(98, tdstmp->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(2, tdstmp->getNAttributes());
    delete tdstmp;
    tdstmp = NULL;

    tds1->unsetLog();
    tds1->setSelect(TString("x1<3.20"));
    tds1->exportData(TMPEXPORTFILENAME, "", "x1>2.0");
    tds1->clearSelect();
    tdstmp = new URANIE::DataServer::TDataServer();
    tdstmp->fileDataRead(TMPEXPORTFILENAME);

    CPPUNIT_ASSERT_EQUAL(43, tdstmp->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(2, tdstmp->getNAttributes());
    delete tdstmp;
    tdstmp = NULL;
    
    tdstmp = new URANIE::DataServer::TDataServer();
    URANIE::DataServer::TDataServer *tdstmp2 =
            new URANIE::DataServer::TDataServer();
    Double_t* Xtmp = new Double_t[4];
    for (Int_t i=0; i < 3; i++)
      {
        tdstmp2->addAttribute(new URANIE::DataServer::TUniformDistribution(Form("x%d",i+1), 0., 1.));
	tdstmp2->getAttribute(Form("x%d",i+1))->delShare();
      }
    Xtmp[0] = 1.0;
    Xtmp[1] = 6.8767122877391517743462e-01; 
    Xtmp[2] = 2.9431307193909306274548e-01;
    Xtmp[3] = 8.9538841337475472137442e-01;
    tdstmp2->createTuple();
    tdstmp2->getTuple()->Fill(Xtmp);
    tdstmp2->exportData(TMPEXPORTFILENAME, "*", "", "precision=22");
    tdstmp->fileDataRead(TMPEXPORTFILENAME);
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdstmp->getValue("x1",0),
            6.8767122877391517743462e-01, DBL_PRECISION_COMPARE22);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdstmp->getValue("x2",0),
            2.9431307193909306274548e-01, DBL_PRECISION_COMPARE22);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdstmp->getValue("x3",0),
            8.9538841337475472137442e-01, DBL_PRECISION_COMPARE22);
    delete tdstmp;
    tdstmp = NULL;
    delete tdstmp2;
    tdstmp2 = NULL;
    delete[] Xtmp;
    
}

void TDataServerTest::testReloadFileVectorAndString()
{
    
    delete tds1; tds1 = new URANIE::DataServer::TDataServer();
    tds1->fileDataRead(STRINGANDVECTOR2);
    //test another filedataread without option
    CPPUNIT_ASSERT_THROW(tds1->fileDataRead(STRINGANDVECTOR2), URANIE::Exceptions::UErrorExceptions);

    //test with reload option
    tds1->fileDataRead(STRINGANDVECTOR2,false,true);	
    tds1->exportData(Form("bis_%s", STRINGANDVECTOR2),"*","","%i") ;
    Int_t nret = UranieTests::areFilesEqual(TString("redump_myDataStrVector2.dat.ref"),
                                            Form("bis_%s", STRINGANDVECTOR2), TString("#DATE:"));
    CPPUNIT_ASSERT_EQUAL(0, nret);

    //Reload another file 
    CPPUNIT_ASSERT_THROW(tds1->fileDataRead(STRINGANDVECTOR,false,true), URANIE::Exceptions::UErrorExceptions);
}

void TDataServerTest::testPreaddFileVectorAndString()
{

    delete tds1; tds1 = new URANIE::DataServer::TDataServer();
    tds1->addAttribute("x"); tds1->addAttribute("y");
    tds1->getAttribute("x")->setUnit("s"); tds1->getAttribute("y")->setUnit("s^{2}");
    //Don't precise the pre-add step
    CPPUNIT_ASSERT_THROW(tds1->fileDataRead(STRINGANDVECTOR2), URANIE::Exceptions::UErrorExceptions);
    //Do it properly
    tds1->fileDataRead(STRINGANDVECTOR2,false,true);
    tds1->exportData(Form("bis_%s", STRINGANDVECTOR2),"*","","%i") ;

    Int_t nret = UranieTests::areFilesEqual(TString("redump_myDataStrVector2.dat.ref"),
                                            Form("bis_%s", STRINGANDVECTOR2), TString("#DATE:"));
    CPPUNIT_ASSERT_EQUAL(0, nret);

    delete tds1; tds1 = new URANIE::DataServer::TDataServer();
    tds1->addAttribute("x"); tds1->addAttribute("toto");
    //Wrong attribute
    CPPUNIT_ASSERT_THROW(tds1->fileDataRead(STRINGANDVECTOR2), URANIE::Exceptions::UErrorExceptions);

    delete tds1; tds1 = new URANIE::DataServer::TDataServer();
    tds1->addAttribute("x"); tds1->addAttribute(new URANIE::DataServer::TAttribute("y",URANIE::DataServer::TAttribute::kVector));
    tds1->getAttribute("y")->delShare();
    //Wrong kind of attribute
    CPPUNIT_ASSERT_THROW(tds1->fileDataRead(STRINGANDVECTOR2), URANIE::Exceptions::UErrorExceptions);
}

void TDataServerTest::testReadAndExportDataWithVectorAndString()
{

  tds1->fileDataRead(STRINGANDVECTOR2);
  tds1->exportData(TMPEXPORTFILENAME);
  TList *tlist;
  
  URANIE::DataServer::TDataServer *tdstmp =  new URANIE::DataServer::TDataServer();
  tdstmp->fileDataRead(TMPEXPORTFILENAME);

  tlist = tdstmp->getListOfAttributes();
  CPPUNIT_ASSERT_EQUAL(5, tlist->GetEntries());
  delete tlist; tlist=NULL;
  tlist = tdstmp->getListOfAttributes("*",false);
  CPPUNIT_ASSERT_EQUAL(8, tlist->GetEntries());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(5, tdstmp->getValue("size__yfactor",9), DBL_PRECISION_COMPARE );
  CPPUNIT_ASSERT_DOUBLES_EQUAL(0, tdstmp->getValue("size__yfactor",5), DBL_PRECISION_COMPARE );

  CPPUNIT_ASSERT( tdstmp->getAttribute("x")->getUnit().EqualTo("s") );
  CPPUNIT_ASSERT( tdstmp->getAttribute("xfactor")->getUnit().EqualTo("") );
  CPPUNIT_ASSERT( tdstmp->getAttribute("y")->getUnit().EqualTo("s^{2}") );

  CPPUNIT_ASSERT_DOUBLES_EQUAL(1, tdstmp->getValue("xfactor",0,0), DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT( ("odd"==tdstmp->getString("parity",0)) );
  
  CPPUNIT_ASSERT_DOUBLES_EQUAL(25, tdstmp->getValue("yfactor", tdstmp->getNPatterns()-1,2), DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT( ("odd"==tdstmp->getString("parity",tdstmp->getNPatterns()-1)) );
  delete tlist; tlist=NULL;
  delete tdstmp; tdstmp = NULL;

  tds1->exportData(TMPEXPORTFILENAME,"parity:yfactor","parity==\"even\"");
  tdstmp =  new URANIE::DataServer::TDataServer(); tdstmp->fileDataRead(TMPEXPORTFILENAME);
  tlist = tdstmp->getListOfAttributes();
  CPPUNIT_ASSERT_EQUAL(2, tlist->GetEntries());
  delete tlist; tlist=NULL;
  tlist = tdstmp->getListOfAttributes("*",false);
  CPPUNIT_ASSERT_EQUAL(4, tlist->GetEntries());
  CPPUNIT_ASSERT_EQUAL(5, tdstmp->getNPatterns());
  delete tlist;tlist=NULL;
  delete tdstmp; tdstmp = NULL;
  
  tds1->exportData(TMPEXPORTFILENAME,"parity:yfactor","yfactor>10");
  tdstmp =  new URANIE::DataServer::TDataServer(); tdstmp->fileDataRead(TMPEXPORTFILENAME);
  CPPUNIT_ASSERT_EQUAL(4, tdstmp->getNPatterns());
  delete tdstmp; tdstmp = NULL;

  tds1->exportData(TMPEXPORTFILENAME,"parity:yfactor","yfactor[2]>10");
  tdstmp =  new URANIE::DataServer::TDataServer(); tdstmp->fileDataRead(TMPEXPORTFILENAME);
  CPPUNIT_ASSERT_EQUAL(2, tdstmp->getNPatterns());
  delete tdstmp; tdstmp = NULL;
 
}

#ifdef WITH_JSONCPP
void TDataServerTest::testFileDataReadJSon()
{
    TString sDataBase = "example1.json";
    tdsJSON = new URANIE::DataServer::TDataServer();
    tdsJSON->unsetLog();
    tdsJSON->fileDataReadJSon(sDataBase.Data());

    CPPUNIT_ASSERT_EQUAL(3, tdsJSON->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(3, tdsJSON->getNAttributes());

    CPPUNIT_ASSERT(tdsJSON->isAttribute("P"));
    CPPUNIT_ASSERT(tdsJSON->isAttribute("u"));
    CPPUNIT_ASSERT(tdsJSON->isAttribute("weight"));
    CPPUNIT_ASSERT(tdsJSON->isAttribute("x3") == kFALSE);
    CPPUNIT_ASSERT(  !strcmp(tdsJSON->getAttribute("P")->GetTitle(), "Pressure"));

}

void TDataServerTest::testExportDataJson()
{
    TString sDataBase = "example1.json";
    tdsJSON = new URANIE::DataServer::TDataServer();
    tdsJSON->unsetLog();
    tdsJSON->fileDataReadJSon(sDataBase.Data());
    tdsJSON->exportDataJSon(Form("bis_%s", sDataBase.Data()));
    delete tdsJSON; tdsJSON=NULL;

    tdsJSON = new URANIE::DataServer::TDataServer();
    tdsJSON->fileDataReadJSon(Form("bis_%s", sDataBase.Data()));

    CPPUNIT_ASSERT_EQUAL(3, tdsJSON->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(3, tdsJSON->getNAttributes());

    CPPUNIT_ASSERT(tdsJSON->isAttribute("u"));
    CPPUNIT_ASSERT(tdsJSON->isAttribute("weight"));
}

void TDataServerTest::testReadAndExportDataJsonMultiType()
{
    tdsJSON = new URANIE::DataServer::TDataServer();
    tdsJSON->fileDataRead(STRINGANDVECTOR2);
    tdsJSON->exportDataJSon(Form("bis_%s", STRINGANDVECTOR2)) ;
    delete tdsJSON; tdsJSON=NULL;

    tdsJSON = new URANIE::DataServer::TDataServer("myTDS2","_title_");
    tdsJSON->fileDataReadJSon(Form("bis_%s", STRINGANDVECTOR2));
    tdsJSON->exportData(Form("bis_bis_%s", STRINGANDVECTOR2),"*","","%i") ;

    Int_t nret = UranieTests::areFilesEqual(TString("redump_myDataStrVector2.dat.ref"),
                                            Form("bis_bis_%s", STRINGANDVECTOR2), TString("#DATE:"));
    CPPUNIT_ASSERT_EQUAL(0, nret);
    
}

void TDataServerTest::testThrowExceptionsInJSon()
{
    
#ifdef WIN32
    URANIE_IGNORED_TEST() ; // No equivalent of sed in windows... Removed for this specific case
#endif
    
    tdsJSON = new URANIE::DataServer::TDataServer();
    tdsJSON->fileDataRead(STRINGANDVECTOR2);
    tdsJSON->exportDataJSon(Form("bad_%s", STRINGANDVECTOR2)) ;
    delete tdsJSON; tdsJSON=NULL; tdsJSON = new URANIE::DataServer::TDataServer();

    gSystem->Exec(" sed 's/short_names/short_name/' bad_myDataStrVector2.dat > BadJsonFile.json ");
    CPPUNIT_ASSERT_THROW(tds1->fileDataReadJSon("BadJsonFile.json"), URANIE::Exceptions::UErrorExceptions);
    delete tdsJSON; tdsJSON=NULL; tdsJSON = new URANIE::DataServer::TDataServer();

    gSystem->Exec(" sed 's/metadata/metadatta/' bad_myDataStrVector2.dat > BadJsonFile.json ");
    CPPUNIT_ASSERT_THROW(tds1->fileDataReadJSon("BadJsonFile.json"), URANIE::Exceptions::UErrorExceptions);
    delete tdsJSON; tdsJSON=NULL; tdsJSON = new URANIE::DataServer::TDataServer();

    gSystem->Exec(" sed 's/items/ittems/' bad_myDataStrVector2.dat > BadJsonFile.json ");
    CPPUNIT_ASSERT_THROW(tds1->fileDataReadJSon("BadJsonFile.json"), URANIE::Exceptions::UErrorExceptions);
    delete tdsJSON; tdsJSON=NULL; tdsJSON = new URANIE::DataServer::TDataServer();

    gSystem->Exec(" sed '/\"x\" : 5/d' bad_myDataStrVector2.dat > BadJsonFile.json ");
    CPPUNIT_ASSERT_THROW(tds1->fileDataReadJSon("BadJsonFile.json"), URANIE::Exceptions::UErrorExceptions);
    
    UranieTests::removeFile(_workingTestPath.Data(), "BadJsonFile.json");
    UranieTests::removeFile(_workingTestPath.Data(), Form("bad_%s", STRINGANDVECTOR2));
}
#endif


void TDataServerTest::testPatternsCardinalWithCarriageReturn()
{
    tds2->fileDataRead(GEYSERFILENAME);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(
            "Nb of Patterns with a file with carriage return", 272,
            tds2->getNPatterns());
}

void TDataServerTest::testPatternsCardinalWithoutCarriageReturn()
{
    tds2->fileDataRead(GEYSERFILENAMEWITHOUTCARRIAGERETURN);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(
            "Nb of Patterns with file without any carriage return", 272,
            tds2->getNPatterns());
}

void TDataServerTest::testPatternsCardinalWithWhiteLineAtEof()
{
    tds2->fileDataRead(GEYSERFILENAMEWITHWHITELINESORCARRIAGERETURNATEOF);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(
            "Nb of Patterns of a file with a white line at the end of the file",
            272, tds2->getNPatterns());
}

void TDataServerTest::testPatternsCardinalWithWhiteLineInsideData()
{
    tds2->fileDataRead(GEYSERWITHCARRIAGERETURNINSIDEDATA);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(
            "Nb of Patterns of a file with a carriage return inside data:", 270,
            tds2->getNPatterns());

    tds4->fileDataRead(GEYSERWITHWHITELINESINSIDEDATA);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(
            "Nb of Patterns of a file with a white line inside data:", 270,
            tds4->getNPatterns());
}

void TDataServerTest::testComputeCorrelationMatrix1()
{

    tds2->fileDataRead(RAOTOUTENBURGFILENAME);
    CPPUNIT_ASSERT_EQUAL(5, tds2->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(10, tds2->getNPatterns());

    // Compute the statistic of all the attributes
    tds2->unsetLog();
    tds2->computeStatistic("*");

    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds2->getAttribute("x1")->getMean(), 6.2,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds2->getAttribute("x1")->getStd(),
            2.93636207273937, DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds2->getAttribute("x2")->getMean(), 13.2,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds2->getAttribute("x2")->getStd(),
            6.64663657632774, DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds2->getAttribute("x3")->getMean(), 3.3,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds2->getAttribute("x3")->getStd(),
            7.84644293082328, DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds2->getAttribute("x4")->getMean(), 23.1,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds2->getAttribute("x4")->getStd(),
            29.4710781011561, DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds2->getAttribute("y")->getMean(), 45.5,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds2->getAttribute("y")->getStd(),
            30.9237413289171, DBL_PRECISION_COMPARE);

    // Compute the correlation matrix
    tds2->unsetLog();
    TMatrixD matcorr = tds2->computeCorrelationMatrix("*");

    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(1, 2), 0.971236462201965,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(1, 3), -0.66840285239337,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(1, 4), 0.651995439261572,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(1, 0), 0.740305928800358,
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(2, 3), -0.597819949799314,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(2, 4), 0.526843638943986,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(2, 0), 0.627618512636776,
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(3, 4), -0.841009960960375,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(3, 0), -0.780072155760794,
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(matcorr(4, 0), 0.977603097437009,
            DBL_PRECISION_COMPARE);


    tds1->fileDataRead("myDataSimpleCorrel.dat");
    TMatrixD matcorr2 = tds1->computeCorrelationMatrix("*");

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1, matcorr2(0,0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1, matcorr2(1,1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1, matcorr2(2,2), DBL_PRECISION_COMPARE);
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0, matcorr2(0,1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.92424733098579, matcorr2(0,2), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0, matcorr2(1,2), DBL_PRECISION_COMPARE);
    
}

void TDataServerTest::testComputeCorrelationMatrix2()
{
    tds2->fileDataRead(RAOTOUTENBURGFILENAME);

    // Compute the statistic of all the attributes
    tds2->unsetLog();
    tds2->computeStatistic("*");

    // Compute the correlation matrix on ranks
    tds2->unsetLog();
    TMatrixD matcorr = tds2->computeCorrelationMatrix("*", "", "rank");

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.915151515151515, matcorr(1, 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.963636363636364, matcorr(1, 2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.660606060606061, matcorr(1, 3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.745454545454545, matcorr(1, 4),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.672727272727273, matcorr(2, 3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.781818181818182, matcorr(2, 4),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.903030303030303, matcorr(2, 0),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.890909090909091, matcorr(3, 4),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.709090909090909, matcorr(3, 0),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.866666666666666, matcorr(4, 0),
            DBL_PRECISION_COMPARE);
}

void TDataServerTest::testComputeCorrelationMatrix3()
{
    tds2->fileDataRead(RAOTOUTENBURGFILENAME);

    // Compute the statistic of all the attributes
    tds2->unsetLog();
    tds2->computeStatistic("*");

    // Compute the correlation matrix on ranks with full var list in entry.
    tds2->unsetLog();
    TMatrixD matcorr = tds2->computeCorrelationMatrix("y:x1:x2:x3:x4", "",
            "rank");

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.915151515151515, matcorr(1, 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.963636363636364, matcorr(1, 2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.660606060606061, matcorr(1, 3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.745454545454545, matcorr(1, 4),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.903030303030303, matcorr(2, 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.672727272727273, matcorr(2, 3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.781818181818182, matcorr(2, 4),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.709090909090909, matcorr(3, 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.890909090909091, matcorr(3, 4),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.866666666666666, matcorr(4, 0),
            DBL_PRECISION_COMPARE);
}

void TDataServerTest::testComputeCorrelationMatrix4()
{
    tds2->fileDataRead(RAOTOUTENBURGFILENAME);

    // Compute the statistic of all the attributes
    tds2->unsetLog();
    tds2->computeStatistic("*");

    // Compute the correlation matrix on ranks with a sub var list in entry. y:x1:x2:x3:x4
    tds2->unsetLog();
    TMatrixD matcorr = tds2->computeCorrelationMatrix("y:x2:x4", "", "rank");

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.781818181818182, matcorr(1, 2),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.903030303030303, matcorr(1, 0),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.866666666666666, matcorr(2, 0),
            DBL_PRECISION_COMPARE);
}

void TDataServerTest::testComputeCAH()
{
    tds2->fileDataRead("BirthDeath.dat");

    Int_t npat = tds2->getNPatterns();

    // Compute the CAH with kWardExact algorithm
    tds2->unsetLog();
    TMatrixD matCAH = tds2->computeCAH("death:birth",
            URANIE::DataServer::TDataServer::kWardExact);

    CPPUNIT_ASSERT_EQUAL(34, npat);
    CPPUNIT_ASSERT_EQUAL(2 * npat - 1, matCAH.GetNrows());
    CPPUNIT_ASSERT_EQUAL(7, matCAH.GetNcols());

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, matCAH(npat, 3), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_EQUAL(21, (Int_t)(matCAH(npat, 5)));
    CPPUNIT_ASSERT_EQUAL(26, (Int_t)(matCAH(npat, 6)));

    CPPUNIT_ASSERT_EQUAL(2, (Int_t)(matCAH(2 * npat - 3, 5)));
    CPPUNIT_ASSERT_EQUAL(6, (Int_t)(matCAH(2 * npat - 3, 6)));

    CPPUNIT_ASSERT_DOUBLES_EQUAL(6.09574748980547, matCAH(2 * npat - 2, 3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_EQUAL(0, (Int_t)(matCAH(2 * npat - 2, 5)));
    CPPUNIT_ASSERT_EQUAL(2, (Int_t)(matCAH(2 * npat - 2, 6)));
    
//     delete tds2;
}

void TDataServerTest::testFillDataServer()
{  
  tds1->addAttribute("E");
    tds1->createTuple();
    CPPUNIT_ASSERT_EQUAL(1, tds1->getNAttributes());

    for (Int_t i = 0; i < youngsmodulus_nPattern; i++)
        tds1->getTuple()->Fill(i, E[i]);
    CPPUNIT_ASSERT_EQUAL( youngsmodulus_nPattern, tds1->getNPatterns());
    tds1->computeStatistic("E");

    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getAttribute("E")->getMean(), 29575.609,
            0.001);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getAttribute("E")->getStd(), 1507.0,
            1.0);

}

void TDataServerTest::testBootstrap()
{
    tds1->fileDataRead(GEYSERFILENAME);
    tds1->bootstrap(-1, "seed=123456");
    TEventList *tel = tds1->getTuple()->GetEventList();
    CPPUNIT_ASSERT_EQUAL(272, tel->GetSize());
    CPPUNIT_ASSERT_EQUAL(159, tds1->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(34, (Int_t)(tel->GetEntry(0)));
    CPPUNIT_ASSERT_EQUAL(140, (Int_t)(tel->GetEntry(1)));
    CPPUNIT_ASSERT_EQUAL(117, (Int_t)(tel->GetEntry(136)));
    CPPUNIT_ASSERT_EQUAL(143, (Int_t)(tel->GetEntry(tel->GetN() - 2)));
    CPPUNIT_ASSERT_EQUAL(265, (Int_t)(tel->GetEntry(tel->GetN() - 1)));

    tds1->computeStatistic("x1:x2");

    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getAttribute("x1")->getMean(),
            3.38332352941177, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getAttribute("x1")->getStd(),
            1.16679613932162, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getAttribute("x2")->getMean(),
            69.5625000000001, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getAttribute("x2")->getStd(),
            13.5868040902728, DBL_PRECISION_COMPARE);

    delete tds1;
    tds1 = NULL;

    tds1 = new URANIE::DataServer::TDataServer("tds1", "my title 1");
    tds1->fileDataRead(GEYSERFILENAME);
    tds1->bootstrap();
    CPPUNIT_ASSERT(
            (155 < tds1->getNPatterns()) && (tds1->getNPatterns() < 190));
}

void TDataServerTest::testfillConstantAttribute()
{
    tds2->fileDataRead(FLOWRATEUNIFORDESIGN);
    CPPUNIT_ASSERT_EQUAL(9, tds2->getNAttributes());
    
    URANIE::DataServer::UEntry *tmpConst = new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::String, "foofoo");
    string pouet = "pouet";
    tmpConst->pushEntry(&pouet,false);
    tds2->fillConstantAttribute("newstring",tmpConst);
    delete tmpConst;

    CPPUNIT_ASSERT_EQUAL(10, tds2->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds2->isAttribute("newstring"));

    tmpConst = new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Vector, "Tutu");
    tmpConst->pushEntry(1.0);
    tmpConst->pushEntry(2.0);
    tmpConst->pushEntry(3.0);
    tmpConst->pushEntry(4.0);
    tds2->fillConstantAttribute("newvector",tmpConst);
    delete tmpConst;

    CPPUNIT_ASSERT_EQUAL(11, tds2->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds2->isAttribute("newvector"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds2->isAttribute( (SIZENAME+"__"+"newvector").c_str() ));
    tds2->computeStatistic( (SIZENAME+"__"+"newvector").c_str() );
    CPPUNIT_ASSERT_DOUBLES_EQUAL(4, tds2->getAttribute( (SIZENAME+"__"+"newvector").c_str() )->getMean(),
            DBL_PRECISION_COMPARE);

    tmpConst = new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Double, "Titi");
    tmpConst->pushEntry(1.0);
    tds2->fillConstantAttribute("newdouble",tmpConst);

    CPPUNIT_ASSERT_EQUAL(12, tds2->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds2->isAttribute("newdouble"));
    tds2->computeStatistic( "newdouble" );
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1, tds2->getAttribute( "newdouble" )->getMean(),
            DBL_PRECISION_COMPARE);
    delete tmpConst;
    
}

void TDataServerTest::testGetValue1()
{
    tds2->fileDataRead(FLOWRATEUNIFORDESIGN);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds2->getValue("rw", 0), 0.05,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds2->getValue("r", 0), 33366.67,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds2->getValue("ystar", 0), 26.18,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds2->getValue("rw", 15), 0.0967,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds2->getValue("r", 15), 16733.33,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds2->getValue("ystar", 15), 97.98,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds2->getValue("rw", 31), 0.15,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds2->getValue("r", 31), 100.00,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds2->getValue("ystar", 31), 157.59,
            DBL_PRECISION_COMPARE);
    
//     delete tds2;
}

void TDataServerTest::testGetValue2()
{
    tds2->fileDataRead(FLOWRATEUNIFORDESIGN_IT);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds2->getValue("rw", 0), 0.05,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds2->getValue("r", 0), 33366.67,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds2->getValue("ystar", 0), 26.18,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds2->getValue("rw", 15), 0.0967,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds2->getValue("r", 15), 16733.33,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds2->getValue("ystar", 15), 97.98,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds2->getValue("rw", 31), 0.15,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds2->getValue("r", 31), 100.00,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds2->getValue("ystar", 31), 157.59,
            DBL_PRECISION_COMPARE);
}

void TDataServerTest::testGetValueRange()
{
    tds2->fileDataRead(FLOWRATEUNIFORDESIGN);
    CPPUNIT_ASSERT_NO_THROW( tds2->getValue("ystar",31) );
    CPPUNIT_ASSERT_THROW( tds2->getValue("ystar",32), URANIE::Exceptions::UErrorExceptions );
    CPPUNIT_ASSERT_NO_THROW( tds2->getValue("ystar",0) );
    CPPUNIT_ASSERT_THROW( tds2->getValue("ystar",-1), URANIE::Exceptions::UErrorExceptions );
}

void TDataServerTest::testLeafIteratorPositionCreateTuple()
{
    tds2->addAttribute("E");
    tds2->createTuple();

    for (Int_t i = 0; i < youngsmodulus_nPattern; i++)
        tds2->getTuple()->Fill(i, E[i]);
    TTree* tupleTDS = (TNtupleD*) tds2->getTuple();
    TObjArray * Leaves = tupleTDS->GetListOfLeaves();
    TBranch* leaf;
    TString slabel;
    leaf = (TBranch *) Leaves->At(0);
    slabel = leaf->GetName();
    CPPUNIT_ASSERT(slabel == "_tds___n__iter__");

}

void TDataServerTest::testLeafIteratorPosition_FileWithoutIterator()
{
    tds2->fileDataRead("flowrateUniformDesign.dat");

    TTree* tupleTDS = (TNtupleD*) tds2->getTuple();
    TObjArray * Leaves = tupleTDS->GetListOfLeaves();
    TBranch* leaf;
    TString slabel;
    leaf = (TBranch *) Leaves->At(0);
    slabel = leaf->GetName();
    CPPUNIT_ASSERT(slabel == "flowrateborehole__n__iter__");
//     delete tds2;
}

void TDataServerTest::testLeafIteratorPosition_FileWithIterator()
{
    tds2->fileDataRead("flowrateUniformDesign_withIterator.dat");
    TTree* tupleTDS = (TNtupleD*) tds2->getTuple();
    TObjArray * Leaves = tupleTDS->GetListOfLeaves();
    TBranch* leaf;
    TString slabel;
    leaf = (TBranch *) Leaves->At(0);
    slabel = leaf->GetName();
    CPPUNIT_ASSERT(slabel == "flowrateborehole__n__iter__");
}

void TDataServerTest::testLeafIteratorPosition_Merge2Files()
{//Changed to keep same test
    tds2->fileDataRead("flowrateUniformDesign_withIterator.dat");
    tds4->fileDataRead("flowrateUniformDesign.dat");
    tds2->merge(tds4, "*", kTRUE);
    tds2->exportData("_flowrateUniformDesignMerge_.dat",
            "rw:r:flowrateborehole__n__iter__:ystar");
    tds5->fileDataRead("_flowrateUniformDesignMerge_.dat");
    TTree* tupleTDS = (TNtupleD*) tds5->getTuple();
    TObjArray * Leaves = tupleTDS->GetListOfLeaves();
    TBranch* leaf;
    TString slabel;
    leaf = (TBranch *) Leaves->At(0);
    slabel = leaf->GetName();
    CPPUNIT_ASSERT(slabel == "flowrateborehole__n__iter__");

}

void TDataServerTest::testdrawMethods()
{
    tdsgeyser->fileDataRead(GEYSERFILENAME);
    tdsgeyser->addAttribute("xnorm", "sqrt(x1*x1+x2*x2)");

    TCanvas *c1=NULL;

    //DrawProfile
    c1 = new TCanvas("c1", "c1", 0, 0, 500, 450);
    tdsgeyser->drawProfile("xnorm:x1");
    delete c1;

    c1 = new TCanvas("c1", "c1", 0, 0, 500, 450);
    tdsgeyser->drawProfile("xnorm:x1", "", "samenclass=50");
    delete c1;

    //DrawBoxPlot
    c1 = new TCanvas("c1", "c1", 0, 0, 500, 450);
    tdsgeyser->drawBoxPlot("xnorm", "", "");
    delete c1;

    //DrawPairs
    c1 = new TCanvas("c1", "c1", 0, 0, 500, 450);
    tdsgeyser->drawPairs();
    delete c1;

    //DrawCDF
    c1 = new TCanvas("c1", "c1", 0, 0, 500, 450);
    tdsgeyser->drawCDF("xnorm", "", "");
    delete c1;

    c1 = new TCanvas("c1", "c1", 0, 0, 500, 450);
    tdsgeyser->drawCDF("xnorm", "", "ccdfhist");
    delete c1;

    //DrawTufte
    c1 = new TCanvas("c1", "c1", 0, 0, 500, 450);
    tdsgeyser->drawTufte("xnorm:x1");
    delete c1;

    c1 = new TCanvas("c1", "c1", 0, 0, 500, 450);
    tdsgeyser->drawTufte("xnorm:x1", "", "optstatscatter");
    delete c1;
    
    //DrawScatterPlot
    c1 = new TCanvas("c1", "c1", 0, 0, 500, 450);
    tdsgeyser->drawScatterplot("xnorm:x1");
    delete c1;

    //DrawECDF
    c1 = new TCanvas("c1", "c1", 0, 0, 500, 450);
    tdsgeyser->drawECDF("x1");
    delete c1;
    
    //DrawWithNclass
    c1 = new TCanvas("c1", "c1", 0, 0, 500, 450);
    tdsgeyser->draw("x1","","nclass=root");
    tdsgeyser->draw("x1","","nclass=sturges");
    tdsgeyser->draw("x1","","nclass=fd");
    tdsgeyser->draw("x1","","nclass=scott");
    tdsgeyser->draw("x1","","nclass=32");
    delete c1;
    
    //DrawCobWeb
    c1 = new TCanvas("c1", "c1", 0, 0, 500, 450);
    tdsgeyser->drawCobWeb("xnorm:x1");
    delete c1;
}

void TDataServerTest::testdrawMethodsNoCanvases()
{
    tdsgeyser->fileDataRead(GEYSERFILENAME);
    tdsgeyser->addAttribute("xnorm", "sqrt(x1*x1+x2*x2)");

    //DrawProfile
    tdsgeyser->drawProfile("xnorm:x1");
    tdsgeyser->drawProfile("xnorm:x1", "", "samenclass=50");

    //DrawBoxPlot
    tdsgeyser->drawBoxPlot("xnorm", "", "");

    //DrawPairs
    tdsgeyser->drawPairs();
    
    //DrawCDF
    tdsgeyser->drawCDF("xnorm", "", "");
    tdsgeyser->drawCDF("xnorm", "", "ccdfhist");

    //DrawTufte
    tdsgeyser->drawTufte("xnorm:x1");
    tdsgeyser->drawTufte("xnorm:x1", "", "optstatscatter");

    //DrawScatterPlot
    tdsgeyser->drawScatterplot("xnorm:x1");

    //DrawECDF
    tdsgeyser->drawECDF("x1");
    
    //DrawWithNclass
    tdsgeyser->draw("x1","","nclass=root");
    tdsgeyser->draw("x1","","nclass=sturges");
    tdsgeyser->draw("x1","","nclass=fd");
    tdsgeyser->draw("x1","","nclass=scott");
    tdsgeyser->draw("x1","","nclass=32");

    //DrawCobWeb
    tdsgeyser->drawCobWeb("xnorm:x1");    
}

void TDataServerTest::testDrawWithSelection()
{
    URANIE::DataServer::TDataServer *tds =
            new URANIE::DataServer::TDataServer();
    tds->fileDataRead("flowrateUniformDesign.dat");

    tds->draw("rw:r", "", "goff");
    Int_t nvisu = (Int_t) tds->GetTuple()->GetSelectedRows();
    CPPUNIT_ASSERT_EQUAL(nvisu, 32);

    tds->draw("rw:r", "rw<0.1", "goff");
    nvisu = (Int_t) tds->GetTuple()->GetSelectedRows();
    CPPUNIT_ASSERT_EQUAL(nvisu, 16);

    tds->setSelect("rw>0.08");
    tds->draw("rw:r", "", "goff");
    nvisu = (Int_t) tds->GetTuple()->GetSelectedRows();
    CPPUNIT_ASSERT_EQUAL(nvisu, 22);

    tds->clearSelect();
    tds->setSelect("rw>0.08");
    tds->draw("rw:r", "rw<0.12", "goff");
    nvisu = (Int_t) tds->GetTuple()->GetSelectedRows();
    CPPUNIT_ASSERT_EQUAL(nvisu, 12);

    tds->clearSelect();
    tds->setSelect("");
    tds->draw("rw:r", "rw<0.12", "goff");
    nvisu = (Int_t) tds->GetTuple()->GetSelectedRows();
    CPPUNIT_ASSERT_EQUAL(nvisu, 22);

    tds->clearSelect();
    tds->setSelect("");
    tds->draw("rw:r", "rw<0.12 && rw>0.08", "goff");
    nvisu = (Int_t) tds->GetTuple()->GetSelectedRows();
    CPPUNIT_ASSERT_EQUAL(nvisu, 12);

    delete tds;
}

void TDataServerTest::testGetNPatterns()
{
    CPPUNIT_ASSERT_EQUAL(0, tds1->getNPatterns());
    tds1->fileDataRead(GEYSERFILENAME);
    CPPUNIT_ASSERT_EQUAL(272, tds1->getNPatterns());
    tds1->createTuple();
    CPPUNIT_ASSERT_EQUAL(0, tds1->getNPatterns());
}

void TDataServerTest::testGetIteratorName()
{
    CPPUNIT_ASSERT(tds2->getIteratorName() == TString("_tds___n__iter__"));
    URANIE::DataServer::TDataServer * tdsIteratorname =
            new URANIE::DataServer::TDataServer("toto", "the toto title");
    CPPUNIT_ASSERT(
            tdsIteratorname->getIteratorName() == TString("toto__n__iter__"));
    delete tdsIteratorname;
    tds2->fileDataRead(FLOWRATEUNIFORDESIGN);
    CPPUNIT_ASSERT(
            tds2->getIteratorName() == TString("flowrateborehole__n__iter__"));
    tds4->fileDataRead(GEYSERFILENAME);
    CPPUNIT_ASSERT(tds4->getIteratorName() == TString("geyser__n__iter__"));
}

void TDataServerTest::testGetDataTreeName()
{
    CPPUNIT_ASSERT(tds2->getDataTreeName() == TString("_tds___data__tree__"));
    URANIE::DataServer::TDataServer * tdsDataTreeName =
            new URANIE::DataServer::TDataServer("tdsName", "the tdsName title");
    CPPUNIT_ASSERT(
            tdsDataTreeName->getDataTreeName()
                    == TString("tdsName__data__tree__"));
    delete tdsDataTreeName;
    tdsDataTreeName = NULL;
    tds2->fileDataRead(FLOWRATEUNIFORDESIGN);
    CPPUNIT_ASSERT(
            tds2->getDataTreeName()
                    == TString("flowrateborehole__data__tree__"));
    tds4->fileDataRead(GEYSERFILENAME);
    CPPUNIT_ASSERT(tds4->getDataTreeName() == TString("geyser__data__tree__"));
}

void TDataServerTest::testGetArchiveFileName()
{
  CPPUNIT_ASSERT(!strcmp(tds1->getArchiveFileName(), "_dataserver_.root"));
  tds2->fileDataRead(FLOWRATEUNIFORDESIGN);

  char *name = gSystem->ConcatFileName(gSystem->pwd(), "flowrateUniformDesign.root");
  CPPUNIT_ASSERT( !strcmp(tds2->getArchiveFileName(), name) );

  delete[] name;

}

void TDataServerTest::testGetOrigin()
{

  CPPUNIT_ASSERT(
		 tds1->getOrigin() == URANIE::DataServer::TDataServer::kUnknown);
  tds1->fileDataRead(FLOWRATEUNIFORDESIGN);
  CPPUNIT_ASSERT(
		 tds1->getOrigin() == URANIE::DataServer::TDataServer::kASCIIFile);

  tds2->ntupleDataRead("flowrateUniformDesign.root", "tds1__data__tree__",
		       "rw");
  CPPUNIT_ASSERT(tds2->getOrigin() == URANIE::DataServer::TDataServer::kTree);
}


void TDataServerTest::testGetListOfAttributes()
{

    tds1->fileDataRead(GEYSERFILENAME);
    TList *tlist = tds1->getListOfAttributes();
    CPPUNIT_ASSERT_EQUAL(2, tlist->GetEntries());

    tds2->fileDataRead(GEYSERFILENAMENOCOLUMNNAMES);
    delete tlist; tlist=tds2->getListOfAttributes();
    CPPUNIT_ASSERT_EQUAL(2, tlist->GetEntries());

    tds3->fileDataRead(STRINGANDVECTOR);
    delete tlist; tlist=tds3->getListOfAttributes();
    CPPUNIT_ASSERT_EQUAL(4, tlist->GetEntries());

    delete tlist; tlist=tds3->getListOfAttributes("*",false);
    CPPUNIT_ASSERT_EQUAL(6, tlist->GetEntries());
    delete tlist; tlist=NULL;
}


void TDataServerTest::testGetNAttributes()
{
    CPPUNIT_ASSERT_EQUAL(0, tds1->getNAttributes());

    tds1->fileDataRead(GEYSERFILENAME);
    CPPUNIT_ASSERT_EQUAL(2, tds1->getNAttributes());

    tds2->fileDataRead(GEYSERFILENAMENOCOLUMNNAMES);
    CPPUNIT_ASSERT_EQUAL(2, tds2->getNAttributes());

    tds3->fileDataRead(STRINGANDVECTOR);
    CPPUNIT_ASSERT_EQUAL(4, tds3->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(6, tds3->getNAttributes(false));
}

void TDataServerTest::testGetInputOutputAttributes()
{
    tds1->fileDataRead(FLOWRATEUNIFORDESIGN);
    CPPUNIT_ASSERT_EQUAL(9, tds1->getNInputAttributes());
    CPPUNIT_ASSERT_EQUAL(9, tds1->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(0, tds1->getNOutputAttributes());
    tds1->getAttribute("ystar")->setOutput();
    CPPUNIT_ASSERT_EQUAL(8, tds1->getNInputAttributes());
    CPPUNIT_ASSERT_EQUAL(1, tds1->getNOutputAttributes());
}

void TDataServerTest::testGetDataSpecification()
{
    CPPUNIT_ASSERT(tds1->getDataSpecification() == NULL);
    tds1->fileDataRead(FLOWRATEUNIFORDESIGN);
    CPPUNIT_ASSERT(
            tds1->getDataSpecification()->GetName()
                    == TString("uheader__tds1"));

    tds2->fileDataRead(FLOWRATEUNIFORDESIGN);
    CPPUNIT_ASSERT(
            tds2->getDataSpecification()->GetName()
                    == TString("uheader__flowrateborehole"));

}

void TDataServerTest::testGetCut()
{
    tds1->fileDataRead(GEYSERFILENAME);
    tds1->setCut(TCut("x1<=1.817"));
    CPPUNIT_ASSERT(tds1->getNPatterns() == 253);
    tds1->clearCut();
    tds1->setCut(TString("x1<=1.817"));
    CPPUNIT_ASSERT(tds1->getNPatterns() == 253);

    tds1->clearCut();
    CPPUNIT_ASSERT(TCut("") == tds1->getCut());
    CPPUNIT_ASSERT(272 == tds1->getNPatterns());
}

void TDataServerTest::testSelect()
{
    tds1->fileDataRead(GEYSERFILENAME);

    tds1->setSelect("x1 < 3.");
    CPPUNIT_ASSERT_EQUAL(97, tds1->getNPatterns());

    tds1->clearSelect();
    tds1->setSelect("!(x1 < 3.)");
    CPPUNIT_ASSERT_EQUAL(175, (int) tds1->getTuple()->GetEntries("!(x1 < 3.)"));

    tds1->clearSelect();
    CPPUNIT_ASSERT_EQUAL(272, tds1->getNPatterns());
}

void TDataServerTest::testNtupleDataRead()
{
    tds1->fileDataRead(FLOWRATEUNIFORDESIGN);

    tds2->unsetLog();
    tds2->ntupleDataRead("flowrateUniformDesign.root", "tds1__data__tree__");
    CPPUNIT_ASSERT(tds2->getNPatterns() == 32);

    tds4->unsetLog();
    tds4->ntupleDataRead("flowrateUniformDesign.root", "tds1__data__tree__",
            "rw", "rw<0.09");
    CPPUNIT_ASSERT(tds4->getNPatterns() == 20);

    tds1->unsetLog();
    tds1->setCut(TString("rw<0.09"));
    CPPUNIT_ASSERT(tds1->getNPatterns() == 20);
}

void TDataServerTest::testNtupleDataReadStringAndVector()
{
  tds3->fileDataRead(STRINGANDVECTOR);
  tds1->ntupleDataRead("myDataStrVector.root","tds3__data__tree__");
  
  CPPUNIT_ASSERT_EQUAL(4, tds1->getNAttributes());
  CPPUNIT_ASSERT_EQUAL(6, tds1->getNAttributes(false));
  CPPUNIT_ASSERT_EQUAL(11, tds1->getNPatterns());

  vector<double> vec = tds1->getVector("xfactor",9);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(8, vec.at(3), DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(5,vec.size(), DBL_PRECISION_COMPARE);

  tds2->ntupleDataRead("myDataStrVector.root","tds3__data__tree__","parity:xfactor","parity==\"odd\"");
  CPPUNIT_ASSERT_EQUAL(2, tds2->getNAttributes());
  CPPUNIT_ASSERT_EQUAL(4, tds2->getNAttributes(false));
  CPPUNIT_ASSERT_EQUAL(5, tds2->getNPatterns());
  
}

void TDataServerTest::testLoadTree()
{
    cout << endl;
    gROOT->ProcessLine(".x basic.C");
    TNtuple * tuple = dynamic_cast<TNtuple *>(gROOT->FindObject("ntuple"));

    tds2->unsetLog();
    tds2->loadTree(tuple);
    CPPUNIT_ASSERT(tds2->getNPatterns() == 1000);
    CPPUNIT_ASSERT(
            tds2->getDataSpecification()->getVariablesNames()
                    == TString("x:y:z:ntuple__n__iter__"));
}

void TDataServerTest::testExportDataHeader()
{
    tds2->fileDataRead(GEYSERFILENAME);
    tds2->exportDataHeader("geyser.h");
    CPPUNIT_ASSERT(gSystem->AccessPathName("geyser.h", kFileExists) == 0);

    tds1->fileDataRead(STRINGANDVECTOR);
    tds1->exportDataHeader("stringandvector.h");
    CPPUNIT_ASSERT(gSystem->AccessPathName("stringandvector.h", kFileExists) == 0);
}

void TDataServerTest::testExportDataNeMo()
{
    tds2->fileDataRead(GEYSERFILENAME);
    tds2->exportDataNeMo("geyser.nemo", "x1", "x2");
    CPPUNIT_ASSERT(gSystem->AccessPathName("geyser.nemo", kFileExists) == 0);

    tds1->fileDataRead(STRINGANDVECTOR);
    tds1->exportDataNeMo("stringandvector.nemo", "x:parity:xfactor", "y");
    CPPUNIT_ASSERT(gSystem->AccessPathName("stringandvector.nemo", kFileExists) == 0);

}

void TDataServerTest::testExportDataContext()
{
    tds2->fileDataRead(GEYSERFILENAME);

    tds2->exportDataContext("geyser.tex");
    CPPUNIT_ASSERT(gSystem->AccessPathName("geyser.tex", kFileExists) == 0);

    Int_t nret = -1;
#ifdef WIN32
    nret = gSystem->Exec("fc geyser.tex geyser1.tex.ref");
    CPPUNIT_ASSERT_EQUAL(1, nret);
#else
    nret =
            gSystem->Exec(
                    "diff --ignore-matching-lines='%% DateTime' --ignore-matching-lines='%% File' geyser.tex geyser1.tex.ref");
    CPPUNIT_ASSERT_EQUAL(0, nret);
#endif

    tds2->exportDataContext("geyser.tex", "x2", "x1<3.2");
    CPPUNIT_ASSERT(gSystem->AccessPathName("geyser.tex", kFileExists) == 0);

    nret = -1;
#ifdef WIN32
    nret = gSystem->Exec("fc geyser.tex geyser2.tex.ref");
    CPPUNIT_ASSERT_EQUAL(1, nret);
#else
    nret =
            gSystem->Exec(
                    "diff --ignore-matching-lines='%% DateTime' --ignore-matching-lines='%% File' geyser.tex geyser2.tex.ref");
    CPPUNIT_ASSERT_EQUAL(0, nret);
#endif
}

void TDataServerTest::testAddAndDelAttributes()
{
    TString sExpected("x1"), sSet("x2"), sTitleSet("#Delta P"), sUnitySet(
            "bar");

    tds1->fileDataRead(GEYSERFILENAME);
    tds2->fileDataRead(GEYSERFILENAME);
    tds3->fileDataRead(GEYSERFILENAME);

    Int_t natt = tds1->getNAttributes();
    CPPUNIT_ASSERT_EQUAL(natt, tds1->getNAttributes());
    tds1->delAttribute(sExpected);
    CPPUNIT_ASSERT_EQUAL(natt - 1, tds1->getNAttributes());

    tds2->delAttribute(sSet);
    CPPUNIT_ASSERT_EQUAL(natt - 1, tds2->getNAttributes());
    CPPUNIT_ASSERT(tds2->isAttribute(sSet) == kFALSE);
    CPPUNIT_ASSERT(tds2->isAttribute(sExpected));

    tds3->addAttribute("x3", "x1*x2");
    tds3->addAttribute("x4", "x2*x2", sTitleSet);
    tds3->addAttribute("x5", "x2*x2*x2", sTitleSet, sUnitySet);
    CPPUNIT_ASSERT(tds3->isAttribute("x3"));

    CPPUNIT_ASSERT_DOUBLES_EQUAL(284.4, tds3->getValue("x3", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(97.2, tds3->getValue("x3", 1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(246.642, tds3->getValue("x3", 2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT(tds3->isAttribute("x4"));
    CPPUNIT_ASSERT(tds3->isAttribute("x5"));
    CPPUNIT_ASSERT_EQUAL(5, tds3->getNAttributes());
    CPPUNIT_ASSERT(!strcmp(tds3->getAttribute("x4")->GetTitle(), sTitleSet));
    CPPUNIT_ASSERT(!strcmp(tds3->getAttribute("x5")->getUnity(), sUnitySet));
}

void TDataServerTest::testAddAttributeWithData()
{
    tds1->fileDataRead(GEYSERFILENAME);

    Int_t nP = tds1->getNPatterns();
    vector<double> unite;
    for(int i=0; i<nP; i++)
        unite.push_back(i);

    CPPUNIT_ASSERT(tds1->isAttribute("unite") == kFALSE);
    tds1->addAttributeUsingData("unite",&unite[0],unite.size());
    CPPUNIT_ASSERT(tds1->isAttribute("unite") == kTRUE);

    tds1->getTuple()->Draw("x1","","goff");
    double *test=tds1->getTuple()->GetV1();
    
    CPPUNIT_ASSERT(tds1->isAttribute("copyx1") == kFALSE);
    tds1->addAttributeUsingData("copyx1", test, nP);
    CPPUNIT_ASSERT(tds1->isAttribute("copyx1") == kTRUE);

    tds1->addAttribute("diff","x1-copyx1");
    tds1->computeStatistic("diff");
       
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tds1->getAttribute("diff")->getMean(),
                                 DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tds1->getAttribute("diff")->getStd(),
                                 DBL_PRECISION_COMPARE);

    delete tds2;
    tds2 = new URANIE::DataServer::TDataServer("tds2","foo");

    tds1->getTuple()->Draw("x1","","goff");
    test=tds1->getTuple()->GetV1();

    tds2->addAttributeUsingData("copyx1", test, nP);
    tds2->merge(tds1, "x1", kFALSE);
    
    tds2->addAttribute("diff","x1-copyx1");
    tds2->computeStatistic("diff");
       
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tds2->getAttribute("diff")->getMean(),
                                 DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tds2->getAttribute("diff")->getStd(),
                                 DBL_PRECISION_COMPARE);
    
}

void TDataServerTest::testAddAndDelAttributesWithVectors()
{
  tds1->fileDataRead(STRINGANDVECTOR2);

  CPPUNIT_ASSERT_EQUAL(5, tds1->getNAttributes());
  CPPUNIT_ASSERT_EQUAL(8, tds1->getNAttributes(false));
  tds1->delAttribute("size__yfactor");
  CPPUNIT_ASSERT_EQUAL(5, tds1->getNAttributes());
  CPPUNIT_ASSERT_EQUAL(7, tds1->getNAttributes(false));
  tds1->delAttribute("y");
  CPPUNIT_ASSERT_EQUAL(4, tds1->getNAttributes());
  CPPUNIT_ASSERT_EQUAL(6, tds1->getNAttributes(false));

  tds1->addAttribute("produit","x*xfactor");
  tds1->addAttribute("sum","x+xfactor");
  tds1->addAttribute("produitv","yfactor*xfactor");
  tds1->addAttribute("sumv","yfactor+xfactor");

  CPPUNIT_ASSERT( tds1->getValue("size__produit",0)==2 );
  CPPUNIT_ASSERT( tds1->getValue("size__sum",0)==2 );
  CPPUNIT_ASSERT( tds1->getValue("size__produitv",0)==2 );
  CPPUNIT_ASSERT( tds1->getValue("size__sumv",0)==2 );

  CPPUNIT_ASSERT_DOUBLES_EQUAL(-5, tds1->getValue("produit",0,0), DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(-4, tds1->getValue("sum",0,0), DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(1, tds1->getValue("produitv",0,0), DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(2, tds1->getValue("sumv",0,0), DBL_PRECISION_COMPARE);

  CPPUNIT_ASSERT_DOUBLES_EQUAL(25, tds1->getValue("produit",tds1->getNPatterns()-1,1), DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(10, tds1->getValue("sum",tds1->getNPatterns()-1,1), DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(25, tds1->getValue("produitv",tds1->getNPatterns()-1,1), DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(10, tds1->getValue("sumv",tds1->getNPatterns()-1,1), DBL_PRECISION_COMPARE);
}

void TDataServerTest::testFillAttribute()
{
    tds1->fileDataRead(GEYSERFILENAME);

    tds1->addAttribute("x3");
    tds1->fillAttribute("x3", "x1*x2");

    CPPUNIT_ASSERT(tds1->isAttribute("x3"));
    CPPUNIT_ASSERT_DOUBLES_EQUAL(284.4, tds1->getValue("x3", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(97.2, tds1->getValue("x3", 1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(246.642, tds1->getValue("x3", 2),
            DBL_PRECISION_COMPARE);

}

void TDataServerTest::testGetAttributeIndex()
{
    tds1->fileDataRead(GEYSERFILENAME);
    CPPUNIT_ASSERT_EQUAL(tds1->getAttributeIndex("x1"), 0);
    CPPUNIT_ASSERT_EQUAL(tds1->getAttributeIndex("x2"), 1);
    CPPUNIT_ASSERT_EQUAL(tds1->getAttributeIndex(tds1->getAttribute("x1")), 0);
    CPPUNIT_ASSERT_EQUAL(tds1->getAttributeIndex(tds1->getAttribute("x2")), 1);
}

void TDataServerTest::testGetAttributeFromIndex()
{
    tds1->fileDataRead("flowrateUniformDesign.dat");
    CPPUNIT_ASSERT(TString("rw") == tds1->getAttribute(0)->GetName());
    CPPUNIT_ASSERT(TString("hl") == tds1->getAttribute(5)->GetName());
}

void TDataServerTest::testMerge()
{
    tds2->fileDataRead("tds1.dat");
    tds4->fileDataRead("tds2.dat");
    tds2->merge(tds4, "*", kTRUE);
    CPPUNIT_ASSERT(
            tds2->getDataSpecification()->getVariablesNames()
                    == TString("x:dy:z:theta:tds1_1__n__iter__:x2:y:u:ua"));
    tds1->fileDataRead(GEYSERFILENAME);
    tds3->fileDataRead(GEYSERFILENAME);
    tds1->merge(tds3, "*", kTRUE);
    CPPUNIT_ASSERT(
            tds1->getDataSpecification()->getVariablesNames()
                    == TString("x1:x2:tds1__n__iter__:tds3.x1:tds3.x2"));
}

void TDataServerTest::testMergeWithVectors()
{
  tds1->fileDataRead(STRINGANDVECTOR);
  tds3->fileDataRead(STRINGANDVECTOR2);
  tds1->merge(tds3);
//  CPPUNIT_ASSERT( tds1->getDataSpecification()->getVariablesNames().EqualTo("x:y:parity:xfactor:size__xfactor:tds1__n__iter__:tds3.x:tds3.xfactor:tds3.size__xfactor:tds3.parity:tds3.y:yfactor:size__yfactor"));
  CPPUNIT_ASSERT( tds1->getDataSpecification()->isComparableToTDSP( TString("x(D):y(D):parity(S):xfactor(V):size__xfactor(D):tds3.x(D):tds3.xfactor(V):tds3.size__xfactor(D):tds3.parity(S):tds3.y(D):yfactor(V):size__yfactor(D)") ) );
  CPPUNIT_ASSERT_EQUAL(9, tds1->getNAttributes());
  CPPUNIT_ASSERT_EQUAL(13, tds1->getNAttributes(false));

  tds2->fileDataRead(STRINGANDVECTOR);
  tds2->merge(tds3, "yfactor");
  CPPUNIT_ASSERT( tds2->getDataSpecification()->getVariablesNames().EqualTo("x:y:parity:xfactor:size__xfactor:myTDS__n__iter__:yfactor:size__yfactor"));
  CPPUNIT_ASSERT_EQUAL(5, tds2->getNAttributes());
  CPPUNIT_ASSERT_EQUAL(8, tds2->getNAttributes(false));
}


void TDataServerTest::testMergeWithIterators()
{
  tds4->fileDataRead("DummyNAME.dat");
  tds5->fileDataRead("DummyNAME2.dat");
  tds4->merge(tds5);

  for(int i=0; i<tds4->getNPatterns(); i++)
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds4->getValue("x",i),tds4->getValue("cho_1.x",i), DBL_PRECISION_COMPARE);

}

void TDataServerTest::testIsAttribute()
{
    tds2->fileDataRead(GEYSERFILENAME);
    CPPUNIT_ASSERT(tds2->isAttribute("x1") == kTRUE);
    CPPUNIT_ASSERT(tds2->isAttribute("z") == kFALSE);
    tds2->delAttribute("x2");
    CPPUNIT_ASSERT(tds2->isAttribute("x2") == kFALSE);
}

void TDataServerTest::testSetIterator()
{
    tds1->fileDataRead("flowrateUniformDesign.dat");
    tds1->setIterator("tds1__n__iter__");
    CPPUNIT_ASSERT(
            tds1->getAttribute("tds1__n__iter__")->getOrigin()
                    == URANIE::DataServer::TAttribute::kIterator);
}

void TDataServerTest::testCreateTuple()
{
    tds2->fileDataRead(GEYSERFILENAME);
    tds2->createTuple();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0, tds2->getNPatterns(),
            DBL_PRECISION_COMPARE);
}

void TDataServerTest::testVariousNormalisation()
{
  tds2->fileDataRead("myDataSimpleNorm.dat");
  bool gl=false;
  string type[2]={"NG","G"};
  for(unsigned int i=0; i<2; i++)
    {
      tds2->normalize("rank",("_Z"+type[i]).c_str(), URANIE::DataServer::TDataServer::kZeroOne,gl);
      tds2->normalize("rank",("_CR"+type[i]).c_str(), URANIE::DataServer::TDataServer::kCR,gl); 
      tds2->normalize("rank",("_CE"+type[i]).c_str(), URANIE::DataServer::TDataServer::kCentered,gl);
      tds2->normalize("rank",("_M"+type[i]).c_str(), URANIE::DataServer::TDataServer::kMinusOneOne,gl);	
      gl=true;
    }

  string t[8]={"ZNG","CRNG","CENG","MNG","ZG","CRG","CEG","MG"};
  double v[8]={0.25,-0.632455532067352,-2,-0.5,0.333333333333,-0.495433752250095,-1.5,-0.33333333333};
  for(unsigned int it=0; it<8; it++)
    CPPUNIT_ASSERT_DOUBLES_EQUAL(v[it], tds2->getValue(("rank_"+t[it]).c_str(),1,1), DBL_PRECISION_COMPARE);
}

void TDataServerTest::testComputeStatistic()
{
    tds1->fileDataRead("flowrateUniformDesign.dat");
    tds1->computeStatistic("rw");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.1, tds1->getAttribute("rw")->getMean(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0312212706649727,
            tds1->getAttribute("rw")->getStd(), DBL_PRECISION_COMPARE);

    tds1->computeStatistic("rw", "rw<0.1");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0733375, tds1->getAttribute("rw")->getMean(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0157790525697838,
            tds1->getAttribute("rw")->getStd(), DBL_PRECISION_COMPARE);

    tds1->computeStatistic("r:hu:l");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(25050., tds1->getAttribute("r")->getMean(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(18894.1977765683,
            tds1->getAttribute("r")->getStd(), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1050., tds1->getAttribute("hu")->getMean(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(39.9077277664849,
            tds1->getAttribute("hu")->getStd(), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1400., tds1->getAttribute("l")->getMean(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(186.236062479161,
            tds1->getAttribute("l")->getStd(), DBL_PRECISION_COMPARE);

    tds1->computeStatistic("r:hu:l", "rw<0.1");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(22970.833125,
            tds1->getAttribute("r")->getMean(), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(18093.9281541158,
            tds1->getAttribute("r")->getStd(), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1059.641875,
            tds1->getAttribute("hu")->getMean(), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(36.7507518687259,
            tds1->getAttribute("hu")->getStd(), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1390., tds1->getAttribute("l")->getMean(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(204.222101317822,
            tds1->getAttribute("l")->getStd(), DBL_PRECISION_COMPARE);

    tds2->fileDataRead("myDataSimpleNorm.dat");
    tds2->computeStatistic("x:rank");
    double refMean[3]={4,5,4.5}, refStd[3]={3.1622776601684, 3.1622776601684, 3.0276503540975};
    double refMinimum[]={0,1,0}, refMaximum[3]={8,9,9}; 
    for(unsigned int ival=0; ival<3; ival++)
       {
	 CPPUNIT_ASSERT_DOUBLES_EQUAL(refMean[ival], tds2->getAttribute("rank")->getMean(ival),
				      DBL_PRECISION_COMPARE);
	 CPPUNIT_ASSERT_DOUBLES_EQUAL(refStd[ival], tds2->getAttribute("rank")->getStd(ival),
				      DBL_PRECISION_COMPARE);
	 CPPUNIT_ASSERT_DOUBLES_EQUAL(refMinimum[ival], tds2->getAttribute("rank")->getMinimum(ival),
				      DBL_PRECISION_COMPARE);
	 CPPUNIT_ASSERT_DOUBLES_EQUAL(refMaximum[ival], tds2->getAttribute("rank")->getMaximum(ival),
				      DBL_PRECISION_COMPARE);
       }
    
    // TDS with min and max values of different sign
    tds3->fileDataRead("problem2Classes_001000.dat");
    tds3->computeStatistic();
    CPPUNIT_ASSERT_EQUAL( 12, tds3->getNAttributes());
    CPPUNIT_ASSERT_EQUAL( 1000, tds3->getNPatterns());

    Double_t dminR, dminU;
    Double_t dmaxR, dmaxU;
    TString sx;

    for(Int_t i=0; i<tds3->getNAttributes(); i++) {
        sx = tds3->getAttribute(i)->GetName();
        dminR = tds3->getTuple()->GetMinimum(sx);
        dmaxR = tds3->getTuple()->GetMaximum(sx);
        dminU = tds3->getAttribute(i)->getMinimum();
        dmaxU = tds3->getAttribute(i)->getMaximum();
        //cout << " ** i(" << i << ") name[" << sx << "] Min U[" <<  dminU << "] R[" << dminR << "] Max U[" << dmaxU  << "] R[" << dmaxR << "]" << endl;
        CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE( Form("attribute[%s] Min Uranie[%f] ROOT[%f]", sx.Data(), dminU, dminR), dminR, dminU, DBL_PRECISION_COMPARE);
        CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE( Form("attribute[%s] Max Uranie[%f] ROOT[%f]", sx.Data(), dmaxU, dmaxR), dmaxR, dmaxU, DBL_PRECISION_COMPARE);
    }
}


void TDataServerTest::testComputeStatisticOnABranchWithSeveralLeaves()
{
    tds1->addAttribute("E");
    tds1->createTuple();

    for (Int_t i = 0; i < 10; i++)
        tds1->getTuple()->Fill(i, i+1);
   
    tds1->computeStatistic("E");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getAttribute("E")->getMean(), 5.5,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getAttribute("E")->getStd(), 3.0276503541,
            DBL_PRECISION_COMPARE);
    
    Double_t *valOut = new Double_t[2];
    
    tds1->addAttribute("F");
    tds1->addAttribute("G");    
    
    TBranch* branchAttribut = tds1->getTuple()->Branch("FG", valOut, "F/D:G/D") ;
    branchAttribut->Print();
    
    for (Int_t i = 0; i < 10; i++)
    {
        valOut[0] = i+2.0; 
        valOut[1] = 2*(i+1.0);
        cout<<branchAttribut->Fill()<<endl;
    }
    
    tds1->getTuple()->ResetBranchAddress(branchAttribut);
    
    
    tds1->computeStatistic();
    
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getAttribute("F")->getMean(), 6.5,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getAttribute("F")->getStd(), 3.0276503541,
            DBL_PRECISION_COMPARE);
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getAttribute("G")->getMean(), 11.0,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getAttribute("G")->getStd(), 6.05530070819,
            DBL_PRECISION_COMPARE);
    
    tds1->getTuple()->ResetBranchAddress(branchAttribut);

    delete[] valOut;
}

void TDataServerTest::testComputeQuantilesVector()
{
  tds2->fileDataRead("myDataSimpleNorm.dat");
  const int n=2; double proba[n]={0.5,0.75}; double value[n];
  tds2->computeQuantile("rank",n,proba,value);
  vector<double> vec;
  tds2->getAttribute("rank")->getQuantileVector(proba[0],vec);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(4, vec.at(0), DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(5, vec.at(1), DBL_PRECISION_COMPARE);    
  tds2->getAttribute("rank")->getQuantileVector(proba[1],vec);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(6, vec.at(0), DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(7, vec.at(1), DBL_PRECISION_COMPARE);    
}

void TDataServerTest::testEstimateQuantilesVector()
{
  tds2->fileDataRead("myDataSimpleNorm.dat");
  const int n=2; double proba[n]={0.5,0.75}; double value[n];
  tds2->estimateQuantile("rank",n,proba,value);
  vector<double> vec;
  tds2->getAttribute("rank")->getQuantileVector(proba[0],vec);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(4, vec.at(0), DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(5, vec.at(1), DBL_PRECISION_COMPARE);    
  // tds2->getAttribute("rank")->getQuantileVector(proba[1],vec);
  // CPPUNIT_ASSERT_DOUBLES_EQUAL(6, vec.at(0), DBL_PRECISION_COMPARE);
  // CPPUNIT_ASSERT_DOUBLES_EQUAL(7, vec.at(1), DBL_PRECISION_COMPARE);    
}

void TDataServerTest::testComputeIndexQuantileWilks()
{
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.,
            tds1->computeIndexQuantileWilks(0.1, 0.5, 10),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(4.,
            tds1->computeIndexQuantileWilks(0.4, 0.5, 10),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(5.,
            tds1->computeIndexQuantileWilks(0.5, 0.5, 10),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(6.,
            tds1->computeIndexQuantileWilks(0.6, 0.5, 10),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(9.,
            tds1->computeIndexQuantileWilks(0.9, 0.5, 10),
            DBL_PRECISION_COMPARE);
}

void TDataServerTest::testComputeInverseNormalCDF()
{
#if ROOT_VERSION_CODE <= ROOT_VERSION(5,18,99)
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-1.28155159971612,
            tds1->computeInverseNormalCDF(0.1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.253347148540183,
            tds1->computeInverseNormalCDF(0.4), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tds1->computeInverseNormalCDF(0.5),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.253347148540183,
            tds1->computeInverseNormalCDF(0.6), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.28155159971612,
            tds1->computeInverseNormalCDF(0.9), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.781551599716118,
            tds1->computeInverseNormalCDF(0.1, 0.5, 1.), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.246652851459817,
            tds1->computeInverseNormalCDF(0.4, 0.5, 1.), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5,
            tds1->computeInverseNormalCDF(0.5, 0.5, 1.), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.753347148540183,
            tds1->computeInverseNormalCDF(0.6, 0.5, 1.), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.78155159971612,
            tds1->computeInverseNormalCDF(0.9, 0.5, 1.), DBL_PRECISION_COMPARE);
#else
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -1.2815515655446, tds1->computeInverseNormalCDF(0.1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -0.2533471031358, tds1->computeInverseNormalCDF(0.4), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.0, tds1->computeInverseNormalCDF(0.5), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.2533471031358, tds1->computeInverseNormalCDF(0.6), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.2815515655446, tds1->computeInverseNormalCDF(0.9), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL( -0.7815515655446, tds1->computeInverseNormalCDF(0.1,0.5,1.), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.2466528968642, tds1->computeInverseNormalCDF(0.4,0.5,1.), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.5, tds1->computeInverseNormalCDF(0.5,0.5,1.), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.7533471031358, tds1->computeInverseNormalCDF(0.6,0.5,1.), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.7815515655446, tds1->computeInverseNormalCDF(0.9,0.5,1.), DBL_PRECISION_COMPARE);
#endif

}

void TDataServerTest::testGetTuple()
{
    tds1->fileDataRead(GEYSERFILENAME);
    TTree* tuple = (TNtupleD*) tds1->getTuple();
    TObjArray * Leaves = tuple->GetListOfLeaves();
    TBranch* leaf;

    leaf = (TBranch *) Leaves->At(0);
    CPPUNIT_ASSERT(!strcmp(leaf->GetName(), "tds1__n__iter__"));

    leaf = (TBranch *) Leaves->At(1);
    CPPUNIT_ASSERT(!strcmp(leaf->GetName(), "x1"));

    leaf = (TBranch *) Leaves->At(2);
    CPPUNIT_ASSERT(!strcmp(leaf->GetName(), "x2"));
}

void TDataServerTest::testSetTuple()
{
    TNtupleD * tupleD = new TNtupleD();
    tds3->setTuple(tupleD);
    CPPUNIT_ASSERT(tds3->getNPatterns() == 0);
    delete tupleD;
    
    tds1->fileDataRead(GEYSERFILENAME);
    URANIE::DataServer::TDSNtupleD * tuple = tds1->getTuple();
    tds2->deleteTuple();
    tds2->setTuple(tuple);
    CPPUNIT_ASSERT(tds2->getNPatterns() == 272);
    tds2->deleteTuple();
    tds2->setTuple((URANIE::DataServer::TDSNtupleD *) NULL);
}

void TDataServerTest::testComputeRank()
{
    tds1->fileDataRead("shortGeyser.dat");
    tds1->computeRank();

    CPPUNIT_ASSERT(tds1->getValue("Rk_x1", 0) == 4);
    CPPUNIT_ASSERT(tds1->getValue("Rk_x1", 1) == 1);
    CPPUNIT_ASSERT(tds1->getValue("Rk_x1", 2) == 3);
    CPPUNIT_ASSERT(tds1->getValue("Rk_x1", 3) == 2);

    CPPUNIT_ASSERT(tds1->getValue("Rk_x2", 0) == 4);
    CPPUNIT_ASSERT(tds1->getValue("Rk_x2", 1) == 1);
    CPPUNIT_ASSERT(tds1->getValue("Rk_x2", 2) == 3);
    CPPUNIT_ASSERT(tds1->getValue("Rk_x2", 3) == 2);

    tds2->fileDataRead("myDataSimpleNorm.dat");
    tds2->computeRank("rank");

    for(unsigned int ival=0; ival<5; ival++)
      {
	CPPUNIT_ASSERT_DOUBLES_EQUAL(ival+1, tds2->getValue("Rk_rank",ival,0), DBL_PRECISION_COMPARE);
	CPPUNIT_ASSERT_DOUBLES_EQUAL(ival+1, tds2->getValue("Rk_rank",ival,1), DBL_PRECISION_COMPARE);
      }
}

void TDataServerTest::testGetVIssues()
{
    //URANIE_IGNORED_TEST() ; 
    // Creating the dataserver 
    tds2->addAttribute("x");
    tds2->addAttribute( new URANIE::DataServer::TAttribute("vec",URANIE::DataServer::TAttribute::kVector) );
    tds2->getAttribute("vec")->delShare();
    
    tds2->createTuple();

    stringstream sstr;
    
    const int N = 1100000;
    const int MaxSize=3;
    unsigned int size=0;

    for(unsigned int i=0; i<N; i++)
    {
	sstr.str("");

        sstr << i << "  "<<gRandom->Uniform(1,2)<<" ";
        size=MaxSize;//gRandom->Uniform(0,MaxSize);
	for(unsigned int j=0; j<size; j++)
            sstr << gRandom->Gaus(0,1) << ((j!=(size-1)) ? "," : "");

        //cout<<sstr.str()<<endl;
        tds2->getTuple()->Fill( sstr.str() );
    }

    TCanvas *can= new TCanvas("tree","tree",10,32,600,600);
    can->Divide(2,2);

    can->cd(1);

    // ============ Check Quantiles
    double proba=0.95, quant=0, quantref=1.645; 
    vector<double> allquant;

    tds2->computeQuantile("vec",proba,quant);
    tds2->getAttribute("vec")->getQuantileVector(proba, allquant);
    for(unsigned int i=0; i<allquant.size(); i++)
        CPPUNIT_ASSERT_DOUBLES_EQUAL(quantref, allquant.at(i), 0.015);

    tds2->getAttribute("vec")->clearVectors();
    
    tds2->estimateQuantile("vec",proba,quant,0.5);    
    tds2->getAttribute("vec")->getQuantileVector(proba, allquant);
    for(unsigned int i=0; i<allquant.size(); i++)
       CPPUNIT_ASSERT_DOUBLES_EQUAL(quantref, allquant.at(i), 0.015);

    can->cd(3);
    tds2->drawECDF("x");
    //
    can->cd(4);
    tds2->draw("x","1>0","nclass=3");


    // ============ Check Statistic
    tds2->computeStatistic("vec");
    
    // ============ Check Getters from TDS2NtupleD
    double *data = tds2->getTuple()->getBranchData("x");
    CPPUNIT_ASSERT( data[1000001]!= 0);

    double *myX = new double[N];
    tds2->getTuple()->copyBranchData(myX,N,"x");
    CPPUNIT_ASSERT( myX[1000001]!= 0 );
    delete[] myX;
    
}

////////////////////////////////
///                          ///
///       Exceptions         ///
///                          ///
////////////////////////////////


void TDataServerTest::throwExceptionFileDataReadDataServerNotEmpty1()
{
    tds1->addAttribute("var1");
    CPPUNIT_ASSERT_THROW(tds1->fileDataRead(GEYSERFILENAME),
            URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionFileDataReadDataServerNotEmpty2()
{
    tds1->fileDataRead(GEYSERFILENAME);
    CPPUNIT_ASSERT_THROW(tds1->fileDataRead(GEYSERFILENAME),
            URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionFileDataReadFileFileNotExist()
{
    CPPUNIT_ASSERT_THROW(tds1->fileDataRead("FileDoesNotExist.dat"),
            URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionFileDataReadNoLineHeader()
{
    CPPUNIT_ASSERT_THROW(tds1->fileDataRead(GEYSERFILENAMENOLINEHEADER),
            URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionFileDataReadTypeHeaderNotCompleteType()
{
    CPPUNIT_ASSERT_THROW(tds1->fileDataRead(STRINGANDVECTORNOTCOMPLETETYPE),
            URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionFileDataReadTypeHeaderMissingField()
{
    CPPUNIT_ASSERT_THROW(tds1->fileDataRead(STRINGANDVECTORMISSINGFIELD),
            URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionFileDataReadTypeHeaderTooMuchField()
{
    CPPUNIT_ASSERT_THROW(tds1->fileDataRead(STRINGANDVECTORTOOMUCHFIELD),
            URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionSetIterator()
{
    tds1->fileDataRead(GEYSERFILENAME);
    CPPUNIT_ASSERT_THROW(tds1->setIterator("VarNotExist"),
            URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionGetNInputAttributes()
{
    CPPUNIT_ASSERT_THROW(tds1->getNInputAttributes(),
            URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionGetNOutputAttributes()
{
    CPPUNIT_ASSERT_THROW(tds1->getNOutputAttributes(),
            URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionNtupleDataReadWithInexistingRootFile()
{
    CPPUNIT_ASSERT_THROW(tds1->ntupleDataRead("fake.root", "flowrate"),
            URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionNtupleDataReadWithNoExistingTuple()
{
    tds1->fileDataRead("flowrateUniformDesign.dat");
    CPPUNIT_ASSERT_THROW(tds4->ntupleDataRead("_dataserver_.root", "faketree"),
            URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionNtupleDataReadWithNotEmptyTDataspecification()
{
    tds1->fileDataRead("flowrateUniformDesign.dat");
    CPPUNIT_ASSERT_THROW(tds1->ntupleDataRead("_dataserver_.root", "flowrate"),
            URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionNtupleDataReadWithEmptyExistingTree()
{
  TFile *fin=new TFile("pouet.root","RECREATE");
  TTree *t = new TTree("youhou","youpi");
  t->SetDirectory(0);
  t->Write();
  fin->Close();
  CPPUNIT_ASSERT_THROW(tds1->ntupleDataRead("pouet.root","youhou"),
		       URANIE::Exceptions::UExceptions);
  delete fin;
  delete t;
}

void TDataServerTest::throwExceptionExportDataWithEmptyTDataserver()
{
    CPPUNIT_ASSERT_THROW(tds1->exportData("fakeFile.dat"),
            URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionExportDataWithTDSNull()
{
    tds1->addAttribute("x");
    CPPUNIT_ASSERT_THROW(tds1->exportData("fakeFile.dat"),
            URANIE::Exceptions::UErrorExceptions);
}

void TDataServerTest::throwExceptionExportDataHeaderWithEmptyTDataserver()
{
    CPPUNIT_ASSERT_THROW(tds1->exportDataHeader("fakeFile.h"),
            URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionExportDataHeaderWithNoPatterns()
{
    tds2->addAttribute("x");
    tds2->addAttribute("y");
    CPPUNIT_ASSERT_THROW(tds2->exportDataHeader("fakeFile.h", "x:y"),
            URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionExportDataNeMoWithEmptyTDataserver()
{
    CPPUNIT_ASSERT_THROW(tds1->exportDataNeMo("fakeFile.nemo", "x1:x2", "y"),
            URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionExportDataNeMoWithNoPatterns()
{
    tds2->addAttribute("x");
    tds2->addAttribute("y");
    tds2->addAttribute("z");
    CPPUNIT_ASSERT_THROW(tds2->exportDataNeMo("fakeFile.nemo", "x:y", "z"),
            URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionExportDataContextWithEmptyTDataserver()
{
    CPPUNIT_ASSERT_THROW(tds1->exportDataContext("fakeFile.tex"),
            URANIE::Exceptions::UErrorExceptions);
}

void TDataServerTest::throwExceptionExportDataContextWithNoPatterns()
{
    tds2->addAttribute("x");
    tds2->addAttribute("y");
    tds2->createTuple();
    CPPUNIT_ASSERT_THROW(tds2->exportDataContext("fakeFile.tex"),
            URANIE::Exceptions::UErrorExceptions);
}

void TDataServerTest::throwExceptionAddAttributeWithFormula()
{
    // The exception for an invalid formula is now thrown by fillAttribute
    //CPPUNIT_ASSERT_THROW(tds1->addAttribute("z","x*x") , URANIE::Exceptions::UErrorExceptions );

    tds2->fileDataRead(GEYSERFILENAME);
    // Creating a TAttributeFormula with an empty formula
    CPPUNIT_ASSERT_THROW(tds2->addAttribute("x3", " "),
            URANIE::Exceptions::UErrorExceptions);
    //CPPUNIT_ASSERT_THROW( tds2->addAttribute("x3", "u*x1+x2") , URANIE::Exceptions::UErrorExceptions );
    // Creating a TAttributeFormula with the name of an existing attribute
    CPPUNIT_ASSERT_THROW(tds2->addAttribute("x1", "x2*x2"),
            URANIE::Exceptions::UErrorExceptions);
}


void TDataServerTest::throwExceptionAddAttributeWithData()
{

    tds2->fileDataRead(GEYSERFILENAME);

    Int_t nP = tds2->getNPatterns();
    vector<double> toobig, toosmall, good;
    for(int i=0; i<nP+1; i++)
    {
        if(i<(nP-1))
            toosmall.push_back(i);
        if(i<nP)
            good.push_back(i);
        toobig.push_back(i);
    }
    
    // Too small
    CPPUNIT_ASSERT_THROW(tds2->addAttributeUsingData("toosmall",&toosmall[0],toosmall.size()),
                         URANIE::Exceptions::UErrorExceptions);

    // Too big
    CPPUNIT_ASSERT_THROW(tds2->addAttributeUsingData("toobig",&toobig[0],toobig.size()),
                         URANIE::Exceptions::UErrorExceptions);

    double *arr=NULL;
    // Empty array
    CPPUNIT_ASSERT_THROW(tds2->addAttributeUsingData("goodbutempty", arr,good.size()),
                         URANIE::Exceptions::UErrorExceptions);


    tds2->addAttributeUsingData("good", &good.at(0), good.size());
    // existing name
    CPPUNIT_ASSERT_THROW(tds2->addAttributeUsingData("good", &good.at(0), good.size()),
                         URANIE::Exceptions::UErrorExceptions);

    delete tds2;
    tds2 = new URANIE::DataServer::TDataServer();    
    tds2->addAttribute("toto");

    // adding data when other attribute does not have some (no datatree so far)
    CPPUNIT_ASSERT_THROW(tds2->addAttributeUsingData("good", &good.at(0), good.size()),
                         URANIE::Exceptions::UErrorExceptions);

    
}

void TDataServerTest::throwExceptionFillAttribute()
{
    tds1->addAttribute("x", "1");
    // Filling an attribute without a data tree
    CPPUNIT_ASSERT_THROW(tds1->fillAttribute("x"),
            URANIE::Exceptions::UErrorExceptions);

    tds2->fileDataRead(GEYSERFILENAME);
    tds2->addAttribute("x3");
    tds2->addAttribute("x4","x1+x2");
    // Filling an unknown attribute    
    CPPUNIT_ASSERT_THROW(tds2->fillAttribute("z"),
            URANIE::Exceptions::UErrorExceptions);
    // Filling a TAttribute without a formula    
    CPPUNIT_ASSERT_THROW(tds2->fillAttribute("x3"),
            URANIE::Exceptions::UErrorExceptions);
    // Filling a TAttribute with an invalid formula 
    CPPUNIT_ASSERT_THROW(tds2->fillAttribute("x3", "u*x1+x2"),
            URANIE::Exceptions::UErrorExceptions);
    // Filling a TAttributeFormula with another formula    
    CPPUNIT_ASSERT_THROW(tds2->fillAttribute("x4","x1-x2"),
            URANIE::Exceptions::UErrorExceptions);
    // Filling an attribute already filled 
    CPPUNIT_ASSERT_THROW(tds2->fillAttribute("x1", "x2*x2"),
            URANIE::Exceptions::UErrorExceptions);

}


void TDataServerTest::throwExceptionDelAttribute()
{
    tds2->fileDataRead(GEYSERFILENAME);
    CPPUNIT_ASSERT_THROW(tds2->delAttribute("z"),
            URANIE::Exceptions::UErrorExceptions);
}

void TDataServerTest::throwExceptionGetValueWithStringOrVector()
{
  tds1->fileDataRead(STRINGANDVECTOR);
  CPPUNIT_ASSERT_THROW(tds1->getValue("fakeAttribute",0),
		       URANIE::Exceptions::UErrorExceptions);
  CPPUNIT_ASSERT_THROW(tds1->getValue("xfactor",0,10),
		       URANIE::Exceptions::UErrorExceptions);
  CPPUNIT_ASSERT_THROW(tds1->getValue("xfactor",0),
		       URANIE::Exceptions::UErrorExceptions);
  CPPUNIT_ASSERT_THROW(tds1->getValue("parity",0),
		       URANIE::Exceptions::UErrorExceptions);
}

void TDataServerTest::throwExceptionGetVectorWithStringOrDouble()
{
  tds1->fileDataRead(STRINGANDVECTOR);
  CPPUNIT_ASSERT_THROW(tds1->getVector("fakeAttribute",0),
		       URANIE::Exceptions::UErrorExceptions);
  CPPUNIT_ASSERT_THROW(tds1->getVector("x",0),
		       URANIE::Exceptions::UErrorExceptions);
  CPPUNIT_ASSERT_THROW(tds1->getVector("parity",0),
		       URANIE::Exceptions::UErrorExceptions);
}

void TDataServerTest::throwExceptionGetStringWithVectorOrDouble()
{
  tds1->fileDataRead(STRINGANDVECTOR);
  CPPUNIT_ASSERT_THROW(tds1->getString("fakeAttribute",0),
		       URANIE::Exceptions::UErrorExceptions);
  CPPUNIT_ASSERT_THROW(tds1->getString("x",0),
		       URANIE::Exceptions::UErrorExceptions);
  CPPUNIT_ASSERT_THROW(tds1->getString("xfactor",0),
		       URANIE::Exceptions::UErrorExceptions);
}

void TDataServerTest::throwExceptionGetAttributeIndexFromName()
{
    tds1->fileDataRead(GEYSERFILENAME);
    CPPUNIT_ASSERT_THROW(tds1->getAttributeIndex("fakeAttribute"),
            URANIE::Exceptions::UErrorExceptions);
}

void TDataServerTest::throwExceptionGetAttributeIndexFromAttribute()
{
    tds1->fileDataRead(GEYSERFILENAME);
    URANIE::DataServer::TAttribute * fakeattribute = new URANIE::DataServer::TAttribute("fakeAttribute");
    CPPUNIT_ASSERT_THROW(tds1->getAttributeIndex(fakeattribute),
            URANIE::Exceptions::UErrorExceptions);
    delete fakeattribute;
}

void TDataServerTest::throwExceptionGetAttributeWithEmptyDataspecification()
{
    CPPUNIT_ASSERT_THROW(tds1->getAttribute(1),
            URANIE::Exceptions::UErrorExceptions);
}

void TDataServerTest::throwExceptionGetAttributeWithIndexSuperiorToNbAttributes()
{
    tds1->fileDataRead("flowrateUniformDesign.dat");
    CPPUNIT_ASSERT_THROW(tds1->getAttribute(30),
            URANIE::Exceptions::UErrorExceptions);
}

void TDataServerTest::throwExceptionGetAttributeFromIndexWithNegativeIndex()
{
    tds1->fileDataRead("flowrateUniformDesign.dat");
    Int_t nbDataspecificationAttributes =
            tds1->getDataSpecification()->getNAttributes();
    CPPUNIT_ASSERT_THROW(tds1->getAttribute(nbDataspecificationAttributes),
            URANIE::Exceptions::UErrorExceptions);
}

void TDataServerTest::throwExceptionMerge2filesOfDifferentNbPatterns()
{
    tds1->fileDataRead("flowrateUniformDesign.dat");
    tds4->fileDataRead(GEYSERFILENAME);
    CPPUNIT_ASSERT_THROW(tds1->merge(tds4, "*", kTRUE),
            URANIE::Exceptions::UErrorExceptions);
}

void TDataServerTest::throwExceptionComputeStatistic()
{
  tds1->fileDataRead(GEYSERFILENAME);
  CPPUNIT_ASSERT_THROW(tds1->computeStatistic("x1:x2:x3"),
		       URANIE::Exceptions::UErrorExceptions);
}

void TDataServerTest::throwExceptionComputeRank()
{
    tds2->fileDataRead(GEYSERFILENAME);
    CPPUNIT_ASSERT_THROW(tds2->computeRank("x3"),
            URANIE::Exceptions::UErrorExceptions);
}

void TDataServerTest::throwExceptionComputeInverseNormalCDF()
{
    CPPUNIT_ASSERT_THROW(tds2->computeInverseNormalCDF(-2., 0.5, 2.),
            URANIE::Exceptions::UErrorExceptions);
}

void TDataServerTest::throwExceptionComputeIndexQuantileWilksForBadProbability()
{
    CPPUNIT_ASSERT_THROW(tds2->computeIndexQuantileWilks(2., 0.5, 10),
            URANIE::Exceptions::UErrorExceptions);
}

void TDataServerTest::throwExceptionComputeIndexQuantileWilksForInapropriateConfidenceLevel()
{
    CPPUNIT_ASSERT_THROW(tds2->computeIndexQuantileWilks(0.5, 2., 10),
            URANIE::Exceptions::UErrorExceptions);
}

void TDataServerTest::throwExceptionScan()
{
    CPPUNIT_ASSERT_THROW(tds2->scan("*"), URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionDraw()
{
    CPPUNIT_ASSERT_THROW(tds2->draw("*"), URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionDrawWithOrderTooHigh()
{
    tds2->fileDataRead(GEYSERFILENAME);
    CPPUNIT_ASSERT_THROW(tds2->draw(200), URANIE::Exceptions::UExceptions);
//     delete tds2;
}
void TDataServerTest::throwExceptionDrawWithOrderAndEmptyTds()
{
    CPPUNIT_ASSERT_THROW(tds2->draw(200), URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionDrawHistogram()
{
    CPPUNIT_ASSERT_THROW(tds2->drawHistogram("*"),
            URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionDrawIter()
{
    CPPUNIT_ASSERT_THROW(tds2->drawIter("*"), URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionDrawECDF()
{
    CPPUNIT_ASSERT_THROW(tds2->drawECDF("*"), URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionDrawECDFWithMultiVariables()
{
    tds2->fileDataRead(GEYSERFILENAME);
    CPPUNIT_ASSERT_THROW(tds2->drawECDF("x1:x2"),
            URANIE::Exceptions::UExceptions);
//     delete tds2;
}

void TDataServerTest::throwExceptionDrawCDF()
{
    CPPUNIT_ASSERT_THROW(tds2->drawCDF("*"), URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionDrawCDFWithMultiVariables()
{
    tds2->fileDataRead(GEYSERFILENAME);
    CPPUNIT_ASSERT_THROW(tds2->drawCDF("x1:x2"),
            URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionDrawBoxPlot()
{
    CPPUNIT_ASSERT_THROW(tds2->drawBoxPlot("x1"),
            URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionDrawTufte()
{
    CPPUNIT_ASSERT_THROW(tds2->drawTufte("*"), URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionDrawTufteWithBadNumberOfVariables()
{
    CPPUNIT_ASSERT_THROW(tds2->drawTufte("x1:x2"),
            URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionDrawScatterPlotWithBadNumberOfVariables()
{
    CPPUNIT_ASSERT_THROW(tds2->drawScatterplot("*"),
            URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionDrawProfile()
{
    CPPUNIT_ASSERT_THROW(tds2->drawProfile("*"),
            URANIE::Exceptions::UExceptions);
}

#ifdef WITH_MATHMORE
void TDataServerTest::throwExceptionDrawPPPlot()
{
    CPPUNIT_ASSERT_THROW(tds2->drawPPPlot("*","normal",150), URANIE::Exceptions::UErrorExceptions);

    tds2->fileDataRead(FLOWRATEUNIFORDESIGN);

    CPPUNIT_ASSERT_THROW(tds2->drawPPPlot("rw3","normal",150), URANIE::Exceptions::UErrorExceptions);

    CPPUNIT_ASSERT_THROW(tds2->drawPPPlot("*","normal",150), URANIE::Exceptions::UErrorExceptions);   
    
    CPPUNIT_ASSERT_THROW(tds2->drawPPPlot("x:y","normal",150), URANIE::Exceptions::UErrorExceptions);

    CPPUNIT_ASSERT_THROW(tds2->drawPPPlot("rw","triangular",150), URANIE::Exceptions::UErrorExceptions);
}
void TDataServerTest::throwExceptionDrawQQPlot()
{
    CPPUNIT_ASSERT_THROW(tds2->drawQQPlot("*","normal",150), URANIE::Exceptions::UErrorExceptions);

    tds2->fileDataRead(FLOWRATEUNIFORDESIGN);

    CPPUNIT_ASSERT_THROW(tds2->drawQQPlot("rw3","normal",150), URANIE::Exceptions::UErrorExceptions);

    CPPUNIT_ASSERT_THROW(tds2->drawQQPlot("*","normal",150), URANIE::Exceptions::UErrorExceptions);   
    
    CPPUNIT_ASSERT_THROW(tds2->drawQQPlot("x:y","normal",150), URANIE::Exceptions::UErrorExceptions);

    CPPUNIT_ASSERT_THROW(tds2->drawQQPlot("rw","triangular",150), URANIE::Exceptions::UErrorExceptions);

}
#endif

void TDataServerTest::throwExceptionStartViewer()
{
    CPPUNIT_ASSERT_THROW(tds2->startViewer(), URANIE::Exceptions::UExceptions);
}

void TDataServerTest::throwExceptionGetTuple()
{
    CPPUNIT_ASSERT_THROW(tds2->getTuple(),
            URANIE::Exceptions::UWarningExceptions);
}

void TDataServerTest::throwExceptionfillConstantAttribute()
{
    tds2->fileDataRead(FLOWRATEUNIFORDESIGN);
    CPPUNIT_ASSERT_EQUAL(9, tds2->getNAttributes());
    
    URANIE::DataServer::UEntry *tmpConst = new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::String, "foofoo");
    string pouet = "pouet";
    tmpConst->pushEntry(&pouet,false);
    CPPUNIT_ASSERT_THROW(tds2->fillConstantAttribute("kw",tmpConst), URANIE::Exceptions::UErrorExceptions);
    delete tmpConst;

    tmpConst = new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Vector, "foofoo");
    tmpConst->pushEntry(0.0);
    CPPUNIT_ASSERT_THROW(tds2->fillConstantAttribute("kw",tmpConst), URANIE::Exceptions::UErrorExceptions);
    delete tmpConst;

    tmpConst = new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Double, "foofoo");
    tmpConst->pushEntry(1.5);
    CPPUNIT_ASSERT_THROW(tds2->fillConstantAttribute("kw",tmpConst), URANIE::Exceptions::UErrorExceptions);
    delete tmpConst;

    tmpConst = new URANIE::DataServer::UEntry();
    CPPUNIT_ASSERT_THROW(tds2->fillConstantAttribute("pouet",tmpConst), URANIE::Exceptions::UErrorExceptions);
    delete tmpConst;
    
}
