#include "TestTPCA.h"
#include "TPCA.h"
#include "TDataServer.h"
#ifdef WITH_DATAFRAME
#include <ROOT/RDataFrame.hxx>
#endif
#include "TFile.h"
#include <string>

const double epsilon_factor = 1e3 * numeric_limits<double>::epsilon();

const string marks_input = "Notes.dat";
const string marks_variables = "marks_variables.root";
const string marks_results = "marks_results.root";

const string decathlon_input = "decathlon.dat";
const string decathlon_results = "decathlon_results.root";
const string decathlon_variables = "decathlon_variables.root";

void TPCATest::throwNullTDataServerPointer(){
    unique_ptr<URANIE::DataServer::TDataServer> data_ptr;
    URANIE::DataServer::TPCA pca(data_ptr.get(), "");
}

void TPCATest::throwEmptyTDataServer() {
    URANIE::DataServer::TDataServer data;
    URANIE::DataServer::TPCA pca(&data, "");
}

void TPCATest::throwStringAttribute() {
    URANIE::DataServer::TDataServer data;
    data.fileDataRead(marks_input.data());
    CPPUNIT_ASSERT_EQUAL(TString("S"), data.getAttribute(0)->getDataTypeStr());
    URANIE::DataServer::TPCA pca(&data, data.getAttribute(0)->GetName());
    pca.compute();
}

void TPCATest::testDestruction() {
    URANIE::DataServer::TDataServer data;
    data.fileDataRead(marks_input.data());
    URANIE::DataServer::TPCA pca(&data, data.getStringListOfAttributes().data());
}

#ifdef WITH_DATAFRAME
void compare_double(ROOT::RDataFrame& expected, ROOT::RDataFrame& actual,
    const string& column_name) {
    auto expected_column = expected.Take<double>(column_name);
    auto actual_column = actual.Take<double>(column_name);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(column_name + " size",
        expected_column.GetValue().size(), actual_column.GetValue().size());

    for(size_t i_value=0; i_value<expected_column.GetValue().size(); i_value++)
        CPPUNIT_ASSERT_DOUBLES_EQUAL_MESSAGE(column_name + "[" + to_string(i_value) + "]",
            expected_column.GetValue()[i_value], actual_column.GetValue()[i_value],
            abs(expected_column.GetValue()[i_value] * epsilon_factor));
}

void compare_string(ROOT::RDataFrame& expected, ROOT::RDataFrame& actual,
    const string& column_name) {
    auto expected_column = expected.Take<string>(column_name);
    auto actual_column = actual.Take<string>(column_name);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(column_name + " size",
        expected_column.GetValue().size(), actual_column.GetValue().size());

    for(size_t i_value=0; i_value<expected_column.GetValue().size(); i_value++)
        CPPUNIT_ASSERT_EQUAL_MESSAGE(column_name + "[" + to_string(i_value) + "]",
            expected_column.GetValue()[i_value], actual_column.GetValue()[i_value]);
}

void compare_columns(ROOT::RDataFrame& expected, ROOT::RDataFrame& actual) {
    CPPUNIT_ASSERT_EQUAL_MESSAGE("GetColumnNames().size()",
        expected.GetColumnNames().size(), actual.GetColumnNames().size());
    for(size_t i_column=0; i_column<expected.GetColumnNames().size(); ++i_column)
        CPPUNIT_ASSERT_EQUAL_MESSAGE("GetColumnNames()[" + to_string(i_column) + "]",
            expected.GetColumnNames()[i_column], actual.GetColumnNames()[i_column]);

#if ROOT_VERSION_CODE >= ROOT_VERSION(6, 16, 0)
    for(const auto& column_name: expected.GetColumnNames())
        if (expected.GetColumnType(column_name) == "Double_t")
            compare_double(expected, actual, column_name);
        else if (expected.GetColumnType(column_name) == "string")
            compare_string(expected, actual, column_name);
        else
            throw invalid_argument(column_name + " type unknown: " +
                expected.GetColumnType(column_name));
#else
    cout << endl;
    cout << "ROOT version: " + string(gROOT->GetVersion()) + " is too low." << endl;
    cout << "Comparison is skipped for:" << endl;
    for(const auto& column_name: expected.GetColumnNames())
        cout << '\t' << column_name << endl;
#endif
}

void TPCATest::testMarksResultsAndVariables() {
    URANIE::DataServer::TDataServer data;
    data.fileDataRead(marks_input.data());
    URANIE::DataServer::TPCA pca(&data, data.getStringListOfAttributes().data());
    pca.compute();

    ROOT::RDataFrame results_actual(*pca.getResultNtupleD());
    ROOT::RDataFrame results_expected(pca.getResultNtupleD()->GetName(), marks_results);
    compare_columns(results_expected, results_actual);

    ROOT::RDataFrame variables_actual(*pca.getVariableResultNtupleD());
    ROOT::RDataFrame variables_expected(pca.getVariableResultNtupleD()->GetName(), marks_variables);
    compare_columns(variables_expected, variables_actual);
}

void TPCATest::testDecathlonResultsAndVariables() {
    URANIE::DataServer::TDataServer data;
    data.fileDataRead(decathlon_input.data());
    const vector<string> to_delete {"Rank", "Points", "Competition"};
    for(const auto& name: to_delete)
        data.delAttribute(name);
    URANIE::DataServer::TPCA pca(&data, data.getStringListOfAttributes().data());
    pca.compute();

    ROOT::RDataFrame results_actual(*pca.getResultNtupleD());
    ROOT::RDataFrame results_expected(pca.getResultNtupleD()->GetName(), decathlon_results);
    compare_columns(results_expected, results_actual);

    ROOT::RDataFrame variables_actual(*pca.getVariableResultNtupleD());
    ROOT::RDataFrame variables_expected(pca.getVariableResultNtupleD()->GetName(), decathlon_variables);
    compare_columns(variables_expected, variables_actual);
}
#endif
