//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
// TCluster lun aoû 4 14:21:10 CEST 2008
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TCluster.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mon august 4th 14:21:10 CEST 2008
 
 \brief Implementation of the class URANIE::Launcher::TCluster
 
 */
#include "TCluster.h"
#define BUF_SIZE 2048

#include "TObjString.h"

using std::cout;
using std::endl;

ClassImp (URANIE::Launcher::TCluster)

namespace URANIE
{
namespace Launcher
{
TCluster::TCluster(const char* name, const char *nodeList, Int_t nnode) :
        TNamed(name, nodeList), _blog(kFALSE), _nnode(nnode), _slistOfNodes("")
{
    if (_blog)
        cout << " ** Constructor TCluster::TCluster [" << GetName()
                << "] Title[" << GetTitle() << "]  Node[" << _nnode << "]"
                << endl;
}

#ifdef WITH_LIBSSH
TCluster::TCluster(URANIE::DataServer::TDataServer * tds, TCode * mycode) :
    TNamed("clustername","clustertitle"), _code(mycode), _tds(tds)
{

}

TCluster::TCluster(TCode * mycode) :
    TNamed("clustername","clustertitle"), _code(mycode), _tds(NULL)    
{

}
#endif



TCluster::~TCluster()
{
    if (_blog)
        cout << " ** Destroy TCluster [" << GetName() << "] Title["
                << GetTitle() << "]" << endl;
}

///////////////////////////////////////////////////////////
//  SSH connection
///////////////////////////////////////////////////////////
#ifdef WITH_LIBSSH
void TCluster::setCluster(TCluster *cluster)
{
    TString aux;
    _sClusterName= cluster->_sClusterName.Copy();
    _sClusterIP = cluster->_sClusterIP.Copy();
    _sOutputHeader= cluster->_sOutputHeader.Copy();
    _sSkeletonHeader=cluster->_sSkeletonHeader.Copy();
    _sRemoteWorkingDirectory= cluster->_sRemoteWorkingDirectory.Copy();
    map<TString, TString> mp2(cluster->_sListClusterDirectives);
    _sListClusterDirectives = mp2;

    _nAllocatedCores =  Int_t(cluster->_nAllocatedCores);
    _sClusterUser= cluster->_sClusterUser.Copy();
    _sClusterUserAuthMethod= cluster->_sClusterUserAuthMethod.Copy();
    _sJobName = cluster->_sJobName.Copy();
}


Int_t TCluster::connectSSH(TString cluster_name ,  TString  user,  TString authentification_method)
{
    char * username;

    Int_t rc;
    ssh_options_set(*_ptrsshSession, SSH_OPTIONS_HOST, cluster_name.Data());
    if (user.Length() == 0)
    {
        ssh_options_get(*_ptrsshSession, SSH_OPTIONS_USER, &username);
        user=TString(username);
    }
    else
    {
        ssh_options_set(*_ptrsshSession, SSH_OPTIONS_USER, user.Data());
    }
    rc = ssh_connect(*_ptrsshSession);
    if (rc != SSH_OK)
    {
        ssh_disconnect(*_ptrsshSession);
        cout << "Unable to connect , wrong hostname " << endl;

    }
    else
    {
        if (authentification_method.Length() > 0 or authentification_method =="public_key")
        {
            rc = ssh_userauth_publickey_auto(*_ptrsshSession, NULL, NULL);
            if (rc != SSH_AUTH_SUCCESS)
            {
                cout << "Unable to connect, wrong public key or username" << endl;
            }
        }
        else
        {
            //keyboard? iterative? other?
            cout << "Not implemented" << endl;
        }
    }

    return rc;


}

Int_t TCluster::verify_knownhostSSH()
{
    //enum ssh_known_hosts_e state;
    unsigned char *hash = NULL;
    ssh_key srv_pubkey = NULL;
    size_t hlen;
    char buf[10];
    char *hexa;
    char *p;
    Int_t cmp;
    Int_t rc;
    Int_t state = ssh_session_is_known_server(*_ptrsshSession);
    switch (state)
    {
        case SSH_KNOWN_HOSTS_OK:
            return 0;
        case SSH_KNOWN_HOSTS_CHANGED:
            cout << "Host key for server changed: it is now:" << endl;
            ssh_print_hexa("Public key hash", hash, hlen);
            cout << "for security reasons, connection will be stopped" << endl;
            ssh_clean_pubkey_hash(&hash);
            return -1;
        case SSH_KNOWN_HOSTS_OTHER:
            cout << "The host key for this server was not found but an other"
                    "type of key exists." << endl;
            ssh_clean_pubkey_hash(&hash);
            return -1;
        case SSH_KNOWN_HOSTS_NOT_FOUND:
            cout << "Could not find known host file." ;
            cout << "If you accept the host key here, file will be automatically created." << endl;
        case SSH_KNOWN_HOSTS_UNKNOWN:
            hexa = ssh_get_hexa(hash, hlen);
            cout << "The server is unknown. Do you trust the host key?" << endl;
            cout << "Public key hash: %s" << hexa << endl;
            ssh_string_free_char(hexa);
            ssh_clean_pubkey_hash(&hash);
            p = fgets(buf, sizeof(buf), stdin);
            if (p == NULL) {
                return -1;
            }
            cmp = strncasecmp(buf, "yes", 3);
            if (cmp != 0) {
                return -1;
            }
            rc = ssh_session_update_known_hosts(*_ptrsshSession);
            if (rc < 0) {
                cout <<  "Error \n" << strerror(errno) << endl;
                return -1;
            }
            break;
        case SSH_KNOWN_HOSTS_ERROR:
            cout <<  "Error %s" << ssh_get_error(*_ptrsshSession) << endl;
            ssh_clean_pubkey_hash(&hash);
            return -1;
    }
    return 0;
}


Int_t TCluster::openSFTPChannel()
{
    if (_sConnectionProtocol.EqualTo("ecaccess")){
        return SSH_FX_OK;
    }
    Int_t rc;
    (*_ptrsftp) = sftp_new(*_ptrsshSession);
    if (*_ptrsftp == NULL)
    {
        cout << "Error allocating SFTP session: %s" << ssh_get_error(*_ptrsshSession) << endl;
        return SSH_ERROR;
    }

    rc = sftp_init(*_ptrsftp);
    if (rc != SSH_OK)
    {
        cout << "Error initializing SFTP session: %s." << sftp_get_error(*_ptrsftp) << endl;
        sftp_free(*_ptrsftp);
        return rc;
    }
    return SSH_FX_OK;
}
Int_t TCluster::closeSFTPChannel()
{
    if (_sConnectionProtocol.EqualTo("ecaccess")){
        return SSH_FX_OK;
    }
    sftp_free(*_ptrsftp);

    return SSH_FX_OK;
}

Int_t TCluster::sendHeader()
{

    Int_t rc;
    //RECIEVE(REMOTE)

    Int_t access_type = O_RDWR | O_CREAT | O_TRUNC;

    sftp_file file = sftp_open(*_ptrsftp,gSystem->ConcatFileName(_sRemoteWorkingDirectory,_sSplitName),access_type,0755);
    if (file == NULL)
    {
        cout << "(REMOTE)Can't open file for writing: " << ssh_get_error(*_ptrsshSession) << endl;
        return SSH_ERROR;
    }
    //SEND local
    ifstream infile(gSystem->ConcatFileName(_sWorkingDirectory,_sSplitName), ios::binary);
    while (infile)
    {
        char buffer[BUF_SIZE];
        infile.read(buffer, sizeof(buffer));
        if (infile.gcount() > 0)
        {
            ssize_t nwritten = sftp_write(file, buffer, infile.gcount());
            if (nwritten != infile.gcount()) {
                cout <<  "(LOCAL)Can't write data to file: " << ssh_get_error(*_ptrsshSession) << endl;
                sftp_close(file);
                return 1;
            }
        }
    }
    sftp_close(file);
    return SSH_OK;
}

Int_t TCluster::sendCompiledBinary(list<TString> filenames)
{
    Int_t rc;

    Int_t access_type = O_RDWR | O_CREAT | O_TRUNC;
    sftp_mkdir(*_ptrsftp,gSystem->ConcatFileName(_sRemoteWorkingDirectory,"bin"),0755);
    auto it = filenames.begin();
    while ( it != filenames.end() )
    {
        //RECIEVE(REMOTE)
        sftp_file file = sftp_open(*_ptrsftp,gSystem->ConcatFileName(_sRemoteWorkingDirectory,"bin/"+(*it)),access_type,0755);
        if (file == NULL)
        {
            cout << "(REMOTE)Can't open file for writing: " << ssh_get_error(*_ptrsshSession) << endl;
            return SSH_ERROR;
        }
        //SEND local
        ifstream infile(gSystem->ConcatFileName(_sWorkingDirectory,"bin/"+(*it)), ios::binary);
        while (infile)
        {
            char buffer[BUF_SIZE];
            infile.read(buffer, sizeof(buffer));
            if (infile.gcount() > 0) {
                ssize_t nwritten = sftp_write(file, buffer, infile.gcount());
                if (nwritten != infile.gcount()) {
                    cout << "(LOCAL)Can't write data to file: " << ssh_get_error(*_ptrsshSession) << endl;
                    sftp_close(file);
                    return 1;
                }
            }
        }
        sftp_close(file);
        it++;
    }
    _sUserBinary = kTRUE;
    return SSH_OK;

}

///////////////
// Cluster
///////////////
void TCluster::substituteDirectives(TString infilePath, TString outFilePath, std::map<TString,TString> ListOfDirectives)
{
    std::ifstream infile(infilePath);
    std::ofstream os(outFilePath, ios::trunc );
    std::string line;
    TString auxLine;
    while (std::getline(infile, line))
    {
        auxLine = TString(line);
        std::map<TString, TString>::iterator it = ListOfDirectives.begin();
        while(it != ListOfDirectives.end())
        {
            if (auxLine.Contains(it->first.Data()))
            {
                if ( (auxLine.Contains("-o") || auxLine.Contains("-e") || auxLine.Contains("--error") || auxLine.Contains("--output") ) && !it->second.Contains(_sRemoteWorkingDirectory) )
                {
                    it->second = it->second.Strip(TString::kTrailing,' ');
                    it->second = it->second.Strip(TString::kLeading,' ');
                    it->second=_sRemoteWorkingDirectory+"/"+it->second;

                }
                replaceHeaderKey(line, it->first.Data(), it->second.Data());
            }
            it++;
        }

        os << line << endl;
    }
    infile.close();
    os.close();

}
void TCluster::addClusterDirective(TString directive, TString value)
{

    directive = "@"+directive+"@";
    _sListClusterDirectives.insert(std::make_pair(directive,value));
}

void TCluster::preTreatment(Option_t *option)
{

    TString soption(option);
    soption.ToLower();
    _sSshSession = ssh_new();
    _ptrsshSession = &_sSshSession;
    _ptrsftp= &_sFtpSession;

    if (_sConnectionProtocol.EqualTo("ssh"))
    {
        if (connectSSH(getClusterIP(),_sClusterUser,_sClusterUserAuthMethod) != SSH_OK)     {
            cout << "Error" << endl;
            exit(-1);
        }
        if (verify_knownhostSSH() != SSH_OK)
        {
            cout << "Error" << endl;
            exit(-1);
        }
        if (openSFTPChannel() != SSH_OK)
        {
            cout << "Error" << endl;
            exit(-1);
        }
        if (!_sConnected)
            setRemoteWorkingDirectory();
        if (closeSFTPChannel() != SSH_OK)
        {
            cout << "Error" << endl;
            exit(-1);
        }
    }
}

void TCluster::execRemoteCommand(TString command)
{
    char buffer[BUF_SIZE];
    _sSshOutput = "";
    ssize_t nbytes;
    ssh_channel channelsingle = ssh_channel_new(_sSshSession);
    if (channelsingle == NULL)
    cout << "Channel failed to create, command did not execute" << endl;
    else
        {
        if (ssh_channel_open_session(channelsingle) != SSH_OK)
            cout << "Channel failed to open" << endl;
        else
            {
            if (ssh_channel_request_exec(channelsingle, command.Data()) != SSH_OK)
                cout << "Something went wrong, command did not execute" << endl;
        }
    }
    do
    {
      nbytes = ssh_channel_read(channelsingle, buffer, sizeof(buffer),0);
      _sSshOutput+=TString(buffer);
    } while(nbytes >0);

    ssh_channel_send_eof(channelsingle);
    ssh_channel_close(channelsingle);
    ssh_channel_free(channelsingle);

}
void TCluster::setWorkingDirectory(TString str)
{


    TString strDirectory = str;
    try
    {
        if ((strDirectory.EndsWith("\\")) || (strDirectory.EndsWith("/")))
        {
            strDirectory.Remove(strDirectory.Length() - 1);
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                                                         Form(
                                                                 "TLauncher::setWorkingDirectory method\n the working directory \"%s\" is redefined without last backslash/slash \"%s\"",
                                                                 str.Data(), strDirectory.Data()));

        }
    } catch (URANIE::Exceptions::UExceptions& ue)
    {
        ue.printMessage();
    }
    _sWorkingDirectory = strDirectory;
    _code->setWorkingDirectory(_sWorkingDirectory);
    if (_blog)
        cout << " *** _sworkingDirectory [" << _sWorkingDirectory << "]"
             << endl;
}
void TCluster::setRemoteWorkingDirectory()
{
    //REMOTE
    if (_sConnectionProtocol == "ssh")
    {
        sftp_mkdir(*_ptrsftp,_sRemoteWorkingDirectory,0755);
        sftp_mkdir(*_ptrsftp,gSystem->ConcatFileName(_sRemoteWorkingDirectory,"URANIE"),0755);
        //sftp_rmdir(*_ptrsftp,gSystem->ConcatFileName(gSystem->ConcatFileName(_sRemoteWorkingDirectory,"URANIE"),_sJobName));
        TString command = (TString("rm -rf ")+gSystem->ConcatFileName(gSystem->ConcatFileName(_sRemoteWorkingDirectory,"URANIE"),_sJobName));
        execRemoteCommand(command);
        sftp_mkdir(*_ptrsftp,gSystem->ConcatFileName(gSystem->ConcatFileName(_sRemoteWorkingDirectory,"URANIE"),_sJobName),0755);
    }
    else if (_sConnectionProtocol == "ecaccess") {
        gSystem->Exec(TString("ecaccess-file-mdelete -force ") + gSystem->ConcatFileName(gSystem->ConcatFileName(_sRemoteWorkingDirectory,"URANIE"),_sJobName+"/*"));
        gSystem->Exec("ecaccess-file-mkdir " + _sRemoteWorkingDirectory);
        gSystem->Exec(TString("ecaccess-file-mkdir ") + gSystem->ConcatFileName(_sRemoteWorkingDirectory,"URANIE") );

    }
    _sRemoteWorkingDirectory=gSystem->ConcatFileName(gSystem->ConcatFileName(_sRemoteWorkingDirectory,"URANIE"),_sJobName);

}

Int_t TCluster::sendInputFiles(Int_t startPat ,Int_t endPat)
{
    TString remoteWorkingJob=_sRemoteWorkingDirectory;
    TString command;
    Int_t access_type = O_RDWR | O_CREAT | O_TRUNC;
    const char * remoteFolder;
    const char * localFolder;

    for (Int_t indx = startPat; indx < endPat; indx++)
    {
        remoteFolder = gSystem->ConcatFileName(_sRemoteWorkingDirectory,
                                               TString("UranieLauncher_") + std::to_string(indx + 1).c_str());
        sftp_mkdir(*_ptrsftp,remoteFolder, 0755);
    }
    for (Int_t indx = startPat; indx < endPat; indx++)
    {
        remoteFolder = gSystem->ConcatFileName(_sRemoteWorkingDirectory,
                                               TString("UranieLauncher_") + std::to_string(indx + 1).c_str());
        remoteFolder= gSystem->ConcatFileName(remoteFolder,basename(_code->getInputFile(0)->getFileName()));
        sftp_file file = sftp_open(*_ptrsftp,remoteFolder, access_type, 0755);
        if (file == NULL) {
            cout << "(REMOTE)Can't open file for writing: " << ssh_get_error(*_ptrsshSession) << endl;
            delete[] remoteFolder;
            remoteFolder = NULL;
            return SSH_ERROR;
        }
        //SEND
        localFolder = gSystem->ConcatFileName(_sWorkingDirectory,
                                                           TString("UranieLauncher_") + std::to_string(indx + 1).c_str()); //_code->getReferenceDirectory();

        localFolder=gSystem->ConcatFileName(localFolder,basename(_code->getInputFile(0)->getFileName()));
        ifstream infile(localFolder, ios::binary);
        while (infile)
        {
            char buffer[BUF_SIZE];
            infile.read(buffer, sizeof(buffer));
            if (infile.gcount() > 0) {
                ssize_t nwritten = sftp_write(file, buffer, infile.gcount());
                if (nwritten != infile.gcount()) {
                    cout << "(LOCAL)Can't write data to file: " << ssh_get_error(*_ptrsshSession) << endl;
                    sftp_close(file);
                    delete[] localFolder;
                    localFolder = NULL;
                    return 1;
                }
            }
        }
        sftp_close(file);
    }
    return SSH_OK;

}

Int_t TCluster::retrieveInputFiles(const char * launcherFolder)
{
    sftp_dir dir;
    sftp_attributes attributes;
    int access_type = O_RDONLY;
    TString remoteDir = Form("%s/%s", _sRemoteWorkingDirectory.Data(), launcherFolder);
    dir = sftp_opendir(*_ptrsftp, remoteDir );
    if (!dir)
    {
        cout << "(REMOTE)Can't open dir for reading: " << ssh_get_error(*_ptrsshSession) << endl;
    }
    while ((attributes = sftp_readdir(*_ptrsftp, dir)) != NULL)
    {
        TString filename = attributes->name;
        if (filename.EndsWith(".in") || filename.EndsWith("."))
            continue;
        sftp_file file = sftp_open(*_ptrsftp,gSystem->ConcatFileName(remoteDir, filename.Data()), access_type, 0755);
        if (file == nullptr)
        {
            cout << "(REMOTE)Can't open file for reading: " << ssh_get_error(*_ptrsshSession) << endl;
            delete[] launcherFolder;
            launcherFolder = NULL;
            cout << SSH_ERROR << endl ;
        }
        ofstream outfile(gSystem->ConcatFileName(_sWorkingDirectory,gSystem->ConcatFileName(launcherFolder, filename.Data())),ios::binary | ios::out);
        while (outfile)
        {
            char buffer[BUF_SIZE];
            ssize_t nbytes = sftp_read(file, buffer, sizeof(buffer));
            if (nbytes == 0)
                break; // EOF
            else if (nbytes < 0)
            {
                cout << "Error while reading file:" << ssh_get_error(*_ptrsshSession) << endl;
                sftp_close(file);
            }
            outfile.write(buffer, nbytes);
        }
        sftp_attributes_free(attributes);
        sftp_close(file);
    }
    sftp_closedir(dir);
    return 1;

}
TString TCluster::submitJob(TString command)
{
    execRemoteCommand(command);
     if (_sSshOutput.Contains("Submitted batch job"))
    {
        TObjArray * aux= _sSshOutput.Tokenize(" ");
        TString jobid = TString(((TObjString *)(aux->Last()))->String());  //std::string jobid = TString(((TObjString *)(aux->Last()))->String()).Data();
        while (!jobid.IsDigit())
        {
            jobid.Chop();
        }
        return jobid;
        //setJobID(jobid);
    } else
        {
        cout << "[CRITICAL] Unable to retrieve JOBID closing URANIE." << endl << endl;
        exit(-1);
     }
}
#endif

///////////////////////////////////////////////////////////
//  Construct and return the list of node
///////////////////////////////////////////////////////////

void TCluster::addNodeName(const TString& snodename, Int_t n)
{
    Int_t ntimes = n > 0 ? n : _nnode;

    for (Int_t i = 0; i < ntimes; i++)
    {
        _slistOfNodes += snodename.Data();
        _slistOfNodes += " ";
    }
}

Int_t TCluster::getProcsPerJob(Char_t separator, const Char_t* procs) const
{
    const Char_t* procptr = procs;
    Int_t procsperjob = 1;
    //counts the number of separators in
    // the string to define number of procs per job
    //WARNING : it works only if string is properly structured
    // such as 'host1,host2,host3' with separator ==','
    while (*procptr != '\0')
    {
        if (*procptr == separator)
            procsperjob++;
        procptr++;
    }
    return procsperjob;
}

#ifdef WITH_LIBSSH
void TCluster::replaceHeaderKey(std::string& source, const std::string& from, const std::string& to)
{
        std::string newString;
        newString.reserve(source.length()-from.length()+to.length());  // avoids a few memory allocations

        std::string::size_type lastPos = 0;
        std::string::size_type findPos;

        while(std::string::npos != (findPos = source.find(from, lastPos)))
    {
        newString.append(source, lastPos, findPos - lastPos);
        newString += to;
        lastPos = findPos + from.length();
    }

    // Care for the rest after last occurrence
    newString += source.substr(lastPos);
    source.swap(newString);
}
#endif
////
// Virtual Functions
///



///////////////////////////////////////////////////////////
//  Log
///////////////////////////////////////////////////////////
void TCluster::printLog(Option_t *option)
{
    cout << " ** TCluster::printLog[" << option << "]" << endl;
    cout << " ** Name[" << GetName() << "] Title[" << GetTitle() << "] Node["
            << _nnode << "]" << endl;
    cout << " ** _slistOfNodes[" << _slistOfNodes << "]" << endl;
}

} // Fin du namespace Launcher
} // Fin du namespace URANIE

