//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TCodeFMU.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Created : 2021/09/10 at 12:10:53

 \brief Implementation of the class URANIE::Launcher::TCodeFMU
 */

// ROOT
#include "TRegexp.h"

// URANIE
#include "TCodeFMU.h"

#include <sstream>


ClassImp (URANIE::Launcher::TCodeFMU)

namespace URANIE {
namespace Launcher {

// -----------------------------------------------------------------------
//  Default constructor
// -----------------------------------------------------------------------
TCodeFMU::TCodeFMU ():
    TCode(), _sFMUFile(""), _sFMUWorkingDirectory(""),
    _logLevel(log_level_verbose),
    _dStart(0.0), _dStop(1.0), _dStep(1.0e-3), _dTolerance(1.0e-4),
    _bStart(kFALSE), _bStop(kFALSE), _bStep(kFALSE), _bTolerance(kFALSE),
    _binitIFirst(kTRUE),
    _fmi1Import(nullptr), _fmi2Import(nullptr),
    _sGenerationTool(""), _sGenerationDateTime("")
{
    if (_blog)
        cout << "** TCodeFMU Default Constructor with name[" << GetName() << "] t       itle[" << GetTitle() << "]" << endl;
}

// -----------------------------------------------------------------------
//  Constructor with name AND title
// -----------------------------------------------------------------------
TCodeFMU::TCodeFMU(const TString &name, const TString &title) :
    TCode(name, title), _sFMUFile(""), _sFMUWorkingDirectory(""),
    _logLevel(log_level_verbose),
    _dStart(0.0), _dStop(1.0), _dStep(1.0e-3), _dTolerance(1.0e-4),
    _bStart(kFALSE), _bStop(kFALSE), _bStep(kFALSE), _bTolerance(kFALSE),
    _binitIFirst(kTRUE),
    _fmi1Import(nullptr), _fmi2Import(nullptr),
  _sGenerationTool(""), _sGenerationDateTime("")
{
    FUNCT_TRACE();
    if (_blog)
        cout << "** TCodeFMU constructor with name[" << GetName() << "] title[" << GetTitle() << "]" << endl;
}

// -----------------------------------------------------------------------
//  Constructor with a TDataServer
// -----------------------------------------------------------------------
TCodeFMU::TCodeFMU(URANIE::DataServer::TDataServer *tds, TString scmd, Option_t *option) :
    TCode(tds, "pwd && this string is note used && ls -alhtr"),
     _sFMUFile(""), _sFMUWorkingDirectory(""),
    _logLevel(log_level_verbose),
    _dStart(0.0), _dStop(1.0), _dStep(1.0e-3), _dTolerance(1.0e-4),
    _bStart(kFALSE), _bStop(kFALSE), _bStep(kFALSE), _bTolerance(kFALSE),
    _binitIFirst(kTRUE),
    _context_FMI(nullptr), _fmi1Import(nullptr), _fmi2Import(nullptr),
    _sGenerationTool(""), _sGenerationDateTime("")
{
    FUNCT_TRACE();
    try {
        if ( _tds == nullptr)
            throw URANIE::Exceptions::UErrorExceptions(
                    __FILE__,
                    __LINE__,
                    Form("Method TCodeFMU::%s :\n The type TDS is empty.",
                         __func__));
        if (_blog)
            cout << " *** TCodeFMU::Constructor with TDS [" << tds->GetName() << "] name[" << GetName() << "] title[" << GetTitle() << "]" << endl;

        // Verify the FMU File
        if (gSystem->IsAbsoluteFileName(scmd.Data())) {
            _sFMUFile = scmd;
        } else {
            // buffer used to prevent memleak, as gSystem::ConcatFileName result must be deleted
            char *foo=NULL;
            foo = gSystem->ConcatFileName(_sReferenceDirectory.Data(), scmd.Data());
            _sFMUFile = foo;
            delete[] foo; foo=NULL;
        }

        if (_blog)
            cout << " *** _sFMUFile[" << _sFMUFile << "]" << endl;


        TString soption = option;
        soption.ToLower();
        if (_blog)
            cout << " *** soption[" << soption << "]" << endl;
        treatOption(soption);

        if (_blog)
            cout << " *** End of TCodeFMU::Constructor" << endl;
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}


TCodeFMU::~TCodeFMU ()
{
    FUNCT_TRACE();
    if (_blog)
        cout << " *** ==============" << endl <<
                " *** Method [TCodeFMU::" << __func__  << "] Name[" << GetName() << "]" << endl;
    cout << " *** pwd[" << gSystem->pwd() << "]" << endl;

    if (_blog)
        cout << " *** End of Method [TCodeFMU::" << __func__  <<
                "] Name[" << GetName() << "]" << endl <<
                " *** ==============" << endl;
}

//  
// Methods
//  

// ----------------------------------------------------------------------
//  Set the FMU File
// ----------------------------------------------------------------------
void TCodeFMU::setFMUFile(TString sFMUFile, Option_t *option)
{
    FUNCT_TRACE();
    if (_blog)
        cout << " *** Method [TCodeFMU::" << __func__  << "] with FMUFile[" << sFMUFile << "] Option[" << option << "]" << endl;

    // Add the FMU Input File
    addInputFile( new TInputFileFMU(_sFMUFile));

    if (_blog) {
        _listOfInputFiles->Print("*");

        TInputFile *tfmufile = getInputFile(_sFMUFile);
        tfmufile->printLog();
    }

    // Add the Output File as the same file than the FMU Input file
    addOutputFile( new TOutputFileFMU (getInputFile(_sFMUFile)->getFileName()));

    if (_blog)
        cout << " *** End of Method [TCodeFMU::" << __func__  << "] with FMUFile[" << sFMUFile << "] Option[" << option << "]" << endl;
}


// ----------------------------------------------------------------------
//  initIFirst Method
// ----------------------------------------------------------------------
void TCodeFMU::initIFirst(Option_t *option)
{
    TString soption = option ;
    soption.ToLower();
    FUNCT_TRACE();

    if (_blog)
        cout << " *** Method TCodeFMU::" << __func__ <<
                " with soption[" << soption << "]" <<
                " _nCurrentSimulation["<< _nCurrentSimulation << "]" << endl;

    // At least apply one time this method
    setWorkingDirectory(_sWorkingDirectory);
    if ( _blog)
        cout << " *** _sWorkingDirectory[" << _sWorkingDirectory << "]" << endl;

    try {
        // Create the _sWorkingDirectory if it doesn't exist
        createDirectory(_sWorkingDirectory);

        // Create the "UranieResult" directory if it doesn't exist
        createDirectory(_sWorkingDirectory, "UranieResults");

        // Create the launching DEFAULT_WORKING_DIRECTORY directory if it doesn't exist
        createDirectory(_sWorkingDirectory, DEFAULT_FMU_WORKING_DIRECTORY);

        // set the _sFMUWorkingDirector attribute
        _sFMUWorkingDirectory = gSystem->pwd();

        if ( _blog) {
            cout << endl << " *** =======================" << endl;
            cout << " *** _sFMUFile(" << _sFMUFile << ")" << endl;
            cout << " *** _sFMUWorkingDirectory(" << _sFMUWorkingDirectory << ")" << endl;
        }

        TInputFileFMU * tipfmu = dynamic_cast<TInputFileFMU *>(getInputFile(_sFMUFile));

        if ( !tipfmu )
            throw URANIE::Exceptions::UErrorExceptions(
                    __FILE__,
                    __LINE__,
                    Form("TCodeFMU::%s Method\n The FMU Input file \"%s\" not present",
                         __func__, _sFMUFile.Data()));

        if ( _blog)
            tipfmu->printLog();

        // Init the TInputFileFMU at WorkingDirectory
        tipfmu->setWorkingDirectory(_sFMUWorkingDirectory);
        if (_blog)
            cout << " *** _logLevel [" << _logLevel << "]" << endl;
        switch (_logLevel) {
        case log_level_nothing:
            tipfmu->getCallbackFMI()->log_level = jm_log_level_nothing;
            break;
        case log_level_fatal:
            tipfmu->getCallbackFMI()->log_level = jm_log_level_fatal;
            break;
        case log_level_error:
            tipfmu->getCallbackFMI()->log_level = jm_log_level_error;
            break;
        case log_level_warning:
            tipfmu->getCallbackFMI()->log_level = jm_log_level_warning;
            break;
        case log_level_info:
            tipfmu->getCallbackFMI()->log_level = jm_log_level_info;
            break;
        case log_level_verbose:
            tipfmu->getCallbackFMI()->log_level = jm_log_level_verbose;
            break;
        case log_level_debug:
            tipfmu->getCallbackFMI()->log_level = jm_log_level_debug;
            break;
        case log_level_all:
            tipfmu->getCallbackFMI()->log_level = jm_log_level_all;
            break;
        default:
            break;
        }

        tipfmu->init();
        if ( _blog)
            cout << endl << " ******* " << endl;
        fmi_version_enu_t eVersion = fmi_version_unknown_enu;
        if ( _blog)
            cout << " *** eVersion(" << eVersion << ")" << endl;

        try {

            // Get the FMI informations
            _dFMUStart     = tipfmu->getStartTime();
            _dFMUStop      = tipfmu->getStopTime();
            _dFMUStep      = tipfmu->getStepTime();
            _dFMUTolerance = tipfmu->getTolerance();
            eVersion       = tipfmu->getFMUVersion();

            // in case of not defined by user
            if (!_bStart)
                _dStart = _dFMUStart;
            if (!_bStop)
                _dStop = _dFMUStop;
            if(!_bStep)
                _dStep = _dFMUStep;
            if(!_bTolerance)
                _dTolerance = _dFMUTolerance;

            throw URANIE::Exceptions::UInfoExceptions(
                    __FILE__,
                    __LINE__,
                    Form("Method TCodeFMU::%s :\n The values used for the simulation are :\n"
                         "  Start   = %e\n"
                         "  Stop    = %e\n"
                         "  Tolearnce = %e\n"
                         "  Step      = %e",
                         __func__, _dStart, _dStop, _dTolerance, _dStep));

        } catch (URANIE::Exceptions::UInfoExceptions& ue)
        {
            ue.printMessage();
        }


        if ( _blog)
            cout << " *** eVersion(" << eVersion << ") string[" << fmi_version_to_string(eVersion) << "]" << endl;

        TOutputFileFMU *topfmi = dynamic_cast<TOutputFileFMU *>(getOutputFile(_sFMUFile));

        if ( !topfmi ) {
            // Add the FMU file as TOutputFileFMU
            addOutputFile( new TOutputFileFMU(_sFMUFile) );

            topfmi = dynamic_cast<TOutputFileFMU *>(getOutputFile(_sFMUFile));

            if ( !topfmi )
                throw URANIE::Exceptions::UErrorExceptions(
                        __FILE__,
                        __LINE__,
                        Form("TCodeFMU::%s Method\n The Output file \"%s\" can be added as Output file",
                             __func__, _sFMUFile.Data()));

        }


        topfmi->setFMUVersion(eVersion);
        switch (eVersion) {
        case fmi_version_1_enu:
            topfmi->setFMU1Import( tipfmu->getFMU1Import());
            break;
        case fmi_version_2_0_enu:
            topfmi->setFMU2Import( tipfmu->getFMU2Import());
            break;
        case fmi_version_unknown_enu:
            throw URANIE::Exceptions::UErrorExceptions(
                    __FILE__,
                    __LINE__,
                    Form("TCodeFMU::%s Method\n The FMI Version is already unknown.", __func__));
            break;
        default:
            throw URANIE::Exceptions::UErrorExceptions(
                    __FILE__,
                    __LINE__,
                    Form("TCodeFMU::%s Method\n Only versions 1.0 and 2.0 are supported so far. \n"
                         " Here, for the FMUFile \"%s\", the version is %d.",
                         __func__, _sFMUFile.Data(), eVersion));
            break;
        }
        if ( _blog)
            cout << " *** topfmi->getNOutputVector()[" << topfmi->getNOutputVector() << "]" << endl;

        if ( topfmi->getNOutputVector() > 0 ) {
            URANIE::DataServer::TAttribute *attTime = new URANIE::DataServer::TAttribute("__time__", URANIE::DataServer::TAttribute::kVector);
            topfmi->addAttribute(attTime);
            attTime->delShare();
            tipfmu->setListOutputVector( topfmi->getListOutputVector());
        }

        if ( _blog)
            tipfmu->printLog();

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }


    // End of initIFirst
    _binitIFirst = kFALSE;

    if ( _blog )
        cout << " *** End of Method TCodeFMU::" << __func__ << " with soption[" << soption << "]" << endl;

}

// ----------------------------------------------------------------------
//  Init Method
// ----------------------------------------------------------------------
void TCodeFMU::init(Option_t *option)
{
    FUNCT_TRACE();
    TString soption = option ;
    soption.ToLower();


    if (_blog)
        cout << " *** Method TCodeFMU::" << __func__ << " with soption[" << soption << "]_nCurrentSimulation["<< _nCurrentSimulation << "]" << endl;


    try {
        if (_binitIFirst)
            initIFirst(option);

        // Create the launching "UranieLauncher_i" directory
        createDirectory(_sWorkingDirectory, Form("UranieLauncher_%d", _nCurrentSimulation));
        if (_blog)
            cout << " --- PWD[" << gSystem->pwd() << "]" << endl;

        // Copy Files step
        copyFiles(option);
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    if (_blog)
        cout << " *** End of Method TCodeFMU::" << __func__ << " with soption[" << soption << "]" << endl;
}

// ----------------------------------------------------------------------
//  Run Method
// ----------------------------------------------------------------------
void TCodeFMU::run(Option_t *option)
{
    FUNCT_TRACE();
    TString soption = option ;
    soption.ToLower();

    if (_blog)
        cout << " ********************************" << endl <<
                " *** Method TCodeFMU::" << __func__ <<
                " with soption[" << soption << "]" << endl <<
                " *** _nCurrentSimulation[" << _nCurrentSimulation << "]" << endl;


    // Case of never passed into the TCodeFMU::init method
    if ( _binitIFirst )
        init(option);

    try {
        TInputFileFMU * tipfmu = dynamic_cast<TInputFileFMU *>(getInputFile(_sFMUFile));

        if ( !tipfmu )
            throw URANIE::Exceptions::UErrorExceptions(
                    __FILE__,
                    __LINE__,
                    Form("TCodeFMU::%s Method\n The FMU Input file \"%s\" not found",
                         __func__, _sFMUFile.Data()));


        if (_blog)
            cout << " *** Version returned from FMU:" << tipfmu->getFMUVersionString() << endl;
        if (_blog)
            cout << " *** _logLevel [" << _logLevel << "]" << endl;
        switch (_logLevel) {
        case log_level_nothing:
            tipfmu->getCallbackFMI()->log_level = jm_log_level_nothing;
            break;
        case log_level_fatal:
            tipfmu->getCallbackFMI()->log_level = jm_log_level_fatal;
            break;
        case log_level_error:
            tipfmu->getCallbackFMI()->log_level = jm_log_level_error;
            break;
        case log_level_warning:
            tipfmu->getCallbackFMI()->log_level = jm_log_level_warning;
            break;
        case log_level_info:
            tipfmu->getCallbackFMI()->log_level = jm_log_level_info;
            break;
        case log_level_verbose:
            tipfmu->getCallbackFMI()->log_level = jm_log_level_verbose;
            break;
        case log_level_debug:
            tipfmu->getCallbackFMI()->log_level = jm_log_level_debug;
            break;
        case log_level_all:
            tipfmu->getCallbackFMI()->log_level = jm_log_level_all;
            break;
        default:
            break;
        }

//        tipfmu->getCallbackFMI()->log_level = jm_log_level_all;

        tipfmu->setStartTime(_dStart);
        tipfmu->setStopTime(_dStop);
        tipfmu->setStepTime(_dStep);
        tipfmu->setTolerance(_dTolerance);

        if (_blog) {
            cout << " ***    Default experiment setup : Start[" << _dFMUStart << "]" <<
                    " Stop[" << _dFMUStop << "]" <<
                    " Step[" << _dFMUStep << "]" <<
                    " Tolerance[" << _dFMUTolerance << "]" << endl;
            cout << " *** Simulation experiment setup : Start[" << _dStart << "]" <<
                    " Stop[" << _dStop << "]" <<
                    " Step[" << _dStep << "]" <<
                    " Tolerance[" << _dTolerance << "]" << endl;

            cout << endl << " ===== Begin  Simulate ==== " << endl;
        }
        tipfmu->runFMI(option);
        if (_blog)
            cout << " ===== End of Simulate ==== " << endl << endl;

        // Add informations to the TOuputFMU file from the TInputFMU file
        if (_blog)
            cout << " ===== Pass info Input -> Output  ==== " << endl << endl;
        TOutputFileFMU *topfmi = dynamic_cast<TOutputFileFMU *>(getOutputFile(_sFMUFile));

        if ( !topfmi )
            throw URANIE::Exceptions::UErrorExceptions(
                    __FILE__,
                    __LINE__,
                    Form("TCodeFMU::%s Method\n The FMU Output file \"%s\" not found",
                         __func__, _sFMUFile.Data()));

        topfmi->setVectorValues( tipfmu->getVectorValues());
        if (_blog)
            cout << " ===== End of Pass info ==== " << endl << endl;

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    if (_blog)
        cout << endl <<
                " *** End of Method TCodeFMU::" << __func__ <<
                " with soption[" << soption << "]" << endl <<
                " *** _nCurrentSimulation[" << _nCurrentSimulation << "]" << endl <<
                " ********************************" << endl;
}

// ----------------------------------------------------------------------
//  Terminate Method
// ----------------------------------------------------------------------
void TCodeFMU::terminate(Option_t *option)
{
    FUNCT_TRACE();
    TString soption = option ;
    soption.ToLower();

    if (_blog)
        cout << " ** Method TCodeFMU::" << __func__ << " with soption[" << soption << "]" << endl;

    // Clean
    if (_blog)
        cout << " *** clean ... : ";
    TInputFileFMU * tipfmu = dynamic_cast<TInputFileFMU *>(getInputFile(_sFMUFile));
    fmi_version_enu_t eVersion = fmi_version_unknown_enu;
    eVersion       = tipfmu->getFMUVersion();


    switch (eVersion) {
    case fmi_version_1_enu: {
        fmi1_import_destroy_dllfmu(_fmi1Import);

        if (_fmi1Import != nullptr)
            fmi1_import_free(_fmi1Import);
        _fmi1Import=nullptr;
        break;
    }
    case fmi_version_2_0_enu: {
        fmi2_import_destroy_dllfmu(_fmi2Import);

        if (_fmi2Import != nullptr)
            fmi2_import_free(_fmi2Import);
        _fmi2Import=nullptr;
        break;
    }
    default:
        break;
    }

   gSystem->cd(_sReferenceDirectory.Data());

   if (_blog)
       cout << " *** End of Method TCodeFMU::" << __func__ << " with soption[" << soption << "]" << endl;
}

// ----------------------------------------------------------------------
// Printing Log files
// ----------------------------------------------------------------------
void TCodeFMU::printLog(Option_t *option)
{
    FUNCT_TRACE();
    TCode::printLog(option);
    cout << " *** _sFMUFile[" << _sFMUFile << "]" << endl;
    cout << " *** _sFMUWorkingDirectory[" << _sFMUWorkingDirectory << "]" << endl;
    cout << " *** _sVersion [" << _sVersion << "]" << endl;
    cout << " *** _logLevel [" << _logLevel << "]" << endl;
    cout << " *** _sGenerationTool [" << _sGenerationTool << "]" <<
            " _sGenerationDateTime [" << _sGenerationDateTime << "]" << endl;
    cout << " ***    Default experiment setup : Start[" << _dFMUStart << "]" <<
            " Stop[" << _dFMUStop << "]" <<
            " Step[" << _dFMUStep << "]" <<
            " Tolerance[" << _dFMUTolerance << "]" << endl;
    cout << " *** Simulation experiment setup : Start[" << _dStart << "]" <<
            " Stop[" << _dStop << "]" <<
            " Step[" << _dStep << "]" <<
            " Tolerance[" << _dTolerance << "]" << endl;

}

// Accessor methods
//  


// Other methods
//  



// ----------------------------------------------------------------------
//  treat the TString option
// ----------------------------------------------------------------------
void TCodeFMU::treatOption(TString soption)
{
    if (_blog)
        cout << " *** Method [TCodeFMU::" << __func__  << "] with Option[" << soption << "]" << endl;

    // Replace all " ="
    soption.ReplaceAll(" =", "=");
    soption.ReplaceAll("  =", "=");
    soption.ReplaceAll("   =", "=");
    soption.ReplaceAll("= ", "=");
    soption.ReplaceAll("=  ", "=");
    soption.ReplaceAll("=   ", "=");
    soption.ReplaceAll(" ", ",");
    soption.ReplaceAll(",,", ",");
    soption.ReplaceAll(",", " ");
    if (_blog)
        cout << " *** soption[" << soption << "]" << endl;

    istringstream inStream((char *) soption.Data());
    string inName;

    while (inStream >> inName) {
        if (_blog)
            cout << " *** case of inName[" << inName << "]";
        TString sCrtOption(inName);
        if (_blog)
            cout << " --- sCrtOption[" << sCrtOption << "]" << endl;

        // Case of LogLevel
        if ( sCrtOption.Contains("loglevel=")) {
            if (_blog)
                cout << " *** contains LogLevel FMI option with _logLevel [" << _logLevel << "]" << endl;
            TRegexp reg("loglevel[ ]*=[ ]*[a-z]*");
            TString out = sCrtOption(reg);
            TString snclass = out.Remove(0, 9);
            if (_blog)
                cout << " *** --- snclass[" << snclass << "]" << endl;
            if ( snclass.EqualTo("nothing")) {
                _logLevel = log_level_nothing;
            } else if (snclass.EqualTo("fatal")) {
                _logLevel = log_level_fatal;
            } else if (snclass.EqualTo("error") ) {
                _logLevel = log_level_error;
            } else if (snclass.EqualTo("warning") ) {
                _logLevel = log_level_warning;
            } else if (snclass.EqualTo("info") ) {
                _logLevel = log_level_info;
            } else if (snclass.EqualTo("verbose") ) {
                _logLevel = log_level_verbose;
            } else if (snclass.EqualTo("debug") ) {
                _logLevel = log_level_debug;
            } else if (snclass.EqualTo("all") ) {
                _logLevel = log_level_all;
            } else  {
                cout << " *** ERROR\n     Keyword [" << snclass << "] not known between [nothing|fatal|error|warning|info|verbose|debug|all] AND not treated" << endl;
            }
            if (_blog)
                cout << " *** final _logLevel [" << _logLevel << "]" << endl;
        } // end of Case of logLevel
    }
}

} // End of namespace Launcher
} // End of namespace URANIE
// End of File

