//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TInputFileFMU.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Created : 2021/09/10 at 13:42:45

 \brief Implementation of the class URANIE::Launcher::TInputFileFMU
 */
#include <iostream>
#include<fstream>
using std::cout;
using std::endl;

// ROOT
#include "TMath.h"

// URANIE
#include "UExceptions.h"
#include "TInputFileFMU.h"



ClassImp (URANIE::Launcher::TInputFileFMU)

namespace URANIE {
namespace Launcher {

// -----------------------------------------------------------------------
//  Constructor with filename
// -----------------------------------------------------------------------
TInputFileFMU::TInputFileFMU(TString str) :
        TInputFile(str), _sFMUWorkingDirectory(DEFAULT_FMU_WORKING_DIRECTORY),
        _callbacks_FMI(0), _context_FMI(nullptr), _sVersion(fmi_version_unknown_enu),
        _fmi1Import(nullptr), _fmi2Import(nullptr),
        _sPlatformName(""), _sModelName(""), _sModelVersion(""), _sGUID(""),
        _dFMUStep(-1.0),
        _vecOutputs(0), _vecVROutputs(0), _no(0)
{
    FUNCT_TRACE();
    if (_blog)
        cout << " *** Constructor TInputFileFMU file[" << str << "]" << endl;
    SetTitle(Form("TInputFileFMU with name[%s]", str.Data()));
    setSeparatorCharacter(" \t\r\n");
    _sstrbeg = _sstrend = "";
    _svecbeg = _svecend = ""; _svecdel = " ";
    if(_blog)
        cout << " *** getFileName(" << getFileName() << ")" << endl;

    // Create the callback
    _callbacks_FMI = (jm_callbacks *) malloc(sizeof(jm_callbacks));
    _callbacks_FMI->malloc = malloc;
    _callbacks_FMI->calloc = calloc;
    _callbacks_FMI->realloc = realloc;
    _callbacks_FMI->free = free;
    _callbacks_FMI->logger = jm_default_logger; // importlogger; jm_default_logger
    _callbacks_FMI->log_level = jm_log_level_all;
    _callbacks_FMI->context = 0;

    // init the _sFMUWorkingDirectory attribute
    setWorkingDirectory();
}

TInputFileFMU::~TInputFileFMU()
{
    if (_blog)
        cout << " ********************************" << endl <<
                " *** Method [TInputFileFMU::" << __func__  << "] Name[" << GetName() << "]" << endl;

    // Clean
    jm_status_enu_t eStatusFMI;
    if (0) {
        eStatusFMI = fmi_import_rmdir(_callbacks_FMI, _sFMUWorkingDirectory.Data());
        if (eStatusFMI == jm_status_error) {
            printf("Could not remove directory[%s].\n", _sFMUWorkingDirectory.Data());
        }
    }

    if (_fmi1Import != nullptr)
        fmi1_import_free(_fmi1Import);
    _fmi1Import=nullptr;

    if (_fmi2Import != nullptr)
        fmi2_import_free(_fmi2Import);
    _fmi2Import=nullptr;

    if ( _context_FMI != nullptr)
        fmi_import_free_context(_context_FMI);
    _context_FMI = nullptr;

    if ( _callbacks_FMI != nullptr )
        free( _callbacks_FMI);

    if (_blog)
        cout << " *** End of Method [TInputFileFMU::" << __func__  << "] Name[" << GetName() << "]" << endl <<
                " ********************************" << endl;
}

// ----------------------------------------------------------------------
//  addAttribute Method
// ----------------------------------------------------------------------
void TInputFileFMU::addAttribute(URANIE::DataServer::TAttribute* att)
{
    FUNCT_TRACE();

    if (_blog)
        cout << " ********************************" << endl <<
                " *** Method [TInputFileFMU::" << __func__  << "]" <<
                "      attribute[" << att->GetName() << "]" <<
                "      TOutputFileFMU::FileName[" << _sFileName << "]" << endl;


    try
    {
        if ((att->getKeyList().size() == 0))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__,
                                                       __LINE__,
                                                       Form("TInputFileFMU::addAttribute ERROR:\n"
                                                            " the attribute \"%s\" has keys or positions (lines and fields) not defined",
                                                            att->GetName()));
        if ( att->getDataType() == URANIE::DataServer::TAttribute::kVector )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__,
                                                       __LINE__,
                                                       Form("TInputFileFMU::addAttribute ERROR:\n"
                                                            " the type of attribute \"%s\" is \"kVector\" not take into account.",
                                                            att->GetName()));
        if ( att->getDataType() == URANIE::DataServer::TAttribute::kString )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__,
                                                       __LINE__,
                                                       Form("TInputFileFMU::addAttribute ERROR:\n"
                                                            " the type of attribute \"%s\" is \"kString\" not take into account.",
                                                            att->GetName()));
        if ( att->getDataType() == URANIE::DataServer::TAttribute::kCategory )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__,
                                                       __LINE__,
                                                       Form("TInputFileFMU::addAttribute ERROR:\n"
                                                            " the type of attribute \"%s\" is \"kCategory\" not take into account.",
                                                            att->GetName()));
        // Finally, add the attribute if not founded
        if( find(_attributes.begin(), _attributes.end(), att) == _attributes.end() )
            TCodeFile::addAttribute(att);
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    if (_blog)
        cout << endl <<
                " *** End of Method [TInputFileFMU::" << __func__  << "]" <<
                "      attribute[" << att->GetName() << "]" <<
                "      TInputFileFMU::FileName[" << _sFileName << "]" << endl <<
                " ********************************" << endl;


} // End of addAttribute Method

// ----------------------------------------------------------------------
//  Init Method
// ----------------------------------------------------------------------
void TInputFileFMU::init()
{
    FUNCT_TRACE();
    if (_blog)
        cout << " *** Method TInputFileFMU::" << __func__ << endl;
    try
    {

        if ( _sVersion == fmi_version_unknown_enu ) {
            // init the FMI
            initFMI();

            if ( ( _sVersion == fmi_version_1_enu ) ||
                 ( _sVersion == fmi_version_2_0_enu ) ) {
                // Make the searchIndexeAttribute action
                Int_t natt = (Int_t)(_attributes.size());
                if (_blog)
                    cout << " *** NumberOfAttributes[" << natt << "] in FileName[" << GetName() << "]" << endl;
                for (Int_t i = 0; i < natt; i++) {
                    searchIndexeAttribute(_attributes[i]);
                }
            } // End of _sVersion == [ fmi_1 || fmi_2 ]
        } // End of _sVersion == fmi_version_unknown_enu
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    if (_blog)
        cout << " *** End of Method TInputFileFMU::" << __func__ << endl;
}

// ----------------------------------------------------------------------
//  terminate Method
// ----------------------------------------------------------------------
void TInputFileFMU::terminate()
{
    if (_blog)
        cout << " *** Method TInputFileFMU::" << __func__ << endl;
}

// ----------------------------------------------------------------------
//  initFMI Method
// ----------------------------------------------------------------------
void TInputFileFMU::initFMI(Option_t *option)
{
    TString soption = option ;
    soption.ToLower();
    FUNCT_TRACE();
    if (_blog)
        cout << " *** Method TInputFileFMU::" << __func__ << " with soption[" << soption << "]" << endl;

    if (_blog)
        cout << " *** function[" << __func__ << "] _callbacks_FMI->log_level(" << _callbacks_FMI->log_level << ")" << endl;

    // create the FMI context
    _context_FMI = fmi_import_allocate_context(_callbacks_FMI);

    if (_blog)
        cout << " *** _sFMUWorkingDirectory(" << _sFMUWorkingDirectory << ")" << endl;
    try {

        if (gSystem->AccessPathName(getFileName(), kFileExists))
          throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                Form("TInputFileFMU::%s ERROR :\n File \"%s\" not found.",
                                                     __func__, getFileName().Data()));


        // Creates the directory if it doesn't exist
        void * openFileResult = gSystem->OpenDirectory(_sFMUWorkingDirectory.Data());
        try {
            if (openFileResult == 0) {
                throw URANIE::Exceptions::UInfoExceptions(
                            __FILE__,
                            __LINE__,
                            Form("TInputFileFMU::initFMI Method\n The directory \"%s\" does not exist\n URANIE creates it for you", _sFMUWorkingDirectory.Data()));
            }
        } catch (URANIE::Exceptions::UExceptions& ue)
        {
            ue.printMessage();
            if (gSystem->mkdir(_sFMUWorkingDirectory, kTRUE) == -1)
                throw URANIE::Exceptions::UErrorExceptions(
                        __FILE__,
                        __LINE__,
                        Form("TInputFileFMU::%s ERROR\n : The directory \"%s\" can not be created",
                             __func__, _sFMUWorkingDirectory.Data()));
        }
        gSystem->FreeDirectory(openFileResult);

        // Get the FMI version in enum type
        _sVersion = fmi_import_get_fmi_version(_context_FMI, getFileName().Data(), _sFMUWorkingDirectory.Data());

        if (_blog)
            cout << " *** _sVersion(" << _sVersion << ") string[" << fmi_version_to_string(_sVersion) << "]" << endl;

        // init the FMU pointer
        switch (_sVersion) {
        case fmi_version_1_enu: {
            if (_blog)
                cout << endl << " *** FMU case of FMI v1.0" << endl;
            init_fmi1(option);
            break;
        }
        case fmi_version_2_0_enu: {
            if (_blog)
                cout << endl << " *** FMU Case of FMI v2.0" << endl;
            init_fmi2_0(option);
            break;
        }
        default:
            throw URANIE::Exceptions::UErrorExceptions(
                        __FILE__,
                        __LINE__,
                        Form("TInputFileFMU::%s Method\n The FMI Version[%d] not yet implemented.",
                             __func__, _sVersion));
            break;
        } // End of switch _sVersion
    } catch (URANIE::Exceptions::UExceptions& ue)
    {
        ue.printMessage();
        if (ue.getMessageType() > URANIE::Exceptions::UExceptions::kWARNING)
            throw ue;
    }
    if (_blog)
        cout << " *** End Of Method TInputFileFMU::" << __func__ << " with soption[" << soption << "]" << endl;
}

// ----------------------------------------------------------------------
//  init_fmi1 Method
// ----------------------------------------------------------------------
void TInputFileFMU::init_fmi1(Option_t *option)
{
    FUNCT_TRACE();
    TString soption = option ;
    soption.ToLower();
    if (_blog)
        cout << " *** Method TInputFileFMU::" << __func__ << " with soption[" << soption << "]" << endl;

    try {

        _fmi1Import = fmi1_import_parse_xml(_context_FMI, _sFMUWorkingDirectory.Data());

        if(!_fmi1Import) {
            printf("Error parsing XML, exiting\n");
            do_exit(CTEST_RETURN_FAIL);
        }

        // Get the FMI General informations
        _sModelName = fmi1_import_get_model_name(_fmi1Import);
        if (_blog)
            cout << " ***      _sModelName[" << _sModelName << "] (Original)" << endl;

        _sModelName.ReplaceAll(" ", "_");
        _sModelName.ReplaceAll(",", "_");
        if (_blog)
            cout << " *** _sModelName[" << _sModelName << "] (After replace characters)" << endl;

        _sModelVersion = fmi1_import_get_model_version(_fmi1Import);
        if (_blog)
            cout << " *** _sModelVersion[" << _sModelVersion << "]" << endl;

        _sGUID = fmi1_import_get_GUID(_fmi1Import);
        if (_blog)
            cout << " *** _sGUID[" << _sGUID << "]" << endl;

        _sGenerationTool     = fmi1_import_get_generation_tool(_fmi1Import);
        _sGenerationDateTime = fmi1_import_get_generation_date_and_time(_fmi1Import);

        // Get Default Expeiment Values from XML file
        _dFMUStart = fmi1_import_get_default_experiment_start(_fmi1Import);
        _dFMUStop = fmi1_import_get_default_experiment_stop(_fmi1Import);
        _dFMUTolerance = fmi1_import_get_default_experiment_tolerance(_fmi1Import);

        // attribut computed
        _dFMUStep = ( _dFMUStop - _dFMUStart) / 100.0;

        fmi1_import_model_counts_t  mc;
        if (_blog)
            printf(" *** Model name: %s\n", _sModelName.Data());
        fmi1_import_collect_model_counts(_fmi1Import, &mc);
        if (_blog) {
            cout << "  *** num_constants  [" << mc.num_constants << "]" << endl;
            cout << "  *** num_discrete   [" << mc.num_discrete << "]" << endl;
            cout << "  *** num_continuous [" << mc.num_continuous << "]" << endl;
        }

        fmi1_fmu_kind_enu_t fmukind = fmi1_fmu_kind_enu_unknown;

        fmukind = fmi1_import_get_fmu_kind(_fmi1Import);
        if (_blog)
            cout << " *** fmukind[" << fmukind << "] string[" << fmi1_fmu_kind_to_string(fmukind) << "]" << endl;

        if ( fmukind != fmi1_fmu_kind_enu_cs_standalone )
            throw URANIE::Exceptions::UErrorExceptions(
                    __FILE__,
                    __LINE__,
                    Form("Method TInputFileFMU::%s :\n The type \"%s\" not supporte: only \"%s\" supported.",
                         __func__,
                         fmi1_fmu_kind_to_string(fmukind),
                         fmi1_fmu_kind_to_string(fmi1_fmu_kind_enu_cs_standalone)));

        const char* modelIdentifier;
        modelIdentifier = fmi1_import_get_model_identifier(_fmi1Import);

        try {
            throw URANIE::Exceptions::UInfoExceptions(
                        __FILE__,
                        __LINE__,
                        Form("Method TInputFileFMU::%s :\n The default Experiment values readed are :\n"
                             "  FMU file \"%s\"\n"
                             "  Version   = FMI %d\n"
                             "  Type      = %s\n"
                             "  Start     = %e\n"
                             "  Stop      = %e\n"
                             "  Tolerance = %e\n"
                             "  Step      = %e (computed)",
                             __func__, _sFileName.Data(), _sVersion, fmi1_fmu_kind_to_string(fmukind), _dFMUStart, _dFMUStop, _dFMUTolerance, _dFMUStep));
        } catch (URANIE::Exceptions::UInfoExceptions& ue)
        {
            ue.printMessage();
        }

        if (_blog)
            printf(" *** Model Identifier \"%s\" with \"%s\" (%d)\n", modelIdentifier, fmi1_fmu_kind_to_string(fmukind), fmukind);

//        const char* modelIdentifier;
//        fmi1_fmu_kind_enu_t fmukind = fmi1_fmu_kind_unknown;

//        if(fmi1_import_get_fmu_kind(_fmi1Import) != fmi1_fmu_kind_cs) {
//            modelIdentifier = fmi1_import_get_model_identifier_ME(_fmi1Import);
//            printf("Model identifier for ME: %s\n", modelIdentifier);
//            fmukind = fmi1_fmu_kind_me;
//        }
//        else if(fmi1_import_get_fmu_kind(_fmi1Import) != fmi1_fmu_kind_me) {
//            modelIdentifier = fmi1_import_get_model_identifier_CS(_fmi1Import);
//            printf("Model identifier for CS: %s\n", modelIdentifier);
//            fmukind = fmi1_fmu_kind_cs;
//        }
//        else {
//            printf("Unxepected FMU kind, exiting\n");
//        }
        if (_blog)
            printf(" *** Model GUID \"%s\"\n", _sGUID.Data());

        // Create FMI pointer
        fmi1_callback_functions_t callBackFunctions;
        callBackFunctions.logger = fmi1_log_forwarding;  //fmilogger;  fmi1_default_callback_logger  fmi1_log_forwarding
        callBackFunctions.allocateMemory = calloc;
        callBackFunctions.freeMemory = free;
        callBackFunctions.stepFinished = 0;
//        callBackFunctions.componentEnvironment = 0;

        if (_blog) {
            cout << " --- _sFMUWorkingDirectory[" << _sFMUWorkingDirectory << "]" << endl;
            cout << " --- PWD[" << gSystem->pwd() << "]" << endl;
        }

        jm_status_enu_t status_FMI = fmi1_import_create_dllfmu(_fmi1Import, callBackFunctions, 1);
        if (status_FMI == jm_status_error) {
            throw URANIE::Exceptions::UErrorExceptions(
                        __FILE__,
                        __LINE__,
                        Form("Method TInputFileFMU::%s :\n Could not create the DLL loading mechanism(C-API).", __func__));
        } else {
            if (_blog)
                cout << " *** fmi1_import_create_dllfmu   OK" << endl;
        }

        _sPlatformName = fmi1_import_get_types_platform(_fmi1Import);
        if (_blog)
            cout << " *** _sPlatformName[" << _sPlatformName << "]" << endl;

        // unused information
        TString sModelStandardVersion = fmi1_import_get_model_standard_version(_fmi1Import);
        if (_blog)
            cout << " *** sModelStandardVersion[" << sModelStandardVersion << "]" << endl;

    } catch (URANIE::Exceptions::UErrorExceptions& ue) {
        ue.printMessage();
        throw ue;
    }


    if (_blog)
        cout << " *** End of Method TInputFileFMU::" << __func__ << " with soption[" << soption << "]" << endl;
}

// ----------------------------------------------------------------------
//  init_fmi2_0 Method
// ----------------------------------------------------------------------
void TInputFileFMU::init_fmi2_0(Option_t *option)
{
    TString soption = option ;
    soption.ToLower();
    FUNCT_TRACE();
    if (_blog)
        cout << " *** Method TInputFileFMU::" << __func__ << " with soption[" << soption << "]" << endl;

    try {

        _fmi2Import = fmi2_import_parse_xml(_context_FMI, _sFMUWorkingDirectory.Data(), 0);

        if(!_fmi2Import) {
            printf("Error parsing XML, exiting\n");
            do_exit(CTEST_RETURN_FAIL);
        }

        // Get the FMI General informations
        _sModelName = fmi2_import_get_model_name(_fmi2Import);
        if (_blog)
            cout << " ***      _sModelName[" << _sModelName << "] (Original)" << endl;

        _sModelName.ReplaceAll(" ", "_");
        _sModelName.ReplaceAll(",", "_");
        if (_blog)
            cout << " *** _sModelName[" << _sModelName << "] (After replace characters)" << endl;

        _sModelVersion = fmi2_import_get_model_version(_fmi2Import);
        if (_blog)
            cout << " *** _sModelVersion[" << _sModelVersion << "]" << endl;

        _sGUID = fmi2_import_get_GUID(_fmi2Import);
        if (_blog)
            cout << " *** _sGUID[" << _sGUID << "]" << endl;

        _sGenerationTool     = fmi2_import_get_generation_tool(_fmi2Import);
        _sGenerationDateTime = fmi2_import_get_generation_date_and_time(_fmi2Import);

        _dFMUStart = fmi2_import_get_default_experiment_start(_fmi2Import);
        _dFMUStop = fmi2_import_get_default_experiment_stop(_fmi2Import);
        _dFMUTolerance = fmi2_import_get_default_experiment_tolerance(_fmi2Import);
        _dFMUStep = fmi2_import_get_default_experiment_step(_fmi2Import);


        fmi2_import_model_counts_t  mc;

        if (_blog)
            printf(" *** Model name: %s\n", _sModelName.Data());
        fmi2_import_collect_model_counts(_fmi2Import, &mc);
        if (_blog) {
            cout << "  *** num_constants  [" << mc.num_constants << "]" << endl;
            cout << "  *** num_fixed      [" << mc.num_fixed << "]" << endl;
            cout << "  *** num_tunable    [" << mc.num_tunable << "]" << endl;
            cout << "  *** num_discrete   [" << mc.num_discrete << "]" << endl;
            cout << "  *** num_continuous [" << mc.num_continuous << "]" << endl;
            cout << "  *** num_independent[" << mc.num_independent << "]" << endl;
        }

        fmi2_fmu_kind_enu_t fmukind = fmi2_fmu_kind_unknown;
        fmukind = fmi2_import_get_fmu_kind(_fmi2Import);
        if (_blog)
            cout << " *** fmukind[" << fmukind << "] string[" << fmi2_fmu_kind_to_string(fmukind) << "]" << endl;

        if ( fmukind != fmi2_fmu_kind_cs )
            throw URANIE::Exceptions::UErrorExceptions(
                    __FILE__,
                    __LINE__,
                    Form("Method TInputFileFMU::%s :\n The type \"%s\" not supporte: only \"%s\" supported.",
                         __func__,
                         fmi2_fmu_kind_to_string(fmukind),
                         fmi2_fmu_kind_to_string(fmi2_fmu_kind_cs)));

        const char* modelIdentifier;
        if( fmukind == fmi2_fmu_kind_cs)
            modelIdentifier = fmi2_import_get_model_identifier_CS(_fmi2Import);

//        if(fmukind == fmi2_fmu_kind_me) {
//            modelIdentifier = fmi2_import_get_model_identifier_ME(_fmi2Import);
//        }

        // Print informations
        try {
            throw URANIE::Exceptions::UInfoExceptions(
                        __FILE__,
                        __LINE__,
                        Form("Method TInputFileFMU::%s :\n The default Experiment values readed are :\n"
                             "  FMU file \"%s\"\n"
                             "  Version   = FMI %d\n"
                             "  Type      = %s\n"
                             "  Start     = %e\n"
                             "  Stop      = %e\n"
                             "  Tolerance = %e\n"
                             "  Step      = %e",
                             __func__, _sFileName.Data(), _sVersion, fmi2_fmu_kind_to_string(fmukind), _dFMUStart, _dFMUStop, _dFMUTolerance, _dFMUStep));

            if (_blog)
                printf(" *** Model Identifier \"%s\" with \"%s\" (%d)\n", modelIdentifier, fmi2_fmu_kind_to_string(fmukind), fmukind);

        } catch (URANIE::Exceptions::UInfoExceptions& ue)
        {
            ue.printMessage();
        }


        if (_blog)
            printf(" *** Model GUID \"%s\"\n", _sGUID.Data());

        // Create FMI pointer
        fmi2_callback_functions_t callBackFunctions;
        callBackFunctions.logger = fmi2_log_forwarding;  //fmilogger;  fmi2_default_callback_logger  fmi2_log_forwarding
        callBackFunctions.allocateMemory = calloc;
        callBackFunctions.freeMemory = free;
        callBackFunctions.stepFinished = 0;
        callBackFunctions.componentEnvironment = 0;

        if (_blog) {
            cout << " --- _sFMUWorkingDirectory[" << _sFMUWorkingDirectory << "]" << endl;
            cout << " --- PWD[" << gSystem->pwd() << "]" << endl;
        }

        jm_status_enu_t status_FMI = fmi2_import_create_dllfmu(_fmi2Import, fmukind, &callBackFunctions);
        if (status_FMI == jm_status_error) {
            throw URANIE::Exceptions::UErrorExceptions(
                    __FILE__,
                    __LINE__,
                    Form("Method TInputFileFMU::%s :\n Could not create the DLL loading mechanism(C-API).", __func__));
        } else {
            if (_blog)
                cout << " *** fmi2_import_create_dllfmu   OK" << endl;
        }

        _sPlatformName = fmi2_import_get_types_platform(_fmi2Import);
        if (_blog)
            cout << " *** _sPlatformName[" << _sPlatformName << "]" << endl;

        // unused information
        TString sModelStandardVersion = fmi2_import_get_model_standard_version(_fmi2Import);
        if (_blog)
            cout << " *** sModelStandardVersion[" << sModelStandardVersion << "]" << endl;

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    if (_blog)
        cout << " *** End of Method TInputFileFMU::" << __func__ << " with soption[" << soption << "]" << endl;
}


// ----------------------------------------------------------------------
//  setWorkingDirectory Method
// ----------------------------------------------------------------------
void TInputFileFMU::setWorkingDirectory(TString sPathName)
{
    FUNCT_TRACE();
    if (_blog)
        cout << " *** Method TInputFileFMU::" << __func__ << " with sPathName[" << sPathName << "]" << endl;

    // buffer used to prevent memleak, as gSystem::ConcatFileName result must be deleted
    char *foo=NULL;

    TString sPWD = gSystem->pwd();
    if (_blog)
        cout << " *** sPWD(" << sPWD << ")" << endl;


    if ( sPathName.Length() == 0 ) {
        foo = gSystem->ConcatFileName(sPWD.Data(), DEFAULT_FMU_WORKING_DIRECTORY);
        _sFMUWorkingDirectory=foo;
        delete[] foo; foo=NULL;
    } else { if (gSystem->IsAbsoluteFileName(sPathName.Data())) {
            _sFMUWorkingDirectory = sPathName;
        } else {
            foo = gSystem->ConcatFileName(sPWD.Data(), sPathName.Data());
            _sFMUWorkingDirectory = foo;
            delete[] foo; foo=NULL;
        }
    }

    if ( _blog )
        cout << " *** Final _sFMUWorkingDirectory Directory[" << _sFMUWorkingDirectory << endl;

    // Clean
    if ( !foo )
        delete[] foo;
    foo=NULL;

    if (_blog)
        cout << " *** End of Method TInputFileFMU::" << __func__ << " with sPathName[" << sPathName << "]" << endl;
}

// ----------------------------------------------------------------------
//  simulate_fmi2_cs Method
// ----------------------------------------------------------------------
Int_t TInputFileFMU::simulate_fmi1_cs(Option_t *option)
{
    FUNCT_TRACE();
    TString soption = option ;
    soption.ToLower();
    if (_blog)
        cout << " *** Method TInputFileFMU::" << __func__ << " with soption[" << soption << "]" << endl;


    fmi1_status_t fmi1status;
    jm_status_enu_t jmstatus;
    fmi1_string_t fmuGUID;

    TString sInstanceName = Form("__Uranie_FMI2_CS_%s_%s_%d__", _sModelName.Data(), _sPlatformName.Data(), 1);
    if (_blog)
        cout << " *** sInstanceName[" << sInstanceName << "]" << endl;
    Int_t nLength = sInstanceName.Length();
    if (_blog)
        cout << " *** sInstanceName[" << sInstanceName << "] - nLength[" << nLength << "]" << endl;

    if (_blog) {
        printf(" *** Version returned from FMU:   %s\n", fmi1_import_get_version(_fmi1Import));
        printf(" *** Platform type returned:      %s\n", fmi1_import_get_types_platform(_fmi1Import));
    }

    fmuGUID = fmi1_import_get_GUID(_fmi1Import);
    if (_blog)
        printf(" *** GUID:      %s\n", fmuGUID);

    Double_t tstart = _dFMUStart;
    Double_t tend = _dFMUStop;
    Double_t hstep = _dFMUStep;
    fmi2_boolean_t StopTimeDefined = fmi2_false; // \todo Add this attribute as global

    Double_t tcur = tstart;
    UInt_t nYVector = _no;
    if (_blog)
        cout << " *** nYVector[" << nYVector << "]" << endl;

    Double_t *dVecVal = new Double_t[nYVector+1];
    fmi1_value_reference_t *compare_real_variables_vr = new fmi1_value_reference_t[nYVector];

    for(UInt_t i=0; i<nYVector; i++) {
        if (_blog)
            cout << "    *** i(" << i << "/" << nYVector << ") - _vecVROutputs[i][" << _vecVROutputs[i] << "]";
        compare_real_variables_vr[i] = (fmi1_value_reference_t) (_vecVROutputs[i]);
        if (_blog)
            cout << " - VR[" << compare_real_variables_vr[i] << "]" << endl;
    }

    // Store values
    //vector<vector<double>> vecOutputs;
    _vecOutputs.clear();
    tcur = tstart;

    // time init
    vector<double> vecValue;

    fmi1_import_get_real(_fmi1Import, compare_real_variables_vr, nYVector, dVecVal);
    for(UInt_t i=0; i<nYVector; i++)
        vecValue.push_back(dVecVal[i]);
    vecValue.push_back(tcur);

    if (_blog) {
        cout << " *** time[" << tcur << "/" << tend << "][" << nYVector << "][" << _vecOutputs.size() << "]    ::";
        for(UInt_t i=0; i<vecValue.size(); i++) {
            if ( i == 0 )
                cout << " (";
            else
                cout << ", ";
            cout << vecValue[i];
            if ( i == nYVector )
                cout << ")" << endl;
        }
    }
    _vecOutputs.push_back(vecValue);

    // Loop
    while (tcur < tend) {
        fmi1_boolean_t newStep = fmi1_true;
        fmi1status = fmi1_import_do_step(_fmi1Import, tcur, hstep, newStep);

        dVecVal[nYVector] = tcur + hstep;
        // log the intermediate value
        fmi1_import_get_real(_fmi1Import, compare_real_variables_vr, nYVector, dVecVal);

        vecValue.clear();
        for(UInt_t i=0; i<nYVector; i++) {
            vecValue.push_back(dVecVal[i]);
        }
        vecValue.push_back(tcur + hstep);
        _vecOutputs.push_back(vecValue);

        if (_blog) {
            cout << " --- time[" << tcur + hstep << "/" << tend << "][" << nYVector << "][" << _vecOutputs.size() << "] ::";

            for(UInt_t i=0; i<vecValue.size(); i++) {
                if ( i == 0 )
                    cout << " (";
                else
                    cout << ", ";
                cout << vecValue[i];
                if ( i == nYVector )
                    cout << ")" << endl;
            }
        }

        tcur += hstep;
    }

    if (_blog) {
        printf(" *** Simulation finished\n");
        for (UInt_t k = 0; k < nYVector; k++) {
            fmi1_value_reference_t vr = compare_real_variables_vr[k];
            Double_t rvalue;
            fmi1status = fmi1_import_get_real(_fmi1Import, &vr, 1, &rvalue);
            printf("    *** k(%d/%d) Value[%e]\n", k, nYVector, rvalue);
        }
    }

    // Clean
    delete[] dVecVal; dVecVal = NULL;
    delete[] compare_real_variables_vr; compare_real_variables_vr = NULL;

    if (_blog)
        cout << " *** end of Function[" << __func__ << "] with soption[" << soption << "]" << endl;
    return 0;
}

// ----------------------------------------------------------------------
//  simulate_fmi2_cs Method
// ----------------------------------------------------------------------
Int_t TInputFileFMU::simulate_fmi2_cs(Option_t *option)
{
    FUNCT_TRACE();
    TString soption = option ;
    soption.ToLower();
    if (_blog)
        cout << " *** Method TInputFileFMU::" << __func__ << " with soption[" << soption << "]" << endl;


    fmi2_status_t fmistatus;
    jm_status_enu_t jmstatus;
    fmi2_string_t fmuGUID;


    TString sInstanceName = Form("__Uranie_FMI2_CS_%s_%s_%d__", _sModelName.Data(), _sPlatformName.Data(), 1);
    if (_blog)
        cout << " *** sInstanceName[" << sInstanceName << "]" << endl;
    Int_t nLength = sInstanceName.Length();
    if (_blog)
        cout << " *** sInstanceName[" << sInstanceName << "] - nLength[" << nLength << "]" << endl;

    if (_blog) {
        printf(" *** Version returned from FMU:   %s\n", fmi2_import_get_version(_fmi2Import));
        printf(" *** Platform type returned:      %s\n", fmi2_import_get_types_platform(_fmi2Import));
    }

    fmuGUID = fmi2_import_get_GUID(_fmi2Import);
    if (_blog)
        printf(" *** GUID:      %s\n", fmuGUID);

    Double_t tstart = _dFMUStart;
    Double_t tend = _dFMUStop;
    Double_t hstep = _dFMUStep;
    fmi2_boolean_t StopTimeDefined = fmi2_false; // \todo Add this attribute as global

    Double_t tcur = tstart;
    UInt_t nYVector = _no;
    if (_blog)
        cout << " *** nYVector[" << nYVector << "]" << endl;

    Double_t *dVecVal = new Double_t[nYVector+1];
    fmi2_value_reference_t *compare_real_variables_vr = new fmi2_value_reference_t[nYVector];

    for(UInt_t i=0; i<nYVector; i++) {
        if (_blog)
            cout << "    *** i(" << i << "/" << nYVector << ") - _vecVROutputs[i][" << _vecVROutputs[i] << "]";
        compare_real_variables_vr[i] = (fmi2_value_reference_t) (_vecVROutputs[i]);
        if (_blog)
            cout << " - VR[" << compare_real_variables_vr[i] << "]" << endl;
    }

    // Store values
    //vector<vector<double>> vecOutputs;
    _vecOutputs.clear();
    tcur = tstart;

    // time init
    vector<double> vecValue;

    fmi2_import_get_real(_fmi2Import, compare_real_variables_vr, nYVector, dVecVal);
    for(UInt_t i=0; i<nYVector; i++)
        vecValue.push_back(dVecVal[i]);
    vecValue.push_back(tcur);

    if (_blog) {
        cout << " *** time[" << tcur << "/" << tend << "][" << nYVector << "][" << _vecOutputs.size() << "] ::";
        for(UInt_t i=0; i<vecValue.size(); i++) {
            if ( i == 0 )
                cout << " (";
            else
                cout << ", ";
            cout << vecValue[i];
            if ( i == nYVector )
                cout << ")" << endl;
        }
    }
    _vecOutputs.push_back(vecValue);

    // Loop
    while (tcur < tend) {
        fmi2_boolean_t newStep = fmi2_true;
        fmistatus = fmi2_import_do_step(_fmi2Import, tcur, hstep, newStep);

        dVecVal[nYVector] = tcur + hstep;
        // log the intermediate value
        fmi2_import_get_real(_fmi2Import, compare_real_variables_vr, nYVector, dVecVal);

        vecValue.clear();
        for(UInt_t i=0; i<nYVector; i++)
            vecValue.push_back(dVecVal[i]);
        vecValue.push_back(tcur + hstep);
        _vecOutputs.push_back(vecValue);

        if (_blog) {
            cout << " --- time[" << tcur + hstep << "/" << tend << "][" << nYVector << "][" << _vecOutputs.size() << "] ::";

            for(UInt_t i=0; i<vecValue.size(); i++) {
                if ( i == 0 )
                    cout << " (";
                else
                    cout << ", ";
                cout << vecValue[i];
                if ( i == nYVector )
                    cout << ")" << endl;
            }
        }

        tcur += hstep;
    }

    if (_blog) {
        printf(" *** Simulation finished\n");
        for (UInt_t k = 0; k < nYVector; k++) {
            fmi2_value_reference_t vr = compare_real_variables_vr[k];
            fmi2_real_t rvalue;
            fmistatus = fmi2_import_get_real(_fmi2Import, &vr, 1, &rvalue);
            printf("    *** k(%d/%d) Value[%e]\n", k, nYVector, rvalue);
        }
    }

    // Clean
    delete[] dVecVal; dVecVal = NULL;
    delete[] compare_real_variables_vr; compare_real_variables_vr = NULL;

    if (_blog)
        cout << " *** end of Function[" << __func__ << "] with soption[" << soption << "]" << endl;
    return 0;
}

// ----------------------------------------------------------------------
//  searchIndexeAttribute Method
// ----------------------------------------------------------------------
void TInputFileFMU::searchIndexeAttribute(URANIE::DataServer::TAttribute* att)
{
    FUNCT_TRACE();
    if (_blog) {
        cout << " *** Method [TInputFileFMU::" << __func__  << "]" << endl;
        cout << " *** File[" << getFileName() << "] Attribute[" << att->GetName() << "]" << endl;
    }

    try {
        switch (_sVersion) {
        case fmi_version_unknown_enu:
            throw URANIE::Exceptions::UInfoExceptions(
                        __FILE__,
                        __LINE__,
                        Form("TInputFileFMU::%s ERROR:\n The FMI version (%i) of the file \"%s\" is not defined for attribute[%s].",
                             __func__, _sVersion, _sFileName.Data(), att->GetName()));
            break;
        case fmi_version_unsupported_enu:
            throw URANIE::Exceptions::UErrorExceptions(
                        __FILE__,
                        __LINE__,
                        Form("TInputFileFMU::%s ERROR:\n The FMI version (%i) of the file \"%s\" is not supported.",
                             __func__, _sVersion, _sFileName.Data()));
            break;
        case fmi_version_1_enu:
        case fmi_version_2_0_enu:

            list<URANIE::DataServer::TAttributeFileKey *> KeyList = att->getKeyList();
            for (list<URANIE::DataServer::TAttributeFileKey *>::iterator it = KeyList.begin(); it != KeyList.end(); ++it)
            {
                Int_t nfield = -1;

                TString skey = (*it)->_sKey;

                TString sfilecrt = (*it)->_sFile;
                if (sfilecrt.CompareTo(_sFileName, TString::kExact) == 0) {
                    nfield = -2;

                    if (_blog)
                        cout << " ***  Attribute[" << att->GetName() << "], sfilecrt[" << sfilecrt.Data() << "] :: skey[" << skey << "] nfield[" << nfield << "]" << endl;

                    // Get the FMI Attribute ( 2.0 )
                    if ( _sVersion == fmi_version_2_0_enu ) {
                        fmi2_import_variable_t* fmiAtt = fmi2_import_get_variable_by_name(_fmi2Import, skey.Data());

                        if ( !fmiAtt) {
                            throw URANIE::Exceptions::UErrorExceptions(
                                        __FILE__,
                                        __LINE__,
                                        Form("TInputFileFMU::%s ERROR\n: the key \"%s\" of attribute \"%s\" not found in the file \"%s\".",
                                             __func__, skey.Data(), att->GetName(), _sFileName.Data()));
                        }

                       if (_blog)
                           printf(" ***  att[%s] :: fmi2_import_get_variable_name[%s]", att->GetName(), fmi2_import_get_variable_name(fmiAtt));

                        // Get the ValueReference index
                        nfield = fmi2_import_get_variable_vr(fmiAtt);
                        if (_blog) {
                            cout << " nfield[" << nfield << "]";
                            if ( fmi2_import_get_variable_has_start(fmiAtt) ) {
                                cout << " Start-Value[" << fmi2_import_get_real_variable_start(fmi2_import_get_variable_as_real(fmiAtt)) << "]";
                            }
                            cout << endl;
                        }
                    } // end of _sVersion == fmi_version_2_0_enu

                    // Get the FMI Attribute ( 1.0 )
                    if ( _sVersion == fmi_version_1_enu ) {
                        fmi1_import_variable_t* fmiAtt = fmi1_import_get_variable_by_name(_fmi1Import, skey.Data());

                        if ( !fmiAtt) {
                            throw URANIE::Exceptions::UErrorExceptions(
                                        __FILE__,
                                        __LINE__,
                                        Form("TInputFileFMU::%s ERROR\n: the key \"%s\" of attribute \"%s\" not found in the file \"%s\".",
                                             __func__, skey.Data(), att->GetName(), _sFileName.Data()));
                        }

                        if (_blog)
                            printf(" ***  att[%s] :: fmi1_import_get_variable_name[%s]", att->GetName(), fmi1_import_get_variable_name(fmiAtt));

                        // Get the ValueReference index
                        nfield = fmi1_import_get_variable_vr(fmiAtt);

                        if (_blog) {
                            cout << " nfield[" << nfield << "]";

                            if ( fmi1_import_get_variable_has_start(fmiAtt) ) {
                                cout << " Start-Value[" << fmi1_import_get_real_variable_start(fmi1_import_get_variable_as_real(fmiAtt)) << "]";
                            }
                            cout << endl;
                        }
                    } // end of _sVersion == fmi_version_1_enu

                    //
                    if ( nfield > -1 ) {
                        if (_blog)
                            cout << " ***  Attribute[" << att->GetName() << "], sfilecrt[" << sfilecrt.Data() << "] :: skey[" << skey << "] nfield[" << nfield << "]" << endl;

                        list<URANIE::DataServer::TAttributeFileKey *>::iterator first = KeyList.begin();
                        att->setFieldOfKey(distance(first, it), nfield);
                    }

                }
            }
            break;
        }

    } catch (URANIE::Exceptions::UInfoExceptions& ue)
    {
        // RAS :: ue.printMessage();
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }


    if (_blog)
        cout << " *** End of Method [TInputFileFMU::" << __func__  << "] - " <<
                " File[" << getFileName() << "]" <<
                " Attribute[" << att->GetName() << "]" << endl;
}

// ----------------------------------------------------------------------
//  substitue Method
// ----------------------------------------------------------------------
void TInputFileFMU::substitue(Int_t &n, Double_t *val, Int_t *index)
{
    FUNCT_TRACE();
    UNUSED(n);
    UNUSED(val);
    UNUSED(index);
    if (_blog)
        cout << " *** Method TInputFileFMU::" << __func__  << endl;

    if (_blog)
        cout << " *** End of Method TInputFileFMU::" << __func__  << endl;
}

// ----------------------------------------------------------------------
//  substitue Method
// ----------------------------------------------------------------------
void TInputFileFMU::substitue(Int_t &n, vector<URANIE::DataServer::UEntry*> *ent, Int_t *index)
{
    FUNCT_TRACE();
    if (_blog)
        cout << " *** Method TInputFileFMU::" << __func__  << " n[" << n << "]" << endl;

    // instanciate the FMU code
    if ( n == 1 ) {

        TString sInstanceName = Form("__Uranie_FMI%d_CS_%s_%s_%d__", _sVersion, _sModelName.Data(), _sPlatformName.Data(), n);
        if ( _blog)
            cout << " *** sInstanceName[" << sInstanceName << "]" << endl;
        Int_t nLength = sInstanceName.Length();
        if ( _blog)
            cout << " *** sInstanceName[" << sInstanceName << "] - nLength[" << nLength << "]" << endl;


        Double_t tstart = 0.0;
        Double_t tend = 10.0;
        Double_t tcur = tstart;
        Double_t hstep = 0.01;
        Double_t dRelTol = 1e-4;


        Bool_t StopTimeDefined = kFALSE; /// fmi2_false; // \todo Add this attribute as global
        TString sFMULocation = "";

        jm_status_enu_t jmstatus;

        try {
            switch (_sVersion) {
            case fmi_version_unknown_enu:
                throw URANIE::Exceptions::UInfoExceptions(
                            __FILE__,
                            __LINE__,
                            Form("TInputFileFMU::%s ERROR:\n The FMI version (%i) of the file \"%s\" for n[%d] is not defined.",
                                 __func__, _sVersion, _sFileName.Data(), n));
                break;
            case fmi_version_unsupported_enu:
                throw URANIE::Exceptions::UErrorExceptions(
                            __FILE__,
                            __LINE__,
                            Form("TInputFileFMU::%s ERROR:\n The FMI version (%i) of the file \"%s\" for n[%d] is not supported.",
                                 __func__, _sVersion, _sFileName.Data(), n));
                break;
            case fmi_version_1_enu:
            {
                if (_blog)
                    cout << " *** case of FMI 1" << endl;

                fmi1_status_t fmistatus;

                fmi1_string_t fmu1GUID;
                fmi1_string_t mimeType = "";
                fmi1_boolean_t bIsVisible = fmi1_false;
                fmi1_real_t timeout = 0.0;
                fmi1_boolean_t interactive = fmi1_false;
                /*	fmi1_boolean_t loggingOn = fmi1_true; */

                if (_blog) {
                    printf(" *** Version returned from FMU:   %s\n", fmi1_import_get_version(_fmi1Import));
                    printf(" *** Platform type returned:      %s\n", fmi1_import_get_types_platform(_fmi1Import));
                }

                fmu1GUID = fmi1_import_get_GUID(_fmi1Import);
                if (_blog)
                    printf(" *** GUID:      %s\n", fmu1GUID);

                jmstatus = jm_status_success;
                //jmstatus = fmi1_import_instantiate_model(_fmi1Import, sInstanceName.Data());
                jmstatus = fmi1_import_instantiate_slave(_fmi1Import,
                                                         sInstanceName.Data(),
                                                         sFMULocation.Data(),
                                                         mimeType,
                                                         timeout,
                                                         bIsVisible,
                                                         interactive);

                if (jmstatus == jm_status_error) {
                    printf("fmi_import_instantiate failed [%s]\n", sInstanceName.Data());
                    do_exit(CTEST_RETURN_FAIL);
                }


                fmistatus = fmi1_import_initialize_slave(_fmi1Import,
                                                         tstart,
                                                         StopTimeDefined,
                                                         tend);
                if(fmistatus != fmi1_status_ok) {
                    printf("fmi1_import_initialize_slave failed\n");
                    do_exit(CTEST_RETURN_FAIL);
                }

//                printf(" *** Begin fmi1_import_enter_initialization_mode\n");
//                for (Int_t k = 0; k < 5; k++) {
//                    fmi1_value_reference_t vr = k; // compare_real_variables_vr[k];
//                    fmi1_real_t rvalue;
//                    fmistatus = fmi1_import_get_real(_fmi1Import, &vr, 1, &rvalue);
//                    printf("    **** instantiate    k(%d/%d) Value[%e]\n", k, 5, rvalue);
//                }
                break;
            }
            case fmi_version_2_0_enu:
            {

                fmi2_status_t fmistatus;
                fmi2_string_t fmuGUID;
                fmi2_boolean_t bIsVisible = fmi2_false;

                if (_blog) {
                    printf(" *** Version returned from FMU:   %s\n", fmi2_import_get_version(_fmi2Import));
                    printf(" *** Platform type returned:      %s\n", fmi2_import_get_types_platform(_fmi2Import));
                }
                fmuGUID = fmi2_import_get_GUID(_fmi2Import);
                if (_blog)
                    printf(" *** GUID:      %s\n", fmuGUID);

                jmstatus = fmi2_import_instantiate(_fmi2Import, sInstanceName.Data(), fmi2_cosimulation, sFMULocation.Data(), bIsVisible);
                if (jmstatus == jm_status_error) {
                    printf("fmi2_import_instantiate failed\n");
                    do_exit(CTEST_RETURN_FAIL);
                }

                fmistatus = fmi2_import_setup_experiment(_fmi2Import, fmi2_true, dRelTol, tstart, StopTimeDefined, tend);
                if(fmistatus != fmi2_status_ok) {
                    printf("fmi2_import_setup_experiment failed\n");
                    do_exit(CTEST_RETURN_FAIL);
                }

                fmistatus = fmi2_import_enter_initialization_mode(_fmi2Import);
                if(fmistatus != fmi2_status_ok) {
                    printf("fmi2_import_enter_initialization_mode failed\n");
                    do_exit(CTEST_RETURN_FAIL);
                }

//                printf(" *** Begin fmi2_import_enter_initialization_mode\n");
//                for (Int_t k = 0; k < 5; k++) {
//                    fmi2_value_reference_t vr = k; // compare_real_variables_vr[k];
//                    fmi2_real_t rvalue;
//                    fmistatus = fmi2_import_get_real(_fmi2Import, &vr, 1, &rvalue);
//                    printf("    **** instantiate    k(%d/%d) Value[%e]\n", k, 5, rvalue);
//                }

                break;
            }
            default:
                cout << " ****** default case  :: RAS" << endl;
                break;
            } // End of Switch _sVersion
        } catch (URANIE::Exceptions::UInfoExceptions& ue)
        {
            // RAS :: ue.printMessage();
        } catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            throw ue;
        }
    } // End of n == 1


    // Current value
    TString beg, end;
    Int_t nIndex;
    TString sValue;
    Double_t dValue;
    Int_t natt = getNAttributes();

    Double_t *inputValues = new Double_t[natt];
    fmi2_value_reference_t *inputIndexes = new fmi2_value_reference_t[natt];

    for (Int_t i = 0; i < natt; i++) {
        if (_blog)
            cout << " *** " << i << "/" << natt <<
                    " name[" << _attributes[i]->GetName() << "]" << endl;
        getBegEndProperties( _attributes[i], beg, end);

        list<URANIE::DataServer::TAttributeFileKey *> KeyList = _attributes[i]->getKeyList();
        for (list<URANIE::DataServer::TAttributeFileKey *>::iterator it = KeyList.begin(); it != KeyList.end(); ++it)
        {
            TString sfilename = gSystem->BaseName((*it)->_sFile.Data());
            nIndex = -1;
            sValue ="";

            if (_blog)
                cout << "   *** TInputFileFMU::substitue\n   *** PWD[" << gSystem->pwd() << "]" << endl <<
                        "   *** sfilename[" << sfilename << "]" << endl;

            TString sfilecrt = (*it)->_sFile;
            if (_blog)
                cout << "   *** sfilecrt[" << sfilecrt  << "]" << endl;
            TString sformattosubstitute = TString((*it)->_sFormat);
            if (_blog)
                cout << "   *** sformattosubstitute[" << sformattosubstitute  << "]";

            nIndex = (*it)->_sEventList->GetEntry(0);
            if (_blog)
                cout << " - nIndex[" << nIndex  << "]";
            inputIndexes[i] = nIndex;
            if (_blog)
                cout << " - inputIndexes[" << inputIndexes[i]  << "]";

            sValue = ent->at( ((index==NULL) ? n : index[n - 1]) )->dumpStr( _svecdel, beg, end, sformattosubstitute.Data());
            dValue = atof(sValue);
            inputValues[i] = dValue;
            if (_blog)
                cout << " - sValue[" << sValue  << "] double[" << dValue  << "] inputValues[" << i << "][" << inputValues[i]  << "]" << endl;
        }
        n++;
    }



    try {
        switch (_sVersion) {
        case fmi_version_1_enu:

            fmi1_status_t fmi1status;
            // Set the initialisation values ( Only for Real input)
            fmi1status = fmi1_import_set_real(_fmi1Import, inputIndexes, natt, inputValues);
            if(fmi1status != fmi1_status_ok) {
                printf("fmi1_import_set_real failed\n");
                do_exit(CTEST_RETURN_FAIL);
            }

            if (_blog) {
                printf(" *** Initialisation Step\n");
                for (Int_t i = 0; i < natt; i++) {
                    fmi1_value_reference_t vr = inputIndexes[i];
                    fmi1_real_t rvalue;
                    fmi1status = fmi1_import_get_real(_fmi1Import, &vr, 1, &rvalue);
                    printf("    *** FMI1 (%d/%d) Index(%d)Value[%e]\n", i, natt, inputIndexes[i], rvalue);
                }
            }

            break;
        case fmi_version_2_0_enu:
            fmi2_status_t fmistatus;

            // Set the initialisation values ( Only for Real input)
            fmistatus = fmi2_import_set_real(_fmi2Import, inputIndexes, natt, inputValues);
            if(fmistatus != fmi2_status_ok) {
                printf("fmi2_import_set_real failed\n");
                do_exit(CTEST_RETURN_FAIL);
            }


            fmistatus = fmi2_import_exit_initialization_mode(_fmi2Import);
            if(fmistatus != fmi2_status_ok) {
                printf("fmi2_import_exit_initialization_mode failed\n");
                do_exit(CTEST_RETURN_FAIL);
            }

            if (_blog) {
                printf(" *** Initialisation Step\n");
                for (Int_t i = 0; i < natt; i++) {
                    fmi2_value_reference_t vr = inputIndexes[i];
                    fmi2_real_t rvalue;
                    fmistatus = fmi2_import_get_real(_fmi2Import, &vr, 1, &rvalue);
                    printf("    ***FMI2 (%d/%d) Index(%d)Value[%e]\n", i, natt, inputIndexes[i], rvalue);
                }
            }
            break;
        default:
            break;
        }
    } catch (URANIE::Exceptions::UInfoExceptions& ue)
    {
        // RAS :: ue.printMessage();
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    // clean
    delete[] inputValues; inputValues = NULL;
    delete[] inputIndexes; inputIndexes = NULL;

    if (_blog)
        cout << " *** End of Method [TInputFileFMU::" << __func__  << "] with n[" << n << "]" << endl;
}

// ----------------------------------------------------------------------
//  runFMI Method
// ----------------------------------------------------------------------
void TInputFileFMU::runFMI(Option_t *option)
{
    FUNCT_TRACE();
    TString soption = option;
    soption.ToLower();
    if (_blog)
        cout << " *** Method [TInputFileFMU::" << __func__  << "] with soption[" << soption << "]" << endl;

    try {
        switch (_sVersion) {
        case fmi_version_unknown_enu:
            throw URANIE::Exceptions::UInfoExceptions(
                        __FILE__,
                        __LINE__,
                        Form("TInputFileFMU::%s ERROR:\n The FMI version (%i) of the file \"%s\" is not defined.",
                             __func__, _sVersion, _sFileName.Data()));
            break;
        case fmi_version_unsupported_enu:
            throw URANIE::Exceptions::UErrorExceptions(
                        __FILE__,
                        __LINE__,
                        Form("TInputFileFMU::%s ERROR:\n The FMI version (%i) of the file \"%s\" is not supported.",
                             __func__, _sVersion, _sFileName.Data()));
            break;
        case fmi_version_1_enu: {
            if ( _fmi1Import )
                simulate_fmi1_cs(option);
            break;
        }
        case fmi_version_2_0_enu: {
            if ( _fmi2Import ) {
                simulate_fmi2_cs(option);
            }
            break;
        }
        default:
            cout << " *** Default case :: RAS" << endl;
            break;
        }
    } catch (URANIE::Exceptions::UInfoExceptions& ue)
    {
        // RAS :: ue.printMessage();
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    if (_blog)
        cout << " *** End of Method [TInputFileFMU::" << __func__  << "] with soption[" << option << "]" << endl;
}

// ----------------------------------------------------------------------
//  getFMUVersionString Method
// ----------------------------------------------------------------------
const char* TInputFileFMU::getFMUVersionString()
{
    FUNCT_TRACE();
    if (_blog)
        cout << " *** Method [TInputFileFMU::" << __func__  << "]" << endl;

    switch (_sVersion) {
    case fmi_version_1_enu:
        return fmi1_import_get_version(_fmi1Import);
    case fmi_version_2_0_enu:
        return fmi2_import_get_version(_fmi2Import);
    default:
        return "Unknown Version";
        break;
    }
}

// ----------------------------------------------------------------------
//  setListOutputVector Method
// ----------------------------------------------------------------------
void TInputFileFMU::setListOutputVector(TArrayI *arr)
{
    FUNCT_TRACE();
    if (_blog)
        cout << " *** Method [TInputFileFMU::" << __func__  << "] with array size[" << arr->GetSize() << "]" << endl;


    if ( arr ) {
        if (_blog)
            cout << " ***** setListOutputVector array is NOT NULLL array size[" << arr->GetSize() << "]" << endl;
        // Clear
        _vecVROutputs.clear();

        for(Int_t i=0; i<arr->GetSize(); i++)
            _vecVROutputs.push_back(arr->At(i));
    }

    _no = _vecVROutputs.size();
    if (_blog)
        cout << " *** _no[" << _no << "]" << endl;

    if (_blog)
        cout << " *** End of Method [TInputFileFMU::" << __func__  << "]" << endl;
}

// ----------------------------------------------------------------------
//  getAttributeInKey Method
// ----------------------------------------------------------------------
TString TInputFileFMU::getAttributeInKey(TString skey)
{
    FUNCT_TRACE();
    if (_blog)
        cout << " *** Method [TInputFileFMU::" << __func__  << "] skey[" << skey << "]" << endl;
    TString sattname = skey;

    //if (_blog)
    cout << " *** TInputFileFMU::getAttributeInKey[" << skey << "] sattname[" << sattname << "] NOT YET IMPLEMENTED" << endl;

    if (_blog)
        cout << " *** End of Method [TInputFileFMU::" << __func__  << "] skey[" << skey << "]" << endl;

    return sattname;
}

void TInputFileFMU::printLog(Option_t *option)
{
    FUNCT_TRACE();
    cout << " *** TInputFileFMU::printLog with option[" << option << "]" << endl;
    cout << " *** _sFileName[" << _sFileName << "]" << endl;
    cout << " --- _cComments[" << _cComments << "] ---  Not used in this class ---" << endl;
    cout << " --- _cSeparators[" << _cSeparators << "] ---  Not used in this class ---" << endl;
    cout << " *** _sFMUWorkingDirectory[" << _sFMUWorkingDirectory << "]" << endl;
    cout << " ******************** General FMI informations" << endl;
    cout << " *** _context_FMI[" << _context_FMI << "]" << endl;
    cout << " *** _sVersion type : enum [" << _sVersion << "], string[" << fmi_version_to_string(_sVersion) << "]" << endl;
    cout << " *** _sPlatformName[" << _sPlatformName << "]" << endl;
    cout << " *** _sModelName[" << _sModelName << "] _sModelVersion[" << _sModelVersion << "]" << endl;
    cout << " *** _sGUID[" << _sGUID << "]" << endl;
    cout << " **********************************************" << endl;

    Int_t nAttribute = (Int_t)(_attributes.size());
    cout << " *** nAttribute[" << nAttribute << "]" << endl;

    TString soption = TString(option);
    soption.ToLower();
    cout << " *** soption[" << soption << "]" << endl;

    if (soption.Contains("att")) {
        for (Int_t indx = 0; indx < nAttribute; indx++) {
            cout << "   *** index[" << indx + 1 << "/" << nAttribute << "]" << endl;
            _attributes[indx]->printLog();
        }
    }
        cout << " *** End of Method [TInputFileFMU::" << __func__  << "]" << endl;
}

} // End of namespace Launcher
} // End of namespace URANIE
// End of File

