//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TInputFileFMU.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Created : 2021/09/10 at 13:42:45

 \brief Interface file for the class URANIE::Launcher::TInputFileFMU
 */

/**
 * @class URANIE::Launcher::TInputFileFMU
 *
 * @brief Description of the class TInputFileFMU.
 *
 * To be written by the developper.
 *
 */
#ifndef TInputFileFMU_H
#define TInputFileFMU_H

// ROOT
#include "Rtypes.h"
#include "TString.h"

// URANIE
#include "TAttribute.h"
#include "TInputFile.h"

// FMI
#ifdef WIN32
#include <Windows4Root.h>
#endif
#include "fmilib.h"

namespace URANIE {
namespace Launcher {

class TInputFileFMU: public TInputFile
{
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    TInputFileFMU(TString str);
    //! Default destructor
    virtual ~TInputFileFMU();
    //@}

    // Public attribute accessor methods
    //
    //---------------------------------------------
    /** @name Setting and Getting attributs
     **/
    //@{
    /*!
     * \brief Set the working directory
     * \param str
     */
    void setWorkingDirectory(TString sPathName = "");

    /*!
     * \brief Get the working directory
     * \return
     */
    TString getWorkingDirectory() {
        return _sFMUWorkingDirectory;
    };

    Double_t getStartTime() {
        return _dFMUStart;
    }
    void setStartTime(Double_t dval) {
        _dFMUStart = dval;
    }

    Double_t getStopTime() {
        return _dFMUStop;
    }
    void setStopTime(Double_t dval) {
        _dFMUStop = dval;
    }

    Double_t getStepTime() {
        return _dFMUStep;
    }
    void setStepTime(Double_t dval) {
        _dFMUStep = dval;
    }

    Double_t getTolerance() {
        return _dFMUTolerance;
    }
    void setTolerance(Double_t dval) {
        _dFMUTolerance = dval;
    }



    //@}


    //---------------------------------------------
    /** @name Methods to perform initialisation and terminate states
     **/
    //@{
    //! The init step
    /*!
      \warning not yet used
    */
    virtual void init();

    //! The terminate step
    /*!
      \warning not yet used
    */
    virtual void terminate();
    //@}


    //---------------------------------------------
    /** @name File properties
     **/
    //@{
    /*!
     * \brief Gets the type of the file. This method is used to find InputFile of type 'Recreate'
     * \return
     */
    Bool_t isFileToCopy()
	{
            return kFALSE;
	}
    //! Return the name of the attribute for the key
    /*!
      The name of the attribute is the string after the last :"@" character
    */
    TString getAttributeInKey(TString skey);

    /*!
     * \brief addAttribute
     */
    virtual void addAttribute(URANIE::DataServer::TAttribute*);


    //@}

    //---------------------------------------------
    /** @name Implement the abstract pure methods
     **/
    //@{
    //! search the indexes of the attribute in the input file
    /*!
      \param att (URANIE::DataServer::TAttribute*) the attribute to search the indexes
    */
    virtual void searchIndexeAttribute(URANIE::DataServer::TAttribute* att);

    void substitue(Int_t &n, Double_t *val, Int_t *index = NULL);

    //! Substitute the values of the attributes linked in the input file and write it
    /*!
      The first value is defined par the parameter \em n and the values are stored in the parameter \em val.
      After the method, the parameter is increased by the number of attributes contained in the input file.
      \param n (Int_t &) the index of the first value in the vector of data \e val
      \param val (Double_t *) the pointer of the data
      \param index (Int_t *)the pointer of the index of data
    */
    void substitue(Int_t &n, vector<URANIE::DataServer::UEntry*> *ent, Int_t *index = NULL);
    //@}

    // Public FMI functions
    //
    //---------------------------------------------
    /** @name methods associated to the FMI Library
     **/
    //@{
    void setContext(fmi_import_context_t* fmi_ctxt)
    {
        _context_FMI = fmi_ctxt;
    }

    fmi_import_context_t* getContext()
    {
        return _context_FMI;
    }

    jm_callbacks* getCallbackFMI()
    {
        return _callbacks_FMI;
    }

    fmi1_import_t* getFMU1Import()
    {
        return _fmi1Import;
    }

    fmi2_import_t* getFMU2Import()
    {
        return _fmi2Import;
    }


    void setVersion(fmi_version_enu_t fmi_version)
    {
        _sVersion = fmi_version;
    }

    fmi_version_enu_t getFMUVersion()
    {
        return _sVersion;
    }
    const char* getFMUVersionString();

    void runFMI(Option_t *option="");

    /*!
     * \brief simulate_fmi1_cs
     * \param fmu
     * \return
     */
    Int_t simulate_fmi1_cs(Option_t *option="");

    /*!
     * \brief test_simulate_fmi2_cs
     * \param fmu
     * \return
     */
    Int_t simulate_fmi2_cs(Option_t *option="");

    void initFMI(Option_t *option="");

    void do_exit(int code)
    {
            printf("Press 'Enter' to exit\n");
            /* getchar(); */
            exit(code);
    }


    /*!
     * \brief init_fmi_v2_0
     * \param option
     */
    void init_fmi2_0(Option_t *option = "");

    /*!
     * \brief init_fmi1
     * \param option
     */
    void init_fmi1(Option_t *option = "");

    vector<vector<double>> getVectorValues()
    {
        return _vecOutputs;
    }

    void setListOutputVector(TArrayI *arr);
    //@}

    //---------------------------------------------
    /** @name The Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    //---------------------------------------------
    /** @name protected attributes
     *
     **/
    //@{
protected:
    TString _sFMUWorkingDirectory;      ///< the FMU import directory
    // FMI structures
    jm_callbacks *_callbacks_FMI;
    fmi_import_context_t* _context_FMI; ///< The FMU context
    fmi_version_enu_t _sVersion;        ///< The FMU version
    fmi1_import_t* _fmi1Import;         ///< The FMU 1.0 Object
    fmi2_import_t* _fmi2Import;         ///< The FMU 2.0 Object

    // FMI General informations
    TString _sGenerationTool;            ///< The FMU generation tool
    TString _sGenerationDateTime;        ///< The FMU generation date and time

    TString _sPlatformName;        ///< The
    TString _sModelName;        ///< The
    TString _sModelVersion;        ///< The
    TString _sGUID;        ///< The


    Double_t _dFMUStart;                 ///< The start time for default experiment as specified in the XML file
    Double_t _dFMUStop;                  ///< The stop time for default experiment as specified in the XML file
    Double_t _dFMUTolerance;             ///< The tolerance for default experiment as specified in the XML file (not used)
    Double_t _dFMUStep;                  ///< The step for default experiment as specified in the XML file
    vector<vector<double>> _vecOutputs;  ///< the vector of stored kVector attributes
    vector<int> _vecVROutputs;           ///< The vector of VR of the kVector Output attributes
    UInt_t _no;                          ///< The number of  kVector Output attributes

    //@}



    ClassDef(URANIE::Launcher::TInputFileFMU, ID_LAUNCHER)
};

} // End of namespace Launcher
} // End of namespace URANIE

#endif // TInputFileFMU_H
// End of File
