//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TInputFileKey.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mon oct 9th 14:50:04 CEST 2006
 
 \brief Implementation of the class URANIE::DLauncher::TInputFileKey

 \par Modifications
 \li F. Gaudier (2010-12-14 11:38:09)
 - resolve the DI111 and DI113

 
 */
// ROOT
#include "TSystem.h"
#include "TMath.h"

// URANIE
#include "TInputFileKey.h"

// Club
#ifdef OLDCLUB
#include "club/TamponAscii.h"
#include "club/IterateurChamp.h"
#include "club/IterateurLigne.h"
#include "club/TamponAscii.h"
#include "club/TamponTexte.h"
#include "club/ClubErreurs.h"
#else
#include "TamponTexte.h"
#endif
ClassImp (URANIE::Launcher::TInputFileKey)

namespace URANIE
{
namespace Launcher
{

TInputFileKey::TInputFileKey(TString str) :
    TInputFile(str)
{
    if (_blog)
        cout << "constructing TInputFileKey file[" << str << "]" << endl;
    _sstrbeg = _sstrend = "";
    _svecbeg = "["; _svecend = "]"; _svecdel = ",";
}

TInputFileKey::~TInputFileKey()
{
    if (_blog)
        cout << " ** deletes TInputFileKey [" << getFileName() << "]" << endl;
}


void TInputFileKey::init()
{
    if (_blog)
        cout << "URANIE::Launcher::init()" << endl;
}

void TInputFileKey::terminate()
{
    if (_blog)
        cout << "URANIE::Launcher::terminate()" << endl;
}

void TInputFileKey::searchIndexeAttribute(URANIE::DataServer::TAttribute* att)
{
    list<URANIE::DataServer::TAttributeFileKey *> KeyList = att->getKeyList();
    for (list<URANIE::DataServer::TAttributeFileKey *>::iterator it =
	     KeyList.begin(); it != KeyList.end(); ++it)
    {
        int nfield = -1;

        TString skey = (*it)->_sKey;

        TString sfilecrt = (*it)->_sFile;
        if (sfilecrt.CompareTo(_sFileName, TString::kExact) == 0)
        {
            try
            {
                TamponTexte fnomFichier((*it)->_sFile.Data(), 0);

                // Define the comment and separator character
                fnomFichier.initCarCommentaires(_cComments);
                fnomFichier.initSeparateurs(_cSeparators);
#ifndef OLDCLUB
		fnomFichier.initFieldSeparateur(_cFieldSeparator);
#endif
                // Elimination des lignes de commentaires
                //fnomFichier.elimineBlancs();

                fnomFichier.elimineCommentaires();

                // Test si le fichier ainsi obtenu est vide
                if (fnomFichier.estVide())
                {
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__,__LINE__,
							       Form("TInputFileKey::searchIndexeAttribute ERROR\n The file [%s] is empty",
                                                                    this->GetName()));
                }

                if (_blog)
                {
                    cout << " ** Opening file " << fnomFichier.nom()
			 << " ended." << endl;
                    cout << " ** fichier " << fnomFichier.nom() << " : "
			 << fnomFichier.nombreChamps() << " champs" << endl;
                    cout << " ** _cComments [" << _cComments << "]" << endl;
                }

                // acces direct
                Int_t nchamps = fnomFichier.nombreChamps();
                for (Int_t i = 1; i <= nchamps; i++)
                {
                    const char* champ = fnomFichier.champ(i);
                    TString strchamp(champ);
#ifndef OLDCLUB
                    delete[] champ;
#endif
		    if (_blog)
                        cout << "champ[" << i << "] valeur[" << strchamp
			     << "], skey[" << skey << "]      -------- \n";

                    if (strchamp.CompareTo(skey.Data(), TString::kExact) == 0)
                    {
                        if (_blog)
                            cout << "********** champ[" << i << "] valeur["
				 << strchamp << "] skey[" << skey << "]";
                        list<URANIE::DataServer::TAttributeFileKey *>::iterator first =
			    KeyList.begin();
                        att->setFieldOfKey(distance(first, it), i + 1);
                        nfield = i + 1;
                        i = nchamps;
                        if (_blog)
                        {
                            const char* champi = fnomFichier.champ(nfield);
                            cout << "---------- champ[" << nfield - 1
				 << "] valeur[" << strchamp << "] ["
				 << champi << "]" << endl;
#ifndef OLDCLUB
                            delete[] champi;
#endif
                        }
                    }
                }
                if (nfield == -1)
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__,
							       __LINE__,
							       Form(
								   "TInputFileKey::searchIndexeAttribute ERROR\n The key [%s] in the TInputFileKey[%s] not found",
								   skey.Data(), ((*it)->_sFile).Data()));
            }
#ifdef OLDCLUB
            catch (ClubErreurs ce)
            {
                ce.affiche();
                throw ce;
            }
#else
            catch (TamponTexteError& ex)
            {
                ex.printMessage();
                throw ex;
            } catch (URANIE::Exceptions::UErrorExceptions& ue)
            {
                cout << "message 1" << endl;
                ue.printMessage();
                throw ue;
            }
#endif
        }
    }

}

void TInputFileKey::substitue(Int_t &n, vector<URANIE::DataServer::UEntry*> *ent, Int_t *index)
{
    if (_blog)
        cout << "** TInputFileKey::substitue ****" << endl;

    TString beg, end;
    for (Int_t i = 0; i < getAttributesNumber(); i++)
    {
        list<URANIE::DataServer::TAttributeFileKey *> KeyList =  _attributes[i]->getKeyList();
        for (list<URANIE::DataServer::TAttributeFileKey *>::iterator it = KeyList.begin(); it != KeyList.end(); ++it)
        {
            TString sfilename = gSystem->BaseName((*it)->_sFile.Data());

            if (_blog)
                cout << "\nTInputFileKey::substitue sfilename[" << sfilename  << "]" << endl;

            TString sfilecrt = (*it)->_sFile;
            if (sfilecrt.CompareTo(_sFileName, TString::kExact) == 0)
            {
                try
                {
                    TamponTexte fnomFichier(sfilename, 0);
                    // Define the comment and separator character
                    fnomFichier.initCarCommentaires(_cComments);
                    fnomFichier.initSeparateurs(_cSeparators);
#ifndef OLDCLUB
		    fnomFichier.initFieldSeparateur(_cFieldSeparator);
#endif
                    // Elimination des lignes de commentaires
                    fnomFichier.elimineCommentaires();

                    // Elimination des lignes vides
                    //fnomFichier.elimineBlancs();

                    // Test si le fichier ainsi obtenu est vide
                    if (fnomFichier.estVide())
                    {
                        throw URANIE::Exceptions::UErrorExceptions(__FILE__,__LINE__,
                                                                   Form("TInputFileKey::searchIndexeAttribute ERROR\n The file [%s] is empty",
                                                                        this->GetName()));
                    }

                    TString sformattosubstitute = TString((*it)->_sFormat);

                    if (_blog)
                        cout << " ********************** key["
			     << Int_t((*it)->_sEventList->GetEntry(0))
			     << "] value[" << ent->at( ((index==NULL) ? n : index[n - 1]) )->dumpStr() << "] FormatSubstitute["
			     << sformattosubstitute << "]" << endl;

		    
		    getBegEndProperties( _attributes[i], beg, end);
		    
		    fnomFichier.remplaceChamp(
			Int_t((*it)->_sEventList->GetEntry(0)),
			ent->at( ((index==NULL) ? n : index[n - 1]) )->dumpStr( _svecdel, beg, end, (*it)->_sFormat).c_str());

                    if (_blog)
                        cout << " Write the file ********************** "
			     << endl;
                    fnomFichier.ecrit();
                }
#ifdef OLDCLUB
                catch (ClubErreurs ce)
                {
                    throw ("TAttribute::substitue Error");
                }
#else
                catch (TamponTexteError& ex)
                {
                    ex.printMessage();
                    throw("TAttribute::substitue Error");
                } catch (URANIE::Exceptions::UErrorExceptions& ue)
                {
                    ue.printMessage();
                    throw("TAttribute::subsitue error");
                }
#endif
            }

        }
        n++;
    }
}

void TInputFileKey::substitue(Int_t &n, Double_t *val, Int_t *index)
{
    if (_blog)
        cout << "** TInputFileKey::substitue ****" << endl;
    for (Int_t i = 0; i < getAttributesNumber(); i++)
    {
        list<URANIE::DataServer::TAttributeFileKey *> KeyList =
	    _attributes[i]->getKeyList();
        for (list<URANIE::DataServer::TAttributeFileKey *>::iterator it =
		 KeyList.begin(); it != KeyList.end(); ++it)
        {
            TString sfilename = gSystem->BaseName((*it)->_sFile.Data());

            if (_blog)
                cout << "\nTInputFileKey::substitue sfilename[" << sfilename
		     << "]" << endl;

            TString sfilecrt = (*it)->_sFile;
            if (sfilecrt.CompareTo(_sFileName, TString::kExact) == 0)
            {
                try
                {
                    TamponTexte fnomFichier(sfilename, 0);
                    // Define the comment and separator character
                    fnomFichier.initCarCommentaires(_cComments);
                    fnomFichier.initSeparateurs(_cSeparators);
#ifndef OLDCLUB
		    fnomFichier.initFieldSeparateur(_cFieldSeparator);
#endif 
                    // Elimination des lignes de commentaires
                    fnomFichier.elimineCommentaires();

                    // Elimination des lignes vides
                    //fnomFichier.elimineBlancs();

                    // Test si le fichier ainsi obtenu est vide
                    if (fnomFichier.estVide())
                    {
                        throw "URANIE::Launcher::TCode::substitue the file is empty";
                    }

                    TString sformattosubstitute = TString((*it)->_sFormat);
                    Double_t dValue = 0.0;
                    // in case of the vector of index is not empty, use it to catch the value
                    if (index != NULL)
                    {
                        dValue = val[index[n - 1]];
                        if (_blog)
                            cout << " **** sfilename[" << sfilename << "] n["
				 << n << "] index[n][" << index[n - 1]
				 << "] dValue[" << dValue << "]" << endl;
                    }
                    else
                    {
                        dValue = val[n];
                        if (_blog)
                            cout << " **** sfilename[" << sfilename << "] n["
				 << n << "] index[n][NULL] dValue[" << dValue
				 << "]" << endl;
                    }
                    if (_blog)
                        cout << " ********************** key["
			     << Int_t((*it)->_sEventList->GetEntry(0))
			     << "] value[" << dValue << "] FormatSubstitute["
			     << sformattosubstitute << "]" << endl;

                    if (sformattosubstitute.Contains("%d"))
                    {
                        Int_t nval = (Int_t)(TMath::Ceil(dValue));
                        fnomFichier.remplaceChamp(
			    Int_t((*it)->_sEventList->GetEntry(0)),
			    Form(sformattosubstitute.Data(), nval));
                    }
                    else
                        fnomFichier.remplaceChamp(
			    Int_t((*it)->_sEventList->GetEntry(0)),
			    Form(sformattosubstitute.Data(), dValue));

                    if (_blog)
                        cout << " Write the file ********************** "
			     << endl;
                    fnomFichier.ecrit();
                }
#ifdef OLDCLUB
                catch (ClubErreurs ce)
                {
                    throw ("TAttribute::substitue Error");
                }
#else
                catch (TamponTexteError& ex)
                {
                    ex.printMessage();
                    throw("TAttribute::substitue Error");
                } catch (URANIE::Exceptions::UErrorExceptions& ue)
                {
                    ue.printMessage();
                    throw("TAttribute::subsitue error");
                }
#endif
            }

        }
        n++;
    }
}

void TInputFileKey::printLog(Option_t *option)
{
    cout << "TInputFileKey::printLog[" << option << "]" << endl;
    cout << "_sFileName[" << _sFileName << "]" << endl;
    cout << "_cComments[" << _cComments << "]" << endl;
    cout << "_cSeparators[" << _cSeparators << "]" << endl;
    cout << "_cFieldSeparator[" << _cFieldSeparator << "]" << endl;
    Int_t natt = (Int_t)(_attributes.size());
    cout << " nAttribute[" << natt << "]" << endl;
    for (Int_t indx = 0; indx < natt; indx++)
    {
        cout << "index[" << indx + 1 << "/" << natt << "]" << endl;
        _attributes[indx]->printLog();
    }
}

} // Fin du namespace Launcher
} // Fin du namespace URANIE

