//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TInputFileRecreate.cxx
 \author Cedric Forgues <uranietm@cea.fr>
 \date   thu sep 24th 15:40:28 CEST 2009
 
 \brief Implementation of the class URANIE::Launcher::TInputFileRecreate
 
 \par Modifications
 \li F. Gaudier (2010-12-14 11:38:09)
 - resolve the DI111 and DI113

 */
#include <iostream>
#include<fstream>
using std::cout;
using std::endl;

// ROOT
#include "TMath.h"

// URANIE
#include "UExceptions.h"
#include "TInputFileRecreate.h"

ClassImp (URANIE::Launcher::TInputFileRecreate)

namespace URANIE
{
namespace Launcher
{

TInputFileRecreate::TInputFileRecreate(TString str) :
        TInputFile(str), _cHeaderComments(""), _cFooterComments("")
{
    if (_blog)
        cout << "constructing TInputFileRecreate file[" << str << "]" << endl;
    SetTitle(Form("TInputFileRecreate with name[%s]", str.Data()));
    setSeparatorCharacter(" \t\r\n");
    _svecdel=","; _svecbeg = _svecend = "";
    _sstrbeg = _sstrend = "";
}

TInputFileRecreate::~TInputFileRecreate()
{
    if (_blog)
        cout << "destroying TInputFileRecreate" << endl;
}

bool TInputFileRecreate::IsAWhiteLine(TString line)
{
    for (Int_t i = 0; i < Int_t(sizeof(line)); i++)
    {
        if (line[i] == ' ')
            continue;
        else if (line[i] == '\0' || line[i] == '\n')
            return true;
        else
            return false;
    }
    return true;
}

void TInputFileRecreate::searchIndexeAttribute( URANIE::DataServer::TAttribute* att)
{
    UNUSED(att); //Function not yet implemented but here has it comes from pure virtual
    if (_blog)
    {
        cout
                << "TInputFileRecreate::searchIndexeAttribute(URANIE::DataServer::TAttribute*) launch\t";
        cout << "TCodeFile::addAttribute(URANIE::DataServer::TAttribute*)."
                << endl;
    }
}

void TInputFileRecreate::substitue(Int_t &n, Double_t *val, Int_t *index)
{
    if (_blog)
        cout << "** TInputFileRecreate::substitue ****" << endl;
    std::ofstream f(gSystem->BaseName(getFileName()), std::ios::trunc);
    if (getHeader().Length() > 0)
        f << getHeader() << endl;
    for (Int_t i = 0; i < getAttributesNumber(); i++)
    {
        if (_blog)
            cout << " ** " << i << "/" << getAttributesNumber() << " -- name["
                    << _attributes[i]->GetName() << "]" << endl;
        list<URANIE::DataServer::TAttributeFileKey *> KeyList =
                _attributes[i]->getKeyList();
        for (list<URANIE::DataServer::TAttributeFileKey *>::iterator it =
                KeyList.begin(); it != KeyList.end(); ++it)
        {
            switch ((*it)->_sFileType)
            {
            case URANIE::DataServer::TAttributeFileKey::kNewKey:
                if (index != NULL)
                    f << (*it)->_sKey << " = "
                            << Form(((*it)->_sFormat).Data(), val[index[n - 1]])
                            << " ;" << endl;
                else
                    f << (*it)->_sKey << " = "
                            << Form(((*it)->_sFormat).Data(), val[n]) << " ;"
                            << endl;
                break;
            case URANIE::DataServer::TAttributeFileKey::kNewRow:
            case URANIE::DataServer::TAttributeFileKey::kNewLine:
                if (index != NULL)
                    f << Form(((*it)->_sFormat).Data(), val[index[n - 1]])
                            << " ";
                else
                    f << Form(((*it)->_sFormat).Data(), val[n]) << " ";
                break;
            case URANIE::DataServer::TAttributeFileKey::kNewColumn:
                if (index != NULL)
                    f << Form(((*it)->_sFormat).Data(), val[index[n - 1]])
                            << endl;
                else
                    f << Form(((*it)->_sFormat).Data(), val[n]) << endl;
                break;
	    default:
		break;
            }
        }
        n++;
    }
    if (getFooter().Length() > 0)
        f << endl << getFooter() << endl;
    f << endl;
    f.close();
}

void TInputFileRecreate::substitue(Int_t &n, vector<URANIE::DataServer::UEntry*> *ent, Int_t *index)
{
    
    if (_blog)
        cout << "** TInputFileRecreate::substitue ****" << endl;

    ofstream f(gSystem->BaseName(getFileName()), std::ios::trunc);
    if (getHeader().Length() > 0)
        f << getHeader() << endl;

    bool thingstobewritten=true;
    bool moretowrite;
    unsigned int nentry=0;
    int nvalue=n;
    TString beg, end;
    while( thingstobewritten )
    {
	thingstobewritten = false;
	n=nvalue;
	for (Int_t i = 0; i < getAttributesNumber(); i++)
	{
	    if (_blog)
		cout << " ** " << i << "/" << getAttributesNumber() << " -- name["<< _attributes[i]->GetName() << "], index being[" << ((index==NULL)? n : index[n-1]) << "]" << endl;

	    list<URANIE::DataServer::TAttributeFileKey *> KeyList = _attributes[i]->getKeyList();
	    getBegEndProperties( _attributes[i], beg, end);
	    for (list<URANIE::DataServer::TAttributeFileKey *>::iterator it = KeyList.begin(); it != KeyList.end(); ++it)
	    {
		if(_blog)
		    cout<<"  ** Using _sFileType "<<(*it)->_sFileType<<endl;
		
		switch ((*it)->_sFileType)
		{
		case URANIE::DataServer::TAttributeFileKey::kNewKey:
		    
		    if (index != NULL)
			f << (*it)->_sKey << " = " << ent->at(index[n-1])->dumpStr(_svecdel, beg, end, (*it)->_sFormat) << " ;" << endl;
		    else
			f << (*it)->_sKey << " = " << ent->at(n)->dumpStr(_svecdel, beg, end, (*it)->_sFormat) << " ;" << endl;
		    
		    break;
		case URANIE::DataServer::TAttributeFileKey::kNewRow:
		case URANIE::DataServer::TAttributeFileKey::kNewLine:

		    if (index != NULL)
			f << ent->at(index[n-1])->dumpStr(nentry, moretowrite, beg, end, (*it)->_sFormat)  << ((i==getAttributesNumber()-1) ? "\n" : _cFieldSeparator);
		    else
			f << ent->at(n)->dumpStr(nentry, moretowrite, beg, end, (*it)->_sFormat) << ((i==getAttributesNumber()-1) ? "\n" : _cFieldSeparator); 

		    thingstobewritten |= moretowrite;

		    break;
		case URANIE::DataServer::TAttributeFileKey::kNewTDS:

		    if(_blog)
			cout<<"Coucou je suis la "<<ent->at(((index==NULL)? n : index[n-1]))->dumpStr(_svecdel, beg, end, (*it)->_sFormat) << endl;
		    
		    if (index != NULL)
			f << ent->at(index[n-1])->dumpStr(_svecdel, beg, end, (*it)->_sFormat) << ((i==getAttributesNumber()-1) ? "\n" : _cFieldSeparator);
		    else
			f << ent->at(n)->dumpStr(_svecdel, beg, end, (*it)->_sFormat) << ((i==getAttributesNumber()-1) ? "\n" : _cFieldSeparator);
		    
		    break;
		case URANIE::DataServer::TAttributeFileKey::kNewColumn:

		    if (index != NULL)
			f << ent->at(index[n-1])->dumpStr(" ", beg, end, (*it)->_sFormat) << endl;
		    else
			f << ent->at(n)->dumpStr(" ", beg, end, (*it)->_sFormat) << endl; 

		    break;
		default:
		    break;
		}
	    }
	    n++;
	}
	nentry++;
    }
    if (getFooter().Length() > 0)
        f << endl << getFooter() << endl;

    f.close();
}

void TInputFileRecreate::printLog(Option_t *option)
{
    cout << "TInputFileRecreate::printLog[" << option << "]" << endl;
    cout << "_sFileName[" << _sFileName << "]" << endl;
    cout << "_cComments[" << _cComments << "]" << endl;
    cout << "_cSeparators[" << _cSeparators << "]" << endl;
    cout << "_cHeaderComments[" << _cHeaderComments << "]" << endl;
    cout << "_cFooterComments[" << _cFooterComments << "]" << endl;
    Int_t nAttribute = (Int_t)(_attributes.size());

    TString soption = TString(option);
    soption.ToLower();

    if (soption.Contains("att"))
    {
        for (Int_t indx = 0; indx < nAttribute; indx++)
        {
            cout << "index[" << indx + 1 << "/" << nAttribute << "]" << endl;
            _attributes[indx]->printLog();
        }
    }
}

} // Fin du namespace Launcher
} // Fin du namespace URANIE

