//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TLauncher.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mon oct 9th 11:37:19 CEST 2006
 
 \brief Interface of the class URANIE::Launcher::TLauncher
 
 */

/**
 
 * @class URANIE::Launcher::TLauncher
 *
 * @brief Description of the class TLauncher.
 *
 * To be written by the developper.
 *
 */
#ifndef TLAUNCHER_H
#define TLAUNCHER_H

#define MAX_PATTERN 980

//ROOT
#include "Rtypes.h"
#include "TString.h"

// Uranie
#include "Launcher.h"
#include "TCode.h"
#include "TDataServer.h"
#include "UMessageLogger.h"
#include "UExceptions.h"

#include <iostream>
#include <fstream>
#include <iomanip>

using std::cout;
using std::endl;
using std::cerr;
using std::setw;

namespace URANIE
{
namespace Launcher
{
class TLauncher: public TNamed
{
protected:
    // the mutable declaration is needed to use the logger in const methods
    URANIE::DataServer::UMessageLogger *_fLogger; // message logger

    //! Set the "only final file" flag
    /*!
     The output file is not created at every iteration, but only at the end. To be used only for (very) quick 
     code when re-running it is not an issue
     The method is put in protected so that this option is changed only by option "noIntermediateSaved" in run()
     */
    void setNoIntermediateSaved(Bool_t bbool = kTRUE)
    {
        _bnoIntermediateSaved = bbool;
    }
    
public:
    TCode* _code;
    TString _sdraw;             // Variables to visualize
    TString _sopt;              // Option of visualisation
    TString _scut;              // Selection of visualisation
    TString _sdirectory;        ///< Current directory
    TString _sworkingDirectory; ///< The Working directory
    TString _sPreProCommand;     ///< The prepro command before each parralel computation
    TString _sPostProCommand;    ///< The prepro command after  each parralel computation
    Bool_t _bsave;        ///< Saves all the computation files in each directory
    Bool_t _bclean;     ///< Cleans the Working Directory before runing the code
    Bool_t _blog;               ///< Boolean to edit the log
    Bool_t _DoLogger;               ///< Boolean to edit the log
    Bool_t _DoMultiDump;               ///< Boolean to edit the log
    Bool_t _bdrawProgressBar; ///< Boolean to know if the progress bar has to be drawn
    Bool_t _bnoIntermediateSaved; ///< Boolean to decide whether to keep or not the intermediate state
    Bool_t _bemptyMasterNode;
    Int_t _nprocs;             ///< The number of process for one job
    Int_t _nMaxSave;           ///< The authorized number of saves
    Int_t _delay; /// temporisation in seconds between first and last execution on a batch system 
    TList *_listOfSaveFile; ///< The list of output files to save with the number of run
    URANIE::DataServer::TDataServer *_tds;      //! Pointer vers un TDS
    vector<URANIE::DataServer::UEntry*> _InputVector, _OutputVector;
    
    //---------------------------------------------
    /** @name Private Methods to manipulate the code
     **/
    //@{
    //! Builts Results directory, no available options
    virtual void createResultsDirectory(Option_t *option = "");
    //! Create the ntupleD to store the output results
    /*!
      This function is separated from the main loop to be reused in the TLauncherByStep (keep as much common code as possible)
      \param sntlname: name of the tuple
      \param slabelvarname: name ot the branches to be created
      \param slabelvartype: type ot the branches to be created
     */
    URANIE::DataServer::TDSNtupleD* createNtupleDFile(TString &sntlname, TString &slabelvarname, TString &slabelvartype);

    //! Create fill correctly the UEntry vectors
    /*!
      This function is separated from the main loop to be reused in the TLauncherByStep (keep as much common code as possible)
      \param in: address of the vector of UEntry for inputs
      \param out: address of the vector of UEntry for outputs
     */
    void initiateTheEntryVector();
    //@}

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    TLauncher();
    //! Constructor with a dataserver
    TLauncher(URANIE::DataServer::TDataServer *tds, TCode *code, Option_t *option="");
    //! Default destructor
    virtual ~TLauncher();
    //@}

    //---------------------------------------------
    /** @name Manipulate the code
     **/
    //@{
    //! Runs phase of the code
    virtual void run(Option_t *option = "");

protected:
    //! Init phase of the code
    virtual void init(Option_t *option = "");
    //! Terminate phase of the code
    virtual void terminate(Option_t *option = "");
    //! Cleans the working directory of the code
    virtual void clean(Option_t *option = "");
    //@}

    //---------------------------------------------
    /** @name Setting and Getting attributes
     **/
    //@{
    //! Sets the \em working \em directory where to realize the computations
    /*!
     The \em working \em directory does not end with a slash (\) or a backslash (/) symbols.
     \code
     lanceur->setWorkingDirectory("/home/uranie/tmp/UranieLauncher/myCode");
     \endcode
     \param str (TString) the working directory where to realize the computations.
     */
public:
    void setWorkingDirectory(TString str);
    //! Gets the working directory
    /*!
     \return (TString) the working directory
     */
    TString getWorkingDirectory()
    {
        return _sworkingDirectory;
    }

    //! Set the \it "prepro" command to execute before each parallel computation
    /*!
     \brief Set the \it "prepro" command to execute before each parallel computation
     \param sCommand (TString) The \it "prepro" command to execute beafore each parallel computation
     */
    void setPreProCommand(TString sCommand) { _sPreProCommand = sCommand;}

    TString getPreProCommand() { return _sPreProCommand;}

    //! Set the \it "postpro" command to execute after each parallel computation
    /*!
     \brief Set the \it "postpro" command to execute after each parallel computation
     \param sCommand (TString) The \it "prepro" command to execute beafore each parallel computation
     */
    void setPostProCommand(TString sCommand) { _sPostProCommand = sCommand;}
    TString getPostProCommand() { return _sPostProCommand;}

    //! Activates the save mode and sets number of max saves to perform. 
    /*!
     A positive value stands for the explicit value of max saves.
     -1 allows to set max number of saves to the number of patterns.
     \param nSave (Int_t) [-1] Defines the number of saved patterns (working subdirectory).
     The default value [-1] means that all patterns are launched in a different working directory \em UranieLauncher_i
     */
    void setSave(Int_t nSave = -1);

    //! Desactivates the save mode and unsets the specified number of max saves.
    void unsetSave()
    {
        _nMaxSave = MAX_PATTERN;
        _bsave = kFALSE;
    }

    //! Get the save flag
    /*!
     \return (Bool_t) the \em save flag (boolean)
     */
    Bool_t getSave()
    {
        return _bsave;
    }

    //! Set the clean flag
    /*!
     The clean flag means that before launching, all the working directory are cleaned,
     and, if \em save == FALSE, after the launching all the working directory are also cleaned.
     */
    void setClean(Bool_t bbool = kTRUE)
    {
        _bclean = bbool;
    }

    //! Get the \em clean flag
    /*!
     \return (Bool_t) the \em clean flag (boolean)
     */
    Bool_t getClean()
    {
        return _bclean;
    }

    //! Set the "draw progress bar" flag
    /*!
     The progress bar will be drawn if and only if the "draw progress bar"
     flag is true.
     */
    void setDrawProgressBar(Bool_t bbool = kTRUE)
    {
        _bdrawProgressBar = bbool;
    }

    //! Get the \em clean flag
    /*!
     \return (Bool_t) the \em clean flag (boolean)
     */
    Bool_t getDrawProgressBar()
    {
        return _bdrawProgressBar;
    }

    //! Get the \em clean flag
    /*!
     \return (Bool_t) the \em clean flag (boolean)
     */
    Bool_t getNoIntermediateSaved()
    {
      return _bnoIntermediateSaved;
    }
    
    //! Defines the output file to save during the launching process
    /*!
     \param sfile (TString) the name of the file to store during the process
     */
    void addSaveFile(TString sfile);
    //! Activates and defines the graphics during the launching
    /*!
     By default, there is no graphic during the launchin process.
     To activate it, you must defines the informations of the graphics : the X, Y and Z attributes, the select and the options of the graphic.
     \param str (TString) the attributes to visualize ("Z:Y:X")
     \param select (TString) [""] the \em "select" of patterns to visualize
     \param opt (TString) [""] the options of the graphic
     */
    void setVarDraw(TString str, TString select = "", TString opt = "")
    {
        _sdraw = str;
        _scut = select;
        _sopt = opt;
    }

    //@}

    //---------------------------------------------
    /** @name Number of Process per Jobs
     **/
    //@{
    inline void setProcsPerJob(int procs)
    {
        _nprocs = procs;
    }
    inline int getProcsPerJob()
    {
        return _nprocs;
    }
    /** @name Temporisation in seconds for large number of procs **/
    void setDelay(int delay)
    {
        _delay = delay;
    }
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    void setLog()
    {
        _blog = kTRUE;
    }
    void unsetLog()
    {
        _blog = kFALSE;
    }
    void changeLog()
    {
        _blog = _blog ? kFALSE : kTRUE;
    }
    Bool_t getLog()
    {
        return _blog;
    }
    virtual void printLog(Option_t *option = "");
    //@}
    /** @name MultiLauncher strategy
     **/
    //@{
    void setEmptyMasterNode()
    {
        _bemptyMasterNode = kTRUE;
    }
    void unsetEmptyMasterNode()
    {
        _bemptyMasterNode = kFALSE;
    }
    //@}

    ClassDef(URANIE::Launcher::TLauncher, ID_LAUNCHER)
    //Classe de
};

} // Fin du namespace Launcher
} // Fin du namespace URANIE
#endif
