//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TLauncher.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mon oct 9 11:37:19 CEST 2006
 
 \brief Implementation de la classe URANIE::Launcher::TLauncher
 
 */

#include "TLauncherMulti.h"
#include "TSlurmCluster.h"
#include "TLSFCluster.h"
#include "TSGECluster.h"
#include "TLocalHostCluster.h"
#include "TLoadLevelerCluster.h"

#include <sys/types.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <unistd.h>
#include <fstream>

#include "TSystem.h"

#include "TFile.h"
#include "TChain.h"

using std::cout;
using std::cerr;
using std::endl;

ClassImp (URANIE::Launcher::TLauncherMulti);

namespace URANIE
{
namespace Launcher
{

TLauncherMulti::TLauncherMulti() :
        TNamed("_uranie_TLauncherMulti_", "multi lancher uranie"),
        //_edistrib(kUnknown),
        _blog(kFALSE), _bemptyMasterNode(kFALSE), njobs_(0), nb_avail_procs_(1)
{
    _edistrib = NULL;
    createHostsList("local=2");
    if (_blog)
        cout << "end of TLauncherMulti () constructor" << endl;
}
TLauncherMulti::TLauncherMulti(const char* name, const char* title,
                               Option_t *opt, Bool_t emptyMasterNode, Int_t delay, bool dumpmulti) :
        TNamed(name, title),
        //_edistrib(kUnknown),
        _blog(kFALSE), _bemptyMasterNode(emptyMasterNode), _bmultidump(dumpmulti), njobs_(0), nb_avail_procs_(1), delay_(delay)
{
    _edistrib = NULL;
    createHostsList(opt);
    if (_blog)
        cout << "end of TLauncherMulti (name,title,opt) constructor" << endl;
}

TLauncherMulti::~TLauncherMulti()
{
    if (_edistrib)
        delete _edistrib;
    _edistrib = NULL;
}

void TLauncherMulti::createHostsList(Option_t *opt)
{
    _masterpid = getpid();
    if (_blog)
        cerr << " *** _masterpid:" << _masterpid << endl;

    TString sListOfHosts = "";

    if (_blog)
        cerr << " ** sListOfHosts[" << sListOfHosts << "]" << endl;

    if ((!_edistrib) && TSlurmCluster::eligible())
    {
        _edistrib = new TSlurmCluster();
        if (_blog)
            _edistrib->setLog();
        sListOfHosts = _edistrib->getListOfNodes();
        cerr << " ***************** SLURM_NODELIST = " << _edistrib->GetTitle()
                << endl;
        cerr << " ***************** SLURM_NODELIST = " << sListOfHosts << endl;
        cout << " ***************** SLURM_NODELIST = " << sListOfHosts << endl;
    }
    
    if ((!_edistrib) && TLoadLevelerCluster::eligible())
    {
        _edistrib = new TLoadLevelerCluster();
        sListOfHosts = _edistrib->getListOfNodes();
        cerr << " ***************** TLoadLevelerCluster = [" << _edistrib->GetTitle()
                << "]" << endl;
        cerr << " ***************** TLoadLevelerCluster = [" << sListOfHosts << "]"
                << endl;
        cout << " ***************** TLoadLevelerCluster = [" << sListOfHosts << "]"
                << endl;
    }

    if ((!_edistrib) && TLSFCluster::eligible())
    {
        _edistrib = new TLSFCluster();
        sListOfHosts = _edistrib->getListOfNodes();
        if (_blog)
        {
            cerr << " ***************** LSF_HOSTS = " << _edistrib->GetTitle()
                    << endl;
            cerr << " ***************** LSF_HOSTS = " << sListOfHosts << endl;
            cout << " ***************** LSF_HOSTS = " << sListOfHosts << endl;
        }
    }

    if ((!_edistrib) && TSGECluster::eligible())
    {
        _edistrib = new TSGECluster();
        sListOfHosts = _edistrib->getListOfNodes();
        cerr << " ***************** SGE_HOSTFILE = [" << _edistrib->GetTitle()
                << "]" << endl;
        cerr << " ***************** SGE_HOSTFILE = [" << sListOfHosts << "]"
                << endl;
        cout << " ***************** SGE_HOSTFILE = [" << sListOfHosts << "]"
                << endl;
    }

    if ((!_edistrib) && TLocalHostCluster::eligible(opt))
    {
        _edistrib = new TLocalHostCluster();
        ((TLocalHostCluster*)_edistrib)->setMultiDump(_bmultidump);
        sListOfHosts = _edistrib->getListOfNodes();
        cerr << " ***************** LocalHost = [" << _edistrib->GetTitle()
                << "]" << endl;
        cerr << " ***************** LocalHost = [" << sListOfHosts << "]"
                << endl;
        cout << " ***************** LocalHost = [" << sListOfHosts << "]"
                << endl;
    }

    // In case of any TLocalHost::eligible is OK, we are in sequential mode
    if (_edistrib == NULL)
    {
        if (_blog)
            cout << " ** Mode : Sequential" << endl;
        return;
    }

    if (_blog)
        cout << " ** Mode : Distribution " << endl;

    istringstream hostsTmp((char *) sListOfHosts.Data());

    while (!hostsTmp.eof())
    {
        string tmp;
        hostsTmp >> tmp;
        if (tmp.size())
            hostslist_.push_back(tmp);
    }
    // eliminating the procs on the master node for large jobs (bottleneck problem)
    if (_bemptyMasterNode)
    {
        string master = *(hostslist_.begin());
        list<string>::iterator iter = hostslist_.begin();
        while (iter != hostslist_.end())
        {
            if (strcmp(iter->c_str(), master.c_str()) == 0)
                iter = hostslist_.erase(iter);
            else
                iter++;
        }
    }
    nb_avail_procs_ = hostslist_.size();
    nb_total_procs_ = hostslist_.size();
    //cout << "nombre de pros dispos : " << nb_avail_procs_ << endl;

    if (_blog)
    {
        cout << " ** MASTER JOB LAUNCHED, pid = " << _masterpid << " ** Proc["
                << nb_avail_procs_ << "]" << endl;
        cout << " ** LIST OF HOSTS **" << _edistrib->getListOfNodes() << endl;
        //system("uname -n"); //Masking warning complaining about not using returned value
        list<string>::iterator iterHost, iterEnd;
        for (iterHost = hostslist_.begin(), iterEnd = hostslist_.end();
                iterHost != iterEnd; iterHost++)
            cout << "\t" << *iterHost;
    }
}

void TLauncherMulti::defineJob(const char* exe, int njobs, int nprocs)
{
    exe_ = " ";
    exe_ += exe;

    njobs_ = njobs;
    procsperjob_ = nprocs;
    if (0)
    {
        cout << "\texe : " << exe_ << endl;
        cout << "\tnjobs : " << njobs_ << endl;
        cout << "\tnombre de pros dispos : " << nb_avail_procs_ << endl;
    }

}

void TLauncherMulti::waitForProcs()
{
    list<job>::iterator iterJob, iterJobEnd;
    job tmpJob, tmpJob2;
    int status, testpid = 0;
    int waitingTest = 0;

    while (waitingTest == 0)
    {
        list < string > tmpHostlist;
        list<string>::iterator iterHost, iterEnd;
        list<string>::iterator iterost, iterend;
        string name;
        for (iterJob = runningjobs_.begin(), iterJobEnd = runningjobs_.end();
                iterJob != iterJobEnd; iterJob++)
        {
            tmpJob2 = *iterJob;
            testpid = waitpid(tmpJob2.pid, &status, WNOHANG);
            // FGA::WARNING 2011-07-19 14:37:49
            //      This test was operated by V. Bergeaud, but it is not well fine 
            //the test for end of job is based on the two following events :
            // 1/ tespid==tmpJob2.pid means that a signal has been detected      
            // 2/ WIFEXITED(status) means that this signal
            //     is a real termination (and not just a stop)
            //        if (testpid==tmpJob2.pid && WIFEXITED(status)!=0) {
            // End Of FGA::WARNING 2011-07-19 14:37:49
            if (testpid == tmpJob2.pid)
            {
                tmpHostlist = tmpJob2.hostlist;
                for (iterHost = tmpHostlist.begin(), iterEnd = tmpHostlist.end();
                        iterHost != iterEnd; iterHost++)
                    hostslist_.push_back(*iterHost);
                //buildMasterTree(tmpJob2.testnumber); //REMPLACE PAR MERGEALL()
                runningjobs_.erase(iterJob);
                nb_avail_procs_ += procsperjob_;
                if (_blog)
                    cout
                            << " ** waitForProcs method:: nb avail procs after job finish ["
                            << nb_avail_procs_ << "]" << endl;
                waitingTest = 1;
                if (_blog)
                {
                    cout << "  In waitProcs:[";
                    for (iterost = hostslist_.begin(), iterend =
                            hostslist_.end(); iterost != iterend; iterost++)
                        cout << "\t" << *iterost;
                    cout << "]" << endl;
                }
                break;
            }
            if (waitingTest)
                continue;
        }

    } // End while(waitingTest==0)

}

void TLauncherMulti::createJob(int iteration)
{
    list<job>::iterator iterJob;
    job tmpJob, tmpJob2;

    if ((nb_avail_procs_ < procsperjob_))
    {
        // ------------------------------------------------
        // ---------- IF NO PROCS AVAILABLE, WAIT ---------
        // ------------------------------------------------
//         cout << "WAIT FOR JOB ENDED, pid = " << getpid() << " Iteration[" << iteration << "]" << endl;
        waitForProcs();

    } // End if (nb_avail_procs_ < procsperjob_)
    list<string>::iterator iterHost, iterEnd;

    // ---------------------------------------------
    // ---------- PROCS AVAILABLE, CREATE NEW ------
    // ---------------------------------------------
    char mpiexe[2500];
    char procs[2000];
    char singlehost[200];
    list < string > tmpHosts;
    list<string>::iterator iter = hostslist_.begin();
    pid_t pid;
    strcpy(procs, "");
    tmpHosts.clear();
    for (int nproc = 1; nproc <= procsperjob_; nproc++)
    {
        tmpHosts.push_back(*iter);
        strcpy(singlehost, iter->c_str());
        strcat(procs, singlehost);
        if (nproc < procsperjob_)
            strcat(procs, ",");
        iter++;
        hostslist_.pop_front();
    }
    tmpJob.hostlist = tmpHosts;

    pid = fork();
    // FGA::Remarque Probleme rencontre 05/08/2008 avec le cluster Platine
    // Alors Julien derouillat du support applicatif du CCRT propose de "ralentir" ce systeme pour ne pas avoir de probleme
    // Tests realise par fga le 07/08/2008 : test effectue avec le code "flowreate" avec sampling =lhs jusqu'a 900 pour ne pas avoir le problem "too many open files" avec nproc=8,16,32,64,128
    // End Of FGA::Remarque

    long microsecondsDelay = delay_ * 1000000 / nb_total_procs_;

    usleep(microsecondsDelay);

    /*if (pid == -1) {
     IF FORK NOT SUCCEED
     }*/

    if (getpid() == _masterpid)
    {
        // -----------------
        // ----- PERE ------
        // -----------------
        njobs_--;
        nb_avail_procs_ -= procsperjob_;
        if (_blog)
            cout
                    << " ** TLauncherMulti::createJob method :: fork with nb_avail ["
                    << nb_avail_procs_ << "] slots" << endl;
        tmpJob.pid = pid;
        tmpJob.testnumber = iteration;
        runningjobs_.push_back(tmpJob);

    } // END if (getpid()==pid_pere)
    else
    {
        // -----------------
        // ----- FILS ------
        // -----------------
        if( _bmultidump )
        {
            cout << " ** TLauncherMulti::createJob method :: Job Created, pid = "
                 << getpid() << " Iteration[" << iteration + 1 << "] host["
                 << procs << "]" << endl;
            cerr << " ** TLauncherMulti::createJob method :: Job Created, pid = "
                 << getpid() << " Iteration[" << iteration + 1 << "] host["
                 << procs << "]" << endl;
        }
        
        if (_edistrib)
            _edistrib->getCommandPrefix(mpiexe, procs);
        exe_ = mpiexe;
        if (_blog)
        {
            cout << " ** createJob method :: Debug de la Ligne de commande["
                    << exe_ << "]" << endl;
            cerr << " ** createJob method :: Debug de la Ligne de commande["
                    << exe_ << "]" << endl;
        }
        njobs_ = 0;

    } // END else

}

void TLauncherMulti::run(TCode* code, Option_t* option)
{
    TString tmp = code->getCommand();
    exe_ += " ";
    exe_ += tmp;
    code->setCommand(exe_);
    if( _bmultidump )
    {
        cout << " ** TLauncherMulti::run method :: Command line ["
             << code->getCommand() << "]" << endl;
        cerr << " ** TLauncherMulti::run method :: Command line ["
             << code->getCommand() << "]" << endl;
    }
    code->run(option);
}

void TLauncherMulti::waitAll()
{
    list<job>::iterator iterJob;
    job tmpJob2;
    int status, testpid = 0;
    int waitingTest;

    if (getpid() == _masterpid)
    {

        while (runningjobs_.size())
        {
            iterJob = runningjobs_.begin();
            tmpJob2 = *iterJob;
            waitingTest = 0;
            while (waitingTest == 0)
            {
                testpid = waitpid(tmpJob2.pid, &status, WNOHANG);
                if (testpid == tmpJob2.pid)
                {
                    //buildMasterTree(tmpJob2.testnumber); //REMPLACE PAR MERGEALL()
                    waitingTest = 1;
                }
            }
            runningjobs_.pop_front();
        }

    }

}

void TLauncherMulti::mergeAll(const int niteration)
{
  TString expname = GetName();
    cerr << " **  TLauncherMulti::mergeAll Name[" << expname << "]" << endl;

    cout << "TLauncherMulti::mergeAll(" << niteration << ")" << endl;
    cerr << "TLauncherMulti::mergeAll(" << niteration << ")" << endl;
    cout << "expname[" << expname << "]" << endl;
    cerr << "expname[" << expname << "]" << endl;
    TChain ch(expname);
    TString sfilsfile;

//       TList list;
    for (int i = 0; i < niteration; i++)
    {
        sfilsfile = _sworkingDirectory + "/UranieResults/uranielauncher_";
        sfilsfile += i;
        sfilsfile += ".root";
        std::ifstream file(sfilsfile);
        if (file.fail())
        {
            cout << "Skip " << sfilsfile
                    << " for merging because it doesn't exist!" << endl;
            cerr << "Skip " << sfilsfile
                    << " for merging because it doesn't exist!" << endl;
        }
        else
        {
            ch.Add(sfilsfile);
        }
    }

    sfilsfile = _sworkingDirectory + "/UranieResults/uranielauncher.root";
    ch.Merge(sfilsfile);
}

void TLauncherMulti::display()
{

    std::ofstream os(_sworkingDirectory + "/UranieResults/display.C");

    os << "void display() {" << endl << endl;
    ;
    // On ouvre le fichier ROOT "uranielauncher.root" en mode update
    os << "\tTFile *rt = new TFile(\"uranielauncher.root\",\"read\");" << endl;
    // Informations de ce fichier
    os << "\trt->Print();" << endl;
    // Listing du contenu du fichier
    os << "\trt->ls();" << endl;
    os << endl << "}" << endl;

    os.close();

}

// ----------------------------------------------------------------------
// Printing Log
// ----------------------------------------------------------------------
//---------------------------------------------
void TLauncherMulti::printLog(Option_t *option)
{
    cout << "TLauncherMulti::printLog[" << option << "]" << endl;
    cout << "Name[" << GetName() << "] Title[" << GetTitle() << "]" << endl;
    cout << " _edistrib[" << _edistrib << "]" << endl;
    cout << " njobs_[" << njobs_ << "]" << endl;
    cout << " nb_avail_procs_[" << nb_avail_procs_ << "]" << endl;
    cout << " _sworkingDirectory[" << _sworkingDirectory << "]" << endl;
}

} // Fin du namespace LAUNCHER
} // Fin du namespace URANIE

