//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TOutputFileDataServer.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   thu jun 3rd 16:06:57 CEST 2007
 
 \brief Interface of the class URANIE::Launcher::TOutputFileDataServer
 
 */
/**
 
 * @class URANIE::Launcher::TOutputFileDataServer
 *
 * @brief Description of the class TOutputFileDataServer.
 *
 This class enables to get scalars written by column, in an ASCII file.

 \verbatim
 1.5697507E-19  1.9395840E-19
 2.9158929E-16  3.6028774E-16
 3.5310144E-14  4.3629214E-14
 1.0450831E-12  1.2913046E-12
 1.3153202E-11  1.6252097E-11
 9.3449165E-11  1.1546579E-10
 4.4158568E-10  5.4562327E-10
 1.5389996E-09  1.9015880E-09
 4.2396255E-09  5.2384812E-09
 1.7353132E-07  2.1441529E-07
 7.4187462E-08  9.1666024E-08
 7.1243931E-09  8.8028997E-09
 3.5554371E-10  4.3930981E-10
 1.2805967E-11  1.5823053E-11
 1.3971713E-12  1.7263449E-12
 5.8563991E-12  7.2361665E-12
 2.2953027E-11  2.8360760E-11
 6.7645112E-11  8.3582301E-11
 2.6601454E-09  3.2868757E-09
 3.6597383E-09  4.5219726E-09
 4.1768415E-09  5.1609055E-09
 4.7367488E-09  5.8527267E-09
 5.3706906E-09  6.6360251E-09
 6.0915699E-09  7.5267437E-09
 6.9160828E-09  8.5455119E-09
 7.8621030E-09  9.7144142E-09
 \endverbatim
 This file contains 26 attributes and for each attributes, it contains two values (like temporal attributes).

 \par Example code source
 Let the file \b "data.out" shown in the last section.
 Suppose that we want to store the four attributes in the lines 1, 15, 22 and 26.
 For the \b out22 attribute, we set the default value to \b -1.0

 \verbatim
 TOutputFileDataServer * fout = new TOutputFileDataServer("data.out");
 
 fout->addAttribute( new TAttribute("out1"), 1);
 fout->addAttribute( new TAttribute("out15"), 15);

 TAttribute *attdose =  new TAttribute("out22");
 attdose->setDefaultValue(-1.0);
 fout->addAttribute(attdose, 22);
 
 fout->addAttribute( new TAttribute("out26"), 26);
 \endverbatim

 \see TOutputFile, \see TOutputFileRow

 */
#ifndef TOUTPUTFILEDATASERVER_H
#define TOUTPUTFILEDATASERVER_H

#include <iostream>

//ROOT
#include "Rtypes.h"

// Uranie
#include "Launcher.h"
#include "TOutputFile.h"
#include "UEntry.h"

namespace URANIE
{
namespace Launcher
{
class TOutputFileDataServer: public TOutputFile
{
private:
    Int_t _nLines;                ///< The number of lines in the file.
    Int_t _nColumns;              ///< The number of columnsin the file
    Int_t _nLength;       ///< The number of temporal data (_nColumns) to store.
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    TOutputFileDataServer(TString str);
    //! Default destructor
    virtual ~TOutputFileDataServer();
    //@}

    //---------------------------------------------
    /** @name Add attribut to the File
     **/
    //@{
    //!   Add the attribute and the line number in the file
    /*!
     The line number begins to 1.
     \par Log
     Modified since URANIE v2.0/1. 1 arg was added.
     */
    void addAttribute(URANIE::DataServer::TAttribute* att, Int_t nline = 0);
    //! Add the attribute by this name and the nline number in the file
    /*!
     The line number begins to 1.
     */
    void addAttribute(TString name, Int_t nline = 0);
    //@}

    /** @name Computes Output
     **/
    //@{
    //! Looking for output usindouble only and transmitting matrix (not used by launcher anymore)
    Bool_t searchOutput(Int_t &indcrt, Double_t *val, TList *lstOut = NULL);
    //! Looking for output using a map to be able to cope with vectors/strings
    Bool_t searchOutput(vector<URANIE::DataServer::UEntry*> *out);
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Launcher::TOutputFileDataServer, ID_LAUNCHER)
    //Classe de
};

} // Fin du namespace Launcher
} // Fin du namespace URANIE
#endif
