//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TOutputFileFMU.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Created : 2021/10/14 at 11:52:12

 \brief Interface of the class URANIE::Launcher::TOutputFileFMU
 
 */
/**
 
 @class URANIE::Launcher::TOutputFileFMU

 @brief Description of the class TOutputFileFMU.

 This classe enables to raise information from an FMU file

 */
#ifndef TOutputFileFMU_H
#define TOutputFileFMU_H

// ROOT
#include "Rtypes.h"
#include "TString.h"
#include "TList.h"
#include "TArrayI.h"

// URANIE
#include "Launcher.h"
#include "TOutputFile.h"

// FMI
#ifdef WIN32
#include <Windows4Root.h>
#endif
#include "fmilib.h"

#include <vector>

namespace URANIE
{
namespace Launcher
{
class TOutputFileFMU: public TOutputFile
{
public:

    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    /*!
     \param str Name of the XML output file
     */
    TOutputFileFMU(TString str);
    //! Default destructor
    virtual ~TOutputFileFMU();
    //@}

    // Public attribute accessor methods
    //
    //---------------------------------------------
    /** @name Setting and Getting attributs
     **/
    //@{

    Int_t getNOutputVector()
    {
        FUNCT_TRACE();
        return _nOutputVector;
    }
    TArrayI * getListOutputVector()
    {
        FUNCT_TRACE();
        if (_blog)
            cout << " *** method TOutputFileFMU::" << __func__ << " array size[" << _lstOfVector->GetSize() << "]" << endl;
        return _lstOfVector;
    }
    //@}

    //---------------------------------------------
    /** @name Add attribut to the File
     **/
    //@{
    //!   Add the attribute in the file
    /*!
     \param att (URANIE::DataServer::TAttribute*) the pointer to the attribute
     \param sKey (TString) the sKey for the attribute in the file.
     */
    void addAttribute(URANIE::DataServer::TAttribute* att, TString sKey = "");
    /*!
     \param sname (TString)  the name of the attribute to created.
     \param sKey (TString) the sKey for the attribute in the file.
     */
    void addAttribute(TString sname, TString sKey = "");


    void searchIndexeOutputAttribute(URANIE::DataServer::TAttribute* attOutput);

    void addDefaultOutput();

    void setVectorValues(vector<vector<double>> vvd)
    {
        _vecOutputs = vvd;
//        cout << " **** Method TOutputFileFMU::" << __func__ << " with vector size[" << _vecOutputs.size()<< "]"
//                " AND item 0 size[" << _vecOutputs[0].size() << "]" << endl;
    }



    //@}

    //---------------------------------------------
    /** @name Computes Output
     **/
    //@{
    //! Looking for output usindouble only and transmitting matrix (not used by launcher anymore)
    Bool_t searchOutput(Int_t &n, Double_t *val, TList *lstOut = NULL);
    //! Looking for output using a map to be able to cope with vectors/strings
    Bool_t searchOutput(vector<URANIE::DataServer::UEntry*> *out);
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{

    void setFMU1Import( fmi1_import_t* crtFMU1Import);

    void setFMU2Import( fmi2_import_t* crtFMU2Import);

    void setFMUVersion(fmi_version_enu_t theFMUVersion)
    {
        _sVersion = theFMUVersion;
    }


    void printLog(Option_t *option = "");
    //@}

    //---------------------------------------------
    /** @name protected attributes
     *
     **/
    //@{
protected:
    TString _spwd;                      ///< The relative path to suppress
    // FMI structures
    fmi_version_enu_t _sVersion;        ///< The FMU version
    fmi1_import_t* _fmi1Import;         ///< The FMU 1.0 Object
    fmi2_import_t* _fmi2Import;         ///< The FMU 2.0 Object

    Int_t _nOutputVector;               ///< the number of Output which is kVector
    Int_t _nCrtOutputVector;            ///< the curent number of Output which is kVector
    TArrayI *_lstOfVector;              ///< The list of RV for kVectors
    vector<vector<double>> _vecOutputs;  ///< the vector of stored kVector attributes

    // FMI General informations

    //@}




    ClassDef(URANIE::Launcher::TOutputFileFMU, ID_LAUNCHER)
    // End of Class
};

} // End of namespace Launcher
} // End of namespace URANIE

#endif // TInputFileFMU_H
// End of File
