//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TOutputFileRow.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mon sept 3rd 14:31:26 CEST 2007

 \brief Implementation of the class URANIE::Launcher::TOutputFileRow

 \par Modifications
\li Jean-Baptiste Blanchard 2016/04/08
 - Add the searchOutput with a map to read strings/vectors
 \li F. Gaudier (2010-08-06 11:25:44)
 - change the deprecated setKey method by setFileKey method
 \li Fabrice Gaudier (2010/03/22)
 - read the good format of the row data
 - 3 good methods \b addAttribute and 1 deprecated method

 */

//Uranie
#include "TOutputFileRow.h"
#include "UExceptions.h"

// ROOT
#include "TMatrixT.h"

// Club
#ifdef OLDCLUB
#include "club/TamponTexte.h"
#include "club/IterateurChamp.h"
#include "club/ClubErreurs.h"
#else
#include "TamponTexte.h"
#endif

using std::cout;
using std::endl;

ClassImp (URANIE::Launcher::TOutputFileRow)

namespace URANIE
{
namespace Launcher
{

TOutputFileRow::TOutputFileRow(TString str) :
    TOutputFile(str), _nLines(0), _nColumns(0)
{
    if (_blog)
	cout << "constructing TOutputFileRow file[" << str << "]" << endl;
    setCommentCharacter("#");
    setSeparatorCharacter(" \t");
}

TOutputFileRow::~TOutputFileRow()
{
    if (_blog)
	cout << "destroys TOutputFileRow" << endl;
}

/////////////////////
//// Add attribut to the Output File
/////////////////////
void TOutputFileRow::addAttribute(TString str, Int_t nrow)
{
    URANIE::DataServer::TAttribute *att = new URANIE::DataServer::TAttribute(str);
    att->delShare();
    addAttribute(att, nrow);
}

void TOutputFileRow::addAttribute(URANIE::DataServer::TAttribute* att,
				  Int_t nrow)
{
    Int_t nrowcrt = nrow;

    if (nrow < 1)
    {
        nrowcrt = 1;
        Int_t natt = getAttributesNumber();
        if (natt > 0)
	{
            list<URANIE::DataServer::TAttributeFileKey *> KeyList =
		_attributes[natt - 1]->getKeyList();
            list<URANIE::DataServer::TAttributeFileKey *>::iterator it =
		KeyList.begin();
            TEventList * eventfield = (*it)->_sEventList;
            nrowcrt = (Int_t) eventfield->GetEntry(0) + 1;
	}
    }

    att->setFileKey(_sFileName, att->GetName());
    att->setFieldOfKey(0, nrowcrt);
    TCodeFile::addAttribute(att);
}

/////////////////////
//// searchOutput
/////////////////////
Bool_t TOutputFileRow::searchOutput(Int_t &indcrt, Double_t *val, TList *lstOut)
{
    // Get the working file
    if (_blog)
    {
        cout << endl
	     << "************************\n*** TOutputFileRow::searchOutput indcrt["
	     << indcrt << "]" << endl;
        cout << " *** PWD[" << gSystem->pwd() << "]" << endl;
        cout << " *** _sFileName[" << _sFileName << "]" << endl;
    }

    // In case of the output file is not presented
    // put the default value
    //Number of attributes
    Int_t natt = getAttributesNumber();
    if (_blog)
    {
        cout << " *** natt[" << natt << "]" << endl;
    }

    TMatrixT<double> *matOut = new TMatrixT<double>(1, natt);

    // Default Value
    // In case of no output
    Double_t valdef = 1.23456789;

    Double_t * tempVal = matOut->GetMatrixArray();
    for (Int_t i = 0; i < natt; i++)
    {
        valdef = 1.23456789;
        _attributes[i]->getDefaultValue(valdef);
        tempVal[i] = valdef;
    }
    if (_blog)
    {
        cout << endl << " *** Default Value";
        matOut->Print();
        cout << " *** End Of Default Value" << endl;
    }

    try
    {
        TamponTexte fnomFichier(_sFileName, 0);
        fnomFichier.initSeparateurs(" \t");
        fnomFichier.initCarCommentaires(_cComments);
#ifndef OLDCLUB
	fnomFichier.initFieldSeparateur(_cFieldSeparator);
#endif
        // Elimination des lignes de commentaires
        fnomFichier.elimineCommentaires();
        // Elimination des lignes vides
        fnomFichier.elimineBlancs();
        // Test si le fichier ainsi obtenu est vide
        if (fnomFichier.estVide())
	{
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 Form(
							     "TOutputFileRow::searchOutput method\nthe file is empty [%s] is empty",
							     _sFileName.Data()));
	}

        // Set the number of lines in this file
        _nLines = fnomFichier.nombreLignes();
        _nColumns = fnomFichier.nombreChampsLigne(1);

        // Case of several lines
        if (_blog)
	{
            cout << " *** _nLines[" << _nLines << "]" << endl;
            cout << " *** _nColumns[" << _nColumns << "]" << endl;
	}

	delete matOut; 
        matOut = new TMatrixT<double>(_nLines, natt);
        Double_t * theVal = matOut->GetMatrixArray();

        Int_t ncrt = 0;
        Int_t nindcrt = indcrt;

        // Default Value for all the matrix
        for (Int_t ichamp = 0; ichamp < natt; ichamp++)
	{
            ncrt = ichamp;
            valdef = 1.23456789;
            _attributes[ichamp]->getDefaultValue(valdef);
            for (Int_t ilength = 0; ilength < _nLines; ilength++)
	    {
                if (ilength == 0)
		    val[nindcrt++] = valdef;
                theVal[ncrt] = valdef;
                ncrt += natt;
	    }
	}
        if (_blog)
	{
            cout << endl << " *** Default Value For the Matrix";
            matOut->Print();
            cout << " *** End Of Default Value For the Matrix" << endl;
	}

        // Case of all the matrix
        nindcrt = indcrt;
        for (Int_t iout = 1; iout <= natt; iout++)
	{
            ncrt = iout - 1;
            list<URANIE::DataServer::TAttributeFileKey *> KeyList =
		_attributes[iout - 1]->getKeyList();
            list<URANIE::DataServer::TAttributeFileKey *>::iterator it =
		KeyList.begin();
            TEventList * eventfield = (*it)->_sEventList;
            Int_t ichamp = (Int_t) eventfield->GetEntry(0);

            if (ichamp > _nColumns)
		throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							   Form(
							       "TOutputFileRow::searchOutput method\nthe output file [%s] contains only [%d] columns - not [%d] for attribute [%s]",
							       _sFileName.Data(), _nColumns, ichamp,
							       _attributes[iout - 1]->GetName()));

            for (Int_t ilength = 1; ilength <= _nLines; ilength++)
	    {
                const char* champ = fnomFichier.champLigne(ichamp, ilength);
                if (_blog)
		    cout << " *** iout(" << iout << "/" << natt << " :: "
			 << _attributes[iout - 1]->GetName() << ") Row["
			 << ichamp << "]  ** Field(" << ilength << "/"
			 << _nLines << ") value[" << champ << "]" << endl;
                if (ilength == 1)
		    val[indcrt++] = atof(champ);
                theVal[ncrt] = atof(champ);
#ifndef OLDCLUB
                delete[] champ;
#endif
                ncrt += natt;
	    }
	}

        // Pour le retour
        if (lstOut != NULL)
	{
            if (_blog)
	    {
                cout << endl << " *** Returned Value For the Matrix";
                matOut->Print();
                cout << " *** End Of Returned Value For the Matrix" << endl;
	    }
            lstOut->Add(matOut);
	}
	else
	{
	    if(matOut!=NULL)
		delete matOut;
	    matOut=NULL;
	}
        // Clean
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
	if(matOut!=NULL)
	    delete matOut;
	matOut=NULL;
        throw(ue);
    }
#ifdef OLDCLUB
    catch (ClubErreurs ce)
    {
        ce.correction();
        if( lstOut != NULL )
	{
            if ( _blog )
	    {
                cout << " *** catch ClubErreurs" << endl;
                matOut->Print();
                cout << " *** End Of catch ClubErreurs" << endl;
	    }
            lstOut->Add(matOut);
	}
        return kFALSE;
        //throw ("TOutputFileRow::searchOutput Error");
    }
#else
    catch (TamponTexteError&)
    {
        if (lstOut != NULL)
	{
            if (_blog)
	    {
                cout << " *** catch ClubErreurs" << endl;
                matOut->Print();
                cout << " *** End Of catch ClubErreurs" << endl;
	    }
            lstOut->Add(matOut);
	}
        return kFALSE;
    }
#endif
    if (_blog)
	cout << "*** End Of TOutputFileRow::searchOutput indcrt[" << indcrt
	     << "]" << endl << "************************" << endl;
    return kTRUE;
}


/////////////////////
//// searchOutput
/////////////////////
Bool_t TOutputFileRow::searchOutput(vector<URANIE::DataServer::UEntry*> *out)
{
    // Get the working file
    if (_blog)
    {
        cout << endl
	     << "************************\n*** TOutputFileRow::searchOutput " << endl;
        cout << " *** PWD[" << gSystem->pwd() << "]" << endl;
        cout << " *** _sFileName[" << _sFileName << "]" << endl;
    }

    // In case of the output file is not presented
    // put the default value
    //Number of attributes
    Int_t natt = getAttributesNumber();
    if (_blog)
	cout << " *** natt[" << natt << "]" << endl;

    // Create a default map containing just the default of Attribute of 1.23456789
    // // Default Value
    for (Int_t i = 0; i < natt; i++)
    {
        TString def=_attributes[i]->getDefault();
        out->at(i)->fillEntry( ((def=="" && out->at(i)->tag==URANIE::DataServer::UEntry::Double ) ? "1.23456789" : def) );
    }

    if (_blog)
    	cout << endl << " *** Default Value  "<< "" << " *** End Of Default Value" << endl;

    try
    {
        TamponTexte fnomFichier(_sFileName, 0);
        fnomFichier.initSeparateurs(_cSeparators);
        fnomFichier.initCarCommentaires(_cComments);
#ifndef OLDCLUB
	fnomFichier.initFieldSeparateur(_cFieldSeparator);
#endif
        // Elimination des lignes de commentaires
        fnomFichier.elimineCommentaires();
        // Elimination des lignes vides
        fnomFichier.elimineBlancs();
        // Test si le fichier ainsi obtenu est vide
        if (fnomFichier.estVide())
	{
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 Form(
							     "TOutputFileRow::searchOutput method\nthe file is empty [%s] is empty",
							     _sFileName.Data()));
	}

        // Set the number of lines in this file
        _nLines = fnomFichier.nombreLignes();
        _nColumns = fnomFichier.nombreChampsLigne(1);

        // Case of several lines
        if (_blog)
	{
            cout << " *** _nLines[" << _nLines << "]" << endl;
            cout << " *** _nColumns[" << _nColumns << "]" << endl;
	}

        for (Int_t iout = 1; iout <= natt; iout++)
	{
            list<URANIE::DataServer::TAttributeFileKey *> KeyList = _attributes[iout - 1]->getKeyList();
            list<URANIE::DataServer::TAttributeFileKey *>::iterator it = KeyList.begin();
            TEventList * eventfield = (*it)->_sEventList;
            Int_t ichamp = (Int_t) eventfield->GetEntry(0);

            if (ichamp > _nColumns)
		throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							   Form(
							       "TOutputFileRow::searchOutput method\nthe output file [%s] contains only [%d] columns - not [%d] for attribute [%s]",
							       _sFileName.Data(), _nColumns, ichamp,
							       _attributes[iout - 1]->GetName()));

	    for (Int_t ilength = 1; ilength <= ((_attributes[iout - 1]->getDataTypeStr()=="V") ? _nLines : 1); ilength++)
	    {
		const char* champ = fnomFichier.champLigne(ichamp, ilength);
		TString strchamp(champ);
#ifndef OLDCLUB
		delete[] champ;
#endif
		if(  _attributes[iout - 1]->getDataTypeStr()!="V" || (_attributes[iout - 1]->getDataTypeStr()=="V" && strchamp!="") )
		    out->at(iout-1)->pushEntry( strchamp );
	    }
	}


    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw(ue);
    }
#ifndef OLDCLUB 
    catch (TamponTexteError&)
    {
	if (_blog)
	    cout << " *** End Of catch ClubErreurs" << endl;
        return kFALSE;
    }
#else
   catch (ClubErreurs ce)
    {
        if (_blog)
            cout << " *** End Of catch ClubErreurs" << endl;

        return kFALSE;
    }

#endif

    if (_blog)
	cout << "*** End Of TOutputFileRow::searchOutput ************************" << endl;
    return kTRUE;
}


void TOutputFileRow::printLog(Option_t *option)
{
    cout << "*** TOutputFileRow::printLog[" << option << "]" << endl;
    cout << " *** _sFileName[" << _sFileName << "]" << endl;
    cout << " *** _nColumns[" << _nColumns << "]" << endl;
    cout << " *** _nLines[" << _nLines << "]" << endl;
    TOutputFile::printLog(option);
    cout << "*** End of TOutputFileRow::printLog[" << option << "]" << endl;
}

} // Fin du namespace Launcher
} // Fin du namespace URANIE

