//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
//////////////////////////////////////////////////////////////
/*!
 \file   TSGECluster.h
 \author Cedric Forgues <uranietm@cea.fr>
 \date   tue 28  16:10 CEST 2009
 
 \brief Interface of class URANIE::Launcher::TSGECluster
 
 */
/**
 
 * @class URANIE::Launcher::TSGECluster
 *
 * @brief Description of the class TSGECluster.
 *
 Uranie looks for the environnement variable \b  PE_HOSTFILE to activate the SGE batch.
 This variable defines an ASCII file 
 \code
 PE_HOSTFILE=/opt/sge/default/spool/mars028/active_jobs/322763.1/pe_hostfile
 \endcode
 which contains the list of nodes with the format:
 \code
 ::::::::::::::
 more /opt/sge/default/spool/mars028/active_jobs/322763.1/pe_hostfile
 ::::::::::::::
 mars028.cluster 8 express_par@mars028.cluster <NULL>
 mars034.cluster 8 express_par@mars034.cluster <NULL>
 mars029.cluster 2 express_par@mars029.cluster <NULL>
 ::::::::::::::
 \endcode
 where we request 18 slots where we have 8 slots of the node \b mars028.cluster, 8 slots of the node \b mars034.cluster and 2 slots of the node \b mars029.cluster.

 Then, the internal list of nodes is as follow:
 \code
 mars028.cluster mars028.cluster mars028.cluster mars028.cluster mars028.cluster mars028.cluster mars028.cluster mars028.cluster mars034.cluster mars034.cluster mars034.cluster mars034.cluster mars034.cluster mars034.cluster mars034.cluster mars034.cluster mars029.cluster mars029.cluster
 \endcode
 and we create ASCII files which contains an only node for each nodes:
 \code
 ::::::::::::::
 more /tmp/322763.1.express_par/mars028.cluster.txt
 ::::::::::::::
 mars028.cluster
 ::::::::::::::
 more /tmp/322763.1.express_par/mars029.cluster.txt
 ::::::::::::::
 mars029.cluster
 ::::::::::::::
 more /tmp/322763.1.express_par/mars034.cluster.txt
 ::::::::::::::
 mars034.cluster
 ::::::::::::::
 \endcode
 and use the \it mpirun options \b machinefile and \b np
 \code
 mpirun  -machinefile /tmp/322763.1.express_par/mars028.cluster.txt -np 1  leveleLine
 \endcode

 */
#ifndef TSGECLUSTER_H
#define TSGECLUSTER_H

#include <iostream>
#include <sstream>
#include <fstream>

using std::cout;
using std::endl;
using std::istringstream;
using std::string;

//ROOT
#include "Rtypes.h"

// Uranie
#include "TCluster.h"

namespace URANIE
{
namespace Launcher
{
class TSGECluster: public TCluster
{
private:
    static TString sRawListOfHosts;
    static const char *rawNodeList();
public:
    static bool eligible();
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    TSGECluster();
    //! Default destructor
    virtual ~TSGECluster();
    //@}

    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Constructs and returns the list of nodes
    virtual void init();
    virtual void getCommandPrefix(char *, char *);
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    void setLog()
    {
        _blog = kTRUE;
    }
    void unsetLog()
    {
        _blog = kFALSE;
    }
    void changeLog()
    {
        _blog = _blog ? kFALSE : kTRUE;
    }
    Bool_t getLog()
    {
        return _blog;
    }
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Launcher::TSGECluster, ID_LAUNCHER)
    //Classe de
};

} // Fin du namespace Launcher
} // Fin du namespace URANIE
#endif
