//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
// TSlurmCluster lun aoû 4 14:19:19 CEST 2008
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TSlurmCluster.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mon august 4 14:19:19 CEST 2008
 
 \brief Implementation of the class URANIE::Launcher::TSlurmCluster
 
 */

#include "TSlurmCluster.h"

#include "TSystem.h"
#include <vector>
#include <TRegexp.h>
#include "TDataServer.h"

ClassImp (URANIE::Launcher::TSlurmCluster)

namespace URANIE
{
namespace Launcher
{

// class method
bool TSlurmCluster::eligible()
{
    const char *tmp = gSystem->Getenv("SLURM_NODELIST");
    return (tmp != NULL && tmp[0]);
}

const char *TSlurmCluster::nodeList()
{
    return gSystem->Getenv("SLURM_NODELIST");
}

// Constructor
TSlurmCluster::TSlurmCluster() :
    TCluster("SLURM CLUSTER", TSlurmCluster::nodeList(), 1)
{
    if (_blog)
        cout << " ** Constructor TSlurmCluster::TSlurmCluster [" << GetName()
             << "] Title[" << GetTitle() << "]" << endl;
    init();

#ifdef WITH_LIBSSH
    _edistrib=EDistrib::kSLURM;
#endif

}


#ifdef WITH_LIBSSH
//Constructor Remote
TSlurmCluster::TSlurmCluster(URANIE::DataServer::TDataServer * tds, TCode * mycode) :
    TCluster(tds,mycode)
{
    _edistrib=EDistrib::kSLURM;
}

TSlurmCluster::TSlurmCluster(TCode * mycode) :
    TCluster(mycode)
{
    _edistrib=EDistrib::kSLURM;
}
#endif

TSlurmCluster::~TSlurmCluster()
{
    if (_blog)
        cout << " ** Destroy TSlurmCluster [" << GetName() << "] Title["
             << GetTitle() << "]" << endl;
}

///////////////////////////////////////////////////////////
//  Construct and return the list of node
///////////////////////////////////////////////////////////
/*!
  Constructs and returns the list of nodes
  SLURM node lists are built such as :
  myhost[10,13-15] for 4 nodes myhost10, myhost13, myhost14, myhost15
  myhost10 for one node myhost10

  The SLURM_TASKS_PER_NODE gives the number of tasks on each of the node
*/
void TSlurmCluster::init()
{

    TString tasks = gSystem->Getenv("SLURM_TASKS_PER_NODE");
    TString snodes = gSystem->Getenv("SLURM_NODELIST");

    tasks.ReplaceAll(",", " ");
    istringstream is((char*) tasks.Data());
    string task_number;
    std::vector<int> numbers;   //vector containing the number of tasks per node
    while (is >> task_number)
    {
        size_t pos1 = task_number.find('x');
        if (pos1 == string::npos) // no multiplier
            numbers.push_back(atoi(task_number.c_str()));
        else
        {
            size_t pos2 = task_number.find(')');
            string smultiplier = task_number.substr(pos1 + 1, pos2 - pos1 - 1);
            string snumber = task_number.substr(0, pos1 - 1);
            Int_t number = atoi(snumber.c_str());
            Int_t multiplier = atoi(smultiplier.c_str());
            for (int i = 0; i < multiplier; i++)
                numbers.push_back(number);
        }
    }

    if (_blog)
        cout << " ** TCluster::getListOfNodes Node[" << snodes << "]" << endl;
    if (snodes.CountChar('[') == 0) //list of nodes contains only one node
    {
        for (Int_t i = 0; i < numbers[0]; i++)
            addNodeName(snodes);
    }
    else  //several nodes exist
    {
        UInt_t inumber = 0; //index for the number of tasks per node 
        Int_t nbegin = snodes.Index('[');
        Int_t nend = snodes.Index(']');
        TString ssnodes = snodes(nbegin + 1, nend - 1 - nbegin);
        TString hostname = snodes(0, nbegin);
        if (_blog)
            cout << " ** snode[" << snodes << "] ssnode[" << ssnodes << "]"
                 << endl;
        ssnodes.ReplaceAll(",", " ");
        istringstream inStream((char *) ssnodes.Data());
        string inName;
        size_t found;
        while (inStream >> inName)
        {
            if (_blog)
                cout << " ** inName[" << inName << "] -------- ";
            found = inName.find("-");
            if (found != string::npos)
            { // in irene[10-14,16], 10-14 has been found
                if (_blog)
                    cout << "found at: " << int(found);
                inName.replace(int(found), 1, " ");
                if (_blog)
                    cout << " **** inName[" << inName << "]" << endl;
                istringstream inStream2(inName);
                string sfirst, slast;
                inStream2 >> sfirst;
		Int_t nLengthFirst = sfirst.length();
                TString sFormatNum = Form("%%%d%dd",0,nLengthFirst);
                Int_t nfirst = atoi(sfirst.c_str());
                inStream2 >> slast;
                Int_t nlast = atoi(slast.c_str());
                if (_blog)
                    cout << " ** ****** First\"" << sfirst << "\" nfirst["
                         << nfirst << "] -- last\"" << slast << "\" nlast["
                         << nlast << "]" << endl;
                for (Int_t icrt = nfirst; icrt <= nlast; icrt++)
                {
                    for (Int_t i = 0; i < numbers[inumber]; i++)
                    {
                        TString tmp(hostname);
                        //tmp += icrt;
			tmp += Form(sFormatNum,icrt);
                        addNodeName(tmp);
                    }
                    inumber++;
                }

            }
            else
            { // in irene[10-14,16], 16 has been found
                if (_blog)
                    std::cout << "not found" << endl;
                for (Int_t i = 0; i < numbers[inumber]; i++)
                {
                    TString tmp(hostname);
                    tmp.Append(inName.c_str());
                    addNodeName(tmp);
                }
                inumber++;
            }
        }
        if (inumber != numbers.size())
        {
            std::cerr
                << " TSlurmcluster : incoherence between SLURM_TASKS_PER_NODE "
                << endl;
            std::cerr << " and SLURM_NODELIST";
            exit(0);
        }
    }

    if (_blog)
        std::cout << " TCluster::getListOfNodes _slistOfNodes["
                  << TCluster::getListOfNodes() << "]" << endl;
    TCluster::getListOfNodes();
}

void TSlurmCluster::getCommandPrefix(char *mpiexe, char* procs)
{
    if (_blog)
    {
        cout << "MPIRUN FOR SLURM" << endl;
        cout << "procs = " << procs << endl;
    }

    Char_t npstring[10];
    // getProcsPerJob reads the list of hosts and uses separator to count 
    // available nodes
    sprintf(npstring, " -np %d ", getProcsPerJob(',', procs));

    strcpy(mpiexe, "mpirun ");
    strcat(mpiexe, npstring);
    strcat(mpiexe, " -host ");
    strcat(mpiexe, procs);
#ifdef WITH_MPIOLD
    strcat(mpiexe, " --bind-to-none");
#else
    strcat(mpiexe, " --bind-to none");
#endif

    if (_blog)
    {
        cout << "TSlurmCluster mpiexe: " << mpiexe << endl;
    }
}

/////////////////////////////////////
/// SSH
////////////////////////////////////
#ifdef WITH_LIBSSH
void TSlurmCluster::setCluster(TString cluster_name)
{
    _sClusterName = cluster_name;
    if (cluster_name == "marenostrum4")
    {
        _sClusterIP="mn2.bsc.es";
    }
    else if (cluster_name == "power9")
    {
        _sClusterIP="plogin2.bsc.es";
    }
    _eType="kSLURM";
}

void TSlurmCluster::selectHeader(TString SkeletonHeaderName,TString localHeaderFolderPath)
{

    if (!localHeaderFolderPath.EndsWith("/"))
        localHeaderFolderPath.Append("/");
    _slocalHeaderFolderPath=localHeaderFolderPath;

    if (getClusterName().Length() > 0)
    {
        cout << "Selecting a predefined header skeleton." << endl;
        if (SkeletonHeaderName.EqualTo("") || SkeletonHeaderName.Length() <= 0 || SkeletonHeaderName.EqualTo("./")  )
        {
            if (getClusterName().EqualTo("marenostrum4"))
                SkeletonHeaderName = TString("mn4_slurm_skeleton.in");
            else if (getClusterName().EqualTo("power9"))
                SkeletonHeaderName = TString("p9_slurm_skeleton.in");
            else
            {
                cout << "Invalid cluster or not implemented" << endl ;
                return;
            }

        }
        setSkeleton(SkeletonHeaderName);
    }
    else{
        cout << "Error, first one must set a clustername" << endl ;
        return;
    }
}

void TSlurmCluster::generateHeaderSplit(Option_t *option)
{
    TString soption =option;
    soption.ToLower();
    std::vector<TRegexp> regular;
    TRegexp regularStart("start=[0-9]+");
    TRegexp regularEnd("end=[0-9]+");
    TString outStart = soption(regularStart);
    TString outEnd = soption(regularEnd);
    TString start = outStart.Remove(0, 6);
    TString end = outEnd.Remove(0, 4);

    _sSplitName=start+"_to_"+end+"_"+getOutputHeaderName();

    //std::ifstream infile();
    substituteDirectives(gSystem->ConcatFileName(_slocalHeaderFolderPath,_sSkeletonHeader),gSystem->ConcatFileName(_sWorkingDirectory,_sSplitName),  _sListClusterDirectives);

    std::ofstream os(gSystem->ConcatFileName(_sWorkingDirectory,_sSplitName), ios::app );

    TString workingDirectory;

    if  (_sRemoteWorkingDirectory.BeginsWith('.'))
        os << "workingDirectory=\"\"$HOME\"/" <<  _sRemoteWorkingDirectory.Data()  << "\"" << endl;
    else
        os << "workingDirectory=\"" <<  _sRemoteWorkingDirectory.Data() << "\"" <<  endl;
    //os << "set -x \n" << endl;
    if (_sUserBinary)
    {
        os << "#Configuration of UserBinary" <<endl;
        os << "export PATH=$PATH:$workingDirectory/bin" <<endl;
        os << "export LIBRARY_PATH=$LIBRARY_PATH:$workingDirectory/lib" <<endl;
        os << "export CPATH=$CPATH:$workingDirectory/include" <<endl;
        os << "export LD_LIBRARY_PATH=$LD_LIBRARY_PATH:$workingDirectory/lib" <<endl;
        os << "#END Configuration" <<endl;
    }
    os << "sdir=\"" << "nothing" << "\"" <<  endl;
    os << "fixedRetries=" << _sRetries << endl;
    os << "ipat=" << start << endl;
    os << "ninput=" << end << endl;
    os << "command=\""<<_code->getCommand() << "\"" << endl;
    os << "ncores=" << _nAllocatedCores  << endl;
    os << "SECONDS=0" << endl;
    //os << "echo $workingDirectory" <<  endl;
    os << "for (( ipat=$ipat; ipat<$ninput; ipat++ ))" << endl;
    os << "do" << endl;
        os << "     sdir=\"\"$workingDirectory\""<< "/UranieLauncher_" << "$(($ipat+1))" << "\"" <<  endl;
        os << "     cd "  << "$sdir" << endl;//os << "     mpirun -E '--exclusive' -n " << _nAllocatedCores << " "<< "$(command)" << " &" << endl;
        //os << "     mpirun  -n " << _nAllocatedCores << " "<< "$command" << " &" << endl;
        os << "     fileExist=$(ls -l | grep '\\.in\\|\\.dat' | wc -l)" << endl;
        os << "     retries=$fixedRetries" << endl;
        os << "     while [ $fileExist -eq 0 ]" << endl;
        os << "     do" << endl;
        os << "         echo \"Waiting 1m for recv an input file, retries left:\" ";
            os << "     \"$retries\"" << endl;
            os << "     ((retries--))" << endl;
            os << "     sleep 1m"  << endl;
            os << "     cd "  << "$sdir" << endl;
            os << "     fileExist=$(ls -l | grep '*.in \\| *.dat' | wc -l)" << endl;
            os << "     if [ $retries -eq 0 ]; then" << endl;
            os << "         break; fi" << endl;
        os << "     done " << endl;
        os << "     eval mpirun -n $ncores" <<  " $command" << " &" << endl;
         //os << "      srun " << "$command" << " &" << endl;
        os << "      sleep 0.15s"  << endl;
    os << "done" << endl;
    os << "wait" << endl;
    os << "echo \"Completed in $SECONDS seconds\" " << endl;
    os.close();
}
TString TSlurmCluster::getJobStatusCommand(TString JobIDs)
{
        return  "sacct -j " + JobIDs + " -o state,jobid%15  -Xp | sed -n \'2,$p\' "; // GET COMMAND FROM SLURM ETC
}
TString TSlurmCluster::checkJobStatus(TString JobId, TString Command)
{
    execRemoteCommand(Command);
    if ( _sSshOutput.Contains("COMPLETED|"+JobId) )
    {
        return "COMPLETED";
    } else if ( _sSshOutput.Contains("FAILED|"+JobId) || _sSshOutput.Contains("CANCELLED|"+JobId) || _sSshOutput.Contains("NODE_FAIL|"+JobId) ||_sSshOutput.Contains("PREEMPTED|"+JobId) ||_sSshOutput.Contains("SUSPENDED|"+JobId) ||_sSshOutput.Contains("TIMEOUT|"+JobId) )
        return "FAILED";
    else
    {
        return "RUNNING";
    }
}
void TSlurmCluster::generateHeader(Option_t *option)
{

    _sConnected = kTRUE;
    generateHeaderSplit(option);
    if (openSFTPChannel() != SSH_OK)
    {
        cout << "SFTP channel couldn't be open" << endl;
        exit(-1);
    }
    if (sendHeader()  != SSH_OK)
    {
        cout << "Header couldn't be send" << endl;
        exit(-1);
    }
    if (closeSFTPChannel() != SSH_OK)
    {
        cout << "SFTP couldn't be closed" << endl;
        exit(-1);
    }
}

#endif

///////////////////////////////////////////////////////////
//  Log
///////////////////////////////////////////////////////////
void TSlurmCluster::printLog(Option_t *option)
{
    cout << "TSlurmCluster::printLog[" << option << "]" << endl;
    TCluster::printLog(option);
}

} // Fin du namespace Launcher
} // Fin du namespace URANIE

