//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
// TCalibration
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TCalibration.h
 \author Jean-Baptiste Blanchard
 \date   2020/07/17

 \brief Interface of class URANIE::Calibration::TCalibration

 */
/**

 * @class URANIE::Calibration::TCalibration
 *
 * @brief Description of the class TCalibration.
 *

 *
 */
#ifndef TCALIBRATION_H
#define TCALIBRATION_H

// Root
#include "TMatrixD.h"

#include <iostream>
#include <sstream>
using std::cout;
using std::endl;

// Uranie
#include "TDataServer.h"
#include "TDSNtupleD.h"
#include "TRun.h"
#include "UExceptions.h"
#include "Calibration.h"
#include "TDistanceFunction.h"
#include "TCode.h"

namespace URANIE {
namespace Calibration {
class TCalibration: public TNamed {

   
    //---------------------------------------------
    /** @name attributes
     *
     **/
    //@{
public:
    
    enum ELauncher
    {
        kCode, kFunction, kRun, kUnknown
    };

    URANIE::DataServer::TDataServer *_tdsPar;         ///< TDS containing parameters properties (parameters that should be calibrated)
    URANIE::DataServer::TDataServer *_tdsObs;         ///< TDS containing observations used for calibration
    URANIE::DataServer::TDataServer *_tdsEval;        ///< TDS containing a priori / a posteriori evaluations
    
    ELauncher _nLauncher;                             ///< The type of launcher
    TString _sFunctionName;                           ///< The Name of the evaluatuor
    TString _sEI;                                     ///< The Name of input
    TString _sEO;                                     ///< The Name of output
    URANIE::Launcher::TCode *_code;                   ///< The tcode
    URANIE::Relauncher::TRun *_run;                   ///< Pointer to the runner to be used
    void (*_pFunction)(double *, double *);           ///< Function pointer
    vector<URANIE::DataServer::TStochasticAttribute*>  _vatt; ///< internal vector of stochastic attribute for some methods

protected:

    URANIE::Calibration::TDistanceFunction *_dFunc;   ///< Pointer to chosen distance function
    URANIE::DataServer::TDSNtupleD         *_evTuple; ///< Pointer to the eval ntuple

    TList *_listOfParameters;                         ///< List of the parameters to be calibrated

    TCanvas *_canvas;                                 ///< Canvas object to deal with
    TObjArray *_drawingGarbageCollector;              ///< Garbage collector for prints.
    
    int _nSam;                                        ///< The number of sample in a posteriori distributions
    int _nObs;                                        ///< The number of observations in the reference database
    int _nIterMax;                                    ///< The maximum number of iteration allowed (meaning total number of code estimation is _nIterMax * _nObs)
    int _nPar;                                        ///< Dimension of the parameters
    int _nVar;                                        ///< Dimension of the output and references to be compared with
    
    int   _nSeed;                                     ///< The seed of the random generator
    TString _sMethodName;                             ///< The method name
    TString _referenceName;                           ///< The reference name
    TString _outputName;                              ///< The output name
    vector<string> _vrefName;                         ///< The reference names
    vector<string> _voutName;                         ///< The output names
    TString _weightName;                              ///< The weight name
    TMatrixD _mObsCovMat;                             ///< Observation Covariance matrix
    
    bool   _buseMatrix;                               ///< Use matrix instead of vectors in the Distance Function
    bool   _bsaveAll;                                 ///< Whether all evaluations should be saved, not only a priori and a posteriori
    bool   _bdontKeepAgreement;                       ///< Remove the agreement attribute from the tdsPar object
    bool   _buseMode;                                 ///< Use Mode instead of Mean 
    Bool_t _blog;                                     ///< Boolean for edit the log
    //@}
    
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{

    /*! 
     * \brief Default constructor with the runner argument: it contains the assessor to be used
     * \param tds : the dataserver that contains no data but only one attribute per parameter to be calibrated
     * \param run : the runner that contains the assessor to be used on the reference data
     * \param ns : number of sample to be generated (depending on the method)
     * \param option : see parseOption below
     */
    TCalibration(URANIE::DataServer::TDataServer *tds, URANIE::Relauncher::TRun *run, int ns=1, Option_t * option = "");
    
    /*! 
     * \brief Default Calibration constructor with the function argument: it contains the assessor to be used
     * \param tds : the dataserver that contains no data but only one attribute per parameter to be calibrated
     * \param fcn : the pointer to a function
     * \param ns : number of sample to be generated (depending on the method)
     * \param varexpinput : the input variable for the function in the correct order (both input and parameters)
     * \param varexpoutput : the output of the function in the correct order 
     */
    TCalibration(URANIE::DataServer::TDataServer *tds, void (*fcn)(Double_t*,Double_t*), const char *varexpinput, const char *varexpoutput, int ns = 1, Option_t * option = "");

    /*! 
     * \brief Default Calibration constructor with the function argument: it contains the assessor to be used
     * \param tds : the dataserver that contains no data but only one attribute per parameter to be calibrated
     * \param fcn : the name of the function
     * \param ns : number of sample to be generated (depending on the method)
     * \param varexpinput : the input variable for the function in the correct order (both input and parameters)
     * \param varexpoutput : the output of the function in the correct order 
     */
    TCalibration(URANIE::DataServer::TDataServer *tds, const char *fcn, const char *varexpinput, const char *varexpoutput, int ns = 1, Option_t * option = "");

    /*! 
     * \brief Default Calibration constructor with the code argument: it contains the assessor to be used
     * \param tds : the dataserver that contains no data but only one attribute per parameter to be calibrated
     * \param code : the code object that will be runned
     * \param ns : number of sample to be generated (depending on the method)
     */
    TCalibration(URANIE::DataServer::TDataServer *tds, URANIE::Launcher::TCode *fcode, int ns = 1, const char *option = "");
    
    //! Default destructor
    virtual ~TCalibration();
    //@}

    //---------------------------------------------
    /** @name Setting and Getting attributs
     **/
    //@{
    Int_t getID()
    {
        return ID_CALIBRATION;
    }

    //! Set the seed of the random generator if one is used
    void setSeed(UInt_t nval)
    {
        _nSeed = nval;
    }

    //! Get the seed of the random generator if one is used
    UInt_t getSeed()
    {
        return _nSeed;
    }

    //! Get the method name
    const char * getMethodName()
    {
        return _sMethodName.Data();
    }

    /*!
     * \brief Set the observatiton covariance matrix
     * \param mat : a _nObs * _nObs covariance matrix
     */
    void setObservationCovarianceMatrix(TMatrixD &mat);

    //! Return the distance function
    URANIE::Calibration::TDistanceFunction *getDistanceFunction() { return _dFunc; }   
    
    /*!
     * \brief Get the number of parameters to be calibrated
     * \return The number of parameters ( _nPar )
     */
    Int_t getNPar() {
        return _nPar;
    }

    //! Get the tds in which evaluation will be performed
    URANIE::DataServer::TDataServer *getEvaluationTDS() {
        return _tdsEval;
    }
    
    //@}
    
    //---------------------------------------------
    /** @name Initialisation methods
     **/
    //@{

    /*!
     * \brief Set the distance function and some needed informations
     * \param funcName: name of the distance function chosen among the already implemented ones:
     \li "LS" for least square
     \li "weighterLS" for weighted least square
     \li "relativeLS" for relative least square
     \li "L1" for L1 norm 
     \li "Mahalanobis" for Mahalanobis distance function
     * \param tdsRef : the dataserver that contains all needed information detailled below 
     \li input: the input attributes used of the assessors in run
     \li reference: the reference attributes that will be used to compared the newly done estimations for every iterations
     \li weight: a weight to reweight every event, one-by-one (an experimental uncertainty for instance)
     */
    virtual void setDistanceAndReference(const char *funcName, URANIE::DataServer::TDataServer *tdsRef, const char *input, const char *reference, const char *weight="");

    /*!
     * \brief Set the distance function and some needed informations
     * \param distFunc: a pointer to a TDistanceFunction object (not usual, recommended only when dealing with a home-made distance unction compiled on the spot)
     * \param tdsRef : the dataserver that contains all needed information detailled below 
     \li input: the input attributes used of the assessors in run
     \li reference: the reference attributes that will be used to compared the newly done estimations for every iterations
     \li weight: a weight to reweight every event, one-by-one (an experimental uncertainty for instance)
     */
    virtual void setDistanceAndReference(URANIE::Calibration::TDistanceFunction *distFunc, URANIE::DataServer::TDataServer *tdsRef, const char *input, const char *reference, const char *weight="");
    //@}
    
    //---------------------------------------------
    /** @name launching method
     **/
    //@{

    /* 
     * \brief common method to estimate the parameters once all variables are set
     * \param option : see parseOption
     */
    void estimateParameters(Option_t *option="");

    /* 
     * \brief common method to estimate the residue on the observed dataserver given a specific set of parameter values
     * \param resName: the name of the parameters to be stored
     * \param n_theta: the number of parameters
     * \param theta: the values of the parameters
     */
    void estimateCustomResidues(string resName, int n_theta, double *theta);

    
    //!Check the consistency of the formation provided
    /* 
     * \warning Should only be used by user if one wants to provide its own DistanceFunction
     */
    void checkReference(URANIE::DataServer::TDataServer *tdsRef, const char *input, const char *output, const char *weight);    
    
    //@}

    //---------------------------------------------
    /** @name drawing method
     **/
    //@{

    /*!     
     * brief Draws the parameters (only skeleton, at this level it is empty)
     * \param sTitre: title of the plots if specified
     * \param variable: list of variables that should be plotted
     * \param select: selection pattern, if this is requested and possible
     * \param option: possible option for this plot
     */
    virtual void drawParameters(TString sTitre, const char *variable = "*", const char *select = "1>0", Option_t * option = "");// { UNUSED(sTitre, variable, select, option); cout<<"The method drawParameters is not relevant (and so not implemented) for the class "<<this->ClassName()<<endl; }

    /*!     
     * brief Draws the residuals meaning the difference between the reference and the estimation a priori and a posteriori
     * Depending whether the method use point-wise or distribution-wise the apriori and a posteriori estimatiton are :
     \li from a single point for TMinimisation: initialisaiton for a priori and result for a posteriori
     \li the mode and mean for TLinearBayesian as the hypothesis are gaussian prior and analytical posterior
     \li the mode from TMetropHasting for a prior and the mean_of the a posteriorie parameters distributions, or the MAP (mode a posteriori) estimated from an internal smoothing
     * \param sTitre: title of the plots if specified
     * \param variable: list of variables that should be plotted
     * \param select: selection pattern, if this is requested (on the number of events from the reference for instance or with respect to input value for diagnostic)
     * \param option: possible option for this plot
     \li "nonewcanvas": don't create a new canvas, but use the one active 
     \li "apriori": draw only a priori residuals (default is both)
     \li "aposteriori": draw only a posteriori residuals (default is both)
     \li "vertical" : if there are several parameters to be shown put them on top of another instead of side by side
    */
    void drawResidues(TString sTitre, const char *variable = "*", const char *select="1>0", Option_t * option = "");
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    void setLog()
    {
        _blog = kTRUE;
    }
    void unsetLog()
    {
        _blog = kFALSE;
    }
    void changeLog()
    {
        _blog = _blog ? kFALSE : kTRUE;
    }
    Bool_t getLog()
    {
        return _blog;
    }
    virtual void printLog(Option_t *option = "");
    //@}

    
protected:
    //---------------------------------------------
    /** @name Setting and Getting attributs
     **/
    //@{

    //! Set the Method name
    void setMethodName(const char *str)
    {
        _sMethodName = TString(str);
    }
    
    //@}

    //---------------------------------------------
    /** @name Internal methods
     **/
    //@{

    /*!
     * \brief internal method in which the estimation is performed for all inheriting classes 
     */
    virtual void computeParameters(Option_t *option="")=0;
    virtual void checktdsParContent()=0;
    
    //!Create a canvas if needed
    void checkCanvasCreation(bool newcan);

    //!Initialise some common inputs
    void initInputs();

    //!Initialise some common inputs
    void initResults(vector<string> *ParsedLines);

    //!Compute the a posteriori residual for many-solutions method
    void computeAPosterioriForDistribution();

    //@}
    //---------------------------------------------
    /** @name Parse the option
     **/
    //@{
    //! Read the possible options
    /*!
      Possible options are (common to all methods)
     \li "savealleval": keep all estimation in the _tdsEval (Extremely dangerous as it will keep _nIter branches with _nObs estimations. 
     This option might slow down the process.
     \li "usematrix"  : option use to specify that computation is done with matrices instead of vector<vector<double> >.
     Shoud be used only if one is using an home-made TDistanceFunction
     */
    virtual void parseOption(Option_t *option = "");
    //@}

    ClassDef(URANIE::Calibration::TCalibration, ID_CALIBRATION)

};

} // Fin du namespace Calibration
} // Fin du namespace URANIE

#endif
