//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*! \file TMetropHasting.cxx
 *
 */

#include "TH1.h"
#include "TCanvas.h"

//ROOT
#include "TMath.h"
#include "TROOT.h"
#include "TRandom3.h"

//URANIE
#include "TMetropHasting.h"
#include "TStochAttribut.h"

ClassImp (URANIE::Calibration::TMetropHasting)

namespace URANIE
{
namespace Calibration
{

// -----------------------------------------------------------------------
//  Constructors & Initialisation
// -----------------------------------------------------------------------

TMetropHasting::TMetropHasting(URANIE::DataServer::TDataServer *tds, URANIE::Relauncher::TRun *run, int ns, Option_t *option) :
    TCalibration(tds, run, ns, option), _burnin(-1), _lag(-1), _nbDump(1000), _lowAccRange(-1.), _higAccRange(-1.), _mbGDVersion(false), _bcleaningAtt(false)
{
    if (_blog)
        cout << "constructor TMetropHasting by TDataServer[" << tds->GetName() << "]"
	     << endl;

    parseOption(option);
    
    checktdsParContent();
 
}

TMetropHasting::TMetropHasting(URANIE::DataServer::TDataServer *tds, void (*fcn)(Double_t*,Double_t*), const char *varexpinput, const char *varexpoutput, int ns, Option_t *option):
    TCalibration(tds, fcn, varexpinput, varexpoutput, ns, option), _burnin(-1), _lag(-1), _nbDump(1000), _lowAccRange(-1.), _higAccRange(-1.), _mbGDVersion(false), _bcleaningAtt(false)
{

    parseOption(option);

    checktdsParContent();
}

TMetropHasting::TMetropHasting(URANIE::DataServer::TDataServer *tds, const char *fcn, const char *varexpinput, const char *varexpoutput, int ns, Option_t *option):
    TCalibration(tds, fcn, varexpinput, varexpoutput, ns, option), _burnin(-1), _lag(-1),_nbDump(1000), _lowAccRange(-1.), _higAccRange(-1.), _mbGDVersion(false), _bcleaningAtt(false)
{

    parseOption(option);

    checktdsParContent();
}

TMetropHasting::TMetropHasting(URANIE::DataServer::TDataServer *tds, URANIE::Launcher::TCode *fcode, int ns, Option_t *option):
    TCalibration(tds, fcode, ns, option), _burnin(-1), _lag(-1),_nbDump(1000), _lowAccRange(-1.), _higAccRange(-1.), _mbGDVersion(false), _bcleaningAtt(false)
{
    TString opt = option;
    opt.ToLower();
    
    checktdsParContent();
}

void TMetropHasting::checktdsParContent()
{

    // Dans le cas d'un plan SRS
    setMethodName("MetropolisHasting");

    if(_blog)
        XYZ<<"In checktdsParContent"<<endl;

    // Get the number of attributes
    _nPar=_tdsPar->getNAttributes();
    TIter next(_listOfParameters);
    URANIE::DataServer::TAttribute *att=NULL;
    URANIE::DataServer::TStochasticAttribute *stoatt=NULL;
    try
    {
        while ((att = (URANIE::DataServer::TAttribute *) next()))
        {
            stoatt = dynamic_cast<URANIE::DataServer::TStochasticAttribute *>(att);
            if(stoatt==0)
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TMetropHasting::constructor\n\
 The provided dataserver must contains stochastic attributes. Attribute [%s] is not one of them", att->GetName()));

            _vatt.push_back( stoatt );
            if (_blog)
                cout << " ** Storing " << att->GetName() << "] in our attribute vector"<<endl;
        }
        
        if( _vatt.size()==0 )
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       "TMetropHasting: construction with tds not containing any TStochasticAttribute");

        // number of attributes different from number of stochastic attributes
	if( int(_vatt.size()) != _nPar )
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form("TMetropHasting: construction with _nPar [%d] different from _vatt content [%lu]", _nPar, _vatt.size()));
    }
    catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    
}

void TMetropHasting::setAcceptationRatioRange(double lower, double higher)
{
    try{
	if( lower >= higher )
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form("TMetropHasting::setAcceptationRatioRange. Lower bound [%g] should be lower than the higher one [%g]", lower, higher));
        if( lower > 1.0 )
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form("TMetropHasting::setAcceptationRatioRange. Lower bound [%g] is greater than 1.0. It should not be.", lower));
        if( lower < 0.0 )
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form("TMetropHasting::setAcceptationRatioRange. Lower bound [%g] is lower than 0.0. It should not be.", lower));
        if( higher > 1.0 )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form("TMetropHasting::setAcceptationRatioRange. Higher bound [%g] is greater than 1.0. It should not be.", higher));
        if( higher < 0.0 )
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form("TMetropHasting::setAcceptationRatioRange. Higher bound [%g] is higher than 0.0. It should not be.", higher));

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    _lowAccRange=lower;
    _higAccRange=higher;
}

void TMetropHasting::setInitialisation(int n, double *values, double *standDev)
{
    try{
	if( n != _nPar )
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form("TMetropHasting::setInitialisation. Arrays should contain %d elements. The number provided is %d",
							    _nPar,n));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    try{
        // initialise the parameters values
        for (int iatt=0;iatt<_nPar; iatt++)
        {
            _values.push_back(values[iatt]);
            if( standDev[iatt] < 0 )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TMetropHasting::setInitialisation. Standard deviation provided is negative [%g]. This cannot be !",
                                                                standDev[iatt]));
            _vstd.push_back(standDev[iatt]);
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    
}

void TMetropHasting::setInitialisation(vector<double> values, vector<double> standDev)
{
    try{
	if( (int(values.size()) != _nPar) || (int(standDev.size()) != _nPar) )
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form("TMetropHasting::setInitialisation. Vectors should contain [%d] elements. Value and standard deviation does only have [%lu] and [%lu]",
							    _nPar,values.size(),standDev.size()));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    setInitialisation(_nPar, &values[0], &standDev[0]);
}


// -----------------------------------------------------------------------
//  Destructor
// -----------------------------------------------------------------------

TMetropHasting::~TMetropHasting()
{
    if (_blog)
        cout << "************ Destructor URANIE::Calibration::TMetropHasting" << endl;

    if (_blog)
        cout << "************ End of Destructor URANIE::Calibration::TMetropHasting"<< endl;
}


// -----------------------------------------------------------------------
//  Setters
// -----------------------------------------------------------------------

void TMetropHasting::setDistanceAndReference(URANIE::Calibration::TDistanceFunction *distFunc, URANIE::DataServer::TDataServer *tdsRef, const char *input, const char *output, const char *weight)
{
    UNUSED(distFunc);
    TCalibration::setDistanceAndReference("weightedls", tdsRef, input, output, weight);
}

void TMetropHasting::setDistanceAndReference(const char *funcName, URANIE::DataServer::TDataServer *tdsRef, const char *input, const char *output, const char *weight)
{
    UNUSED(funcName);
    TCalibration::setDistanceAndReference("weightedls", tdsRef, input, output, weight);
}
   

// -----------------------------------------------------------------------
//  Compute the parameters
// -----------------------------------------------------------------------

void TMetropHasting::computeParameters(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    
    parseOption(option); 

    int iatt=0;

    if(_blog)
        XYZ<<"In computeParameters"<<endl;

    //Add attributes corresponding to the values of acceptation ratio per components
    for (iatt=0;iatt<_nPar; iatt++)
    {
	_tdsPar->addAttribute(new URANIE::DataServer::TAttribute(Form("acceptation_ratio_%s",_vatt.at(iatt)->GetName())));
	_tdsPar->getAttribute(Form("acceptation_ratio_%s",_vatt.at(iatt)->GetName()))->delShare();
    }

    if(!_bcleaningAtt)
    {
	//Add the log prior as an attibute
	_tdsPar->addAttribute(new URANIE::DataServer::TAttribute("log_prior"));
	_tdsPar->getAttribute("log_prior")->delShare();
	//Add the log likelihood as an attribute
	_tdsPar->addAttribute(new URANIE::DataServer::TAttribute("log_likelihood"));
	_tdsPar->getAttribute("log_likelihood")->delShare();
	//Add the unormalized log posterior as an attribute
	_tdsPar->addAttribute(new URANIE::DataServer::TAttribute("log_post_unormalized"));
	_tdsPar->getAttribute("log_post_unormalized")->delShare();
    }

    // delete the TDSNtupleD if it exists
    try
    {
        if (_tdsPar->hasTuple())
            throw URANIE::Exceptions::UInfoExceptions(__FILE__, __LINE__,
                                                      Form(
                                                          "TMetropHasting::computeParameters\n\
the TDS [%s] contains data we need to empty it ! ",
                                                          _tdsPar->GetName()));
    } catch (URANIE::Exceptions::UInfoExceptions& ue)
    {
        ue.printMessage();
	_tdsPar->deleteTuple();
    }

    TRandom3 *rand = new TRandom3();
    if(_nSeed != 0)
	rand->SetSeed(_nSeed);

    if( _values.size()==0 && _vstd.size()==0)
    {
	// initialise the parameters values
	for (iatt=0;iatt<_nPar; iatt++)
	{
	    _values.push_back(_vatt.at(iatt)->getInvCDF(rand->Rndm()));
	    _vstd.push_back( _vatt.at(iatt)->getTheoreticalStdDev());
	}
    }

    // Coucou
    double log_prior_value=0, log_lik_value=0, loglik_prop=0.;
    logPriorPdf(log_prior_value);
    _dFunc->eval(&_values[0],&log_lik_value);
    log_lik_value*=-0.5;
    
    if(_mbGDVersion)
    {
        _tdsPar->addAttribute("yes_or_no");

        // DEaling with agreement
        int baseSize=2*_nPar+6;
        _tdsPar->delAttribute("agreement");        
        _tdsPar->addAttribute("agreement");    

        // Create the database
        vector<double> dBase(baseSize,0.);
        for(iatt=0; iatt<_nPar; iatt++)
        {
            dBase[iatt+1]=_values.at(iatt);
            dBase[_nPar+1+iatt]=1.;
        }

        dBase[2*_nPar+1]=log_prior_value; 
        dBase[2*_nPar+2]=log_lik_value; 
        dBase[2*_nPar+3]=log_lik_value+log_prior_value;
        dBase[2*_nPar+4]=1.;

        double retval=0;
        _dFunc->eval(&_values[0], &retval);
        dBase[2*_nPar+5]=retval;

        _tdsPar->createTuple();
        _tdsPar->getTuple()->Fill(&dBase[0]);

        vector<int> accept(_nPar, 0);
        vector<int> reject(_nPar, 0);
        vector<int> LenBurn(_nPar, 0);

        // Initialise
        int Niter=1;    
        vector<bool> vOK(_nPar, false);
        vector<bool> vBatchDone(_nPar, false);
        bool PrintDone=false;
        int LenBatch=100;
        
        for (int ievt=0; ievt<_nSam; ievt++)
        {                                	
            for(iatt=0; iatt<_nPar; iatt++)
            {
                if ( ( Niter%_nbDump==0 ) && ! PrintDone  )
                {
                    cout<<Niter<<" events done"<<endl;
                    PrintDone=true;
                }

                _values[iatt]=rand->Gaus( dBase[iatt+1], _vstd[iatt] );

                if (_blog)                    
                    printf("From (%6.6f,%6.6f) => %6.6f\n",dBase[iatt+1],_vstd[iatt],_values.at(iatt));
                
                if ( ( _values[iatt] < _vatt.at(iatt)->getLowerBound() )  || ( _values[iatt] > _vatt.at(iatt)->getUpperBound() ) )
                {
                    dBase[0]=double(Niter);
                    dBase[2*_nPar+4]=0.;
                    reject[iatt]+=1;
                    dBase[_nPar+1+iatt]=float(accept[iatt])/float(reject[iatt]+accept[iatt]);
                    _values[iatt]=dBase[iatt+1];
                    _tdsPar->getTuple()->Fill(&dBase[0]);
                    Niter+=1;
                    continue;
                }                   

                log_prior_value=0.;
                // Recompute the proba with new values.
                logPriorPdf(log_prior_value);                        
                _dFunc->eval(&_values[0], &loglik_prop);
                dBase[2*_nPar+5]=loglik_prop;
                log_lik_value*=-0.5;

                // Recompute the ratio	    
                double value_prop=loglik_prop+log_prior_value;
                double log_ratio=value_prop-dBase[2*_nPar+3];
                double log_rand=log(rand->Rndm());

                //cout<<" Niter "<<Niter<<" Value  "<<_values[0]<<" ==> logratio "<<log_ratio<<"  value_prop "<<value_prop<<" - dbase  "<<dBase[2*_nPar+3]<<"  logrand  "<<log_rand<<endl;
                if ( log_ratio > log_rand )
                {
                    //cout<<" ACCEPTED "<<Niter<<endl;
                    dBase[0]=double(Niter);
                    accept[iatt]+=1;
                    dBase[2*_nPar+1]=log_prior_value;
                    dBase[2*_nPar+2]=loglik_prop;
                    dBase[2*_nPar+3]=value_prop;
                    dBase[2*_nPar+4]=1;
                    dBase[iatt+1]=_values[iatt];
                    dBase[_nPar+1+iatt]=float(accept[iatt])/float(reject[iatt]+accept[iatt]);
                    PrintDone=false;
                    // cout<<"Info are ";
                    // for(int info=0; info<2*_nPar+4; info++)
                    //     cout<<info<<":"<<dBase[info]<<", ";
                    // cout<<endl;                    
                    _tdsPar->getTuple()->Fill(&dBase[0]);
                }
                else
                {
                    //cout<<" REJECTED"<<endl;;
                    dBase[0]=double(Niter);
                    dBase[2*_nPar+4]=0;                   
                    reject[iatt]+=1;
                    _values[iatt]=dBase[iatt+1];
                    dBase[_nPar+1+iatt]=float(accept[iatt])/float(reject[iatt]+accept[iatt]);
                }

                Niter++;
                vOK[iatt]=true;
                vBatchDone[iatt]=false;
                if ( ievt%LenBatch==0  && _lowAccRange != -1. && _higAccRange !=-1)
                {
                    if( dBase[_nPar+1+iatt] < _lowAccRange )
                    {
                     	LenBurn[iatt]=double(Niter);
                        accept[iatt]=0;
                        reject[iatt]=0;
                        _vstd[iatt]=0.9*_vstd[iatt];
                        dBase[_nPar+1+iatt]=0.;
                    }
                    else if( dBase[_nPar+1+iatt] > _higAccRange )
                    {
                        LenBurn[iatt]=double(Niter);
                        accept[iatt]=0;
                        reject[iatt]=0;
                        _vstd[iatt]=1.1*_vstd[iatt];
                        dBase[_nPar+1+iatt]=0.;
                    }
                    vBatchDone[iatt]=true;
                }

                // if(reject[iatt]>=10)
                // 	return;
            }// _nPar
        }// _nS

        int burn=TMath::MaxElement(_nPar,&LenBurn[0]);
        cout<<"la longueur de chauffe pour un taux d'acceptation acceptable est de %d"<<burn<<endl;
    }
    else // JBB
    {
        // Dealing with agreement
        int baseSize=2*_nPar+5;
        _tdsPar->delAttribute("agreement");        
        _tdsPar->addAttribute("agreement");    
        
        // Create the database
        vector<double> dBase(baseSize,0.);
        for(iatt=0; iatt<_nPar; iatt++)
        {
            dBase[iatt+1]=_values.at(iatt);
            dBase[_nPar+1+iatt]=1.;
        }

        dBase[2*_nPar+1]=log_prior_value; 
        dBase[2*_nPar+2]=log_lik_value; 
        dBase[2*_nPar+3]=log_lik_value+log_prior_value;

        double retval=0;
        _dFunc->eval(&_values[0], &retval);
        dBase[2*_nPar+4]=retval;

        _tdsPar->createTuple();
        // XYZ<<"DBase content: ";
        // for(unsigned int ip=0; ip<2*_nPar+4; ip++)
        //     cout<<dBase.at(ip)<<", ";
        // cout<<endl;
        
        _tdsPar->getTuple()->Fill(&dBase[0]);

        vector<int> accept(_nPar, 0);
        vector<int> reject(_nPar, 0);

        // The proposal distribution is specified through a TRamdom3 object
    
        // Initialise
        int Niter=1;    
        vector<bool> vOK(_nPar, false);
        vector<bool> vBatchDone(_nPar, false);
        bool PrintDone=false;
        int LenBatch=100;

        while( Niter < _nSam )
        {                                	
            for(iatt=0; iatt<_nPar; iatt++)
            {

                if ( ( Niter%_nbDump==0 ) && ! PrintDone  )
                {
                    cout<<Niter<<" events done"<<endl;
                    PrintDone=true;
                }

                _values[iatt]=rand->Gaus( dBase[iatt+1], _vstd[iatt] );

                //printf("From (%6.6f,%6.6f) => %6.6f\n",dBase[iatt+1],_vstd[iatt],_values.at(iatt));
                if ( ( _values[iatt] < _vatt.at(iatt)->getLowerBound() )  || ( _values[iatt] > _vatt.at(iatt)->getUpperBound() ) )
                {
                    reject[iatt]+=1;
                    _values[iatt]=dBase[iatt+1];
                    dBase[_nPar+1+iatt]=float(accept[iatt])/float(reject[iatt]+accept[iatt]);
                    continue;
                }                   

                // Recompute the proba with new values.
                logPriorPdf(log_prior_value);                        
                _dFunc->eval(&_values[0], &loglik_prop);
                dBase[2*_nPar+4]=loglik_prop;
                loglik_prop*=-0.5;

                // Recompute the ratio	    
                double value_prop=loglik_prop+log_prior_value;
                double log_ratio=value_prop-dBase[2*_nPar+3];
                double log_rand=log(rand->Rndm());
                dBase[0]=float(Niter);
                //cout<<" Niter "<<Niter<<" Value  "<<_values[0]<<" ==> logratio "<<log_ratio<<"  value_prop "<<value_prop<<" - dbase  "<<dBase[2*_nPar+3]<<"  logrand  "<<log_rand<<endl;
                if ( log_ratio > log_rand )
                {
                    //cout<<" ACCEPTED "<<Niter<<endl;
                    accept[iatt]+=1;
                    dBase[2*_nPar+1]=log_prior_value;
                    dBase[2*_nPar+2]=loglik_prop;
                    dBase[2*_nPar+3]=value_prop;
                    dBase[iatt+1]=_values[iatt];
                    dBase[_nPar+1+iatt]=float(accept[iatt])/float(reject[iatt]+accept[iatt]);
                    Niter+=1;
                    PrintDone=false;

                    // XYZ<<"DBase content: ";
                    // for(unsigned int ip=0; ip<2*_nPar+4; ip++)
                    //     cout<<dBase.at(ip)<<", ";
                    // cout<<endl;
                    _tdsPar->getTuple()->Fill(&dBase[0]);
                }
                else
                {
                    //cout<<" REJECTED"<<endl;;
                    reject[iatt]+=1;
                    _values[iatt]=dBase[iatt+1];
                    dBase[_nPar+1+iatt]=float(accept[iatt])/float(reject[iatt]+accept[iatt]);
                }

                vOK[iatt]=true;
                vBatchDone[iatt]=false;
                //dBase[0]=float(Niter);
                if ( (Niter-2-iatt)%LenBatch==0 && ! vBatchDone[iatt] && _lowAccRange != -1 && _higAccRange != -1)
                {
                    if( dBase[_nPar+1+iatt] < _lowAccRange )
                    {
                        // accept[iatt]=0;
                        // reject[iatt]=0;
                        _vstd[iatt]=0.9*_vstd[iatt];
                        // dBase[_nPar+1+iatt]=0.;
                    }
                    else if( dBase[_nPar+1+iatt] > _higAccRange )
                    {
                        // accept[iatt]=0;
                        // reject[iatt]=0;
                        _vstd[iatt]=1.1*_vstd[iatt];
                        // dBase[_nPar+1+iatt]=0.;
                    }
                    vBatchDone[iatt]=true;
                }

                // if(reject[iatt]>=10)
                // 	return;
            }

        }
    }
    delete rand;
    
    computeAPosterioriForDistribution();
}

// -----------------------------------------------------------------------
//  Visualisation and diagnostic
// -----------------------------------------------------------------------

void TMetropHasting::drawTrace(TString sTitre, const char *variable, const char *select, Option_t *option)
{
    TString soption = option;
    TString cleanoption = option;
    soption.ToLower();

    if ( _blog )
        cout << " ** TMetropHasting::drawTrace for method[" << _sMethodName << "] option[" << option << "]" << endl;
    
    if (_tdsPar->getTuple()->GetEntries() == 0) {
        cout << "TMatropHAsting::drawTrace Error, the tdsPar is empty" << endl;
        return;
    }

    bool newcan=false;
    if (!soption.Contains("nonewcanvas"))
	newcan=true;
    else
        cleanoption.ReplaceAll("nonewcanvas","");

    checkCanvasCreation(newcan);

    string svariable=variable;
    vector<string> variables;
    if( svariable=="*" )
        for(unsigned int ivar=0; ivar<_vatt.size(); ivar++)
            variables.push_back( string(_vatt.at(ivar)->GetName() ) );
    else
        URANIE::DataServer::UEntry::StringParser(svariable, ":", variables);

    if (soption.Contains("vertical"))
    {
        _canvas->Divide( variables.size(), 1);
        cleanoption.ReplaceAll("vertical","");
    }
    else
        _canvas->Divide(1, variables.size());    

    URANIE::DataServer::TAttribute *att=NULL;
    TH1 *h=NULL;
    string sselect=select;
    string cut = getDefaultCut() + ((sselect!="" && sselect!="1>0") ? ((getDefaultCut()!="")?" && ":"")+sselect : "");

    for(unsigned int ivar=0; ivar<variables.size(); ivar++)
    {
        _canvas->cd(ivar+1);

        _tdsPar->getTuple()->Draw(Form("%s:%s",variables.at(ivar).c_str(), _tdsPar->getIteratorName()), cut.c_str(), (TString("l")+cleanoption).Data());

        att=(URANIE::DataServer::TAttribute*)_listOfParameters->FindObject(variables.at(ivar).c_str());

        h = (TH1*)gPad->GetPrimitive("htemp");
        h->SetName( (variables.at(ivar)+"_trace").c_str() );
        h->SetTitle( ((sTitre=="") ? (variables.at(ivar)+"_trace").c_str() : sTitre.Data() ) );
        h->GetXaxis()->SetTitle("Nb_{Calc}");
        h->GetYaxis()->SetTitle( ((att!=NULL)?att->getTitle():"") );
    }
}

void TMetropHasting::drawAcceptationRatio(TString sTitre, const char *variable, const char *select, Option_t *option)
{
    TString soption = option;
    TString cleanoption = option;
    soption.ToLower();

    if ( _blog )
        cout << " ** TMetropHasting::drawAcceptationRatio for method[" << _sMethodName << "] option[" << option << "]" << endl;
    
    if (_tdsPar->getTuple()->GetEntries() == 0) {
        cout << "TMatropHAsting::drawAcceptationRatio Error, the tdsPar is empty" << endl;
        return;
    }

    bool newcan=false;
    if (!soption.Contains("nonewcanvas"))
	newcan=true;
    else
        cleanoption.ReplaceAll("nonewcanvas","");

    checkCanvasCreation(newcan);

    string svariable=variable;
    vector<string> variables;
    if( svariable=="*" )
        for(unsigned int ivar=0; ivar<_vatt.size(); ivar++)
            variables.push_back( string(_vatt.at(ivar)->GetName() ) );
    else
        URANIE::DataServer::UEntry::StringParser(svariable, ":", variables);

    if (soption.Contains("vertical"))
    {
        _canvas->Divide( variables.size(), 1);
        cleanoption.ReplaceAll("vertical","");
    }
    else
        _canvas->Divide(1, variables.size());    

    URANIE::DataServer::TAttribute *att=NULL;
    TH1 *h=NULL;
    string sselect=select;
    string cut = getDefaultCut() + ((sselect!="" && sselect!="1>0") ? ((getDefaultCut()!="")?" && ":"")+sselect : "");
    for(unsigned int ivar=0; ivar<variables.size(); ivar++)
    {
        _canvas->cd(ivar+1);

        _tdsPar->getTuple()->Draw(Form("acceptation_ratio_%s:%s",variables.at(ivar).c_str(), _tdsPar->getIteratorName()), cut.c_str(), (TString("l")+cleanoption).Data());

        att=(URANIE::DataServer::TAttribute*)_listOfParameters->FindObject(variables.at(ivar).c_str());

        h = (TH1*)gPad->GetPrimitive("htemp");
        h->SetName( (variables.at(ivar)+"_trace").c_str() );
        h->SetTitle( ((sTitre=="") ? (variables.at(ivar)+"_trace").c_str() : sTitre.Data() ) );
        h->GetXaxis()->SetTitle("Nb_{Calc}");
        h->GetYaxis()->SetTitle( ((att!=NULL)?att->getTitle():"") );
    }
}

void TMetropHasting::drawParameters(TString sTitre, const char *variable, const char *select, Option_t *option)
{
    TString soption = option;
    TString cleanoption = option;
    soption.ToLower();

    if ( _blog )
        cout << " ** TMetropHasting::drawParameters for method[" << _sMethodName << "] option[" << option << "]" << endl;
    
    if (_tdsPar->getTuple()->GetEntries() == 0) {
        cout << "TMetropHasting::drawParameters Error, the tdsPar is empty" << endl;
        return;
    }

    bool newcan=false;
    if (!soption.Contains("nonewcanvas"))
	newcan=true;
    else
        cleanoption.ReplaceAll("nonewcanvas","");
    
    checkCanvasCreation(newcan);

    string svariable=variable;
    vector<string> variables;
    if( svariable=="*" )
        for(unsigned int ivar=0; ivar<_vatt.size(); ivar++)
            variables.push_back( string(_vatt.at(ivar)->GetName() ) );
    else
        URANIE::DataServer::UEntry::StringParser(svariable, ":", variables);

    if( !soption.Contains("same") )
    {
        if (soption.Contains("vertical"))
        {
            _canvas->Divide(1, variables.size());
            cleanoption.ReplaceAll("vertical","");
        }
        else
            _canvas->Divide(variables.size(), 1);
    }
    
    URANIE::DataServer::TAttribute *att=NULL;
    TH1 *h=NULL;
    string sselect=select;
    string cut = getDefaultCut() + ((sselect!="" && sselect!="1>0") ? ((getDefaultCut()!="")?" && ":"")+sselect : "");
    for(unsigned int ivar=0; ivar<variables.size(); ivar++)
    {
        _canvas->cd(ivar+1);

        att=(URANIE::DataServer::TAttribute*)_listOfParameters->FindObject(variables.at(ivar).c_str());

        _tdsPar->getTuple()->Draw( variables.at(ivar).c_str(), cut.c_str(), cleanoption.Data() );
        h = (TH1*)gPad->GetPrimitive("htemp");
        h->SetName( ("h"+variables.at(ivar)+"_aposteriori").c_str() );
        h->SetTitle( ((sTitre=="") ? ("h"+variables.at(ivar)+"_aposteriori").c_str() : sTitre.Data() ) );
        h->GetXaxis()->SetTitle(att->getTitle());
    }
}
	

void TMetropHasting::getAutoCorrelation(vector<int> l, vector<double> *out, int cut)
{
    out->clear();
    out->resize(l.size()*_vatt.size());
    double *data;
    int np=0, arrsize=0;
    for(unsigned int ip=0; ip<_vatt.size(); ip++)
    {
	np=_tdsPar->getTuple()->Draw(_vatt.at(ip)->GetName(), Form("(%s > %d)",_tdsPar->getIteratorName(),cut),"goff");
	data=_tdsPar->getTuple()->GetV1();
	double mY=TMath::Mean(np, data), sdY=TMath::RMS(np, data);

	for(unsigned int il=0; il<l.size(); il++)
	{
            arrsize=np-l.at(il);
	    int Nech=np-l.size();
	    double cross=0.;
	    for(int i=0; i<arrsize; i++)
		cross+= (data[i]-mY)/sdY * (data[i+l.at(il)]-mY)/sdY;

	    cross*=1./(Nech-1);
	    out->at(l.size()*ip+il)=cross;
	}
    }
}


// -----------------------------------------------------------------------
//  internal methodprintLog
// -----------------------------------------------------------------------

void TMetropHasting::logPriorPdf(double &ret)
{

    ret=0;
    for(unsigned int iatt=0; iatt<_vatt.size(); iatt++)
	ret+=log(_vatt.at(iatt)->getPDF(_values.at(iatt)));
    
}


// -----------------------------------------------------------------------
//  printLog
// -----------------------------------------------------------------------
void TMetropHasting::parseOption(Option_t *option)
{

    TCalibration::parseOption(option);
    
    TString soption=option;
    soption.ToLower();
    
    if( soption.Contains("clean") )
	_bcleaningAtt=true;

    if( soption.Contains("gd") )
	_mbGDVersion=true;


}

void TMetropHasting::printLog(Option_t *option)
{
    cout << "TMetropHasting::printLog[" << option << "] Name[" << GetName()
         << "] Title[" << GetTitle() << "] " << endl;
    cout << "** _nSam   [" << _nSam << "]" << endl;
    cout << "** _nPar   [" << _nPar << "]" << endl;
    cout << "** _sMethodName   [" << _sMethodName << "]" << endl;


    cout << "End Of TMetropHasting::printLog" << endl;
}



}
}

