//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
#ifndef TMetropHasting_H
#define TMetropHasting_H

// Uranie
#include "TCalibration.h"

namespace URANIE
{
namespace Calibration
{

class TMetropHasting: public URANIE::Calibration::TCalibration
{

public:
    
    // Attributes
    vector<double>                                       _values;         ///< Vector of values to be tested
    vector<double>                                       _vstd;           ///< Vector of standard deviation

    int                                                  _burnin;         ///< The warm-up or burn-in
    int                                                  _lag;            ///< The lag value
    int                                                  _nbDump;         ///< Frequency to which the algo dump a line
    double                                               _lowAccRange;    ///< loweracceptation ratio bound to increase _vstd
    double                                               _higAccRange;    ///< higheracceptation ratio bound to decrease _vstd
    bool                                                 _mbGDVersion;    ///< Use the Guillaume Damblin
    bool                                                 _bcleaningAtt;   ///< Do not store the underlying att

    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{

    /*!
      Constructor usual where nS is set to be the number of sample produced by the algorithm.
      \param tds: the pointer of the TDataServer. It must contains attribuutes that inherit from TStochasticAttribute
      \param run: the runner that would be used to produce evaluations 
      \param nS:  size of the sample to be generated [100 by default]
    */    
    TMetropHasting(URANIE::DataServer::TDataServer *tds, URANIE::Relauncher::TRun *run, int nS=100, Option_t *option = "lhs");

    /*! 
     * \brief Default Calibration constructor with the function argument: it contains the assessor to be used
     * \param tds : the dataserver that contains no data but only one attribute per parameter to be calibrated
     * \param fcn : the pointer to a function
     * \param ns : number of sample to be generated (depending on the method)
     * \param varexpinput : the input variable for the function in the correct order (both input and parameters)
     * \param varexpoutput : the output of the function in the correct order 
     */

    TMetropHasting(URANIE::DataServer::TDataServer *tds, void (*fcn)(Double_t*,Double_t*), const char *varexpinput, const char *varexpoutput, int ns = 100, Option_t *option = "");

    /*! 
     * \brief Default Calibration constructor with the function argument: it contains the assessor to be used
     * \param tds : the dataserver that contains no data but only one attribute per parameter to be calibrated
     * \param fcn : the name of the function
     * \param ns : number of sample to be generated (depending on the method)
     * \param varexpinput : the input variable for the function in the correct order (both input and parameters)
     * \param varexpoutput : the output of the function in the correct order 
     */
    TMetropHasting(URANIE::DataServer::TDataServer *tds, const char *fcn, const char *varexpinput, const char *varexpoutput, int ns = 100, Option_t *option = "");

    /*! 
     * \brief Default Calibration constructor with the code argument: it contains the assessor to be used
     * \param tds : the dataserver that contains no data but only one attribute per parameter to be calibrated
     * \param code : the code object that will be runned
     * \param ns : number of sample to be generated (depending on the method)
     */
    TMetropHasting(URANIE::DataServer::TDataServer *tds, URANIE::Launcher::TCode *fcode, int ns = 100, Option_t *option = "");
    
    //! Default destructor
    virtual ~TMetropHasting();
    //@}

    //---------------------------------------------
    /** @name evaluate the events
     **/
    //@{
    
    /*! 
     * \brief Generate the sample 
     * \param option: possble options are
     * \li "clean": remove internal branche from the tdsPar object at the end
     * \li "gd": another implementation of this algorithm
     */
    void computeParameters(Option_t *option = "");

    /*!
     * Check the content of the input dataserver
     */ 
    void checktdsParContent();
    //@}


    //---------------------------------------------
    /** @name Set reference information and initialisation methods
     **/
    //@{

    void setBurnin( int burn ){ _burnin = burn; }
    void setLag( int lag ){ _lag = lag; }
    void setNbDump( int nbDump ){ _nbDump = nbDump; }

    void setAcceptationRatioRange( double lower, double higher);

    string getDefaultCut(){
        string cut = "";
        if( _burnin != -1 )
            cut += string(Form("%s > %d",_tdsPar->getIteratorName(), _burnin));
        if( _lag != -1 )
            cut += ((cut=="")? "": " && " ) + string(Form("%s %% %d",_tdsPar->getIteratorName(), _lag));

        if(cut!="")
            cout<<"TMetropHasting drawing method apply default cut ["<<cut<<"]"<<endl;
        return cut;
    }

    void clearDefaultCut(){ setBurnin(-1); setLag(-1); }

    /*!     
     * \brief Initialise the parameters 
     * \param n: number of parameters
     * \param values: array of values (size must equal to n)
     * \param standDev: array of standard deviation (size must equal to n)
     */
    void setInitialisation(int n, double *values, double *standDev);

    /*!     
     * \brief Initialise the parameters 
     * \param values: vector of values (size must equal to n)
     * \param standDev: vector of standard deviation (size must equal to n)
     */
    void setInitialisation(vector<double> values, vector<double> standDev);

    /*!
     * \brief Set the distance function and some needed informations
     * \param funcName: name of the distance function chosen among the already implemented ones:
     \li "LS" for least square
     \li "weighterLS" for weighted least square
     \li "relativeLS" for relative least square
     \li "L1" for L1 norm 
     \li "Mahalanobis" for Mahalanobis distance function
     * \param tdsRef : the dataserver that contains all needed information detailled below 
     \li input: the input attributes used of the assessors in run
     \li reference: the reference attributes that will be used to compared the newly done estimations for every iterations
     \li weight: a weight to reweight every event, one-by-one (an experimental uncertainty for instance)
    */
    void setDistanceAndReference(const char *funcName, URANIE::DataServer::TDataServer *tdsRef, const char *input, const char *output, const char *weight="");

    /*!
     * \brief Set the distance function and some needed informations
     * \param distFunc: a pointer to a TDistanceFunction object (not usual, recommended only when dealing with a home-made distance unction compiled on the spot)
     * \param tdsRef : the dataserver that contains all needed information detailled below 
     \li input: the input attributes used of the assessors in run
     \li reference: the reference attributes that will be used to compared the newly done estimations for every iterations
     \li weight: a weight to reweight every event, one-by-one (an experimental uncertainty for instance)
    */
    void setDistanceAndReference(URANIE::Calibration::TDistanceFunction *distFunc, URANIE::DataServer::TDataServer *tdsRef, const char *input, const char *output, const char *weight="");
    
    //@}
    
    //---------------------------------------------
    /** @name Visualisation and diagnostic
     **/
    //@{
     
    /*!     
     * \brief Draws the evolution of parameters as a function of the iterator
     * \param sTitre: title of the plot if specified
     * \param variable: list of variables that should be plotted
     * \param select: select that can be applied on the parameters or the iterator (burnin, selection for autocorrelation)
     * \param option: possible option for this plot
     \li "nonewcanvas": don't create a new canvas, but use the one active 
    */
    void drawTrace(TString sTitre, const char *variable="*", const char *select="1>0", Option_t *option="");

    /*!     
     * \brief Draws the evolution of acceptation ratio  as a function of the iterator
     * \param sTitre: title of the plot if specified
     * \param variable: list of variables that should be plotted
     * \param select: select that can be applied on the parameters or the iterator (burnin, selection for autocorrelation)
     * \param option: possible option for this plot
     \li "nonewcanvas": don't create a new canvas, but use the one active 
    */
    void drawAcceptationRatio(TString sTitre, const char *variable="*", const char *select="1>0", Option_t *option="");

    /*!     
     * \brief Draws the parameters as distributions
     * The estimateParamters method has computed the parameters, so this one  plots the parameters as distributions
     * \param sTitre: title of the plots if specified
     * \param variable: list of variables that should be plotted
     * \param select: select that can be applied on the parameters or the iterator (burnin, selection for autocorrelation)
     * \param option: possible option for this plot
     \li "nonewcanvas": don't create a new canvas, but use the one active 
     \li "vertical" : if there are several parameters to be shown put them on top of another instead of side by side
    */
    void drawParameters(TString sTitre, const char *variable = "*", const char *select = "1>0", Option_t * option = "");

    /*!     
     * \brief Compute the autocorrelation 
     * \param l: vector of steps (reading 1 events out of l)
     * \param out: vector of estimations (one per step)
     * \param cut: burn-up value (events ignored from 0 to cut)
     */
    void getAutoCorrelation(vector<int> l, vector<double> *out, int cut=0);
    //@}
    
    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    //! Prints the log
    virtual void printLog(Option_t *option = "");
    //@}
    
protected:
    //---------------------------------------------
    /** @name private methods
     **/
    //@{

    /*!
     * \brief Logarithm of the prior
     */
    void logPriorPdf(double &ret);

    //! Read the possible options
    /*! On top of the other, one can call 
      \li "savealleval": keep all estimation in the _tdsEval (Extremely dangerous as it will keep _nIter branches with _nObs estimations. 
      This option might slow down the process.
      \li "usematrix"  : option use to specify that computation is done with matrices instead of vector<vector<double> >.
      Shoud be used only if one is using an home-made TDistanceFunction
    */
    void parseOption(Option_t *option = "");
    //@}
    //@}

    ClassDef(URANIE::Calibration::TMetropHasting, ID_CALIBRATION)
    //Classe
};
}
}


#endif
