//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TStandardDistanceFunction.h
 \author Jean-Baptiste Blanchard (jean-baptiste.blanchard@cea.fr)
 \date   

 \brief Interface de la classe URANIE::Calibration::TStandardDistanceFunction

 */
/**

 */
#ifndef TSTANDARDDISTANCEFUNCTION_H
#define TSTANDARDDISTANCEFUNCTION_H    

//// URANIE
#include "TDistanceFunction.h"

namespace URANIE
{
namespace Calibration
{

/**
 * \brief Description of the class TLSDistanceFunction
 * \n
 * The distance is estimated as 
 * \f[ Dist = \sum_{i=1}^{n_{Var}} \alpha_i \times \sqrt{ \sum_{j=1}^{n_{Obs}} (\mathbf{y^{j}_{i,Obs}} - \mathbf{y^{j}_{i,Est}})^{2}} \f] where
 * \li \f$n_{Var}\f$ and \f$n_{Obs}\f$ are respectively the number of output variables and the number of events in the reference dataset
 * \li \f$\mathbf{y^{j}_{i,Obs}}\f$ and \f$\mathbf{y^{j}_{i,Est}}\f$ are respectively the reference and estimated value of output \f$ i \f$ for event \f$ j \f$
 * \li \f$ \alpha_i \f$  is a weight that might be used to ponderate one variable with respect to the others
 */
class TLSDistanceFunction: public URANIE::Calibration::TDistanceFunction
{
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    /*! 
     * \brief Default constructor
     * @copydoc hide_DistanceConstructor
     */
    TLSDistanceFunction(URANIE::DataServer::TDataServer *tdsEval, URANIE::DataServer::TDataServer *tdsObs, const char *reference, const char *output, const char *weight="");
    //! Default destructor
    virtual ~TLSDistanceFunction();
    //@}

    //---------------------------------------------
    /** @name Estimate
     **/
    //@{            
    /*! 
      \brief Evaluate the distance function 
    */
    Int_t localeval(Double_t* res);
    //@}

    ClassDef(URANIE::Calibration::TLSDistanceFunction, ID_CALIBRATION)
};

/**
 * \brief Description of the class TWeightedLSDistanceFunction
 * \n
 * The distance is estimated as 
 * \f[ Dist = \sum_{i=1}^{n_{Var}} \alpha_i \times \sqrt{ \sum_{j=1}^{n_{Obs}} \beta_{j} (\mathbf{y^{j}_{i,Obs}} - \mathbf{y^{j}_{i,Est}})^{2}} \f] where
 * \li \f$n_{Var}\f$ and \f$n_{Obs}\f$ are respectively the number of output variables and the number of events in the reference dataset
 * \li \f$\mathbf{y^{j}_{i,Obs}}\f$ and \f$\mathbf{y^{j}_{i,Est}}\f$ are respectively the reference and estimated value of output \f$ i \f$ for event \f$ j \f$
 * \li \f$ \alpha_i \f$  is a weight that might be used to ponderate one variable with respect to the others
 * \li \f$ \beta_j \f$  is a weight that might be used to ponderate one event with respect to the others (uncertainty for instance)
 */
class TWeightedLSDistanceFunction: public URANIE::Calibration::TDistanceFunction
{
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    /*! 
     * \brief Default constructor
     * @copydoc hide_DistanceConstructor
     */
    TWeightedLSDistanceFunction(URANIE::DataServer::TDataServer *tdsEval, URANIE::DataServer::TDataServer *tdsObs, const char *reference, const char *output, const char *weight="");
    //! Default destructor
    virtual ~TWeightedLSDistanceFunction();
    //@}

    //---------------------------------------------
    /** @name Estimate
     **/
    //@{            
    /*! Evaluate the distance function
     */
    Int_t localeval(Double_t* res);
    //@}

    ClassDef(URANIE::Calibration::TWeightedLSDistanceFunction, ID_CALIBRATION)
};

/**
 * \brief Description of the class TRelativeLSDistanceFunction
 * \n
 * The distance is estimated as 
 * \f[ Dist = \sum_{i=1}^{n_{Var}} \alpha_i \times \sqrt{ \sum_{j=1}^{n_{Obs}}  \frac{(\mathbf{y^{j}_{i,Obs}} - \mathbf{y^{j}_{i,Est}})^{2}}{(\mathbf{y^{j}_{i,Obs}})^{2}}} \f] where
 * \li \f$n_{Var}\f$ and \f$n_{Obs}\f$ are respectively the number of output variables and the number of events in the reference dataset
 * \li \f$\mathbf{y^{j}_{i,Obs}}\f$ and \f$\mathbf{y^{j}_{i,Est}}\f$ are respectively the reference and estimated value of output \f$ i \f$ for event \f$ j \f$
 * \li \f$ \alpha_i \f$  is a weight that might be used to ponderate one variable with respect to the others
 */
class TRelativeLSDistanceFunction: public URANIE::Calibration::TDistanceFunction
{
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    /*! 
     * \brief Default constructor
     * @copydoc hide_DistanceConstructor
     */
    TRelativeLSDistanceFunction(URANIE::DataServer::TDataServer *tdsEval, URANIE::DataServer::TDataServer *tdsObs, const char *reference, const char *output, const char *weight="");
    //! Default destructor
    virtual ~TRelativeLSDistanceFunction();
    //@}

    //---------------------------------------------
    /** @name Estimate
     **/
    //@{            
    /*! Evaluate the distance function
     */
    Int_t localeval(Double_t* res);
    //@}

    ClassDef(URANIE::Calibration::TRelativeLSDistanceFunction, ID_CALIBRATION)
};

/**
 * \brief Description of the class TL1DistanceFunction
 * \n
 * The distance is estimated as 
 * \f[ Dist = \sum_{i=1}^{n_{Var}} \alpha_i \times ( \sum_{j=1}^{n_{Obs}}  | \mathbf{y^{j}_{i,Obs}} - \mathbf{y^{j}_{i,Est}} | ) \f] where
 * \li \f$n_{Var}\f$ and \f$n_{Obs}\f$ are respectively the number of output variables and the number of events in the reference dataset
 * \li \f$\mathbf{y^{j}_{i,Obs}}\f$ and \f$\mathbf{y^{j}_{i,Est}}\f$ are respectively the reference and estimated value of output \f$ i \f$ for event \f$ j \f$
 * \li \f$ \alpha_i \f$  is a weight that might be used to ponderate one variable with respect to the others
 */
class TL1DistanceFunction: public URANIE::Calibration::TDistanceFunction
{
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    /*! 
     * \brief Default constructor
     * @copydoc hide_DistanceConstructor
     */
    TL1DistanceFunction(URANIE::DataServer::TDataServer *tdsEval, URANIE::DataServer::TDataServer *tdsObs, const char *reference,  const char *output, const char *weight="");
    //! Default destructor
    virtual ~TL1DistanceFunction();
    //@}

    //---------------------------------------------
    /** @name Estimate
     **/
    //@{            
    /*! Evaluate the distance function
     */
    Int_t localeval(Double_t* res);
    //@}

    ClassDef(URANIE::Calibration::TL1DistanceFunction, ID_CALIBRATION)
};

/**
 * \brief Description of the class TMahalanobisDistanceFunction
 * \n
 * The distance is estimated as 
 * \f[ Dist = \sum_{i=1}^{n_{Var}} \alpha_i \times \sqrt{ (\mathbf{y_{i,Obs}} - \mathbf{y_{i,Est}})^{T} \Omega^{-1} (\mathbf{y_{i,Obs}} - \mathbf{y_{i,Est}}) } \f] where 
 * \li \f$n_{Var}\f$ is the number of output variables 
 * \li \f$\mathbf{y_{i,Obs}}\f$ and \f$\mathbf{y_{i,Est}}\f$ are respectively the reference and estimated matrix of output \f$ i \f$ 
 * \li \f$ \Omega \f$ is the covariance matrix of the events (from the reference file)
 * \li \f$ \alpha_i \f$  is a weight that might be used to ponderate one variable with respect to the others
 */
class TMahalanobisDistanceFunction: public URANIE::Calibration::TDistanceFunction
{
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    /*! 
     * \brief Default constructor
     * @copydoc hide_DistanceConstructor
     */         
    TMahalanobisDistanceFunction(URANIE::DataServer::TDataServer *tdsEval, URANIE::DataServer::TDataServer *tdsObs, const char *reference, const char *output, const char *weight="");
    //! Default destructor
    virtual ~TMahalanobisDistanceFunction();
    //@}

    //---------------------------------------------
    /** @name Estimate
     **/
    //@{            
    /*! Evaluate the distance function
     */
    Int_t localeval(Double_t* res);
    //@}

    ClassDef(URANIE::Calibration::TMahalanobisDistanceFunction, ID_CALIBRATION)
};

} // Fin du namespace Calibration
} // Fin du namespace URANIE
#endif
