//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTMinimisation.cxx
 \author Jean-Baptiste Blanchard
 \date   2020/08/27

 \brief  test file for Linear bayesian classes

 */

// URANIE
#include "UExceptions.h"
#include "TestTMinimisation.h"
#include "TUniformDistribution.h"
#include "TTriangularDistribution.h"
#include "TestUserFunctions.h"
#include "code_paths.h"
#include "uranie_commands.h"

#ifdef WITH_NLOPT
#include "TNloptDirect.h"
#endif

#ifdef WITH_VIZIR
#include "TVizirGenetic.h"
#include "TVizir2.h"
#endif

// CPPUNIT
#include <cppunit/extensions/HelperMacros.h>
#include <cppunit/Test.h>

#define DBL_PRECISION_FINE 1.e-5
#define DBL_PRECISION_MID 5.e-3
#define DBL_PRECISION_LARGE 1.2e-2
#define Flow1DFile "Ex2DoE_n100_sd1.75.dat"
#define DEL(x) if(x!=NULL){delete x; x=NULL;} 

// Constructor Destructor
TMinimisationTest::TMinimisationTest():
    _tds(NULL),_tdsRef(NULL), _tmi(NULL), _seqRun(NULL), _evalJit(NULL), _evalCode(NULL), _inp(NULL), _outp(NULL), _norm(NULL), _out(NULL)
{
    _workingTestPath = gSystem->pwd();
}

TMinimisationTest::~TMinimisationTest()
{
    tearDown();   
}


void TMinimisationTest::cleanWorkingDirectories()
{
    UranieTests::removeFile(_workingTestPath.Data(),"_dataserver_.root");
    UranieTests::removeFile(_workingTestPath.Data(),"Ex2DoE_n100_sd1.root");
#ifndef WIN32
    UranieTests::removeFolder(_workingTestPath.Data(), "URA_*");
    UranieTests::removeFolder(_workingTestPath.Data(), "URANIE*");
#endif

}


// pre and post treatment
void TMinimisationTest::setUp()
{
    _tds = new TDataServer("tds","pouet");
    _norm = new TNormalDistribution("norm", 0., 1.0);
    
    _out=new TAttribute("out");

    _tdsRef = new TDataServer("tdsRef","pouet");
}

void TMinimisationTest::setUpException()
{
    _evalJit = new TCJitEval(EmptyFunction);
    _evalJit->addInput(_norm);
    _evalJit->addOutput(_out);

    _seqRun = new TSequentialRun(_evalJit);
}

void TMinimisationTest::setUpFlow1D(bool function)
{

    _tdsRef->fileDataRead(Flow1DFile);
    
    // define the parameters
    double mu=750.6, std=0.5;
    _tds->addAttribute( new URANIE::DataServer::TNormalDistribution("hl", mu, std) );
    double binf_search=749., bsup_search=751.;
    _tds->getAttribute("hl")->setBounds( binf_search, bsup_search);
    _tds->getAttribute( "hl" )->delShare();

    if(function)
    {
        // Create interface to assessors
        _evalJit= new TCJitEval(flowrateModel);
        // Add inputs in the correct order
        _evalJit->addInput(_tds->getAttribute("hl"));
        _evalJit->addInput(_tdsRef->getAttribute("rw"));
        _evalJit->addInput(_tdsRef->getAttribute("l"));
    
        //Add the output: the cost value 
        _evalJit->addOutput(_out);

        _seqRun = new TSequentialRun(_evalJit);    
    }
    else
    {
        // Create interface to assessors
        _evalCode= new TCodeEval((UranieTests::code_path("flowrate") + " -s -k").data());
    
        // Add inputs in the correct orde
        _inp = new TKeyScript("flowrate_input_with_keys.in");
        _inp->addInput(_tds->getAttribute("hl"),"Hl");    
        _inp->addInput(_tdsRef->getAttribute("rw"),"Rw");
        _inp->addInput(_tdsRef->getAttribute("l"),"L");

        // Add inputs in the correct orde
        _outp = new TKeyResult("_output_flowrate_withKey_.dat");
        _outp->addOutput(_out, "yhat" );    
    
        //Add the output: the cost value 
        _evalCode->addInputFile(_inp);
        _evalCode->addOutputFile(_outp);

        _seqRun = new TSequentialRun(_evalCode);    
    }
}

void TMinimisationTest::setUpFlow2D()
{

    _tdsRef->fileDataRead(Flow1DFile);
    
    // define the parameters
    _tds->addAttribute( new URANIE::DataServer::TUniformDistribution("hu", 1048., 1052.) );
    _tds->getAttribute("hu")->setDefaultValue(1049.2);
    _tds->getAttribute( "hu" )->delShare();
    _tds->addAttribute( new URANIE::DataServer::TUniformDistribution("hl", 748., 752.) );
    _tds->getAttribute("hl")->setDefaultValue(751.8);
    _tds->getAttribute( "hl" )->delShare();

    // Create interface to assessors
    _evalJit= new TCJitEval(flowrateModel);
    // Add inputs in the correct order
    _evalJit->addInput(_tds->getAttribute("hu"));
    _evalJit->addInput(_tds->getAttribute("hl"));
    _evalJit->addInput(_tdsRef->getAttribute("rw"));
    _evalJit->addInput(_tdsRef->getAttribute("l"));
    
    //Add the output: the cost value 
    _evalJit->addOutput(_out);
    
    _seqRun = new TSequentialRun(_evalJit);    

}

void TMinimisationTest::tearDown()
{
    DEL(_evalJit);
    DEL(_evalCode);
    DEL(_inp);
    DEL(_outp);
    DEL(_seqRun);
    DEL(_tmi);
    DEL(_tdsRef);
    DEL(_tds);
    DEL(_norm);
    DEL(_out);

    cleanWorkingDirectories();
}

#ifdef WITH_NLOPT
void TMinimisationTest::runFlow1DNoPlots()
{
    setUpFlow1D(true);

    _tmi = new TMinimisation(_tds,_seqRun,1);
    _tmi->setDistanceAndReference("relativeLS",_tdsRef,"rw:l","Qexp");
    _tds->getAttribute("hl")->setDefaultValue(750.8);
    URANIE::Reoptimizer::TNloptSubplexe solv;
    _tmi->setOptimProperties(&solv);
    _tmi->estimateParameters();

    int nb=_tds->getTuple()->Draw("hl","","goff");
    double *data = _tds->getTuple()->GetV1();
    
    CPPUNIT_ASSERT_EQUAL(1, nb);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(750.06875, data[0], DBL_PRECISION_FINE);
    
}

void TMinimisationTest::runFlow1DNoPlotsCode()
{
    setUpFlow1D(false);
   
    _tmi = new TMinimisation(_tds,_seqRun,1);
    _tmi->setDistanceAndReference("relativeLS",_tdsRef,"rw:l","Qexp");
    _tds->getAttribute("hl")->setDefaultValue(750.8);

    URANIE::Reoptimizer::TNloptSubplexe solv;
    _tmi->setOptimProperties(&solv);
    _tmi->estimateParameters();

    _tds->scan("*");
    int nb=_tds->getTuple()->Draw("hl","","goff");
    double *data = _tds->getTuple()->GetV1();
    
    CPPUNIT_ASSERT_EQUAL(1, nb);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(750.06875, data[0], DBL_PRECISION_FINE);
    
}
#endif

#ifdef WITH_VIZIR
void TMinimisationTest::runFlow1DVizirNoPlots()
{
    setUpFlow1D(true);
   
    _tmi = new TMinimisation(_tds,_seqRun,1);
    _tmi->setDistanceAndReference("relativeLS",_tdsRef,"rw:l","Qexp");
    _tds->getAttribute("hl")->setDefaultValue(750.8);

    URANIE::Reoptimizer::TVizirGenetic solv;
    int Nb=150;
    solv.setSize(Nb,15000);
    _tmi->setOptimProperties(&solv);
    //((URANIE::Reoptimizer::TVizir2*)_tmi->getOptimMaster())->setFixedSeed();
    _tmi->estimateParameters();

    int nb=_tds->getTuple()->Draw("hl","","goff");
    _tds->computeStatistic();

    cout<<" : pouet resultat  "<<setprecision(12)
        <<_tds->getAttribute("hl")->getMean()<<"/"<<750.065<<" (3.5e-2)   "<<_tds->getAttribute("hl")->getStd()<<"/"<<0.041<<" ("<<DBL_PRECISION_LARGE<<")"<<endl;
    CPPUNIT_ASSERT_EQUAL(Nb, nb);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(750.065, _tds->getAttribute("hl")->getMean(), 3.5e-2 );
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.041, _tds->getAttribute("hl")->getStd(),  DBL_PRECISION_LARGE );
    
}

void TMinimisationTest::runFlow2DVizirNoPlots()
{
    setUpFlow2D();
   
    _tmi = new TMinimisation(_tds,_seqRun,1);
    _tmi->setDistanceAndReference("relativeLS",_tdsRef,"rw:l","Qexp");

    int ns=150;
    URANIE::Reoptimizer::TVizirGenetic solv;
    solv.setSize(ns,15000);

    _tmi->setOptimProperties(&solv);
    //((URANIE::Reoptimizer::TVizir2*)_tmi->getOptimMaster())->setFixedSeed();
    _tmi->getOptimMaster()->setTolerance(1e-7);
    _tmi->estimateParameters();

    int nb=_tds->getTuple()->Draw("hl","","goff");
    _tds->computeStatistic();
    
    TMatrixD corr=_tds->computeCorrelationMatrix("hu:hl");

    cout<<" : pouet resultat "<<setprecision(12)
        <<_tds->getAttribute("hl")->getMean()<<"/"<<750<<" (1.5)   "
        <<_tds->getAttribute("hl")->getStd()<<"/"<<0.52<<" (0.32)   "
        <<_tds->getAttribute("hu")->getMean()<<"/"<<1049.764<<" (0.005)    "
        <<_tds->getAttribute("hu")->getStd()<<"/"<<5e-5<<" (3.5e-5)   "
        <<corr(0,1)<<"/"<<0.88<<" (0.12)"<<endl;
    CPPUNIT_ASSERT_EQUAL(ns, nb);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(750., _tds->getAttribute("hl")->getMean(), 1.5);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.52, _tds->getAttribute("hl")->getStd(), 0.35);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1049.764, _tds->getAttribute("hu")->getMean(), 0.005);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(5.e-05, _tds->getAttribute("hu")->getStd(), 3.5e-5);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.88, corr(0,1), 0.22);

}
#endif

void TMinimisationTest::throwExceptionWrongConstruction()
{
    setUpException();

    // no tds
    CPPUNIT_ASSERT_THROW(new TMinimisation(_tds, "flowrateModel", "hl:rw:l","out",10), URANIE::Exceptions::UErrorExceptions);

    // tds without attributes
    CPPUNIT_ASSERT_THROW(new TMinimisation(_tds, flowrateModel,"hlrw:l","out", 10), URANIE::Exceptions::UErrorExceptions);

    // tds youhou
    CPPUNIT_ASSERT_THROW(new TMinimisation(_tds, (URANIE::Launcher::TCode*)NULL, 10), URANIE::Exceptions::UErrorExceptions);

}

void TMinimisationTest::throwExceptionConstruction()
{
    setUpException();

    // no tds
    CPPUNIT_ASSERT_THROW(new TMinimisation(NULL, (URANIE::Relauncher::TRun*)NULL, 10), URANIE::Exceptions::UErrorExceptions);

    // tds without attributes
    CPPUNIT_ASSERT_THROW(new TMinimisation(_tds, (URANIE::Relauncher::TRun*)NULL, 10), URANIE::Exceptions::UErrorExceptions);

    // wrong kind of attribute
    _tds->addAttribute("tutu");
    CPPUNIT_ASSERT_THROW(new TMinimisation(_tds, (URANIE::Relauncher::TRun*)NULL, 10), URANIE::Exceptions::UErrorExceptions);

    _tds->delAttribute("tutu");
    // wrong kind of laws
    _tds->addAttribute(new TTriangularDistribution("triang",0., 1., 0.5) );
    _tds->getAttribute("triang")->delShare();
    CPPUNIT_ASSERT_THROW(new TMinimisation(_tds, (URANIE::Relauncher::TRun*)NULL, 10), URANIE::Exceptions::UErrorExceptions);

}

#ifdef WITH_NLOPT
void TMinimisationTest::throwExceptionOptimproperties()
{
    setUpFlow1D(true);

    // empty solver
    _tmi=new TMinimisation(_tds, _seqRun, 10);
    CPPUNIT_ASSERT_THROW(_tmi->setOptimProperties((URANIE::Reoptimizer::TNloptSubplexe*)NULL);, URANIE::Exceptions::UErrorExceptions);

    // no initialisation
    _tmi->setDistanceAndReference("relativeLS",_tdsRef,"rw:l","Qexp");
    URANIE::Reoptimizer::TNloptSubplexe solv;
    _tmi->setOptimProperties(&solv);
    CPPUNIT_ASSERT_THROW(_tmi->estimateParameters(), URANIE::Exceptions::UErrorExceptions);
}
#endif

void TMinimisationTest::throwExceptionReference()
{
    setUpException();

    _tds->addAttribute(_norm);
    // wrong kind of attribute
    _tmi=new TMinimisation(_tds, _seqRun, 10);

    // no tdsref
    CPPUNIT_ASSERT_THROW(_tmi->setDistanceAndReference("toutou",NULL,"piou","piou","piou"), URANIE::Exceptions::UErrorExceptions);

    // no tdsref empty
    CPPUNIT_ASSERT_THROW(_tmi->setDistanceAndReference("toutou",_tdsRef,"piou","piou","piou"), URANIE::Exceptions::UErrorExceptions);

    _tdsRef->fileDataRead(Flow1DFile);
    // wrong input
    CPPUNIT_ASSERT_THROW(_tmi->setDistanceAndReference("LS",_tdsRef,"rw:lout","Qexp","piou"), URANIE::Exceptions::UErrorExceptions);
    // wrong reference
    CPPUNIT_ASSERT_THROW(_tmi->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexpmero","piou"), URANIE::Exceptions::UErrorExceptions);
    // wrong weight
    CPPUNIT_ASSERT_THROW(_tmi->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexp","piou"), URANIE::Exceptions::UErrorExceptions);
    // wrong reference number
    CPPUNIT_ASSERT_THROW(_tmi->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexp:sd_eps","sd_eps"), URANIE::Exceptions::UErrorExceptions);
    
    // wrong function name
    CPPUNIT_ASSERT_THROW(_tmi->setDistanceAndReference("toutou",_tdsRef,"rw:l","Qexp"), URANIE::Exceptions::UErrorExceptions);

}
