//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
  \file   TestTRejectionABC.cxx
  \author Rudy Chocat
  \date   2020/10/29

  \brief  test file for rejection abc classes

*/

// ROOT
#include "TSystem.h"

// URANIE
#include "UExceptions.h"
#include "TestTRejectionABC.h"
#include "TTriangularDistribution.h"
#include "TestUserFunctions.h"
#include "code_paths.h"
#include "uranie_commands.h"

// CPPUNIT
#include <cppunit/extensions/HelperMacros.h>
#include <cppunit/Test.h>

#define DBL_PRECISION_FINE 1.e-5
#define DBL_PRECISION_LARGE 1.e-2
#define Flow1DFile "Ex2DoE_n100_sd1.75.dat"
#define DEL(x) if(x!=NULL){delete x; x=NULL;} 

// Constructor Destructor
TRejectionABCTest::TRejectionABCTest():
    _tds(NULL),_tdsRef(NULL), _tabc(NULL), _seqRun(NULL), _thrRun(NULL), _evalJit(NULL), _evalCode(NULL), _code(NULL),
    _outRow(NULL), _inp(NULL), _outp(NULL), _norm(NULL), _out(NULL), _threaded(false)
{
    _workingTestPath = gSystem->pwd();
}

TRejectionABCTest::~TRejectionABCTest()
{
    tearDown();
}

void TRejectionABCTest::cleanWorkingDirectories()
{
    UranieTests::removeFile(_workingTestPath.Data(),"_dataserver_.root");
    UranieTests::removeFile(_workingTestPath.Data(),"Ex2DoE_n100_sd1.root");
#ifndef WIN32
    UranieTests::removeFolder(_workingTestPath.Data(), "URA_*");
    UranieTests::removeFolder(_workingTestPath.Data(), "URANIE*");
#endif

}

// pre and post treatment
void TRejectionABCTest::setUp()
{
    _tds = new TDataServer("tds","pouet");
    _norm = new TNormalDistribution("norm", 0., 1.0);
    
    _out=new TAttribute("out");

    _tdsRef = new TDataServer("tdsRef","pouet");
}

void TRejectionABCTest::setUpException()
{
    _evalJit = new TCJitEval(EmptyFunction);
    _evalJit->addInput(_norm);
    _evalJit->addOutput(_out);

    _seqRun = new TSequentialRun(_evalJit);
    
}

void TRejectionABCTest::setUpFlow1D(string type)
{

    _tdsRef->fileDataRead(Flow1DFile);
        
    // define the parameters
    _tds->addAttribute( new URANIE::DataServer::TUniformDistribution("hl", 700, 760) );
    _tds->getAttribute( "hl" )->delShare();

    if(type=="TCJitEval")
    {
        // Create interface to assessors
        _evalJit= new TCJitEval(flowrateModel);
        // Add inputs in the correct order
        _evalJit->addInput(_tds->getAttribute("hl"));
        _evalJit->addInput(_tdsRef->getAttribute("rw"));
        _evalJit->addInput(_tdsRef->getAttribute("l"));
    
        //Add the output: the cost value 
        _evalJit->addOutput(_out);

        _seqRun = new TSequentialRun(_evalJit);    
    }
    else if(type=="TCodeEval")
    {
        // Create interface to assessors
        _evalCode= new TCodeEval((UranieTests::code_path("flowrate") + " -s -k").data());

        // Add inputs in the correct orde
        _inp = new TKeyScript("flowrate_input_with_keys.in");
        _inp->addInput(_tds->getAttribute("hl"),"Hl");    
        _inp->addInput(_tdsRef->getAttribute("rw"),"Rw");
        _inp->addInput(_tdsRef->getAttribute("l"),"L");

        // Add inputs in the correct orde
        _outp = new TKeyResult("_output_flowrate_withKey_.dat");
        _outp->addOutput(_out, "yhat" );    
    
        //Add the output: the cost value 
        _evalCode->addInputFile(_inp);
        _evalCode->addOutputFile(_outp);

        if( ! _threaded )
            _seqRun = new TSequentialRun(_evalCode);
        else
            _thrRun = new TThreadedRun(_evalCode,4);    
    }
    else if(type=="TCode")
    { 
        TString sIn = TString("flowrate_input_with_keys.in");

        // Set the reference input file and the key for each input attributes
        _tdsRef->getAttribute("rw")->setFileKey(sIn, "Rw");
        _tdsRef->getAttribute("l")->setFileKey(sIn, "L");
        _tds->getAttribute("hl")->setFileKey(sIn, "Hl");
 
        // The output file of the code
        _outRow = new TOutputFileRow("_output_flowrate_withRow_.dat");
        // The attribute in the output file
        _outRow->addAttribute(_out);
    
        // Instanciation de mon code
        _code= new TCode(_tdsRef, UranieTests::code_path("flowrate") + " -s -k");
        // on ajoute le fichier de sortie du code
        _code->addOutputFile( _outRow );

    }
}

void TRejectionABCTest::tearDown()
{
    DEL(_evalJit);
    DEL(_evalCode);
    DEL(_code);
    DEL(_outRow);
    DEL(_inp);
    DEL(_outp);
    DEL(_seqRun);
    DEL(_thrRun);
    DEL(_tabc);
    DEL(_tdsRef);
    DEL(_tds);
    DEL(_norm);
    DEL(_out);
    _threaded=false;

    cleanWorkingDirectories();
}

void TRejectionABCTest::runFlow1DCJitNoPlots()
{
    setUpFlow1D("TCJitEval");

    _tabc = new URANIE::Calibration::TRejectionABC(_tds, _seqRun, 100, "");
    _tabc->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexp");
    _tabc->setSeed(123456);
    _tabc->estimateParameters();

    _tds->computeStatistic();
    
    int nb=_tds->getTuple()->Draw("hl","","goff");
    int nbeval=_tabc->getNEval();
    
    cout<<setprecision(12)<<_tds->getAttribute("hl")->getMean()<<"  "<<_tds->getAttribute("hl")->getStd()<<endl;
    CPPUNIT_ASSERT_EQUAL(100, nb);
    CPPUNIT_ASSERT_EQUAL(10000, nbeval);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(749.736392433, _tds->getAttribute("hl")->getMean(), DBL_PRECISION_FINE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.206096600626, _tds->getAttribute("hl")->getStd(), DBL_PRECISION_FINE);
}

void TRejectionABCTest::runFlow1DCJitNoPlotsSmallNumber()
{
    setUpFlow1D("TCJitEval");

    _tabc = new URANIE::Calibration::TRejectionABC(_tds, _seqRun, 10, "");
    _tabc->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexp");
    _tabc->setSeed(123456);
    _tabc->estimateParameters();

    _tds->computeStatistic();
    
    int nb=_tds->getTuple()->Draw("hl","","goff");
    int nbeval=_tabc->getNEval();
    
    cout<<setprecision(12)<<_tds->getAttribute("hl")->getMean()<<"  "<<_tds->getAttribute("hl")->getStd()<<endl;
    CPPUNIT_ASSERT_EQUAL(10, nb);
    CPPUNIT_ASSERT_EQUAL(1000, nbeval);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(749.706254761, _tds->getAttribute("hl")->getMean(), DBL_PRECISION_FINE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.504397974141, _tds->getAttribute("hl")->getStd(), DBL_PRECISION_FINE);
}

void TRejectionABCTest::runFlow1DCJitNoPlotsHigherPercentile()
{
    setUpFlow1D("TCJitEval");

    _tabc = new URANIE::Calibration::TRejectionABC(_tds, _seqRun, 100, "");
    _tabc->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexp");
    _tabc->setPercentile(0.1);
    _tabc->setSeed(123456);
    _tabc->estimateParameters();

    _tds->computeStatistic();
    
    int nb=_tds->getTuple()->Draw("hl","","goff");
    int nbeval=_tabc->getNEval();
    
    cout<<setprecision(12)<<_tds->getAttribute("hl")->getMean()<<"  "<<_tds->getAttribute("hl")->getStd()<<endl;
    CPPUNIT_ASSERT_EQUAL(100, nb);
    CPPUNIT_ASSERT_EQUAL(1000, nbeval);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(749.900878634, _tds->getAttribute("hl")->getMean(), DBL_PRECISION_FINE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.30696222066, _tds->getAttribute("hl")->getStd(), DBL_PRECISION_FINE);
}

void TRejectionABCTest::runFlow1DCJitNoPlotsHigherPercentileNoise()
{
    setUpFlow1D("TCJitEval");

    _tdsRef->addAttribute("sd_rudy","0.1");
    _tabc = new URANIE::Calibration::TRejectionABC(_tds, _seqRun, 100, "");
    _tabc->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexp");
    _tabc->setPercentile(0.1);
    _tabc->setSeed(123456);
    _tabc->setGaussianNoise("sd_rudy");
    _tabc->getDistanceFunction()->setSeed(123456);
    _tabc->estimateParameters();

    _tds->computeStatistic();
    
    int nb=_tds->getTuple()->Draw("hl","","goff");
    int nbeval=_tabc->getNEval();
    
    cout<<setprecision(12)<<_tds->getAttribute("hl")->getMean()<<"  "<<_tds->getAttribute("hl")->getStd()<<endl;
    CPPUNIT_ASSERT_EQUAL(100, nb);
    CPPUNIT_ASSERT_EQUAL(1000, nbeval);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(749.978396928, _tds->getAttribute("hl")->getMean(), DBL_PRECISION_FINE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.30128364761, _tds->getAttribute("hl")->getStd(), DBL_PRECISION_FINE);
}

void TRejectionABCTest::runFlow1DCJitNoPlotsHigherPercentileStdNoise()
{
    setUpFlow1D("TCJitEval");
    
    _tabc = new URANIE::Calibration::TRejectionABC(_tds, _seqRun, 100, "");
    _tabc->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexp");
    _tabc->setPercentile(0.1);
    _tabc->setSeed(123456);
    _tabc->setGaussianNoise("sd_eps");
    _tabc->getDistanceFunction()->setSeed(123456);
    _tabc->estimateParameters();

    _tds->computeStatistic();
    
    int nb=_tds->getTuple()->Draw("hl","","goff");
    int nbeval=_tabc->getNEval();
    
    cout<<setprecision(12)<<_tds->getAttribute("hl")->getMean()<<"  "<<_tds->getAttribute("hl")->getStd()<<endl;
    CPPUNIT_ASSERT_EQUAL(100, nb);
    CPPUNIT_ASSERT_EQUAL(1000, nbeval);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(750.249547464, _tds->getAttribute("hl")->getMean(), DBL_PRECISION_FINE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.68118728872, _tds->getAttribute("hl")->getStd(), DBL_PRECISION_FINE);
}

void TRejectionABCTest::runFlow1DCJitNoPlotsSmallNumberHigherPercentile()
{
    setUpFlow1D("TCJitEval");

    _tabc = new URANIE::Calibration::TRejectionABC(_tds, _seqRun, 10, "");
    _tabc->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexp");
    _tabc->setPercentile(0.1);
    _tabc->setSeed(123456);
    _tabc->estimateParameters();

    _tds->computeStatistic();
    
    int nb=_tds->getTuple()->Draw("hl","","goff");
    int nbeval=_tabc->getNEval();
    
    cout<<setprecision(12)<<_tds->getAttribute("hl")->getMean()<<"  "<<_tds->getAttribute("hl")->getStd()<<endl;
    CPPUNIT_ASSERT_EQUAL(10, nb);
    CPPUNIT_ASSERT_EQUAL(100, nbeval);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(749.899569596, _tds->getAttribute("hl")->getMean(), DBL_PRECISION_FINE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.98386304636, _tds->getAttribute("hl")->getStd(), DBL_PRECISION_FINE);
}

void TRejectionABCTest::runFlow1DCodeEvalNoPlotsSmallNumberHigherPercentile()
{
    setUpFlow1D("TCodeEval");

    _tabc = new URANIE::Calibration::TRejectionABC(_tds, _seqRun, 10, "");
    _tabc->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexp");
    _tabc->setSeed(123456);
    _tabc->setPercentile(0.1);
    _tabc->estimateParameters();

    _tds->computeStatistic();
    
    int nb=_tds->getTuple()->Draw("hl","","goff");
    int nbeval=_tabc->getNEval();
    
    cout<<setprecision(12)<<_tds->getAttribute("hl")->getMean()<<"  "<<_tds->getAttribute("hl")->getStd()<<endl;
    CPPUNIT_ASSERT_EQUAL(10, nb);
    CPPUNIT_ASSERT_EQUAL(100, nbeval);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(749.899569596, _tds->getAttribute("hl")->getMean(), DBL_PRECISION_FINE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.98386304636, _tds->getAttribute("hl")->getStd(), DBL_PRECISION_FINE);
}

void TRejectionABCTest::runFlow1DCodeEvalThreadedNoPlotsSmallNumberHigherPercentile()
{
    _threaded=true;
    setUpFlow1D("TCodeEval");

    _tabc = new URANIE::Calibration::TRejectionABC(_tds, _thrRun, 10, "");
    _tabc->setPercentile(0.1);
    _tabc->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexp");
    _tabc->setSeed(123456);
    _tabc->estimateParameters();

    _tds->computeStatistic();
    
    int nb=_tds->getTuple()->Draw("hl","","goff");
    int nbeval=_tabc->getNEval();
    
    cout<<setprecision(12)<<_tds->getAttribute("hl")->getMean()<<"  "<<_tds->getAttribute("hl")->getStd()<<endl;
    CPPUNIT_ASSERT_EQUAL(10, nb);
    CPPUNIT_ASSERT_EQUAL(100, nbeval);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(749.899569596, _tds->getAttribute("hl")->getMean(), DBL_PRECISION_FINE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.98386304636, _tds->getAttribute("hl")->getStd(), DBL_PRECISION_FINE);
}

void TRejectionABCTest::runFlow1DCodeNoPlotsSmallNumberHigherPercentile()
{
    setUpFlow1D("TCode");

    _tabc = new URANIE::Calibration::TRejectionABC(_tds, _code, 10, "");
    _tabc->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexp");
    _tabc->setPercentile(0.1);
    _tabc->setSeed(123456);
    _tabc->estimateParameters();
    
    _tds->computeStatistic();
    
    int nb=_tds->getTuple()->Draw("hl","","goff");
    int nbeval=_tabc->getNEval();
    
    cout<<setprecision(12)<<_tds->getAttribute("hl")->getMean()<<"  "<<_tds->getAttribute("hl")->getStd()<<endl;
    CPPUNIT_ASSERT_EQUAL(10, nb);
    CPPUNIT_ASSERT_EQUAL(100, nbeval);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(749.899569596, _tds->getAttribute("hl")->getMean(), DBL_PRECISION_FINE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.98386304636, _tds->getAttribute("hl")->getStd(), DBL_PRECISION_FINE);
}

void TRejectionABCTest::runFlow1DPointerFunctionNoPlots()
{
    setUpFlow1D("Function");

    TString inFile="myflow.C";
    gROOT->LoadMacro( gSystem->PrependPathName(gSystem->pwd(), inFile) );
    
    _tabc = new URANIE::Calibration::TRejectionABC(_tds, "flowrateModel" , "hl:rw:l", "out",100, "");
    _tabc->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexp");
    _tabc->setSeed(123456);
    _tabc->estimateParameters();

    _tds->computeStatistic();
    
     int nb=_tds->getTuple()->Draw("hl","","goff");
    int nbeval=_tabc->getNEval();
    
    cout<<setprecision(12)<<_tds->getAttribute("hl")->getMean()<<"  "<<_tds->getAttribute("hl")->getStd()<<endl;
    CPPUNIT_ASSERT_EQUAL(100, nb);
    CPPUNIT_ASSERT_EQUAL(10000, nbeval);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(749.736392433, _tds->getAttribute("hl")->getMean(), DBL_PRECISION_FINE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.206096600626, _tds->getAttribute("hl")->getStd(), DBL_PRECISION_FINE);
}

void TRejectionABCTest::runFlow1DFunctionNoPlots()
{
    setUpFlow1D("Function");

   _tabc = new URANIE::Calibration::TRejectionABC(_tds, flowrateModel , "hl:rw:l", "out",100, "");
    _tabc->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexp");
    _tabc->setSeed(123456);
    _tabc->estimateParameters();
    _tds->computeStatistic();
    
    int nb=_tds->getTuple()->Draw("hl","","goff");
    int nbeval=_tabc->getNEval();
    
    cout<<setprecision(12)<<_tds->getAttribute("hl")->getMean()<<"  "<<_tds->getAttribute("hl")->getStd()<<endl;
    CPPUNIT_ASSERT_EQUAL(100, nb);
    CPPUNIT_ASSERT_EQUAL(10000, nbeval);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(749.736392433, _tds->getAttribute("hl")->getMean(), DBL_PRECISION_FINE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.206096600626, _tds->getAttribute("hl")->getStd(), DBL_PRECISION_FINE);
}

void TRejectionABCTest::throwExceptionConstruction()
{
    setUpException();
    
    // no tds
    CPPUNIT_ASSERT_THROW(new TRejectionABC(NULL, (URANIE::Relauncher::TRun*)NULL, 10), URANIE::Exceptions::UErrorExceptions);

    // tds without attributes
    CPPUNIT_ASSERT_THROW(new TRejectionABC(_tds, (URANIE::Relauncher::TRun*)NULL, 10), URANIE::Exceptions::UErrorExceptions);

    // empty runner
    CPPUNIT_ASSERT_THROW(new TRejectionABC(_tds, (URANIE::Relauncher::TRun*)NULL, 10), URANIE::Exceptions::UErrorExceptions);
    
    // wrong kind of attribute
    _tds->addAttribute("tutu");
    CPPUNIT_ASSERT_THROW(new TRejectionABC(_tds, _seqRun, 10), URANIE::Exceptions::UErrorExceptions);

    // wrong kind of laws
    CPPUNIT_ASSERT_THROW(new TRejectionABC(_tds, (URANIE::Relauncher::TRun*)NULL, 10), URANIE::Exceptions::UErrorExceptions);

}


void TRejectionABCTest::throwExceptionPercentile()
{
    setUpException();

    _tds->addAttribute(_norm);
    // wrong kind of attribute
    _tabc=new TRejectionABC(_tds, _seqRun, 10);

//     double v[1]={10}, s[1]={-1};
    CPPUNIT_ASSERT_THROW(_tabc->setPercentile(-0.1);, URANIE::Exceptions::UErrorExceptions);
CPPUNIT_ASSERT_THROW(_tabc->setPercentile(1.2);, URANIE::Exceptions::UErrorExceptions);
}


void TRejectionABCTest::throwExceptionGaussianNoise()
{
    setUpException();

    _tds->addAttribute(_norm);
    // wrong kind of attribute
    _tabc=new TRejectionABC(_tds, _seqRun, 10);

    _tdsRef->fileDataRead(Flow1DFile);
    _tdsRef->addAttribute("sd_rudy","0.1");
    // too early
    CPPUNIT_ASSERT_THROW(_tabc->setGaussianNoise("sd_rudy");, URANIE::Exceptions::UErrorExceptions);
             
    _tabc->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexp");
        
    //   wrong name
    CPPUNIT_ASSERT_THROW(_tabc->setGaussianNoise("tutu");, URANIE::Exceptions::UErrorExceptions);
    
    // wrong numner of variable
    CPPUNIT_ASSERT_THROW(_tabc->setGaussianNoise("sd_rudy:sd_rudy");, URANIE::Exceptions::UErrorExceptions);

}



void TRejectionABCTest::throwExceptionReference()
{
    setUpException();

    _tds->addAttribute(_norm);
//     _tds->getListOfAttributes()->ls();
    // wrong kind of attribute
     _tabc=new TRejectionABC(_tds, _seqRun, 10);

    // no tdsref
    CPPUNIT_ASSERT_THROW(_tabc->setDistanceAndReference("toutou",NULL,"piou","piou","piou"), URANIE::Exceptions::UErrorExceptions);

    // no tdsref empty
    CPPUNIT_ASSERT_THROW(_tabc->setDistanceAndReference("toutou",_tdsRef,"piou","piou","piou"), URANIE::Exceptions::UErrorExceptions);

    _tdsRef->fileDataRead(Flow1DFile);
    // wrong input
    CPPUNIT_ASSERT_THROW(_tabc->setDistanceAndReference("LS",_tdsRef,"rw:lout","Qexp","piou"), URANIE::Exceptions::UErrorExceptions);
    // wrong reference
    CPPUNIT_ASSERT_THROW(_tabc->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexpmero","piou"), URANIE::Exceptions::UErrorExceptions);
    // wrong weight
    CPPUNIT_ASSERT_THROW(_tabc->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexp","piou"), URANIE::Exceptions::UErrorExceptions);
    // wrong reference number
    CPPUNIT_ASSERT_THROW(_tabc->setDistanceAndReference("LS",_tdsRef,"rw:l","Qexp:sd_eps","sd_eps"), URANIE::Exceptions::UErrorExceptions);
    
    // wrong function name
    CPPUNIT_ASSERT_THROW(_tabc->setDistanceAndReference("toutou", _tdsRef, "rw:l", "Qexp", "bisou"), URANIE::Exceptions::UErrorExceptions);

}
