// --- C++ ---
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2019 CEA/DEN
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   TEgoKBModeler.cxx
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2019-11-12
 
 \brief  kriging believe ego modeler
*/


#include "TEgoKBModeler.h"

ClassImp(URANIE::MetaModelOptim::TEgoKBModeler);

// constructeur
URANIE::MetaModelOptim::TEgoKBModeler::TEgoKBModeler() :
    TEgoModeler(),
    _tds2(NULL), _mm1(NULL), _mm2(NULL)
{
    // default model
    _correlat = "matern7/2";
    _tendance = "const";
    _regul = 0.0;
    // default solver
    _obj = "ML";
    _algo = "BOBYQA";
    //_algo = "Praxis";
    _doesize = 300;
    _maxeval = 1000;
    
}

URANIE::MetaModelOptim::TEgoKBModeler::~TEgoKBModeler()
{
    if (_mm1) delete _mm1;
    if (_tds2) delete _tds2;
    if (_mm2) delete _mm2;
}

// parametrage
void URANIE::MetaModelOptim::TEgoKBModeler::setModel(const char *model, const char *trend, double regularisation)
{
    _correlat = model;
    _tendance = trend;
    if (regularisation) _regul = regularisation;
}

void URANIE::MetaModelOptim::TEgoKBModeler::setSolver(const char *objective, const char *algo, int screening, 
        int optsize)
{
    _obj = objective;
    _algo = algo;
    if (screening) _doesize = screening;
    if (optsize) _maxeval = optsize;
}


// helpers
/* remplissage d'un tds avec
 * - les points déjà calculés donc connues
 * - les points en cours de calcul estimés
 ***/
double URANIE::MetaModelOptim::TEgoKBModeler::fillWithExpectedValue(vector<char> &mark, 
        vector<vector<URANIE::DataServer::UEntry*>> &sampl,
        URANIE::DataServer::TDataServer &tds2, URANIE::Modeler::TKriging *krig)
{
    int i, j, tag;
    double fmin;
    double tmpout[2];
    vector<double> tmp(_NinDat + 2);

    fmin = 1.0e20;
    for (j=0; j<_NC; j++) {
        switch (mark[j]) {
        case PENDING :
            for (i=0; i<_NinDat+1; i++) tmp[i] = *sampl[j][i]->val.dou;
            krig->eval(&tmp[1], tmpout);
#ifdef ALPHA
            tmp[_NinDat+1] = tmpout[0] + ALPHA * sqrt(tmpout[1]);
            //tmp[_NinDat+1] = tmpout[0] + ALPHA * (ALEAS - rnd.Rndm()) * sqrt(tmpout[1]);
#else
            tmp[_NinDat+1] = tmpout[0];
#endif
            tag = 0;
            break;
        case EVALUATED :
            for (i=0; i<_NinDat+1; i++) tmp[i] = *sampl[j][i]->val.dou;
            {
                auto out = new vector<URANIE::DataServer::UEntry*> (
                    sampl[j].begin()+_NinDat+1, sampl[j].end());
                tmp[_NinDat+1] = _crit->from(out);
                delete out;
            }
            tag = 0;
            break;
        default :
            tag = 1;
            break;
        }
        if (tag) continue;
        tds2.getTuple()->Fill(tmp.data());
        if (tmp[_NinDat+1] < fmin) fmin = tmp[_NinDat+1];
    }
    //tds2.exportData("firstPlus.res");
    return fmin;
}

// protocole 
void URANIE::MetaModelOptim::TEgoKBModeler::build(URANIE::DataServer::TDataServer *tds,
    URANIE::Reoptimizer::TFitGetter *crit,
    vector<URANIE::DataServer::TAttribute*>& in, URANIE::DataServer::TAttribute *out,
    URANIE::DataServer::TAttribute *yvar, double *fmin, int esize)
{
    int i;
    
    _tds = tds;
    _crit = crit;
    _in = in;
    _y = out;
    _yvar = yvar;
    _fmin = fmin;
    _NC = esize;
    
    // input output descriptor
    _indescr="";
    for (i=0; in[i]; i++) {
        _indescr += (i) ? TString(":")+TString(in[i]->GetName()) :
                        TString(in[i]->GetName()) ;
    }
    _NinDat = i;
    _outdescr = out->GetName();
    
    // meta modele 1 with calculated value
    _mm1 = new URANIE::Modeler::TGPBuilder(tds, _indescr.Data(), _outdescr, 
            _correlat, _tendance);
    if (_regul != 0.0) _mm1->setRegularisation(_regul);
    
    // tds2 fake sampling for mm2
    _tds2 = new URANIE::DataServer::TDataServer("tmptds", "temporaire ego tds");
    _tds2->keepFinalTuple(kFALSE);
    for (i=0; in[i] != NULL; i++) _tds2->addAttribute(in[i]);
    _tds2->addAttribute(_y);
    _tds2->createTuple();
    
    // mm2 completed with estimated value
    _mm2 = new URANIE::Modeler::TGPBuilder(_tds2, _indescr.Data(), _outdescr, 
            _correlat, _tendance);
    //mm2.setHasMeasurementError();
    if (_regul != 0.0) _mm2->setRegularisation(_regul);
    //_mm2->setLengthRange(2.0e-2, 4.0);
    
}

URANIE::Modeler::TKriging *URANIE::MetaModelOptim::TEgoKBModeler::next(vector<char> &mark,
    vector<vector<URANIE::DataServer::UEntry*>> &sampl, int dodoe)
{
    int j;
    URANIE::Modeler::TKriging *krig, *krig2;
    
    _mm1->updateObservations();
    switch (dodoe) {
    case 1 :
        _mm1->findOptimalParameters(_obj, 0, _algo, _maxeval, kFALSE);
        break;
    case 2 :
        _mm1->findOptimalParameters(_obj, 0, "subplexe", 1, kFALSE);
        break;
    default:
        _mm1->findOptimalParameters(_obj, _doesize, _algo, _maxeval, kTRUE);
        break;
    }

    // metamodele1
    try {
        krig = _mm1->buildGP();
    }
    catch (URANIE::Exceptions::UErrorExceptions& ue) {
        ue.printMessage();
        return NULL;
    }
    for (j=0; _in[j] != NULL; j++) krig->addInput(_in[j]);
    krig->addOutput(_y);
    krig->addOutput(_yvar);

    // fill tds2
    _tds2->deleteTuple();
    _tds2->createTuple();
    *_fmin = fillWithExpectedValue(mark, sampl, *_tds2, krig);
    delete krig;

    // metamodel2
    _mm2->updateObservations();
    //if (dodoe) _mm2->findOptimalParameters(_obj, _doesize, _algo, _maxeval, kTRUE);
    //else _mm2->findOptimalParameters(_obj, 0, _algo, _maxeval, kFALSE);
    _mm2->getCorrFunction()->setParameters(_mm1->getCorrFunction()->getParameters());
    _mm2->findOptimalParameters(_obj, 0, "subplexe", 1, kFALSE);

    try {
        krig2 = _mm2->buildGP();
    }
    catch (URANIE::Exceptions::UErrorExceptions& ue) {
        ue.printMessage();
        return NULL;
    }
    for (j=0; _in[j] != NULL; j++) krig2->addInput(_in[j]);
    krig2->addOutput(_y);
    krig2->addOutput(_yvar);

    return krig2;    
}
