// --- C++ ---
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2019 CEA/DEN
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   THJMA.cxx
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2019-11-21
 
 \brief  HJMA search algorithm
*/


#include "THJMA.h"

// ROOT
#include "TLeaf.h"

// Uranie

#ifdef WIN32
#include <process.h>
#endif

//#define TRACE

//#define PAS 0.2
//#define STEP 8

//#define PAS 0.464159
//#define nSEARCH 9
//#define nMAJ 6
//#define nRAFINE 6

#define PAS 0.316228
#define nSEARCH 6
#define nMAJ 4
#define nRAFINE 4

ClassImp(URANIE::MetaModelOptim::THJMA);



/*** constructor ***/
URANIE::MetaModelOptim::THJMA::THJMA(URANIE::DataServer::TDataServer *tds, URANIE::Relauncher::TRun *run) :
    URANIE::Reoptimizer::TOptimShare(tds, run), 
    _y(NULL), _crit(NULL), _valid(NULL), _global(NULL),
    _mode(NULL), _dom(NULL)
{
    // init
    _first = 1;
    _update = 0;
    // default
    _Ncall= 1000;
    _Ncut = 8;
    _Nstep = 10;
    _Nfirst = 40;

#ifndef WIN32
    _rnd.SetSeed(getpid());
#else
    _rnd.SetSeed(_getpid());
#endif
}

URANIE::MetaModelOptim::THJMA::~THJMA()
{
    int i, sz;
    
    if(_dom) delete[] _dom;
    if (_global) {
        sz = _NinDat * _Ncut + 1;
        for (i=0; i<sz; i++) delete[] _global[i];
        delete[] _global;
    }
    if (_mode) {
        for (auto soon : *_mode) delete[] soon;
        delete _mode;
    }
    if (_crit) delete _crit;
    if (_valid) delete _valid;
}

/* setup */

void URANIE::MetaModelOptim::THJMA::addObjective(URANIE::DataServer::TAttribute *obj,
        URANIE::Reoptimizer::TOptimFit *fit)
{
    if (!_crit) {
        _y = obj;
        _crit = new URANIE::Reoptimizer::TFitGetter(obj, _fun->getOutputList(), fit);
    }
    else std::cout << "HJMA is single objective\n";
}

void URANIE::MetaModelOptim::THJMA::setUsable(URANIE::Reoptimizer::TOptimFit *signif)
{
    if (_y) _valid = new 
            URANIE::Reoptimizer::TFitGetter(_y, _fun->getOutputList(), signif);
    else std::cout << "objective is undefined\n";
}

void URANIE::MetaModelOptim::THJMA::setSize(int first, int step)
{
    _Nfirst = first;
    _Nstep = step;
}


/*** protected ***/
void URANIE::MetaModelOptim::THJMA::build() 
{
    int i;
    URANIE::DataServer::TAttribute *att;

    // search domain
    _dom = new struct domain[_NinDat];
    TList *attrlist = _tds->getListOfAttributes();
    TIter next(attrlist);
    
    for (i=0; i<_NinDat; i++) {
        att = (URANIE::DataServer::TAttribute *) next();
        _dom[i].mini = att->getLowerBound();
        _dom[i].maxi = att->getUpperBound();
        _dom[i].it = _dom[i].maxi - _dom[i].mini;
    }
    delete attrlist;
    
    // global search screening tab
    unsigned long ui, sz = _Ncut*_NinDat+1;
    std::cout << sz << " hjma build\n";
    _global = new double*[sz];
    for (ui=0; ui<sz; ui++) {
        _global[ui] = new double[_NinDat+_NoutDat+1];
    }
    
    // global map
    std::cout << "hjma build end\n";
    
}

inline double URANIE::MetaModelOptim::THJMA::evaluate_pt(struct hjma_pt &pt)
{
    long j;
    double ret;
    
    inputForCode(pt.xy, pt.codi);
    _run->put(0, pt.codi, pt.codo);
    _run->get(j);
    ret = pt.xy[_NinDat] = _crit->from(pt.codo);
    return ret;
}

inline void URANIE::MetaModelOptim::THJMA::evaluate_best(struct hjma_work &self)
{
    (void) evaluate_pt(*self.best);
}

inline bool URANIE::MetaModelOptim::THJMA::evaluate_current(struct hjma_work &self)
{
    double mini;
    struct hjma_pt *tmp;
    bool ret;
    
    mini = evaluate_pt(*self.cur);
    ret = (mini < self.best->xy[_NinDat]) ;
    if (ret) {
        // swap pointers
        tmp = self.best;
        self.best = self.cur;
        self.cur = tmp;
    }
    return ret;
}

/* global search */
void URANIE::MetaModelOptim::THJMA::globalsearch(struct hjma_work &self)
{
    int i, k;
    double base, step;
    //int sz = 2 *_NinDat + 1;
    vector<double> tmp(_NinDat);
    
    // initial point
    for (i=0; i<_NinDat; i++) tmp[i] = _dom[i].it*_rnd.Rndm() +  _dom[i].mini;
    //decode(tmp, _global[0]);
    for (k=0; k<_NinDat; k++) self.best->xy[k] = tmp[k];
    evaluate_best(self);
    
    // OAT grid
    for (i=0; i<_NinDat; i++) { //dim
        step = _dom[i].it/(_Ncut+0.9999);
        base = tmp[i] + step;
        for (k=0; k<_NinDat; k++) self.cur->xy[k] = tmp[k];
        while (base <= _dom[i].maxi) {
            self.cur->xy[i] = base;
            if (evaluate_current(self)) {
                for (k=0; k<_NinDat; k++) self.cur->xy[k] = tmp[k];
            }
            base += step;
        }
        base = tmp[i] - step;
        for (k=0; k<_NinDat; k++) self.cur->xy[k] = tmp[k];
        while (base >= _dom[i].mini) {
            self.cur->xy[i] = base;
            if (evaluate_current(self)) {
                for (k=0; k<_NinDat; k++) self.cur->xy[k] = tmp[k];
            }
            base -= step;
        }
    }
}

/* Hooke-Jeeves local search */
int URANIE::MetaModelOptim::THJMA::hjonestep(struct hjma_work &self, double *dep, double abstep)
{
    int i, j;
    double vmemo, step;
    int ret = 0;
    
    for (i=0; i<_NinDat; i++) {
        //dep[i] = 0.0;
        self.cur->xy[i] = self.best->xy[i];
    }
    
    for (i=0; i<_NinDat; i++) {
        vmemo = self.best->xy[i];
        step = abstep * _dom[i].it;
        dep[i] = 0.0;
        // up
        self.cur->xy[i] += step;
        if (self.cur->xy[i] > _dom[i].maxi) self.cur->xy[i] = _dom[i].maxi;
        if (evaluate_current(self)) {
            for (j=0; j<_NinDat; j++) self.cur->xy[j] = self.best->xy[j];
            if (self.cur->xy[i] != _dom[i].maxi) {
                dep[i] = step;
                ret = 1;
            }
            continue;
        }
        // down
        self.cur->xy[i] = vmemo - step;
        if (self.cur->xy[i] < _dom[i].mini) self.cur->xy[i] = _dom[i].mini;
        if (evaluate_current(self)) {
            for (j=0; j<_NinDat; j++) self.cur->xy[j] = self.best->xy[j];
            if (self.cur->xy[i] != _dom[i].mini) {
                dep[i] = -step;
                ret = 1;
            }
            continue;
        }
        // stay
        self.cur->xy[i] = vmemo;
    }
    return ret;
}


void URANIE::MetaModelOptim::THJMA::hjtwostep(struct hjma_work &self, double *dep)
{
    int i, dec, out;
    double *orig;
    vector<double> vorig ;
    vorig.resize(_NinDat);
    orig = &vorig[0];
    
    // memorise initial point
    for (i=0; i<_NinDat; i++) orig[i] = self.best->xy[i];
    dec = 1;
    do {
        out = 0;
        for (i=0; i<_NinDat; i++) {
            self.cur->xy[i] = orig[i] + dec*dep[i];
            if (self.cur->xy[i] < _dom[i].mini || 
                self.cur->xy[i] > _dom[i].maxi) 
                out = 1;
        }
        if (out) break;
        if (evaluate_current(self)) dec++;
        else break;
    } while(1);
    // copy
#ifdef TRACE
    std::cout << dec <<  " ";
#endif
}

//#define ARTICLE
int URANIE::MetaModelOptim::THJMA::hjsearch(struct hjma_work &self, double prox, int loop, int dup)
{
    double coef;
    int i, tag;
    vector<double> dir(_NinDat);
    
#ifdef ARTICLE
    double coef0 = 0.4;
    coef = coef0;
    for (i=0; i<5; i++) {
#else
    coef = PAS; 
    for (i=0; i<loop; i++) {
#endif
        do {
            tag = hjonestep(self, dir.data(), prox);
            if (tag) hjtwostep(self, dir.data());
        } while (tag);
        if (dup && i && duplicate(self, prox)) return 0;
#ifdef TRACE
        std::cout << std::endl << prox << ":";
#endif
        prox *= coef;
#ifdef ARTICLE
        coef *= coef0;
#endif
    }
    if (_valid && _valid->from(self.best->codo) > 0) return 0;
    else return 1;
}

// test duplication des modalités
int URANIE::MetaModelOptim::THJMA::duplicate(struct hjma_work &self, double coef)
{
    int i, near;
    double delta, step;
    
    step = coef * sqrt(_NinDat);
    near = 0;
    for (auto soon : *_mode) {
        near = 1;
        for (i=0; i<_NinDat; i++) {
            delta = fabs(soon[i] - self.best->xy[i]);
            if (delta > step*_dom[i].it) {
                near = 0;
                break;
            }
        }
        if (near) break;
    }
    return near;
}

// ajout de la modalités
void URANIE::MetaModelOptim::THJMA::addSolution(struct hjma_work &self)
{
    int i, inew;
    double *tmp = new double [_NinDat+1];
    
    inew = _mode->size();
    // memory
    for (i=0; i<_NinDat; i++) tmp[i] = self.best->xy[i];
    tmp[_NinDat] = _crit->from(self.best->codo);
    _mode->push_back(tmp);
    // tds
    auto tmpi = new std::vector<URANIE::DataServer::UEntry*>(self.best->codi->begin(),
            self.best->codi->end()); 
    auto tmpo = new std::vector<URANIE::DataServer::UEntry*>(self.best->codo->begin(),
            self.best->codo->end());
    auto all = new std::vector<URANIE::DataServer::UEntry*>();
    
    allocDataEntries(*all);
    all->at(0)->pushEntry((double) inew);
    auto tmpa = new std::vector<URANIE::DataServer::UEntry*>(all->begin()+1, all->end());
    allFromCode(tmpi, tmpo, tmpa);
    _tds->getTuple()->Fill(all);
    
    for (auto it : *all) delete it;
    delete all;
}

int URANIE::MetaModelOptim::THJMA::getBest(struct hjma_work &self)
{
    int i;
    double *ret;
    
    if (_mode->size() == 0) return 0; 
    // search
    ret = (*_mode)[0];
    for (auto solu: *_mode) {
        if (solu[_NinDat] < ret[_NinDat]) ret = solu;
    }
    // copy
    for (i=0; i<_NinDat; i++) self.best->xy[i] = ret[i];
    return 1;
}

/*** public ***/
void URANIE::MetaModelOptim::THJMA::fitnessChanged()
{
    _update = 1;
}

// solve
void URANIE::MetaModelOptim::THJMA::solverLoop()
{
    int i, loop;
    std::vector<double> one, two;
    std::vector<double *> *oldmode;
    std::vector<std::vector<URANIE::DataServer::UEntry*>> codi, codo;
    struct hjma_work self;
    
    // alloc
    one.resize(_NinDat+1); two.resize(_NinDat+1);
    codi.resize(2); codo.resize(2);
    allocInCodeEntries(2, codi);
    allocOutCodeEntries(2, codo);
    // self init
    self.one.xy = &one[0]; self.one.codi = &codi[0]; self.one.codo = &codo[0];
    self.two.xy = &two[0]; self.two.codi = &codi[1]; self.two.codo = &codo[1];
    self.best = &self.one; self.cur = &self.two;
    // work
    if (_first) {
        std::cout << "hjma first\n" ;
        fillDS();
        _tds->createTuple();
        
        build();
        _mode = new std::vector<double *>();
        loop = _Nfirst;
        _first = 0;
    }
    else if (_update) {
        // update
        std::cout << "hjma next " << _mode->size() << std::endl;
        _tds->createTuple();

        oldmode = _mode;
        _mode = new std::vector<double *>();
        for (auto solu : *oldmode) {
            for (i=0; i<_NinDat; i++) self.best->xy[i] = solu[i] ;
            (void) evaluate_best(self);
            if (hjsearch(self, 0.005, nMAJ, 1)) addSolution(self);
            delete[] solu;
        }
        delete oldmode;
        _update = 0;
        loop = _Nstep;
    }
    else loop = _Nstep; // rafine
    
    // more search
    for (i=0; i<loop; i++) {
        globalsearch(self);
        if (hjsearch(self, 0.05, nSEARCH, 1)) addSolution(self);
    }
    
    // rafine best
    if ( getBest(self) ) {
        (void) hjsearch(self, 5.e-5, nRAFINE, 0);
    }
    //return best;
    //std::cout << _mode->size() << "\nhjma end\n";
    //_tds->createTuple();
    //for (auto solu : *_mode) _tds->getTuple()->Fill(solu);
    for (i=0; i<2; i++) {
        for (auto it : codi[i]) delete it;
        codi[i].clear();
        for (auto it : codo[i]) delete it;
        codo[i].clear();
    }
#ifdef TRACE
    for (auto solu : *_mode) {
        for (i=0; i<_NinDat+_NoutDat+1; i++) std::cout << solu[i] << " ";
        std::cout << std::endl;
    }
#endif
}

// result
double *URANIE::MetaModelOptim::THJMA::item(int i)
{
    unsigned long j = i;
    
    if (j < _mode->size()) return (*_mode)[j];
    else return NULL;
}

