/*//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////*/

#include <stdlib.h>
#include <math.h>
#include "linpack_d.h"
#ifdef WIN32
#include "libpgauss_Export.h"
#else
#define libpgauss_EXPORT
#endif



/* Calcul des erreurs par LOO			*/
/* Methode gpLoo0				*/
/* Input Kin : matrice de variance		*/
/* Input yin : vecteur des observations		*/
/* Input n : dimension Kin[n x n], yin[n]	*/
/* Output loo[2] : critere + variance estimee 	*/
/* loo[0] : eqm	et loo[1] = variance estimee	*/
/* Si loo[1] = -1 => echec Cholesky	  	*/
libpgauss_EXPORT int gpLoo0(double *Kin, double *yin, int n, double *loo) {
  void gpDposl ( double a[], int lda, int n, double b[] );
  int i, j, info;
  double dT, determinant[2], *errLoo, *sg2Loo, vt;

  /* Recopie des donnees qui seront modifiees */
  double *K, *pK, *y, *py;
  
  K = calloc(sizeof(double),n*n);
  y = calloc(sizeof(double),n);
  errLoo = calloc(sizeof(double),n);
  sg2Loo = calloc(sizeof(double),n);
  for(pK = K,i=0;i<n*n;i++) *pK++ = *Kin++;
  for(py = y,i=0;i<n;i++)   *py++ = *yin++;
  /*******************************/

  info = dpofa (K, n, n);        /* decomposition Choleski K = Lt L*/
  if (info == 0) {
    dposl (K,n,n,y);             /* resolution lineaire => e = T * y */
    dtrdi(K,n,n,determinant,11); /* inverse d'une triangulaire superieure */
                                 /* inverse de L */
    for(i=0;i<n;i++) {
      dT=0.;
      for(j=i;j<n;j++) dT += K[i+j*n]*K[i+j*n];
      errLoo[i] = y[i] / dT;
      sg2Loo[i] = 1. / dT;
    }
    for(loo[0]=0., loo[1]=0., i=0;i<n;i++) {
      vt      = errLoo[i] * errLoo[i];
      loo[0] += vt;
      loo[1] += vt / sg2Loo[i];
    }
    loo[0] /= n;
    loo[1] /= n;
  }
  else {
    /* printf("gpLoo0 : The leading minor %d is not positive\n", info); */
    loo[0] = -1.;
    loo[1] = -1.;
  }
  free(K);  K=NULL;
  free(y);  y=NULL;
  free(errLoo);   errLoo = NULL;
  free(sg2Loo);   sg2Loo = NULL;
  return info;
}

