/*//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "linpack_d.h"
#ifdef WIN32
#include "libpgauss_Export.h"
#else
#define libpgauss_EXPORT
#endif

/* Calcul des erreurs par LOO                           */
/* Methode gpLoohBayes                                  */
/* Input BetaPrior : moyenne a priori			*/
/* Input Qprior : matrice de variance a priori		*/
/* Input Hin : matrice des regresseurs			*/
/* Input Kin : matrice de variance                      */
/* Input yin : vecteur des observations                 */
/* Input n : dimension Kin[n x n], yin[n]               */
/* Input p : dimension Hin[n x p]                       */
/* Output loo[2] : critere + variance estimee 		*/
/* loo[0] : eqm	et loo[1] = variance estimee		*/
/* Si loo[1] = -1 => echec Cholesky	  		*/
libpgauss_EXPORT int gpLoohBayes(double *BetaPrior, double *Qprior, double *Hin, double *Kin, double *yin, int n, int p, double *loo) {
  void gpDposl ( double a[], int lda, int n, double b[] );
  int prodMatAtA(double A[], int n, int p, double C[]);
  int gpInvCholesky(double *K, int n, double *iK);
  int i, j, k, jn, info;
  double xa, xk, dT, determinant[2], *errLoo, *sg2Loo, vt;

  double * iQprior = calloc(sizeof(double),p*p);
  double * v  = calloc(sizeof(double),n);
  double * W  = calloc(sizeof(double),p*p);
  double * Q  = calloc(sizeof(double),n*p);
  double * At = calloc(sizeof(double),n*p);
  double * ay = calloc(sizeof(double),p);

  /* Recopie des donnees qui seront modifiees */
  double *H, *pH, *K, *pK, *y;
  H = calloc(sizeof(double),n*p);
  K = calloc(sizeof(double),n*n);
  y = calloc(sizeof(double),n);
  errLoo = calloc(sizeof(double),n);
  sg2Loo = calloc(sizeof(double),n);
  for(pH = H,i=0;i<n*p;i++) *pH++ = *Hin++;
  for(pK = K,i=0;i<n*n;i++) *pK++ = *Kin++;
  /*******************************/

  gpInvCholesky(Qprior, p, iQprior);
  /* Calcul de y - H * BetaPrior */
  for(i=0;i<n;i++) {
    y[i] = yin[i];
    for(j=0;j<p;j++) y[i] -= H[i+j*n] * BetaPrior[j];
  }
  info = dpofa (K, n, n); /* Decomposition Cholesky K = Lt L*/
  if (info == 0) {
    for(i=0;i<n;i++) v[i]=y[i];
    dposl (K, n, n, v);                            /* Resolution y = K v */
    for (j=0;j<p;j++) gpDposl(K, n, n, &(H[j*n])); /* Resolution H = L(-t) M */
   
    prodMatAtA(H, n, p, W);                /* calcul de W = MtM              */
    for(i=0;i<p*p;i++) W[i] += iQprior[i]; /* calcul de W = iQprior + MtM    */
    info = dpofa (W, p, p);                /* W triangulaire superieure */
    dtrdi(W,p,p,determinant,11);           /* inverse de W */
    /* Calcul de Q = M * iW */
    for(i=0;i<n;i++) {
      for(j=0;j<p;j++) {
        jn = j*n;
        Q[i+jn]=0.;
        for(k=0;k<=j;k++) Q[i+jn] += H[i+k*n] * W[k+j*p];
      }
    }
    dtrdi(K,n,n,determinant,11); /* calcul de iL */
    /* calcul de At = iL * Q */
    for(i=0;i<n;i++) {
      for(j=0;j<p;j++) {
        jn = j*n;
        for(k=i;k<n;k++) At[i+jn] += K[i+k*n] * Q[k+jn];
     }
    }
    /* Calcul de yt * At */
    for(j=0;j<p;j++) {
      ay[j]=0.; jn=j*n;
      for(i=0;i<n;i++) ay[j] += y[i] * At[i+jn];
    }
    /* Elements diagonaux de inv(K)*/
    for(i=0;i<n;i++) {
      dT = 0.;
      for(j=i;j<n;j++) { xk = K[i+j*n];  dT += xk * xk; }
      for(j=0;j<p;j++) { xa = At[i+j*n]; dT -= xa * xa; }
      errLoo[i] = v[i];
      for(j=0;j<p;j++) errLoo[i] -= At[i+j*n] * ay[j];
      errLoo[i] /= dT;
      sg2Loo[i]  = 1. / dT;
     }
     for(loo[0] = 0., loo[1] = 0., i=0;i<n;i++) {
       vt      = errLoo[i] * errLoo[i];
       loo[0] += vt;
       loo[1] += vt / sg2Loo[i];
     }
     loo[0] /= n;
     loo[1] /= n;
  }
  else {
    /* printf("gpLoohBayes : The leading minor %d is not positive\n", info); */
    loo[0] = -1.;
    loo[1] = -1.;
  }
    free(iQprior);    iQprior = NULL;
    free(v);        v = NULL;
    free(W);      W = NULL;
    free(Q);       Q = NULL;
    free(At);       At = NULL;
    free(ay);        ay = NULL;
    free(H);        H = NULL;
    free(K);        K = NULL;
    free(y);        y = NULL;
    free(errLoo);        errLoo = NULL;
    free(sg2Loo);        sg2Loo = NULL;
  return info;
}

