/*//////////////////////////////////////////////////////////////
 // Copyright (C) 2013-2022 CEA/DES
 //
 // This program is free software: you can redistribute it and/or modify
 // it under the terms of the GNU Lesser General Public License as published
 // by the Free Software Foundation, either version 3 of the License, or any
 // later version.
 //
 // This program is distributed in the hope that it will be useful,
 // but WITHOUT ANY WARRANTY; without even the implied warranty of
 // MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 // GNU Lesser General Public License for more details.
 //
 // You should have received a copy of the GNU Lesser General Public License
 // along with this program. If not, see <http://www.gnu.org/licenses/>.
 //////////////////////////////////////////////////////////////
 //////////////////////////////////////////////////////////////
 // $Id$
 // $Author$
 // $Date$
 // $Revision$
 // $State$
 //////////////////////////////////////////////////////////////*/
/*!
 \file   gpML0.c
 \author Jean-Marc Martinez <jean-marc.martinez@cea.fr>
 \date   2012/07/19 15:38:00

 \brief
 */
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#ifdef WIN32
#include "libpgauss_Export.h"
#else
#define libpgauss_EXPORT
#endif

/* Calcul du moins log de la vraisemblance                */
/* Methode gpML0                                          */
/* Input Kin : matrice de variance (array double Fortran) */
/* Input yin : vecteur des observations (array double)    */
/* Input n : dimension Kin[n x n], yin[n]                     */
/* cas = 0 general sinon variance en facteur              */
/* Output ml[2] : critere + variance si cas != 0      */
/* Si ml[1] = -1 => echec lors decomposition Cholesky      */

libpgauss_EXPORT int gpML0(double *Kin, double *yin, int n, double *ml, int cas)
{
    void gpDposl(double a[], int lda, int n, double b[]);
    int dpofa(double a[], int lda, int n);
    int i, info;
    double v1, v2;

    /* Recopie des donnees qui seront modifiees */
    double *K, *pK, *y, *py;
	
    K = calloc(sizeof(double), n * n);
    y = calloc(sizeof(double), n);
    for (pK = K, i = 0; i < n * n; i++)
        *pK++ = *Kin++;
    for (py = y, i = 0; i < n; i++)
        *py++ = *yin++;
    /*******************************/

    info = dpofa(K, n, n); /* Decomposition Choleski K = L Lt   */
    /* L matrice triangulaire superieure */
    if (info == 0)
    {
        gpDposl(K, n, n, y); /* Resolution y = Lt x */
        for (v1 = 0., i = 0; i < n; i++)
            v1 += y[i] * y[i];
        for (v2 = 0., i = 0; i < n; i++)
            v2 += log(K[i + i * n]);/* ce n'est plus K mais L */
        if (cas == 0)
        { /* general */
            ml[0] = v1 + 2. * v2;
            ml[1] = 0.; /* pas utilise dans ce cas */
        }
        else
        { /* variance en facteur */
            ml[0] = n * log(v1) + 2. * v2;
            ml[1] = v1 / n;
        }
    }
    else
    {
        /* printf("gpML0 : The leading minor %d is not positive\n", info); */
        ml[0] = 0.;
        ml[1] = -1.; /* Cholesky fails */
    }
    free(K);        K = NULL;
    free(y);        y = NULL;
    return info;
}

