/*//////////////////////////////////////////////////////////////
 // Copyright (C) 2013-2022 CEA/DES
 //
 // This program is free software: you can redistribute it and/or modify
 // it under the terms of the GNU Lesser General Public License as published
 // by the Free Software Foundation, either version 3 of the License, or any
 // later version.
 //
 // This program is distributed in the hope that it will be useful,
 // but WITHOUT ANY WARRANTY; without even the implied warranty of
 // MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 // GNU Lesser General Public License for more details.
 //
 // You should have received a copy of the GNU Lesser General Public License
 // along with this program. If not, see <http://www.gnu.org/licenses/>.
 //////////////////////////////////////////////////////////////
 //////////////////////////////////////////////////////////////
 // $Id$
 // $Author$
 // $Date$
 // $Revision$
 // $State$
 //////////////////////////////////////////////////////////////*/
/*!
 \file   gpMod1.c
 \author Jean-Marc Martinez <jean-marc.martinez@cea.fr>
 \date   2012/07/19 15:38:00

 \brief
 */
#include <stdlib.h>
#include <math.h>
#include "linpack_d.h"
#ifdef WIN32
#include "libpgauss_Export.h"
#else
#define libpgauss_EXPORT
#endif

/* Calcul de hbeta, vbeta, gamma et de iL             */
/* Methode gpMod1                                     */
/* Input Kin : matrice de variance                    */
/* Input yin : vecteur des observations               */
/* Input n : dimension K[n x n], y[n]                 */
/* Output hbeta : estimation de beta                  */
/* Output vbeta : variance de hbeta                  */
/* Output gamma : vecteur de taille n              */
/* Output iL   : inverse de L                         */
libpgauss_EXPORT int gpMod1(double *Kin, double *yin, int n, double *hbeta, double *vbeta,
        double *gamma, double *iL)
{
    void gpDposl(double a[], int lda, int n, double b[]);
    int i, j, info;
    double *w, l2v, beta, determinant[2], *v;
    double *K, *pK, *y, *py;
    w = calloc(sizeof(double), n);
    v = calloc(sizeof(double), n);

    /* Recopie des donnees qui seront modifiees */
    K = calloc(sizeof(double), n * n);
    y = calloc(sizeof(double), n);
    for (pK = K, i = 0; i < n * n; i++)
        *pK++ = *Kin++;
    for (py = y, i = 0; i < n; i++)
        *py++ = *yin++;
    /*******************************/

    info = dpofa(K, n, n); /* decomposition Cholesky K = Lt L*/
    if (info == 0)
    {
        for (i = 0; i < n; i++)
            w[i] = y[i];
        gpDposl(K, n, n, w); /* Resolution y = Lt w puis y=w*/
        for (i = 0; i < n; i++)
            v[i] = 1.;
        gpDposl(K, n, n, v); /* Resolution 1 = Lt z */
        for (l2v = 0., i = 0; i < n; i++)
            l2v += v[i] * v[i]; /* carre de la norme L2 de v = z */
        for (beta = 0., i = 0; i < n; i++)
            beta += w[i] * v[i];
        *hbeta = beta / l2v;
        *vbeta = 1. / l2v;
        for (i = 0; i < n; i++)
            gamma[i] = y[i] - *hbeta;
        dposl(K, n, n, gamma); /* resolution lineaire => gamma = inv(K) * (y - hbeta)*/
        dtrdi(K, n, n, determinant, 11); /* inverse d'une triangulaire superieure */
        /* inverse de L */
        for (i = 0; i < n; i++)
        {
            for (j = 0; j < i; j++)
                iL[i + j * n] = 0.;
            for (j = i; j < n; j++)
                iL[i + j * n] = K[i + j * n];
        }
    }
    free(w);  w=NULL;
    free(v);  v=NULL;
    free(K);  K=NULL;
    free(y);  y=NULL;
    return info;
}

