/*//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////*/

#include <stdlib.h>
#include <math.h>
#include "linpack_d.h"
#ifdef WIN32
#include "libpgauss_Export.h"
#else
#define libpgauss_EXPORT
#endif

/*
	VERIFIE 25/06/2012
*/
/* Calcul de des predictions - avec tendance deterministe       */
/* Methode gpPredh                                      	*/
/* Input Hp : matrice new x p des regresseurs			*/
/* Input Kp : matrice new x n de covariances    		*/
/* Input variance : variance du processus			*/
/* Input hbeta : coefficients de la tendance			*/
/* Input iL : matrice n x n calcule par gpModh    	    	*/
/* Input gamma : vecteur de taille n calcule par gpModh         */
/* Input iR : matrice p x p calculee par gpModh    	    	*/
/* Input M  : matrice n x p calculee par gpModh         	*/
/* Input new : nombre de predictions                  		*/
/* Input n : nombre de points utilises pour la selection	*/
/* Input p : nombre de regresseurs				*/
/* Output yp : vecteur des new predictions			*/
/* Output vp : vecteur des new variances			*/
libpgauss_EXPORT void gpPredh(double *Hp, double *Kp, double variance, double *hbeta, double *gamma, double *iL, double *iR, double *M, int new, int n, int p, double *yp, double *vp) {
  int i, j, k, jn, jp;
  double vt1, vt2, vt3, *pgamma, *snew, *hMtsnew;

  snew    = calloc(sizeof(double),n);
  hMtsnew = calloc(sizeof(double),p);
  /* calcul des predictions */
  for(i=0;i<new;i++) {
    pgamma = gamma;
    for(vt1=0.,j=0;j<n;j++) vt1 += Kp[i+j*new] * *pgamma++;
    for(vt2=0.,j=0;j<p;j++) vt2 += hbeta[j] * Hp[i+j*new];
    yp[i] = vt2 + vt1;
  }
  /* calcul des variances */
  for(i=0;i<new;i++) {
    /* premiere partie : algo identitique a celle sans tendance deterministe */
    vp[i] = variance;
    for(j=0;j<n;j++) {
      jn = j * n;
      for(vt2=0.,k=0;k<=j;k++) vt2 += Kp[i+k*new] * iL[k+jn]; /* iL est triangulaire superieure */
      snew[j] = vt2; /* vecteur L^{-T} r */
      vp[i] -= vt2 * vt2;
    }
    /* seconde partie : Calcul de R(-1) [h - Mt snew] */
    for(j=0;j<p;j++) {
      hMtsnew[j] = Hp[i+j*new];
      jn = j * n;
      for(k=0;k<n;k++) hMtsnew[j] -= M[k+jn] * snew[k];
    }
    for(j=0;j<p;j++) {
      jp = j * p;
      for(vt3=0.,k=0;k<=j;k++) vt3 += iR[k+jp] * hMtsnew[k]; /* iR triangulaire superieure */
      vp[i] += vt3 * vt3;
    }
    if (vp[i] < 0.) vp[i] = 0.;
  }
  free(snew);    snew = NULL;
  free(hMtsnew); hMtsnew = NULL;
}
