//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TExponentialCorrFunction.cxx
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date   

 \brief Implementation of the class  URANIE::Modeler::TExponentialCorrFunction

 */
#include <iostream>

//// ROOT
#include "TMath.h"

//// URANIE
#include "TExponentialCorrFunction.h"

using std::cout;
using std::endl;

ClassImp (URANIE::Modeler::TExponentialCorrFunction)

namespace URANIE
{
namespace Modeler
{
// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------
TExponentialCorrFunction::TExponentialCorrFunction(Int_t nbDim, Double_t* p,
        Double_t* corrLength)
{
    nbDimensions  = nbDim;
    nbParameters  = nbDimensions * 2;
    nbCorrLengths = nbDimensions;
    parameters    = new Double_t[nbParameters];
    _lPower       = new Double_t[nbDimensions];
    _lCorrLengths = new Double_t[nbCorrLengths];

    Int_t offSet = nbParameters - nbCorrLengths;
    if(p)
    {
        for (Int_t i = 0; i < offSet; i++)
        {
            parameters[i] = p[i];
            _lPower[i] = p[i];
        }
    }
    else
    {
        for (Int_t i = 0; i < offSet; i++)
        {
            parameters[i] = 1.0;
            _lPower[i] = 1.0;
        }
    }
    if(corrLength)
    {
        for (Int_t i = offSet; i < nbParameters; i++)
        {
            parameters[i] = corrLength[i - offSet];
            _lCorrLengths[i - offSet] = corrLength[i - offSet];
        }
    }
    else
    {
        for (Int_t i = offSet; i < nbParameters; i++)
        {
            parameters[i] = 0.1;
            _lCorrLengths[i - offSet] = 0.1;
        }
    }
}

// -----------------------------------------------------------------------
//  Destructor
// -----------------------------------------------------------------------

TExponentialCorrFunction::~TExponentialCorrFunction()
{
    delete[] parameters;
    parameters = NULL;
    delete[] _lPower;
    _lPower = NULL;
    delete[] _lCorrLengths;
    _lCorrLengths = NULL;
    if (_blog)
        cout << "detruit TExponentialCorrFunction" << endl;
}

// -----------------------------------------------------------------------
//  Set methods
// -----------------------------------------------------------------------

void TExponentialCorrFunction::setParameters(Double_t* newParams)
{
    Int_t offSet = nbParameters / 2;
    for (Int_t i = 0; i < offSet; i++)
    {
        parameters[i] = newParams[i];
        _lPower[i] = newParams[i];
    }
    for (Int_t i = offSet; i < nbParameters; i++)
    {
        parameters[i] = newParams[i];
        _lCorrLengths[i - offSet] = newParams[i];
    }
}

void TExponentialCorrFunction::setPower(Double_t* p)
{
    for (Int_t i = 0; i < nbParameters / 2; i++)
    {
        parameters[i] = p[i];
        _lPower[i] = p[i];
    }
}

void TExponentialCorrFunction::setCorrLengths(Double_t *corrLength)
{
    Int_t offSet = nbParameters / 2;
    for (Int_t i = offSet; i < nbParameters; i++)
    {
        parameters[i] = corrLength[i - offSet];
        _lCorrLengths[i - offSet] = corrLength[i - offSet];
    }
}

// -----------------------------------------------------------------------
// Clone
// -----------------------------------------------------------------------

TCorrelationFunction* TExponentialCorrFunction::clone()
{
    TExponentialCorrFunction* cfClone = new TExponentialCorrFunction(nbDimensions, _lPower, _lCorrLengths);
    return cfClone;
}

// -----------------------------------------------------------------------
// Evaluation
// -----------------------------------------------------------------------

void TExponentialCorrFunction::eval(Int_t n1, Double_t* x1, Int_t n2, Double_t* x2, Double_t* C)
{
    if (_blog)
        std::cout << "** TExponentialCorrFunction::eval **: p[0]  = " << _lPower[0] << "; l[0] = " << _lCorrLengths[0] << std::endl;

    Int_t errorCode = gpCorExponential(x1, x2, n1, n2, nbDimensions, _lCorrLengths, _lPower, C);
    try
    {
        if (errorCode == 1)
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TExponentialCorrFunction::eval: evaluation failed");
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

}

// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------
void TExponentialCorrFunction::printLog(Option_t *option)
{
    cout << "*******************************" << endl;
    cout << " ** TExponentialCorrFunction::printLog[" << option << "]" << endl;
    cout << " ** Number of parameters = [" << nbParameters << "]" << endl;
    for (Int_t i = 0; i < nbParameters / 2; i++)
        cout << " ** power " << i << " = [" << parameters[i] << "]" << endl;
    cout << "*******************************" << endl;
    for (Int_t i = nbParameters / 2; i < nbParameters; i++)
        cout << " ** correlation length " << i << " = [" << parameters[i] << "]"
                << endl;
    cout << "*******************************" << endl;
}

} // Fin du namespace Modeler
} // Fin du namespace URANIE

