//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TGPLooCostFunction.cxx
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date   

 \brief Implementation of the class  URANIE::Modeler::TGPLooCostFunction

 */

//// ROOT
#include "TObjArray.h"

//// URANIE
#include "TGPLooCostFunction.h"
#include "TAttribute.h"

using std::cout;
using std::endl;
using namespace URANIE::DataServer;

ClassImp (URANIE::Modeler::TGPLooCostFunction)

namespace URANIE
{
namespace Modeler
{
// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------
TGPLooCostFunction::TGPLooCostFunction(TGPBuilder* gpb): TGPCostFunction(gpb)
{
    // Information transfer
    _BetaPrior = gpBuilder->BetaPrior;
    _QPrior    = gpBuilder->QPrior;
    
}

// -----------------------------------------------------------------------
//  Destructor
// -----------------------------------------------------------------------

TGPLooCostFunction::~TGPLooCostFunction()
{
    if (_blog)
        cout << "detruit TGPLooCostFunction" << endl;
}

URANIE::Relauncher::TEval *TGPLooCostFunction::rebuild(int rank, Bool_t chef)
{
    UNUSED(rank,chef);
    TGPLooCostFunction* ret = new TGPLooCostFunction(*this);
    ret->duplicateGPB();
    return (URANIE::Relauncher::TEval *) ret;
}
Bool_t TGPLooCostFunction::unbuild()
{
    delete gpBuilder;
    return kTRUE;
}

// -----------------------------------------------------------------------
//  Eval
// -----------------------------------------------------------------------
Int_t TGPLooCostFunction::eval(vector<URANIE::DataServer::UEntry*> *in, vector<URANIE::DataServer::UEntry*> *out, int ncalc)
{
    UNUSED(ncalc);
    double *inp = new double[in->size()];
	double *outp = new double[out->size()];
    for(unsigned int ival=0; ival<in->size(); ival++) inp[ival] = *in->at(ival)->val.dou;

    int admis=eval(inp, outp);
    
    for(unsigned int ival=0; ival<out->size(); ival++) *out->at(ival)->val.dou = outp[ival];
	
	delete [] inp; delete [] outp;
    return ((admis!=0) ? getOutDimension(out) : admis);
}

Int_t TGPLooCostFunction::eval(Double_t* theta, Double_t* res, int ncalc)
{
    UNUSED(ncalc);
    Int_t returnCode;
    Int_t nbCorrFuncParams;

    if (gpBuilder->hasCorrFunction())
    {
        // Update the parameters of the correlation function            
        nbCorrFuncParams = gpBuilder->getCorrFunction()->getNParameters();
        gpBuilder->getCorrFunction()->setParameters(theta);
    }
    else
        nbCorrFuncParams = 0;
    
    if (gpBuilder->hasMeasurementError())
    {
        // In this case, the array theta contains nbCorrFuncParams+1 elements        
        if(gpBuilder->hasVarianceInFactor())
            // The last element is the parameter alpha
            gpBuilder->setAlpha(theta[nbCorrFuncParams]);
        else
            // The last element is the variance of the GP
            gpBuilder->setVariance(theta[nbCorrFuncParams]);
    }

    // Update the covariance matrix
    gpBuilder->computeCovarianceMatrix();

    // Compute the cost
    if (gpBuilder->hasTrend())
    {
        if (gpBuilder->usePrior())
            returnCode = gpLoohBayes(_BetaPrior, _QPrior, _H, _K, _yObs, _n, _p, res);
        else
            returnCode = gpLooh(_H, _K, _yObs, _n, _p, res);
    }
    else
        returnCode = gpLoo0(_K, _yObs, _n, res);

    if (returnCode != 0)
    {
        res[0] = 1e99;
        res[1] = 1e99;
    }

    return 1;
}

} // Fin du namespace Modeler
} // Fin du namespace URANIE

