//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TKernel.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2012-11-26 21:50:54
 
 \brief Interface de la classe URANIE::Modeler::TKernel
 
 */
/**
  * @class URANIE::Modeler::TKernel
  *
  * @brief Description of the class TKernel
  *
 */
#ifndef TKERNEL_H
#define TKERNEL_H

/// ROOT
#include "TMatrixDSymfwd.h"

/// Uranie
#include "TModeler.h"

namespace URANIE
{
namespace Modeler
{
class TKernel: public TModeler
{
public:
    enum EKernelType
    {
        kUnknown,
        kLinear,
        kPoly,
        kSigmoid,
        kRBF,
        kLaplace,
        kSpline1,
        kBernoulli1
    };
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    /*!
      *\brief constructor with the TDataServer and the architecture
      *
      * \param tds (URANIE::DataServer::TDataServer *) the dataserver
      * \param architecture (TString) The architecture (input and output attributes) of the model separated by the caracter ":"
      * \param option (Option_t *)[""]  The option to pass
     */
    TKernel(URANIE::DataServer::TDataServer *tds, TString architecture, Option_t *option = "");

    /*!
      * \brief Default constructor with input and output attributes
      *
      * \param tds (URANIE::DataServer::TDataServer *) the dataserver
      * \param varexpinput(const char*)[""] The list of input attributes to pass to the function separated by the caracter ":"
      * \param varexpoutput (const char *)[""]  The list of output attributes separated by the caracter ":"
      * \param option (Option_t *)[""]  The option to pass
     */
    TKernel(URANIE::DataServer::TDataServer *tds, const char *varexpinput,
            const char *varexpoutput, Option_t * option = "");
    //! Default destructor
    virtual ~TKernel();
    //@}

    //---------------------------------------------
    /** @name Get and Set methods
     **/
    //@{
    //! Set the type of kernel without parameter
    /*!
     \param nKernel (EKernelType) The type of kernel
     */
    void setKernel(EKernelType nKernel);
    /*!
     \param nKernel (EKernelType) The type of kernel
     \param dparam (Double_t) the First and only one parameter
     */
    void setKernel(EKernelType nKernel, Double_t dparam1);
    //! Set the type of kernel with two parameters
    /*!
     \param nKernel (EKernelType) The type of kernel
     \param dparam1 (Double_t) First parameter
     \param dparam2 (Double_t) Second parameter
     */
    void setKernel(EKernelType nKernel, Double_t dparam1, Double_t dparam2);
    //! Set the type of kernel with three parameters
    /*!
     \param nKernel (EKernelType) The type of kernel
     \param dparam1 (Double_t) First parameter
     \param dparam2 (Double_t) Second parameter
     \param dparam3 (Double_t) Third parameter
     */
    void setKernel(EKernelType nKernel, Double_t dparam1, Double_t dparam2,
            Double_t dparam3);
    //! Return the type of Y distribution
    EKernelType getKernel()
    {
        return _nKernelType;
    }
    //! Return First Parameter
    Double_t getParameter1()
    {
        return _dParameter1;
    }
    //! Return Second Parameter
    Double_t getParameter2()
    {
        return _dParameter2;
    }
    //! Set the Lambda parameter
    void setLambda(Double_t dlambda)
    {
        if (dlambda > 0)
            _dLambda = dlambda;
    }
    //! Get the Lambda
    Double_t getLambda()
    {
        return _dLambda;
    }
    //! Get the matrix _matK
    TMatrixDSym getMatrixK() { return _matK;}
    //@}

    //---------------------------------------------
    /** @name Estimate
     **/
    //@{
    /*!
     * \brief estimate
     *
     * \param option
     */
    void estimate(Option_t * option = "");

    /*!
     * \brief Fill the matrix of Inputs attributes and the vector of output attribute (nY=1)
     *
     * \todo To deplace in the TModeler Class because it is the same for TLinearRegression.
     */
    void fillData();

    /*!
     * \brief compute the matrix K
     *
     * \param option
     */
    void computeK(Option_t * option = "");
    //@}

    //---------------------------------------------
    /** @name Export the model in extrenal langage
     **/
    //@{
    //! Export the model in C++ langage in a file
    /*!
     \warning Don't use this method. Use the main method TModeler::exportFunction.
     \param sourcefile (std::ofstream *) the pointer to the file to export the modeler.
     */
    void exportModelCplusplus(std::ofstream * sourcefile) const;
    //! Export the model in Fortran langage in a file
    /*!
     \warning Don't use this method. Use the main method TModeler::exportFunction.
     \param sourcefile (std::ofstream *) the pointer to the file to export the modeler.
     */
    void exportModelFortran(std::ofstream * sourcefile) const;
    
    /*!
      * \brief Export the model in a PMML file (not yet implemented)
      *
      * \param sourcefile (std::ofstream *) the pointer to the file to export the modeler.
      *
      * \warning Don't use this method. Use the main method TModeler::exportFunction.
     */
    void exportModelPMML(const char* file = "", 
                         const char* name = "",Option_t *option = "") const
    {
        TString soption = option;
        soption.ToLower();
	UNUSED(file,name);
        std::cerr
                << "   --- Method TKernel::exportModelPMML(std::ofstream * sourcefile) not yet implemented"
                << endl;
    };
    
    //! Export the model in Python langage in a file (not yet implemented)
    /*!
     \warning Don't use this method. Use the main method TModeler::exportFunction.
     \param sourcefile (std::ofstream *) the pointer to the file to export the modeler.
     */
    void exportModelPython(std::ofstream * sourcefile) const
    {
	UNUSED(sourcefile);
        std::cerr
                << "   --- Method TKernel::exportModelPython(std::ofstream * sourcefile) not yet implemented"
                << endl;
    };
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    void printLog(Option_t *option = "");
    //@}

    //---------------------------------------------
    /** @name Private and Public attributes
     *
     **/
    //@{
private:
    EKernelType _nKernelType;        ///< The type of Kernel
    Double_t _dParameter1;           ///< The First Parameter    (gamma)   [1.0]
    Double_t _dParameter2;           ///< The Second Parameter   (coeff)   [0.0]
    Double_t _dParameter3;           ///< The Third Parameter    (degree)  [1.0]
    Int_t _nParameter;               ///< The number of parameter

    TMatrixD _matX;                 ///< The matrix of inputs  attributes
    TVectorD _vecY;                 ///< The vector of Y (outputs)

    TMatrixDSym _matK;              ///< The matrix of K (inputs)
    Double_t _dLambda;               ///< The lambda parameter (default 1.e-6)
    //@}

    ClassDef(URANIE::Modeler::TKernel, ID_MODELER)
};
} // End Of namespace Modeler
} // End Of namespace URANIE
#endif

