//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TMatern52CorrFunction.cxx
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date   

 \brief Implementation of the class  URANIE::Modeler::TMatern52CorrFunction

 */
#include <iostream>

//// ROOT
#include "TMath.h"

//// URANIE
#include "TMatern52CorrFunction.h"

using std::cout;
using std::endl;

ClassImp (URANIE::Modeler::TMatern52CorrFunction)

namespace URANIE
{
namespace Modeler
{
// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------
TMatern52CorrFunction::TMatern52CorrFunction(Int_t nbDim, Double_t* corrLength)
{
    nbDimensions  = nbDim;
    nbParameters  = nbDimensions;
    nbCorrLengths = nbDimensions;
    _lCorrLengths = new Double_t[nbCorrLengths];

    if(corrLength)
    {
        for (Int_t i = 0; i < nbParameters; i++)
            _lCorrLengths[i] = corrLength[i];
    }
    else
    {
        for (Int_t i = 0; i < nbParameters; i++)
            _lCorrLengths[i] = 0.1;
    }   
    parameters = _lCorrLengths;
}

// -----------------------------------------------------------------------
//  Destructor
// -----------------------------------------------------------------------

TMatern52CorrFunction::~TMatern52CorrFunction()
{
    delete[] _lCorrLengths;
    _lCorrLengths = NULL;
    if (_blog)
        cout << "detruit TMatern52CorrFunction" << endl;
}

// -----------------------------------------------------------------------
//  Set methods
// -----------------------------------------------------------------------

void TMatern52CorrFunction::setParameters(Double_t* newParams)
{
    for (Int_t i = 0; i < nbParameters; i++)
        _lCorrLengths[i] = newParams[i];
}

void TMatern52CorrFunction::setCorrLengths(Double_t *corrLength)
{
    for (Int_t i = 0; i < nbParameters; i++)
        _lCorrLengths[i] = corrLength[i];
}

// -----------------------------------------------------------------------
// Clone
// -----------------------------------------------------------------------

TCorrelationFunction* TMatern52CorrFunction::clone()
{
    TMatern52CorrFunction* cfClone = new TMatern52CorrFunction(nbDimensions, _lCorrLengths);
    return cfClone;
}

// -----------------------------------------------------------------------
// Evaluation
// -----------------------------------------------------------------------

void TMatern52CorrFunction::eval(Int_t n1, Double_t* x1, Int_t n2, Double_t* x2, Double_t* C)
{
    if (_blog)
        std::cout << "** TMatern52CorrFunction::eval **: " << "l[0] = " << _lCorrLengths[0] << std::endl;

    Int_t errorCode = gpCorMatern52(x1, x2, n1, n2, nbDimensions, _lCorrLengths, C);

    try
    {
        if (errorCode == 1)
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TMatern52CorrFunction::eval: bessik function failed to converge.");
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------
void TMatern52CorrFunction::printLog(Option_t *option)
{
    cout << "*******************************" << endl;
    cout << " ** TMatern52CorrFunction::printLog[" << option << "]" << endl;
    cout << " ** Number of parameters = [" << nbParameters << "]" << endl;
    for (Int_t i = 1; i < nbParameters; i++)
        cout << " ** correlation length " << i << " = [" << parameters[i] << "]"
                << endl;
    cout << "*******************************" << endl;
}

} // Fin du namespace Modeler
} // Fin du namespace URANIE

