//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TPolynomialChaos.cxx
 \author Maud Salvoldi (OCC)
 \date   23 March 2011
 
 \brief Implementation of the class URANIE::Modeler::TPolynomialChaos

 
 */
#include <sstream>
#include <string>

// ROOT
#include "TMath.h"

// URANIE
#include "TPolynomialChaos.h"
#include "TNisp.h"
#include "UTimer.h"
#include "UExceptions.h"
#include "TLogNormalDistribution.h"
#include "TExponentialDistribution.h"

// NISP
#include "nisp_pc.h"
#include "nisp_util.h"

#define HERMITE  1
#define LEGENDRE 2
#define LAGUERRE 3
#define JACOBI   4

ClassImp (URANIE::Modeler::TPolynomialChaos)

namespace URANIE
{
namespace Modeler
{

  //default constructor from a set of stochastic variables
  TPolynomialChaos::TPolynomialChaos(TDataServer *tds, TNisp *nisp,
				     TString soutput) :
    TNamed("TPolynomialChaos", "Chaos Polynomial functionalities"),
    _fLogger(NULL), _nisp(nisp), _degree(1), _pc(NULL), _nx(0), _np(0), _ny(0),  
    _bStoreYHat(kTRUE), _degreeResults(NULL), _autoDegreeFactor(1.5),  
    _minAutoDeg(-1), _maxAutoDeg(-1), _bestAutoDeg(-1), _verrloo(NULL),
    _listAttOut(NULL), _blog(kFALSE)
  {
    try
      {
        if (tds == NULL)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     "TPolynomialChaos::constructor - Error: the pointer of \"tds\" is empty");
        if (nisp == NULL)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     "TPolynomialChaos::constructor - Error: the pointer of \"nisp\" is empty!");
        if (tds != nisp->_tds)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     "TPolynomialChaos::constructor - Error: The tds and the nisp->_tds are not the same!");
        _pc = new PolynomialChaos(nisp->_gvx);        
        _nx = nisp->_gvx->nx;
        _listAttOut = new TDataSpecification();
        if (_blog)
	  cout << "liste des attibuts: " << soutput << endl;
        //List of Output Attributs
        if (soutput != "")
	  {
            if (_blog)
	      cout
		<< " ** TPolynomialChaos::constructor a list of output attributes:"
		<< soutput.Data() << endl;
            TList * listOfAttributes = tds->getListOfAttributes(soutput.Data());
            TAttribute *att = NULL;
            TIter next(listOfAttributes);
      
            while ((att = (TAttribute *) next()))
	      _listAttOut->addAttribute(att);
	    delete listOfAttributes;
	  }
        else
	  {
            if (_blog)
	      cout << " ** TPolynomialChaos::constructor: all the attributs"
		   << endl;
            TList * listOfAttributes = tds->getListOfAttributes("*");
            TAttribute *att = NULL;
            TIter next(listOfAttributes);
            while ((att = (TAttribute *) next()))
	      {
                if (att->isOutput())
		  _listAttOut->addAttribute(att);
	      } //while
	    delete listOfAttributes;
	  } // end else
        _ny = _listAttOut->getNAttributes();
        _np = (Int_t)(tds->getTuple()->GetEntries());
        if (_np != _nisp->getNPatterns())
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     "TPolynomialChaos::constructor - Error: number of simulation are not the same in the tds object and in nisp object");
        //Nisp recupere les realisations du plan d'experience
        getTarget();
        
        _pc->SetDegree(_degree);
        for (Int_t jny = 1; jny <= _pc->ny; jny++)
	  {
            for (Int_t i = 0; i <= _pc->p; i++)
	      {
                _pc->beta[jny][i] = 0;
	      }
	  }
        //print the log
        if (_blog)
	  _listAttOut->printLog();

      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
        if (_listAttOut != NULL)
	  {
            delete _listAttOut;
            _listAttOut = NULL;
	  }

        if (_pc != NULL)
	  {
            delete _pc;
            _pc = NULL;
	  }

        throw ue;
      }

  }

  //constructor from a TPolynomialChaos object
  TPolynomialChaos::TPolynomialChaos(const TPolynomialChaos & TPC) :
    TNamed("TPolynomialChaos", "Chaos Polynomial functionalities"),
    _fLogger(NULL)
  {
    // public:
    _nisp = TPC._nisp;
    _degree = TPC._degree;
    _nx = TPC._nx;
    _autoDegreeFactor = TPC._autoDegreeFactor;
    _minAutoDeg = TPC._minAutoDeg;
    _maxAutoDeg = TPC._maxAutoDeg;
    _bestAutoDeg = TPC._bestAutoDeg;
    _degreeResults = TPC._degreeResults;
    _verrloo = TPC._verrloo;
    //       int* var_sto = new int[_nx];
    //       cout<<"TPC._pc->nx = "<<TPC._pc->nx<<"  ,  _nx = "<<_nx<<endl;
    //       _pc = new PolynomialChaos( TPC._pc, var_sto, _nx);
    SetRandomVariable* gpx = new SetRandomVariable();
    for (Int_t i = 0; i < _nx; i++)
      {
        
        if (TPC._pc->typol[i] == "Hermite")
	  gpx->AddRandomVariable(new RandomVariable("Normale"));
        else if (TPC._pc->typol[i] == "Legendre")
	  gpx->AddRandomVariable(new RandomVariable("Uniforme"));
        else if (TPC._pc->typol[i] == "Laguerre")
	  gpx->AddRandomVariable(new RandomVariable("Exponentielle"));
        else
	  {
	    try
	      {
		throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							   "TPolynomialChaos::TPolynomialChaos copy constructor - ERROR: law  is not available");
            
	      }
	    catch (URANIE::Exceptions::UErrorExceptions& ue)
	      {
		ue.printMessage();
             
		//Clean///////////
		if(gpx)
		  {
		    delete gpx;
		    gpx = NULL;
		  }
		//////////////////
		throw ue;
		exit(-1);
	      }
	  }
      }
    _pc = new PolynomialChaos(gpx);
    _np = TPC._np;
    _ny = TPC._ny;
    // private:
    _listAttOut = new TDataSpecification(TPC._listAttOut);
    _bStoreYHat = TPC._bStoreYHat;
    _blog = TPC._blog;
    getTarget();

    _pc->SetDegree(_degree);
    for (Int_t jny = 1; jny <= TPC._pc->ny; jny++)
      {
        for (Int_t i = 0; i <= _pc->p; i++)
	  {
            _pc->beta[jny][i] = TPC._pc->beta[jny][i];
	  }
      }
    _pc->SetAnova();

    
    
    
    //Clean///////////
    if(gpx)
      {
        delete gpx;
        gpx = NULL;
      }
    //////////////////
  }

  //destructor
  TPolynomialChaos::~TPolynomialChaos()
  {
    _nx = 0;
    _ny = 0;
    _np = 0;
    _degree = 0;

    if (_listAttOut != NULL)
      {
        delete _listAttOut;
        _listAttOut = NULL;
      }

    if (_pc != NULL)
      {
        delete _pc;
        _pc = NULL;
      }

    if(  _degreeResults )
      delete _degreeResults;

    if( _verrloo )
      delete _verrloo;

    if( _fLogger!=NULL)
      delete _fLogger;
    _fLogger=NULL;
    
    _blog = kFALSE;
  }
  //Set the degree
  void TPolynomialChaos::setDegree(Int_t degree)
  {
    try
      {
        if (degree < 0)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     "TPolynomialChaos::setDegree - Error: invalid value for degree (Need to be >0)");
        _degree = degree;
        _pc->SetDegree(degree);
      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
        throw ue;
        exit(-1);
      }
  }
  //
  void TPolynomialChaos::getTarget()
  {
    _pc->SetDimensionOutput(_ny);
    _pc->SetSizeTarget(_np);
    TDataServer * tds = _nisp->_tds;
    Int_t jcrt = 0;
    while (jcrt < _ny)
      {
        TLeaf *loutput = tds->getTuple()->GetLeaf(
						  _listAttOut->getAttribute(jcrt)->GetName());
        for (Int_t k = 0; k < _np; k++)
	  {
            tds->getTuple()->GetEntry(k);
            _pc->SetTarget(k + 1, jcrt + 1, loutput->GetValue());
	  } //end  for
        jcrt++;
      }
  }
  //Computation of output in relation with input : input[0...nx-1]
  void TPolynomialChaos::computeOutput(double *input)
  {
    Double_t * inputNisp = new Double_t[_nx+1];
    Int_t i;
    //les tableaux dans Nisp commence par 1 donc on transfome le tableau uranie en tableau nisp
    for (i = 0; i < _nx; i++)
      inputNisp[i + 1] = input[i];
    _pc->ComputeOutput(inputNisp);
    
    // Clean
    delete[] inputNisp;
  }
  //Get an output
  Double_t TPolynomialChaos::getOutput(TString jname)
  {
    Double_t val;
    Int_t j;
    try
      {
        if (jname != "")
	  {
            if (_listAttOut->isAttribute(jname))
	      {
                j = _listAttOut->getAttributeIndex(jname);
                val = getOutput(j);
	      }
            else
	      {
                throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							     "TPolynomialChaos::getOuput - WARNING: this attribut is unknown");
	      }
	  }
        else
	  val = getOutput(0);
      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }
    return val;
  }
  //Get an output
  Double_t TPolynomialChaos::getOutput(Int_t j)
  {
    Double_t val;
    try
      {
        if (j + 1 < 1)
	  {
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 "TPolynomialChaos::getOutput - WARNING: rank of the output <1");
	  }
        if (j + 1 > _ny)
	  {
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 "TPolynomialChaos::getOutput - WARNING: rank of the output exceed");
	  }
        val = _pc->GetOutput(j + 1);
      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }
    return val;
  }
  
  //! Get the Mean squared uncertainty for the output with rank
  Double_t TPolynomialChaos::getEqmLoo(Int_t rank)
  {
    try
      {
        if (rank+1 < 1)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     "TPolynomialChaos::getEqmLoo - Error: invalid value for rank (Need to be >=0)");

	if (rank + 1 > _ny)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form("TPolynomialChaos::getEqmLoo - ERROR: rank of the output should be < %d",
							  _ny));
      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
        throw ue;
        exit(-1);
      }
    return _pc->GetEqmLoo( rank + 1 );
  }

  //! Get the Leave-One-Out uncertainty for the input (input) and the output (rank)
  Double_t TPolynomialChaos::getErrLoo(Int_t input, Int_t rank)
  {
    try
      {
	if (rank+1 < 1)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     "TPolynomialChaos::getErrLoo - Error: invalid value for rank (Need to be >=0)");
	if (rank + 1 > _ny)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form("TPolynomialChaos::getErrLoo - ERROR: rank of the output should be < %d",
							  _ny));
	if (input+1 < 1)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     "TPolynomialChaos::getErrLoo - Error: invalid value for the index of the input (Need to be >=0)");

	if (input+1 > _nisp->getNPatterns() )
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form("TPolynomialChaos::getErrLoo - Error: invalid value for the index of the input (Need to be <%d)",
							  _nisp->getNPatterns()));

      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
        throw ue;
        exit(-1);
      }
    return _pc->GetErrLoo( input+1, rank + 1 );
  }
  
  //Computation of coefficients by method type
  void TPolynomialChaos::computeChaosExpansion(TString type, Option_t *option)
  {
    // Treat the option
    TString soption = TString(option);
    soption.ToLower();
    if (soption.Contains("nostoreyhat"))
      _bStoreYHat = kFALSE;
    int npat=_nisp->_tds->getNPatterns();

    if ( soption.Contains("autodegree") || _minAutoDeg != -1 )
      {
	try
	  {
	    if(!type.EqualTo("Regression"))
	      throw URANIE::Exceptions::UWarningExceptions(__FILE__,__LINE__,
							   "TPolynomialChaos::computeChaosExpension  WARNING: Cannot use degree auto-determination with Integration");
	    
	    if( _ny != 1 )
	      throw URANIE::Exceptions::UWarningExceptions(__FILE__,__LINE__,
							   "TPolynomialChaos::computeChaosExpension  WARNING: Cannot use degree auto-determination with more than one output");
	    
	    automatisedDegree(option);
	    
	  }catch(URANIE::Exceptions::UWarningExceptions &ue)
	  {
	    ue.printMessage();
	  }
      }
    try
      {
        if (type != "Integration" && type != "Regression")
	  throw URANIE::Exceptions::UErrorExceptions(
						     __FILE__,
						     __LINE__,
						     Form("TPolynomialChaos::computeChaosexpension\n Invalid name [%s] of the method ( Integration | Regression )",
							  type.Data ()));
	_pc->ComputeChaosExpansion(_nisp->_gvx, (string) type);

	if ( _bStoreYHat) {
	  // Build Yhat for each Output attributes
	  Int_t jcrt = 0;
	  TString syhat = "";
	  TString sListOfYHat = "";
	  for( jcrt = 0; jcrt < _ny; jcrt++) {
	    syhat = Form("%shat", _listAttOut->getAttribute(jcrt)->GetName());
	    sListOfYHat += Form("%s/D", syhat.Data ());
	    if ( jcrt < _ny-1 )
	      sListOfYHat +=  ":";
	    if ( _blog )
	      cout << " ** Output(" << jcrt << "/" << _ny << ") syhat[" << syhat << "] sListOfYHat["
		   << sListOfYHat << "]" << endl;
	    _nisp->_tds->addAttributeByNameTitle(syhat, Form("#hat{%s}", _listAttOut->getAttribute(jcrt)->GetTitle()));
	  }
	  Double_t *tabValue = new Double_t[_ny];
	  TBranch *tbr = _nisp->_tds->getTuple()->Branch("TPC", tabValue, sListOfYHat);

	  // Loop on the DoE
	  for(Int_t ipat=0; ipat<npat; ipat++) {
	    _pc->ComputeOutput( _nisp->_gvx->x[ipat+1] );

	    for( jcrt = 0; jcrt < _ny; jcrt++)
	      tabValue[jcrt] = _pc->GetOutput (jcrt+1);
	    tbr->Fill();
	  }
	  // resest branch address                 
	  _nisp->_tds->getTuple()->ResetBranchAddress(tbr);
	  // Clean
	  delete[] tabValue;
	}
      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }

  }

  void TPolynomialChaos::automatisedDegree(Option_t *option)
  {
    TString soption = option;
    soption.ToLower();

    Bool_t bdrawProgressBar=kTRUE;
    if (soption.Contains("noprogressbar"))
        bdrawProgressBar = kFALSE;

    unsigned int maxDeg=1;
    if ( _minAutoDeg == -1 )
      {
	//Getting the highest degree possible to be tested
	double limit=_np / _autoDegreeFactor;
	for(maxDeg=1; maxDeg<100; maxDeg++)
	  {
	    double NbCoeff = factorial(_nx + maxDeg) / (factorial(maxDeg) * factorial(_nx));
	    if( NbCoeff > limit )
	      {
		maxDeg--;
		break;
	      }
	  }
      }  
    
    //Cleaning and redoing a TDSNtuple to store the results
    if(  _degreeResults )
      delete _degreeResults;

    //Construction by hand of the tree that will contains results
    _degreeResults = new TTree( Form("%s_%s",GetName(),"DegreeOptimisation"), "title ?");
    _degval=0;
    _degreeResults->Branch("degree",&_degval, "degree/I");
    _eqmval=0;
    _degreeResults->Branch("eqmloo",&_eqmval, "eqmloo/D");
    _verrloo = new vector<double>(_np,0);
    _degreeResults->Branch("errloo", "vector<double>", &_verrloo, 320000);

    int length = ((_maxAutoDeg==-1) ? maxDeg : _maxAutoDeg) - ((_minAutoDeg==-1) ? 1 : _minAutoDeg) +1;
    URANIE::DataServer::UTimer timer(length);//, "Uranie TPolynomialChoas::AutomatisedRegression");    
    double minEqmVal=1e15;
    for(unsigned int ideg=((_minAutoDeg==-1) ? 1 : _minAutoDeg); ideg<=((_maxAutoDeg==-1) ? maxDeg : _maxAutoDeg); ideg++)
      {

#ifndef WIN32
        if (bdrawProgressBar)
            timer.DrawProgressBar(ideg);
#endif
	_pc->SetDegree( ideg );
	for (Int_t jny = 1; jny <= _pc->ny; jny++)
	  {
            for (Int_t i = 0; i <= _pc->p; i++)
	      {
                _pc->beta[jny][i] = 0;
	      }
	  }
	
    	_pc->ComputeChaosExpansion(_nisp->_gvx, "Regression" );

	_degval=ideg;
	_eqmval=getEqmLoo();
	if( _eqmval < minEqmVal )
	  {
	    minEqmVal = _eqmval;
	    _bestAutoDeg = _degval;
	  }

	for(int i=0; i<_np; i++)
	  _verrloo->at(i) = getErrLoo(i);
	
	_degreeResults->Fill();

	if( _pc != NULL)
	  {
	    delete _pc;
	   _pc = new PolynomialChaos(_nisp->_gvx);
	   getTarget();
	  }
      }

    _pc->SetDegree( _bestAutoDeg );
    for (Int_t jny = 1; jny <= _pc->ny; jny++)
      {
	for (Int_t i = 0; i <= _pc->p; i++)
	  {
	    _pc->beta[jny][i] = 0;
	  }
      }
    
#ifndef WIN32
    if (bdrawProgressBar)
        timer.DrawProgressBar( length  );
#endif
    _fLogger = new URANIE::DataServer::UMessageLogger("TPolynomialChaos");
    (*_fLogger) << Endl<< "TPolynomialChaos:: Auto-determination of the degree done " << Endl;
    (*_fLogger) << "                The best estimated one is "<<_bestAutoDeg<<Endl;
    
  }

  double TPolynomialChaos::factorial(int n)
  {
    return ((n==1||n==0) ? 1 : n*factorial(n-1));
  }

  void TPolynomialChaos::setAutoDegreeFactor(double autodeg)
  {
    try
      {
	if( autodeg < 1)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     "TPolynomialChaos::setAutoDegreeFactor - ERROR: This factor should be >= 1. ");

	
        _autoDegreeFactor = autodeg;
      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
	throw ue;
      }
  }

  void TPolynomialChaos::setAutoDegreeBoundaries(int amin, int amax)
  {
    try
      {
	if( amin < 1)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form("TPolynomialChaos::setAutoDegreeBoundaries - ERROR: Min value should be >=1 (%d give)",amin));

	if( amax != -1 && amin > amax)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form("TPolynomialChaos::setAutoDegreeBoundaries - ERROR: When setting Max value (%d), it should be greater than a Min one (%d)",amax,amin));

	double limit=_np / _autoDegreeFactor;
	for(_maxAutoDeg=1; _maxAutoDeg<100; _maxAutoDeg++)
	  {
	    double NbCoeff = factorial(_nx + _maxAutoDeg) / (factorial(_maxAutoDeg) * factorial(_nx));
	    if( NbCoeff > limit )
	      {
		_maxAutoDeg--;
		break;
	      }
	  }
	
	if( amax ==-1 && amin > _maxAutoDeg )
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form("TPolynomialChaos::setAutoDegreeBoundaries - ERROR: Given the number of patterns (%d), the autoDegreeFactor(%f) and the min value (%d), the maximumu autorised degree (%d) is lower than the requested min one.",_np, _autoDegreeFactor, amin, _maxAutoDeg));
	
	if( amax > _maxAutoDeg )
	  throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
						       Form("TPolynomialChaos::setAutoDegreeBoundaries - WARNING: Given the number of patterns (%d) and the autoDegreeFactor(%f) the maximumu autorised degree (%d) is lower than the requested one (%d)... Max is set to %d",_np, _autoDegreeFactor, _maxAutoDeg, amax, _maxAutoDeg));
 
	  
	_maxAutoDeg=amax;
	
      }catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
        ue.printMessage();
      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
        throw(ue);
      }
        
    _minAutoDeg=amin;
  }
  
  // Get the mean of attribut given by index
  Double_t TPolynomialChaos::getMean(Int_t j)
  {
    Double_t val;
    try
      {
        if (j + 1 < 1)
	  {
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 "TPolynomialChaos::getMean - ERROR: rank of the output < 1");
	  }
        if (j + 1 > _ny)
	  {
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 "TPolynomialChaos::getMean - ERROR: rank of the output exceed");
	  }
        val = _pc->GetMean(j + 1);
      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }
    return val;
  }
  // Get the meanof attribut give by name
  Double_t TPolynomialChaos::getMean(TString name)
  {
    Double_t val;
    Int_t j;
    try
      {
        if (name == "")
	  val = getMean(0);
        else
	  {
            if (_listAttOut->isAttribute(name))
	      {
                j = _listAttOut->getAttributeIndex(name);
                val = getMean(j);
	      }
            else
	      {
                throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							     "TPolynomialChaos::getMean - Error: the attribut name is unknown");
	      }
	  }
      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }
    return val;
  }
  //Get the variance of attribut give by name
  Double_t TPolynomialChaos::getVariance(TString name)
  {
    Double_t val;
    Int_t j;
    try
      {
        if (name == "")
	  val = getVariance(0);
        else
	  {
            if (_listAttOut->isAttribute(name))
	      {
                j = _listAttOut->getAttributeIndex(name);
                val = getVariance(j);
	      }
            else
	      {
                throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							     "TPolynomialChaos::getVariance - Error: the attribut name is unknown");
	      }
	  }
      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }
    return val;
  }
  //Get the variance of attribut give by index
  Double_t TPolynomialChaos::getVariance(Int_t j)
  {
    Double_t val;
    try
      {
        if (j + 1 < 1)
	  {
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 "TPolynomialChaos::getVariance - ERROR: rank of the output < 1");
	  }
        if (j + 1 > _ny)
	  {
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 "TPolynomialChaos::getVariance - ERROR: rank of the output exceed");
	  }
        val = _pc->GetVariance(j + 1);
      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }
    return val;
  }
  //Get the Covariance by name of the attributes
  Double_t TPolynomialChaos::getCovariance(TString xname, TString yname)
  {
    Double_t val = -1.0;
    Int_t i, j;
    try
      {
        if (_listAttOut->isAttribute(yname) && _listAttOut->isAttribute(xname))
	  {
            j = _listAttOut->getAttributeIndex(yname);
            i = _listAttOut->getAttributeIndex(xname);
            val = getCovariance(i, j);
	  }
        else
	  {
            if (!_listAttOut->isAttribute(yname))
	      throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							   "TPolynomialChaos::getCovariance - Error: this output is unknown");
            if (!_listAttOut->isAttribute(xname))
	      throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							   "TPolynomialChaos::getCovariance - Error: this output is unknown ");

	  }
      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }
    return val;
  }
  //Get the Covariance by index of the attributes
  Double_t TPolynomialChaos::getCovariance(Int_t i, Int_t j)
  {
    Double_t val;
    try
      {
        if (j + 1 < 1)
	  {
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 "TPolynomialChaos::getCovariance - ERROR: rank of the output < 1");
	  }
        if (j + 1 > _ny)
	  {
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 "TPolynomialChaos::getCovariance - ERROR: rank of the output exceed");
	  }
        if (i + 1 < 1)
	  {
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 "TPolynomialChaos::getCovariance - ERROR: rank of the output < 1");
	  }
        if (i + 1 > _ny)
	  {
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 "TPolynomialChaos::getCovariance - ERROR: rank of the output exceed");
	  }
        val = _pc->GetCovariance(i + 1, j + 1);
      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }
    return val;
  }
  //Get the correlation by name of the attributes
  Double_t TPolynomialChaos::getCorrelation(TString xname, TString yname)
  {
    Double_t val = -1.0;
    Int_t i, j;
    try
      {
        if (_listAttOut->isAttribute(yname) && _listAttOut->isAttribute(xname))
	  {
            j = _listAttOut->getAttributeIndex(yname);
            i = _listAttOut->getAttributeIndex(xname);
            val = getCorrelation(i, j);
	  }
        else
	  {
            if (!_listAttOut->isAttribute(yname))
	      throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							   "TPolynomialChaos::getCorrelation - Error: the name of the output is unknown");
            if (!_listAttOut->isAttribute(xname))
	      throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							   "TPolynomialChaos::getCorrelation - Error: the name of the output is unknown");
	  }
      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }
    return val;
  }
  //Get the correlation by index
  Double_t TPolynomialChaos::getCorrelation(Int_t i, Int_t j)
  {
    Double_t val;
    try
      {
        if (j + 1 < 1)
	  {
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 "TPolynomialChaos::getCorrelation - ERROR: rank of the output < 1");
	  }
        if (j + 1 > _ny)
	  {
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 "TPolynomialChaos::getCorrelation - ERROR: rank of the output exceed");
	  }
        if (i + 1 < 1)
	  {
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 "TPolynomialChaos::getCorrelation - ERROR: rank of the output <1");
	  }
        if (i + 1 > _ny)
	  {
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 "TPolynomialChaos::getCorrelation - ERROR: rank of the output exceed ");
	  }
        val = _pc->GetCorrelation(i + 1, j + 1);
      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }
    return val;
  }
  //First indice of sensitivity  by attributes name
  Double_t TPolynomialChaos::getIndexFirstOrder(TString xname, TString yname)
  {
    Double_t val = -1.0;
    Int_t i, j;
    try
      {
        if (yname == "")
	  yname = _listAttOut->getAttribute(0)->GetName();
        if (_listAttOut->isAttribute(yname)
	    && _nisp->_listAttIn->isAttribute(xname))
	  {
            j = _listAttOut->getAttributeIndex(yname);
            i = _nisp->_listAttIn->getAttributeIndex(xname);
            val = getIndexFirstOrder(i, j);
	  }
        else
	  {
            if (!_listAttOut->isAttribute(yname))
	      throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							   "TPolynomialChaos::getIndexFirstOrder - Warning: the output name is unknown");
            if (!_nisp->_listAttIn->isAttribute(xname))
	      throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							   "TPolynomialChaos::getIndexFirstOrder - Warning: the name of the input is unknown");
	  }
      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }
    return val;
  }
  //First indice of sensitivity by attributes index
  Double_t TPolynomialChaos::getIndexFirstOrder(Int_t i, Int_t j)
  {
    Double_t val;
    try
      {
        if (j + 1 < 1)
	  {
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 "TPolynomialChaos::getIndexFirstOrder - ERROR: rank of the output  1");
	  }
        if (j + 1 > _ny)
	  {
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 "TPolynomialChaos::getIndexFisrtOrder - ERROR: rank of the output exceed");
	  }
        if (i + 1 < 1)
	  {
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 "TPolynomialChaos::getIndexFirstOrder - ERROR: rank of the input  < 1");
	  }
        if (i + 1 > _nx)
	  {
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 "TPolynomialChaos::getIndexFirstOrder - ERROR: rank of the input exceed ");
	  }
        val = _pc->GetIndiceFirstOrder(i + 1, j + 1);
      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }
    return val;
  }
  //Total indice of sensitivity  by attributes name
  Double_t TPolynomialChaos::getIndexTotalOrder(TString xname, TString yname)
  {
    Double_t val = -1.0;
    Int_t i, j;
    try
      {
        if (yname == "")
	  yname = _listAttOut->getAttribute(0)->GetName();
        if (_listAttOut->isAttribute(yname)
	    && _nisp->_listAttIn->isAttribute(xname))
	  {
            j = _listAttOut->getAttributeIndex(yname);
            i = _nisp->_listAttIn->getAttributeIndex(xname);
            val = getIndexTotalOrder(i, j);
	  }
        else
	  {
            if (!_listAttOut->isAttribute(yname))
	      throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							   "TPolynomialChaos::getIndexTotalOrder - Warning: the output is unknown");
            if (!_nisp->_listAttIn->isAttribute(xname))
	      throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							   "TPolynomialChaos::getIndexTotalOrder - Warning: the input is unknown");
	  }
      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }
    return val;
  }
  //Total indice of sensitivity  by attributes index
  Double_t TPolynomialChaos::getIndexTotalOrder(Int_t i, Int_t j)
  {
    Double_t val;
    try
      {
        if (j + 1 < 1)
	  {
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 "TPolynomialChaos::getIndexTotalOrder - ERROR: rank of the output < 1");
	  }
        if (j + 1 > _ny)
	  {
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 "TPolynomialChaos::getIndexTotalOrder - ERROR: rank of the output exceed");
	  }
        if (i + 1 < 1)
	  {
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 "TPolynomialChaos::getIndexTotalOrder - ERROR: rank of the input  < 1");
	  }
        if (i + 1 > _nx)
	  {
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 "TPolynomialChaos::getIndexTotalOrder - ERROR: rank of the input exceed");
	  }
        val = _pc->GetIndiceTotalOrder(i + 1, j + 1);
      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }
    return val;
  }
  //Indice of sensitivity of a set of variables given by name
  Double_t TPolynomialChaos::getIndex(TString sinput, TString yname)
  {
    //initialization
    Double_t val;
    Int_t j, i;
    Bool_t ok = kFALSE;
    TList * listOfInput = NULL;
    try
      {
        setGroupEmpty();
        //ouput name
        if (yname == "")
	  {
            yname = _listAttOut->getAttribute(0)->GetName();
            ok = kTRUE;
	  }
        else
	  {
            if (_listAttOut->isAttribute(yname))
	      ok = kTRUE;
            else
	      {
                throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							     "TPolynomialChaos::getIndex - Warning: the output is unknown");
	      }
	  }

        //calculation of the indice
        if (ok == kTRUE)
	  {
            j = _listAttOut->getAttributeIndex(yname);
            if (sinput != "")
	      {
                if (_blog)
		  cout
		    << " ** TPolynomialChaos::getIndex: list of input attributes:"
		    << sinput.Data() << endl;

		try{
		  listOfInput=_nisp->_tds->getListOfAttributes(sinput.Data());
		}catch(URANIE::Exceptions::UErrorExceptions &ue)
		  {
		    ue.printMessage();
		    delete listOfInput;
		    throw ue;
		  }
                TAttribute *att = NULL;
                TIter next(listOfInput);
                while ((att = (TAttribute *) next()))
		  {
                    if (_blog)
		      cout << "Attribute name : " << att->GetName() << endl;
		    try{
		      i = _nisp->_listAttIn->getAttributeIndex(att->GetName());
		      setGroupAddVar(i);
		    }catch(URANIE::Exceptions::UErrorExceptions &ue)
		      {
			ue.printMessage();
			delete listOfInput;
			throw ue;
		      }
		  } //end while
                val = _pc->GetGroupIndice(j + 1);
		delete listOfInput;
	      }
            else
	      {
                if (_blog)
		  cout
		    << " ** TPolynomialChaos::getIndex: All the input attributes"
		    << endl;

                for(Int_t ivarInput=0;ivarInput<_pc->GetDimensionInput();ivarInput++)
		  {
                    setGroupAddVar(ivarInput);
		  } 
                val = _pc->GetGroupIndice(j + 1);
	      } //else
	  } //end if ok
      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
        ue.printMessage();
	delete listOfInput; listOfInput=NULL;
        throw ue;
      }
    return val;
  }
  //Indice of interaction sensitivity of a set of variables given by name
  Double_t TPolynomialChaos::getIndexInteraction(TString sinput, TString yname)
  {
    Double_t val;
    Int_t j, i;
    Bool_t ok = kFALSE;
    try
      {
        setGroupEmpty();
        //ouput name
        if (yname == "")
	  {
            yname = _listAttOut->getAttribute(0)->GetName();
            ok = kTRUE;
	  }
        else
	  {
            if (_listAttOut->isAttribute(yname))
	      ok = kTRUE;
            else
	      {
                throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							     "TPolynomialChaos::getIndexInterraction - Warning: the output is unknown");
	      }
	  }
        //calculation of the indice
        if (ok == kTRUE)
	  {
            j = _listAttOut->getAttributeIndex(yname);
            if (sinput != "")
	      {
                if (_blog)
		  cout
		    << " ** TPolynomialChaos::getIndexInteraction: list of input attributes:"
		    << sinput.Data() << endl;
                TList * listOfInput = _nisp->_tds->getListOfAttributes(
								       sinput.Data());
                TAttribute *att = NULL;
                TIter next(listOfInput);
                while ((att = (TAttribute *) next()))
		  {
                    if (_blog)
		      cout << "Attribute name : " << att->GetName() << endl;
		    try{
		      i = _nisp->_listAttIn->getAttributeIndex(att->GetName());
		      setGroupAddVar(i);
		    }catch(URANIE::Exceptions::UErrorExceptions &ue)
		      {
			ue.printMessage();
			delete listOfInput;
			throw ue;
		      }
		  }
                val = _pc->GetGroupIndiceInteraction(j + 1);
		delete listOfInput;
	      }
            else
	      {
                if (_blog)
		  cout
		    << " ** TPolynomialChaos::getIndexInteraction: All the input attributes"
		    << endl;

                for(Int_t ivarInput=0;ivarInput<_pc->GetDimensionInput();ivarInput++)
		  {
                    setGroupAddVar(ivarInput);
		  } 
                val = _pc->GetGroupIndiceInteraction(j + 1);
	      }
	  }
      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }
    return val;
  }
  //Add a random variable (rank i) in the set
  void TPolynomialChaos::setGroupAddVar(Int_t i)
  {
    try
      {
        if (i + 1 < 1)
	  throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
						       "TPolynomialChaos::setGroupAddVar - ERROR: rank of the input < 1");
        if (i + 1 > _nx)
	  throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
						       "TPolynomialChaos::setGroupAddVar - ERROR: rank of the input exceed");
        _pc->SetGroupAddVar(i + 1);
      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }
  }
  //Freeing the set
  void TPolynomialChaos::setGroupEmpty()
  {
    _pc->SetGroupEmpty();
  }

  void TPolynomialChaos::setCoefficient(Int_t noutput, Int_t num, Double_t coef)
  {
    if (_pc == NULL)
      cout << "pc null ////////////////////////////" << endl;
    try
      {
        if (noutput > _pc->ny)
	  {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       "TPolynomialChaos::setCoefficient - ERROR: number of output must be <= _pc->ny");
	  }
        if (num > _pc->p)
	  {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       "TPolynomialChaos::setCoefficient - ERROR: number of PC must be <= _pc->p");
	  }
        _pc->beta[noutput][num] = coef;
      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }
  }

  Double_t TPolynomialChaos::getCoefficient(Int_t noutput, Int_t num)
  {
    try
      {
        if (noutput > _pc->ny)
	  {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       "TPolynomialChaos::getCoefficient - ERROR: number of output must be <= _pc->ny");
	  }
        if (num > _pc->p)
	  {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       "TPolynomialChaos::getCoefficient - ERROR: number of PC must be <= _pc->p");
	  }
        return _pc->beta[noutput][num];
      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }
  }

  //Same that aboce method but with a name in paremeter in place of a index number
  void TPolynomialChaos::getAnovaOrdered(Double_t seuil, TString yname)
  {
    Int_t j;
    try
      {
        if (yname == "")
	  getAnovaOrdered(seuil, 0);
        else
	  {
            if (_listAttOut->isAttribute(yname))
	      {
                j = _listAttOut->getAttributeIndex(yname);
                getAnovaOrdered(seuil, j);
	      }
            else
	      throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							   "TPolynomialChaos::getAnovaOrdered - Warning: the output is unknown ");
	  }
      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }
  }
  //Same that method  getAnovaOrderedCoefficients(Double_t seuil, Int_t j) with a name in paremeter in place of a index number
  void TPolynomialChaos::getAnovaOrderedCoefficients(Double_t seuil,
						     TString yname)
  {
    Int_t j;
    try
      {
        if (yname == "")
	  getAnovaOrderedCoefficients(seuil, 0);
        else
	  {
            if (_listAttOut->isAttribute(yname))
	      {
                j = _listAttOut->getAttributeIndex(yname);
                getAnovaOrderedCoefficients(seuil, j);
	      }
            else
	      throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							   "TPolynomialChaos::getAnovaOrdered - Warning: the output is unknown ");
	  }
      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }
  }
  //Number of input
  Int_t TPolynomialChaos::getDimensionInput()
  {
    Int_t val;
    val = _nx;
    return val;
  }
  //Number of output
  Int_t TPolynomialChaos::getDimensionOutput()
  {
    Int_t val;
    val = _ny;
    return val;
  }
  //Number of the coefficients of the polynomial chaos
  Int_t TPolynomialChaos::getDimensionExpansion()
  {
    Int_t p;
    p = _pc->GetDimensionExpansion();
    return p;
  }
  //Get the degree
  Int_t TPolynomialChaos::getDegree()
  {
    _degree = _pc->GetDegree();
    return _degree;

  }

  //Same that method  getSample(Int_t k, Int_t j) with the output name in parameter in place of a index number
  Double_t TPolynomialChaos::getSample(Int_t k, TString yname)
  {
    Double_t val;
    Int_t j;
    try
      {
        if (_listAttOut->isAttribute(yname))
	  {
            j = _listAttOut->getAttributeIndex(yname);
            val = getSample(k, j);
	  }
        else
	  {
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 "TPolynomilaChaos::getSample - Warning: the output is unknown ");
	  }
      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }
    return val;
  }
  //Quantile of order alpha of the output yname (name)
  Double_t TPolynomialChaos::getQuantile(Double_t alpha, TString yname)
  {
    Double_t val;
    Int_t j;
    try
      {
        if (_listAttOut->isAttribute(yname))
	  {
            j = _listAttOut->getAttributeIndex(yname);
            val = getQuantile(alpha, j);
	  }
        else
	  {
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 "TPolynomilaChaos::getQuantile - Warning: the output is unknown");
	  }
      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }
    return val;
  }
  //Quantile of order alpha of the output j (index)
  Double_t TPolynomialChaos::getQuantile(Double_t alpha, Int_t j)
  {
    Double_t val;
    val = _pc->GetQuantile(alpha, j + 1);
    return val;
  }
  //Wilks's Quantile of order alpha with confidence beta of the ouput yname (name)
  Double_t TPolynomialChaos::getQuantileWilks(Double_t alpha, Double_t beta,
					      TString yname)
  {
    Double_t val;
    Int_t j;
    try
      {
        if (_listAttOut->isAttribute(yname))
	  {
            j = _listAttOut->getAttributeIndex(yname);
            val = getQuantileWilks(alpha, beta, j);
	  }
        else
	  {
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 "TPolynomialChaos::getQuantileWilks - Warning: the output is unknown ");
	  }
      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }
    return val;
  }
  //Wilks's Quantile of order alpha with confidence beta of the ouput j (index)
  Double_t TPolynomialChaos::getQuantileWilks(Double_t alpha, Double_t beta,
					      Int_t j)
  {
    Double_t val;
    val = _pc->GetQuantileWilks(alpha, beta, j + 1);
    return val;
  }
  //Probability of a value overrun of the ouput yname (name)
  Double_t TPolynomialChaos::getInvQuantile(Double_t seuil, TString yname)
  {
    Double_t val;
    Int_t j;
    try
      {
        if (_listAttOut->isAttribute(yname))
	  {
            j = _listAttOut->getAttributeIndex(yname);
            val = getInvQuantile(seuil, j);
	  }
        else
	  {
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 "TPolynomialChaos::getInvQuantile - Warning: the output is unknown");
	  }
      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }
    return val;
  }
  //Probability of a value overrun of the output j (index)
  Double_t TPolynomialChaos::getInvQuantile(Double_t seuil, Int_t j)
  {
    Double_t val;
    val = _pc->GetInvQuantile(seuil, j + 1);
    return val;
  }
  //get the coefficient n� k of the output name
  Double_t TPolynomialChaos::getCoefficient(Int_t k, TString yname)
  {
    Double_t val = 0;
    Int_t j, p;
    p = getDimensionExpansion();
    try
      {
        if (k >= p)
	  throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
						       "TPolynomialChaos::getCoefficient - Warning: rank of the coefficient exceed");
        if (k < 0)
	  throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
						       "TPolynomialChaos::getCoefficient - Warning: rank of the coefficient <0");
        if (yname == "")
	  val = _pc->beta[1][k];
        else
	  {
            if (_listAttOut->isAttribute(yname))
	      {
                j = _listAttOut->getAttributeIndex(yname);
                val = _pc->beta[j + 1][k];
	      }
            else
	      {
                throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							     "TPolynomialChaos::getCoefficient - Warning: the output is unknown ");
	      }
	  }
      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }
    return val;
  }
  //Writing the program C++ of the polynomial chaos in a file
  void TPolynomialChaos::exportFunction(const char *file, const char *name)
  {
    //initialization
    TString lg = "c";
    lg.ToLower();
    TString sComment = "**";
    TString sFile = TString(file);
    if (sFile.Length() == 0)
      sFile = "nisp";
    if (lg.BeginsWith("c"))
      {
        sFile += TString(".C");
        sComment = "  //";
      }
    TString sName = TString(name);
    // console message
    cout << endl << "*******************************" << endl;
    cout << "*** exportFunction lang[" << lg << "] file[" << sFile << "] name["
	 << name << "]" << endl;
    // openning the file
    std::ofstream * sourcefile = new ofstream(sFile.Data());
    //include    
    *sourcefile << "#include <math.h>" << endl;
    *sourcefile << "#include \"TMath.h\"" << endl << endl;
    //call to Nisp export methods
    WriteCoef(_pc, *sourcefile, (char*) (name));
    WriteIndiceMultiple(_pc, *sourcefile, (char*) (name));
    int flaghermite = 0;
    int flaglegendre = 0;
    int flaglaguerre = 0;
    int flagjacobi   = 0;
    for (int i = 1; i <= _nx; i++)
      {
        if (_pc->funpol[i - 1] == HERMITE)
	  {
            if (!flaghermite)
	      WriteCodeHermite(*sourcefile, (char*) (name));
            flaghermite = 1;
	  }
        else if (_pc->funpol[i - 1] == LEGENDRE)
	  {
            if (!flaglegendre)
	      WriteCodeLegendre(*sourcefile, (char*) (name));
            flaglegendre = 1;
	  }
        else if (_pc->funpol[i - 1] == LAGUERRE)
	  {
            if (!flaglaguerre)
	      WriteCodeLaguerre(*sourcefile, (char*) (name));
            flaglaguerre = 1;
	  }
        else if (_pc->funpol[i - 1] == JACOBI)
	  {
	    if (!flagjacobi) {
	      *sourcefile << "////// lgamma Function from ROOT mathcore library" << endl;
	      //	    Modification :: Write onself the Jacobi polynomial function with the TMath::LnGamma ROOT Function insted of the lgamma C Function
	      //	    *sourcefile << "#define lgamma(double x) { return ::ROOT::Math::lgamma(x);}" << endl;
	      //	    WriteCodeJacobi(*sourcefile, (char*) (name), _pc->gpx->va[i - 1]->a , _pc->gpx->va[i - 1]->b);  ///gpx->va[i - 1]->a, gpx->va[i - 1]->b);

	      *sourcefile << "void " << name << "_jacobi(double *phi,double x, int no) {" << endl;
	      *sourcefile << "   double ai, bi, ci, di, c1, c2, c3;" << endl;
	      *sourcefile << "   double a = " << _pc->gpx->va[i - 1]->a << ";" << endl;
	      *sourcefile << "   double b = " << _pc->gpx->va[i - 1]->b << ";" << endl;
	      *sourcefile << "   double alpha = b - 1.;" << endl;
	      *sourcefile << "   double beta  = a - 1.;" << endl;
	      *sourcefile << "   int i;" << endl;
	      *sourcefile << "   x=2.*x-1.;" << endl;
	      *sourcefile << "   phi[0]=1.; if(no>0)  phi[1]=(alpha-beta)/2. + x * (1. +(alpha+beta)/2.);" << endl;
	      *sourcefile << "   for(i=1;i<no;i++) {" << endl;
	      *sourcefile << "      ai = 2 * (i + 1.) * (i + alpha + beta + 1.) * (2 * i + alpha + beta);"<< endl;
	      *sourcefile << "      bi = (2 * i + alpha + beta + 1.) * (alpha - beta) * (alpha + beta);"  << endl;
	      *sourcefile << "      ci = (2. * i + alpha + beta) * (2. * i + alpha + beta + 1.0) * (2. * i + alpha + beta + 2.0);" << endl;
	      *sourcefile << "      di = 2 * (i + alpha) * (i + beta) * (2 * i + alpha + beta + 2.);" << endl;
	      *sourcefile << "      phi[i+1]= ( (bi + x * ci) * phi[i] - di * phi[i-1] ) / ai;" << endl;
	      *sourcefile << "   }" << endl;
	      *sourcefile << "   for(i=0;i<=no;i++) {" << endl;
	      *sourcefile << "      c1 = 1. / (2. * i + a + b - 1.);" << endl;
	      *sourcefile << "      c2 = TMath::LnGamma(i + a) + TMath::LnGamma(i + b) - TMath::LnGamma(i + 1.) - TMath::LnGamma(i + a + b - 1.);" << endl;
	      *sourcefile << "      c3 = TMath::LnGamma(a + b) - TMath::LnGamma(a) - TMath::LnGamma(b);" << endl;
	      *sourcefile << "      phi[i]  = phi[i] / sqrt(c1 * exp(c2 + c3));" << endl;
	      *sourcefile << "   }" << endl;
	      *sourcefile << "}" << endl;




	    }
            flagjacobi = 1;
	  }
      }
    // Prototype of the function
    if (lg.BeginsWith("c"))
      {
        *sourcefile << "void " << sName << "(double *x, double *y)" << endl;
        *sourcefile << "{" << endl;
      }
    // Header
    // product and time
    time_t seconds;
    time(&seconds);
    *sourcefile << sComment << "//////////////////////////// " << endl;
    *sourcefile << sComment << "" << endl;
    *sourcefile << sComment
		<< "     *********************************************" << endl;
    *sourcefile << sComment << "     ** Uranie v" << URANIE_RELEASE << endl;
    *sourcefile << sComment << "     ** Export Nisp : " << GetName() << endl;
    *sourcefile << sComment << "     ** Date : " << ctime(&seconds);
    *sourcefile << sComment
		<< "     *********************************************" << endl;
    *sourcefile << sComment << "" << endl;
    // comment and title
    *sourcefile << sComment << "" << endl;
    *sourcefile << sComment
		<< "     *********************************************" << endl;
    *sourcefile << sComment << "     ** TDataServer : "
		<< _nisp->_tds->GetName() << " **" << endl;
    *sourcefile << sComment << "     **" << endl;
    *sourcefile << sComment << "     ** " << _nisp->_tds->GetTitle() << " **"
		<< endl;
    *sourcefile << sComment
		<< "     *********************************************" << endl;
    *sourcefile << sComment << "" << endl;
    *sourcefile << sComment << "" << endl;
    *sourcefile << sComment << "//////////////////////////// " << endl;
    // C programm
    if (lg.BeginsWith("c"))
      {
        // Constant
        *sourcefile << sComment << "" << endl;
        *sourcefile << sComment << "     Constantes" << endl;
        *sourcefile << sComment << "" << endl;
        *sourcefile << "   int i,j,k,nx,ny,no,p;" << endl;
        *sourcefile << "   nx=" << _nx << ";ny=" << _ny << ";" << "no="
		    << _degree << ";p=" << getDimensionExpansion()-1 << ";" << endl;
        *sourcefile << "   double xi[" << _nx << "];" << endl;
        // Treat the normalization ( x -> psi_x)
        writeCodeCToDenormalizeInput(sourcefile);
        // Code source
        *sourcefile << sComment << "" << endl;
        *sourcefile << sComment << "     Code" << endl;
        *sourcefile << sComment << "" << endl;
        WriteCodeSrc(_pc, *sourcefile, (char*) (name));
      }
    if (lg.BeginsWith("c"))
      *sourcefile << "}" << endl;
    //end of the file
    sourcefile->close();
    delete sourcefile;
    sourcefile = NULL;

    cout << "*** End Of exportFunction" << endl;
    cout << "*******************************" << endl;
  }
  void TPolynomialChaos::printLog(Option_t *option)
  {
    cout << " ** TPolynomialCgaos:printLog[" << option << "]" << endl;
    cout << "      Number of simulation       :" << _np << endl;
    cout << "      Number of input variable   :" << _nx << endl;
    cout << "      Number of ouput variable   :" << _ny << endl;
    cout << "      Degree maximum             :" << _degree << endl;
    _pc->GetLog();
    cout << "fin de TPolynomialChaos::printLog[" << option << "]" << endl;

  }

  //    PRIVATE FUNCTIONS
  void TPolynomialChaos::writeCodeCToDenormalizeInput(ofstream * sourcefile)
  {
    TString sComment = "  // ";
    cout << " ** TPolynomialCgaos:writeCodeCToDenormalizeInput" << endl;
    *sourcefile << sComment << "" << endl;
    *sourcefile << sComment << "     Treat the normalization ( x -> psi_x)" << endl;
    *sourcefile << sComment << "" << endl;
    URANIE::DataServer::TAttribute *att;
    TStochasticAttribute *stoatt;
    //ofstream::scientific;
    try
      {
        for (int i = 0; i < _nx; i++) {
	  att = (URANIE::DataServer::TAttribute *) _nisp->_listAttIn->getAttribute (i);
	  stoatt = dynamic_cast<TStochasticAttribute *>(att);
	  switch (stoatt->getLawType())
            {
            case TStochasticAttribute::kNormal:
	      {
                if ( stoatt->getStochasticBasis () == TStochasticAttribute::kHermite ) {
		  *sourcefile << "  xi[" << i << "] = ( x[" << i << "] - " << stoatt->getTheoreticalMean() << ") / ( " << stoatt->getTheoreticalStdDev() << " );  // Attribute [" << att->GetName () << "] Normal"  << endl;
                } else {
		  throw URANIE::Exceptions::UErrorExceptions(__FILE__,
							     __LINE__, Form("TPolynomialChaos::writeCodeCToDenormalizeInput\n Normal Law [%s] and type [%d] not Hermite", stoatt->GetName (), stoatt->getLawType()));
                }
                break;
	      }
            case TStochasticAttribute::kUniform:
	      {
                if ( stoatt->getStochasticBasis () == TStochasticAttribute::kLegendre ) {
		  *sourcefile << "  xi[" << i << "] = ( x[" << i << "] - " << stoatt->getLowerBound() << ") / ( " << stoatt->getUpperBound() << " - " << stoatt->getLowerBound() << ");  // Attribute [" << att->GetName () << "] Uniform"  << endl;
                } else {
		  throw URANIE::Exceptions::UErrorExceptions(__FILE__,
							     __LINE__, Form("TPolynomialChaos::writeCodeCToDenormalizeInput\n Uniform Law [%s] and type [%d] not Legendre", stoatt->GetName (), stoatt->getLawType()));
                }
                break;
	      }
            case TStochasticAttribute::kExponential:
	      {
                if ( stoatt->getStochasticBasis () == TStochasticAttribute::kLaguerre ) {
		  *sourcefile << "  xi[" << i << "] = x[" << i << "] / " << ((TExponentialDistribution *)stoatt)->getParameterLambda() << ";  // Attribute [" << att->GetName () << "] Exponential"  << endl;
                } else {
		  throw URANIE::Exceptions::UErrorExceptions(__FILE__,
							     __LINE__, Form("TPolynomialChaos::writeCodeCToDenormalizeInput\n Exponential Law [%s] and type [%d] not Laguerre", stoatt->GetName (), stoatt->getLawType()));
                }
                break;
	      }
            case TStochasticAttribute::kBeta:
	      {
                if ( stoatt->getStochasticBasis () == TStochasticAttribute::kJacobi ) {
		  *sourcefile << "  xi[" << i << "] = ( x[" << i << "] - " << stoatt->getLowerBound() << ") / ( " << stoatt->getUpperBound() << " - " << stoatt->getLowerBound() << ");  // Attribute [" << att->GetName () << "] Beta"  << endl;
                } else {
		  throw URANIE::Exceptions::UErrorExceptions(__FILE__,
							     __LINE__, Form("TPolynomialChaos::writeCodeCToDenormalizeInput\n Beta Law [%s] and type [%d] not Jacobi", stoatt->GetName (), stoatt->getLawType()));
                }
                break;
	      }
            case TStochasticAttribute::kLogUniform:
	      {
                if ( stoatt->getStochasticBasis () == TStochasticAttribute::kLegendre ) {
		  *sourcefile << "  xi[" << i << "] = ( log( x[" << i << "] ) - " << TMath::Log(stoatt->getLowerBound()) << " ) / ( " << TMath::Log(stoatt->getUpperBound()) << " - " << TMath::Log(stoatt->getLowerBound()) << " );  // Attribute [" << att->GetName () << "] LogUniform"  << endl;
                } else {
		  throw URANIE::Exceptions::UErrorExceptions(__FILE__,
							     __LINE__, Form("TPolynomialChaos::writeCodeCToDenormalizeInput\n LogUniform Law [%s] and type [%d] not Legendre", stoatt->GetName (), stoatt->getLawType()));
                }
                break;
	      }
            case TStochasticAttribute::kLogNormal:
	      {
                if ( stoatt->getStochasticBasis () == TStochasticAttribute::kHermite ) { //// getParameterXMin  getTheoreticalMean getUnderlyingNormalParameterMu
		  *sourcefile << "  xi[" << i << "] = ( log( x[" << i << "] - " << ((TLogNormalDistribution *)stoatt)->getParameterXMin() << " ) - " << ((TLogNormalDistribution *)stoatt)->getUnderlyingNormalParameterMu() << " ) / ( " << 1.0 * ((TLogNormalDistribution *)stoatt)->getUnderlyingNormalParameterSigma() << " );  // Attribute [" << att->GetName () << "] LogNormal"  << endl;
                } else {
		  throw URANIE::Exceptions::UErrorExceptions(__FILE__,
							     __LINE__, Form("TPolynomialChaos::writeCodeCToDenormalizeInput\n LogNormal Law [%s] and type [%d] not Hermite", stoatt->GetName (), stoatt->getLawType()));
                }
                break;
	      }
            default:
	      {
                throw URANIE::Exceptions::UErrorExceptions(__FILE__,
                                                           __LINE__, Form("TPolynomialChaos::writeCodeCToDenormalizeInput\n Law [%s] type [%d] not yet implemented", stoatt->GetName (), stoatt->getLawType()));
	      }
            } //end switch
        } // end for
      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }
    cout << " ** End of TPolynomialChaos::writeCodeCToDenormalizeInput" << endl;
  }

  //---------------------------------TO DO------------------------------
  // Polynomial chaos is a random variable : a realisation of this random variable (GetOutput())
  void TPolynomialChaos::realisation()
  {

  }
  //Save the PC in  a file
  void TPolynomialChaos::save(Char_t *file)
  {
    _pc->Save(file);
  }
  //Save the PC in  a file
  void TPolynomialChaos::save(string file)
  {
      char pouet[250];
      strcpy(pouet,file.c_str());
      _pc->Save(pouet);
  }
  //Preparation on the ANOVA decomposition (for Sobol'indices)
  void TPolynomialChaos::setAnova()
  {
    _pc->SetAnova();
  }

  //constructor from a file
  TPolynomialChaos::TPolynomialChaos(const char* ct):
    TNamed("TPolynomialChaos",
	   Form("Chaos Polynomial functionalities from file\"%s\"", ct)),
    _nisp(NULL), _np(0), _bStoreYHat(kTRUE), _degreeResults(NULL),
    _autoDegreeFactor(1.5), _minAutoDeg(-1), _maxAutoDeg(-1), _bestAutoDeg(-1),
    _verrloo(NULL), _blog(kFALSE)
  {
    try
      {
        throw URANIE::Exceptions::UInfoExceptions(__FILE__, __LINE__,
						  Form("TPolynomialChaos::TPolynomialChaos constructor from file \"%s\" is building!!!", ct));
      } catch (URANIE::Exceptions::UInfoExceptions& ue)
      {
        ue.printMessage();
      }

    _pc = new PolynomialChaos((char *)ct);
    _nx = _pc->GetDimensionInput();
    _ny = _pc->GetDimensionOutput();
    _degree = _pc->GetDegree();
  }

  //read a set of target from a file
  void TPolynomialChaos::readTarget(Char_t *target)
  {
      UNUSED(target);
  }
  //Set an input
  void TPolynomialChaos::setInput(Int_t nt, Double_t dt)
  {
      UNUSED(nt,dt);
  }
  //Propagation of input which has been specified SetInput()
  void TPolynomialChaos::propagateInput()
  {

  }
  //Propagation of dt[1...nx]
  void TPolynomialChaos::propagateInput(Double_t *dt)
  {
      UNUSED(dt);
  }
  //Edition of the ANOVA decomposition
  void TPolynomialChaos::getAnova(Int_t nt)
  {
      UNUSED(nt);
  }
  //Edition of the ANOVA ordered decomposition
  void TPolynomialChaos::getAnovaOrdered(Double_t seuil, Int_t j)
  {
    try
      {
        if (j + 1 < 1)
	  throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
						       "TPolynomialChaos::getAnovaOrdered - ERROR: rank of the output < 1");
        if (j + 1 > _ny)
	  throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
						       "TPolynomialChaos::getAnovaOrder - ERROR: rank of the output exceed");
        _pc->GetAnovaOrdered(seuil, j + 1);
      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }
  }
  //Edition of the ANOVA ordered decomposition / coefficients
  void TPolynomialChaos::getAnovaOrderedCoefficients(Double_t seuil, Int_t j)
  {
      UNUSED(seuil,j);
  }
  //Build a sample for statitical analysis (quantile)
  void TPolynomialChaos::generateSample(TString type, Int_t np, Int_t order)
  {
    _pc->BuildSample((string) type, np, order);

  }
  //The value of the output j for the example k
  Double_t TPolynomialChaos::getSample(Int_t k, Int_t j)
  {
    Double_t val;
    try
      {
        if (j + 1 < 1)
	  {
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 "TPolynomialChaos::getSample - ERROR: rank of the output  < 1");
	  }
        if (j + 1 > _ny)
	  {
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 "TPolynomialChaos::getSample - ERROR: rank of the output exceed");
	  }
        val = _pc->GetSample(k, j + 1);
      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }
    return val;
  }
}      //end of uranie namespace
}      //end of uranie namespace
