//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
  \file   TkNN.h
  \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
  \date   Wed Jul  8 22:24:51 2015
  
  \brief  header of the kNN class
  
  
*/
/*!
  \class URANIE::Modeler::TkNN

  \brief k Nearsest Neighbor algorithm

  The TDS in the constructor is the data set to the reference points.

*/
#ifndef TKNN_H
#define TKNN_H


// Uranie
#include "TModeler.h"


namespace URANIE {
  namespace Modeler {
    enum EDistance { kEuclidean, kL1, kManhattan, kChebyshev, kMahalanobis };  ///< the distance
    class TkNN : public TModeler {
    public:
      Double_t     _dr2a;                             ///< The Adjusted R2 quality
      Int_t        _nK;                               ///< the number of neighboor to take into account
      EDistance    _nDistance;                        ///! The distance normalized for input and output
      TString _sDraw;                                 ///< Variables to visualize during the learning process
      URANIE::DataServer::TDataServer *_tdsQuery;     ///< the TDS contins the query points
      
      // Operations
    public:
    //---------------------------------------------
      /** @name Constructor and Destructor
       **/
      //@{
      //! Default constructor with the TDataServer
      TkNN(URANIE::DataServer::TDataServer *tds, TString architecture, Option_t *option ="");
      //! Default constructor with input and output attributes
      /*
	\param tds (URANIE::DataServer::TDataServer *) the dataserver
	\param varexpinput(const char*)[""] The list of input attributes to pass to the function separated by the caracter ":"
	\param varexpoutput (const char *)[""]  The list of output attributes separated by the caracter ":"
	\param 
       */
      TkNN (URANIE::DataServer::TDataServer *tds, const char *varexpinput, const char *varexpoutput, Int_t nkernel, Option_t * option="");
     //! Default destructor
      virtual ~TkNN();
      //@}
      
      //---------------------------------------------
      /** @name Setter and Getter function
       **/
      //@{
      //! Set the number of neighbor (\it k)
      /*!
	\param k (Int_t) k is the number of neighbor into the reference points to select for a query point
      */
      void setK( Int_t k) { _nK = k;}
      //! Get the number of neighbor (\it k)
      Int_t getK() { return _nK;}
      //! Set the set of query points from a dataserver
      /*!
	\param tds (URANIE::DataServer::TDataServer) the TDS which contains the set of query points
      */
      void setQueryPoints(URANIE::DataServer::TDataServer *tds);
      
      //@}
      
      //---------------------------------------------
      /** @name Estimate
       **/
      //@{
      //! draw
      void draw(const char * vars, Option_t * option = "");
      //! make the learning
#ifdef WITH_GPUANN
      void estimate(Option_t * option = "", Bool_t useGPU = true);
#else
      void estimate(Option_t * option = "", Bool_t useGPU = false);
#endif
      //@}

      //---------------------------------------------
      /** @name Export the model in extrenal langage
       **/
      //@{
      //! Export the model in C++ langage in a file
      /*!
	\warning Don't use this method. Use the main method TModeler::exportFunction.
	\param sourcefile (ofstream *) the pointer to the file to export the modeler.
      */
      void exportModelCplusplus(std::ofstream * sourcefile) const;
      //! Export the model in Fortran langage in a file
      /*!
	\warning Don't use this method. Use the main method TModeler::exportFunction.
	\param sourcefile (ofstream *) the pointer to the file to export the modeler.
      */
      void exportModelFortran(std::ofstream * sourcefile) const;

      //! Export the model as a PMML function (not implemented)
      void exportModelPMML(const char* file = "",
                           const char* name = "",Option_t *option = "") const
      {
          TString soption = option;
          soption.ToLower();
	  UNUSED(file,name);
          std::cerr
                  << "   --- Method TkNN::exportModelPMML(ofstream * sourcefile) not yet implemented"
                  << endl;
          throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
              "TkNN::exportFunction\n The PMML language is not yet implemented.");
      };
      //! Export the model as a Python function (not implemented)
      void exportModelPython(std::ofstream * sourcefile) const
      {
	  UNUSED(sourcefile);
          std::cerr
                  << "   --- Method TkNN::exportModelPython(ofstream * sourcefile) not yet implemented"
                  << endl;
          throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
              "TkNN::exportFunction\n The Python language is not yet implemented.");
      };
      //@}

      

      //---------------------------------------------
      /** @name Printing Log
       **/
      //@{
      virtual void printLog(Option_t *option ="");
      //@}

    private:
      void estimate_cpu(Option_t * option = "");

      void LSort(int n, double* array, int* index, Int_t nbthreads);

      void beginLoop(Int_t iq, Int_t nsq);
      void endLoop(Int_t iq, int nsr, double* vecDistance, int* vecIndex, Float_t *vecReferenceOutput, Double_t *tabValue, TBranch *tby, Int_t nbthreads);

#ifdef WITH_GPUANN
    void estimate_GPU(Option_t * option = "");
#endif


      ClassDef(URANIE::Modeler::TkNN, ID_MODELER)	//Classe de
	};
    
  } // Fin du namespace Modeler
} // Fin du namespace URANIE
#endif
