//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TkPCA.cxx
 \b Created : 2019-07-22 07:41:46
 \author Jean-Baptiste Blanchard <jean-baptiste.blanchard@cea.fr>
 \brief Interface file for the class URANIE::Modeler::TkPCA
 
 Execute the Principal Component Analysis from a TDS.
 
 */
//ROOT
#include "TMath.h"
#include "TMatrixTSym.h"
#include "TDecompSVD.h"
#include "TMatrixDSymEigen.h"
#include "TPad.h"
#include "TH2.h"
#include "TLatex.h"

// Uranie
#include "TkPCA.h"
#include "UExceptions.h"

ClassImp (URANIE::Modeler::TkPCA)

namespace URANIE
{
namespace Modeler
{

// -----------------------------------------------------------------------
//  Constructor 
// -----------------------------------------------------------------------
TkPCA::TkPCA(URANIE::DataServer::TDataServer *tds, const char *varexp,
             Option_t * option) :
    _blog(kFALSE), _CentData(true), _CentGram(true), _bCompletedFile(true), _tds(tds), _ker(NULL), _nS(0),
    _vecEigenValues(1), _vecPctEigenValues(1) // _ntdResult(NULL), _vecEigenValues(1), _matEigenVectors(1, 1)
{
    TString soption = option;
    soption.ToLower();
    try
    {
        if (_tds == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TkPCA::constructor ERROR : the pointer of \"tds\" is empty");

        //  Get the list of Input attributes
        _listOfInputAttributes = _tds->getListOfAttributes(varexp);
        Int_t natt = _listOfInputAttributes->GetEntries();
        _nS = _tds->getNPatterns();

        if(soption.Contains("nocentgram"))
            _CentGram=false;
        if(soption.Contains("nocentdata"))
            _CentData=false;
        if(soption.Contains("nocomp"))
            _bCompletedFile=false;

        TString Variables=varexp;
        if(_CentData==true)
        {
            string varia = (string)varexp;
            cout<<"Modified variables "<<varia<<endl;
            TString delim=":";
            char mydelim=delim.Data()[0];
            istringstream tempor(varia); string test;
            Variables="";
            while(std::getline(tempor, test, mydelim))
                if( test != " " && test != "")
                {
                    if(_blog)
                        cout<<"variable under normalisation process : ["<<test<<"]"<<endl;
                    _tds->normalize(test.c_str(),"CR",URANIE::DataServer::TDataServer::kCR,true);
                    Variables += test+"CR:";
                }
            Variables.Remove(Variables.Length()-1,1);
        }
    
        if (_blog)
        {
            cout << " ** natt[" << natt << "], _nS[" << _nS << "]" << endl;
            _listOfInputAttributes->Print();
        }

        _ker = new URANIE::Modeler::TKernel(_tds, Variables.Data(), "", "nointercept,nostoreyhat");

        // compute statistic
        _tds->computeStatistic(Variables.Data());

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

}

// -----------------------------------------------------------------------
//  Destructor
// -----------------------------------------------------------------------
TkPCA::~TkPCA()
{
    cout << " ** Destroy TkPCA" << endl;
    
    if (_listOfInputAttributes)
        delete _listOfInputAttributes;
    _listOfInputAttributes = NULL;

    if (_ker)
        delete _ker;
    _ker = NULL;

    if(_ntdResult != NULL)
        delete _ntdResult;
    _ntdResult=NULL;
    
}

// -----------------------------------------------------------------------
//  Computation
// -----------------------------------------------------------------------
void TkPCA::generateKCenteredMatrix(Option_t * option)
{
    TString soption = option;
    soption.ToLower();

    TMatrixDSym _1N;
    _1N.ResizeTo(_nS, _nS);
    _1N += 1.0 / (1.0*_nS);

    _matKCent.ResizeTo(_nS, _nS);
    _matKCent += _matK - _1N*_matK - _matK*_1N + _1N*_matK*_1N;
    
    if ( _blog )
        _matKCent.Print();
}

// -----------------------------------------------------------------------
//  Computation
// -----------------------------------------------------------------------
void TkPCA::compute(Option_t* option)
{
    if (_blog)
    {
        cout << " ** Compute kPCA[" << option << "]" << endl;
        //_matK.Print();
    }

    _ker->computeK();

    _matK.ResizeTo(_nS,_nS);
    _matK += _ker->getMatrixK();

    double dSumOfL = _matK.Sum();

    if ( _blog ) {
        cout << " ***  Traditionnal _matK with dSumOfL[" << dSumOfL << "]" << endl;
        _matK.Print();
    }

    if( _CentGram )
        generateKCenteredMatrix("");

    TDecompSVD toto( ((_CentGram)?_matKCent:_matK) );
    bool ok=false;
    ok=toto.Decompose();

    cout<<"[91mDecompose ? "<<ok<<"[0m"<<endl;
    TMatrixD U = toto.GetU();
    TMatrixD V = toto.GetV();
    TVectorD Sprime(_nS);
    Sprime = toto.GetSig();
    Sprime.Sqr();

    // Store the eigen values
    _vecEigenValues.ResizeTo(_nS);
    _vecEigenValues += Sprime;

    // Store the percentages
    Sprime*= 100./Sprime.Norm1();
    _vecPctEigenValues.ResizeTo(_nS);
    _vecPctEigenValues = Sprime;

    double sum=0;
    for(int ii=0; ii<_nS; ii++)
        sum+=Sprime[ii];

    if(_blog)
    {
        cout<<"Matrix U"<<endl;
        U.Print();
        cout<<"Matrix V"<<endl;        
        V.Print();
        cout<<"Singular Values"<<endl;
        Sprime.Print( );
    }

    TMatrixD pouet3(_nS,_nS);
    pouet3.TMult(V,((_CentGram)?_matKCent:_matK) );
    pouet3.T();
    if(_blog)
    {
        cout<<"Matrix Pouet3"<<endl;
        pouet3.Print();
    }

    TString BranchesName="";
    for(int ii=0; ii<_nS; ii++)
    {
        if(ii!=0)
            BranchesName+=":";
        BranchesName += Form("kPC_%d",ii+1);
    }
    BranchesName+=":i:eigen:eigen_pct:sum_eigen_pct";
    
    _ntdResult = new URANIE::DataServer::TDSNtupleD("kPCA_proj","kPCA", BranchesName.Data());
    vector<double> val;
    val.resize(_nS+4);
    int jj=0;
    double Eigensum=0;
    for(int ii=0; ii<_nS; ii++)
    {
        for(jj=0; jj<_nS; jj++)
        {
            val[jj]=pouet3(ii,jj);
        }
        val[jj]=ii+1;
        jj++;
        val[jj]=_vecEigenValues[ii];
        jj++;
        val[jj]=_vecPctEigenValues[ii];
        Eigensum+=_vecPctEigenValues[ii];
        jj++;
        val[jj]=Eigensum;
        
        _ntdResult->Fill(&val[0]);
    }

    if(_bCompletedFile)
    {
        string var=_tds->getStringListOfAttributes(false,true);
        _ntdResult->Merger(_tds->getTuple(), var, var);
    }
}

// -----------------------------------------------------------------------
//  drawkPCA
// -----------------------------------------------------------------------
void TkPCA::drawkPCA(Int_t indX, Int_t indY, TString scut, Option_t* option)
{
    if (_blog)
        cout << " ** draw kPCA (" << indX << ", " << indY << ") Option["
                << option << "]" << endl;

    try
    {
        if (indX > _nS)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TkPCA::drawkPCA method\n The first index \"indx\" %d is greater than the number ofattributes (%d)",
                            indX, _nS));
        if (indY > _nS)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TkPCA::drawkPCA method\n The second index \"indY\" %d is greater than the number ofattributes (%d)",
                            indY, _nS));
        if (_ntdResult == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TkPCA::drawkPCA method\n The \"TkPCA::compute()\" method must be call before the \"TkPCA::drawkPCA()\" method");

        // Draw the Graph
        // _ntdResult->SetMarkerColor(3);
        TString sopt = option;
        sopt.ToLower();
        if( !sopt.Contains("same") )
        {
            _ntdResult->Draw(Form("kPC_%d:kPC_%d", indY, indX), scut.Data());

            // Label of the Axis with Percentage of variance explained by the component
            Double_t dPctX = _vecPctEigenValues[indX-1];
            Double_t dPctY = _vecPctEigenValues[indY-1];

            // if (_blog)
            cout << " ** PCT X(" << dPctX << ") -- Y(" << dPctY << ")" << endl;
            TH2F *htemp = (TH2F*) gPad->GetPrimitive("htemp");
            htemp->SetXTitle(Form("Component %d (%5.2f %%)", indX, dPctX));
            htemp->SetYTitle(Form("Component %d (%5.2f %%)", indY, dPctY));
            htemp->SetTitle("");
        }
        
        TString soption = option;
        if(soption != "")
        {            
            if( _tds->isAttribute(soption) )
            {

                _ntdResult->Draw(Form("kPC_%d:kPC_%d", indY, indX),scut.Data(),"goff");

                Double_t* valY = _ntdResult->GetV1();
                Double_t* valX = _ntdResult->GetV2();

                int nS=_tds->getNPatterns();
            
                TGraph *gr = new TGraph(nS);
                TLatex *tl = 0;
                TString sname = "";
                for (Int_t i = 0; i < nS; i++)
                {
                    sname = _tds->getString(soption,i);
                    
                    //if (_blog)
                        cout << " ** sname[" << sname << "]   x(" << valX[i] << ") Y("
                             << valY[i] << ")" << endl;
                    tl = new TLatex(valX[i] * 1.01, valY[i] * 1.01, sname);
                    tl->SetTextAngle(0);
                    tl->SetTextSize(0.05);
                    tl->SetTextColor(kBlack);
                    gr->GetListOfFunctions()->Add(tl);
                }
                gr->Draw("same");

                gPad->Modified();
                gPad->Update();
            }

            else
            {
                _ntdResult->Draw(Form("kPC_%d:kPC_%d", indY, indX), scut.Data(), soption);

                gPad->Modified();
                gPad->Update();
            }
            
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}


// -----------------------------------------------------------------------
//  Log
// -----------------------------------------------------------------------
void TkPCA::printLog(Option_t *option)
{
    cout << " ** TkPCA::printLog[" << option << "]" << endl;
    cout << " ** Name[" << GetName() << "] Title[" << GetTitle() << "]" << endl;
    cout << " ** TDS Name[" << _tds->GetName() << "] Title[" << _tds->GetTitle()
            << "]" << endl;
    cout << " ** List of Attributes size["
            << _listOfInputAttributes->GetEntries() << "]" << endl;
    _listOfInputAttributes->Print();
    cout << " ** Correlation Matrix" << endl;
    _matK.Print();

}

} // End Of namespace Modeler
} // End Of namespace URANIE

