//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TkPCA.h
 \b Created : 2019-07-22 07:41:46
 \author Jean-Baptiste Blanchard <jean-baptiste.blanchard@cea.fr>
 \brief Interface file for the class URANIE::Modeler::TkPCA
 
 Execute the Principal Component Analysis from a TDS.

 */
/*!
 @class URANIE::Modeler::TkPCA

 @brief Execute the Principal Component Analysis from a TDS.


 The Uranie API is
 \code
 TDataServer * tds = new TDataServer("tds", "my TDS");
 tds->fileDataRead("jambu.dat");
 
 TkPCA * tpca = new TkPCA(tds, "Prof:Tran:Mena:Enfa:Cour:Toil:Rep:Somm:Tele:Lois");
 tpca->compute();
 
 tpca->drawkPCA(1, 2);
 tpca->drawLoading(1,2);

 \endcode

 \see http://root.cern.ch/root/html/TPrincipal.html

 */
#ifndef TkPCA_H
#define TkPCA_H
//ROOT
#include "Rtypes.h"
#include "TNamed.h"
#include "TMatrixTBase.h"
#include "TVectorT.h"

//URANIE
#include "Modeler.h"
#include "TDataServer.h"
#include "TDSNtupleD.h"
#include "TKernel.h"

namespace URANIE
{
namespace Modeler
{
class TkPCA: public TNamed
{
public:
    Bool_t _blog;                             ///< Boolean for edit the log
    Bool_t _CentData;                         ///< Centering Data matrix
    Bool_t _CentGram;                         ///< Centering Gram matrix
    Bool_t _bCompletedFile;                   ///< Add original data to final results (for plotting purpose)
    URANIE::DataServer::TDataServer *_tds;     //! Pointer to a TDS
    URANIE::DataServer::TDSNtupleD * _ntdResult;          ///< The NTupleD of results (for variable interpretation)
    TKernel *_ker;                             ///< Kernel constructor pointer
    // TNtupleD * _ntdLoading;                    ///< The NTupleD of Loading
    // TNtupleD * _ntdLoadingCircle;     ///< The NTupleD of Loading for the circle
    TList *_listOfInputAttributes;             ///< The list of attributes
    
    TMatrixD _matK;                         ///< The Gram matrix
    TMatrixD _matKCent;                     ///< The Centered and transformed Gram Matrix
    int _nS;                                ///< Sample size
    TVectorD _vecEigenValues;                 ///< The vector of Eigenvalues
    TVectorD _vecPctEigenValues; ///< The vector of sum (in percent) of the Eigenvalues
    // TMatrixD _matEigenVectors;                ///< The matrix of Eigenvectors

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor with TDS and attributes
    /*!
     \param tds (TDataServer *) the pointer of the TDataServer.
     \param varexp (char*)[""] The list of attributes to analyse separated by the caracter ":" ( default "*" all attributes)
     \param option (Option_t*) [""]  option  (not used actually)
     */

    TkPCA(URANIE::DataServer::TDataServer *tds, const char *varexp = "*",
            Option_t * option = "");

    //! Default destructor
    virtual ~TkPCA();
    //@}

    //---------------------------------------------
    /** @name Get and Set methods
     **/
    //@{
    //! Return the NtupleD of results
    /*!
     The NTupleD contains :
     \li an iterator (\b i)
     \li the eigenvalue (\b eigen)
     \li the eigenvalue in percentage (\b eigen_pct)
     \li the cumulative eigenvalue in percentage (\b sum_eigen_pct)

     \code
     TDataServer * tds = new TDataServer("tds", "my TDS");
     tds->fileDataRead("jambu.dat");
     
     TkPCA * tpca = new TkPCA(tds, "Prof:Tran:Mena:Enfa:Cour:Toil:Rep:Somm:Tele:Lois");
     tpca->compute();

     TNtupleD *ntd = tpca->getResultNtupleD();
     ntd->Scan("*");
     \endcode
     
     \verbatim
     ************************************************************
     *    Row   *         i *     eigen * eigen_pct * sum_eigen *
     ************************************************************
     *        0 *         1 * 4.5886688 * 45.886688 * 45.886688 *
     *        1 *         2 * 2.1198425 * 21.198425 * 67.085114 *
     *        2 *         3 * 1.3209783 * 13.209783 * 80.294897 *
     *        3 *         4 * 1.1952550 * 11.952550 * 92.247448 *
     *        4 *         5 * 0.4684105 * 4.6841051 * 96.931553 *
     *        5 *         6 * 0.1990474 * 1.9904742 * 98.922027 *
     *        6 *         7 * 0.0468131 * 0.4681319 * 99.390159 *
     *        7 *         8 * 0.0370651 * 0.3706510 * 99.760810 *
     *        8 *         9 * 0.0239189 * 0.2391893 *       100 *
     *        9 *        10 * 2.720e-16 * 2.720e-15 *       100 *
     ************************************************************
     \endverbatim
     */
    URANIE::DataServer::TDSNtupleD* getResultNtupleD(){
        return _ntdResult;
    }

     //! Set the type of kernel without parameter
    /*!
     \param nKernel (EKernelType) The type of kernel
     */
    void setKernel(TKernel::EKernelType nKernel){
        _ker->setKernel(nKernel);
    }    
    /*!
     \param nKernel (EKernelType) The type of kernel
     \param dparam (Double_t) the First and only one parameter
     */
    void setKernel(TKernel::EKernelType nKernel, Double_t dparam1){
        _ker->setKernel(nKernel, dparam1);
    }
    //! Set the type of kernel with two parameters
    /*!
     \param nKernel (EKernelType) The type of kernel
     \param dparam1 (Double_t) First parameter
     \param dparam2 (Double_t) Second parameter
     */
    void setKernel(TKernel::EKernelType nKernel, Double_t dparam1, Double_t dparam2){
        _ker->setKernel(nKernel, dparam1, dparam2);
    }
    //! Set the type of kernel with three parameters
    /*!
     \param nKernel (EKernelType) The type of kernel
     \param dparam1 (Double_t) First parameter
     \param dparam2 (Double_t) Second parameter
     \param dparam3 (Double_t) Third parameter
     */
    void setKernel(TKernel::EKernelType nKernel, Double_t dparam1, Double_t dparam2, Double_t dparam3){
        _ker->setKernel(nKernel, dparam1, dparam2, dparam3);
    }
    
    //! Return the NtupleD of results
    /*!
     The TNtupleD contains :
     \li an iterator (\b i)
     \li the projection of origin variable onto PC
     \li the quality of this projection through the squared cosine
     \li the contribution to the definition of each and every PC
    */
    // TDSNtupleD* getVariableResultNtupleD()
    // {
    //     return _ntdVarResult;
    // }
    //@}

    //---------------------------------------------
    /** @name Computation
     **/
    //@{
    //! The compute method
    /*!
     \param option (Option_t*) [""]  option  (not used actually)
     */
    void compute(Option_t* option = "");

    void generateKCenteredMatrix(Option_t * option);

    //@}

    //---------------------------------------------
    /** @name The Draw methods
     **/
    //@{
    //! Draw the Component
    /*!
     \param indX (Int_t) : the X component 
     \param indY (Int_t) : the Y component
     \param option (Option_t*) [""]  option  (not used actually)
     */
    void drawkPCA(Int_t indX, Int_t indY, TString scut="", Option_t* option = "");
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    void setLog()
    {
        _blog = kTRUE;
    }
    void unsetLog()
    {
        _blog = kFALSE;
    }
    void changeLog()
    {
        _blog = _blog ? kFALSE : kTRUE;
    }
    Bool_t getLog()
    {
        return _blog;
    }
    virtual void printLog(Option_t *option = "RAS");
    //@}

    ClassDef(URANIE::Modeler::TkPCA, ID_MODELER)
    //Definition of an variable

};

} // Fin du namespace Modeler
} // Fin du namespace URANIE

#endif
