//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   UMLP.C
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Thu Sep  8 16:06:44 2005
 
 \brief  Implementation of the class UMLP
 
 
 */
#include "UMLP.h"
#include <fstream>
#include <ostream>
#include <iostream>
#include <iomanip>

#include "TString.h"
#include "TMath.h"


#include "newmatio.h"              // to help namespace with VC++ 5
#ifdef use_namespace
using namespace RBD_LIBRARIES;
#endif

#define Act_TanH(sum)         ( tanh(sum) )
#define Act_Logistic(sum) ( 1.0 / ( 1.0 + exp( -1.0 * sum )) )

namespace URANIE
{
namespace Modeler
{

UMLP::UMLP(int input, int hidden, int output) :
  _input(input), _hidden(hidden), _sHidden(Form("%d", hidden)), _output(output),
  _nActivationFunction(kLOGISTIC), _nProlemType(kRegression), _lstOfSynapses(NULL), _tds_is_new(false), _tds(NULL), _bLog(kFALSE)
{
    init();
}

UMLP::UMLP(int input, const char* shidden, int output) :
  _input(input), _hidden(-1), _sHidden(shidden), _output(output),
  _nActivationFunction(kLOGISTIC), _nProlemType(kRegression), _lstOfSynapses(NULL), _tds_is_new(false), _tds(NULL), _bLog(kFALSE)
{
    init();
}

UMLP::UMLP(const char* file) :
  _nActivationFunction(kLOGISTIC), _nProlemType(kRegression), _lstOfSynapses(NULL), _tds_is_new(false), _tds(NULL), _bLog(kFALSE)
{
    int nbw;
    if ( _bLog )
        cout << endl << " ** UMLP::UMLP[" << file << "]" << endl;
    ifstream input(file);
    input >> _input >> _hidden >> _output >> nbw;
    _sHidden = Form("%d", _hidden);
    input.close();
    if ( _bLog )
        cout << " **i(" << _input << ", h(" << _hidden << "), sh[" << _sHidden << "], o(" << _output << "), nw(" << nbw<<")" << endl;
    init();

    try
    {
        if ( nbw != _nbweight )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("UMLP::UMLP ERROR :\n The number of weights in the file (%s)(%d) is not coherent with the input(%d), hidden(%d), output(%d) and computed (%d) numbers", file, nbw, _input, _hidden, _output, _nbweight));
    }catch(URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        if( _lstOfSynapses!=NULL)
            delete _lstOfSynapses;
        _lstOfSynapses = NULL;
        throw ue;
    }

    loadWeights(file);


}

  UMLP::~UMLP()
  {
    if(_tds!=NULL && _tds_is_new)
      delete _tds;
    _tds=NULL;

    if ( _lstOfSynapses != NULL ) {
        delete _lstOfSynapses; _lstOfSynapses = NULL;
    }
  }
  
void UMLP::init()
{
    try
      {
        if ( _input < 1 )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        Form("UMLP::init ERROR :\n the number of Input must be positive and (%d) given", _input));
        if ( _output < 1 )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        Form("UMLP::init ERROR :\n the number of Output must be positive and (%d) given", _output));
     }catch(URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }

    Int_t nRank = 1;
    Int_t nLastIndex = 0;
    Int_t nCrtNeurons;
    Int_t nLastNNeurons = _input;
    USynapses *us = NULL;

    _lstOfSynapses = new TList();
    _lstOfSynapses->SetOwner(kTRUE);
    
    if ( _bLog )
        cout << "  *** _sHidden[" << _sHidden << "] in UMLP::init" << endl;
    string inName;

    TString sCrtHidden = _sHidden;
    sCrtHidden.ReplaceAll(",", " ");
    istringstream inStream((char *) sCrtHidden.Data());

    try
    {
        // Hidden Layers
        while (inStream >> inName)
        {
            if ( _bLog )
                cout << "     *** inName[" << inName << "]" << endl;
            nCrtNeurons = atoi(inName.c_str());
            if ( nCrtNeurons < 1 )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("UMLP::init ERROR :\n the number of Input must be positive and (%d) given in the hidden layer[%s]", nCrtNeurons, _sHidden.data()));
            if ( _bLog )
                cout << "     *** nCrtNeurons[" << nCrtNeurons << "]" << endl;
            us = new USynapses(nLastNNeurons, nCrtNeurons, USynapses::kLOGISTIC, nRank++, nLastIndex+1);
            _lstOfSynapses->Add(us);

            // for the next step
            nLastIndex = us->getLastIndexOfWeights();
            nLastNNeurons = nCrtNeurons;
        }
    }catch(URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        if( _lstOfSynapses!=NULL)
            delete _lstOfSynapses;
        _lstOfSynapses = NULL;
        throw ue;
    }

    //  The output Layer
    us = new USynapses(nLastNNeurons, _output, USynapses::kIDENTITY, -1, nLastIndex+1);
    nLastIndex = us->getLastIndexOfWeights();
    _lstOfSynapses->Add(us);

    _nbweight = nLastIndex;
    if ( _bLog )
        cout << "  **** _nbweight[" << _nbweight << "] nLastIndex[" << nLastIndex << "]" << endl;

    Int_t nNeurons = _input;
    TIter nextSynapses(_lstOfSynapses);
    while ((us = (USynapses *) nextSynapses()))
    {
        if ( _bLog )
            us->printLog();
        nNeurons += us->getNOutput();
    }
    _nbneurons  = nNeurons;
    if ( _bLog )
        cout << "  **** _nbneurons[" << _nbneurons << "]" << endl;


    // Les caracteriqtiques sauvees pour chaque poids
    _weights.ReSize(_nbweight);
    _deriveweights.ReSize(_nbweight);
    _derivepartialweights.ReSize(_nbweight);

    //   CatchAll {
    //     cout << Exception::what() << endl;
    //   }

    // Les caracteriqtiques sauvees pour chaque neurones
    _act.ReSize(_nbneurons);
    _layer.ReSize(_nbneurons);
    _potentiel.ReSize(_nbneurons);
    _derive.ReSize(_nbneurons);
    _delta.ReSize(_nbneurons);
    _deltapartial.ReSize(_nbneurons);

    for(Int_t ii=0; ii<_input; ii++)
        _layer(ii+1) = 0;
    for(Int_t ii=0; ii<_hidden; ii++)
        _layer(_input+ii+1) = 1;
    for(Int_t ii=0; ii<_output; ii++)
        _layer(_input+_hidden+ii+1) = -1;



    _delta = 0.0;

    // Statistiques sur les entrees et sorties
    _minValues.ReSize(_input + _output);
    _maxValues.ReSize(_input + _output);

    time(&_seed);
}
void UMLP::initWeights(time_t seed)
{
    if ( _bLog )
        cout << endl << " ** initWeights seed[" << seed << "]" << endl;
    if (seed != 0)
        srand((unsigned int)seed);

    for (int i = 0; i < _nbweight; i++)
    {
        _weights(i + 1) = -1.0 + 2.0 * rand() / (RAND_MAX + 1.0);
        if ( _bLog )
            cout << "initWeights i(" << i+1 << ")[" << _weights(i+1) << "]" << endl;
    }
    if ( _bLog )
        cout << " ** End Of initWeights seed[" << seed << "]" << endl << endl;
}

void UMLP::setWeights(ColumnVector weights)
{
    _weights = weights;
}

void UMLP::dumpWeights(const char* file)
{
    ostream * output;
    output = new std::ofstream(file);
    // Tailles entrees et couche cachee puis Nombre de poids
    *output << _input << " " << _hidden << " " << _output << " " << _nbweight
            << endl;

    // Les valeurs Min/Max
    int nVar = _input + _output;
    // minimum
    for (int i = 0; i < nVar; i++)
//     *output << _tds->getAttribute(i)->getMinimum() << " ";
        *output << _minValues(i + 1) << " ";
    *output << endl;
    // minimum
    for (int i = 0; i < nVar; i++)
        //      *output << _tds->getAttribute(i)->getMaximum() << " ";
        *output << _maxValues(i + 1) << " ";
    *output << endl;

    for (int i = 1; i <= _nbweight; i++)
    { // 
        *output << _weights(i) << endl;
    }

    ((std::ofstream *) output)->close();
    delete output;
    output = NULL;
}


void UMLP::setActivationFunction(EActivation nFunction)
{
    _nActivationFunction = nFunction;

    if ( _bLog )
        cout << "  *** UMLP::setActivationFunction[" << _nActivationFunction << "] for hidden Layer" << endl;

    if ( _lstOfSynapses ) {
        TIter nextSynapses(_lstOfSynapses);
        USynapses *us;
        while ((us = (USynapses *) nextSynapses()))
        {
            if ( us->getRank() != -1 ) {
                switch (_nActivationFunction) {
                case kTANH:
                    us->setActivationFunction(USynapses::kTANH);
                    break;
                case kLOGISTIC:
                    us->setActivationFunction(USynapses::kLOGISTIC);
                    break;
                }
                if ( _bLog )
                    us->printLog();
            }
        }
    }

}

void UMLP::setProblem( EProblem nType) {
    _nProlemType = nType;

    if ( _bLog )
        cout << "  *** UMLP::setProblem[" << _nProlemType << "] for output Layer" << endl;

    if ( _lstOfSynapses ) {
        TIter nextSynapses(_lstOfSynapses);
        USynapses *us;
        while ((us = (USynapses *) nextSynapses()))
        {
            if ( us->getRank() == -1 ) {
                switch (_nProlemType) {
                case kRegression:
                    us->setActivationFunction(USynapses::kIDENTITY);
                    break;
                case kClassification:
                    us->setActivationFunction(USynapses::kLOGISTIC);
                    break;
                }
                if ( _bLog )
                    us->printLog();
            }
        }
    }
}


//! Exportation au format SNNS dyu reseau de neurones
/*!
 \warning A ne pas confondre avec la methode \b dumpWeights
 qui est la sauvegarde interne a cet objet UMLP.
 
 \param file Nom du fichier d'exportation
 */void UMLP::exportWeights(const char* file)
{
    std::ofstream output(file);

    // Header SNNS
    output << "SNNS network definition file V1.4-3D" << endl;
    time_t clock = 1;
    (void) time((time_t *) &clock);
    output << "generated at " << ctime((time_t *) &clock) << endl;

    // Caracteristique du reseau
    output << "network name : UMLP" << endl;
    output << "source files :" << endl;
    output << "no. of units : " << _nbneurons << endl;
    int nbcon = _hidden * (_input + _output);
    output << "no. of connections : " << nbcon << endl;
    output << "no. of unit types : 0" << endl;
    output << "no. of site types : 0" << endl << endl << endl;

    output << "learning function : Std_Backpropagation" << endl;
    output << "update function   : Topological_Order" << endl << endl << endl;

    // unit default section :
    output << "unit default section :" << endl << endl;
    output
            << "act      | bias     | st | subnet | layer | act func     | out func"
            << endl;
    output
            << "---------|----------|----|--------|-------|--------------|-------------"
            << endl;
    output
            << " 0.00000 |  0.00000 | h  |      0 |     1 | Act_TanH     | Out_Identity "
            << endl;
    output
            << "---------|----------|----|--------|-------|--------------|-------------"
            << endl << endl << endl;
    // unit definition section :
    output << "unit definition section :" << endl << endl;
    output
            << "no. | typeName | unitName | act      | bias     | st | position | act func     | out func | sites"
            << endl;
    output
            << "----|----------|----------|----------|----------|----|----------|--------------|----------|-------"
            << endl;
    int ind = 1;
    int CrtW;
    int CoeffLoc = 3;
    for (int i = 0; i < _input; i++)
    {
        output << setw(3) << ind++
                << " |          | unit     | -0.00000 |  0.00000 | i  | 2,"
                << 2 + CoeffLoc * i << ",-26229    | Act_Identity |          | "
                << endl;
    }
    for (int i = 0; i < _hidden; i++)
    {
        CrtW = i * (_input + 2) + 3;
        output << setw(3) << ind++ << " |          | unit     | -0.00000 | "
                << setw(8) << _weights(CrtW) << " | h  | 8," << 2 + CoeffLoc * i
                << ",-26229    |||" << endl;
    }
    for (int i = 0; i < _output; i++)
    {
        output << setw(3) << ind++ << " |          | unit     | -0.00000 | "
                << setw(8) << _weights(1) << " | o  | 14," << 2 + CoeffLoc * i
                << ",-26229    | Act_Identity |          | " << endl;
    }
    output
            << "----|----------|----------|----------|----------|----|----------|--------------|----------|-------"
            << endl << endl << endl;
    // connection definition section :
    output << "connection definition section :" << endl << endl;
    output << "target | site | source:weight" << endl;
    output
            << "-------|------|---------------------------------------------------------------------------------------------------------------------"
            << endl;
    //// couche cachee
    for (int member = 0; member < _hidden; member++)
    {
        output << setw(5) << _input + member + 1 << "  |      |";
        CrtW = member * (_input + 2) + 4;
        for (int source = 0; source < _input; source++)
        {
            output << setw(3) << source + 1 << ":" << setw(7)
                    << _weights(CrtW++);
            if (source < _input - 1)
                output << ",";
        }
        output << endl;
    }
    //// couche sortie 
    for (int member = 0; member < _output; member++)
    {
        output << setw(5) << _input + _hidden + member + 1 << "  |      |";
        for (int source = 0; source < _hidden; source++)
        {
            CrtW = source * (_input + 2) + 2;
            output << setw(3) << _input + source + 1 << ":" << setw(7)
                    << _weights(CrtW);
            if (source < _hidden - 1)
                output << ",";
        }
        output << endl;
    }

    output
            << "-------|------|---------------------------------------------------------------------------------------------------------------------"
            << endl << endl;

}

void UMLP::exportFunction(const char* lang, const char* file, const char* name,
        int nIC)
{
    TString lg = lang;
    lg.ToUpper();

    if (lg == "FORTRAN")
    {
        exportFunctionFortran(file, name, nIC);
    }

    if (lg == "C++")
    {
        exportFunctionCplusplus(file, name, nIC);
    }

}
void UMLP::exportFunctionCplusplus(const char* file, const char* name,
        int nIC) const
{
    UNUSED(nIC);
    
    int i ;//, source , member, CrtW, nCrt;

    int nVar = _tds->getNAttributes();

    std::ofstream sourcefile(file);
//   sourcefile << endl;
//   sourcefile << "#define Act_TanH(sum)         ( tanh(sum) )" << endl;
//   sourcefile << endl;

    sourcefile << "void " << name << "(double *param, Double_t *res)" << endl;
    sourcefile << "{" << endl;

    // product and time
    time_t seconds;
    time(&seconds);
    sourcefile << "  //" << endl;
    sourcefile << "  //     *********************************************"
            << endl;
    sourcefile << "  //     ** Uranie  " << ctime(&seconds);
    sourcefile << "  //     *********************************************"
            << endl;
    sourcefile << "  //" << endl;

    // comment and title
    sourcefile << "  //" << endl;
    sourcefile << "  //     *********************************************"
            << endl;
    sourcefile << "  //     ** " << _comment << " **" << endl;
    sourcefile << "  //     **" << endl;
    sourcefile << "  //     ** " << _title << " **" << endl;
    sourcefile << "  //     *********************************************"
            << endl;
    sourcefile << "  //" << endl;
    // inputs title 
    sourcefile << "  //    INPUT : " << endl;
    int nInput = 0, nOuput = 0;
    for (i = 0; i < nVar; i++)
        if (_tds->getAttribute(i)->getOrigin()
                == URANIE::DataServer::TAttribute::kAttribute)
        {
            sourcefile << "  //             "
                    << _tds->getAttribute(i)->GetName() << " : " << endl;
            nInput++;
        }
    // outputs title
    sourcefile << "  //" << endl;
    sourcefile << "  //    OUTPUT : " << endl;
    for (i = 0; i < nVar; i++)
        if (_tds->getAttribute(i)->getOrigin()
                == URANIE::DataServer::TAttribute::kAttribute)
        {
            sourcefile << "  //             "
                    << _tds->getAttribute(i)->GetName() << " : " << endl;
            nOuput++;
        }
    sourcefile << "  //" << endl;
    sourcefile << "  ////////////////////////////// " << endl;

    sourcefile << "  int nInput   = " << nInput << ";" << endl;
    sourcefile << "  int nOuput   = " << nOuput << ";" << endl;
    sourcefile << "  int nWeights = " << _nbweight << ";" << endl;
    sourcefile << "  int nHidden  = " << _hidden << ";" << endl;
    sourcefile << "  const int nNeurones  = " << _nbneurons << ";" << endl;
    sourcefile << "  double act[nNeurones];" << endl;

    // Normalisation des donnees
    sourcefile << endl;
    sourcefile << "  // --- Pretraitment of the inputs and outputs" << endl;

    sourcefile << "  double minInput[] = {" << endl << "  ";
    for (i = 0; i < nVar; i++)
    {
        if (_tds->getAttribute(i)->getOrigin()
                == URANIE::DataServer::TAttribute::kAttribute)
        {
            sourcefile << _minValues(i + 1) << ", ";
            if ((i + 1) % 5 == 0)
                sourcefile << endl << "  ";
        }
    }
    if (_input % 5 != 0)
        sourcefile << endl;
    sourcefile << "  };" << endl;

    sourcefile << "  double minOuput[] = {" << endl << "  ";
    for (i = 0; i < nVar; i++)
    {
        if (_tds->getAttribute(i)->getOrigin()
                == URANIE::DataServer::TAttribute::kAttribute)
        {
            sourcefile << _minValues(i + 1) << ", ";
            if ((i + 1) % 5 == 0)
                sourcefile << endl << "  ";
        }
    }
    if (_output % 5 != 0)
        sourcefile << endl;
    sourcefile << "  };" << endl;

    sourcefile << "  double maxInput[] = {" << endl << "  ";
    for (i = 0; i < nVar; i++)
    {
        if (_tds->getAttribute(i)->getOrigin()
                == URANIE::DataServer::TAttribute::kAttribute)
        {
            sourcefile << _maxValues(i + 1) << ", ";
            if ((i + 1) % 5 == 0)
                sourcefile << endl << "  ";
        }
    }
    if (_input % 5 != 0)
        sourcefile << endl;
    sourcefile << "  };" << endl;

    sourcefile << "  double maxOuput[] = {" << endl << "  ";
    for (i = 0; i < nVar; i++)
    {
        if (_tds->getAttribute(i)->getOrigin()
                == URANIE::DataServer::TAttribute::kAttribute)
        {
            sourcefile << _maxValues(i + 1) << ", ";
            if ((i + 1) % 5 == 0)
                sourcefile << endl << "  ";
        }
    }
    if (_output % 5 != 0)
        sourcefile << endl;
    sourcefile << "  };" << endl;

    // Vecteur des poids
    sourcefile << endl;
    sourcefile << "  // --- Values of the weights" << endl;
    sourcefile << "  double valW[] = {" << endl << "  ";

    for ( i = 0; i < _nbweight; i++)
    {
        sourcefile << _weights(i + 1) << ", ";
        if ((i + 1) % 5 == 0)
            sourcefile << endl << "  ";
    }
    sourcefile << endl << "  };" << endl;

    sourcefile << "  // Constantes";
    sourcefile << endl;
    sourcefile << "  int indNeurone = 0;" << endl;
    sourcefile << "  int CrtW;" << endl;
    sourcefile << "  double sum;" << endl;

    // couche  entree
    sourcefile << endl;
    sourcefile << "  // couche  entree" << endl;
    sourcefile << "  for(int i = 0; i < nInput; i++) {" << endl;
    sourcefile
            << "     act[indNeurone++] = 2.0 * ( param[i] - minInput[i] ) / ( maxInput[i] - minInput[i] ) - 1.0;"
            << endl;
    sourcefile << "  }" << endl;

    // couche cachee
    sourcefile << endl;
    sourcefile << "  // couche cachee" << endl;
    sourcefile << "  for (int member = 0; member < nHidden; member++) {"
            << endl;
    sourcefile << "     int CrtW = member * ( nInput + 2) + 2;" << endl;
    sourcefile << "     sum = valW[CrtW++];" << endl;
    sourcefile << "     for (int source = 0; source < nInput; source++) {"
            << endl;
    sourcefile << "         sum += act[source] * valW[CrtW++];" << endl;
    sourcefile << "       }" << endl;
    sourcefile << "       act[indNeurone++] = tanh(sum);" << endl;
    sourcefile << "  }" << endl;

    // couche sortie
    sourcefile << endl;
    sourcefile << "  // couche sortie" << endl;
    sourcefile << "  for (int member = 0; member < nOuput; member++) {" << endl;
    sourcefile << "    sum = valW[0];" << endl;
    sourcefile << "    for (int source = 0; source < nHidden; source++) {"
            << endl;
    sourcefile << "      CrtW = source * ( nInput + 2) + 1;" << endl;
    sourcefile << "      sum += act[nInput+source] * valW[CrtW];" << endl;
    sourcefile << "    }" << endl;
    sourcefile << "    act[indNeurone++] = sum;" << endl;
    sourcefile
            << "    res[member] = minOuput[member] + 0.5 * ( maxOuput[member] - minOuput[member] ) * ( sum + 1.0);"
            << endl;
    sourcefile << "  }" << endl;

    sourcefile << "}" << endl;
    sourcefile.close();

}

void UMLP::exportFunctionFortran(const char* file, const char* name,
        int nIC) const
{
    int i, member, source, CrtW, nCrt;

    std::ofstream sourcefile(file);
    sourcefile << "      SUBROUTINE " << name << "(";

    int nVar = _tds->getNAttributes();
    // inputs
    for (i = 0; i < nVar; i++)
//         if (_tds->getAttribute(i)->getOrigin()
//                 == URANIE::DataServer::TAttribute::kAttribute)
        if (_tds->getAttribute(i)->isInput())
            sourcefile << _tds->getAttribute(i)->GetName() << ",";

    // outputs
    int nOut = 1;
  
    for (i = 0; i < nVar; i++)
    {
//         if (_tds->getAttribute(i)->getOrigin()
//                 == URANIE::DataServer::TAttribute::kAttribute)
          if (_tds->getAttribute(i)->isOutput())
        {
            nOut++;
            sourcefile << _tds->getAttribute(i)->GetName();
            if (nOut < _output)
                sourcefile << ", ";
            else
                sourcefile << ")" << endl;
        }
    }
    sourcefile << endl;


    // product and time
    time_t seconds;
    time(&seconds);
    sourcefile << "C" << endl;
    sourcefile << "C     *********************************************" << endl;
    sourcefile << "C     ** Uranie  " << ctime(&seconds);
    sourcefile << "C     *********************************************" << endl;
    sourcefile << "C" << endl;

    // comment and title
    sourcefile << "C" << endl;
    sourcefile << "C     *********************************************" << endl;
    sourcefile << "C     ** " << _comment << " **" << endl;
    sourcefile << "C     **" << endl;
    sourcefile << "C     ** " << _title << " **" << endl;
    sourcefile << "C     *********************************************" << endl;
    sourcefile << "C" << endl;
    // inputs title 
    sourcefile << "C*    ENTREE : " << endl;
    for (i = 0; i < nVar; i++)
        if (_tds->getAttribute(i)->getOrigin()
                == URANIE::DataServer::TAttribute::kAttribute)
            sourcefile << "C*             " << _tds->getAttribute(i)->GetName()
                    << " : " << endl;
    // outputs title
    sourcefile << "C" << endl;
    sourcefile << "C*    SORTIE : " << endl;
    for (i = 0; i < nVar; i++)
        if (_tds->getAttribute(i)->getOrigin()
                == URANIE::DataServer::TAttribute::kAttribute)
            sourcefile << "C*             " << _tds->getAttribute(i)->GetName()
                    << " : " << endl;
    sourcefile << "C" << endl;
    sourcefile
            << "C**********************************************************************"
            << endl;
    sourcefile << "      IMPLICIT DOUBLE PRECISION (A-H,N-Z)\n";
    for (i = 0; i < nVar; i++)
    {
        sourcefile << "      DOUBLE PRECISION "
                << _tds->getAttribute(i)->GetName() << endl;
    }
    if (nIC != 0)
    {
        sourcefile << endl;
        sourcefile << "C --- Case with the confidence interval" << endl;
        sourcefile << "      DOUBLE PRECISION ddIC" << endl;
        sourcefile << "      DOUBLE PRECISION ddPartialW" << endl;
        sourcefile << "      DOUBLE PRECISION valTemp" << endl;
        sourcefile << "      DIMENSION  vecJacobien (" << _nbweight << ")"
                << endl;
        sourcefile << "      DIMENSION  vecTemp (" << _nbweight << ")" << endl;
        sourcefile << "      DIMENSION  mattJJinv (" << _nbweight << ","
                << _nbweight << ")" << endl;
    }

    // Normalisation des donnees
    sourcefile << endl;
    sourcefile << "C --- Pretraitment of the inputs" << endl;
    for (i = 0; i < nVar; i++)
        if (_tds->getAttribute(i)->getOrigin()
                == URANIE::DataServer::TAttribute::kAttribute)
        {
            sourcefile << "      Xnorm" << _tds->getAttribute(i)->GetName()
                    << " = ";
            sourcefile << "2.D0 * ( " << _tds->getAttribute(i)->GetName()
                    << " - " << _minValues(i + 1) << "D0 ) / "
                    << _maxValues(i + 1) - _minValues(i + 1) << "D0 - 1.0"
                    << endl;
        }

    // Vecteur des poids
    sourcefile << endl;
    sourcefile << "C --- Values of the weights" << endl;
    for ( i = 0; i < _nbweight; i++)
        sourcefile << "      W" << i + 1 << " = " << _weights(i + 1) << endl;

    if (nIC != 0)
    {
        // Pour le calcul de l'intervalle de confiance de prediction
        // matrice (tZ Z)-1 
        sourcefile << endl;
        sourcefile << "C --- Values of the matrix (tZZ)-1" << endl;
        for (int ii = 0; ii < _nbweight; ii++)
        {
            sourcefile << "C --- Colonne " << ii + 1 << endl;
            for (int jj = 0; jj < _nbweight; jj++)
            {
                sourcefile << "      mattJJinv(" << ii + 1 << "," << jj + 1
                        << ") = " << _tZZinv(ii + 1, jj + 1) << endl;
            }
        }
    }

    // Boucle sur les neurones caches
    sourcefile << endl;
    for (member = 0; member < _hidden; member++)
    {
        sourcefile << "C --- hidden neural number " << member + 1 << endl;
        CrtW = member * (_input + 2) + 3;
        sourcefile << "      VAct" << member + 1 << " = W" << CrtW++;
        for (source = 0; source < _input; source++)
        {
            sourcefile << " + W" << CrtW++ << " * Xnorm"
                    << _tds->getAttribute(source)->GetName();
            if (source % 5 == 2)
                sourcefile << endl << "     1           ";
        }
        sourcefile << endl;
        sourcefile << "      VPot" << member + 1
                << " = 2.D0 / (1.D0 + DEXP(-2.D0 * VAct" << member + 1
                << ")) - 1.D0" << endl;
    }

    // Ouput of the model
    sourcefile << endl;
    sourcefile << "C --- Output" << endl;
    sourcefile << "      Out = W1";
    for (source = 0; source < _hidden; source++)
    {
        CrtW = source * (_input + 2) + 2;
        sourcefile << " + W" << CrtW << " * VPot" << source + 1;
        if (source % 3 == 2)
            sourcefile << endl << "     1      ";
    }

    if (nIC != 0)
    {
        sourcefile << endl << endl
                << "C --- Compute the confidence interval ddIC" << endl;
        sourcefile << "C --- Compute the jacobien vecteur " << endl;
        // Le seul neurone de la couche de sortie (biais)
        sourcefile << "      vecJacobien(1) = 1.0D0" << endl;
        // Boucle sur les neurones de la couche cachee
        for (source = 0; source < _hidden; source++)
        {
            sourcefile << "C --- Cas of the hidden neuron " << source + 1
                    << endl;
            // Le poids entre le neurone cache et le neurone de la sortie
            CrtW = source * (_input + 2) + 2;
            sourcefile << "      ddPartialW = 1.0D0 - VPot" << source + 1
                    << " * VPot" << source + 1 << endl;
            sourcefile << "      vecJacobien(" << CrtW++ << ") = VPot"
                    << source + 1 << endl;
            // Le bais du neurone cache
            nCrt = CrtW - 1;
            sourcefile << "      vecJacobien(" << CrtW++ << ") = W" << nCrt
                    << " * ddPartialW" << endl;
            for (member = 0; member < _input; member++)
                sourcefile << "      vecJacobien(" << CrtW++ << ") = W" << nCrt
                        << " * ddPartialW * Xnorm"
                        << _tds->getAttribute(member)->GetName() << endl;
        }
        sourcefile << "C --- Compute the product matriw Z (tJJ)-1 Z" << endl;
        sourcefile << "C ---  where Z is the jacobien vecteur" << endl;
        sourcefile << "      DO 100 I = 1, " << _nbweight << endl;
        sourcefile << "      valTemp = 0.0D0" << endl;
        sourcefile << "      DO 200 J = 1, " << _nbweight << endl;
        sourcefile
                << " 200  valTemp = valTemp + vecJacobien(J) * mattJJinv(J, I)"
                << endl;
        sourcefile << " 100  vecTemp(I) = valTemp" << endl;
        sourcefile << "      valTemp = 0.0D0" << endl;
        sourcefile << "      DO 300 I = 1, " << _nbweight << endl;
        sourcefile << " 300  valTemp = valTemp + vecJacobien(I) * vecTemp(I)"
                << endl;

        // Prendre la racine carre du produit matriciel
        sourcefile << "      ddIC = DSQRT(valTemp)";
        sourcefile << endl;
    }
    // Denormalisation de la sortie
    sourcefile << endl;
    sourcefile << "C --- Pretraitment of the output" << endl;
    if (nIC != 0)
        sourcefile << "C ---  and sum the confidence interval" << endl;
    sourcefile << "      " << _tds->getAttribute(_input)->GetName() << " = "
            << _minValues(_input) << "D0 + "
            << (_maxValues(_input) - _minValues(_input)) / 2.0
            << "D0 * ( Out + 1.D0)";
    if (nIC != 0)
    {
        if (nIC == 1)
            sourcefile << endl << "     1      + " << _dStudentQuantile << " * "
                    << _dsigmahat << " * ddIC";
        else
            sourcefile << endl << "     1      - " << _dStudentQuantile << " * "
                    << _dsigmahat << " * ddIC";
    }
    
    sourcefile << endl;
    sourcefile << "C" << endl;
    sourcefile << "      RETURN" << endl;
    sourcefile << "      END" << endl;

    sourcefile.close();

}

//! Loads the weights of the network from an \b ASCII file.
/*!
 The network consists reseau in a single layer hidden by a single output ( \f$ y \in \Re\f$)
 The format of the file is
 \li first line contains 3 integers : Ninputs Nhidden NbWeights
 \li then a line by weight

 \warning The order of weights is a little special.
 It follows the convention introduced by nicolas Lecler (DAM/DP2I) in his macro \em Excel for TA.
 - the deviation of the output neurone \par
 - the weight between the first neurone of the hidden layer and the output \par
 - the deviation of the first neurone of the hidden layer \par
 - the weight between the first neurone of the input layer and the first neurone of the hidden layer \par
 - the weight between the second neurone of the input layer and the first neurone of the hidden layer \par
 - and so on

 \note a \em dot scheme may be more pertinent , but it remains to be done
 
 \param file File for saving the weights to be loaded
 */
void UMLP::loadWeights(const char* file)
{
    if (1)
        cout << "UMLP::loadWeights( " << file << "]" << endl;
    ifstream input(file);
    int ni, nh, no, nbw;
    input >> ni >> nh >> no >> nbw;
    //cout << ni << " ," << nh << ", " << no << ", " << nbw << endl;
    if ((ni != _input) || (nh != _hidden) || (nbw != _nbweight))
    {
        cout << "***** Error UMLP::loadWeights parameters not coherent :"
                << endl;
        cout << "        entree[" << _input << "] versus [" << ni << "] **"
                << endl;
        cout << "        cachee[" << _hidden << "] versus [" << nh << "] **"
                << endl;
        cout << "         poids[" << _nbweight << "] versus [" << nbw << "] **"
                << endl;
        return;
    }
    int nVar = ni + no;
    _tds = new URANIE::DataServer::TDataServer();
    _tds_is_new=true;
    
    _minValues.ReSize(nVar);
    _maxValues.ReSize(nVar);
    for (int i = 0; i < nVar; i++)
        input >> _minValues(i + 1);
    for (int i = 0; i < nVar; i++)
        input >> _maxValues(i + 1);
    if (0)
    {
        for (int i = 0; i < nVar; i++)
        {
            cout << "i[" << i << "]" << _minValues(i + 1) << " - "
                    << _maxValues(i + 1) << endl;
            ;
        }
    }

    try
      {
	for (int i = 0; i < nVar; i++)
	  {
	    if (i < ni)
	      {
                _tds->addAttribute(Form("inp%d", i+1), _minValues(i + 1), _maxValues(i + 1));
                _tds->getAttribute(Form("inp%d", i+1))->setInput();
	      }
	    else
	      {
                _tds->addAttribute(Form("out%d", i+1), _minValues(i + 1), _maxValues(i + 1));
                _tds->getAttribute(Form("out%d", i+1))->setOutput();
              }
	  }
      }catch(URANIE::Exceptions::UErrorExceptions& ue)
      {
	ue.printMessage();
	if(_tds!=NULL && _tds_is_new)
	  delete _tds;
	_tds=NULL;
        if( _lstOfSynapses!=NULL)
            delete _lstOfSynapses;
        _lstOfSynapses = NULL;
	throw ue;
      }

    for (int i = 1; i <= _nbweight; i++)
    { //
        //cout << i << endl;
        input >> _weights(i);
    }
}

//! Simulation of the network of neurons for an input Pattern
void UMLP::evaluateInput(double *input)
{
    int member, source, CrtW ; //, cible;
    double sum, act;

    int indNeurone = 1;
    // couche  entree
    for (member = 0; member < _input; member++)
    {
        _act(indNeurone) = input[member];
        _potentiel(indNeurone) = _act(indNeurone);
        indNeurone++;
    }

    // couche cachee
    for (member = 0; member < _hidden; member++)
    {
        CrtW = member * (_input + 2) + 3;
        sum = _weights(CrtW++);
        for (source = 0; source < _input; source++)
        {
            sum += _act(source + 1) * _weights(CrtW++);
        }
        _potentiel(indNeurone) = sum;
        act=0.0;
        switch (_nActivationFunction)
        {
        case kTANH:
        {
            act = TMath::TanH(sum);
            break;
        }
        case kLOGISTIC:
        {
            act = Act_Logistic(sum);
            break;
        }
        }
        _act(indNeurone) = act;
        indNeurone++;
    }

    // couche sortie 
    for (member = 0; member < _output; member++)
    {
        sum = _weights(1);
        for (source = 0; source < _hidden; source++)
        {
            CrtW = source * (_input + 2) + 2;
            sum += _act(_input + source + 1) * _weights(CrtW);
        }
        _potentiel(indNeurone) = sum;
        _act(indNeurone) = sum;
    }
}
//! Simulation of the network of neurons for a given pattern.
/*!
 \em Apart from the forward, this function gives the values of the \em feedforward
 for the computations of the derivatives.

 \note The activation functions fonctions ere \em \b identity for the input and output layers,
 and \em \b Tanh for the hidden layer.
 
 \param pointer to the data ( input and output) to be simulated
 */
void UMLP::evaluate(double *in)
{

    int member, source, CrtW=0; //, cible;
    double sum, act, dactvitity, dpotential, dderivation;

    //setLog();

    if ( _bLog )
        cout << endl<< endl;

    int indNeurone = 1;
    // couche  entree
    for (member = 0; member < _input; member++)
    {
        _act(indNeurone) = in[member];
        _potentiel(indNeurone) = _act(indNeurone);
        _derive(indNeurone++) = 1.0;
    }

    if ( 1 ) {
        USynapses * us;
        // Loop by Synapses for the Forward
        TIter nextSynapses(_lstOfSynapses);
        Int_t nIndexFirstNeuron = 1;
        Int_t nIndexCrtNeuron = 1;
        Int_t nIndexTargetNeuron = 0;

        Int_t ni, no, nr;
        // Forward step
        if ( _bLog ) cout << " *** FORWARD STEP ***" << endl << endl;
        while ((us = (USynapses *) nextSynapses()))
        {
            ni = us->getNInput();
            no = us->getNOutput();
            nr = us->getRank();

            nIndexTargetNeuron = nIndexFirstNeuron+ni;

            if ( _bLog ) cout << " *** Rank[" << nr << "] ni[" << ni << "] no[" << no << "] --- nIndexFirstNeuron[" << nIndexFirstNeuron << "] nIndexTargetNeuron[" << nIndexTargetNeuron << "]" << endl;
            for(member = 0; member<no; member++) {
                nIndexCrtNeuron = nIndexFirstNeuron;

                CrtW = us->getBiasIndex(member+1);

                if ( _bLog ) cout << "    **** nIndexCrtNeuron[" << nIndexCrtNeuron << "] BIAS CrtW[" << CrtW-1 << "]" << endl;
                sum = _weights(CrtW++);
                if ( _bLog ) cout << "    **** hi(" << member << ") BIAS CrtW[" << CrtW-1 << "]  Weight[" << _weights(CrtW-1) <<"] sum[" << sum <<"]" << endl;

                for (source = 0; source < ni; source++)
                {
                    sum += _act(nIndexCrtNeuron) * _weights(CrtW++);
                    if ( _bLog ) cout << "      ** hi(" << member << ") so(" << source << ") nIndexCrtNeuron[" << nIndexCrtNeuron << "] act[" << _act(nIndexCrtNeuron) << "] CrtW[" << CrtW-1 << "] Weight[" << _weights(CrtW-1) <<"] sum[" << sum <<"]" << endl;
                    nIndexCrtNeuron++;
                }
                if ( _bLog ) cout << "         -- nIndexTargetNeuron[" << nIndexTargetNeuron << "] - ";
                _potentiel(nIndexTargetNeuron) = sum;
                dpotential = sum;
                us->getActivityDerivate(dpotential, dactvitity, dderivation);

                _act(nIndexTargetNeuron) = dactvitity;
                _derive(nIndexTargetNeuron) = dderivation;
                if ( _bLog ) cout << " _potentiel[" << _potentiel(nIndexTargetNeuron) << "] _act[" << _act(nIndexTargetNeuron) << "] _derive[" << _derive(nIndexTargetNeuron) << "]" << endl;

                if ( us->getRank() == -1 ) {
                    if ( _bLog ) cout << " *** output Layer " << endl;
                    if ( _bLog ) cout << "     *** ou(" << member << ") nIndexTargetNeuron[" << nIndexTargetNeuron << "]" << endl;
                    _delta(nIndexTargetNeuron) = _act(nIndexTargetNeuron) - in[_input+member];
                    _deltapartial(nIndexTargetNeuron) = _derive(nIndexTargetNeuron);
                    if ( _bLog ) cout << "     *** out(" << member << ") act[" <<  _act(nIndexTargetNeuron) << "] in[" << in[_input+member]<<"] _delta[" << _delta(nIndexTargetNeuron) <<"] _deltapartial[" << _deltapartial(nIndexTargetNeuron) <<"]" << endl;

                    if ( 0 ) {
                        for( Int_t ii=0; ii< _input; ii++) {
                            cout<< "   *** ii(" << ii << "/" << _input << ") X[" << in[ii]<< "] ";
                        }
                        cout << " - ";
                        for( Int_t ii=0; ii< _output; ii++) {
                            cout<< " oo(" << ii << "/" << _output << ") Ystar[" << in[_input+ii]<< "] ";
                        }
                        cout << " - ";
                        for( Int_t ii=0; ii< _output; ii++) {
                            cout<< " *** (" << ii << "/" << _output << ") Yhat[" << _act(nIndexTargetNeuron)<< "] ";
                        }
                        cout << endl;
                    }
                }
                nIndexTargetNeuron++;
            }
            nIndexFirstNeuron += ni;
        }
        if ( _bLog ) cout << " *** END OF FORWARD STEP ***" << endl << endl;


        // Backward step
        TListIter backwardSynapses(_lstOfSynapses, kIterBackward);

        if ( _bLog ) cout << endl << endl << " *** BACKWARD STEP ***" << endl;
        nIndexTargetNeuron--;
        if ( _bLog ) cout << " *** nIndexFirstNeuron[" << nIndexFirstNeuron << "] nIndexTargetNeuron[" << nIndexTargetNeuron << "]" << endl;

        while ((us = (USynapses *) backwardSynapses()))
        {
            //us->printLog();
            ni = us->getNInput();
            no = us->getNOutput();
            nr = us->getRank();


            if ( _bLog ) cout << " *** output Layer (of the synapse matrix) to compute Derivate for Bias" << endl;
            for(member = no; member > 0; member--) {
                if ( _bLog ) cout << "     *** ou(" << member << ") nIndexTargetNeuron[" << nIndexTargetNeuron << "]" << endl;
                CrtW = us->getBiasIndex(member);
                if ( _bLog ) cout << "     BIAS CrtW[" << CrtW << "]" << endl;
                _deriveweights(CrtW) = _delta(nIndexTargetNeuron);
                _derivepartialweights(CrtW) = _deltapartial(nIndexTargetNeuron);
                nIndexTargetNeuron--;
            }

            // now compute the delta for the input layer (of the synapse matrix)
            for (member = ni; member > 0; member--) {
                sum=0.;
                if ( _bLog ) cout << "     *** sum(" << sum << ") nIndexTargetNeuron[" << nIndexTargetNeuron << "] nIndexFirstNeuron[" << nIndexFirstNeuron << "]" << endl;
                nIndexCrtNeuron = nIndexFirstNeuron;
                for(source = 0; source < no; source++) {
                    CrtW = us->getWeightIndex(member, source+1);
                    sum += _delta(nIndexCrtNeuron) *  _weights(CrtW);
                    if ( _bLog ) cout << "      ** in(" << member << ") so(" << source+1 << ") nIndexCrtNeuron[" << nIndexCrtNeuron << "] _delta[" << _delta(nIndexCrtNeuron) << "] CrtW[" << CrtW << "] Weight[" << _weights(CrtW) <<"] sum[" << sum <<"]" << endl;
                    _deriveweights(CrtW) = _delta(nIndexCrtNeuron) * _act(nIndexTargetNeuron);
                    _derivepartialweights(CrtW) = _deltapartial(nIndexCrtNeuron) * _act(nIndexTargetNeuron);
                    nIndexCrtNeuron++;
                }
                if ( _bLog ) cout << "         *** nIndexTargetNeuron[" << nIndexTargetNeuron << "] sum(" << sum << ") _derive[" << _derive(nIndexTargetNeuron) << "]" << endl;
                _delta(nIndexTargetNeuron) = _derive(nIndexTargetNeuron) * sum;
                _deltapartial(nIndexTargetNeuron) = _derive(nIndexTargetNeuron) * _weights(CrtW);
                if ( _bLog ) cout << "         *** nIndexTargetNeuron[" << nIndexTargetNeuron << "] _delta(" << _delta(nIndexTargetNeuron) << ") _deltapartial[" << _deltapartial(nIndexTargetNeuron) << "]" << endl;
                nIndexTargetNeuron--;
            }
            nIndexFirstNeuron -= ni;
            nIndexTargetNeuron += ni;
            if ( _bLog ) cout << endl << endl;

        }
        if ( _bLog ) cout << " *** END OF BACKWARD STEP ***" << endl << endl;


    }

    // Calcul de la derive de l'erreur par rapport au poids entre le neurone courant cache et la sortie
    if ( _bLog )  {
        cout << "   ************************************ NEW VERSION ******************************* " << endl;
        for (int i = 1; i <= _nbneurons; i++)
        { // _nbneurons
            cout << scientific << "[" << i << "] layer[" << _layer(i) << "] pot[" << _potentiel(i) << "] act[" << _act(i)
                 << "] delta[" << _delta(i) << "] derive[" << _derive(i) << "]"
                 << endl;
        }
        for (int i = 1; i <=  _nbweight; i++)
        { // _nbweight
            cout << scientific << "weight[" << i << "] value[" << _weights(i)
                 << "] deriveweights[" << setprecision(14) << _deriveweights(i) << "] _derivepartialweights[" << _derivepartialweights(i) << "]" << endl;
        }
        cout << "   ************************************ NEW VERSION ******************************* " << endl << endl;
    }



}

void UMLP::evaluateWeightsJacobian()
{

    int member, source, CrtW;

    // Le seul neurone de la couche de sortie (biais)
    _derivepartialweights(1) = _deltapartial(_input + _hidden + 1);

    for (member = 0; member < _hidden; member++)
    {
        // poids entre le neurone cache et le neurone de sortie
        CrtW = member * (_input + 2) + 2;
        //sum = _delta[_input+_hidden] * _weights(CrtW);
        // Calcul de la derive de la sortie par rapport au poids entre le neurone courant cache et la sortie (jacobien)
        // C'est le produit du deltapartial du neurone de sortie et de l'activation du neurone cache 
        _derivepartialweights(CrtW) = _deltapartial(_input + _hidden + 1)
                * _act(_input + member + 1);
        // le deltapartial du neurone cache
        _deltapartial(_input + member + 1) = _derive(_input + member + 1)
                * _weights(CrtW);
        // Calcul de la derive de l'erreur par rapport au biais du neurone cache courant
        // C'est le delta du neurone cache courant
        //CrtW = member * ( _input + 2) + 3;
        CrtW++; // cas du biais
        _derivepartialweights(CrtW) = _deltapartial(_input + member + 1);

        // cas de la couche d'entre
        for (source = 0; source < _input; source++)
        {
            CrtW++;
            _derivepartialweights(CrtW) = _deltapartial(_input + member + 1)
                    * _act(source + 1);

        }
    }

    // Calcul de la derive de l'erreur par rapport au poids entre le neurone courant cache et la sortie
    for (int i = 1; i <= 0; i++)
    { // _nbneurons
        cout << "[" << i + 1 << "] pot[" << _potentiel(i) << "] act[" << _act(i)
                << "] delta[" << _delta(i) << "] derive[" << _derive(i) << "]"
                << endl;
    }
    for (int i = 1; i <= 0; i++)
    { // _nbweight
        cout << "weight[" << i << "] value[" << _weights(i)
                << "] derivepartialweights(" << _derivepartialweights(i) << "]"
                << endl;
    }

}

void UMLP::evaluateWeightsJacobian(double *in)
{
    evaluate(in);
    evaluateWeightsJacobian();
}

double UMLP::getEcart(double *in)
{
    evaluate(in);
    double error = 0.0;
    Double_t dCrtError = 0.0;

    Int_t nIndexOutput = _nbneurons;
    nIndexOutput -= _output;
    if ( _bLog ) cout << "  *** nIndexOutput[" << nIndexOutput << "] _nbneurons[" << _nbneurons << "] " << endl;

    for (int i = 0; i < _output; i++)
    {
        double hat = _act(nIndexOutput + i+1);
        double star = in[_input + i];

        switch (_nProlemType) {
        case kRegression:
            dCrtError = (hat - star);
            error += 0.5 * dCrtError *dCrtError ;
            break;
        case kClassification:
            dCrtError = star * TMath::Log(hat) + ( 1.0 - star) * TMath::Log(1.0 - hat);
            error += -1.0 * dCrtError;
            break;
        }
        if ( _bLog ) cout << setprecision(12) << "   *** UMLP::getEcart hat[" << hat << "] star[" << star << "] dCrtError[" << dCrtError << "] error[" << error << "]" << endl;
    }
    return error;

}

double UMLP::getWeightsDerivate(int i)
{
    try
    {
        if ( i < 1 )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("UMLP::getWeightsDerivate ERROR :\n The index start to 1 and (%d) is given", i));
        if ( i > _nbweight )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("UMLP::getWeightsDerivate ERROR :\n The index must be lesser or equal to the number of weights [%d] and (%d) is given", _nbweight, i));

        return _deriveweights(i);
    }catch(URANIE::Exceptions::UErrorExceptions& ue) {
        ue.printMessage();
        if( _lstOfSynapses!=NULL)
            delete _lstOfSynapses;
        _lstOfSynapses = NULL;
        throw ue;
    }
}

double UMLP::getPartialWeightsDerivate(int i)
{
    try
    {
        if ( i < 1 )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("UMLP::getPartialWeightsDerivate ERROR :\n The index start to 1 and (%d) is given", i));
        if ( i > _nbweight )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("UMLP::getPartialWeightsDerivate ERROR :\n The index must be lesser or equal to the number of weights [%d] and (%d) is given", _nbweight, i));
        return _derivepartialweights(i);
    }catch(URANIE::Exceptions::UErrorExceptions& ue) {
        ue.printMessage();
        if( _lstOfSynapses!=NULL)
            delete _lstOfSynapses;
        _lstOfSynapses = NULL;
        throw ue;
    }
}

void UMLP::setMatrixtZZinv(Matrix & mattZZinv)
{
    _tZZinv.ReSize(_nbweight, _nbweight);
    _tZZinv = mattZZinv;
}

void UMLP::printLog()
{
    cout << "****Log UMLP[" << _input << "] [" << _hidden << "] [" << _output
            << "] ";
    cout << "_nbneurons[" << _nbneurons << "] _nbweight[" << _nbweight << "]"
            << endl;
    for (int i = 0; i < _nbweight; i++)
        cout << "  weight (" << i+1 << "/" << _nbweight << ") Value[" << _weights(i + 1) << "]" << endl;
    for (int i = 0; i < _input+_output; i++)
        cout << "  att(" << i+1 << "/" << _input+_output << ") Min[" << _minValues(i+1) << "] Max[" << _maxValues(i+1)<< "]" << endl;
    if ( _lstOfSynapses ) {
        TIter nextSynapses(_lstOfSynapses);
        USynapses *us;
        while ((us = (USynapses *) nextSynapses())) {
            us->printLog();

        }
    }
}

} // Fin du namespace Modeler
} // Fin du namespace URANIE
