/////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   USynapses.h
 \b Created : Fri Dec  1 23:35:21 2017
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \brief Interface file for the class URANIE::Modeler::USynapses

 \par Current version
 $Date$
 $Revision$
 
 
 */
#ifndef USYNAPSES_H
#define USYNAPSES_H
#include <string>

//ROOT
#include "Rtypes.h"
#include "TROOT.h"
#include "TNamed.h"
#include "TString.h"

//URANIE
#include "Modeler.h"
#include "UExceptions.h"

//STL
#include <iostream>
using std::cout;

//URANIE


/*!
 @class URANIE::Modeler::USynapses

 @brief Description of a matrix of synapses between 2 layers in ANN.

 Avalaible informations are:
 \li name : the name of the matrix of synapses
 \li title : the title of the matrix of synapses
 \li _nRank : the rank of the matrix of synapses
 \li _nInputNeurons : the number of input neurones
 \li _nOutputNeurons : the number of output neurones
 \li _nFirstIndexOfWeights : the first index of weights
 - ordering
 - taxonomy
 - normalization

 For creation of variables to use in all incertainties and sensitivities studies.

 To define a pointer \em px on a non-bounded variable \f$ x\f$.
 \code
 USynapses *uSynapses = new USynapses(2, 4, 8, );
 \endcode

 */

namespace URANIE {
namespace Modeler {

class USynapses: public TNamed {
public:
    enum EActivationFunction {
        kTANH, kLOGISTIC, kIDENTITY, kUNKNOWN
    };

protected:
    //---------------------------------------------
    /** @name Numerical informations
       **/
    //@{
    Int_t _nInputNeurons;         ///< The number of input neurones
    Int_t _nOutputNeurons;        ///< The number of output neurones
    Int_t _nRank;                 ///< The rank of the layer
    Int_t _nFirstIndexOfWeights;  ///< The first index of weights
    Int_t _nLastIndexOfWeights;   ///< The last index of weights
    EActivationFunction _nOuputActivationFunction;    ///! The activation function for the outputs

    Bool_t _bLog;                 ///< Log printing
    //@}

public:
    //---------------------------------------------
    /** @name Constructor and Destructor

       **/
    //@{
    //! Default constructor
    USynapses();
    //! Ordinal constructor
    USynapses(Int_t ni, Int_t no, EActivationFunction type,  Int_t nrank=0, Int_t nFirstIndex=0);
    //! Default destructor
    virtual ~USynapses();
    //@}

    //---------------------------------------------
    /** @name Operations
       **/
    //@{
    //! init the structure
    void init();
    //@}

    //---------------------------------------------
    /** @name Setting and Getting informations
       **/
    //@{
    //! Returns the number of inputs
    Int_t getNInput() { return _nInputNeurons;}
    //! Returns the number of outputs
    Int_t getNOutput() { return _nOutputNeurons;}
    //! Returns the rank
    Int_t getRank() { return _nRank;}
    //! Returns the first index of weights
    Int_t getFirstIndexOfWeights() {
        return _nFirstIndexOfWeights;
    }
    //! Returns the last index of weights
    Int_t getLastIndexOfWeights() {
        return _nLastIndexOfWeights;
    }
    //! Returns the number of weights
    Int_t getNWeights() {
        return _nLastIndexOfWeights - _nFirstIndexOfWeights+1;
    }
    //! Returns the index of bias for output
    Int_t getBiasIndex(Int_t no);
    //! Returns the index of weight between input and output
    Int_t getWeightIndex(Int_t ni, Int_t no);

    //! Returns the activity and derivates
    void getActivityDerivate(Double_t pot, Double_t &dact, Double_t &dderivate);

    //! Defines the activation function
    void setActivationFunction(EActivationFunction ntype ) {
        _nOuputActivationFunction = ntype; }
    //! Returns the type of activation function in the output
    EActivationFunction getOutputActivationFunction() {
        return _nOuputActivationFunction;
    }

    //! Defines the title of the matrix of synapses
    void setTitle(const char* title) {
        SetTitle(title);
    }
    //! Get the title of the matrix of synapses
    TString getTitle() {
        return TString(GetTitle());
    }
    //@}


    //---------------------------------------------
    /** @name Printing Log
   **/
    //@{
    void setLog() {
        _bLog = kTRUE;
    }
    void unsetLog() {
        _bLog = kFALSE;
    }
    void changeLog() {
        _bLog = _bLog ? kFALSE : kTRUE;
    }
    Bool_t getLog() {
        return _bLog;
    }
    virtual void printLog();
    //@}

    ClassDef(URANIE::Modeler::USynapses, ID_MODELER)
    //Classe def

}; // End Of Class Definition

} // End Of namespace Modeler
} // End Of namespace URANIE

#endif
