//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////

/** ****************************************************** **
                     kNN (GPU) - Implementation
    ** ****************************************************** **/

#include <cublas.h>

#include <stdio.h>
#include <stdlib.h>

#include "kNN_gpu.h"
#include "errors_gpu.h"
#include "engine_gpu.h"

#include "kernels_gpu.h"

namespace URANIE {
  namespace GPU {
  
    /* ************************************* */
    /*         kNN          */
    /* ************************************* */
  
    kNN::kNN(int dim, int k, EDistance t):
      _nX(dim), _nK(k), _nRef(0), matRef_host(0), vecRef_host(0),
      _nQuery(0), matQuery_host(0), vecQuery_host(0),
      _Distance(t)
    {
      initialize();
      char *line = new char[100];

      
      Reporting::InformationLog( "No training done yet.", __FILE__, __LINE__ );
      sprintf(line, "Dimension %d", _nX);
      Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );
      sprintf(line, "Number of NN %d", _nK);
      Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );
      
      sprintf(line, "nMaxThreadsPerBlock of NN %d", URANIE::GPU::Engine::_nMaxThreadsPerBlock);
      Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );

      // Clean
      delete[] line;
    }
    
    kNN::~kNN()
    {
        //       if( matQuery_device ) {
        // 	cublasFree (matQuery_device);
        //       }

        Reporting::InformationLog( "kNN::~kNN Destructor", __FILE__, __LINE__ );
        // Clean
        if( vecQuery_host ) {
            Reporting::InformationLog( " - Free the  vecQuery_host pointer", __FILE__, __LINE__ );
            free(vecQuery_host);
        }

        // Shutdown common (if shared equal to 1)
        GPU::Engine::Shutdown();
    }

    void kNN::initialize()
    {
      // Initialize common (if not already done)
      GPU::Engine::Init();      
    }

    void kNN::setReferencePoints(int nsr, Real *vecReferencePoints, Real *vecReferenceOuput)
    {
        _nRef = nsr;
        matRef_host = vecReferencePoints;
        vecRef_host = vecReferenceOuput;

        // Just a log
//        for(int index=0; index < _nRef; index++) {
//            printf("   - Ref. Pat. (%04d) : Out [%f] : ", index, vecRef_host[index]);
//            for( int indexDim=0; indexDim<_nX; indexDim++) {
//                printf(" (%02d)[%f] ", indexDim, vecReferencePoints[index*_nX+indexDim]);
//            }
//            printf("\n");
//        }
//        printf("\n");

      
      char *line = new char[100];
      Reporting::InformationLog( "kNN::setReferencePoints.", __FILE__, __LINE__ );
      sprintf(line, " - Number of references Points %d", _nRef);
      Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );

//       cublasStatus stat;
      
//       // Allocation on the Device and Transfer Data
//       printf("\n - Matrix matRef_device    : Device  ... ");
//       stat = cublasAlloc (_nRef*_nX, sizeof(*matRef_host), (void**)&matRef_device);
//       if (stat != CUBLAS_STATUS_SUCCESS) {
//         printf ("device memory matRef_device allocation failed");
//         cublasShutdown();
//         return ;
//       }
//       printf(" Ok - Set ...");
//       stat = cublasSetMatrix (_nX, _nRef, sizeof(*matRef_host), matRef_host, _nX, matRef_device, _nX);
//       if (stat != CUBLAS_STATUS_SUCCESS) {
//         printf ("data matRef_device download failed");
//         cublasFree (matRef_device);
//         cublasShutdown();
//         return ;
//       }
//       printf(" Done1\n");
      
//       URANIE::GPU::Engine::UpdateFreeMemory();

      // Clean
      delete[] line;
    }




    void kNN::setQueryPoints(int nsq, Real *vecQueryPoints)
    {
      _nQuery = nsq;
      matQuery_host = vecQueryPoints;
      
      vecQuery_host   = (Real *) malloc(_nQuery * sizeof(Real));
      for (int i=0 ; i<_nQuery; i++)
          vecQuery_host[i] = 0.0f;

      char *line = new char[100];
      Reporting::InformationLog( "kNN::setQueryPoints.", __FILE__, __LINE__ );
      sprintf(line, " - Number of Query Points %d", _nQuery);
      Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );

      // Clean
      delete[] line;
    }

    
    void kNN::estimate()
    {
      char *line = new char[100];
      
      // Determine maximum number of query that can be treated
//       max_nb_query_traited = ( memory_free * MAX_PART_OF_FREE_MEMORY_USED - size_of_float * ref_width*height ) / ( size_of_float * (height + ref_width) + size_of_int * k);
//       max_nb_query_traited = min( query_width, (max_nb_query_traited / 16) * 16 );
//       printf(" - max_nb_query_traited(%d)\n", max_nb_query_traited);
      sprintf(line, " - kNN::run :: Memory Total (%.0f MBytes)", URANIE::GPU::Engine::_nMemoryTotal);
      Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );
      
      sprintf(line, " - kNN::run :: Memory free (%.0f MBytes)", URANIE::GPU::Engine::_nMemoryFree);
      Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );


      // Variables for duration evaluation
      cudaEvent_t start, stop;
      cudaEventCreate(&start);
      cudaEventCreate(&stop);
      float elapsed_time;
      
      int iterations = 1;
     // Display informations
      printf("Number of reference points      : %6d\n", _nRef   );
      printf("Number of query points          : %6d\n", _nQuery );
      printf("Dimension of points             : %4d\n", _nX     );
      printf("Number of neighbors to consider : %4d\n", _nK     );
      printf("Number of iterations            : %4d\n", iterations     );
      printf("Processing kNN search           :"                );
      float* 	dist   = (float *) malloc(_nQuery * _nK * sizeof(float));
      int*	ind    = (int *)   malloc(_nQuery * _nK * sizeof(float));

      // Init 
      for (int i=0 ; i<_nQuery * _nK ; i++) dist[i]  = -1.0;
      for (int i=0 ; i<_nQuery * _nK ; i++) ind[i]   = -1;

      
      // Call kNN search CUDA
      cudaEventRecord(start, 0);
      for (int i=0; i<iterations; i++) {
        printf("\n - Iteration(%d/%d)\n", i, iterations);
    kernel_kNN(matRef_host, _nRef, matQuery_host, _nQuery, _nX, _nK, dist, ind);

    int  nIndexCrt; // = (Int_t)(vecIndex[ik]);
    Real dDistCrt = 0.0f;
//    for(int index=0; index < _nQuery; index++) {
//        printf(" - Query (%05d) :\n", index+1);
//        for( int indexDim=0; indexDim<_nK; indexDim++) {
//            nIndexCrt = ind[_nQuery*indexDim+index];
//            dDistCrt = dist[_nQuery*indexDim+index];
//            printf("\t Index(%04d) : Rank(%04d) - nIndexCrt (%04d) Dist(%f)\n", indexDim, indexDim*_nQuery+index, nIndexCrt, dDistCrt);
//        }
//    }
//    for(int index=0; index < _nQuery * _nK; index++) {
//        printf(" * Index(%04d / %04d) - nIndexCrt (%04d) Dist(%f)\n", index, _nQuery * _nK, ind[index], dist[index]);
//    }

    if ( 0 ) {
	  for(int index=0; index < _nRef; index++) {
            printf("   - Reference Pattern (%04d) : ", index+1);
            for( int indexDim=0; indexDim<_nX; indexDim++) {
                printf(" (%02d)[%f] ", indexDim, matRef_host[index*_nX+indexDim]);
            }
            printf("- Out [%f]\n", vecRef_host[index]);
	  }
	  printf("\n");
	}
	  
	Real dsumw = 0.0f;
	Real dyhat = 0.0f;
	Real dWeight = 0.0f;
	

        if ( 1 ) {
            for(int index=0; index < _nQuery; index++) {
//                printf(" - Query (%05d) : ", index+1);
//                for( int indexDim=0; indexDim<_nX; indexDim++) { // _nX
//                    printf("(%02d)[%f] ", indexDim, matQuery_host[index*_nX+indexDim]);
//                }

                dsumw = 0.0f;
                dyhat = 0.0f;
//                printf("- kNN Index / Distance ");
                for( int indexDim=0; indexDim<_nK; indexDim++) {
                    nIndexCrt = ind[indexDim*_nQuery+index];
//                    dWeight = 1.0f / sqrt(dist[indexDim*_nQuery+index]*dist[indexDim*_nQuery+index]);
                    dWeight = 1.0f / dist[indexDim*_nQuery+index];
                    dsumw += dWeight;
                    dyhat += vecRef_host[nIndexCrt-1] * dWeight;
//                    printf("[%d]y(%f)d(%f)w(%f)s(%f)YY(%f) ", nIndexCrt,  vecRef_host[nIndexCrt-1], dist[indexDim*_nQuery+index], dWeight, dsumw, dyhat); ///    ind[indexDim*_nQuery+index], dist[indexDim*_nQuery+index]);
                }
                vecQuery_host[index] = (Real)(dyhat / dsumw);
//                printf("- Out(%f)\n", vecQuery_host[index]);
            }
        }
      }
      
      cudaEventRecord(stop, 0);
      cudaEventSynchronize(stop);
      cudaEventElapsedTime(&elapsed_time, start, stop);
      printf(" - Estimation done in %f s for %d iterations (%f s by iteration)\n", elapsed_time/1000, iterations, elapsed_time/(iterations*1000));
      
      // Destroy cuda event object and free memory
      cudaEventDestroy(start);
      cudaEventDestroy(stop);

      // Clean
      free(dist);
      free(ind);
      delete[] line;
    }

    void kNN::printLog()
    {
      char *line = new char[100];

      Reporting::InformationLog( "No training done yet.", __FILE__, __LINE__ );

      sprintf(line, "Dimension %d", _nX);
      Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );

      sprintf(line, "Number of NN %d", _nK);
      Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );

      sprintf(line, "Number of Reference points %d", _nRef);
      Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );
      for(int index=0; index < _nRef*_nX; index++) {
          printf(" - Reference :: index(%04d) Value(%f)\n", index, matRef_host[index]);
      }

      sprintf(line, "Number of Query points %d", _nQuery);
      Reporting::InformationLog( std::string(line) , __FILE__, __LINE__ );
      for(int index=0; index < _nQuery*_nX; index++) {
          printf(" - Query :: index(%04d) Value(%f)\n", index, matQuery_host[index]);
      }

      // Clean
      delete[] line;
    }
  } // namespace GPU

} // namespace URANIE

// End of File
