//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////

#ifndef __NN_ACTIVATION_FUNCTIONS_INCLUDE__
#define __NN_ACTIVATION_FUNCTIONS_INCLUDE__
#include <cmath>

/* This flag, when set, allows use __global, __host__ and __device__ keywords in math
   and helper functions (and forbids static variables) */
#ifndef __USE_CUDA_FUNC_QUALIFIERS__
  #define __AFunctionQualifier__ inline
#else
  #define __AFunctionQualifier__ inline __host__ __device__
#endif

namespace NeuralNetworkLevenbergMarquardt
{
	namespace ActivationFunction
	{
		static const Real zero = static_cast<Real>(0.0);
        static const Real one  = static_cast<Real>(1.0);
        static const Real two  = static_cast<Real>(2.0);
        
        typedef Real (*function)( Real );
        #ifndef __USE_CUDA_FUNC_QUALIFIERS__
            #define NNfExp__ std::exp
            #define NNfTanh__ std::tanh
            #define NNfLog__ std::log
            #define NNfSqrt__ std::sqrt
        #else // CUDA
            #define NNfExp__ expf
            #define NNfTanh__ tanhf
            #define NNfLog__ logf // TODO: CUDA: __logf less accurate than logf, but faster
            #define NNfSqrt__ sqrtf
        #endif

        /**! Sigmoid function.
            @param x    real, argument of the function
            @return        Sigmoid(x)=1/(1+exp(-x))
        **/
        __AFunctionQualifier__ Real Sigmoid( Real x )
        {
            Real y = NNfExp__( -x );
            y = one / (one + y);
            return y;
        }
        
        /**! Sigmoid derivative, using the sigmoid value.
            @param s    real, result of the function: s = Sigmoid(x)
            @return     Sigmoid'(x)=s(1-s)=Sigmoid(x)(1-Sigmoid(x))
        **/
        __AFunctionQualifier__ Real dSigmoidWithFunctionValue( Real s )
        {
            return s*(one-s);
        }
        
        /**! Sigmoid derivative.
            @param x    real, argument of the function
            @return     Sigmoid'(x)=Sigmoid(x)(1-Sigmoid(x))
        **/
        __AFunctionQualifier__ Real dSigmoid( Real x )
        {
            Real y = Sigmoid(x);
            return dSigmoidWithFunctionValue(y);
        }
        
        /**! Tanh (hyperbolic tangente) function.
            @param x    real, argument of the function
            @return        tanh(x)
        **/
        __AFunctionQualifier__ Real Tanh( Real x )
        {
            return NNfTanh__(x);// TODO: use CUDA tanh() or tanhf()
        }
        
        /**! Tanh derivative, using the tanh value.
            @param t    real, result of the function: t = tanh(x)
            @return     Tanh'(x)=1-t^2=1-Tanh^2(x)
        **/
        __AFunctionQualifier__ Real dTanhWithFunctionValue( Real t )
        {
            return one-t*t;
        }
        
        /**! Tanh derivative.
            @param x    real, argument of the function
            @return     Tanh'(x)=1-Tanh^2(x)
        **/
        __AFunctionQualifier__ Real dTanh( Real x )
        {
            x = NNfTanh__(x);
            return dTanhWithFunctionValue(x);
        }
        
        /**! Gaussian (kernel) function.
            @param x    real, argument of the function
            @return        Gauss(x)=exp(-x^2)
        **/
        __AFunctionQualifier__ Real Gauss( Real x )
        {
            x *= x;
            x *= -one;
            return NNfExp__(x);
        }

        /**! Gaussian derivative.
            @param x    real, argument of the function
            @return        Gauss'(x)=-2x.exp(-x^2)
        **/
        __AFunctionQualifier__ Real dGauss( Real x )
        {
            Real y = Gauss(x);
            y *= -two;
            y *= x;
            return y;
        }
        
        __AFunctionQualifier__ Real dGaussWithFunctionValue( Real g )
        {
            Real y = NNfSqrt__(-NNfLog__(g));
            y *= -two;
            y *= g;
            return y;
        }
    }
}

#endif
