//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////

#ifndef __NN_ERRORS_INCLUDE__
#define __NN_ERRORS_INCLUDE__
#include <exception>
#include <string>
#include <iostream>

namespace NeuralNetworkLevenbergMarquardt
{
    namespace Reporting
    {
        // Compile-time assertion
        /* Use of CompileTimeError<X>, where X is a constant expression, will cause a compile-time error unless E evaluates to
        a nonzero integral value. */
        /**! \brief Compile-time assertions.
          Use of COMPILE_ASSERT( E, SomeClearErrorMessage ) will cause a compile-time
          (understandable) error unless E evaluates to a nonzero integral value. **/
        template<bool> struct CompileTimeError;
        /**! \brief Compile-time assertions: template specialization (so that verified assertions are quietly ignored). **/
        template<> struct CompileTimeError<true> {};
        /**! \brief Compile-time assertions: the actual macro, checking at compilatiob time if a condition is met. **/
        #define COMPILE_ASSERT(expr, msg) \
        { NeuralNetworkLevenbergMarquardt::Reporting::CompileTimeError<((expr) != 0)> ERROR_##msg; (void)ERROR_##msg; } 
          
        class Exception: public std::exception
        {
            public:
                Exception( const std::string err, const char* filename, size_t line );    
                virtual ~Exception() throw();
                virtual const char* what() const throw();
            
            protected:
                const std::string mError;
                const std::string mFilename;
                const size_t mLine;
                std::string mFullError;

        };
        
        /**!\brief Exceptions specific to the cuBLAS errors **/
        class cublasException: public Exception
        {
            public:
                cublasException( const std::string err, const char* filename, size_t line )
                	: Exception( err, filename, line ) {}
                virtual ~cublasException() throw() {};
        };
        
        inline void Warn( const std::string& warning, const char* filename, size_t line )
        {
            std::cerr << "--Warning:\t" << warning << " (in \"" << filename << "\", l. " << line << ")" << std::endl;
        }
        inline void Err( const std::string err, const char* filename, size_t line )
        {
            throw Exception( err, filename, line );
        }
    } //--> Reporting
    
    namespace GPU
    {
      /**!\brief Check if something went wrong during a cuBLAS call, and throw an exception if something did.
      The second argument is only non-default for the cuBLAS helper functions (which return a status,
      to be then passed as parameter): as for the core functions, they don't return any value and we
      have to check via cublasGetError(). **/
#ifndef NDEBUG
      void checkCuBLASError( const std::string message, int statusCode = -1 );
#else
      inline void checkCuBLASError( const std::string message, int statusCode = -1 ){}
#endif

      /**!\brief Check if something went wrong during a CUDA call, and throw an exception if something did.
       @param kernelName identifier of the kernel wrapper in which the error occurred. **/
#ifndef NDEBUG
      void cudaCheckError( const std::string kernelName );
#else
      inline void cudaCheckError( const std::string kernelName ){}
#endif
    } //--> GPU

} //--> NeuralNetworkLevenbergMarquardt

#endif
