//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////

/** ****************************************************** **
               Kernels (GPU) - Implementation
 ** ****************************************************** **/

/* To allow use __global, __host__ and __device__ keywords in math
  and helper functions (and forbid static variables) */
#define __USE_CUDA_FUNC_QUALIFIERS__

#include <cuda.h>
#include "utils.h"
#include "errors.h"
#include "multilayerperceptron.h"
#include "kernels_gpu.h"
#include "activation_functions.h"

#include <stdio.h>

#ifndef CPU_PROCESSING_ONLY

    namespace NeuralNetworkLevenbergMarquardt
    {
      namespace GPU
      {
#ifndef NDEBUG
      inline void cudaCheckError( const std::string kernelName )
      {
          cudaError_t error = cudaGetLastError();
          if( error != cudaSuccess )
              Reporting::Err( "Error in CUDA kernel \""+ kernelName +"\": "+cudaGetErrorString(error), __FILE__, __LINE__ );
      }
#endif

        namespace Kernels
        {
          /**!\brief Kernel: Fills a NxN identity matrix, represented by a N-dimensional unit vector (the diagonal only) **/
          __global__ void identityMatrix( const int N, Real* id )
          {
              const int i = blockDim.x * blockIdx.x + threadIdx.x;
              if( i < N )
              {
                  id[ i ] = ActivationFunction::one; // References either 1.0 or 1.0f, depending on the actual type of "Real"
              }
          }

          /**!\brief Kernel: performs sample normalization (on N teacher outputs) **/
          __global__ void normalizeSamples( const int N, const NeuronType T, Real* t, const Real min, const Real max )
          {
              namespace pp = Preprocessing;
              const int i = blockDim.x * blockIdx.x + threadIdx.x;
              if( i < N )
              {
                  switch(T)
                  {
                      case kGauss:
                      case kLinear:
                      case kOff:
                          break;
                      case kSigmoid:
                          t[i] = pp::normalize<pp::kZeroOne>( t[i], min, max );
                          break;
                      case kTanh:
                          t[i] = pp::normalize<pp::kMinusOneOne>( t[i], min, max );
                          break;
                  }
              }
          }

          /**!\brief Kernel: applies appropriate activation function (on each component of a Nx(K+1) matrix).
              Applies the function to the N elements of the first K columns,  and set bias equal to 1. 
              to the elements of the (K+1)th column. **/
          __global__ void activationFunction( const int NK, const int N, const NeuronType T, Real* values )
          {
              const int i = blockDim.x * blockIdx.x + threadIdx.x;
              if( i < NK )
              {
                  switch(T)
                  {
                      case kLinear:
                          break;
                      case kSigmoid:
                          values[i] = ActivationFunction::Sigmoid( values[i] );
                          break;
                      case kTanh:
                          values[i] = ActivationFunction::Tanh( values[i] );
                          break;
                      case kGauss:
                          values[i] = ActivationFunction::Gauss( values[i] );
                          break;
                      case kOff:
                          values[i] = ActivationFunction::zero;
                          break;
                  }
              }
              else if( i < NK+N ) // Last column ((K+1)th) - bias column
              {
                values[i] = ActivationFunction::one;
              }
          }

          /**!\brief Kernel: computes the NxR Jacobian matrix, element by element. **/
            __global__ void jacobianElement( const int M, const int K, const int N,
                                             const NeuronType T,
                                             const Real* hiddenOutput,
                                             const Real* weightsOutput,
                                             const Real* inputValues,
                                             Real* jacobian )
          {
              const int n = blockDim.x * blockIdx.x + threadIdx.x;
              int k = blockDim.y * blockIdx.y + threadIdx.y;
              const int m = k/K; // blockDim.z must be 1.
              k = k%K;
              
              namespace AF = ActivationFunction;
              if( n < N && m < M+1 && k < K ) // M+1: +1 for the bias
              {
                Real dk = weightsOutput[k];     // sum_{p=0}^{P-1} w(p,k) (P=1)
                switch(T)            // [ f'(a_k)*sum_{p=0}^{P-1} w(p,k) ]
                {
                  case kLinear:
                    break;
                  case kSigmoid:
                    // Faster then dSigmoid( x ), since we already have computed the sigmoid
                    dk *= AF::dSigmoidWithFunctionValue( hiddenOutput[ n + k*N ] );
                    break;
                  case kTanh:
                    dk *= AF::dTanhWithFunctionValue( hiddenOutput[ n + k*N ] );
                    break;
                  case kGauss:
                    dk *= AF::dGaussWithFunctionValue( hiddenOutput[ n + k*N ] );
                    break;
                  case kOff:
                    dk = ActivationFunction::zero;
                    break;
                }
                // [ f'(a_k)*sum_{p=0}^{P-1} w(p,k) ]*x_m = f'(a_k)*w(0,k)*x_m
                jacobian[ n + (m*K+k)*N ] = dk*inputValues[ n + m*N ];
              }
          }
        } // namespace Kernels
      
        namespace Wrappers
        {
          void kIdentityMatrix( dim3 numBlocks, dim3 threadsPerBlock, const int N, Real* id )
          {
            Kernels::identityMatrix<<<numBlocks, threadsPerBlock>>>( N, id );
            cudaCheckError( __FUNCTION__ );
          }
          
          void kNormalizeSamples( dim3 numBlocks, dim3 threadsPerBlock, const int N, const NeuronType T, Real* t,
                                      const Real min, const Real max )
          {
            Kernels::normalizeSamples<<<numBlocks, threadsPerBlock>>>( N, T, t, min, max );
            cudaCheckError( __FUNCTION__ );
          }
          
          void kActivationFunction( dim3 numBlocks, dim3 threadsPerBlock,
                                        const int NK, const int N, const NeuronType T, Real* values )
          {
            Kernels::activationFunction<<<numBlocks, threadsPerBlock>>>( NK, N, T, values );
            cudaCheckError( __FUNCTION__ );
          }
          
          void kJacobianElement( const int M, const int K, const int N,
                                  const NeuronType T,
                                     const Real* hiddenOutput, const Real* weightsOutput,
                                     const Real* inputValues, Real* jacobian )
          {
          
          	/* Problem: the 3D grid is made of blocks whose z dimension can only be 1 (cf. CUDA doc)
          	 * and a block can contain at most 512, 1024 or 2048 threads (depending on the device). We cannot
             * use 3D to store (n,k,m), because "big" M (eg, >= 10) would lead to a CUDA error (too many
             * threads per block). A workaround is to use 2D blocks, with indexes being (n, m*K+k), and to get
             * the tuple (n, m, k) with euclidean division by K in the kernel itself. */
                dim3 threadsPerBlock( 64, 4 );
                dim3 numBlocks( (N+threadsPerBlock.x-1)/threadsPerBlock.x,
                                ((M+1)*K+threadsPerBlock.y-1)/threadsPerBlock.y,
                                1 ); // Third dimension here can only be 1 (CUDA restriction)
            Kernels::jacobianElement<<<numBlocks, threadsPerBlock>>>( M, K, N, T, hiddenOutput, weightsOutput, inputValues, jacobian );
            cudaCheckError( __FUNCTION__ );
          }
        } // namespace Wrappers
      
      } // namespace GPU

    } // namespace NeuralNetworkLevenbergMarquardt
#endif // #ifndef CPU_PROCESSING_ONLY
