//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////

#ifndef __NN_UTILS_INCLUDE__
#define __NN_UTILS_INCLUDE__

#include "errors.h" // For reporting::Err
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/format.hpp>             // Error formatting printf-style
//#include <boost/numeric/ublas/io.hpp>   // To output the vectors with std::cout

#ifdef NDEBUG
    #define DEBUG()
    #undef BOOST_UBLAS_TYPE_CHECK_MIN
    #undef BOOST_UBLAS_TYPE_CHECK_EPSILON
    #define BOOST_UBLAS_TYPE_CHECK 0
#else
    #include <iostream>
    #include <iomanip>  // std::setfill, std::setw ...
    #define DEBUG() std::cout << "DEBUG:\t" << __FILE__ << ":" << __LINE__  << " - " << __FUNCTION__ << "()" << std::endl;
#endif

/* This flag, when set, allows use __global, __host__ and __device__ keywords in math
   and helper functions (and forbids static variables) */
#ifndef __USE_CUDA_FUNC_QUALIFIERS__
  #define __PPFunctionQualifier__ 
  #define __PPFunctionStatic__ static
#else
  #define __PPFunctionQualifier__ __host__ __device__
  #define __PPFunctionStatic__  // No static in device functions
#endif

namespace NeuralNetworkLevenbergMarquardt
{
    // Type we use for real numbers (typically double or float)
    #ifdef CPU_PROCESSING_ONLY
        //typedef double Real;  // For the sake of GPU/CPU comparison, both should be single-point precision
        typedef float Real;
        typedef int Int;
    #else
        typedef float Real;     //!< GPU => float (this is an implication, not a pretty arrow)
        typedef int Int;
    #endif

    namespace Debug
    {
        /**!\brief Displays a host-side matrix.
        @param mat matrix to display
        @param out stream where to output the matrix **/
        template<typename T, typename A> void printMatrix( const boost::numeric::ublas::matrix<T,A>& mat, std::ostream& out = std::cout )
        {
            out << "Displaying[CPU]::" << mat.size1() << "x" << mat.size2() << " matrix:" << std::endl;
            for( size_t i = 0; i < mat.size1(); ++i )
            {
                for( size_t j = 0; j < mat.size2(); ++j )
                {
                    out << mat(i,j) << "\t";
                }
                out << std::endl;
            }
        }

        /**!\brief Displays a host-side vector.
        @param vec vector to display
        @param out stream where to output the vector **/
        template<typename T> void printVector( const boost::numeric::ublas::vector<T>& vec, std::ostream& out = std::cout )
        {
            out << "Displaying[CPU]::" << vec.size() << "-dimensional vector (in row form):" << std::endl;
            out << "[ ";
            for( size_t i = 0; i < vec.size(); ++i )
            {
                out << vec(i) << " ";
            }
            out << "]" << std::endl;
        }
        
        /**!\brief Displays a device-side matrix (or vector)
        @param M number of rows (and leading dimension)
        @param N number of columns
        @param deviceMatrix pointer in GPU memory to the column-major matrix to display.
        @param out stream where to output the matrix **/
        void printMatrixGPU( const size_t M, const size_t N, const Real* deviceMatrix, std::ostream& out = std::cout );
        // Defined in cppimplementations.cpp
    }
    
    namespace Preprocessing
    {
        enum NormalizationType
        {
            kZeroOne,
            kMinusOneOne
        };

        template<NormalizationType T> __PPFunctionQualifier__ Real normalize( Real value, Real min, Real max )
        {
            __PPFunctionStatic__ const Real b = 1.0;
            __PPFunctionStatic__ const Real a = ( T == kZeroOne ) ? 0.0 : -1.0;
            if( max > min )
                return (value - min)/(max-min)*(b-a)+a;
            return a;    
        }
        template<NormalizationType T> __PPFunctionQualifier__  Real denormalize( Real nvalue, Real min, Real max )
        {
            __PPFunctionStatic__ const Real b = 1.0;
            __PPFunctionStatic__ const Real a = ( T == kZeroOne ) ? 0.0 : -1.0;
            if( max > min )
                return (nvalue-a)/(b-a)*(max-min)+min;
            else if( nvalue < a || nvalue > b )
              Reporting::Err( "denormalize(): Value to restore is not in the normalized range", __FILE__, __LINE__ );
            else if( min > max )
              Reporting::Err( "denormalize(): having min <= max is usually a good idea.", __FILE__, __LINE__ );
            return a;
        }
    }
    
    namespace GPU
    {
      /**!\brief Initializes/shuts down all required subsystems.
        Doesn't do anything if CPU_PROCESSING_ONLY is defined. **/
      class Engine
      {
      public:
        static void Init();
        static void Shutdown();
      private:
        explicit Engine(){} // No instanciation.
        static bool systemInitialized;
      };
    }

    namespace Random
    {
      /**!\brief Returns a random real, following a N(0,1) distribution. */
      Real StandardGaussianValue();
      void RandomPermutation(size_t len, Int * list);
    }
}

#endif
