//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   nisp_util.cpp
 \author Jean-Marc Martinez <jean-marc.martinez@cea.fr>
 \date   2011/04/07 14:33:00

 \brief
 */
#include <iostream>
#include <sstream>
#include <fstream>
#include <string>
#include <vector>
#include <math.h>
#ifndef WIN32 
#include <unistd.h> // pour le type pid_t
#else
#include <io.h>
#include <process.h>
#include "utilWIN32.h"
#endif
#include <algorithm>
#include "nisp_pc.h"

using namespace std;

#include "nisp_util.h"

double RandUni();
double RandNor();
double RandExp();
double RandBeta(double, double);
double InvFcn(double);
double alngam ( double a, int *ifault );
double xinbta ( double p, double q, double beta, double alpha, int *ifault );
double betain ( double x, double p, double q, double beta, int *ifault );

// PDF
double Normale(double mean, double std, double null1, double null2)
{
    (void)(null1); (void)(null2);
    return (mean + std * RandNor());
}
double LogNormale(double mean, double std, double min, double null2)
{
    double mu = log(mean) - std * std / 2.;
    double zero = 0.;
    return (min + exp(Normale(mu, std, zero, null2))); // mu = moyenne de la lognormale mais std = sigma de la normale
}
double Uniforme(double min, double max, double null1, double null2)
{
    (void)(null1); (void)(null2);
    return (min + (max - min) * RandUni());
}
double LogUniforme(double min, double max, double null1, double null2)
{
    return (exp(Uniforme(log(min), log(max), null1, null2))); // min et max de la variable
}
double Exponentielle(double lambda, double min, double null1, double null2)
{
    (void)(null1); (void)(null2);
    return (min + RandExp() / lambda);
}
double Beta(double a, double b, double min, double max) {
  return(min + (max - min) * RandBeta(a,b));
}
// mere est uniforme dans [0,1]
double PdfChange(RandomVariable *fils, double z)
{
    double y;
    double a = fils->a;
    double b = fils->b;
    string typefils = fils->type;
    if ((z < 0.) || (z > 1.))
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(PdfChange) : value " << z << "is not in [0,1]" << endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    if (typefils == "Uniforme")
        y = a + (b - a) * z;
    else if (typefils == "Normale")
    {
        y = a + b * InvFcn(z);
    }
    else if (typefils == "Exponentielle")
        y = fils->b + (-1. / a) * log(z);
    else if (typefils == "LogNormale")
        y = fils->c + exp(log(a) - b * b / 2. + b * InvFcn(z));
    else if (typefils == "LogUniforme")
        y = exp(log(a) + log(b / a) * z);
    else if(typefils=="Beta")
    {
      int ifault;
      double beta_log = alngam ( a, &ifault ) + alngam ( b, &ifault ) - alngam ( a + b, &ifault );
      y	= fils->c + (fils->d - fils->c) * xinbta ( a, b, beta_log, z, &ifault );
    }
    else
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(PdfChange) : law " << typefils << " unknown" << endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    return (y);
}
double PdfChange(RandomVariable *fils, RandomVariable *mere, double x)
{
    double y;
    double a = fils->a;
    double b = fils->b;
    double c = fils->c;
    double d = fils->d;
    string typefils = fils->type;
    string typemere = mere->type;
    ostringstream msg;
    if (typemere == "Normale")
    {            // Mere est Normale
        double mean = mere->a;
        double sigma = mere->b;
        double z = (x - mean) / sigma; // z suit une loi normale centree reduite
        if (typefils == "Normale")
            y = a + b * z;
        else if (typefils == "Uniforme")
            y = a + (b - a) * (erfc(-z / sqrt(2.)) / 2.);
        else if (typefils == "Exponentielle")
            y = b -(1. / a) * log((erfc(-z / sqrt(2.)) / 2.));
        else if (typefils == "LogNormale")
            y = c + exp(log(a) - b * b / 2. + b * z);
        else if (typefils == "LogUniforme")
            y = exp(log(a) + log(b / a) * (erfc(-z / sqrt(2.)) / 2.));
	else if(typefils=="Beta")
	{
	  int ifault;
	  double u = erfc(-z/sqrt(2.))/2.; // u suit une loi uniforme dans [0,1]
	  double beta_log = alngam ( a, &ifault ) + alngam ( b, &ifault ) - alngam ( a + b, &ifault );
	  y = fils->c + (fils->d - fils->c) * xinbta ( a, b, beta_log, u, &ifault );
	}
        else
        {
            msg << "NISP - ERROR" << endl;
            msg << "Nisp(PdfChange) : Transformation Normale -> " << typefils
                    << " not available" << endl;
            NispOutPutMsg(msg.str(), 1);
            return (0.);
        }
    }
    else if (typemere == "LogNormale")
    {    // Mere est LogNormale
        if (x < mere->c)
        {
            msg << "NISP - ERROR" << endl;
            msg << "Nisp(PdfChange) : law (mother)is LogNormale and value " << x
                    << " < min" << endl;
            NispOutPutMsg(msg.str(), 1);
            return (0.);
        }
        double mean = mere->a;
        double sigma = mere->b;
        double mu = log(mean) - sigma * sigma / 2.;
        double z = (log(x - mere->c) - mu) / sigma; // z suit une loi normale centree reduite
        if (typefils == "Normale")
            y = a + b * z;
        else if (typefils == "Uniforme")
            y = a + (b - a) * (erfc(-z / sqrt(2.)) / 2.);
        else if (typefils == "Exponentielle")
            y = b -(1. / a) * log((erfc(-z / sqrt(2.)) / 2.));
        else if (typefils == "LogNormale")
            y = c + exp(log(a) - b * b / 2. + b * z);
        else if (typefils == "LogUniforme")
            y = exp(log(a) + log(b / a) * (erfc(-z / sqrt(2.)) / 2.));
	else if (typefils=="Beta")
	{
	  int ifault;
	  double u = erfc(-z/sqrt(2.))/2.; // u suit une loi uniforme dans [0,1]
	  double beta_log = alngam ( a, &ifault ) + alngam ( b, &ifault ) - alngam ( a + b, &ifault );
	  y = fils->c + (fils->d - fils->c) * xinbta ( a, b, beta_log, u, &ifault );
	}
	else
        {
            msg << "NISP - ERROR" << endl;
            msg << "Nisp(PdfChange) : Transformation LogNormale -> " << typefils
                    << " not available" << endl;
            NispOutPutMsg(msg.str(), 1);
            return (0.);
        }
    }
    else if (typemere == "Uniforme")
    {      // Mere est Uniforme
        double min = mere->a;
        double max = mere->b;
        if ((x < min) || (x > max))
        {
            msg << "NISP - ERROR" << endl;
            msg << "Nisp(PdfChange) : value " << x << " must be in [" << min
                    << "," << max << "]" << endl;
            NispOutPutMsg(msg.str(), 1);
            return (0.);
        }
        double z = (x - min) / (max - min);     // z suit une loi uniforme [0,1]
        if (typefils == "Uniforme")
            y = a + (b - a) * z;
        else if (typefils == "Normale")
            y = a + b * InvFcn(z);
        else if (typefils == "Exponentielle")
            y = b + (-1. / a) * log(z);
        else if (typefils == "LogNormale")
            y = c + exp(log(a) - b * b / 2. + b * InvFcn(z));
        else if (typefils == "LogUniforme")
            y = exp(log(a) + log(b / a) * z);
	else if (typefils=="Beta")
	{
	  int ifault;
	  double beta_log = alngam ( a, &ifault ) + alngam ( b, &ifault ) - alngam ( a + b, &ifault );
	  y = fils->c + (fils->d - fils->c) * xinbta ( a, b, beta_log, z, &ifault );
	}
        else
        {
            msg << "NISP - ERROR" << endl;
            msg << "Nisp(PdfChange) : Transformation Uniform -> " << typefils
                    << " not available" << endl;
            NispOutPutMsg(msg.str(), 1);
            return (0.);
        }
    }
    else if (typemere == "LogUniforme")
    {                // Mere est LogUniforme
        double min = mere->a;
        double max = mere->b;
        if (x <= 0.)
        {
            msg << "NISP - ERROR" << endl;
            msg << "Nisp(PdfChange) : law (mother)is LogUniform and value " << x
                    << " < 0" << endl;
            NispOutPutMsg(msg.str(), 1);
            return (0.);
        }
        double z = (log(x) - log(min)) / (log(max) - log(min)); // z suit une loi uniforme [0,1]
        if ((z < 0.) || (z > 1.))
        {
            msg << "NISP - ERROR" << endl;
            msg
                    << "Nisp(PdfChange) : law (mother) is LogUniform and error on value "
                    << x << endl;
            NispOutPutMsg(msg.str(), 1);
            return (0.);
        }
        if (typefils == "Uniforme")
            y = a + (b - a) * z;
        else if (typefils == "Normale")
            y = a + b * InvFcn(z);
        else if (typefils == "Exponentielle")
            y = b + (-1. / a) * log(z);
        else if (typefils == "LogNormale")
            y = c + exp(log(a) - b * b / 2. + b * InvFcn(z));
        else if (typefils == "LogUniforme")
            y = exp(log(a) + log(b / a) * z);
	else if (typefils=="Beta")
	{
	  int ifault;
	  double beta_log = alngam ( a, &ifault ) + alngam ( b, &ifault ) - alngam ( a + b, &ifault );
	  y = fils->c + (fils->d - fils->c) * xinbta ( a, b, beta_log, z, &ifault );
	}
        else
        {
            msg << "NISP - ERROR" << endl;
            msg << "Nisp(PdfChange) : Transformation LogUniform -> " << typefils
                    << " not available" << endl;
            NispOutPutMsg(msg.str(), 1);
            return (0.);
        }
    }
    else if (typemere == "Exponentielle")
    { // Mere est Exponentielle
        double lambda = mere->a;
	x = x - mere->b;
        if (x < 0)
        {
            msg << "NISP - ERROR" << endl;
            msg << "Nisp(PdfChange) : law (mother)is Exponential and value "
                    << x << " <= min" << endl;
            NispOutPutMsg(msg.str(), 1);
            return (0.);
        }
        double z = exp(-x / lambda); // z suit une loi uniforme [0,1] 31/03/2011
        if (typefils == "Uniforme")
            y = a + (b - a) * z;
        else if (typefils == "Normale")
            y = a + b * InvFcn(z);
        else if (typefils == "Exponentielle")
            y = b + (-1. / a) * log(z);
        else if (typefils == "LogNormale")
            y = c + exp(log(a) - b * b / 2. + b * InvFcn(z));
        else if (typefils == "LogUniforme")
            y = exp(log(a) + log(b / a) * z);
	else if (typefils=="Beta")
	{
	  int ifault;
	  double beta_log = alngam ( a, &ifault ) + alngam ( b, &ifault ) - alngam ( a + b, &ifault );
	  y = c + (d - c) * xinbta ( a, b, beta_log, z, &ifault );
	}
        else
        {
            msg << "NISP - ERROR" << endl;
            msg << "Nisp(PdfChange) : Transformation Exponential -> "
                    << typefils << " not available" << endl;
            NispOutPutMsg(msg.str(), 1);
            return (0.);
        }
    }
    else if (typemere == "Beta")
    { // Mere est Beta
      double z = (x - mere->c) / (mere->d - mere->c); // z suit une loi beta (a,b) dans [0,1];
      int ifault;
      double beta_log = alngam ( mere->a, &ifault ) + alngam ( mere->b, &ifault ) - alngam ( mere->a + mere->b, &ifault );
      double u = betain(z, mere->a, mere->b, beta_log, &ifault); // u suit une loi uniforme dans [0,1]
      if (typefils == "Uniforme")
            y = a + (b - a) * u;
        else if (typefils == "Normale")
            y = a + b * InvFcn(u);
        else if (typefils == "Exponentielle")
            y = b + (-1. / a) * log(u);
        else if (typefils == "LogNormale")
            y = c + exp(log(a) - b * b / 2. + b * InvFcn(u));
        else if (typefils == "LogUniforme")
            y = exp(log(a) + log(b / a) * u);
	else if (typefils=="Beta")
	{
	  int ifault;
	  double beta_log = alngam ( a, &ifault ) + alngam ( b, &ifault ) - alngam ( a + b, &ifault );
	  y = c + (d - c) * xinbta ( a, b, beta_log, u, &ifault );
	}
        else
        {
            msg << "NISP - ERROR" << endl;
            msg << "Nisp(PdfChange) : Transformation Beta -> "
                    << typefils << " not available" << endl;
            NispOutPutMsg(msg.str(), 1);
            return (0.);
        }
    }
    else
    {
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(PdfChange) : law " << typemere << " unknown" << endl;
        NispOutPutMsg(msg.str(), 1);
        return (0.);
    }
    return (y);
}
// Monte Carlo
void montecarlo(SetRandomVariable *gva, int np)
{
    gva->degre_max = -1;
    int nx = gva->nx;
    if (nx < 1)
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(montecarlo) : stochastic dimension " << nx << " < 1"
                << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    gva->np = np;
    gva->x = dmatrix(np + 1, nx + 1);
    gva->w = dvector(np + 1);
    int k, j;
    double pas = 1. / (double) np;
    for (k = 1; k <= np; k++)
        for (j = 1; j <= nx; j++)
            gva->x[k][j] = gva->va[j - 1]->GetValue();
    for (k = 1; k <= np; k++)
        gva->w[k] = pas;
}
// Latin Hypercube Sampling
void lhs(SetRandomVariable *gva, int np)
{
    gva->degre_max = -1;
    int nx = gva->nx;
    if (nx < 1)
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(lhs) : stochastic dimension " << nx << " < 1" << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    gva->np = np;
    gva->x = dmatrix(np + 1, nx + 1);
    gva->w = dvector(np + 1);
    int k, j;
    double pas = 1. / (double) np;
    vector<double> v(np);
    for (k = 0; k < np; k++)
        v[k] = (0.5 + (double) k) * pas;
    for (j = 1; j <= nx; j++)
    {
        random_shuffle(v.begin(), v.end());
        for (k = 1; k <= np; k++)
            gva->x[k][j] = PdfChange(gva->va[j - 1], v[k - 1]);
    }
    for (k = 1; k <= np; k++)
        gva->w[k] = pas;
}
// Quasi Monte Carlo Sobol sequence (Numerical Recipies)
void qmcSobol(SetRandomVariable *gva, int np)
{
    gva->degre_max = -1;
    int nx = gva->nx;
    if (nx < 1)
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(qmcSobol) : stochastic dimension " << nx << " < 1" << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    gva->np = np;
    gva->x = dmatrix(np + 1, nx + 1);
    gva->w = dvector(np + 1);

    void i8_sobol(int dim_num, long long int *seed, double quasi[]);
    long long int seed = 0;

    double * x = dvector(nx);
    double pas = 1. / (double) np;
    int k, j;
    i8_sobol(nx, &seed, x); // on passe le premier point qui est toujours x[i]=0 i=0,1,...,nx-1
    for (k = 1; k <= np; k++)
    {
        i8_sobol(nx, &seed, x);
        for (j = 1; j <= nx; j++)
            gva->x[k][j] = PdfChange(gva->va[j - 1], x[j - 1]);
        gva->w[k] = pas;
    }
    free_dvector(x);
}
// Latin Hypercube Sampling critere Max Min
void lhsmaxmin(SetRandomVariable *gva, int np, int ne)
{
    int i, j, k, ii, kk, k1;
    double **xt, dmin, dmaxmin, d, dx;
    int nx = gva->nx;
    if (nx < 1)
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(lhsmaxmin) : stochastic dimension " << nx << " < 1"
                << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    gva->np = np;
    gva->x = dmatrix(np + 1, nx + 1);
    gva->w = dvector(np + 1);
    gva->degre_max = -1;
    vector<double> v(np);
    double pas = 1. / ((double) np);
    for (k = 0; k < np; k++)
        v[k] = (0.5 + (double) k) * pas;
    dmaxmin = 0.;
    xt = dmatrix(np + 1, nx + 1);
    for (int b = 1; b <= ne; b++)
    {
        for (j = 1; j <= nx; j++)
        {
            random_shuffle(v.begin(), v.end());
            for (k = 1; k <= np; k++)
                xt[k][j] = v[k - 1];
        }
        // MaxMin
        dmin = 1.e45;
        for (k = 1; k <= np; k++)
        {
            for (k1 = k + 1; k1 <= np; k1++)
            {
                for (d = 0., i = 1; i <= nx; i++)
                {
                    dx = xt[k][i] - xt[k1][i];
                    d += dx * dx;
                }
                if (d < dmin)
                    dmin = d;
            }
        }
        if (dmaxmin < dmin)
        { // Sauvegarde du plan Maximisant la plus petite distance entre 2 points
            dmaxmin = dmin;
            for (ii = 1; ii <= nx; ii++)
                for (kk = 1; kk <= np; kk++)
                    gva->x[kk][ii] = xt[kk][ii];
        }
    }
    // Changements de variable
    for (i = 1; i <= nx; i++)
        for (k = 1; k <= np; k++)
            gva->x[k][i] = PdfChange(gva->va[i - 1], gva->x[k][i]);
    free_dmatrix(xt, np + 1);
    for (k = 1; k <= np; k++)
        gva->w[k] = pas;
}
// Test des variables stochastiques
void teststo(string type, double a, double b, double c, double d)
{
    int test = 0;
    if (type == "Normale")
    {
        if ((a == 0.) && (b = 1.))
            test = 1;
    }
    else if (type == "Uniforme")
    {
        if ((a == 0.) && (b = 1.))
            test = 1;
    }
    else if (type == "Exponentielle")
    {
        if ((a == 1.) && (b = 1.))
            test = 1;
    }
    else if (type == "Beta")
    {
        if ((c == 0.) && (d == 1.))
            test = 1;
    }
    if (!test)
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(teststo) : variables are not stochastic " << type << " ";
	if (type == "Beta") 
          msg << c << " " << d << endl;
	else
	  msg << a << " " << b << endl;
        NispOutPutMsg(msg.str(), 1);
    }
}
// Test des variables stochastiques uniformes
void testnor(string type, double a, double b)
{
    int test = 0;
    if (type == "Uniforme")
    {
        if ((a == 0.) && (b = 1.))
            test = 1;
    }
    if (!test)
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(teststo) : variables are not stochastic-uniform " << type
                << " " << a << " " << b << endl;
        NispOutPutMsg(msg.str(), 1);
    }
}
// Quadrature Tensorisee
void quadrature(SetRandomVariable *gva, int degre)
{
    int nx = gva->nx;
    int nq = degre + 1; // le nombre de points doit etre egal au degre + 1

    if (nx < 1)
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(quadrature) : stochastic dimension = " << nx << " < 1"
                << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    if (degre < 0)
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(quadrature) : degree of quadrature == " << degre << " < 0"
                << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    // degr�
    gva->degre_max = degre;
    // On verifie que les variables aleatoires sont normalisees
    int j, np;
    for (j = 1; j <= nx; j++)
    {
        teststo(gva->va[j - 1]->type, gva->va[j - 1]->a, gva->va[j - 1]->b, gva->va[j - 1]->c, gva->va[j - 1]->d);
    }
    for (np = 1, j = 1; j <= nx; j++)
        np = np * nq;
    gva->np = np;
    gva->x = dmatrix(np + 1, nx + 1);
    gva->w = dvector(np + 1);
    // Calcul la quadrature tensorisee
    void Quadrature(double x[], double w[], int nq, string type, double a, double b);
    void QuadratureTensorise(double **yt, double *wt, double **xq, double **wq,
            int nx, int nq, int nt);
    double **xqt, **wqt;
    xqt = dmatrix(nx + 1, nq + 1);
    wqt = dmatrix(nx + 1, nq + 1);
    for (j = 1; j <= nx; j++)
        Quadrature(xqt[j], wqt[j], nq, gva->va[j - 1]->type, gva->va[j - 1]->a, gva->va[j - 1]->b);
    QuadratureTensorise(gva->x, gva->w, xqt, wqt, nx, nq, np);
    free_dmatrix(xqt, nx + 1);
    free_dmatrix(wqt, nx + 1);
}
// Quadrature Tensorisee Incomplete (Smolyak, Petras)
// Le niveau correspond au degre max des PC qu'on peut calculer
void petras(SetRandomVariable *gva, int degre)
{
    int niveau = degre;
    if (degre < 1)
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(petras) : degree of cubature == " << degre << " < 1"
                << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    if (niveau > 40)
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(petras) : level of cubature == " << niveau
                << " > 40 (value max of the level)" << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    int k, icpt, i, j, nnx, q, nquad;
    double *ps, *ws;
    void size_smolyak(int *, int *, int *);
    void quad_smolyak(int *, int *, double *, double *);
    int nx = gva->nx;
    if (nx < 1)
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(petras) : stochastic dimension = " << nx << " < 1" << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    // Degre maximum
    gva->degre_max = degre;
    // On verifie que les variables aleatoires sont normalisees
    for (j = 1; j <= nx; j++)
        teststo(gva->va[j - 1]->type, gva->va[j - 1]->a, gva->va[j - 1]->b, gva->va[j - 1]->c, gva->va[j - 1]->d);
    nnx = nx;
    q = niveau + nnx;
    size_smolyak(&nnx, &q, &nquad);
    ws = dvector(nquad + 1);
    ps = dvector(nquad * nnx + 1);
    quad_smolyak(&nx, &q, ps, ws);
    gva->np = nquad;
    gva->x = dmatrix(nquad + 1, nx + 1);
    gva->w = dvector(nquad + 1);
    for (k = 1, icpt = 0, i = 0; i < nquad; i++)
    {
        for (j = 1; j <= gva->nx; j++)
            gva->x[k][j] = PdfChange(gva->va[j - 1], ps[icpt++]);
        gva->w[k] = ws[i];
        k++;
    }
    free_dvector(ws);
    free_dvector(ps);
}
// Calcul des coefficients par integration (quadrature de Gauss)
void CoefficientsIntegration(double **beta, double **Z, double *w, double **y,
        int p, int np, int ny)
{
    int i, j, k;
    double s;
    for (j = 1; j <= ny; j++)
    {
        for (k = 0; k <= p; k++)
        {
            for (s = 0., i = 1; i <= np; i++)
                s += Z[i][k] * w[i] * y[i][j];
            beta[j][k] = s;
        }
    }
}

void CoefficientsRegression(double **beta, double *Z, double **y, double **errLoo, double *eqmLoo, int p, int np,
        int ny)
{
    int i, j, k;

    int m = np;
    int n = p + 1;
    int lda = m;
    int ldv = n;
    int ldu = m;
    int job = 21;

    //double s[m+n], e[m+n], u[m*n], v[n*n], work[m],tmp[n];
    double * s = new double[m + n];
    double * e = new double[m + n];
    double * u = new double[m * n];
    double * v = new double[n * n];
    double * work = new double[m];
    double * tmp = new double[n];
    double * lev = new double[m];

    int dsvdc(double a[], int lda, int m, int n, double s[], double e[],
            double u[], int ldu, double v[], int ldv, double work[], int job);

    // Verification via Scilab
//     ofstream sortie((char *) "verifScilab.dat", ios::out);
//     if (!sortie)
//     {
//         ostringstream msg;
//         msg << "NISP - ERROR" << endl;
//         msg << "Nisp(xx): Problem when open file tata" << endl;
//         NispOutPutMsg(msg.str(), 1);
//         return;
//     }
//     sortie << "Z = [";
//     for (i = 1; i <= m; i++) {
//       for(j = 1; j <= n; j++) sortie << Z[(i-1) + (j-1)*m] << " ";
//       sortie << endl;
//     }
//     sortie << "];" << endl;
//     sortie << "y = [";
//     for (i = 1; i <= m; i++) sortie << y[i][1] << " ";
//     sortie << endl;
// 
//     sortie << "];" << endl;
//     
//     sortie << "y2 = [";
//     for (i = 1; i <= m; i++) sortie << y[i][2] << " ";
//     sortie << endl;
// 
//     sortie << "];" << endl;
//     
//     sortie.close();
	
    //////////////////////////////////////////////
    
    dsvdc(Z, lda, m, n, s, e, u, ldu, v, ldv, work, job);

    // Transformation des vecteurs de la matrice V (regularisation)
    double coef, alpha = 1.e-6;
    for (j = 1; j <= n; j++)
    {
        coef = 0.;
        if (s[j - 1])
            coef = s[j - 1] / (s[j - 1] * s[j - 1] + alpha);
        for (i = 1; i <= n; i++)
            v[(i - 1) + (j - 1) * n] = coef * v[(i - 1) + (j - 1) * n];
    }
    double st;
    double tolmin = 1.e-8;
    double tolmax = 1. - tolmin;
    // calcul des leviers communs à toutes les sorties
    // U(i,j) = u[(i-1)+(j-1)*m]
    for (i = 1; i <= m; i++) {
      for (st=0., j = 1; j <= n; j++) st += u[(i - 1) + (j - 1) * m] * u[(i - 1) + (j - 1) * m];
      // On contraint les leviers dans [tolmin > 0., tolmax < 1.]
      if      ( st > tolmax ) st = tolmax;
      else if ( st < tolmin ) st = tolmin;
      lev[i - 1] = st;
    }
    
    // Calcul des leviers verifié le 18/02/2016 à 14h30
    
    for (j = 1; j <= ny; j++)
    {
        for (i = 1; i <= n; i++)
        {
            for (st = 0., k = 1; k <= m; k++) st += u[(k - 1) + (i - 1) * m] * y[k][j];
            tmp[i - 1] = st;
        }
        for (k = 1; k <= n; k++)
        {
            for (st = 0., i = 1; i <= n; i++)
                st += v[(k - 1) + (i - 1) * n] * tmp[i - 1];
            beta[j][k - 1] = st;
        }
        
        // JMM 18/02/2016
        // Calcul de U * (Ut * y) = U * tmp
        for (eqmLoo[j]=0., k = 1; k <= m; k++)
	{
	  st=0.;
	  for (st = 0., i = 1; i <= n; i++) st += u[(k - 1) + (i - 1) * m] * tmp[i-1];
	  errLoo[k][j] = (y[k][j] - st) / ( 1. - lev[k - 1]);
	  eqmLoo[j]   += errLoo[k][j] * errLoo[k][j];
	}
	eqmLoo[j] = eqmLoo[j] / m;
    }

    delete[] s;
    delete[] e;
    delete[] u;
    delete[] v;
    delete[] work;
    delete[] tmp;
    delete[] lev;
}

// Utiliser par la methode pc::sobol()
int Puissance2(int n)
{
    int i, ip;
    for (ip = 1, i = 1; i <= n; i++)
        ip = ip * 2;
    return (ip);
}
// Lecture des simulations
void ReadTarget2(double **y, int np, int ny, char *fichier)
{
    int k, j, nlignes, ncolonnes;
    ifstream entree(fichier, ios::in);
    if (!entree)
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(ReadTarget2) : problem to open the file " << fichier
                << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    string str;
    entree >> str >> nlignes;
    entree >> str >> ncolonnes;
    if (nlignes != np)
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(ReadTarget2) : number of simulations " << nlignes << " != "
                << np << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    if (ncolonnes != ny)
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(ReadTarget2) : number of output " << ncolonnes << " != "
                << ny << endl;
        NispOutPutMsg(msg.str(), 1);
        return;
    }
    for (k = 1; k <= nlignes; k++)
    {
        for (j = 1; j <= ncolonnes; j++)
            entree >> y[k][j];
    }
    entree.close();
}
// Utilitaires
int CalculP(int nx, int no)
{
    long min, max, num, den, i, p;
    if (nx > no)
    {
        min = (long) no;
        max = (long) nx;
    }
    else
    {
        min = (long) nx;
        max = (long) no;
    }
    for (num = 1, i = max + 1; i <= max + min; i++)
        num *= i;
    for (den = 1, i = 2; i <= min; i++)
        den *= i;
    p = num / den - 1;
    return ((int) p);
}

// Probleme grandes valeurs
double binomial(int n, int j)
{
    int k;
    double num, den;
    for (num = 1., k = n - j + 1; k <= n; k++)
        num = num * k;
    for (den = 1., k = 1; k <= j; k++)
        den = den * k;
    return (num / den);
}

