//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   nisp_va.cpp
 \author Jean-Marc Martinez <jean-marc.martinez@cea.fr>
 \date   2011/04/07 14:33:00

 \brief
 */
#include <iostream>
#include <sstream>
#include <fstream>
#include <string>
#include <vector>

using namespace std;

#include "nisp_va.h"
#include "nisp_util.h"

double Normale(double, double, double, double);
double LogNormale(double, double, double, double);
double Uniforme(double, double, double, double);
double LogUniforme(double, double, double, double);
double Exponentielle(double, double, double, double);
double Beta(double, double, double, double);

// ===========
// M�thodes RandomVariable
// ===========
RandomVariable::RandomVariable(string name)
{
    type = name;
    if (type == "Normale")
    {
        function = Normale;
        a = 0.;
        b = 1.;
	c = 0.;
	d = 1.;
    }
    else if (type == "Uniforme")
    {
        function = Uniforme;
        a = 0.;
        b = 1.;
	c = 0.;
	d = 1.;
    }
    else if (type == "Exponentielle")
    {
        function = Exponentielle;
        a = 1.;
        b = 0.;
	c = 0.;
	d = 1.;
    }
    else if (type == "LogNormale")
    {
        function = LogNormale;
        a = 0.1;
        b = 1.;
	c = 0.;
	d = 1.;
    }
    else if (type == "LogUniforme")
    {
        function = LogUniforme;
        a = 0.1;
        b = 1.;
	c = 0.;
	d = 1.;
    }
    else if (type == "Beta")
    {
        function = Beta;
        a = 3.;
        b = 3.;
	c = 0.;
	d = 1.;
    }
    else
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(RandomVariable::RandomVariable) : law " << type
                << " unknown" << endl;
        NispOutPutMsg(msg.str(), 1);
    }
}
RandomVariable::~RandomVariable()
{
}
RandomVariable::RandomVariable(string name, double x)
{
    type = name;
    a = x;
    b = 0.;
    if (type == "Exponentielle")
        function = Exponentielle;
    else
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(RandomVariable::RandomVariable) : law " << type
                << " and one parameter" << endl;
        NispOutPutMsg(msg.str(), 1);
    }
}

RandomVariable::RandomVariable(string name, double a1, double b1, double c1, double d1)
{
    type = name;
    a = a1;
    b = b1;
    c = c1;
    d = d1;
    ostringstream msg;
    if ((a1 <= 0.) || (b1 <= 0.))
        {
	    msg << "NISP - ERROR" << endl;
            msg << "Nisp(RandomVariable::RandomVariable) : law " << type
                    << " and one of parameters are " << a1 << " , " << b1
                    << " <=0" << endl;
            NispOutPutMsg(msg.str(), 1);
        }
    if (d1 < c1)
        {
	    msg << "NISP - ERROR" << endl;
            msg << "Nisp(RandomVariable::RandomVariable) : law " << type
                    << " and its min  = " << c1 << " > its max = " << d1 << endl;
            NispOutPutMsg(msg.str(), 1);
        }
    
    if (type == "Beta")
        function = Beta;
    else
    {
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(RandomVariable::RandomVariable) : law " << type
                << " and four parameter" << endl;
        NispOutPutMsg(msg.str(), 1);
    }
}

RandomVariable::RandomVariable(string name, double x, double y)
{ // tester les valeurs a=min < b=max loi uniforme par exemple ...
    type = name;
    a = x;
    b = y;
    ostringstream msg;
    if (type == "Normale")
    {
        function = Normale;
        if (y < 0)
        {
            msg << "NISP - ERROR" << endl;
            msg << "Nisp(RandomVariable::RandomVariable) : law " << type
                    << " and standard error " << b << " < 0 " << endl;
            NispOutPutMsg(msg.str(), 1);
        }
    }
    else if (type == "Uniforme")
    {
        function = Uniforme;
        if (x >= y)
        {
            msg << "NISP - ERROR" << endl;
            msg << "Nisp(RandomVariable::RandomVariable) : law " << type
                    << " and min = " << x << " >= max = " << y << endl;
            NispOutPutMsg(msg.str(), 1);
        }
    }
    else if (type == "Exponentielle")
    {
        function = Exponentielle;
	if (x <= 0.)
        {
            msg << "NISP - ERROR" << endl;
            msg << "Nisp(RandomVariable::RandomVariable) : law " << type
                    << " and one of parameters are " << x << " <=0" << endl;
            NispOutPutMsg(msg.str(), 1);
        }
        b = y;
        c = 0.;
	d = 1.;
    }
    else if (type == "Beta")
    {
        function = Beta;
	if ((x <= 0.) || (y <= 0.))
        {
            msg << "NISP - ERROR" << endl;
            msg << "Nisp(RandomVariable::RandomVariable) : law " << type
                    << " and one of parameters are " << x << " , " << y
                    << " <=0" << endl;
            NispOutPutMsg(msg.str(), 1);
        }
        c = 0.;
	d = 1.;
    }
    else if (type == "LogNormale")
    {
        function = LogNormale;
        if ((x <= 0.) || (y <= 0.))
        {
            msg << "NISP - ERROR" << endl;
            msg << "Nisp(RandomVariable::RandomVariable) : law " << type
                    << " and one of parameters are " << x << " , " << y
                    << " <=0" << endl;
            NispOutPutMsg(msg.str(), 1);
        }
        c=0.;
        d=0.;
    }
    else if (type == "LogUniforme")
    {
        function = LogUniforme;
        if ((x <= 0.) || (y <= 0.))
        {
            msg << "NISP - ERROR" << endl;
            msg << "Nisp(RandomVariable::RandomVariable) : law " << type
                    << " and one of parameters are " << x << " , " << y
                    << " <=0" << endl;
            NispOutPutMsg(msg.str(), 1);
        }
        if (x > y)
        {
            msg << "NISP - ERROR" << endl;
            msg << "Nisp(RandomVariable::RandomVariable) : law " << type
                    << " and min = " << x << " > max = " << y << endl;
            NispOutPutMsg(msg.str(), 1);
        }
        c=0.;
	d=0.;
    }
    else
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(RandomVariable::RandomVariable) : law " << type
                << " and 2 parameters" << endl;
        NispOutPutMsg(msg.str(), 1);
    }
}

RandomVariable::RandomVariable(string name, double x, double y, double min)
{
    type = name;
    a = x;
    b = y;
    c = min;
    d = 0.;
    ostringstream msg;
    if (type == "LogNormale")
    {
      function = LogNormale;
      if ((x <= 0.) || (y <= 0.))
        {
            msg << "NISP - ERROR" << endl;
            msg << "Nisp(RandomVariable::RandomVariable) : law " << type
                    << " and one of parameters are " << x << " , " << y
                    << " <=0" << endl;
            NispOutPutMsg(msg.str(), 1);
        }
    }
    else
    {
        ostringstream msg;
        msg << "NISP - ERROR" << endl;
        msg << "Nisp(RandomVariable::RandomVariable) : law " << type
                << " and 3 parameters" << endl;
        NispOutPutMsg(msg.str(), 1);
    }
}


double RandomVariable::GetValue()
{
    return (function(a, b, c ,d));
}
double RandomVariable::GetValue(RandomVariable *mere, double x)
{
    double PdfChange(RandomVariable *, RandomVariable *, double);
    return (PdfChange(this, mere, x));
}
void RandomVariable::GetLog()
{
    ostringstream msg;
    msg << "***********************************************" << endl;
    msg << "Nisp(RandomVariable::GetLog) for RandomVariable" << endl;
    msg << type << " : " << a << " : " << b << " : " << c << " : " << d << endl;
    msg << "***********************************************" << endl;
    NispOutPutMsg(msg.str(), 0);
}

