//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTGPBuilder.cxx
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 \date   $Date$

 \brief  Test suite for class URANIE::Modeler::TGPBuilder and its children classes.



 */

#define DBL_PRECISION_COMPARE 1.e-10
#define DBL_PRECISION_COMPARE2 1.e-16
        
#include "TestTGPBuilder.h"
#include "TString.h"  
#include "TRandom3.h"
#ifdef ROOT5
#include "Api.h"
#endif

#include "TGPBuilder.h"
#include "TGPMLCostFunction.h"
#include "TGPReMLCostFunction.h"
#include "TGPLooCostFunction.h"
#include "TIsotropicGaussianCorrFunction.h"
#include "TMaternIIICorrFunction.h"
#include "TMaternIICorrFunction.h"
#include "TMaternICorrFunction.h"
#include "TMatern52CorrFunction.h"
#include "TMatern72CorrFunction.h"
#include "TExponentialCorrFunction.h"

#include "TUniformDistribution.h"
#include "TSampling.h"
#include "TLauncherFunction.h"
#include "uranie_commands.h"

#include <iostream>
#include <string>
#include <iomanip>
#include <cmath>
using namespace std;
using namespace URANIE::Modeler;
using namespace URANIE::DataServer;
using namespace URANIE::Sampler;
using namespace URANIE::Launcher;

TGPBuilderTest::TGPBuilderTest()
{

}

TGPBuilderTest::~TGPBuilderTest()
{

}

void TGPBuilderTest::setUp()
{
    tdsObs = new TDataServer("tdsObs","");
    _workingTestPath = gSystem->pwd();
    
    tdsObs->fileDataRead("utf-1D-train.dat");
}

void TGPBuilderTest::tearDown()
{
    delete tdsObs;
    tdsObs=NULL;
    UranieTests::removeFile(_workingTestPath.Data(),"krig-model.dat");
    UranieTests::removeFile(_workingTestPath.Data(),"gpBuilder-optim.dat");
    UranieTests::removeFile(_workingTestPath.Data(),"gpBuilder-screening.dat");
}

////////////////////////////////
///                          ///
///         Tests            ///
///                          ///
////////////////////////////////

void TGPBuilderTest::testConstruction()
{
    Double_t *mat;
    TGPBuilder *gpb;
    
    // test the default constructor
    gpb = new TGPBuilder();    
    CPPUNIT_ASSERT( !gpb->hasCorrFunction() );
    CPPUNIT_ASSERT(  gpb->hasVarianceInFactor() );
    CPPUNIT_ASSERT( !gpb->hasMeasurementError() );    
    CPPUNIT_ASSERT( !gpb->hasTrend() );
    CPPUNIT_ASSERT( !gpb->usePrior() );
    delete gpb; gpb = NULL;

    // test the constructor with the correlation function object
    TIsotropicGaussianCorrFunction *corrFunc = new TIsotropicGaussianCorrFunction(1, 0.5);
    gpb = new TGPBuilder(tdsObs, "x1", "y", corrFunc);
    CPPUNIT_ASSERT(  gpb->hasCorrFunction() );
    CPPUNIT_ASSERT(  gpb->hasVarianceInFactor() );
    CPPUNIT_ASSERT( !gpb->hasMeasurementError() );    
    CPPUNIT_ASSERT( !gpb->hasTrend() );
    CPPUNIT_ASSERT( !gpb->usePrior() );
                       
    CPPUNIT_ASSERT(  gpb->getCorrFunction() != NULL );     
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.5, gpb->getCorrFunction()->getParameters()[0], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_EQUAL( 1, gpb->getNInputs());
    CPPUNIT_ASSERT_EQUAL( 6, gpb->getNObs());
    CPPUNIT_ASSERT_EQUAL( 0, gpb->getNTrendParams());
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.6715423728261041, gpb->getObsInputs()[0],      DBL_PRECISION_COMPARE); // this is a normalised value, not equal to the one in tdsObs.
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.3431918,          gpb->getObsOutput()[0],     DBL_PRECISION_COMPARE); // this is the same value as in tdsObs
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.02215734,         gpb->getObsNormParams()[0], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.9915253,          gpb->getObsNormParams()[1], DBL_PRECISION_COMPARE);
    
    mat = gpb->getCovarianceMatrix();
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.0,                 mat[0], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.9795727921033551,  mat[1], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.8118517087245244,  mat[2], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.16465932540717815, mat[3], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.6495096071351415,  mat[4], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.3575280679165563,  mat[5], DBL_PRECISION_COMPARE);
    
    mat = gpb->getCorrelationMatrix();
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.0,                 mat[0], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.9795727921033551,  mat[1], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.8118517087245244,  mat[2], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.16465932540717815, mat[3], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.6495096071351415,  mat[4], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.3575280679165563,  mat[5], DBL_PRECISION_COMPARE);

    mat = gpb->getMeasurementErrorCovMatrix();    
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.0, mat[0], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.0, mat[1], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.0, mat[2], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.0, mat[3], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.0, mat[4], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.0, mat[5], DBL_PRECISION_COMPARE);
        
    CPPUNIT_ASSERT( gpb->getTrendMatrix()      == NULL ); 
    CPPUNIT_ASSERT( gpb->getTrendFormula()          == NULL );       
    CPPUNIT_ASSERT( gpb->getTrendPriorMeans()  == NULL );
    CPPUNIT_ASSERT( gpb->getTrendPriorCovMat() == NULL );
    CPPUNIT_ASSERT( strcmp(gpb->getTrendString().Data(), "") == 0 );
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.0, gpb->getVariance(),       DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.0, gpb->getAlpha(),          DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.0, gpb->getRegularisation(), DBL_PRECISION_COMPARE);

    delete gpb; gpb = NULL;
    
    // test the constructor with the correlation function string            
    gpb = new TGPBuilder(tdsObs, "x1", "y", "isogauss");
    CPPUNIT_ASSERT(  gpb->hasCorrFunction() );
    CPPUNIT_ASSERT(  gpb->hasVarianceInFactor() );
    CPPUNIT_ASSERT( !gpb->hasMeasurementError() );    
    CPPUNIT_ASSERT( !gpb->hasTrend() );
    CPPUNIT_ASSERT( !gpb->usePrior() );
    
    CPPUNIT_ASSERT(  gpb->getCorrFunction() != NULL );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.1, gpb->getCorrFunction()->getParameters()[0], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_EQUAL( 1, gpb->getNInputs() );
    CPPUNIT_ASSERT_EQUAL( 6, gpb->getNObs() );
    CPPUNIT_ASSERT_EQUAL( 0, gpb->getNTrendParams() );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.6715423728261041, gpb->getObsInputs()[0],      DBL_PRECISION_COMPARE); // this is a normalised value, not equal to the one in tdsObs.
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.3431918,          gpb->getObsOutput()[0],     DBL_PRECISION_COMPARE); // this is the same value as in tdsObs
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.02215734,         gpb->getObsNormParams()[0], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.9915253,          gpb->getObsNormParams()[1], DBL_PRECISION_COMPARE);
    
    mat = gpb->getCovarianceMatrix();
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.0,                    mat[0], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.5969223627713421,     mat[1], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.005456544824590714,   mat[2], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 2.5981145523491532e-20, mat[3], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 2.0636663452106608e-05, mat[4], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 6.803831497717578e-12,  mat[5], DBL_PRECISION_COMPARE);
    
    mat = gpb->getCorrelationMatrix();
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.0,                    mat[0], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.5969223627713421,     mat[1], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.005456544824590714,   mat[2], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 2.5981145523491532e-20, mat[3], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 2.0636663452106608e-05, mat[4], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 6.803831497717578e-12,  mat[5], DBL_PRECISION_COMPARE);

    mat = gpb->getMeasurementErrorCovMatrix();    
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.0, mat[0], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.0, mat[1], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.0, mat[2], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.0, mat[3], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.0, mat[4], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.0, mat[5], DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT( gpb->getTrendMatrix()      == NULL );
    CPPUNIT_ASSERT( gpb->getTrendFormula()          == NULL );  
    CPPUNIT_ASSERT( gpb->getTrendPriorMeans()  == NULL );
    CPPUNIT_ASSERT( gpb->getTrendPriorCovMat() == NULL );
    CPPUNIT_ASSERT( strcmp(gpb->getTrendString().Data(), "") == 0 );
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.0, gpb->getVariance(),       DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.0, gpb->getAlpha(),          DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.0, gpb->getRegularisation(), DBL_PRECISION_COMPARE);
    
    delete gpb;
    delete corrFunc;
}

void TGPBuilderTest::testTrend()
{
    Double_t *mat;
    TObjArray *array;
    TGPBuilder *gpb;
    
    // GPBuilder with a constant trend
    gpb = new TGPBuilder(tdsObs, "x1", "y", "isogauss", "const");
    CPPUNIT_ASSERT( gpb->hasTrend() );
    
    CPPUNIT_ASSERT( gpb->getTrendString() == "const");
    CPPUNIT_ASSERT_EQUAL( 1, gpb->getNTrendParams() );

    array = gpb->getTrendFormula();
    CPPUNIT_ASSERT_EQUAL( 1, array->GetEntries() );
    CPPUNIT_ASSERT( strcmp(array->At(0)->GetName(), "1") == 0 );        
    
    mat = gpb->getTrendMatrix();
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.0, mat[0], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.0, mat[1], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.0, mat[2], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.0, mat[3], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.0, mat[4], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.0, mat[5], DBL_PRECISION_COMPARE);
                                         
    delete gpb; gpb = NULL;

    // GPBuilder with a linear trend
    gpb = new TGPBuilder(tdsObs, "x1", "y", "isogauss", "linear");
    CPPUNIT_ASSERT( gpb->hasTrend() );
    
    CPPUNIT_ASSERT( gpb->getTrendString() == "linear");
    CPPUNIT_ASSERT_EQUAL( 2, gpb->getNTrendParams() );

    array = gpb->getTrendFormula();
    CPPUNIT_ASSERT_EQUAL( 2, array->GetEntries() );
    CPPUNIT_ASSERT( strcmp(array->At(0)->GetName(), "1")  == 0 );
    CPPUNIT_ASSERT( strcmp(array->At(1)->GetName(), "x1") == 0 );
    
    mat = gpb->getTrendMatrix();
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.0,                mat[0], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.6715423728261041, mat[6], DBL_PRECISION_COMPARE);
    
    delete gpb; gpb = NULL;

    // GPBuilder with a custom trend
    gpb = new TGPBuilder(tdsObs, "x1", "y", "isogauss", "1:x1**2:sin(x1)");
    CPPUNIT_ASSERT( gpb->hasTrend() );
    
    CPPUNIT_ASSERT( gpb->getTrendString() == "1:x1**2:sin(x1)");
    CPPUNIT_ASSERT_EQUAL( 3, gpb->getNTrendParams() );
                                 
    array = gpb->getTrendFormula();
    CPPUNIT_ASSERT_EQUAL( 3, array->GetEntries() );
    CPPUNIT_ASSERT( strcmp(array->At(0)->GetName(), "1"      ) == 0 );
    CPPUNIT_ASSERT( strcmp(array->At(1)->GetName(), "x1**2"  ) == 0 );
    CPPUNIT_ASSERT( strcmp(array->At(2)->GetName(), "sin(x1)") == 0 );
    
    mat = gpb->getTrendMatrix();
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.0,                 mat[0],  DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.45310265064099997, mat[6],  DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.6234355211040987,  mat[12], DBL_PRECISION_COMPARE);
    
    delete gpb;    
}

void TGPBuilderTest::testBayesian()
{      
    Double_t *mat;
    TGPBuilder *gpb;
        
    // Construction of the GP builder    
    gpb = new TGPBuilder(tdsObs, "x1", "y", "isogauss", "linear");

    // Setting and checking of bayesian parameters    
    Double_t betaPrior[2] = {2.0, 0.0};
    Double_t qPrior[4]    = {1.0, 0.0, 0.0, 1.0};
    gpb->setPriorData(betaPrior, qPrior);    
    CPPUNIT_ASSERT( gpb->usePrior() );
    
    mat = gpb->getTrendPriorMeans();
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 2.0, mat[0], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.0, mat[1], DBL_PRECISION_COMPARE);
    
    mat = gpb->getTrendPriorCovMat();
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.0, mat[0], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.0, mat[1], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.0, mat[2], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.0, mat[3], DBL_PRECISION_COMPARE);

    // Construction and checking of a model using a priori knowledge
    TKriging* krig = gpb->buildGP();
    
    mat = krig->getGPParams();
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -1.17878263576558306,   mat[0], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(  0.300322748344696744,  mat[1], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -1.74195071357504316,   mat[2], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(  0.907577288491262624,  mat[3], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(  1.12603664208771215,   mat[4], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(  0.0747499242060974817, mat[5], DBL_PRECISION_COMPARE);    
    
    mat = krig->getTrendParams();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(  1.48795325378914156,   mat[0], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -0.202170992542105354,  mat[1], DBL_PRECISION_COMPARE);
    
    mat = krig->getTrendParamsCovMatrix();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(  2.52170044278196948e-01, mat[0], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -2.18985144429413664e-01, mat[1], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -2.18985144429413664e-01, mat[2], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(  5.40604677445862936e-01, mat[3], DBL_PRECISION_COMPARE);
    
    delete krig;
    delete gpb;    
}

void TGPBuilderTest::testAlpha()
{
    Double_t *mat;
    TGPBuilder *gpb;
        
    // Construction of the GP builder    
    gpb = new TGPBuilder(tdsObs, "x1", "y", "isogauss", "const");
    gpb->setHasMeasurementError();
    
    CPPUNIT_ASSERT( gpb->hasVarianceInFactor() );
    CPPUNIT_ASSERT( gpb->hasMeasurementError() );        

    // Setting of alpha and checking of the effects on the kriging model    
    gpb->setAlpha(0.1);
    TKriging* krig = gpb->buildGP();                  

    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.1, krig->getMeasurementErrorVariance(), DBL_PRECISION_COMPARE);
    
    mat = krig->getGPParams();
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -0.958467680514433429, mat[0], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(  0.20078834795275291,  mat[1], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -1.48234651616997959,  mat[2], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(  1.00033981176913445,  mat[3], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(  1.0238483198822701,   mat[4], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(  0.215837717080258756, mat[5], DBL_PRECISION_COMPARE);    

    delete krig;
    delete gpb;    
    
}

void TGPBuilderTest::testVMes()
{              
    Double_t *mat;
    TGPBuilder *gpb;
        
    // Construction of the GP builder    
    gpb = new TGPBuilder(tdsObs, "x1", "y", "isogauss", "const");
    gpb->setHasMeasurementError();
    
    CPPUNIT_ASSERT( gpb->hasVarianceInFactor() );
    CPPUNIT_ASSERT( gpb->hasMeasurementError() );        
    
    // Setting of the measurement error variance and checking of the effects
    gpb->setMeasurementErrorVariance(0.1);
    CPPUNIT_ASSERT( !gpb->hasVarianceInFactor() );    
    
    TKriging* krig = gpb->buildGP();                  

    CPPUNIT_ASSERT_DOUBLES_EQUAL( -1.0, krig->getMeasurementErrorVariance(), DBL_PRECISION_COMPARE);
    
    mat = krig->getGPParams();
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -0.958467680514433429, mat[0], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(  0.20078834795275291,  mat[1], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -1.48234651616997959,  mat[2], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(  1.00033981176913445,  mat[3], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(  1.0238483198822701,   mat[4], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(  0.215837717080258756, mat[5], DBL_PRECISION_COMPARE);    

    delete krig;
    delete gpb;    
    
}

void TGPBuilderTest::testSetMeasurementErrorCovMatrix()
{         
    Int_t nS = 6;
    Double_t *covMes = new Double_t[nS*nS];
    Double_t *mat;
    TGPBuilder *gpb;
        
    // Construction of the GP builder    
    gpb = new TGPBuilder(tdsObs, "x1", "y", "isogauss", "const");
    gpb->setHasMeasurementError();
    
    CPPUNIT_ASSERT( gpb->hasVarianceInFactor() );
    CPPUNIT_ASSERT( gpb->hasMeasurementError() );        
    
    for(Int_t iter=0;iter<nS*nS;iter++) 
        covMes[iter] = 0.0;
    for(Int_t iter=0;iter<nS;iter++) 
        covMes[iter*(nS+1)] = 0.1;
    
    // Setting of the measurement error variance and checking of the effects
    gpb->setMeasurementErrorCovMatrix(covMes);
    CPPUNIT_ASSERT( !gpb->hasVarianceInFactor() );    
    
    TKriging* krig = gpb->buildGP();                  

    CPPUNIT_ASSERT_DOUBLES_EQUAL( -1.0, krig->getMeasurementErrorVariance(), DBL_PRECISION_COMPARE);
    
    mat = krig->getGPParams();
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -0.958467680514433429, mat[0], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(  0.20078834795275291,  mat[1], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -1.48234651616997959,  mat[2], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(  1.00033981176913445,  mat[3], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(  1.0238483198822701,   mat[4], DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(  0.215837717080258756, mat[5], DBL_PRECISION_COMPARE);    

    delete krig;
    delete gpb;    
    delete[] covMes;
}


void TGPBuilderTest::testFindOptimalParam()
{
    TGPBuilder *gpb=NULL;
    TKriging *krig=NULL;


    // Find optimal parameter: 
    // no trend, no measurement error, no regularisation, 
    // ML cost function, no screening, NelderMead optimiser, 1000 run max, reset on.        
    gpb = new TGPBuilder(tdsObs, "x1", "y", "isogauss");
    gpb->findOptimalParameters("ML", 0, "NelderMead");
    //gpb->findOptimalParameters("ML", 0, "");
    krig = gpb->buildGP();
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 3.65686223283    , krig->getVariance()      , DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.491115509218985, krig->getCorrLengths()[0], DBL_PRECISION_COMPARE );   
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -5.94601187463 , krig->getGPParams()[0], DBL_PRECISION_COMPARE );
	
    delete krig; krig = NULL;
    delete gpb;  gpb  = NULL;

    // Find optimal parameter: 
    // linear trend, no measurement error, no regularisation, 
    // ML cost function, no screening, NelderMead optimiser, 1000 run max, reset on.        
    gpb = new TGPBuilder(tdsObs, "x1", "y", "isogauss", "linear");
    gpb->findOptimalParameters("ML", 0, "NelderMead");
    //gpb->findOptimalParameters("ML", 0, "");
    krig = gpb->buildGP();
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.68327350961    , krig->getVariance()      , DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.428347750498486, krig->getCorrLengths()[0], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.58715598328    , krig->getTrendParams()[0], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -0.0702569489215 , krig->getTrendParams()[1], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -5.88014756141   , krig->getGPParams()[0], DBL_PRECISION_COMPARE );

    delete krig; krig = NULL;
    delete gpb;  gpb  = NULL;

    // Find optimal parameter: 
    // linear trend, no measurement error, no regularisation, 
    // ReML cost function, no screening, NelderMead optimiser, 1000 run max, reset on.        
    gpb = new TGPBuilder(tdsObs, "x1", "y", "isogauss", "linear");
    gpb->findOptimalParameters("ReML", 0, "NelderMead");
    //gpb->findOptimalParameters("ReML", 0, "");
    krig = gpb->buildGP();
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 4.19154219765    , krig->getVariance()             , DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.48661703602961 , krig->getCorrLengths()[0], DBL_PRECISION_COMPARE );    
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.47037018446    , krig->getTrendParams()[0], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -0.0995859135223 , krig->getTrendParams()[1], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -5.6767123036    , krig->getGPParams()[0], DBL_PRECISION_COMPARE );

    delete krig; krig = NULL;
    delete gpb;  gpb  = NULL;

    // Find optimal parameter: 
    // linear trend, no measurement error, no regularisation, 
    // Loo cost function, no screening, NelderMead optimiser, 1000 run max, reset on.        
    gpb = new TGPBuilder(tdsObs, "x1", "y", "isogauss", "linear");
    gpb->findOptimalParameters("Loo", 0, "NelderMead");
    //gpb->findOptimalParameters("Loo", 0, "");
    krig = gpb->buildGP();
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.130138878453   , krig->getVariance()             , DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.406769136918203, krig->getCorrLengths()[0], DBL_PRECISION_COMPARE );    
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.59824759467    , krig->getTrendParams()[0], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -0.0581805388525 , krig->getTrendParams()[1], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -54.1624875703   , krig->getGPParams()[0], DBL_PRECISION_COMPARE );

    delete krig; krig = NULL;
    delete gpb;  gpb  = NULL;

    // Find optimal parameter: 
    // linear trend, no measurement error, regularisation, 
    // ML cost function, no screening, NelderMead optimiser, 1000 run max, reset on.        
    gpb = new TGPBuilder(tdsObs, "x1", "y", "isogauss", "linear");
    gpb->setRegularisation(1e-3);
    gpb->findOptimalParameters("ML", 0, "NelderMead");
    //gpb->findOptimalParameters("ML", 0, "");
    krig = gpb->buildGP();
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.43900643756    , krig->getVariance()      , DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.405925673195195, krig->getCorrLengths()[0], DBL_PRECISION_COMPARE );    
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.57190206842    , krig->getTrendParams()[0], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.0134875116779  , krig->getTrendParams()[1], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -3.22258939937   , krig->getGPParams()[0], DBL_PRECISION_COMPARE );

    delete krig; krig = NULL;
    delete gpb;  gpb  = NULL;

    // Find optimal parameter: 
    // linear trend, measurement error, no regularisation, 
    // ML cost function, no screening, NelderMead optimiser, 1000 run max, reset on.        
    gpb = new TGPBuilder(tdsObs, "x1", "y", "isogauss", "linear");
    gpb->setHasMeasurementError();
    gpb->findOptimalParameters("ML", 0, "NelderMead");
    //gpb->findOptimalParameters("ML", 0, "");
    krig = gpb->buildGP();
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.68271349638     , krig->getVariance()                , DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.68271349638e-06 , krig->getMeasurementErrorVariance(), DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.428304591774088 , krig->getCorrLengths()[0], DBL_PRECISION_COMPARE );    
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.58711005277     , krig->getTrendParams()[0], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -0.0700196982443  , krig->getTrendParams()[1], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -5.87334538274    , krig->getGPParams()[0], DBL_PRECISION_COMPARE );

    delete krig; krig = NULL;
    delete gpb;  gpb  = NULL;

    // Find optimal parameter: 
    // linear trend, user defined measurement error, no regularisation, 
    // ML cost function, no screening, NelderMead optimiser, 1000 run max, reset on.        
    gpb = new TGPBuilder(tdsObs, "x1", "y", "isogauss", "linear");
    gpb->setMeasurementErrorVariance(0.01);    
    gpb->findOptimalParameters("ML", 0, "NelderMead");
    //gpb->findOptimalParameters("ML", 0, "");
    krig = gpb->buildGP();
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.55274088697   , krig->getVariance()                , DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -1.0            , krig->getMeasurementErrorVariance(), DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.396084775804263, krig->getCorrLengths()[0], DBL_PRECISION_COMPARE );    
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.56250110491   , krig->getTrendParams()[0], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.0701784322164 , krig->getTrendParams()[1], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -1.22798527828  , krig->getGPParams()[0], DBL_PRECISION_COMPARE );

    delete krig; krig = NULL;
    delete gpb;  gpb  = NULL;
    
    // Find optimal parameter: 
    // linear trend, no measurement error, no regularisation, a priori knowledge 
    // ReML cost function, no screening, NelderMead optimiser, 1000 run max, reset on.
    Double_t betaPrior[2] = {1.0, 0.0};
    Double_t qPrior[4]    = {0.25, 0.0, 0.0, 0.01};
    gpb = new TGPBuilder(tdsObs, "x1", "y", "isogauss", "linear");
    gpb->setPriorData(betaPrior, qPrior);
    gpb->findOptimalParameters("ReML", 0, "NelderMead");
    //gpb->findOptimalParameters("ReML", 0, "");
    krig = gpb->buildGP();
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 4.19154219765     , krig->getVariance()                , DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.0               , krig->getMeasurementErrorVariance(), DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.48661703602961  , krig->getCorrLengths()[0], DBL_PRECISION_COMPARE );    
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.04929037645     , krig->getTrendParams()[0], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.000790735107588 , krig->getTrendParams()[1], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -5.76568450108    , krig->getGPParams()[0], DBL_PRECISION_COMPARE );

    delete krig; krig = NULL;
    delete gpb;  gpb  = NULL;
    
    // Find optimal parameter: 
    // no trend, no measurement error, no regularisation,  
    // ML cost function, 100 screening, NelderMead optimiser, 1000 run max, reset on.
    gpb = new TGPBuilder(tdsObs, "x1", "y", "isogauss");
    CPPUNIT_ASSERT_NO_THROW( gpb->findOptimalParameters("ML", 100, "NelderMead") );
    //CPPUNIT_ASSERT_NO_THROW( gpb->findOptimalParameters("ML", 100, "") );

    // Manual setting of parameters
    Double_t params[1] = {0.5};
    gpb->getCorrFunction()->setParameters(params);
    // Find optimal parameter: 
    // no trend, no measurement error, no regularisation,  
    // ML cost function, no screening, NelderMead optimiser, 1000 run max, reset off.    
    gpb->findOptimalParameters("ML", 0, "NelderMead", 1000, kFALSE);   
    //gpb->findOptimalParameters("ML", 0, "", 1000, kFALSE);  
    krig = gpb->buildGP();
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 3.6554640835    , krig->getVariance()             , DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.491061076936074, krig->getCorrLengths()[0], DBL_PRECISION_COMPARE );    
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -5.94293605949  , krig->getGPParams()[0], DBL_PRECISION_COMPARE );


    delete krig; krig = NULL;
    delete gpb;  gpb  = NULL;
    
   // Find optimal parameter: 
    // linear trend, no measurement error, no regularisation, 
    // ReML cost function, no screening, NelderMead optimiser, 1000 run max, reset on.        
    gpb = new TGPBuilder(tdsObs, "x1", "y", "maternIII", "linear");
    gpb->findOptimalParameters("ReML", 0, "NelderMead");
    //gpb->findOptimalParameters("ReML", 0, "");
    krig = gpb->buildGP();
   
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(  5.33028322137853   , krig->getVariance()             , DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.572161927517284 , krig->getCorrLengths()[0], DBL_PRECISION_COMPARE );    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(  1.61005197196293   , krig->getTrendParams()[0], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -0.0800027766867485 , krig->getTrendParams()[1], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL(  -5.14359799026159   , krig->getGPParams()[0], DBL_PRECISION_COMPARE );


    delete krig; krig = NULL;
    delete gpb;  gpb  = NULL;
    
   // Find optimal parameter: 
    // linear trend, no measurement error, no regularisation, 
    // ReML cost function, no screening, NelderMead optimiser, 1000 run max, reset on.        
    gpb = new TGPBuilder(tdsObs, "x1", "y", "maternI", "linear");
    gpb->findOptimalParameters("ReML", 0, "NelderMead");
    //gpb->findOptimalParameters("ReML", 0, "");
    krig = gpb->buildGP();
    
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(  5.33028322137853   , krig->getVariance()             , DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.572161927517284 , krig->getCorrLengths()[0], DBL_PRECISION_COMPARE );    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(  1.61005197196293   , krig->getTrendParams()[0], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -0.0800027766867485 , krig->getTrendParams()[1], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL(  -5.14359799026159   , krig->getGPParams()[0], DBL_PRECISION_COMPARE );


    delete krig; krig = NULL;
    delete gpb;  gpb  = NULL;
    
   // Find optimal parameter: 
    // linear trend, no measurement error, no regularisation, 
    // ReML cost function, no screening, NelderMead optimiser, 1000 run max, reset on.        
    gpb = new TGPBuilder(tdsObs, "x1", "y", "exp", "linear");
    gpb->findOptimalParameters("ReML", 0, "NelderMead");
    //gpb->findOptimalParameters("ReML", 0, "");
    krig = gpb->buildGP();
    
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 4.19248533419316    , krig->getVariance()             , DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 0.486639335856841 , krig->getCorrLengths()[0], DBL_PRECISION_COMPARE );    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(  1.47029614515789   , krig->getTrendParams()[0], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -0.099595144003217 , krig->getTrendParams()[1], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( -5.6772547134265  , krig->getGPParams()[0], DBL_PRECISION_COMPARE );


    delete krig; krig = NULL;
    delete gpb;  gpb  = NULL; 
}

void TGPBuilderTest::testExportModelData()
{
	cout<<endl;
    TGPBuilder  *gpb1, *gpb2 = NULL;
    TKriging    *k1, *k2 = NULL;
    TDataServer *tdsK;
    
    // Case with no trend
    gpb1 = new TGPBuilder(tdsObs, "x1", "y", "isogauss");
    gpb1->findOptimalParameters("ML", 100, "Bobyqa");
    //gpb1->findOptimalParameters("ML", 100, "");
	
    k1 = gpb1->buildGP();
    CPPUNIT_ASSERT_NO_THROW( gpb1->exportGPData("krig-model.dat") );
    
    tdsK = new TDataServer();
    CPPUNIT_ASSERT_NO_THROW( tdsK->fileDataRead("krig-model.dat") );
    CPPUNIT_ASSERT_NO_THROW( gpb2 = new TGPBuilder(tdsK) );
    CPPUNIT_ASSERT_NO_THROW( k2 = gpb2->buildGP() );
    
    // Comparison between the two models
    CPPUNIT_ASSERT_DOUBLES_EQUAL( k1->getVariance(), k2->getVariance(), DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( k1->getCorrLengths()[0] , k2->getCorrLengths()[0], DBL_PRECISION_COMPARE );    
    CPPUNIT_ASSERT_DOUBLES_EQUAL( k1->getGPParams()[0] , k2->getGPParams()[0], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( k1->getGPParams()[1] , k2->getGPParams()[1], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( k1->getGPParams()[2] , k2->getGPParams()[2], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( k1->getGPParams()[3] , k2->getGPParams()[3], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( k1->getGPParams()[4] , k2->getGPParams()[4], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( k1->getGPParams()[5] , k2->getGPParams()[5], DBL_PRECISION_COMPARE );    
    
    delete k1;   k1   = NULL;
    delete k2;   k2   = NULL;
    delete tdsK; tdsK = NULL;
    delete gpb2; gpb2 = NULL;
    delete gpb1; gpb1 = NULL;
    
    // Case with trend
    gpb1 = new TGPBuilder(tdsObs, "x1", "y", "isogauss", "linear");
    gpb1->findOptimalParameters("ML", 100, "bobyqa");
    //gpb1->findOptimalParameters("ML", 100, "");
    k1 = gpb1->buildGP();
    
    CPPUNIT_ASSERT_NO_THROW( gpb1->exportGPData("krig-model.dat") );
    
    tdsK = new TDataServer();
    CPPUNIT_ASSERT_NO_THROW( tdsK->fileDataRead("krig-model.dat") );
    CPPUNIT_ASSERT_NO_THROW( gpb2 = new TGPBuilder(tdsK) );
    CPPUNIT_ASSERT_NO_THROW( k2 = gpb2->buildGP() );
    
    // Comparison between the two models
    CPPUNIT_ASSERT_DOUBLES_EQUAL( k1->getVariance(), k2->getVariance(), DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( k1->getCorrLengths()[0] , k2->getCorrLengths()[0], DBL_PRECISION_COMPARE );    
    CPPUNIT_ASSERT_DOUBLES_EQUAL( k1->getGPParams()[0] , k2->getGPParams()[0], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( k1->getGPParams()[1] , k2->getGPParams()[1], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( k1->getGPParams()[2] , k2->getGPParams()[2], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( k1->getGPParams()[3] , k2->getGPParams()[3], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( k1->getGPParams()[4] , k2->getGPParams()[4], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( k1->getGPParams()[5] , k2->getGPParams()[5], DBL_PRECISION_COMPARE );    
    
    delete k1;   k1   = NULL;
    delete k2;   k2   = NULL;
    delete tdsK; tdsK = NULL;
    delete gpb2; gpb2 = NULL;
    delete gpb1; gpb1 = NULL;
    
    // Case with trend and measurement error
    gpb1 = new TGPBuilder(tdsObs, "x1", "y", "isogauss");
    gpb1->setHasMeasurementError(kTRUE);
    gpb1->setAlpha(0.1); // we set alpha manually because its optimal value is 0.0 (train data are noiseless).
    gpb1->findOptimalParameters("ML", 100, "bobyqa");
    //gpb1->findOptimalParameters("ML", 100, "");
    k1 = gpb1->buildGP();
    
    CPPUNIT_ASSERT_NO_THROW( gpb1->exportGPData("krig-model.dat") );
    
    tdsK = new TDataServer();
    CPPUNIT_ASSERT_NO_THROW( tdsK->fileDataRead("krig-model.dat") );
    CPPUNIT_ASSERT_NO_THROW( gpb2 = new TGPBuilder(tdsK) );
    CPPUNIT_ASSERT_NO_THROW( k2 = gpb2->buildGP() );
    
    // Comparison between the two models
    CPPUNIT_ASSERT_DOUBLES_EQUAL( k1->getVariance(), k2->getVariance(), DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( k1->getMeasurementErrorVariance(), k2->getMeasurementErrorVariance(), DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( k1->getCorrLengths()[0] , k2->getCorrLengths()[0], DBL_PRECISION_COMPARE );    
    CPPUNIT_ASSERT_DOUBLES_EQUAL( k1->getGPParams()[0] , k2->getGPParams()[0], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( k1->getGPParams()[1] , k2->getGPParams()[1], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( k1->getGPParams()[2] , k2->getGPParams()[2], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( k1->getGPParams()[3] , k2->getGPParams()[3], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( k1->getGPParams()[4] , k2->getGPParams()[4], DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( k1->getGPParams()[5] , k2->getGPParams()[5], DBL_PRECISION_COMPARE );    
    
    delete k1;   k1   = NULL;
    delete k2;   k2   = NULL;
    delete tdsK; tdsK = NULL;
    delete gpb2; gpb2 = NULL;
    delete gpb1; gpb1 = NULL; 
}

void TGPBuilderTest::testUpdateObservations()
{
    TGPBuilder  *gpb;
    TKriging    *k1, *k2;
    TDataServer *tdsObs2;
    
    tdsObs2 = new TDataServer("tdsObs2","");
    tdsObs2->fileDataRead("utf-1D-train.dat");
    
    // Create the GPB
    gpb = new TGPBuilder(tdsObs2, "x1", "y", "isogauss");
    gpb->findOptimalParameters("ML", 100, "bobyqa");
    //gpb->findOptimalParameters("ML", 100, "");
    k1 = gpb->buildGP();
    
    // update the data server
    TDSNtupleD *t = tdsObs2->getTuple();
    Double_t line[3] = {7, 4.698548e-03, 2.413597e+00};
    t->Fill(line);
    
    // Update the gpb
    gpb->updateObservations();
    k2 = gpb->buildGP();

    // Comparison between the two models
    CPPUNIT_ASSERT_DOUBLES_EQUAL( k1->getVariance(), k2->getVariance(), DBL_PRECISION_COMPARE );
    CPPUNIT_ASSERT_DOUBLES_EQUAL( k1->getCorrLengths()[0] , k2->getCorrLengths()[0], DBL_PRECISION_COMPARE );    

    delete k1;      k1 = NULL;
    delete k2;      k2 = NULL;
    delete gpb;     gpb = NULL; 
    delete tdsObs2; tdsObs2 = NULL;
    
}


void TGPBuilderTest::testSetSeed()
{
    TGPBuilder *gpb;
    TKriging *krig;
    
    
    // Find optimal parameter: 
    // no trend, no measurement error, no regularisation,  
    // ML cost function, 100 screening, NelderMead optimiser, 1000 run max, reset on.
    gpb = new TGPBuilder(tdsObs, "x1", "y", "isogauss");
    //set seed at 12345 value
    gpb->setSeed(12345);
    CPPUNIT_ASSERT_EQUAL(12345, gpb->getSeed());
	
    CPPUNIT_ASSERT_NO_THROW( gpb->findOptimalParameters("ML", 100, "NelderMead") );
    //CPPUNIT_ASSERT_NO_THROW( gpb->findOptimalParameters("ML", 100, "") );
    krig = gpb->buildGP();
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.49110727227384598503, krig->getCorrLengths()[0], DBL_PRECISION_COMPARE2 );  
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.50662626839228930287, gpb->getCorrFunction()->getCorrLengths()[0], DBL_PRECISION_COMPARE2 );    
    
    delete krig; 
    krig = NULL;
    
    // Manual setting of parameters
    Double_t params[1] = {0.5};
    gpb->getCorrFunction()->setParameters(params);
    // Find optimal parameter: 
    // no trend, no measurement error, no regularisation,  
    // ML cost function, no screening, NelderMead optimiser, 1000 run max, reset off.    
    gpb->findOptimalParameters("ML", 0, "NelderMead", 1000, kFALSE);   
    //gpb->findOptimalParameters("ML", 0, "", 1000, kFALSE);  
    krig = gpb->buildGP();
    
#ifdef WIN32
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.6554640835008296307, krig->getVariance()             , 10E-4  );
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.49106107693607420339, krig->getCorrLengths()[0], 10E-4  );   
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-5.9429360594879527824, krig->getGPParams()[0], 10E-4 );
#else
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.6554640835008296307, krig->getVariance()             , DBL_PRECISION_COMPARE2 );
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.49106107693607420339, krig->getCorrLengths()[0], DBL_PRECISION_COMPARE2 );    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-5.9429360594879527824, krig->getGPParams()[0], DBL_PRECISION_COMPARE2 ); 
#endif	

    delete krig;
    krig = NULL;
    delete gpb;  
    gpb = NULL; 
}


void TGPBuilderTest::testFindOptimalParamBetweenBounds()
{
  gROOT->LoadMacro("Dirichlet.C");
#ifdef ROOT5
  G__ClassInfo globalscopeDirichlet;
  G__MethodInfo methodDirichlet;
  long dummyDirichlet;
  methodDirichlet = globalscopeDirichlet.GetMethod("Dirichlet", "double*,double*", &dummyDirichlet);
  void * _p2fDirichlet = methodDirichlet.PointerToFunc();
#endif
  
  Int_t Dim = 4;
  Double_t *Alpha = new Double_t[Dim];
  for (Int_t i=0; i < Dim; i++) Alpha[i] = 1. / (i + 1.);
  
  // Noms des variables
  TString sInput="x1";
  for (Int_t i=1; i < Dim; i++) sInput=sInput+Form(":x%d",i+1);
  TString sOutput="y";
  
  TDataServer * tds = new TDataServer("tds", "Tds for use case Dirichlet");
  for (Int_t i=0; i < Dim; i++)
    {
      tds->addAttribute(new TUniformDistribution(Form("x%d",i+1), 0., 1.));
      tds->getAttribute(Form("x%d",i+1))->delShare();
    }

  Int_t nSimulation = 50;
  TSampling *sampling = new TSampling(tds, "lhs", nSimulation);
  sampling->setSeed(1234);
  sampling->generateSample();

#ifdef ROOT5
  TLauncherFunction * tlf = new TLauncherFunction(tds, _p2fDirichlet,sInput,sOutput);
#else
  TLauncherFunction * tlf = new TLauncherFunction(tds, "Dirichlet", sInput, sOutput);
#endif
  tlf->setDrawProgressBar(kFALSE);
  tlf->run();
  
  // Construction d'un nouveau plan
  TDataServer * tdsPool = new TDataServer("tdsPoll", "Tds for use case Dirichlet");
  tdsPool->fileDataRead("samplePool10000LHS.dat");
  
  // Construction du TGPBuilder
  TGPBuilder *gpb = new TGPBuilder(tds, sInput, sOutput, "gauss", "const");
  gpb->setRegularisation(1.e-6);
  gpb->setSeed( 62589 );
  gpb->findOptimalParameters("ML", 1000, "bfgs");  
  //gpb->findOptimalParameters("ML", 1000, "");      
    
  // construction du premier modele
  TKriging *krig = gpb->buildGP();
  krig->printLog();

  // Recalcul du krigeage
  // mise a jour des donnees du GPBuilder
  Double_t* XX;
  TDataServer * tdsBis = new TDataServer("tds", "Tds for use case Dirichlet");
  tdsBis->fileDataRead("ReSample.dat");
  TString attName;
  
  Int_t kDoe = 10;
  for (Int_t k=0; k < kDoe; k++) {
    cout << "Iter = " << k+1 << endl;
    
    XX = new Double_t[Dim+2];
    for(Int_t ktt = 0; ktt<Dim+1; ktt++)
        {
            attName = tdsBis->getAttribute(ktt)->GetName();
            XX[ktt+1] = tdsBis->getValue(attName,k);
        }
    XX[0] = tdsBis->getValue(tdsBis->getIteratorName(),k);
    tds->getTuple()->Fill(XX);
    
    gpb->updateObservations();
    
    // Relance de l'optimisation
    CPPUNIT_ASSERT_NO_THROW(gpb->findOptimalParameters("ML", 0, "BOBYQA", 500, kFALSE)); // ne pas oublier le kFALSE pour reprendre l'optimisation au point actuel.
    //CPPUNIT_ASSERT_NO_THROW(gpb->findOptimalParameters("ML", 0, "", 500, kFALSE)); // ne pas oublier le kFALSE pour reprendre l'optimisation au point actuel.
    // construction du second modele
    delete krig;
    krig = gpb->buildGP();
    cout << "RMSE LOO = " << krig->getLooRMSE() << endl;
    cout << "Q2   LOO = " << krig->getLooQ2()   << endl;
    delete[] XX;
  }
    
	delete [] Alpha;
    delete krig;
    delete gpb;
    delete sampling;
    delete tlf;
    delete tds;
    delete tdsPool;
    delete tdsBis;
}

void TGPBuilderTest::testsetTolerance()
{
    TGPBuilder *gpb;
    //Bool_t hasCorrFuntion = kFALSE;
      
    gpb = new TGPBuilder(tdsObs, "x1", "y", "isogauss");
    gpb->setTolerance(1e-4);
    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0001, gpb->getTolerance(),
				 DBL_PRECISION_COMPARE);

    delete gpb;
    
    
}

void TGPBuilderTest::testsetCorrFunction()
{
    TGPBuilder *gpb;
    Bool_t hasCorrFuntion = kFALSE;
      
    gpb = new TGPBuilder(tdsObs, "x1", "y", "isogauss");
    
    if(dynamic_cast<TIsotropicGaussianCorrFunction *> (gpb->getCorrFunction()) != NULL )
        hasCorrFuntion = kTRUE;        
    CPPUNIT_ASSERT(hasCorrFuntion);
    hasCorrFuntion = kFALSE;    

    gpb->setCorrFunction("maternIII");
    
    if(dynamic_cast<TMaternIIICorrFunction *> (gpb->getCorrFunction()) != NULL )
        hasCorrFuntion = kTRUE;        
    CPPUNIT_ASSERT(hasCorrFuntion);
    hasCorrFuntion = kFALSE;    

    gpb->setCorrFunction("maternII");
    
    if(dynamic_cast<TMaternIICorrFunction *> (gpb->getCorrFunction()) != NULL )
        hasCorrFuntion = kTRUE;        
    CPPUNIT_ASSERT(hasCorrFuntion);
    hasCorrFuntion = kFALSE;    

    gpb->setCorrFunction("maternI");
    
    if(dynamic_cast<TMaternICorrFunction *> (gpb->getCorrFunction()) != NULL )
        hasCorrFuntion = kTRUE;        
    CPPUNIT_ASSERT(hasCorrFuntion);
    hasCorrFuntion = kFALSE;    

    gpb->setCorrFunction("matern52");
    
    if(dynamic_cast<TMatern52CorrFunction *> (gpb->getCorrFunction()) != NULL )
        hasCorrFuntion = kTRUE;        
    CPPUNIT_ASSERT(hasCorrFuntion);
    hasCorrFuntion = kFALSE;    

    gpb->setCorrFunction("matern72");
    
    if(dynamic_cast<TMatern72CorrFunction *> (gpb->getCorrFunction()) != NULL )
        hasCorrFuntion = kTRUE;        
    CPPUNIT_ASSERT(hasCorrFuntion);
    hasCorrFuntion = kFALSE;    

    gpb->setCorrFunction("TExponentialCorrFunction");
    
    if(dynamic_cast<TExponentialCorrFunction *> (gpb->getCorrFunction()) != NULL )
        hasCorrFuntion = kTRUE;        
    CPPUNIT_ASSERT(hasCorrFuntion);
    hasCorrFuntion = kFALSE;    
    delete gpb;
    
    
}


void TGPBuilderTest::throwExceptionSetCorrFunction()
{
    TGPBuilder *gpb = new TGPBuilder(tdsObs, "x1", "y", "isogauss");
    
    CPPUNIT_ASSERT_THROW(gpb->setCorrFunction("toto");,
            URANIE::Exceptions::UErrorExceptions);
    
    delete gpb;
    
}

void TGPBuilderTest::throwExceptionConstructorIncompatibleDataServer()
{    
    CPPUNIT_ASSERT_THROW(new TGPBuilder(tdsObs),
            URANIE::Exceptions::UErrorExceptions);
    
    TDataServer* tds1 = new TDataServer();
    tds1->fileDataRead("krig-WrongModel1.dat");
    
    CPPUNIT_ASSERT_THROW(new TGPBuilder(tds1),
            URANIE::Exceptions::UErrorExceptions);
    
    TDataServer* tds2 = new TDataServer();
    tds2->fileDataRead("krig-WrongModel2.dat");
    
    CPPUNIT_ASSERT_THROW(new TGPBuilder(tds2),
            URANIE::Exceptions::UErrorExceptions);
    
    delete tds1;
    delete tds2;
    
}

void TGPBuilderTest::throwExceptionUnknownOptimisationAlgorithm()
{
    TGPBuilder *gpb = new TGPBuilder(tdsObs, "x1", "y", "isogauss");
    
    CPPUNIT_ASSERT_THROW(gpb->findOptimalParameters("ML", 1000, "toto"),
            URANIE::Exceptions::UErrorExceptions);
    
    delete gpb;
    
}

void TGPBuilderTest::throwExceptionSetUsePrior()
{
    TGPBuilder *gpb = new TGPBuilder(tdsObs, "x1", "y", "isogauss");
    
    CPPUNIT_ASSERT_THROW(gpb->setUsePrior(kTRUE),
            URANIE::Exceptions::UErrorExceptions);
    delete gpb;
}


void TGPBuilderTest::throwExceptionUnsupportedCriterion()
{
    TGPBuilder *gpb = new TGPBuilder(tdsObs, "x1", "y", "isogauss");
    
    CPPUNIT_ASSERT_THROW(gpb->findOptimalParameters("toto", 1000, "bfgs"),
            URANIE::Exceptions::UErrorExceptions);
    delete gpb;
}


void TGPBuilderTest::throwExceptionReMLCriterionWithoutTrend()
{
    TGPBuilder *gpb = new TGPBuilder(tdsObs, "x1", "y", "isogauss");
    
    CPPUNIT_ASSERT_THROW(gpb->findOptimalParameters("ReML", 1000, "bfgs"),
            URANIE::Exceptions::UErrorExceptions);
    delete gpb;
}

void TGPBuilderTest::throwExceptionMLCriterionWithPriorData()
{
    TGPBuilder *gpb = new TGPBuilder(tdsObs, "x1", "y", "isogauss","linear");
    // Setting and checking of bayesian parameters    
    Double_t betaPrior[2] = {2.0, 0.0};
    Double_t qPrior[4]    = {1.0, 0.0, 0.0, 1.0};
    gpb->setPriorData(betaPrior, qPrior);
    CPPUNIT_ASSERT_THROW(gpb->findOptimalParameters("ML", 1000, "bfgs"),
            URANIE::Exceptions::UErrorExceptions);
    delete gpb;
}

void TGPBuilderTest::throwExceptionConstructorError()
{    
    TDataServer* tds1 = new TDataServer();
    tds1->fileDataRead("cafards.dat");
    CPPUNIT_ASSERT_THROW(new TGPBuilder(tds1, "dose", "total:morts", "isogauss","linear"),
            URANIE::Exceptions::UErrorExceptions);
       
    
    CPPUNIT_ASSERT_THROW( new TGPBuilder(tdsObs, "x1", " ", "isogauss","linear"),
            URANIE::Exceptions::UErrorExceptions);
    
    CPPUNIT_ASSERT_THROW( new TGPBuilder(tdsObs, " ", "y", "isogauss","linear"),
            URANIE::Exceptions::UErrorExceptions);

    delete tds1;
    
}

void TGPBuilderTest::throwExceptionsetLengthRange()
{
    TGPBuilder *gpb = new TGPBuilder(tdsObs, "x1", "y", "isogauss","linear");
    CPPUNIT_ASSERT_THROW(gpb->setLengthRange(-1.0, 0.5), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(gpb->setLengthRange(0.0, -0.1), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(gpb->setLengthRange(0.001, 0.0005), URANIE::Exceptions::UErrorExceptions);
    delete gpb;
}

void TGPBuilderTest::throwExceptionsetLengthMaxScreeningSF()
{
    TGPBuilder *gpb = new TGPBuilder(tdsObs, "x1", "y", "isogauss","linear");
    CPPUNIT_ASSERT_THROW(gpb->setLengthMaxScreeningSF(-1.0), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(gpb->setLengthMaxScreeningSF(-0.1), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(gpb->setLengthMaxScreeningSF(1.01), URANIE::Exceptions::UErrorExceptions);
    delete gpb;
}

