//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TestTPolynomialChaos.cxx
 \author Maud Salvoldi (OCC)
 \date   08/04/2011

 \brief

 */
// ROOT
#include "TMath.h"
#include "TString.h"
#include "Rtypes.h"
#ifdef ROOT5
#include "Api.h"
#endif

// cppunit
#include <iostream>
#define DBL_PRECISION_COMPARE 1.e-8

//uranie
#include "TestTPolynomialChaos.h"
#include "TDataServer.h"
#include "TSampling.h"
#include "TSampler.h"
#include "TUniformDistribution.h"
#include "TNormalDistribution.h"
#include "TBetaDistribution.h"
#include "TLauncherFunction.h"
#include "TLauncher.h"
#include "TCIntEval.h"
#include "TLauncher2.h"
#include "TPetras.h"
#include "TestUserFunctions.h"
#include "uranie_commands.h"

#ifndef ROOT5
#include "UserFunctions.C"
#endif

#include <cmath>
#include <iomanip>

using namespace std;

TPolynomialChaosTest::TPolynomialChaosTest() :
        _tds(NULL), _nisp(NULL), _pc(NULL), _tlf(NULL)
{
    _workingTestPath = gSystem->pwd();
    
    fstream file("NispIshigami.C");
    if (!file.fail())
        remove("NispIshigami.C");

#ifdef ROOT5
    gROOT->LoadMacro("UserFunctions.C");
#endif
}

TPolynomialChaosTest::~TPolynomialChaosTest()
{
    fstream file("NispIshigami.C");
    if (!file.fail())
        remove("NispIshigami.C");
}


void TPolynomialChaosTest::setUp()
{
    
    _tds = new URANIE::DataServer::TDataServer();
    _tds->unsetLog();
    _tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x1",
                    -1. * TMath::Pi(), TMath::Pi()));
    _tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x2",
                    -1. * TMath::Pi(), TMath::Pi()));
    _tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x3",
                    -1. * TMath::Pi(), TMath::Pi()));
    _tds->getAttribute("x1")->delShare();
    _tds->getAttribute("x2")->delShare();
    _tds->getAttribute("x3")->delShare();
    
    _nisp = new URANIE::Modeler::TNisp(_tds);
    _nisp->unsetLog();
    _nisp->generateSample("Petras", 8);
#ifdef ROOT5
    _tlf = new URANIE::Launcher::TLauncherFunction(_tds, "Ishigami",
            "", "Ishigami:Lineaire");
#else
    _tlf = new URANIE::Launcher::TLauncherFunction(_tds, &Ishigami,
            "", "Ishigami:Lineaire");
#endif
    _tlf->unsetLog();
    _tlf->setDrawProgressBar(kFALSE);
    _tlf->run();
    
    _pc = new URANIE::Modeler::TPolynomialChaos(_tds, _nisp);
    _pc->unsetLog();
}
void TPolynomialChaosTest::tearDown()
{
   if (_tlf != NULL)
    {
        delete _tlf;
        _tlf = NULL;
    }
    if (_nisp != NULL)
    {
        delete _nisp;
        _nisp = NULL;
    }
    if (_tds != NULL)
    {
        delete _tds;
        _tds = NULL;
    }
    
    if (_pc != NULL)
        delete _pc;
    _pc = NULL;
    
    UranieTests::removeFile(_workingTestPath.Data(),"filePChaos_Integration.C");
    UranieTests::removeFile(_workingTestPath.Data(),"functionPChaos.C");
}

//////////Units tests///////
void TPolynomialChaosTest::testgetDegree()
{
    CPPUNIT_ASSERT_EQUAL(1, _pc->getDegree());
    CPPUNIT_ASSERT( _pc->_bStoreYHat );
}
void TPolynomialChaosTest::testsetDegree()
{
    _pc->setDegree(8);
    CPPUNIT_ASSERT_EQUAL(8, _pc->getDegree());

}
void TPolynomialChaosTest::throwExceptionsetDegree()
{
    CPPUNIT_ASSERT_THROW(_pc->setDegree(-1),
            URANIE::Exceptions::UErrorExceptions);
}
void TPolynomialChaosTest::throwExceptionTPolynomialChaosTdsNull()
{
    URANIE::DataServer::TDataServer * tdsnull = NULL;
    CPPUNIT_ASSERT_THROW(new URANIE::Modeler::TPolynomialChaos(tdsnull, _nisp),
            URANIE::Exceptions::UErrorExceptions);
}
void TPolynomialChaosTest::throwExceptionTPolynomialChaosTNispNull()
{
    URANIE::Modeler::TNisp *nispnull = NULL;
    CPPUNIT_ASSERT_THROW(new URANIE::Modeler::TPolynomialChaos(_tds, nispnull),
            URANIE::Exceptions::UErrorExceptions);
}
void TPolynomialChaosTest::throwExceptionTPolynomialChaosTdsNotSame()
{
    URANIE::DataServer::TDataServer *tds2 = new URANIE::DataServer::TDataServer(
            "tds2ishigami", "Ex2. Ishigami");
    tds2->unsetLog();
    tds2->addAttribute(
            new URANIE::DataServer::TUniformDistribution("xx1",
                    -1. * TMath::Pi(), TMath::Pi()));
    tds2->addAttribute(
            new URANIE::DataServer::TUniformDistribution("xx2",
                    -1. * TMath::Pi(), TMath::Pi()));
 
   tds2->getAttribute("xx1")->delShare();
   tds2->getAttribute("xx2")->delShare();
    
    CPPUNIT_ASSERT_THROW(new URANIE::Modeler::TPolynomialChaos(tds2, _nisp),
            URANIE::Exceptions::UErrorExceptions);

    delete tds2;
    
}
void TPolynomialChaosTest::throwExceptionTPolynomialChaosOutputNotExist()
{
    CPPUNIT_ASSERT_THROW(
            new URANIE::Modeler::TPolynomialChaos(_tds, _nisp, "Ishigami:toto"),
            URANIE::Exceptions::UErrorExceptions);
}
void TPolynomialChaosTest::testTPolynomialChaosListOutput()
{
    URANIE::Modeler::TPolynomialChaos *pc2 =
            new URANIE::Modeler::TPolynomialChaos(_tds, _nisp,
                    "Ishigami:Lineaire");
    pc2->setDegree(8);
    pc2->computeChaosExpansion("Integration");

    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration", "nostoreyhat");

    //control the first coefficient of the output Ishigami
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_pc->getCoefficient(1, "Ishigami"),
            pc2->getCoefficient(1, "Ishigami"), DBL_PRECISION_COMPARE);

    delete pc2;
}
void TPolynomialChaosTest::testComputeOutputGetOuput()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    Double_t xi[4];
    xi[0] = 0.8;
    xi[1] = 0.2;
    xi[2] = 0.7;

    _pc->computeOutput(xi);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.88495559, _pc->getOutput("Lineaire"),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(7.55107666, _pc->getOutput(0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_pc->getOutput(0), _pc->getOutput("Ishigami"),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_pc->getOutput(1), _pc->getOutput("Lineaire"),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_pc->getOutput(0), _pc->getOutput(),
            DBL_PRECISION_COMPARE);
}
void TPolynomialChaosTest::throwExceptionGetOuputJNegative()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    Double_t xi[4];
    xi[0] = 0.8;
    xi[1] = 0.2;
    xi[2] = 0.7;
    _pc->computeOutput(xi);
    CPPUNIT_ASSERT_THROW(_pc->getOutput(-2),
            URANIE::Exceptions::UWarningExceptions);
}
void TPolynomialChaosTest::throwExceptionGetOuputJExceed()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    Double_t xi[4];
    xi[0] = 0.8;
    xi[1] = 0.2;
    xi[2] = 0.7;
    _pc->computeOutput(xi);
    CPPUNIT_ASSERT_THROW(_pc->getOutput(4),
            URANIE::Exceptions::UWarningExceptions);
}
void TPolynomialChaosTest::throwExceptionGetOuputNamefalse()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    Double_t xi[4];
    xi[0] = 0.8;
    xi[1] = 0.2;
    xi[2] = 0.7;
    _pc->computeOutput(xi);
    CPPUNIT_ASSERT_THROW(_pc->getOutput("toto"),
            URANIE::Exceptions::UWarningExceptions);
}
void TPolynomialChaosTest::testcomputeChaosExpansionIntegration()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");

    //control the first coefficient of the output Ishigami
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.625417827,
            _pc->getCoefficient(1, "Ishigami"), DBL_PRECISION_COMPARE);
    //control the last coefficient of the output Ishigami
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.00157948, _pc->getCoefficient(164),
            DBL_PRECISION_COMPARE);
}

void TPolynomialChaosTest::testcomputeChaosExpansionIntegrationWithCopy()
{
    URANIE::Modeler::TPolynomialChaos * pcCopy = new URANIE::Modeler::TPolynomialChaos(_pc[0]);
    pcCopy->setDegree(8);
    pcCopy->computeChaosExpansion("Integration");

    //control the first coefficient of the output Ishigami
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.625417827,
            pcCopy->getCoefficient(1, "Ishigami"), DBL_PRECISION_COMPARE);
    //control the last coefficient of the output Ishigami
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.00157948, pcCopy->getCoefficient(164),
            DBL_PRECISION_COMPARE);
    delete pcCopy;
}

void TPolynomialChaosTest::testcomputeChaosExpansionIntegrationOnCopyWithLaguerre()
{
    URANIE::DataServer::TDataServer* tds = new URANIE::DataServer::TDataServer("Pouet","Foo");
    tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x1",
                    -1. * TMath::Pi(), TMath::Pi()));
    tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x2",
                    -1. * TMath::Pi(), TMath::Pi()));
    tds->getAttribute("x1")->delShare();
    tds->getAttribute("x2")->delShare();
	
    URANIE::DataServer::TUniformDistribution* tud = new URANIE::DataServer::TUniformDistribution("x3",-1*TMath::Pi(),TMath::Pi());
    tud->setStochasticBasis(URANIE::DataServer::TStochasticAttribute::kLaguerre);
    tds->addAttribute( tud );
    tud->delShare();
    URANIE::Modeler::TNisp* nisp = new URANIE::Modeler::TNisp(tds);
    nisp->unsetLog();
    nisp->generateSample("Petras", 8);
    URANIE::Launcher::TLauncherFunction* tlf = 
#ifdef ROOT5
        new URANIE::Launcher::TLauncherFunction(tds, "Ishigami","", "Ishigami:Lineaire");
#else
        new URANIE::Launcher::TLauncherFunction(tds, &Ishigami, "", "Ishigami:Lineaire");
#endif
    tlf->setDrawProgressBar(kFALSE);
    tlf->run();
    URANIE::Modeler::TPolynomialChaos* pc = new URANIE::Modeler::TPolynomialChaos(tds, nisp);
    pc->unsetLog();

    URANIE::Modeler::TPolynomialChaos * pcCopy = new URANIE::Modeler::TPolynomialChaos(pc[0]);
    pcCopy->unsetLog();
    pcCopy->setDegree(8);
    pcCopy->computeChaosExpansion("Integration");

    //control the first coefficient of the output Ishigami
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.625417827,
            pcCopy->getCoefficient(1, "Ishigami"), DBL_PRECISION_COMPARE);
    //control the last coefficient of the output Ishigami
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.00295695068, pcCopy->getCoefficient(164),
            DBL_PRECISION_COMPARE);

    delete pcCopy;
    delete pc;
    delete tlf;
    delete nisp;
    delete tds;
}

void TPolynomialChaosTest::testcomputeChaosExpansionIntegrationWithTPetras()
{
    URANIE::DataServer::TDataServer* tds = new URANIE::DataServer::TDataServer("Pouet","Foo");
    tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x1",
                    -1. * TMath::Pi(), TMath::Pi()));
    tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x2",
                    -1. * TMath::Pi(), TMath::Pi()));
    tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x3",
                    -1. * TMath::Pi(), TMath::Pi()));

    tds->getAttribute("x1")->delShare();
    tds->getAttribute("x2")->delShare();
    tds->getAttribute("x3")->delShare();

    URANIE::Modeler::TNisp* nisp = new URANIE::Modeler::TNisp(tds);
    URANIE::Sampler::TPetras *sampler = new URANIE::Sampler::TPetras(tds,"",8);
    sampler->generateSample();
    nisp->setSample("Petras");
    nisp->setLevel(8);
    URANIE::Launcher::TLauncherFunction* tlf = 
#ifdef ROOT5
        new URANIE::Launcher::TLauncherFunction(tds, "Ishigami", "", "Ishigami:Lineaire");
#else
        new URANIE::Launcher::TLauncherFunction(tds, &Ishigami, "", "Ishigami:Lineaire");
#endif
    tlf->setDrawProgressBar(kFALSE);
    tlf->run();
    URANIE::Modeler::TPolynomialChaos * pc = new URANIE::Modeler::TPolynomialChaos(tds,nisp);
    pc->setDegree(8);
    pc->computeChaosExpansion("Integration");

    //control the first coefficient of the output Ishigami
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.625417827,
            pc->getCoefficient(1, "Ishigami"), DBL_PRECISION_COMPARE);
    //control the last coefficient of the output Ishigami
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.00157948, pc->getCoefficient(164),
            DBL_PRECISION_COMPARE);

    delete pc;
    delete sampler;
    delete tlf;
    delete nisp;
    delete tds;

}

void TPolynomialChaosTest::testcomputeChaosExpansionIntegrationWithImportedDatas()
{

    /////////////////////////////////////
    // Lecture des simulations realisees
    /////////////////////////////////////
    URANIE::DataServer::TDataServer * tds = new URANIE::DataServer::TDataServer("PolynomeMultiD", "Polynômes de Chaos");
    tds->fileDataRead("Quadrature.dat");

    // Creation d'un nouveau TDataServer
    URANIE::DataServer::TDataServer * tdsNisp = new URANIE::DataServer::TDataServer("PolynomeMultiD", "Polynômes de Chaos - Nisp");
    tdsNisp->addAttribute( new URANIE::DataServer::TNormalDistribution("x1", 1., 0.5));
    tdsNisp->addAttribute( new URANIE::DataServer::TUniformDistribution("x2", 1, 2));
    tdsNisp->addAttribute( new URANIE::DataServer::TBetaDistribution("x3",2., 2.));

    tdsNisp->getAttribute("x1")->delShare();
    tdsNisp->getAttribute("x2")->delShare();
    tdsNisp->getAttribute("x3")->delShare();
    tdsNisp->addAttribute("uWeight");
    tdsNisp->addAttribute("y");

    //////////////////
    // Objet TNisp  //
    //////////////////
    URANIE::Modeler::TNisp * nisp = new URANIE::Modeler::TNisp(tdsNisp);
    tdsNisp->setTuple( tds->getTuple() );
    nisp->setSample("Quadrature","savegvx");

    Int_t Level = 4;
    nisp->setLevel(Level);

    ////////////////////////
    // Polynomes de Chaos
    ////////////////////////
    URANIE::Modeler::TPolynomialChaos * pc = new URANIE::Modeler::TPolynomialChaos(tdsNisp, nisp, "y");
    pc->setDegree(Level);
    pc->computeChaosExpansion("Integration");


    pc->exportFunction("filePChaos_Integration","PChaos_Integration");
    gROOT->LoadMacro("filePChaos_Integration.C");
    
#ifdef ROOT5
    G__ClassInfo globalscope;
    G__MethodInfo method;
    long dummy;
    method = globalscope.GetMethod("PChaos_Integration", "double*,double*", &dummy);
    void * PChaos_Integration = method.PointerToFunc();

    URANIE::Relauncher::TCIntEval fEval(PChaos_Integration);
#else
    URANIE::Relauncher::TCIntEval fEval("PChaos_Integration");
#endif
    URANIE::Relauncher::TLauncher2 *lpcg = new URANIE::Relauncher::TLauncher2(tdsNisp, &fEval, "x1:x2:x3", "yEstim");
    lpcg->solverLoop();

    Int_t npat = tds->getNPatterns();
    CPPUNIT_ASSERT_EQUAL(125, npat);

    Int_t npatHalf = (Int_t)(npat/2);

    CPPUNIT_ASSERT( npatHalf < npat );

    //compare real values with estimated values of the output with a 1% margin
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsNisp->getValue("y", 0), tdsNisp->getValue("yEstim", 0),
            tdsNisp->getValue("y", 0)*0.01);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsNisp->getValue("y", npatHalf), tdsNisp->getValue("yEstim", npatHalf),
            tdsNisp->getValue("y", npatHalf)*0.01);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsNisp->getValue("y", npat-1), tdsNisp->getValue("yEstim", npat-1),
            tdsNisp->getValue("y", npat-1)*0.01);


    if (lpcg)
        delete lpcg;
    lpcg = NULL;

    if (pc)
        delete pc;
    pc = NULL;
    if (nisp)
        delete nisp;
    nisp = NULL;

    if (tdsNisp)
        delete tdsNisp;
    tdsNisp = NULL;

    delete tds;

}

void TPolynomialChaosTest::testcomputeChaosExpansionRegression()
{
    URANIE::DataServer::TDataServer *tds =
            new URANIE::DataServer::TDataServer("Pouet","Foo");
    tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x1",
                    -1. * TMath::Pi(), TMath::Pi()));
    tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x2",
                    -1. * TMath::Pi(), TMath::Pi()));
    tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x3",
                    -1. * TMath::Pi(), TMath::Pi()));
    tds->getAttribute("x1")->delShare();
    tds->getAttribute("x2")->delShare();
    tds->getAttribute("x3")->delShare();
    
    URANIE::Modeler::TNisp * nisp = new URANIE::Modeler::TNisp(tds);
    nisp->generateSample("QmcSobol", 512);
    URANIE::Launcher::TLauncherFunction *tlf =
#ifdef ROOT5
            new URANIE::Launcher::TLauncherFunction(tds, "Ishigami",
#else
            new URANIE::Launcher::TLauncherFunction(tds, &Ishigami,
#endif
                    "", "Ishigami:Lineaire");
    tlf->setDrawProgressBar(kFALSE);
    tlf->run();
    URANIE::Modeler::TPolynomialChaos *pc =
            new URANIE::Modeler::TPolynomialChaos(tds, nisp);
    pc->setDegree(6);
    pc->computeChaosExpansion("Regression");

    //control the first coefficient of the output Ishigami
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.621145162, pc->getCoefficient(1, "Ishigami"),
            DBL_PRECISION_COMPARE);
    //control the last coefficient of the output Ishigami
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.00703984, pc->getCoefficient(83),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.43358793843314,pc->getEqmLoo(), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-1.594268238229,pc->getErrLoo(20), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.63457882732163,pc->getErrLoo(500), DBL_PRECISION_COMPARE);


    if (tlf)
        delete tlf;
    tlf = NULL;
  if (pc)
    delete pc;
  pc = NULL;
    if (nisp)
        delete nisp;
    nisp = NULL;

    if (tds)
        delete tds;
    tds = NULL;
}

void TPolynomialChaosTest::testcomputeChaosExpansionRegressionAutomatised()
{
  URANIE::DataServer::TDataServer *tds = new URANIE::DataServer::TDataServer("another_tdsishigami","Ex. Ishigami");
  tds->addAttribute( new URANIE::DataServer::TUniformDistribution("x1", -1. * TMath::Pi(), TMath::Pi()));
  tds->addAttribute( new URANIE::DataServer::TUniformDistribution("x2", -1. * TMath::Pi(), TMath::Pi()));
  tds->addAttribute( new URANIE::DataServer::TUniformDistribution("x3", -1. * TMath::Pi(), TMath::Pi()));
  tds->getAttribute("x1")->delShare();
  tds->getAttribute("x2")->delShare();
  tds->getAttribute("x3")->delShare();

  // tds->createTuple();
  tds->fileDataRead("NispLhsForAutomatisedTest.dat", false, true);
  //  tds->Scan("*");
  URANIE::Modeler::TNisp * nisp = new URANIE::Modeler::TNisp(tds);
  nisp->setLevel(300);
  nisp->setSample("Lhs");
  //  nisp->generateSample("Lhs", 1000);
#ifdef ROOT5
  URANIE::Launcher::TLauncherFunction *tlf = new URANIE::Launcher::TLauncherFunction(tds, "Ishigami", "", "Ishigami:Lineaire");
#else
  URANIE::Launcher::TLauncherFunction *tlf = 
        new URANIE::Launcher::TLauncherFunction(tds, &Ishigami, "", "Ishigami:Lineaire");
#endif
  tlf->setDrawProgressBar(kFALSE);
  tlf->run();

  URANIE::Modeler::TPolynomialChaos *pc =  new URANIE::Modeler::TPolynomialChaos(tds, nisp, "Ishigami");

  pc->computeChaosExpansion("Regression","autoDegree noprogressbar");

  TTree *ntuple = pc->getAutoDegreeResults();
  stringstream sstr; sstr << pc->getBestAutoDegree();
  int nbEnt=(int)ntuple->Draw("errloo",("degree=="+sstr.str()).c_str(),"goff");
  double *data = ntuple->GetV1();

  CPPUNIT_ASSERT_EQUAL(8, pc->getBestAutoDegree());
  CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0828286595688912, pc->getEqmLoo(), DBL_PRECISION_COMPARE);
  CPPUNIT_ASSERT_DOUBLES_EQUAL(0.288263839310111, TMath::RMS(nbEnt,data), 1e-6);

  delete tlf;
  delete pc;
  delete nisp;
  delete tds;

}

void TPolynomialChaosTest::testcomputeChaosExpansionRegressionWithImportedDatas()
{
    URANIE::DataServer::TDataServer *tds =
            new URANIE::DataServer::TDataServer();
    tds->fileDataRead("Lhs_4ParametresLearning.dat");

    // Creation d'un nouveau TDataServer
    URANIE::DataServer::TDataServer * tdsNisp = new URANIE::DataServer::TDataServer("Trio_U_MC2", "Cas_4_Parametres Nisp");
    tdsNisp->addAttribute( new URANIE::DataServer::TNormalDistribution("T.entree",245,5));
    tdsNisp->addAttribute( new URANIE::DataServer::TNormalDistribution("Debit.mas",0.66,0.05));
    tdsNisp->addAttribute( new URANIE::DataServer::TUniformDistribution("Diam.aiguille",0.01225,0.01589));
    tdsNisp->addAttribute( new URANIE::DataServer::TUniformDistribution("Ep.gaine",0.0005,0.0056));
    tdsNisp->addAttribute("T.sortie");
    tdsNisp->getAttribute("T.sortie")->setOutput();

    tdsNisp->getAttribute("T.entree")->delShare();
    tdsNisp->getAttribute("Debit.mas")->delShare();
    tdsNisp->getAttribute("Diam.aiguille")->delShare();
    tdsNisp->getAttribute("Ep.gaine")->delShare();

    //////////////////
    // Objet TNisp  //
    //////////////////
    URANIE::Modeler::TNisp * nisp = new URANIE::Modeler::TNisp(tdsNisp);
    tdsNisp->setTuple( tds->getTuple() ); // Transfert des donnees
    nisp->setSample("Lhs","savegvx");

    ////////////////////////
    // Polynomes de Chaos
    ////////////////////////
    Int_t degre = 5;
    URANIE::Modeler::TPolynomialChaos * pc = new URANIE::Modeler::TPolynomialChaos(tdsNisp, nisp, "T.sortie");
    pc->setDegree(degre);
    pc->computeChaosExpansion("Regression");

    pc->exportFunction("functionPChaos","PChaos_Tsortie");

    gROOT->LoadMacro("functionPChaos.C");
    
#ifdef ROOT5
    G__ClassInfo globalscope;
    G__MethodInfo method;
    long dummy;
    method = globalscope.GetMethod("PChaos_Tsortie", "double*,double*", &dummy);
    void * PChaos_Tsortie = method.PointerToFunc();
    
    URANIE::Relauncher::TCIntEval fEval(PChaos_Tsortie);
#else
    URANIE::Relauncher::TCIntEval fEval("PChaos_Tsortie");
#endif
    TString sLabelPsi = "T.entree:Debit.mas:Diam.aiguille:Ep.gaine";
    
    URANIE::Relauncher::TLauncher2 *lpcg = new URANIE::Relauncher::TLauncher2(tdsNisp, &fEval, sLabelPsi, "yEstim");
    lpcg->solverLoop();

    //compare real values with estimated values of the output with a 1% margin
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsNisp->getValue("T.sortie", 0), tdsNisp->getValue("yEstim", 0),
            tdsNisp->getValue("T.sortie", 0)*0.01);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsNisp->getValue("T.sortie", 999), tdsNisp->getValue("yEstim", 999),
            tdsNisp->getValue("T.sortie", 999)*0.01);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsNisp->getValue("T.sortie", 1999), tdsNisp->getValue("yEstim", 1999),
            tdsNisp->getValue("T.sortie", 1999)*0.01);


    if (lpcg)
        delete lpcg;
    lpcg = NULL;

    if (pc)
        delete pc;
    pc = NULL;
    if (nisp)
        delete nisp;
    nisp = NULL;

    if (tdsNisp)
      delete tdsNisp;
    tdsNisp = NULL;

    delete tds;
}

void TPolynomialChaosTest::throwExceptioncomputeChaosExpansion()
{
    _pc->setDegree(8);
    CPPUNIT_ASSERT_THROW(_pc->computeChaosExpansion("toto"),
            URANIE::Exceptions::UErrorExceptions);

}
void TPolynomialChaosTest::testgetMean()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");

    //Mean of the output Ishigami
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.5, _pc->getMean("Ishigami"),
            DBL_PRECISION_COMPARE);

    //Mean of the ouput by default
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.5, _pc->getMean(), DBL_PRECISION_COMPARE);

    //Mean of the first output
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.5, _pc->getMean(0), DBL_PRECISION_COMPARE);

    //Mean of the first ouptut is the same as the mean of Ishigami output
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_pc->getMean("Ishigami"), _pc->getMean(0),
            DBL_PRECISION_COMPARE);
}
void TPolynomialChaosTest::throwExceptiongetMeanFalseName()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    CPPUNIT_ASSERT_THROW(_pc->getMean("toto"),
            URANIE::Exceptions::UWarningExceptions);
}
void TPolynomialChaosTest::throwExceptiongetMeanJNegative()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    CPPUNIT_ASSERT_THROW(_pc->getMean(-1),
            URANIE::Exceptions::UWarningExceptions);
}
void TPolynomialChaosTest::throwExceptiongetMeanJExceed()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    CPPUNIT_ASSERT_THROW(_pc->getMean(4),
            URANIE::Exceptions::UWarningExceptions);
}
void TPolynomialChaosTest::testgetVariance()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");

    //Variance of the output Ishigami
    CPPUNIT_ASSERT_DOUBLES_EQUAL(13.840567914, _pc->getVariance("Ishigami"),
            DBL_PRECISION_COMPARE);

    //Variance of the ouput by default
    CPPUNIT_ASSERT_DOUBLES_EQUAL(13.840567914, _pc->getVariance(),
            DBL_PRECISION_COMPARE);

    //Variance of the first output
    CPPUNIT_ASSERT_DOUBLES_EQUAL(13.840567914, _pc->getVariance(0),
            DBL_PRECISION_COMPARE);

    //Variance of the first ouptut is the same as the Variance of Ishigami output
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_pc->getVariance("Ishigami"),
            _pc->getVariance(0), DBL_PRECISION_COMPARE);
}
void TPolynomialChaosTest::throwExceptiongetVarianceFalseName()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    CPPUNIT_ASSERT_THROW(_pc->getVariance("toto"),
            URANIE::Exceptions::UWarningExceptions);
}
void TPolynomialChaosTest::throwExceptiongetVarianceJNegative()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    CPPUNIT_ASSERT_THROW(_pc->getVariance(-1),
            URANIE::Exceptions::UWarningExceptions);
}
void TPolynomialChaosTest::throwExceptiongetVarianceJExceed()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    CPPUNIT_ASSERT_THROW(_pc->getVariance(4),
            URANIE::Exceptions::UWarningExceptions);
}
void TPolynomialChaosTest::testgetDimensionInput()
{
    CPPUNIT_ASSERT_EQUAL(3, _pc->getDimensionInput());
}
void TPolynomialChaosTest::testgetDimensionOutput()
{
    CPPUNIT_ASSERT_EQUAL(2, _pc->getDimensionOutput());
}
void TPolynomialChaosTest::testgetDimensionExpansion()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    CPPUNIT_ASSERT_EQUAL(165, _pc->getDimensionExpansion());

}
void TPolynomialChaosTest::testexportFunction()
{
    time_t seconds;
    string line, date;
    Bool_t bStop = kFALSE;
    Int_t test = 0;
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    _pc->exportFunction("NispIshigami", "NispIshigami");
    fstream file("NispIshigami.C");
    time(&seconds);
    date = string(ctime(&seconds));
    while (bStop == kFALSE)
    {
        getline(file, line);
        if (line.find("  //     ** Date : ") != string::npos)
        {
            bStop = kTRUE;
            //verify that the date are the same actually and in the file
            if ((date[0] == line[19]) && (date[1] == line[20])
                    && (date[2] == line[21]) && (date[3] == line[22])
                    && (date[4] == line[23]) && (date[5] == line[24])
                    && (date[6] == line[25]) && (date[7] == line[26])
                    && (date[8] == line[27]) && (date[9] == line[28])
                    && (date[10] == line[29]) && (date[11] == line[30])
                    && (date[12] == line[31]) && (date[13] == line[32])
                    && (date[14] == line[33]) && (date[15] == line[34])
                    && (date[16] == line[35]) && (date[17] == line[36])
                    && (date[18] == line[37]) && (date[19] == line[38])
                    && (date[20] == line[39]) && (date[21] == line[40])
                    && (date[22] == line[41]) && (date[23] == line[42]))
                test = 1;

//             if(test==0)
//             {
//                 Int_t pos = date.find_last_of(":");
//                 Int_t pos2 = line.find_last_of(":");
//                 Int_t date_seconds, line_seconds;
//                 date_seconds =  (date[pos-5]*10+date[pos-4])*3600
//                             +(date[pos-2]*10+date[pos-1])*60
//                             +(date[pos+1]*10+date[pos+2]);
//
//                 line_seconds =  (line[pos2-5]*10+line[pos2-4])*3600
//                             +(line[pos2-2]*10+line[pos2-1])*60
//                             +(line[pos2+1]*10+line[pos2+2]);
//
//
//                 if(fabs(date_seconds - line_seconds + 0.0)<3)
//                     test = 1;
//             }
        }
        if (file.eof())
            bStop = kTRUE;
    }
    CPPUNIT_ASSERT_EQUAL(1, test);
}
void TPolynomialChaosTest::testgetCorrelation()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");

    //correlation between the first ouput and itself
    CPPUNIT_ASSERT_EQUAL(1.0, _pc->getCorrelation("Ishigami", "Ishigami"));
    //correlation with index is the same like correlation with name
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_pc->getCorrelation(0, 1),
            _pc->getCorrelation("Ishigami", "Lineaire"), DBL_PRECISION_COMPARE);
    //correlation between the first and the second output
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.11676805,
            _pc->getCorrelation("Ishigami", "Lineaire"), DBL_PRECISION_COMPARE);
}
void TPolynomialChaosTest::throwExceptiongetCorrelationJnameFalse()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    //false name for the second output
    CPPUNIT_ASSERT_THROW(_pc->getCorrelation("Ishigami", "toto"),
            URANIE::Exceptions::UWarningExceptions);
}
void TPolynomialChaosTest::throwExceptiongetCorrelationInameFalse()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    //false name for the first output
    CPPUNIT_ASSERT_THROW(_pc->getCorrelation("toto", "Ishigami"),
            URANIE::Exceptions::UWarningExceptions);

}
void TPolynomialChaosTest::throwExceptiongetCorrelationInegative()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    //error index for the first output
    CPPUNIT_ASSERT_THROW(_pc->getCorrelation(-1, 0),
            URANIE::Exceptions::UWarningExceptions);
}
void TPolynomialChaosTest::throwExceptiongetCorrelationJnegative()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    //error index for the second output
    CPPUNIT_ASSERT_THROW(_pc->getCorrelation(0, -1),
            URANIE::Exceptions::UWarningExceptions);
}
void TPolynomialChaosTest::throwExceptiongetCorrelationIexceed()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    //error index for the first output
    CPPUNIT_ASSERT_THROW(_pc->getCorrelation(2, 0),
            URANIE::Exceptions::UWarningExceptions);
}
void TPolynomialChaosTest::throwExceptiongetCorrelationJexceed()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    //error index for the second output
    CPPUNIT_ASSERT_THROW(_pc->getCorrelation(1, 2),
            URANIE::Exceptions::UWarningExceptions);
}
void TPolynomialChaosTest::testgetCovariance()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");

    //covariance between the first ouput and itself is the variance
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_pc->getVariance("Ishigami"),
            _pc->getCovariance("Ishigami", "Ishigami"), DBL_PRECISION_COMPARE);
    //covariance with index is the same like correlation with name
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_pc->getCovariance(0, 1),
            _pc->getCovariance("Ishigami", "Lineaire"), DBL_PRECISION_COMPARE);
    //correlation between the first and the second output
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.94818182,
            _pc->getCovariance("Ishigami", "Lineaire"), DBL_PRECISION_COMPARE);

}
void TPolynomialChaosTest::throwExceptiongetCovarianceJnameFalse()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    //false name of the second output
    CPPUNIT_ASSERT_THROW(_pc->getCovariance("Ishigami", "toto"),
            URANIE::Exceptions::UWarningExceptions);

}
void TPolynomialChaosTest::throwExceptiongetCovarianceInameFalse()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    //false name of the first output
    CPPUNIT_ASSERT_THROW(_pc->getCovariance("toto", "Ishigami"),
            URANIE::Exceptions::UWarningExceptions);
}
void TPolynomialChaosTest::throwExceptiongetCovarianceInegative()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    //false index of the first output
    CPPUNIT_ASSERT_THROW(_pc->getCovariance(-1, 0),
            URANIE::Exceptions::UWarningExceptions);
}
void TPolynomialChaosTest::throwExceptiongetCovarianceJnegative()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    //false index of the second output
    CPPUNIT_ASSERT_THROW(_pc->getCovariance(1, -2),
            URANIE::Exceptions::UWarningExceptions);
}
void TPolynomialChaosTest::throwExceptiongetCovarianceIexceed()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    //false index of the second output
    CPPUNIT_ASSERT_THROW(_pc->getCovariance(1, 2),
            URANIE::Exceptions::UWarningExceptions);
}
void TPolynomialChaosTest::throwExceptiongetCovarianceJexceed()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    //false index of the first output
    CPPUNIT_ASSERT_THROW(_pc->getCovariance(3, 1),
            URANIE::Exceptions::UWarningExceptions);
}
void TPolynomialChaosTest::testgetIndexFirstOrder()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    //index for the couple (X1,Ishigami)
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.31399700,
            _pc->getIndexFirstOrder("x1", "Ishigami"), DBL_PRECISION_COMPARE);
    //index for the couple (X3,Ishigami)
    CPPUNIT_ASSERT_DOUBLES_EQUAL(6.50043e-07,
            _pc->getIndexFirstOrder("x3", "Ishigami"), DBL_PRECISION_COMPARE);
    //index for the couple (X1,Lineaire)
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.07142857,
            _pc->getIndexFirstOrder("x1", "Lineaire"), DBL_PRECISION_COMPARE);
    //index for the couple (X3,Lineaire)
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.64285714,
            _pc->getIndexFirstOrder("x3", "Lineaire"), DBL_PRECISION_COMPARE);
    //index for the couple (X1,default output) ie (X1,Ishigami)
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_pc->getIndexFirstOrder("x1", "Ishigami"),
            _pc->getIndexFirstOrder("x1"), DBL_PRECISION_COMPARE);
    //index for the couple (0,0) ie (X1,Ishigami)
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_pc->getIndexFirstOrder("x1", "Ishigami"),
            _pc->getIndexFirstOrder(0, 0), DBL_PRECISION_COMPARE);
    //index for the couple (0,default output) ie (X1,Ishigami)
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_pc->getIndexFirstOrder("x1", "Ishigami"),
            _pc->getIndexFirstOrder(0), DBL_PRECISION_COMPARE);
    //index of th couple (2,1) ie (x3,Lineaire)
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_pc->getIndexFirstOrder("x3", "Lineaire"),
            _pc->getIndexFirstOrder(2, 1), DBL_PRECISION_COMPARE);
    //index of th couple (2,default ouput) ie (x3,Ishigami)
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_pc->getIndexFirstOrder("x3", "Ishigami"),
            _pc->getIndexFirstOrder(2), DBL_PRECISION_COMPARE);
}
void TPolynomialChaosTest::throwExceptiongetIndexFirstOrderJnameFalse()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    //false name of the output
    CPPUNIT_ASSERT_THROW(_pc->getIndexFirstOrder("x1", "toto"),
            URANIE::Exceptions::UWarningExceptions);

}
void TPolynomialChaosTest::throwExceptiongetIndexFirstOrderInameFalse()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    //false name of the input
    CPPUNIT_ASSERT_THROW(_pc->getIndexFirstOrder("toto", "Ishigami"),
            URANIE::Exceptions::UWarningExceptions);

}
void TPolynomialChaosTest::throwExceptiongetIndexFirstOrderInegative()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    //Index of the input negative
    CPPUNIT_ASSERT_THROW(_pc->getIndexFirstOrder(-1, 0),
            URANIE::Exceptions::UWarningExceptions);
}
void TPolynomialChaosTest::throwExceptiongetIndexFirstOrderJnegative()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    //Index of the output negative
    CPPUNIT_ASSERT_THROW(_pc->getIndexFirstOrder(0, -1),
            URANIE::Exceptions::UWarningExceptions);
}
void TPolynomialChaosTest::throwExceptiongetIndexFirstOrderIexceed()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    //Index of the input exceed
    CPPUNIT_ASSERT_THROW(_pc->getIndexFirstOrder(3, 0),
            URANIE::Exceptions::UWarningExceptions);
}
void TPolynomialChaosTest::throwExceptiongetIndexFirstOrderJexceed()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    //Index of the output exceed
    CPPUNIT_ASSERT_THROW(_pc->getIndexFirstOrder(2, 2),
            URANIE::Exceptions::UWarningExceptions);
}
void TPolynomialChaosTest::testgetIndexTotalOrder()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    //index for the couple (X1,Ishigami)
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.55756751,
            _pc->getIndexTotalOrder("x1", "Ishigami"), DBL_PRECISION_COMPARE);
    //index for the couple (X3,Ishigami)
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.243571163,
            _pc->getIndexTotalOrder("x3", "Ishigami"), DBL_PRECISION_COMPARE);
    //index for the couple (X1,Lineaire)
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.071428571,
            _pc->getIndexTotalOrder("x1", "Lineaire"), DBL_PRECISION_COMPARE);
    //index for the couple (X3,Lineaire)
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.642857143,
            _pc->getIndexTotalOrder("x3", "Lineaire"), DBL_PRECISION_COMPARE);
    //index for the couple (X1,default output) ie (X1,Ishigami)
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_pc->getIndexTotalOrder("x1", "Ishigami"),
            _pc->getIndexTotalOrder("x1"), DBL_PRECISION_COMPARE);
    //index for the couple (0,0) ie (X1,Ishigami)
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_pc->getIndexTotalOrder("x1", "Ishigami"),
            _pc->getIndexTotalOrder(0, 0), DBL_PRECISION_COMPARE);
    //index for the couple (0,default output) ie (X1,Ishigami)
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_pc->getIndexTotalOrder("x1", "Ishigami"),
            _pc->getIndexTotalOrder(0), DBL_PRECISION_COMPARE);
    //index of th couple (2,1) ie (x3,Lineaire)
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_pc->getIndexTotalOrder("x3", "Lineaire"),
            _pc->getIndexTotalOrder(2, 1), DBL_PRECISION_COMPARE);
    //index of th couple (2,default ouput) ie (x3,Ishigami)
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_pc->getIndexTotalOrder("x3", "Ishigami"),
            _pc->getIndexTotalOrder(2), DBL_PRECISION_COMPARE);
}
void TPolynomialChaosTest::throwExceptiongetIndexTotalOrderJnameFalse()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    //false name of the output
    CPPUNIT_ASSERT_THROW(_pc->getIndexTotalOrder("x1", "toto"),
            URANIE::Exceptions::UWarningExceptions);

}
void TPolynomialChaosTest::throwExceptiongetIndexTotalOrderInameFalse()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    //false name of the output
    CPPUNIT_ASSERT_THROW(_pc->getIndexTotalOrder("toto", "Ishigami"),
            URANIE::Exceptions::UWarningExceptions);
}
void TPolynomialChaosTest::throwExceptiongetIndexTotalOrderInegative()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    //false name of the output
    CPPUNIT_ASSERT_THROW(_pc->getIndexTotalOrder(-1, 0),
            URANIE::Exceptions::UWarningExceptions);
}
void TPolynomialChaosTest::throwExceptiongetIndexTotalOrderJnegative()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    //false name of the output
    CPPUNIT_ASSERT_THROW(_pc->getIndexTotalOrder(0, -1),
            URANIE::Exceptions::UWarningExceptions);
}
void TPolynomialChaosTest::throwExceptiongetIndexTotalOrderIexceed()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    //false name of the output
    CPPUNIT_ASSERT_THROW(_pc->getIndexTotalOrder(4, 0),
            URANIE::Exceptions::UWarningExceptions);
}
void TPolynomialChaosTest::throwExceptiongetIndexTotalOrderJexceed()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    //false name of the output
    CPPUNIT_ASSERT_THROW(_pc->getIndexTotalOrder(2, 2),
            URANIE::Exceptions::UWarningExceptions);
}
void TPolynomialChaosTest::testgetIndex()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    //Sensitivity Index for ("x1:x3","Ishigami")
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.55752254, _pc->getIndex("x1:x3", "Ishigami"),
            DBL_PRECISION_COMPARE);
    // Sensitivity Index for ("x1:x2:x3","Ishigami")
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, _pc->getIndex("x1:x2:x3", "Ishigami"),
            DBL_PRECISION_COMPARE);
    //Default sensitivity index
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_pc->getIndex("x1:x2:x3", "Ishigami"),
            _pc->getIndex(), DBL_PRECISION_COMPARE);
    //Sensitivity index of the delault ouput for the group "x1:x3")
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_pc->getIndex("x1:x3", "Ishigami"),
            _pc->getIndex("x1:x3"), DBL_PRECISION_COMPARE);
}
void TPolynomialChaosTest::throwExceptiongetIndexJnameFalse()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    CPPUNIT_ASSERT_THROW(_pc->getIndex("x1:x2:x3", "toto"),
            URANIE::Exceptions::UWarningExceptions);

}
void TPolynomialChaosTest::throwExceptiongetIndexSInputFalse()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    CPPUNIT_ASSERT_THROW(_pc->getIndex("x1:toto:x3", "Ishigami"),
            URANIE::Exceptions::UErrorExceptions);

}
void TPolynomialChaosTest::testgetIndexInteraction()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    //Interactive sensitivity Index for ("x1:x3","Ishigami")
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.24352489,
            _pc->getIndexInteraction("x1:x3", "Ishigami"),
            DBL_PRECISION_COMPARE);
    //Interactive sensitivity Index for ("x1:x2:x3","Ishigami")
    CPPUNIT_ASSERT_DOUBLES_EQUAL(4.084619869e-32,
            _pc->getIndexInteraction("x1:x2:x3", "Ishigami"),
            DBL_PRECISION_COMPARE);
    //Default Interactivesensitivity index
    CPPUNIT_ASSERT_DOUBLES_EQUAL(
            _pc->getIndexInteraction("x1:x2:x3", "Ishigami"),
            _pc->getIndexInteraction(), DBL_PRECISION_COMPARE);
    //Interactive sensitivity index of the delault ouput for the group "x1:x3")
    CPPUNIT_ASSERT_DOUBLES_EQUAL(_pc->getIndexInteraction("x1:x3", "Ishigami"),
            _pc->getIndexInteraction("x1:x3"), DBL_PRECISION_COMPARE);

}
void TPolynomialChaosTest::throwExceptiongetIndexInteractionJnameFalse()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    CPPUNIT_ASSERT_THROW(_pc->getIndexInteraction("x1:x2:x3", "toto"),
            URANIE::Exceptions::UWarningExceptions);
}
void TPolynomialChaosTest::throwExceptiongetIndexInteractionSInputFalse()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    CPPUNIT_ASSERT_THROW(_pc->getIndexInteraction("x1:toto:x3", "Ishigami"),
            URANIE::Exceptions::UErrorExceptions);

}
void TPolynomialChaosTest::getAnovaOrdered()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    Double_t seuil = 0.98;
    CPPUNIT_ASSERT_NO_THROW(_pc->getAnovaOrdered(seuil, "Ishigami"));
    CPPUNIT_ASSERT_NO_THROW(_pc->getAnovaOrdered(seuil));
}
void TPolynomialChaosTest::throwExceptiongetAnovaOrdered()
{
    _pc->setDegree(8);
    _pc->computeChaosExpansion("Integration");
    Double_t seuil = 0.98;
    CPPUNIT_ASSERT_THROW(_pc->getAnovaOrdered(seuil, "toto"),
            URANIE::Exceptions::UWarningExceptions);
}

void TPolynomialChaosTest::throwExceptionNPatternsDifferentNispTds()
{
    Double_t* XXX = new Double_t[7];
    for(int i=0;i<7;i++)
        XXX[i]=0;
    _tds->getTuple()->Fill(XXX);
    CPPUNIT_ASSERT_THROW(URANIE::Modeler::TPolynomialChaos(_tds, _nisp),
                         URANIE::Exceptions::UErrorExceptions);

    delete[] XXX;

}

void TPolynomialChaosTest::throwExceptionConstructorByCopyWithIndisponibleLaw()
{
    URANIE::DataServer::TDataServer* tds = new URANIE::DataServer::TDataServer();
    tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x1",
                    -1. * TMath::Pi(), TMath::Pi()));
    tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x2",
                    -1. * TMath::Pi(), TMath::Pi()));
    tds->getAttribute("x1")->delShare();
    tds->getAttribute("x2")->delShare();
  
    URANIE::DataServer::TBetaDistribution* tud = new URANIE::DataServer::TBetaDistribution("x3",1,1);
    tds->addAttribute( tud );
    tud->delShare();
    URANIE::Modeler::TNisp* nisp = new URANIE::Modeler::TNisp(tds);
    nisp->generateSample("Petras", 4);
    URANIE::Launcher::TLauncherFunction* tlf = 
#ifdef ROOT5
        new URANIE::Launcher::TLauncherFunction(tds, "Ishigami", "", "Ishigami:Lineaire");
#else
        new URANIE::Launcher::TLauncherFunction(tds, &Ishigami, "", "Ishigami:Lineaire");
#endif
    tlf->setDrawProgressBar(kFALSE);
    tlf->run();
    URANIE::Modeler::TPolynomialChaos* pc = new URANIE::Modeler::TPolynomialChaos(tds, nisp);
    pc->unsetLog();
    CPPUNIT_ASSERT_THROW(URANIE::Modeler::TPolynomialChaos pcCopy(pc[0]),
                         URANIE::Exceptions::UErrorExceptions);


    delete pc;
    delete tlf;
    delete nisp;
    delete tds;
}


void TPolynomialChaosTest::throwExceptionSetGroupAddVarWithRankTooHigh()
{
    _tds->addAttribute( new URANIE::DataServer::TUniformDistribution("x4",1,1));
    _tds->getAttribute("x4")->delShare();
    _nisp->_listAttIn->addAttribute(new URANIE::DataServer::TUniformDistribution("x4",1,1));
    _nisp->_listAttIn->getAttribute("x4")->delShare();
    CPPUNIT_ASSERT_THROW(_pc->getIndex("x4"),
            URANIE::Exceptions::UWarningExceptions);
}


void TPolynomialChaosTest::throwExceptionSetCoefficientNumberOfOutputTooHigh()
{

    CPPUNIT_ASSERT_THROW(_pc->setCoefficient(3, 0, 0.5),
            URANIE::Exceptions::UErrorExceptions);
}

void TPolynomialChaosTest::throwExceptionSetCoefficientNumberOfPCTooHigh()
{

    CPPUNIT_ASSERT_THROW(_pc->setCoefficient(1, 1000, 0.5),
            URANIE::Exceptions::UErrorExceptions);
}

void TPolynomialChaosTest::throwExceptionGetCoefficientNumberOfOutputTooHigh()
{

    CPPUNIT_ASSERT_THROW(_pc->getCoefficient(3, 0),
            URANIE::Exceptions::UErrorExceptions);
}

void TPolynomialChaosTest::throwExceptionGetCoefficientNumberOfPCTooHigh()
{

    CPPUNIT_ASSERT_THROW(_pc->getCoefficient(1, 1000),
            URANIE::Exceptions::UErrorExceptions);
}

void TPolynomialChaosTest::throwExceptiongetAnovaOrderedCoefficients()
{

    CPPUNIT_ASSERT_THROW(_pc->getAnovaOrderedCoefficients(0.9, "toto"),
            URANIE::Exceptions::UWarningExceptions);
}

void TPolynomialChaosTest::throwExceptiongetSampleWrongOutputName()
{

    CPPUNIT_ASSERT_THROW(_pc->getSample(1, "toto"),
            URANIE::Exceptions::UWarningExceptions);
}

// void TPolynomialChaosTest::testGetSampleWithOutputName()
// {
//
//     cout<<_pc->getSample(1, "Ishigami")<<endl;
// }

void TPolynomialChaosTest::throwExceptiongetQuantileWrongOutputName()
{

    CPPUNIT_ASSERT_THROW(_pc->getQuantile(0.5, "toto"),
            URANIE::Exceptions::UWarningExceptions);

    CPPUNIT_ASSERT_THROW(_pc->getQuantileWilks(0.5, 0.1, "toto"),
            URANIE::Exceptions::UWarningExceptions);

    CPPUNIT_ASSERT_THROW(_pc->getInvQuantile(0.5, "toto"),
            URANIE::Exceptions::UWarningExceptions);
}

void TPolynomialChaosTest::throwExceptionGetCoefficientWithOutputName()
{

    CPPUNIT_ASSERT_THROW(_pc->getCoefficient(1000, "Ishigami"),
            URANIE::Exceptions::UWarningExceptions);

    CPPUNIT_ASSERT_THROW(_pc->getCoefficient(-1, "Ishigami"),
            URANIE::Exceptions::UWarningExceptions);

    CPPUNIT_ASSERT_THROW(_pc->getCoefficient(1, "toto"),
            URANIE::Exceptions::UWarningExceptions);
}

void TPolynomialChaosTest::testGetQuantiles()
{
    URANIE::DataServer::TDataServer* tds = new URANIE::DataServer::TDataServer();
    tds->addAttribute(
            new URANIE::DataServer::TNormalDistribution("x1", 0 , 0.7071068));
    tds->addAttribute(
            new URANIE::DataServer::TNormalDistribution("x2", 0 , 0.7071068));

    tds->getAttribute("x1")->delShare();
    tds->getAttribute("x2")->delShare();

    URANIE::Modeler::TNisp* nisp = new URANIE::Modeler::TNisp(tds);
    URANIE::Sampler::TPetras *sampler = new URANIE::Sampler::TPetras(tds,"",2);
    sampler->generateSample();
    nisp->setSample("Petras");
    nisp->setLevel(2);

    URANIE::Launcher::TLauncherFunction* tlf = new URANIE::Launcher::TLauncherFunction(tds, fSum,"", "fSum");
    tlf->setDrawProgressBar(kFALSE);
    tlf->run();
    URANIE::Modeler::TPolynomialChaos * pc = new URANIE::Modeler::TPolynomialChaos(tds,nisp);
    pc->unsetLog();
    pc->setDegree(1);
    pc->computeChaosExpansion("Integration");
    InitSeed(1);
    pc->generateSample("Lhs",20000);
// cout<<"###################################################"<<endl;
// cout<<setprecision(15)<<pc->getQuantile(0.95, 0) << "    " << pc->getQuantile(0.5, 0) << endl;

// cout<<setprecision(15)<<pc->getQuantileWilks(0.95,0.9, 0) << "    " << pc->getQuantileWilks(0.5,0.9, 0) << endl;

// cout<<setprecision(15)<<pc->getInvQuantile(0.0, 0) << "    " << pc->getInvQuantile(0.5, 0) << endl;
#ifndef WIN32
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.32096476540663 ,pc->getQuantile(0.95, 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0083983870635938 ,pc->getQuantile(0.5, 0), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.32096476540663 ,pc->getQuantile(0.95, "fSum"), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0083983870635938 ,pc->getQuantile(0.5, "fSum"), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.33664065009463,pc->getQuantileWilks(0.95,0.9, 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0183749710292739,pc->getQuantileWilks(0.5,0.9, 0), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.33664065009463 ,pc->getQuantileWilks(0.95,0.9, "fSum"), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0183749710292739,pc->getQuantileWilks(0.5,0.9, "fSum"), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.49605 ,pc->getInvQuantile(0.0, 0), 0.0001);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.7295 ,pc->getInvQuantile(0.5, 0), 0.0001);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.49605,pc->getInvQuantile(0.0, "fSum"), 0.0001);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.7295 ,pc->getInvQuantile(0.5, "fSum"), 0.0001);
#else
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.3569680444339471, pc->getQuantile(0.95, 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.0058677754615533706, pc->getQuantile(0.5, 0), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.3569680444339471, pc->getQuantile(0.95, "fSum"), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.0058677754615533706, pc->getQuantile(0.5, "fSum"), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.3737043824765067, pc->getQuantileWilks(0.95,0.9, 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.003672720321017231, pc->getQuantileWilks(0.5,0.9, 0), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.3737043824765067, pc->getQuantileWilks(0.95,0.9, "fSum"), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.003672720321017231, pc->getQuantileWilks(0.5,0.9, "fSum"), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.50305, pc->getInvQuantile(0.0, 0), 0.0001);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.73325, pc->getInvQuantile(0.5, 0), 0.0001);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.50305, pc->getInvQuantile(0.0, "fSum"), 0.0001);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.73325, pc->getInvQuantile(0.5, "fSum"), 0.0001);
#endif
    free_dmatrix(pc->_pc->sample,pc->_pc->ny);
    free_imatrix(pc->_pc->sampleindx, pc->_pc->ny);
    delete pc;
    delete sampler;
    delete tlf;
    delete nisp;
    delete tds;

}


void TPolynomialChaosTest::throwExceptionGetAnovaOrdered()
{

    CPPUNIT_ASSERT_THROW(_pc->getAnovaOrdered(0.95, -1),
            URANIE::Exceptions::UWarningExceptions);

    CPPUNIT_ASSERT_THROW(_pc->getAnovaOrdered(0.95, 2),
            URANIE::Exceptions::UWarningExceptions);


}

void TPolynomialChaosTest::throwExceptionGetSample()
{

    CPPUNIT_ASSERT_THROW(_pc->getSample(1, -1),
            URANIE::Exceptions::UWarningExceptions);

    CPPUNIT_ASSERT_THROW(_pc->getSample(1, 2),
            URANIE::Exceptions::UWarningExceptions);


}

void TPolynomialChaosTest::throwExceptionGetEqmLoo()
{
    CPPUNIT_ASSERT_THROW(_pc->getEqmLoo(-1), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(_pc->getEqmLoo(1000), URANIE::Exceptions::UErrorExceptions);
}

void TPolynomialChaosTest::throwExceptionGetErrLoo()
{
    CPPUNIT_ASSERT_THROW(_pc->getErrLoo(-1), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(_pc->getErrLoo(10000), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(_pc->getErrLoo(5,1000), URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(_pc->getErrLoo(5,-1), URANIE::Exceptions::UErrorExceptions);
}

void TPolynomialChaosTest::throwExceptionSetAutoDegreeBoundaries()
{

  URANIE::DataServer::TDataServer *tds = new URANIE::DataServer::TDataServer();
  tds->addAttribute( new URANIE::DataServer::TUniformDistribution("x1", -1. * TMath::Pi(), TMath::Pi()));
  tds->addAttribute( new URANIE::DataServer::TUniformDistribution("x2", -1. * TMath::Pi(), TMath::Pi()));
  tds->addAttribute( new URANIE::DataServer::TUniformDistribution("x3", -1. * TMath::Pi(), TMath::Pi()));
  tds->getAttribute("x1")->delShare();
  tds->getAttribute("x2")->delShare();
  tds->getAttribute("x3")->delShare();
  URANIE::Modeler::TNisp *nisp = new URANIE::Modeler::TNisp(tds);
  Int_t level = 250;
  nisp->generateSample("Lhs",level);
#ifdef ROOT5
  URANIE::Launcher::TLauncherFunction * tlf = new  URANIE::Launcher::TLauncherFunction(tds, "Ishigami","","Ishigami:Lineaire");
#else
  URANIE::Launcher::TLauncherFunction * tlf = 
        new URANIE::Launcher::TLauncherFunction(tds, &Ishigami, "","Ishigami:Lineaire");
#endif
  tlf->setDrawProgressBar(kFALSE);
  tlf->run();
  URANIE::Modeler::TPolynomialChaos * pc = new URANIE::Modeler::TPolynomialChaos(tds,nisp);
  CPPUNIT_ASSERT_THROW(pc->setAutoDegreeBoundaries(-1), URANIE::Exceptions::UErrorExceptions);
  CPPUNIT_ASSERT_THROW(pc->setAutoDegreeBoundaries(3,2), URANIE::Exceptions::UErrorExceptions);
  CPPUNIT_ASSERT_THROW(pc->setAutoDegreeBoundaries(9), URANIE::Exceptions::UErrorExceptions);
  delete pc;
  delete tlf;
  delete nisp;
  delete tds;
}

void TPolynomialChaosTest::throwExceptionSetAutoDegreeFactor()
{
    CPPUNIT_ASSERT_THROW(_pc->setAutoDegreeFactor(0.9), URANIE::Exceptions::UErrorExceptions);
}
