//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TMultiGenAnalyticalFunction.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   lun sep 20 13:46:01 2010
 
 \brief Analytical Function to mimize whith the new TMinuit2 wrapper

 \see http://root.cern.ch/drupal/content/numerical-minimization
 \see http://root.cern.ch/viewvc/trunk/math/minuit2/test/testNewMinimizer.cxx?view=markup

 */
#ifndef TMULTIGENANALYTICALFUNCTION_H
#define TMULTIGENANALYTICALFUNCTION_H

using namespace std;

// ROOT
#include "Math/IFunction.h"
#include "TMethodCall.h"
#include "TTreeFormula.h"
#include "TNtupleD.h"

// URANIE
#include "Optimizer.h"
#include "TDataServer.h"

namespace URANIE
{
namespace Optimizer
{
class TMultiGenAnalyticalFunction: public ROOT::Math::IBaseFunctionMultiDim
{

public:
    //---------------------------------------------
    /** @name Constructor and Desctructor
     **/
    //@{
    //! Constructor with a dataserver
#ifdef ROOT5
    TMultiGenAnalyticalFunction(URANIE::DataServer::TDataServer *tds, void * fcn,
				TString sinput, TString soutput, TString scost = "");
#else
    TMultiGenAnalyticalFunction(URANIE::DataServer::TDataServer *tds, void  (*fcn)(Double_t*, Double_t*),
				TString sinput, TString soutput, TString scost = "");
#endif
     //! Constructor with a dataserver and function name only
    TMultiGenAnalyticalFunction(URANIE::DataServer::TDataServer *tds, const char* fcn,
				TString sinput, TString soutput, TString scost = "");
    //! Desctructor
    ~TMultiGenAnalyticalFunction();

    //---------------------------------------------
    /** @name Methods
     **/
    //@{
    //! The init method
    void init();
    //! The clean method
    void clean();
    //@}

    //---------------------------------------------
    /** @name pure virtual methods
     **/
    //@{
    //! The clone method
    TMultiGenAnalyticalFunction * Clone() const
    {
#ifdef ROOT5
        return new TMultiGenAnalyticalFunction(_tds, (void *) _fMethod, _sinput,
                _soutput, _sCost);
#else
        return new TMultiGenAnalyticalFunction(_tds, _fMethodCall->GetMethodName(), _sinput,
                _soutput, _sCost);
#endif
    }
    //! the NDIM method
    unsigned int NDim() const
    {
        return _ninput;
    }
    //@}

    //---------------------------------------------
    /** @name Setting and Getting attributs
     **/
    //@{
    //! Return the pointer to the TmethodCall attribut
    TMethodCall *getMethodCall() const
    {
        return _fMethodCall;
    }
    //@}
    
     //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    void setLog()
    {
        _blog = kTRUE;
    }
    void unsetLog()
    {
        _blog = kFALSE;
    }
    
    void changeLog()
    {
        _blog = _blog ? kFALSE : kTRUE;
    }
    Bool_t getLog()
    {
        return _blog;
    }
    
    //@}
    

public:
    Double_t *getOutputValues() { return _doutputValue;}

private:
    //---------------------------------------------
    /** @name pure virtual methods
     **/
    //@{
    //! the DoEval method
    double DoEval(const double * x) const;
    //@}

    Bool_t _blog;                            ///< Boolean for edit the log
    int _ninput;
    TString _sinput;                          ///< Name of the input attributes
    double * _dinputValue;
    double * _xin, *_codevalue;
    int *_codeindex ;
    int _noutput;
    int _ninformula;
    int _noutformula;
    TString _soutput;                         ///< Name of the output attributes
    double * _doutputValue;                       ///< The output attribute name
    TString _sCost; ///< The name of the selected cost (the number must be equal to 1)
    Int_t _nCostIndex; ///< the current index for the selected cost in the output list of attribute
    TString _sWorkingDirectory;             ///< The Working directory
    URANIE::DataServer::TDataServer *_tds;              //! Pointeur vers un TDS
    Int_t _CostInOutput;                           ///< Is cost in output of the function or computed on top
    void (*_fMethod)(Double_t *gin, Double_t *f); // A pointer for the analytical function
    TMethodCall *_fMethodCall; ///< Pointer to MethodCall in case of interpreted function
    TNtupleD *_informTree; 
    vector<TTreeFormula*> _vinForm;
    
    ClassDef(URANIE::Optimizer::TMultiGenAnalyticalFunction, ID_OPTIMIZER)
    //Classe de

};
} // Fin du namespace Optimizer
} // Fin du namespace URANIE

R__EXTERN URANIE::Optimizer::TMultiGenAnalyticalFunction *gTMGAF;

#endif
