//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////

/*!
 \file  TMultiGenSumOfSquares.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date     
 \brief Implementation of the class URANIE::Optimizer::TMultiGenSumOfSquares

 \par Modifications
 \li M. Salvoldi (Areva/OCC) 2011/09/28
 - add a IFDEF WIN32 to replace rm -fr by del /Q on Windows platform
 
 */

#include <sstream>
#include <string>

// ROOT
//fga 2015/01/12 #include "Api.h"
#include "TPad.h"

// URANIE
#include "TMultiGenSumOfSquares.h"
#include "UExceptions.h"

URANIE::Optimizer::TMultiGenSumOfSquares *gTMGCSOS;

ClassImp (URANIE::Optimizer::TMultiGenSumOfSquares)

namespace URANIE
{
namespace Optimizer
{

////////////////
// Constructor 
////////////////
TMultiGenSumOfSquares::TMultiGenSumOfSquares(URANIE::DataServer::TDataServer *tds, URANIE::Launcher::TCode *code) :
  _objectives(0), _blog(kFALSE), _ninput(0), _dinputValue(0), _noutput(0),
  _doutputValue(0), _dSumOfWeight(0.0), _tds(tds), _code(code), _listOfDataOutput(NULL),  _sCriteria("")
{
    try
    {
        if (_tds == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TMultiGenSumOfSquares::constructor\n The pFointer of \"TDataServer\" is empty");
        if (_code == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TMultiGenSumOfSquares::constructor\n The pointer of \"TCode\" is empty");

        _ninput = _tds->getNInputAttributes();
        _noutput = _tds->getNOutputAttributes();

        cout << " ** _ninput [ " << _ninput << "]" << endl;
        cout << " ** _noutput [ " << _noutput << "]" << endl;

        // Erase the output Attributes of the Constructor
        for (Int_t i = _tds->getNAttributes() - 1; i > -1; i--)
        {
            URANIE::DataServer::TAttribute *att = _tds->getAttribute(i);
            cout << " ** Attribute[" << att->GetName() << "]";
            if (att->isOutput())
            {
                cout << "-- Del";
                _tds->delAttribute(att->GetName());
            }
            cout << endl;
        }

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

}

//////////////////////////////////
// Destructor
//////////////////////////////////
TMultiGenSumOfSquares::~TMultiGenSumOfSquares()
{
    delete[] _dinputValue;
    _dinputValue = NULL;
    delete[] _doutputValue;
    _doutputValue = NULL;
    
    for(vector<TObjective*>::iterator it=_objectives.begin(); it!=_objectives.end(); it++)
      {
	(*it)->delShare();
	if( (*it)->getShare()==0)
	  delete (*it);
      }
    _objectives.clear();

    //    _listOfDataOutput->Delete();
    delete _listOfDataOutput;
    _listOfDataOutput=NULL;
}

//////////////////////////////////
// addObjective
//////////////////////////////////
void TMultiGenSumOfSquares::addObjective(TObjective *tobj)
{
    try
    {
        if (tobj == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TMultiGenSumOfSquares::addObjective\n The pointer of \"TObjective\" is empty");
	tobj->addShare();
        _objectives.push_back(tobj);

        _tds->addAttribute(tobj->GetName());
        _tds->getAttribute(tobj->GetName())->setOutput();

        // looking for the cost attribute (sum of objectives)
        _sCriteria = "";
        Int_t nObjectives = (Int_t)(_objectives.size());
        _dSumOfWeight = 0.0;

        cout << " ** list of Objectives size(" << nObjectives
                << ") _dSumOfWeight(" << _dSumOfWeight << ")" << endl;
        for (Int_t indx = 0; indx < nObjectives; indx++)
        {
            cout << " ** Objective i[" << indx << "/" << nObjectives
                    << "] Name[" << _objectives[indx]->GetName() << "] weight( "
                    << _objectives[indx]->getWeight() << ")" << endl;
            _sCriteria = _objectives[indx]->GetName();
            if (_objectives[indx]->isActive())
                _dSumOfWeight += _objectives[indx]->getWeight();
        }

        if (nObjectives > 1)
        {
            _sCriteria = "__cost__";
            if (nObjectives == 2)
            {
                _tds->addAttribute("__cost__");
                _tds->getAttribute("__cost__")->setOutput();
            }
        }

        cout << " ** End Of list of Objectives _dSumOfWeight[" << _dSumOfWeight
                << "]" << endl;
        cout << " ** _sCriteria [ " << _sCriteria << "]" << endl;
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

//////////////////////////////////
// init
//////////////////////////////////
void TMultiGenSumOfSquares::init()
{

    cout << endl << "*************************************" << endl
            << " *** TMultiGenSumOfSquares::init" << endl;

    // We recreate the TDS for integrate Objective and Final Objective;
    _tds->createTuple();

    if (_blog)
        _tds->getTuple()->Print();

    // The size of the input vector is increased by two
    // - one for the iterator
    // - one for the criteria for drawing
    _dinputValue = new double[_ninput + _objectives.size() + 2];
    _doutputValue = new double[_noutput + 1];

    clean();

    _listOfDataOutput = new TList();

    cout << " *** End Of TMultiGenSumOfSquares::init" << endl
            << "*************************************" << endl;
}

//////////////////////////////////
// DoEval
//////////////////////////////////
double TMultiGenSumOfSquares::DoEval(const double * par) const
{
    Int_t niter = (Int_t) _tds->getTuple()->GetEntries();
    Double_t fval = -0.123456789;
    _tds->getAttribute(_sCriteria)->getDefaultValue(fval);
    //    cout << " ** fval [ " << fval << "]" << endl;

    if (_code != NULL)
    {
        if (_blog)
        {
            cout << endl << " ****** Step[" << niter << "]" << endl;
        }
        _dinputValue[0] = 1.0 + niter;
        // Remark: If we want to save all computations, we need a constant Bool_t _bSave,
        // and call the the method _code->init() at each new computation.
        _code->init();
        // Pretraitment of the data
        Double_t valcrt = 1.23456789;
        Int_t ncpt = 0;

        // fga::modif le 2008-10-01 07:49:37
        // attributes kConstant must be forgotten
        for (int i = 0; i < _ninput; i++)
        {
            // end of fga::modif le 2008-10-01 07:49:37
            URANIE::DataServer::TAttribute *att = _tds->getAttribute(i);
            if (_code->getLog())
                cout << " ** " << i << "/" << _ninput << " name ["
                        << att->GetName() << "]";
            URANIE::DataServer::TAttribute::EOrigin norigin = att->getOrigin();
            switch (norigin)
            {
            case URANIE::DataServer::TAttribute::kConstant:
                if (_code->getLog())
                    cout << " Origin[kConstant]";
                att->getDefaultValue(valcrt);
                break;
            case URANIE::DataServer::TAttribute::kAttribute:
                if (_code->getLog())
                    cout << " Origin[kAttribute]";
                valcrt = par[ncpt++];
                break;
	    default:
		break;
            }
            _dinputValue[1 + i] = valcrt;
            if (_code->getLog())
                cout << " val[" << _dinputValue[1 + i] << "]" << endl;
        }
        int ind = 1;
        _code->preTraitment(ind, _dinputValue);

        // Launch the code
        _code->run();
        // Post-traitment of the data
        if (_blog)
            cout << "run finished" <<endl ;
        ind = 0;
        _code->postTraitment(ind, _doutputValue, "", _listOfDataOutput);
        if (_blog)
            cout << "postTraitment finished" <<endl ;
        if (_blog)
        {
            cout << endl << "*************************************" << endl
                    << " *** TMultiGenSumOfSquares operator () _noutput["
                    << _noutput << "] ind[" << ind << "]" << endl;
            for (Int_t ii = 0; ii <= _noutput; ii++)
                cout << " ** ii(" << ii << "/" << _noutput << ") val("
                        << _doutputValue[ii] << ")" << endl;
            cout << " *** End Of TMultiGenSumOfSquares DoEval" << endl
                    << "*************************************" << endl;
        }

        Int_t noutmatrix = _listOfDataOutput->GetEntries();
        if (_blog)
            cout << " ** noutmatrix[" << noutmatrix << "]" << endl;
        TMatrixT<double> *lastmat =
                (TMatrixT<double> *) _listOfDataOutput->Last();
        Int_t nrows = lastmat->GetNrows();
        Int_t ncols = lastmat->GetNcols();
        if (_blog)
            cout << " ** nrows[" << nrows << "] ncols[" << ncols << "]" << endl;
        Double_t *allvalues;// = new Double_t[nrows];
        Double_t *dvalues = new Double_t[nrows];
        if (_blog)
            cout << " **********************************************************" << endl;
        Double_t dObjectives = 0.;
        Double_t dvalcrt;
        Int_t nObjectives = (Int_t)(_objectives.size());

        if (_blog)
           cout << " ** nObjectives[" << nObjectives << "]" << endl;

        for (Int_t indx = 0; indx < nObjectives; indx++)
        {
            // Output Matrix for the objective
            TMatrixT<double> *outMatrix =
                    (TMatrixT<double> *) _listOfDataOutput->At(indx);


            ncols = outMatrix->GetNcols();

            if (_blog) {
                cout << " ** objective(" << indx << "/"<<nObjectives <<") matrix(" << outMatrix->GetNrows() << ", " << ncols <<")" << endl;
                outMatrix->Print("mmmmmm");
            }

            allvalues = outMatrix->GetMatrixArray();

            for(Int_t jobjective=0; jobjective < ncols; jobjective++)
            {
                if (_blog)
                    cout << "   *** jobjective("<< jobjective << "/" << ncols <<")"<<endl;
                for (Int_t indy = 0; indy < nrows; indy++) {
                    dvalues[indy] = allvalues[ncols*indy+jobjective];
                    if (_blog)
                        cout << " *** indx(" << indx << ") jobjective(" << jobjective << ") indy(" << indy << ") Value(" << dvalues[indy] << ")" << endl;
                }

                dvalcrt = _objectives[indx]->getObjective(dvalues, nrows);
                _dinputValue[_ninput + indx + 1] = dvalcrt; //_objectives[indx]->getObjectiveWithoutWeight();
                dObjectives += dvalcrt;
                if (_blog)
                    cout << " ** Objective [" << indx + 1 << "/" << nObjectives
                        << "] name[" << _objectives[indx]->GetName()
                        << "] dweight[" << _objectives[indx]->getWeight()
                        << "] val[" << dvalcrt << "] Sum [" << dObjectives
                        << "]" << endl;
                // increment the number of objectives found in the current matrix
                indx++;
            }
        }
        dObjectives /= _dSumOfWeight; //(Double_t)(nObjectives);
        if (_blog)
        {
            cout << " **********************************************************" << endl;
            cout << " ** Objective [" << dObjectives << "] Weight["
                    << _dSumOfWeight << "/ " << nObjectives << "]" << endl;
            cout << " **********************************************************" << endl;
        }

	_listOfDataOutput->Delete();
        fval = dObjectives;

        //Fill the data
        _dinputValue[_ninput + nObjectives + 1] = fval;
        _tds->getTuple()->Fill(_dinputValue);

        // Export the TDS in the uranieLauncher working directory
        if (_blog)
            cout << " ** _code->getWorkingDirectory("<<_code->getWorkingDirectory() << ")"<< endl;

        TString srootfile = _code->getWorkingDirectory();
        _tds->exportData(
                Form("%s/UranieResults/_%s_iteration_%d_.dat", srootfile.Data(),
                        _tds->GetName(), niter));

        // Save the TDSNtupleD in a RootFile
        if (1)
        {
            srootfile += "/UranieResults/uranielauncher.root";
            if (niter % 5 == 0)
            {
                TString ssavefile = srootfile + TString(".bak");
                gSystem->CopyFile(srootfile.Data(), ssavefile.Data(), kTRUE);
            }
            TFile * rootfile = TFile::Open(srootfile.Data(), "RECREATE");
            _tds->getTuple()->SetDirectory(0);
            _tds->getTuple()->Write();
            rootfile->Close();
            _tds->getTuple()->SetDirectory(gDirectory);
        }

        // Visualisation
        if (1)
        {
            _tds->Draw(
                    Form("%s:%s", _sCriteria.Data(), _tds->getIteratorName()),
                    "", "lp");
            gPad->Modified();
            gPad->Update();
        }
        delete[] dvalues;
    }
    
    
    return fval;
}

//////////////////////////////////
// clean
//////////////////////////////////
void TMultiGenSumOfSquares::clean()
{
    cout << " *** TMultiGenSumOfSquares::clean" << endl;
    TString sworkindirectory = _code->getWorkingDirectory();

    // Place in the good directory
    if (sworkindirectory.Length() != 0)
        gSystem->cd(sworkindirectory);

    // Clean the subdirectory "UranieLauncher_*"
#ifdef WIN32
    // gSystem->Exec("rd /s /q UranieLauncher_*");
	gSystem->Exec( Form("for /d %%i in ( %s\\UranieLauncher_*) do @rmdir /s /q %%i", gSystem->pwd()));
#else
    gSystem->Exec("rm -fr UranieLauncher_*");
#endif
}
} // Fin du namespace Optimizer
} // Fin du namespace URANIE
