//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TOptimizer.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   tue jul 3 10:38:56 CEST 2007
 
 \brief Interface of class URANIE::Optimize::TOptimizer
 
 */
/**
 
 * @class URANIE::Optimizer::TOptimizer
 *
 * @brief Description of the class TOptimizer.
 *
 * To be written by the developper.
 *
 */
#ifndef TOPTIMIZER_H
#define TOPTIMIZER_H

#include <iostream>

//ROOT
#include "Rtypes.h"
#include "TNamed.h"
#include "TMethodCall.h"

// Uranie
#include "Optimizer.h"
#include "TCode.h"
#include "TDataServer.h"
#include "TObjective.h"
#include "TObjectiveUncertainty.h"

using namespace std;

namespace URANIE
{
namespace Optimizer
{
class TOptimizer: public TNamed
{
public:
    enum EOptimType
    {
        kMinimizeCode, kMinimizeFunction, kSumOfSquare
    };
    enum EOptimMethod
    {
        kMigrad, kSimplex, kCombined, kScan
    };

private:
    TString _sdirectory;                        ///< Current directory
    Bool_t _bsave;        ///< Saves all the computation files in each directory
    Bool_t _bclean;    ///< Cleans the Working Directory before running the code
    Bool_t _blog;                              ///< Boolean to edit the log
    Int_t _nPrintLevel; ///< The level of Debg between \f$ [1, 5]\f$ without test
    URANIE::DataServer::TDataServer *_tds;      //! Pointer vers un TDS
    URANIE::Launcher::TCode *_code;     //! Pointer vers un TCode
    Int_t _nMaxIterations;                  ///< The max number of iterations
    Int_t _nMaxFunctionCalls;      ///< The max number to call the code/function
    Double_t _dTolerance;                      ///< The tolerance parameter
    vector<TObjective *> _objectives;           ///< vector of objectives
    TMethodCall *_fMethodCall; ///< Pointer to MethodCall in case of interpreted function
#ifdef ROOT5
    void *_fMethod;                    //! A pointer for the analytical function
#else
    void (*_fMethod)(Double_t*, Double_t*);                    //! A pointer for the analytical function
#endif
    TString _sFunctionName;              //! Function name 
    EOptimType _optimProblem;                   ///< The problem of optimisation
    EOptimMethod _optimMethod;                  ///< The method of optimisation
    TString _sInput;                       ///< The string of input attributes
    TString _sOutput; ///< The string of output attributes (the number must be equal to 1)
    TString _sCost; ///< The name od the selected cost (the number must be equal to 1)
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Constructor with a dataserver
    TOptimizer(URANIE::DataServer::TDataServer *tds,
            URANIE::Launcher::TCode *code);
    //! Constructor by a TDataServer, a function and  list of input and output (optional)
#ifdef ROOT5
    TOptimizer(URANIE::DataServer::TDataServer *tds, void* fcn, TString sinput =
            "", TString soutput = "");
#else
    TOptimizer(URANIE::DataServer::TDataServer *tds, void (*fcn)(Double_t*, Double_t*),
               TString sinput, TString soutput);
#endif
    //! Constructor by a TDataServer, a function  name and  list of input and output (optional)
    TOptimizer(URANIE::DataServer::TDataServer *tds, const char* fcn, TString sinput =
	       "", TString soutput = "");
    //! Default destructor
    virtual ~TOptimizer();
    //@}

    //---------------------------------------------
    /** @name Set/Get parameters
     **/
    //@{
    //! Set the tolerance parameter
    void setTolerance(Double_t dtol)
    {
        _dTolerance = dtol;
    }
    //! Get the tolerance parameter
    Double_t getTolerance()
    {
        return _dTolerance;
    }
    //! Set the maximum number of iterations
    void setMaxIterations(Int_t nmax)
    {
        _nMaxIterations = nmax;
    }
    //! Get the maximum number of iterations
    Int_t getMaxIterations()
    {
        return _nMaxIterations;
    }
    //! Set the maximum number of evaluations
    void setMaxFunctionCalls(Int_t nmax)
    {
        _nMaxFunctionCalls = nmax;
    }
    //! Get the maximum numbers of evaluations
    Int_t getMaxFunctionCalls()
    {
        return _nMaxFunctionCalls;
    }
    //@}

    //---------------------------------------------
    /** @name Set/Get the method of Optimization
     These classe is a wrapper with the \b TMinuit library interfaced in \b ROOT which contains several algorithm of optimization :
     \li \em kMigrad : the \b Migrad  algorithm; it is a quasi Newton algorithm. It is the default algorithm selected in \b Uranie.
     \li \em kSimplex : the \b Simplex algorithm
     \li \em kCombined :
     \li \em kScan : 
     **/
    //@{
    //! Select the algorithm of optimization
    void setMethod(EOptimMethod method = kMigrad)
    {
        _optimMethod = method;
    }
    //! Return the algorithm of optimization
    EOptimMethod getMethod()
    {
        return _optimMethod;
    }
    //! Return the type of optimization
    EOptimType getOptimType()
    {
        return _optimProblem;
    }
    //! Set the print Level of Debug between \f$ [1, 5]\f$
    void setPrintLevel(int level)
    {
        _nPrintLevel = level;
    }
    //! Get the print Level of Debug between \f$ [1, 5]\f$
    int getPrintLevel() const
    {
        return _nPrintLevel;
    }
    //@}

    //---------------------------------------------
    /** @name Manipulate the code
     **/
    //@{
    //! Init phase of the code
    virtual void init(Option_t *option = "");
    //! Runs phase of the code (TMinuit2)
    virtual void optimize(Option_t *option = "");
    //! Terminates phase of the code
    virtual void terminate(Option_t *option = "");
    //! Cleans the working directory of the code
    virtual void clean(Option_t *option = "");
    //@}

    //---------------------------------------------
    /** @name Manipulate the Objectives
     In the case of calibration of code, we can define one or several \em "Objectives" to specify the criteria to minimize.
     The criteria to minimize is a sum of weighted objectives:
     \f[
     C = \sum_{i=1}^{nO}{\omega_i Obj_{i}}
     \f]
     By default, all added objectives are \em "active" and the weight \f$ \omega = 1.0\f$.
     **/
    //@{
    //! Add the objective in the TOptimizer object : Sum of square of the error
    /*!
     The objective is computed with formula 
     \f[
     obj = \alpha \sum_{i=1}^{n}{(y^{\star}_{i} - \hat{y}_{i})^{2}}
     \f]
     \param name (TString) the name of the TObjective
     \param tds (TDataServer *) The pointer of the TDataServer object which contains the attribute \f$ y^{\star}\f$
     \param ystar (TString) the name of the \f$ y^{\star}\f$ attribute
     \param outfile (TOutputFile *) The pointer of the TOutputFile object which contains the attribute \f$ \hat{y}\f$
     \param yhat (TString) the name of the \f$ \hat{y}\f$ attribute
     \param weight (Double_t) The weight of the objective [1.0]
     */
    void addObjective(TString name, URANIE::DataServer::TDataServer *tds,
            TString ystar, URANIE::Launcher::TOutputFile *outfile, TString yhat,
            Double_t weight = 1.0);
    //! Add the objective in the TOptimizer object : Sum of square of the weighted error
    /*!
     The objective is computed with formula 
     \f[
     obj = \alpha \sum_{i=1}^{n}{\frac{(y^{\star}_{i} - \hat{y}_{i})^{2}}{\sigma_{i}^{2}}}
     \f]
     \param name (TString) the name of the TObjective
     \param tds (TDataServer *) The pointer of the TDataServer object which contains the attribute \f$ y^{\star}\f$
     \param ystar (TString) the name of the \f$ y^{\star}\f$ attribute
     \param sigma (TString) the name of the weight \f$ \sigma\f$ attribute
     \param outfile (TOutputFile *) The pointer of the TOutputFile object which contains the attribute \f$ \hat{y}\f$
     \param yhat (TString) the name of the \f$ \hat{y}\f$ attribute
     \param alpha (Double_t) The weight of the objective [1.0]
     */
    void addObjective(TString name, URANIE::DataServer::TDataServer *tds,
            TString ystar, TString sigma,
            URANIE::Launcher::TOutputFile *outfile, TString yhat,
            Double_t alpha = 1.0);
    //! Unactive the objective given by these name
    void unactiveObjective(TString name);
    //! Active the objective given by these name
    void activeObjective(TString name);
    //! Invert the state of all the objectives
    void invertObjectives();
    //! get the vector containing the objectives
    vector< URANIE::Optimizer::TObjective * > getObjectives()
    {
        return _objectives;
    }
    //! Add input and output variables
    /*!
      Add output attributes defined from other attributes

      \param soutput (TString): list of output variable to be created 
     */
    void addOutputVariable(TString soutput);
    //! Select an cost in the Output string
    void selectCost(TString scost);
    //! Get the selected Select an cost in the Output string
    TString getSelectedCost()
    {
        return _sCost;
    }
    //@}

    //---------------------------------------------
    /** @name Manipulate the parameters
     At any time, we can change the state of the parameters; it can be fix to the default value or it can be variable in these range.
     **/
    //@{
    //! Fix the parameter given by these name to the default value
    void fixParameter(TString name);
    //! Unfix the parameter given by these name.
    /*!
     These parameter became a \em "variable" attribute
     */
    void unfixParameter(TString name);
    //! Invert the state of all the parameters
    /*!
     Then, fixed parameters became \em "variable" attribute and \em vice \em versa.
     */
    void invertParameters();
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    //! Print the state of the algorithm of optimization
    void printState(Option_t *option = "");
    void setLog()
    {
        _blog = kTRUE;
    }
    void unsetLog()
    {
        _blog = kFALSE;
    }
    void changeLog()
    {
        _blog = _blog ? kFALSE : kTRUE;
    }
    Bool_t getLog()
    {
        return _blog;
    }
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Optimizer::TOptimizer, ID_OPTIMIZER)
    //Classe de
};

} // Fin du namespace Optimizer
} // Fin du namespace URANIE
#endif
