//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TVizir.cxx
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2009-06-02.11.27.49
 
 \brief interface uranie pour vizir
 
 */

#include <sstream>
#include <string>

// ROOT
#include "TRegexp.h"
//fga 2015/01/12 #include "Api.h"
#include "TPad.h"
#include "TH1F.h"

// URANIE
#include "TVizir.h"
#include "TVizir_function.h"
#include "TInputFile.h"

#include "vizir.h"



extern vzrfitnessbuild_t *vzrcint_class(URANIE::Optimizer::TVizir *);

ClassImp (URANIE::Optimizer::TVizir)

namespace URANIE
{
namespace Optimizer
{

// Constructeur

//       TVizir::TVizir() : 
//     TNamed("Vizir", "The Minimization package with Evolutionnary Algorithm"),
//     _tds(NULL),
//     precision(0.001),
//     vizir(NULL)
//       {
//     population = 200;
//     duree = 100000;
//     survie = zygotie = mutation = -1.0;
//     contrainte = cost = -1;
//     fDecode = fSoft = fHard = (TMethodCall *) NULL;
//       }

TVizir::TVizir(URANIE::DataServer::TDataServer *tds, int pop) :
        TNamed("Vizir", "The Minimization package with Evolutionnary Algorithm"),
	population(pop), _blog(kFALSE), _tds(tds), _sdirectory(gSystem->pwd()), _sdraw(0),
	_sopt(0), _scut(0), _nNbGenerationToDraw(2)
{
    try{
        if(_tds == NULL)
        {   
            throw URANIE::Exceptions::UErrorExceptions(__FILE__,
                                __LINE__,
                                Form("TVizir::constructor ERROR : the TDS is null!"));
        }
        if(population <= 0)
             throw URANIE::Exceptions::UErrorExceptions(__FILE__,
                                __LINE__,
                                Form("TVizir::constructor ERROR : the size of the population specified [%d] is not acceptable!",population));
    }
    catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            throw ue;
            exit(-1);
        }
    
    
    precision = 0.001;
    duree = 100000;
    survie = zygotie = mutation = -1.0;
    contrainte = cost = hide = -1;
    fDecode = fSoft = fHard = (TMethodCall *) NULL;
    _code = (URANIE::Launcher::TCode *) NULL;
    vizir = (vzr_t *) NULL;
    intern = param = _tds->getNAttributes();
}

TVizir::~TVizir()
{
    if (fDecode)
        delete fDecode;
    fDecode = NULL;
    if (fHard)
        delete fHard;
    fHard = NULL;
    if (fSoft)
        delete fSoft;
    fSoft = NULL;
    if (vizir)
        vizir1_vzr_dealloc(vizir);
}

// void TVizir::init(URANIE::DataServer::TDataServer *tds, int pop)
// {
// }

void TVizir::setDecode(const char * funcname, Int_t phys)
{
    try{
//fga 2015/01/12        if (!fun)
//fga 2015/01/12            throw URANIE::Exceptions::UErrorExceptions(__FILE__,
//fga 2015/01/12                                __LINE__,
//fga 2015/01/12                                Form("TVizir::setDecode ERROR : the decode function is null!"));
//fga 2015/01/12        char *funcname = G__p2f2funcname(fun);
        if (funcname)
        {
            // creation TMethodCall
            fDecode = new TMethodCall();
            fDecode->InitWithPrototype(funcname,
                    "Int_t*,Int_t*,Double_t*,Double_t*");
            if (phys)
                intern = phys;
        }
        else
            throw URANIE::Exceptions::UErrorExceptions(__FILE__,
                                __LINE__,
                                Form("TVizir::setDecode ERROR : the decode function is unknown!"));
        if (! fDecode->IsValid() )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__,
                                __LINE__,
                                Form("TVizir::setDecode ERROR : the decode function is not valid!"));
    }
    catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            if(fDecode!=NULL)
            {
                delete fDecode;
                fDecode=NULL;
            }
            throw ue;
            exit(-1);
        }
}

void TVizir::setHardConstraint(const char *funcname)
{
    try{
//fga 2015/01/12        if (!fun)
//fga 2015/01/12            throw URANIE::Exceptions::UErrorExceptions(__FILE__,
//fga 2015/01/12                                __LINE__,
 //fga 2015/01/12                               Form("TVizir::setHardConstraint ERROR : the hardConstraint function is null!"));
 //fga 2015/01/12       char *funcname = G__p2f2funcname(fun);
        if (funcname)
        {
            // creation TMethodCall
            fHard = new TMethodCall();
            fHard->InitWithPrototype(funcname, "Double_t*,Int_t*");
        }
        else
            throw URANIE::Exceptions::UErrorExceptions(__FILE__,
                                __LINE__,
                                Form("TVizir::setHardConstraint ERROR : the hardConstraint function is unknown!"));
    }
    catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            throw ue;
            exit(-1);
        }
}

void TVizir::rearangeInput(TList *listOfInputAttributes, TString &labels)
{
    Int_t nvars = listOfInputAttributes->GetEntries();
    minmax.ResizeTo(nvars, 2);
    for (int i = 0; i < nvars; i++)
    {
        URANIE::DataServer::TAttribute *att =
                (URANIE::DataServer::TAttribute *) listOfInputAttributes->At(i);
        minmax(i, 0) = att->getLowerBound();
        minmax(i, 1) = att->getUpperBound();
        labels += TString(":") + att->GetName();
    }
    if (_blog)
        minmax.Print("*");

}

void TVizir::createTuple(TString &labels)
{
    // rang(s) eventuel(s)
    if (cost == 1)
    {
        if (contrainte != 0)
        {
            //ajouter rang contrainte et domi
            _tds->addAttribute("rgadmin");
            labels += TString(":rgadmin");
        }
    }
    _tds->addAttribute("rgpareto") ;
    labels += TString(":rgpareto");
    // generation
    _tds->addAttribute("generation") ;
    labels += TString(":generation");

    // Create the tuple from the TDataServer object
    _tds->createTuple();
}

void TVizir::setConstraintAndCost(const char *funcname, const char* label, int nbconst,
        int hiden, const char* input, Bool_t bAddTheAttributes)
{

//fga 2015/01/12    if (!fun)
//fga 2015/01/12        return;
//fga 2015/01/12    char *funcname = G__p2f2funcname(fun);
    if (!funcname)
        return;

    TString sListOfCostAttributes(label);
    int nbcout = sListOfCostAttributes.CountChar(':') + 1;

    // creation TMethodCall
    fSoft = new TMethodCall();
    fSoft->InitWithPrototype(funcname, "Double_t*,Double_t*");

    // Add all the attributes in the TDS
    if (bAddTheAttributes)
    {
        cost = nbcout - nbconst - hiden;
        contrainte = nbconst;
        hide = hiden;
        // preparation NTuple
        TString labels("tdsvzr__n");

        // entrées
        TList *listOfInputAttributes = _tds->getListOfAttributes(input);
        rearangeInput(listOfInputAttributes, labels);
        /** mise a jour data server **/
        // sorties
        labels += TString(":") + label;
        sListOfCostAttributes += TString(" ");
        sListOfCostAttributes.ReplaceAll(":", " ");
        std::istringstream inStream((char*) sListOfCostAttributes.Data());
        std::string inName;
        while (inStream >> inName)
        {
            TString svar(inName.c_str());
            _tds->addAttribute(svar) ;
        }
        // extra et creation tuple
        createTuple(labels);
	delete listOfInputAttributes;
	listOfInputAttributes=NULL;
    }

}

void TVizir::setCost(const char *funcname, const char* label, int hiden, const char *input)
{
    setConstraintAndCost(funcname, label, 0, hiden, input);
}

void TVizir::setConstraintAndCost(URANIE::Launcher::TCode *cod, int nbconst,
        int hiden)
{
    if (!cod)
        return;

    // mémorisation TCode
    _code = cod;

    _code->initFileType();
    //     setLog();
    if (0)
     {
    //     try
    //     {
    //         // Boucle sur les TAttribut du TDataServer
    //         Int_t natt = _tds->getNAttributes();
    //         URANIE::DataServer::TAttribute *att = 0;
    //         for (Int_t iatt = 0; iatt < natt; iatt++)
    //         {
    //             att = _tds->getAttribute(iatt);
    //             if (_blog)
    //                 att->printLog();

    //             TString sfile = att->getFileNameOfKey();
    //             if (sfile.Length() > 0)
    //             {
    //                 if (_blog)
    //                     cout << "sfile[" << sfile << "]" << endl;
    //                 URANIE::Launcher::TInputFile *tif = _code->getInputFile(
    //                         sfile);
    //                 if (tif != NULL)
    //                 {
    //                     if (_blog)
    //                         cout << "FILENAME[" << tif->getFileName() << "]"
    //                                 << endl;
    //                     tif->addAttribute(att);
    //                 }
    //                 else
    //                 {
    //                     throw URANIE::Exceptions::UExceptions(__FILE__,
    //                             __LINE__,
    //                             Form(
    //                                     "TLauncher::constructor ERROR : for the attribute \"%s\", the file \"%s\" is not found",
    //                                     att->GetName(), sfile.Data()));
    //                 }
    //             }
    //         }
    //     } catch (URANIE::Exceptions::UExceptions& ue)
    //     {
    //         if (ue.getMessageType() == URANIE::Exceptions::UExceptions::kDEBUG)
    //             ue.printMessage();
    //         if (ue.getMessageType() > URANIE::Exceptions::UExceptions::kWARNING)
    //         {
    //             ue.printMessage();
    //             exit(-1);
    //         }
    //     }
    }
    //unsetLog();
    //     _code->printLog();

    // preparation NTuple
    TString labels("tdsvzr__n");
    // entrées
    TList *listOfInput = new TList();
    //     cout << " ** TVizir::setConstraintAndCost NInputfiles[" << _code->getNInputFiles() << "]" << endl;
    for (Int_t ifile = 0; ifile < _code->getNInputFiles(); ifile++)
    {
        //       cout << "   ** i[" << ifile << " / " << _code->getNInputFiles() << "] -- natt[" << _code->getNInputAttributes(ifile) <<"]" << endl;
        for (Int_t ivar = 0; ivar < _code->getNInputAttributes(ifile); ivar++)
        {
            //      cout << "   ** i[" << ifile << " / " << _code->getNInputFiles() << "] --  iatt[" << "/" << _code->getNInputAttributes(ifile) <<"]" << endl;
            listOfInput->Add((TObject*) _code->getInputAttribute(ifile, ivar));
        }
    }
    std::cout << std::endl;
    rearangeInput(listOfInput, labels);
    // sorties
    Int_t nbcout = 0;
    Int_t noutfile = _code->getNOutputFiles();

    for (Int_t ioutfile = 0; ioutfile < noutfile; ioutfile++)
    {
        Int_t nattfile = _code->getNOutputAttributes(ioutfile);
        for (Int_t iattfile = 0; iattfile < nattfile; iattfile++)
        {
            URANIE::DataServer::TAttribute * att = _code->getAttribute(ioutfile,
                    iattfile);
            _tds->addAttribute(att);
            labels += TString(":") + att->GetName();
            nbcout++;
        }
    }
    cost = nbcout - nbconst - hiden;
    contrainte = nbconst;
    hide = hiden;
    // extra et creation tuple
    createTuple(labels);
    delete listOfInput;
    listOfInput=NULL;
}

void TVizir::setCost(URANIE::Launcher::TCode *cod, int hiden)
{
    setConstraintAndCost(cod, 0, hiden);
}

void TVizir::decode(Double_t *coded, Double_t *phys, TMethodCall *dec)
{
    if (dec)
    {
        // decodage specifique
        void *args[4];
        Double_t result;

        // TODO param phys ??
        args[0] = (void *) &intern;
        args[1] = (void *) &param;
        args[2] = (void *) coded;
        args[3] = (void *) phys;
        //std::cout << "decode " ;
        dec->SetParamPtrs(args, 4);
        //std::cout << "." << std::endl;
        dec->Execute(result);
    }
    else
    {
        // default
        Int_t i;
        Double_t lower, upper;

        for (i = 0; i < param; i++)
        {
            lower = minmax(i, 0);
            upper = minmax(i, 1);
            phys[i] = coded[i] * (upper - lower) + lower;
        }
    }
    //for (int i=0; i<param; i++) std::cout << phys[i] << " "; std::cout << std::endl;
}

URANIE::Launcher::TCode *TVizir::copyCode()
{
    if (!_code)
        return NULL;

    URANIE::Launcher::TCode *ret;

	Double_t* bidon = new Double_t[param + 1];

    bidon[0] = 0;
    for (Int_t i = 0; i < param; i++)
        bidon[i + 1] = minmax[i][0];
    _code->init();
    _code->preTraitment(param, bidon);
    ret = new URANIE::Launcher::TCode(*_code);
    // ret = (URANIE::Launcher::TCode *) _code->Clone();
    // ret->setCommand( Form("cd ../UranieLauncher_%d;", nid) + ret->getCommand() );
    // ret->printLog();

	delete[] bidon;

    return ret;
}

int TVizir::evaluer(Double_t *in, Double_t *out, TMethodCall *decod,
        TMethodCall* hard, TMethodCall *soft)
{

	Double_t* phys = new Double_t[param];

    Int_t admis;
    void *args[2];
    Double_t result;

   
    // decodage
    decode(in, phys, decod);
   
    // contrainte hard
    admis = 1;
    if (hard)
    {
        args[0] = (void *) phys;
        args[1] = (void *) &admis;
        //std::cout << "hardConstraint " ;
        hard->SetParamPtrs(args, 2);
        hard->Execute(result);
        //std::cout << admis << "." << std::endl;
    }
    // contrainte et cost
    if (admis)
    {
        if (soft)
        {
            args[0] = (void *) phys;
            args[1] = (void *) out;
            //std::cout << "cost " ;
            soft->SetParamPtrs(args, 2);
            soft->Execute(result);
            //std::cout << "." << std::endl;
        }
        else
            admis = 0; // throw Error
    }
    

	delete[] phys;

	return admis;
}

int TVizir::evaluer(Double_t *in, Double_t *out, TMethodCall *decod,
        TMethodCall* hard, URANIE::Launcher::TCode *soft)
{

	Double_t* phys = new Double_t[param + 1];

    Int_t admis, nb;
    void *args[2];
    Double_t result;

    // decodage
    phys[0] = 0;
    decode(in, &phys[1], decod);
    
    // contrainte hard
    admis = 1;
    if (hard)
    {
        args[0] = (void *) phys;
        args[1] = (void *) &admis;
        hard->SetParamPtrs(args, 2);
        hard->Execute(result);
    }
    // contrainte et cost
    if (admis)
    {
        if (soft)
        {
            soft->preTraitment(param, phys);
            soft->run();
            // output without index
            nb = 1;
            admis = (soft->postTraitment(nb, out - 1)) ? 1 : 0;
            // std::cout << out[0] << " " << out[1] << std::endl;
        }
        else
            admis = 0; // throw Error
    }
  

	delete[] phys;
   
	return admis;
}

void TVizir::fitnessChanged(void)
{
    if (vizir)
        vizir1_vzr_reevaluate(vizir);
}

// solver
void TVizir::optimize(Option_t *option)
{
    TString soption = TString(option);

    if (_blog)
        cout << " ** TVizir::optimize [" << option << "]" << endl;

    if (vizir)
    {
        if (_blog)
            printLog("TVizir::optimize recall");
        // std::cout << "Search is soon Done" << std::endl;
        // return;
        vizir->maxevalu = duree;
    }
    else
    {
        if (_blog)
            printLog("TVizir::optimize step 1");
        /*** initialisation ***/
        // initialisation selon le type
        if (cost == 1)
        {
            if (contrainte == 0)
            {
/*cout<<endl<<"######################## vizir1_vzr_mono "<<endl;  */                  
                vizir = vizir1_vzr_mono(intern, precision);
            }
            else
            {
/*cout<<endl<<"######################## vizir1_vzr_monoConstraint "<<contrainte<<endl;   */             
                vizir = vizir1_vzr_monoConstraint(intern, precision, contrainte);
            }
        }
        else
        {
            if (contrainte == 0)
            {
// cout<<endl<<"######################## vizir1_vzr_multi "<<endl;                   
                vizir = vizir1_vzr_multi(intern, cost);
            }
            else
            {
// cout<<endl<<"######################## vizir1_vzr_multiConstraint "<<contrainte<<endl;                    
                vizir = vizir1_vzr_multiConstraint(intern, cost, contrainte);
            }
        }
        if (hide)
            vizir->nbhide = hide;
        // parametre
        if (survie > 0.0)
            vizir1_vzr_setSurvivalRate(vizir, survie);
        if (mutation >= 0.0)
            vizir1_vzr_setMutationRate(vizir, mutation);
        if (zygotie >= 0.0)
            vizir1_vzr_setHomozygoteRate(vizir, zygotie);
        // fonction
        if (_code)
            vizir1_vzr_setFitness(vizir, vzrprog_class(this));
        else
            vizir1_vzr_setFitness(vizir, vzrcint_class(this));
        if (soption.Contains("saveall"))
            vizir->generation = vzrsaveall(this);
        vizir1_vzr_build(vizir, duree, population);
    }

    /*** execution ***/
    // for (int i=0; i<1; i++) vizir1_vzr_start(vizir);
    vizir1_vzr_run(vizir);

    // Fill the results in the TDataServer
    Results();

    // Return in the current directory
    gSystem->cd(_sdirectory);

}

// resultats
void TVizir::Results()
{
    if (_blog)
        cout << " ** TVizir::Results" << endl;

    if (!vizir)
        return;
    else
    {
        Int_t nX = _tds->getNAttributes();
        Int_t nG = vizir1_vzr_generationCount(vizir);
        if (_blog)
            cout << " ** _nx[" << nX << "]" << endl;

        Double_t * dval = new Double_t[nX + 1];
        for (Int_t ipat = 0; ipat < population; ipat++)
        {

            // Iterator
            dval[0] = 1.0 * ipat;

            // Input attributes
            Double_t *gene = vizir1_vzr_phenotype(vizir, ipat);

			Double_t* par = new Double_t[param];

            decode(gene, par, fDecode);
            for (Int_t jvar = 0; jvar < param; jvar++)
            {
                dval[1 + jvar] = par[jvar];
            }

            // Constraints and cost
            Double_t *phys = vizir1_vzr_objective(vizir, ipat);
            
            for (Int_t jvar = 0; jvar < vizir->nbcrit; jvar++)
            {
                dval[1 + param + jvar] = phys[jvar];
            }
            // generation
            dval[1 + param + vizir->nbcrit] = 1.0 * nG;

            // Fill all the data
            _tds->getTuple()->Fill(dval);

			delete[] par;

        }
        _tds->getNPatterns();

        // Draw the graph with the selection
        if ((_sdraw.Length() > 0) && (_nNbGenerationToDraw > 0))
        {
            Int_t ncolor = 0;
            Int_t vecColor[5] =
            { kRed, kOrange, kYellow, kGreen, kBlack };

            TString sopt = "";
            TString svarcut = "";
            Int_t nloopinit = 1;
            Int_t nloopend = _nNbGenerationToDraw;
            if (nG < nloopend + 1)
            {
                nloopinit = nloopend = nG;
                ncolor = nG - 1;
                if (nG > 1)
                    sopt = TString("same");
            }

            for (Int_t i = nloopinit; i <= nloopend; i++)
            {
                Int_t nIndCrt = nG - nloopend + i;

//           cout << " ** nG[" << nG<<"] i[" << i << "/" << nloopend << "] :: nIndCrt[" << nIndCrt << "] color[" << vecColor[ncolor] << "]" << endl;

                svarcut = Form("generation==%d", nIndCrt);
                _tds->getTuple()->SetMarkerColor(vecColor[ncolor++]);
                _tds->getTuple()->Draw(_sdraw.Data(), svarcut.Data(),
                        sopt.Data());

                sopt = TString("same");
            }

            TH1F *htemp = (TH1F*) gPad->GetPrimitive("htemp");
            htemp->SetTitle(Form("%s {%s}", _sdraw.Data(), svarcut.Data()));

            gPad->Modified();
            gPad->Update();
//          gPad->SaveAs(Form("_vizir_%s_optim_generation%03d.png", _tds->GetName(), nG));
//         gPad->SaveAs(Form("_vizir_%s_optim_generation%03d.gif", _tds->GetName(), nG));

            // Reset the color of the TDS
            _tds->getTuple()->SetMarkerColor(kRed);
        }

        
        delete[] dval;
    }

}

////////////////////////////////////////////////////////////
//  setNbGenerationToDraw
////////////////////////////////////////////////////////////
void TVizir::setNbGenerationToDraw(Int_t n)
{
    try
    {
        if (n < 0)
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                    Form("TVizir::setNbGenerationToDraw method\nThe number of generation to draw must be positive\nYour request is %d and then we set to  0",
                         n));
        if (n > MAXNBGENERATIONS)
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                    Form("TVizir::setNbGenerationToDraw method\nThe number of generation to draw is limited to %d\nYour request is %d and then we set to %d",
                         MAXNBGENERATIONS, n, MAXNBGENERATIONS));
        _nNbGenerationToDraw = n;
    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        if (n < 0)
            _nNbGenerationToDraw = 0;
        if (n > MAXNBGENERATIONS)
            _nNbGenerationToDraw = MAXNBGENERATIONS;
        throw ue;
    }
}

////////////////////////////////////////////////////////////
//  setVarDraw
////////////////////////////////////////////////////////////
void TVizir::setVarDraw(const char* varexp, const char* selection,
        Option_t* option)
{
    //(TString str, TString select , TString opt)
    _sdraw = TString(varexp);       // str;
    _scut = TString(selection);    // select;
    _sopt = TString(option);       // opt;

    // Treat the options
    TString soption = _sopt;
    soption.ToLower();
    soption.ReplaceAll(" ", "");
    if (_blog)
        cout << " ** soption[" << soption << "]" << endl;

    TRegexp reg("generation=[0-9]*");
    if (soption.Contains(reg))
    {
        TString out = soption(reg);
        if (_blog)
            cout << " ** out[" << out << "]" << endl;
        TString snbgeneration = out.Remove(0, 11);
        if (_blog)
            cout << " ** snbgeneration[" << snbgeneration << "]" << endl;

        Int_t nbgeneration = snbgeneration.Atoi();
        if (_blog)
            cout << " ** nbgeneration[" << nbgeneration << "]" << endl;
        try
        {
            if (nbgeneration < 0)
                throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                        Form("TVizir::setVarDraw method\nThe number of last generation is non-positive \nYour request is %d with \"%s\" and then we set to 0",
                             nbgeneration, _sopt.Data()));
        } catch (URANIE::Exceptions::UWarningExceptions& ue)
        {
            ue.printMessage();
            _nNbGenerationToDraw = 0;
            throw ue;
        }

        try
        {
            if (nbgeneration > MAXNBGENERATIONS)
                throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                        Form("TVizir::setVarDraw method\nThe number of last generation is limited to %d\nYour request is %d with \"%s\" and then we set to %d",
                             MAXNBGENERATIONS, nbgeneration, _sopt.Data(), MAXNBGENERATIONS));
            _nNbGenerationToDraw = nbgeneration;
        } catch (URANIE::Exceptions::UWarningExceptions& ue)
        {
            ue.printMessage();
            _nNbGenerationToDraw = MAXNBGENERATIONS;
            throw ue;
        }
    }
    else
    {
        try
        {
           throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                        Form("TVizir::setVarDraw method\nThe option [%s] is incorrect",
                              _sopt.Data()));
        } catch (URANIE::Exceptions::UWarningExceptions& ue)
        {
            ue.printMessage();
            _nNbGenerationToDraw = MAXNBGENERATIONS;
            throw ue;
        }
    }
}

////////////////////////////////////////////////////////////
//  printLog
////////////////////////////////////////////////////////////
void TVizir::printLog(Option_t *option)
{
    cout << " ** TVizir::printLog[" << option << "]" << endl;
    cout << " ** Name[" << GetName() << "] Title[" << GetTitle() << "]" << endl;
    cout << " ** population[" << population << "] duree[" << duree << "]"
            << endl;
    cout << " ** intern[" << intern << "] param[" << param << "] contrainte["
            << contrainte << "] cost[" << cost << "]" << endl;
    cout << " ** survie[" << survie << "] zygotie[" << zygotie << "] mutation["
            << mutation << "] precision[" << precision << "]" << endl;
    cout << " ** _sdirectory[" << _sdirectory << "]" << endl;
    cout << " ** _sdraw[" << _sdraw << "] _scut[" << _scut << "] _sopt["
            << _sopt << "] _nNbGenerationToDraw[" << _nNbGenerationToDraw << "]"
            << endl;
    cout << " ** End Of TVizir::printLog[" << option << "]" << endl;

}

} // Fin du namespace Optimizer
} // Fin du namespace URANIE

