//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTVizir.cxx
 \author Maud Salvoldi (OCC/Areva group)
 \date   $Date$
 
 \brief  Fichier de test de la classe URANIE::Optimizer::TVizir
 
 Cree le : jeu dec 22 11:27:34 CET 2011
 
 */

#define DBL_PRECISION_COMPARE 1.e-12
#define EPSILON 0.01 

// Uranie
#include "TestTVizir.h"
#include "UDisableTest.h"
#include "TOutputFileKey.h"
#include "code_paths.h"
#include "uranie_commands.h"

// ROOT
#include "TString.h"
#include "TROOT.h"
#include "TSystem.h" 
//fga 2015/01/12 #include "Api.h"

#include <math.h>
#include <iostream>
#include <iomanip>

using namespace std;

TVizirTest::TVizirTest() :
        _tds(NULL), _tds1(NULL), _tds2(NULL), _vzr(NULL), _foutDataServer(NULL), _tcode(
                NULL)
{
    _workingTestPath = gSystem->pwd();
    gROOT->LoadMacro("TestUserFunctions.C");
}

TVizirTest::~TVizirTest()
{

}
void TVizirTest::setUp()
{
    gSystem->cd(_workingTestPath);

    // Backup of the file input_rosenbrock_with_keys.dat
    gSystem->CopyFile("input_rosenbrock_with_keys.dat","input_rosenbrock_with_keys.dat_bkp",kTRUE);
    
    _vzr = NULL;

    _tds = new URANIE::DataServer::TDataServer("testTVizir",
            "test of classes TVizir");
    _tds->addAttribute("x", 0.5, 2.0);
    _tds->addAttribute("y", 0.5, 1.0);

    _tds1 = new URANIE::DataServer::TDataServer("tdsOATMinMaxFlowrate",
            "OATMinMax Design of experiments for Flowrate");
    _tds1->addAttribute("rw", 0.05, 0.15);
    _tds1->addAttribute("r", 100.0, 50000.0);
    _tds1->addAttribute("tu", 63070.0, 115600.0);
    _tds1->addAttribute("tl", 63.1, 116.0);
    _tds1->addAttribute("hu", 990.0, 1110.0);
    _tds1->addAttribute("hl", 700.0, 820.0);
    _tds1->addAttribute("l", 1120.0, 1680.0);
    _tds1->addAttribute("kw", 9855.0, 12045.0);

    TString sJDDReference = _workingTestPath;
#ifdef WIN32
    sJDDReference += "\\";
#else
    sJDDReference += "/";
#endif  
    sJDDReference += "flowrate_input_with_values_rows.in";
    _tds1->getAttribute("rw")->setFileKey(sJDDReference, "Rw", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds1->getAttribute("r")->setFileKey(sJDDReference, "R", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds1->getAttribute("tu")->setFileKey(sJDDReference, "Tu", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds1->getAttribute("tl")->setFileKey(sJDDReference, "Tl", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds1->getAttribute("hu")->setFileKey(sJDDReference, "Hu", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds1->getAttribute("hl")->setFileKey(sJDDReference, "Hl", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds1->getAttribute("l")->setFileKey(sJDDReference, "L", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds1->getAttribute("kw")->setFileKey(sJDDReference, "Kw", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);

    _foutDataServer = new URANIE::Launcher::TOutputFileDataServer(
            "_output_flowrate_withRow_.dat");
    _foutDataServer->addAttribute("yhat");
    _foutDataServer->addAttribute("d");

#ifdef WIN32
    _tcode = new URANIE::Launcher::TCode(_tds1, "flowrate.exe -r 1>NUL");
#else
    _tcode = new URANIE::Launcher::TCode(_tds1, "flowrate -r >> /dev/null");
#endif
    _tcode->addOutputFile(_foutDataServer);

    _tds2 = new URANIE::DataServer::TDataServer("tdsvzr", "tds for vizir test");
    _tds2->addAttribute("fond", 0., 256.);
    for (Int_t i = 1; i < 8; i++)
        _tds2->addAttribute(Form("zone%d", i), 0., 256.);
    
    TString sJDDRosenbrockReference = gSystem->UnixPathName(Form( "%s/input_rosenbrock_with_keys.dat",_workingTestPath.Data()));
    
    _tdsRosenbrock = new URANIE::DataServer::TDataServer("tdsRosenbrock", "Optimize Code externe Rosenbrock via TDataServer");
    _tdsRosenbrock->addAttribute("x", -1.5, 1.5);
    _tdsRosenbrock->addAttribute("y", -1.5, 1.5);
    
    _tdsRosenbrock->getAttribute("x")->setFileKey(sJDDRosenbrockReference);
    _tdsRosenbrock->getAttribute("y")->setFileKey(sJDDRosenbrockReference);
    
    _fOutputFile = new URANIE::Launcher::TOutputFileKey("_output_rosenbrock_with_keys_.dat");
    _fOutputFile->addAttribute("fval");
    
}

void TVizirTest::tearDown()
{
  
    if (_vzr != NULL)
    {
        delete _vzr;
        _vzr = NULL;
    }
    
    if (_tcode != NULL)
    {
      //        _tcode->setLog();
        delete _tcode;
        _tcode = NULL;
    }
       
    if (_foutDataServer != NULL)
    {
      //        _foutDataServer->setLog();
        delete _foutDataServer;
        _foutDataServer = NULL;
    }
   
    if (_fOutputFile != NULL)
    {
      //        _fOutputFile->setLog();
        delete _fOutputFile;
        _fOutputFile = NULL;
    }
   
    if (_tds1 != NULL)
    {
      //        _tds1->setLog();
        delete _tds1;
        _tds1 = NULL;
    }
    
    if (_tds2 != NULL)
    {
      //        _tds2->setLog();
        delete _tds2;
        _tds2 = NULL;
    }
    if (_tds != NULL)
    {
      //        _tds->setLog();
        delete _tds;
        _tds = NULL;
    }

    if (_tdsRosenbrock != NULL)
    {
      //        _tdsRosenbrock->setLog();
        delete _tdsRosenbrock;
        _tdsRosenbrock = NULL;
    }
    
    gSystem->CopyFile("input_rosenbrock_with_keys.dat_bkp","input_rosenbrock_with_keys.dat",kTRUE);
    UranieTests::removeFile(gSystem->pwd(),"input_rosenbrock_with_keys.dat_bkp");
}



// void funAllCost(Double_t *param, Double_t *res)
// {
//     Double_t x = param[0];
//     Double_t y = param[1];
//     res[0] = (x - 1.0) * (x - 1.0) + (y - 1.0) * (y - 1.0) + 1.0;
//     res[1] = 1. / (x * x + y * y + 1.0);
// }
// 
// 
// 
// void contHard1(double * paramt, Int_t *res)
// {
//     *res = 0;
//     if (paramt[0] <= paramt[1] + 1.0)
//         *res = 1;
// }
// 
// 
// 
// void mydecode(int *rien, int *size, double *intern, double *user)
// {
//     int i;
// 
//     user[0] = 256.0;
//     for (i = 1; i < *size; i++)
//         user[i] = 256.0 * intern[i - 1];
// }
// 
// 
// 
// double contraste(int size, double *sol, const char **connexion)
// {
//     int i, j;
//     double min, diff;
// 
//     min = 256.0;
//     for (i = 0; i < size; i++)
//     {
//         for (j = i + 1; j < size; j++)
//         {
//             if (connexion[i][j])
//             {
//                 diff = fabs(sol[i] - sol[j]);
//                 if (diff < min)
//                     min = diff;
//             }
//         }
//     }
//     return min;
// }
// void myfitness(double *in, double *out)
// {
//     static const char tgr[8 * 8] =
//     {
//     /*  0  1  2  3  4  5  6  7 */
//     0, 1, 1, 1, 0, 0, 1, 1, // 0
//             1, 0, 1, 1, 1, 0, 0, 0, // 1
//             1, 1, 0, 0, 1, 0, 0, 0, // 2
//             1, 1, 0, 0, 0, 1, 1, 0, // 3
//             0, 1, 1, 0, 0, 1, 0, 1, // 4
//             0, 0, 0, 1, 1, 0, 1, 0, // 5
//             1, 0, 0, 1, 0, 1, 0, 1, // 6
//             1, 0, 0, 0, 1, 0, 1, 0  // 7
//             };
//     static const char *tangram[8] =
//     { &tgr[0], &tgr[8], &tgr[2 * 8], &tgr[3 * 8], &tgr[4 * 8], &tgr[5 * 8],
//             &tgr[6 * 8], &tgr[7 * 8] };
// 
//     out[0] = -contraste(8, in, tangram);
// }
////////////////////////////////
///                          ///                            
///           test           ///
///                          ///
////////////////////////////////
void TVizirTest::testTVizirConstructor()
{
    _vzr = new URANIE::Optimizer::TVizir(_tds, 10000);
    _vzr->unsetLog();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(10000, _vzr->population,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.001, _vzr->precision, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-1.0, _vzr->survie, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-1.0, _vzr->zygotie, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-1.0, _vzr->mutation, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_EQUAL(100000, _vzr->duree);
    CPPUNIT_ASSERT_EQUAL(-1, _vzr->contrainte);
    CPPUNIT_ASSERT_EQUAL(-1, _vzr->cost);
    CPPUNIT_ASSERT_EQUAL(-1, _vzr->hide);

}
void TVizirTest::testTVizirsetPrecision()
{
    _vzr = new URANIE::Optimizer::TVizir(_tds, 10000);
    _vzr->unsetLog();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.001, _vzr->precision, DBL_PRECISION_COMPARE);
    _vzr->setPrecision(0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.1, _vzr->precision, DBL_PRECISION_COMPARE);
}
void TVizirTest::testTVizirsetSurvivalRate()
{
    _vzr = new URANIE::Optimizer::TVizir(_tds, 10000);
    _vzr->unsetLog();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(-1.0, _vzr->survie, DBL_PRECISION_COMPARE);
    _vzr->setSurvivalRate(1.2);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.2, _vzr->survie, DBL_PRECISION_COMPARE);
}
void TVizirTest::testTVizirsetHomozygoteRate()
{
    _vzr = new URANIE::Optimizer::TVizir(_tds, 10000);
    _vzr->unsetLog();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(-1.0, _vzr->zygotie, DBL_PRECISION_COMPARE);
    _vzr->setHomozygoteRate(3.4);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.4, _vzr->zygotie, DBL_PRECISION_COMPARE);
}
void TVizirTest::testTVizirsetMutationRate()
{
    _vzr = new URANIE::Optimizer::TVizir(_tds, 10000);
    _vzr->unsetLog();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(-1.0, _vzr->mutation, DBL_PRECISION_COMPARE);
    _vzr->setMutationRate(2.5);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.5, _vzr->mutation, DBL_PRECISION_COMPARE);
}
void TVizirTest::testTVizirsetMaxEvaluation()
{
    _vzr = new URANIE::Optimizer::TVizir(_tds, 10000);
    _vzr->unsetLog();

    CPPUNIT_ASSERT_EQUAL(100000, _vzr->duree);
    _vzr->setMaxEvaluation(5000);
    CPPUNIT_ASSERT_EQUAL(5000, _vzr->duree);
}
void TVizirTest::testTVizirGetLastGenerationIndex()
{
    
//fga 2015/01/12     // Get the void * from the "funAllCost" analytical function
//fga 2015/01/12     G__ClassInfo globalscope;
//fga 2015/01/12     G__MethodInfo method;
//fga 2015/01/12     long dummy;
//fga 2015/01/12     // resolve function overloading
//fga 2015/01/12     method = globalscope.GetMethod("funAllCost", "double*,double*", &dummy);
//fga 2015/01/12     void * p2ffunAllCost = method.PointerToFunc();
    
//fga 2015/01/12    // Get the void * from the "contHard1" analytical function
//fga 2015/01/12    G__ClassInfo globalscope2;
//fga 2015/01/12    G__MethodInfo method2;
//fga 2015/01/12     long dummy2;
//fga 2015/01/12     // resolve function overloading
//fga 2015/01/12     method2 = globalscope2.GetMethod("contHard1", "double*,int*", &dummy2);
//fga 2015/01/12     void * p2fcontHard1 = method2.PointerToFunc();

    _vzr = new URANIE::Optimizer::TVizir(_tds, 1000);
    _vzr->unsetLog();
    _vzr->setCost("funAllCost", "c1:c2");
    _vzr->setHardConstraint("contHard1");
    _vzr->setSurvivalRate(0.85);
    _vzr->setMaxEvaluation(1400);
    _vzr->optimize("");
   
    CPPUNIT_ASSERT_EQUAL(2, _vzr->getLastGenerationIndex());
}
void TVizirTest::testTVizirSetDecode()
{

//fga 2015/01/12     // Get the void * from the "mydecodeImage" analytical function
//fga 2015/01/12     G__ClassInfo globalscope;
//fga 2015/01/12     G__MethodInfo method;
//fga 2015/01/12     long dummy;
//fga 2015/01/12     // resolve function overloading
//fga 2015/01/12     method = globalscope.GetMethod("mydecodeImage", "in*,int*,double*,double*", &dummy);
//fga 2015/01/12     void * p2fmydecode = method.PointerToFunc();

    _vzr = new URANIE::Optimizer::TVizir(_tds2, 400);
    _vzr->unsetLog();
    _vzr->setDecode("mydecodeImage", 7);
    
    CPPUNIT_ASSERT_EQUAL(7,_vzr->intern);
    CPPUNIT_ASSERT(_vzr->fDecode->IsValid());
    
}

void TVizirTest::testTVizirSetHardConstraint()
{
    
//fga 2015/01/12     // Get the void * from the "contHard1" analytical function
//fga 2015/01/12     G__ClassInfo globalscope;
//fga 2015/01/12     G__MethodInfo method;
//fga 2015/01/12     long dummy;
//fga 2015/01/12     // resolve function overloading
//fga 2015/01/12     method = globalscope.GetMethod("contHard1", "double*,int*", &dummy);
//fga 2015/01/12     void * p2fcontHard1 = method.PointerToFunc();


    _vzr = new URANIE::Optimizer::TVizir(_tds, 1000);
    _vzr->unsetLog();
    _vzr->setHardConstraint("contHard1");
    
    CPPUNIT_ASSERT(_vzr->fHard->IsValid());

}
void TVizirTest::testTVizirSetCost()
{

    _vzr = new URANIE::Optimizer::TVizir(_tds1, 400);
    _vzr->unsetLog();
    _vzr->setCost(_tcode, 0);
   
    TString tmpstr = _vzr->_code->getCommand();
    CPPUNIT_ASSERT(tmpstr.Contains("flowrate"));
    CPPUNIT_ASSERT_EQUAL(2,_vzr->cost);
    CPPUNIT_ASSERT_EQUAL(0,_vzr->hide);
    CPPUNIT_ASSERT_EQUAL(0,_vzr->contrainte);
    CPPUNIT_ASSERT(_vzr->_tds->isAttribute("yhat")); 
    CPPUNIT_ASSERT(_vzr->_tds->isAttribute("d"));
    
    
//fga 2015/01/12     // Get the void * from the "funAllCost2" analytical function
//fga 2015/01/12     G__ClassInfo globalscope;
//fga 2015/01/12     G__MethodInfo method;
//fga 2015/01/12     long dummy;
//fga 2015/01/12     // resolve function overloading
//fga 2015/01/12     method = globalscope.GetMethod("funAllCost2", "double*,double*", &dummy);
//fga 2015/01/12     void * p2ffunAllCost2 = method.PointerToFunc();

    delete _vzr;
    _vzr = new URANIE::Optimizer::TVizir(_tds, 10000);
    _vzr->unsetLog();
    _vzr->setCost("funAllCost2", "c1:c2");
    
    
    CPPUNIT_ASSERT(_vzr->fSoft->IsValid());
    CPPUNIT_ASSERT_EQUAL(2,_vzr->cost);
    CPPUNIT_ASSERT_EQUAL(0,_vzr->hide);
    CPPUNIT_ASSERT_EQUAL(0,_vzr->contrainte);
    CPPUNIT_ASSERT(_vzr->_tds->isAttribute("c1")); 
    CPPUNIT_ASSERT(_vzr->_tds->isAttribute("c2"));

}
void TVizirTest::testTVizirSetConstraintAndCost()
{

    _vzr = new URANIE::Optimizer::TVizir(_tds1, 400);
    _vzr->unsetLog();
    _vzr->setConstraintAndCost(_tcode, 0, 0);
    
    TString tmpstr = _vzr->_code->getCommand();
    CPPUNIT_ASSERT(tmpstr.Contains("flowrate"));
    CPPUNIT_ASSERT_EQUAL(2,_vzr->cost);
    CPPUNIT_ASSERT_EQUAL(0,_vzr->hide);
    CPPUNIT_ASSERT_EQUAL(0,_vzr->contrainte);
    CPPUNIT_ASSERT(_vzr->_tds->isAttribute("yhat")); 
    CPPUNIT_ASSERT(_vzr->_tds->isAttribute("d"));
    

//fga 2015/01/12     // Get the void * from the "funCostAndConstraint" analytical function
//fga 2015/01/12     G__ClassInfo globalscope;
//fga 2015/01/12     G__MethodInfo method;
//fga 2015/01/12     long dummy;
//fga 2015/01/12     // resolve function overloading
//fga 2015/01/12     method = globalscope.GetMethod("funCostAndConstraint2", "double*,double*", &dummy);
//fga 2015/01/12     void * p2ffunCostAndConstraint = method.PointerToFunc();

    delete _vzr;
    _vzr = new URANIE::Optimizer::TVizir(_tds, 10000);
    _vzr->unsetLog();
    _vzr->setConstraintAndCost("funCostAndConstraint2","c1:c2:fc1",1);
    
    CPPUNIT_ASSERT(_vzr->fSoft->IsValid());
    CPPUNIT_ASSERT_EQUAL(2,_vzr->cost);
    CPPUNIT_ASSERT_EQUAL(0,_vzr->hide);
    CPPUNIT_ASSERT_EQUAL(1,_vzr->contrainte);
    CPPUNIT_ASSERT(_vzr->_tds->isAttribute("c1")); 
    CPPUNIT_ASSERT(_vzr->_tds->isAttribute("c2"));
    CPPUNIT_ASSERT(_vzr->_tds->isAttribute("fc1"));
}
void TVizirTest::testTVizirSetNbGenerationToDraw()
{
    _vzr = new URANIE::Optimizer::TVizir(_tds, 10000);
    _vzr->unsetLog();
    _vzr->setNbGenerationToDraw(3);
    CPPUNIT_ASSERT_EQUAL(3,_vzr->getNbGenerationToDraw());
}
void TVizirTest::testTVizirSetVarDraw()
{
    _vzr = new URANIE::Optimizer::TVizir(_tds, 10000);
    _vzr->unsetLog();
    _vzr->setVarDraw("y:x", "x>0", "generation=3");

     TString tmpstr = _vzr->_sdraw;
     CPPUNIT_ASSERT(tmpstr.EqualTo("y:x"));
     tmpstr = _vzr->_scut;
     CPPUNIT_ASSERT(tmpstr.EqualTo("x>0"));
     tmpstr = _vzr->_sopt;
     CPPUNIT_ASSERT(tmpstr.EqualTo("generation=3"));
     CPPUNIT_ASSERT_EQUAL(3,_vzr->_nNbGenerationToDraw);
     
}

void TVizirTest::testTVizirSetLog()
{
    _vzr = new URANIE::Optimizer::TVizir(_tds, 10000);
    _vzr->unsetLog();    
    CPPUNIT_ASSERT(!_vzr->getLog());
    
    _vzr->setLog();    
    CPPUNIT_ASSERT(_vzr->getLog());
    
    _vzr->changeLog();    
    CPPUNIT_ASSERT(!_vzr->getLog());
    
    _vzr->changeLog();    
    CPPUNIT_ASSERT(_vzr->getLog());
    
}

void TVizirTest::testTVizirEvaluer()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_INBUILDING_TEST()
    ;

//fga 2015/01/12     // Get the void * from the "mydecode" analytical function
//fga 2015/01/12     G__ClassInfo globalscope;
//fga 2015/01/12     G__MethodInfo method;
//fga 2015/01/12     long dummy;
//fga 2015/01/12     // resolve function overloading
//fga 2015/01/12     method = globalscope.GetMethod("mydecode", "int*,int*,double*,double*", &dummy);
//fga 2015/01/12     void * p2fmydecode = method.PointerToFunc();
    
//fga 2015/01/12     // Get the void * from the "myfitness" analytical function
//fga 2015/01/12     G__ClassInfo globalscope2;
//fga 2015/01/12     G__MethodInfo method2;
//fga 2015/01/12     long dummy2;
//fga 2015/01/12     // resolve function overloading
//fga 2015/01/12     method2 = globalscope2.GetMethod("myfitness", "double*,double*", &dummy2);
//fga 2015/01/12     void * p2fmyfitness = method2.PointerToFunc();

    _vzr = new URANIE::Optimizer::TVizir(_tds2, 400);
    _vzr->unsetLog();
    _vzr->setDecode("mydecode", 7);
    _vzr->setCost("myfitness", "contrast");
    _vzr->setPrecision(0.001);
    
//fga 2015/01/12     char *funcname = G__p2f2funcname(p2fmydecode);
//fga 2015/01/12     char *funcname2 = G__p2f2funcname(p2fmyfitness);
    
    // creation TMethodCall
    TMethodCall*  fDecode = new TMethodCall();
    fDecode->InitWithPrototype("mydecode",
            "Int_t*,Int_t*,Double_t*,Double_t*");
    TMethodCall*  fSoft = new TMethodCall();
    fSoft->InitWithPrototype("myfitness",
            "double *in, double *out");
    
    double* in = new double[8];
    double* out = new double[8];
    
    for (Int_t i=0; i<8; i++)
        in[i]=i*0.1;
    _vzr->evaluer(in,out,fDecode,(TMethodCall*) NULL,fSoft);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(-25.6,out[0],DBL_PRECISION_COMPARE);
     
    delete[] in;
    delete[] out;
    delete fDecode;
    delete fSoft;
    

}
void TVizirTest::testTVizirfitnessChanged()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_INBUILDING_TEST();

    _vzr = new URANIE::Optimizer::TVizir(_tds2, 400);
//     _vzr->setDecode((void *) &mydecode, 7);
//     _vzr->setCost((void*) &myfitness, "contrast");
//     _vzr->setPrecision(0.001);
//     _vzr->optimize();
//     CPPUNIT_ASSERT_NO_THROW(_vzr->fitnessChanged());

}
void TVizirTest::testTVizirOptimizeMonoCriterionWithoutConstraint()
{
 
    using namespace URANIE::DataServer;
    using namespace URANIE::Launcher;
    using namespace URANIE::Optimizer;
    
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_INBUILDING_TEST();


    if(_tcode != NULL)
    {
        delete _tcode;
        _tcode = NULL;
    }
    
#ifdef WIN32
    _tcode = new TCode(_tdsRosenbrock,"rosenbrock.exe -v -k 1>NUL");
#else  
    _tcode = new TCode(_tdsRosenbrock, Form("%s/../../teST/coDE/rosENBROK/rosenbrock -v -k",
                        _workingTestPath.Data()));
#endif
    
    // Add the output file
    _tcode->addOutputFile( _fOutputFile );
    

    _vzr = new TVizir(_tdsRosenbrock, 200);
    _vzr->unsetLog();
    _vzr->setCost(_tcode, 0);
    _vzr->setPrecision(0.0001);
    _vzr->optimize();
    
    

    
    Double_t min = 1.e+10;
    Int_t minIndex = -1;
    
    for(Int_t it = 0 ; it < _tdsRosenbrock->getNPatterns() ; it++)
    {
       
        if((fabs(_tdsRosenbrock->getValue("fval",it)) < min))
        {
            min = fabs(_tdsRosenbrock->getValue("fval",it));
            minIndex = it;
        }
           
    }
    
    CPPUNIT_ASSERT(minIndex!=-1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0,min, 1.e-2);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0,_tdsRosenbrock->getValue("x",minIndex)  ,4*1.e-2);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0,_tdsRosenbrock->getValue("y",minIndex) ,4*1.e-2);
    
//     _tdsRosenbrock->exportData("_vizir_code_rosenbrock_.dat");    
    
    CPPUNIT_ASSERT(_tdsRosenbrock->isAttribute("fval"));
    CPPUNIT_ASSERT(_tdsRosenbrock->isAttribute("rgpareto"));
    CPPUNIT_ASSERT(_tdsRosenbrock->isAttribute("generation"));
    
}

void TVizirTest::testTVizirOptimizeMonoCriterionWithConstraint()
{
    
    using namespace URANIE::DataServer;

    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_INBUILDING_TEST();

    _vzr = new URANIE::Optimizer::TVizir(_tdsRosenbrock, 1000);
    _vzr->unsetLog();
    
//fga 2015/01/12     // Get the void * from the "rosenbrockConstraint" analytical function
//fga 2015/01/12     G__ClassInfo globalscope;
//fga 2015/01/12     G__MethodInfo method;
//fga 2015/01/12     long dummy;
//fga 2015/01/12     // resolve function overloading
//fga 2015/01/12     method = globalscope.GetMethod("rosenbrockConstraint", "double*,double*", &dummy);
//fga 2015/01/12     void * p2fRosenbrockConstraint = method.PointerToFunc();
   
//fga 2015/01/12     // Get the void * from the "contHard2" analytical function
//fga 2015/01/12     G__ClassInfo globalscope2;
//fga 2015/01/12     G__MethodInfo method2;
//fga 2015/01/12     long dummy2;
//fga 2015/01/12     // resolve function overloading
//fga 2015/01/12     method2 = globalscope2.GetMethod("contHard2", "double*,int*", &dummy2);
//fga 2015/01/12     void * p2fcontHard2 = method2.PointerToFunc();
    
//fga 2015/01/12     // Get the void * from the "mydecode2" analytical function
//fga 2015/01/12     G__ClassInfo globalscope3;
//fga 2015/01/12     G__MethodInfo method3;
//fga 2015/01/12     long dummy3;
//fga 2015/01/12     // resolve function overloading
//fga 2015/01/12     method3 = globalscope3.GetMethod("mydecode2", "int*, int*, double*, double*", &dummy3);
//fga 2015/01/12     void * p2fmydecode2 = method3.PointerToFunc();
    
    
    _vzr->setDecode("mydecode2", 2);
    _vzr->setHardConstraint("contHard2");
    _vzr->setConstraintAndCost("rosenbrockConstraint","const1:rosenbrock",1);
    _vzr->setPrecision(1.e-7);
    _vzr->optimize();
    
    Bool_t bOK = kTRUE;
    Int_t nAccepted = 0;
    Double_t min = 1.e+10;
    Int_t minIndex = -1;
 
    for(Int_t it = 0 ; it < _tdsRosenbrock->getNPatterns() ; it++)
    {
        if(_tdsRosenbrock->getValue("const1",it) > 0.5)
            nAccepted++;
        
        if((fabs(_tdsRosenbrock->getValue("rosenbrock",it)) < min) && (_tdsRosenbrock->getValue("const1",it) > 0.5))
        {
            min = fabs(_tdsRosenbrock->getValue("rosenbrock",it));
            minIndex = it;
        }
           
    }
    
    CPPUNIT_ASSERT(minIndex!=-1);

    if(min > 1.e-4) bOK = kFALSE;
        
    if(fabs(_tdsRosenbrock->getValue("x",minIndex) -1.0) > 1.e-2) bOK = kFALSE;
        
    if(fabs(_tdsRosenbrock->getValue("y",minIndex) -1.0) > 1.e-2) bOK = kFALSE;
           
            
    
    if(nAccepted == 0)  bOK = kFALSE;
   
    
//     _tdsRosenbrock->exportData("_vizir_mono_contraint_.dat");
  
    CPPUNIT_ASSERT(bOK);
    CPPUNIT_ASSERT(_tdsRosenbrock->isAttribute("rosenbrock"));
    CPPUNIT_ASSERT(_tdsRosenbrock->isAttribute("const1"));
    CPPUNIT_ASSERT(_tdsRosenbrock->isAttribute("rgadmin"));
    CPPUNIT_ASSERT(_tdsRosenbrock->isAttribute("rgpareto"));
    CPPUNIT_ASSERT(_tdsRosenbrock->isAttribute("generation"));
    
}

void TVizirTest::testTVizirOptimizeMultiCriterionWithoutConstraint()
{
    
    using namespace URANIE::DataServer;
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_INBUILDING_TEST();

    // Les variables de l'etude
    TAttribute *x = new TAttribute("x", -1.5, 1.5);
    x->setDefaultValue(0.5);
    x->setStepValue(0.01);
    x->delShare();
    
    TAttribute *y = new TAttribute("y", -1.5,1.5);
    y->setDefaultValue(0.5);
    y->setStepValue(0.01);
    y->delShare();
    
    TDataServer *tds = new TDataServer("tds", "Optimize via TDataServer");
    tds->addAttribute(x);
    tds->addAttribute(y);

    _vzr = new URANIE::Optimizer::TVizir(tds, 2000);
    _vzr->unsetLog();
    
//fga 2015/01/12     // Get the void * from the "funAllCost2" analytical function
//fga 2015/01/12     G__ClassInfo globalscope;
//fga 2015/01/12     G__MethodInfo method;
//fga 2015/01/12     long dummy;
//fga 2015/01/12     // resolve function overloading
//fga 2015/01/12     method = globalscope.GetMethod("funAllCost2", "double*,double*", &dummy);
//fga 2015/01/12     void * p2fFunAllCost2 = method.PointerToFunc();
    
    _vzr->setCost("funAllCost2", "c1:c2");
    _vzr->setPrecision(1.e-7);
    
    _vzr->optimize();
    
    
    for(Int_t it = 0 ; it < tds->getNPatterns() ; it++)
    {
        CPPUNIT_ASSERT(fabs(tds->getValue("x",it)-tds->getValue("y",it)) < 5*1.e-1);
        CPPUNIT_ASSERT(tds->getValue("x",it) < 0.04);
        CPPUNIT_ASSERT(tds->getValue("y",it) < 0.04);

           
    }
    
//     tds->exportData("_vizir_multi_without_constraint_.dat");
    
    CPPUNIT_ASSERT(tds->isAttribute("c1"));
    CPPUNIT_ASSERT(tds->isAttribute("c2"));
    CPPUNIT_ASSERT(tds->isAttribute("rgpareto"));
    CPPUNIT_ASSERT(tds->isAttribute("generation"));
    
    
    
    delete tds;
    
}

void TVizirTest::testTVizirOptimizeMultiCriterionWithConstraint()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_INBUILDING_TEST();

   
//fga 2015/01/12     // Get the void * from the "funCostAndConstraint2" analytical function
//fga 2015/01/12     G__ClassInfo globalscope;
//fga 2015/01/12     G__MethodInfo method;
//fga 2015/01/12     long dummy;
//fga 2015/01/12     // resolve function overloading
//fga 2015/01/12     method = globalscope.GetMethod("funCostAndConstraint2", "double*,double*", &dummy);
//fga 2015/01/12     void * p2ffunCostAndConstraint = method.PointerToFunc();
    
    

    _vzr = new URANIE::Optimizer::TVizir(_tds, 1000);
    _vzr->unsetLog();
    _vzr->setConstraintAndCost("funCostAndConstraint2","fc1:c1:c2",1);
    _vzr->setPrecision(1.e-4);
    _vzr->optimize("saveall");
    
    Int_t nPatterns = _tds->getNPatterns();
    Int_t nGeneration = (Int_t)(_tds->getValue("generation",nPatterns-1));
    for(Int_t it = 0 ; it < nPatterns ; it++)
    {
        if(_tds->getValue("generation",it) == nGeneration)
        {
            CPPUNIT_ASSERT( _tds->getValue("x",it) > 1 - 0.03);
            CPPUNIT_ASSERT( _tds->getValue("y",it) > 1 - 0.03);
        }
           
    }
    
//     _tds->exportData("_vizir_multi_with_constraint_.dat");

    CPPUNIT_ASSERT(_tds->isAttribute("c1"));
    CPPUNIT_ASSERT(_tds->isAttribute("c2"));
    CPPUNIT_ASSERT(_tds->isAttribute("fc1"));
    CPPUNIT_ASSERT(_tds->isAttribute("rgpareto"));
    CPPUNIT_ASSERT(_tds->isAttribute("generation"));
    
}

void TVizirTest::testTVizirDecode()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_INBUILDING_TEST();

    
//fga 2015/01/12     // Get the void * from the "rosenbrock" analytical function
//fga 2015/01/12     G__ClassInfo globalscope;
//fga 2015/01/12     G__MethodInfo method;
//fga 2015/01/12     long dummy;
//fga 2015/01/12     // resolve function overloading
//fga 2015/01/12     method = globalscope.GetMethod("mydecodeImage", "int*,int*,double*,double*", &dummy);
//fga 2015/01/12     void * p2fmydecodeImage = method.PointerToFunc();

//fga 2015/01/12     char *funcname = G__p2f2funcname(p2fmydecodeImage);
    
    int i;
    Int_t TAILLEPOP = 400;
    Int_t NBGENE = 7;
    //Double_t PRECISION = 0.001;
    /* init */
   URANIE::DataServer::TDataServer *tdsvzr = new URANIE::DataServer::TDataServer("tdsvzr", "tds for vizir test");
    tdsvzr->addAttribute("fond", 0., 256.);
    for (i=1; i<NBGENE+1; i++) 
        tdsvzr->addAttribute(Form("zone%d", i), 0., 256.);

    URANIE::Optimizer::TVizir *vzr = new URANIE::Optimizer::TVizir(tdsvzr, TAILLEPOP);
    vzr->unsetLog();
    vzr->setDecode("mydecodeImage", NBGENE);
    
    vzr->minmax.ResizeTo(NBGENE+1, 2);
    for (int it = 0; it < NBGENE+1; it++)
    {
       URANIE::DataServer::TAttribute *att = (URANIE::DataServer::TAttribute *) tdsvzr->getAttribute(it);
       vzr->minmax(it, 0) = att->getLowerBound();
       vzr->minmax(it, 1) = att->getUpperBound();
     }
    
    double* in = new double[8];
    double* out = new double[8];
    
    

    // creation TMethodCall
    TMethodCall*  fDecode = new TMethodCall();
    fDecode->InitWithPrototype("mydecodeImage",
            "Int_t*,Int_t*,Double_t*,Double_t*");

    for (i=0; i<8; i++)
        in[i]=i*0.1;

    // mode avec fDecode preciser
    vzr->decode(in,out,fDecode);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(256.0,out[0],DBL_PRECISION_COMPARE);
    for (i=1; i<NBGENE+1; i++) 
        CPPUNIT_ASSERT_DOUBLES_EQUAL(256.0*in[i-1],out[i],DBL_PRECISION_COMPARE);
   
    //mode par defaut
    vzr->decode(in,out,NULL);
    for (i=0; i<NBGENE+1; i++) 
        CPPUNIT_ASSERT_DOUBLES_EQUAL(256.0*in[i],out[i],DBL_PRECISION_COMPARE);
       
    
   delete[] in;
   delete[] out;
   delete fDecode;
   delete vzr;
   delete tdsvzr;
    
}

void TVizirTest::testTVizirEvaluerCode()
{
    
  using namespace URANIE::DataServer;
  using namespace URANIE::Launcher;
  using namespace URANIE::Optimizer;
 
  //Directory where we can find references input files
  TString sRepertoireReference = gSystem->pwd();
  
  // The reference input file
  TString sJDDReference = gSystem->pwd();
  sJDDReference += TString("/input_rosenbrock_with_keys.dat");
  
  // The x attribute of the use case
  TAttribute *x = new TAttribute("x", -1.5, 1.5);
  x->setFileKey(sJDDReference);
  x->delShare();
  
  // The y attribute of the use case
  TAttribute *y = new TAttribute("y", -1.5,1.5);
  y->setFileKey(sJDDReference);
  y->delShare();

  // Define the DataServer and add the two attributes
  TDataServer *tdsRosenbrock = new TDataServer("tdsRosenbrock", "Optimize Code externe Rosenbrock via TDataServer");
  tdsRosenbrock->addAttribute(x);
  tdsRosenbrock->addAttribute(y);
  
  // The output file of the code where values are stored in (key = value) format
  TOutputFileKey *fOutputFile = new TOutputFileKey("_output_rosenbrock_with_keys_.dat");
  fOutputFile->addAttribute("fval");
  
  // Create an TCode object for my code
  TCode *myRosenbrockCode;
  myRosenbrockCode = new TCode(tdsRosenbrock, UranieTests::code_path("rosenbrock") + " -v -k");
  
  myRosenbrockCode->setWorkingDirectory(gSystem->UnixPathName(
            gSystem->pwd()
                    + TString("/tmpLanceurUranie/OptimizerUranie/rosenbrock")));
  // Add the output file
  myRosenbrockCode->addOutputFile( fOutputFile );
  
  
  TVizir *vzr = new TVizir(tdsRosenbrock, 200);
  vzr->unsetLog();
  vzr->setCost(myRosenbrockCode, 1);
 
//fga 2015/01/12   // Get the void * from the "contHard2" analytical function
//fga 2015/01/12   G__ClassInfo globalscope2;
//fga 2015/01/12   G__MethodInfo method2;
//fga 2015/01/12   long dummy2;
//fga 2015/01/12   // resolve function overloading
//fga 2015/01/12   method2 = globalscope2.GetMethod("contHard2", "double*,int*", &dummy2);
//fga 2015/01/12   void * p2fcontHard2 = method2.PointerToFunc();
//fga 2015/01/12   char *funcname = G__p2f2funcname(p2fcontHard2);
    
  // creation TMethodCall
  TMethodCall*  fHard = new TMethodCall();
  fHard->InitWithPrototype("contHard2",
            "double*,int*");
   
  
  
  double* in = new double[2];
  double* out = new double[2];
  Int_t admis = 66666;
  in[0] = 0.0;
  in[1] = 0.0;
  admis = vzr->evaluer(in, out, (TMethodCall *)NULL,(TMethodCall *)NULL, myRosenbrockCode);  

  CPPUNIT_ASSERT_DOUBLES_EQUAL(146.875,out[0],DBL_PRECISION_COMPARE);
  in[0] = 1.0;
  in[1] = 1.0;
  admis = vzr->evaluer(in, out, (TMethodCall *)NULL,(TMethodCall *)NULL, myRosenbrockCode);  

  CPPUNIT_ASSERT_DOUBLES_EQUAL(5.875,out[0],DBL_PRECISION_COMPARE);
  in[0] = 0.3;
  in[1] = 0.7;
  admis = vzr->evaluer(in, out, (TMethodCall *)NULL,(TMethodCall *)NULL, myRosenbrockCode);  

  CPPUNIT_ASSERT_DOUBLES_EQUAL(3.136,out[0],DBL_PRECISION_COMPARE);
  
  delete[] in;
  delete[] out;
  delete fOutputFile;
  delete myRosenbrockCode;
  delete tdsRosenbrock;
  delete fHard;
  delete vzr;
  
}


////////////////////////////////
///                          ///                            
///     throwexception       ///
///                          ///
////////////////////////////////
void TVizirTest::throwExceptionTVizirConstructorTdsNull()
{

    URANIE::DataServer::TDataServer *tdsnull = NULL;

    CPPUNIT_ASSERT_THROW(new URANIE::Optimizer::TVizir(tdsnull, 1000),
            URANIE::Exceptions::UErrorExceptions);
}

void TVizirTest::throwExceptionTVizirConstructorFalseNpop()
{

 CPPUNIT_ASSERT_THROW(new URANIE::Optimizer::TVizir(_tds,0),URANIE::Exceptions::UErrorExceptions); 
}

void TVizirTest::throwExceptionTVizirSetDecodeFalseFun()
{
    _vzr = new URANIE::Optimizer::TVizir(_tds2, 400);
    _vzr->unsetLog();
    void * toto = NULL;
    void * titi = new double;
    CPPUNIT_ASSERT_THROW(_vzr->setDecode("", 7),
        URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(_vzr->setDecode("titi", 7),
        URANIE::Exceptions::UErrorExceptions);

    delete (double*) titi;
}
void TVizirTest::throwExceptionTVizirsetHardConstraintFalseFun()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_INBUILDING_TEST()
    ;

    _vzr = new URANIE::Optimizer::TVizir(_tds2, 400);
    _vzr->unsetLog();
    void * toto = NULL;
    void * titi = new double;
    CPPUNIT_ASSERT_THROW(_vzr->setHardConstraint(""),
        URANIE::Exceptions::UErrorExceptions);
    CPPUNIT_ASSERT_THROW(_vzr->setHardConstraint("titi"),
        URANIE::Exceptions::UErrorExceptions);

    delete (double*) titi;
}
void TVizirTest::throwExceptionTVizirsetConstraintAndCostFalseFun()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_INBUILDING_TEST()
    ;
}
void TVizirTest::throwExceptionTVizirsetConstraintAndCostFalseInput()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_INBUILDING_TEST()
    ;
}
void TVizirTest::throwExceptionTVizirsetConstraintAndCostFalseNHidden()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_INBUILDING_TEST()
    ;
}
void TVizirTest::throwExceptionTVizirsetConstraintAndCostFalseNbConst()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_INBUILDING_TEST()
    ;
}
void TVizirTest::throwExceptionTVizirsetConstraintAndCostFalseTCode()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_INBUILDING_TEST()
    ;
}
void TVizirTest::throwExceptionTVizirevaluerDecodNull()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_INBUILDING_TEST()
    ;
}
void TVizirTest::throwExceptionTVizirevaluerHardNull()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_INBUILDING_TEST()
    ;
}
void TVizirTest::throwExceptionTVizirevaluerSoftNull()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_INBUILDING_TEST()
    ;
}
void TVizirTest::throwExceptionTVizirevaluerTCodeNull()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_INBUILDING_TEST()
    ;
}

void TVizirTest::throwExceptionTVizirSetNbGenerationToDraw()
{
    _vzr = new URANIE::Optimizer::TVizir(_tds, 10000);
    _vzr->unsetLog();
    
    CPPUNIT_ASSERT_THROW(_vzr->setNbGenerationToDraw(-1),
            URANIE::Exceptions::UWarningExceptions);
    
    CPPUNIT_ASSERT_THROW(_vzr->setNbGenerationToDraw(10),
            URANIE::Exceptions::UWarningExceptions);
    
}



void TVizirTest::throwExceptionTVizirSetVarDraw()
{
    _vzr = new URANIE::Optimizer::TVizir(_tds, 10000);
    _vzr->unsetLog();

    CPPUNIT_ASSERT_THROW(_vzr->setVarDraw("y:x", "", "generation=10"),
            URANIE::Exceptions::UWarningExceptions);
    
    CPPUNIT_ASSERT_THROW(_vzr->setVarDraw("y:x", "", "generation=1000000000000000000000000000000000000000000000000000000000000"),
            URANIE::Exceptions::UWarningExceptions);
    
    CPPUNIT_ASSERT_THROW(_vzr->setVarDraw("y:x", "", "toto"),
            URANIE::Exceptions::UWarningExceptions);

}



void TVizirTest::throwExceptionTViziroptimizeFalseOption()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_INBUILDING_TEST()
    ;
}
