//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////

// ROOT
#include "TCanvas.h"
#include "TH2.h"
#include "TLegend.h"
#include "TLatex.h"
#include "TFile.h"
#include "TMath.h"

// URANIE
#include "vzr_class.h"
#include "TAttribute.h"
#include "UExceptions.h"

using namespace URANIE::DataServer;

#include "Api.h"
#define NL 5000        /* nombre de valeurs du tableau al�atoire */
#define NN 100
#define cmax 101   /* utilise dans rang, doit etre sup a nbcrit */

VizirCommon *gVizir;

ClassImp (VizirCommon)

VizirCommon::VizirCommon() :
        TNamed("Vizir", "The Minimization package by Genetic algorithms"), _sdraw(
                0), _scut(0), _sopt(0), _nIter(0), fFitness(NULL), fDecode(
                NULL), fCout(NULL), fHardCont(NULL)
{
    _blog = kFALSE;
    bImageZoneAdmissible = kFALSE;
    bZoneAdmissible = kFALSE;
    nmode = MONO;
    nbgen = 0;
    nbcout = 0;
    nbcont = 0;
    _dprec = 0.01;
    fMethodCall = 0;
    _datatree = 0;
    _ntZoneAdmissible = 0;
    _ntImageZoneAdmissible = 0;
    initDefault();
}

VizirCommon::VizirCommon(int mode, int gene, int cout, int cont, double prec,
        int (*fct)(double*, double*, void*), void *arg) :
        TNamed("Vizir", "The Minimization package by Genetic algorithms"), _sdraw(
                0), _scut(0), _sopt(0), _nIter(0), fFitness(NULL), fDecode(
                NULL), fCout(NULL), fHardCont(NULL)
{
    _blog = kFALSE;
    bImageZoneAdmissible = kFALSE;
    bZoneAdmissible = kFALSE;
    nmode = mode;
    nbgen = gene;
    nbcout = cout;
    nbcont = cont;
    _dprec = prec;
    fMethodCall = 0;
    initDefault();
    _datatree = 0;
    _ntZoneAdmissible = 0;
    _ntImageZoneAdmissible = 0;
}
VizirCommon::VizirCommon(int mode, int gene, int cout, int cont, double prec) :
        TNamed("Vizir", "The Minimization package by Genetic algorithms"), _sdraw(
                0), _scut(0), _sopt(0), _nIter(0), fFitness(NULL), fDecode(
                NULL), fCout(NULL), fHardCont(NULL)
{
    _blog = kFALSE;
    bImageZoneAdmissible = kFALSE;
    bZoneAdmissible = kFALSE;
    nmode = mode;
    nbgen = gene;
    nbcout = cout;
    nbcont = cont;
    _dprec = prec;
    fMethodCall = 0;
    _datatree = 0;
    _ntZoneAdmissible = NULL;
    _ntImageZoneAdmissible = NULL;
//   for (Int_t i =0; i<10; i++) {
//     fCoutCall[i] = new TMethodCall();
//     fHardContCall[i] = new TMethodCall();
//   }
    _dataSpecification = new URANIE::DataServer::TDataSpecification();
    //_dataSpecification->setRelationName("Parameter of the minimization");
    _coutSpecification = new URANIE::DataServer::TDataSpecification();
    //_coutSpecification->setRelationName("Cost of the minimization");
    initDefault();
}

VizirCommon::~VizirCommon()
{
    cout << "*** Info [Delete VizirCommon] ***" << endl;
    if (_ntZoneAdmissible)
    {
        delete _ntZoneAdmissible;
        _ntZoneAdmissible = NULL;
    }
    if (_ntImageZoneAdmissible)
    {
        delete _ntImageZoneAdmissible;
        _ntImageZoneAdmissible = NULL;
    }
    cout << "*** fin de delete VizirCommon ***" << endl;
}

void VizirCommon::setPointP(Double_t px, Double_t py)
{
    _xPointP = px;
    _yPointP = py;
}

void VizirCommon::Draw(const char* varexp, const char* selection,
        Option_t* option)
{
    if (_datatree)
    {
        _datatree->Draw(varexp, selection, option);
        gPad->Modified();
        gPad->Update();
    }
    else
        cout << "VizirCommon::Draw: Error pas de tree _datatree" << endl;
}

void VizirCommon::drawLastGeneration(const char* varexp, Option_t* option)
{
    Draw(varexp, Form("(generation==%d) && (rangdomi==0)", getNbGeneration()),
            option);
}

void VizirCommon::Scan(const char *varexp, const char *selection,
        Option_t *option)
{
    if (_datatree)
        _datatree->Scan(varexp, selection, option);
    else
        cout << "VizirCommon::Scan: Error pas de tree _datatree" << endl;
}

void VizirCommon::NtDraw(TNtupleD *nt, TString xVar, TString yVar,
        TString sTitre, Option_t *option)
{
    nt->Draw(Form("%s:%s", yVar.Data(), xVar.Data()), "", option);

    // Les titres
    TH1F *htemp = (TH1F*) gPad->GetPrimitive("htemp");
    if (htemp)
    {
        htemp->SetXTitle(xVar);
        htemp->SetYTitle(yVar);
        htemp->SetTitle(sTitre);
    }

}

void VizirCommon::drawFrontPareto(Int_t nflag)
{
    Option_t *option = "";
    Int_t ntitle;
    if (nflag == 0)
    {
        ntitle = drawImageZoneAdmissible();
        if (ntitle == 1)
            option = "same";
    }

    TString sVar;
    if (nbcout > 1)
        sVar = _coutSpecification->getAttribute(1)->GetName() + TString(":");
    sVar += _coutSpecification->getAttribute(0)->GetName();

    _datatree->SetMarkerColor(1);
    _datatree->Draw(sVar, Form("(generation==%d) && (rangdomi==0)", _nIter),
            option);
    _datatree->SetMarkerColor(2);

    TH1F *htemp = (TH1F*) gPad->GetPrimitive("htemp");
    if (htemp)
    {
        htemp->SetTitle("Set of Pareto Optimal Solutions");
        if (ntitle == 0)
        {
            htemp->SetXTitle(_coutSpecification->getAttribute(0)->GetName());
            htemp->SetYTitle(_coutSpecification->getAttribute(1)->GetName());
        }
    }
}

void VizirCommon::drawZonePareto(Int_t nflag)
{
    Option_t *option = "";
    if (nflag == 0)
    {

        if (drawZoneAdmissible(1))
            option = "same";

    }

    if (nflag == 2)
    {
        drawZoneRecherche();
        option = "same";
    }

    TString sVar;
    if (nbgen > 1)
        sVar = _dataSpecification->getAttribute(1)->GetName() + TString(":");
    sVar += _dataSpecification->getAttribute(0)->GetName();

    _datatree->SetMarkerColor(1);
    _datatree->Draw(sVar, Form("(generation==%d) && (rangdomi==0)", _nIter),
            option);
    _datatree->SetMarkerColor(2);

    TH1F *htemp = (TH1F*) gPad->GetPrimitive("htemp");
    if (htemp)
        htemp->SetTitle("Search Area");
}

void VizirCommon::drawCriteres(Int_t i, Int_t j)
{
    TMatrixDRow vX = TMatrixDRow(pop_val, i);
    TMatrixDRow vY = TMatrixDRow(pop_val, j);

    TGraph *gr = new TGraph(itaille, vX.GetPtr(), vY.GetPtr());
    gr->SetTitle(Form("Critere %d vs %d", j, i));
    gr->SetMarkerStyle(21);
    gr->SetMarkerColor(kRed);
    gr->SetMarkerSize(0.3);
    gr->Draw("ap");

}

void VizirCommon::drawZoneRecherche(Int_t nx, Int_t ny, Int_t nColor)
{
    if (nx > nbgen)
        throw URANIE::Exceptions::UExceptions(__FILE__, __LINE__,
                Form(
                        "VizirCommon::drawZoneRecherche ERROR : the index nx [%d] is greater than number of parameter [%d]",
                        nx, nbgen));
    if (nx < 0)
        throw URANIE::Exceptions::UExceptions(__FILE__, __LINE__,
                Form(
                        "VizirCommon::drawZoneRecherche ERROR : the index nx [%d] is lesser than 0",
                        nx));
    if (ny > nbgen)
        throw URANIE::Exceptions::UExceptions(__FILE__, __LINE__,
                Form(
                        "VizirCommon::drawZoneRecherche ERROR : the index ny [%d] is greater than number of parameter [%d]",
                        ny, nbgen));
    if (ny < 0)
        throw URANIE::Exceptions::UExceptions(__FILE__, __LINE__,
                Form(
                        "VizirCommon::drawZoneRecherche ERROR : the index ny [%d] is lesser than 0",
                        ny));
    if (nbgen < 2)
        return;

    Float_t xlower =
            (Float_t) _dataSpecification->getAttribute(nx)->getLowerBound();
    Float_t xupper =
            (Float_t) _dataSpecification->getAttribute(nx)->getUpperBound();

    Float_t ylower =
            (Float_t) _dataSpecification->getAttribute(ny)->getLowerBound();
    Float_t yupper =
            (Float_t) _dataSpecification->getAttribute(ny)->getUpperBound();

    Float_t x[5] =
    { xlower, xupper, xupper, xlower, xlower };
    Float_t y[5] =
    { ylower, ylower, yupper, yupper, ylower };

    TGraph *grSearch = new TGraph(5, x, y);
    grSearch->SetFillColor(nColor);
    grSearch->SetTitle("Search Area");
    grSearch->GetXaxis()->SetTitle(
            _dataSpecification->getAttribute(nx)->GetName());
    grSearch->GetYaxis()->SetTitle(
            _dataSpecification->getAttribute(ny)->GetName());
    grSearch->Draw("AF");
}

Int_t VizirCommon::drawZoneAdmissible(Int_t nflag)
{
    if (nbgen < 2)
        return 0;

    if (ncrtnbhardcont == 0)
    {
        cout
                << "DrawZoneAdmissible: pas de contrainte hard donc zone admissible est la zone de recherche"
                << endl;
        return 1;
    }

    Int_t nx = 0;
    Int_t ny = 1;
    TString sx = _dataSpecification->getAttribute(nx)->GetName();
    TString sy = _dataSpecification->getAttribute(ny)->GetName();

    Float_t xlower =
            (Float_t) _dataSpecification->getAttribute(nx)->getLowerBound();
    Float_t xupper =
            (Float_t) _dataSpecification->getAttribute(nx)->getUpperBound();
    Float_t ylower =
            (Float_t) _dataSpecification->getAttribute(ny)->getLowerBound();
    Float_t yupper =
            (Float_t) _dataSpecification->getAttribute(ny)->getUpperBound();

    if (bZoneAdmissible == kFALSE)
    {
        Int_t nX = 500;
        Int_t nY = 500;

        Double_t xpas = (xupper - xlower) / (nX + 1.0);
        Double_t ypas = (yupper - ylower) / (nY + 1.0);

        TVectorD par;
        par.ResizeTo(nbgen);

        // Cr�ation des labels des variables
        TString sLabelVariable = "iter:";
        sLabelVariable += Form("%s:", sx.Data());
        sLabelVariable += Form("%s", sy.Data());

        _ntZoneAdmissible = new URANIE::DataServer::TDSNtupleD("ZoneAdmissible",
                "ZoneAdmissible", sLabelVariable);

        Int_t nCrt = 1;
        for (Int_t i = 0; i < nX; i++)
        {
            for (Int_t j = 0; j < nY; j++)
            {
                Int_t nAdmissible = 1;
                par[0] = xlower + i * xpas;
                par[1] = ylower + j * ypas;
                (*fHardCont)(par.GetMatrixArray(), nAdmissible);
                if (nAdmissible == 1)
                {
                    _ntZoneAdmissible->Fill(nCrt * 1.0, par[0], par[1]);
                    nCrt++;
                }
            }
        }
        if (_blog)
            _ntZoneAdmissible->Scan("*");
        _ntZoneAdmissible->SetMarkerColor(2);
        _ntZoneAdmissible->SetMarkerSize(4);
        _ntZoneAdmissible->SetMarkerStyle(7);

//     TFile tf("_vizir_optimizer.root", "recreate");
//     _ntZoneAdmissible->Write("", TObject::kOverwrite);
//     tf.Close();

        bZoneAdmissible = kTRUE;
    }

    Option_t *option = "";
    if (nflag == 0)
    {
        // Dessine la zone de recherche
        Int_t nColor = 42;
        Float_t x[5] =
        { xlower, xupper, xupper, xlower, xlower };
        Float_t y[5] =
        { ylower, ylower, yupper, yupper, ylower };

        TGraph *grSearch = new TGraph(5, x, y);
        grSearch->SetFillColor(nColor);
        grSearch->SetTitle("Zone Admissible");
        grSearch->GetXaxis()->SetTitle(
                _dataSpecification->getAttribute(nx)->GetName());
        grSearch->GetYaxis()->SetTitle(
                _dataSpecification->getAttribute(ny)->GetName());
        grSearch->Draw("AF");
        option = "same";
    }

    NtDraw(_ntZoneAdmissible, sx, sy, "Zone Admissible", option);

    TLatex *t = new TLatex();
    t->SetTextFont(32);
    t->SetTextColor(1);
    t->SetTextSize(0.05);
    t->SetTextAlign(12);
    if (nflag == 0)
        t->DrawLatex(0.2, 0.8, "Zone de Recherche ");
    t->DrawLatex(0.7, 0.10, "Zone Admissible ");
    return 1;
}

Int_t VizirCommon::drawImageZoneAdmissible(Int_t nflag)
{
    if (nbgen < 2)
        return 0;

    if (ncrtnbhardcont == 0)
    {
        cout << "DrawImageZoneAdmissible : pas de contrainte hard" << endl;
        return 0;
    }

    Int_t nx = 0;
    Int_t ny = 1;
//   TString sx = TString(fCoutCall[nx]->GetMethodName());
    TString sx = _coutSpecification->getAttribute(nx)->GetName(); //TString(fCoutCall[nx]->GetMethodName());

//   TString sy = TString(fCoutCall[ny]->GetMethodName());
    TString sy = _coutSpecification->getAttribute(ny)->GetName();

    if (bImageZoneAdmissible == kFALSE)
    {
        Int_t nX = 500;
        Int_t nY = 500;

        Double_t xlower = _dataSpecification->getAttribute(nx)->getLowerBound();
        Double_t xupper = _dataSpecification->getAttribute(nx)->getUpperBound();
        Double_t xpas = (xupper - xlower) / (nX + 1.0);

        Double_t ylower = _dataSpecification->getAttribute(ny)->getLowerBound();
        Double_t yupper = _dataSpecification->getAttribute(ny)->getUpperBound();
        Double_t ypas = (yupper - ylower) / (nY + 1.0);

        Int_t ncrits = nbcout;
        Double_t *crits = new Double_t[ncrits];
        TVectorD par;
        par.ResizeTo(nbgen);

        // Cr�ation des labels des variables
        TString sLabelVariable = "iter:";
        sLabelVariable += Form("%s:", sx.Data());
        sLabelVariable += Form("%s", sy.Data());
        _ntImageZoneAdmissible = new URANIE::DataServer::TDSNtupleD(
                "ImageZoneAdmissible", "ImageZoneAdmissible", sLabelVariable);
        Int_t nCrt = 1;

        for (Int_t i = 0; i < nX; i++)
        {
            for (Int_t j = 0; j < nY; j++)
            {
                Int_t nAdmissible = 1;
                par[0] = xlower + i * xpas;
                par[1] = ylower + j * ypas;
                (*fHardCont)(par.GetMatrixArray(), nAdmissible);
                if (nAdmissible == 1)
                {
                    // Calcul des couts et des contraintes
                    (*fCout)(par.GetMatrixArray(), crits);
                    _ntImageZoneAdmissible->Fill(nCrt * 1.0, crits[0],
                            crits[1]);
                    nCrt++;
                }
            }
        }
        _ntImageZoneAdmissible->SetMarkerColor(2);
        _ntImageZoneAdmissible->SetMarkerSize(4);
        _ntImageZoneAdmissible->SetMarkerStyle(7);
        bImageZoneAdmissible = kTRUE;
        delete[] crits;
        crits = NULL;
    }

    NtDraw(_ntImageZoneAdmissible, sx, sy, "Image de la Zone Admissible");

    // Ajout du point
    if (nflag > 0)
    {
        Double_t dx, dy;

        if (nflag == 5)
        {
            dx = 4.7326903;
            dy = 2.7219929;
        }
        else
        {
            dx = _xPointP;
            dy = _yPointP;
        }
        // L�gende
        TLegend *leg = new TLegend(0.52, 0.65, 0.89, 0.85);
        leg->SetBorderSize(0);
        leg->SetFillColor(10);
        leg->SetTextAlign(22);
        leg->SetTextSize(0.04);

        Float_t *dum = 0;
        TGraph *grpt = new TGraph(1, dum, dum);
        grpt->SetMarkerColor(4);
        grpt->SetMarkerStyle(8);
        grpt->SetPoint(0, dx, dy);
        grpt->Draw("Psame");
        leg->AddEntry(grpt, Form("Point P ( %.3f, %.3f)", dx, dy), "P");
        leg->Draw();

        // Ajout des axes
        if (nflag > 1)
        {
            TH2F *htemp = (TH2F*) gPad->GetPrimitive("htemp");
            if (htemp)
            {
                Double_t xmin = htemp->GetXaxis()->GetXmin();
                Double_t xmax = htemp->GetXaxis()->GetXmax();
                Double_t ymin = htemp->GetYaxis()->GetXmin();
                Double_t ymax = htemp->GetYaxis()->GetXmax();
                // Droite en X
                TGraph *graph = new TGraph(2, dum, dum);
                graph->SetLineWidth(1);
                graph->SetLineColor(1);
                graph->SetLineStyle(1);   // 1=solid, 2=dash
                graph->SetPoint(0, dx, ymin);
                graph->SetPoint(1, dx, ymax);
                graph->Draw("Lsame");
                // Droite en Y
                graph = new TGraph(2, dum, dum);
                graph->SetLineWidth(1);
                graph->SetLineColor(1);
                graph->SetLineStyle(1);   // 1=solid, 2=dash
                graph->SetPoint(0, xmin, dy);
                graph->SetPoint(1, xmax, dy);
                graph->Draw("Lsame");

                // solution domin� par P
                if (nflag > 2)
                {
                    Int_t nColordomine = 3;
                    _ntImageZoneAdmissible->SetMarkerColor(nColordomine);
                    TString sdraw = Form("%s:%s", sy.Data(), sx.Data());
                    TString sselect = Form(" ( %s > %f ) & ( %s > %f ) ",
                            sx.Data(), dx, sy.Data(), dy);
                    _ntImageZoneAdmissible->Draw(sdraw, sselect, "same");
                    _ntImageZoneAdmissible->SetMarkerColor(2);
                    TGraph *grdomine = new TGraph(1, dum, dum);
                    grdomine->SetFillColor(nColordomine);
                    grdomine->SetPoint(0, dx, dy);
                    leg->AddEntry(grdomine, "Points Domin�s par P", "F");

                    leg->Draw();
                    // solution qui domine P
                    if (nflag > 3)
                    {
                        Int_t nColordominant = 5;
                        _ntImageZoneAdmissible->SetMarkerColor(nColordominant);
                        TString sdraw = Form("%s:%s", sy.Data(), sx.Data());
                        TString sselect = Form(" ( %s < %f ) & ( %s < %f ) ",
                                sx.Data(), dx, sy.Data(), dy);
                        _ntImageZoneAdmissible->Draw(sdraw, sselect, "same");
                        _ntImageZoneAdmissible->SetMarkerColor(2);
                        TGraph *grdomine = new TGraph(1, dum, dum);
                        grdomine->SetFillColor(nColordominant);
                        grdomine->SetPoint(0, dx, dy);
                        leg->AddEntry(grdomine, "Points qui dominent P", "F");
                        leg->Draw();
                    }
                }
            }
        }
    }
    return 1;

}

void VizirCommon::optimize(Option_t *option)
{

    if (_blog)
        cout << "VizirCommon::optimize[" << option << "]" << endl;

    // Initialisation du tuple
    createTuple();

    // Cette fonction devrait �tre mise dans une m�thode initPopulation();
    // Ainsi, on pourra faire �voluer une population � partir de n'importe quelle population initiale.
    do
        creation_aleatoire();
    while (evaluation_initiale());
    classe(nbcout);
    saveGeneration();

    // A partir de la pop initiale (diploide : 2 chromosome par individu), c'est le premier qui est le param�tre
    if (_blog)
    {
        cout << "fin de evoluer" << endl;
        pop_val.Print();
        descendant.Print();
        cout << "fin difin fin de evoluer" << endl;
    }

    Int_t nIndexVarDrawX = 0, nIndexVarDrawY = 1;
    TString sVar;
    TString sVarCout;

    if (_sdraw.Length() > 0)
    {
        sVar = _sdraw;
    }
    else
    {
        if (nbgen > 1)
            sVar = _dataSpecification->getAttribute(1)->GetName()
                    + TString(":");
        sVar += _dataSpecification->getAttribute(0)->GetName();
    }
    if (_blog)
        cout << "sVar[" << sVar << "] ******** _sdraw[" << _sdraw << "]"
                << endl;

    TString svardraw(sVar);
    svardraw.ReplaceAll(":", " ");
    svardraw += " ";

    istringstream inStream((char *) svardraw.Data());
    string inName;
    Int_t nInputs = 0;
    while (inStream >> inName)
    {
        URANIE::DataServer::TAttribute *tatt = _dataSpecification->getAttribute(
                TString(inName.c_str()));
        if (tatt)
        {
            Int_t nindex = _dataSpecification->getAttributeIndex(tatt);
            if (_blog)
                cout << nInputs << " -- " << inName << " -- nindex[" << nindex
                        << "]" << endl;
            if (nInputs == 0)
                nIndexVarDrawY = nindex;
            if (nInputs == 1)
                nIndexVarDrawX = nindex;
            nInputs++;
        }
        else
        {
            throw URANIE::Exceptions::UExceptions(__FILE__, __LINE__,
                    Form(
                            "VizirCommon::optimize ERROR : the variable \"%s\" does not exist in the parameter",
                            inName.c_str()));
        }
    }

    if (_blog)
    {
        cout << "nIndexVarDrawX[" << nIndexVarDrawX << "]" << endl;
        cout << "nIndexVarDrawY[" << nIndexVarDrawY << "]" << endl;
    }
    TCanvas *Canvas;
    if (gPad)
        Canvas = (TCanvas *) (gPad);
    else
        Canvas = new TCanvas();
    Canvas->Clear();
    Canvas->Divide(2, 1);

    if (nbcout > 1)
        sVarCout = _coutSpecification->getAttribute(1)->GetName()
                + TString(":");
    sVarCout += _coutSpecification->getAttribute(0)->GetName();
    if (_blog)
        cout << " ** sVarCout[" << sVarCout << "]" << endl;
    TString sOptionZoneAdmissible = "";

    // Boucle
    if (parents())
    {
        do
        {
            _nIter++;
            Int_t nCrt = 0;
            cout << endl << "Generation N� : " << _nIter
                    << " Nombre d'enfants a creer " << N - nbsurv << endl;
            do
            {
                naissance();
            } while (evaluation());

            _datatree->SetMarkerColor(4);
            Canvas->cd(1);
            TString sTitle1 = "Zone de Recherche";
            //drawZoneAdmissible();
            //       NtDraw(_ntZoneAdmissible, _dataSpecification->getAttribute(0)->GetName(), _dataSpecification->getAttribute(1)->GetName(),
            //          , "");
            if (bZoneAdmissible)
            {
                _ntZoneAdmissible->Draw(sVar);
                TH1F *htemp = (TH1F*) gPad->GetPrimitive("htemp");
                TString sTitle1 = htemp->GetTitle();
                htemp->SetTitle(Form("%s : iter %d", sTitle1.Data(), _nIter));
            }
            else
            {
                drawZoneRecherche(nIndexVarDrawX, nIndexVarDrawY);
            }
            gPad->SetTitle(Form("%s : iter %d", sTitle1.Data(), _nIter));
            Draw(sVar, Form("(generation==%d)", _nIter), "same");

            Canvas->cd(2);
            if (bZoneAdmissible)
            {
                drawImageZoneAdmissible();
                Draw(sVarCout, Form("(generation==%d)", _nIter), "same");
                sOptionZoneAdmissible = "same";
            }

            // Draw the graph with the selection
            _datatree->SetMarkerStyle(8);
            _datatree->SetMarkerSize(1.);
            _datatree->SetMarkerColor(1);
            Canvas->cd(1);
            Draw(sVar, Form("(generation==%d) && (rangdomi==0)", _nIter),
                    "same");

            Canvas->cd(2);
            Draw(sVarCout, Form("(generation==%d) && (rangdomi==0)", _nIter),
                    sOptionZoneAdmissible);
            _datatree->SetMarkerSize(0.50);

            Canvas->Modified();
            Canvas->Update();
            printf("\t Nb d'evaluation : %d taille de la Z.P. :  %d\n", nbevalu,
                    _nbelite);

        } while (parents());
//     } while( _nIter < 2 );
    }
    /********************************/
    /*                              */
    /*  Fin de Boucle Principale    */
    /*                              */
    /********************************/

    // Save all the population
//   TFile tf("_vizir_optimizer_all.root", "recreate");
//   _datatree->Write("", TObject::kOverwrite);
//   tf.Close();
    if (bZoneAdmissible)
    {
        Canvas->cd(1);
        drawZonePareto();

        Canvas->cd(2);
        drawFrontPareto();

        Canvas->Modified();
        Canvas->Update();
    }

}

// On construit un nouvel individu � partir de la matrice courante \em descendant.
// 2 types possibles :
//  - 1.  homozygote (chromosome 1 = chromosome2) avec une proba \em phono
// En retour, la matrice courante \em descendant est modifi�e
void VizirCommon::pheno()
{
    Double_t temp;
    Int_t i;
    /* on construit un homozygote */
    if (geneal(1) < phomo)
    {
        for (i = 0; i < nbgen; i++)
        {
            Double_t dval = geneal(1);
            temp = dval * descendant(0, i) + (1.0 - dval) * descendant(1, i);
            descendant(0, i) = descendant(1, i) = temp;
        }
    }
    else
    {
        // On construit un h�t�rozygote
        for (i = 0; i < nbgen; i++)
        {
            if (geneal(1) >= 0.5)
            {
                temp = descendant(0, i);
                descendant(0, i) = descendant(1, i);
                descendant(1, i) = temp;
            }
        }
    }
}

Int_t VizirCommon::parents()
{
    Int_t i;
    Int_t ipareto = 0;
    Int_t cardad = 0;

    switch (nmode)
    {

    case MULTI: // Mode MultiCrit�re
    {
        classe(nbcout);
        for (i = 0; i < N; i++)
        {
//     if ( _blog ) {
//       cout << i << "[" << rangdomi[i] << " - " << ipareto << "]" << endl;
//     }
            if (rangdomi[i] == 0)
                ipareto++;
        }

    }
        break;
    case MONO: // Mode monocrit�re
        /* les individus sont classes directement suivant la valeur */
        /*  de leur fonction objectif                               */
    {
        for (i = 0; i < N; i++)
            rangdomi[i] = pop_val[0][i];
    }
        break;
    }

    //
    tridicho(); /*modifie rangdomi, pop-val,population */
    /*------------- test d'arret  -----------*/
//   if( nmode == 1 )
    _nbelite = ipareto + cardad;

    if (_blog)
    {
        printf("\t fmin : %e, fmax : %e \n", rangdomi[0],
                rangdomi[itaille - 1]);
        cout << "_nbelite[" << _nbelite << "]" << endl;
        cout << "itaille-1[" << itaille - 1 << "]" << endl;
        cout << "rangdomi[itaille-1][" << rangdomi[itaille - 1] << "]" << endl;
        cout << "rangdomi[0][" << rangdomi[0] << "]" << endl;
        cout << "_dprec[" << _dprec << "]" << endl;
    }
    /*------------- test d'arret  -----------*/
    /* la population est optimale ou le nb de calcul max est atteint */
    if ((rangdomi[itaille - 1] - rangdomi[0] < _dprec) || (nbevalu >= maxevalu)
            || (_nbelite >= itaille))
        return (0);
    /* -------------test d'arret   ----------*/

    //    sous_population[nbsspop-2]=_nbelite;
    nbsurv = _nbelite + (int) floor(alpha * (N - _nbelite)); /* nb de survivant pour la generation */
    //    sous_population[nbsspop-1]=nbsurv;
    rgmax = (int) (N - alpha * (N - _nbelite)); /* rang max pour le probleme ne contenant que les contraintes
     qu'un individu conserve ne pourra pas depasse   */
    fitseuil = rangdomi[nbsurv - 1];

//   if ( _nIter > 25 )
//     return (0) ; // arret de principe

    return (1); /* on doit construire une nouvelle generation */

}

void InteractiveFCNCOUT(Double_t *gin, Double_t *f)
{
//*-*-*-*-*-*-*Static function called when SetFCN is called in interactive mode
//*-*          ===============================================
    Int_t ncrit = gVizir->GetNbCrit();
    Int_t ngene = gVizir->GetNbGene();
    Bool_t blog = gVizir->getLog();

    cout << endl << " ** InteractiveFCNCOUT ngene[" << ngene << "] ncrit "
            << ncrit << "]" << endl;
    cout << " ** x[";
    for (Int_t i = 0; i < ngene; i++)
        cout << gin[i] << ", ";
    cout << "]" << endl;

    if (0)
    {
        // Decodage des entr�es
        TMethodCall *fdecode = gVizir->GetDecodeCall();
        Double_t *par = new Double_t[ngene];
        Long_t args_decode[3];
        args_decode[0] = (Long_t) & ngene;
        args_decode[1] = (Long_t) gin;
        args_decode[2] = (Long_t) par;
        fdecode->SetParamPtrs(args_decode);
        Double_t result;
        fdecode->Execute(result);
        if (blog)
        {
            cout << endl << "InteractiveFCNCOUT[" << ncrit << "] ngene["
                    << ngene << "]";
            cout << "-don[" << gin[0] << " , " << gin[1] << "]";
            cout << "-decodedon[" << par[0] << " , " << par[1] << "]" << endl;
            ;
        }

        delete[] par;
        par = NULL;
    }
    Double_t fval;
    // Boucle sur le calcul des couts
    for (Int_t i = 0; i < ncrit; i++)
    {
        TMethodCall *m = gVizir->GetMethodCallCout(i);
        fval = -1.0;
        if (!m)
            return;
        Long_t args[2];
        args[0] = (Long_t) gin;
        args[1] = (Long_t) & fval;
        m->SetParamPtrs(args);
        Double_t result;
        m->Execute(result);
        cout << " ** Result[" << result << "] ***fval[" << fval << "]" << endl;
        f[i] = fval;
        if (blog)
        {
            cout << "Cout : # " << i << " / " << ncrit;
            cout << " -Res[" << fval << "]";
            cout << endl;
        }
    }
}

void InteractiveFCNMultiCost(Double_t *gin, Double_t *f)
{
//*-*-*-*-*-*-*Static function called when SetFCN is called in interactive mode
//*-*          ===============================================
    Int_t ncrit = gVizir->GetNbCrit();
    Int_t ngene = gVizir->GetNbGene();
    Bool_t blog = gVizir->getLog();
    if (blog)
    {
        cout << endl << " ** InteractiveFCNMultiCost ngene[" << ngene
                << "] ncrit " << ncrit << "]" << endl;
        cout << " ** x[";
        for (Int_t i = 0; i < ngene; i++)
            cout << gin[i] << ", ";
        cout << "]" << endl;
    }

    // Launch the cost function ( multi output)
    TMethodCall *m = gVizir->GetMethodCallCout(0);
    if (!m)
        return;
    Long_t args[2];
    args[0] = (Long_t) gin;
    args[1] = (Long_t) f;
    m->SetParamPtrs(args);
    Double_t result;
    m->Execute(result);
    if (blog)
    {
        cout << endl << " ** y[";
        for (Int_t i = 0; i < ncrit; i++)
            cout << f[i] << ", ";
        cout << "]" << endl;
    }
}

void VizirCommon::addCost(void *fcn)
{
    if (!fcn)
        return;
    if (ncrtnbcout > nbcout)
        return;
    char *funcname = G__p2f2funcname(fcn);
    if (funcname)
    {
        if (_blog)
        {
            cout << "VizirCommon::addCost [" << funcname << "]" << ncrtnbcout
                    << "/" << nbcout << endl;
        }
        fCoutCall[ncrtnbcout] = new TMethodCall();
        fCoutCall[ncrtnbcout]->InitWithPrototype(funcname,
                "Double_t*,Double_t*");
        TAttribute *tcout = new TAttribute(funcname);
	tcout->delShare();
        _coutSpecification->addAttribute(tcout);
    }
    ncrtnbcout++;
    fCout = InteractiveFCNCOUT;
    gVizir = this; //required by InteractiveFCNm
}

void VizirCommon::addCost(void *fcn, const char *svarexp)
{
    if (_blog)
    {
        cout << endl << "****** VizirCommon::addCost" << endl;
        cout << " **   svarexp[" << svarexp << "]" << endl;
    }
    if (!fcn)
        return;
    TString sListOfCostAttributes(svarexp);
    Int_t ncost = sListOfCostAttributes.CountChar(':') + 1;
    if (ncost > nbcout)
        return;

    char *funcname = G__p2f2funcname(fcn);
    if (funcname)
    {
        if (_blog)
            cout << " ** multi objective VizirCommon::addCost [" << funcname
                    << "]" << ncost << "/" << nbcout << endl;

        fCoutCall[0] = new TMethodCall();
        fCoutCall[0]->InitWithPrototype(funcname, "Double_t*,Double_t*");

        sListOfCostAttributes += TString(" ");
        sListOfCostAttributes.ReplaceAll(":", " ");
        if (_blog)
            cout << " ** sListOfCostAttributes[" << sListOfCostAttributes << "]"
                    << endl;

        istringstream inStream((char *) sListOfCostAttributes.Data());
        string inName;
        while (inStream >> inName)
        {
            TString svar(inName.c_str());
            if (_blog)
                cout << " ** svar[" << svar << "]" << endl;
            _coutSpecification->addAttribute(new TAttribute(svar));
	    _coutSpecification->getAttribute(svar)->delShare();
        }
        if (_blog)
            _coutSpecification->printLog();

    }
    ncrtnbcout += ncost;
    fCout = InteractiveFCNMultiCost;
    gVizir = this; //required by InteractiveFCNm

}

//______________________________________________________________________________
void InteractiveHardCont(Double_t *param, Int_t &ret)
{
    Int_t nhardcont = gVizir->GetNbHardCont();
    Int_t i = 0;
    do
    {
        TMethodCall *m = gVizir->GetHardContCall(i);
        if (!m)
            return;
        Long_t args[2];
        args[0] = (Long_t) param;
        args[1] = (Long_t) & ret;
        m->SetParamPtrs(args);
        Double_t result;
        m->Execute(result);
        i++;
    } while ((i < nhardcont) && (ret == 1));
}

void VizirCommon::addHardConstraint(void *fcn)
{
    if (!fcn)
        return;
    char *funcname = G__p2f2funcname(fcn);
    if (funcname)
    {
        fHardContCall[ncrtnbhardcont] = new TMethodCall();
        fHardContCall[ncrtnbhardcont]->InitWithPrototype(funcname,
                "Double_t*,Int_t&");
    }
    ncrtnbhardcont++;
    fHardCont = InteractiveHardCont;
    gVizir = this; //required by InteractiveFCNm
}
//______________________________________________________________________________
void InteractiveFCNDecode(Int_t &npar, Double_t *descendant, Double_t *param)
{
//*-*-*-*-*-*-*Static function called when FCNDecode is called in interactive mode
//*-*          ===============================================

    TMethodCall *m = gVizir->GetDecodeCall();
    if (!m)
        return;

    Long_t args[3];
    args[0] = (Long_t) & npar;
    args[1] = (Long_t) descendant;
    args[2] = (Long_t) param;
    m->SetParamPtrs(args);
    Double_t result;
    m->Execute(result);
}
void VizirCommon::SetDecode(void *fcn)
{
    //*-*-*-*-*-*-*To set the address of the decode function*-*-*-*-*-*-*-*
    //*-*          ===============================================
    //     this function is called by CINT 
    //*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*

    if (!fcn)
        return;

    char *funcname = G__p2f2funcname(fcn);
    if (funcname)
    {
        fDecodeCall = new TMethodCall();
        fDecodeCall->InitWithPrototype(funcname, "Int_t&,Double_t*,Double_t*");
    }
    fDecode = InteractiveFCNDecode;
    gVizir = this; //required by InteractiveFCNm
}
//______________________________________________________________________________
void InteractiveFitness(Int_t &npar, Double_t *descendant, Double_t &f,
        Double_t *val, void *info)
{
    //*-*-*-*-*-*-*Static function called when SetFCN is called in interactive mode
    //*-*          ===============================================

    TMethodCall *m = gVizir->GetMethodCall();
    if (!m)
        return;

    Long_t args[5];
    args[0] = (Long_t) & npar;
    args[1] = (Long_t) descendant;
    args[2] = (Long_t) & f;
    args[3] = (Long_t) val;
    args[4] = (Long_t) info;
    m->SetParamPtrs(args);
    Double_t result;
    m->Execute(result);
}

void VizirCommon::SetFitness(void *fitness)
{
    //*-*-*-*-*-*-*To set the address of the minimization function*-*-*-*-*-*-*-*
    //*-*          ===============================================
    //     this function is called by CINT instead of the function above
    //*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*

    if (!fitness)
        return;

    char *funcname = G__p2f2funcname(fitness);
    if (funcname)
    {
        fMethodCall = new TMethodCall();
        fMethodCall->InitWithPrototype(funcname,
                "Int_t &,Double_t*,Double_t&,Double_t*,void*");
    }
    fFitness = InteractiveFitness;
    gVizir = this; //required by InteractiveFCNm
}

void VizirCommon::printLog(Option_t *option)
{
    cout << "*******************************" << endl;
    cout << "** VizirCommon::printLog[" << option << "]  ******" << endl;
    cout << "**   Nom[" << GetName() << "] Title[" << GetTitle() << "]" << endl;
    cout << "**   mode[" << nmode << "]" << endl;
    cout << "**   nbgen[" << nbgen << "]" << endl;
    cout << "**   nbcout[" << nbcout << "]" << endl;
    cout << "**   nbcont[" << nbcont << "]" << endl;
    cout << "**   nbevalu[" << nbevalu << "]" << endl;
    cout << "**   maxevalu[" << maxevalu << "]" << endl;
    cout << "**   itaille[" << itaille << "]" << endl;
    cout << "**   N[" << N << "]" << endl;
    cout << "**   consang[" << consang << "]" << endl;
    if (_dataSpecification)
    {
        cout << "**   DataSpecification" << endl;
        _dataSpecification->printLog();
    }
    if (_coutSpecification)
    {
        cout << "**   CoutSpecification" << endl;
        _coutSpecification->printLog();
    }
    if (_datatree)
    {
        cout << "**   Tree" << endl;
        _datatree->Print();
        _datatree->ls();
    }
    cout << "*******************************" << endl;

}

Int_t VizirCommon::getNbZonePareto()
{
    _datatree->Draw(_dataSpecification->getAttribute(0)->GetName(),
            Form("(generation==%d) && (rangdomi==0)", _nIter), "goff");
    return (Int_t) _datatree->GetSelectedRows();
}

Double_t *VizirCommon::getZonePareto(TString sX)
{
    _datatree->Draw(sX, Form("(generation==%d) && (rangdomi==0)", _nIter),
            "goff");
    Int_t nPoints = (Int_t) _datatree->GetSelectedRows();
    Double_t *ZP = new Double_t[nPoints];
    ZP = _datatree->GetV1();
    Double_t *ZPx = new Double_t[nPoints];

    Int_t *index = new Int_t[nPoints];
    TMath::Sort(nPoints, _datatree->GetV1(), index, kFALSE);
    for (Int_t nIndex = 0; nIndex < nPoints; nIndex++)
    {
        ZPx[nIndex] = ZP[index[nIndex]];
    }

    delete[] ZP;
    ZP = NULL;
    delete[] index;
    index = NULL;

    return ZPx;
}

void VizirCommon::initDefault()
{

    /* Mise � zero des compteurs de base */
    ipop = 0;
    nbevalu = 0;

    nCrtInp = 0;
    ncrtnbcout = 0;
    ncrtnbhardcont = 0;
    /* valeur par defaut */

    alpha = 0.2;
    phomo = 0.5;
    pmut = 0.01;
    bias = 0.5;
    prop = 1.01;
    setTaille(100000, 100);
}

void VizirCommon::setTauxSurvie(Double_t val)
{
    if (val > 0. && val < 1.)
        alpha = val;
}
void VizirCommon::setHomozigotProb(Double_t val)
{
    if (val > 0. && val < 1.)
        phomo = val;
}
void VizirCommon::setMutationProb(Double_t val)
{
    if (val > 0. && val < 1.)
        pmut = val;
}
void VizirCommon::setNaissanceBiais(Double_t val)
{
    if (val > 0. && val < 1.)
        bias = val;
}
void VizirCommon::setOutlierFacteur(Double_t val)
{
    if (val > 1.)
        prop = val;
}

void VizirCommon::setTaille(Int_t taille_pop)
{
    setTaille(100000, taille_pop);
}
void VizirCommon::setTaille(Int_t max_eval, Int_t taille_pop)
{
    maxevalu = max_eval;
    itaille = taille_pop;
    N = (Int_t)(prop * taille_pop);
    consang = (Int_t) floor(alpha * N);

    // Allocation des vecteurs
    rangdomi.ResizeTo(N);

    // Allocation des matrices
    descendant.ResizeTo(2, nbgen);
    pop_val.ResizeTo(nbcout + nbcont, N);
    descendant_val.ResizeTo(nbcout + nbcont, N);
    int i, j; // Juste pour la d�finition de la macro NEWmatrice3D
    NEWmatrice3D(population, N, 2, nbgen);

    // Initialisation du seed
    geneal(0.0);

}

////////////////////////////////////////////////////////
//
// D�but des fonctions privates
//
////////////////////////////////////////////////////////

void VizirCommon::creation_aleatoire()
{
    for (Int_t i = 0; i < nbgen; i++)
    {
        descendant(0, i) = geneal(1.0); // 2.0 + i + 
        descendant(1, i) = geneal(1.0); // 3.0 + i + 
    }
    if (_blog)
    {
        cout << "creation_aleatoire" << endl;
        descendant.Print("creation_aleatoire");
    }

}

double VizirCommon::geneal(double valeur)
{
    int j;
    static int IX1, IX2, IX3;
    double trans;
    static double r[NL];
    long M1 = 259200, M2 = 134456, M3 = 243000;
    long IA1 = 7141, IA2 = 8121, IA3 = 4561;
    long IC1 = 54773, IC2 = 28411, IC3 = 51349;
    double RM1 = 3.8580247e-6, RM2 = 7.4373773e-6;
    char line[30];

    /*initialisation du generateur: pour valeur=0.0*/
    if (valeur == 0.0)
    {
        srand((unsigned) time(NULL));
        sprintf(line, "%d", rand());
        sscanf(line, "%3d", &IX1);

        /*calcul du tableau r[j] de nombres aleatoires*/
        IX1 = (IC1 - IX1) % M1;
        IX1 = (IA1 * IX1 + IC1) % M1;
        IX2 = IX1 % M2;
        IX3 = IX1 % M3;
        for (j = 0; j < NL; j++)
        {
            IX1 = (IA1 * IX1 + IC1) % M1;
            IX2 = (IA2 * IX2 + IC2) % M2;
            r[j] = (IX1 + IX2 * RM2) * RM1 * 1.0e6;
        }
    }

    /*determination de l'indice j du nombre aleatoire*/
    do
    {
        IX3 = (IA3 * IX3 + IC3) % M3;
        j = 1 + (NL * IX3) / M3;
    } while (j >= NL || j < 0);

    /*transmission du nombre aleatoire a la fonction appelante*/
    trans = r[j] * 1.0e-6 * valeur;

    /*calcul du nouveau nombre aleatoire d'indice j*/
    IX1 = (IA1 * IX1 + IC1) % M1;
    IX2 = (IA2 * IX2 + IC2) % M2;
    r[j] = (IX1 + IX2 * RM2) * RM1 * 1.0e6;

    return (trans);
}

int VizirCommon::evaluation_initiale()
{

    Int_t i;
    nbevalu++;
    if (_blog)
        cout << "VizirCommon::evaluation_initiale nbevalu[" << nbevalu << "]"
                << endl;
    Int_t ncrits = nbcout + nbcont;
    if (_blog)
        cout << "VizirCommon::evaluation_initiale ncrits[" << ncrits << "]"
                << endl;
    //   if ( nbevalu > 20 )
    //     return 0;
    Double_t *valCriteria = new Double_t[ncrits];
    if (_blog)
    {
        cout << "valCriteria" << endl;
    }
    TVectorD par(nbgen);
    if (_blog)
    {
        cout << "Par vector" << endl;
        par.Print();
    }

    // a ce jour, je suis seulement en multicrit�re (01/04/2005)
    // donc aucun test si le point v�rifie les contraintes dites hards
    // --> Le test des contraintes hard est ok (le Wed Apr 27 23:32:26 2005)
    // L'individu sur lequel est �valu� les crit�res est le premier chromosome de \em descendant
    TMatrixDRow ind = TMatrixDRow(descendant, 0);

    if (fDecode)
    {
        (*fDecode)(nbgen, ind.GetPtr(), par.GetMatrixArray());
    }
    else
    {
        DecodeDefault(ind.GetPtr(), par.GetMatrixArray());
    }
    if (_blog)
    {
        cout << "Decode vector" << endl;
        par.Print();
    }
    // Test d'admissibilit� de la solution
    Int_t nAdmissible = 1;
    if (fHardCont)
        (*fHardCont)(par.GetMatrixArray(), nAdmissible);

    if (nAdmissible == 1)
    {

        // Calcul des couts et des contraintes
        if (fCout)
            (*fCout)(par.GetMatrixArray(), valCriteria);

        if (_blog)
        {
            cout << endl << "nadmis[ " << nadmis << "] nbevalu[" << nbevalu
                    << "] crits[" << ncrits << "]";
            for (i = 0; i < ncrits; i++)
                cout << " ***** cout[" << i + 1 << "][" << valCriteria[i]
                        << "] - ";
            cout << " *****" << endl;
        }

        nadmis++;
        for (i = 0; i < nbgen; i++)
        {
            population[ipop][0][i] = descendant[0][i];
            population[ipop][1][i] = descendant[1][i];
        }
        for (i = 0; i < ncrits; i++)
            pop_val(i, ipop) = valCriteria[i];
        ipop++;
        if (ipop >= N)
        {
            ipop = 0;
            rgmax = (Int_t)(N - alpha * N);
            return (0);
        }
    }

    // Clean
    delete[] valCriteria;
    valCriteria = NULL;

    return 1;
}

void VizirCommon::classe(Int_t ncri)
{
    // Initialisation � 0 du vecteur
    rangdomi.Zero();

    for (Int_t i = 1; i < N; i++)
    {
        for (Int_t j = 0; j < i; j++)
        {
            Int_t h = 0;
            Int_t c = 1;
            for (Int_t k = 0; k < ncri; k++)
            {
                if (pop_val[k][i] > pop_val[k][j])
                    c = 0;
                else if (pop_val[k][i] < pop_val[k][j])
                    c = cmax;
                else if (pop_val[k][i] == pop_val[k][j])
                    c = 1;
                h = h + c;
            }
            if (h < ncri)
                rangdomi[i]++;
            else
            {
                for (Int_t k = 1; k <= ncri; k++)
                {
                    if (h == (cmax - 1) * k + ncri)
                        rangdomi[j]++;
                }
            }
        }
    }
}

void VizirCommon::tridicho()
{
    Int_t nbcrit = nbcout + nbcont;
    Double_t temp;

    for (Int_t h = N / 2; h > 0; h /= 2)
        for (Int_t u = h; u < N; u++)
            for (Int_t v = u - h; v >= 0 && rangdomi[v] > rangdomi[v + h]; v -=
                    h)
            {
                temp = rangdomi[v];
                rangdomi[v] = rangdomi[v + h];
                rangdomi[v + h] = temp;
                for (Int_t k = 0; k < nbcrit; k++)
                {
                    temp = pop_val[k][v];
                    pop_val[k][v] = pop_val[k][v + h];
                    pop_val[k][v + h] = temp;
                }
                for (Int_t i = 0; i < 2; i++)
                    for (Int_t k = 0; k < nbgen; k++)
                    {
                        temp = population[v][i][k];
                        population[v][i][k] = population[v + h][i][k];
                        population[v + h][i][k] = temp;
                    }
            }
}

void VizirCommon::naissance()
{
    Int_t ichoix, parent1, parent2;
    /* le descendant sera issu d'une mutation ou d'un croisement */
    /*************************************************************/
    /* mutation : l'op�ration de mutation est la creation aleatoire d'un individu */
    if (geneal(1.) < pmut)
    {
        creation_aleatoire();
    }
    else
    {
        /*************************************************************/
        /* croissement - choix de la sous-population : 
         soit un element de l'elite (s'il y en a assez)
         plus bias est grand plus on choisit les parents dans les elites 
         soit un quelconque des surbvivants */
        ichoix = nbsurv;
        /*   if(geneal(1)>bias||_nbelite>consang) ichoix=_nbelite;  */
        if (geneal(1) < bias && _nbelite > consang)
            ichoix = _nbelite;
        parent1 = (Int_t)(floor(geneal(ichoix) + 0.5));
        do
        { /* strategie implantee : un des parents est tjs pris parmi les survivants */
            parent2 = (Int_t)(floor(geneal(nbsurv) + 0.5));
//       cout << "parent2[" << parent2 << "] parent1[" << parent1 << "]" << endl;
        } while (parent1 == parent2); /* le clonage est interdit !! */
        /* tirage des alleles de l'enfant */
        for (Int_t i = 0; i < nbgen; i++)
        {
            descendant[0][i] =
                    population[parent1][(int) floor(geneal(1) + 0.5)][i];
            descendant[1][i] =
                    population[parent2][(int) floor(geneal(1) + 0.5)][i];
        }
        /* si le descendant est homozygote descendant sera modifie lors du calcul de son phenotype   */
    }

}

int VizirCommon::evaluation()
{
    Double_t fitval;
    pheno();
    nbevalu++;

    Int_t ncrits = nbcout + nbcont;
    Double_t *crits = new Double_t[ncrits];
    TVectorD par;
    par.ResizeTo(nbgen);

    TMatrixDRow ind = TMatrixDRow(descendant, 0);

    if (fDecode)
    {
        (*fDecode)(nbgen, ind.GetPtr(), par.GetMatrixArray());
    }
    else
    {
        DecodeDefault(ind.GetPtr(), par.GetMatrixArray());
    }

    // Test d'admissibilit� de la solution
    Int_t nAdmissible = 1;
    if (fHardCont)
        (*fHardCont)(par.GetMatrixArray(), nAdmissible);

    if (nAdmissible == 1)
    {

        // Calcul des couts et des contraintes
        if (fCout)
            (*fCout)(par.GetMatrixArray(), crits);

        nadmis++;
        switch (nmode)
        {
        case MULTI: // Mode MultiCrit�re
        {
            fitval = (double) rang(crits, pop_val, nbcout);
        }
            break;
        case MONO: // Mode Monocrit�re
        {
            fitval = crits[0];
        }
            break;
        }

        // tritement 
        if (fitval <= fitseuil)
            if (stock(crits))
                return (0);

    }
    return (1);

}

int VizirCommon::rang(double *valcour, TMatrixD valcomp, Int_t ncri)
{
    Int_t d = 0;
    if ( cmax < ncri)
        printf(
                "******************** \n  ATTENTION : modifier la valeur de cmax dans rang (de multicritere.c) \n");
    for (Int_t j = 0; j < N; j++)
    {
        Int_t h = 0;
        Int_t c = 1;
        for (Int_t k = 0; k < ncri; k++)
        {
            if (valcour[k] > valcomp[k][j])
                c = 0;
            else if (valcour[k] < valcomp[k][j])
                c = cmax;
            else if (valcour[k] == valcomp[k][j])
                c = 1;
            h = h + c;
        }
        if (h < ncri)
            d++;
    }
    return (d);
}

int VizirCommon::stock(double *valcour)
{
    /*   ajout du descendant a la population   *******************/
    /************     ATTENTION   ****** **************************/
    /**********   nbsurv est le nombre de survivants **/
    for (Int_t i = 0; i < nbgen; i++)
    {
        population[nbsurv + ipop][0][i] = descendant[0][i];
        population[nbsurv + ipop][1][i] = descendant[1][i];
    }
    /* descendant_val:  tableau temporaire contenant                          */
    /*                  les valeurs des criteres des descendants stockes      */
    for (Int_t i = 0; i < nbcout + nbcont; i++)
        descendant_val[i][ipop] = valcour[i];
    ipop++; /*** nombre d'individu cree pendant la generation **********/
    if (ipop < N - nbsurv)
        return (0); /* la population n'est pas completement reconstruite */
    for (Int_t i = 0; i < N - nbsurv; i++)
        for (Int_t j = 0; j < nbcout + nbcont; j++)
            pop_val[j][nbsurv + i] = descendant_val[j][i];
    saveGeneration();
    ipop = 0;
    return (1);
}

void VizirCommon::addAttribute(TAttribute *tattribut)
{
    nCrtInp++;
    if (nCrtInp > nbgen)
    {
        cout << "***Error[VizirCommon::addAttribute] Trop de param�tres ["
                << nCrtInp << "] pour   " << nbgen
                << " d�finis par le constructeur" << endl;
        exit(0);
    }
    _dataSpecification->addAttribute(tattribut);

}

void VizirCommon::createTuple()
{
    char hname[512];
    TString sLabelVariable = "generation";

    // Les param�tres
    for (Int_t j = 0; j < nbgen; j++)
    {
        sLabelVariable += Form(":%s",
                _dataSpecification->getAttribute(j)->GetName());
    }

    // Les fontions
    for (Int_t i = 0; i < nbcout; i++)
    {
        sLabelVariable += Form(":%s",
                _coutSpecification->getAttribute(i)->GetName());
    }
    // Les contraintes 
    for (Int_t i = 0; i < nbcont; i++)
    {
        sprintf(hname, ":cont%d", i + 1);
        sLabelVariable += hname;
    }
    // les infos
    sLabelVariable += ":rangdomi";

    if (_blog)
    {
        cout << "sLabelVariable[" << sLabelVariable << "]" << endl;
    }
    _datatree = new TNtuple(GetName(), GetTitle(), sLabelVariable);
    _datatree->SetMarkerColor(2);
    _datatree->SetMarkerSize(4);
    _datatree->SetMarkerStyle(7);

}

void VizirCommon::DecodeDefault(Double_t *descendant, Double_t *param)
{
    for (Int_t j = 0; j < nbgen; j++)
    {
        Double_t lower = _dataSpecification->getAttribute(j)->getLowerBound();
        Double_t upper = _dataSpecification->getAttribute(j)->getUpperBound();
        param[j] = descendant[j] * (upper - lower) + lower;
//     if ( _blog ) {
//       cout << "name[" << _dataSpecification->getAttribute(j)->GetName() << "] lower[ " << lower << "] [" << upper << "] val[" << descendant[j] << "] ==>[" << param[j] << "]" << endl;
//     }
    }
}

void VizirCommon::saveGeneration()
{
    Int_t nInfo = _datatree->GetNvar();
    Float_t *dInfo = new Float_t[nInfo];
    //  Double_t gin[nbgen];
    TVectorD gin;
    gin.ResizeTo(nbgen);
    TVectorD par;
    par.ResizeTo(nbgen);

    for (Int_t i = 0; i < N; i++)
    {
        Int_t nIter = 0;
        dInfo[nIter++] = (Float_t) _nIter;

        // Decode
        for (Int_t j = 0; j < nbgen; j++)
        {
            gin[j] = population[i][0][j];
        }
        if (fDecode)
        {
            (*fDecode)(nbgen, gin.GetMatrixArray(), par.GetMatrixArray());
        }
        else
        {
            DecodeDefault(gin.GetMatrixArray(), par.GetMatrixArray());
        }

        for (Int_t j = 0; j < nbgen; j++)
        {
            dInfo[nIter++] = (Float_t) par[j];
        }

        // fonction
        for (Int_t j = 0; j < nbcout; j++)
        {
            dInfo[nIter++] = (Float_t) pop_val(j, i);
        }
        dInfo[nIter++] = (Float_t) rangdomi[i];
        _datatree->Fill(dInfo);
    }
    delete[] dInfo;
    dInfo = NULL;
}

ClassImp (VizirMulti)

VizirMulti::VizirMulti() :
        VizirCommon()
{
}
;
VizirMulti::VizirMulti(int gene, int cout) :
        VizirCommon(MULTI, gene, cout, 0, 0.1)
{
}
;

VizirMulti::VizirMulti(URANIE::DataServer::TDataServer *tds, int cout) :
        VizirCommon(MULTI, tds->getNAttributes(), cout, 0, 0.1)
{

    for (Int_t i = 0; i < nbgen; i++)
    {
        addAttribute(tds->getAttribute(i));
    }

}
;

VizirMulti::VizirMulti(URANIE::DataServer::TDataServer *tds, Int_t nSize,
        int cout) :
        VizirCommon(MULTI, tds->getNAttributes(), cout, 0, 0.1)
{

    for (Int_t i = 0; i < nbgen; i++)
    {
        addAttribute(tds->getAttribute(i));
    }

    setTaille(nSize);

}
;

VizirMulti::VizirMulti(int gene, int cout, int (*fct)(double*, double*, void*),
        void*arg) :
        VizirCommon(MULTI, gene, cout, 0, 0.1, fct, arg)
{
}

ClassImp (VizirMono)

VizirMono::VizirMono() :
        VizirCommon()
{
}
;
VizirMono::VizirMono(int gene) :
        VizirCommon(MONO, gene, 1, 0, 0.1)
{
}
;
