//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.5 $
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TNlopt.h
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date    2012-03-20.16.04.27

 \brief interface algorithme optimisation de la bibliotheque NLOpt
 */

#ifndef __NLOPT__
#define __NLOPT__

#include <vector>
#include <map>

//ROOT
#include "Rtypes.h"

//Uranie
#include "Reoptimizer.h"
#include "TNloptShare.h"

#include "DataServer.h"
#include "TEval.h"
#include "TOptimFit.h"
#include "TFitGetter.h"

struct grad_work {
    vector<URANIE::DataServer::UEntry *> *grad;
    vector<vector<URANIE::DataServer::UEntry *>> *in, *out;
    int count;
};


namespace URANIE
{
namespace Reoptimizer
{

class TNlopt: public TNloptShare
{
protected:
    std::map<int, struct grad_work> _job;
    int _coef;
    std::vector<std::vector<URANIE::DataServer::UEntry *>> _codi, _codo, _codg;

    virtual void build();
    // pseudo runner
    void put(long, std::vector<URANIE::DataServer::UEntry*>*, 
        std::vector<URANIE::DataServer::UEntry*>*, int =0);
    void put(long, std::vector<URANIE::DataServer::UEntry*>*, 
        std::vector<URANIE::DataServer::UEntry*>*,
        std::vector<URANIE::DataServer::UEntry*>*, int =0);
    virtual Bool_t readyToGet();
    // solver
    void gradFromCode(double *code, double *solv);
    // abstract from TNloptShare
    void buildSolver(int siz, struct nlopt_work *);
    void cleanSolver(struct nlopt_work *);
    void put(struct nlopt_work *, Bool_t, const double *);
    virtual int get(long &);
#ifdef UNSHARE
    void solver(struct nlopt_work *, int, double*,
            vector<URANIE::DataServer::UEntry*>&,
            vector<URANIE::DataServer::UEntry*>&,
            vector<URANIE::DataServer::UEntry*>&
        );
    void calcul(struct nlopt_work *, Bool_t, const double *);
    void fillOptimum(struct nlopt_work *);
    void fill(struct nlopt_work *, URANIE::DataServer::TDataServer *tds, int id);
#endif

public:
    class NotASingleEval
    {
     public:
	NotASingleEval(){std::cout<<"[91m <ERROR>"<<std::endl<<" <ERROR> This eval is not working (either not properly defined, or might have crashed)"<<std::endl<<" <ERROR> [0m"<<std::endl;}
    };
    
    TNlopt(URANIE::DataServer::TDataServer *tds, URANIE::Relauncher::TRun *run,
            TNloptSolver *algo);
    TNlopt(URANIE::DataServer::TDataServer *tds, URANIE::Relauncher::TEval *fun,
            TNloptSolver *algo);
    virtual ~TNlopt();

    // initialisation
    virtual void addObjective(URANIE::DataServer::TAttribute *, TOptimFit *f =
            NULL);
    void addEquality(URANIE::DataServer::TAttribute *, TOptimFit *f = NULL);
#ifdef UNSHARE
    void setStartingPoint(double *pt);
    void setStartingPoint(int nb, double *pt);
    void setMaximumEval(int max)
    {
        _max = max;
    }
#endif    
    // callback methode
    int critere(struct nlopt_work*, const double *ini, double *out,
            double *grad);
    int inequality(struct nlopt_work*, const double *in, double *out,
            double *grad);
    int equality(struct nlopt_work*, const double *in, double *out,
            double *grad);
    void run(struct nlopt_work *);
#ifdef UNSHARE
    int _retOptim;
    int getReturnOptim()
	{
	    return _retOptim;
	}

    virtual void solverLoop();
#endif 

    ClassDef(URANIE::Reoptimizer::TNlopt, ID_REOPTIMIZER)
};

double nlopt_crit(unsigned, const double*, double*, void *);
void nlopt_inequal(unsigned, double *, unsigned, const double*, double*, void*);
void nlopt_equal(unsigned, double *, unsigned, const double*, double*, void*);
void nlopt_run(void*);

}
} //namespace
#endif

