//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.7 $
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TNlopt.cxx
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date    2012-03-20.16.04.36

 \brief   interface algorithme optimisation de la bibliotheque NLOpt
 */

#include "TNloptShare.h"

#include <stdlib.h>

#include "nlopt.h"
#include "loop.h"

#include "TMasterVar.h"
#include "TFitGetter.h"

#ifdef WITH_PCL
#define COROUTINE
#endif

#ifdef COROUTINE
#include "pcl.h"

#define StackSIZE 131072
//#define StackSIZE 1048576
#endif

#define SUIVI
#define IsRUNNING 16384

#define EPS 1.0e-5
#define ZERO 1.e-20

ClassImp (URANIE::Reoptimizer::TNloptSolver);
ClassImp (URANIE::Reoptimizer::TNloptShare);

#include "nlopt_work.h"

namespace URANIE
{
namespace Reoptimizer
{

//TNloptShare
TNloptShare::TNloptShare(URANIE::DataServer::TDataServer *tds,
        URANIE::Relauncher::TRun *Run, TNloptSolver *solv) :
    TOptimShare(tds, Run),  _solv(solv), _retOptim(-99)
{
    create();
}

TNloptShare::TNloptShare(URANIE::DataServer::TDataServer *tds,
        URANIE::Relauncher::TEval *fun, TNloptSolver *solv) :
    TOptimShare(tds, fun),  _solv(solv), _retOptim(-99)
{
    create();
}

TNloptShare::~TNloptShare()
{
    TFitGetter *one;
    TIter next(&_equal);

    while ((one = (TFitGetter *) next()))
        delete one;
    delete _crit;
    _crit = NULL;
    for (size_t i = 0; i < _guess.size(); i++)
    {
        delete[] _guess[i];
        //_guess.erase(_guess.begin()+i);
    }
}

/* constructeur helpers */
void TNloptShare::create()
{
    //_coef = 0;
    /* defaut */
    _max = 10000;
    /* blank */
    _crit = NULL;
    /* init */
    _id = 0;
}

void TNloptShare::build()
{
    fillDS();
    _tds->createTuple();
    if (_tdstrc) initTrace();
    //_coef = _NinCod * 2 + 1;
}

/* objective and constraints */
#ifdef UNSHARE
void TNloptShare::addObjective(URANIE::DataServer::TAttribute *obj, TOptimFit *fit)
{
    if (!_crit)
        _crit = new TFitGetter(obj, _fun->getOutputList(), fit);
    else
        std::cout << "NLOpt is single objective\n";
}

void TNloptShare::addEquality(URANIE::DataServer::TAttribute *obj, TOptimFit *fit)
{
    TFitGetter *one = new TFitGetter(obj, _fun->getOutputList(), fit);
    _equal.Add(one);
}
#endif

void TNloptShare::setStartingPoint(double *pt)
{
    try{
        throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                                                     "TNloptShare::setStartingPoint WARNING : this function with only a (double*) signature is not recommended anymore.\nPlease, use the setStartingPoint(int nb, double *p) that check the size of the provided array and check that it matches the number of inputs.\nFor C++ you can do: vector<double> p{0.2,0.3}; opt.setStartingPoint(p.size(),&p[0]);.\nFor python you can use numpy array to do: p=numpy.array([0.2,0.3]); opt.setStartingPoint(len(p),p);");
    }catch(URANIE::Exceptions::UWarningExceptions &ue)
    {
	ue.printMessage();
    }

    double* copy = new double[_NinDat];

    for (int i = 0; i < _NinDat; i++)
        copy[i] = pt[i];

    _guess.push_back(copy);
}

void TNloptShare::setStartingPoint(int nb, double *pt)
{
    try{
        if(nb != _NinDat)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TNloptShare::setStartingPoint ERROR : the starting point provided as dimension [%d], while the code as [%d] inputs",
                                                            nb,_NinDat));
    }catch(URANIE::Exceptions::UErrorExceptions &ue)
    {
	ue.printMessage();
	throw ue;
    }
    
    double* copy = new double[_NinDat];

    for (int i = 0; i < _NinDat; i++)
        copy[i] = pt[i];

    _guess.push_back(copy);
}

/* nlopt callback */
double nlopt_crit(unsigned n, const double *in, double *grad, void *info)
{
    UNUSED(n);
    double out;
    struct nlopt_work *work = (struct nlopt_work *) info;

    work->self->critere(work, in, &out, grad);
    return out;
}

void nlopt_inequal(unsigned m, double *out, unsigned n, const double *in,
        double *grad, void *info)
{
    UNUSED(n,m);
    struct nlopt_work *work = (struct nlopt_work *) info;

    work->self->inequality(work, in, out, grad);
}

void nlopt_equal(unsigned m, double *out, unsigned n, const double *in,
        double *grad, void *info)
{
    UNUSED(n,m);
    struct nlopt_work *work = (struct nlopt_work *) info;

    work->self->equality(work, in, out, grad);
}

/* coroutine */
void nlopt_run(void * info)
{
    struct nlopt_work *work = (struct nlopt_work*) info;

    work->self->run(work);
}

/* callback associated method */
void TNloptShare::calcul(struct nlopt_work *self, Bool_t grad, const double *in)
{
#ifdef SUIVI
    static int a = 0;
#endif
    int i, j;
    long id;
	id = 0;

    for EACH(i, _NinDat) if (in[i] != self->last[i]) break;

    //if (i != _NinDat || (self->lastg == kFALSE && grad == kTRUE))
    if (i != _NinDat || self->lastg != grad)
    {
#ifdef SUIVI
        if (grad)
        {
            if (!(a % 10))
                printf("!");
            else if (!(a % 5))
                printf(";");
            else
                printf(",");
        }
        else
        {
            if (!(a % 10))
                printf("|");
            else if (!(a % 5))
                printf(":");
            else
                printf(".");
        }        
        if (!(++a % 50))
            printf("%04d\n", a);
        fflush(stdout);
#endif
        for EACH(j, _NinDat) self->last[j] = in[j];
        put(self, grad, in);
        self->lastg = grad;
#ifdef COROUTINE
        co_resume();
#else
        self->noerr = get(id);
#endif
        if (_tdstrc) {
            fill(self, _tdstrc, _idtrc);
            _idtrc++;
        }
    }
}

#ifdef UNSHARE
void TNloptShare::gradFromCode(double *code, double *solv)
{
    URANIE::Relauncher::TMasterVar *var;
    TIter next(&_param);
    int j = 0;

    while ((var = (URANIE::Relauncher::TMasterVar *) next()))
    {
        var->get(j, solv, code);
        j++;
    }
}

int TNloptShare::critere(struct nlopt_work *work, const double *in, double *out,
        double *grad)
{
    if (grad) {
        double* tmp = new double[_NinCod];

        calcul(work, kTRUE, in);
        //out[0] = _crit->from(_NinCod, work->work + _NinCod, work->grad, tmp);
        out[0] = _crit->from(_NinCod, work->codo, work->codg, tmp);
        gradFromCode(tmp, grad);
	    delete[] tmp;
	}
    else {
        calcul(work, kFALSE, in);
        out[0] = _crit->from(work->codo);
	}
    return work->noerr;
}

int TNloptShare::inequality(struct nlopt_work *work, const double *in, double *out,
        double *grad)
{
    TFitGetter *one;
    TIter next(&_const);
    int i = 0;

    if (grad) {
        double* tmp = new double[_NinCod];

        calcul(work, kTRUE, in);
        while ((one = (TFitGetter *) next()))
        {
            //out[i] = one->from(_NinCod, work->work + _NinCod, work->grad, tmp);
            out[i] = one->from(_NinCod, work->codo, work->codg, tmp);
            gradFromCode(tmp, grad + i * _NinDat);
            i++;
        }
	    delete[] tmp;
	}
    else {
        calcul(work, kFALSE, in);
        while ((one = (TFitGetter *) next()))
            out[i++] = one->from(work->codo);
    }
    return work->noerr;
}

int TNloptShare::equality(struct nlopt_work *work, const double *in, double *out,
        double *grad)
{
    TFitGetter *one;
    TIter next(&_equal);
    int i = 0;

    if (grad) {
        double* tmp = new double[_NinCod];

        calcul(work, kTRUE, in);
        while ((one = (TFitGetter *) next()))
        {
            //out[i] = one->from(_NinCod, work->work + _NinCod, work->grad, tmp);
            out[i] = one->from(_NinCod, work->codo, work->codg, tmp);
            gradFromCode(tmp, grad + i * _NinDat);
            i++;
        }
	    delete[] tmp;
	}
	else {
        calcul(work, kFALSE, in);
        while ((one = (TFitGetter *) next()))
            out[i++] = one->from(work->codo);
	}
    return work->noerr;
}

/* runner indirection */
void TNloptShare::put(struct nlopt_work *self, Bool_t grad, const double *in)
{
    if (grad) {
        for (auto val: *self->codo ) val->clearEntry();
        for (auto val: *self->codg ) val->clearEntry();
        inputForCode(in, self->codi);
        put(self->id, self->codi, self->codo, self->codg);
    }
    else {
        for (auto val: *self->codo ) val->clearEntry();
        inputForCode(in, self->codi);
        put(self->id, self->codi, self->codo);
    }
}

void TNloptShare::put(long id, 
    vector<URANIE::DataServer::UEntry *> *in, 
    vector<URANIE::DataServer::UEntry *> *out, int ncalc)
{
    struct grad_work job =
    { NULL, NULL, NULL, 1 };

    _job[id] = job;
    _run->put(id * _coef, in, out, ncalc);
}

void TNloptShare::put(long id, 
    vector<URANIE::DataServer::UEntry *> *in, 
    vector<URANIE::DataServer::UEntry *> *out, 
    vector<URANIE::DataServer::UEntry *> *grad, int ncalc)
{
    int i, j;
    struct grad_work job =
    { grad, NULL, NULL, _coef };

    job.in = new vector<vector<URANIE::DataServer::UEntry *>>(); //new double[2 * _NinCod * _NinCod];
    job.in->resize(2 * _NinCod);
    allocInCodeEntries(2 * _NinCod, *job.in);
    job.out = new vector<vector<URANIE::DataServer::UEntry *>>(); //new double[2 * _NinCod * _NoutCod];
    job.out->resize(2 * _NinCod);
    allocOutCodeEntries(2 * _NinCod, *job.out);

    _job[id] = job;

    /* generation plan OAT */
    /* copie conforme */
    for (i = 0; i < 2 * _NinCod; i++)
        for (j = 0; j < _NinCod; j++)
            (*job.in)[i][j]->copyEntry( (*in)[j]);
    /* OAT diff */
    for (i = 0; i < _NinCod; i++)
    {
        if (fabs(*(*job.in)[2 * i][i]->val.dou) < ZERO)
        {
            (*job.in)[2 * i][i]->pushEntry( ZERO);
            (*job.in)[2 * i + 1][i]->pushEntry( -ZERO);
        }
        else
        {
            (*job.in)[2 * i][i]->pushEntry( *(*in)[i]->val.dou * (1. + EPS));
            (*job.in)[2 * i + 1][i]->pushEntry( *(*in)[i]->val.dou * (1. - EPS));
        }
    }
#ifdef TRACE
    for (i=0; i<2*_NinCod; i++)
    {
        for (j=0; j<_NinCod; j++)
        std::cout << (*job.in)[i][j]->val.dou << " ";
        std::cout << std::endl;
    }
#endif

    /* submit des jobs */
    _run->put(id * _coef, in, out, ncalc);
    for (i = 0; i < 2 * _NinCod; i++)
        _run->put(id * _coef + i + 1, &(*job.in)[i], &(*job.out)[i], ncalc);
    
}

Bool_t TNloptShare::readyToGet()
{   
    Bool_t ready;
    int ret;
    long id, gid;
    
    // TODO
    while ((ready =_run->readyToGet())) {
        ret = get(gid);
        if (ret == 0)
            throw new NotASingleEval();
        id = gid / _coef;
        _job[id].count--;
        if (!_job[id].count) break;
    }
    
    return ready;
}

int TNloptShare::get(long &id)
{
    int ret, i, j;
    long gid;

    // if readyToGet called
    for (const auto ii : _job) {
        id = ii.first;
        if (ii.second.count == 0) break;
    }
    while (_job[id].count) {
        ret = _run->get(gid);
        if (ret == 0)
            throw new NotASingleEval();
        id = gid / _coef;
        _job[id].count--;
    }
    
    if (_job[id].grad)
    {
        /* gradient estimation */
        for (i = 0; i < _NoutCod; i++)
        {
            if ((*_job[id].grad)[i]->tag == URANIE::DataServer::UEntry::Vector) { 
                auto tmp = new vector<double>();
                tmp->resize(_NinCod);
                for (j = 0; j < _NinCod; j++)
                {
                    (*tmp)[j] = 
                        ( *(*_job[id].out)[2 * j][i]->val.dou - 
                          *(*_job[id].out)[2 * j + 1][i]->val.dou )
                        / (*(*_job[id].in)[2 * j][j]->val.dou -
                          *(*_job[id].in)[2 * j + 1][j]->val.dou );
                }
                (*_job[id].grad)[i]->assignEntry (tmp);
            }
        }
        /* temp data cleanup */
        for (i=0; i<2*_NinCod; i++) {
            for (auto it : (*_job[id].in)[i]) delete it;
            (*_job[id].in)[i].clear();
            for (auto it : (*_job[id].out)[i]) delete it;
            (*_job[id].out)[i].clear();
        }
        delete _job[id].in;
        delete _job[id].out;
    }
    _job.erase(id);

    return 1;
}
#endif

/* solve */
void TNloptShare::solver(struct nlopt_work *ret, int id, double *guess,
        vector<URANIE::DataServer::UEntry*>& codi,
        vector<URANIE::DataServer::UEntry*>& codo,
        vector<URANIE::DataServer::UEntry*>& codg
    )
{
    nlopt_opt solv, subsolv;
    long i;
    unsigned nineq = _const.GetEntries();
    unsigned nequ = _equal.GetEntries();
    double* top = new double[_NinDat];
    double* low = new double[_NinDat];

    //std::cout << _NinDat << " " << nineq << " " << nequ << std::endl;
    // solver
    solv = _solv->build(_NinDat, nineq, nequ, _prec, &subsolv);
    // bounds
    TList *inputs = _tds->getListOfAttributes();
    for EACH(i, _NinDat)
    {
        URANIE::DataServer::TAttribute *att = (URANIE::DataServer::TAttribute *) inputs->At(i);
        //std::cout << att->GetName() << ": lower = " << att->getLowerBound() << ", upper = " << att->getUpperBound() << std::endl;
        low[i] = att->getLowerBound();
        top[i] = att->getUpperBound();
    }
    delete inputs;
    nlopt_set_lower_bounds(solv, low);
    nlopt_set_upper_bounds(solv, top);

    // fun
    nlopt_set_min_objective(solv, nlopt_crit, (void *) ret);
    if (nineq)
    {
        double* tol = new double[nineq];
        for EACH(i, nineq) tol[i] = 1e-7;
        nlopt_add_inequality_mconstraint(solv, nineq, nlopt_inequal,
                (void *) ret, tol);
        delete[] tol;
    }
    if (nequ)
    {
        double* tol = new double[nequ];
        for EACH(i, nequ) tol[i] = 1e-4;
        nlopt_add_equality_mconstraint(solv, nequ, nlopt_equal, (void *) ret,
                tol);
        delete[] tol;
    }
    // stop
    nlopt_set_maxeval(solv, _max);

    ret->self = this;
    ret->state = IsRUNNING;
    ret->solv = solv;
    ret->subsolv = subsolv;
    ret->id = id;
    ret->guess = guess;
    ret->last = (double *) calloc(_NinDat, sizeof(double));
    ret->work = (double *) calloc(_NinCod + _NoutCod, sizeof(double));
    ret->grad = (double *) calloc(_NinCod * _NoutCod, sizeof(double));
    ret->codo = &codo;
    ret->codi = &codi;
    ret->codg = &codg;
    ret->lastg = kFALSE;
            
	
	delete[] top;
	delete[] low;
}

#ifdef SHARE
void TNloptShare::buildSolver(int siz, struct nlopt_work *opt)
{
    int i;
    
    // data
    _codi.resize(siz);
    allocInCodeEntries(siz, _codi);
    _codo.resize(siz);
    allocOutCodeEntries(siz, _codo);
    _codg.resize(siz);
    allocGradCodeEntries(siz, _codg);

    // solver
    for (i=0; i<siz; i++)
    {
        solver(&(opt[i]), i, _guess[i], _codi[i], _codo[i], _codg[i]);
    }
    
}

void TNloptShare::cleanSolver(struct nlopt_work *opt)
{
    free(opt->last);
    free(opt->work);
    free(opt->grad);
    for (auto it : *opt->codi) delete it;
    opt->codi->clear();
    for (auto it : *opt->codo) delete it;
    opt->codo->clear();
    for (auto it : *opt->codg) delete it;
    opt->codg->clear();
    nlopt_destroy(opt->solv);
    if (opt->subsolv) nlopt_destroy(opt->subsolv);
}
#endif

void TNloptShare::run(struct nlopt_work *opt)
{
    int ret;

    ret = nlopt_optimize(opt->solv, opt->guess, &(opt->minf));

    switch (ret)
    {
    case NLOPT_FAILURE:
    case NLOPT_INVALID_ARGS:
    case NLOPT_OUT_OF_MEMORY:
    case NLOPT_FORCED_STOP:
        std::cout << "\n nlopt failed " << ret << " !\n";
        break;
    case NLOPT_ROUNDOFF_LIMITED:
        std::cout << "\n roundoff\n";
        fillOptimum(opt);
        break;
    case NLOPT_MAXEVAL_REACHED:
        std::cout << "\n max eval reached\n";
        fillOptimum(opt);
        break;
    default:
        std::cout << std::endl;
        fillOptimum(opt);
    }
    opt->state = ret;
    _retOptim = ret;

}

void TNloptShare::solverLoop()
{
    int i;

    // verif
    if (!_guess.size())
    {
        std::cout << "Need a Starting Point\n";
        return;
    }
    if (!_crit)
    {
        std::cout << "Need an Objective\n";
        return;
    }
    build();

    int siz = _guess.size();
    // solver
    struct nlopt_work * opt = new struct nlopt_work[siz];
    buildSolver(siz, opt);
#ifdef COROUTINE
    long id;
    int compt, ret;
    // coroutine
    char **stack = new char*[siz];
    coroutine_t * fibre = new coroutine_t[siz];
	// coroutine_t * current;

    for (i=0; i<siz; i++)
    {
        stack[i] = (char*) calloc(StackSIZE, sizeof(char));
        fibre[i] = co_create(nlopt_run, &(opt[i]), stack[i], StackSIZE);
        co_call(fibre[i]);
    }
    compt = siz;
    while (compt)
    {
        ret = get(id);
        opt[id].noerr = ret;
        co_call(fibre[id]);
        if (opt[id].state != IsRUNNING)
        {
            compt--;
        }
    }
    for (i=0; i<siz; i++)
    {
        cleanSolver(&opt[i]);
        free(stack[i]);
        //co_delete(fibre[i]);
    }
    delete[] fibre;
    delete[] stack;

#else
    // data
    for (i = 0; i < siz; i++)
    {
        run(&opt[i]);
        cleanSolver(&opt[i]);
    }

#endif
    delete[] opt;
    
    fillConstant();
}

void TNloptShare::fillOptimum(struct nlopt_work *self)
{
    calcul(self, kFALSE, self->guess);
    fill(self, _tds, _id);
    _id++;
}

void TNloptShare::fill(struct nlopt_work *self, URANIE::DataServer::TDataServer *tds, int id)
{
    auto tmpi = new vector<URANIE::DataServer::UEntry*>(self->codi->begin(), self->codi->end());
    auto tmpo = new vector<URANIE::DataServer::UEntry*>(self->codo->begin(), self->codo->end());
    vector<vector<URANIE::DataServer::UEntry*>> all;
    all.resize(1);
    allocDataEntries(1, all);
    auto tmpa = new vector<URANIE::DataServer::UEntry*>(all[0].begin()+1, all[0].end());
    
    all[0][0]->pushEntry((double) id);
    allFromCode(tmpi, tmpo, tmpa);
    tds->getTuple()->Fill(&all.at(0));
	
    for (auto val : all[0]) delete val;
    all[0].clear();
}

}
} //namespace
