////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!\file   TestTNlopt.cxx
 *   \author Gilles Arnaud <gilles.arnaud@cea.fr>
 *   \date   jeu mai 23 13:36:02 GMT 2013  
 * 
 *   \brief  test file for Nlopt classes
 * 
 * */

#include "TestTNlopt.h"

// CPPUNIT
#include <cppunit/extensions/HelperMacros.h>
#include <cppunit/Test.h>

#define DBL_PREC_FINE 5.e-6
#define DBL_PREC_STD  5.e-4
#define DBL_PREC_MED  2.e-3
#define DBL_PREC_LARGE 2.5e-2

// ROOT
#include "Rtypes.h"
//#include "Api.h"

//URANIE
#include "UExceptions.h"
#include "TDataServer.h"
#include "TAttribute.h"
#include "TNormalDistribution.h"

#include "TCJitEval.h"
#include "TComposeEval.h"
//#include "TCIntEval.h"
#include "TSequentialRun.h"
#include "TThreadedRun.h"
#include "TGradientEstimationRun.h"

#include "TNlopt.h"
#include "TNloptGradient.h"
#include "TNloptCobyla.h"
#include "TNloptDirect.h"

#include "uranie_commands.h"
// local
#include "TestUserFunctions.h"

#ifdef WITH_MPI
#include "mpi_commands.h"
#endif

using namespace URANIE::DataServer;
using namespace URANIE::Relauncher;
using namespace URANIE::Reoptimizer;

/* possible solver
 * direct
 #define DIRECT TNloptCobyla
 * lagragien direct
 #define DIRECT TNloptBobyqa //doesn't work on this problem
 #define DIRECT TNloptPraxis
 #define DIRECT TNloptNelderMead
 #define DIRECT TNloptSubplexe
 * gradient
 #define GRADIENT TNloptMMA
 #define GRADIENT TNloptSLSQP
 * lagragien gradient
 #define GRADIENT TNloptLBFGS
 #define GRADIENT TNloptNewton
 #define GRADIENT TNloptVariableMetric
 *
 * */

// Constructor Destructor
TNloptTest::TNloptTest()
{
#ifdef ROOT5
    // creation of CInt function pointers
    G__ClassInfo globalscope;
    G__MethodInfo method;
    long dummy;

    method = globalscope.GetMethod("rosenbrock", "Double_t*,Double_t*", &dummy);
    _p2fRosenbrock = method.PointerToFunc();
#endif
}

TNloptTest::~TNloptTest()
{
    UranieTests::removeFolder(gSystem->pwd(), "URA_*");
}

// pre and post treatemnt
void TNloptTest::setUp()
{
}

void TNloptTest::tearDown()
{
}


// TEST SUITE
/*** rosenbrock problem ***/
/* direct */
void TNloptTest::roseBobyqa()
{
    TNloptBobyqa solv;
    rosenbrock(solv);
}

void TNloptTest::roseSimplex()
{
    TNloptNelderMead solv;
    rosenbrockVec(solv);
}

void TNloptTest::rosePraxis()
{
    TNloptPraxis solv;
    rosenbrock2(solv);
}

/* gradient */
void TNloptTest::roseLBFGS()
{
    TNloptLBFGS solv;
    rosenbrockGrad(solv);
}

void TNloptTest::roseNewton()
{
    TNloptNewton solv;
    rosenbrock(solv);
}

// test suite helpers
void TNloptTest::rosenbrockGrad(TNloptSolver &solv)
{
    TAttribute x0("x0", -4.0, 4.0);
    TAttribute x1("x1", -4.0, 4.0);
    TAttribute x2("x2", -4.0, 4.0);
    TAttribute x3("x3", -4.0, 4.0);
    TAttribute x4("x4", -4.0, 4.0);
    TAttribute x5("x5", -4.0, 4.0);
    TAttribute rose("rose");
    TAttribute drose("drose", URANIE::DataServer::TAttribute::kVector);

    //double start[] = {-0.5, 2., 2., 2., 2., 2.};
    double start[] =
    { -0.5, -0.5, -0.5, -0.5, -0.5, -0.5 };

    TCJitEval code(frosenbrockGrad);
    code.setInputs(6, &x0, &x1, &x2, &x3, &x4, &x5);
    code.setOutputs(2, &rose, &drose);

    TDataServer tds("vizirDemo", "Param de l'opt vizir pour la barre");
    tds.keepFinalTuple(false);
    code.addAllInputs(&tds);

    TNloptGradient opt(&tds, &code, &solv);
    opt.addObjective(&rose, &drose);
    opt.setTolerance(1.e-9);

    /* resolution */
    opt.setStartingPoint(6,start);
    opt.solverLoop();

    /* solution */
    tds.getTuple()->Scan("*");
    validateRosenbrock(tds, 8);

}

void TNloptTest::rosenbrockVec(TNloptSolver &solv)
{
    TAttribute x0("x0", -4.0, 4.0);
    TAttribute x1("x1", -4.0, 4.0);
    TAttribute x2("x2", -4.0, 4.0);
    TAttribute x3("x3", -4.0, 4.0);
    TAttribute x4("x4", -4.0, 4.0);
    TAttribute x5("x5", -4.0, 4.0);
    TAttribute rose("rose");
    TAttribute drose("drose", URANIE::DataServer::TAttribute::kVector);

    //double start[] = {-0.5, 2., 2., 2., 2., 2.};
    double start[] =
    { -0.5, -0.5, -0.5, -0.5, -0.5, -0.5 };

    TCJitEval code(frosenbrockVec);
    code.setInputs(6, &x0, &x1, &x2, &x3, &x4, &x5);
    code.setOutputs(1, &drose);
    
    TCJitEval cum(frosenbrockCummul);
    cum.addInput(&drose);
    cum.addOutput(&rose);
    
    TComposeEval all;
    all.addEval(&code);
    all.addEval(&cum);

    TDataServer tds("vizirDemo", "Param de l'opt vizir pour la barre");
    tds.keepFinalTuple(false);
    code.addAllInputs(&tds);

    TNlopt opt(&tds, &all, &solv);
    opt.addObjective(&rose);
    opt.setTolerance(1.e-9);

    /* resolution */
    opt.setStartingPoint(6,start);
    opt.solverLoop();

    /* solution */
    tds.getTuple()->Scan("*");
    validateRosenbrock(tds, 8);

}

void TNloptTest::rosenbrock(TNloptSolver &solv)
{
    TAttribute x0("x0", -4.0, 4.0);
    TAttribute x1("x1", -4.0, 4.0);
    TAttribute x2("x2", -4.0, 4.0);
    TAttribute x3("x3", -4.0, 4.0);
    TAttribute x4("x4", -4.0, 4.0);
    TAttribute x5("x5", -4.0, 4.0);
    TAttribute rose("rose");

    //double start[] = {-0.5, 2., 2., 2., 2., 2.};
    double start[] =
    { -0.5, -0.5, -0.5, -0.5, -0.5, -0.5 };

    TCJitEval code(frosenbrock);
    code.setInputs(6, &x0, &x1, &x2, &x3, &x4, &x5);
    code.setOutputs(1, &rose);

    TDataServer tds("vizirDemo", "Param de l'opt vizir pour la barre");
    tds.keepFinalTuple(false);
    code.addAllInputs(&tds);

    TNlopt opt(&tds, &code, &solv);
    opt.addObjective(&rose);
    opt.setTolerance(1.e-9);

    /* resolution */
    opt.setStartingPoint(6,start);
    opt.solverLoop();

    /* solution */
    tds.getTuple()->Scan("*");
    validateRosenbrock(tds, 7);

}

void TNloptTest::rosenbrock2(TNloptSolver &solv)
{
    TAttribute x0("x0", -4.0, 4.0);
    TAttribute x1("x1", -4.0, 4.0);
    TAttribute x2("x2", -4.0, 4.0);
    TAttribute x3("x3", -4.0, 4.0);
    TAttribute x4("x4", -4.0, 4.0);
    TAttribute x5("x5", -4.0, 4.0);
    TAttribute rose("rose");

    //double start[] = {-0.5, 2., 2., 2., 2., 2.};
    double start[] =
    { -0.5, -0.5, -0.5, -0.5, -0.5, -0.5 };

    TCJitEval code(frosenbrock);
    code.setInputs(6, &x0, &x1, &x2, &x3, &x4, &x5);
    code.setOutputs(1, &rose);

    TSequentialRun run(&code);
    run.startSlave();
    if (run.onMaster()) {
        TDataServer tds("vizirDemo", "Param de l'opt vizir pour la barre");
        tds.keepFinalTuple(false);
        code.addAllInputs(&tds);
        TDataServer trc("NloptTrc", "trace de recherche"); 
        trc.keepFinalTuple(false);
        
        TNlopt opt(&tds, &code, &solv);
        opt.addObjective(&rose);
        opt.setTolerance(1.e-9);
        opt.setTrace(&trc);

        /* resolution */
        opt.setStartingPoint(6,start);
        opt.solverLoop();

        /* solution */
        tds.getTuple()->Scan("*");
        validateRosenbrock(tds, 7);
        validateRosenbrock2(trc);
    }

}

void TNloptTest::validateRosenbrock(TDataServer &tds, int siz)
{
    // only one result
    long nb = tds.getNPatterns();
    CPPUNIT_ASSERT_EQUAL(1L, nb);
    // attributes
    // x0, x1, x2, x3, x4, x5, rose
    CPPUNIT_ASSERT_EQUAL(siz, tds.getNAttributes());
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("x0"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("x3"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("x5"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("rose"));
    // values
    // theo 1, 1, 1, 1, 1, 1, 0
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0., tds.getValue("rose", 0), DBL_PREC_FINE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1., tds.getValue("x0", 0), DBL_PREC_STD);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1., tds.getValue("x5", 0), DBL_PREC_LARGE);

}

void TNloptTest::validateRosenbrock2(TDataServer &tds)
{
    // only one result
    long nb = tds.getNPatterns();
    //CPPUNIT_ASSERT_EQUAL(1L, nb);
    // attributes
    // x0, x1, x2, x3, x4, x5, rose
    CPPUNIT_ASSERT_EQUAL(7, tds.getNAttributes());
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("x1"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("x2"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("x4"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("rose"));
    // values
    // theo 1, 1, 1, 1, 1, 1, 0
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0., tds.getValue("rose", nb-2), DBL_PREC_FINE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1., tds.getValue("x1", nb-2), DBL_PREC_STD);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1., tds.getValue("x4", nb-2), DBL_PREC_LARGE);

}

/*** Constraint Problem ***/
/* direct */
void TNloptTest::constCobyla()
{
    TNloptCobyla solv;
    constraint(solv);
}

/* gradient */
void TNloptTest::constMMA()
{
    TNloptMMA solv;
    constraintGrad(solv);
}

void TNloptTest::constSLSQP()
{
    TNloptSLSQP solv;
    constraintGrad(solv);
}

void TNloptTest::constMMADiff()
{
    TNloptMMA solv;
    constraintPlus(solv);
}

void TNloptTest::constSLSQPDiff()
{
    TNloptSLSQP solv;
    constraint(solv);
}

// direct with lagragien
void TNloptTest::constSubplexe()
{
    TNloptSubplexe solv;
    constraint(solv);
}

// gradient with lagragien
void TNloptTest::constVariableMetric()
{
    TNloptVariableMetric solv;
    constraintGrad(solv);
}

void TNloptTest::constVariableMetricDiff()
{
    TNloptVariableMetric solv;
    constraintPlus(solv);
}

// test suite helpers
void TNloptTest::constraint(TNloptSolver &solv)
{
    /* inputs */
    TAttribute x1("x1", -3.0, 3.0), x2("x2", -3.0, 3.0), x3("x3", -3.0, 3.0);
    /* outputs */
    TAttribute cont1("c1"), cont2("c2"), cont3("c3"), obj("obj");
    /* starting point */
    double start[][3] =
    {
    { 1.5, 1.5, 1.5 },
    { -1.5, -1.5, -1.5 } };

    /* code */
    //TCJitGradEval code(&fselect);
    TCJitEval code(&fselect);
    code.setInputs(3, &x1, &x2, &x3);
    code.setOutputs(4, &obj, &cont1, &cont2, &cont3);

    /* run */
    TThreadedRun run(&code, 3);
    //TSequentialRun run(&code);
    //TMpiRun run(&code);
    run.startSlave();
    if (run.onMaster())
    {
        /* tds */
        TDataServer tds("toto", "tds for vizir test");
        tds.keepFinalTuple(false);
        // reverse output test
        tds.addAttribute(&x3);
        tds.addAttribute(&x2);
        tds.addAttribute(&x1);
        /* optimizer */
        TNlopt nlo(&tds, &run, &solv);
        TGreaterFit it(1.0);
        nlo.addConstraint(&cont3, &it);
        nlo.addConstraint(&cont2, &it);
        nlo.addConstraint(&cont1, &it);
        nlo.addObjective(&obj);

        /* resolution */
        nlo.setStartingPoint(3,start[0]);
        nlo.setStartingPoint(3,start[1]);
        nlo.solverLoop();

        run.stopSlave();
        /* resultats */
        tds.getTuple()->Scan("*");
        validateConstraint(tds, 7);
    }

}

void TNloptTest::constraintGrad(TNloptSolver &solv)
{
    /* inputs */
    TAttribute x1("x1", -3.0, 3.0), x2("x2", -3.0, 3.0), x3("x3", -3.0, 3.0);
    /* outputs */
    TAttribute cont1("c1"), cont2("c2"), cont3("c3"), obj("obj");
    TAttribute dobj("dobj", URANIE::DataServer::TAttribute::kVector);
    TAttribute dcnt1("dc1", URANIE::DataServer::TAttribute::kVector);
    TAttribute dcnt2("dc2", URANIE::DataServer::TAttribute::kVector);
    TAttribute dcnt3("dc3", URANIE::DataServer::TAttribute::kVector);
    /* starting point */
    double start[][3] =
    {
    { 1.5, 1.5, 1.5 },
    { -1.5, -1.5, -1.5 } };

    /* code */
    TCJitEval code(&fselectd);
    code.setInputs(3, &x1, &x2, &x3);
    code.setOutputs(8, &obj, &dobj, &cont1, &dcnt1, &cont2, &dcnt2,
        &cont3, &dcnt3);

    /* run */
    TThreadedRun run(&code, 3);
    //TSequentialRun run(&code);
    //TMpiRun run(&code);
    run.startSlave();
    if (run.onMaster())
    {
        /* tds */
        TDataServer tds("toto", "tds for vizir test");
        tds.keepFinalTuple(false);
        // reverse output test
        tds.addAttribute(&x3);
        tds.addAttribute(&x2);
        tds.addAttribute(&x1);
        /* optimizer */
        TNloptGradient nlo(&tds, &run, &solv);
        TGreaterFit it(1.0);
        nlo.addConstraint(&cont3, &dcnt3, &it);
        nlo.addConstraint(&cont2, &dcnt2, &it);
        nlo.addConstraint(&cont1, &dcnt1, &it);
        nlo.addObjective(&obj, &dobj);

        /* resolution */
        nlo.setStartingPoint(3,start[0]);
        nlo.setStartingPoint(3,start[1]);
        nlo.solverLoop();

        run.stopSlave();
        /* resultats */
        tds.getTuple()->Scan("*");
        validateConstraint(tds, 11);
    }

}

void TNloptTest::constraintPlus(TNloptSolver &solv)
{
    /* inputs */
    TAttribute x1("x1", -3.0, 3.0), x2("x2", -3.0, 3.0), x3("x3", -3.0, 3.0);
    /* outputs */
    TAttribute cont1("c1"), cont2("c2"), cont3("c3"), obj("obj"), obj1("obj1"), obj2("obj2");
    /* starting point */
    double start[][3] =
    {
    { 1.5, 1.5, 1.5 },
    { -1.5, -1.5, -1.5 } };

    /* code */
    TCJitEval code(&select);
    code.setInputs(3, &x1, &x2, &x3);
    code.setOutputs(6, &obj1, &obj2, &obj, &cont1, &cont2, &cont3);

    /* run */
    TThreadedRun run(&code, 3);
    //TGradientEstimationRun run(&run_);
    //TSequentialRun run(&code);
    //TMpiRun run(&code);
    run.startSlave();
    if (run.onMaster())
    {
        /* tds */
        TDataServer tds("toto", "tds for vizir test");
        tds.keepFinalTuple(false);
        // reverse output test
        tds.addAttribute(&x3);
        tds.addAttribute(&x2);
        tds.addAttribute(&x1);
        /* optimizer */
        TNlopt nlo(&tds, &run, &solv);
        TGreaterFit it(1.0);
        nlo.addConstraint(&cont3, &it);
        nlo.addConstraint(&cont2, &it);
        nlo.addConstraint(&cont1, &it);
        nlo.addObjective(&obj);

        /* resolution */
        nlo.setStartingPoint(3,start[0]);
        nlo.setStartingPoint(3,start[1]);
        nlo.solverLoop();

        run.stopSlave();
        /* resultats */
        tds.getTuple()->Scan("*");
        validateConstraint(tds, 9);
    }

}

void TNloptTest::validateConstraint(TDataServer &tds, int natt)
{
    int i;
    long siz;

    // only one result
    siz = tds.getNPatterns();
    CPPUNIT_ASSERT_EQUAL(2L, siz);
    // attributes
    // x1, x2, x3, c1, c2, c3, obj
    CPPUNIT_ASSERT_EQUAL(natt, tds.getNAttributes());
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("x1"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("x3"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("c2"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("c3"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("obj"));
    // values 
    // global: 0.5948127, 0.7296318, 0.7296318 -> 0.2556757
    // local : -0.467707, 0.625, 0625 -> 2.3839557
    for (i = 0; i < 2; i++)
    {
        if (tds.getValue("x3", i) > 0)
        {
            // global found
            CPPUNIT_ASSERT_DOUBLES_EQUAL(0.2556757, tds.getValue("obj", i),
                    DBL_PREC_STD);
            CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5948127, tds.getValue("x3", i),
                    DBL_PREC_LARGE);
            CPPUNIT_ASSERT_DOUBLES_EQUAL(0.7296318, tds.getValue("x1", i),
                    DBL_PREC_LARGE);
        }
        else
        {
            // local found
            CPPUNIT_ASSERT_DOUBLES_EQUAL(2.3839557, tds.getValue("obj", i),
                    DBL_PREC_MED);
            CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.467707, tds.getValue("x3", i),
                    DBL_PREC_LARGE);
            CPPUNIT_ASSERT_DOUBLES_EQUAL(0.625, tds.getValue("x1", i),
                    DBL_PREC_LARGE);
        }
        CPPUNIT_ASSERT_DOUBLES_EQUAL(tds.getValue("x1", i),
                tds.getValue("x2", i), DBL_PREC_STD);
    }

}


void TNloptTest::validateBar()
{

    TDataServer tds("OptimDemo","test");
    tds.keepFinalTuple(false);
    tds.fileDataRead("_hollowBar_exported.dat",false);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.55, tds.getValue("sect", 0), DBL_PREC_MED);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tds.getValue("def", 0), DBL_PREC_MED);

    UranieTests::removeFile(gSystem->pwd(),"_hollowBar_exported.dat");
}

#ifdef WITH_MPI    
void TNloptTest::barMPIMMAGradDCJit()
{

    Int_t nProc = 4;
    // Run the macro with MPI
    gSystem->Exec(UranieTests::mpi_command(nProc, "nloptmpi MMA evalType 1").data());
    validateBar();     
}
void TNloptTest::barMPIMMAGradDCInt()
{
    Int_t nProc = 4;
    // Run the macro with MPI
    gSystem->Exec(UranieTests::mpi_command(nProc, "nloptmpi MMA evalType 0").data());
    validateBar();     
}
void TNloptTest::barMPIMMAGradDCodeLine()
{
    Int_t nProc = 4;
    // Run the macro with MPI
    gSystem->Exec(UranieTests::mpi_command(nProc, "nloptmpi MMA evalType 2").data());
    validateBar();     
}
void TNloptTest::barMPIMMAGradDCodeFlat()
{
    Int_t nProc = 4;
    // Run the macro with MPI
    gSystem->Exec(UranieTests::mpi_command(nProc, "nloptmpi MMA evalType 3").data());
    validateBar();     
}
void TNloptTest::barMPIMMAGradDCodeKey()
{
    Int_t nProc = 4;
    // Run the macro with MPI
    gSystem->Exec(UranieTests::mpi_command(nProc, "nloptmpi MMA evalType 4").data());
    validateBar();     
}

void TNloptTest::barMPISubplexeGradDCJit()
{
    Int_t nProc = 4;
    // Run the macro with MPI
    gSystem->Exec(UranieTests::mpi_command(nProc, "nloptmpi evalType 1").data());
    validateBar();     
}
void TNloptTest::barMPISubplexeGradDCInt()
{
    Int_t nProc = 4;
    // Run the macro with MPI
    gSystem->Exec(UranieTests::mpi_command(nProc, "nloptmpi evalType 0").data());
    validateBar();     
}
void TNloptTest::barMPISubplexeGradDCodeLine()
{
    Int_t nProc = 4;
    // Run the macro with MPI
    gSystem->Exec(UranieTests::mpi_command(nProc, "nloptmpi evalType 2").data());
    validateBar();     
}
void TNloptTest::barMPISubplexeGradDCodeFlat()
{
    Int_t nProc = 4;
    // Run the macro with MPI
    gSystem->Exec(UranieTests::mpi_command(nProc, "nloptmpi evalType 3").data());
    validateBar();     
}
void TNloptTest::barMPISubplexeGradDCodeKey()
{
    Int_t nProc = 4;
    // Run the macro with MPI
    gSystem->Exec(UranieTests::mpi_command(nProc, "nloptmpi evalType 4").data());
    validateBar();     
}

void TNloptTest::barMPISubplexeNoGradDCJit()
{
    Int_t nProc = 4;
    // Run the macro with MPI
    gSystem->Exec(UranieTests::mpi_command(nProc, "nloptmpi evalType 1").data());
    validateBar();     
}
void TNloptTest::barMPISubplexeNoGradDCInt()
{
    Int_t nProc = 4;
    // Run the macro with MPI
    gSystem->Exec(UranieTests::mpi_command(nProc, "nloptmpi evalType 0").data());
    validateBar();     
}
void TNloptTest::barMPISubplexeNoGradDCodeLine()
{
    Int_t nProc = 4;
    // Run the macro with MPI
    gSystem->Exec(UranieTests::mpi_command(nProc, "nloptmpi evalType 2").data());
    validateBar();     
}
void TNloptTest::barMPISubplexeNoGradDCodeFlat()
{
    Int_t nProc = 4;
    // Run the macro with MPI
    gSystem->Exec(UranieTests::mpi_command(nProc, "nloptmpi evalType 3").data());
    validateBar();     
}
void TNloptTest::barMPISubplexeNoGradDCodeKey()
{
    Int_t nProc = 4;
    // Run the macro with MPI
    gSystem->Exec(UranieTests::mpi_command(nProc, "nloptmpi evalType 4").data());
    validateBar();     
}

void TNloptTest::barMPISubplexeInitInConstCJit()
{
    Int_t nProc = 4;
    // Run the macro with MPI
    gSystem->Exec(UranieTests::mpi_command(nProc, "nloptmpi initInConst evalType 1").data());
    validateBar();     
}
void TNloptTest::barMPISubplexeInitInConstCInt()
{
    Int_t nProc = 4;
    // Run the macro with MPI
    gSystem->Exec(UranieTests::mpi_command(nProc, "nloptmpi initInConst evalType 0").data());
    validateBar();     
}
void TNloptTest::barMPISubplexeInitInConstCodeLine()
{
    Int_t nProc = 4;
    // Run the macro with MPI
    gSystem->Exec(UranieTests::mpi_command(nProc, "nloptmpi initInConst evalType 2").data());
    validateBar();     
}
void TNloptTest::barMPISubplexeInitInConstCodeFlat()
{
    Int_t nProc = 4;
    // Run the macro with MPI
    gSystem->Exec(UranieTests::mpi_command(nProc, "nloptmpi initInConst evalType 3").data());
    validateBar();     
}
void TNloptTest::barMPISubplexeInitInConstCodeKey()
{
    Int_t nProc = 4;
    // Run the macro with MPI
    gSystem->Exec(UranieTests::mpi_command(nProc, "nloptmpi initInConst evalType 4").data());
    validateBar();     
}
#endif
