//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestUserFunctions.cxx
 \author Gilles Arnaud <gilles.arnaud@cea.fr>
 \date   mar mai 21 13:36:02 GMT 2013  

 \brief Implement the user function 

 */

// ROOT
#include "TROOT.h"
#include "TMath.h"

#include "UEntry.h"

/* Rosenbrock optimization probleme
 * */

#define SIZE 6
#define VAL(tab, i) (*(*tab)[i]->val.dou)

int frosenbrock(Double_t *in, Double_t *res)
{
    double ret, next;
    int i;

    ret = (in[0] - 1.) * (in[0] - 1.);
    for (i = 1; i < SIZE; i++)
    {
        next = in[i - 1] * in[i - 1] - in[i];
        ret += next * next;
    }
    res[0] = ret;
    return 1;
}

int frosenbrockGrad(vector<URANIE::DataServer::UEntry *> *in, 
    vector<URANIE::DataServer::UEntry*> *out)
{
    double ret, next;
    vector<double> *grad;
    int i;

    // rose (out[0])
    ret = (VAL(in, 0) - 1.) * (VAL(in,0) - 1.);
    for (i = 1; i < SIZE; i++)
    {
        next = VAL(in,i - 1) * VAL(in,i - 1) - VAL(in, i);
        ret += next * next;
    }
    VAL(out, 0) = ret;
    //*(*out)[0]->val.dou = ret;
    // rose' (out[1])
    grad = (*out)[1]->val.vec;
    grad->push_back(2 * (VAL(in, 0) - 1.));
    for (i = 1; i < SIZE; i++)
        grad->push_back(0.);
    for (i = 1; i < SIZE; i++) {
        grad->at(i - 1) += 4 * VAL(in, i - 1) * (VAL(in, i-1) * VAL(in, i-1) - VAL(in, i));
        grad->at(i) += 2 * (VAL(in, i) - VAL(in, i-1) * VAL(in, i-1));
    }
    
    return 1;
}

int frosenbrockVec(vector<URANIE::DataServer::UEntry *> *in, 
    vector<URANIE::DataServer::UEntry*> *out)
{
    vector<double> *vec;
    double ret, next;
    int i;

    vec = (*out)[0]->val.vec;
    vec->push_back ((VAL(in, 0) - 1.) * (VAL(in, 0) - 1.));
    for (i = 1; i < SIZE; i++)
    {
        next = VAL(in, i - 1) * VAL(in, i - 1) - VAL(in, i);
        vec->push_back(next * next);
    }
    return 1;
}

int frosenbrockCummul(vector<URANIE::DataServer::UEntry *> *in, 
    vector<URANIE::DataServer::UEntry*> *out)
{
    vector<double> *vin;
    double ret;
    
    vin = (*in)[0]->val.vec;
    ret = 0;
    for (auto it : *vin) ret += it;
    VAL(out, 0) = ret;
    return 1;
}

/* a constraint problem
 * */

/* differents cas de saturation des contraintes
 #define ZZZ 0.4 // 3 saturées
 #define ZZZ 0.55 // 2 saturées
 #define ZZZ 0.65 // aucune saturées
 * */
#define ZZZ 0.55 // 2 saturées

int select(double *in, double *res)
{
    double x, y, z;

    x = in[0];
    y = in[1];
    z = in[2];
    /* critere */
    // cumul des distances aux points (1,0.5,ZZZ) et (0.5,1,ZZZ)
    res[0] = (x - 1.) * (x - 1.) + (y - .5) * (y - .5) + (z - ZZZ) * (z - ZZZ);
    res[1] = (x - .5) * (x - .5) + (y - 1.) * (y - 1.) + (z - ZZZ) * (z - ZZZ);
    res[2] = res[0] + res[1];
    /* contraintes */
    res[3] = x * x + y * y + z * z; // distance au point (0,0,0)
    res[4] = (x - 1.5) * (x - 1.5) + (y - 0.5) * (y - 0.5) + z * z; // point (1.5,0.5,0)
    res[5] = (x - 0.5) * (x - 0.5) + (y - 1.5) * (y - 1.5) + z * z; // point (0.5,1.5,0)
    return 1;

}

int fselect(double *in, double *res)
{
    double x, y, z;

    x = in[0];
    y = in[1];
    z = in[2];
    /* critere */
    // cumul des distances aux points (1,0.5,ZZZ) et (0.5,1,ZZZ)
    res[0] = (x - 1.) * (x - 1.) + (y - .5) * (y - 0.5)
            + 2 * (z - ZZZ) * (z - ZZZ) + (x - 0.5) * (x - .5)
            + (y - 1.) * (y - 1.);
    /* contraintes */
    res[1] = x * x + y * y + z * z; // distance au point (0,0,0)
    res[2] = (x - 1.5) * (x - 1.5) + (y - 0.5) * (y - 0.5) + z * z; // point (1.5,0.5,0)
    res[3] = (x - 0.5) * (x - 0.5) + (y - 1.5) * (y - 1.5) + z * z; // point (0.5,1.5,0)
    
    return 1;
}

int fselectd(vector<URANIE::DataServer::UEntry*> *in, 
    vector<URANIE::DataServer::UEntry *>  *res)
{
    double x, y, z;
    vector<double> *grad;

    x = VAL(in, 0);
    y = VAL(in, 1);
    z = VAL(in, 2);
    /* critere */
    // cumul des distances aux points (1,0.5,ZZZ) et (0.5,1,ZZZ)
    VAL(res, 0) = (x - 1.) * (x - 1.) + (y - .5) * (y - 0.5)
            + 2 * (z - ZZZ) * (z - ZZZ) + (x - 0.5) * (x - .5)
            + (y - 1.) * (y - 1.);
    /* contraintes */
    VAL(res, 2) = x * x + y * y + z * z; // distance au point (0,0,0)
    VAL(res, 4) = (x - 1.5) * (x - 1.5) + (y - 0.5) * (y - 0.5) + z * z; // point (1.5,0.5,0)
    VAL(res, 6) = (x - 0.5) * (x - 0.5) + (y - 1.5) * (y - 1.5) + z * z; // point (0.5,1.5,0)
    /* gradient */
    // dcrit
    grad = res->at(1)->val.vec;
    grad->push_back (2 * (x - 1.) + 2 * (x - .5));
    grad->push_back (2 * (y - .5) + 2 * (y - 1.));
    grad->push_back (4 * (z - ZZZ));
    // dcont
    grad = res->at(3)->val.vec;
    grad->push_back(2 * x);
    grad->push_back(2 * y);
    grad->push_back( 2 * z);

    grad = res->at(5)->val.vec;
    grad->push_back( 2 * (x - 1.5));
    grad->push_back( 2 * (y - .5));
    grad->push_back( 2 * z);

    grad = res->at(7)->val.vec;
    grad->push_back( 2 * (x - .5));
    grad->push_back( 2 * (y - 1.5));
    grad->push_back( 2 * z);
    
    return 1;
}

int fselect_vec(vector<URANIE::DataServer::UEntry*> *in, 
    vector<URANIE::DataServer::UEntry *>  *res)
{
    double x, y, z;
    vector<double> *cont;

    x = VAL(in, 0);
    y = VAL(in, 1);
    z = VAL(in, 2);
    /* critere */
    // cumul des distances aux points (1,0.5,ZZZ) et (0.5,1,ZZZ)
    VAL(res, 0) = (x - 1.) * (x - 1.) + (y - .5) * (y - 0.5)
            + 2 * (z - ZZZ) * (z - ZZZ) + (x - 0.5) * (x - .5)
            + (y - 1.) * (y - 1.);
    /* contraintes */
    cont = res->at(1)->val.vec;
    cont->push_back( x * x + y * y + z * z); // distance au point (0,0,0)
    cont->push_back((x - 1.5) * (x - 1.5) + (y - 0.5) * (y - 0.5) + z * z); // point (1.5,0.5,0)
    cont->push_back((x - 0.5) * (x - 0.5) + (y - 1.5) * (y - 1.5) + z * z); // point (0.5,1.5,0)
    return 1;
}

int fselect_min(vector<URANIE::DataServer::UEntry *> *in, 
    vector<URANIE::DataServer::UEntry*> *out)
{
    vector<double> *vin;
    double ret;
    
    vin = (*in)[0]->val.vec;
    ret = vin->at(0);
    for (auto it : *vin) if (it < ret) ret = it;
    VAL(out, 0) = ret;
    return 1;
}


void add(double *in, double *out)
{
    out[0] = in[0] + in[1];
}
