/* --- Ansi C --- */
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2016-2022 CEA/DES
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   vzr_mating.c
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2016-07-13
 
 \brief  choix des parents
*/

#include "vzr_mating.h"

#include "vizir/vzr_aleas.h"
#include "vizir/vzrpop_util.h"
#include "vizir/vzr_ibea.h"
#include "vizir/vzr_moeadd.h"

#define ISOL
#define BIAIS 0.5
#define ETRANGER 0.9

static INLINE int first(struct vzrpop *pop)
{
    int ret, ichoix, tmp;
    
    /* choix du premier parent
     * soit un element de l'elite (s'il y en a assez)
     *   plus bias est grand plus on choisit dans les elites 
     * soit un quelconque des survivants */
    ichoix = (pop->elite > pop->step && mt_real1() < BIAIS) ?
            pop->elite : pop->level ;
    ret = mt_int(ichoix);
#ifdef ISOL
    tmp = mt_int(ichoix);
    if (pop->item[pop->rg[tmp]].isol > pop->item[pop->rg[ret]].isol) ret = tmp;
#endif
    return ret;
}

static struct vzrcouple couple_std(struct vzrpop *pop)
{
    struct vzrcouple ret;
    int tmp;
    
    /* premier parent */
    ret.pa = first(pop);

    /* le second parent est tjs pris parmi les survivants */
    /* deux autres individus */
    do ret.ma = mt_int(pop->level);
    while (ret.pa == ret.ma);
    do tmp = mt_int(pop->level);
    while (tmp == ret.pa || tmp == ret.ma);
    /* tournoi sur classement */
    if (tmp < ret.ma) ret.ma = tmp;
#ifdef RANKED
    if (ret.pa > ret.ma) {
        tmp = ret.pa; ret.pa = ret.ma; ret.ma = tmp;
    }
#endif
    ret.pa = pop->rg[ret.pa];
    ret.ma = pop->rg[ret.ma];
    return ret;
}

static struct vzrcouple couple_voisin(struct vzrpop *pop)
{
    struct vzrcouple ret;
    struct voisin *self = (struct voisin *) pop->info;
    
    /* premier parent */
    ret.pa = first(pop);
    
    if (mt_real1() < ETRANGER) {
        /* second parent est un voisin */
        ret.ma = self->ref[pop->rg[ret.pa]][mt_int(self->prof)];
    }
    else {
        /* second parent est un survivant */
        do ret.ma = mt_int(pop->level) ;
        while (ret.ma == ret.pa);
        ret.ma = pop->rg[ret.ma];
    }
    ret.pa = pop->rg[ret.pa];
    return ret;
}

static struct vzrcouple couple_ibea(struct vzrpop *pop)
{
    struct vzrcouple ret;
    int tmp;
    double indij, indma, indtmp;
    
    struct ibea *self = (struct ibea *) pop->info;
    
    /* premier parent */
    ret.pa = first(pop);

    /* le second parent */
    /* deux individus survivants */
    do ret.ma = mt_int(pop->level);
    while (ret.pa == ret.ma);
    do tmp = mt_int(pop->level);
    while (tmp == ret.pa || tmp == ret.ma);
    /* comparaison sur la proximite via l'indicateur ibea */
    ibea_indicator(self, &pop->item[pop->rg[ret.pa]], &pop->item[pop->rg[ret.ma]], pop
        , &indij, &indma);
    ibea_indicator(self, &pop->item[pop->rg[ret.pa]], &pop->item[pop->rg[tmp]], pop
        , &indij, &indtmp);
    /* tournoi */
    if (indtmp < indma) ret.ma = tmp;

    ret.pa = pop->rg[ret.pa];
    ret.ma = pop->rg[ret.ma];
    return ret;
}

static struct vzrcouple couple_ibea2(struct vzrpop *pop)
{
    struct vzrcouple ret;
    int tmp, i;
    double done, dtwo, dtmp;
    
    struct ibea *self = (struct ibea *) pop->info;
    
    /* premier parent */
    ret.pa = first(pop);

    /* le second parent */
    /* deux individus survivants */
    do ret.ma = mt_int(pop->level);
    while (ret.pa == ret.ma);
    do tmp = mt_int(pop->level);
    while (tmp == ret.pa || tmp == ret.ma);
    /* comparaison sur la proximite */
    done = 0;
    for EACH(i, pop->valid+pop->cost) {
        dtmp = (pop->item[pop->rg[ret.pa]].front[i] - pop->item[pop->rg[ret.ma]].front[i])
                / self->it[i];
        done += dtmp*dtmp;
    }
    dtwo = 0;
    for EACH(i, pop->valid+pop->cost) {
        dtmp = (pop->item[pop->rg[ret.pa]].front[i] - pop->item[pop->rg[tmp]].front[i])
                / self->it[i];
        dtwo += dtmp*dtmp;
    }
    /* tournoi */
    if ( dtwo < done) ret.ma = tmp;

    ret.pa = pop->rg[ret.pa];
    ret.ma = pop->rg[ret.ma];
    return ret;
}

static struct vzrcouple couple_proche(struct vzrpop *pop)
{
    struct vzrcouple ret;
    int tmp, i;
    double done, dtwo, dtmp;
    
    struct voisin *self = (struct voisin *) pop->info;
    
    /* premier parent */
    ret.pa = first(pop);

    /* le second parent */
    /* deux individus survivants */
    do ret.ma = mt_int(pop->level);
    while (ret.pa == ret.ma);
    do tmp = mt_int(pop->level);
    while (tmp == ret.pa || tmp == ret.ma);
    /* comparaison sur la proximite */
    done = 0;
    for EACH(i, pop->valid+pop->cost) {
        dtmp = (pop->item[pop->rg[ret.pa]].front[i] - pop->item[pop->rg[ret.ma]].front[i])
                / self->norm[i];
        done += dtmp*dtmp;
    }
    dtwo = 0;
    for EACH(i, pop->valid+pop->cost) {
        dtmp = (pop->item[pop->rg[ret.pa]].front[i] - pop->item[pop->rg[tmp]].front[i])
                / self->norm[i];
        dtwo += dtmp*dtmp;
    }
    /* tournoi */
    if ( dtwo < done) ret.ma = tmp;

    ret.pa = pop->rg[ret.pa];
    ret.ma = pop->rg[ret.ma];
    return ret;
}

static struct vzrcouple couple_moeadd(struct vzrpop *pop)
{
    int group, nbv, vs, i, j, k;
    struct vzrcouple ret;
    struct ddknn_info *self = (struct ddknn_info*) pop->info;
    VZRTAB(int, vois, self->near.prof);
    
    /* premier parent */
    ret.pa = first(pop);
    
    /* filtrage voisin */
    group = pop->item[pop->rg[ret.pa]].group;
    nbv = 0;
    for EACH(i, self->near.prof) {
        j = self->near.ref[group][i];
        if (self->inner[j]) {
            vois[nbv] = j;
            nbv++;
        }
    }
    /* second parent */
    if (nbv && mt_real1() < ETRANGER) {
        /* choix voisin */
        vs = vois[mt_int(nbv)];
        i = mt_int(self->inner[vs]);
        k = j = 0;
        do {
            while (pop->item[pop->rg[j]].group != vs) j++;
        } while (k++ != i);
        ret.ma = j;
    }
    else {
         /* choix survivant */
        do ret.ma = mt_int(pop->level) ;
        while (ret.ma == ret.pa);
    }
    ret.pa = pop->rg[ret.pa];
    ret.ma = pop->rg[ret.ma];

    VZRTAB_OUT(vois);
    return ret;
}

/* declaration des structs */
struct vzrcoupling vzrcoupling_std = { None, &couple_std };
struct vzrcoupling vzrcoupling_voisin = { Knn, &couple_voisin};
struct vzrcoupling vzrcoupling_ibea = {Ibea, &couple_ibea};
struct vzrcoupling vzrcoupling_ibea2 = {Ibea, &couple_ibea2};
struct vzrcoupling vzrcoupling_proche = {Knn, &couple_proche};
struct vzrcoupling vzrcoupling_moeadd = {Ddknn, &couple_moeadd};
