/* --- Ansi C --- */
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2008-2022 CEA/DES
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   vzr_naitre.c
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2008-03-19
 
 \brief  naissance des indivividus
*/

#include <stdio.h>
#include <stdlib.h>

#include "vzr_naitre.h"

#include "vzr_util.h"
#include "vzr_croiser.h"

#include "vizir/vzr_aleas.h"
#include "vizir/vzrfifo_api.h"
#include "vizir/vzrpop_api.h"

/** vzrcouple **/
/* standard */
static void accoupler(void *rien, struct vizir *self, struct vzrwork *wself)
{
    struct vzrcouple parent;
    int i;
    struct vzrpop *pop = &self->base.pop;

    pthread_mutex_lock(&(self->base.lock));
    /* choix des parents */
    parent = self->mate->mate(&(self->base.pop));
    /* recopie */
    for(i=0;i<pop->gene;i++) {
        wself->pere.domin[i] = pop->item[parent.pa].dip.domin[i];
        wself->pere.reces[i] = pop->item[parent.pa].dip.reces[i];
        wself->mere.domin[i] = pop->item[parent.ma].dip.domin[i];
        wself->mere.reces[i] = pop->item[parent.ma].dip.reces[i];
    }
    pthread_mutex_unlock(&(self->base.lock));
    
}

static void dealloc(struct vzrmating *obj) {}

struct vzrmating *vzrcouple_create(void)
{
    static struct vzrmating ret = {(void*) NULL, &accoupler, &dealloc };
    return &ret;
}

/* sexué */
static void copuler(void *male, struct vizir *female, struct vzrwork *wself)
{
    int     i, j, ichoix, parent;
    struct vizir *ind[2] = { female,  (struct vizir *) male};
    vzrdiploid_t allel[2] = { &(wself->pere), &(wself->mere) };
    struct vzrpop *pop;    

    for (i=0; i<2; i++) {
        pop = &ind[i]->base.pop;
        pthread_mutex_lock(&(ind[i]->base.lock));
        ichoix = ( i && pop->elite > ind[i]->consang && mt_real1() < ind[i]->bias) ?
            pop->elite : pop->stop;
        parent = mt_int(ichoix);
        for (j=0; j<pop->gene; j++) {
            allel[i]->domin[j] = pop->item[pop->rg[parent]].dip.domin[j];
            allel[i]->reces[j] = pop->item[pop->rg[parent]].dip.reces[j];
        }
        pthread_mutex_unlock(&(ind[i]->base.lock));
    }
}

struct vzrmating *vzrsexue_create(struct vizir *other)
{
    struct vzrmating *ret;

    ret = malloc(sizeof(struct vzrmating));
    ret->data = (void *) other;
    ret->couple = &copuler;
    ret->dealloc = (void (*)(struct vzrmating *)) &free;

    return ret;
}

/** generique **/
int reevaluation(struct vizir *self, struct vzrwork *wself)
{
    int ret;

    pthread_mutex_lock(&(self->base.lock));
    ret = vzrpop_again(&(self->base.pop), &(wself->child));
    pthread_mutex_unlock(&(self->base.lock));
    return ret;
}

int immigration(struct vizir *self, struct vzrwork *wself)
{
    int ret, i;

    ret = 0;
    if ( self->base.migr && 
            vzrfifo_get(self->base.migr, wself->child.dip.domin) 
        )  
    {
        ret = 1;
        for EACH(i, self->base.pop.gene) 
            wself->child.dip.reces[i] = wself->child.dip.domin[i];
    }
    return ret;
}

void procreer(struct vizir*self, struct vzrwork* wself)
{
    int     i;
    double  seuil;

    /* le descendant sera issu d'un croisement avec ou sans mutation*/
    seuil = mt_real1();
    if((!wself->oper->mutation) && seuil<self->pmut) { 
        /* mutation sans operateur specifique
         * on ecrase un allele du pere par un allele aleatoire
         * */
        PRO_CALL(wself->oper, sample, &(wself->child.dip));
        for (i=0; i<self->base.pop.gene; i++)
            wself->pere.reces[i] = wself->child.dip.reces[i];
    }
    /* tirage des alleles de l'enfant */
    PRO_CALL(wself->oper, allele, &(wself->pere), wself->child.dip.domin);
    PRO_CALL(wself->oper, allele, &(wself->mere), wself->child.dip.reces);
    /* mutation (avec operateur specifique) */
    if (wself->oper->mutation && seuil<self->pmut) {
        PRO_CALL(wself->oper, mutation, &(wself->child.dip));
    }
    /* homozygotie et dominance sont traite lors du calcul du phenotype   */
    if(mt_real1()<self->phomo)  { 
        /* on construit un homozygote */
        PRO_CALL(wself->oper, homozygote, &(wself->child.dip));
    }
    else {
        /* on construit le phenotype */
        PRO_CALL(wself->oper, heterozygote, &(wself->child.dip));
    }
}


