/* --- Ansi C --- */
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2010-2022 CEA/DES
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   nms_core.c
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2010-04-13
 
 \brief  boucle principale du simplex
*/


#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "nms_core.h"

#include "vizir/vzr_macro.h"
#include "vizir/vzritem.h"

#include "vizir/vzrbase_api.h"
#include "vizir/vzrfifo_api.h"
#include "vizir/vzr_heap.h"
#include "vizir/vzr_aleas.h"
#include "vizir/vzrcore_api.h"
#include "vizir/vzrpop_api.h"
#include "vizir/vzr_survie.h"

#define DIST 0.25
#define CDIST 16.0
#define BOUND_BIAS
#define TRACE

/* etat */
#define BOOT 0
#define REFLECT 1
#define EXPAND 2
#define CONTRACT 3
#define REFRACT 4
#define WAIT 5
#define IMMIG 64

//#define EPSILON 1.e-10

#define MEMO_LOCK pthread_mutex_lock(&(self->base.lock))
#define MEMO_UNLOCK pthread_mutex_unlock(&(self->base.lock))

/* local structure
 * */
struct nmswork {
    struct nms *self;
    void (*save)(struct nms *, char*);
    int etat;
    double low, top, fitone;
    vzrgene_t *one, *ref;
    struct vzritem two;
};

/* utilitaires geometriques 
 * */
static void symetrique(int size, vzrgene_t *un, vzrgene_t *cent, vzrgene_t *two)
{
    int i;
#ifdef BOUND_BIAS
    int j;
    double coef, tmp;

    coef = 1.0;
    j = -1;
    for EACH(i, size) {
        if (cent[i] == un[i]) continue; // NaN
        tmp = (cent[i] > un[i] ) ? 
                (1.0 - cent[i]) / (cent[i] - un[i]) :
                cent[i] / (un[i] - cent[i]);
        if (tmp < coef) {
            coef = tmp;
            j = i;
        }
    }

    if (coef == 1.0) { // in domain
        for EACH(i, size) two[i] = 2 * cent[i] - un[i];
    }
    else if (coef > 0.9) { // scotch on bounds
        for EACH(i, size)
            two[i] = cent[i] + coef * (cent[i] - un[i]);
        if (j != -1) two[j] = (cent[j] > un[j]) ? 1. : 0. ; // roundoff
    }
    else { // atenuated rebond
#endif
        for EACH(i, size) {
            two[i] = 2 * cent[i] - un[i];
            if (two[i] > 1.) two[i] = 1.5 - two[i] / 2;
            else if (two[i] < 0.) two[i] = two[i] / -2;
        }
#ifdef BOUND_BIAS
    }
#endif
}

static INLINE void milieu(int size, vzrgene_t *un, vzrgene_t *deux, vzrgene_t *cent)
{   
    int i;
    for EACH(i, size) cent[i] = (un[i] + deux[i])/2;
}

static INLINE void sample(int size, vzrgene_t *cent)
{   
    int i;
    for EACH(i, size) cent[i] = mt_real1();
}

/* creation du simplexe */
static void getbarycentre(struct vzrpop *pop, int but, int top, vzrgene_t *res, double *lowfit, double *topfit)
{
    int i, ri, j;
    double fit;
    VZRTAB(double, dist, pop->stop+pop->step);
    VZRTAB(int, id, top);
    static int one=2;
#ifdef DIST
    vzrgene_t a;
    double near;
    int ratio = (mt_real1() < DIST) ;
#endif

    /* selection */
    for EACH(i, top) {
        id[i] = ri = pop->rg[i];
        /* part aleatoire */
        dist[ri] = mt_real1();
#ifdef DIST
        if (ratio) {
            near = 0.;
            for EACH(j, pop->gene) {
                a = pop->item[ri].zone[j] - pop->item[pop->rg[but]].zone[j];
                near += a*a;
            }
            dist[ri] -= CDIST * near ;
        }
#endif
    }
    if (but < top) {
        id[0] = pop->rg[but];
        id[but] = pop->rg[0];
        //dist[but] = dist[0];
        vzr_rank_double(top-1, top-pop->gene-1, id+1, dist+1);
    } else {
        if (one) {
            printf("bary %d, %d, %d\n", top, pop->gene, top-pop->gene-1);
            one--;
        } 
        vzr_rank_double(top, top-pop->gene-1, id, dist);
    }
    VZRTAB_OUT(dist);
    /* coordonnées du barycentre */
    for EACH(i, pop->gene) {
        res[i] = 0.;
        for EACH(j, pop->gene) res[i] += pop->item[id[top-j-1]].zone[i];
        res[i] /= pop->gene;
    }
    /* fitness intervale */
    *topfit = *lowfit = pop->item[id[top-1]].rank;
    for FROMEACH(i, 1, pop->gene) {
        fit = pop->item[id[top-i-1]].rank;
        if (fit < *lowfit) *lowfit = fit;
        else if (fit > *topfit) *topfit = fit;
    }
    VZRTAB_OUT(id);
}

static void getsimplex(struct vzrpop *pop, struct nmswork *wself)
{
    int i, id, yaid, top;

    top = pop->stop / 2;
    /* worst item tournoi */
    /* TODO pbl en mode continuous */
    id = mt_int(top) + top ;
    yaid = mt_int(top) + top ;
    if (pop->item[pop->rg[yaid]].isol > pop->item[pop->rg[id]].isol) id = yaid;
    for EACH(i, pop->gene) wself->one[i] = pop->item[pop->rg[id]].zone[i];
    wself->fitone = pop->item[pop->rg[id]].rank;
    /* choice set */
    //top = id;
    //while (wself->fitone == pop->item[top-1].rank && top > pop->low) top--;
    /* other items */
    //printf("%d %d %d %d\n", id, top, pop->low, pop->stop);
    getbarycentre(pop, id, top, wself->ref, &(wself->low), &(wself->top));
}

static void nms_add(struct nms *self, struct vzritem *item)
{
    struct vzrpop *pop = &self->base.pop;
    
    if (vzrpop_add(pop, item)) {
        vzrpop_cycle(pop);
        if (!self->base.quiet)
            printf("%s  Generation : %d, elite %d rang max %d\n"
                ,self->base.id, pop->cycle, pop->elite
                , pop->item[pop->rg[pop->level-1]].rank);
    }
}

/* protocole
 * */
static void nmswork_subscribe(void *core)
{
    struct nmswork *wself = (struct nmswork *) core;
    struct nms *self = wself->self;

    pthread_mutex_lock(&(self->base.lock));
    (void) vzrbase_first(&(self->base));
    wself->etat = (self->base.pop.elite) ? REFLECT : BOOT;
    
    if (!self->base.quiet) 
        printf("Simplex %s %d\n", self->base.id, self->base.slave);
    pthread_mutex_unlock(&(self->base.lock));
}

static void nmswork_unsubscribe(void *core)
{
    struct nmswork *wself = (struct nmswork *) core;
    struct nms *self = wself->self;

    pthread_mutex_lock(&(self->base.lock));
    if (self->base.slave == 1) {
        printf("Simplex %s converge %d\n", self->base.id, self->base.eval);
        //VZRCORE_SAVE(wself, self);
    }
    if (self->base.pop.dealloc_data && wself->two.data) 
        self->base.pop.dealloc_data(wself->two.data);
    wself->etat = WAIT;
    self->base.slave--;
    pthread_mutex_unlock(&(self->base.lock));
}

static int nmswork_get(void *core, vzrgene_t*in)
{
    struct nmswork *wself = (struct nmswork *) core;
    struct nms *self = wself->self;

    pthread_mutex_lock(&(self->base.lock));
    if (vzrbase_end(&(self->base)) ) {
        pthread_mutex_unlock(&(self->base.lock));
        return 1;
    }
    else {
        if (self->base.pop.again) {
            struct vzritem item;

            item.flag = APLOID;
            item.zone = in;
            item.front = NULL;
            item.data = NULL;
            vzrpop_again(&(self->base.pop), &item);
            pthread_mutex_unlock(&(self->base.lock));
            wself->etat |= IMMIG;
        }
        else if (self->base.migr && vzrfifo_get(self->base.migr, in)) {
            pthread_mutex_unlock(&(self->base.lock));
            wself->etat |= IMMIG;
        }
        else {
            pthread_mutex_unlock(&(self->base.lock));
            switch (wself->etat) {
            case BOOT :
                sample(self->base.pop.gene, in);
                break;
            case REFLECT :
                pthread_mutex_lock(&(self->base.lock)); 
                getsimplex(&self->base.pop, wself);
                pthread_mutex_unlock(&(self->base.lock));
                symetrique(self->base.pop.gene, wself->one, wself->ref, in);
                break;
            case EXPAND :
                symetrique(self->base.pop.gene, wself->ref, wself->two.zone, in);
                break;
            case REFRACT :
                milieu(self->base.pop.gene, wself->one, wself->ref, in);
                break;
            case CONTRACT :
                milieu(self->base.pop.gene, wself->two.zone, wself->ref, in);
                if (self->base.pop.dealloc_data && wself->two.data) { 
                    self->base.pop.dealloc_data(wself->two.data);
                    wself->two.data = NULL;
                }
                break;
            }
        }
        return 0;
    }
}

static int nmswork_put(void *core, int good, vzrgene_t *in, double *out, void *dat)
{
    struct nmswork *wself = (struct nmswork *) core;
    struct nms *self = wself->self;
    int viable, ret;
    struct vzritem item;

    item.flag = APLOID;
    item.group = 0;
    item.front = out;
    item.zone = in;
    item.data = dat;

    pthread_mutex_lock(&(self->base.lock));
    self->base.eval++;
    /* evaluation individu */
    ret = 0;
    if (good) {
        viable = vzrpop_viable(&(self->base.pop), &item);
        /* maj */
        if (wself->etat & IMMIG) {
            if (viable) nms_add(self, &item);
            else {
                if (self->base.pop.dealloc_data && dat) self->base.pop.dealloc_data(dat);
            } 
            wself->etat &= ~IMMIG;
        }
        else {
            switch (wself->etat) {
            case BOOT:
                nms_add(self, &item);
                if (self->base.pop.level == self->base.pop.stop) wself->etat = REFLECT;
                break;
            case REFLECT:
                /* next state */
                if (item.rank <= wself->low) {
                    wself->etat = EXPAND;
                }
                else if (item.rank > wself->top) {
                    wself->etat = (item.rank > wself->fitone)? REFRACT : CONTRACT ;
                }
                /* action */
                if (wself->etat == REFLECT) {
                    nms_add(self, &item);
                }
                else if (wself->etat == REFRACT) {
                    if (self->base.pop.dealloc_data && item.data) 
                        self->base.pop.dealloc_data(item.data);
                }
                else {
                    /* memorize reflection point */
                    vzritem_cp(&(self->base.pop), &item, &(wself->two));
                }
                break;
            case EXPAND:
                if (item.rank <= wself->two.rank) { 
                    nms_add(self, &item);
                    if (self->base.pop.dealloc_data && wself->two.data) {
                        self->base.pop.dealloc_data(wself->two.data);
                        wself->two.data = NULL;
                    }
                }
                else {
                    nms_add(self, &(wself->two));
                    wself->two.data = NULL;
                    if (self->base.pop.dealloc_data && item.data) 
                        self->base.pop.dealloc_data(item.data);
                }
                wself->etat = REFLECT;
                break;
            case REFRACT:
            case CONTRACT:
                if (item.rank < wself->fitone) nms_add(self, &item);
                else {
                    if (self->base.pop.dealloc_data && item.data) 
                        self->base.pop.dealloc_data(item.data);
                }
                wself->etat = REFLECT;
                break;
            }
        }
    } 
    else { /* unevaluable item */
        if (wself->etat & IMMIG) {
            wself->etat &= ~IMMIG;
        }
        else {
            switch (wself->etat) {
            case BOOT:
                if (self->base.pop.level == self->base.pop.stop) wself->etat = REFLECT;
                break;
            case EXPAND:
                nms_add(self, &(wself->two));
                wself->two.data = NULL;
            case REFRACT:
                wself->etat = REFLECT;
                break;
            case REFLECT:
            case CONTRACT:
                wself->etat = REFRACT;
                break;
            }
        }
    }

    /* test convergence */
    ret = vzrpop_end(&(self->base.pop));
    pthread_mutex_unlock(&(self->base.lock));
    return ret;
}

static void nmswork_dealloc(struct vzrcore *core)
{
    struct nmswork *wself = (struct nmswork *) core->data;
    
    free(wself->two.front);
    free(wself->ref);
    free(wself);
    free(core);
}

struct vzrcore *vzrcore_nms(struct nms *self)
{
    struct nmswork *dat;
    struct vzrcore *ret;

    /* data */
    dat = (struct nmswork *) malloc(sizeof(struct nmswork));
    dat->self = self;
    dat->etat = WAIT;
    dat->ref = (vzrgene_t *) calloc(3*self->base.pop.gene, sizeof(vzrgene_t));
    dat->one = dat->ref + self->base.pop.gene;
    dat->two.flag = APLOID;
    dat->two.zone = dat->ref + 2 * self->base.pop.gene;
    dat->two.data = NULL;
    dat->two.front = (double *) calloc(self->base.pop.all, sizeof(double));

    /* protocole */
    ret = (struct vzrcore *) malloc(sizeof(struct vzrcore));
    ret->data = (void *) dat;
    ret->get = &nmswork_get;
    ret->put = &nmswork_put;
    ret->subscribe = &nmswork_subscribe;
    ret->unsubscribe = &nmswork_unsubscribe;
    ret->dealloc = &nmswork_dealloc;

    return ret;
}


/* mainloop
 * */
void nms_run(struct nms *self)
{
    vzrcore_run(vzrcore_nms(self));
}

void nms_start(struct nms *self)
{
    vzrcore_start(vzrcore_nms(self));
}

void nms_wait(struct nms *self)
{
    vzrbase_wait(&(self->base));
}

int nms_running(struct nms *self)
{
    return vzrbase_running(&(self->base));
}
