/* --- Ansi C --- */
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2010-2022 CEA/DES
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   pso_core.c
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2010-03-04
 
 \brief  boucle de résolution
*/

#include <stdio.h>
#include <stdlib.h>

#include "pso_core.h"

#include "vizir/vzr_macro.h"
#include "vizir/vzrcore_api.h"
#include "vizir/vzrbase_api.h"
#include "vizir/vzrfifo_api.h"
#include "vizir/vzrpop_api.h"
#include "vizir/vzritem.h"
//#include "pso_guide.h"

//#include "swarm/pso_guide.h"

/* state 
 * particle is in use
 **/
#define FREE 0
#define CALC 1

/* last 
 * last call is succesfull
 **/
#define GOOD 1
#define BAD 0

struct psowork {
    struct psoswarm *self;
    struct psopart *part;
    int immig;
};

/* protocole
 * */
static void psowork_subscribe(void *core)
{
    struct psowork *wself = (struct psowork *) core;
    struct psoswarm *self = wself->self;
    struct psopart *part = wself->part;
    int i;

    pthread_mutex_lock(&(self->base.lock));
    if ( vzrbase_first(&(self->base)) ) {
        /* specifique */
        self->partc = 0;
        for EACH(i, self->nbpart) {
            part = &(self->swarm[i]);
            part->state = FREE;
            part->last = GOOD;
            part->memo = 0;
            //vzrpop_reset(&(part->local));
        }
    }
    
    if (!self->base.quiet) 
        printf("Swarm   %s %d\n", self->base.id, self->base.slave-1);
    pthread_mutex_unlock(&(self->base.lock));
}

static void psowork_unsubscribe(void *core)
{
    struct psowork *wself = (struct psowork *) core;
    struct psoswarm *self = wself->self;

    pthread_mutex_lock(&(self->base.lock));
    if (self->base.slave == 1) {
        printf("Swarm %s converge %d\n", self->base.id, self->base.eval);
        //VZRCORE_SAVE(wself, self);
    }
    self->base.slave--;
    pthread_mutex_unlock(&(self->base.lock));
}

static int psowork_get(void *core, vzrgene_t *next)
{
    struct psowork *wself = (struct psowork *) core;
    struct psoswarm *self = wself->self;
    struct psopart *part = wself->part;
 

    pthread_mutex_lock(&(self->base.lock));
    if (vzrbase_end(&(self->base)) ) {
        pthread_mutex_unlock(&(self->base.lock));
        return 1;
    }
    else {
        /* next eval */
        wself->immig = 0;
        if (self->base.pop.again) {
            struct vzritem item;
            
            item.flag=APLOID;
            item.zone = next;
            item.front = NULL;
            item.data = NULL;
            vzrpop_again(&(self->base.pop), &item);
            pthread_mutex_unlock(&(self->base.lock));
            wself->immig = 1;
            wself->part = part = &(self->swarm[item.group]); //item.group]);
        }
        else if (self->base.migr && vzrfifo_get(self->base.migr, next)) {
            pthread_mutex_unlock(&(self->base.lock));
            wself->immig = 1;
            part = &(self->swarm[0]);
        }
        else {
            int locb;
            VZRTAB(vzrgene_t, loc, self->base.pop.gene);
            VZRTAB(vzrgene_t, glob, self->base.pop.gene);

           /* choix particule */
            do {
                self->partc++;
                if (self->partc == self->nbpart) self->partc = 0;
            } while (self->swarm[self->partc].state == CALC); 
            wself->part = part = &(self->swarm[self->partc]);
            part->state = CALC;
            /* choix attracteur */
            locb = self->mate->local(part, &(self->base.pop), loc);
            if (locb) self->mate->social(part, &(self->base.pop), loc, glob);
            pthread_mutex_unlock(&(self->base.lock));
            /* move */
            if (locb)
                if (part->last) 
                    PRO_CALL(part->move, goodmove, loc, glob, next);
                else
                    PRO_CALL(part->move, badmove, loc, glob, next);
            else PRO_CALL(part->move, sample, next);
            
            VZRTAB_OUT(loc);
            VZRTAB_OUT(glob);
        }
        return 0;
    }
}

static int psowork_put(void *core, int good, vzrgene_t *next, double *out, void *dat)
{
    struct psowork *wself = (struct psowork *) core;
    struct psoswarm *self = wself->self;
    struct psopart *part = wself->part;
    struct vzrpop *pop = &self->base.pop;
    int ret;
    struct vzritem item;

    item.flag = APLOID;
    item.group = part->id;
    item.front = out;
    item.zone = next;
    item.data = dat;

    pthread_mutex_lock(&(self->base.lock));
    self->base.eval++;
    /* integration individu */
    if (good) {
        if (vzrpop_viable(pop, &item)) {
            part->last = GOOD;
            /* ajout global */
            if (vzrpop_add(pop, &item)) {
                vzrpop_cycle(pop);
                if (!self->base.quiet) 
                    printf("%s  Generation : %d, elite %d rang max %d eval %d\n"
                        , self->base.id, pop->cycle, pop->elite
                        , pop->item[pop->rg[pop->level-1]].rank
                        , self->base.eval);
            }
        }
        else {
            if (pop->dealloc_data && dat) pop->dealloc_data(dat);
            part->last = BAD;
        }
    }
    else part->last = BAD;
    
    if (!wself->immig) part->state = FREE;
    /* test convergence */
    ret = vzrpop_end(&(self->base.pop));
    /* reloop */
    pthread_mutex_unlock(&(self->base.lock));
    return ret;
}


static void psowork_dealloc(struct vzrcore *core)
{
    free(core->data);
    free(core);
}
    
struct vzrcore *vzrcore_pso(struct psoswarm *self)
{
    struct vzrcore *ret;
    struct psowork *dat;

    /* data */
    dat = (struct psowork *) malloc(sizeof(struct psowork));
    dat->self = self;
    dat->part = NULL;
    dat->immig = 0;

    /* protocole */
    ret = (struct vzrcore *) malloc(sizeof(struct vzrcore));
    ret->data = (void *) dat;
    ret->get = &psowork_get;
    ret->put = &psowork_put;
    ret->subscribe = &psowork_subscribe;
    ret->unsubscribe = &psowork_unsubscribe;
    ret->dealloc = &psowork_dealloc;

    return ret;
}

/* mainloop
 * */
void pso_run(struct psoswarm *self)
{
    vzrcore_run(vzrcore_pso(self));
}

void pso_start(struct psoswarm *self)
{
    vzrcore_start(vzrcore_pso(self));
}

void pso_wait(struct psoswarm *self)
{
    vzrbase_wait(&(self->base));
}

int pso_running(struct psoswarm *self)
{
    return vzrbase_running(&(self->base));
}


