/* --- Ansi C --- */
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   pso_guide.c
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2013-11-14
 
 \brief  choix des points attracteurs
*/

#include "pso_guide.h"

#include "vizir/vzritem.h"
#include "vizir/vzr_macro.h"
#include "vizir/vzr_aleas.h"

//#define DIST 0.4
#define DIST 0.5
#define IPROF 8
#define SPROF 16

/*** recherche de la solution guide individuelle ***/
static INLINE int propre(struct psopart *self, struct vzrpop *pop, int siz, int *index)
{
    int i, j;
    
    i = j = 0;
    while (j < siz && i < pop->level) {
        if (pop->item[pop->rg[i]].group == self->id ) index[j++] = i;
        i++;
    }
    if (!j) index[0] = self->id;
    return j;
}

/* la iéme meilleure solution globale */
static int pso_local_ieme(struct psopart *self, struct vzrpop *pop, vzrgene_t *out)
{
    int i;

    if (pop->level == pop->stop) {
        /* copie */
        for EACH(i, pop->gene) {
            out[i] = pop->item[pop->rg[self->id]].zone[i];
        }
        return 1;
    }
    else return 0;
}

/* la meilleure solution individuelle */
static int pso_local_best(struct psopart *self, struct vzrpop *pop, vzrgene_t *out)
{
    int i, j;

    if (pop->level == pop->stop) {
        /* recherche */
        i = 0;
        while (pop->item[pop->rg[i]].group != self->id && i < pop->level) i++;
        if (i == pop->level) i = self->id; // not found
        /* copie */
        for EACH(j, pop->gene) {
            out[j] = pop->item[pop->rg[i]].zone[j];
        }
        return 1;
    }
    else return 0;
}

/* une des meilleures solutions individuelle */
static int pso_local_good(struct psopart *self, struct vzrpop *pop, vzrgene_t *out)
{
    int i, j, ours[IPROF];

    if (pop->level == pop->stop) {
        if (self->last && pop->item[pop->rg[self->memo]].group == self->id) {
            /* keep last guide */
            i = self->memo;
        }
        else {
            /* new guide */
            /* first solutions */
            j = propre(self, pop, IPROF, ours);
            /* choice */
            self->memo = i = (j) ? ours[mt_int(j)] : self->id ;
        }
        /* copie */
        for EACH(j, pop->gene) {
            out[j] = pop->item[pop->rg[i]].zone[j];
        }
        return 1;
    }
    else return 0;
}

/* tournoi sur l'indicateur d'isolement */
static int pso_local_tournoi(struct psopart *self, struct vzrpop *pop, vzrgene_t *out)
{
    int i, id, id2, siz, ours[IPROF];

    if (pop->level == pop->stop) {
        if (self->last && pop->item[pop->rg[self->memo]].group == self->id) {
            /* keep last guide */
            id = self->memo;
        }
        else {
            /* new guide */
            siz = propre(self, pop, IPROF, ours);
            //if (siz == 1) printf("local %d for %d\n", siz, IPROF);
            /* choice */
            if (siz > 1) {
                id = ours[mt_int(siz)];
                for EACH(i, 2) {
                    id2 = ours[mt_int(siz)];
                    if (pop->item[pop->rg[id2]].isol > pop->item[pop->rg[id]].isol) 
                        id = id2;
                }
            }
            else id = ours[0];
            self->memo = id;
        }
        /* copie */
        for EACH(i, pop->gene) {
            out[i] = pop->item[pop->rg[id]].zone[i];
        }
        return 1;
    }
    else return 0;
}


/*** recherche de la solution guide sociale ***/
static INLINE int voisin(struct psopart *self, struct vzrpop *pop, int siz, int *index)
{
    int i, j, v1, v2;
    
    v1 = (self->id + 1) % self->part;
    v2 = (self->id + self->part - 1) % self->part;
    i = j = 0;
    while (j < siz && i < pop->level) {
        if (pop->item[pop->rg[i]].group == v1 || pop->item[pop->rg[i]].group == v2) 
            index[j++] = i;
        i++;
    }
    if (!j) index[0] = v2;
    return j;
}

static void pso_social_one(struct psopart *self, struct vzrpop *pop, vzrgene_t *one, vzrgene_t *out)
{
    int i, j, yours[SPROF];
    
    /* first solutions */
    j = voisin(self, pop, SPROF, yours);
    /* choice */
    i = (j) ? yours[mt_int(j)] : yours[0] ;
    /* copie */
    for EACH(i, pop->gene) {
        out[j] = pop->item[pop->rg[i]].zone[j];
    }
}

static void pso_social_near(struct psopart *self, struct vzrpop *pop, vzrgene_t *one, vzrgene_t *out)
{
    int i, j, id, id2, siz, known[SPROF];
    double tmp, d, d2;

    siz = voisin(self, pop, SPROF, known);
    //if (siz == 1) printf("social %d for %d\n", siz, SPROF);
    if (siz > 1) {
        id = known[mt_int(siz)];
        if (mt_real1() < DIST) {
            /* proximité */
            d = 0.0;
            for EACH(j, pop->gene) {
                tmp = one[j] - pop->item[pop->rg[id]].zone[j];
                d += tmp * tmp;
            }
#ifdef NEW
            for EACH(i, siz) {
                id2 = i; 
#else
            for EACH(i, 1) {
                id2 = known[mt_int(siz)];
#endif
                if (id2 == id) continue;
                d2 = 0.0;
                for EACH(j, pop->gene) {
                    tmp = one[j] - pop->item[pop->rg[id2]].zone[j];
                    d2 += tmp * tmp;
                }
                if (d2 < d) {
                    id = id2;
                    d = d2;
                }
            }
        } 
        else {
            /* isolation */
            for EACH(i, 1) {
                id2 = known[mt_int(siz)]; 
                if (pop->item[pop->rg[id2]].isol > pop->item[pop->rg[id]].isol) id = id2;
            }
        }
    }
    else id = known[0];

    /* copie */
    for EACH(i, pop->gene) {
        out[i] = pop->item[pop->rg[id]].zone[i];
    }
}

static void pso_social_tournoi(struct psopart *self, struct vzrpop *pop, vzrgene_t *one, vzrgene_t *out)
{
    int i, id, id2, siz, known[SPROF];

    /* choix */
    siz = voisin(self, pop, SPROF, known);
    if (siz > 1) {
        id = known[mt_int(siz)];
        /* other */
        for EACH(i, 1) {
            id2 = known[mt_int(siz)]; 
            if (id2 < id) id = id2;
        }
    }
    else id = known[0];

    /* copie */
    for EACH(i, pop->gene) {
        out[i] = pop->item[pop->rg[id]].zone[i];
    }
}

/** struct **/
struct psomate psomate_std = {None, pso_local_tournoi, pso_social_near};
struct psomate psomate_tournoi = {None, pso_local_tournoi, pso_social_tournoi};
struct psomate psomate_elite = {None, pso_local_best, pso_social_one};
struct psomate psomate_other = {None, pso_local_good, pso_social_tournoi};
struct psomate psomate_ieme = {None, pso_local_ieme, pso_social_one};
