/* --- Ansi C --- */
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2016-2022 CEA/DES
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   vzr_fleischer.c
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2016-02-05
 
 \brief  hypervolume par Fleischer algorithm
*/

#include "vzr_fleischer.h"

#include <stdlib.h>

#include "vzr_config.h"
#include "vzr_macro.h"
#include "vzr_alloc.h"

/* local sort */
struct heap_comp {
    int one;
    struct vzritem *item;
};

static INLINE int hs_comp(struct heap_comp *hs, int i, int j)
{
    return ( hs->item[i].front[hs->one] < hs->item[j].front[hs->one]
    /*        || ( hs->item[i].front[hs->one] == hs->item[j].front[hs->one] && i < j ) */
        );
}

#include "vizir/vzr_heapsort.h"

static int vzr_rank_by_obj(int size, int *idx, int objone, struct vzritem *item)
{
    struct heap_comp comp = {objone, item};
    return hs_rank(idx, size, 0, &comp);
}

static double contribution(double *nadir, int *ref, int lsize, int **lifo
        , int asize, int **axes, struct vzrpop *pop, double hv)
{
    int i, j, k, kk, kkk;
    int tag, dom;
    double local, top;
    int *nouv;
    double *pt;
    VZRTAB(int, cur, pop->cost);
    VZRTAB(double, val, pop->cost);
    
    hv = 0.;
    while (lsize) {
        /* pop point courant */
        lsize--;
        for EACH(k, pop->cost) cur[k] = lifo[lsize][k];
        /* contribution a l'hypervolume */
        local = 1.;
        for EACH(k, pop->cost) {
            top = (cur[k] == asize-1) ? 
                    nadir[k] : pop->item[axes[k][cur[k] + 1]].front[pop->valid+k];
            local *= (top - pop->item[axes[k][cur[k]]].front[pop->valid+k]);
        }
        hv += local;
    
        /* recursion */
        for EACH(k, pop->cost) {
            /* test recursion */
            if (cur[k] == asize-1) continue; //bordure
            /* coordonnées du point suivant */
            for EACH(kk, pop->cost) lifo[lsize][kk] = cur[kk]; 
            lifo[lsize][k] = cur[k]+1;
            /** test domine **/
            tag = 1;
            nouv = lifo[lsize];
            /* points lifo (test indice) */
            for EACH(i, lsize) {
                dom = 1;
                for EACH(kk, pop->cost) {
                    if (nouv[kk] < lifo[i][kk]) {
                        dom = 0;
                        break;
                    }
                }
                if (dom) {
                    tag = 0;
                    break;
                }
            }
            if (tag == 0) continue;
            /* points pop (test valeur) */
            for EACH(kk, pop->cost) {
                val[kk] = pop->item[axes[kk][nouv[kk]]].front[pop->valid+kk];
            }
            for EACH(kk, pop->cost) {
                for (j=nouv[kk]; j>ref[kk]; j--) {
                    pt = pop->item[axes[kk][j]].front + pop->valid;
                    dom = 1;
                    for EACH(kkk, pop->cost) {
                        if (val[kkk] < pt[kkk]) {
                            dom = 0;
                            break;
                        }
                    }
                    if (dom) {
                        tag = 0;
                        break;
                    }
                }
                if (!tag) break;
            }
            if (tag == 0) continue;
            /* push suivant */
            lsize++;
        }
    }
    VZRTAB_OUT(cur);
    VZRTAB_OUT(val);
    
    return hv;
}

/* fleischer algorithm */
double fleischer(struct vzrpop *pop, double *ref)
{
    int i, j, k;
    int size, tag, now;
    int **axes, **lifo;
    double ret;
    VZRTAB(int, cur, pop->cost);
    NEWmatrice(int, axes, pop->cost, pop->stop);
    NEWmatrice(int, lifo, pop->cost, pop->cost);

    /* filtre des points valides */
    size = 0;
    for EACH(j, pop->level) {
        /* non pareto optimal */
        if (pop->item[pop->rg[j]].rank != 0) continue;
        /* out of ref */
        tag = 0;
        for EACH(k, pop->cost) 
            if (pop->item[pop->rg[j]].front[pop->valid+k] > ref[k]) tag=1;
        if (tag) continue;
        /* good */
        for EACH(k, pop->cost) axes[k][size] = pop->rg[j];
        size++;
    }
    /* premier classement quadrillage par directions */
    for EACH(k, pop->cost) vzr_rank_by_obj(size, axes[k], pop->valid+k, pop->item);
    
    /* calcul de l'hypervolume */
    ret = 0.;
    for (i=size; i>0; i--) {
        /* choix heuristiques du prochain point */
        //now = axes[0][0]; // really bad choice
        //now = axes[0][mt_int(i)];
        //now = axes[0][i-1];
        now = axes[i%pop->cost][i-1];
        /* traitement cube */
        for EACH(k, pop->cost) {
            for(j=0; axes[k][j] != now; j++);
            cur[k] = lifo[0][k] = j;
        }
        ret += contribution(ref, cur, 1, lifo, i, axes, pop, 0.);
        /* mise a jour quadrillage */
        for EACH(k, pop->cost) {
            for (j=0; axes[k][j] != now; j++);
            for (; j<i-1; j++) axes[k][j] = axes[k][j+1];
        }
    }
    DELmatrice(lifo);
    DELmatrice(axes);
    return ret;
}
