/* --- Ansi C --- */
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2014-2022 CEA/DES
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   vzr_hypervol.c
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2014-05-15
 
 \brief  calcul de l'indicateur d'hyper-volume
*/


#include "vizir/vzr_hypervol.h"

#include "vizir/vzr_macro.h"

#ifdef TRACE
#include <stdio.h>
#endif

/* local sort */
struct heap_comp {
    int one, two, three;
    struct vzritem *item;
};

static INLINE int hs_comp(struct heap_comp *hs, int i, int j)
{
    return (hs->item[i].front[hs->one] < hs->item[j].front[hs->one]
        || (   
            hs->item[i].front[hs->one] == hs->item[j].front[hs->one] &&
            hs->item[i].front[hs->two] < hs->item[j].front[hs->two] 
        )
        || (   
            hs->item[i].front[hs->one] == hs->item[j].front[hs->one] &&
            hs->item[i].front[hs->two] == hs->item[j].front[hs->two] &&
            hs->item[i].front[hs->three] < hs->item[j].front[hs->three] 
        )
    );
}

#include "vizir/vzr_heapsort.h"

static int vzr_rank_by_obj(int size, int *idx, int obj1, int obj2, int obj3, struct vzritem *item)
{
    struct heap_comp comp = {obj1, obj2 , obj3, item};
    return hs_rank(idx, size, 0, &comp);
}

/* 3D indicator */
double hyper_volume(struct vzrpop *pop, struct ref3D *ref)
{
    int i, j, k;
    int dom, memi, prec, un, deux;
    double indic, surf, f2, f3;
    VZRTAB(int, id, pop->level);
    VZRTAB(int, id2, pop->level);
    
    /* premier classement */
    for EACH(i, pop->level) id[i] = pop->rg[i];
    vzr_rank_by_obj(pop->level, id, ref->id1, ref->id2, ref->id3, pop->item);

    /* */
    prec = 0;
    indic = 0.0;
    surf = 0.0;
    for EACH(i, pop->level) {
        /* filtrage domaine */
        if (pop->item[id[i]].front[ref->id1] > ref->f1) continue;
        if (pop->item[id[i]].front[ref->id2] > ref->f2) continue;
        if (pop->item[id[i]].front[ref->id3] > ref->f3) continue;
        /* filtrage pareto */
        dom = 0;
        for EACH(j, prec) {
            if (pop->item[id[i]].front[ref->id2] >= pop->item[id2[j]].front[ref->id2]
                    && pop->item[id[i]].front[ref->id3] >= pop->item[id2[j]].front[ref->id3]) {
                dom = 1;
                break;
            }
        }
        if (dom) continue;
        //if (pop->item[id[i]].rank) printf("%d ", id[i]);
        /* calcul indic */
        if (prec) 
            indic += surf * (pop->item[id[i]].front[ref->id1] -
                     pop->item[id[memi]].front[ref->id1]);
#ifdef TRACE
        printf("%d indic %g\n", i, indic);
#endif
        /** insertion point valide **/
        j = 0;
        /* positionnement */
        while (j < prec &&
            pop->item[id2[j]].front[ref->id2] < pop->item[id[i]].front[ref->id2]) {
            j++;
        }
        /* suppression points dominés */
        k=j;
        f3 = (j) ? pop->item[id2[j-1]].front[ref->id3] : ref->f3 ;
        while (k < prec &&
            pop->item[id2[k]].front[ref->id3] > pop->item[id[i]].front[ref->id3]) {
            f2 = (k+1 == prec) ? ref->f2 : pop->item[id2[k+1]].front[ref->id2];
            surf -= (f2 - pop->item[id2[k]].front[ref->id2]) *
                    (f3 - pop->item[id2[k]].front[ref->id3]);
            k++;
        }
        /* ajout point */
        f2 = (k == prec) ? ref->f2 : pop->item[id2[k]].front[ref->id2];
        surf += (f2 - pop->item[id[i]].front[ref->id2]) *
                (f3 - pop->item[id[i]].front[ref->id3]);
        /* traitement reste */
        un = id2[k]; /* needed if j=k */
        id2[j++] = id[i];
        while (k < prec) {
            k++;
            deux = id2[k];
            id2[j++] = un;
            un = deux;
        }
        /* next */
        prec = j;
        memi = i;
    }
    indic += surf * (ref->f1 - pop->item[id[memi]].front[ref->id1]);
#ifdef TRACE
    printf("final indic %g \n", indic);
#endif
    VZRTAB_OUT(id2);
    VZRTAB_OUT(id);
    return indic;
}
