/* --- Ansi C --- */
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2016-2022 CEA/DES
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   vzr_knee.c
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2016-02-25
 
 \brief  knee point evolutionary algorithm
*/

#include <stdlib.h>
#include <math.h>

#include "vzrsort.h"

#include "vzr_macro.h"
#include "vzr_alloc.h"

//#include "vzr_pair.h"
#include "vzrpop_api.h"
#include "vzrpop_util.h"
#include "vzr_moga.h"

//#define TRACE
#ifdef TRACE
#include <stdio.h>
#endif

/* local defs */

#define PROF 16
#define T_KNEE 0.7

#define VZRTMP -1
#define VZRNOTK 0
#define VZRKNEE 1

struct knee {
    //struct voisin v;
    int prof;
    struct xplan xp;
    double near, taux;
};

/**** utilitaires ****/
/* knee dedicated sort */
struct heap_comp {
    struct vzritem *val;
};

static INLINE int hs_comp(struct heap_comp *hs, int i, int j)
{
    return (
        hs->val[i].rank < hs->val[j].rank
        || (
            hs->val[i].rank == hs->val[j].rank &&
            hs->val[i].info > hs->val[j].info
        )
        || (  
            hs->val[i].rank == hs->val[j].rank &&
            hs->val[i].info == hs->val[j].info &&
            hs->val[i].isol > hs->val[j].isol
        )
    );
}

#include "vizir/vzr_heapsort.h"

int vzr_rank_knee(int size, int top, int *idx, struct vzritem *val)
{
    struct heap_comp comp = {val};
    return hs_rank(idx, size, top, &comp);
}

/* utilitaire */
static void knee_indicator(struct xplan *self, struct vzrpop *pop, int size, double *norm)
{
    int d, i;
    double cum;
    struct vzritem *obj;

    /* plan */
    vzrpop_pareto_plan(pop, size, self, norm);
    
    /* knee indicateur */
    for EACH(i, size) {
        obj = &(pop->item[pop->rg[i]]);
        cum = 0.;
        for EACH(d, self->dim) cum += obj->front[d+pop->valid] * self->plan[d];
        obj->isol = cum;
    }
    
}

/**** protocoles ****/
/* viable
 * using moga_viable
 ****/

/* premier indicateur (rank)
 * using pareto_egal_rank
 ****/

/* second indicateur
 * distance au plan des extremes
 ****/
static void knee_indic(void *info, int size, struct vzrpop *pop)
{
    int i, j, k;
    int nbok, nbko;
    double tauxeff, isol;
    struct knee *self = (struct knee*) info;
    struct voisin *knn = (struct voisin*) pop->info;
    
    /* knee indicator */
    knee_indicator(&self->xp, pop, size, NULL);
    /* TODO voisinage plan */
    //knee_indicator(&self->xp, pop, size, norm);
        
    /* proche voisin */
    vzrpop_proche_voisin(pop, size, knn);  
    //vzrpop_proche_voisin3(pop, size, norm, &self->v);  
    //VZRTAB_OUT(norm);
    
    /* knee mark */
    for EACH(i, size) pop->item[pop->rg[i]].info = VZRTMP;
    nbok = nbko = 0;
    while (nbok+nbko < size) {
        i = 0;
        while (pop->item[pop->rg[i]].info != VZRTMP) i++;
        for FROMEACH(j, i+1, size) {
            if (pop->item[pop->rg[j]].info != VZRTMP) continue;
            if ( pop->item[pop->rg[j]].rank < pop->item[pop->rg[i]].rank || 
                    (
                        pop->item[pop->rg[j]].rank == pop->item[pop->rg[i]].rank && 
                        pop->item[pop->rg[j]].isol < pop->item[pop->rg[i]].isol
                    )
                ) i = j;
        }
        /* knee point */
        pop->item[pop->rg[i]].info = VZRKNEE;
        nbok ++;
        /* not knee point */
        for EACH(j, knn->prof) {
            if (knn->dist[pop->rg[i]][j] < self->near 
                    && pop->item[knn->ref[pop->rg[i]][j]].info == VZRTMP
                ) {
                pop->item[knn->ref[pop->rg[i]][j]].info = VZRNOTK;
                nbko++;
            }
        }
    }
    
    /* mise a jour de la proximite */
    tauxeff = ((double) nbok) / (nbok+nbko);
#ifdef TRACE
    printf("knee %d vs %d = %g : %g\n", nbok, nbko, tauxeff, self->near);
#endif
    self->near *= exp(- (1. - tauxeff / self->taux) / self->xp.dim );
    //self->near *= exp(- (1. - tauxeff / self->taux) );
    
    /* isol distance du k plus proche voisin */
    /* TODO distance pondere */
    for EACH(i, size) {
        isol = 0;
        for EACH(k, knn->prof) {
            isol += knn->dist[pop->rg[i]][k];
        }
        pop->item[pop->rg[i]].isol = isol;
    }
}

/* suivie
 * classement des individus
 ****/
static void knee_survive(void *info, int size, int soon, struct vzrpop *pop)
{
    //int i, j, k;
    //struct knee *self = (struct knee *) info;
    
    /* select */
    // TODO tournoi
    
    /* last sort */
    (void) vzr_rank_knee(size-soon, 0, pop->rg, pop->item);
}

/* dealloc */
static void knee_build(void *info, struct vzrpop *pop)
{
    int i;
    struct knee *dat = (struct knee *) info;
    struct voisin *knn;
    
    /* voisin */
    knn = (struct voisin *) malloc(sizeof(struct voisin));
    knn->prof = dat->prof;
    knn->size = pop->stop + pop->step;
    NEWvecteur(double, knn->norm, pop->all);
    NEWmatrice(int, knn->ref, knn->size, knn->prof);
    NEWmatrice(double, knn->dist, knn->size, knn->prof);
    /* plan */
    dat->xp.dim = pop->cost;
    NEWmatrice(double, dat->xp.rot, dat->xp.dim, dat->xp.dim);
    dat->xp.plan = dat->xp.rot[dat->xp.dim - 1];
    /* pop */
    pop->genre = Knn;
    pop->info = knn;
}

static void knee_dealloc(struct vzrsort *self, struct vzrpop *pop)
{
    struct knee *dat = (struct knee *) self->data;
    struct voisin *knn = (struct voisin *) pop->info;
    
    if (knn->ref) {
        DELvecteur(knn->norm);
        DELmatrice(knn->ref);
        DELmatrice(knn->dist);
        //DELmatrice(dat->xp.rot);
    }
    free(knn);
    free(dat);
    free(self);
}

/**** constructeur ****/
struct vzrsort *create_knee_full(int prof, double taux)
{
    struct knee *dat;
    struct vzrsort *ret;
    
    /* data */
    dat = (struct knee *) malloc(sizeof(struct knee));
    dat->prof = prof;
    /* plan */
    dat->xp.dim = 0;
    dat->xp.rot = NULL;
    dat->xp.plan = NULL;
    /* knee */
    dat->taux = taux;
    dat->near = 1.;
    
    /* interface */
    ret = (struct vzrsort *) malloc(sizeof(struct vzrsort));
    ret->data = (void *) dat;
    ret->build = &knee_build;
    ret->valid = &moga_viable;
    ret->rank = &pareto_egal_rank;
    ret->isol = &knee_indic;
    ret->survive = &knee_survive;
    ret->dealloc = &knee_dealloc;
    
    return ret;
}

struct vzrsort *create_knee()
{
    return create_knee_full(PROF, T_KNEE);
}
