/* --- Ansi C --- */
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2017-2022 CEA/DES
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   vzr_modal.c
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2017-08-25
 
 \brief  recherche de mode
*/

#include <stdlib.h>

#include "vzrsort.h"

#include "vzr_macro.h"
#include "vzr_alloc.h"

#include "vzrpop_api.h"
#include "vzrpop_util.h"
#include "vzr_pareto.h"
#include "vzr_moga.h"
#include "vzr_pair.h"

//#define TRACE
#ifdef TRACE
#include <stdio.h>
#endif

/* local def */
#define OLD
#define PROF 16

/*
struct modal {
    int prof;
};
*/

/**** utilitaires ****/
/* modal dedicated sort */
struct heap_comp {
    struct vzritem *val;
};

static INLINE int hs_comp(struct heap_comp *hs, int i, int j)
{
    return (
        hs->val[i].rank < hs->val[j].rank
        || (
            hs->val[i].rank == hs->val[j].rank &&
            hs->val[i].info < hs->val[j].info
        )
        || (  
            hs->val[i].rank == hs->val[j].rank &&
            hs->val[i].info == hs->val[j].info &&
            hs->val[i].isol > hs->val[j].isol
        )
    );
}

#include "vizir/vzr_heapsort.h"

static int vzr_rank_modal(int size, int top, int *idx, struct vzritem *val)
{
    struct heap_comp comp = {val};
    return hs_rank(idx, size, top, &comp);
}


/**** protocole ****/
/* viable
 * always viable
 ****/
static int modal_viable(void *info, struct vzritem *item, struct vzrpop *pop)
{
    return 1;
}

/* calcul du rang
 * dominance dans le voisinage
 * TODO utilisation pareto_egal_rank pour egalité
 ****/
 

/* isolement
 ****/
static void modal_isol(void *info, int size, struct vzrpop *pop)
{
    int ii, i, pp, p;
    struct voisin *knn = (struct voisin *) pop->info;
    
    /* proche */
#ifdef TRACE
    printf("dup %d\n", size);
#endif
    vzrpop_proche_individu(pop, size, knn);
    
    /* local dominance */
    for EACH(ii, size) {
        i = pop->rg[ii];
        pop->item[i].info = knn->prof;
        //pop->item[i].isol = 0.0;
        for EACH(pp, knn->prof) {
            p = knn->ref[i][pp];
            switch ((*pop->darwin)(pop, pop->item[i].front, pop->item[p].front)) {
            case ONE_DOM :
                pop->item[i].info -= 1;
            case TWO_DOM :
                break;
            case EGAL :
                // TODO
            case BOTH_DOM :
                break;
            }
        }
    }
    
    /* local dominance */
    for EACH(ii, size) {
        i = pop->rg[ii];
        pop->item[i].rank = pop->item[i].info;
        for EACH(pp, knn->prof) {
            p = knn->ref[i][pp];
            if ( pop->item[i].rank > pop->item[p].info ) {
                pop->item[i].rank = pop->item[p].info ;
            }
        }
    }
    
    /* isolement */
    for EACH(ii, size) {
        i = pop->rg[ii];
        pop->item[i].isol = knn->dist[i][0];
    }
}

/* survie
 * use standard survie
 ****/
#ifdef OLD
static void modal_survive(void *info, int size, int soon, struct vzrpop *pop)
{
    //int i, j, k;
    //struct knee *self = (struct knee *) info;
    
    /* select */
    // TODO tournoi
    
    /* last sort */
    (void) vzr_rank_modal(size-soon, 0, pop->rg, pop->item);
}

#else

static void modal_pair_survive(void *info, int size, int soon, struct vzrpop *self)
{
#ifdef V0
    int imax, imin, itemp, i, j;
#else
    int imax, imin, itemp;
#endif
    struct local *loc = (struct local *) info;
    VZRTAB_IN(int, simil);

    /*** phase 1 : élimination par fitness ***/
    /* premier classement par rang puis isolement*/
    (void) vzr_rank_modal(size-soon, 0, self->rg, self->item);

#ifdef V0 // version moit moit
    /* recherche seuil fitness/isol */
    imin = size - self->step ;
    imax = (soon >= self->step) ? imin : size - soon;
    itemp = (imin+imax+1) / 2;
    for (i=itemp; i > imin &&
            self->item[self->rg[i-1]].rank == self->item[self->rg[itemp]].rank &&
            self->item[self->rg[i-1]].info == self->item[self->rg[itemp]].info ;
         i--);
    for (j=itemp; j < imax &&
            self->item[self->rg[j+1]].rank == self->item[self->rg[itemp]].rank &&
            self->item[self->rg[j+1]].info == self->item[self->rg[itemp]].info ;
         j++);
    imax = (itemp-i > j-itemp)? j : i ;
#else // version last class
    /* recherche seuil fitness/isol */
    imin = size - self->step - 1;
    imax = (soon >= self->step) ? imin : size - soon;
    for (itemp = imin+1; itemp < imax &&
            self->item[self->rg[itemp]].rank == self->item[self->rg[imin]].rank &&
            self->item[self->rg[itemp]].info == self->item[self->rg[imin]].info ;
         itemp++);
    imax = itemp ;
#endif
         
#ifdef TRACE    
    printf("pair dupli %d simil %d / %d\n", soon, imax-imin-1, self->step); 
#endif

    /*** phase 2 : élimination par similarité ***/
    /* apairage des points */
    pair_affectation((struct voisin *) self->info, loc, imax, self) ;    
    /* classement par similarités en V */
    VZRTAB_NEW(int, simil, loc->done);
    modal_vorder(loc, imax, self, simil);
    
    /*** phase 3 : last sort ***/
    (void) vzr_rank_modal(size - self->step, 0, self->rg, self->item);
#ifdef TRACE
    if (imax == self->level + self->step) {
        for EACH(itemp, self->level) {
            if (itemp%20==19) printf("%d\n", self->item[self->rg[itemp]].rank);
            else printf("%d  ", self->item[self->rg[itemp]].rank);
        }
    }
#endif
    /*** finalize : maj voisin ***/
    //vzrpop_reorder_voisin(self, size, idx, (struct voisin *) loc);
	VZRTAB_OUT(simil);
	

}
#endif
 
/* dealloc */
static void modal_build(void *info, struct vzrpop *pop)
{
    int i, siz;
    struct local *dat = (struct local *) info;
    struct voisin *knn;
    
    /* voisin */
    knn = (struct voisin *) malloc(sizeof(struct voisin));
    knn->prof = dat->prof;
    knn->size = pop->stop + pop->step;
    knn->norm = NULL;
    NEWmatrice(int, knn->ref, knn->size, knn->prof);
    NEWmatrice(double, knn->dist, knn->size, knn->prof);
    
    siz = (knn->size+1)/2;
    NEWvecteur(struct pair, dat->fit, siz);
    
    /* pop */
    pop->genre = Knn;
    pop->info = knn;
}

static void modal_dealloc(struct vzrsort *self, struct vzrpop *pop)
{
    struct local *dat = (struct local *) self->data;
    struct voisin *knn = (struct voisin *) pop->info;
    
    if (knn->ref) {
        DELmatrice(knn->ref);
        DELmatrice(knn->dist);
        DELvecteur(dat->fit);
    }
    free(knn);
    free(dat);
    free(self);
}

/**** constructeur ****/
struct vzrsort *create_modal_full(int prof)
{
    struct local *dat;
    struct vzrsort *ret;
    
    /* data */
    dat = (struct local *) malloc(sizeof(struct local));
    dat->prof = prof;
    
    /* interface */
    ret = (struct vzrsort *) malloc(sizeof(struct vzrsort));
    ret->data = (void*) dat;
    ret->build = &modal_build;
    ret->valid = &modal_viable;
    ret->rank = &pareto_egal_rank;
    ret->isol = &modal_isol;
#ifdef OLD
    ret->survive = &modal_survive;
#else
    ret->survive = &modal_pair_survive;
#endif
    ret->dealloc = &modal_dealloc;
    
    return ret;
}

struct vzrsort *create_modal()
{
    return create_modal_full(PROF);
}

