/* --- Ansi C --- */
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2016-2022 CEA/DES
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   vzr_moeadd.c
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2016-05-26
 
 \brief  methode MOEA/DD
*/

#include "vzrsort.h"

#include <stdlib.h>
#include <math.h>

#include "vzr_macro.h"
#include "vzr_alloc.h"

#include "vzr_moeadd.h"
#include "vzrpop_api.h"
#include "vzrpop_util.h"
#include "vzr_moga.h"
#include "vzr_heap.h"

#define THETA 5.
#define KPPV_MOEA 20
#define H_MOEA 7

/* utilitaires */
struct heap_comp {
    struct vzritem *val;
};

static INLINE int hs_comp(struct heap_comp *hs, int i, int j)
{
    //return (hs->val[i].rank < hs->val[j].rank);
    return (hs->val[i].rank < hs->val[j].rank ||
            ( 
                hs->val[i].rank == hs->val[j].rank &&
                hs->val[i].isol < hs->val[j].isol
            )
        );
}

#include "vizir/vzr_heapsort.h"

int vzr_rank_item_pgi(int size, int top, int *idx, struct vzritem *val)
{
    struct heap_comp comp = {val};
    return hs_rank(idx, size, top, &comp);
}

/*** util ***/
static int fill_pond(double **tab, int lig, int col0, int coln, int rest)
{
    int i, j, cum, done;
    
    if (rest == 0) {
        for (i=col0; i<coln+1; i++) tab[lig][i] = 0.;
        return 1;
    }
    else if (col0 == coln) {
        tab[lig][coln] = rest;
        return 1;
    }
    else {
        cum = 0;
        for (i=0; i<rest+1; i++) {
            done = fill_pond(tab, lig+cum, col0+1, coln, rest-i);
            for (j=0; j<done; j++) tab[lig+cum+j][col0] = i;
            cum += done;
        }
        return cum;
    }
}

static void proche_voisin_dir(struct voisin *self, int nbobj, double **tab )
{
    int i, j, k;
    double di, tmp;
    VZRTAB(char, fill, self->size);

    /* initialisation */
    for EACH(i, self->size) fill[i] = 0;

    /* calcul des couts (distances) */
    for FROMEACH(i, 1, self->size) {
        for EACH(j, i) {
            di = 0.;
            for EACH(k, nbobj) {
                tmp = tab[i][k] - tab[j][k];
                di += tmp * tmp;
            }
            fill_voisin(self, fill, i, j, di);
        }
    }
#ifdef TRACE2
    for EACH(i, siz) {
        printf("[%d, %d/%d]", i, fill[i], self->prof);
        for EACH(j, self->prof) printf("%d %g,", self->ref[i][j], self->dist[i][j]);
        printf("\n");
    }
#endif
    VZRTAB_OUT(fill);
}
  
static void build_moeadd(void *info, struct vzrpop *pop)
//int nbobj, int nbcut1, int nbcut2, int nbvois)
{
    struct moeadd *ret = (struct moeadd *) info;
    struct ddknn_info *dat;
    int i, j, outsiz, siz;
    double norm;
    
    //dat->inner = NULL;
    /* nombre de directions */
    outsiz = moeadd_pop_size(pop->cost, ret->cut1, 0);
    siz = moeadd_pop_size(pop->cost, 0, ret->cut2) + outsiz;
    //printf("MOEADD %d \n", a);
        
    /* initialisation */
    dat = (struct ddknn_info *) malloc(sizeof(struct ddknn_info));
    dat->near.size = siz;
    dat->near.prof = ret->prof;
    //ret.near.prof = nbvois;
    NEWmatrice(int, dat->near.ref, siz, dat->near.prof);
    NEWmatrice(double, dat->near.dist, siz, dat->near.prof);
    NEWvecteur(int, dat->inner, siz);
    pop->genre = Ddknn;
    pop->info = (void *) dat;
    
    ret->objsize = pop->cost;
    NEWvecteur(double, ret->pgi, siz);
    //for EACH(j, a) ret.inner[i] = 0;
    NEWmatrice(double, ret->weight, siz, pop->cost);
    /* vecteurs de poids */
    if (ret->cut1) fill_pond(ret->weight, 0, 0, pop->cost-1, ret->cut1);
    if (ret->cut2) fill_pond(ret->weight+outsiz, 0, 0, pop->cost-1, ret->cut2);
    /* inside layer transformation */
    for FROMEACH(i, outsiz, siz) {
        for EACH(j, pop->cost) ret->weight[i][j] += ret->weight[i][j] + 1;
    }
    /* normalisation */
    for EACH(i, siz) {
        norm = 0.;
        for EACH(j, pop->cost) norm += ret->weight[i][j] * ret->weight[i][j];
        norm = sqrt(norm);
        for EACH(j, pop->cost) ret->weight[i][j] /= norm;
    }
    /* voisinage */
    proche_voisin_dir(&dat->near, pop->cost, ret->weight);
    
}

/*** protocole ***/
/* viable
 * using moga_viable
 ****/

/* premier indicateur (rank)
 * using pareto_egal_rank
 ****/

/* second indicateur
 ****/
static void moeadd_indic(void *info, int size, struct vzrpop *pop)
{
    int i, ii, d, k;
    double tmp, d1, d2, pbi;
    struct moeadd *self = (struct moeadd *) info;
    struct ddknn_info *knn = (struct ddknn_info *) pop->info;
    VZRTAB(double, ideal, pop->cost);
    VZRTAB(double, norm, pop->cost); //self->near.norm unsize
    
    /* point ideal */
    vzrpop_ideal_it(pop, size, ideal, norm);
    
    /* PBI indicator */
    for EACH(ii, size) {
        i = pop->rg[ii];
        for EACH(d, knn->near.size) {
            d1 = d2 = 0.;
            for EACH(k, pop->cost) {
                tmp = (pop->item[i].front[k+pop->valid] - ideal[k]) 
                        * self->weight[d][k] / norm[k];
                d1 += tmp*tmp;
            }
            d1 = sqrt(d1);
            for EACH(k, pop->cost) {
                tmp = (pop->item[i].front[k+pop->valid] - 
                        ( ideal[k] + d1 * self->weight[d][k] * norm[k])) 
                        / norm[k];
                d2 += tmp*tmp;
            }
            pbi = d1 + THETA * sqrt(d2);
            if (!d || pbi < pop->item[i].isol) {
                pop->item[i].group = d;
                pop->item[i].isol = pbi;
            }
        }
    }
    
    /* cleanup */
    VZRTAB_OUT(norm);
    VZRTAB_OUT(ideal);
}

/* classement des individus
 ****/
static void moeadd_trie(void *info, int size, int soon, struct vzrpop *pop)
{
    int i, ii, isel, d, lev, top, grp;
    double isol;
    struct moeadd *self = (struct moeadd *) info;
    struct ddknn_info *knn = (struct ddknn_info *) pop->info;
    
    /* classement standard */
    (void) vzr_rank_item_pgi(size-soon, 0, pop->rg, pop->item);
    
    top = size-soon-1;
    if (size-soon > pop->stop) {
        /* cuting rank level */
        lev = pop->item[pop->rg[pop->stop-1]].rank;
        while (pop->item[pop->rg[top]].rank != lev) top--;
    }
    
    if (top >= pop->stop-1) {
        /* densites des regions */
        for EACH(d, knn->near.size) {
            knn->inner[d] = 0;
            self->pgi[d] = 0.;
        }
        for EACH(i, top+1) {
            knn->inner[pop->item[pop->rg[i]].group]++;
            self->pgi[pop->item[pop->rg[i]].group] += pop->item[pop->rg[i]].isol;
        }
    
        /* boucle retrait dans les regions denses */
        for (;top >= pop->stop; top--) {
            /* choix region */
            grp = 0;
            for FROMEACH(d, 1, knn->near.size) {
                if (knn->inner[d] > knn->inner[grp]) grp = d;
                else if (knn->inner[d] == knn->inner[grp] &&
                        self->pgi[d] > self->pgi[grp]) grp = d;
            }
            
            /* choix individu dans region */
            i = top;
            while (pop->item[pop->rg[i]].group != grp) i--;

            /* retrait */
            knn->inner[grp]--;
            self->pgi[grp] -= pop->item[pop->rg[i]].isol;
            
            isel = pop->rg[i];
            for (ii=i; ii<top; ii++) pop->rg[ii] = pop->rg[ii+1];
            pop->rg[top] = isel;
        }
        
        /* indicateur isolement */
        for EACH(d, knn->near.size) {
            isol = 1. + 10. * knn->inner[d];
            for EACH(i, knn->near.prof) {
                isol += knn->inner[knn->near.ref[d][i]];
            }
            self->pgi[d] = (1. + 10. + knn->near.prof) / isol;
        }
        for EACH(i, pop->stop) 
            pop->item[pop->rg[i]].isol = self->pgi[pop->item[pop->rg[i]].group];
        
        /* last sort */
        (void) vzr_rank_item(pop->stop, 0, pop->rg, pop->item);
    }

}

/* dealloc */
static void moeadd_dealloc(struct vzrsort *self, struct vzrpop *pop)
{
    struct moeadd *dat = (struct moeadd *) self->data;
    struct ddknn_info *info = (struct ddknn_info *) pop->info;
    
    DELmatrice(info->near.ref);
    DELmatrice(info->near.dist);
    DELvecteur(info->inner);
    DELvecteur(dat->pgi);
    DELmatrice(dat->weight);
    free(dat);
    free(info);
    free(self);
}

/*** constructeur ***/
int moeadd_pop_size(int nbobj, int nbcut1, int nbcut2)
{
    int i, j, a, ret, nom, den, nbcut[2] = {nbcut1, nbcut2};
    
    /* nombre de directions */
    ret = 0;
    for EACH(i, 2) {
        if (nbcut[i] == 0) continue;
        a = 1;
        nom = nbcut[i] + nbobj - 1;
        den = 1;
        for EACH(j, nbobj-1) {
            a = (a * nom) / den;
            nom--;
            den++;
        }
        ret += a;
    }
    return ret;
}

struct vzrsort *create_moeadd_full(int outsize, int insize, int nearsize)
{
    struct moeadd *dat;
    struct vzrsort *ret;
    
    /* data */
    dat = (struct moeadd *) malloc(sizeof(struct moeadd));
    /* initialisation */
    dat->cut1 = outsize;
    dat->cut2 = insize;
    dat->objsize = 0;
    dat->prof = nearsize;
    dat->pgi = NULL;
    dat->weight = NULL;
    
    /* interface */
    ret = (struct vzrsort *) malloc(sizeof(struct vzrsort));
    ret->data = (void*) dat;
    ret->build = &build_moeadd;
    ret->valid = &moga_viable;
    ret->rank = &pareto_egal_rank;
    ret->isol = &moeadd_indic;
    ret->survive = &moeadd_trie;
    ret->dealloc = &moeadd_dealloc;
    
    return ret;
}

struct vzrsort *create_moeadd(int outsize, int insize)
{
    return create_moeadd_full(outsize, insize, KPPV_MOEA);
}
