/* --- Ansi C --- */
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2008-2022 CEA/DES
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   vzr_moga.c
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2008-04-03
 
 \brief  operateur de survie utilisant le rang de pareto
*/

#include <stdlib.h>
#include <stdio.h>

#include "vzr_moga.h"
#include "vzrsort.h"

#include "vzr_macro.h"

#include "vzr_pareto.h"
#include "vzr_aleas.h"
#include "vzr_survie.h"
#include "vzr_heap.h"

/* utilitaires */
struct heap_comp {
    int id;
    struct vzritem *val;
};

static INLINE int hs_comp(struct heap_comp *hs, int i, int j)
{
    return (hs->val[i].front[hs->id] < hs->val[j].front[hs->id] );
}

#include "vzr_heapsort.h"

static int vzr_rank_crowd(int size, int ref, int *idx, struct vzritem *val)
{
    struct heap_comp comp = {ref, val};
    return hs_rank(idx, size, 0, &comp);
}

/**** fonction du protocole ****/
/** build **/
void moga_build(void *info, struct vzrpop *pop) {}

/** viable **/
int moga_viable(void *info, struct vzritem *item, struct vzrpop *self)
{
    int j, rank;
    
    rank = 0;
    for EACH(j, self->level) {
        if ((*self->darwin)(self, item->front, self->item[self->rg[j]].front) == TWO_DOM)
            rank++;
    }
    item->rank = rank;

    return (self->level == self->stop && self->item[self->rg[self->level-1]].rank != -1) ? 
            rank <= self->item[self->rg[self->level-1]].rank : 1;
}

int spea_viable(void *info, struct vzritem *item, struct vzrpop *self)
{
    int j, rank;
    
    rank = 0;
    for EACH(j, self->level) {
        if ((*self->darwin)(self, item->front, self->item[self->rg[j]].front) == TWO_DOM)
            rank += self->item[self->rg[j]].info ;
    }
    item->rank = rank;

    return (self->level == self->stop && self->item[self->rg[self->level-1]].rank != -1) ? 
            rank <= self->item[self->rg[self->level-1]].rank : 1;
}

/** calcul du rang **/
void pareto_rank(void *info, int size, struct vzrpop *pop)
{
    int i,j;

    for EACH(i, size) {
        pop->item[i].rank = 0;
        pop->item[i].isol = 0.0;
    }

    for FROMEACH(i, 1, size) {
        for EACH(j, i) {
            switch ((*pop->darwin)(pop, pop->item[i].front, pop->item[j].front)) {
            case ONE_DOM:
                pop->item[j].rank += 1;
                break;
            case TWO_DOM:
                pop->item[i].rank += 1;
            case EGAL:
            case BOTH_DOM:
                break;
            }
        }
    }
}

void pareto_egal_rank(void *info, int size, struct vzrpop *pop)
{
    int i,j,k;
    VZRTAB(short, tmp, size);
    
    for EACH(i, size) {
        pop->item[i].rank = 0;
        pop->item[i].isol = 0.0;
    }

    for FROMEACH(i, 1, size) {
        k = 0;
        for EACH(j, i) {
            if (pop->item[j].rank == -1) continue;
            switch ((*pop->darwin)(pop, pop->item[i].front, pop->item[j].front)) {
            case ONE_DOM:
                tmp[k++] = j;
                break;
            case TWO_DOM:
                pop->item[i].rank += 1;
                break;
            case EGAL:
                pop->item[i].rank = -1;
            case BOTH_DOM:
                break;
            }
            if (pop->item[i].rank == -1) break;
        }
        if (pop->item[i].rank != -1) {
            for EACH(j, k) pop->item[tmp[j]].rank += 1;
        }
    }
    VZRTAB_OUT(tmp);
}

void spea_rank(void *info, int size, struct vzrpop *pop)
{
    int i,j,k;
    VZRTAB(short, tmp, size);
    //short force[size];
    
    for EACH(i, size) {
        pop->item[i].info = pop->item[i].rank = 0;
        pop->item[i].isol = 0.0;
    }

    /* calcul de la force */
    for FROMEACH(i, 1, size) {
        k = 0;
        for EACH(j, i) {
            if (pop->item[j].info == -1) continue;
            switch ((*pop->darwin)(pop, pop->item[i].front, pop->item[j].front)) {
            case ONE_DOM:
                pop->item[i].info += 1;
                break;
            case TWO_DOM:
                tmp[k++] = j;
                break;
            case EGAL:
                pop->item[i].info = -1;
            case BOTH_DOM:
                break;
            }
            if (pop->item[i].info == -1) break;
        }
        if (pop->item[i].info != -1) {
            for EACH(j, k) pop->item[tmp[j]].info += 1;
        }
    }
    VZRTAB_OUT(tmp);
    
    /* calcul de l'indicateur */
    for FROMEACH(i, 1, size) {
        if (pop->item[i].info == -1) {
            pop->item[i].rank = -1;
            continue;
        }
        for EACH(j,i) {
            if (pop->item[j].info == -1) continue;
            switch ((*pop->darwin)(pop, pop->item[i].front, pop->item[j].front)) {
            case ONE_DOM :
                pop->item[j].rank += pop->item[i].info;
                break;
            case TWO_DOM :
                pop->item[i].rank += pop->item[j].info;
                break;
            default :
                break;
            }
        }
    }
}

static void strength_rank(void *info, int size, struct vzrpop *pop)
{
    int i,j,k,l;
    VZRTAB(short, tmp, size);

    /* init */
    for EACH(i, size) {
        pop->item[i].info = pop->item[i].rank = 0;
        pop->item[i].isol = 0.0;
    }
    /* loop */
    for FROMEACH(i, 1, size) {
        k = l = 0;
        for EACH(j, i) {
            switch ((*pop->darwin)(pop, pop->item[i].front, pop->item[j].front)) {
            case ONE_DOM:
                tmp[k++] = j;
                pop->item[i].info += 1;
                break;
            case TWO_DOM:
                tmp[k++] = j-i;
                pop->item[i].rank += 1;
                break;
            case EGAL :
                pop->item[i].rank = -1;
                pop->item[i].info = 0;
            case BOTH_DOM :
                break;
            }
            if (pop->item[i].rank == -1) break;
        }
        if (pop->item[i].rank != -1) {
            for EACH(j, k) {
                if (tmp[j] < 0) pop->item[tmp[j]+i].info += 1; 
                else pop->item[tmp[j]].rank += 1;
            }
        }
    }
    VZRTAB_OUT(tmp);
}

/* calcul de l'indicateur d'isolement */
static void random_isol(void *info, int size, struct vzrpop *pop)
{
    int i;

    for EACH(i, size) {
        pop->item[pop->rg[i]].isol = mt_real1();
    }
}

static void strength_isol(void *info, int size, struct vzrpop *pop)
{
    int i;
    struct vzritem *ind;

    /* isol */
    for EACH(i, size) {
        ind = &(pop->item[pop->rg[i]]);
        ind->isol = 1. / (1. + ind->info + ind->rank);
    }
}

static void crowd_isol(void *info, int size, struct vzrpop *pop)
{
    int i, j;

    VZRTAB(int, work, size);
    double plage, diff;

    /* calcul du rang */
    for EACH(i, size) work[i] = pop->rg[i];

    /* calcul de la crowding distance */
    for EACH(i, pop->all) {
        vzr_rank_crowd(size, i, work, pop->item);
        plage = pop->item[work[size-1]].front[i] - pop->item[work[0]].front[i];
        /* bounds */
        pop->item[work[size-1]].isol += 1.0;
        pop->item[work[0]].isol += 1.0;
        /* others */
        for FROMEACH(j, 1, size) {
            diff = (pop->item[work[j]].front[i] - pop->item[work[j-1]].front[i]) / plage;
            pop->item[work[j-1]].isol += diff;
            pop->item[work[j]].isol += diff;
        }
    }
	VZRTAB_OUT(work);
}

/* dealloc */
static void dealloc(struct vzrsort *self, struct vzrpop *pop)
{
    free(self);
}

/** constructeur **/
struct vzrsort *create_moga()
{
    struct vzrsort *ret;

    ret = (struct vzrsort *) malloc(sizeof(struct vzrsort));
    ret->data = NULL;
    ret->build = &moga_build;
    ret->valid = &moga_viable;
    ret->rank = &pareto_rank;
    ret->isol = &random_isol;
    ret->survive = &pareto_survive;
    ret->dealloc = &dealloc;
    return ret;
}

struct vzrsort *create_moga2()
{
    struct vzrsort *ret;

    ret = (struct vzrsort *) malloc(sizeof(struct vzrsort));
    ret->data = NULL;
    ret->build = &moga_build;
    ret->valid = &moga_viable;
    ret->rank = &pareto_egal_rank;
    ret->isol = &random_isol;
    ret->survive = &pareto_survive;
    ret->dealloc = &dealloc;
    return ret;
}

struct vzrsort *create_crowding()
{
    struct vzrsort *ret;

    ret = (struct vzrsort *) malloc(sizeof(struct vzrsort));
    ret->data = NULL;
    ret->build = &moga_build;
    ret->valid = &moga_viable;
    ret->rank = &pareto_egal_rank;
    ret->isol = &crowd_isol;
    ret->survive = &pareto_survive;
    ret->dealloc = &dealloc;
    return ret;
}

struct vzrsort *create_strength()
{
    struct vzrsort *ret;

    ret = (struct vzrsort *) malloc(sizeof(struct vzrsort));
    ret->data = NULL;
    ret->build = &moga_build;
    ret->valid = &moga_viable;
    ret->rank = &strength_rank;
    ret->isol = &strength_isol;
    ret->survive = &pareto_survive;
    ret->dealloc = &dealloc;
    return ret;
}

struct vzrsort *create_spea_rand()
{
    struct vzrsort *ret;

    ret = (struct vzrsort *) malloc(sizeof(struct vzrsort));
    ret->data = NULL;
    ret->build = &moga_build;
    ret->valid = &spea_viable;
    ret->rank = &spea_rank;
    ret->isol = &random_isol;
    ret->survive = &pareto_survive;
    ret->dealloc = &dealloc;
    return ret;
}

struct vzrsort *create_spea_crowd()
{
    struct vzrsort *ret;

    ret = (struct vzrsort *) malloc(sizeof(struct vzrsort));
    ret->data = NULL;
    ret->build = &moga_build;
    ret->valid = &spea_viable;
    ret->rank = &spea_rank;
    ret->isol = &crowd_isol;
    ret->survive = &pareto_survive;
    ret->dealloc = &dealloc;
    return ret;
}

