/* --- Ansi C --- */
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2008-2022 CEA/DES
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   vzr_pareto.c
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2008-04-03
 
 \brief  survie dans le cas multi contraint (version oep)
*/


#include "vzr_pareto.h"

#include "vzrpop_api.h"

/** utilitaire **/
/* calcul de la dominance
 * cas mono critére
 * */
int mono_rank(struct vzrpop *viz, double *one, double *two)
{
    int   mark;

    if (one[0] == two[0]) mark = EGAL;
    else if (one[0] - two[0] > viz->precis) mark = TWO_DOM;
    else if (two[0] - one[0] > viz->precis) mark = ONE_DOM;
    else mark = BOTH_DOM;
    
    return mark;
}


/* calcul de la dominance
 * cas multi critére
 * */
int multi_rank(struct vzrpop *viz, double *one, double *two)
{
    int   mark;
    int     i;

    mark = EGAL;
    for (i=0; i<viz->cost; i++) {
        if (two[i] < one[i] ) mark |= TWO_DOM;
        else if ( one[i] < two[i]) mark |= ONE_DOM;
        if (mark == BOTH_DOM) break;
    }
    return mark;
}
   
/* calcul de la dominance
 * cas multi critére avec contrainte (face admissible)
 * */
int multi_strict_rank(struct vzrpop *viz, double *one, double *two)
{
    int   mark;
    int     i;

    mark = DIFF;
    for (i=0; i<viz->valid; i++) {
        if ( one[i] > two[i] ) {
            if (one[i] > 0) mark |= TWO_DOM;
            mark &= ~DIFF;
        }
        else if ( two[i] > one[i]) {
            if (two[i] > 0) mark |= ONE_DOM;
            mark &= ~DIFF;
        }
        if (mark == BOTH_DOM) break;
    }
    if (!(mark & BOTH_DOM)) {
        for (i=viz->valid; i<viz->valid+viz->cost; i++) {
            if (two[i] < one[i]) mark |= TWO_DOM;
            else if (one[i] < two[i]) mark |= ONE_DOM;
            if (mark == BOTH_DOM) break;
        }
        if (mark == EGAL) mark = BOTH_DOM;
        mark &= ~DIFF;
    }
    return mark;
}

/* calcul de la dominance
 * cas single critére avec contrainte (face admissible)
 * */
int mono_strict_rank(struct vzrpop *viz, double *one, double *two)
{
    int   mark;
    int     i;

    mark = DIFF;
    for (i=0; i<viz->valid; i++) {
        if ( one[i] > two[i] ) {
            if (one[i] > 0) mark |= TWO_DOM;
            mark &= ~DIFF;
        }
        else if ( two[i] > one[i]) {
            if (two[i] > 0) mark |= ONE_DOM;
            mark &= ~DIFF;
        }
        if (mark == BOTH_DOM) break;
    }
    if (!(mark & BOTH_DOM)) {
        // i=viz->valid; i<viz->valid+viz->cost; i++) {
        if (one[i] == two[i] && mark & DIFF) mark = EGAL; 
        else if (one[i] - two[i] > viz->precis) mark = TWO_DOM;
        else if (two[i] - one[i] > viz->precis) mark = ONE_DOM;
        else mark = BOTH_DOM ;
    }
    return mark;
}

/* calcul de la dominance
 * cas multi critére avec contrainte (face contrainte)
 * */
int multi_soft_rank(struct vzrpop *viz, double *one, double *two)
{
    int mark, admis;
    int i;
    
    /* test admissible */
    admis = DIFF ;
    for (i=0; i<viz->valid; i++) {
        if (one[i] > 0.0) admis |= ONE_DOM;
        if (two[i] > 0.0) admis |= TWO_DOM;
        if (one[i] != two[i]) admis &= ~DIFF;
        //if (admis == BOTH_DOM) break;
    }
    if (admis & DIFF) {
        for (i=viz->valid; i<viz->valid+viz->cost; i++) {
            if (one[i] != two[i]) {
                admis &= ~DIFF;
                break;
            }
        }
    }
    /* dominance */
    if (admis & DIFF) mark = EGAL;
    else {
        switch (admis) {
        case BOTH_DOM :
            mark = EGAL;
            for (i=0; i<viz->valid; i++) {
                if ( one[i] > 0 && one[i] > two[i] ) mark |= TWO_DOM;
                else if ( two[i] > 0 && two[i] > one[i]) mark |= ONE_DOM;
                if (mark == BOTH_DOM) break;
            }
            if (mark == EGAL) mark = BOTH_DOM;
            break;
        case EGAL:
            mark = EGAL;
            for (i=viz->valid; i<viz->valid+viz->cost; i++) {
                if (two[i] < one[i]) mark |= TWO_DOM;
                else if (one[i] < two[i]) mark |= ONE_DOM;
                if (mark == BOTH_DOM) break;
            }
            if (mark == EGAL) mark = BOTH_DOM;
            break;
        case ONE_DOM:
            mark = TWO_DOM;
            for (i=viz->valid; i<viz->valid+viz->cost; i++) {
                if (one[i] < two[i]) {
                    mark |= ONE_DOM;
                    break;
                }
            }
            break;
        case TWO_DOM :
            mark = ONE_DOM;
            for (i=viz->valid; i<viz->valid+viz->cost; i++) {
                if (two[i] < one[i]) {
                    mark |= TWO_DOM;
                    break;
                }
            }
        }
    }
    return mark;
}

/* calcul de la dominance
 * cas single critére avec contrainte (face contrainte)
 * */
int mono_soft_rank(struct vzrpop *viz, double *one, double *two)
{
    int mark, admis;
    int i, id;
    
    id = viz->valid;
    /* test admissible */
    admis = DIFF ;
    for (i=0; i<viz->valid; i++) {
        if (one[i] > 0.0) admis |= ONE_DOM;
        if (two[i] > 0.0) admis |= TWO_DOM;
        if (one[i] != two[i]) admis &= ~DIFF;
        //if (admis == BOTH_DOM) break;
    }
    if (admis & DIFF) {
        for (i=viz->valid; i<viz->valid+viz->cost; i++) {
            if (one[i] != two[i]) {
                admis &= ~DIFF;
                break;
            }
        }
    }
    /* dominance */
    if (admis & DIFF) mark = EGAL;
    else {
        switch (admis) {
        case BOTH_DOM :
            mark = EGAL;
            for (i=0; i<id; i++) {
                if ( one[i] > 0 && one[i] > two[i] ) mark |= TWO_DOM;
                else if ( two[i] > 0 && two[i] > one[i]) mark |= ONE_DOM;
                if (mark == BOTH_DOM) break;
            }
            break;
        case EGAL:
            mark = EGAL;
            if (two[id] - one[id] > viz->precis ) mark |= ONE_DOM;
            else if (one[id] - two[id] > viz->precis) mark |= TWO_DOM;
            break;
        case ONE_DOM:
            mark = TWO_DOM;
            if (two[id] - one[id] > viz->precis ) mark |= ONE_DOM;
            break;
        case TWO_DOM :
            mark = ONE_DOM;
            if (one[id] - two[id] > viz->precis) mark |= TWO_DOM;
            break;
        }
    }
    return mark;
}

/* calcul de la dominance
 * cas admissiblité
 * */
int admissible_rank(struct vzrpop *viz, double *one, double *two)
{
    int mark, admis;
    int i, id;
    
    id = viz->valid;
    /* test admissible */
    admis = DIFF ;
    for (i=0; i<viz->valid; i++) {
        if (one[i] > 0.0) admis |= ONE_DOM;
        if (two[i] > 0.0) admis |= TWO_DOM;
        if (one[i] != two[i]) admis &= ~DIFF;
    }
    /* dominance */
    if (admis & DIFF) mark = EGAL;
    else {
        switch (admis) {
            case BOTH_DOM :
                mark = EGAL;
                for (i=0; i<id; i++) {
                    if ( one[i] > 0 && one[i] >= two[i] ) mark |= TWO_DOM;
                    else if ( two[i] > 0 && two[i] >= one[i]) mark |= ONE_DOM;
                    if (mark == BOTH_DOM) break;
                }
                break;
            case EGAL:
                mark = BOTH_DOM;
                break;
            case ONE_DOM:
                mark = TWO_DOM;
                break;
            case TWO_DOM :
                mark = ONE_DOM;
                break;
        }
    }
    return mark;
}

