/* --- Ansi C --- */
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   vzrpop_api.c
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2013-09-24
 
 \brief  gestion des populations
*/

#include <stdlib.h>
#include <stdio.h>

#include "vizir/vzrpop_api.h"

#include "vizir/vzr_macro.h"
#include "vizir/vzr_alloc.h"

#include "vizir/vzritem.h"
#include "vizir/vzr_pareto.h"
#include "vizir/vzrsort.h"
#include "vizir/vzr_survie.h"
#include "vizir/vzr_heap.h"
#include "vizir/vzr_aleas.h"

//#define DEBUG
#define ASSERT

#define SEUIL 20
//#define SEUIL 10000 // no seuil
#define RATIO 50
#define PROF 16
#define K_IBEA 0.02
#define T_KNEE 0.7
#define KPPV_MOEA 20
#define H_MOEA 7

void vzrpop_init(struct vzrpop *self, int nbin, int nbcont, int nbout, double prec)
{
    self->gene = nbin;
    self->valid = nbcont;
    self->cost = nbout;
    self->hiden = 0;
    self->precis = prec;
    self->policy = VZR_DEFLT_POLICY ;
    // self->policy = VZR_SPEA ;
    self->genre = None;
    self->sort = NULL;
    self->darwin = NULL;
    self->info = NULL;
    self->dealloc_data = NULL;
    
}

static void vzrpop_build(struct vzrpop *self, int stop, int step)
{
    int i;
    
    /* constant */
    self->step = step;
    self->stop = stop;

    self->all = self->valid + self->cost + self->hiden ;

    /* ranking */
    if (self->valid == 0) 
        self->darwin = (self->cost == 1)? &mono_rank : &multi_rank ;
    else if (self->cost == 0)
        self->darwin = &admissible_rank;
    else {
        switch (self->policy & VZR_POLICY_MASK) {
        case VZR_DEFLT_POLICY :
            self->darwin = (self->cost == 1)? &mono_soft_rank : &multi_strict_rank ;
            break;
        case VZR_SOFT_POLICY :
            self->darwin = (self->cost == 1)? &mono_soft_rank : &multi_soft_rank ;
            break;
        case VZR_STRICT_POLICY :
            self->darwin = (self->cost == 1)? &mono_strict_rank : &multi_strict_rank ;
            break;
        }
    }
    /* survie */
    if (! self->sort) self->sort = create_moga();
    PRO_CALL(self->sort, build, self);
    /* working */
    self->level = self->cycle = 0;
    self->cur = self->step;
    self->again = 0;
    self->low = self->elite = 0;
    NEWvecteur(int, self->rg, stop+step);
    for EACH(i, stop+step) self->rg[i] = i;
}

int vzrpop_aploid_build(struct vzrpop *self, int stop, int step)
{
    vzrgene_t *zon;
    double *fron;
    int i, siz;

    vzrpop_build(self, stop, step);
    /* population */
    siz = stop + step;
    NEWvecteur(struct vzritem, self->item, siz);
    NEWvecteur(vzrgene_t, zon, siz * self->gene);
    NEWvecteur(double, fron, siz * self->all);
    for EACH(i, siz) {
        self->item[i].flag = APLOID;
        self->item[i].rank = 0;
        self->item[i].info = 0;
        self->item[i].cycle = 0;
        self->item[i].isol = 0.;
        self->item[i].front = fron + i*self->all;
        self->item[i].data = NULL;
        self->item[i].zone = zon + i*self->gene;
    }
    return 1;

}

int vzrpop_diploid_build(struct vzrpop *self, int stop, int step)
{
    vzrgene_t *zon;
    double *fron;
    int i, siz;

    vzrpop_build(self, stop, step);
    /* population */
    siz = stop + step;
    NEWvecteur(struct vzritem, self->item, siz);
    NEWvecteur(vzrgene_t, zon, 2* siz * self->gene);
    NEWvecteur(double, fron, siz * self->all);
    for EACH(i, siz) {
        self->item[i].flag = DIPLOID;
        self->item[i].rank = 0;
        self->item[i].info = 0;
        self->item[i].cycle = 0;
        self->item[i].isol = 0.;
        self->item[i].front = fron + i*self->all;
        self->item[i].data = NULL;
        self->item[i].dip.domin = zon + 2*i*self->gene;
        self->item[i].dip.reces = zon + (2*i+1)*self->gene;
    }
    return 1;

}

void vzrpop_aploid_dealloc(struct vzrpop *self)
{
    int i;
    
    if (self->dealloc_data) {
        for EACH(i, self->stop+self->step) {
            if (self->item[i].data) self->dealloc_data(self->item[i].data);
        }
    }
    PRO_DEALLOC2(self->sort, self);
    DELvecteur(self->item[0].front);
    DELvecteur(self->item[0].zone);
    DELvecteur(self->item);
    DELvecteur(self->rg);
}

/* dump and resume */
void vzrpop_dump(struct vzrpop *self, char *nfile)
{
    FILE *file;
    int siz = self->stop + self->step;
    short isdip = self->item[0].flag & DIPLOID;
    
    file = fopen(nfile, "wb");
    
    /* meta data */
    fwrite(self, sizeof(struct vzrpop), 1, file);
    fwrite(&isdip, sizeof(short), 1, file);
    
    /* item */
    fwrite(self->rg, sizeof(int), siz, file);
    fwrite(self->item, sizeof(struct vzritem), siz, file);
    fwrite(self->item[0].front, sizeof(double), self->all*siz, file);
    if (isdip) fwrite(self->item[0].zone, sizeof(vzrgene_t), 2*siz*self->gene, file);
    else fwrite(self->item[0].zone, sizeof(vzrgene_t), siz*self->gene, file);
    
    fclose(file);
}

void vzrpop_resume(struct vzrpop *self, int dip, char *nfile)
{
    FILE *file;
    struct vzrpop clone;
    short isdip;
    int i, siz;
    double *fron;
    vzrgene_t *zon;
        
    file = fopen(nfile, "rb");
    
    /* meta data */
    fread(&clone, sizeof(struct vzrpop), 1, file);
    fread(&isdip, sizeof(short), 1, file);
    /* test compatible */
    if ( self->gene != clone.gene || self->valid != clone.valid ||
            self->cost != clone.cost || self->hiden != clone.hiden ||
            dip != isdip  
        ) exit(1);
    
    vzrpop_build(self, clone.stop, clone.step);
    
    /* population */
    siz = self->stop + self->step;
    fread(self->rg, sizeof(int), siz, file);
    NEWvecteur(struct vzritem, self->item, siz);
    fread(self->item, sizeof(struct vzritem), siz, file);
    NEWvecteur(double, fron, siz * self->all);
    fread(fron, sizeof(double), self->all*siz, file);
    if (isdip) {
        NEWvecteur(vzrgene_t, zon, 2*siz*self->gene);
        fread(zon, sizeof(vzrgene_t), 2*siz*self->gene, file);
        for EACH(i, siz) {
            self->item[i].data = NULL;
            self->item[i].front = fron + i*self->all;
            self->item[i].dip.domin = zon + 2*i*self->gene;
            self->item[i].dip.reces = zon + (2*i+1)*self->gene;
        }
    }
    else {
        NEWvecteur(vzrgene_t, zon, siz*self->gene);
        fread(zon, sizeof(vzrgene_t), siz*self->gene, file);
        for EACH(i, siz) {
            self->item[i].data = NULL;
            self->item[i].front = fron + i*self->all;
            self->item[i].zone = zon + i*self->gene;
        }
    }
    
    fclose(file);
    
    self->cur = clone.cur; self->cycle = clone.cycle;
    self->level = clone.level; self->again = clone.again;
    self->low = clone.low; self->elite = clone.elite;
    //printf("resume cur %d lev %d re %d low %d elit %d\n", self->cur, self->level,
    //        self->again, self->low, self->elite);
}

/* init */
void vzrpop_reset(struct vzrpop *self)
{
    //self->level = self->cycle = 0;
    //self->cur = self->step;
    //self->low = self->stop;
}

/* mise a jour population */
int vzrpop_viable(struct vzrpop *self, struct vzritem *item)
{
#ifdef DEBUG
    int i, j;
    // test nan
    for EACH(i, self->all) {
        if (!(item->front[i] < 0 || item->front[i] >=0)) {
            printf("pb00 %d\n", self->cur);
            for EACH(j, self->gene) printf("%.16g  ", item->domin[j]);
            printf("->");
            for EACH(j, self->all) printf("  %g", item->front[j]);
            printf("\n");
        }
    }
#endif
    return PRO_CALL(self->sort, valid, item, self);
}

int vzrpop_add(struct vzrpop *self, struct vzritem *item)
{
    struct vzritem *dest;

    item->cycle = self->cycle;
    /* recopie */
    self->cur--;
    dest = &(self->item[self->rg[self->cur]]);
    if (self->dealloc_data && dest->data) self->dealloc_data(dest->data);
    vzritem_cp(self, item, dest);

    return (self->cur == self->level);
}

void vzrpop_rerank(struct vzrpop *self, int siz)
{
    int low, top, i;
    
    /* recalcul des rangs */
    PRO_CALL(self->sort, rank, siz, self);
    /* doublon */
    low=0; top=siz;
    for EACH(i, siz) {
        if (self->item[i].rank == -1) self->rg[--top] = i;
        else self->rg[low++] = i;
    }
    /* calcul de l'indicateur d'isolement */
    PRO_CALL(self->sort, isol, top, self);
    /* survie */
    PRO_CALL(self->sort, survive, siz, siz-top, self);
}

#ifdef HIDE
void vzrpop_reorder(struct vzrpop *self, int size, int *fit)
{
    int i, dest, src, imem;
    VZRTAB(int, mark, size);
    DIPLOID_VAR(swap, self);

#ifdef TRACE
    for EACH(i, size) printf("%d  ",fit[i]);
    printf("\n");
#endif
    for EACH(i, size) mark[i] = 0;
    for EACH(i, size) {
        if (mark[i]) continue;
        if (i == fit[i]) continue;
        dest = imem = i;
        src = fit[dest];
        /* memo */
        vzritem_cp(self, self->item+i, &swap);
        do {
            /* copie */
            vzritem_cp(self, self->item+src, self->item+dest);
            /* marque */
            mark[dest] = 1;
            /* next */
            dest = src;
            src = fit[dest];
        } while (src != imem);
        /* last copie */
        vzritem_cp(self, &swap, self->item+dest);
        mark[dest] = 1;
    }
    VZRTAB_OUT(mark);
    VAR_OUT(swap);
}
#endif

void vzrpop_cycle(struct vzrpop *self)
{
    int i;
#ifdef ASSERT
    int low;
#endif
    double rg;
    int siz = (self->again != 0)? self->again : self->level;

    /** rerank **/
    vzrpop_rerank(self, siz+self->step);
    /* recopie */
    //vzrpop_reorder(self, siz, idx);
    //VZRTAB_OUT(idx);

    /** finalize **/
    /* elite */
    i = 0;
    while (i < siz && self->item[self->rg[i]].rank == 0) i ++;
    self->elite = i;
#ifdef ASSERT // rank==0 is first
    low = 0;
    for EACH(i, siz) if (self->item[self->rg[i]].rank == 0) low++;
    if (low != self->elite) {
        printf("ASSERT elite %d, %d\n", self->elite, low);
        for FROMEACH(i, self->elite, siz) {
            if (i%10==9) printf("%d\n", self->item[self->rg[i]].rank);
            else printf("%d  ", self->item[self->rg[i]].rank);
        } 
    }
#endif
    /* bad */
    i = self->gene+1;
    rg = self->item[self->rg[i]].rank;
    while (i<self->stop && self->item[self->rg[i]].rank == rg) i++;
    self->low = i-1;
    //printf("-- %d %d %g\n", self.gene, self->pop.low, rg);
    /* cycle */
    self->cycle++;
    if (self->again != 0) {
        self->level = self->again;
        self->again = 0;
    } else if (self->level != self->stop) {
        self->level += self->step;
        if (self->level > self->stop) self->level = self->stop;
    }
    self->cur = self->level + self->step;
}

/* select */
int vzrpop_again(struct vzrpop *self, struct vzritem *item)
{
    if(self->again != 0) {
        vzritem_incp(self, &(self->item[self->rg[self->cur-1]]), item);
        //printf(".");
        return 1;
    }
    else return 0;
}

