/* --- Ansi C --- */
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2016-2022 CEA/DES
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   vzrpop_util.c
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2016-04-21
 
 \brief  utilitaire sur la population
*/

//#define L1
#define TRACE

#include <math.h>
#ifdef TRACE
#include <stdio.h>
#endif

#include "vzrpop_util.h"

#include "vzr_macro.h"
#include "vzrpop_api.h"


/* normalisation via minmax */
void vzrpop_minmaxnorm(struct vzrpop *pop, int siz, double *norm)
{
    int i, j;
    double min, max;
    
    /* coef normalisation */
    for EACH(j, pop->all) {
        /* parcours id */
        min = max = pop->item[pop->rg[0]].front[j];
        for FROMEACH(i, 1, siz) {
            if (pop->item[pop->rg[i]].front[j] < min) 
                min = pop->item[pop->rg[i]].front[j];
            else if (pop->item[pop->rg[i]].front[j] > max) 
                max = pop->item[pop->rg[i]].front[j];
        }
        norm[j] = (max == min) ? 1 : max - min;
    }
#ifdef TRACE2  
    for EACH(j, pop->all) printf("%g  ", norm[j]);
    printf("\n");
#endif
}

/* recherche du point ideal et normalisation sur les objectifs */
void vzrpop_ideal_it(struct vzrpop *pop, int siz, double *ideal, double *norm)
{
    int i, j;
    double min, max;
    
    /* coef normalisation */
    for FROMEACH(j, pop->valid, pop->valid+pop->cost) {
        /* parcours id */
        min = max = pop->item[pop->rg[0]].front[j];
        for FROMEACH(i, 1, siz) {
            if (pop->item[pop->rg[i]].front[j] < min) 
                min = pop->item[pop->rg[i]].front[j];
            else if (pop->item[pop->rg[i]].front[j] > max) 
                max = pop->item[pop->rg[i]].front[j];
        }
        ideal[j-pop->valid] = min;
        norm[j-pop->valid] = (max == min) ? 1 : max - min;
    }
}

/* plus proche voisin */
static INLINE void insert_voisin(double *dist, int *ref, double dnew, short idnew, 
        char low, char siz)
{
    int i, j;
    
    i = low;
    j = ((low+1)<<1)-1;
    while (j < siz) {
        if (j+1 < siz && dist[j] < dist[j+1]) j++;
        if (dnew < dist[j]) {
            dist[i] = dist[j];
            ref[i] = ref[j];
            i = j;
            j = ((j+1)<<1)-1;
        }
        else break;
    }
    dist[i] = dnew;
    ref[i] = idnew;
}
  
static char put_voisin(double *dist, int *ref, char fill, char prof, 
        double d, short id)
{
    if (fill < prof) { /* premier voisin */
        insert_voisin(dist, ref, d, id, prof-fill-1, prof);
        fill++;
    }
    else { /* voisin suivant */
        if (d < dist[0]) insert_voisin(dist, ref, d, id, 0, prof);
    }
    return fill;
}

void fill_voisin(struct voisin *self, char *fill, int i, int j, double dij)
{
    fill[i] = put_voisin(self->dist[i], self->ref[i], fill[i], self->prof, dij, j);
    fill[j] = put_voisin(self->dist[j], self->ref[j], fill[j], self->prof, dij, i);
} 

void lastsort_voisin(struct voisin *self, int i)
{
    int j, ri;
    double di;
    
    for (j = self->prof-1; j >=0; j--) {
        di = self->dist[i][j];
        ri = self->ref[i][j];
        self->dist[i][j] = self->dist[i][0];
        self->ref[i][j] = self->ref[i][0];
        insert_voisin(self->dist[i], self->ref[i], di, ri, 0, j);
    }
}

/** espace objectifs **/
void vzrpop_proche_voisin3(struct vzrpop *pop, int siz, double *norm, struct voisin *self)
{
    int i, j, k, ii, jj;
    double di, tmp;
    VZRTAB(char, fill, self->size);

    /* initialisation */
    for EACH(i, self->size) fill[i] = 0;

    /* calcul des couts (distances) */
    for FROMEACH(ii, 1, siz) {
        for EACH(jj, ii) {
            di = 0.;
            i = pop->rg[ii];
            j = pop->rg[jj];
            //for EACH(k, pop->all) { // rank may be unsane
            for FROMEACH(k, pop->valid, pop->cost+pop->valid) { // rank may be unsane
                tmp = (pop->item[i].front[k] > pop->item[j].front[k])?
                    (pop->item[i].front[k] - pop->item[j].front[k]) 
                        / norm[k] :
                    (pop->item[j].front[k] - pop->item[i].front[k])
                        / norm[k];
#ifdef L1
                di += tmp ;
#else
                if (tmp > di) di = tmp;
#endif
            }
            fill_voisin(self, fill, i, j, di);
        }
    }
#ifdef TRACE2
    for EACH(i, siz) {
        printf("[%d, %d/%d]", i, fill[i], self->prof);
        for EACH(j, self->prof) printf("%d %g,", self->ref[i][j], self->dist[i][j]);
        printf("\n");
    }
#endif
    VZRTAB_OUT(fill);
}

void vzrpop_proche_voisin(struct vzrpop *pop, int siz, struct voisin *self)
{
    /* normalisation */
    vzrpop_minmaxnorm(pop, siz, self->norm);
    
    /* distance */
    vzrpop_proche_voisin3(pop, siz, self->norm, self);

}

void vzrpop_proche_voisin2(struct vzrpop *pop, int siz, struct voisin *self)
{
    int ii, i;

    /* normalisation */
    vzrpop_minmaxnorm(pop, siz, self->norm);
    
    /* distance */
    vzrpop_proche_voisin3(pop, siz, self->norm, self);
    
    /* last sort */
    for EACH(ii, siz) {
        i = pop->rg[ii];
        lastsort_voisin(self, i);
    }
}

/** espace variables **/
void vzrpop_proche_individu(struct vzrpop *pop, int siz, struct voisin *self)
{
    int i, j, k, ii, jj;
    double di, tmp;
    VZRTAB(char, fill, self->size);

    /* initialisation */
    for EACH(i, self->size) fill[i] = 0;

    /* calcul des couts (distances) */
    for FROMEACH(ii, 1, siz) {
        for EACH(jj, ii) {
            di = 0.;
            i = pop->rg[ii];
            j = pop->rg[jj];
            //for EACH(k, pop->all) { // rank may be unsane
            for EACH(k, pop->gene) { // rank may be unsane
                tmp = (pop->item[i].zone[k] > pop->item[j].zone[k]) ?
                    (pop->item[i].zone[k] - pop->item[j].zone[k]) :
                    (pop->item[j].zone[k] - pop->item[i].zone[k]) ;
#ifdef L1
                di += tmp ;
#else
                if (tmp > di) di = tmp;
#endif
            }
            fill_voisin(self, fill, i, j, di);
        }
    }
#ifdef TRACE2
    for EACH(i, siz) {
        printf("[%d, %d/%d]", i, fill[i], self->prof);
        for EACH(j, self->prof) printf("%d %g,", self->ref[i][j], self->dist[i][j]);
        printf("\n");
    }
#endif
    VZRTAB_OUT(fill);
}



#ifdef HIDE
void vzrpop_reorder_voisin(struct vzrpop *pop, int siz, int *id, struct voisin *self)
{
    int i, j, dest, src, imem;
    VZRTAB(char, mark, siz);
    VZRTAB(int, swap, self->prof);
    
    /* TODO utilisee self->level a la place de siz ?? */ 
    /* mise a jour des index des voisins */
    for EACH(i, siz) {
        for EACH(j, self->prof) self->ref[i][j] = id[self->ref[i][j]];
    }
    
    /* reorder */
    for EACH(i, siz) mark[i] = 0;
    for EACH(i, siz) {
        if (mark[i]) continue;
        if (i == id[i]) continue;
        dest = imem = i;
        src = id[i];
        /* memo (we just save ref (not dist)) */
        for EACH(j, self->prof) swap[j] = self->ref[dest][j];
        do  {
            /* copie */
            for EACH(j, self->prof) self->ref[dest][j] = self->ref[src][j];
            /* mark */
            mark[dest] = 1;
            /* next */
            dest = src;
            src = id[dest];
        } while (src != imem);
        /* last copie */
        for EACH(j, self->prof) self->ref[dest][j] = swap[j];
        mark[dest] = 1;
    }
    
    VZRTAB_OUT(swap);
    VZRTAB_OUT(mark);
}
#endif

/* pareto plan rotation */
/* vecteur perpendiculaire au plan */
static void GramSchmidt(int size, double **rot)
{
    int i, j, k;
    double proj, norm;

    /* gram-schmidt orthonormalisation */
    for EACH(j, size)
    {
        for EACH(i, j)
        {
            /* projection on vi */
            proj = 0.;
            //denom = 0.;
            for EACH(k, size)
            {
                proj += rot[j][k] * rot[i][k];
                //denom += rot[i*size+k]*rot[i*size+k];
            }
            //proj /= denom;
            /* remove component in direction vi */
            for EACH(k, size)
            {
                rot[j][k] -= proj * rot[i][k];
            }
        }
        /* normalisation */
        norm = 0;
        for EACH(i, size)
        {
            norm += rot[j][i] * rot[j][i];
        }
        norm = sqrt(norm);
        for EACH(i, size)
        {
            rot[j][i] /= norm;
        }
    }
}

//static void knee_indicator(struct xplan *self, struct vzrpop *pop, int size, int *idx)
void vzrpop_pareto_plan(struct vzrpop *pop, int size, struct xplan *self, double *norm)
{
    int d, i, j, incr;
    double cst, tmp;
    VZRTAB(int, mono, self->dim);

    incr = pop->valid;
    /* solutions mono objectif */
    for EACH(d, self->dim) {
        i = 0;
        while (pop->item[pop->rg[i]].rank != 0) i++;
        mono[d] = pop->rg[i];
        for (; i<size; i++) {
            if (
                    pop->item[pop->rg[i]].rank == 0
                    && pop->item[pop->rg[i]].front[d+incr] < pop->item[mono[d]].front[d+incr] 
                ) 
                mono[d]=pop->rg[i];
        }
    }
#ifdef TRACE2
    for EACH(i, self->dim) printf("ideal[%d] is %d with %g\n", i, 
            mono[i], pop->item[mono[i]].front[i+incr]);
#endif

    /* n-1 first vector is vecteur x_i->x_i+1 */
    for EACH(i, self->dim-1) {
        for EACH(j, self->dim) {
            self->rot[i][j] = pop->item[mono[i+1]].front[j+incr] - 
                    pop->item[mono[i]].front[j+incr];
        }
    }
    /* last is vector ideal->x_0 */
    for EACH(i, self->dim) {
        //self->rot[self->dim-1][i] = pop->item[mono[0]].front[i+incr] - 
        //      pop->item[mono[i]].front[i+incr];
        self->rot[self->dim-1][i] = 1.;
    }

    /* plan */
    GramSchmidt(self->dim, self->rot);
    for EACH(i, self->dim) if (self->plan[i] < 0) printf("%d-\n", i);
#ifdef TRACE2
    for EACH(i, self->dim) {
        for EACH(j, self->dim) printf("%g ", self->rot[i][j]);
        printf("\n");
    }
#endif    
    
    if (norm) {
        cst = 0.;
        for EACH(d, self->dim) cst += pop->item[mono[0]].front[d+incr] * self->plan[d];
#ifdef TRACE2        
        for EACH(d, self->dim) printf("%g  ", self->plan[d]);
        printf("%g  Plan\n", cst);
#endif        
        for EACH(d, self->dim) {
            tmp = cst;
            for EACH(i, self->dim) {
                if (d == i) continue;
                tmp -= self->plan[i] * pop->item[mono[i]].front[i+incr];
            }
            norm[d] = (tmp / self->plan[d]) - pop->item[mono[d]].front[d+incr];
        }
#ifdef TRACE2
        for EACH(d, self->dim) printf("%g  ", norm[d]);
        printf(" Norm\n");
#endif
    }
    
    VZRTAB_OUT(mono);
}

