/*  _______________________________________________________________________

 DAKOTA: Design Analysis Kit for Optimization and Terascale Applications
 Copyright (c) 2001, Sandia National Laboratories.
 This software is distributed under the GNU General Public License.
 For more information, see the README file in the top Dakota directory.
 _______________________________________________________________________ */

//      Description:  This is the header file for the LatinHypercube
//        object. The LatinHypercube is used to generate a
//        distribution of random sample values.
//
//      Limitations and abilities:
//        The following distributions can be generated:
//
//        Distribution          Inputs
//        1. Normal             mean and coefficient of variation
//        2. Lognormal          mean and coefficient of variation
//        3. Weibull            mean and coefficient of variation
//        4. Uniform            Lower Limit and Upper Limit
//        5. Loguniform         Lower Limit and Upper Limit
//        6. Constant           mean = value, coef of variation = 0
//
//      Maxima (these can be changed in lhs.h file)
//        Max sample size = 1000
//        Max number of random variables to create a sample for = 50
//
//      Note: most of this code was originally Fortran code where
//        all arrays were 1 dimensional. This use of 1-dim arrays to
//        represent 2-dim arrays is still used in this C++ code.
//
//        Also, for portablity to the Mac all output that is not for 
//        debugging must use the Tcl_Channel structure defined by  
//        tcl/tk and it related functions to print to files.
//**************************************************************************
#ifndef LATIN_HYPERCUBE_H
#define LATIN_HYPERCUBE_H

class LHSInput;

const int MAXPOSDEF = 10;
const int NONRANDOM = 0;
const int RANDOM = 1;

struct stats
{
    double mean;
    double k;
    double std;
    double min;
    double max;
};

struct lhrank
{
    lhrank()
    {
    }
    ;
    ~lhrank();
    double *xv;
    double *rxv;
    int *iwk;
    double *final;
    void SetSize(long int nobv, long int nvar); // allocate space for vectors
};

struct workc
{
    workc()
    {
    }
    ~workc();
    double *q;
    double *s;
    double *uq;
    void SetSize(long int nvar); // allocate space for vectors
};

struct cmatr
{
    cmatr()
    {
    }
    ;
    ~cmatr();
    double *corr;
    double *rcorr;
    double *ucorr;
    int *icm;
    int ncm;
    void SetSize(long int nvar); // allocate space for vectors
};

class LatinHypercube
{
public:
    LatinHypercube(LHSInput *InPtr);
    ~LatinHypercube();
    void GenSample();
    double * GetSample()
    {
        return _sampx;
    }
    double * GetRange()
    {
        return _rank.final;
    }

    double GetSample(int i, int j)
    {
        return _sampx[(j - 1) * _nobv + i - 1];
    }
    void SetSampleType(int type)
    {
        _sampletype = type;
    }
    int SetUserCorrMatrix(double *matrix);
    void TestCorrMatrix();
    int addCorr(int i, int j, double value);
    void SetSeed(int seed)
    {
        _seed = -seed;
        _originalseed = seed;
    }
    void SetSaveRang()
    {
        _saveRang = 1;
    }
    double getRang(int i, int j)
    {
        return _rank.final[(j - 1) * _nobv + i - 1];
    }
    void UnsetSaveRang()
    {
        _saveRang = 0;
    }
    int GetSaveRang()
    {
        return _saveRang;
    }
    // Gestion de la log
    void SetLog(int value = 1)
    {
        _bLog = value;
    }
    void UnsetLog()
    {
        _bLog = 0;
    }
    int GetLog()
    {
        return _bLog;
    }

    //

    void LOG(const char *texte);
#ifdef root
    char * GetVariablesName();
#endif
    long int GetN_of_V()
    {
        return _nvar;
    }
    long int GetN_of_O()
    {
        return _nobv;
    }

    void normal(int j, long int n, double a, double b, int type, double pmin,
            double pmax);
    void uniform(int j, long int n, double a, double b, int type);
    void triangular(int j, long int n, double a, double b, double c, int type,
            double pmin, double pmax);
    void weibull(int j, long int n, double alpha, double beta, double xmin);
    void exponential(int j, long int n, double lambda, double xmin);
    void constant(int j, long int n, double a);
    void user_defined(int j, long int n, char *fname);
    void Mix();
    void compute_Stats();
    void PrintHistogram(int numbins);
    void Output_Plot_Data(const char *file);
    void Print_sample(const char *filename)
    {
        print_matrix_to_file(_sampx, filename);
    }
    void print_Output(const char *filename);
    void printf_matrix(double *, const char *, const char *filename);
    void printf_matrix_dlhs(const char *filename);
    void printf_cmatrix(double *, const char *, const char *filename);
    void print_matrix(double *, char *);
    void print_cmatrix(double *, char *);
    void print_cmatrix_root(double *, char *, const char *filename);
    void printf_matrix_root(double *, char *, const char *filename);
    void init_root_file(const char *filename);
    void close_root_file(const char *filename);
    void print_rank(long int);
    void printlog();
protected:
    double finvnor(double x);
    double rierfc1(double y);
    void corcal();
    void rankcorcal();
    void matinv(double *cm, long int size);
    void ranker();
    void hpsrt();
    void print_matrix_to_file(double *mat, const char *filename);

    void d2_to_d1(double **a2, double *a1);
    void d1_to_d2(double **a2, double *a1);
    void print_2d(double **a2);

    long int LOC(long int i, long int j)
    {
        return (j - 1) * _nobv + i - 1;
    }
    long int LOC1(long int i, long int j)
    {
        return j + (i * i - i) / 2 - 1;
    }

    LHSInput *_InPtr;
    int _seed;
    int _originalseed;
    int _randomPairing;
    int _sampletype;  // if 1 random sample; if 0 non random sample
    long int _nobv;        // number of observations
    long int _nvar;        // number of variables
    int _ucm;         // use user corr matrix flag
    int _saveRang;      // bool�en pour sauvegarder les rangs
    int _bLog;          // indice sur le niveau de log � afficher
    long int _nvx;
    lhrank _rank; // arrays used for ranking sample
    double *_sampx;   // holds sample vectors and used as temp array
    double *_workxx;  // save sample vectors
    workc _workc;     // q and s matrix
    cmatr _cmatr;     // stores default and user corr matrix
    stats *_sinfo;
    int nbins;        // histogram bin count
    long int **bins;	 // array for histogram

    //  original fortran sub. Not used
    void chlsky(double *cm, double *q, long int size);

private:
};

#endif
