//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*! \file TConstrLHS.cxx
 *  \author Jean-Baptiste Blanchard <jean-baptiste.blanchard@cea.fr>
 *  \date   $date$
 *  \brief Implementation of the class URANIE::Sampler::TConstrLHS. (based on the algorithm by Guillaume Damblin)
 *
 *  Created the : Thu Apr 13 13:37:23 2004
 *
 */
#include <ctime>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <algorithm>

////// ROOT
#include "TMath.h"
#include "TDecompLU.h"
#include "TDatime.h"
#include "TROOT.h"
#include "TRandom3.h"

////// URANIE
#include "TConstrLHS.h"

ClassImp (URANIE::Sampler::TConstrLHS)

namespace URANIE
{
namespace Sampler
{
////////////////////////////////////////////////////////////
//  Constructeur : from TDataServer + Methode + size
////////////////////////////////////////////////////////////
TConstrLHS::TConstrLHS(URANIE::DataServer::TDataServer *tds, int sampleSize):
    TSamplerStochastic(tds, "lhs", sampleSize), _fConstrain(NULL)
{
    if (_blog)
        cout << "constructor TConstrLHS using TDataServer["
             << tds->GetName() << "]" << endl;

    //Check whether a tree exists or not
    try
    {
	if( tds->getTuple() != NULL )
	{
	    _bAlreadyFilled = (tds->getTuple()->GetEntries() != 0) ;
            int nEntries = tds->getTuple()->GetEntries();
	    if( _nS != nEntries )
		throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							   Form("TConstrLHS::TConstrLHS : The number of requested sample [%i] is different from the one stored [%i]", 
                                                                _nS, nEntries));
            
	    TList *attList = _tds->getListOfAttributes();
	    _nX = attList->GetEntries();
	    delete attList;
	}
    }catch( URANIE::Exceptions::UWarningExceptions& ue)
    {
        _bAlreadyFilled = false;
    }catch( URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }  
	

    //Create the sampler for initial sampling    
    _sampling = ( (_bAlreadyFilled ) ? NULL : new TSampling(tds, "lhs", sampleSize));

    _sMethod=TString("constr_lhs");

    //Resizing the methods
    _currentLHS.ResizeTo(_nS,_nX);
    _currentPerm.ResizeTo(_nS,_nS);

    _paramValues.clear();
    
    _ns=_nS;
    _nx=_nX;

    _nConstr=0;

        
}

void TConstrLHS::initElements()
{
        
    // Initialise the matrix of solution by rows (nConstr vector containing nS elements)
    _rowSol.clear();
    _rowSol.resize(_nConstr, std::vector<int>(_nS,0.));
        
    // Initialise the matrix of solution by columns (nConstr vector containing nS elements)
    _colSol.clear();
    _colSol.resize(_nConstr, std::vector<int>(_nS,0.));
        
    // Initialise the matrix of diagonal value (nConstr vector containing nS elements)
    _diag.clear();
    _diag.resize(_nConstr, std::vector<int>(_nS,0.));
        
    // Initialise the vector of modifications done by permutation (nConstr vector containing no elements)
    _modified.clear();
    _modified.resize(_nConstr);
        
    // Initialise the vector of constraint matrixsolution by rows (nConstr Matrix of size _nS x _nS
    _currentConst.clear();
    _currentConst.resize(_nConstr, TMatrixD(_ns,_ns));
        
}

TConstrLHS::~TConstrLHS()
{
    if (_blog)
        cout << "************ Destroying URANIE::Sampler::TConstrLHS"
                << endl;

    if( _sampling != NULL)
	delete _sampling;
    _sampling=NULL;

}

// add the constraint function and its parameters
void TConstrLHS::addConstraint( void (*fConstrain)(double *, double*), int nConstr, int nbparam, int *param )
{
        
    _nConstr = nConstr;
    _fConstrain = fConstrain;
    _nParam = nbparam;
    try
    {
        if( _nParam != 2*_nConstr )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TConstrLHS::addConstraint: The number of input parameters [%i] should be [%i], twice the number of constraints [%i]", \
                                                            _nParam,_nConstr*2, _nConstr));
    
        _paramValues.resize(_nParam,-1);
        vector<int> colonne; colonne.clear();
        vector<int>::iterator iti;
        for(unsigned int ip=0; ip<_nParam; ip++)
        {        

            if( param[ip] >= _nX || param[ip] < 0 )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TConstrLHS::addConstraint : \n\
The parameter provided in the list [%i] is not correct given the number of variables in the dataserver [%d]\n\
The parameter identifier should be lying within [0,%i]",
                                                                param[ip], _nX, _nX-1));
            
            _paramValues.at(ip) = param[ip];

            if( ip % 2 == 1)
            {
                iti = std::find(colonne.begin(), colonne.end(), param[ip] );
                if( iti != colonne.end() )
                {
                    int pos=iti-colonne.begin();
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                               Form("TConstrLHS::addConstraint : The second variable of the %d-th constraint [%s] is also the second \
variable of the %d-th constraint.\nThis is dangerous as both constraints will reorganise the [%s] column of the doe and their permutations might cancel themself.\n Please\
re-organise your constraints keeping in mind that the first variable is always untouched in term of permutations for the constraint you're defining.",\
                                                                    ip/2, _tds->getAttribute(param[ip])->GetName(), pos, _tds->getAttribute(param[ip])->GetName()));
                }

                colonne.push_back( param[ip] );
            }
            
        }
    }catch( URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }  
    
    _nbdiag.resize(_nConstr, 0);
        
}

// setSeed
void TConstrLHS::setSeed(Int_t ind )
{

    if( _sampling != NULL)
	_sampling->setSeed(ind);

    TSamplerStochastic::setSeed(ind);
    
}

// Printer
void TConstrLHS::printMatrix(TMatrixD &dist)
{
    int size=3;
    unsigned int Nrow=dist.GetNrows();
    cout<<"The current matrix is: "<<endl;
    cout<<setw(size)<<"    "<<setw(size)<<"[04m";
    for(unsigned int i=0 ; i<Nrow ; i++)
        cout << setw(size) << i ;
    cout<<" [0m";
    cout<<endl;
    for(unsigned int i=0 ; i<Nrow ; i++)
    {
        cout << setw(size) << i << "|";
        for(unsigned int j=0 ; j<Nrow ; j++)
            cout<< setw(size) << dist(i,j);
        cout<<endl;
    }
    cout<<endl;
   
}

// compute the current constraint matrix
void TConstrLHS::getConstMatrix()
{
    initElements();
            
    // simple pointer to read data
    double *arr=NULL;
    
    vector< vector<double> > mcol;
    //mcol.resize(_nParam);
    for(unsigned int ip=0; ip<_nParam; ip++)
    {
        TMatrixD localcol=_currentLHS.GetSub(0, _nS-1, _paramValues.at(ip), _paramValues.at(ip));
        arr = localcol.GetMatrixArray();
        mcol.push_back(vector<double>(arr, arr+_nS));        
    }
    
    vector<double> vi, vj; vi.resize(_nParam); vj.resize(_nConstr);
            
    for (unsigned int i=0 ; i<_ns ; i++)
    {
        for (unsigned int j=0 ; j<_ns; j++)
        {

            // Compute the constraint for the (i,j) element of the nS x nS constraint matrices
            for(unsigned int ic=0; ic<_nConstr; ic++)
            {
                vi[2*ic]   = mcol.at(2*ic).at(i);
                vi[2*ic+1] = mcol.at(2*ic+1).at(j);
            }
            _fConstrain(&vi[0], &vj[0]);

            for(unsigned int ic=0; ic<_nConstr; ic++)
            {         
                if (i==j)
                {
                    _diag.at(ic).at(j)+=vj[ic];
                    if(vj[ic]==1) _nbdiag.at(ic)++;
                }
                
                _currentConst.at(ic)(i,j)=vj[ic];
                
                _colSol.at(ic).at(j)+=vj[ic];
                _rowSol.at(ic).at(i)+=vj[ic];
            }
        }
    }

    for(unsigned int ic=0; ic<_nConstr; ic++)
    {

        // Check that there is no row for which no columns are fulfilling the constraints (obvious problem) 
        vector<int>::iterator iter = std::find(_rowSol.at(ic).begin(), _rowSol.at(ic).end(), 0);
        if ( iter != _rowSol.at(ic).end() )
        {
            cout<<"This design-of-experiments is not meant to be constrained with the "<<ic<<"-th constraint provided. row "
                << iter - _rowSol.at(ic).begin() <<" has no solutions !!"<<endl;
            _nbdiag.at(ic)=-1;
            
            if( _blog && _nS < 75 )
                printMatrix(_currentConst.at(ic));
            return;
        }

        // Check that there is a solution to match row and columns
        int maxbpm=maxBPM(_currentConst.at(ic));
        if (  maxBPM(_currentConst.at(ic)) != _nS )
        {
            cout << "This design-of-experiments is not meant to be constrained with the " << ic << "-th constraint provided. There are only "
                 << maxbpm << " events which can be sorted (out of " << _nS << ")" << endl;
            _nbdiag.at(ic)=-1;
            
            if( _blog && _nS < 75 )
                printMatrix(_currentConst.at(ic));
        }
        else
            cout << "The maxBPM for this doe and the " << ic << "-th constraint with variables (" << _tds->getAttribute(_paramValues.at(2*ic))->GetName()
                 << ", "<< _tds->getAttribute(_paramValues.at(2*ic+1))->GetName() << ") is " << maxbpm << endl;

    
        if(_blog && _nS <= 75)
        {
            cout << "COL 1 {";
            for (unsigned int i=0 ; i<_ns ; i++)
                cout << mcol.at(2*ic).at(i) << ((i!=_ns-1)?", ":"}\n");
        
            cout << "COL 2 {";
            for (unsigned int i=0 ; i<_ns ; i++)
                cout << mcol.at(2*ic+1).at(i) << ((i!=_ns-1)?", ":"}\n");

            cout << "The lhs is " << endl;
            _currentLHS.Print();

            printMatrix(_currentConst.at(ic));

            cout << "COL SOLUTIONS {";
            for (unsigned int i=0 ; i<_ns ; i++)
                cout << _colSol.at(ic).at(i) << ((i!=_ns-1)?", ":"}\n");

            cout << "ROW SOLUTIONS {";
            for (unsigned int i=0 ; i<_ns ; i++)
                cout << _rowSol.at(ic).at(i) << ((i!=_ns-1)?", ":"}\n");

            cout << "DIAG SOLUTIONS {";
            for (unsigned int i=0 ; i<_ns ; i++)
                cout << _diag.at(ic).at(i) << ((i!=_ns-1)?", ":"}\n");
        } // if _blog

        try
        {
            double colRMS = TMath::RMS( _colSol.at(ic).begin(), _colSol.at(ic).end());
            double rowRMS = TMath::RMS( _rowSol.at(ic).begin(), _rowSol.at(ic).end());
            
            if( colRMS > rowRMS )
                throw URANIE::Exceptions::UInfoExceptions(__FILE__, __LINE__,
                                                           Form("TConstrLHS::getConstMatrix :\n\
Information for constraint [%i]\n\
An empirical assumption, given both the number of solutions looking at [%s] and at [%s], tends to say that an inversion of the constraint might speed up the process.\n\
If you have written for instance x1-x2<0, you might considerer x2-x1>0 (if this is possible for you with respect to other constraints for instance).",\
                                                                ic, _tds->getAttribute(_paramValues.at(2*ic))->GetName(), _tds->getAttribute(_paramValues.at(2*ic+1))->GetName()));
        }catch( URANIE::Exceptions::UInfoExceptions& ue)
        {
            ue.printMessage();
        }
        
    } // loop over nConstr to check content of 
        
}

// A DFS based recursive function that returns true if a matching for vertex u is possible
bool TConstrLHS::bpm(TMatrixD &lhs, int u,  int *seen, int *matchR)
{
    // Try every job one by one
    for (int v = 0; v < _nS; v++)
    {
        // If applicant u is interested in job v and v is not visited
        if (lhs(u,v) && !seen[v])
        {
            // Mark v as visited
            seen[v] = true;
 
            // If job 'v' is not assigned to an
            if (matchR[v] < 0 || bpm(lhs, matchR[v],
                                     seen, matchR))
            {
                matchR[v] = u;
                return true;
            }
        }
    }
    return false;
}
 
// Returns maximum number of matching from _nS to _nS
int TConstrLHS::maxBPM(TMatrixD &lhs)
{
    // An array to keep track of the applicants assigned to jobs.
    // The value of matchR[i] is the applicant number assigned to job i,
    // the value -1 indicates nobody is  assigned.
    vector<int> matchR(_nS,-1);
 
    // Count of jobs assigned to applicants
    int result = 0;
    for (int u = 0; u < _nS; u++)
    {
        // Mark all jobs as not seen for next applicant.
        vector<int> seen(_nS, 0);
 
        // Find if the applicant 'u' can get a job
        if (bpm(lhs, u, &seen[0], &matchR[0]))
            result++;
    }
    return result;
}


void TConstrLHS::findNextPermutation()
{
 
    std::vector<int>::iterator iti;
    std::vector<double>::iterator itd;
    for( unsigned int ic=0; ic<_nConstr; ic++)
    {
    
        if( _nbdiag.at(ic) == _nS )
            continue;
        // else if( _iPerm < 45)
        // {
        //     if(true) //_blog)
        //         cout << "In Permutation " << _iPerm << ", only " << _nbdiag.at(ic) << " diag are non-zero " ;
        // }
        // else
        // {
        //     // printMatrix(_currentConst.at(ic));
        //     for(unsigned int i=0; i<_diag.at(ic).size(); i++)
        //     {
        //         if(_diag.at(ic).at(i)==0)
        //             cout<<" elements "<<i<<endl;
        //     }
        //     exit(-1);
        // }
        
        if(_blog)
            printMatrix(_currentConst.at(ic));
        
        // Looking for the first 0 diagonale element, meaning the one with the least number of solutions
        multimap<int,int> rowSol_rowPos;
        iti = std::find(_diag.at(ic).begin(), _diag.at(ic).end(), 0);
        while( iti != _diag.at(ic).end() )
        {
            rowSol_rowPos.insert( make_pair( _rowSol.at(ic).at(iti-_diag.at(ic).begin()), iti-_diag.at(ic).begin() ));
            iti = std::find(iti+1, _diag.at(ic).end(), 0);
        }
    
        if(_blog)
        {
            cout << rowSol_rowPos.size() << " problems :: (NB COL SOLUTION, ROW POS) are ";
            for (auto const& x : rowSol_rowPos)        
                cout << " (" << x.first << "," << x.second << ") ";
            cout << endl;
        }
        int MyRow = rowSol_rowPos.begin()->second;
        
        // Looking for the first 0 diagonale element, meaning the one with the least number of solutions
        vector<double> rowContent; rowContent.resize(_nS);
        _currentConst.at(ic).ExtractRow(MyRow,0, &rowContent[0]);
    
        if(_blog)
        {
            cout << "SELECTED ROW CONTENT {";
            for (unsigned int i=0 ; i<_ns ; i++)
                cout << rowContent.at(i) << ((i!=_ns-1)?", ":"}\n");
        }
        
        //multimap<int,int, greater <int>> colSol_colPos;
        multimap<int,int> colSol_colPos;    
        itd = std::find(rowContent.begin(), rowContent.end(), 1.);
        while( itd != rowContent.end() )
        {        
            colSol_colPos.insert(make_pair(_colSol.at(ic).at(itd-rowContent.begin()), itd-rowContent.begin()));
            itd = std::find(itd+1, rowContent.end(), 1.);
        }
    
        if( _blog )
        {
            cout << colSol_colPos.size() << " solutions :: (NB SOLUTIOSN, COL POS) are ";
            for (auto const& x : colSol_colPos)        
                cout << " (" << x.first << "," << x.second << ") ";
            cout << endl;
        }
        
        int MyCol=-1;
        vector<int> bestsolutions; bestsolutions.clear();
        vector<int> allsolutions; allsolutions.clear();
        bool previous;

        if( _modified.at(ic).count(MyRow) == 0 )
            _modified.at(ic)[MyRow].clear();
        
        for (auto const& x : colSol_colPos)
        {
            MyCol = x.second;
            iti = std::find(_modified.at(ic)[MyRow].begin(), _modified.at(ic)[MyRow].end(), MyCol);
            
            previous=(iti!=_modified.at(ic)[MyRow].end()); 
            
            allsolutions.push_back(MyCol);

            if(_blog)
                XYZ << " Looking at row " << MyRow << " considering col (" << MyCol << ") ==> " << _currentConst.at(ic)(MyCol, MyRow)
                     << " in modified ? " << previous << " nbsol " << x.first << endl;
            
            if( previous )
                continue;

            if(_blog)
                cout << " Looking at row " << MyRow << " considering col (" << MyCol << ") ==> " << _currentConst.at(ic)(MyCol, MyRow) << endl;
                
            bestsolutions.push_back(MyCol);
            if( _currentConst.at(ic)(MyCol, MyRow) == 1 )
                break;
        }
        
        if( _currentConst.at(ic)(MyCol, MyRow) == 0 )
        {
            if( bestsolutions.size() != 0 )
                MyCol = bestsolutions.at(gRandom->Integer(bestsolutions.size()));//back();//at(0);
            else
                MyCol = allsolutions.at(gRandom->Integer(allsolutions.size()));
        }
    
        // _modified.at(ic).push_back(MyCol);
        _modified.at(ic)[MyRow].push_back(MyCol);

        performPermutation(ic, MyRow, MyCol, _paramValues.at(2*ic+1));
        
        int MyColSol=-1;
        for (auto const& x : colSol_colPos)
            if( x.second == MyCol )
                MyColSol = x.first;
                    
    }

}

// perform the permutation 
void TConstrLHS::performPermutation(int Rcon, int Rrow, int Rcol, int Rvar)
{

    if(_blog)
        XYZ << "Permutation " << _iPerm << "  determined from constraint " << Rcon << " (only " << _nbdiag.at(Rcon) << " non-zero elements);"
            << "In line " << Rrow << "  will swap column " << Rrow << " and " << Rcol  << endl;

    // Change the content of the colSol vector
    double oldVal = _currentLHS(Rrow,Rvar);
    _currentLHS(Rrow,Rvar) = _currentLHS(Rcol,Rvar);
    _currentLHS(Rcol,Rvar) = oldVal;

    bool isrow=false, iscol=false;
    // Run the permutation for constraint matrix and row/col vectors
    for(unsigned int ic=0; ic<_nConstr; ic++)
    {
        // Initialise to false in case of many constraint
        isrow=iscol=false;
        
        if( Rvar == _paramValues.at(2*ic))
            isrow=true;
        
        else if( Rvar == _paramValues.at(2*ic+1))
            iscol=true;

        if( !isrow && !iscol)
            continue;
        
        // Create the permutation matrix to change the constraint matrix
        _currentPerm.UnitMatrix();

        _currentPerm(Rrow, Rrow)=0;
        _currentPerm(Rcol,Rcol)=0;

        _currentPerm(Rrow,Rcol)=1;
        _currentPerm(Rcol, Rrow)=1;

        if(iscol)
            _currentConst.at(ic)=_currentConst.at(ic)*_currentPerm;
        else if(isrow)
            _currentConst.at(ic)=_currentPerm*_currentConst.at(ic);
        
        if(_blog)
            printMatrix(_currentConst.at(ic));

        // Change the content of the colSol vector
        if(iscol)
        {
            int oldval = _colSol.at(ic).at(Rrow);
            _colSol.at(ic).at(Rrow) = _colSol.at(ic).at(Rcol);
            _colSol.at(ic).at(Rcol) = oldval;
        }
        if(isrow)
        {
            int oldval = _rowSol.at(ic).at(Rcol);
            _rowSol.at(ic).at(Rcol) = _rowSol.at(ic).at(Rrow);
            _rowSol.at(ic).at(Rrow) = oldval;
        }
        
        // Change the content of the diag vector and compute new nbdiag
        _nbdiag.at(ic)=0;
        for(unsigned int i=0; i<_ns; i++)
        {
            _diag.at(ic).at(i) = _currentConst.at(ic)(i,i);
            _nbdiag.at(ic) += _currentConst.at(ic)(i,i);
        }
          
    }
    
}

// Generate the sample
void TConstrLHS::generateSample(Option_t *option)
{
        
    try
    {
        if( _fConstrain == NULL )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TConstrLHS::generateTuple: The constraint function is not defined !!");
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
            
    TString soption = option;
    soption.ToLower();

    parseOption(option);
   
    if( !_bAlreadyFilled )
    {
	_sampling->generateSample();
    }
        
    _currentLHS = _tds->getMatrix();
        
    // Start the job !
    getConstMatrix();
        
    try
    {
        vector<int>::iterator it;
        int nredo=0;
        it = std::find( _nbdiag.begin(), _nbdiag.end(), -1); 
        if( it != _nbdiag.end() && _bAlreadyFilled )
        {
            int pos=it-_nbdiag.begin();
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TConstrLHS::generateTuple: The provided sample cannot fulfill the %d-th constraint provided", pos));
        }
        
        while ( it != _nbdiag.end() && nredo < 5)
        {            
            _sampling->generateSample();
            _currentLHS = _tds->getMatrix();
            getConstMatrix();

            it = std::find( _nbdiag.begin(), _nbdiag.end(), -1);
            nredo++;
        }

        if(nredo==5)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TConstrLHS::generateTuple: %d samples have been generated but none of them manage to fulfill the constraints", nredo)); 
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    
    int sum=0;
    for (auto& n : _nbdiag)
        sum += n;

    _iPerm=0;
    while ( sum != (int)_nConstr * _nS)
    {
        findNextPermutation();
        
        if (!(_iPerm % 10))
            printf("|");
        else if (!(_iPerm % 5))
            printf(":");
        else
            printf(".");
        
        if (!(++_iPerm % 50))
            printf("%04d\n", _iPerm);

        fflush(stdout);

        sum=0;
        for (auto& n : _nbdiag)
            sum += n;
        
    }

    // Finalise line
    printf(" Done ! \n");
    fflush(stdout);
    
    // Fill the TTree
    Double_t * valPattern = new Double_t[_nx + 1];
    _tds->deleteTuple();
    createTuple();
    
    
    if (_blog)
        cout << "Filling the TTree..." << endl;

    
    for (unsigned int is = 0; is < _ns; is++)
    {
        valPattern[0] = (Double_t)(is + 1);
        for (unsigned int ix = 0; ix < _nx; ix++)
            valPattern[ix + 1] = _currentLHS(is, ix);
        _ntsample->Fill(valPattern);
    }
    delete[] valPattern;
    valPattern = NULL;
    
    
    if (_blog)
        cout << "TTree filling done" << endl;
    
    
    // Affect the current TTree to the TDataServer
    _tds->setTuple(_ntsample);


    // Fill the formula attributes
    if ( !_bAlreadyFilled )
	fillOtherAttributes();
}

void TConstrLHS::fillOtherAttributes()
{
    _sampling->fillOtherAttributes();
}

///////////////////////////////////////////////////////////
//  Create the URANIE::DataServer::TDSTNtupleD of data
///////////////////////////////////////////////////////////
void TConstrLHS::createTuple()
{
    try
    {
        TString sLabelVariable = _tds->getIteratorName();
        if (_blog)
            cout << endl
                 << "**************************\n*** TConstrLHS::createTuple sLabelVariable ["
                 << sLabelVariable << "]" << endl;

        if (_nX != _lstOfAttributesToSample->GetEntries() && !_bAlreadyFilled )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form(
                                                           "TConstrLHS::createTuple: The number of stochastic attributes (%d) has been modified (%d).",
                                                           _nX, _lstOfAttributesToSample->GetEntries()));
        if (_blog)
            cout << " ** TConstrLHS::createTuple Size _nX [" << _nX << "]"
                 << endl;

        // Loop on stochastic attributes to have their names
        URANIE::DataServer::TAttribute *att;
	TList *attList = _tds->getListOfAttributes();
        TIter next(((!_bAlreadyFilled) ? _lstOfAttributesToSample : attList ) );
        while ((att = (URANIE::DataServer::TAttribute *) next()))
        {
            sLabelVariable += Form(":%s", att->GetName());
            if (_blog)
                cout << " ** stoatt[" << att->GetName() << "] sLabelVariable["
                     << sLabelVariable << "]" << endl;
        }

        if (_blog)
            cout << " ** Create the TTree (" << sLabelVariable
                 << ") of data ...";

	if(_ntsample!=NULL)
            delete _ntsample;
	
        _ntsample = new URANIE::DataServer::TDSNtupleD(Form("Sampling_%s", GetName()), GetTitle(), sLabelVariable);

        if (_blog)
            cout
                << "*** End Of TConstrLHS::createTuple\n**************************"
                << endl;

        delete attList;
        
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

//////////////////////////////////////////////////////////
//  Printing Log
//////////////////////////////////////////////////////////
void TConstrLHS::printLog(Option_t *option)
{
    cout << "** TConstrLHS::printLog  [" << option << "]" << endl;
    cout << "** _nS   [" << _nS << "]" << endl;
    cout << "** _nx   [" << _nx << "]" << endl;

}

} // Fin du namespace Sampler
} // Fin du namespace URANIE
// fin du fichier $RCSfile$.
