//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TEllipticalCopula.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mer oct 17 17:36:39 CEST 2012
 
 \brief Implementat de la classe URANIE::Sampler::TEllipticalCopula
 
 */

////// ROOT
#include "TMath.h"

////// URANIE
#include "TEllipticalCopula.h"
#include "UExceptions.h"

ClassImp (URANIE::Sampler::TEllipticalCopula)

namespace URANIE
{
namespace Sampler
{
TEllipticalCopula::TEllipticalCopula(URANIE::DataServer::TDataServer *tds,
        Option_t *option, Int_t nCalcul) :
        TCopula(tds, "EllipticalCopula", nCalcul)
{
    TString soption = option;
    soption.ToLower();
    if (_blog)
        cout << " ** constructor TEllipticalCopula" << endl;

    // Init the correlation matrix to identity 
    _matCorrelation.ResizeTo(_nX, _nX);
    _matCorrelation.UnitMatrix();
}

TEllipticalCopula::~TEllipticalCopula()
{
    if (_blog)
        cout << " ** destroy TEllipticalCopula" << endl;
}

////////////////////////////////////////////////////////////
//  Correlation factor
////////////////////////////////////////////////////////////
void TEllipticalCopula::setUserCorrelation(URANIE::DataServer::TAttribute *x,
        URANIE::DataServer::TAttribute *y, Double_t value)
{
    try
    {
        if (fabs(value) > 1.0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TEllipticalCopula::setUserCorrelation\n The correlation factor between (\"%s\", \"%s\") is greater 1.0 : [%f]",
                            x->GetName(), y->GetName(), value));
        Int_t indx = _tds->getAttributeIndex(x);
        Int_t indy = _tds->getAttributeIndex(y);

        setUserCorrelation(indx - 1, indy - 1, value);
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

void TEllipticalCopula::setUserCorrelation(TString xname, TString yname,
        double value)
{
    try
    {
        if (fabs(value) > 1.0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TEllipticalCopula::setUserCorrelation ERROR : the correlation factor  between (\"%s\", \"%s\") is greater 1.0 : [%f]",
                            xname.Data(), yname.Data(), value));
        Int_t indx = _tds->getAttributeIndex(xname);
        Int_t indy = _tds->getAttributeIndex(yname);

        setUserCorrelation(indx - 1, indy - 1, value);
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

void TEllipticalCopula::setUserCorrelation(Int_t indx, Int_t indy,
        Double_t value)
{
    try
    {
        if (fabs(value) > 1.0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TEllipticalCopula::setUserCorrelation ERROR : the absolute value of the correlation factor between attributes (%d, %d) is greater 1.0 : [%f]",
                            indx, indy, value));
        if (indx == indy)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TEllipticalCopula::setUserCorrelation ERROR : the two indices are equal: i = %d, j= %d]",
                            indx, indy));

        if ((indx >= _nX) || (indy >= _nX) || (indx < 0) || (indy < 0))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TEllipticalCopula::setUserCorrelation ERROR : index i [%d] or j [%d] is out of range",
                            indx, indy));

        // Update the correlation matrix
        _matCorrelation(indx, indy) = value;
        _matCorrelation(indy, indx) = value;

        if (_blog)
        {
            cout << " Updated correlation matrix:" << endl;
            _matCorrelation.Print();
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

////////////////////////////////////////////////////////////
//  Printing Log
////////////////////////////////////////////////////////////
void TEllipticalCopula::printLog(Option_t *option)
{
    cout << "TEllipticalCopula::printLog[" << option << "]" << endl;
}

} // Fin du namespace Sampler
} // Fin du namespace URANIE

