//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
// THaltonSequence mar jui 25 13:03:02 CEST 2006
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   THaltonSequence.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mar jul 25 13:03:02 CEST 2006
 
 \brief Interface de la classe URANIE::Sampler::THaltonSequence
 
 */
/**
 
 * @class URANIE::Sampler::THaltonSequence

 * @brief Description of the class THaltonSequence.
 *
 * \section  halton_description Description
 * (source Numerical Recipes - Chapter 7-7 page 309)
 * \par
 * In one dimension, the jth number \f$ H_j\f$  in the sequence is obtained by the followings steps :
 * \li Write \f$ j\f$ as a number in base \f$ b\f$, where \f$ b\f$ is some prime.
 * For example, \f$ j=17\f$ in base \f$ b=3\f$ is 122.
 * \li Reverse the digits and put a radix point (\em i.e. a decimal point base \f$ b\f$) in
 * front of the sequence. In our example, we get 0.221 base 3. It is the result \f$ H_j\f$.
 * \par
 * To get a sequence of n-tuples in n-space, you make each component a Halton sequence with
 * a different prime base \f$ b\f$. Typically, we can use the n primes.
 *
 * \section halton_implementation Implementation
 *
 * The code original is in the library \b seq_beta_04.21.01 de Duchemol and co
 * 
 */
#ifndef THALTONSEQUENCE_H
#define THALTONSEQUENCE_H
#include "TSequence.h"

#include <iostream>
using std::cout;
using std::endl;

namespace URANIE
{
namespace Sampler
{
class THaltonSequence: public TSequence
{
private:
    Double_t *Radical; //! Radical
    Int_t *Base;    //! The first n prime numbers for the base
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    THaltonSequence(ULong_t ndim);
    //! Default destructor
    virtual ~THaltonSequence();
    //@}

    void FirstPrimes(const ULong_t n);
    //---------------------------------------------
    /** @name Abstract method
     **/
    //@{
    void setInstance(const ULong_t n);
    void init();
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Sampler::THaltonSequence, ID_SAMPLER)
    //Classe de
};

} // Fin du namespace Sampler
} // Fin du namespace URANIE
#endif
