//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TImportanceSampling.h
 \b Created : mar jui 12 11:23:07 2011 \n
 \author Maud Salvoldi <ANP/OCC> \n
 \brief Interface file for the class URANIE::DataServer::TImportanceSampling

 \par Current version
 $Date$
 $Revision$
 
 
 */
#ifndef TImportanceSampling_H
#define TImportanceSampling_H

//ROOT
#include "Rtypes.h"
#include "TLeaf.h"
#include "TMath.h"

// uranie
#include "TDataServer.h"
#include "UExceptions.h"
#include "TStochAttribut.h"
#include "TSamplerStochastic.h"
#include "TSampling.h"

namespace URANIE
{
namespace Sampler
{
class TImportanceSampling: public TSamplerStochastic
{

public:
    enum EQuantileMethod
    {
        kQuantile1, kQuantile2, kQuantile3, kQuantile4, kQuantileUnknown
    };
    //! new tds build by the constructor
    URANIE::DataServer::TDataServer * _tds_IS;
    //!attribut which be subsitute
    URANIE::DataServer::TStochasticAttribute * _x;
    //!New law/attribut
    URANIE::DataServer::TStochasticAttribute * _is_x;
    //! mean 
    Double_t _mean;
    //! standart deviation 
    Double_t _std;
    //! probability of threshold crossing
    Double_t _proba;

protected:
    //! Boolean for edit the log.
    Bool_t _blog;
    //! THe TSampling object
    TSampling *_sampling;

    //////////////methode of the class
public:
    /** @name Constructor and Destructor
     **/
    //@{
    //! Constructor
    TImportanceSampling(URANIE::DataServer::TDataServer *tds, TString xname,
            URANIE::DataServer::TStochasticAttribute *att, Int_t ns);
    //! Destructor
    virtual ~TImportanceSampling();
    //@}

    //---------------------------------------------
    /** @name Setting and Getting global informations
     **/
    //@{
    /*! Return the tds
     \brief This method returns the new tds build by the contructor of the object TIportanceSampling.
     \brief This tds is a copy of the iital tds where a attribut was replace by a new attribut which follows
     the new law given in parameter of the constructor.
     \brief The inital attrinut replaced is attached to the new attribute.
     */
    URANIE::DataServer::TDataServer *getTDS()
    {
        return _tds_IS;
    }
    //@}

    //---------------------------------------------
    /** @name Compute Statistical parameters
     **/
    //@{
    /*! Return mean
     \brief This method computes the mean of the ouput attribut given in parameter.
     \param u : name of the output attribut
     */
    Double_t computeMean(TString u);
    /*! Return the standard deviation
     \brief This method computes and returns the stantard deviation of the ouput attribut given in parameter.
     \param u : name of the output attribut
     */
    Double_t computeStd(TString u);
    /*! Return the probability of threshold crossing 
     \brief This method computes and returns the probability of threshold crossing of the ouput attribut given in parameter.
     \param u : name of the output attribut,
     \param dThreshold : the threshold parameter
     */
    Double_t computeThreshold(TString u, Double_t dThreshold);
    /*! Return the \f$\alpha\f$-quantile of the attribute
     This method computes and returns the \f$\alpha\f$-quantile of the attribute given in parameter.
     \param sVar : name of the attribute
     \param dAlpha : the parameter \f$\alpha\f$.
     */
    Double_t computeQuantile(TString sVar, Double_t dAlpha,
            EQuantileMethod nMethod = kQuantile2);

    //@}

    //---------------------------------------------
    /** @name Generates the sample
     **/
    //@{
    //! Generates the Sample
    void generateSample(Option_t *option = "");
    //@}

    //---------------------------------------------
    /** @name Treats the TAttributes
     **/
    //@{
    //! Creates the List of attributes to simulate
    void createListOfAttributes();
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    void setLog()
    {
        _blog = kTRUE;
    }
    void unsetLog()
    {
        _blog = kFALSE;
    }
    void changeLog()
    {
        _blog = _blog ? kFALSE : kTRUE;
    }
    Bool_t getLog()
    {
        return _blog;
    }
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Sampler::TImportanceSampling, ID_DATASERVER)
    //Definition of an variable

};

} // Fin du namespace DataServer
} // Fin du names

#endif
