//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*! \file TMaxiMinLHS.h
 *  \author Jean-Baptiste Blanchard <jean-baptiste.blanchard@cea.fr>
 *  \date   $Date$
 *  \brief Interface of the class URANIE::Sampler::TMaxiMinLHS.
 *
 *  Created: Fri Oct  1 13:37:23 2004
 *
 */

/**
 * @class URANIE::Sampler::TMaxiMinLHS
 *
 * @brief Description of the class TMaxiMinLHS.

 This class is used to produce a maximin lhs doe. It starts either with an empy tds or 
 an already filled one, with the hypothesis that it has been filled with an LHS doe 

 An simulated annealing algorithm is implemented here to deal with the permutation.
 This implementation has be done by Guillaume Damblin.
 */

#ifndef TMAXIMINLHS_H
#define TMAXIMINLHS_H

#ifndef ROOT_TMatrixD
#include "TMatrixD.h"
#endif
#ifndef ROOT_TVectorD
#include "TVectorD.h"
#endif
#ifndef ROOT_TNtuple
#include "TNtuple.h"
#endif
#ifndef ROOT_TString
#include "TString.h"
#endif
#ifndef ROOT_TNtupleD
#include "TNtupleD.h"
#endif
#ifndef ROOT_TNamed
#include "TNamed.h"
#endif

// Uranie
#include "TSamplerStochastic.h"
#include "TSampling.h"
#include "TAttribute.h"
#include "TStochAttribut.h"
#include "TDataServer.h"

namespace URANIE
{
namespace Sampler
{
class TMaxiMinLHS: public TSamplerStochastic
{
    // Associations

    // Attributes
public:
    /** @name main parameters
     **/
    double _SAinitTemp; ///< Initial temperature of the simulated Annealing algorithm for maximin lhs
    double _SACoeff; ///< Temperature coefficient of the simulated Annealing algorithm for maximin lhs
    unsigned int _SAN; ///<  Number of iteration the simulated Annealing algorithm for maximin lhs
    unsigned int _SAinnerN; ///< Inner Number of iteration before changing the temperature of the simulated Annealing algorithm for maximin lhs
    //@}
    /** @name internal variables
     **/
    bool _bAlreadyFilled; ///< If the tree is already filled

    TMatrixD _currentLHS, _varyingLHS, _currentDist, _varyingDist, _fp;
    TSampling *_sampling; ///< Pointer to a TSampling object if tds is empty;
    unsigned int _ns, _nx; ///< Local counter for number of samples and variables
    double *_vi, *_vj; ///< arrays used throughout many calculations
    //@}
    // Operations
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Creates a TMaxiMinLHS object from a TDataServer, definig the SA properties  
    /*!

      none of these is given, "lhs" is chosen.                
     
      \param tds        (TDataServer *) pointer to the TDataServer.
      \param sampleSize (Int_t) number of values to generate for each random variable.
      \param sainitTemp (double): initial temperature for simulated annealing algo (see methodology for explanation)
      \param saCoeff (double): coefficient used to decrease temperature for simulated annealing algo (see methodology for explanation)
      \param saN (int): number of iterations in the outer loop for simulated annealing algo (see methodology for explanation)
      \param sinneraN (int): number of iterations in the inner loop for simulated annealing algo (see methodology for explanation)
     
      \exception UErrorExceptions if sampleSize <= 0
    */
    TMaxiMinLHS(URANIE::DataServer::TDataServer *tds, int sampleSize, double sainitTemp, double saCoeff, int saN, int sainnerN);

    //! Default destructor
    virtual ~TMaxiMinLHS();
    //@}

    //---------------------------------------------            
    /** @name Generate the sample
     **/
    
    //! Generates the sample 
    /*!
      Generates the random data and fills the TDataServer with it.                
      \param option (Option_t *) option string (Default = "")
    */
    void generateSample(Option_t *option = "");

    //! Set the random generator seed
    void setSeed(Int_t ind = 0);

    //! transform the sample
    /*!
      Tranform the sample from the probability space to the physical one depending on the boolean
      \param RefMat : the reference TMatrixD containing the information
      \param toProbaSpace: a boolean to state whether transformation should be done to the probability space or from it
     */
    void transformMatrix(TMatrixD &RefMat, bool toProbaSpace);
    
    //! Fills the TAttributeFormula and totally correlated attributes 
    virtual void fillOtherAttributes();
    //@}


    //---------------------------------------------
    /** @name Method for the LHS_maximin producation
     **/
    //@{
    //! Compute the minimal distance between points
    /*!
      Compute the minimal distance between points
     
      \param lhs (TMatrixD) the lhs plan under study
    */
   static Double_t getMinDist(const TMatrixD &lhs);

    //! Compute the distance matrix
    /*!
      Compute the distance matrix

      \param Dist (TMatrixD) the distance matrix of the lhs plan under study
      \param lhs (TMatrixD) the lhs plan under study
    */
    static void getDistMatrix(TMatrixD& Dist, const TMatrixD &lhs);

    //! Compute the Lp  norm
    /*!
      Compute the Lp norm

      \param lhs (TMatrixD) the distance matrix of the lhs plan under study
    */
    Double_t getLPNorm(const TMatrixD &Dist_);

    //! update the Lp  norm
    /*!
      Update the Lp norm

      \param lhs (TMatrixD) the distance matrix of the lhs plan under study
    */
    Double_t getUpdatedLPNorm(Double_t LP_norm_, unsigned int nrowRand1_, unsigned int nrowRand2_);
    
    //! Generate the new LHS
    /*!
      Generate a new lhs doe by doing all the permutation allowed by the Simulated Annealing algorithm.
      The method requires four parameters that shoud be set by setSAProperties()

      \param lhs (TMatrixD) the lhs plan under study
    */
    TMatrixD maximinSA_LHS(const TMatrixD &lhs);

    //! Set the Simulated Annealing algorithm properties
    /*! 
      Set the Simulated Annealing algorithm properties

      \param sainitTemp (double)  Initial temperature of the simulated Annealing algorithm for maximin lhs
      \param saCoeff (double)  Temperature coefficient of the simulated Annealing algorithm for maximin lhs
      \param saN (double)   Number of iteration of the simulated Annealing algorithm for maximin lhs
      \param sainnerN (double)  Inner Number of iteration before changing the temperature of the simulated Annealing algorithm for maximin lhs
    */
    void setSAProperties(double &sainitTemp, double &saCoeff, int &saN, int &sainnerN);
    
    //! Permute the content of the lhs plan
    /*!
      Permute the content of the lhs plan

      \param lhs_new (TMatrixD) the new lhs after permutation
      \param lhs (TMatrixD) the lhs plan under study
    */
    void Permutation(unsigned int & ncolRand, unsigned int & nrowRand1, unsigned int & nrowRand2);
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{                
    //! Prints the log
    /*!
      \param option (Option_t *) option string (Default = "")
    */
    virtual void printLog(Option_t *option = "");
    void createTuple();
    //@}

    ClassDef(URANIE::Sampler::TMaxiMinLHS, ID_SAMPLER)
    //Definition d'un plan d'exp�rience
};

} // Fin du namespace Sampler
} // Fin du namespace URANIE
#endif
// fin du fichier $RCSfile$.
