//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TOATSampling.h
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr> 
 \date   2011.02.24

 \brief  "One factor At a Time" design of experiment. It has been developped 
 mainly for finite differences and sensitivity computation.


 */
////// ROOT
#include "TPRegexp.h"
#include "TMath.h"
#include "TObjString.h"

////// URANIE
#include "TOATSampling.h"
#include "TAttribute.h"      // for DLE_MAX ? 
#include "TStochAttribut.h"
#include "UExceptions.h" 

using namespace URANIE::DataServer;

ClassImp (URANIE::Sampler::TOATSampling)

namespace URANIE
{
namespace Sampler
{
//////////////////////////////////////////////////////////////
//  Constructor : from TDataServer + Methode + size
////////////////////////////////////////////////////////////
TOATSampling::TOATSampling(URANIE::DataServer::TDataServer *tds,
        Option_t *option) :
        TSamplerDeterminist(tds, "", 1), _smode(option), _dStepPrecision(dStepPrecision)
{
    setOption(option);

    // Create the internal attributes which identify which set of nominal values 
    // is used (__nominal_set__) and which attribute is modified (__modified_att__)            
    TAttribute *attNominalSet = new TAttribute("__nominal_set__",					       
            "Index of the set of nominal value");
    attNominalSet->delShare();
    TAttribute *attModifiedAtt = new TAttribute("__modified_att__",
            "Index of the modified attribute");
    attModifiedAtt->delShare();
    //attNominalSet->setOrigin(TAttribute::kInternal);
    //attModifiedAtt->setOrigin(TAttribute::kInternal);

    _tds->addAttribute(attNominalSet);
    _tds->addAttribute(attModifiedAtt);

    try
    {
        // verify that the attributes have a default value
        Double_t defValue;
        for (Int_t i = 0; i < _nX; i++)
        {
            URANIE::DataServer::TAttribute *att =
                    (URANIE::DataServer::TAttribute *) _lstOfAttributesToSample->At(
                            i);
            if (att->getDefaultValue(defValue))
                continue;
            else
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        Form(
                                "*** ERROR in TOATSampling::TOATSampling: Attribute %s has no default value",
                                att->GetName()));
        }

        // Check that the attributes are compatible with the chosen mode             
        if (_smode.Contains("minmax"))
        {
            // verify that the attributes have finite bounds
            for (Int_t i = 0; i < _nX; i++)
            {
                URANIE::DataServer::TStochasticAttribute *att =
                        (URANIE::DataServer::TStochasticAttribute *) _lstOfAttributesToSample->At(
                                i);
                if (att->getLowerBound()
                        > (-1.0 * DLE_MAX)&& att->getUpperBound() < DLE_MAX)
                    continue;
                else
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__,
                            __LINE__,
                            Form(
                                    "*** ERROR in TOATSampling::TOATSampling: minmax mode can't be applied: attribute %s has at least one infinite boundary.",
                                    att->GetName()));
            }
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

TOATSampling::~TOATSampling()
{
    if (_blog)
        cout << "*** Destroy The TOATSampling object" << endl;
}

void TOATSampling::setOption(Option_t *option)
{
    _smode = option;
    _smode.ToLower();
    _smode.ReplaceAll(" ", "");

    // The following regular expressions match the "steps=X" and "steps=X%" 
    // options and allow to extract the value of X.
    TPMERegexp stepRegExp("steps=([0-9]+\\.?[0-9]*)");
    TPMERegexp stepRegExpPercent("steps=([0-9]+\\.?[0-9]*)%");

    try
    {
        if (_smode.Contains("minmax"))
        {
            _smode = "minmax";
            _bisPercent = kFALSE;
            _dstepSize = 0.0;
        }
        else if (stepRegExpPercent.Match(_smode) > 0)
        {
            _smode = stepRegExpPercent[0];
            _bisPercent = kTRUE;
            _dstepSize = stepRegExpPercent[1].Atof();
        }
        else if (stepRegExp.Match(_smode) > 0)
        {
            _smode = stepRegExp[0];
            _bisPercent = kFALSE;
            _dstepSize = stepRegExp[1].Atof();
        }
        else
        {
            _smode = "steps=1%";
            _bisPercent = kTRUE;
            _dstepSize = 1.0;
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                    "*** WARNING in TOATSampling::setOption: Option parameter does not contain any recognized step value. Using 1% of the default values");
        }
    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
    }

}

/*******************************************/
/*           Generate the sample           */
/*******************************************/
void TOATSampling::createTuple()
{

    TString sLabelVariable = _tds->getIteratorName();
    if (_blog)
        cout << endl
                << "**************************\n*** TOATSampling::createTuple sLabelVariable ["
                << sLabelVariable << "]" << endl;

    try
    {
        if (_nX != _lstOfAttributesToSample->GetEntries())
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TOATSampling::createTuple[%d] The size of stochastic (%d) attributes have been modified.",
                            _nX, _lstOfAttributesToSample->GetEntries()));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    if (_blog)
        cout << " ** TOATSampling::createTuple Size _nX [" << _nX << "]"
                << endl;

    // Loop on stochastic attributes to have their names
    URANIE::DataServer::TAttribute *att;
    TIter next(_lstOfAttributesToSample);
    while ((att = (URANIE::DataServer::TAttribute *) next()))
    {
        sLabelVariable += Form(":%s", att->GetName());
        if (_blog)
            cout << " ** stoatt[" << att->GetName() << "] sLabelVariable["
                    << sLabelVariable << "]" << endl;
    }
    // Add two internal attributes to identify which set of nominal values 
    // is used (__nominal_set__) and which attribute is modified (__modified_att__)
    sLabelVariable += ":__nominal_set__:__modified_att__";

    if (_blog)
        cout << " ** Create the TTree (" << sLabelVariable << ") of data ...";

    // Create the TTree of data
    _ntsample = new URANIE::DataServer::TDSNtupleD(
            Form("Sampling_%s", GetName()), GetTitle(), sLabelVariable);

    if (_blog)
        cout
                << "*** End Of TOATSampling::createTuple\n**************************"
                << endl;

}

void TOATSampling::generateSample(Option_t *option)
{
    // The only option available is the list of input attributes, i.e. 
    // the attributes we want to modify.
    TString sInputs = "";
    TString soption = option;
    soption.ReplaceAll(" ", "");

    TPMERegexp inputRegExp("in[puts]*=(.+)[;,$]?");
    if (inputRegExp.Match(soption) > 0)
        sInputs = inputRegExp[1];

    if (_blog)
        cout << " ** TOATSampling::GenerateSample soption[" << soption << "]"
                << endl;

    // Create the TTree of data
    try
    {
      createTuple();
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        throw ue;
    }

    if (_nX == 0)
        return;
    if (_blog)
        cout << "nvar[" << _nX << "]" << endl;

    // Initialisation of the attributes values to their default value.
    // NOTE: if needed, one could give the choice to the user to use 
    //       another nominal value, like the theoretical median.
    // (Nicolas Gilardi, 2011.02.25)
    Double_t *valpattern = new Double_t[_nX + 3];
    Double_t defValue;

    valpattern[0] = 1.0;
    for (Int_t i = 0; i < _nX; i++)
    {
        URANIE::DataServer::TAttribute *att =
                (URANIE::DataServer::TAttribute *) _lstOfAttributesToSample->At(
                        i);
        att->getDefaultValue(defValue);
        valpattern[i + 1] = defValue;
    }
    // Fill the internal attributes. Setting "__modified_att__" to 0 
    // means that this line contains the nominal values of the attributes.
    // NOTE: for the moment, only one set of nominal values can be used.
    //       This is the reason why "__nominal_set__" is always set to 1.0.
    // (Nicolas Gilardi, 2011.06.10)
    valpattern[_nX + 1] = 1.0;
    valpattern[_nX + 2] = 0.0;

    if (_blog)
    {
        cout << "valpattern = [" << valpattern[0];
        for (Int_t k = 1; k < _nX + 3; k++)
            cout << ", " << valpattern[k];
        cout << "]" << endl;
    }

    _ntsample->Fill(valpattern);

    // Case of the various OAT sampling modes.
    // NOTE: in the present implementation there is some unnecessary code 
    //       duplication. This is done for the sake of readability and 
    //       extensibility, as future sampling modes may not follow the 
    //       "two realisations per attribute" prototype.
    // (Nicolas Gilardi, 2011.02.25)
    try
    {
        if (_smode.Contains("minmax"))
        {
            if (_blog)
                cout << "working in [minmax] mode..." << endl;

            for (Int_t i = 0; i < _nX; i++)
            {
                // The attribute to modify is retrieved from the list of 
                // attributes to sample.
                URANIE::DataServer::TAttribute *att =
                        (URANIE::DataServer::TAttribute *) _lstOfAttributesToSample->At(
                                i);

                // If this attribute is not in the input list, we skip it.
                if (sInputs.Contains(att->GetName()) || sInputs == "")
                {
                    // The nominal value of this attribute, already stored in the 
                    // valpattern array, is saved.
                    Double_t nominalValue = valpattern[i + 1];

                    // Set the values of the internal attributes
                    valpattern[_nX + 1] = 1.0;
                    valpattern[_nX + 2] = i + 1.0;

                    // The nominal value is replaced first by the lower bound, 
                    // then by the upper bound of the attribute's distribution.
                    // The other values are unchanged, except for the iterator.
                    valpattern[0] = 2.0 * (i + 1);
                    valpattern[i + 1] = att->getLowerBound();
                    _ntsample->Fill(valpattern);

                    valpattern[0] = 2.0 * (i + 1) + 1.0;
                    valpattern[i + 1] = att->getUpperBound();
                    _ntsample->Fill(valpattern);

                    // The valpattern array is restored to its initial state by 
                    // putting back the nominal value of the attribute. 
                    valpattern[i + 1] = nominalValue;
                }
                else if (_blog)
                {
                    cout << "Attribute name: " << att->GetName()
                            << "; sInputs: " << sInputs << endl;
                }

            }
        }
        else if (_smode.Contains("steps"))
        {
            if (_blog)
                cout << "working in [steps] mode..." << endl;

            for (Int_t i = 0; i < _nX; i++)
            {
                // If this attribute is not in the input list, we skip it.
                URANIE::DataServer::TAttribute *att =
                        (URANIE::DataServer::TAttribute *) _lstOfAttributesToSample->At(
                                i);
                if (sInputs.Contains(att->GetName()) || sInputs == "")
                {
                    // The nominal value of this attribute, already stored in the 
                    // valpattern array, is saved.
                    Double_t nominalValue = valpattern[i + 1];

                    // Set the values of the internal attributes
                    valpattern[_nX + 1] = 1.0;
                    valpattern[_nX + 2] = i + 1.0;

                    // The real step value is computed. It can be relative to 
                    // the attribute's absolute or a percentage of the nominal value. 
                    Double_t step;
                    if (_bisPercent)
                        step = _dstepSize / 100.0 * nominalValue;
                    else
                        step = _dstepSize;

                    // Case of Step increment is lower than StepPrecision
                    //   then step = getStepValue ELSE dStepPrecision parameter
                    if (( TMath::Abs(step) < _dStepPrecision ) &&
                            ( !att->getStepValue(step) )) {
                            step = _dStepPrecision;
                    }

                    // The nominal value is first decreased by one "step", then 
                    // increased by one "step". The other values are unchanged, 
                    // except for the iterator.  
                    valpattern[0] = 2.0 * (i + 1);
                    valpattern[i + 1] = TMath::Max(att->getLowerBound(),
                            nominalValue - step);
                    _ntsample->Fill(valpattern);

                    valpattern[0] = 2.0 * (i + 1) + 1.0;
                    valpattern[i + 1] = TMath::Min(att->getUpperBound(),
                            nominalValue + step);
                    _ntsample->Fill(valpattern);

                    // The valpattern array is restored to its initial state by 
                    // putting back the nominal value of the attribute. 
                    valpattern[i + 1] = nominalValue;
                }
                else if (_blog)
                {
                    cout << "Attribute name: " << att->GetName()
                            << "; sInputs: " << sInputs << endl;
                }

            }
        }
        else
        {
            if (_blog)
                cout << "unknown mode: " << _smode << endl;

            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "*** ERROR in TOATSampling::generateSample: The given OAT sampling mode [%s] is not yet implemented or unknown",
                            _smode.Data()));
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    // Affect the current TTree to the TDataServer
    if (_blog)
        cout << "Affect the TTree to the tds..." << endl;

    _tds->setTuple(_ntsample);

    // Fill the formula attributes
    if (_blog)
        cout << "Fill the other attributes..." << endl;
    fillOtherAttributes();

    delete[] valpattern;
    valpattern = NULL;
}

////////////////////////////////////////////////////////////
//  Printing Log
////////////////////////////////////////////////////////////
void TOATSampling::printLog(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    cout << "TOATSampling::printLog" << endl;
    cout << "  _smode[" << _smode << "]" << endl;
    cout << "  _dstepSize[" << _dstepSize << "]" << endl;
    TSampler::printLog();
    cout << "End of TOATSampling::printLog" << endl;
}

} // Fin du namespace Sampler
} // Fin du namespace URANIE
// fin du fichier $RCSfile$.
