//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TOATSampling.h
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr> 
 \date   2011.02.24

 \brief  This class creates a "One factor At a Time" design of experiment. It has been developped 
 mainly for finite differences sensitivity computation.


 */
#ifndef TOATSampling_H
#define TOATSampling_H
/*!

 @class URANIE::Sampler::TOATSampling

 @brief This class creates a "One factor At a Time" design of experiment. It has 
 been developped mainly for finite differences sensitivity computation.

 The principle of the OAT design of experiment is to modify only one attribute at 
 a time, all the other factors being kept at a nominal (default) value. 
 This is the simplest way to compute partial derivatives or sensitivities. For the 
 latter application however, it might not be the best solution as this approach 
 does not explore multi-factorial effects.       

 The first point of the design always corresponds to all the attributes set to their 
 nominal value. Then, depending on the application, the following options are available: 
 \li "minmax": the varying attribute will take its maximum and minimum values. 
 All the others are given their nominal value.
 \li "steps=x": the varying attribute will move around its nominal value by a given 
 step "x", usually small. "x" can either be a real value (e.g. "steps=3.1415") or a 
 percentage of the nominal value (e.g. "steps=0.123%"). This variation step will 
 be applied to all the attributes. Please note that if the modified value  
 is outside of the attribute's range of variation, the value of the relevant boundary 
 will be used instead.

 As a result, an OAT design of experiment will provide 2*n+1 realisations of the 
 input variables, n being the number of attributes to sample.  

 \note When the sampling is created, two attributes are added:

 \li __nominal_set__: so far it is always set to 1. This attribute will be used 
 when the variables can vary around more than one nominal value.
 \li __modified_att__: for a given realisation of the sampling, its value is the 
 index of the attribute that is modified. This index starts 
 from 1. if __modified_att__ is 0, it means that all the 
 attributes are set to their nominal value.
 */
#include <iostream>
using std::cout;
using std::endl;

#include "TSamplerDeterminist.h"
#include "TSequence.h"

#define dStepPrecision 1.0e-16


namespace URANIE
{
namespace Sampler
{
class TOATSampling: public TSamplerDeterminist
{

    // Attributes
private:
    Double_t _dstepSize;
    TString _smode;
    Bool_t _bisPercent;

    Double_t _dStepPrecision;

    // Operations
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //!  Constructor of the OAT sampler.
    /*!
     \param tds (URANIE::DataServer::TDataServer*) Data server containing the attributes. 
     \param option (Option_t*) If option contains "minmax", the min and max values of the 
     attributes are used. 
     If it contains "steps=X", the attributes values are the default 
     value plus or minus X. 
     If it contains "steps=X%", the attributes values are the default 
     value plus or minus X percent of the default value (Default = "steps=1%"). 
     
     \exception UErrorExceptions if option is "minmax" and some attributes are unbounded.
     \exception UErrorExceptions if option is "steps" and some attributes have no default value.
     */
    TOATSampling(URANIE::DataServer::TDataServer *tds, Option_t *option =
            "steps=1%");

    //! Default destructor
    virtual ~TOATSampling();
    //@}

    //---------------------------------------------
    /** @name Getting and setting parameters
     **/
    //@{
    //! Get the step size parameter
    virtual Double_t getStepSizeParameter()
    {
        return _dstepSize;
    }
    //! Get OAT sampling mode
    virtual TString getSamplingMode()
    {
        return _smode;
    }
    //! Get the boolean
    //!
    virtual Bool_t isPercentMode()
    {
        return _bisPercent;
    }

    //! Set the step size parameter
    /*!
     This function sets the values of the _dstepSize and _smode variables after parsing the option string.
     
     \param option (Option_t *) If option contains "minmax", the min and max values of the 
     attributes are used. 
     If it contains "steps=X", the attributes values are the default 
     value plus or minus X. 
     If it contains "steps=X%", the attributes values are the default 
     value plus or minus X percent of the default value (Default = "steps=1%"). 
     */
    virtual void setOption(Option_t *option);

    /*!
     * \brief getStepPrecision
     * \return (Double_t) the step precision value where to change with the step value in the TAttribute
     */
    Double_t getStepPrecision() {
        return _dStepPrecision;
    }

    void setStepPrecision(Double_t dval) {
        _dStepPrecision = TMath::Max(0.0, dval);
    }

    //@}

    //---------------------------------------------
    /** @name Generation of the sample
     **/
    //@{
    //! Create the data tree for the data server
    virtual void createTuple();

    //! Generates the sample
    /*!
     This function creates the OAT sampling data.
     
     \param option (Option_t *) If option contains "in=" or "inputs=" followed by list of attribute 
     names separated by colons, only the attributes listed will vary during 
     the OAT sampling (provided they exist in the TDS, of course...).
     */
    virtual void generateSample(Option_t *option = "");
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    //! Prints the log
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Sampler::TOATSampling, ID_SAMPLER)
};
} // Fin du namespace Sampler
} // Fin du namespace URANIE

#endif
// fin du fichier $RCSfile$.
