//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TOATSamplingMultiNominal.h
 \author Nicolas Gilardi <nicolas.gilardi@cea.fr> 
 \date   2012.08.10

 \brief  This class creates a "One factor At a Time" design of experiment. It has been developped 
 mainly for finite differences sensitivity computation.


 */
#ifndef TOATSamplingMultiNominal_H
#define TOATSamplingMultiNominal_H
/*!

 @class URANIE::Sampler::TOATSamplingMultiNominal

 @brief This class has been renamed. Please use TOATDesign instead.

 The principle of the OAT design of experiment is to modify only one attribute at 
 a time, all the other factors being kept at a nominal value. 
 This is the simplest way to compute partial derivatives or sensitivities. For the 
 latter application however, it might not be the best solution as this approach 
 does not explore multi-factorial and non-linear effects.  

 The present implementation is an improvement of TOATSampling. The user has now 
 two ways of defining the nominal values of the attributes: by setting the default 
 value for each attribute or by providing a tds with a set of nominal values. In 
 the latter case, more than one set of nominal values can be provided, and an OAT 
 sampling will be generated for each one of them.

 The generated values are sampled inside an interval, centered on the nominal 
 value. The width of this interval is defined as the "range". For each changing 
 attribute, the user has to define the range, either as a numerical value, or as 
 the name of an attribute of the tds. In both cases, an option allows the 
 resulting value of the range to be interpreted as a percentage of the nominal 
 value.

 Two modes of data generation are available. The "regular" mode will generate data 
 regularily spaced over the interval. The "random" mode will generate random 
 values inside the interval. In the "random" mode, the changing attributes must 
 be TStochasticAttribute objects. This means that (at the moment) this mode can't 
 be used on attributes loaded from a file.   

 The number of generated values is not limited. It just needs to be greater or 
 equal to 2. And if the sampling is "regular", it must also be even. But if an 
 odd value is given, it will be reduced to an even one.

 Here is a small example script, showing a regular OAT sampling on 2 attributes 
 whose nominal values are read from a file:

 ----- file myNominalValues.dat -----
 \verbatim
 #COLUMN_NAMES: x1 | rx1 | x2

 3.0  0.4  2.0
 5.0  0.2  3.0

 \endverbatim
 -----  data generation script  -----
 \code
 {
 // Construction of the data server
 URANIE::DataServer::TDataServer tds("myTDS","OAT sampling TDS");
 // Load the nominal values
 tds.fileDataRead("myNominalValues.dat");
 
 // Create the sampler
 URANIE::Sampler::TOATSamplingMultiNominal s(&tds, "regular", 4);
 
 // Set the range of attribute x1. It will be the value of attribute rx1.
 s.setRange("x1","rx1"); 
 // Set the range of attribute x2. It will be 10.0% of x2's nominal value.
 s.setRange("x2",10.0,kTRUE); 
 // Generate an OAT sampling where only x1 and x2 will vary
 s.generateSample();
 
 // Save the results to a file 
 tds.exportData("myOATSampling.dat");
 }
 \endcode
 -----  file myOATSampling.dat  -----
 \verbatim
 #COLUMN_NAMES: x1| rx1| x2| __nominal_set__| __modified_att__| myTDS__n__iter__

 3.00   0.4   2.000   1.0   -1.0    1.0 
 2.80   0.4   2.000   1.0    0.0    2.0 
 3.20   0.4   2.000   1.0    0.0    3.0 
 2.90   0.4   2.000   1.0    0.0    4.0 
 3.10   0.4   2.000   1.0    0.0    5.0 
 3.00   0.4   1.900   1.0    2.0    6.0 
 3.00   0.4   2.100   1.0    2.0    7.0 
 3.00   0.4   1.950   1.0    2.0    8.0 
 3.00   0.4   2.050   1.0    2.0    9.0 
 5.00   0.2   3.000   2.0   -1.0   10.0 
 4.90   0.2   3.000   2.0    0.0   11.0 
 5.10   0.2   3.000   2.0    0.0   12.0 
 4.95   0.2   3.000   2.0    0.0   13.0 
 5.05   0.2   3.000   2.0    0.0   14.0 
 5.00   0.2   2.850   2.0    2.0   15.0 
 5.00   0.2   3.150   2.0    2.0   16.0 
 5.00   0.2   2.925   2.0    2.0   17.0 
 5.00   0.2   3.075   2.0    2.0   18.0  
 \endverbatim

 The attribute __nominal_set__ refers to the set of nominal values used to generate 
 the data. The attribute __modified_att__ gives the index of the attribute whose 
 value has been modified. If it is -1, this means that all the values are nominal.
 */

#include <iostream>
#include <map>
using namespace std;

#include "TSamplerStochastic.h"
#include "TSequence.h"
#include "TMap.h"

namespace URANIE
{
namespace Sampler
{
class TOATSamplingMultiNominal: public TSamplerStochastic
{

    // Attributes
private:
    TString _sMode;    ///< Sampling mode. So far, can be "random" or "regular".
    TString _sGenMode; ///< Data generation mode. Only used in the random mode context, can be "lhs", or "srs".
    Int_t _nNbModif; ///< Number of modification for each nominal value of each changing attribute.
    TList _changingAttList;        ///< List of changing attributes.
    TMap _infoRangeMap; ///< A list mapping the names of the changing attributes and their corresponding range information.
    URANIE::DataServer::TDSNtupleD* _nominalValueTuple; ///< A tuple storing the nominal values of the attributes.
    Bool_t _bHasData; ///< Boolean value set to TRUE if the data server _tds contains data.

    // Operations
protected:
    //---------------------------------------------
    /** @name State machine
     For the stochastic methods, two states can be created \em init and \em terminate
     which are executed just before the generation of the sample for the first one, 
     and after the generation for the second one. They can be used for pre and post 
     treatment purpose.
     **/

    //@{
    //! Initialise the sampling procedure.
    /*!
     This function creates a tuple containing the nominal values of all 
     the attributes of the data server. It also creates two new 
     attributes:
     - __nominal_set__: tells the user which set of nominal value is used;
     - __modified_att__: tells which attribute has been modified.
     
     \exception UErrorExceptions if the original data server is empty and one of its attributes has no default value.            
     */
    virtual void init();

    //! The post-processing step. Nothing is done.
    virtual void terminate()
    {
    }
    //@}
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //!  Constructor of the OAT sampler.
    /*!
     \param tds (URANIE::DataServer::TDataServer*) Data server containing the attributes. 
     \param mode (TString): the sampling mode. It can be either:
     - "regular" (or "reg"): generate regularily spaced values around the nominal value.
     - "lhs": generate random values using the "lhs" sampling method
     - "srs": generate random values using the "srs" sampling method
     - "random" (or "rand"): same as "srs".
     (Default = "regular").    
     \param number (Int_t): the number of values to generate for each 
     nominal value of a given attribute. If _sMode 
     is "regular", number must be even and equal 
     or greater than 2. Otherwise, _nNbModif is 
     set to max(2, number-1).
     \exception UErrorExceptions if "number" is negative or null
     \exception UErrorExceptions if "number" is odd and "mode" is "regular".
     \exception UErrorExceptions if "mode" is unknown.  
     \deprecated This class has been renamed. Please use TOATDesign instead.
     */
    TOATSamplingMultiNominal(URANIE::DataServer::TDataServer *tds,
            TString mode = "regular", Int_t number = 2);

    //! Default destructor
    virtual ~TOATSamplingMultiNominal();
    //@}

    //---------------------------------------------
    /** @name Getting and setting parameters
     **/
    //@{
    //! Returns TRUE if the internal data server contains data.
    virtual Bool_t getHasData()
    {
        return _bHasData;
    }

    //! Get the number of modifications per attribute and nominal values.
    virtual Double_t getNumberOfModifications()
    {
        return _nNbModif;
    }

    //! Get OAT sampling mode.
    virtual TString getSamplingMode()
    {
        return _sMode;
    }

    //! Get random generation method.
    virtual TString getGenerationMode()
    {
        return _sGenMode;
    }

    //! Get the list of changing attributes.
    virtual TList* getChangingAttList()
    {
        return &_changingAttList;
    }

    //! Get the list mapping the names of the changing attributes and their corresponding range information.
    virtual TMap* getRangeInformation()
    {
        return &_infoRangeMap;
    }

    //! Get the tuple storing the nominal values of the attributes.
    virtual URANIE::DataServer::TDSNtupleD* getNominalValueTuple()
    {
        return _nominalValueTuple;
    }

    //! Set the sampling mode and number of modifications
    /*!
     This function sets the values of the _sMode and _nNbModif variables.            
     \param mode (TString): the sampling mode. It can be either:
     - "regular": generate regularily spaced values around the nominal value.
     - "lhs":     generate random values using the "lhs" sampling method
     - "srs":     generate random values using the "srs" sampling method
     - "random":  same as "srs".
     \param number (Int_t): the number of values to generate for each 
     nominal value of a given attribute. If _sMode 
     is "regular", number must be even and equal 
     or greater than 2. Otherwise, _nNbModif is 
     set to max(2, number-1).
     \exception UErrorExceptions if "mode" is unknown.                   
     \exception UErrorExceptions if "number" is odd and "mode" is "regular".
     */
    virtual void setOption(TString mode, Int_t number);

    //! Set the range of the sampling for a changing attribute.
    /*!
     This function sets the value of the range inside which the values of 
     the changing attribute will be sampled. If this value is R, and the 
     nominal value is N, the new values will be sampled in the interval 
     [N-R/2; N+R/2]. The range can also be interpreted as a percentage p 
     of the nominal value, in which case R = p * N / 100.
     \note a function setRange must be called for _every_ changing 
     attribute. There is no default range value.
     
     \param changingAttName (TString): name of the changing attribute to sample.
     \param rangeValue     (Double_t): the range value, i.e. the width of the 
     interval around the nominal value where 
     the sampling will take place.
     \param is_percent       (Bool_t): if True, the rangeValue parameter will 
     be interpreted as a percentage of the 
     nominal value of the changing attribute.
     (Default = kFALSE).  
     \exception UErrorExceptions if no attribute changingAttName is found in the data server.
     \exception UErrorExceptions if the data server is empty and the attribute has no default value.
     \exception UErrorExceptions if range is negative or null.  
     \exception UErrorExceptions if mode is random and the attribute isn't a TStochasticAttribute.
     \exception UErrorExceptions if a range was already set for the attribute.
     */
    void setRange(TString changingAttName, Double_t rangeValue,
            Bool_t is_percent = kFALSE);

    //! Set the name of the attribute holding the range of the sampling for a changing attribute.
    /*!
     This function associate the name of two attributes of the data server, 
     the first being the changing attribute, the second holding the size 
     of the range of the sampling. In this case, the value of the range 
     will be read directly from the data server or the range attribute's 
     default value.
     \note a function setRange must be called for _every_ changing 
     attribute. There is no default range value.
     
     \param changingAttName (TString): name of the changing attribute to sample.
     \param rangeAttName    (TString): name of the attribute holding the range value.
     \param is_percent       (Bool_t): if True, the rangeValue parameter will 
     be interpreted as a percentage of the 
     nominal value of the changing attribute.
     (Default = kFALSE).
     \exception UErrorExceptions if no attribute rangeAttName is found in the data server.
     \exception UErrorExceptions if no attribute changingAttName is found in the data server.
     \exception UErrorExceptions if the data server is empty and the attribute has no default value.
     \exception UErrorExceptions if the data server is empty and the range attribute has no default value.  
     \exception UErrorExceptions if the data server is empty and the range attribute's default value is negative or null.
     \exception UErrorExceptions if mode is random and the attribute isn't a TStochasticAttribute.
     \exception UErrorExceptions if a range was already set for the attribute.
     */
    void setRange(TString changingAttName, TString rangeAttName,
            Bool_t is_percent = kFALSE);
    //@}

    //---------------------------------------------
    /** @name Generation of the sample
     **/
    //@{
    //! Generates the sample
    /*!
     This function creates the OAT sampling data.            
     \param option (Option_t *): this parameter is not used at the moment. (Default = "").
     */
    virtual void generateSample(Option_t *option = "");
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    //! Prints the log
    /*! Prints some internal informations about the object.
     \param option (Option_t *): this parameter is not used at the moment. (Default = "").
     */
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Sampler::TOATSamplingMultiNominal, ID_SAMPLER)
};
} // Fin du namespace Sampler
} // Fin du namespace URANIE

#endif
// fin du fichier $RCSfile$.
