//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TPetras.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   tue apr 8 15:09:37 CEST 2008

 \brief Implementation of the class URANIE::Sampler::TPetras

 */

////// URANIE
#include "TPetras.h"
#include "TStochAttribut.h"

void size_smolyak(int *, int *, int *);
void quad_smolyak(int *, int *, double *, double *);

ClassImp (URANIE::Sampler::TPetras)

namespace URANIE
{
namespace Sampler
{
TPetras::TPetras(URANIE::DataServer::TDataServer *tds, Option_t *option,
        Int_t ns) :
        TSamplerDeterminist(tds, "Petras", 1), _nQ(ns)
{
    TString soption = option;
    soption.ToLower();
    if (_blog)
        cout << " ** Build TPetras nX[" << _nX << "] Level[" << _nQ << "]"
                << endl;

    try
      {
	if( _nQ <=0 )
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__,__LINE__,
						     Form("TPetras::Construction. The level must be strictly greater than 0 (%d provided)",_nQ));
      }catch(URANIE::Exceptions::UErrorExceptions& ue)
      {
	ue.printMessage();
	throw ue;
      }    
}
TPetras::~TPetras()
{
    if (_blog)
        cout << " ** Destroy TPetras" << endl;
}

/*******************************************/
/*           Generate the sample           */
/*******************************************/
void TPetras::generateSample(Option_t *option)
{
    TString soption = option;
    soption.ToLower();
    if (_blog)
        cout << "TPetras::GenerateSample" << endl;
    cout << "TPetras::GenerateSample nX[" << _nX << "] niveau[" << _nQ << "]"<<endl;

    Int_t nvar = _nX;

    //Double_t *xmin= new Double_t[_nX];
    //Double_t *xrange=new Double_t[_nX];

    //// Loop on stochastic attributes to have PMin and PMax values
    URANIE::DataServer::TStochasticAttribute *stoatt;
    TIter next(_lstOfAttributesToSample);
    Int_t j = 0;
    //while ( (stoatt = (URANIE::DataServer::TStochasticAttribute *) next()) ) {
    //    xmin[j] = stoatt->getPMin();
    //    xrange[j] = stoatt->getPMax() - xmin[j];
    //    if ( _blog ) 
    //        cout << "var[" << j << "] name[" << stoatt->GetName() << "] min[" << xmin[j] << "] range["<< xrange[j] <<"]" << endl;
    //    j++;
    //}

    // Add an attribute for the weigts
    URANIE::DataServer::TAttribute *pw = new URANIE::DataServer::TAttribute(
            "uWeight");
    pw->delShare();
    
    _lstOfAttributesToSample->Add(pw);
    _nX = _lstOfAttributesToSample->GetEntries();
    _tds->addAttribute(pw);

    // Create the TTree of data
    createTuple();

    Int_t nquad;
    _nQ += nvar;
    size_smolyak(&nvar, &_nQ, &nquad);

    if (_blog)
        cout << " *** nquad[" << nquad << "]" << endl;

    Double_t * valpattern = new Double_t[_nX + 2];
    Double_t *ws = new Double_t[nquad];
    Double_t *ps = new Double_t[nquad * nvar];

    quad_smolyak(&nvar, &_nQ, ps, ws);

    // Loop on the patterns
    Double_t dproba;
    Int_t icrt = 0;
    for (Int_t i = 0; i < nquad; i++)
    {
        if (_blog)
            cout << i << "/" << nquad << " Weight[" << ws[i] << "]" << endl;
        valpattern[0] = (Double_t)(i + 1);

        for (j = 0; j < nvar; j++)
        {
            dproba = ps[icrt++];
            stoatt =
                    (URANIE::DataServer::TStochasticAttribute *) (_lstOfAttributesToSample->At(
                            j));
            //valpattern[j+1] = stoatt->getInvCDF(xmin[j] + xrange[j] * dproba);
            valpattern[j + 1] = stoatt->getInvCDF(dproba);
            if (_blog)
                cout << " (" << stoatt->GetName() << ")[" << dproba << "]("
                        << valpattern[j + 1] << ") ---- ";
        }
        valpattern[nvar + 1] = ws[i];
        _ntsample->Fill(valpattern);
        if (_blog)
            cout << endl << endl;
    }

    // Clean
    delete[] ws;
    ws = NULL;
    delete[] ps;
    ps = NULL;

    // Affect the current TTree to the TDataServer
    _tds->setTuple(_ntsample);
    // Fill the formula attributes
    fillOtherAttributes();

    // Clean
    _ntsample->ResetBranchAddresses();
    delete[] valpattern;
    valpattern = NULL;
}

void TPetras::printLog(Option_t *option)
{
    cout << "TPetras::printLog[" << option << "]" << endl;
}

} // Fin du namespace Sampler
} // Fin du namespace URANIE
// fin du fichier $RCSfile$.
