//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TQMC.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Mon Jul 24 14:15:38 2006

 \brief  Interface for sampling by QMC


 */
////// ROOT
#include "TRandom3.h"

////// URANIE
#include "TQMC.h"
#include "TStochAttribut.h"
#include "THaltonSequence.h"
#include "TSobolSequence.h"
#include "UExceptions.h"

ClassImp (URANIE::Sampler::TQMC)

namespace URANIE
{
namespace Sampler
{
//////////////////////////////////////////////////////////////
//  Constructor : from TDataServer + Methode + size
////////////////////////////////////////////////////////////
TQMC::TQMC(URANIE::DataServer::TDataServer *tds, Option_t *option, Int_t ns) :
        TSamplerDeterminist(tds, "qMC", ns), _sSequence(option)
{

    _sSequence.ToLower();
    if (_sSequence.CompareTo("halton", TString::kExact) == 0)
    {
        // Cas d'une séquence de Halton
        _Sequence = new THaltonSequence(_nX);
    }
    else if (_sSequence.CompareTo("sobol", TString::kExact) == 0)
    {
        // Cas d'une séquence de Sobol
        _Sequence = new TSobolSequence(_nX);
    }
    else
    {
        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                Form(
                        "TQMC::TQMC constructor ERROR : the sequence \"%s\" is unknown",
                        _sSequence.Data()));
    }

    setMethodName(Form("qMC_%s", _sSequence.Data()));
    SetNameTitle(Form("Sampling_qMC_%s_%d", _sSequence.Data(), ns),
            Form("Uranie %d sample with qMC method %s", ns, _sSequence.Data()));
}

TQMC::~TQMC()
{
    if (_Sequence != NULL)
    {
        delete _Sequence;
        _Sequence = NULL;
    }
    if (_blog)
        cout << " ** Destroy TQMC" << endl;
}

/*******************************************/
/*           Generate the sample           */
/*******************************************/
void TQMC::generateSample(Option_t *option)
{
    // Manage the options
    TString soption = option;
    soption.ToLower();
    Bool_t bshift = kFALSE;

    if (soption.Contains("shift"))
        bshift = kTRUE;

    _Sequence->init();

    if (_blog)
        cout << " ** TQMC::GenerateSample soption[" << soption << "]" << endl;

    // Create the TTree of data
    createTuple();

    if (_nX == 0)
        return;
    if (_blog)
        cout << "nvar[" << _nX << "]" << endl;

    Double_t *valpattern = new Double_t[_nX + 1];
    //Double_t *xmin= new Double_t[_nX];
    //Double_t *xrange= new Double_t[_nX];
    Double_t *xU = new Double_t[_nX];

    if (bshift)
    {
        TRandom3 * rand = new TRandom3(0);
        for (Int_t i = 0; i < _nX; i++)
            xU[i] = rand->Rndm();
        delete rand;
        rand = NULL;
    }
    else
    {
        for (Int_t i = 0; i < _nX; i++)
            xU[i] = 0.0;
    }

    if (_blog)
        for (Int_t i = 0; i < _nX; i++)
            cout << " ** (" << i << ") xu[" << xU[i] << "]" << endl;

    //// Loop on stochastic attributes to check that no infinite distri are unbounded with Halton.
    URANIE::DataServer::TStochasticAttribute *stoatt=NULL;
    TIter next(_lstOfAttributesToSample);
    Int_t j = 0;
    if(_sSequence.CompareTo("halton", TString::kExact) == 0)
    {
        TString AttFound="";
        URANIE::DataServer::TInfiniteDistribution* test=NULL;
        while ( (stoatt = (URANIE::DataServer::TStochasticAttribute *) next()) )
        {
            test=dynamic_cast<URANIE::DataServer::TInfiniteDistribution *>(stoatt);
            if ( test != 0 && !test->isLowerBounded() ) 
                AttFound+=test->GetName()+TString(":");
        }
        try{
            if( AttFound != "")
            {
                AttFound=AttFound.Strip(TString::EStripType::kTrailing,':');
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("Halton sequence is not meant to be runned with unbounded infinite stochastic laws [%s].\n\
The first location of Halton sequence being 0., it might indeed be meaningless (depending on the threshold chosen from one law to the other).\n\
This method was deprecated from v4.3.0 and is now set as an error. Please, use a Sobol sequence define a lower bound.",AttFound.Data()));
            }
        } catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            // Clean
            delete[] valpattern;
            valpattern = NULL;
            delete[] xU;
            xU = NULL;
            throw ue;
        }
    }

    // Loop on the patterns
    Double_t dproba, dxcrt;
    for (Int_t i = 0; i < _nS; i++)
    {
        _Sequence->setInstance((ULong_t)(i));
        valpattern[0] = (Double_t)(i + 1);
        j = 0;
        next.Reset();
        // Loop on stochastic attributes storing in the TList
        while ((stoatt = (URANIE::DataServer::TStochasticAttribute *) next()))
        {
            dproba = _Sequence->getComponent(j);
            //dxcrt = xmin[j] + xrange[j] * dproba;
            dxcrt = dproba;
            if (_blog)
                cout << " ** (" << i << ")(" << j << ") xu(" << xU[j]
                        << ") ** dproba[" << dproba << "] dxcrt[" << dxcrt
                        << "]";
            dxcrt += xU[j];
            if (_blog)
                cout << " --------- dxcrt[" << dxcrt << "]";
            if (dxcrt > 1.0)
                dxcrt -= 1.0;
            if (_blog)
                cout << " ********* dxcrt[" << dxcrt << "]" << endl;
            valpattern[j + 1] = stoatt->getInvCDF(dxcrt);
            if (_blog)
                cout << "** i[" << i << "] -- proba [" << dproba
                        << "] -- value[" << valpattern[j + 1] << "]" << endl;
            j++;
        }
        _ntsample->Fill(valpattern);
    }

    // Affect the current TTree to the TDataServer
    if(_tds->hasTuple())
      delete _tds->getTuple();
    _tds->setTuple(_ntsample);
    // Fill the formula attributes
    fillOtherAttributes();

    // Clean
    delete[] valpattern;
    valpattern = NULL;
    delete[] xU;
    xU = NULL;
}

////////////////////////////////////////////////////////////
//  Printing Log
////////////////////////////////////////////////////////////
void TQMC::printLog(Option_t *option)
{
    cout << "TQMC::printLog[" << option << "] Name[" << GetName() << "] Title["
            << GetTitle() << "]" << endl;
    cout << "_sSequence[" << _sSequence << "]" << endl;
    if (_Sequence != NULL)
        _Sequence->printLog();
    TSampler::printLog();
    cout << "End of TQMC::printLog" << endl;
}

} // Fin du namespace Sampler
} // Fin du namespace URANIE
// fin du fichier $RCSfile$.
