//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TQMC.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Mon Jul 24 14:15:38 2006
 
 \brief  Interface for the sampling by QMC
 
 
 */
#ifndef TQMC_H
#define TQMC_H
/*!
 
 @class URANIE::Sampler::TQMC

 @brief Description of a variable (field).

 Both currently sequences implemented are :
 \li "sobol"
 \li "halton"

 \section qmc_exemple Example of use

 Firstly, a pointer to a TDataServer is constructed 
 \code
 TDataServer *tdsqmc = new TDataServer("tdsQMC", "Test for qMC");
 tdsqmc->addAttribute( new TAttribute("x_{1}", -1.0 , 2.) );
 tdsqmc->addAttribute( new TAttribute("x_{2}",  0.0 , 3.) );
 tdsqmc->addAttribute( new TAttribute("x_{3}",  1.0, 1.5) );
 \endcode

 Then, a pointer to an object TQMC of the chosen sequence ("sobol"|"halton") is constructed from this TDataServer , and the  \b generateSample() method is applied

 \code
 TQMC * qmc = new TQMC(tdsqmc, "sobol", 300);
 qmc->generateSample();
 \endcode



 \section qmc_source Bibliography source
 The  bibliography is :
 \li Eric Thiemard's thesis intitulated Sur le Calcul et la Majoration de la disrepance a l'Origine - Lausanne, EPFL, 2000 (in french)
 \li Lectures Notes in statistics intitulated Random and Quasi-Random Points Sets \par
 He proposes a theorical presentation  of the <b>Digital Point Sets</b> of Gerhard Larcher.Anyway, this book is mostly oriented to the tests for checking that the generated points are really random ones.


 */
#include <iostream>
using std::cout;
using std::endl;

#include "TSamplerDeterminist.h"
#include "TSequence.h"

namespace URANIE
{
namespace Sampler
{
class TQMC: public TSamplerDeterminist
{

    // Attributes
private:
    TSequence * _Sequence;
    TString _sSequence;

    // Operations
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //!  Constructor from a TDataServer, the name of the method and the size of the sample
    TQMC(URANIE::DataServer::TDataServer *tds, Option_t *option, Int_t nCalcul);

    //! Default destructor
    virtual ~TQMC();
    //@}

    //---------------------------------------------
    /** @name Treats the TAttributes
     **/
    //@{
    //! Creates the TNtupleD of data with only the TStochasticAttributes
//        void createTuple();
    //@}
    //---------------------------------------------
    /** @name Generation of the sample
     **/
    //@{
    //! Generates the sample
    virtual void generateSample(Option_t *option = "");
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    //! Prints the log
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Sampler::TQMC, ID_SAMPLER)
    //Classe de g�n�ration par Quasi Monte-Carlo
};
} // Fin du namespace Sampler
} // Fin du namespace URANIE

#endif
// fin du fichier $RCSfile$.
