//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*! \file TRafuSampling.cxx
 *  \author Nicolas Gilardi <nicolas.gilardi@cea.fr>
 *  \date   $date$
 *  \brief Implementation of the class URANIE::Sampler::TRafuSampling.
 *
 *  Created the : Fri Oct  1 13:37:23 2004
 *
 */
#include <ctime>
#include <iostream>
#include <fstream>
using namespace std;

////// ROOT
#include "TROOT.h"
#include "TObjString.h"

////// URANIE
#include "TRafuSampling.h"
#include "TStochAttribut.h"
#include "TNormalDistribution.h"
#include "TLogNormalDistribution.h"
#include "TUniformDistribution.h"
#include "TTriangularDistribution.h"
#include "TTriangularPossibility.h"
#include "TTrapeziumPossibility.h"
#include "TTriangularPossibility.h"
#include "TPossibilityAttribute.h"

ClassImp (URANIE::Sampler::TRafuSampling)

namespace URANIE
{
namespace Sampler
{
////////////////////////////////////////////////////////////
//  Constructeur : from TDataServer + Methode + size
////////////////////////////////////////////////////////////
TRafuSampling::TRafuSampling(URANIE::DataServer::TDataServer *tds,
        Option_t *option, Int_t samplingSize) :
        TSampler(tds, option, samplingSize)
{
    if (_blog)
        cout << "constructor TRafuSampling using TDataServer[" << tds->GetName()
                << "]" << endl;

    TString opt = option;
    opt.ToLower();
    opt.ReplaceAll(" ", "");

    TString sProbaSampling;
    TString sPossibSampling;

    // Set the sampling method for random variables
    if (opt.Contains("proba=srs"))
    {
        _nProbaSampling = kSRS;
        sProbaSampling = "srs";
    }
    else if (opt.Contains("proba=lhs"))
    {
        _nProbaSampling = kLHS;
        sProbaSampling = "lhs";
    }
    else
    {
        _nProbaSampling = kLHS;
        sProbaSampling = "lhs";
    }

    // Set the sampling method for fuzy variables
    if (opt.Contains("possib=random_interval"))
    {
        _nPossibSampling = kRandom;
        sPossibSampling = "random";
    }
    else if (opt.Contains("possib=interval_sampling0"))
    {
        _nPossibSampling = kSampling0;
        sPossibSampling = "sampling0";
    }
    else if (opt.Contains("possib=interval_sampling1"))
    {
        _nPossibSampling = kSampling1;
        sPossibSampling = "sampling1";
    }
    else
    {
        _nPossibSampling = kRandom;
        sPossibSampling = "random";
    }

    if ((_nX == 0) && (_nPossibSampling != kRandom))
        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                Form(
                        "TRafuSampling::TRafuSampling: if no stochastic attribute is provided, the possibility sampling must be random (current value: [interval_%s])",
                        sPossibSampling.Data()));

    SetNameTitle(
            Form("%s_Sampling_%s_%s_%d", _tds->GetName(), sProbaSampling.Data(),
                    sPossibSampling.Data(), samplingSize),
            Form(
                    "Uranie %d sample with sampling proba method %s and possib method %s",
                    samplingSize, sProbaSampling.Data(),
                    sPossibSampling.Data()));
    createListOfPossibilityAttributes();
    _sunsetInputFileName = Form("%s.su", GetName());
    _sunsetInfFileName = Form("%s_inf.csv", GetName());
    _sunsetSupFileName = Form("%s_sup.csv", GetName());
}

TRafuSampling::~TRafuSampling()
{
    if (_blog)
        cout << "************ Destroying URANIE::Sampler::TRafuSampling"
                << endl;
}

/*******************************************/
/*           Generate the sample           */
/*******************************************/
void TRafuSampling::generateSample(Option_t *option)
{
    if (_blog)
        cout << endl
                << "**************************\n*** TRafuSampling::generateSample"
                << endl;

    TString soption = option;
    soption.ToLower();
    
    fillSunsetInputFile();

    // Launch Sunset 
    gSystem->Exec(
            Form("sunset.exe %s > /dev/null", _sunsetInputFileName.Data()));

    // Read the output files from Sunset and store them in a TTree
    createTuple();

    Double_t n_iter_increment = 1.0;
    Double_t values[_ntsample->GetNbranches()];
    ifstream fCSV;
    Double_t rafu_dataset_values[2] =
    { -1.0, 1.0 };
    const char * fileNames[2] =
    { _sunsetInfFileName.Data(), _sunsetSupFileName.Data() };

    for (Int_t idataset = 0; idataset < 2; idataset++)
    {
        if (_blog)
            cout << "  Opening file " << idataset << ": " << fileNames[idataset]
                    << endl;

        fCSV.open(fileNames[idataset]);
        // read the first two lines
        Int_t lineNum = 1;
        char c;
        while (lineNum < 3)
        {
            fCSV.get(c);
            if (c == '\n')
                lineNum++;
        }
        // retrieve the data, line by line
        TString sLine;
        while (1)
        {
            fCSV >> sLine;
            if (fCSV.eof())
                break;
            else
            {
                TObjArray *listOfValues = sLine.Tokenize(",");

                TObjString *sValue = (TObjString *) listOfValues->At(0);
                TString sTmp = sValue->GetString();
                // correction of the Sunset counter
                values[0] = sTmp.Atof() + n_iter_increment;

                for (Int_t ival = 1; ival < listOfValues->GetEntries(); ival++)
                {
                    sValue = (TObjString *) listOfValues->At(ival);
                    sTmp = sValue->GetString();
                    values[ival] = sTmp.Atof();
                }
                // add the __rafu__dataset__ variable
                values[_ntsample->GetNbranches() - 1] =
                        rafu_dataset_values[idataset];

                _ntsample->Fill(values);
            }
        }
        fCSV.close();
        n_iter_increment = values[0] + 1.0;

        if (_blog)
            cout << "File " << idataset << ": " << fileNames[idataset]
                    << " is closed." << endl;
    }

    if (_blog)
        cout << "  Storing the data in the Data Server" << endl;

    // Add the __rafu__dataset__ attribute to the TDataServer
    URANIE::DataServer::TAttribute *rafuAtt =
            new URANIE::DataServer::TAttribute(
                    Form("%s__rafu__dataset__", _tds->GetName()),
                    "RaFu dataset indicator (inf = -1, sup = +1)");
    //rafuAtt->setOrigin(URANIE::DataServer::TAttribute::kInternal);
    rafuAtt->delShare();
    _tds->addAttribute(rafuAtt);
    

    // Store the data in the TDataServer 
    _tds->setTuple(_ntsample);

    if (_blog)
        cout
                << "*** End Of TRafuSampling::generateSample\n**************************"
                << endl;

}

void TRafuSampling::createTuple()
{
    if (_blog)
        cout << endl
                << "**************************\n*** TRafuSampling::createTuple"
                << endl;

    ifstream fInf;

    fInf.open(_sunsetInfFileName.Data());

    if (_blog)
        cout << " ** Retrieve variables names from file [" << _sunsetInfFileName
                << "]" << endl;

    // Read the first line
    char c = ' ';
    while (c != '\n')
        fInf.get(c);

    // Build the labels of the variables for the TDSNtupleD
    TString sLabelVariables;
    fInf >> sLabelVariables;
    fInf.close();

    sLabelVariables = Form("%s__n__iter__%s:%s__rafu__dataset__",
            _tds->GetName(), sLabelVariables.Data(), _tds->GetName());
    sLabelVariables.ReplaceAll(",", ":");

    // Build the TDSNtupleD

    if (_blog)
        cout << " ** Create the TTree with variables [" << sLabelVariables
                << "]" << endl;

    // WARNING: In order to respect some Sunset constraints, the order 
    //          in which the attributes appear in the TDataServer 
    //          can be different from the order in Sunset's 
    //          output files, and thus in the TTree. This should be 
    //          totally transparent, but it's good to know. 
    _ntsample = new URANIE::DataServer::TDSNtupleD(
            Form("RafuSampling_%s", GetName()), GetTitle(), sLabelVariables);

    if (_blog)
        cout
                << "*** End Of TRafuSampling::createTuple\n**************************"
                << endl;

}

void TRafuSampling::fillSunsetInputFile()
{
    Bool_t bConcat = kFALSE;

    TString sProbaModuleName = "Sample_proba";
    TString sPossibModuleName = "Sample_possib";

    std::ofstream fSunsetInputFile;
    fSunsetInputFile.open(_sunsetInputFileName.Data(), std::ios::out);

    // Step 1: build the first lines

    fSunsetInputFile << "MODULE PEL_Application" << endl;
    fSunsetInputFile << "  MODULE Actions" << endl;

    // Step 2: add the modules of the probability data 
    if (_nX > 0)
    {
        fSunsetInputFile << Form("    MODULE %s", sProbaModuleName.Data())
                << endl;
        fSunsetInputFile << "      type=\"random_sampling\"" << endl;
        if (_nProbaSampling == kSRS)
            fSunsetInputFile << "      method=\"SRS\"" << endl;
        else if (_nProbaSampling == kLHS)
            fSunsetInputFile << "      method=\"LHS\"" << endl;
        else
            fSunsetInputFile << "      method=\"LHS\"" << endl;

        fSunsetInputFile << Form("      size= %d", _nS) << endl;
        fSunsetInputFile << "      MODULE Variables" << endl;

        for (Int_t i = 0; i < _nX; i++)
        {
            URANIE::DataServer::TStochasticAttribute *att =
                    (URANIE::DataServer::TStochasticAttribute *) _lstOfAttributesToSample->At(
                            i);

            fSunsetInputFile << Form("        MODULE %s", att->GetName())
                    << endl;
            fSunsetInputFile << buildProbabilityModule(att, "          ")
                    << endl;
            fSunsetInputFile << Form("        END MODULE %s", att->GetName())
                    << endl;
        }

        fSunsetInputFile << "      END MODULE Variables" << endl;
        fSunsetInputFile << Form("    END MODULE %s", sProbaModuleName.Data())
                << endl;
    }

    // Step 3: add the modules of the possibility variables                
    fSunsetInputFile << Form("    MODULE %s", sPossibModuleName.Data()) << endl;
    if (_nPossibSampling == kRandom)
        fSunsetInputFile << "      type= \"random_interval\"" << endl;
    else if (_nPossibSampling == kSampling0)
    {
        fSunsetInputFile << "      type= \"interval_sampling\"" << endl;
        fSunsetInputFile << "      alpha_cut= < 0.0 >" << endl;
    }
    else if (_nPossibSampling == kSampling1)
    {
        fSunsetInputFile << "      type= \"interval_sampling\"" << endl;
        fSunsetInputFile << "      alpha_cut= < 1.0 >" << endl;
    }
    else
        fSunsetInputFile << "      type= \"random_interval\"" << endl;

    fSunsetInputFile << Form("      size= %d", _nS) << endl;
    fSunsetInputFile << "      MODULE Variables" << endl;

    for (Int_t i = 0; i < _nPossib; i++)
    {
        URANIE::DataServer::TPossibilityAttribute *att =
                (URANIE::DataServer::TPossibilityAttribute *) _lstOfPossibilityAttributesToSample->At(
                        i);

        fSunsetInputFile << Form("        MODULE %s", att->GetName()) << endl;
        fSunsetInputFile << buildPossibilityModule(att, "          ") << endl;
        fSunsetInputFile << Form("        END MODULE %s", att->GetName())
                << endl;
    }

    fSunsetInputFile << "      END MODULE Variables" << endl;
    fSunsetInputFile << Form("    END MODULE %s", sPossibModuleName.Data())
            << endl;

    // Step 4: if necessary, ask to concatenate the data
    if (_nX > 0 && _nPossibSampling == kRandom)
    {
        bConcat = kTRUE;
        fSunsetInputFile << "    MODULE Concat_inf" << endl;
        fSunsetInputFile << "      type = \"concatenation\"" << endl;
        fSunsetInputFile
                << Form("      data = < \"%s/Sample\" \"%s/Sample_inf\" >",
                        sProbaModuleName.Data(), sPossibModuleName.Data())
                << endl;
        fSunsetInputFile << "    END MODULE Concat_inf" << endl;
        fSunsetInputFile << "    MODULE Concat_sup" << endl;
        fSunsetInputFile << "      type = \"concatenation\"" << endl;
        fSunsetInputFile
                << Form("      data = < \"%s/Sample\" \"%s/Sample_sup\" >",
                        sProbaModuleName.Data(), sPossibModuleName.Data())
                << endl;
        fSunsetInputFile << "    END MODULE Concat_sup" << endl;
    }
    else if (_nPossibSampling == kSampling0 || _nPossibSampling == kSampling1)
    {
        bConcat = kTRUE;
        fSunsetInputFile << "    MODULE Concat_inf" << endl;
        fSunsetInputFile << "      type = \"product\"" << endl;
        fSunsetInputFile
                << Form("      data = < \"%s/Sample\" \"%s/Sample_inf\" >",
                        sProbaModuleName.Data(), sPossibModuleName.Data())
                << endl;
        fSunsetInputFile << "    END MODULE Concat_inf" << endl;
        fSunsetInputFile << "    MODULE Concat_sup" << endl;
        fSunsetInputFile << "      type = \"product\"" << endl;
        fSunsetInputFile
                << Form("      data = < \"%s/Sample\" \"%s/Sample_sup\" >",
                        sProbaModuleName.Data(), sPossibModuleName.Data())
                << endl;
        fSunsetInputFile << "    END MODULE Concat_sup" << endl;
    }

    // Step 5: ask for the construction of the csv files            
    fSunsetInputFile << "    MODULE csv_inf" << endl;
    fSunsetInputFile << "      type = \"csv_saver\"" << endl;
    if (bConcat)
        fSunsetInputFile << "      data = \"Concat_inf/Output\"" << endl;
    else
        fSunsetInputFile
                << Form("      data = \"%s/Sample_inf\"",
                        sPossibModuleName.Data()) << endl;
    fSunsetInputFile << Form("      filename = \"%s_inf\"", GetName()) << endl;
    fSunsetInputFile << "    END MODULE csv_inf" << endl;
    fSunsetInputFile << "    MODULE csv_sup" << endl;
    fSunsetInputFile << "      type = \"csv_saver\"" << endl;
    if (bConcat)
        fSunsetInputFile << "      data = \"Concat_sup/Output\"" << endl;
    else
        fSunsetInputFile
                << Form("      data = \"%s/Sample_sup\"",
                        sPossibModuleName.Data()) << endl;
    fSunsetInputFile << Form("      filename = \"%s_sup\"", GetName()) << endl;
    fSunsetInputFile << "    END MODULE csv_sup" << endl;

    // Step 6: finish the input file
    fSunsetInputFile << "  END MODULE Actions" << endl;
    fSunsetInputFile << "  concrete_name = \"SUNSET\"" << endl;
    fSunsetInputFile << "END MODULE PEL_Application" << endl;

    fSunsetInputFile.close();

}

void TRafuSampling::createListOfPossibilityAttributes()
{
    if (_blog)
        cout << " ** TRafuSampling::createListOfPossibilityAttributes" << endl;

    _lstOfPossibilityAttributesToSample = new TList();

    // Loop on the attributes of the TDataServer to select only TPossibilityAttribute objects
    URANIE::DataServer::TAttribute *att = 0;
    URANIE::DataServer::TPossibilityAttribute *posatt;

    TList *attList = _tds->getListOfAttributes();
    for (Int_t iatt = 0; iatt < attList->GetEntries(); iatt++)
    {

        if (_blog)
            cout << "iatt [" << iatt << "], n entries ["
                    << attList->GetEntries() << "]" << endl;

        att = (URANIE::DataServer::TAttribute*) attList->At(iatt);
        if (_blog)
            cout << " ** att [" << att->GetName() << "]";
        // Forcast le pointeur pour savoir s'il provient d'un URANIE::DataServer::TPossibilityAttribute
        posatt = dynamic_cast<URANIE::DataServer::TPossibilityAttribute *>(att);
        if (posatt != NULL)
        {
            if (_blog)
                cout << " ********** Add";
            _lstOfPossibilityAttributesToSample->Add(posatt);
        }
        else
        {
            if (_blog)
                cout << " ********** not added";

        }
        if (_blog)
            cout << endl;
    }
    _nPossib = _lstOfPossibilityAttributesToSample->GetEntries();

    if (_blog)
    {
        cout << " ** Entries : " << _nPossib << endl;
        _lstOfPossibilityAttributesToSample->Print();
    }
    if (_blog)
        cout << " ** End of TRafuSampling::createListOfPossibilityAttributes"
                << endl;

}

TString TRafuSampling::buildProbabilityModule(
        URANIE::DataServer::TStochasticAttribute *att, const char *indentation)
{
    TString sOutputString = "";

    TString sClassName = att->IsA()->GetName();

    if (sClassName.Contains("TNormalDistribution"))
    {
        URANIE::DataServer::TNormalDistribution *normAtt =
                (URANIE::DataServer::TNormalDistribution *) att;

        if ((normAtt->getPMin() > 0.0) || (normAtt->getPMax() < 1.0))
        {
            sOutputString += Form("%stype=\"truncated\"\n", indentation);
            sOutputString += Form("%sname=\"%s\"\n", indentation,
                    normAtt->GetName());
            sOutputString += Form("%sinf_limit=%e\n", indentation,
                    normAtt->getLowerBound());
            sOutputString += Form("%ssup_limit=%e\n", indentation,
                    normAtt->getUpperBound());
            sOutputString += Form("%sMODULE ProbaLaw\n", indentation);
            sOutputString += Form("%s  type=\"normal\"\n", indentation);
            sOutputString += Form("%s  mu=%e\n", indentation,
                    normAtt->getParameterMu());
            sOutputString += Form("%s  sigma=%e\n", indentation,
                    normAtt->getParameterSigma());
            sOutputString += Form("%sEND MODULE ProbaLaw\n", indentation);
        }
        else
        {
            sOutputString += Form("type=\"normal\"\n");
            sOutputString += Form("%sname=\"%s\"\n", indentation,
                    normAtt->GetName());
            sOutputString += Form("%smu=%e\n", indentation,
                    normAtt->getParameterMu());
            sOutputString += Form("%ssigma=%e\n", indentation,
                    normAtt->getParameterSigma());
        }
    }

    else if (sClassName.Contains("TTriangularDistribution"))
    {
        URANIE::DataServer::TTriangularDistribution *triAtt =
                (URANIE::DataServer::TTriangularDistribution *) att;
        sOutputString += Form("%stype=\"triangular\"\n", indentation);
        sOutputString += Form("%sname=\"%s\"\n", indentation,
                triAtt->GetName());
        sOutputString += Form("%sinf=%e\n", indentation,
                triAtt->getLowerBound());
        sOutputString += Form("%smode=%e\n", indentation,
                triAtt->getParameterMode());
        sOutputString += Form("%ssup=%e\n", indentation,
                triAtt->getUpperBound());
    }

    else if (sClassName.Contains("TUniformDistribution"))
    {
        URANIE::DataServer::TUniformDistribution *unifAtt =
                (URANIE::DataServer::TUniformDistribution *) att;
        sOutputString += Form("%stype=\"uniform\"\n", indentation);
        sOutputString += Form("%sname=\"%s\"\n", indentation,
                unifAtt->GetName());
        sOutputString += Form("%sinf=%e\n", indentation,
                unifAtt->getLowerBound());
        sOutputString += Form("%ssup=%e\n", indentation,
                unifAtt->getUpperBound());
    }
//            else if( sClassName.Contains("TUniformByPartsDistribution") )
//            {
//            }
    else if (sClassName.Contains("TLogNormalDistribution"))
    {
        URANIE::DataServer::TLogNormalDistribution *logNormAtt =
                (URANIE::DataServer::TLogNormalDistribution *) att;

        if ((logNormAtt->getPMin() > 0.0) || (logNormAtt->getPMax() < 1.0))
        {
            sOutputString += Form("%stype=\"truncated\"\n", indentation);
            sOutputString += Form("%sname=\"%s\"\n", indentation,
                    logNormAtt->GetName());
            sOutputString += Form("%sinf_limit=%e\n", indentation,
                    logNormAtt->getLowerBound());
            sOutputString += Form("%ssup_limit=%e\n", indentation,
                    logNormAtt->getUpperBound());
            sOutputString += Form("%sMODULE ProbaLaw\n", indentation);
            sOutputString += Form("%s  type=\"log_normal\"\n", indentation);
            sOutputString += Form("%s  normal_mu=%e\n", indentation,
                    logNormAtt->getUnderlyingNormalParameterMu());
            sOutputString += Form("%s  normal_sigma=%e\n", indentation,
                    logNormAtt->getUnderlyingNormalParameterSigma());
            sOutputString += Form("%sEND MODULE ProbaLaw\n", indentation);
        }
        else
        {
            sOutputString += Form("%stype=\"log_normal\"\n", indentation);
            sOutputString += Form("%sname=\"%s\"\n", indentation,
                    logNormAtt->GetName());
            sOutputString += Form("%snormal_mu=%e\n", indentation,
                    logNormAtt->getUnderlyingNormalParameterMu());
            sOutputString += Form("%snormal_sigma=%e\n", indentation,
                    logNormAtt->getUnderlyingNormalParameterSigma());
        }
    }
    else
    {
        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                Form(
                        "TRafuSampling::buildProbabilityModule: the type [%s] of attribute [%s] is not supported by TRafuSampling",
                        sClassName.Data(), att->GetName()));
    }

    return sOutputString;
}

TString TRafuSampling::buildPossibilityModule(
        URANIE::DataServer::TPossibilityAttribute *att, const char *indentation)
{
    TString sOutputString = "";

    TString sClassName = att->IsA()->GetName();
    if (sClassName.Contains("TTrapeziumPossibility"))
    {
        URANIE::DataServer::TTrapeziumPossibility *trapAtt =
                (URANIE::DataServer::TTrapeziumPossibility *) att;

        sOutputString += Form("%sname=\"%s\"\n", indentation,
                trapAtt->GetName());

        if (trapAtt->isIncreasing())
            sOutputString += Form("%smonotonous_relationship=\"increasing\"\n",
                    indentation);
        else
            sOutputString += Form("%smonotonous_relationship=\"decreasing\"\n",
                    indentation);

        sOutputString += Form("%sMODULE PossibLaw\n", indentation);
        sOutputString += Form("%s   type=\"trapezium\"\n", indentation);
        sOutputString += Form("%s   inf=%e\n", indentation,
                trapAtt->getLowerBound());
        sOutputString += Form("%s   mode_inf=%e\n", indentation,
                trapAtt->getLowerMode());
        sOutputString += Form("%s   mode_sup=%e\n", indentation,
                trapAtt->getUpperMode());
        sOutputString += Form("%s   sup=%e\n", indentation,
                trapAtt->getUpperBound());
        sOutputString += Form("%sEND MODULE PossibLaw\n", indentation);

    }
    else if (sClassName.Contains("TTriangularPossibility"))
    {
        URANIE::DataServer::TTriangularPossibility *triAtt =
                (URANIE::DataServer::TTriangularPossibility *) att;

        sOutputString += Form("%sname=\"%s\"\n", indentation,
                triAtt->GetName());

        if (triAtt->isIncreasing())
            sOutputString += Form("%smonotonous_relationship=\"increasing\"\n",
                    indentation);
        else
            sOutputString += Form("%smonotonous_relationship=\"decreasing\"\n",
                    indentation);

        sOutputString += Form("%sMODULE PossibLaw\n", indentation);
        sOutputString += Form("%s   type=\"triangular\"\n", indentation);
        sOutputString += Form("%s   inf=%e\n", indentation,
                triAtt->getLowerBound());
        sOutputString += Form("%s   mode=%e\n", indentation, triAtt->getMode());
        sOutputString += Form("%s   sup=%e\n", indentation,
                triAtt->getUpperBound());
        sOutputString += Form("%sEND MODULE PossibLaw\n", indentation);
    }
//            else if ( sClassName.Contains("TPolygonalPossibility") )
//            {
//               URANIE::DataServer::TPolygonalPossibility *polyAtt = (URANIE::DataServer::TPolygonalPossibility *)att;
//               
//               sOutputString += Form("%sname=\"%s\"\n", indentation, polyAtt->GetName());
//               
//               if (polyAtt->isIncreasing())   
//                   sOutputString += Form("%smonotonous_relationship=\"increasing\"\n", indentation);
//               else
//                   sOutputString += Form("%smonotonous_relationship=\"decreasing\"\n", indentation);
//               
//               sOutputString += Form("%sMODULE PossibLaw\n"     , indentation);
//               sOutputString += Form("%s   type=\"polygonal\"\n", indentation);
//               
//               TString sValues    = "values=<";
//               TString sDensities = "densities=<"; 
//               Double_t *values    = polyAtt->getParameterValues();
//               Double_t *densities = polyAtt->getParameterDensities();
//               for (Int_t ival = 0; ival < polyAtt->getNValues()-1; ival++)
//               {
//                   sValues    += Form("%e ", values[ival]);
//                   sDensities += Form("%e ", densities[ival]);
//               }
//               sValues    += Form("%e>", values[polyAtt->getNValues()-1])
//               sDensities += Form("%e>", densities[polyAtt->getNValues()-1])
//               
//               sOutputString += Form("%s   %s\n", indentation, sValues.Data()   );
//               sOutputString += Form("%s   %s\n", indentation, sDensities.Data());
//               sOutputString += Form("%sEND MODULE PossibLaw\n", indentation);
//            }
    else
    {
        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                Form(
                        "TRafuSampling::buildPossibilityModule: the type [%s] of attribute [%s] is not supported by TRafuSampling",
                        sClassName.Data(), att->GetName()));
    }

    return sOutputString;
}

//////////////////////////////////////////////////////////
//  Printing Log
//////////////////////////////////////////////////////////
void TRafuSampling::printLog(Option_t *option)
{
    cout << "** TRafuSampling::printLog  [" << option << "]" << endl;
    cout << "** _nS   [" << _nS << "]" << endl;
    cout << "** _nX   [" << _nX << "]" << endl;
    cout << "** _nPossib   [" << _nX << "]" << endl;
    cout << "** _sMethod   [" << _sMethod << "]" << endl;
}

} // Fin du namespace Sampler
} // Fin du namespace URANIE
// fin du fichier $RCSfile$.
