//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TSamplingMixture.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2012-10-02 11:25:53
 
 \brief  Interface for the sampling for mixture sampling
 */
#ifndef TSAMPLINGMIXTURE_H
#define TSAMPLINGMIXTURE_H
/*!
 @class URANIE::Sampler::TSamplingMixture
 
 @brief Interface with the \em simplex class developped by B. Secher (LGLS).
 
 The type of points are :Summit|CenterEdge|ThirdEdge|CenterFace|CenterSegment|Centroide

 \dot
 digraph tsampling_mixture {
 node [shape=record, fontname=Helvetica, fontsize=10];
 a [ label="Constructor" URL="\ref TSamplingMixture"];
 b [ label="init C" URL="\ref init"];
 c [ label="generateCandidates C" URL="\ref generateCandidates"];
 d [ label="generateSample C" URL="\ref generateSample"];
 a -> b [ arrowhead="open", style="dashed" ];
 b -> c [ arrowhead="open", style="dashed" ];
 b -> d [ arrowhead="open", style="dashed" ];
 c -> d [ arrowhead="open", style="dashed" ];
 }
 \enddot


 The \b _bCentroide attribute have two meanning in the two following method:
 \li in the \em generateCandidates method, add the Centroide point in the list of candidates points (normal)
 \li but, in the \em generateSample method, add the Centroide point in the sample even if the two algorithms (\it MM or \it MMD)  \bf can \bf not select this point. Then, in this case we have $nS+1$ points in the sample.

 
 */

#include <iostream>
using std::cout;
using std::endl;

////// ROOT
#include "TVectorD.h"

////// URANIE
#include "simplex.h"

#include "TSamplerDeterminist.h"

namespace URANIE
{
namespace Sampler
{
class TSamplingMixture: public TSamplerDeterminist
{

    // Attributes
public:
    enum EType
    {
        kSummit = 10,
        kCenterEdge,
        kThirdEdge,
        kCenterFace,
        kCenterSegment,
        kCentroide
    };
private:
    enum ERelation
    {
        kLower, kUpper
    };

    Double_t _dThreshold;               ///< constraint: \sum{x_{i} = dThreshold
    vector<vector<double> > _planes;              ///< The matrix of contraintes
    Simplex *_simplex;                        ///< Simplex Classe from B. Secher

    Bool_t _bSummits;                                  ///< Flag for the summits
    Bool_t _bCenterEdges;                          ///< Flag for the CenterEdges
    Bool_t _bThirdEdges;                            ///< Flag for the ThirdEdges
    Bool_t _bCenterFaces;                          ///< Flag for the CenterFaces
    Bool_t _bCentroide;                              ///< Flag for the Centroide
    Bool_t _bCenterSegments;                    ///< Flag for the CenterSegments

    Bool_t _bSaveCandidates;              ///< Flag for saving candidates points

    TString _sAttNameTypeOfPoints; ///< Name of the internal attribute for the type of points Etype (kSummit, kCenterEdge, ...)
    TString _sAttNameCandidate; ///< Name of the internal attribute for candidates points(1) else 0

    Int_t _nGenerateSample; ///< The index of times that the generateSample method is call
    Bool_t _bTrace;                    ///< The flag to known is trace is needed
    TString _sTraceFile;                         ///< The name of the trace file

    // Operations
    //! Fill the list of points in the TDS with a type
    void fillListOfCandidatesPoints(list<vector<double> > lst, EType typePoint);
    //! Set The List of Candidates Points
    void setListOfCandidatesPoints();
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //!  Constructor from a TDataServer, the name of the method and the size of the sample
    /*!
     \param tds (TDataServer *) the TDS which contains attributes
     \param ns  (Int_t) The size of the sampling to generate
     \param dThreshold (Double_t) the Threshold for the sum (\f$ \sum{x_i} = dThreshold\f$)
     \param option (Option_t *) The option to pass.
     \li if option constains "save", the candidates points are stored in the TDS.
     */
    TSamplingMixture(URANIE::DataServer::TDataServer *tds, Int_t ns,
            Double_t dThreshold = 1.0, Option_t *option = "");

    //! Default destructor
    virtual ~TSamplingMixture();
    //! The \em init method
    /*!
     This method is called at the end of the constructor.
     
     This method adds the two internal attributes:
     \li \b _sAttNameTypeOfPoints : to save the type of points from \b [Summit|CenterEdge|ThirdEdge|CenterFace|CenterSegment|Centroide];
     \li \b _sAttNameCandidate : (optionnal when \em save request) to known if point is a candidate point (equal to 1) or a solution (equal to 0)

     And, this method creates the constraints from the lower and upper bounds for each attributes. 
     */
    void init();
    //@}

    //---------------------------------------------
    /** @name Setting and Getting global informations
     **/
    //@{
    //! Return the name of the internal attribute to known if point is a candidates point ( equal to 0) or not ( greater or equal to 1)
    /*!
     A filter to known if a point is a candidate point or not.
     \code
     tds->getTuple()->Scan("*", Form("%s>0", tmix->getNameInternalAttributeCandidatesPoints()));
     \endcode
     
     \par
     If the candidates points are stored, then the design of experiment (\em the sampling) are points with this attribute is equal to 0.
     \code
     tds->getTuple()->Scan("*", Form("%s == %d", tmix->getNameInternalAttributeCandidatesPoints(), tmix->getNGeneratedSample()));
     \endcode
     */
    const char* getNameInternalAttributeCandidatesPoints()
    {
        return _sAttNameCandidate.Data();
    }

    //! Return the name of the internal attribute to known the type of points
    /*!
     The type of points are EType : kSummit=10, kCenterEdge, kThirdEdge, kCenterFace, kCenterSegment, kCentroide
     */
    const char* getNameInternalAttributeTypePoints()
    {
        return _sAttNameTypeOfPoints.Data();
    }

    //! Return the number of call of the generateSample method
    /*!
     /return the number of call of the generateSample method
     */
    Int_t getNGeneratedSample()
    {
        return _nGenerateSample;
    }
    //! Activate the trace of the process
    /*!
     the trace is used only in the generateCandidate method
     
     /param straceFile (TString) The name of the trace file (empty by defautt, then use the default name which is \b "_tsmix_trace.log" )
     */
    void setTrace(TString straceFile = "");
    //! Unset trace
    void unsetTrace()
    {
        _bTrace = kFALSE;
    }
    //! Return the trace flag
    Bool_t getTrace()
    {
        return _bTrace;
    }
    //! Return the name of the trace file
    TString getTraceFile();

    //! Set the Flag of summits
    void setSummits()
    {
        _bSummits = kTRUE;
    }
    //! Unset the Flag of summits
    void unsetSummits()
    {
        _bSummits = kFALSE;
    }
    //! Set the Flag of CenterEdges
    void setCenterEdges()
    {
        _bCenterEdges = kTRUE;
    }
    //! Unset the Flag of CenterEdges
    void unsetCenterEdges()
    {
        _bCenterEdges = kFALSE;
    }

    //! Set the Flag of ThirdEdges
    void setThirdEdges()
    {
        _bThirdEdges = kTRUE;
    }
    //! Unset the Flag of ThirdEdges
    void unsetThirdEdges()
    {
        _bThirdEdges = kFALSE;
    }

    //! Set the Flag of CenterFaces
    void setCenterFaces()
    {
        _bCenterFaces = kTRUE;
    }
    //! Unset the Flag of CenterFaces
    void unsetCenterFaces()
    {
        _bCenterFaces = kFALSE;
    }

    //! Set the Flag of Centroide
    void setCentroide()
    {
        _bCentroide = kTRUE;
    }
    //! Unset the Flag of Centroide
    void unsetCentroide()
    {
        _bCentroide = kFALSE;
    }

    //! Set the Flag of CenterSegments
    void setCenterSegments()
    {
        _bCenterSegments = kTRUE;
    }
    //! Unset the Flag of CenterSegments
    void unsetCenterSegments()
    {
        _bCenterSegments = kFALSE;
    }

    //! Set the Flag of save candidates
    void setSaveCandidates()
    {
        _bSaveCandidates = kTRUE;
    }
    //! Unset the Flag of CenterSegments
    void unsetSaveCandidates()
    {
        _bSaveCandidates = kFALSE;
    }
    //! Return isSaveCandidates
    Bool_t isSaveCandidates()
    {
        return _bSaveCandidates;
    }
    //@}

    //---------------------------------------------
    /** @name Treat relational constraints
     **/
    //@{
    //! Add the constraint defined in a vector
    /*!
     The constraint defined by the equation
     \f[
     0 \leq \alpha_0 + \alpha_1 x_1 + \alpha_2 x_2 + \cdots + \alpha_n x_n
     \f]
     is stored in the vector as following \f$vec[0] = \alpha_0\f$, \f$vec[1] = \alpha_1\f$ and so on.
     */
    void addConstraint(TVectorD vecConstraint);
    //@}

    //---------------------------------------------
    /** @name Generation of the sample
     **/
    //@{
    //! Generates the sample
    /*!
     Two alagorithms are implemented
     \li MD (for \it OptimalPointsFromMaximalDistance )
     \li MMD (for \it OptimalPointsFromMaximalMinDistance)
     \param option (Option_t *) The option to pass (MD|MMD) (default MD).
     */
    virtual void generateSample(Option_t *option = "MD");
    //! Generates the list of candidates
    /*!
     \param option (Option_t *) The option to pass.
     */
    void generateCandidates(Option_t *option);
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    //! Prints the log
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Sampler::TSamplingMixture, ID_SAMPLER)
    // Classe to generate mixture sampling
};
} // Fin du namespace Sampler
} // Fin du namespace URANIE

#endif
// fin du fichier $RCSfile$.
