//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
// TSobolSequence Wed Jul 26 20:20:06 CEST 2006
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TSobolSequence.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Wed Jul 26 20:20:06 CEST 2006
 
 \brief Implementation of the class URANIE::Sampler::TSobolSequence
 
 */

#include "TSobolSequence.h"
/* primitive polynomials in binary encoding
 */
static const Int_t primitive_polynomials[SOBOL_MAX_DIMENSION] =
{ 1, 3, 7, 11, 13, 19, 25, 37, 59, 47, 61, 55, 41, 67, 97, 91, 109, 103, 115,
        131, 193, 137, 145, 143, 241, 157, 185, 167, 229, 171, 213, 191, 253,
        203, 211, 239, 247, 285, 369, 299 };

/* degrees of the primitive polynomials */
static const Int_t degree_table[SOBOL_MAX_DIMENSION] =
{ 0, 1, 2, 3, 3, 4, 4, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7,
        7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 8, 8, 8 };

/* initial values for direction tables, following
 * Bratley+Fox, taken from [Sobol+Levitan, preprint 1976]
 */
static const Int_t v_init[8][SOBOL_MAX_DIMENSION] =
{
{ 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
{ 0, 0, 1, 3, 1, 3, 1, 3, 3, 1, 3, 1, 3, 1, 3, 1, 1, 3, 1, 3, 1, 3, 1, 3, 3, 1,
        3, 1, 3, 1, 3, 1, 1, 3, 1, 3, 1, 3, 1, 3 },
{ 0, 0, 0, 7, 5, 1, 3, 3, 7, 5, 5, 7, 7, 1, 3, 3, 7, 5, 1, 1, 5, 3, 3, 1, 7, 5,
        1, 3, 3, 7, 5, 1, 1, 5, 7, 7, 5, 1, 3, 3 },
{ 0, 0, 0, 0, 0, 1, 7, 9, 13, 11, 1, 3, 7, 9, 5, 13, 13, 11, 3, 15, 5, 3, 15, 7,
        9, 13, 9, 1, 11, 7, 5, 15, 1, 15, 11, 5, 3, 1, 7, 9 },
{ 0, 0, 0, 0, 0, 0, 0, 9, 3, 27, 15, 29, 21, 23, 19, 11, 25, 7, 13, 17, 1, 25,
        29, 3, 31, 11, 5, 23, 27, 19, 21, 5, 1, 17, 13, 7, 15, 9, 31, 9 },
{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 37, 33, 7, 5, 11, 39, 63, 27, 17, 15,
        23, 29, 3, 21, 13, 31, 25, 9, 49, 33, 19, 29, 11, 19, 27, 15, 25 },
{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 13, 33, 115, 41, 79,
        17, 29, 119, 75, 73, 105, 7, 59, 65, 21, 3, 113, 61, 89, 45, 107 },
{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7, 23, 39 } };

ClassImp (URANIE::Sampler::TSobolSequence)

namespace URANIE
{
namespace Sampler
{
////////////////////////////////////////////////////////////
//  Constructor with size
////////////////////////////////////////////////////////////
TSobolSequence::TSobolSequence(ULong_t ndim) :
        TSequence(ndim)
{
    try
    {
        if (_nDim >= SOBOL_MAX_DIMENSION)
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TSobolSequence::TSobolSequence constructor ERROR : the dim %lu is greater than MAX[%d]",
                            _nDim, SOBOL_MAX_DIMENSION));
        }
        init();
    } catch (URANIE::Exceptions::UExceptions& ue)
    {
        ue.printMessage();
        if (ue.getMessageType() > URANIE::Exceptions::UExceptions::kWARNING)
            exit(-1);
    }
}
TSobolSequence::~TSobolSequence()
{
    if (_blog)
        cout << "destructs TSobolSequence" << endl;
}

//////////////////////////////////////////////////////////
//  The setInstance method
//////////////////////////////////////////////////////////
void TSobolSequence::setInstance(const ULong_t n)
{
    UNUSED(n); // ? UNUSED ?
    Int_t ell = 0;
    Int_t c = _nCounter;
    while (1)
    {
        ++ell;
        if ((c % 2) == 1)
            c /= 2;
        else
            break;
    }

    /* Check for exhaustion. */
    if (ell > SOBOL_BIT_COUNT)
        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                Form(
                        "TSobolSequence::setInstance ERROR : the ell value %d is greater than SOBOL_BIT_COUNT[%d]",
                        ell, SOBOL_BIT_COUNT));

    for (Int_t i_dim = 0; i_dim < (Int_t) _nDim; i_dim++)
    {
        const int direction_i = (Int_t)(v_direction(ell - 1, i_dim));
        const int old_numerator_i = (Int_t)(_last_numerator_vec(i_dim));
        const int new_numerator_i = old_numerator_i ^ direction_i;
        _last_numerator_vec(i_dim) = new_numerator_i;
        _dValue[i_dim] = new_numerator_i * _last_denominator_inv;
    }

    _nCounter++;
}

//////////////////////////////////////////////////////////
//  The init method
//////////////////////////////////////////////////////////
void TSobolSequence::init()
{
    if (_blog)
        cout << "init TSobolSequence" << endl;
    v_direction.ResizeTo(SOBOL_BIT_COUNT, _nDim);

    /* Initialize direction table in dimension 0. */
    for (Int_t k = 0; k < SOBOL_BIT_COUNT; k++)
        v_direction(k, 0) = 1.;

    Int_t ell;
    /* Initialize in remaining dimensions. */
    for (Int_t i_dim = 1; i_dim < (Int_t) _nDim; i_dim++)
    {
        if (_blog)
            cout << "i_dim[" << i_dim << "]" << endl;
        const Int_t poly_index = i_dim;
        const Int_t degree_i = degree_table[poly_index];
        Int_t includ[8];

        // Initialisation du vecteur
        for (Int_t i = 0; i < 8; i++)
            includ[i] = 0;

        /* Leading elements for dimension i come from v_init[][]. */
        for (Int_t j = 0; j < degree_i; j++)
            v_direction(j, i_dim) = v_init[j][i_dim];

        /* Expand the polynomial bit pattern to separate
         * components of the logical array includ[].
         */
        Int_t p_i = primitive_polynomials[poly_index];
        if (_blog)
            cout << "i_dim[" << i_dim << "] poly_index[" << poly_index
                    << "] degree_i[" << degree_i << "] p_i[" << p_i << "]"
                    << endl;
        for (Int_t k = degree_i - 1; k >= 0; k--)
        {
            includ[k] = ((p_i % 2) == 1);
            if (_blog)
                cout << "k[" << k << "] p_i % 2[" << p_i % 2 << "] p_i[" << p_i
                        << "]includ[k][" << includ[k] << "]" << endl;
            p_i /= 2;
        }

        if (_blog)
        {
            for (Int_t ii = 0; ii < 8; ii++)
                cout << includ[ii] << " ---- ";
            cout << endl;
        }

        /* Calculate remaining elements for this dimension,
         * as explained in Bratley+Fox, section 2.
         */
        for (Int_t j = degree_i; j < SOBOL_BIT_COUNT; j++)
        {
            Int_t newv = (Int_t) v_direction(j - degree_i, i_dim);
            ell = 1;
            for (Int_t k = 0; k < degree_i; k++)
            {
                ell *= 2;
                if (includ[k])
                    newv ^= (ell * (Int_t)(v_direction(j - k - 1, i_dim)));
            }
            v_direction(j, i_dim) = newv;
        }

    }

    if (_blog)
    {
        cout << endl << "***** Print the v_direction matrix **** " << endl;
        v_direction.Print();
    }
    /* Multiply columns of v by appropriate power of 2. */
    ell = 1;
    for (Int_t j = SOBOL_BIT_COUNT - 2; j >= 0; j--)
    {
        ell *= 2;
        if (_blog)
            cout << "j[" << j << "] ell[" << ell << "]" << endl;
        for (Int_t i_dim = 0; i_dim < (Int_t) _nDim; i_dim++)
        {
            v_direction(j, i_dim) *= ell;
        }
    }

    if (_blog)
    {
        cout << endl
                << "***** Print the v_direction matrix after Multiply columns by appropriate power of 2.**** "
                << endl;
        v_direction.Print();
    }

    /* 1/(common denominator of the elements in v_direction) */
    _last_denominator_inv = 1.0 / (2.0 * ell);

    _nCounter = 0;

    /* final setup */
    _last_numerator_vec.ResizeTo(_nDim);
    _last_numerator_vec.Zero();
    if (_blog)
    {
        cout << endl << "***** Print the _last_numerator_vec vector **** "
                << endl;
        _last_numerator_vec.Print();
    }
}

//////////////////////////////////////////////////////////
//  Printing Log
//////////////////////////////////////////////////////////
void TSobolSequence::printLog(Option_t *option)
{
    cout << "TSobolSequence::printLog[" << option << "]" << endl;
}

} // Fin du namespace Sampler
} // Fin du namespace URANIE
// fin du fichier $RCSfile$.
