//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TSpaceFilling.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   Mon Jul 24 14:15:38 2006
 
 \brief  Interface for Space Filling
 
 
 */
#ifndef TSPACEFILLING_H
#define TSPACEFILLING_H
/*!
 
 @class URANIE::Sampler::TSpaceFilling

 @brief Description of a generation of sample by space filling (FAST).

 \section Space_Filling_algorithm Method of sampling
 Two methods are implemented
 \li Cukier (1973) algorithm
 \f[ x_{i} = \bar{x_{i}} \exp^{\bar{\gamma_i} \sin{(\omega_i s)}}\f]
 For normalisation, we take \f$ \bar{x_{i}} = \exp^{-\bar{\gamma_i}}\f$ with \f$ \gamma_i = 1\f$ . 
 \li Saltelli A (1999)
 \f[ x_{i} = \frac{1}{2} + \frac{1}{\pi} \arcsin{(\sin{(\omega_i s)})}\f]
 \li Saltelli B (1999)
 \f[ x_{i} = \frac{1}{2} + \frac{1}{\pi} \arcsin{[\sin{(\omega_i s + \varphi_i)}]}\f]
 where \f$ \varphi_i \in [0,2\pi)\f$ randomly.

 \section Space_Filling_usescase Use case

 First of all, a pointer to a TDataServer is constructed
 \code
 TDataServer *tds = new TDataServer("tdsSpaceFilling", "Test for SpaceFilling");
 tds->addAttribute( new TAttribute("x1", -1.0 , 2.) );
 tds->addAttribute( new TAttribute("x2",  0.0 , 3.) );
 tds->addAttribute( new TAttribute("x3",  1.0, 1.5) );
 \endcode

 Then, a pointer to an object TSpaceFilling is constructed using the TDataServer, of the adopted sequence ("cukier"|"Saltelli A"|"Saltelli B"),
 and the method \b generateSample() is applied


 \code
 TSpaceFilling * tsf = new TSpaceFilling(tdsqmc, "", 300, TSpaceFilling::kCukier);
 tsf->generateSample();
 \endcode

 */

#include "TSamplerDeterminist.h"
#include "TSequence.h"

namespace URANIE
{
namespace Sampler
{
class TSpaceFilling: public TSamplerDeterminist
{

    // Attributes
public:
    enum EFunction
    {
        kCukier, kSaltelliA, kSaltelliB
    };
private:
    Int_t * _nValFreq; //[_nX]       ///< The array of frequences
    EFunction _nFunction;       ///< The function G
    Bool_t _bPermutation; ///< the boolean to permute each attribute (default = kFALSE)

    // Operations
public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //!  Constructor from a TDataServer, the name of the method and the size of the sample
    TSpaceFilling(URANIE::DataServer::TDataServer *tds, Option_t *option,
            Int_t nCalcul, EFunction = TSpaceFilling::kSaltelliA);

    //! Default destructor
    virtual ~TSpaceFilling();
    //@}

    //---------------------------------------------
    /** @name Setting and Getting global informations
     **/
    //@{
    //! Sets the size of the matrice
    void setSize(Int_t n)
    {
        _nS = n;
    }
    //! Set the permutation attribute
    void setPermutation(Bool_t bpermute)
    {
        _bPermutation = bpermute;
    }
    //! Get the permutation attribute
    Bool_t getPermutation()
    {
        return _bPermutation;
    }
    //@}

    //---------------------------------------------
    /** @name Methods for the frequencies
     The Fast method needs a vector a frequencies for each uncertainty attributes
     **/
    //@{
    //! Sets the default frequencies by an algorithm
    void setDefaultFrequencies();
    //! Sets the frequencies by an array
    void setFrequencies(Int_t nfreq, Int_t *valfreq);
    //! Returns the frequency
    Int_t getFrequency(Int_t i)
    {
        return _nValFreq[i];
    }
    //@}

    //---------------------------------------------
    /** @name Methods for the fransformation
     The Fast method needs also a function to transform data in [0,1]
     **/
    //@{
    //! Sets the Transform fonction
    void setTransformFunction(EFunction ntype)
    {
        _nFunction = ntype;
    }
    //! Gets the transform function
    EFunction getTransformFunction()
    {
        return _nFunction;
    }
    //@}

    //---------------------------------------------
    /** @name Generation of the sample
     **/
    //@{
    //! Generates the sample
    /*!
     \todo FGA 2011-12-15 10:10:20 Pass the seed in the option "seed=123456" for the kSaltelliB Transformation
     
     */
    virtual void generateSample(Option_t *option = "");
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    //! Prints the log
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Sampler::TSpaceFilling, ID_SAMPLER)
    //Classe de g�n�ration par Space Filling
};
} // Fin du namespace Sampler
} // Fin du namespace URANIE

#endif
// fin du fichier $RCSfile$.
