//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
#include <iostream>
#include <fstream>
#include <sstream>
#include <math.h>
#include <stdlib.h>
#include "simplex.h"

#define epsilon 1.e-8

Simplex::Simplex(const int ndim, const vector<vector<double> > planes,
        const double k, const bool trace) :
        _ndim(ndim), _k(k), _trace(trace), _refFile("")
{
    if (k <= 0.)
        throw ErrorSimplex("k must be positive!");
    _Nplanes = planes.size();
    calculTeta(planes);
    initSummits();
}

void Simplex::calculTeta(const vector<vector<double> >& planes)
{
    vector<double> vec;
    size_t i, j;

    for (i = 0; i < planes.size(); i++)
    {
        vec.clear();
        for (j = 1; j < planes[i].size(); j++)
            vec.push_back(planes[i][j] + ((planes[i][0]) / _k));
        _teta.push_back(vec);
    }
}

void Simplex::initSummits()
{
    vector<double> vec;
    size_t i, j;

    for (i = 0; i < _ndim; i++)
    {
        vec.clear();
        for (j = 0; j < _ndim; j++)
        {
            if (j != i)
                vec.push_back(0.);
            else
                vec.push_back(_k);
        }
        _summits.push_back(vec);
    }
}

bool Simplex::Summits()
{
    vector < list < vector<double> > ::iterator > toberemoved;
    list < vector<double> > new_summits;
    list<vector<double> >::iterator it1, it2;
    vector<double> new_s, d;
    vector<int> pos;
    set<size_t> cu;
    set<int>::iterator itcu;
    double alpha, beta;
    size_t i, ip, nbp, is1, is2;
    bool inside, outside, OK = true;

    // on considere la contrainte ip correspondant a un hyperplan ip
    for (ip = 0; ip < _Nplanes; ip++)
    {

        // initialisation de la liste des nouveaux sommets et des sommets a supprimer
        new_summits.clear();
        toberemoved.clear();

        // on calcul la position des sommets par rapport au demi-espace d>0 dedans d=0 � la frontiere d<0 dehors
        d = innerProduct(_teta[ip]);

        inside = false;
        outside = false;
        pos.clear();
        for (i = 0; i < d.size(); i++)
        {
            if (d[i] > 0)
            {
                inside = true;
                pos.push_back(1);
            }
            else if (d[i] < 0)
            {
                outside = true;
                pos.push_back(-1);
            }
            else
                pos.push_back(0);
        }

        if (inside && outside)
        {
            // l'hyperplan intersecte le simplex

            // step 2 of Piepel88 article
            // Initial and previously introduced new constraints are checked to see
            //    if they are rendered unnecessary by the introduction of the new constraint.
            // If unnecessary constraints are identified, they are deleted from the list of current constraints after printing a warning message
            checkConstraints(cu);

            dSummits(ip, cu);

            is1 = 0;
            for (it1 = _dsummits.begin(); it1 != _dsummits.end(); it1++)
            {
                is2 = 0;
                for (it2 = _dsummits.begin(); it2 != it1; it2++)
                {

                    // on considere les paires de sommets dont le segment intersecte le nouvel hyperplan
                    if (pos[is1] * pos[is2] < 0)
                    {
                        nbp = 0;
                        for (i = 0; i < (*it1).size(); i++)
                        {
                            if (((*it1)[i] == 0.) && ((*it2)[i] == 0.))
                            {
                                nbp++;
                                if ((nbp >= _ndim - 2)
                                        || ((nbp == 1) && (_ndim == 2)))
                                    break;
                            }
                        }
                        // 2 points sont sur le meme edge si au moins ndim-2 contraintes ont d=0 pour ces 2 points
                        if ((nbp >= _ndim - 2) || ((nbp == 1) && (_ndim == 2)))
                        {
                            alpha = d[is2] / (d[is2] - d[is1]);
                            beta = 1. - alpha;
                            new_s.clear();
                            for (i = 0; i < _ndim; i++)
                                new_s.push_back(
                                        alpha * (*it1)[i] + beta * (*it2)[i]);
                            new_summits.push_back(new_s);
                        }
                    }
                    is2++;
                }
                is1++;
            }

            // suppression des sommets en dehors de la contrainte ip
            is1 = 0;
            for (it1 = _summits.begin(); it1 != _summits.end(); it1++)
                if (pos[is1++] < 0)
                    toberemoved.push_back(it1);
            for (i = 0; i < toberemoved.size(); i++)
                _summits.erase(toberemoved[i]);

            // rajout des nouveaux sommets
            for (it1 = new_summits.begin(); it1 != new_summits.end(); it1++)
                _summits.push_back(*it1);

            cu.insert(ip);

        }
        else if (outside)
            throw ErrorSimplex("Pas de solution");
        else if (_trace)
            cout << "contrainte " << ip << " inutile" << endl;

    }

    if (_trace)
        displayPoints("Vertex", _summits);
    if (_refFile.size() > 0)
        OK = checkPoints("Vertex", _summits);

    dSummits(_Nplanes - 1, cu);

    return OK;
}

vector<double> Simplex::innerProduct(const vector<double>& teta)
{
    double sigma;
    vector<double> vec;
    list<vector<double> >::iterator it;
    size_t i;

    for (it = _summits.begin(); it != _summits.end(); it++)
    {
        if (teta.size() != (*it).size())
        {
            ostringstream oss;
            oss << "Probleme de dimension: " << teta.size() << " != "
                    << (*it).size();
            throw ErrorSimplex(oss.str().c_str());
        }
        sigma = 0.;
        for (i = 0; i < teta.size(); i++)
            sigma += (*it)[i] * teta[i];
        if (fabs(sigma) < epsilon)
            sigma = 0.;
        vec.push_back(sigma);
    }
    return vec;
}

void Simplex::checkConstraints(set<size_t> &cu)
{
    vector<double> dStep2, dStep2i, dStep2j;
    vector<int> ci;
    set<size_t>::iterator itcu1, itcu2;
    size_t i, k;

    ci.clear();
    for (itcu1 = cu.begin(); itcu1 != cu.end(); itcu1++)
    {
        dStep2 = innerProduct(_teta[*itcu1]);

        bool bZero = true;
        for (i = 0; i < dStep2.size(); i++)
        {
            if (fabs(dStep2[i]) > epsilon)
                bZero = false;
        }
        if (bZero)
        {
            if (_trace)
                cout << "contrainte " << *itcu1 << " inutile" << endl;
            ci.push_back(*itcu1);
        }
    }
    for (i = 0; i < ci.size(); i++)
        cu.erase(ci[i]);

    ci.clear();
    for (itcu1 = cu.begin(); itcu1 != cu.end(); itcu1++)
    {
        dStep2i = innerProduct(_teta[*itcu1]);

        for (itcu2 = cu.begin(); itcu2 != cu.end(); itcu2++)
        {
            if (itcu2 == itcu1)
                continue;
            dStep2j = innerProduct(_teta[*itcu2]);
            for (k = 0; k < dStep2i.size(); k++)
            {
                if (fabs(dStep2i[k]) > epsilon)
                    continue;
                if (fabs(dStep2j[k]) > epsilon)
                    break;
            }
            if (k == dStep2i.size())
            {
                ci.push_back(*itcu1);
                if (_trace)
                    cout << "contrainte " << *itcu1 << " inutile" << endl;
                break;
            }
        }
    }
    for (i = 0; i < ci.size(); i++)
        cu.erase(ci[i]);
}

void Simplex::dSummits(const size_t i0, const set<size_t>& cu)
{
    double sigma;
    vector<double> vec, d;
    list<vector<double> >::iterator it;
    set<size_t>::const_iterator itcu;
    size_t i, j;

    if (i0 >= _teta.size())
    {
        ostringstream oss;
        oss << "On sort du tableau teta: " << _teta.size() << " <= " << i0;
        throw ErrorSimplex(oss.str().c_str());
    }

    _dsummits.clear();

    for (it = _summits.begin(); it != _summits.end(); it++)
    {
        // on met les coordonn�es des sommets dans les premiers elements de dsummits => contrainte Xi >= 0
        vec.clear();
        for (i = 0; i < (*it).size(); i++)
            vec.push_back((*it)[i]);
        // on ajoute les contraintes dans dsummits
        for (i = 0; i <= i0; i++)
        {
            bool flag = false;
            for (itcu = cu.begin(); itcu != cu.end(); itcu++)
                if (i == *itcu)
                    flag = true;
            if (flag)
            {
                sigma = 0.;
                for (j = 0; j < _teta[i].size(); j++)
                    sigma += (*it)[j] * _teta[i][j];
                if (fabs(sigma) < epsilon)
                    sigma = 0.;
                vec.push_back(sigma);
            }
        }
        _dsummits.push_back(vec);
    }
}

bool Simplex::CenterEdges()
{
    list<vector<double> >::iterator it1, it2;
    vector<double> new_s;
    double alpha, beta;
    size_t i, nbp;
    bool OK = true;

    for (it1 = _dsummits.begin(); it1 != _dsummits.end(); it1++)
    {
        for (it2 = _dsummits.begin(); it2 != it1; it2++)
        {
            nbp = 0;
            for (i = 0; i < (*it1).size(); i++)
            {
                if (((*it1)[i] == 0.) && ((*it2)[i] == 0.))
                {
                    nbp++;
                    if ((nbp >= _ndim - 2) || ((nbp == 1) && (_ndim == 2)))
                        break;
                }
            }
            // 2 points sont sur le meme edge si au moins ndim-2 contraintes ont d=0 pour ces 2 points
            if ((nbp >= _ndim - 2) || ((nbp == 1) && (_ndim == 2)))
            {

                // Calcul du centre de l'arete
                alpha = 0.5;
                beta = 0.5;
                new_s.clear();
                for (i = 0; i < _ndim; i++)
                    new_s.push_back(alpha * (*it1)[i] + beta * (*it2)[i]);
                _centerEdges.push_back(new_s);
            }
        }
    }
    if (_trace)
        displayPoints("CenterEdges", _centerEdges);
    if (_refFile.size() > 0)
        OK = checkPoints("CentEdge", _centerEdges);

    return OK;
}

bool Simplex::ThirdEdges()
{
    list<vector<double> >::iterator it1, it2;
    vector<double> new_s;
    double alpha, beta;
    size_t i, nbp;
    bool OK = true;

    for (it1 = _dsummits.begin(); it1 != _dsummits.end(); it1++)
    {
        for (it2 = _dsummits.begin(); it2 != it1; it2++)
        {
            nbp = 0;
            for (i = 0; i < (*it1).size(); i++)
            {
                if (((*it1)[i] == 0.) && ((*it2)[i] == 0.))
                {
                    nbp++;
                    if ((nbp >= _ndim - 2) || ((nbp == 1) && (_ndim == 2)))
                        break;
                }
            }
            // 2 points sont sur le meme edge si au moins ndim-2 contraintes ont d=0 pour ces 2 points
            if ((nbp >= _ndim - 2) || ((nbp == 1) && (_ndim == 2)))
            {
                // Calcul des tiers de l'arete
                alpha = 1. / 3.;
                beta = 2. / 3.;
                new_s.clear();
                for (i = 0; i < _ndim; i++)
                    new_s.push_back(alpha * (*it1)[i] + beta * (*it2)[i]);
                _thirdEdges.push_back(new_s);

                alpha = 2. / 3.;
                beta = 1. / 3.;
                new_s.clear();
                for (i = 0; i < _ndim; i++)
                    new_s.push_back(alpha * (*it1)[i] + beta * (*it2)[i]);
                _thirdEdges.push_back(new_s);
            }
        }
    }
    if (_trace)
        displayPoints("ThirdEdges", _thirdEdges);
    if (_refFile.size() > 0)
        OK = checkPoints("ThirdEdge", _thirdEdges);

    return OK;
}

bool Simplex::PointsOnEdges()
{
    list<vector<double> >::iterator it1, it2;
    vector<double> new_s;
    double alpha, beta;
    size_t i, nbp;
    bool OK = true;

    for (it1 = _dsummits.begin(); it1 != _dsummits.end(); it1++)
    {
        for (it2 = _dsummits.begin(); it2 != it1; it2++)
        {
            nbp = 0;
            for (i = 0; i < (*it1).size(); i++)
            {
                if (((*it1)[i] == 0.) && ((*it2)[i] == 0.))
                {
                    nbp++;
                    if ((nbp >= _ndim - 2) || ((nbp == 1) && (_ndim == 2)))
                        break;
                }
            }
            // 2 points sont sur le meme edge si au moins ndim-2 contraintes ont d=0 pour ces 2 points
            if ((nbp >= _ndim - 2) || ((nbp == 1) && (_ndim == 2)))
            {

                // Calcul du centre de l'arete
                alpha = 0.5;
                beta = 0.5;
                new_s.clear();
                for (i = 0; i < _ndim; i++)
                    new_s.push_back(alpha * (*it1)[i] + beta * (*it2)[i]);
                _centerEdges.push_back(new_s);

                // Calcul des tiers de l'arete
                alpha = 1. / 3.;
                beta = 2. / 3.;
                new_s.clear();
                for (i = 0; i < _ndim; i++)
                    new_s.push_back(alpha * (*it1)[i] + beta * (*it2)[i]);
                _thirdEdges.push_back(new_s);

                alpha = 2. / 3.;
                beta = 1. / 3.;
                new_s.clear();
                for (i = 0; i < _ndim; i++)
                    new_s.push_back(alpha * (*it1)[i] + beta * (*it2)[i]);
                _thirdEdges.push_back(new_s);
            }
        }
    }
    if (_trace)
    {
        displayPoints("CenterEdges", _centerEdges);
        displayPoints("ThirdEdges", _thirdEdges);
    }
    if (_refFile.size() > 0)
        OK = checkPoints("CentEdge", _centerEdges)
                && checkPoints("ThirdEdge", _thirdEdges);

    return OK;
}

bool Simplex::CenterFaces()
{
    list < vector<double> > vide;
    list<vector<double> >::iterator it;
    vector < list<vector<double> > > summits_by_face;
    vector<double> center, vec;
    size_t i, j, nc;
    bool OK = true;

    it = _dsummits.begin();
    nc = (*it).size();
    for (i = 0; i < nc; i++)
        summits_by_face.push_back(vide);

    for (it = _dsummits.begin(); it != _dsummits.end(); it++)
    {
        for (i = 0; i < (*it).size(); i++)
        {
            if ((*it)[i] == 0.)
            {
                vec.clear();
                for (j = 0; j < _ndim; j++)
                    vec.push_back((*it)[j]);
                summits_by_face[i].push_back(vec);
            }
        }
    }

    // determination des centres des contraintes utiles
    for (i = 0; i < nc; i++)
    {
        if (summits_by_face[i].size() > 0)
        {
            center.clear();
            for (j = 0; j < _ndim; j++)
                center.push_back(0.);
            for (it = summits_by_face[i].begin();
                    it != summits_by_face[i].end(); it++)
            {
                if ((*it).size() != _ndim)
                {
                    ostringstream oss;
                    oss << "Probleme de dimension: " << _ndim << " != "
                            << (*it).size();
                    throw ErrorSimplex(oss.str().c_str());
                }
                for (j = 0; j < _ndim; j++)
                    center[j] += (*it)[j];
            }
            for (j = 0; j < _ndim; j++)
                center[j] /= summits_by_face[i].size();
            _centerFaces.push_back(center);
        }
    }
    if (_trace)
        displayPoints("CenterFaces", _centerFaces);
    if (_refFile.size() > 0)
        OK = checkPoints("PlaneCent", _centerFaces);

    return OK;
}

void Simplex::Centroide()
{
    list<vector<double> >::iterator it;
    double val;
    int ns;
    size_t i;

    ns = _summits.size();

    for (i = 0; i < _ndim; i++)
    {
        val = 0.;
        for (it = _summits.begin(); it != _summits.end(); it++)
        {
            val += (*it)[i];
        }
        val /= ns;
        _centroide.push_back(val);
    }
    if (_trace)
        displayVector("centroide", _centroide);
}

bool Simplex::CenterSegments()
{
    list<vector<double> >::iterator it;
    vector<double> vec;
    double alpha = 0.5;
    size_t i;
    bool OK = true;

    if (_centroide.size() == 0)
        return false;

    for (it = _summits.begin(); it != _summits.end(); it++)
    {
        vec.clear();
        for (i = 0; i < _ndim; i++)
            vec.push_back(alpha * (*it)[i] + alpha * _centroide[i]);
        _centerSegments.push_back(vec);
    }
    if (_trace)
        displayPoints("CenterSegments", _centerSegments);
    if (_refFile.size() > 0)
        OK = checkPoints("AxialCB", _centerSegments);

    return OK;
}

void Simplex::displayPoints(const string title, list<vector<double> >& points)
{
    list<vector<double> >::iterator it;
    size_t j;
    // affichage des sommets
    if (points.size() < 100)
    {
        cout << title << endl;
        for (it = points.begin(); it != points.end(); it++)
        {
            for (j = 0; j < (*it).size(); j++)
                cout << (*it)[j] << " ";
            cout << endl;
        }
    }
    else
        cout << points.size() << " " << title << endl;
}

void Simplex::displayVector(const string name, const vector<double>& d)
{
    size_t j;

    cout << name << ": ";
    for (j = 0; j < d.size(); j++)
        cout << d[j] << " ";
    cout << endl;
}

bool Simplex::checkPoints(const string type, list<vector<double> >& points)
{
    bool OK;

    string buffer, point;
    istringstream iss;
    size_t id, i, nbp;
    vector<double> vec;
    list < vector<double> > reference;
    list<vector<double> >::iterator it1, it2;
    double val, dist;

    if (_refFile.size() == 0)
        throw ErrorSimplex("Pas de fichier de reference!!!!");

    // lecture des sommets de reference dans le fichier d'entree
    ifstream infile(_refFile.c_str());
    if (!infile)
    {
        ostringstream oss;
        oss << "Fichier " << _refFile << " inexistant";
        throw ErrorSimplex(oss.str().c_str());
    }
    while (getline(infile, buffer))
    {
        for (i = 0; i < buffer.size(); i++)
        {
            size_t pos = buffer.find(",");
            if (pos != string::npos)
                buffer.replace(pos, 1, ".");
        }
        iss.clear();
        iss.str(buffer);
        iss >> id >> point;
        if (point.find(type) != string::npos)
        {
            vec.clear();
            for (i = 0; i < _ndim; i++)
            {
                iss >> val;
                vec.push_back(val);
            }
            reference.push_back(vec);
        }
    }

    // comparaison des sommets de reference et obtenus
    if (points.size() != reference.size())
    {
        ostringstream oss;
        oss << "Pas le meme nombre de " << type << ": " << points.size()
                << " != " << reference.size();
        throw ErrorSimplex(oss.str().c_str());
    }

    nbp = 0;
    for (it2 = reference.begin(); it2 != reference.end(); it2++)
    {
        bool trouve = false;
        for (it1 = points.begin(); it1 != points.end(); it1++)
        {
            dist = distance(*it1, *it2);
            if (dist < epsilon)
            {
                nbp++;
                trouve = true;
            }
        }
        if (!trouve)
            displayVector("Point non trouve", *it2);
    }

    if (nbp == reference.size())
        OK = true;
    else
        OK = false;

    return OK;

}

void Simplex::OptimalPointsFromMaximalMinDistance(const int nbp,
        const bool iter)
{
    double optimisation;
    list<vector<double> >::iterator it;

    _nbp = nbp;
    _index = 0;

    _optimalPoints.clear();

    do
    {
        optimisation = _optimalDistMin;
        OptimalPointsFromList(_summits, "Sommet");
        OptimalPointsFromList(_centerEdges, "CentreArete");
        OptimalPointsFromList(_thirdEdges, "TiersArete");
        OptimalPointsFromList(_centerFaces, "CentreFace");
        if (_centroide.size() > 0)
        {
            OptimalPointsFromList(_centerSegments, "CentreSegment");
            isThisPointOptimal(_centroide, "Centroide");
        }
        if (!iter)
            optimisation = _optimalDistMin;
        cout << " ** Optimisation [" << _optimalDistMin << "]" << endl;
    } while (optimisation != _optimalDistMin);

}

void Simplex::calculateOptimalMin()
{
    size_t i, j;
    double dist;

    if (_optimalPoints.size() != _nbp)
    {
        ostringstream oss;
        oss << "Probleme de nombre de points optimaux: "
                << _optimalPoints.size() << " != " << _nbp;
        throw ErrorSimplex(oss.str().c_str());
    }

    // Calcul la distance minimale de chacun des points avec les autres
    for (i = 0; i < _nbp; i++)
    {
        _optimalDists[i] = 1;
        _optimalIndex[i] = -1;
        for (j = 0; j < _nbp; j++)
        {
            if (i != j)
            {
                dist = distance(_optimalPoints[i], _optimalPoints[j]);
                if (dist < _optimalDists[i])
                {
                    _optimalDists[i] = dist;
                    _optimalIndex[i] = j;
                }
            }
        }
    }
    // Determine la distance minimum des distance mimimum pour chacun des points
    _optimalDistMin = _optimalDists[0];
    _optimalIndexMin1 = 0;
    _optimalIndexMin2 = _optimalIndex[0];
    for (i = 1; i < _nbp; i++)
    {
        if (_optimalDists[i] < _optimalDistMin)
        {
            _optimalDistMin = _optimalDists[i];
            _optimalIndexMin1 = i;
            _optimalIndexMin2 = _optimalIndex[i];
        }
    }
}

void Simplex::isThisPointOptimal(const vector<double>& point, const string type)
{
    size_t i;
    size_t i0=0;                 // point optimal le plus proche du nouveau point
    int i1;            // deuxieme point optimal le plus proche du nouveau point
    double dist, dist1, dist2;
    double distmin; // distance minimale entre le nouveau point et les points optimaux
    double distp; // deuxieme distance minimale entre le nouveau point et les points optimaux
    double distmin1; // distance minimale entre _optimalIndexMin1 et les autres points optimaux 
                     // sauf _optimalIndexMin2 mais avec le nouveau point
    double distmin2; // distance minimale entre _optimalIndexMin2 et les autres points optimaux 
                     // sauf _optimalIndexMin1 mais avec le nouveau point

    if (_optimalPoints.size() != _nbp)
    {
        ostringstream oss;
        oss << "Probleme de nombre de points optimaux: "
                << _optimalPoints.size() << " != " << _nbp;
        throw ErrorSimplex(oss.str().c_str());
    }

    // calcul de la distance minimal entre point et les nbp points: le point le plus proche de point est i0
    distmin = 1.;
    for (i = 0; i < _nbp; i++)
    {
        dist = distance(point, _optimalPoints[i]);
        if (dist < distmin)
        {
            distmin = dist;
            i0 = i;
        }
    }

    if (distmin < _optimalDistMin)
    {
        // il faut eliminer point ou i0
        distp = 1.;
        for (i = 0; i < _nbp; i++)
        {
            if (i != i0)
            {
                dist = distance(point, _optimalPoints[i]);
                if (dist < distp)
                {
                    distp = dist;
                    i1 = i;
                }
            }
        }
        if (distp > _optimalDists[i0])
        {
            // il faut remplacer i0 par point
            _optimalPoints[i0] = point;
            _optimalTypes[i0] = type;
        }
    }
    else
    {
        // il faut eliminer _optimalIndexMin1 ou _optimalIndexMin2
        distmin1 = distance(point, _optimalPoints[_optimalIndexMin1]);
        distmin2 = distance(point, _optimalPoints[_optimalIndexMin2]);
        for (i = 0; i < _nbp; i++)
        {
            if ((i != _optimalIndexMin1) && (i != _optimalIndexMin2))
            {
                dist1 = distance(_optimalPoints[_optimalIndexMin1],
                        _optimalPoints[i]);
                dist2 = distance(_optimalPoints[_optimalIndexMin2],
                        _optimalPoints[i]);
                if (dist1 < distmin1)
                    distmin1 = dist1;
                if (dist2 < distmin2)
                    distmin2 = dist2;
            }
        }
        if (distmin1 < distmin2)
        {
            // il faut remplacer _optimalIndexMin1 par point
            _optimalPoints[_optimalIndexMin1] = point;
            _optimalTypes[_optimalIndexMin1] = type;
        }
        else
        {
            // il faut remplacer _optimalIndexMin2 par point
            _optimalPoints[_optimalIndexMin2] = point;
            _optimalTypes[_optimalIndexMin2] = type;
        }
    }

    // lorsqu'on a nbp points on calcul les distances minimales
    calculateOptimalMin();
}

double Simplex::distance(const vector<double>& P1, const vector<double>& P2)
{
    double val, dist;

    if (P1.size() != _ndim)
    {
        ostringstream oss;
        oss << "Probleme de dimension: " << P1.size() << " != " << _ndim;
        throw ErrorSimplex(oss.str().c_str());
    }

    if (P2.size() != _ndim)
    {
        ostringstream oss;
        oss << "Probleme de dimension: " << P2.size() << " != " << _ndim;
        throw ErrorSimplex(oss.str().c_str());
    }

    dist = 0.;
    for (size_t i = 0; i < _ndim; i++)
    {
        val = P1[i] - P2[i];
        dist += val * val;
    }
    return sqrt(dist);
}

void Simplex::OptimalPointsFromList(list<vector<double> >& points,
        const string type)
{
    list<vector<double> >::iterator it;

    for (it = points.begin(); it != points.end(); it++)
    {
        // on prend d'abord les nbp premiers points
        if (_index < _nbp)
        {
            _optimalPoints.push_back(*it);
            _optimalTypes.push_back(type);
            _optimalDists.push_back(1);
            _optimalIndex.push_back(-1);
        }
        // lorsqu'on a nbp points on calcul les distances minimales
        if (_index == _nbp - 1)
            calculateOptimalMin();
        // Ensuite on voit si chacun des points supplementaire est optimal
        else if (_index >= _nbp)
            isThisPointOptimal(*it, type);
        _index++;
    }
}

void Simplex::OptimalPointsFromMaximalDistance(const int nbp)
{
    list<vector<double> >::iterator it;
    double dist, distmin, maxdistmin;
    int i0=-1, j0=-1;

    _optimalPoints.clear();
    _points.clear();
    for (it = _summits.begin(); it != _summits.end(); it++)
        _points.push_back(*it);
    for (it = _centerEdges.begin(); it != _centerEdges.end(); it++)
        _points.push_back(*it);
    for (it = _thirdEdges.begin(); it != _thirdEdges.end(); it++)
        _points.push_back(*it);
    for (it = _centerFaces.begin(); it != _centerFaces.end(); it++)
        _points.push_back(*it);
    for (it = _centerSegments.begin(); it != _centerSegments.end(); it++)
        _points.push_back(*it);
    if (_centroide.size() > 0)
        _points.push_back(_centroide);

    _maximalDist = 0.;
    for (size_t i = 0; i < _points.size(); i++)
        for (size_t j = 0; j < i; j++)
        {
            dist = distance(_points[i], _points[j]);
            if (dist > _maximalDist)
            {
                _maximalDist = dist;
                i0 = i;
                j0 = j;
            }
        }
    _optimalPoints.push_back(_points[i0]);
    _optimalPoints.push_back(_points[j0]);
    _optimalTypes.push_back(getPointType(i0));
    _optimalTypes.push_back(getPointType(j0));

    for (int k = 0; k < nbp - 2; k++)
    {
        maxdistmin = 0.;
        i0 = -1;
        for (size_t i = 0; i < _points.size(); i++)
        {
            distmin = 1.;
            for (size_t j = 0; j < _optimalPoints.size(); j++)
            {
                dist = distance(_points[i], _optimalPoints[j]);
                if (dist < distmin)
                {
                    distmin = dist;
                }
            }
            if (distmin > maxdistmin)
            {
                i0 = i;
                maxdistmin = distmin;
            }
        }
        _optimalPoints.push_back(_points[i0]);
        _optimalTypes.push_back(getPointType(i0));
    }

}

string Simplex::getPointType(const size_t i)
{
    if (i < _summits.size())
        return "Sommet";
    else if (i < _summits.size() + _centerEdges.size())
        return "CentreArete";
    else if (i < _summits.size() + _centerEdges.size() + _thirdEdges.size())
        return "TiersArete";
    else if (i
            < _summits.size() + _centerEdges.size() + _thirdEdges.size()
                    + _centerFaces.size())
        return "CentreFace";
    else if (i
            < _summits.size() + _centerEdges.size() + _thirdEdges.size()
                    + _centerFaces.size() + _centerSegments.size())
        return "CentreSegment";
    else if (i
            == _summits.size() + _centerEdges.size() + _thirdEdges.size()
                    + _centerFaces.size() + _centerSegments.size())
        return "Centroide";
    else
        return "Error";
}

