//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TestTConstrLHS.cxx
 \author JB Blanchard
 \date   $Date$
 
 \brief  Test File of URANIE::Sampler::TSampler class
 
 */

// Uranie
#include "TestTConstrLHS.h"
#include "TAttribute.h"
#include "TUniformDistribution.h"
#include "UDisableTest.h"
#include "uranie_commands.h"
#include "TSampling.h"

// ROOT
#include "TString.h"

#include <iostream>
#include <math.h>

#define DBL_PRECISION_COMPARE 1.e-10
#define EPSILON 0.04

using namespace std;

TConstrLHSTest::TConstrLHSTest() :
    _tds(NULL), _sampler(NULL), _nS(100), _nseed(1234)
{
    _workingTestPath = gSystem->pwd();
}

TConstrLHSTest::~TConstrLHSTest()
{
    if (_sampler)
    {
        delete _sampler;
        _sampler = NULL;
    }

    if (_tds)
    {
        delete _tds;
        _tds = NULL;
    }

}

void TConstrLHSTest::setUp()
{

    _tds = new URANIE::DataServer::TDataServer("foo","bar");
    Int_t dim=4;

    for (Int_t j=0;j<dim;j++)
    {
 	_tds->addAttribute(new URANIE::DataServer::TUniformDistribution(Form("x_%d",j),0.0,(double)(2*j+1)));
	_tds->getAttribute( Form("x_%d",j) )->delShare();
    }

}

void TConstrLHSTest::tearDown()
{
    delete _sampler;
    _sampler = NULL;

    delete _tds;
    _tds = NULL;

}


void OneConstraint(double *param, double *res)
{

    double p1_1 = param[0], p1_2= param[1] ; // 1-st constraint
    double p2_1 = param[2], p2_2= param[3] ; // 2-nd constraint

    res[0] = ((p1_1 - p1_2<0) ? 0 : 1);
    
}

void TwoConstraints(double *param, double *res)
{

    double p1_1 = param[0], p1_2= param[1] ; // 1-st constraint
    double p2_1 = param[2], p2_2= param[3] ; // 2-nd constraint

    res[0] = ((p1_1 - p1_2<0) ? 0 : 1);
    res[1] = ((p2_1 - p2_2<0) ? 0 : 1);
    
}

void OneConstraintInvert(double *param, double *res)
{

    double p1_1 = param[0], p1_2= param[1] ; // 1-st constraint
    double p2_1 = param[2], p2_2= param[3] ; // 2-nd constraint

    res[0] = ((p1_1 - p1_2>0) ? 0 : 1);
    
}

void TwoConstraintsInvert(double *param, double *res)
{

    double p1_1 = param[0], p1_2= param[1] ; // 1-st constraint
    double p2_1 = param[2], p2_2= param[3] ; // 2-nd constraint

    res[0] = ((p1_1 - p1_2>0) ? 0 : 1);
    res[1] = ((p2_1 - p2_2>0) ? 0 : 1);
    
}

////////////////////////////////
///                          ///                            
///         Tests            ///
///                          ///
////////////////////////////////


void TConstrLHSTest::test1constr4var()
{
    _sampler = new URANIE::Sampler::TConstrLHS(_tds, _nS);
    _sampler->setSeed(_nseed);
    vector<int> inputs = {1,0};
    _sampler->addConstraint(OneConstraint, 1, inputs.size(), &inputs[0]);    
    _sampler->generateSample();

    _tds->exportData("OneConstrLHS.dat");
    
    Int_t nret = UranieTests::areFilesEqual(TString("OneConstrLHS.dat.ref"),
                                            TString("OneConstrLHS.dat"), TString("#DATE:"));

    CPPUNIT_ASSERT_EQUAL(nret, 0);
    UranieTests::removeFile(_workingTestPath.Data(),"OneConstrLHS.dat");

    _tds->addAttribute("xdiff","x_1 - x_0");
    _tds->computeStatistic("xdiff");
    
    // Tests
    CPPUNIT_ASSERT(_tds->getAttribute("xdiff")->getMinimum()>0);
}

void TConstrLHSTest::test1constr4varInvert()
{
    _sampler = new URANIE::Sampler::TConstrLHS(_tds, _nS);
    _sampler->setSeed(_nseed);
    vector<int> inputs = {0,1};
    _sampler->addConstraint(OneConstraintInvert, 1, inputs.size(), &inputs[0]);    
    _sampler->generateSample();

    _tds->addAttribute("xdiff","x_1 - x_0");
    _tds->computeStatistic("xdiff");

    // Tests
    CPPUNIT_ASSERT(_tds->getAttribute("xdiff")->getMinimum()>0);

}

void TConstrLHSTest::test1constr4varNoSeed()
{
    _sampler = new URANIE::Sampler::TConstrLHS(_tds, _nS);
    vector<int> inputs = {1,0};
    _sampler->addConstraint(OneConstraint, 1, inputs.size(), &inputs[0]);    
    _sampler->generateSample();

    _tds->addAttribute("xdiff","x_1 - x_0");
    _tds->computeStatistic("xdiff");

    // Tests
    CPPUNIT_ASSERT(_tds->getAttribute("xdiff")->getMinimum()>0);
}

void TConstrLHSTest::test1constr4varFilledTDS()
{
    URANIE::Sampler::TSampling *sam = new URANIE::Sampler::TSampling(_tds, "lhs", _nS);
    sam->setSeed(_nseed);
    sam->generateSample();    
    delete sam;
    
    _sampler = new URANIE::Sampler::TConstrLHS(_tds, _nS);
    vector<int> inputs = {1,0};
    _sampler->addConstraint(OneConstraint, 1, inputs.size(), &inputs[0]);    
    _sampler->generateSample();

    _tds->addAttribute("xdiff","x_1 - x_0");
    _tds->computeStatistic("xdiff");

    // Tests
    CPPUNIT_ASSERT(_tds->getAttribute("xdiff")->getMinimum()>0);

}

/// 2 CONSTRAINTS

void TConstrLHSTest::test2constr4var()
{
    _sampler = new URANIE::Sampler::TConstrLHS(_tds, _nS);
    _sampler->setSeed(_nseed);
    vector<int> inputs = {1,0,3,2};
    _sampler->addConstraint(TwoConstraints, 2, inputs.size(), &inputs[0]);    
    _sampler->generateSample();

    _tds->exportData("TwoConstrLHS.dat");
    
    Int_t nret = UranieTests::areFilesEqual(TString("TwoConstrLHS.dat.ref"),
                                            TString("TwoConstrLHS.dat"), TString("#DATE:"));

    CPPUNIT_ASSERT_EQUAL(nret, 0);
    UranieTests::removeFile(_workingTestPath.Data(),"TwoConstrLHS.dat");

    _tds->addAttribute("xdiff1","x_1 - x_0");
    _tds->addAttribute("xdiff2","x_3 - x_2");
    _tds->computeStatistic("xdiff1:xdiff2");

    // Tests
    CPPUNIT_ASSERT(_tds->getAttribute("xdiff1")->getMinimum()>0);
    CPPUNIT_ASSERT(_tds->getAttribute("xdiff2")->getMinimum()>0);

}

void TConstrLHSTest::test2constr4varInvert()
{
    _sampler = new URANIE::Sampler::TConstrLHS(_tds, _nS);
    _sampler->setSeed(_nseed);
    vector<int> inputs = {0,1,2,3};
    _sampler->addConstraint(TwoConstraintsInvert, 2, inputs.size(), &inputs[0]);    
    _sampler->generateSample();

    _tds->addAttribute("xdiff1","x_1 - x_0");
    _tds->addAttribute("xdiff2","x_3 - x_2");
    _tds->computeStatistic("xdiff1:xdiff2");

    // Tests
    CPPUNIT_ASSERT(_tds->getAttribute("xdiff1")->getMinimum()>0);
    CPPUNIT_ASSERT(_tds->getAttribute("xdiff2")->getMinimum()>0);
}

void TConstrLHSTest::test2constr4varNoSeed()
{
    _sampler = new URANIE::Sampler::TConstrLHS(_tds, _nS);
    vector<int> inputs = {1,0,3,2};
    _sampler->addConstraint(TwoConstraints, 2 , inputs.size(), &inputs[0]);    
    _sampler->generateSample();

    _tds->addAttribute("xdiff1","x_1 - x_0");
    _tds->addAttribute("xdiff2","x_3 - x_2");
    _tds->computeStatistic("xdiff1:xdiff2");

    // Tests
    CPPUNIT_ASSERT(_tds->getAttribute("xdiff1")->getMinimum()>0);
    CPPUNIT_ASSERT(_tds->getAttribute("xdiff2")->getMinimum()>0);

}

void TConstrLHSTest::test2constr4varFilledTDS()
{
    URANIE::Sampler::TSampling *sam = new URANIE::Sampler::TSampling(_tds, "lhs", _nS);
    sam->setSeed(_nseed);
    sam->generateSample();
    sam->generateSample();
    delete sam;

    _sampler = new URANIE::Sampler::TConstrLHS(_tds, _nS);
    vector<int> inputs = {1, 0, 3, 2};
    _sampler->addConstraint(TwoConstraints, 2, inputs.size(), &inputs[0]);    
    _sampler->generateSample();

    _tds->addAttribute("xdiff1","x_1 - x_0");
    _tds->addAttribute("xdiff2","x_3 - x_2");
    _tds->computeStatistic("xdiff1:xdiff2");

    // Tests
    CPPUNIT_ASSERT(_tds->getAttribute("xdiff1")->getMinimum()>0);
    CPPUNIT_ASSERT(_tds->getAttribute("xdiff2")->getMinimum()>0);
}


//!
void TConstrLHSTest::throwExceptionConstrutorTdsNull()
{
    URANIE::DataServer::TDataServer * tdsnull = NULL;
    CPPUNIT_ASSERT_THROW(new URANIE::Sampler::TConstrLHS(tdsnull, _nS),
                         URANIE::Exceptions::UErrorExceptions);
}

//!
void TConstrLHSTest::throwExceptionWrongNumberOfSample()
{
    CPPUNIT_ASSERT_THROW(new URANIE::Sampler::TConstrLHS(_tds, -1),
                         URANIE::Exceptions::UErrorExceptions);

    CPPUNIT_ASSERT_THROW(new URANIE::Sampler::TConstrLHS(_tds, 0),
                         URANIE::Exceptions::UErrorExceptions);
}

//!
void TConstrLHSTest::throwExceptionWrongNumberOfParameter()
{

    _sampler = new URANIE::Sampler::TConstrLHS(_tds, 100);
    vector<int> inputs = {1,0,2};
    CPPUNIT_ASSERT_THROW(_sampler->addConstraint(OneConstraint, 1, inputs.size(), &inputs[0]),
                         URANIE::Exceptions::UErrorExceptions);

    vector<int> inputs_2 = {1};
    CPPUNIT_ASSERT_THROW(_sampler->addConstraint(OneConstraint, 1, inputs_2.size(), &inputs_2[0]),
                         URANIE::Exceptions::UErrorExceptions);

}

//!
void TConstrLHSTest::throwExceptionWrongParameterValue()
{

    _sampler = new URANIE::Sampler::TConstrLHS(_tds, 100);
    vector<int> inputs = {6,0};
    CPPUNIT_ASSERT_THROW(_sampler->addConstraint(OneConstraint, 1, inputs.size(), &inputs[0]),
                         URANIE::Exceptions::UErrorExceptions);

    vector<int> inputs_2 = {0,-1};
    CPPUNIT_ASSERT_THROW(_sampler->addConstraint(OneConstraint, 1, inputs_2.size(), &inputs_2[0]),
                         URANIE::Exceptions::UErrorExceptions);
    
}

//!
void TConstrLHSTest::throwExceptionSameVariableChangedInMultiConstraints()
{
    _sampler = new URANIE::Sampler::TConstrLHS(_tds, 100);
    vector<int> inputs = {1,0,2,0};
    CPPUNIT_ASSERT_THROW(_sampler->addConstraint(TwoConstraints, 2, inputs.size(), &inputs[0]),
                         URANIE::Exceptions::UErrorExceptions);

}

//!
void TConstrLHSTest::throwExceptionNoConstrainProvided()
{
    _sampler = new URANIE::Sampler::TConstrLHS(_tds, 100);
    CPPUNIT_ASSERT_THROW(_sampler->generateSample(),
                         URANIE::Exceptions::UErrorExceptions);
}


//!
void TConstrLHSTest::throwExceptionProvidedSampleImpossibleConstraint()
{
    int ns=100;
    URANIE::Sampler::TSampling *sam = new URANIE::Sampler::TSampling(_tds, "lhs", ns);
    sam->setSeed(_nseed);
    sam->generateSample();
    delete sam;
    
    _sampler = new URANIE::Sampler::TConstrLHS(_tds, ns);
    vector<int> inputs = {0,3};
    _sampler->addConstraint(OneConstraint, 1, inputs.size(), &inputs[0]);
    CPPUNIT_ASSERT_THROW(_sampler->generateSample(),
                         URANIE::Exceptions::UErrorExceptions);
}

//!
void TConstrLHSTest::throwExceptionImpossibleConstraint()
{
    _sampler = new URANIE::Sampler::TConstrLHS(_tds, 100);
    vector<int> inputs = {0,3};
    _sampler->addConstraint(OneConstraint, 1, inputs.size(), &inputs[0]);
    CPPUNIT_ASSERT_THROW(_sampler->generateSample(),
                         URANIE::Exceptions::UErrorExceptions);
}

//!
void TConstrLHSTest::throwExceptionFilledTDSWrongNS()
{
    URANIE::Sampler::TSampling *sam = new URANIE::Sampler::TSampling(_tds, "lhs", _nS);
    sam->generateSample();
    delete sam;
    
    CPPUNIT_ASSERT_THROW(_sampler = new URANIE::Sampler::TConstrLHS(_tds, _nS+2),
                         URANIE::Exceptions::UErrorExceptions);

}

