//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TestTCopula.cxx
 \author Maud Salvoldi (OCC/Areva group)
 \date   $Date$
 
 \brief  Test File of URANIE::Sampler::TCopula class
 
 Cr�e le : mer nov 02 17:00:27 2011
 
 */

// Uranie
#include "TestTCopula.h"
#include "TAttribute.h"
#include "TAttributeFormula.h"
#include "TUniformDistribution.h"
#include "TNormalDistribution.h"
#include "TWeibullDistribution.h"
#include "TUniformByPartsDistribution.h"
#include "TBetaDistribution.h"
#include "TCauchyDistribution.h"
#include "TExponentialDistribution.h"
#include "TTriangularDistribution.h"
#include "UDisableTest.h"

#include "TClaytonCopula.h"
#include "TFrankCopula.h"
#include "TPlackettCopula.h"
#include "TAMHCopula.h"
#include "TEllipticalCopula.h"
#include "TGaussianCopula.h"

// ROOT
#include "TString.h"

#include <iostream>
#include <math.h>

#define DBL_PRECISION_COMPARE 1.e-10
#define EPSILON 0.05

using namespace std;

TCopulaTest::TCopulaTest() :
        _tds(NULL), _tcop(NULL), _nSize(500)
{
}

TCopulaTest::~TCopulaTest()
{
}
void TCopulaTest::setUp()
{
    // Create a TDataServer
    _tds = new URANIE::DataServer::TDataServer();
    // Fill the DataServer with the three attributes of the study
    _tds->addAttribute(new URANIE::DataServer::TUniformDistribution("y1"));
    _tds->addAttribute(new URANIE::DataServer::TUniformDistribution("y2"));
    _tds->getAttribute("y1")->delShare();
    _tds->getAttribute("y2")->delShare();
	
    _tcop = NULL;
}

void TCopulaTest::tearDown()
{
    if (_tcop)
    {
        delete _tcop;
        _tcop = NULL;
    }
    delete _tds;
    _tds = NULL;
}
void TCopulaTest::testTCopulaClayton()
{
    _tcop = new URANIE::Sampler::TClaytonCopula(_tds, 10.0, _nSize);
    _tcop->generateSample();

    //   _tcop->setLog();
    _tds->computeStatistic();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.51, _tds->getAttribute("y1")->getMean(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.288, _tds->getAttribute("y1")->getStd(),
            EPSILON);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.51, _tds->getAttribute("y2")->getMean(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.288, _tds->getAttribute("y2")->getStd(),
            EPSILON);

}
void TCopulaTest::throwExceptionTCopulaClaytonThetaNegatif()
{
    CPPUNIT_ASSERT_THROW(
            new URANIE::Sampler::TClaytonCopula(_tds, -2.0, _nSize),
            URANIE::Exceptions::UErrorExceptions);
}
void TCopulaTest::testTCopulaFrank()
{
    _tcop = new URANIE::Sampler::TFrankCopula(_tds, 15.5, _nSize);
    _tcop->generateSample();

    //   _tcop->setLog();
    _tds->computeStatistic();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.51, _tds->getAttribute("y1")->getMean(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.288, _tds->getAttribute("y1")->getStd(),
            EPSILON);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.51, _tds->getAttribute("y2")->getMean(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.288, _tds->getAttribute("y2")->getStd(),
            EPSILON);
}
void TCopulaTest::testTCopulaPlackett()
{
    _tcop = new URANIE::Sampler::TPlackettCopula(_tds, 20.0, _nSize);
    _tcop->generateSample();

//   _tcop->setLog();
    _tds->computeStatistic();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.51, _tds->getAttribute("y1")->getMean(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.288, _tds->getAttribute("y1")->getStd(),
            EPSILON);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.51, _tds->getAttribute("y2")->getMean(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.288, _tds->getAttribute("y2")->getStd(),
            EPSILON);

}

void TCopulaTest::testTCopulaPlackettThetaUn()
{
    _tcop = new URANIE::Sampler::TPlackettCopula(_tds, 1.0, _nSize);
    _tcop->generateSample();

    //   _tcop->setLog();
    _tds->computeStatistic();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.51, _tds->getAttribute("y1")->getMean(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.288, _tds->getAttribute("y1")->getStd(),
            EPSILON);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.51, _tds->getAttribute("y2")->getMean(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.288, _tds->getAttribute("y2")->getStd(),
            EPSILON);
}

void TCopulaTest::testTCopulaElliptical()
{

    Double_t corr = 0.0;

    URANIE::Sampler::TEllipticalCopula *tcopE = new URANIE::Sampler::TGaussianCopula(_tds, "", _nSize);

    //correlation with attribute number
    corr = 0.8;
    tcopE->setUserCorrelation(0, 1, corr);
    TMatrixD mat1 = tcopE->getCorrelationMatrix();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(corr, mat1(0,1),DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(corr, mat1(1,0),DBL_PRECISION_COMPARE);

    //correlation with attribute name
    corr = 0.7;
    tcopE->setUserCorrelation("y1", "y2", corr);
    TMatrixD mat2 = tcopE->getCorrelationMatrix();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(corr, mat2(0,1),DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(corr, mat2(1,0),DBL_PRECISION_COMPARE);

    //correlation with attribute object
    corr = 0.6;
    tcopE->setUserCorrelation(_tds->getAttribute("y1"), _tds->getAttribute("y2"), corr);
    TMatrixD mat3 = tcopE->getCorrelationMatrix();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(corr, mat3(0,1),DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(corr, mat3(1,0),DBL_PRECISION_COMPARE);

    tcopE->generateSample();

    _tds->computeStatistic();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.51, _tds->getAttribute("y1")->getMean(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.288, _tds->getAttribute("y1")->getStd(),
            EPSILON);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.51, _tds->getAttribute("y2")->getMean(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.288, _tds->getAttribute("y2")->getStd(),
            EPSILON);

    delete tcopE;
    tcopE = NULL;
}
void TCopulaTest::testTCopulaGaussian()
{
    URANIE::DataServer::TDataServer *tdsG = new URANIE::DataServer::TDataServer();
    tdsG->addAttribute(new URANIE::DataServer::TNormalDistribution("Normal", 0., 1));
    tdsG->addAttribute(new URANIE::DataServer::TUniformDistribution("Uniform", 0., 1.));
    tdsG->addAttribute(new URANIE::DataServer::TWeibullDistribution("Weibull", 2., 1.));
    tdsG->addAttribute(new URANIE::DataServer::TTriangularDistribution("Triang", 5.0, 8., 6.0));
    tdsG->getAttribute("Normal")->delShare();
    tdsG->getAttribute("Uniform")->delShare();
    tdsG->getAttribute("Weibull")->delShare();
    tdsG->getAttribute("Triang")->delShare();
    
    URANIE::Sampler::TGaussianCopula *tcop = new URANIE::Sampler::TGaussianCopula(tdsG, "", 5000);

    tcop->generateSample();

    tdsG->computeStatistic();
    TMatrixD matCorrRk = tdsG->computeCorrelationMatrix();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, fabs(tdsG->getAttribute("Normal")->getMean()),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsG->getAttribute("Normal")->getStd(),
            EPSILON);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5, tdsG->getAttribute("Uniform")->getMean(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.288, tdsG->getAttribute("Uniform")->getStd(),
            EPSILON);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsG->getAttribute("Weibull")->getMean(),
            0.1);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(6.34, tdsG->getAttribute("Triang")->getMean(),
            0.08);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.62, tdsG->getAttribute("Triang")->getStd(),
            EPSILON);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(matCorrRk(0, 0), 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matCorrRk(1, 1), 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matCorrRk(2, 2), 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matCorrRk(3, 3), 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(fabs(matCorrRk(0, 1)), 0.0, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(fabs(matCorrRk(0, 2)), 0.0, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(fabs(matCorrRk(0, 3)), 0.0, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(fabs(matCorrRk(1, 2)), 0.0, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(fabs(matCorrRk(1, 3)), 0.0, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(fabs(matCorrRk(2, 3)), 0.0, 0.1);

    delete tcop;
    tcop = NULL;
    delete tdsG;

}
void TCopulaTest::testTCopulaGaussianWithCorrelation()
{

    Double_t dcorr = 0.8;
    URANIE::DataServer::TDataServer *tdsG = new URANIE::DataServer::TDataServer();
    tdsG->addAttribute(new URANIE::DataServer::TNormalDistribution("Normal", 0., 1));
    tdsG->addAttribute(new URANIE::DataServer::TUniformDistribution("Uniform", 0., 1.));
    tdsG->addAttribute(new URANIE::DataServer::TWeibullDistribution("Weibull", 2., 1.));
    tdsG->addAttribute(new URANIE::DataServer::TTriangularDistribution("Triang", 5.0, 8., 6.0));
  tdsG->getAttribute("Normal")->delShare();
    tdsG->getAttribute("Uniform")->delShare();
    tdsG->getAttribute("Weibull")->delShare();
    tdsG->getAttribute("Triang")->delShare();
    
    URANIE::Sampler::TGaussianCopula *tcop = new URANIE::Sampler::TGaussianCopula(tdsG, "", 5000);

    tcop->setUserCorrelation(0, 1, 0.8);
    tcop->generateSample();

    tdsG->computeStatistic();
    TMatrixD matCorrRk = tdsG->computeCorrelationMatrix();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, fabs(tdsG->getAttribute("Normal")->getMean()),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsG->getAttribute("Normal")->getStd(),
            EPSILON);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5, tdsG->getAttribute("Uniform")->getMean(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.288, tdsG->getAttribute("Uniform")->getStd(),
            EPSILON);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsG->getAttribute("Weibull")->getMean(),
            0.1);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(6.34, tdsG->getAttribute("Triang")->getMean(),
            0.08);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.62, tdsG->getAttribute("Triang")->getStd(),
            EPSILON);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(matCorrRk(0, 0), 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matCorrRk(1, 1), 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matCorrRk(2, 2), 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matCorrRk(3, 3), 1.0, DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(matCorrRk(0, 1), dcorr, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(fabs(matCorrRk(0, 2)), 0.0, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(fabs(matCorrRk(0, 3)), 0.0, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(fabs(matCorrRk(1, 2)), 0.0, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(fabs(matCorrRk(1, 3)), 0.0, 0.1);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(fabs(matCorrRk(2, 3)), 0.0, 0.1);

    delete tcop;
    tcop = NULL;
    delete tdsG;

}

void TCopulaTest::testTCopulaAliMikhailHaq()
{
    _tcop = new URANIE::Sampler::TAMHCopula(_tds, -1.0, _nSize);
    _tcop->generateSample();

//   _tcop->setLog();
    _tds->computeStatistic();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.51, _tds->getAttribute("y1")->getMean(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.288, _tds->getAttribute("y1")->getStd(),
            EPSILON);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.51, _tds->getAttribute("y2")->getMean(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.288, _tds->getAttribute("y2")->getStd(),
            EPSILON);
}

void TCopulaTest::testTCopulaGaussianTAttribut()
{
    int nS = 5000;

    URANIE::DataServer::TDataServer *tdsG = new URANIE::DataServer::TDataServer();
    tdsG->addAttribute(new URANIE::DataServer::TNormalDistribution("x1", 0., 1));
    tdsG->addAttribute(new URANIE::DataServer::TNormalDistribution("x2", 0., 1));
    tdsG->addAttribute(new URANIE::DataServer::TAttributeFormula("z","x1+x2"));
    tdsG->getAttribute("x1")->delShare();
    tdsG->getAttribute("x2")->delShare();
    tdsG->getAttribute("z")->delShare();

    URANIE::Sampler::TGaussianCopula *tcop = new URANIE::Sampler::TGaussianCopula(tdsG, "", nS);

    CPPUNIT_ASSERT_EQUAL(2, tcop->_nX);

    tcop->generateSample();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsG->getValue("x1", 0)+tdsG->getValue("x2", 0), 
				 tdsG->getValue("z", 0),
				 DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsG->getValue("x1", 1)+tdsG->getValue("x2", 1), 
				 tdsG->getValue("z", 1),
				 DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tdsG->getValue("x1", nS-1)+tdsG->getValue("x2", nS-1), 
				 tdsG->getValue("z", nS-1),
				 DBL_PRECISION_COMPARE);

    tdsG->computeStatistic("x1");

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, fabs(tdsG->getAttribute("x1")->getMean()),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsG->getAttribute("x1")->getStd(),
            EPSILON);

    delete tcop;
    tcop = NULL;
    delete tdsG;

}

////////////////////////////////
///                          ///                            
///       Exceptions         ///
///                          ///
////////////////////////////////

void TCopulaTest::throwExceptionTCopulaPlackettThetaNegatif()
{
    CPPUNIT_ASSERT_THROW(
            new URANIE::Sampler::TPlackettCopula(_tds, -2.0, _nSize),
            URANIE::Exceptions::UErrorExceptions);
}

void TCopulaTest::throwExceptionTCopulaAliMikhailHaqThetaSupUn()
{
    CPPUNIT_ASSERT_THROW(new URANIE::Sampler::TAMHCopula(_tds, 2.0, _nSize),
            URANIE::Exceptions::UErrorExceptions);
}

void TCopulaTest::throwExceptionTCopulaAliMikhailHaqThetaInfMoinsUn()
{
    CPPUNIT_ASSERT_THROW(new URANIE::Sampler::TAMHCopula(_tds, -2.0, _nSize),
            URANIE::Exceptions::UErrorExceptions);
}

void TCopulaTest::throwExceptionTCopulaConstructorTdsNull()
{
    CPPUNIT_ASSERT_THROW(
            new URANIE::Sampler::TClaytonCopula(NULL, 10.0, _nSize),
            URANIE::Exceptions::UErrorExceptions);
}

void TCopulaTest::throwExceptionTEllipticalCopulasetUserCorrelationAttributeValueFalse()
{
    URANIE::Sampler::TEllipticalCopula *tcopE = new URANIE::Sampler::TGaussianCopula(_tds, "", _nSize);

    CPPUNIT_ASSERT_THROW(
            tcopE->setUserCorrelation(_tds->getAttribute("y1"),
                    _tds->getAttribute("y2"), 1.95),
            URANIE::Exceptions::UErrorExceptions);

    CPPUNIT_ASSERT_THROW(
            tcopE->setUserCorrelation("y1", "y2", 1.95),
            URANIE::Exceptions::UErrorExceptions);

    CPPUNIT_ASSERT_THROW(
            tcopE->setUserCorrelation(0, 1, 1.95),
            URANIE::Exceptions::UErrorExceptions);

    CPPUNIT_ASSERT_THROW(
            tcopE->setUserCorrelation(_tds->getAttribute("y1"),
                    _tds->getAttribute("y2"), -1.1),
            URANIE::Exceptions::UErrorExceptions);

    CPPUNIT_ASSERT_THROW(
            tcopE->setUserCorrelation("y1", "y2", -1.1),
            URANIE::Exceptions::UErrorExceptions);

    CPPUNIT_ASSERT_THROW(
            tcopE->setUserCorrelation(0, 1, -1.1),
            URANIE::Exceptions::UErrorExceptions);

    delete tcopE;
    tcopE = NULL;
}

void TCopulaTest::throwExceptionTEllipticalCopulasetUserCorrelationSameIndex()
{
    URANIE::Sampler::TEllipticalCopula *tcopE = new URANIE::Sampler::TGaussianCopula(_tds, "", _nSize);

    CPPUNIT_ASSERT_THROW(tcopE->setUserCorrelation(0, 0, 0.95),
            URANIE::Exceptions::UErrorExceptions);

    delete tcopE;
    tcopE = NULL;
}

void TCopulaTest::throwExceptionTEllipticalCopulasetUserCorrelationIndexNegative()
{
    URANIE::Sampler::TEllipticalCopula *tcopE = new URANIE::Sampler::TGaussianCopula(_tds, "", _nSize);

    CPPUNIT_ASSERT_THROW(tcopE->setUserCorrelation(0, -1, 0.95),
            URANIE::Exceptions::UErrorExceptions);

    CPPUNIT_ASSERT_THROW(tcopE->setUserCorrelation(-1, 0, 0.95),
            URANIE::Exceptions::UErrorExceptions);

    delete tcopE;
    tcopE = NULL;
}

void TCopulaTest::throwExceptionTEllipticalCopulasetUserCorrelationIndexSup()
{
    URANIE::Sampler::TEllipticalCopula *tcopE = new URANIE::Sampler::TGaussianCopula(_tds, "", _nSize);

    CPPUNIT_ASSERT_THROW(tcopE->setUserCorrelation(0, 2, 0.95),
            URANIE::Exceptions::UErrorExceptions);

    CPPUNIT_ASSERT_THROW(tcopE->setUserCorrelation(2, 0, 0.95),
            URANIE::Exceptions::UErrorExceptions);

    delete tcopE;
    tcopE = NULL;
}
void TCopulaTest::throwExceptionTCopulaGaussianCorrelationNegative()
{

    URANIE::DataServer::TDataServer *tdsG = new URANIE::DataServer::TDataServer();
    tdsG->addAttribute(new URANIE::DataServer::TNormalDistribution("Normal", 0., 1));
    tdsG->addAttribute(new URANIE::DataServer::TUniformDistribution("Uniform", 0., 1.));
    tdsG->addAttribute(new URANIE::DataServer::TWeibullDistribution("Weibull", 2., 1.));
    tdsG->addAttribute(new URANIE::DataServer::TTriangularDistribution("Triang", 5.0, 8., 6.0));
  tdsG->getAttribute("Normal")->delShare();
    tdsG->getAttribute("Uniform")->delShare();
    tdsG->getAttribute("Weibull")->delShare();
    tdsG->getAttribute("Triang")->delShare();
    
    URANIE::Sampler::TGaussianCopula *tcop = new URANIE::Sampler::TGaussianCopula(tdsG, "", 10);

    tcop->setUserCorrelation(0, 1, 0.9);
    tcop->setUserCorrelation(0, 2, 0.9);
    tcop->setUserCorrelation(1, 2, 0.1);

    CPPUNIT_ASSERT_THROW(tcop->generateSample(), URANIE::Exceptions::UErrorExceptions);

    delete tcop;
    tcop = NULL;
    delete tdsG;
}

