//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TestTMaxiMinLHS.cxx
 \author JB Blanchard
 \date   $Date$
 
 \brief  Test File of URANIE::Sampler::TSampler class
 
 
 */

// Uranie
#include "TestTMaxiMinLHS.h"
#include "TAttribute.h"
#include "TUniformDistribution.h"
#include "UDisableTest.h"
#include "uranie_commands.h"

// ROOT
#include "TString.h"

#include <iostream>
#include <math.h>

#define DBL_PRECISION_COMPARE 1.e-10
#define EPSILON 0.04

using namespace std;

TMaxiMinLHSTest::TMaxiMinLHSTest() :
         _tds(NULL), _sampler(NULL), _nS(30)
{
    _workingTestPath = gSystem->pwd();
}

TMaxiMinLHSTest::~TMaxiMinLHSTest()
{
    if (_sampler)
    {
        delete _sampler;
        _sampler = NULL;
    }

    if (_tds)
    {
        delete _tds;
        _tds = NULL;
    }

}

void TMaxiMinLHSTest::setUp()
{

    _tds = new URANIE::DataServer::TDataServer("foo","bar");
    Int_t dim=5;

    for (Int_t j=0;j<dim;j++)
    {
 	_tds->addAttribute(new URANIE::DataServer::TUniformDistribution(Form("x_%d",j),0.0,1.0));
	_tds->getAttribute( Form("x_%d",j) )->delShare();
    }

}

void TMaxiMinLHSTest::tearDown()
{
    delete _sampler;
    _sampler = NULL;

    delete _tds;
    _tds = NULL;

}

////////////////////////////////
///                          ///                            
///         Tests            ///
///                          ///
////////////////////////////////

void TMaxiMinLHSTest::testMaxiMinLHS()
{
    const Double_t coeff=0.98;
    const Double_t Tinit=0.010;
    const Int_t N=2000;
    const Int_t N_inner=300;
    
    _sampler = new URANIE::Sampler::TMaxiMinLHS(_tds, _nS, Tinit, coeff, N, N_inner);
    _sampler->setSeed(123456);
    _sampler->generateSample();

    _tds->exportData("FirstMaxiMinLHS.dat");
    
    Int_t nret = UranieTests::areFilesEqual(TString("FirstMaxiMinLHS.dat.ref"),
                                            TString("FirstMaxiMinLHS.dat"), TString("#DATE:"));

    CPPUNIT_ASSERT_EQUAL(nret, 0);
    UranieTests::removeFile(_workingTestPath.Data(),"FirstMaxiMinLHS.dat");
}

void TMaxiMinLHSTest::testMaxiMinLHSFilledTDS()
{
    
    _tds->fileDataRead("FirstMaxiMinLHS.dat.ref", false, true);
    
    const Double_t coeff=0.98;
    const Double_t Tinit=0.010;
    const Int_t N=2000;
    const Int_t N_inner=300; 

    _sampler = new URANIE::Sampler::TMaxiMinLHS(_tds, _nS, Tinit, coeff, N, N_inner);
    _sampler->setSeed(123456);
    _sampler->generateSample();

    _tds->exportData("ReRunFirstMaxiMinLHS.dat");
    
    Int_t nret = UranieTests::areFilesEqual(TString("ReRunFirstMaxiMinLHS.dat.ref"),
                                            TString("ReRunFirstMaxiMinLHS.dat"), TString("#DATE:"));
    
    CPPUNIT_ASSERT_EQUAL(nret, 0);
    UranieTests::removeFile(_workingTestPath.Data(),"ReRunFirstMaxiMinLHS.dat");
}

void TMaxiMinLHSTest::throwExceptionTMaxiMinLHSConstrutorTdsNull()
{
    URANIE::DataServer::TDataServer * tdsnull = NULL;
    CPPUNIT_ASSERT_THROW(new URANIE::Sampler::TMaxiMinLHS(tdsnull, _nS,1,0.1,1,1),
            URANIE::Exceptions::UErrorExceptions);
}

void TMaxiMinLHSTest::throwExceptionTMaxiMinLHSWrongTemperature()
{
    CPPUNIT_ASSERT_THROW(new URANIE::Sampler::TMaxiMinLHS( _tds, _nS, -3,0.1,1,1),
            URANIE::Exceptions::UErrorExceptions);
}

void TMaxiMinLHSTest::throwExceptionTMaxiMinLHSWrongTemperatureCoeff()
{
    CPPUNIT_ASSERT_THROW(new URANIE::Sampler::TMaxiMinLHS( _tds, _nS, 1,1,1,1),
            URANIE::Exceptions::UErrorExceptions);

    CPPUNIT_ASSERT_THROW(new URANIE::Sampler::TMaxiMinLHS( _tds, _nS, 1,-1,1,1),
            URANIE::Exceptions::UErrorExceptions);

}

void TMaxiMinLHSTest::throwExceptionTMaxiMinLHSWrongNIter()
{
    CPPUNIT_ASSERT_THROW(new URANIE::Sampler::TMaxiMinLHS( _tds, _nS, 3,0.1,-1,1),
            URANIE::Exceptions::UErrorExceptions);
}

void TMaxiMinLHSTest::throwExceptionTMaxiMinLHSWrongNinnerIter ()
{
    CPPUNIT_ASSERT_THROW(new URANIE::Sampler::TMaxiMinLHS( _tds, _nS, 3,0.1,1,-1),
            URANIE::Exceptions::UErrorExceptions);
}

void TMaxiMinLHSTest::throwExceptionTMaxiMinLHSFilledTDSWrongNS()
{
    delete _tds;
    _tds = new URANIE::DataServer::TDataServer("foo","bar");

    _tds->fileDataRead("FirstMaxiMinLHS.dat.ref");
    
    const Double_t coeff=0.98;
    const Double_t Tinit=0.010;
    const Int_t N=2000;
    const Int_t N_inner=300; 

    CPPUNIT_ASSERT_THROW(new URANIE::Sampler::TMaxiMinLHS(_tds, 200, Tinit, coeff, N, N_inner),
			 URANIE::Exceptions::UErrorExceptions);

}
