//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTOATSampling.cxx
 \author Cedric Forgues <uranietm@cea.fr>
 \date   $Date$

 \brief  Test File of URANIE::Sampler::TOATSampling class

 Cr�e le : mer nov 21 07:59:24 2007

 */

// Uranie
#include "TestTOATSampling.h"
#include "TOATSampling.h"
#include "TDataServer.h"
#include "UDisableTest.h"

// ROOT
#include "TString.h"
#include "TSystem.h"
#include "TMath.h"

#include <iostream>
#include <math.h>

#define DBL_PRECISION_COMPARE 1.e-12
#define EPSILON 0.01

using namespace std;

TOATSamplingTest::TOATSamplingTest()
{
}

TOATSamplingTest::~TOATSamplingTest()
{
}

void TOATSamplingTest::setUp()
{
    att1 = new URANIE::DataServer::TUniformDistribution("x1", 0.0, 2.0);
    att1->setDefaultValue(1.0);

    att2 = new URANIE::DataServer::TUniformDistribution("x2", 1.0, 3.0);
    att2->setDefaultValue(2.0);

    att3 = new URANIE::DataServer::TUniformDistribution("x3", 2.0, 4.0);
    att3->setDefaultValue(3.0);

    attN = new URANIE::DataServer::TNormalDistribution("xN", 4.0, 1.0);
    attN->setDefaultValue(4.0);

    attD = new URANIE::DataServer::TUniformDistribution("xD", 3.0, 5.0);
}

void TOATSamplingTest::tearDown()
{
    delete att1;
    att1 = NULL;
    delete att2;
    att2 = NULL;
    delete att3;
    att3 = NULL;
    delete attN;
    attN = NULL;
    delete attD;
    attD = NULL;
}

////////////////////////////////
///                          ///                            
///         Tests            ///
///                          ///
////////////////////////////////

void TOATSamplingTest::testOATSampling()
{
    URANIE::DataServer::TDataServer *tdsGood;
    URANIE::Sampler::TOATSampling *samp;

    //-----------------------------//
    // constructor without options //
    //-----------------------------//    

    tdsGood = new URANIE::DataServer::TDataServer("tdsGood",
            "TDS with no problems");
    tdsGood->addAttribute(att1);
    tdsGood->addAttribute(att2);
    tdsGood->addAttribute(att3);

    samp = new URANIE::Sampler::TOATSampling(tdsGood);

    CPPUNIT_ASSERT_EQUAL(TString("steps=1%"), samp->getSamplingMode());
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, samp->getStepSizeParameter(),
            DBL_PRECISION_COMPARE);

    samp->generateSample();

    CPPUNIT_ASSERT_EQUAL(7, tdsGood->getNPatterns());

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("x1", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsGood->getValue("x2", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tdsGood->getValue("x3", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__nominal_set__", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tdsGood->getValue("__modified_att__", 0),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.99, tdsGood->getValue("x1", 1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsGood->getValue("x2", 1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tdsGood->getValue("x3", 1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__nominal_set__", 1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__modified_att__", 1),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.01, tdsGood->getValue("x1", 2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsGood->getValue("x2", 2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tdsGood->getValue("x3", 2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__nominal_set__", 2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__modified_att__", 2),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("x1", 3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.98, tdsGood->getValue("x2", 3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tdsGood->getValue("x3", 3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__nominal_set__", 3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsGood->getValue("__modified_att__", 3),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("x1", 4),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.02, tdsGood->getValue("x2", 4),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tdsGood->getValue("x3", 4),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__nominal_set__", 4),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsGood->getValue("__modified_att__", 4),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("x1", 5),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsGood->getValue("x2", 5),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.97, tdsGood->getValue("x3", 5),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__nominal_set__", 5),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tdsGood->getValue("__modified_att__", 5),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("x1", 6),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsGood->getValue("x2", 6),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.03, tdsGood->getValue("x3", 6),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__nominal_set__", 6),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tdsGood->getValue("__modified_att__", 6),
            DBL_PRECISION_COMPARE);

    delete samp;
    samp = NULL;
    delete tdsGood;
    tdsGood = NULL;

    //----------------------------------//
    // constructor with "minmax" option //
    //----------------------------------//

    tdsGood = new URANIE::DataServer::TDataServer("tdsGood",
            "TDS with no problems");
    tdsGood->addAttribute(att1);
    tdsGood->addAttribute(att2);
    tdsGood->addAttribute(att3);

    samp = new URANIE::Sampler::TOATSampling(tdsGood, "minmax");

    CPPUNIT_ASSERT_EQUAL(TString("minmax"), samp->getSamplingMode());
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, samp->getStepSizeParameter(),
            DBL_PRECISION_COMPARE);

    samp->generateSample();

    CPPUNIT_ASSERT_EQUAL(7, tdsGood->getNPatterns());

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("x1", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsGood->getValue("x2", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tdsGood->getValue("x3", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__nominal_set__", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tdsGood->getValue("__modified_att__", 0),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tdsGood->getValue("x1", 1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsGood->getValue("x2", 1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tdsGood->getValue("x3", 1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__nominal_set__", 1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__modified_att__", 1),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsGood->getValue("x1", 2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsGood->getValue("x2", 2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tdsGood->getValue("x3", 2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__nominal_set__", 2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__modified_att__", 2),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("x1", 3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("x2", 3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tdsGood->getValue("x3", 3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__nominal_set__", 3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsGood->getValue("__modified_att__", 3),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("x1", 4),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tdsGood->getValue("x2", 4),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tdsGood->getValue("x3", 4),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__nominal_set__", 4),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsGood->getValue("__modified_att__", 4),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("x1", 5),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsGood->getValue("x2", 5),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsGood->getValue("x3", 5),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__nominal_set__", 5),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tdsGood->getValue("__modified_att__", 5),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("x1", 6),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsGood->getValue("x2", 6),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(4.0, tdsGood->getValue("x3", 6),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__nominal_set__", 6),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tdsGood->getValue("__modified_att__", 6),
            DBL_PRECISION_COMPARE);

    delete samp;
    samp = NULL;
    delete tdsGood;
    tdsGood = NULL;

    //-------------------------------------//
    // constructor with "steps=0.5" option //
    //-------------------------------------//

    tdsGood = new URANIE::DataServer::TDataServer("tdsGood",
            "TDS with no problems");
    tdsGood->addAttribute(att1);
    tdsGood->addAttribute(att2);
    tdsGood->addAttribute(att3);

    samp = new URANIE::Sampler::TOATSampling(tdsGood, "steps=0.5");

    CPPUNIT_ASSERT_EQUAL(TString("steps=0.5"), samp->getSamplingMode());
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5, samp->getStepSizeParameter(),
            DBL_PRECISION_COMPARE);

    samp->generateSample();

    CPPUNIT_ASSERT_EQUAL(7, tdsGood->getNPatterns());

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("x1", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsGood->getValue("x2", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tdsGood->getValue("x3", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__nominal_set__", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tdsGood->getValue("__modified_att__", 0),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5, tdsGood->getValue("x1", 1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsGood->getValue("x2", 1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tdsGood->getValue("x3", 1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__nominal_set__", 1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__modified_att__", 1),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.5, tdsGood->getValue("x1", 2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsGood->getValue("x2", 2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tdsGood->getValue("x3", 2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__nominal_set__", 2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__modified_att__", 2),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("x1", 3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.5, tdsGood->getValue("x2", 3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tdsGood->getValue("x3", 3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__nominal_set__", 3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsGood->getValue("__modified_att__", 3),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("x1", 4),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.5, tdsGood->getValue("x2", 4),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tdsGood->getValue("x3", 4),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__nominal_set__", 4),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsGood->getValue("__modified_att__", 4),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("x1", 5),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsGood->getValue("x2", 5),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.5, tdsGood->getValue("x3", 5),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__nominal_set__", 5),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tdsGood->getValue("__modified_att__", 5),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("x1", 6),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsGood->getValue("x2", 6),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.5, tdsGood->getValue("x3", 6),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__nominal_set__", 6),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tdsGood->getValue("__modified_att__", 6),
            DBL_PRECISION_COMPARE);

    delete samp;
    samp = NULL;
    delete tdsGood;
    tdsGood = NULL;

    //------------------------------//
    // Only two variable attributes //
    //------------------------------//

    tdsGood = new URANIE::DataServer::TDataServer("tdsGood",
            "TDS with no problems");
    tdsGood->addAttribute(att1);
    tdsGood->addAttribute(att2);
    tdsGood->addAttribute(att3);

    samp = new URANIE::Sampler::TOATSampling(tdsGood, "minmax");

    CPPUNIT_ASSERT_EQUAL(TString("minmax"), samp->getSamplingMode());
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, samp->getStepSizeParameter(),
            DBL_PRECISION_COMPARE);

    samp->generateSample("inputs=x2:x3");

    CPPUNIT_ASSERT_EQUAL(5, tdsGood->getNPatterns());

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("x1", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsGood->getValue("x2", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tdsGood->getValue("x3", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__nominal_set__", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tdsGood->getValue("__modified_att__", 0),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("x1", 1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("x2", 1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tdsGood->getValue("x3", 1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__nominal_set__", 1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsGood->getValue("__modified_att__", 1),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("x1", 2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tdsGood->getValue("x2", 2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tdsGood->getValue("x3", 2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__nominal_set__", 2),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsGood->getValue("__modified_att__", 2),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("x1", 3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsGood->getValue("x2", 3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsGood->getValue("x3", 3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__nominal_set__", 3),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tdsGood->getValue("__modified_att__", 3),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("x1", 4),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsGood->getValue("x2", 4),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(4.0, tdsGood->getValue("x3", 4),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsGood->getValue("__nominal_set__", 4),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tdsGood->getValue("__modified_att__", 4),
            DBL_PRECISION_COMPARE);

    delete samp;
    samp = NULL;
    delete tdsGood;
    tdsGood = NULL;

}


void TOATSamplingTest::testOATSamplingWithStepValue()
{
    URANIE::DataServer::TDataServer *tdsTmp =  new URANIE::DataServer::TDataServer("tdsTest2", "TDS with no problems");
    tdsTmp->addAttribute ( new URANIE::DataServer::TUniformDistribution("x1", -1.0 * TMath::Pi(), TMath::Pi()));
    tdsTmp->addAttribute ( new URANIE::DataServer::TUniformDistribution("x2", -1.0 * TMath::Pi(), TMath::Pi()));
    tdsTmp->addAttribute ( new URANIE::DataServer::TUniformDistribution("x3", -1.0 * TMath::Pi(), TMath::Pi()));

    tdsTmp->getAttribute("x1")->setDefaultValue(0.0);
    tdsTmp->getAttribute("x2")->setDefaultValue(0.0);
    tdsTmp->getAttribute("x3")->setDefaultValue(0.0);

    tdsTmp->getAttribute("x1")->delShare();
    tdsTmp->getAttribute("x2")->delShare();
    tdsTmp->getAttribute("x3")->delShare();

    //tdsTmp->getAttribute("x1")->setStepValue(0.01);
    tdsTmp->getAttribute("x2")->setStepValue(0.03);
    //tdsTmp->getAttribute("x3")->setStepValue(0.08);

    URANIE::Sampler::TOATSampling *samp =  new URANIE::Sampler::TOATSampling(tdsTmp, "steps=2%");
    CPPUNIT_ASSERT_EQUAL(TString("steps=2%"), samp->getSamplingMode());
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, samp->getStepSizeParameter(), DBL_PRECISION_COMPARE);

    Double_t dDefaultPrecision = samp->getStepPrecision();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0e-16, dDefaultPrecision, DBL_PRECISION_COMPARE);

    samp->setStepPrecision(1.0e-5);
    dDefaultPrecision = samp->getStepPrecision();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0e-5, dDefaultPrecision, DBL_PRECISION_COMPARE);

    samp->generateSample();
    tdsTmp->scan("*");

    Int_t nX = 3;
    CPPUNIT_ASSERT( tdsTmp->isAttribute ("__nominal_set__"));
    CPPUNIT_ASSERT( tdsTmp->isAttribute ("__modified_att__"));
    CPPUNIT_ASSERT_EQUAL(3+2, tdsTmp->getNAttributes());

    Int_t npat = 2*nX+1;
    CPPUNIT_ASSERT_EQUAL(npat, tdsTmp->getNPatterns());

    // Nominal value
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tdsTmp->getValue("x1", 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tdsTmp->getValue("x2", 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tdsTmp->getValue("x3", 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsTmp->getValue("__nominal_set__", 0), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tdsTmp->getValue("__modified_att__", 0), DBL_PRECISION_COMPARE);

    // 2 and 3 values for x1
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tdsTmp->getValue("x1", 1), dDefaultPrecision);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tdsTmp->getValue("x2", 1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tdsTmp->getValue("x3", 1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsTmp->getValue("__nominal_set__", 1), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsTmp->getValue("__modified_att__", 1), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tdsTmp->getValue("x1", 2), dDefaultPrecision);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tdsTmp->getValue("x2", 2), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tdsTmp->getValue("x3", 2), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsTmp->getValue("__nominal_set__", 2), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsTmp->getValue("__modified_att__", 2), DBL_PRECISION_COMPARE);

    // 4 and 5 values for x2
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tdsTmp->getValue("x1", 3), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(-1.0*0.03, tdsTmp->getValue("x2", 3), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tdsTmp->getValue("x3", 3), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsTmp->getValue("__nominal_set__", 3), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsTmp->getValue("__modified_att__", 3), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tdsTmp->getValue("x1", 4), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0*0.03, tdsTmp->getValue("x2", 4), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tdsTmp->getValue("x3", 4), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsTmp->getValue("__nominal_set__", 4), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsTmp->getValue("__modified_att__", 4), DBL_PRECISION_COMPARE);

    // 6 and 7 values for x3
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tdsTmp->getValue("x1", 5), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tdsTmp->getValue("x2", 5), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tdsTmp->getValue("x3", 5), dDefaultPrecision);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsTmp->getValue("__nominal_set__", 5), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tdsTmp->getValue("__modified_att__", 5), DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tdsTmp->getValue("x1", 6), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tdsTmp->getValue("x2", 6), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, tdsTmp->getValue("x3", 6), dDefaultPrecision);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsTmp->getValue("__nominal_set__", 6), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(3.0, tdsTmp->getValue("__modified_att__", 6), DBL_PRECISION_COMPARE);

    // Clean
    delete samp; samp = nullptr;
    delete tdsTmp; tdsTmp = nullptr;
}


void TOATSamplingTest::testTOATSamplingOptionFalse()
{
    URANIE::Sampler::TOATSampling *samp;
 
    URANIE::DataServer::TDataServer *tdsTtest2 =
      new URANIE::DataServer::TDataServer("tdsTest2",
					  "TDS with no problems");
    tdsTtest2->addAttribute(att2);

    samp =  new URANIE::Sampler::TOATSampling(tdsTtest2, "badoption");
    
    CPPUNIT_ASSERT_EQUAL(TString("steps=1%"), samp->getSamplingMode());
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, samp->getStepSizeParameter(), DBL_PRECISION_COMPARE);

    delete samp;
    samp = NULL;
    delete tdsTtest2;
}

void TOATSamplingTest::testGenerateSampleSizeAttributModified()
{

    // If a new attribut is added to the TDS after the sample creation. 
    // the number of attribut must updated when generating the sample 
    URANIE_INBUILDING_TEST();

    URANIE::DataServer::TDataServer *tdsModif =
            new URANIE::DataServer::TDataServer("tdsModif",
						"TDS modfied");
    tdsModif->addAttribute(att1);
    
    URANIE::Sampler::TOATSampling *samp = new URANIE::Sampler::TOATSampling(
            tdsModif, "minmax");

    CPPUNIT_ASSERT_EQUAL(1, samp->_nX);

    tdsModif->addAttribute(att2);

    samp->generateSample();

    CPPUNIT_ASSERT_EQUAL(2, samp->_nX);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, tdsModif->getValue("x1", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.0, tdsModif->getValue("x2", 0),
            DBL_PRECISION_COMPARE);

    delete samp;
    delete tdsModif;
}


////////////////////////////////
///                          ///                            
///       Exceptions         ///
///                          ///
////////////////////////////////

void TOATSamplingTest::throwExceptionWhenNoDefaultValue()
{
    URANIE::DataServer::TDataServer *tdsNoDefault =
            new URANIE::DataServer::TDataServer("tdsNoDefault",
                    "TDS where one attribute has no default value");
    tdsNoDefault->addAttribute(att1);
    tdsNoDefault->addAttribute(att2);
    tdsNoDefault->addAttribute(att3);
    tdsNoDefault->addAttribute(attD);

    CPPUNIT_ASSERT_THROW( new URANIE::Sampler::TOATSampling(tdsNoDefault),
			  URANIE::Exceptions::UErrorExceptions); 

    delete tdsNoDefault;
}

void TOATSamplingTest::throwExceptionWhenNoBounds()
{
    URANIE::DataServer::TDataServer *tdsNoBounds =
            new URANIE::DataServer::TDataServer("tdsNoBounds",
                    "TDS where one attribute is unbounded");
    tdsNoBounds->addAttribute(att1);
    tdsNoBounds->addAttribute(att2);
    tdsNoBounds->addAttribute(att3);
    tdsNoBounds->addAttribute(attN);

    CPPUNIT_ASSERT_THROW( new URANIE::Sampler::TOATSampling(tdsNoBounds, "minmax"),
			  URANIE::Exceptions::UErrorExceptions); 

    delete tdsNoBounds;
}


