//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TestTQMC.cxx
 \author Maud Salvoldi (OCC/Areva group)
 \date   $Date$
 
 \brief  Test File of URANIE::Sampler::TQMC class
 
 Cr�e le : mer nov 02 17:00:24 2011
 
 */

// Uranie
#include "TestTQMC.h"
#include "TAttribute.h"
#include "TUniformDistribution.h"
#include "TNormalDistribution.h"
#include "TWeibullDistribution.h"
#include "TUniformByPartsDistribution.h"
#include "TBetaDistribution.h"
#include "TCauchyDistribution.h"
#include "UDisableTest.h"

// ROOT
#include "TString.h"

#include <iostream>
#include <math.h>

#define DBL_PRECISION_COMPARE 1.e-10
#define EPSILON 0.04

using namespace std;

TQMCTest::TQMCTest()
{
}

TQMCTest::~TQMCTest()
{
}

void TQMCTest::setUp()
{
  // Create a TDataServer
  _tds1 = new URANIE::DataServer::TDataServer();
  // Fill the DataServer with the three attributes of the study
  _tds1->addAttribute(
		      new URANIE::DataServer::TUniformDistribution("Uniform", 0., 1.));
  _tds1->addAttribute(
		      new URANIE::DataServer::TNormalDistribution("Normal", 0., 1.));
  _tds1->addAttribute(
		      new URANIE::DataServer::TWeibullDistribution("Weibull", 2., 1.));
  _tds1->addAttribute(
		      new URANIE::DataServer::TUniformByPartsDistribution("UnifParts",
									  1.0, 4.0, 2.0));
  _tds1->addAttribute(
		      new URANIE::DataServer::TBetaDistribution("Beta", 1.0, 4.0));
  _tds1->addAttribute(
		      new URANIE::DataServer::TCauchyDistribution("Cauchy", 0.5, 2.0));

  //HelpCleaning
  _tds1->getAttribute("Uniform")->delShare();
  _tds1->getAttribute("Normal")->delShare();
  _tds1->getAttribute("Weibull")->delShare();
  _tds1->getAttribute("UnifParts")->delShare();
  _tds1->getAttribute("Beta")->delShare();
  _tds1->getAttribute("Cauchy")->delShare();
}

void TQMCTest::tearDown()
{
    delete _tds1;
    _tds1 = NULL;
}

////////////////////////////////
///                          ///                            
///         Tests            ///
///                          ///
////////////////////////////////

void TQMCTest::testQMC()
{

    // Generate the sampling from the TDataServer with a Sobol sequence
    URANIE::Sampler::TQMC *qmcSobol = new URANIE::Sampler::TQMC(_tds1, "sobol",
            10000);
//   qmcSobol->setLog();
    qmcSobol->generateSample();
    _tds1->computeStatistic("Uniform:Normal:UnifParts");

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5, _tds1->getAttribute("Uniform")->getMean(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.288,
            _tds1->getAttribute("Uniform")->getStd(), EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, _tds1->getAttribute("Normal")->getMean(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, _tds1->getAttribute("Normal")->getStd(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.250,
            _tds1->getAttribute("UnifParts")->getMean(), 0.001);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.8774,
            _tds1->getAttribute("UnifParts")->getStd(), 0.001);

    delete qmcSobol;
    qmcSobol = NULL;

    // Generate the sampling from the TDataServer with a Halton sequence
    URANIE::Sampler::TQMC *qmcHalton = new URANIE::Sampler::TQMC(_tds1,
            "halton", 10000);
    qmcHalton->unsetLog();
    _tds1->getAttribute("Normal")->setLowerBound(-3.);
    _tds1->getAttribute("Cauchy")->setLowerBound(-3.);
    qmcHalton->generateSample();
    _tds1->computeStatistic("Uniform:UnifParts");
    _tds1->computeStatistic("Normal", Form("%s__n__iter__ != 1",_tds1->GetName()));

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5, _tds1->getAttribute("Uniform")->getMean(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.288,
            _tds1->getAttribute("Uniform")->getStd(), EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.0, _tds1->getAttribute("Normal")->getMean(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, _tds1->getAttribute("Normal")->getStd(),
            EPSILON);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.250,
            _tds1->getAttribute("UnifParts")->getMean(), 0.001);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.878,
            _tds1->getAttribute("UnifParts")->getStd(), 0.01);

    delete qmcHalton;
    qmcHalton = NULL;
}

void TQMCTest::throwExceptionQMCConstructorTdsNull()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    // URANIE_IGNORED_TEST()
    // ;

    URANIE::DataServer::TDataServer * tdsnull = NULL;
    CPPUNIT_ASSERT_THROW(new URANIE::Sampler::TQMC(tdsnull, "halton", 1000),
            URANIE::Exceptions::UErrorExceptions);

}
void TQMCTest::throwExceptionQMCConstructorFalseOption()
{
    CPPUNIT_ASSERT_THROW(new URANIE::Sampler::TQMC(_tds1, "toto", 1000),
            URANIE::Exceptions::UErrorExceptions);
}

void TQMCTest::throwExceptionQMCHaltonUnbounded()
{
    URANIE::Sampler::TQMC pouet(_tds1, "halton", 1000);
    CPPUNIT_ASSERT_THROW(pouet.generateSample(),
                         URANIE::Exceptions::UErrorExceptions);
    
}

void TQMCTest::testQMCgenerateSampleOptionShift()
{
    URANIE::Sampler::TQMC *qmcSobol = new URANIE::Sampler::TQMC(_tds1, "sobol",
            100);
//   qmcSobol->setLog();
    qmcSobol->generateSample("shift");
    _tds1->exportData("_samplerQMCshift_.dat", "*");
    _tds1->computeStatistic("*");

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.4994,
            _tds1->getAttribute("Uniform")->getMean(), 3 * 0.006444);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.29, _tds1->getAttribute("Uniform")->getStd(),
            3 * 0.002341);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(2.250,
            _tds1->getAttribute("UnifParts")->getMean(), 3 * 0.01928);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.8814,
            _tds1->getAttribute("UnifParts")->getStd(), 3 * 0.009316);

    delete qmcSobol;
    qmcSobol = NULL;
}
