//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TestTSamplerField.cxx
 \author Maud Salvoldi (OCC/Areva group)
 \date   $Date$
 
 \brief  Test File of URANIE::Sampler::TCopula class
 
 Cr�e le : mer nov 02 17:00:27 2011
 
 */

// Uranie
#include "TestTSamplerField.h"
#include "TAttribute.h"
#include "TUniformDistribution.h"
#include "TNormalDistribution.h"
#include "TWeibullDistribution.h"
#include "TUniformByPartsDistribution.h"
#include "TBetaDistribution.h"
#include "TCauchyDistribution.h"
#include "UDisableTest.h"

// ROOT
#include "TString.h"

#include <iostream>
#include <math.h>

#define DBL_PRECISION_COMPARE 1.e-9
#define EPSILON 0.03

using namespace std;

TSamplerFieldTest::TSamplerFieldTest() :
        _tds(NULL), _splf(NULL)
{
}

TSamplerFieldTest::~TSamplerFieldTest()
{
}
void TSamplerFieldTest::setUp()
{
    // Create a TDataServer
    _tds = new URANIE::DataServer::TDataServer("TDSField", "Weight as Field");
    // Fill the DataServer with the three attributes of the study
    _tds->addAttribute("x", 1, 51);
    _tds->addAttribute("y", 1, 51);

}

void TSamplerFieldTest::tearDown()
{
    delete _splf;
    _splf = NULL;
    delete _tds;
    _tds = NULL;
}

void TSamplerFieldTest::testSamplerFieldConstructor()
{
    _splf = new URANIE::Sampler::TSamplerField(_tds, "Gauss");
//  _splf->setLog();

    CPPUNIT_ASSERT_EQUAL(400, _splf->GetRandomFunction());
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, _splf->GetVariance(),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_EQUAL(kTRUE, _splf->IsIsotropy());
    TString sname = "Spectral Method";
    CPPUNIT_ASSERT_EQUAL(0, strcmp(sname.Data(), _splf->getMethodName()));
    sname = "Gaussian";
    CPPUNIT_ASSERT_EQUAL(0, strcmp(sname.Data(), _splf->getVariogramName()));

}
void TSamplerFieldTest::throwExceptionTSamplerFieldConstrutorTdsNull()
{
    URANIE::DataServer::TDataServer * tdsnull = NULL;
    CPPUNIT_ASSERT_THROW(new URANIE::Sampler::TSamplerField(tdsnull, "Gauss"),
            URANIE::Exceptions::UErrorExceptions);
}

void TSamplerFieldTest::throwExceptionTSamplerFieldConstrutorTdsNullFalseNumberArgument()
{
    URANIE::DataServer::TDataServer * _tdsone =
            new URANIE::DataServer::TDataServer();
    _tdsone->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x", 0., 1.));

    _tdsone->getAttribute("x")->delShare();
    CPPUNIT_ASSERT_THROW(new URANIE::Sampler::TSamplerField(_tdsone, "Gauss"),
            URANIE::Exceptions::UErrorExceptions);

    _tds->addAttribute("x3", 0., 1.);
    _tds->addAttribute("x4", 0., 1.);
    _tds->addAttribute("x5", 0., 1.);
    _tds->addAttribute("x6", 0., 1.);
    CPPUNIT_ASSERT_THROW(new URANIE::Sampler::TSamplerField(_tds, "Gauss"),
            URANIE::Exceptions::UErrorExceptions);

    delete _tdsone;
    _tdsone=NULL;
}

void TSamplerFieldTest::testSamplerFieldcreateListOfAttributes()
{
    _splf = new URANIE::Sampler::TSamplerField(_tds, "Gauss");
//  _splf->setLog();
    _splf->createListOfAttributes();
    _splf->_lstOfAttributesToSample->Print();

// THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_INBUILDING_TEST()
    ;

}
void TSamplerFieldTest::testSamplerFieldsetMethod()
{
    _splf = new URANIE::Sampler::TSamplerField(_tds, "Gauss");
//  _splf->setLog();
    _splf->setMethod("spectral");
    TString sname = "Spectral Method";
    CPPUNIT_ASSERT_EQUAL(0, strcmp(sname.Data(), _splf->getMethodName()));

    _splf->setMethod("dilution");
    sname = "Dilution Method";
    CPPUNIT_ASSERT_EQUAL(0, strcmp(sname.Data(), _splf->getMethodName()));

    _splf->setMethod("tessellation");
    sname = "Tessellation Method";
    CPPUNIT_ASSERT_EQUAL(0, strcmp(sname.Data(), _splf->getMethodName()));

    _splf->setMethod("turningbands");
    sname = "Turning Bands Method";
    CPPUNIT_ASSERT_EQUAL(0, strcmp(sname.Data(), _splf->getMethodName()));
}
void TSamplerFieldTest::throwExceptionTSamplerFieldsetMethod()
{
    _splf = new URANIE::Sampler::TSamplerField(_tds, "Gauss");
    CPPUNIT_ASSERT_THROW(_splf->setMethod("toto"),
            URANIE::Exceptions::UErrorExceptions);
}
void TSamplerFieldTest::testSamplerFieldsetVariogram()
{
    _splf = new URANIE::Sampler::TSamplerField(_tds, "Gauss");
    _splf->setVariogram("spherical");
    TString sname = "Spherical";
    CPPUNIT_ASSERT_EQUAL(0, strcmp(sname.Data(), _splf->getVariogramName()));

    _splf->setVariogram("exponential");
    sname = "Exponential";
    CPPUNIT_ASSERT_EQUAL(0, strcmp(sname.Data(), _splf->getVariogramName()));

    _splf->setVariogram("stable");
    sname = "Stable";
    CPPUNIT_ASSERT_EQUAL(0, strcmp(sname.Data(), _splf->getVariogramName()));

    _splf->setVariogram("hyperbolic");
    sname = "Hyperbolic";
    CPPUNIT_ASSERT_EQUAL(0, strcmp(sname.Data(), _splf->getVariogramName()));

    _splf->setVariogram("gaussian");
    sname = "Gaussian";
    CPPUNIT_ASSERT_EQUAL(0, strcmp(sname.Data(), _splf->getVariogramName()));

    _splf->setVariogram("sine");
    sname = "SineCardinal";
    CPPUNIT_ASSERT_EQUAL(0, strcmp(sname.Data(), _splf->getVariogramName()));

}
void TSamplerFieldTest::throwExceptionTSamplerFieldsetVariogram()
{
    _splf = new URANIE::Sampler::TSamplerField(_tds, "Gauss");
    _splf->setVariogram("toto");

    TString sname = "Gaussian";
    CPPUNIT_ASSERT_EQUAL(0, strcmp(sname.Data(), _splf->getVariogramName()));

}
void TSamplerFieldTest::testSamplerFieldSetRandomFunction()
{
    _splf = new URANIE::Sampler::TSamplerField(_tds, "Gauss");
    _splf->SetRandomFunction(123);
    CPPUNIT_ASSERT_EQUAL(123, _splf->GetRandomFunction());

}
//void TSamplerFieldTest::throwExceptionTSamplerFieldSetRandomFunction()
//{
//  // THIS TEST HAS BEEN DISABLED
//  // TODO: fix me
//  URANIE_IGNORED_TEST();
//
//  _splf = new URANIE::Sampler::TSamplerField (_tds,"Gauss");
////  _splf->setLog();
//  CPPUNIT_ASSERT_THROW (_splf->SetRandomFunction(0.123),URANIE::Exceptions::UErrorExceptions);
//
//}
void TSamplerFieldTest::testSamplerFieldSetIsotropy()
{
    _splf = new URANIE::Sampler::TSamplerField(_tds, "Gauss");

    CPPUNIT_ASSERT_EQUAL(kTRUE, _splf->IsIsotropy());

    _splf->SetIsotropy(0);
    CPPUNIT_ASSERT_EQUAL(kFALSE, _splf->IsIsotropy());

    _splf->SetIsotropy(kTRUE);
    CPPUNIT_ASSERT_EQUAL(kTRUE, _splf->IsIsotropy());

}
void TSamplerFieldTest::testSamplerFieldSetVariance()
{
    _splf = new URANIE::Sampler::TSamplerField(_tds, "Gauss");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, _splf->GetVariance(), DBL_PRECISION_COMPARE);

    _splf->SetVariance(1.95);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.95, _splf->GetVariance(), DBL_PRECISION_COMPARE);
}

void TSamplerFieldTest::throwExceptionTSamplerFieldSetVariance()
{
     _splf = new URANIE::Sampler::TSamplerField(_tds, "Gauss");
     CPPUNIT_ASSERT_THROW(_splf->SetVariance(-2.0), URANIE::Exceptions::UErrorExceptions);
}
void TSamplerFieldTest::testSamplerFieldSetScaleFactor()
{
    _splf = new URANIE::Sampler::TSamplerField(_tds, "Gauss");
    _splf->SetScaleFactor(0.5);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5, _splf->_vScaleFactor(0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5, _splf->_vScaleFactor(1),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.5, _splf->_vScaleFactor(2),
            DBL_PRECISION_COMPARE);

}

void TSamplerFieldTest::throwExceptionTSamplerFieldSetScaleFactor()
{
     _splf = new URANIE::Sampler::TSamplerField(_tds, "Gauss");
     CPPUNIT_ASSERT_THROW(_splf->SetScaleFactor(-1.0),
            URANIE::Exceptions::UErrorExceptions);
     CPPUNIT_ASSERT_THROW(_splf->SetScaleFactor(0.0),
            URANIE::Exceptions::UErrorExceptions);
}
void TSamplerFieldTest::testSamplerFieldgenerateSampleGauss2D()
{
    // default parameters: Gauss and Spectral
    _tds->getAttribute("x")->setStepValue(2);
    _tds->getAttribute("y")->setStepValue(0.5);
    

    _splf = new URANIE::Sampler::TSamplerField(_tds, "Gauss");
    _splf->setSeed(123);
    _splf->SetScaleFactor(10.0);
    _splf->SetRandomFunction(400);
    _splf->SetVariance(1.0);

    _splf->generateSample();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.000000000, _tds->getValue("x", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(26.00000000, _tds->getValue("x", 12),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(51.00000000, _tds->getValue("x", 24),
            DBL_PRECISION_COMPARE);

    Double_t dval00 = 3.45712228428911;
    Double_t dval12 = 6.28806596424689;
    Double_t dval24 = 4.395128101553;

#if ROOT_VERSION_CODE >= ROOT_VERSION(5,30,01) 
    dval00 = 1.63964593446309;
    dval12 = 0.792380522235937;
    dval24 = 0.331699866685088;
#if ROOT_VERSION_CODE >= ROOT_VERSION(6,24,00)
    dval00 = -1.29078517132732;
    dval12 = 1.00151385720155;
    dval24 = -0.799009104458369;
#endif
#endif


    
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dval00, _tds->getValue("y", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dval12, _tds->getValue("y", 12),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dval24, _tds->getValue("y", 24),
            DBL_PRECISION_COMPARE);
}
void TSamplerFieldTest::testSamplerFieldgenerateSampleGauss4D()
{
    // Test to define for 3 dimensions value sample 
    // URANIE_INBUILDING_TEST();

    // default parameters: Gauss and Spectral
    _tds->getAttribute("x")->setStepValue(2);
    _tds->getAttribute("y")->setStepValue(2);

    _tds->addAttribute("z", 1, 51);
    _tds->addAttribute("value");
    _tds->getAttribute("z")->setStepValue(2);

    _splf = new URANIE::Sampler::TSamplerField(_tds, "Gauss");
    _splf->setSeed(1446815647);
    _splf->SetScaleFactor(10.0);
    _splf->SetRandomFunction(400);
    _splf->SetVariance(1.0);

    _splf->generateSample();
 
    Double_t dxval0 = 1.0;
    Double_t dxval1 = 3.0833333333 ;
    Double_t dxval2 = 5.1666666667;

    Double_t dyval0 = 1.0;
    Double_t dyval1 = 32.25 ;
    Double_t dyval2 = 11.416666667;

    Double_t dzval0 = 1.0;
    Double_t dzval1 = 3.0833333333 ;
    Double_t dzval2 = 7.25;
  
    Double_t dvalue0 = -10.6090019504 ;
    Double_t dvalue1 = 0.520163476385 ;
    Double_t dvalue2 = -2.91276839279;

    CPPUNIT_ASSERT_EQUAL(15625, _tds->getNPatterns()) ;

    CPPUNIT_ASSERT_DOUBLES_EQUAL(dxval0, _tds->getValue("x", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dxval1, _tds->getValue("x", 1001),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dxval2, _tds->getValue("x", 2002),
            DBL_PRECISION_COMPARE);
 
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dyval0, _tds->getValue("y", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dyval1, _tds->getValue("y", 1001),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dyval2, _tds->getValue("y", 2002),
            DBL_PRECISION_COMPARE);
 
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dzval0, _tds->getValue("z", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dzval1, _tds->getValue("z", 1001),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dzval2, _tds->getValue("z", 2002),
            DBL_PRECISION_COMPARE);
 
 }
void TSamplerFieldTest::testSamplerFieldgenerateSampleSinus3D()
{

    // Parameters: SinusCardinal and Spectral
    _tds->addAttribute("Weight");

    _splf = new URANIE::Sampler::TSamplerField(_tds, "SineCardinal");
    _splf->setSeed(123);
    _splf->SetScaleFactor(10.0);
    _splf->setVariogram("sinecardinal");
    _splf->SetRandomFunction(1000);
    _splf->SetVariance(1.0);

    _splf->generateSample();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, _tds->getValue("x", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(47.0, _tds->getValue("x", 250),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(41.0, _tds->getValue("x", 550),
            DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.0, _tds->getValue("y", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(5.0, _tds->getValue("y", 250),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(11.0, _tds->getValue("y", 550),
            DBL_PRECISION_COMPARE);

    Double_t dval000 = -10.3155446004658;
    Double_t dval250 = -10.3155446004658;
    Double_t dval550 = -10.3155446004658;

#if ROOT_VERSION_CODE >= ROOT_VERSION(5,30,01)
    dval000 = 0.658243928487207;
    dval250 = 0.250872536327845;
    dval550 = -0.473095709782478;
#if ROOT_VERSION_CODE >= ROOT_VERSION(6,24,00)
    dval000 = -2.25006604587898;
    dval250 = 0.152502354024074;
    dval550 = -0.632694939712257;
#endif
#endif

    CPPUNIT_ASSERT_DOUBLES_EQUAL(dval000, _tds->getValue("Weight", 0),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dval250, _tds->getValue("Weight", 250),
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dval550, _tds->getValue("Weight", 550),
            DBL_PRECISION_COMPARE);

}

void TSamplerFieldTest::testSamplerFieldgenerateManySampleGauss1D()
{

    delete _tds;
    
    // Create a TDataServer
    _tds = new URANIE::DataServer::TDataServer("TDSField", "Weight as Field");
    // Fill the DataServer with the three attributes of the study
    _tds->addAttribute("x", 1, 51);
    _tds->addAttribute("Weight");
    
    // default parameters: Gauss and Spectral
    _tds->getAttribute("x")->setStepValue(2);    

    _splf = new URANIE::Sampler::TSamplerField(_tds, "Gauss",200);
    _splf->setSeed(123);
    _splf->SetScaleFactor(8.0);
    _splf->SetRandomFunction(400);
    _splf->SetVariance(1.0);

    _splf->generateSample();

    _tds->computeStatistic("Weight");

    vector<double> mean;
    vector<double> stand;
    _tds->getAttribute("Weight")->getStatisticalVector("mean",mean);
    _tds->getAttribute("Weight")->getStatisticalVector("std",stand);
    
    for(unsigned int ival=0; ival<mean.size(); ival++)
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(0.000000000, mean.at(ival), 0.12);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(1.000000000, stand.at(ival), 0.2);
    }

}

void TSamplerFieldTest::testSamplerFieldgenerateManySampleGauss2D()
{
    delete _tds;
    
    // Create a TDataServer
    _tds = new URANIE::DataServer::TDataServer("TDSField", "Weight as Field");
    // Fill the DataServer with the three attributes of the study
    _tds->addAttribute("x", 1, 15);
    _tds->addAttribute("y", 1, 15);
    _tds->addAttribute("Weight");
    
    // default parameters: Gauss and Spectral
    _tds->getAttribute("x")->setStepValue(2);
    _tds->getAttribute("y")->setStepValue(0.5);    

    _splf = new URANIE::Sampler::TSamplerField(_tds, "Gauss",50);
    _splf->setSeed(1234565);
    _splf->SetScaleFactor(8.0);
    _splf->SetRandomFunction(400);
    _splf->SetVariance(0.9);

    _splf->generateSample();

    _tds->computeStatistic("Weight");

    vector<double> mean;
    vector<double> stand;
    _tds->getAttribute("Weight")->getStatisticalVector("mean",mean);
    _tds->getAttribute("Weight")->getStatisticalVector("std",stand);
    
    for(unsigned int ival=0; ival<mean.size(); ival++)
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(0.000000000, mean.at(ival), 0.31);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(1.000000000, stand.at(ival), 0.30);
    }

}

void TSamplerFieldTest::throwExceptionSamplerFieldgenerateSampleNoVarisogram()
{
    _splf = new URANIE::Sampler::TSamplerField(_tds, "hyperbolic");
    _splf->setVariogram("hyperbolic");
    CPPUNIT_ASSERT_THROW(_splf->generateSample(),
            URANIE::Exceptions::UErrorExceptions);
    delete _splf;
    _splf = NULL;

    _splf = new URANIE::Sampler::TSamplerField(_tds, "stable");
    _splf->setVariogram("stable");
    CPPUNIT_ASSERT_THROW(_splf->generateSample(),
            URANIE::Exceptions::UErrorExceptions);
    delete _splf;
    _splf = NULL;

    _splf = new URANIE::Sampler::TSamplerField(_tds, "exponential");
    _splf->setVariogram("exponential");
    CPPUNIT_ASSERT_THROW(_splf->generateSample(),
            URANIE::Exceptions::UErrorExceptions);
    delete _splf;
    _splf = NULL;


    _splf = new URANIE::Sampler::TSamplerField(_tds, "spherical");
    _splf->setVariogram("spherical");
    CPPUNIT_ASSERT_THROW(_splf->generateSample(),
            URANIE::Exceptions::UErrorExceptions);
    delete _splf;
    _splf = NULL;


}
void TSamplerFieldTest::throwExceptionSamplerFieldgenerateSampleNoMethod()
{
    _splf = new URANIE::Sampler::TSamplerField(_tds, "dilution");
    _splf->setMethod("dilution");
    CPPUNIT_ASSERT_THROW(_splf->generateSample(),
            URANIE::Exceptions::UErrorExceptions);
    delete _splf;
    _splf = NULL;

    _splf = new URANIE::Sampler::TSamplerField(_tds, "tessellation");
    _splf->setMethod("tessellation");
    CPPUNIT_ASSERT_THROW(_splf->generateSample(),
            URANIE::Exceptions::UErrorExceptions);
    delete _splf;
    _splf = NULL;

    _splf = new URANIE::Sampler::TSamplerField(_tds, "turningbands");
    _splf->setMethod("turningbands");
    CPPUNIT_ASSERT_THROW(_splf->generateSample(),
            URANIE::Exceptions::UErrorExceptions);
    delete _splf;
    _splf = NULL;


}
