//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TestTSamplingMixture.cxx
 \author Fabrice Gaudier
 \date   2012-10-02 16:58:30
 
 \brief  Test File of URANIE::Sampler::TSamplingMixture class
 
 */

// Uranie
#include "TestTSamplingMixture.h"
#include "TAttribute.h"
#include "UDisableTest.h"
#include "uranie_commands.h"

// ROOT
#include "TString.h"
#include "TVectorD.h"

#include <iostream>
#include <math.h>

#define DBL_PRECISION_COMPARE 1.e-10
#define EPSILON 0.03

using namespace std;

TSamplingMixtureTest::TSamplingMixtureTest() :
        _tds1(NULL), _tmix(NULL), _nS(100)
{
}

TSamplingMixtureTest::~TSamplingMixtureTest()
{
}

void TSamplingMixtureTest::setUp()
{
    // Create a TDataServer
    _tds1 = new URANIE::DataServer::TDataServer("tds1", "my first TDS");
    // Fill the DataServer with the three attributes of the study
    _tds1->addAttribute("x1", 0.45, 0.60);
    _tds1->addAttribute("x2", 0.10, 0.20);
    _tds1->addAttribute("x3", 0.06, 0.20);
    _tds1->addAttribute("x4", 0.02, 0.10);
    _tds1->addAttribute("x5", 0.02, 0.12);
    _tds1->addAttribute("x6", 0.00, 0.042);
    _tds1->addAttribute("x7", 0.00, 0.05);
    _tds1->addAttribute("x8", 0.00, 0.02);

    _tmix = NULL;
}

void TSamplingMixtureTest::tearDown()
{
    UranieTests::removeFile(gSystem->pwd(), "_*_.log"); 

    if (_tmix)
      delete _tmix;
    _tmix = NULL;

    if(_tds1!=NULL)
      delete _tds1;
    _tds1 = NULL;

}

////////////////////////////////
///                          ///
///     Auxiliary methods    ///
///                          ///
////////////////////////////////
void TSamplingMixtureTest::addConstraints()
{
    // Add Constraints
    TVectorD vec(9);

    vec.Zero();
    vec(0) = 0.62;
    vec(1) = -1.0;
    vec(4) = -1.0;
    _tmix->addConstraint(vec);

    vec.Zero();
    vec(1) = 0.38;
    vec(3) = -.620;
    vec(4) = -0.620;
    _tmix->addConstraint(vec);

    vec.Zero();
    vec(3) = 1.25;
    vec(4) = -1.0;
    _tmix->addConstraint(vec);

    vec.Zero();
    vec(4) = 1.;
    vec(6) = -2.4;
    _tmix->addConstraint(vec);

    vec.Zero();
    vec(2) = 1.;
    vec(3) = -0.9;
    _tmix->addConstraint(vec);

    vec.Zero();
    vec(2) = 0.2;
    vec(8) = -1.0;
    _tmix->addConstraint(vec);

    vec.Zero();
    vec(5) = 0.556;
    vec(8) = -1.0;
    _tmix->addConstraint(vec);
}

////////////////////////////////
///                          ///
///         Tests            ///
///                          ///
////////////////////////////////

void TSamplingMixtureTest::testSamplingMixtureWithoutCandidates()
{
    _tmix = new URANIE::Sampler::TSamplingMixture(_tds1, _nS, 0.95);
    addConstraints();
    CPPUNIT_ASSERT_EQUAL(0, _tmix->getNGeneratedSample());

    _tmix->setTrace();
    _tmix->generateSample();

    TString sNameAttCandidatesPoints =
            _tmix->getNameInternalAttributeCandidatesPoints();
    TString sNameAttTypePoints = _tmix->getNameInternalAttributeTypePoints();

    CPPUNIT_ASSERT_EQUAL(_nS + 1, _tds1->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(1, _tmix->getNGeneratedSample());
    CPPUNIT_ASSERT(_tds1->isAttribute(sNameAttTypePoints));
    CPPUNIT_ASSERT(_tds1->isAttribute(sNameAttCandidatesPoints));

    CPPUNIT_ASSERT(_tmix->getTrace());
    CPPUNIT_ASSERT_EQUAL(0,
            strcmp("_tsmix_trace_.log", _tmix->getTraceFile().Data()));

    Int_t nfound = 0;
    gSystem->AccessPathName(_tmix->getTraceFile(), kFileExists) ?
            nfound = 1 : nfound = 0;
    CPPUNIT_ASSERT(nfound == 0);

    // Second research
    Int_t ns2 = 20;
    _tmix->_nS = ns2;
    _tmix->unsetTrace();
    _tmix->generateSample();
    CPPUNIT_ASSERT_EQUAL(2, _tmix->getNGeneratedSample());
    CPPUNIT_ASSERT_EQUAL(ns2 + 1, _tds1->getNPatterns());

    CPPUNIT_ASSERT(!_tmix->getTrace());
    CPPUNIT_ASSERT_EQUAL(0, strcmp("", _tmix->getTraceFile().Data()));
    CPPUNIT_ASSERT_EQUAL(0, _tmix->getTraceFile().Length());
}

void TSamplingMixtureTest::testSamplingMixtureWithCandidates()
{
    _tmix = new URANIE::Sampler::TSamplingMixture(_tds1, _nS, 0.95, "save");
    addConstraints();

    _tmix->generateCandidates("");

    TString sNameAttCandidatesPoints =
            _tmix->getNameInternalAttributeCandidatesPoints();
    TString sNameAttTypePoints = _tmix->getNameInternalAttributeTypePoints();

    CPPUNIT_ASSERT_EQUAL(5144, _tds1->getNPatterns());
    CPPUNIT_ASSERT(_tds1->isAttribute(sNameAttTypePoints));
    CPPUNIT_ASSERT(_tds1->isAttribute(sNameAttCandidatesPoints));

    _tds1->setSelect(
            Form("(%s==%d)", sNameAttTypePoints.Data(),
                    URANIE::Sampler::TSamplingMixture::kSummit));
    CPPUNIT_ASSERT_EQUAL(406, _tds1->getNPatterns());
    _tds1->setSelect(
            Form("(%s==%d)", sNameAttTypePoints.Data(),
                    URANIE::Sampler::TSamplingMixture::kCenterEdge));
    CPPUNIT_ASSERT_EQUAL(1437, _tds1->getNPatterns());
    _tds1->setSelect(
            Form("(%s==%d)", sNameAttTypePoints.Data(),
                    URANIE::Sampler::TSamplingMixture::kThirdEdge));
    CPPUNIT_ASSERT_EQUAL(2874, _tds1->getNPatterns());
    _tds1->setSelect(
            Form("(%s==%d)", sNameAttTypePoints.Data(),
                    URANIE::Sampler::TSamplingMixture::kCenterFace));
    CPPUNIT_ASSERT_EQUAL(20, _tds1->getNPatterns());
    _tds1->setSelect(
            Form("(%s==%d)", sNameAttTypePoints.Data(),
                    URANIE::Sampler::TSamplingMixture::kCenterSegment));
    CPPUNIT_ASSERT_EQUAL(406, _tds1->getNPatterns());
    _tds1->setSelect(
            Form("(%s==%d)", sNameAttTypePoints.Data(),
                    URANIE::Sampler::TSamplingMixture::kCentroide));
    CPPUNIT_ASSERT_EQUAL(1, _tds1->getNPatterns());
    _tds1->clearSelect();

    _tmix->unsetCentroide();
    _tmix->generateSample();

    CPPUNIT_ASSERT_EQUAL(5144 + _nS, _tds1->getNPatterns());

    _tds1->setSelect(Form("(%s==0)", sNameAttCandidatesPoints.Data()));
    CPPUNIT_ASSERT_EQUAL(5144, _tds1->getNPatterns());
    _tds1->setSelect(Form("(%s==1)", sNameAttCandidatesPoints.Data()));
    CPPUNIT_ASSERT_EQUAL(_nS, _tds1->getNPatterns());

    _tds1->setSelect(
            Form("(%s==0)&&(%s==%d)", sNameAttCandidatesPoints.Data(),
                    sNameAttTypePoints.Data(),
                    URANIE::Sampler::TSamplingMixture::kSummit));
    CPPUNIT_ASSERT_EQUAL(406, _tds1->getNPatterns());
    _tds1->setSelect(
            Form("(%s==0)&&(%s==%d)", sNameAttCandidatesPoints.Data(),
                    sNameAttTypePoints.Data(),
                    URANIE::Sampler::TSamplingMixture::kCenterEdge));
    CPPUNIT_ASSERT_EQUAL(1437, _tds1->getNPatterns());
    _tds1->setSelect(
            Form("(%s==0)&&(%s==%d)", sNameAttCandidatesPoints.Data(),
                    sNameAttTypePoints.Data(),
                    URANIE::Sampler::TSamplingMixture::kThirdEdge));
    CPPUNIT_ASSERT_EQUAL(2874, _tds1->getNPatterns());
    _tds1->setSelect(
            Form("(%s==0)&&(%s==%d)", sNameAttCandidatesPoints.Data(),
                    sNameAttTypePoints.Data(),
                    URANIE::Sampler::TSamplingMixture::kCenterFace));
    CPPUNIT_ASSERT_EQUAL(20, _tds1->getNPatterns());
    _tds1->setSelect(
            Form("(%s==0)&&(%s==%d)", sNameAttCandidatesPoints.Data(),
                    sNameAttTypePoints.Data(),
                    URANIE::Sampler::TSamplingMixture::kCenterSegment));
    CPPUNIT_ASSERT_EQUAL(406, _tds1->getNPatterns());
    _tds1->setSelect(
            Form("(%s==0)&&(%s==%d)", sNameAttCandidatesPoints.Data(),
                    sNameAttTypePoints.Data(),
                    URANIE::Sampler::TSamplingMixture::kCentroide));
    CPPUNIT_ASSERT_EQUAL(1, _tds1->getNPatterns());

//   _tds1->setSelect(Form("(%s==0)&&(%s==%d)", sNameAttCandidatesPoints.Data(), sNameAttTypePoints.Data(), URANIE::Sampler::TSamplingMixture::kSummit));
//   CPPUNIT_ASSERT_EQUAL(12, _tds1->getNPatterns());
//   _tds1->setSelect(Form("(%s==0)&&(%s==%d)", sNameAttCandidatesPoints.Data(), sNameAttTypePoints.Data(), URANIE::Sampler::TSamplingMixture::kCenterEdge));
//   CPPUNIT_ASSERT_EQUAL(31, _tds1->getNPatterns());
//   _tds1->setSelect(Form("(%s==0)&&(%s==%d)", sNameAttCandidatesPoints.Data(), sNameAttTypePoints.Data(), URANIE::Sampler::TSamplingMixture::kThirdEdge));
//   CPPUNIT_ASSERT_EQUAL(50, _tds1->getNPatterns());
//   _tds1->setSelect(Form("(%s==0)&&(%s==%d)", sNameAttCandidatesPoints.Data(), sNameAttTypePoints.Data(), URANIE::Sampler::TSamplingMixture::kCenterFace));
//   CPPUNIT_ASSERT_EQUAL(3, _tds1->getNPatterns());
//   _tds1->setSelect(Form("(%s==0)&&(%s==%d)", sNameAttCandidatesPoints.Data(), sNameAttTypePoints.Data(), URANIE::Sampler::TSamplingMixture::kCenterSegment));
//   CPPUNIT_ASSERT_EQUAL(4, _tds1->getNPatterns());
//   _tds1->setSelect(Form("(%s==0)&&(%s==%d)", sNameAttCandidatesPoints.Data(), sNameAttTypePoints.Data(), URANIE::Sampler::TSamplingMixture::kCentroide));
//   CPPUNIT_ASSERT_EQUAL(0, _tds1->getNPatterns());
}

