//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   THSIC.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2018/06/22

 \brief Implement the URANIE::Sensitivity::THSIC class

 */
/// ROOT
#include "TPad.h"
#include "TProfile.h"
#include "TLeaf.h"
#include "TGaxis.h"
#include "TLatex.h"
#include "TLegend.h"
#include "TMap.h"
#include "TRegexp.h"
#include "TObjString.h"

#include "TDirectory.h"
#include "Math/Math.h"
#include "Math/ProbFuncMathCore.h"

/// Uranie
#include "THSIC.h"
#include "TSensitivity.h"
#include "TKernel.h"
#include "TPatternsEventList.h"

using std::cout;
using std::endl;

ClassImp (URANIE::Sensitivity::THSIC)

namespace URANIE
{
    namespace Sensitivity
    {
    THSIC::THSIC():
        TSensitivityTest(),
        _nEstimatedVariance(kUnknown),
        _nBootstrapSample(NDefaultBootstrap), _nThresholdBootstrap(NDefaultThresholdBootstrap), _nThresholdGamma(NDefaultThresholdGamma)
    {
        SetName("HSIC");
        setMethodName("HSic");
        SetTitle("my title of the hSIC indexes");
        cout << " *** Default constrcutor THSIC (empty) ***" << endl;
    }

    THSIC::THSIC(URANIE::DataServer::TDataServer *tds, const char *varexpinput, const char *varexpoutput, Option_t * option) :
        TSensitivityTest(tds, varexpinput, varexpoutput, option),
        _nEstimatedVariance(kStdEmpirical),
        _nBootstrapSample(NDefaultBootstrap), _nThresholdBootstrap(NDefaultThresholdBootstrap), _nThresholdGamma(NDefaultThresholdGamma)
    {
        SetName("HSIC");
        setMethodName("HSic");
        SetTitle("my title of the hSIC indexes");

        if ( _blog ) {
            cout << " *** _nY["<< _nY << "] with [" << _sSelectedOutput.Data() << "]" << endl;
            cout << " *** _sOutputAttribute[" << _sOutputAttribute << "]" << endl;
        }

        try {
            if ( _nS < NMinimumPatterns )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("THSIC constructor \nThe number of patterns [%d] is lesser than the number required [%d]", _nS, NMinimumPatterns));

            if ( _sSelectedOutput.Length() == 0 )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("THSIC constructor \nThe output attribute \"%s\" is empty", _sSelectedOutput.Data()));

        } catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            throw ue;
        }

        parseOption(option);
    }

    THSIC::~THSIC()
    {
        if ( _blog )
            cout << " *** destroy THSIC(" << GetName() << ") Title[" << GetTitle() << "] *** " << endl;
    }


    ///////////////////////////////////////////////////////////
    ///  parseOption
    ///////////////////////////////////////////////////////////
    void THSIC::parseOption(Option_t *option)
    {
        TString soption(option);
        soption.ToLower();

        if( soption.Contains("empiri") )
            _nEstimatedVariance = kStdEmpirical;
        if( soption.Contains("median") )
            _nEstimatedVariance = kMedianeDelta;
    }

    ///////////////////////////////////////////////////////////
    ///  setEstimatedVariance
    ///////////////////////////////////////////////////////////
    void THSIC::setEstimatedVariance(EEstimatedVariance nMethod)
    {
        try {
            if ( nMethod == kUnknown )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("THSIC::setEstimatedVariance\n The method \"kUnknown\" [%d] not authorized.\n We must select one of the \"kStdEmpirical | kMedianeDelta\" methods.", nMethod));
            _nEstimatedVariance = nMethod;
        } catch (URANIE::Exceptions::UErrorExceptions& ue) {
            ue.printMessage();
            throw ue;
        }
    }

    ///////////////////////////////////////////////////////////
    ///  setNBoostrapSample
    ///////////////////////////////////////////////////////////
    void THSIC::setNBoostrapSample(Int_t ns)
    {
        try {
            if ( ns < 1 )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("THSIC::setNBoostrapSample\n The size of bootstrap (%d) must be greater than 1", ns));
            _nBootstrapSample = ns;
        } catch (URANIE::Exceptions::UErrorExceptions& ue) {
            ue.printMessage();
            throw ue;
        }
    }

    ///////////////////////////////////////////////////////////
    ///  setThresholdBootstrap
    ///////////////////////////////////////////////////////////
    void THSIC::setThresholdBootstrap(Int_t n)
    {
        try {
            if ( n < 1 )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("THSIC::setThresholdBootstrap\n The Threshold (%d) must be greater than 1", n));
            _nThresholdBootstrap = n;
            if ( _nThresholdBootstrap > _nThresholdGamma )
                throw URANIE::Exceptions::UInfoExceptions(__FILE__, __LINE__,
                                                          Form("THSIC::setThresholdBootstrap function\n Change the value of the _nThresholdGamma parameter from [%d] to [%d]", _nThresholdGamma, _nThresholdBootstrap));
        } catch (URANIE::Exceptions::UErrorExceptions& ue) {
            ue.printMessage();
            throw ue;
        } catch (URANIE::Exceptions::UInfoExceptions& ue) {
            ue.printMessage();
            _nThresholdGamma = _nThresholdBootstrap;
        }
    }

    ///////////////////////////////////////////////////////////
    ///  setThresholdGamma
    ///////////////////////////////////////////////////////////
    void THSIC::setThresholdGamma(Int_t n)
    {
        try {
            if ( n < 1 )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("THSIC::setThresholdGamma\n The Threshold (%d) must be greater than 1", n));
            _nThresholdGamma = n;
            if ( _nThresholdGamma < _nThresholdBootstrap  )
                throw URANIE::Exceptions::UInfoExceptions(__FILE__, __LINE__,
                                                          Form("THSIC::setThresholdGamma function\n Change the value of the _nThresholdBootstrap parameter from [%d] to [%d]", _nThresholdBootstrap, _nThresholdGamma));

        } catch (URANIE::Exceptions::UErrorExceptions& ue) {
            ue.printMessage();
            throw ue;
        } catch (URANIE::Exceptions::UInfoExceptions& ue) {
            ue.printMessage();
            _nThresholdBootstrap = _nThresholdGamma ;
        }
    }

    ///////////////////////////////////////////////////////////
    ///  permutedBootMatrix
    ///////////////////////////////////////////////////////////
    void THSIC::permutedBootMatrix(TMatrixDSym& theMatK, TList *tlstIndex, TList *tlstMatrix)
    {
        Int_t nsize = tlstIndex->GetEntries();
        Int_t nrows = theMatK.GetNrows();
        Int_t ncols = theMatK.GetNcols() ;

        if ( _blog ) {
            cout << " *** THSIC::permutedBootMatrix *** " << endl;
            cout << " *** theMatK(" << nrows << "," << ncols << ")" << endl;
            cout << " *** tlstIndex nsize [" << nsize << "] / nboot[" << _nBootstrapSample << "]" << endl;
            cout << " *** tlstMatrix(" << tlstMatrix->GetEntries() << ")" << endl;
        }

        // Identity Matrix
        tlstMatrix->Add(&theMatK);

        if ( _blog ) {
            cout << " ***** the matrix theMatK (original)" << endl;
            theMatK.Print("f=%7.3f");
        }
        URANIE::DataServer::TPatternsEventList *pelBoot;

        Int_t jPermuted;
        Int_t kPermuted;
        for(Int_t i=1; i<nsize; i++) {
            if ( _blog )
                cout << endl << " ***** bootstrap i(" << i << "/" << nsize << ")" << endl;
            TMatrixDSym matCrt(nrows);
            matCrt.Zero();

            pelBoot = (URANIE::DataServer::TPatternsEventList *)(tlstIndex->At(i));
            if ( _blog )
                pelBoot->Print("all");

            Int_t npat = pelBoot->GetN();
            if ( _blog )
                cout << " *** theMatK(" << nrows << "," << ncols << ") npat[" << npat << "]" << endl;

            for(Int_t j=0; j < npat; j++) {
                jPermuted = pelBoot->GetEntry(j);
                matCrt(jPermuted, jPermuted) = theMatK(j, j);

                for(Int_t k=0; k< j; k++) {
                    kPermuted = pelBoot->GetEntry(k);
                    if ( _blog )
                        cout << " *** j(" << j << "/" << npat << "] jPermuted[" << jPermuted << "] k[" << k << "/" << j << "] kPermuted[" << kPermuted << "]" << endl;
                    matCrt(jPermuted, kPermuted) = theMatK(j, k);
                    matCrt(kPermuted, jPermuted) = theMatK(k, j);
                }
            }

            if ( _blog ) {
                cout << " ***** the matrix matCrt " << endl;
                matCrt.Print("f=%7.3f");
            }

            TMatrixDSym *matHKyBH = new TMatrixDSym(matCrt);
            matHKyBH->SimilarityT(_matH);
            if ( _blog ) {
                cout << " ***** the matrix matHKyBH " << endl;
                matHKyBH->Print("f=%7.3f");
            }

            tlstMatrix->Add(matHKyBH);
        }

        if ( _blog )
            cout << " *** End Of THSIC::permutedBootMatrix *** " << endl;
    }

    ///////////////////////////////////////////////////////////
    ///  estimateSigma
    ///////////////////////////////////////////////////////////
    Double_t THSIC::estimateSigma(const char *sAtt, Option_t * option)
    {
        TString soption = option;
        soption.ToLower();
        if ( _blog )
            cout << " ** THSIC::estimateSigma Att[" << sAtt << "] method[" << _nEstimatedVariance << "] soption[" << soption << "]" << endl;

        Double_t dSigma = -1.0;

        try {
            switch (_nEstimatedVariance) {
            case kStdEmpirical:
                _tds->computeStatistic(sAtt);
                dSigma = _tds->getAttribute(sAtt)->getStd();
                if ( _blog )
                    cout << " ** Empirical method :: Sigma [" << _tds->getAttribute(sAtt)->getStd() << "]" << endl;
                break;
            case kMedianeDelta: {
                URANIE::DataServer::TDataServer* tdsMedianCrt = new URANIE::DataServer::TDataServer(Form("__tds_median_%s", sAtt), "The crt tds");
                tdsMedianCrt->addAttribute("_the_med_");
                tdsMedianCrt->createTuple();

                Double_t dVali, dValj;
                Int_t iCrt = 0;
                for(Int_t i = 0; i < _nS ; i++) {
                    dVali = _tds->getValue(sAtt, i);
                    for(Int_t j=0; j < i; j++) {
                        dValj = _tds->getValue(sAtt, j);
                        tdsMedianCrt->getTuple()->Fill(iCrt++, TMath::Abs( dVali - dValj));
                    }
                }
                Double_t dProbaMedian=0.5;
                tdsMedianCrt->computeQuantile("_the_med_", dProbaMedian, dSigma);
                if ( _blog )
                    cout << " ** Median method :: iCrt[" << iCrt << "] dProbaMedian[" << dProbaMedian << "] dSigma[" << dSigma << "] Empirique[" << _tds->getAttribute(sAtt)->getStd() << "]" << endl;
                delete tdsMedianCrt; tdsMedianCrt = NULL;
            }
                break;
            default:
                throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                                                             Form("THSIC::estimateSigma\n Define a method \"[empiri|median]\" for attribute \"%s\" to estimate sigma for a Gaussian kernel in the option THSIC constructor.", sAtt));
                break;
            }

            if ( dSigma < 0.0 )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("THSIC::estimateSigma\n Sigma computed [%f] for attribute \"%s\" is negative", dSigma, sAtt));
            if ( dSigma < dEPSILONSigma )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("THSIC::estimateSigma\n Sigma computed [%e] for attribute \"%s\" is lesser then dEPSILONSigma [%e]", dSigma, sAtt, dEPSILONSigma));
        } catch (URANIE::Exceptions::UWarningExceptions& ue) {
            ue.printMessage();
            throw ue;
        } catch (URANIE::Exceptions::UErrorExceptions& ue) {
            ue.printMessage();
            throw ue;
        }

        return dSigma;
    }


    ///////////////////////////////////////////////////////////
    ///  generateHMatrix
    ///////////////////////////////////////////////////////////
    void THSIC::generateHMatrix(Option_t * option)
    {
        TString soption = option;
        soption.ToLower();
        if ( _blog )
            cout << "  ** THSIC::generateHMatrix soption[" << soption << "]" << endl;

        Double_t d1OverN = 1.0 / (1.0*_nS);

        _matH.ResizeTo(_nS, _nS);
        _matH.UnitMatrix();
        _matH -= d1OverN;

        if ( _blog )
            _matH.Print();
    }

    ///////////////////////////////////////////////////////////
    ///  getMatrixKGaussian
    ///////////////////////////////////////////////////////////
    TMatrixDSym THSIC::getMatrixKGaussian(const char *sAtt, Double_t dvariance)
    {
        URANIE::Modeler::TKernel *kerGaussian = new URANIE::Modeler::TKernel(_tds, sAtt, sAtt, "nointercept,nostoreyhat");
        kerGaussian->setKernel(URANIE::Modeler::TKernel::kRBF, dvariance);
        kerGaussian->setLambda(1e-9);
        kerGaussian->computeK();
        TMatrixDSym matKernY = kerGaussian->getMatrixK();

        /// Clean
        delete kerGaussian; kerGaussian = NULL;
        return(matKernY);
    }

    ///////////////////////////////////////////////////////////
    ///  unBiasedHSIC
    ///////////////////////////////////////////////////////////
    Double_t THSIC::unBiasedHSIC( TMatrixD matKx, TMatrixD matKy)
    {
        Double_t dUnBiasedHSIC = 0.0;

        Double_t dTraceKxKy = 0.0;
        Double_t dTraceCrtKxKy;
        Double_t dSumKx = 0.0;
        Double_t dSumKy = 0.0;

        Double_t dSumKxKy = 0.0;
        Double_t dSumCrtKxKy = 0.0;

        for(Int_t i=0; i<_nS; i++) {
            dTraceCrtKxKy = 0.0;
            for(Int_t k=0; k<_nS; k++) {

                dSumCrtKxKy = 0.0;
                for(Int_t l=0; l<_nS; l++) {
                    if ( ( l != i ) && ( l != k ))
                        dSumCrtKxKy += matKx(i, l) * matKy(l, k);
                }
                dSumKxKy += dSumCrtKxKy;

                if ( i == k ) {
                    dTraceCrtKxKy += dSumCrtKxKy; //matKx(i, k) * matKy(k, i);
                } else {
                    dSumKx += matKx(i, k);
                    dSumKy += matKy(i, k);
                }

            }
            dTraceKxKy += dTraceCrtKxKy;
        }

        dUnBiasedHSIC += dTraceKxKy;
        dUnBiasedHSIC += dSumKx * dSumKy / ((_nS-1.)*(_nS-2.));
        dUnBiasedHSIC -= 2. * dSumKxKy / (_nS -2.);
        dUnBiasedHSIC /= (_nS-3.)*_nS;

        return (dUnBiasedHSIC);
    }

    ///////////////////////////////////////////////////////////
    ///  getSumOfMatrixBk
    ///////////////////////////////////////////////////////////
    Double_t THSIC::getSumOfMatrixBk(TMatrixD matHLkH, TMatrixD matHLH)
    {

        Double_t dSumBk = 0.0;

        for (Int_t i = 0; i < _nS; i++) {
            for (Int_t j = 0; j < i; j++) {
                dSumBk += 2.0 * matHLkH(i,j) * matHLH(i, j) * matHLkH(i,j) * matHLH(i, j);
            }
        }

        Double_t d1OverN = 1.0 / (1.0*_nS);
        Double_t d1OverNMinus1 = 1.0 / (1.0*(_nS-1));

        dSumBk *= d1OverN * d1OverNMinus1; ///= ( 1.0 * _nS * ( _nS-1));

        if ( _blog )
            cout << " *** getSumOfMatrixBk dSumBk[" << dSumBk << "]" << endl;

        return(dSumBk);
    }

    ///////////////////////////////////////////////////////////
    ///  computeTrace
    ///////////////////////////////////////////////////////////
    Double_t THSIC::computeTrace(TMatrixD matKx, TMatrixD matHKyH)
    {
        Double_t dTrace = 0.0;
        Double_t dSumCrt;

        for(Int_t i=0; i<_nS; i++) {
            // Compute Only the elements ( Kx * HKyH) [i, i]
            dSumCrt = 0.0;
            for(Int_t k=0; k<_nS; k++) {
                dSumCrt += matKx(i, k) * matHKyH(k, i);
            }
            dTrace += dSumCrt;
        }

        Double_t d1OverN = 1.0 / (1.0*_nS);

        dTrace *= d1OverN * d1OverN;  ///= (1.0*_nS*_nS);
        if ( _blog )
            cout << " *** computeTrace dTrace[" << dTrace << "]" << endl;

        return( dTrace);
    }

    ///////////////////////////////////////////////////////////
    ///  generateSample
    ///////////////////////////////////////////////////////////
    void THSIC::generateSample(Option_t * option)
    {
        try {
            TString soption = option;
            soption.ToLower();
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                                                         Form("THSIC::generateSample\n nothing TODO in these method (option[%s])", soption.Data()));
        }   catch (URANIE::Exceptions::UWarningExceptions& ue) {
            ue.printMessage();
            throw ue;
        }
    }

    ///////////////////////////////////////////////////////////
    ///  drawIndexesHistogram
    ///////////////////////////////////////////////////////////
    void THSIC::drawIndexes(TString sTitle, const char *select, Option_t * option)
    {
        TString soption(option);
        soption.ToLower();

        if ( _blog )
            cout << " *** THSIC::drawIndexes sTitle[" << sTitle << "] select[" << select << "] option[" << soption << "]" << endl;

        /// Case to activate the log (locally)
        if ( soption.Contains("log") )
            setLog();

        /// case to draw the legend
        Bool_t bDrawLegend = kFALSE;
        if ( soption.Contains("legend") )
            bDrawLegend = kTRUE;

        // Case of default output to draw : the first output
        TString sDefaultOutputDraw = Form("%s", _sOutputAttribute);
        if ( _nY > 1 ) {
            URANIE::DataServer::TAttribute *atty = (URANIE::DataServer::TAttribute *) (_listOfOutputAttributes->At(0));
            sDefaultOutputDraw = atty->GetName();
        }
        if ( _blog )
            cout << " *** _nY(" << _nY << ") sDefaultOutputDraw[" << sDefaultOutputDraw << "]" << endl;

        Double_t dThreshold = 1.05;
        TString sSelect = TString(select);
        try {
            if ( _ntresult->GetEntries() != 3 * _nX *_nY  ) {
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("THSIC::drawIndexes method \n The TTree of results does not contain the good number of entry (%d) than (%d)\n Do you have call the computeIndexes() method ?", (Int_t)(_ntresult->GetEntries()), 3 * _nX *_nY));
            }

            if ( ( sSelect.Length() > 0 ) && ( sSelect.Contains("Out==") ) ) {
                    TRegexp regOutput("Out==\".*\"");
                    TString SOutTmp = sSelect(regOutput);
                    sDefaultOutputDraw = SOutTmp.Remove(0, 5);
                    sDefaultOutputDraw.ReplaceAll("\"", "");
            }

            if ( !_tds->isAttribute(sDefaultOutputDraw) )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("THSIC::drawIndexes method \n The Output attribute \"%s\" does not exist in the TDS (%s)",
                                                               sDefaultOutputDraw.Data() , _tds->GetName()));


            if (soption.Contains("pvalues="))
            {
                TRegexp reg("pvalues=[+/-]?[0-9.]*.[0-9]*");
                TString out = soption(reg);
                TString sout = out.Remove(0, 8);
                dThreshold = sout.Atof();
                if ( _blog )
                    cout << " ** dThreshold[" << dThreshold << "] in Regular Expression" << endl;

                if ( dThreshold < 0.0 )
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                               Form("THSIC::drawIndexes method \n The given threshold for the Pvalues is negative [%f]. It must be in 0.0 and 1.0", dThreshold));

                if ( dThreshold > 1.0 ) {
                    dThreshold = 1.05;
                    throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                                                                 Form("THSIC::drawIndexes method \n The given threshold for the Pvalues is greater than 1.0  [%s] .\n Then %f is selected ( no selection ).", out.Data(), dThreshold));
                }


            }
        } catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            throw ue;
        }
        catch (URANIE::Exceptions::UWarningExceptions& ue)
        {
            ue.printMessage();
        }

        Int_t nval = (Int_t) (_ntresult->GetEntries()/_nY);
        Int_t natt = _nX; // 3 indexes stored in the computeIndexes in the order :: dR2HSICXkY , dHSIC, dPValue

        if ( _blog ) {
            cout << "  ** THSIC::drawIndexes" << endl;
            cout << "  ** sTitle[" << sTitle << "]" << endl;
            cout << "  ** select[" << select << "]" << endl;
            cout << "  ** sDefaultOutputDraw[" << sDefaultOutputDraw << "]" << endl;
            cout << "  ** option[" << option << "]" << endl;
            cout << "  ** dThreshold[" << dThreshold << "]" << endl;
            cout << "  ** nval[" << nval << "] natt[" << natt << "]" << endl;
            cout << "  ** _sMethodName[" << _sMethodName << "]" << endl;
        }

        // Clean the TEventList
        if( gROOT->FindObject("evlHSIC") )
            gROOT->Delete("evlHSIC");

//        _ntresult->Draw(">>evlHSIC",  Form("(Algo==\"--pvalues--\")&&(Sobol<%f)&&(%s)", dThreshold, select), "goff");
        _ntresult->Draw(">>evlHSIC",  Form("(Out==\"%s\")&&(Algo==\"--pvalues--\")&&(Sobol<%f)", sDefaultOutputDraw.Data(), dThreshold), "goff");
        TEventList *telHSIC = (TEventList*) gROOT->Get("evlHSIC");
        Int_t nattSelect = telHSIC->GetN();

        try
        {
            if (nattSelect > _nX )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("THSIC::drawIndexes ERROR : the number of results [%d] is greater than the number of inputs [%d]",
                                                                nattSelect, _nX));
        }catch(URANIE::Exceptions::UErrorExceptions &ue)
        {
            ue.printMessage();
            throw ue;
        }

        if ( _blog ) {
            telHSIC->Print("all");
            cout << " ** evlHSIC natt selected [" << nattSelect << "]" << endl;
        }

        // a local variable for the event type
        char etype[500];

        _ntresult->GetLeaf("Inp")->SetAddress(&etype);

        TLeaf *leafSobol = _ntresult->GetLeaf("Sobol");

        if( gROOT->FindObject("R2HSic") )
            gROOT->Delete("R2HSic");
        TH1F *hR2HSic = new TH1F("R2HSic","Option B example",nattSelect,1,nattSelect+1);
        TMap *mapAtt = new TMap();
        mapAtt->SetOwnerKeyValue();
        
        for(Int_t ipat=0; ipat< nattSelect; ipat++) {
            // R2HSic
            _ntresult->GetEntry(telHSIC->GetEntry(ipat)-2);
            if ( _blog )
                cout << " *** (" << ipat << "/" << nattSelect << ") - Out(" << sDefaultOutputDraw << ") etype[" << etype << "] HSic[" << leafSobol->GetValue() << "]" << endl;
            hR2HSic->SetBinContent(ipat+1, leafSobol->GetValue());
            hR2HSic->GetXaxis()->SetBinLabel(ipat+1,etype);

            // PValues :: indexes of the entry stored in the TMap object indexed by the attribute name
            TObjString* key = new TObjString(Form("%s", etype));
            TObjString* value = new TObjString(Form("%d", (int)(telHSIC->GetEntry(ipat))));
            mapAtt->Add(key, value);
        }

        if ( _blog ) {
            hR2HSic->Print("all");
            mapAtt->Print("all");
        }


        hR2HSic->SetTitle("R^{2}_{H_{Sic}} and P_{values}");

        // Title of the graph
        TString sTempTitle = sTitle;

        // The output name
        sTempTitle += Form(" for \"%s\" ", sDefaultOutputDraw.Data());

        if ( dThreshold <1.0 ) {
            sTempTitle += Form(" ( P_{Values} < %.2f )", dThreshold);
        }

        if ( natt > nattSelect )
            sTempTitle += Form(" [ %d/%d attributes ]", nattSelect, natt);
        else
            sTempTitle += Form(" [ %d attributes ]", natt);


        hR2HSic->SetTitle(sTempTitle);
        hR2HSic->LabelsOption(">");
        hR2HSic->SetStats(0);
        hR2HSic->SetBarWidth(0.80);
        hR2HSic->SetBarOffset(0.2);
        hR2HSic->SetFillColor(kBlue); // 38 kBlue
        hR2HSic->SetMinimum(0.0);
        //  hR2HSic->SetMaximum(1.0);
        hR2HSic->Draw("bar2");

        // Set the X/Y axis of the histogramm
        hR2HSic->GetXaxis()->SetTickLength(0.25);
        hR2HSic->GetXaxis()->SetTickSize(0.0);
        if ( !bDrawLegend ) {
            hR2HSic->GetYaxis()->SetLabelColor(hR2HSic->GetFillColor());
            hR2HSic->GetYaxis()->SetAxisColor(hR2HSic->GetFillColor());
        }

        gPad->SetGridy();
        gPad->Modified();
        gPad->Update();

        // PValues with HSIC


        if( gROOT->FindObject("hPValues") )
            gROOT->Delete("hPValues");
        TH1F *hPValues = new TH1F("hPValues","Option B example",nattSelect,1,nattSelect+1);
        for(Int_t ipat=0; ipat< nattSelect; ipat++) {
            if( _blog )
                cout << "  *** ipat(" << ipat << "/" << nattSelect << ") att[" << hR2HSic->GetXaxis()->GetBinLabel(ipat + 1) << "]";
            TString rangeInfo = mapAtt->GetValue(hR2HSic->GetXaxis()->GetBinLabel(ipat + 1))->GetName();
            if( _blog )
                cout << " sRangeInfo[" <<rangeInfo << "]";
            Int_t rangeN = atoi(rangeInfo.Data());
            if( _blog )
                cout << " rangeN[" <<rangeN << "]";
            _ntresult->GetEntry(rangeN);
            hPValues->SetBinContent(ipat+1, 1.0-leafSobol->GetValue());
            //hPValues->GetXaxis()->SetBinLabel(ipat+1,etype);
            if( _blog )
                cout << " etype[" << etype << "] PValues[" << leafSobol->GetValue() << "]" << endl;
        }

        if( _blog )
            hPValues->Print("all");

        hPValues->SetMarkerStyle(8);
        hPValues->SetMarkerSize(1.2);
        hPValues->SetLineColor(kRed); // kRed 38);
        hPValues->SetLineWidth(2);
        hPValues->SetMarkerColor(hPValues->GetLineColor());

        // Scale hPValues to the pad coordinates
        Float_t rightmax = 1.1*hPValues->GetMaximum();
        Float_t scale = gPad->GetUymax()/rightmax;
        hPValues->Scale(scale, "nosw2");
        hPValues->Draw("LP,SAME");

        // Draw an axis on the right side
        TGaxis *axis = new TGaxis(gPad->GetUxmax(),gPad->GetUymin(), gPad->GetUxmax(), gPad->GetUymax(),0,rightmax,510,"+LW", -0.80);
        axis->SetLineColor(hPValues->GetMarkerColor());
        axis->SetLabelColor(hPValues->GetMarkerColor());
        axis->SetTickLength(0.50);
        axis->SetMaxDigits(5);
        axis->Draw();

        if ( bDrawLegend  ) {
            TLegend *legend = new TLegend(0.70,0.70,0.87,0.80,NULL,"brNDC");
            legend->SetBorderSize(0);
            legend->SetLineColor(1);
            legend->SetLineStyle(1);
            legend->SetLineWidth(1);

            legend->AddEntry(hR2HSic,"R^{2}_{H_{Sic}}","f");
            legend->AddEntry(hPValues,"1.0 - P_{values}","lp");
            legend->Draw("b");
        } else {
            // Draw the R2Hsic title axis
            TLatex *R2HSictitle = new TLatex(1.1,1.05*gPad->GetUymax(),"R^{2}_{H_{Sic}}");
            R2HSictitle->Draw();
            R2HSictitle->SetTextSize(0.04);
            R2HSictitle->SetTextColor(kBlack);
            R2HSictitle->SetTextColor(hR2HSic->GetFillColor());
            R2HSictitle->SetTextAngle(0.);
            R2HSictitle->SetTextAlign(31);

            // Draw the PValues title axis
            TLatex *PvaluesHSictitle = new TLatex(nattSelect+1.0,1.05*gPad->GetUymax(),"1.0 - P_{values}");
            PvaluesHSictitle->Draw();
            PvaluesHSictitle->SetTextSize(0.04);
            PvaluesHSictitle->SetTextColor(hPValues->GetMarkerColor());
            PvaluesHSictitle->SetTextAngle(0.);
            PvaluesHSictitle->SetTextAlign(31);
        }

        gPad->Modified();
        gPad->Update();

        if ( soption.Contains("log") )
            unsetLog();

        //clean
        delete mapAtt; mapAtt = NULL;
    }


    ///////////////////////////////////////////////////////////
    ///  computeIndexes
    ///////////////////////////////////////////////////////////
    void THSIC::computeIndexes(Option_t * option)
    {
        cout << endl;
        cout << "*********************************" << endl;
        cout << " *** THSIC::computeIndexes" << endl;
        evaluateIndexes(option);

    }

    ///////////////////////////////////////////////////////////
    ///  evaluateIndexes
    ///////////////////////////////////////////////////////////
    void THSIC::evaluateIndexes(Option_t * option)
    {
        Int_t nInputs = _listOfInputAttributes->GetEntries();
        Int_t nOutputs = _listOfOutputAttributes->GetEntries();

        Int_t nBoot = 0;
        cout << " *** _nS[" << _nS << "]" << endl;
        cout << " *** nInputs[" << _nX << "]" << endl;
        cout << " *** nOutputs[" << nOutputs << "]" << endl;
        cout << " *** nBoot[" << nBoot << "]" << endl;
        TString soption = option;
        soption.ToLower();

        cout << " *** soption[" << soption << "]" << endl;

        Int_t nCrtThreshold = _nS ;
        cout << " *** nCrtThreshold[" << nCrtThreshold << "] / [" << _nThresholdBootstrap << "]" << endl;

        if ( nCrtThreshold < _nThresholdBootstrap ) {
            nBoot = _nBootstrapSample;
            cout << " *** nBoot[" << nBoot << "]" << endl;

            if (soption.Contains("nboot=")) {
                TRegexp reg("nboot=[0-9]*");
                TString out = soption(reg);
                TString snboot = out.Remove(0, 6);
                if ( snboot.IsDigit() ) {
                    nBoot = snboot.Atoi();
                    _nBootstrapSample = nBoot;
                    soption.ReplaceAll(Form("nboot=%d", nBoot), "");
                }
            }
            if ( _blog )
                cout << " *** soption[" << soption << "]   nBoot[" << nBoot << "]" << endl;
        }
        Bool_t bUnbiased = kFALSE;

        if ( soption.Contains("unbiased") ) {
            bUnbiased = kTRUE;
            if ( 1 ) //_blog )
                cout << " *** Unbiased computation" << endl;
        }

        cout << " *** soption[" << soption << "]" << endl;
        cout << " *** nBoot[" << nBoot << "]" << endl;


        Double_t d1OverN = 1.0 / (1.0*_nS);
        Double_t d1OverNMinus1 = 1.0 / (1.0*(_nS-1));
        Double_t d1OverN2 = d1OverN * d1OverN;

        if ( _blog ) {
            cout << " *** Compute statistics of Inputs and Outputs" << endl;
            cout << " ***  _sSelectedInput[" << _sSelectedInput << "]" << endl;
            cout << " *** _sSelectedOutput[" << _sSelectedOutput << "]" << endl;
        }

        if ( _blog )
            cout << " *** Generate the H matrix" << endl;
        generateHMatrix();

        Double_t dCrtSigma;
        Double_t dCrtTheta;
        Double_t dHSICXkXk, dHSICXkY, dHSICYY;
        Double_t dR2HSICXkY;
        Double_t dPValue;

        // To compute the bias and the variance of the HSIC estimator
        Double_t dSumOfL, dSumOfLk, dSumBk;
        Double_t dBiask, dEspTest;
        Double_t dVarTest;
        TMatrixD matBk;

        // To compute the pValues with the Gamma Distribution
        Double_t dAlphaGamma, dBetaGamma;

        TIter nextouput(_listOfOutputAttributes);
        while (URANIE::DataServer::TAttribute *attouput =
               (URANIE::DataServer::TAttribute *) nextouput())
        {
            sprintf(_sOutputAttribute, "%s", attouput->GetName());
            cout << " *** case of output [" << attouput->GetName() << "] _sOutputAttribute[" << _sOutputAttribute <<"]" << endl;

            // Estimate the sigma for the Gaussian Kernel
            dCrtSigma = estimateSigma(attouput->GetName());
            dCrtTheta = 1.0 / (2.0 * dCrtSigma * dCrtSigma);
            if ( _blog )
                cout << " *** dCrtSigma[" << dCrtSigma << "] dCrtTheta[" << dCrtTheta << "]" << endl;

            TMatrixDSym matKernY = getMatrixKGaussian( attouput->GetName(), dCrtTheta );
            dSumOfL = matKernY.Sum();

            if ( _blog ) {
                cout << " ***  Traditionnal matKernY with dSumOfL[" << dSumOfL << "]" << endl;
                matKernY.Print();
            }

            TMatrixDSym matHKyH(matKernY);
            matHKyH.SimilarityT(_matH);
            if ( _blog ) {
                cout << "  ***  Traditionnal computations Matrice for \"H . K_{Y} . H\" " << endl;
                matHKyH.Print();
            }

            // Compute HSIC(Y, Y)
            dHSICYY = computeTrace( matKernY, matHKyH );
            if ( _blog )
                cout << "  *** dHSICYY[" << dHSICYY << "]" << endl;

            Double_t dmuy2 = 0.0;
            Int_t iCrt = 0;
            TList *lstIndexBoot = NULL;
            TList *lstMatrixBoot = NULL;

            if ( nBoot > 0 ) {
                // Use the Bootstrap permutation test pour compute the p-values quantities
                // Generate the nBoot permutations vector
                lstIndexBoot = new TList();
                lstIndexBoot->SetOwner();

                URANIE::DataServer::TPatternsEventList tevid("uranie_HSIC_boot_identity", "the identity boot", _nS);
                for(Int_t i=0; i<_nS; i++)
                    tevid.Enter(i);

                if ( _blog )
                    tevid.Print("all");
                lstIndexBoot->Add(&tevid);

                for(Int_t ib=0; ib < nBoot; ib++) {
                    URANIE::DataServer::TPatternsEventList *pelBoot = new URANIE::DataServer::TPatternsEventList(
                                Form("uranie_HSIC_boot_%d", ib+1),
                                Form("the HSIC Bootstrap sample  %d", ib+1),
                                _nS);
                    for(Int_t i=0; i<_nS; i++)
                        pelBoot->Enter(i);

                    pelBoot->Mix();
                    if ( _blog )
                        pelBoot->Print("all");
                    lstIndexBoot->Add(pelBoot);
                }

                if ( _blog )
                    lstIndexBoot->Print("all");

                lstMatrixBoot = new TList();
                lstMatrixBoot->SetOwner();
                lstMatrixBoot->SetName(Form("Bootstrap matrix %d for %s", nBoot, _tds->GetName()));

                permutedBootMatrix(matKernY, lstIndexBoot, lstMatrixBoot);
            } else {
                // compute dmuy2 for the Gamma approximation
                for(Int_t i=0; i<_nS; i++) {
                    for(Int_t j=0; j<i; j++) {
                        dmuy2 += 2.0 * matKernY(i,j);
                    }
                }

                dmuy2 *= d1OverN;
                dmuy2 *= d1OverNMinus1;
                if (  _blog )
                    cout << " *** dmuy2[" << dmuy2 << "]" << endl<< endl<< endl;
            }

            TIter nextinput(_listOfInputAttributes);
            while (URANIE::DataServer::TAttribute *attinput =
                   (URANIE::DataServer::TAttribute *) nextinput())
            {
                cout << " *** input ("<< iCrt++ << "/" << nInputs << ") [" << attinput->GetName() << "]";

                // Estimate the sigma for the Gaussian Kernel
                dCrtSigma = estimateSigma(attinput->GetName());
                dCrtTheta = 1.0 / (2.0 * dCrtSigma* dCrtSigma);
                if ( _blog )
                    cout << " *** dCrtSigma[" << dCrtSigma << "] dCrtTheta[" << dCrtTheta << "]" << endl;

                TMatrixDSym matKernXk = getMatrixKGaussian( attinput->GetName(), dCrtTheta );
                dSumOfLk = matKernXk.Sum();
                if ( _blog ) {
                    cout << " ***  Traditionnal matKernXk with dSumOfLk[" << dSumOfLk << "]" << endl;
                    matKernXk.Print();
                }

                // Compute HSic ( Xk Y )
                dHSICXkY = computeTrace( matKernXk, matHKyH );
                if ( _blog )
                    cout << " *** dHSICXkY[" << dHSICXkY << "]" << endl;
                dBiask = ( 1.0 - dSumOfLk * d1OverN2 ) * ( 1.0 - dSumOfL * d1OverN2) * d1OverN;
                if ( _blog )
                    cout << " *** dBiask[" << dBiask << "]" << endl;

                TMatrixDSym matHLkH(matKernXk);
                matHLkH.SimilarityT( _matH );
                if ( _blog ) {
                    cout << " ***  Traditionnal computations Matrice for \"H . K_{X} .  H\" " << endl;
                    matHLkH.Print();
                }

                // Compute HSic ( Xk Xk)
                dHSICXkXk = computeTrace( matKernXk, matHLkH);
                if ( _blog )
                    cout << " *** dHSICXkXk[" << dHSICXkXk << "]" << endl;

                // Compute R2_{HSic} ( Xk, Y) (from DaVeiga 2015)
                dR2HSICXkY = dHSICXkY / TMath::Sqrt( dHSICXkXk * dHSICYY );


                if ( nBoot > 0 ) {
                    Double_t dHSICB;
                    Int_t nHSICBgreaterHSIC = 0;
                    Int_t nsize =  lstMatrixBoot->GetEntries();
                    if ( _blog )
                        cout << " *** nBOOT[" << nBoot << "] nsize[" << nsize << "]" << endl;

                    for(Int_t i=1; i< nsize; i++) {
                        if ( _blog )
                            cout << " *** i(" << i << "/" << nsize << ") by the lstMatrixBoot --- x(" << attinput->GetName() << ") --- ";
                        TMatrixD *matHKyBH = (TMatrixD *) (lstMatrixBoot->At(i));
                        if ( _blog )
                            matHKyBH->Print("f=%7.3f");
                        dHSICB = computeTrace( matKernXk, TMatrixD( *matHKyBH));
                        if ( dHSICB > dHSICXkY )
                            nHSICBgreaterHSIC++;

                        if ( _blog )
                            cout << " dHSICB[" << dHSICB << "] dHSICXkY[" << dHSICXkY << "] nHSICBgreaterHSIC[" << nHSICBgreaterHSIC << "]" << endl;

                    }

                    dPValue = ( 1.0 * nHSICBgreaterHSIC) / ( 1.0 * nBoot);
                    cout << " HSic[" << dHSICXkY << "]"
                         << " R2HSic[" << dR2HSICXkY << "]"
                         << " 1.-PValue[" << 1.0-dPValue << "] by bootstrap(" << nBoot << ")" << endl;

                } else {

                    // Compute parameters of the Gamma distribution ( alpha, beta)
                    Double_t dmux2 = 0.0;
                    for(Int_t i=0; i<_nS; i++) {
                        for(Int_t j=0; j<i; j++) {
                            dmux2 += 2.0 * matKernXk(i,j);
                        }
                    }
                    dmux2 *= d1OverN;
                    dmux2 *= d1OverNMinus1;

                    if ( _blog )
                        cout << " *** dmux2[" << dmux2 << "] dmuy2[" << dmuy2 << "]" << endl;

                    // Esperance ( cf Gretton 2005 Theorem 3 page 5)
                    dEspTest = (1+dmux2*dmuy2-dmux2-dmuy2) * d1OverN;
                    if ( _blog )
                        cout << " *** dEspTest[" << dEspTest << "]" << endl;

                    // Variance ( cf Gretton 2005 Theorem 4 page 5)
                    dSumBk = getSumOfMatrixBk( matHLkH, matHKyH);
                    if ( _blog )
                        cout << " *** dSumBk[" << dSumBk << "]" << endl;

                    // dVarTest = dSumBk * 2.0*(_nS-4)*(_nS-5) / ( (1.0*_nS)*(_nS-1)*(_nS-2)*(_nS-3));
                    dVarTest = dSumBk * 2.0 * d1OverN * d1OverNMinus1*(_nS-4)*(_nS-5) / ((_nS-2)*(_nS-3));
                    if ( _blog )
                        cout << " *** dVarTest[" << dVarTest << "]" << endl;

                    try {
                        if ( dVarTest < 0.0 )
                            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                                       Form("THSIC::evaluateIndexes\n The Variance estimated [%f] for the Gamma approximation is negative", dVarTest));
                    } catch (URANIE::Exceptions::UErrorExceptions& ue) {
                        ue.printMessage();
                        throw ue;
                    }

                    // The two parameters of the Gamma distribution (under independant hypothesis)
                    dAlphaGamma = dEspTest * dEspTest / dVarTest;
                    dBetaGamma =  _nS * dVarTest / dEspTest;

                    if ( _blog )
                        cout << " *** dAlphaGamma[" << dAlphaGamma << "] dBetaGamma[" << dBetaGamma << "]" << endl;

                    dPValue = ROOT::Math::gamma_cdf_c( _nS * dHSICXkY, dAlphaGamma, dBetaGamma, 0.0);

                    if ( bUnbiased ) {
                        dHSICXkY =  unBiasedHSIC ( matKernXk, matKernY);
                        if ( _blog )
                            cout << " *** dHSIC unbiased[" << dHSICXkY << "] from function \"unBiasedHSIC(matKernXk, matKernY)\"" << endl;
                    }

                    cout << " HSic[" << dHSICXkY << "]"
                         << " R2HSic[" << dR2HSICXkY << "]"
                         << " 1.-PValue[" << 1.0-dPValue << "]" << endl;
                }

                /// Fill the values
                fillIndex(attinput->GetName(), "R2HSic", dR2HSICXkY, "--first--");
                fillIndex(attinput->GetName(), "HSic", dHSICXkY , "--hsic--", dHSICXkY -3.0 * dBiask, dHSICXkY +3.0 * dBiask);
                fillIndex(attinput->GetName(), "pValues", dPValue, "--pvalues--");
            }
            cout << " *** End Of case output [" << attouput->GetName() << "]" << endl;

            /// Clean
            // if ( nBoot > 0 ) {
            //     gROOT->Delete("uranie_HSIC_boot_*");
            // }

            /// Clean
            if ( lstIndexBoot != NULL )
                delete lstIndexBoot;
            lstIndexBoot = NULL;
            if ( lstMatrixBoot != NULL )
                delete lstMatrixBoot;
            lstMatrixBoot = NULL;
        }

    }

    ///////////////////////////////////////////////////////////
    ///  printSummary
    ///////////////////////////////////////////////////////////
    void THSIC::printSummary(Option_t *option)
    {
        cout << " *** THSIC::printSummary option[" << option << "]" << endl;
        cout << " *** Number of : Patterns[" << _nS << "] Inputs[" << _nX << "] Outputs[" << _nY << "]" << endl;
        cout << " *** _nThresholdBootstrap[" << _nThresholdBootstrap << "] _nThresholdGamma[" << _nThresholdGamma << "]" << endl;
        cout << " *** _nBootstrapSample[" << _nBootstrapSample << "]" << endl;
        Int_t iCrtx = 1;
        Int_t iCrty = 1;
        Int_t iPat = 1;

        // a local variable for the event type
        char charInput[500];
        _ntresult->GetLeaf("Inp")->SetAddress(&charInput);
        char cOutput[500];
        _ntresult->GetLeaf("Out")->SetAddress(&cOutput);

        TLeaf *leafSobol = _ntresult->GetLeaf("Sobol");

        TIter nextouput(_listOfOutputAttributes);
        while (URANIE::DataServer::TAttribute *attouput =
               (URANIE::DataServer::TAttribute *) nextouput())
        {
            cout << " *** case output  ("<< iCrty++ << "/" << _nY << ") [" << attouput->GetName() << "]" << endl;

            iCrtx = 1;
            TIter nextinput(_listOfInputAttributes);
            while (URANIE::DataServer::TAttribute *attinput =
                   (URANIE::DataServer::TAttribute *) nextinput())
            {
                cout << " ***  ("<< iCrtx++ << "/" << _nX << ") :: ";
                _ntresult->GetEntry(iPat);
                cout << " HSic(" << charInput  << ", " << cOutput << ")[" << leafSobol->GetValue() << "]";
                _ntresult->GetEntry(iPat-1);
                cout << " R2HSic[" << leafSobol->GetValue() << "]";
                _ntresult->GetEntry(iPat+1);
                cout << " 1.0-Pvalue[" << 1.0-leafSobol->GetValue() << "]" << endl;
                // Next line
                iPat+=3;
            }
        }
        cout << " *** End Of THSIC::printSummary[" << option << "]" << endl;
    }

    ///////////////////////////////////////////////////////////
    ///  printLog
    ///////////////////////////////////////////////////////////
    void THSIC::printLog(Option_t *option)
    {
        cout << " ** THSIC::printLog[" << option << "]" << endl;
        cout << " ** _nEstimatedVariance[" << _nEstimatedVariance << "]" << endl;
        cout << " ** _nThresholdBootstrap[" << _nThresholdBootstrap << "] _nThresholdGamma[" << _nThresholdGamma << "]" << endl;
        cout << " ** _nBootstrapSample[" << _nBootstrapSample << "]" << endl;
        TSensitivityTest::printLog(option);
        cout << " ** End of THSIC::printLog[" << option << "]" << endl;
    }

    } // End of namespace Sensitivity
} // End of namespace URANIE

