//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TMorris.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   mer sep 13 12:39:50 CEST 2006
 
 \brief Implementation of the class URANIE::Sensitivity::TMorris

 */

//ROOT
#include "TROOT.h"
#include "TDirectory.h"
#include "TH2.h"
#include "TRegexp.h"
#include "TSystem.h"
#include "TMath.h"
#include "TCanvas.h"
#include "TPaveText.h"
#include "TLatex.h"
#include "TGraph.h"
#include "TAxis.h"
#include "TLeaf.h"
#ifndef ROOT_TVectorD
#include "TVectorD.h"
#endif
#ifndef ROOT_TMatrixF
#include "TMatrixF.h"
#endif

// Uranie
#include "UVersion.h"
#include "TMorris.h"
#include "TLauncherFunction.h"
#include "TLauncher.h"
#include "TStochAttribut.h"

using std::cout;
using std::endl;

ClassImp (URANIE::Sensitivity::TMorris)

namespace URANIE
{
namespace Sensitivity
{

// -----------------------------------------------------------------------
//  Constructors with a name of a function
// -----------------------------------------------------------------------
TMorris::TMorris(URANIE::DataServer::TDataServer *tds, const char *fcn,
        Int_t nReplique, Int_t level, Double_t delta, TString sout) :
        TScreeming(tds, fcn, nReplique, "", sout.Data()), _delta(delta), _level(level),
        _ntdatamorris(NULL), _ntmorrisresult(NULL),_rand(NULL)
{
    stdInit();
}

// -----------------------------------------------------------------------
//  Constructors with a function
// -----------------------------------------------------------------------
#ifdef ROOT5
TMorris::TMorris(URANIE::DataServer::TDataServer *tds, void *fcn,
        Int_t nReplique, Int_t level, Double_t delta, TString sout) :
    TScreeming(tds, fcn, nReplique, "", sout.Data()), 
#else
TMorris::TMorris(URANIE::DataServer::TDataServer *tds,
        void (*fcn)(Double_t*, Double_t*), TString sin, TString sout,
        Int_t nReplique, Int_t level, Double_t delta) :
    TScreeming(tds, fcn, sin.Data(), sout.Data(), nReplique), 
#endif
        _delta(delta), _level(level),
        _ntdatamorris(NULL), _ntmorrisresult(NULL),_rand(NULL)
{
    stdInit();
}

// -----------------------------------------------------------------------
//  Constructors with a TDataServer filled
// -----------------------------------------------------------------------
TMorris::TMorris(URANIE::DataServer::TDataServer *tds, const char *varexpinput,
        const char *varexpoutput, Option_t* option) :
        TScreeming(tds, varexpinput, varexpoutput, option),
        _delta(-1.0), _level(-1), _ntdatamorris(NULL), _ntmorrisresult(NULL),_rand(NULL)
{
    if (_blog)
    {
        cout << " ********************************" << endl;
        cout << " ** TMorris constructor with TDS filled" << endl;
    }

//     _sIteratorName = Form("morris__n__iter__%s", _tds->GetName());
    _sIteratorName = "__morris__difference__";
    setMethodName("Morris");

    TString stitle = _tds->GetTitle();
    if (_blog)
        cout << " ** Looking for the delta value title[" << stitle << "] ..."
                << endl;

    if (stitle.Contains("with Morris Delta="))
    {
        TRegexp reg("Delta=[0-9].[0-9]*e[-+]?[0-9]*");
        TString out = stitle(reg);
        if (_blog)
            cout << " ** out[" << out << "]" << endl;
        if (out.Length() > 0)
        {
            out.Remove(0, 6);
            _delta = atof(out.Data());
            if (_blog)
                cout << " ** _delta [" << _delta << "]" << endl;
        }
    }

    if (stitle.Contains("Level="))
    {
        TRegexp reg("Level=[0-9]*");
        TString out = stitle(reg);
        if (_blog)
            cout << " ** out[" << out << "]" << endl;
        if (out.Length() > 0)
        {
            out.Remove(0, 6);
            _level = atoi(out.Data());
            if (_blog)
                cout << " ** _level [" << _level << "]" << endl;
        }
    }

    try {

        // Search the number of replica
        Int_t nx = _tds->getNPatterns ();
        if (_blog)
            cout << " nx[" << nx << "]"<<endl;

        TString sInternalReplica = "__npoints__";
        if ( _tds->isAttribute(sInternalReplica) == kFALSE  )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                      Form("TMorris::constructor with TDS filled ERROR\n the internal attribute \"%s\" is not found in the TDS",
                                                              sInternalReplica.Data()));
        tds->computeStatistic (sInternalReplica);
        _nS = (Int_t) tds->getAttribute (sInternalReplica)->getMaximum()+1;

        if (_delta < 0.0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TMorris::constructor with TDS filled ERROR\n the delta value is not well defined in the title [%s]\n Idea : scientific format x.xxxxe-+xx",
                            stitle.Data()));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

// -----------------------------------------------------------------------
//  Constructors with a TCode
// -----------------------------------------------------------------------
TMorris::TMorris(URANIE::DataServer::TDataServer *tds,
        URANIE::Launcher::TCode *fcode, Int_t nReplique, Int_t level,
        Double_t delta) :
    TScreeming(tds, fcode, nReplique), 
    _delta(delta), _level(level), _ntdatamorris(NULL), _ntmorrisresult(NULL)
{
    stdInit();
}

// -----------------------------------------------------------------------
//  Constructors with a TRun
// -----------------------------------------------------------------------
TMorris::TMorris(URANIE::DataServer::TDataServer *tds,
        URANIE::Relauncher::TRun *run, Int_t nReplique, Int_t level,
        Double_t delta) :
    TScreeming(tds, run, nReplique), 
    _delta(delta), _level(level), _ntdatamorris(NULL), _ntmorrisresult(NULL),_rand(NULL)
{
    stdInit();
}

void TMorris::stdInit()
{
    if (_blog){
        cout << " ********************************" << endl;
        cout << " ** TMorris constructor" << endl;
    }

    _sMethodName = TString("Morris");
    _sIteratorName = "__morris__difference__";

    try {
        if ( _nS  < 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TMorris::constructor with TCode\n the number of replica is negative[%d]", _nS));
        if ( _level  < 2)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form("TMorris::constructor with TCode\n the level[%d] must be greater than 2", _level));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    //// Loop on stochastic attributes to check that no infinite distri are unbounded with Halton.
    URANIE::DataServer::TStochasticAttribute *stoatt=NULL;
    TIter next(_listOfInputAttributes);
    Int_t j = 0;
    TString AttFound="";
    URANIE::DataServer::TInfiniteDistribution* test=NULL;
    while ( (stoatt = (URANIE::DataServer::TStochasticAttribute *) next()) )
    {
        test=dynamic_cast<URANIE::DataServer::TInfiniteDistribution *>(stoatt);
        if ( test != 0 && (!test->isLowerBounded() || !test->isUpperBounded()) ) 
            AttFound+=test->GetName()+TString(":");
    }
    try{
        if( AttFound != "")
        {
            AttFound=AttFound.Strip(TString::EStripType::kTrailing,':');
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form(" Morris method is not meant to be run with unbounded infinite stochastic laws [%s].\n \
The sampling can indeed, AND WILL CERTAINLY, randomly draw a probability of 0 or 1 in the [0,1]^{nX} hypercube (see methodology). \n \
The theoretical answer, given an unbounded infinite stochastic laws, is repectively - infinite or + infinite.\n \
Please, if you want to use infinite stochastic laws, use the setBounds(lower,upper) method to truncate the law.\n \
This would allow you to define the proper physical space in which your problem is defined.\n \
!!!! WARNING !!!:\n \
 Bear in mind that the bound values will very likely be used, so choose them wisely (for a gaussian, few sigma away from the mean but not much).\n \
!!!! WARNING !!!.",AttFound.Data()));
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    if (_blog)
    {
        cout << " ** End of TMorris constructor" << endl;
        cout << " ********************************" << endl;
    }
}

TMorris::~TMorris()
{
    if( gROOT->GetList() != 0 &&  !gROOT->GetList()->IsEmpty())
    {
	if (_ntdatamorris != NULL)
	    delete _ntdatamorris;
	_ntdatamorris = NULL;
    
    
	if (_ntmorrisresult != NULL)
	    delete _ntmorrisresult;
	_ntmorrisresult = NULL;
    }

    if( _rand != NULL)
        delete _rand;
    _rand=NULL;
    
}

/*****************************************************/
/*  Create the TNtuple                               */
/*****************************************************/
void TMorris::createTuple()
{
    if (_blog)
    {
        cout << endl;
        cout << "*******************************" << endl;
        cout << "** TMorris::createTuple  ******" << endl;
    }

    //Les variables internes � la m�thode de Morris
    TString sVariableNames = _tds->getIteratorName();

    URANIE::DataServer::TAttribute* attMorris =
            new URANIE::DataServer::TAttribute("__npoints__");
    attMorris->setOrigin(URANIE::DataServer::TAttribute::kAttribute); // kInternal / kAttribute
    attMorris->delShare();
    _tds->addAttribute(attMorris);
    sVariableNames += Form(":%s", attMorris->GetName());

    attMorris = new URANIE::DataServer::TAttribute("__directions__");
    attMorris->setOrigin(URANIE::DataServer::TAttribute::kAttribute); // kInternal / kAttribute
    attMorris->delShare();
    _tds->addAttribute(attMorris);
    sVariableNames += Form(":%s", attMorris->GetName());

    attMorris = new URANIE::DataServer::TAttribute("__morris__difference__");
    attMorris->setOrigin(URANIE::DataServer::TAttribute::kAttribute); // kInternal / kAttribute
    attMorris->delShare();
    _tds->addAttribute(attMorris);
    sVariableNames += Form(":%s", attMorris->GetName());

    TIter next(_listOfInputAttributes);
    while (URANIE::DataServer::TAttribute *att =
            (URANIE::DataServer::TAttribute *) next())
    {
        sVariableNames += Form(":%s", att->GetName());
    }

    if (_blog)
        cout << " ** Input only :: sVariableNames[" << sVariableNames << "]"
                << endl;

    // Construction du tuple des donn�es d'entr�es
    _ntdatamorris = new TNtupleD(Form("tuple_%s", GetName()), GetTitle(),
            sVariableNames);
    _ntdatamorris->SetMarkerColor(2);
    _ntdatamorris->SetMarkerSize(4);
    _ntdatamorris->SetMarkerStyle(7);

    if (_blog)
    {
        cout << "** End Of TMorris::createTuple  ******" << endl;
        cout << "*******************************" << endl;
    }
}

///////////////////////////////////////////////////////////
//  Generate the Sample 
///////////////////////////////////////////////////////////
void TMorris::generateSample(Option_t * option)
{
    if (_blog) {
        cout << " ***********************************" << endl;
        cout << " ** TMorris::generateSample[" << option << "]" << endl;
    }

    try
      {
	if ((_delta == 0) && (TMath::Odd(_level))) 
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form("TMorris::generateSample method\n The delta value is zero [%.1f] and the level is odd [%d]. In odd level, you must specify a delta number in [1, level-1]",
							  _delta, _level));
      }catch( URANIE::Exceptions::UErrorExceptions &ue)
      {
	ue.printMessage();
	throw ue;
      }

    _nX = _tds->getNAttributes();

    if (_delta == 0)
    {
        SetTitle("classical method of Morris");
        _delta = (Double_t)(1.0 * _level / (2.0 * (_level - 1)));
    }
    else
      {
        Int_t ndelta = Int_t(_delta);
        if (_blog)
            cout << " ** ndelta[" << ndelta << "]"<< endl;
        if ( ndelta < 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form("TMorris::generateSample method\n The value of delta is negative [%d]", ndelta));
        if ( ndelta > _level-1 )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form("TMorris::generateSample method\n The delta value [%d] is greater than level-1 [%d]",
                            ndelta, _level-1));
        _delta = (Double_t)(1.0 * ndelta / (_level - 1.0));
        if (_blog)
            cout << " ** _delta[" << _delta << "]"<< endl;
        SetTitle(Form("Morris method with delta[%f]", _delta));
    }

    _tds->SetTitle(Form("%s - Uranie v%s c%d - with Morris Delta=%e  Level=%d", _tds->GetTitle(),
			URANIE_RELEASE, URANIE_VERSION_CODE, _delta, _level));

    if (_blog)
        cout << " *** TDS Title[" << _tds->GetTitle() << "]" << endl;

    if (_ntdatamorris == NULL)
        createTuple();

    // calcul du vecteur de la grille \f$ \[ 0, 1.0/(p-1), 2.0/(p-1), \cdots, 1.0\]\f$
    Double_t *valLevel = new Double_t[_level];
    for (Int_t iter = 0; iter < _level; iter++)
    {
        valLevel[iter] = (Double_t)(iter) / ((Double_t)(_level - 1));
    }

    Int_t niter = 0;
    // Number of internal attributes ( i.e. attribute with name follows \em __*__.)
    // and the iterator attribute
    Int_t nVarMorris = 4;
    Double_t dproba;

    URANIE::DataServer::TStochasticAttribute *curdist;

    Double_t *data = new Double_t[_nX + 4];

    // Loop on the replica
    _rand = new TRandom2(_nSeed);
    _matB.ResizeTo(_nX + 1, _nX);
    _vecDirection.ResizeTo (_nX+1);

    gRandom->SetSeed (_nSeed);


    for (Int_t ireplique = 0; ireplique < _nS; ireplique++)
    {
        // Compute the orientation matrix
        // and the vector of directions
        if (_blog)
            cout << " ** Replica[" << ireplique + 1 << "]/[" << _nS << "]" << endl;
        computeOrientationMatrix();
        if (_blog)
            cout << endl;

        // Store the trajectory
        data[1] = ireplique;
        for (Int_t jdirection = 0; jdirection < _nX + 1; jdirection++)
        {
            data[2] = _vecDirection(jdirection);
            data[0] = niter;
            data[3] = -1;

            // Generate the Design of Experiment (DoE)
            for (Int_t j = 0; j < _nX; j++) {
                dproba = _matB(jdirection, j); //valLevel[(Int_t)(_matB(jdirection, j))];
                curdist =
                        (URANIE::DataServer::TStochasticAttribute *) (_listOfInputAttributes->At(j));

                data[nVarMorris + j] = curdist->getInvCDF(dproba);

                // Looking for the indexe which is modified
                if ((jdirection > 0)
                        && (_matB(jdirection, j) != _matB(jdirection - 1, j)))
                {
                    data[3] = j;
                }
            }

            // Fill the data
            _ntdatamorris->Fill(data);
            niter++;
        }
    }
    _tds->setTuple(_ntdatamorris);
    if (_blog)
    {
      _ntdatamorris->Scan("*");
        cout << " ** End of TMorris::generateSample" << endl;
        cout << "***********************************" << endl;
    }

    // Clean
    delete[] valLevel;
    valLevel = NULL;
    delete[] data;
    data = NULL;
}

  //Pre-treatment check whether the ntuple is created... If not, produce it
  void TMorris::preTreatment()
  {
    if( !_ntmorrisresult )
      {
	_ntmorrisresult = new  URANIE::DataServer::TDSNtupleD("__uranie__morris__results__", "Result of morris method", 
							      "Input:Output:mu:mustar:sigma", "S:S:D:D:D");
	_ntmorrisresult->SetMarkerColor(2);
	_ntmorrisresult->SetMarkerSize(4);
	_ntmorrisresult->SetMarkerStyle(7);
      }    
  }

///////////////////////////////////////////////////////////
//  evaluateIndexes 
///////////////////////////////////////////////////////////
void TMorris::evaluateIndexes(Option_t * option)
{
    Int_t npat = (Int_t)(_tds->getNPatterns());
    if (_blog) {
        cout << " ***********************************" << endl;
        cout << " ** TMorris::evaluateIndexes[" << option << "] npat[" << npat
                << "]" << endl;
	_tds->getTuple()->Scan("*");
    }

    string soutputy = "";
    TLeaf *ldirections = _tds->getTuple()->GetLeaf("__morris__difference__");
    Double_t dvalcrt, dvalold=0.0, ee;

    Int_t ndirection;

    TString soption = option;
    soption.ToLower();
    bool babsolute = ((soption.Contains("relative") ) ? false : true);
    
    TMatrixD matEE(_nX, 3);
    TMatrixD matRep(_nX,1);
    Int_t iRep;
    TString sinputx;

    Double_t dresultmorris[5];
    Double_t dXDelta;
    Double_t sign_delta ;

    sprintf(_sMethod, "%s", _sMethodName.Data());

    // \warning case on only one output
    soutputy = ((_nElY==1) ? string(_sOutputAttribute ) : string(_sOutputAttribute ).substr(0, string(_sOutputAttribute).find_first_of("[")) );;
    
    if (_blog)
      cout << " ** Case of Output atty [" << soutputy << "]" << endl;

    matEE.Zero();
    matRep.Zero();

    // Procedure to have the correct iterator for Relauncher multi-thread
    double *iter;
    int nbData = _tds->getTuple()->Draw(_tds->getIteratorName(),Form("%s==0",_tds->getIteratorName()),"goff");
    int shift=((nbData==0) ? 1 : 0 );
    _tds->getTuple()->Draw(_tds->getIteratorName(),"","goff");
    iter=_tds->getTuple()->GetV1();

    double locMax=TMath::MaxElement(npat, iter);
    int locMax_int=int(locMax);
    // XYZ<<"Create IteVal with "<<locMax_int<<endl;
    vector<int> IteVal(locMax_int+1,-2);
    map<string,double> DefaultVal;

    // if(_code != NULL)
    // {
    //     for(int ioutfile=0; ioutfile< _code->getNOutputFiles(); ioutfile++)
    //         for(int iattoutfile=0; iattoutfile< _code->getNOutputAttributes(ioutfile); iattoutfile++)
    //         {
    //             URANIE::DataServer::TAttribute *l_output = _code->getOutputFile(ioutfile)->getAttribute(iattoutfile);
    //             double pouetpouet=1.23456789;
    //             if(l_output->hasDefault())
    //                 l_output->getDefaultValue(pouetpouet);
    //             DefaultVal[string(l_output->GetName())]=pouetpouet;
    //             cout<<"Default value of "<<soutputy<<" = "<<DefaultVal[string(l_output->GetName())]<<endl;
    //         }
    // }

    TIter nextOut(_listOfOutputAttributes);
    while (URANIE::DataServer::TAttribute *att =
            (URANIE::DataServer::TAttribute *) nextOut()) {
        double pouetpouet=1.23456789;
        if(att->hasDefault())
            att->getDefaultValue(pouetpouet);
        DefaultVal[string(att->GetName())]=pouetpouet;
        // cout<<"Default value of "<<soutputy<<" = "<<DefaultVal[string(att->GetName())]<<endl;
    }

    for (Int_t ipat = 0; ipat < npat; ipat++)
    {
        if(_blog)
            cout << "TMorris :: filling the vector's [" << iter[ipat]-shift << "]th element with [" << ipat  << "]"<< endl;
        IteVal.at(iter[ipat]-shift)=ipat;
    }

    for (Int_t ipat = 0; ipat < int(IteVal.size()); ipat++) //Untouched by JBB 
    {
	 if (_blog)
            cout << endl << "***************************" << endl
                 << " ** ipat(" << ipat << "/" << npat << ")" << endl;
	
	_tds->getTuple()->GetEntry(IteVal.at(ipat));

	// The direction
	ndirection = (Int_t)(ldirections->GetValue(0));

        if (IteVal.at(ipat) == -2)
        {
            cout<<"This event ["<<ipat<<"] is missing from the Relauncher architecture :: not using this event"<<endl;
            continue;
        }
        
	// the current difference
	dvalcrt = ((_nElY==1) ? _tds->getValue(soutputy.c_str(),IteVal.at(ipat)) : _tds->getValue(soutputy.c_str(),IteVal.at(ipat),_iely));

        //cout<<"This event ["<<ipat<<"/"<<IteVal.at(ipat)<<"] compares moved value ["<<dvalcrt<<"] and reference ["<<dvalold<<"] with direction ["<<ndirection<<"]     ";

        // Removing the code specificities because of TMorris from existing samples
        // if( _code  )
        // {            
        if( dvalcrt == DefaultVal[soutputy] ||  dvalold == DefaultVal[soutputy] )
        {
            cout<<"This event ["<<ipat<<"] is missing from the Launcher architecture :: moved value ["<<dvalcrt<<"] or the reference ["<<dvalold<<"] is equal to default, ("
                <<DefaultVal[soutputy]<<")  :: not using this event"<<endl;
            dvalold=dvalcrt;
            continue;
        }
        // }

                
	if (ndirection > -1)
        {
            if (IteVal.at(ipat-1) == -2)
            {
                cout<<"This event ["<<ipat<<"] has no previous event from the Relauncher architecture :: not using this event"<<endl;
                dvalold=dvalcrt;
                continue;
            }

            //cout<<"This event contains ["<<dvalcrt-dvalold<<"]"<<endl;
                    
            sinputx  = Form("%s", ((URANIE::DataServer::TAttribute *)(_listOfInputAttributes->At(ndirection)))->GetName());
            if (_blog)
                cout << " ** ndirection[" << ndirection << "] Attribute[" << sinputx << "]"
                     << " X Values : Current [" << _tds->getValue(sinputx,IteVal.at(ipat)) << "]"
                     << " Previous[" << _tds->getValue(sinputx,IteVal.at(ipat-1)) << "]"
                     << " Curent iterator["<< _tds->getValue(_tds->getIteratorName(),IteVal.at(ipat)) << "]"
                     << " previous one["<<_tds->getValue(_tds->getIteratorName(),IteVal.at(ipat-1))<<"]" << endl;

	    dXDelta = _tds->getValue(sinputx,IteVal.at(ipat)) -_tds->getValue(sinputx,IteVal.at(ipat-1));

            try{
                if( (_tds->getValue(_tds->getIteratorName(),IteVal.at(ipat)) - _tds->getValue(_tds->getIteratorName(),IteVal.at(ipat-1))) != 1 )
                {                
                    cout << " ** ndirection[" << ndirection << "] Attribute[" << sinputx << "]"
                         << " X Values : Current [" << _tds->getValue(sinputx,IteVal.at(ipat)) << "]"
                         << " Previous[" << _tds->getValue(sinputx,IteVal.at(ipat-1)) << "]"
                         << " Curent iterator["<< _tds->getValue(_tds->getIteratorName(),IteVal.at(ipat)) << "]"
                         << " previous one["<<_tds->getValue(_tds->getIteratorName(),IteVal.at(ipat-1))<<"]" << endl;
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                               "TMorris::computeIndexes. Problem in the sorting of the event");
                    continue;
                }
            }catch(URANIE::Exceptions::UErrorExceptions &ue)
            {
                ue.printMessage();
                throw ue;
            }
                
	    if (_blog)
                cout << " XDelta [" << dXDelta << "]";
	    if ( dXDelta < 0.0 )
                sign_delta = -1.0;
	    else
                sign_delta = 1.0;
	    if (_blog)
                cout << " Sign of XDelta[" << sign_delta << "] " << endl;

	    ee = sign_delta * (dvalcrt - dvalold) / fabs(dXDelta) * ( (babsolute) ?  fabs(dXDelta) / _delta : 1); //    || dXDelta

            //cout<<"  This event contains ["<<dvalcrt-dvalold<<"]  <=>  "<<ee<<endl;
             
	    //                ee = (dvalcrt - dvalold) / dXDelta;
	    if (_blog)
                cout << " ** Y Values : Current [" << dvalcrt << "]"
                     << " Previous [" << dvalold << "]"
                     << " Diff [" << dvalcrt - dvalold << "]"
                     << " _delta[" << _delta << "]"
                     << " XDelta [" << dXDelta << "]"
                     << " - ee[" << ee << "]" 
                     << " - Altee[" << sign_delta * (dvalcrt - dvalold) / _delta << "]" << endl;
              
	    iRep = Int_t(matRep(ndirection,0));
             
	    // case :: mean of ee
	    matEE(ndirection, 0) *= (iRep) / (iRep + 1.0);
	    matEE(ndirection, 0) += ee / (iRep + 1.0);
                
	    // case :: variance=sigma**2 of ee
	    if (iRep > 0)
            {
		matEE(ndirection, 1) *= (iRep - 1.0)
                    / (iRep);
		matEE(ndirection, 1) += (iRep + 1.0)
                    * (ee - matEE(ndirection, 0))
                    * (ee - matEE(ndirection, 0))
                    / (iRep * iRep);
            }

	    // case :: mean of abs(ee)
	    matEE(ndirection, 2) *= (iRep) / (iRep + 1.0);
	    matEE(ndirection, 2) += TMath::Abs(ee) / (iRep + 1.0);

	    if (_blog)
                matEE.Print();
                
	    matRep(ndirection,0) += 1.0;

        }
	else
        {
	    ee = -1.0;
	    if (_blog)
                cout << " ****** ndirection[" << ndirection << "] ee[" << ee
                     << "] dvalcrt[" << dvalcrt << "]" << endl;
        }
	dvalold = dvalcrt;
    }
    if (_blog)
      {
	cout << "** vary[" << soutputy << "]" << endl;
	matEE.Print();
      }

    // The sum of the absolute value of EE (elementary effect)
    Double_t dsum = 0.0;
    for (Int_t i = 0; i < _nX; i++)
      dsum += matEE(i, 2);

    // Fill the indexes
    for (Int_t i = 0; i < _nX; i++)
      {

	dresultmorris[0] = i;
	dresultmorris[1] = _iely;
	dresultmorris[2] = matEE(i, 0);
	dresultmorris[3] = matEE(i, 2);
	dresultmorris[4] = TMath::Sqrt(matEE(i, 1));
	//           dresultmorris[iy*2+1] = matEE(i, 0);
	//           dresultmorris[iy*2+2] = TMath::Sqrt(matEE(i, 1));

	URANIE::DataServer::TAttribute *att =
	  (URANIE::DataServer::TAttribute *) (_listOfInputAttributes->At(i));

	//Fill all results at once in the ntuple
	stringstream sstr; sstr.precision(SSTR_PREC);
	sstr << att->GetName() << " "<< _sOutputAttribute << " " << dresultmorris[2] << " " << dresultmorris[3] << " " << dresultmorris[4];
	_ntmorrisresult->Fill( sstr.str() );

	fillIndex(att->GetName(), "mu", dresultmorris[2], "Morris91");
	fillIndex(att->GetName(), "mustar", dresultmorris[3], "Campolongo07");
	fillIndex(att->GetName(), "sigma", dresultmorris[4], "Morris91");
	fillIndex(att->GetName(), "First", matEE(i, 2) / dsum, "--first--");  // "--first--"
	fillIndex(att->GetName(), "Total", matEE(i, 2) / dsum, "--total--");  // "--total--"

      }
    if (_blog)
      {
	_ntresult->Scan("*");
	_ntmorrisresult->Scan("*");
        cout << " ** End of TMorris::evaluateIndexes" << endl;
        cout << "***********************************" << endl;
      }
}
///////////////////////////////////////////////////////////
//  Compute the orientation matrix
///////////////////////////////////////////////////////////
void TMorris::computeOrientationMatrix()
{
    if ( _blog ) {
        cout << " ** TMorris::computeOrientationMatrix with Delta["<< _delta << "]"<<endl;
        gRandom->Print ("random3");
        cout << " ** nSeed[" << _nSeed << "]" << endl;
    }

    _vecDirection.Zero ();
    _vecDirection(0) = -1.0;

    if ( _blog ) {
        cout << " ** vector of Direction";
        _vecDirection.Print ("Direction");
    }
    TMatrixD matB(_nX+1, _nX);
    matB.Zero ();
    TMatrixD matJ(_nX+1, _nX);
    matJ.UnitMatrix ();

    TMatrixD matDStar(_nX, _nX);
    matDStar.Zero ();

    TMatrixD matPStar(_nX, _nX);
    matPStar.UnitMatrix ();


    // Initialize the matrix B, J and D
    for (Int_t inx = 0; inx < _nX; inx++) {

        for(Int_t j=0; j <inx; j++) {
            matB(inx, j) = 1.0;
            matJ(inx, j) = 1.0;
            matJ(j, inx) = 1.0;
        }
        for(Int_t j=0; j <_nX; j++) {
            matJ(_nX, j) = 1.0;
            matB(_nX, j) = 1.0;
        }

        Double_t drand = gRandom->Uniform(0.0, 1.0);
        if ( _blog )
            cout << " ** drand[" << drand << "]" << endl;
        if ( drand < 0.5 )
            matDStar(inx, inx) = 1.0;
        else
            matDStar(inx, inx) = -1.0;
    }

    if ( _blog ) {
        cout << " ** matrix J";
        matJ.Print ("kkk");
        cout << " ** matrix B";
        matB.Print ("kkk");
        cout << " ** matrix DStar";
        matDStar.Print ("matrix DStar");
        cout << " ** matrix PStar";
        matPStar.Print ("matrix PStar");
    }

    // Transform the matrix PStar in permuatation matrix
    // Create a permutation with the Devroye Algorithm
    Int_t * nind = new Int_t[_nX];
    Int_t is, valtmp, indtmp;
    // Initialize permutation
    for (is = 0; is < _nX; is++)
        nind[is] = is;
    for (is = _nX - 1; is > 0; is--)
    {
        if ( _blog ) {
            for(Int_t i=0; i<_nX; i++ )
                cout << " ** nind[" << i << "] = [" << nind[i] << "]" << endl;
        }
        indtmp = (Int_t) gRandom->Integer(is);
        valtmp = nind[indtmp];
        if (_blog)
            cout << "** is[" << is << "] val [" << nind[is] << "] indtmp["
                    << indtmp << "] val[" << nind[indtmp] << "]";
        nind[indtmp] = nind[is];
        nind[is] = valtmp;
        if (_blog)
            cout << "************* is[" << is << "] val [" << nind[is]
                    << "] indtmp[" << indtmp << "] val[" << nind[indtmp]
                    << "]" << endl;
        for(Int_t i=0; i<matPStar.GetNrows (); i++) {
            Double_t dvaltmp = matPStar( nind[is], i);
            matPStar( nind[is], i) = matPStar( nind[indtmp], i);
            matPStar( nind[indtmp], i) = dvaltmp;
        }
        if ( _blog ) {
            cout << " ** matrix PStar";
            matPStar.Print ("matrix PStar");
        }
    }

    if ( _blog ) {
        cout << " ** matrix PStar";
        matPStar.Print ("matrix PStar");
    }

    // Set the vector of Direction
    for(Int_t i=0; i<matPStar.GetNrows (); i++) {
        for(Int_t j=0; j<matPStar.GetNcols (); j++) {
            if ( matPStar(i, j) == 1 ) {
                _vecDirection(i+1) = 1.0*j;
            }
        }
    }

    TMatrixD matBStar = (_delta / 2.0) * ( ( 2.0 * matB - matJ ) * matDStar + matJ ) * matPStar;
    if ( _blog ) {
        cout << " ** matrix BStar";
        matBStar.Print ("matrix BStar");
    }

    // Compute k0 as \frac{k0}{p-1} + \delta = 1.0
    Int_t nK0 = TMath::FloorNint ((_level - 1.0)*(1.0 - _delta))+1;
    if (_blog)
        cout << " ** nKO=[" << nK0 << "]" << endl;

    // Random the origin point x^\star
    if (_blog)
        cout << " ** xstar [";
    Double_t *xstar = new Double_t[_nX];
    for (Int_t iter = 0; iter < _nX; iter++)
    {
        xstar[iter] = (Double_t)(_rand->Integer(nK0))/((Double_t)(_level - 1));
        if (_blog)
        {
            cout << xstar[iter] << ", value(" << xstar[iter] << ") ";
            if (iter < _nX - 1)
                cout << "- ";
        }
    }
    if (_blog)
        cout << "]" << endl;

    // Generate the design B = x^\star + matB
    for(Int_t i=0; i<_matB.GetNrows (); i++){
        for(Int_t j=0; j<_matB.GetNcols (); j++){
            _matB(i, j) = matBStar(i, j) + xstar[j];
        }
    }

    if ( _blog ) {
        cout << " ** matrix _matB";
        _matB.Print ("matrix _matB");
        cout << " ** vector of Direction";
        _vecDirection.Print ("Direction");
        cout << " ** End Of TMorris::computeOrientationMatrix with Delta["<< _delta << "]"<<endl<<endl<<endl;
    }
    
    // Clean
    delete[] xstar;
    xstar = NULL;
    
    delete[] nind;
    nind = NULL;
    
}
///////////////////////////////////////////////////////////
//  Compute the orientation matrix 
///////////////////////////////////////////////////////////
void TMorris::computeOrientationMatrixOld()
{
    Double_t nlevel = (Double_t)(_level / 2.0);

    _matB.Zero();

    if (_blog)
        cout << " ** xstar [";
    // Random the first point \f$ x^{\star}\f$
    Int_t nborne = (Int_t)(_level / 2);
    Double_t *xstar = new Double_t[_nX];
    for (Int_t iter = 0; iter < _nX; iter++)
    {
        xstar[iter] = _rand->Integer(nborne);
        if (_blog)
        {
            cout << xstar[iter];
            if (iter < _nX - 1)
                cout << ", ";
        }
    }
    if (_blog) {
        cout << "]" << endl;
        cout << " ** with nborne [" << nborne << "]"<< endl;
    }

    // Calcul de \f$ x^{\star} + \Delta \tilde{B}\f$
    UInt_t rand01;
    for (Int_t iterj = 0; iterj < _nX; iterj++)
    {
        // tirage un entier compris entre 0 et 1
        rand01 = _rand->Integer(2);
        for (Int_t iteri = 0; iteri < _nX + 1; iteri++)
        {
            if (iterj < iteri)
                _matB(iteri, iterj) = (1.0 - (Double_t)(rand01)) * nlevel
                        + xstar[iterj];
            else
                _matB(iteri, iterj) = (Double_t)(rand01) * nlevel
                        + xstar[iterj];
        }
    }

    // Permut two columns
    UInt_t randi, randii;
    Int_t nrand = _nX, crtrand = 0;
    Double_t valtemp;
    for (Int_t irand = 0; irand < nrand; irand++)
    {
        // random two integer between 0 and _nX-1
        randi = _rand->Integer(_nX);
        randii = _rand->Integer(_nX);
        if (randi != randii)
        {
            crtrand++;
            for (Int_t iteri = 0; iteri < _nX + 1; iteri++)
            {
                valtemp = _matB(iteri, randi);
                _matB(iteri, randi) = _matB(iteri, randii);
                _matB(iteri, randii) = valtemp;
            }
        }
    }

    // Clean
    delete[] xstar;
    xstar = NULL;

    if (_blog)
        _matB.Print();
}

///////////////////////////////////////////////////////////
//  Draw the indexes
///////////////////////////////////////////////////////////
void TMorris::drawIndexes(TString sTitre, const char *select, Option_t * option, bool internalCall)
{

    if ( _blog )
        cout << " ** TMorris::drawIndexes for method[" << _sMethodName << "]" << endl;

    try {
        if (_ntresult->GetEntries() == 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form("TMorris::drawIndexes method\n the TNtupleD of results for [%s] is empty",
							    this->GetTitle ()));
    } catch (URANIE::Exceptions::UErrorExceptions& ue) {
        ue.printMessage();
        throw ue;
    }

    // Hack to be able to get index plot in Python.
    if( strcmp(select,"")==0 )//&& string(option)=="" )
    {
	if( sTitre == "" || sTitre == "mu" || sTitre == "mustar" )
	{
	    drawIndexes( ((sTitre=="") ? (Option_t*)((TString("mu,")+TString(option)).Data()) : (Option_t*)((sTitre+TString(",")+TString(option)).Data()) ));
	    return;
	}
    }

    
    // Treat the option
    TString sSelectMorrisScore ="";
    TString opt = option;
    opt.ToLower();
    if ( opt.Contains("hist") )
    {
        if ( opt.Contains("mustar") )
            sSelectMorrisScore = "mustar";
        else if ( opt.Contains("sigma") )
            sSelectMorrisScore = "sigma";
        else if ( opt.Contains("mu") )
            sSelectMorrisScore = "mu";
        if ( _blog )
            cout << " ** sSelectMorrisScore[" << sSelectMorrisScore << "]" << endl;
    }

    if( !internalCall )
    {
	bool newcan=true;
	if(opt.Contains("nonewcanv"))
	    newcan=false;
	checkCanvasCreation(newcan);
    }
    
    //////////////////////////////
    TH1F *hDivScore = NULL;
    //TLeaf *leafx = _ntresult->GetLeaf("Sobol");
    URANIE::DataServer::TAttribute *att;

    _ntresult->Draw("Inp>>hDivScore", Form("(Order==\"%s\")&&(Inp!=\"__sum__\")",sSelectMorrisScore.Data()), "goff");
    hDivScore = (TH1F*) gDirectory->Get("hDivScore");
    hDivScore->SetDirectory(0);
    hDivScore->SetStats(0);       	
    hDivScore->SetName (Form("Morris_%s_%i",sSelectMorrisScore.Data(),_drawingGarbageCollector->GetEntries()));


    if (hDivScore) {
	hDivScore->Reset();
	hDivScore->SetBarWidth(0.45);
	hDivScore->SetBarOffset(0.1);
	hDivScore->SetFillColor(49);
	hDivScore->SetTitle(sTitre);
	for (Int_t i = 0; i < _nX; i++)
	{

	    att = (URANIE::DataServer::TAttribute *) (_listOfInputAttributes->At(i));
	    _ntresult->Draw("Sobol",
			    Form("((Inp==\"%s\")&&(Order==\"%s\"))", att->GetName(), sSelectMorrisScore.Data()),
			    "goff");

	    Double_t *vv = _ntresult->GetV1();
	    hDivScore->SetBinContent(i + 1, vv[0]);

	}

	hDivScore->Draw("bar2");
	hDivScore->GetYaxis ()->SetTitle (Form("#%s", sSelectMorrisScore.Data ()));
	hDivScore->LabelsOption(">u");
	_drawingGarbageCollector->Add(hDivScore);
    }
    gPad->SetGridy(1);
    gPad->Modified();
    gPad->Update();


}

///////////////////////////////////////////////////////////
//  Draw the indexes 
///////////////////////////////////////////////////////////
void TMorris::drawIndexes(Option_t * option, bool internalCall)
{

  Int_t npat = (Int_t) _ntmorrisresult->GetEntries();
  if (_blog)
    {
      cout << " *******************************" << endl;
      cout << " ** TMorris::drawIndexes option[" << option << "]  _nX[" << _nX
	   << "] _nY[" << _nY << "] npat[" << npat << "]******" << endl;
    }

  TString opt = option;
  opt.ToLower();
  
  Bool_t bmustar = kFALSE;
  if (opt.Contains("mustar"))
    bmustar = kTRUE;
  
  TIter nextouput(_listOfOutputAttributes);
  Int_t ncrtoutput = 0;

  if( !internalCall )
  {
      bool newcan=true;
      if(opt.Contains("nonewcanv"))
	  newcan=false;
      checkCanvasCreation(newcan);
      _canvas->Divide(_nY);
  }

  TLeaf *lx = NULL;

  if (bmustar)
    lx = _ntmorrisresult->GetLeaf("mustar");
  else
    lx = _ntmorrisresult->GetLeaf("mu");

  TLeaf *ly = _ntmorrisresult->GetLeaf("sigma");
  TLeaf *linp = _ntmorrisresult->GetLeaf("Input");
  TLeaf *lout = _ntmorrisresult->GetLeaf("Output");

  string outputname="";
  TGraph* gr = NULL; 
  TLatex *tl = NULL;
  stringstream sstr;
  while (URANIE::DataServer::TAttribute *attouput =
	 (URANIE::DataServer::TAttribute *) nextouput())
    {
	if (_blog)
	    cout << " ** case output [" << attouput->GetName() << "]" << endl;

	if(!internalCall)
	    _canvas->cd(ncrtoutput + 1);
      
	// Go to the current pad
	outputname=string(attouput->GetName());
	gr = new TGraph();
	gr->SetName(Form("__%s__%i",outputname.c_str(),_drawingGarbageCollector->GetEntries()));

	gr->SetTitle(outputname.c_str());
	// if( TString(_canvas->GetName()).BeginsWith("__sensitivitycan__"))
	_drawingGarbageCollector->Add(gr);
	int nbpts=0;
	for(unsigned int iel=0; iel<_mAttributeSize[ attouput->GetName() ]; iel++)
	{
	    sstr.str(""); sstr << iel;
	    outputname=((_mAttributeSize[ attouput->GetName() ]==1) ?  string(attouput->GetName()): string(attouput->GetName())+"["+sstr.str()+"]");

	    //	  gr.push_back(new TGraph(_nX));
	    //	  gr.back()->SetTitle(outputname.c_str());
	    for (Int_t ipat = 0; ipat < npat; ipat++)
	    {
		_ntmorrisresult->GetEntry(ipat);
		//Int_t ny = (Int_t)(lout->GetValue());
		if (outputname != *(string*)lout->GetValuePointer())
		    continue;

		Double_t dvalx = lx->GetValue();
		Double_t dvaly = ly->GetValue();
		//Int_t nx = (Int_t)(linp->GetValue());
		string sName = *(string*)linp->GetValuePointer();
		if(_mAttributeSize[ attouput->GetName() ]!=1) sName+="["+sstr.str()+"]";
		if (_blog)
		    cout << ipat << " - name[" << sName << "]" << dvalx << " ::::: "
			 << dvaly << endl;
		gr->SetPoint(nbpts, dvalx, dvaly);
		tl = new TLatex(dvalx * 1.01, dvaly * 1.01, sName.c_str());
		tl->SetTextAngle(22);
		tl->SetTextSize(0.05);
		tl->SetTextColor(kBlack);
		//_drawingGarbageCollector->Add(tl);
		gr->GetListOfFunctions()->Add(tl);
		nbpts++;
	    }

	    if( iel==_mAttributeSize[ attouput->GetName() ]-1)
	    {
		gr->SetMarkerColor(_ntresult->GetMarkerColor());
		gr->SetMarkerStyle(_ntresult->GetMarkerStyle());
		gr->SetMarkerSize(_ntresult->GetMarkerSize());

		gr->Draw("ap");
		if (bmustar)
		    gr->GetXaxis()->SetTitle("#mu^{*}");
		else
		    gr->GetXaxis()->SetTitle("#mu");
		gr->GetYaxis()->SetTitle("#sigma");
	
		gPad->Modified();
		gPad->Update();

		ncrtoutput++;
	    }
	}
    }
  if (_blog)
    {
      cout << " ** End of TMorris::drawIndexes" << endl;
      cout << "***********************************" << endl;
    }

}
///////////////////////////////////////////////////////////
//  Draw the indexes
///////////////////////////////////////////////////////////
void TMorris::drawIndexesAll(Option_t * option)
{
    Int_t npat = (Int_t) _ntmorrisresult->GetEntries();
    Int_t nScore = 3;  // mu + mustar + sigma

    if (_blog) {
        cout << " *******************************" << endl;
        cout << " ** TMorris::drawIndexesAll option[" << option << "]  _nX[" << _nX
             << "] _nY[" << _nY << "] npat[" << npat << "] nScore[" << nScore << "] ******" << endl;
    }

    TIter nextouput(_listOfOutputAttributes);
    //Int_t ncrtoutput = 0;

    TString opt = TString(option);
    bool newcan=true;
    if(opt.Contains("nonewcanv"))
	newcan=false;
    
    checkCanvasCreation(newcan);
    _canvas->Divide(3, 3);
    _canvas->Draw ();
    _canvas->Modified (); _canvas->Update ();


    while (URANIE::DataServer::TAttribute *attouput =
	   (URANIE::DataServer::TAttribute *) nextouput())
    {
        if (_blog)
            cout << " ** case output [" << attouput->GetName() << "]" << endl;

        // Histogram of \mu
        _canvas->cd(1);
        this->drawIndexes ("Histogram of #mu", "", "hist,mu", kTRUE);
        // Histogram of \mu^{star}
        _canvas->cd(5);
        this->drawIndexes ("Histogram of #mu^{*}", "", "hist,mustar", kTRUE);
        // Histogram of \sigma
        _canvas->cd(9);
        this->drawIndexes ("Histogram of #sigma", "", "hist,sigma", kTRUE);

	//Bidim plots
	_canvas->cd(7);
	drawIndexes("mu", kTRUE);
        _canvas->cd(8);
        drawIndexes("mustar", kTRUE);

    }
    _canvas->Modified (); _canvas->Update ();
}

///////////////////////////////////////////////////////////
//  Draw the indexes for python
///////////////////////////////////////////////////////////
void TMorris::pyDrawIndexes(Option_t * option)
{
    this->drawIndexes(option);
}

///////////////////////////////////////////////////////////
//  Draw the sample 
///////////////////////////////////////////////////////////
void TMorris::drawSample(TString svar, Int_t nreplique, Option_t *option)
{

    Int_t ncrtreplica = nreplique - 1;
    if (_blog)
        cout << "** TMorris::drawSample[" << svar << "] replica[" << nreplique
                << "] ******" << endl;
    if (svar.Length() == 0)
        svar = Form("%s:%s", _tds->getAttribute(1)->GetName(),
                _tds->getAttribute(0)->GetName());
    if (_blog)
        cout << "** TMorris::drawSample[" << svar << "] ******" << endl;

    //////////////////////
    // Visualisation
    //////////////////////
    TString opt = TString(option);
    bool newcan=true;
    if(opt.Contains("nonewcanv"))
	newcan=false;
    
    checkCanvasCreation(newcan);
    _canvas->Range(0, 0, 25, 18);
    // Canvas->SetFillColor(42);

    TPaveText *pl = new TPaveText(5, 15.3, 20, 17.5, "br");
    _drawingGarbageCollector->Add(pl);
    TString stitlefunction = "";
    if (_listOfOutputAttributes == NULL)
    {
        stitlefunction = Form("%s DataServer", _tds->GetName());
    }
    else
    {
        stitlefunction =
                Form("%s Function",
                        ((URANIE::DataServer::TAttribute *) (_listOfOutputAttributes->At(
                                0)))->GetName());
    }

    pl->InsertText(stitlefunction);
    pl->InsertText(
            Form("%s Method :  Trajectoty[%d]  - Level [%d]",
                    _sMethodName.Data(), _nS, _level));
    if ((nreplique > 0) && (ncrtreplica < _nS))
        pl->InsertText(Form("Replica %d", nreplique));
    //pl->SetFillColor(18);
    pl->SetTextFont(32);
    pl->SetTextColor(50);
    pl->Draw();

    TPad* pad1 = new TPad("pad1", "Determ", 0.10, 0.02, 0.90, 0.83);//, 33);
    pad1->Draw();
    pad1->cd();

    Color_t nmarkcol = _ntdatamorris->GetMarkerColor();
    Color_t nlinecol = _ntdatamorris->GetLineColor();

    _ntdatamorris->SetLineColor(pad1->GetFillColor());
    _ntdatamorris->SetLineColor(nlinecol);

    if ( nreplique < 0 )
    {
        _ntdatamorris->Draw(Form("%s", svar.Data()), "", "p");
        for (Int_t i = 0; i < _nS; i++)
            _ntdatamorris->Draw(Form("%s", svar.Data()), Form("__npoints__==%d", i), "samelp");
    }
    if ((nreplique > 0) && (ncrtreplica < _nS))
    {
        _ntdatamorris->Draw(Form("%s", svar.Data()), Form("__npoints__==%d", ncrtreplica), "p");
        _ntdatamorris->SetLineColor(kBlue);
        _ntdatamorris->Draw(Form("%s", svar.Data()), Form("__npoints__==%d", ncrtreplica), "lpsame");

    }
    _ntdatamorris->SetMarkerColor(nmarkcol + 1);
    _ntdatamorris->Draw(Form("%s", svar.Data()), "__morris__difference__==-1", "same");

    _ntdatamorris->SetMarkerColor(nmarkcol);
    _ntdatamorris->SetLineColor(nlinecol);
}


TTree *TMorris::getResultTuple(bool commonresulttuple)
{
    try{
        if(commonresulttuple)
            throw URANIE::Exceptions::UInfoExceptions(__FILE__, __LINE__,
                                                      "TMorris::getResultTuple. You've asked to get the \"classical\" result file by calling getResultTuple(kTRUE).\n\
Be aware that the fact that in this file the numerical results are stored in a variable named Sobol, as nothing to do with their nature (mu, mustar and sigma).\n\
In order to prevent from any mis-conclusions, we strongly recommend to stick to the classical Morris output file by calling either getMorrisResults() or getResultTuple().");
    }catch (URANIE::Exceptions::UInfoExceptions& ue)
    {
        ue.printMessage();
    }

    return ((commonresulttuple) ? _ntresult : _ntmorrisresult);
}

///////////////////////////////////////////////////////////
//  Log 
///////////////////////////////////////////////////////////
void TMorris::printLog(Option_t *option)
{
    cout << "TMorris::printLog[" << option << "]" << endl;
    cout << "_delta[" << _delta << "]" << endl;
    cout << "_level[" << _level << "]" << endl;
    cout << "Number of replica[" << _nS << "]" << endl;
    URANIE::Sensitivity::TScreeming::printLog(option);
    cout << "TMorris::fin de printLog[" << option << "]" << endl;
}

} // Fin du namespace Sensitivity
} // Fin du namespace URANIE

