//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
// TSensitivity jeu sep 14 09:01:14 CEST 2006
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
  \file   TSensitivity.cxx
  \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
  \date   thur Sept14 09:01:14 CEST 2006

  \brief Implementation of the class URANIE::Sensitivity::TSensitivity


*/

// ROOT
#ifdef ROOT5
#include "Api.h"
#endif
#include "TCanvas.h"
#include "TDirectory.h"
#include "TH2.h"
#include "TLeaf.h"
#include "TLegend.h"
#include "TPie.h"

// Uranie
#include "TSensitivity.h"
#include "TLauncher.h"
#include "TLauncherFunction.h"
#include "TEval.h"
#include "TLauncher2.h"
#ifdef WITH_LIBSSH
#include "TLauncherByStepRemote.h"
#endif

#ifdef ROOT5
URANIE::Sensitivity::TSensitivity *gSensitivity;
#endif

ClassImp (URANIE::Sensitivity::TSensitivity)

namespace URANIE
{
namespace Sensitivity
{

#ifdef ROOT5
//______________________________________________________________________________
void SensitivityInteractiveMethod(Double_t *param, Double_t *out)
{
    ///===================================================================================================//
    ///************** Static function called when FCNDecode is called in interactive mode ****************//
    ///===================================================================================================//
    TMethodCall *m = gSensitivity->GetMethodCall();
    if (!m)
        return;
    Long_t args[2];
    args[0] = (Long_t) param;
    args[1] = (Long_t) out;
    m->SetParamPtrs(args);
    Double_t result;
    m->Execute(result);
}
#endif

// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------
TSensitivity::TSensitivity() :
     _bisInputCorrelated(kFALSE), _bgoingThroughError(kFALSE),
    _tds(NULL), _listOfInputAttributes(NULL), _listOfOutputAttributes(NULL), _sTimeAttribute(""), 
#ifdef WITH_LIBSSH
    _cluster(NULL),
#endif
    _nS(0), _nX(0),  _nY(0), _nElY(0), _nbOut(0), _iOut(0), _iy(0), _iely(0),
    _nLauncher(kUnknown), _sFunctionName(""), _code(NULL), _run(NULL),
    _drawingGarbageCollector(NULL), _nSeed(0),
    _bChosenOutputs(kFALSE), _blog(kFALSE), _bdrawProgressBar(kTRUE), _bnoIntermediateSaved(kFALSE)
#ifdef ROOT5
    ,fMethodCall(NULL),
#else
    ,_pFunction(NULL),
#endif
    _ntresult(NULL) 
{
    if (_blog)
      cout << " *** Default constrcutor TSensitivity (empty) ***" << endl;

    _mAttributeSize.clear();
    _mAttributeElements.clear();
#ifdef ROOT5
    fMethod = SensitivityInteractiveMethod;
#endif
    // _drawingGarbageCollector = new TObjArray();
    // _drawingGarbageCollector->SetOwner(kTRUE);

    initResult();
}

TSensitivity::TSensitivity(URANIE::DataServer::TDataServer *tds, const char *varexpinput, const char *varexpoutput, Option_t * option) :
     _bisInputCorrelated(kFALSE), _bgoingThroughError(kFALSE),
    _tds(tds), _listOfInputAttributes(NULL), _listOfOutputAttributes(NULL), _sTimeAttribute(""), 
#ifdef WITH_LIBSSH
    _cluster(NULL),
#endif
    _nS(0), _nX(0),  _nY(0), _nElY(0), _nbOut(0), _iOut(0), _iy(0), _iely(0),
    _nLauncher(kUnknown), _sFunctionName(""), _code(NULL), _run(NULL),
    _drawingGarbageCollector(NULL), _nSeed(0),
    _bChosenOutputs(kTRUE),  _blog(kFALSE), _bdrawProgressBar(kTRUE), _bnoIntermediateSaved(kFALSE)
#ifdef ROOT5
    ,fMethodCall(NULL), 
#else
    ,_pFunction(NULL),
#endif
								     _ntresult(NULL) 
{
    if (_blog)
    {
        cout << "construit TSensitivity tds[" << _tds->GetName() << "] Option [" << option << "]" << endl;
        cout << "npat[" << _tds->getNPatterns() << "]" << endl;
        cout << "varexpinput[" << varexpinput << "]" << endl;
        cout << "varexpoutput[" << varexpoutput << "]" << endl;
    }
    
    try
    {
        if (_tds == NULL)
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       "TSensitivity::constructor \n The TDS is NULL");
	else if(_tds->getDataSpecification()==NULL)
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       "TSensitivity::constructor \n The TDS is NULL");
	
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    
    try
    {
        _nS = _tds->getNPatterns();
        if (_nS == 0)
	    throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                                                         Form("TSensitivity::constructor   ERROR The TTree is empty [%s] ", GetName()));

    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
    }
    _mAttributeSize.clear();
    _mAttributeElements.clear();
    //  Get the list of Input attributes
    _sSelectedInput = TString(varexpinput);
    try
    {
	_listOfInputAttributes = _tds->getListOfAttributes(varexpinput);
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
	if(_listOfInputAttributes != NULL)
	    delete _listOfInputAttributes;
        _listOfInputAttributes = NULL;
	throw ue;
    }
    _nX = _listOfInputAttributes->GetEntries();

    if (_blog)
    {
        cout << endl << " **  Input Attributes listing  _nX[" << _nX << "]"
	     << endl;
        _listOfInputAttributes->Print();
    }

    try
    {
	//  Get the list of Output attributes
	checkOutputRequested( string(varexpoutput) );
	parseOption(option);
        TString firstOutput = _sSelectedOutput(0,_sSelectedOutput.Index(":"));
        sprintf(_sOutputAttribute, "%s", firstOutput.Data());
        // sprintf(_sOutputAttribute, "%s", ((_sSelectedOutput.Index(":")==-1)?_sSelectedOutput.Data():_sSelectedOutput(0,_sSelectedOutput.Index(":")).Data()) );
        _listOfOutputAttributes = _tds->getListOfAttributes(_sSelectedOutput.Data());
        _nY = _listOfOutputAttributes->GetEntries();
        if (_blog)
	{
            cout << endl << " **  Output Attributes listing  _nY[" << _nY << "]"
		 << endl;
            _listOfOutputAttributes->Print();

	}

	// The output of the function
        //_nY = _sSelectedOutput.CountChar(':') + 1;
	
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
	if(_listOfInputAttributes != NULL)
	    delete _listOfInputAttributes;
        _listOfInputAttributes = NULL;
	if(_listOfOutputAttributes != NULL)
	    delete _listOfOutputAttributes;
        _listOfOutputAttributes = NULL;
        _nY = 0;
        throw ue;
    }

    initResult();
}


#ifdef ROOT5
TSensitivity::TSensitivity(URANIE::DataServer::TDataServer *tds, void *fcn,
        Int_t ns, const char *varexpinput, const char *varexpoutput, Option_t * option) :
#else
TSensitivity::TSensitivity(URANIE::DataServer::TDataServer *tds, 
        void (*fcn)(Double_t*, Double_t*),
        const char *varexpinput, const char *varexpoutput, Int_t ns, Option_t * option) :
#endif
     _bisInputCorrelated(kFALSE), _bgoingThroughError(kFALSE),
    _tds(tds), _listOfInputAttributes(NULL), _listOfOutputAttributes(NULL), _sTimeAttribute(""), 
#ifdef WITH_LIBSSH
    _cluster(NULL),
#endif
    _nS(ns), _nX(0),  _nY(0), _nElY(0),_nbOut(0), _iely(0),
    _nLauncher(kFunction), _sFunctionName(""), _code(NULL), _run(NULL),
     _drawingGarbageCollector(NULL), _nSeed(0),
    _bChosenOutputs(kTRUE), _blog(kFALSE), _bdrawProgressBar(kTRUE), _bnoIntermediateSaved(kFALSE)
#ifdef ROOT5
,fMethodCall(NULL), 
#else
,_pFunction(NULL),
#endif
								      _ntresult(NULL) 
{
    TString soption = option;
    soption.ToLower();
    try
    {
        if (_tds == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TSensitivity::constructor with function\n The TDS is NULL");
	else if(_tds->getDataSpecification()==NULL)
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       "TSensitivity::constructor with function\n The TDS is NULL");
	
	if( _nS <= 0)
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       "TSensitivity::constructor\n The requestes number of sample should be stricly greater than 0");
	
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }


    _mAttributeSize.clear();
    _mAttributeElements.clear();
    if (!fcn)
        return;
#ifdef ROOT5
    char *funcname = G__p2f2funcname(fcn);
    _sMethodName = TString(funcname);
    _sFunctionName = TString(funcname);
    SetName(Form("Sensitivity_%s", funcname));

    if (funcname)
    {
        fMethodCall = new TMethodCall();
        fMethodCall->InitWithPrototype(funcname, "Double_t*,Double_t*");
    }

    try
    {
        if (fMethodCall->IsValid () == kFALSE)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TSensitivity::constructor with function\n The function \"%s\" is not valid", funcname));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }


    fMethod = SensitivityInteractiveMethod;
    gSensitivity = this;
#else
    _sMethodName = TString(varexpoutput);
    _sFunctionName = TString(Form("Eval_%s_by_%s", varexpoutput, varexpinput));
    SetName(Form("Sensitivity_%s", varexpoutput));
    _pFunction = fcn;
#endif


    // Get the current list of attributes
    try
    {
	_listOfInputAttributes = _tds->getListOfAttributes(varexpinput);
    }catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
#ifdef ROOT5
        delete fMethodCall;
#endif
        throw ue;
    }

    initInputs();

    // The output of the function
    checkOutputRequested( string(varexpoutput) );
    if (_sSelectedOutput.Length() == 0)
    {
        _bChosenOutputs=kFALSE;
        _sSelectedOutput = _sMethodName;
    }
    _nY = _sSelectedOutput.CountChar(':') + 1;
                
    initResult();

    TString firstOutput = _sSelectedOutput(0,_sSelectedOutput.Index(":"));
    sprintf(_sOutputAttribute, "%s", firstOutput.Data());

    // _drawingGarbageCollector = new TObjArray();
    // _drawingGarbageCollector->SetOwner(kTRUE);
    
    if (_blog)
        cout << "_sSelectedInput[" << _sSelectedInput << "]" << endl
	     << "_sSelectedOutput[" << _sSelectedOutput << "]" << endl;
}

///To be corrected for consistency
TSensitivity::TSensitivity(URANIE::DataServer::TDataServer *tds, const char *fcn, 
        Int_t ns, const char *varexpinput, const char *varexpoutput, Option_t * option) :
     _bisInputCorrelated(kFALSE), _bgoingThroughError(kFALSE),
    _tds(tds), _listOfInputAttributes(NULL), _listOfOutputAttributes(NULL), _sTimeAttribute(""),
#ifdef WITH_LIBSSH
    _cluster(NULL),
#endif
     _nS(ns),
    _nLauncher(kFunction), _sFunctionName(""), _code(NULL), _run(NULL),
    _drawingGarbageCollector(NULL), _nSeed(0), _blog(kFALSE),
    _bdrawProgressBar(kTRUE), _bnoIntermediateSaved(kFALSE)
#ifdef ROOT5
    ,fMethodCall(NULL),
#else
    ,_pFunction(NULL),
#endif
								  _ntresult(NULL) 
{
    TString soption = option;
    soption.ToLower();

    try
    {
        if (_tds == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TSensitivity::constructor with function\n The TDS is NULL");
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }



    if (!fcn)
        return;
    char *funcname = (char *)fcn; //G__p2f2funcname(fcn);
    _sMethodName = TString(funcname);
    _sFunctionName = TString(funcname);
    SetName(Form("Sensitivity_%s", funcname));

#ifndef ROOT5
    TMethodCall *fMethodCall;
#endif
    if (funcname)
    {
        fMethodCall = new TMethodCall();
        fMethodCall->InitWithPrototype(funcname, "Double_t*,Double_t*");
    }

    try
    {
        if (fMethodCall->IsValid () == kFALSE)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TSensitivity::constructor with function\n The function \"%s\" is not valid", funcname));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

#ifdef ROOT5
    fMethod = SensitivityInteractiveMethod;
    gSensitivity = this;
#else
    delete fMethodCall;
    _pFunction = NULL;
#endif

    // Get the current list of attributes
    try
    {
	_listOfInputAttributes = _tds->getListOfAttributes(varexpinput);
    }catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
	delete fMethodCall;
        throw ue;
    }
    initInputs();

    // The output of the function
    checkOutputRequested( string(varexpoutput) );
    if (_sSelectedOutput.Length() == 0)
    {
        _bChosenOutputs=kFALSE;
        _sSelectedOutput = _sMethodName;
    }
    _nY = _sSelectedOutput.CountChar(':') + 1;
    TString firstOutput = _sSelectedOutput(0,_sSelectedOutput.Index(":"));
    sprintf(_sOutputAttribute, "%s", firstOutput.Data());
        
    initResult();
}


TSensitivity::TSensitivity(URANIE::DataServer::TDataServer *tds, URANIE::Launcher::TCode *code, Int_t ns, Option_t * option) :
     _bisInputCorrelated(kFALSE), _bgoingThroughError(kFALSE),
    _tds(tds), _listOfInputAttributes(NULL), _listOfOutputAttributes(NULL), _sTimeAttribute(""), 
#ifdef WITH_LIBSSH
    _cluster(NULL),
#endif
    _nS(ns), _nElY(0), _nbOut(0), _iely(0),
    _nLauncher(kCode), _sFunctionName(""), _code(code), _run(NULL),
     _drawingGarbageCollector(NULL), _nSeed(0),
     _bChosenOutputs(kFALSE), _blog(kFALSE), _bdrawProgressBar(kTRUE), _bnoIntermediateSaved(kFALSE)
#ifdef ROOT5
    ,fMethodCall(NULL), fMethod(NULL), 
#else
    ,_pFunction(NULL),
#endif
				       _ntresult(NULL) 
{
    try
    {
        if (_tds == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TSensitivity::constructor with code\n The TDS is NULL");
	else if(_tds->getDataSpecification()==NULL)
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       "TSensitivity::constructor with code\n The TDS is NULL");
	
	if( _code==NULL)
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TSensitivity::constructor with code\n The code is NULL");

	if( _code->getTDataServer() != _tds)
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TSensitivity::constructor with code\n The code is created but uses a different TDataServer than the one used here");

	if( _nS <= 0)
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       "TSensitivity::constructor\n The requestes number of sample should be stricly greater than 0");	
	
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    SetName("Sensitivity_Code");
    _mAttributeSize.clear();
    _mAttributeElements.clear();
    // Get the current list of attributes
    _listOfInputAttributes = _tds->getListOfAttributes("");
    initInputs();

    // Output
    Int_t nfile = _code->getNOutputFiles();
    Int_t ncrt = 0;
    if (_blog)
        cout << " ** nfile[" << nfile << "]" << endl;
    for (Int_t i = 0; i < nfile; i++)
    {
        Int_t natt = _code->getNOutputAttributes(i);
        for (Int_t j = 0; j < natt; j++)
        {
            TString satt =
		((URANIE::DataServer::TAttribute *) (_code->getAttribute(i,
									 j)))->GetName();
            if (_blog)
                cout << i << "/" << nfile << " :: " << j << "/" << natt
		     << " Name[" << satt << "]" << endl;
            if (ncrt == 0)
                _sSelectedOutput += Form("%s", satt.Data());
            else
                _sSelectedOutput += Form(":%s", satt.Data());
            ncrt++;
        }
    }

    checkOutputRequested( string(_sSelectedOutput) );
    parseOption( option );
    _nY = _sSelectedOutput.CountChar(':') + 1;
    TString firstOutput = _sSelectedOutput(0,_sSelectedOutput.Index(":"));
    sprintf(_sOutputAttribute, "%s", firstOutput.Data());
    
    initResult();
}

TSensitivity::TSensitivity(URANIE::DataServer::TDataServer *tds, URANIE::Relauncher::TRun *run, Int_t ns, Option_t * option) :
     _bisInputCorrelated(kFALSE), _bgoingThroughError(kFALSE),
    _tds(tds), _listOfInputAttributes(NULL), _listOfOutputAttributes(NULL), _sTimeAttribute(""), 
#ifdef WITH_LIBSSH
    _cluster(NULL),
#endif
    _nS(ns), _nElY(0), _nbOut(0), _iely(0),
     _nLauncher(kRun), _sFunctionName(""), _code(NULL), _run(run),
    _drawingGarbageCollector(NULL), _nSeed(0),
    _bChosenOutputs(kFALSE), _blog(kFALSE), _bdrawProgressBar(kTRUE), _bnoIntermediateSaved(kFALSE)
#ifdef ROOT5
    ,fMethodCall(NULL), fMethod(NULL),
#else
    ,_pFunction(NULL),
#endif
					        _ntresult(NULL) 
{
    try
    {
        if (_tds == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TSensitivity::constructor with code\n The TDS is NULL");
	else if(_tds->getDataSpecification()==NULL)
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       "TSensitivity::constructor with code\n The TDS is NULL");
	
	if( _run==NULL)
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TSensitivity::constructor with run\n The run is NULL");

	if( _nS <= 0)
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       "TSensitivity::constructor\n The requestes number of sample should be stricly greater than 0");	
	
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    
    SetName("Sensitivity_Eval");
    // user function
    _run = run;
    _mAttributeSize.clear();
    _mAttributeElements.clear();

    _listOfInputAttributes = new TList();
    // inputs
    try {
        // _listOfInputAttributes = _tds->getListOfAttributes("");
        TObjArray *inputlist = _run->getEval()->getInputList();
        TIter next(inputlist);
        while (URANIE::DataServer::TAttribute *att =
               (URANIE::DataServer::TAttribute *) next()) {
            _listOfInputAttributes->Add( att );
        }   
    }
    catch (URANIE::Exceptions::UErrorExceptions& ue) {
        ue.printMessage();
        throw ue;
    }
    initInputs();

    // Outputs
    TObjArray *outputlist = _run->getEval()->getOutputList();
    Int_t ncrt = 0;
    TIter next(outputlist);
    while (URANIE::DataServer::TAttribute *att =
            (URANIE::DataServer::TAttribute *) next()) {
        _sSelectedOutput += (ncrt == 0) ? 
            Form("%s", att->GetName()) : Form(":%s", att->GetName());
        ncrt++;
    }

    _nY = ncrt;
    checkOutputRequested( string(_sSelectedOutput) );
    parseOption( option );
    TString firstOutput = _sSelectedOutput(0,_sSelectedOutput.Index(":"));
    sprintf(_sOutputAttribute, "%s", firstOutput.Data());

    initResult();
}

// constructor's helper
void TSensitivity::initInputs()
{
    // Get the current list of attributes
    _nX = _listOfInputAttributes->GetEntries();

    // Inputs
    Int_t ncrt = 0;
    TIter next(_listOfInputAttributes);
    while (URANIE::DataServer::TAttribute *att =
            (URANIE::DataServer::TAttribute *) next())
    {
        if (ncrt == 0)
            _sSelectedInput += Form("%s", att->GetName());
        else
            _sSelectedInput += Form(":%s", att->GetName());
        ncrt++;
    }

}

void TSensitivity::initResult()
{

    _ntresult = new TTree("tmpsobol", "sensitivity sobol");
    _ntresult->Branch("Out", (void *) _sOutputAttribute, (const char *) "Out/C");
    _ntresult->Branch("Inp", (void*) _sInputAttribute, (const char *) "Inp/C");
    _ntresult->Branch("Order", (void*) _sOrder, (const char *) "Order/C");
    _ntresult->Branch("Method", (void*) _sMethod, (const char *) "Method/C");
    _ntresult->Branch("Algo", (void *) _sAlgorithm, (const char *) "Algorithm/C");
    _ntresult->Branch("Sobol", &_valSobolCrt, (const char *) "Sobol/D");
    _ntresult->Branch("CILower", &_valSobolCILower, (const char *) "CILower/D");
    _ntresult->Branch("CIUpper", &_valSobolCIUpper, (const char *) "CIUpper/D");
    _ntresult->SetMarkerColor(2);
    _ntresult->SetMarkerSize(4);
    _ntresult->SetMarkerStyle(7);
    
    _drawingGarbageCollector = new TObjArray();
    _drawingGarbageCollector->SetOwner(kTRUE);

    if (_blog)
        cout << " **  _sSelectedInput[" << _sSelectedInput << "]" << endl
	     << "_sSelectedOutput[" << _sSelectedOutput << "]" << endl;
}

TSensitivity::~TSensitivity()
{
    if (_blog)
        cout << " *** destroy TSensitivity(" << GetName() << ") Title[" << GetTitle() << "] *** " << endl;

    if( gROOT->GetList() != 0 &&  !gROOT->GetList()->IsEmpty())
    {
	// XYZ<<_ntresult<<endl;
        if (_ntresult)
        {
            //        _ntresult->ResetBranchAddresses();
            delete _ntresult;
            _ntresult = NULL;
        }
    }

    if(_listOfInputAttributes!=NULL)
	delete _listOfInputAttributes;
    _listOfInputAttributes=NULL;

    if(_listOfOutputAttributes!=NULL)
	delete _listOfOutputAttributes;
    _listOfOutputAttributes=NULL;

    if( gROOT->GetList() != 0 &&  !gROOT->GetList()->IsEmpty())
    {
	if(_drawingGarbageCollector)
	    delete _drawingGarbageCollector;
	_drawingGarbageCollector=NULL;
    }

    _mAttributeSize.clear();
    _mAttributeElements.clear();
#ifdef ROOT5
    if (fMethodCall != NULL)
    {
        delete fMethodCall;
        fMethodCall = NULL;
    }
#endif

}

void TSensitivity::createTuple(Option_t *option)
{
    if (_blog)
    {
        cout << "***********************************" << endl;
        cout << " ** TSensitivity::CreateTuple[" << option << "]" << endl;
    }

    // Create the tuple if necessary
    try
    {
        _tds->getTuple();
    } catch (URANIE::Exceptions::UWarningExceptions&)
    {
        _tds->createTuple();
    }
    if (_blog)
        _tds->getTuple()->Print();
    _nX = _tds->getNAttributes();

    if (_blog)
    {
        cout << " ** End of TSensitivity::CreateTuple" << endl;
        cout << "***********************************" << endl;
    }
}

void TSensitivity::checkOutputRequested(string attlist, bool fromoption)
{
    vector<string> buffer; 
    URANIE::DataServer::UEntry::StringParser( attlist, ":", buffer);
    if(_blog)
        XYZ<<" checkOutputRequested "<<attlist<<" buffer size "<<buffer.size()<<"  _sSelectedOutput ["<<_sSelectedOutput<<"] _bChosenOutput "<<_bChosenOutputs<<" fromopt "<<fromoption<<endl;
    try{
        if( attlist!="" && _sSelectedOutput!="" && fromoption==true && _bChosenOutputs==false)
        {
            _sSelectedOutput="";
            _bChosenOutputs=true;
        }
        for(unsigned int ibuf=0; ibuf<buffer.size(); ibuf++)
        {
            int ipos = ( (buffer.at(ibuf).find_first_of("[")==string::npos) ? buffer.at(ibuf).size() : buffer.at(ibuf).find_first_of("[") );
            int ilen = ( (buffer.at(ibuf).find_first_of("[")==string::npos) ? 0 : buffer.at(ibuf).find_first_of("]")-1-buffer.at(ibuf).find_first_of("["));            
            string restrBuf = buffer.at(ibuf).substr( 0, ipos);
            string restrInd = ((ilen==0)?"":buffer.at(ibuf).substr( ipos+1, ilen));
            if( (ipos+1-ilen) > MAXLENGTHSTRING)
                throw URANIE::Exceptions::UErrorExceptions(__FILE__,
							   __LINE__,
							   Form(
							       "TSensitivity::definition ERROR The output attribute name [%s] is too long. It size should remain < %d.",
							       restrBuf.c_str(),MAXLENGTHSTRING));
            if(_blog)
                XYZ<<" ipos/ilen "<<ipos<<"/"<<ilen<<"; restInd/Buf "<<restrInd<<"/"<<restrBuf<<" count "<<_mAttributeElements.count(restrBuf)<<"  "<<_sSelectedOutput.Contains(restrBuf)<<endl;
            if( find(_vOutputNames.begin(), _vOutputNames.end(), restrBuf) ==  _vOutputNames.end() )
                _vOutputNames.push_back( restrBuf );
	    
            if( restrInd != "" && _mAttributeElements.count(restrBuf)==0)
            {
                _mAttributeElements.insert( std::pair<string, vector<int> >(restrBuf, vector<int> (1, atoi(restrInd.c_str()) ) ) );
                if( !_sSelectedOutput.Contains(restrBuf) )
                    _sSelectedOutput+=(((_sSelectedOutput.EqualTo(""))?"":":")+TString(restrBuf));
            }
            else if( restrInd != "" && _mAttributeElements.count(restrBuf)!=0 )
            {
                if( find(_mAttributeElements[restrBuf].begin(),_mAttributeElements[restrBuf].end(), atoi(restrInd.c_str())) == _mAttributeElements[restrBuf].end() )
                    _mAttributeElements[restrBuf].push_back( atoi(restrInd.c_str()) );
            }
            else
            {
                // if( !_bChosenOutputs )
                // {
                //     _bChosenOutputs = kTRUE;
                //     _sSelectedOutput = restrBuf;
                // }
                if( !_sSelectedOutput.Contains(restrBuf) ) _sSelectedOutput+=(((_sSelectedOutput.EqualTo(""))?"":":")+TString(restrBuf));
            }
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
	ue.printMessage();
        if( _listOfInputAttributes != NULL )
        {
            delete _listOfInputAttributes;
            _listOfInputAttributes = NULL;
        }
	throw ue;
    }
    if(_blog)
    {
        XYZ<<"  _sSelectedOutput ["<<_sSelectedOutput<<"] _bChosenOutput "<<_bChosenOutputs<<" fromopt "<<fromoption<<"  "<<endl;
        
        for(map<string, vector<int> >::const_iterator it = _mAttributeElements.begin();
            it != _mAttributeElements.end(); ++it)
        {
            XYZ << it->first << " [";
            for(unsigned int iv=0; iv < it->second.size(); iv++)
                cout << it->second.at(iv)<<((iv==( it->second.size()-1))? "]" : ", ") ;
            cout << endl;
        }
    }
    // XYZ<<" map size "<<_mAttributeElements.size()<<" _sSelectedOutput "<<_sSelectedOutput<<endl;
    // for(map<string,vector<int> >::iterator it=_mAttributeElements.begin(); it!=_mAttributeElements.end(); it++)
    // {
    // 	XYZ<<" for string "<<it->first<<" ";
    // 	for(int ival=0; ival<it->second.size(); ival++)
    // 	    cout<<it->second.at(ival)<<((ival==it->second.size()-1)?"":",");
    // 	cout<<endl;
    // }

}
  
void TSensitivity::parseOption(Option_t *option)
{
    if (  _blog )
        cout << "***** TSensitivity::parseOption[" << option << "]" << endl;

    TString soption(option);
    soption.ToLower();
    if( soption.Contains("nointermed"))
	setNoIntermediateSaved();

    vector<string> buffer;
    URANIE::DataServer::UEntry::StringParser(string(soption), ",", buffer);
    string selected="";
    for(unsigned int ibuf=0; ibuf<buffer.size(); ibuf++)
    {
	if( TString(buffer.at(ibuf)).Contains("output=") )
	{
//	    _sSelectedOutput="";   
	    selected = TString( buffer.at(ibuf).substr( buffer.at(ibuf).find_first_of("=")+1, buffer.at(ibuf).size()-buffer.at(ibuf).find_first_of("=")));
	}
    }
    
    checkOutputRequested(selected, true);
	    
}
  

///////////////////////////////////////////////////////////
//  computeIndexes 
///////////////////////////////////////////////////////////
void TSensitivity::computeIndexes(Option_t * option)
{
    if (_blog) {
	cout << endl;
	cout << "*********************************" << endl;
	cout << "** TSensitivity::computeIndexes" << endl;
	cout << " **    _sSelectedInput[" << _sSelectedInput << "]" << endl;
	cout << " **   _sSelectedOutput[" << _sSelectedOutput << "]" << endl;
    }

    parseOption(Form("%s,[computeIndexes]", option));
    try {
	// Generate the sampling
        if (_tds->isAttribute(_sIteratorName) == kFALSE) {
	    generateSample(option);
        }

	bool FoundAttributes= ((_vOutputNames.size()==0) ? false : true);
	for(vector<string>::iterator its=_vOutputNames.begin(); its!= _vOutputNames.end(); its++)
	    if ( _tds->isAttribute( (*its).c_str() ) == kFALSE)
		FoundAttributes=false;
	    
        if( !FoundAttributes ) {
	    // launch the code on the sampling
            switch (_nLauncher) {
            case kUnknown: {
                throw URANIE::Exceptions::UErrorExceptions(__FILE__,
							   __LINE__,
							   Form(
							       "TSensitivity::computeIndexes ERROR for TDataServer \"%s\" : the launcher is unknown.",
							       GetName()));
                break;
	    }
            case kFunction: {
                URANIE::Launcher::TLauncherFunction * tlf = 
#ifndef ROOT5
                        (_pFunction) ?
                        new URANIE::Launcher::TLauncherFunction(_tds,
                                                                    _pFunction, _sSelectedInput,
                                _sSelectedOutput) :
#endif
                        new URANIE::Launcher::TLauncherFunction(_tds,
                                                                    _sFunctionName, _sSelectedInput,
                                _sSelectedOutput) ;
                tlf->setDrawProgressBar(_bdrawProgressBar);
                if (_blog)
		    tlf->setLog();
                tlf->run();
                delete tlf;
                tlf = NULL;
                break;
	    }
            case kCode:
	    {

                if (_blog) {
                    cout << " ** Case of Tcode" << endl;
                    _tds->getTuple()->Print();
		    _tds->getTuple()->Scan("*");
		    }

                URANIE::Launcher::TLauncher * tl = new URANIE::Launcher::TLauncher(_tds, _code);
                // tl->setSave();
                tl->setClean();
                tl->setDrawProgressBar(_bdrawProgressBar);
                if (_blog)
		    tl->setLog();
                tl->run( ((_bnoIntermediateSaved) ? "nointermed" : "") );

                stringstream sstr;
                for(int ioutfile=0; ioutfile< _code->getNOutputFiles(); ioutfile++)
                    for(int iattoutfile=0; iattoutfile< _code->getNOutputAttributes(ioutfile); iattoutfile++)
                    {
                        URANIE::DataServer::TAttribute *l_output = _code->getOutputFile(ioutfile)->getAttribute(iattoutfile);
                        TString pouetpouet="1.23456789";
                        sstr.str(""); sstr << l_output->GetName();
                        if(l_output->hasDefault())
                            pouetpouet=l_output->getDefault();
                        sstr << setprecision(12) << " == " << pouetpouet;
                        int nData = _tds->getTuple()->Draw(l_output->GetName(),sstr.str().c_str(),"goff");
                         try
                         {
                             if( nData != 0)
                             {
                                 if( _bgoingThroughError )
                                     throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                                                                                Form("TSensitivity::computeIndexes The dataserver contains [%d] events for which attributes [%s] is filled with its default value. \nThis might implies failure in code so results interpretation might be subject to problem !!.", nData, l_output->GetName()));
                                 else
                                     throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                                                Form("TSensitivity::computeIndexes The dataserver contains [%d] events for which attributes [%s] is filled with its default value. \nThis might implies failure in code so results interpretation might be subject to problem !!.", nData, l_output->GetName()));
                             }
                         }catch(URANIE::Exceptions::UWarningExceptions &ue)
                         {
                             ue.printMessage();
                         }catch(URANIE::Exceptions::UErrorExceptions &ue)
                         {
                             ue.printMessage();
                             throw ue;
                         }
                         
                    }
                
                delete tl;
                tl = NULL;
                break;
	    }
	    case kCodeRemote:
	    {
#ifdef WITH_LIBSSH		
	        if (_blog)
                {
                    cout << " ** Case of TCodeRemote" << endl;
                    _tds->getTuple()->Print();
		    _tds->getTuple()->Scan("*");
                }

                URANIE::Launcher::TLauncherByStepRemote* tlRem = new URANIE::Launcher::TLauncherByStepRemote(_tds, _code, _cluster);
                
	        tlRem->setSave();
                tlRem->setClean();
                tlRem->setDrawProgressBar(_bdrawProgressBar);
                if (_blog)
                    tlRem->setLog();
                //tlRem->run( ((_bnoIntermediateSaved) ? "nointermed" : "") );
                tlRem->setNumberOfChunks(1);
                tlRem->setJobName("sensitivity");

                tlRem->run();
                tlRem->postTreatment();
                delete tlRem;
                tlRem = NULL;
#endif
                break;
	    }
        case kRun :
            URANIE::Relauncher::TLauncher2 launch(_tds, _run);
            URANIE::DataServer::TDataServer _errTds(Form("_%s_error_tds_.dat", _sMethodName.Data()), "local_err_tds");
            launch.setSaveError( &_errTds );
            launch.solverLoop();

            try
            {
                if( _errTds.getNPatterns() != 0)
                {
                    _errTds.exportData( Form("_%s_error_tds_.dat", _sMethodName.Data()) );
                    if( _bgoingThroughError )
                        throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                                                                     Form("TSensitivity::computeIndexes The failure dataserver is not empty, it contains [%d] events !!! Results interpretation might be subject to problem.\n Have a look at the dumped failure dataserver [%s].", _errTds.getNPatterns(), Form("_%s_error_tds_.dat", _sMethodName.Data())));
                    else
                        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                                   Form("TSensitivity::computeIndexes The failure dataserver is not empty, it contains [%d] events !!! Results interpretation might be subject to problem.\n Have a look at the dumped failure dataserver [%s].", _errTds.getNPatterns(), Form("_%s_error_tds_.dat", _sMethodName.Data())));
                }
            }catch(URANIE::Exceptions::UWarningExceptions &ue)
            {
                ue.printMessage();
            }catch(URANIE::Exceptions::UErrorExceptions &ue)
            {
                ue.printMessage();
                throw ue;
            }
            break;
            
            }
	    TString exportFileName = Form("_%s_launching_.dat", _sMethodName.Data());
	    exportFileName.ToLower();
	    // Mod by JBB so that output file can be used again (sobol__n__iter__tdsName is compulsory in this case)
	    TString VarList = Form("%s:%s", _sSelectedInput.Data(),  _sSelectedOutput.Data());
	    if( _sMethodName.Contains("Sobol") ||_sMethodName.Contains("Morris")) VarList+=":"+_sIteratorName;
	    if( _sMethodName.Contains("Morris") ) VarList+=":__npoints__";
	    if( _sMethodName.Contains("FiniteDifferences") ) VarList+=":__nominal_set__:__modified_att__";
	    _tds->exportData(exportFileName, VarList.Data() );
        }//if (_tds->isAttribute(_sOutputAttribute) == kFALSE)

        delete _listOfOutputAttributes;

	_listOfOutputAttributes = _tds->getListOfAttributes(_sSelectedOutput);
	_nY = 0;

	if (_blog) {
	    cout << endl << " **  Output Attributes listing  _nY[" << _nY << "]"
		 << endl;
	    _listOfOutputAttributes->Print();
	}

	// Run on all output attributes to check if they have to be done and get the number of elements
	_mAttributeSize.clear();
	TIter next(_listOfOutputAttributes);
	string soutput="";
	_nbOut=0;
	while (URANIE::DataServer::TAttribute *att = (URANIE::DataServer::TAttribute *) next())
	{
	    if (_blog)
		cout<<" ** Looking at attribute "<<att->GetName()<<endl;
	    
	    try
	    {
		if( att->getDataTypeStr().EqualTo("S") )
		    throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
								 Form("TSensitivity::computeIndexes is only able to deal with double and vectors. Branch %s has type %s, so nothing will be done for it!",
								      att->GetName(), att->getDataTypeStr().Data()));
		else if(att->getDataTypeStr().EqualTo("V"))
		{
		    string sizebranch=SIZENAME+"__"+string(att->GetName());
		    if(!_tds->isAttribute( sizebranch.c_str() ))
			throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
								     Form("TSensitivity::computeIndexes: vectors are usable when size is an attribute. %s not found so nothing will be done for it!",
									  sizebranch.c_str()));
		    _tds->computeStatistic( sizebranch.c_str() );
		    if( _tds->getAttribute( sizebranch.c_str() )->getMinimum() != _tds->getAttribute( sizebranch.c_str() )->getMaximum() )
			throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
								     Form("TSensitivity::computeIndexes: vectors are usable when size is not changing. %s Std is not 0 (%g) so nothing will be done for it!",
									  sizebranch.c_str(), _tds->getAttribute( sizebranch.c_str() )->getStd() ));
		    
		    _nY++;
		    _mAttributeSize.insert( std::pair<string,unsigned int>(string(att->GetName()),TMath::Nint(_tds->getAttribute( sizebranch.c_str() )->getMean())) );
		    soutput+=string(att->GetName())+":";
		    _nbOut+=_mAttributeSize[string(att->GetName())];
		}
		else
		{
		    _nY++;
		    _mAttributeSize.insert( std::pair<string,unsigned int>(string(att->GetName()),1) );
		    soutput+=string(att->GetName())+":";
		    _nbOut+=_mAttributeSize[string(att->GetName())];
		}
	    }catch(URANIE::Exceptions::UWarningExceptions &ue)
	    {
		ue.printMessage();
	    }	    
	}
	soutput=soutput.substr(0, soutput.size()-1);
	if (_blog)
	    cout<<" **  List of kept variable is "<<soutput<<"  and _nY "<<_nY<<endl;

	vector<string> voutput;
	URANIE::DataServer::UEntry::StringParser(soutput, ":", voutput);
	std::stringstream sstr;

	if (_blog) {
	    cout << endl << "____________________________________" << endl;
	    cout << " *** TSensitivity::computeIndexes   iter[" << _tds->getIteratorName() << "]" << endl;
	}

	//Run over all outputs and subelement if so
	_iOut=0; _iely=0;
	URANIE::DataServer::TAttribute *attY;
	for(_iy=0; _iy<_nY; _iy++)
	{
	    attY = _tds->getAttribute( voutput.at(_iy) );
	    sstr.str(""); sstr << attY->GetName();
	    _nElY=_mAttributeSize[ attY->GetName() ];
	    sprintf(_sOutputAttribute, "%s", sstr.str().c_str() );

	    if(_blog)
		cout<<"Looking at attribute "<< attY->GetName() << "  with a number of element of " << _nElY <<endl;

	    preTreatment();

	    if (_mAttributeElements.find( string(attY->GetName()) )!=_mAttributeElements.end())
		std::sort(_mAttributeElements.find( string(attY->GetName()) )->second.begin(),_mAttributeElements.find( string(attY->GetName()) )->second.end());

	    for( _iely=0; _iely<_nElY; _iely++)
	    {
		//Set the name of the output
		if (_nElY!=1)
		{
		    sstr.str("");
		    sstr << attY->GetName() <<  "[" << _iely << "]";
		    sprintf(_sOutputAttribute, "%s", sstr.str().c_str() );

		    if (_mAttributeElements.find( string(attY->GetName()) )!=_mAttributeElements.end())
		    {
			vector<int> sublist =  _mAttributeElements.find( string(attY->GetName()) )->second;
			if( find( sublist.begin(), sublist.end(), _iely) == sublist.end() )
			    continue;
		    }
		}

		//XYZ<<"evaluateIndexes for "<<attY->GetName()<<" iel "<<_iely<<"  _nS being ["<<_nS<<"] while _nPatterns is ["<<_tds->getNPatterns()<<"]"<<endl;
                // if( _nS != _tds->getNPatterns() )
                //     throw URANIE::Exceptions::UInfoExceptions(__FILE__, __LINE__,
                //                                                  Form("TSensitivity::computeIndexes method:\n The number of patterns in the dataserver [%d] does not match the number of events requested for a proper indexes estimation [%d]", _tds->getNPatterns(), _nS));
                
		evaluateIndexes(option);
	    }

	    postTreatment();
	  
	    // if ( strcmp(this->IsA()->GetName(),"URANIE::Sensitivity::TFiniteDifferences")==0 )
	    //   break;

	}

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
	ue.printMessage();
	throw ue;
    } catch (URANIE::Exceptions::UExceptions& ue)
    {
	ue.printMessage();
	if (ue.getMessageType() > URANIE::Exceptions::UExceptions::kWARNING)
	    exit(-1);
    }

    TString opt = option;
    opt.ToLower();
    if (_blog)
    {
	cout << " ** End of TSensitivity::computeIndexes" << endl;
	cout << "*********************************" << endl << endl;
    }
}

void TSensitivity::setTimeName(TString sname)
{
    try {
        if ( sname.CountChar(':') != 0 )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form("TSensitivity::setTimeName method:\n Only one attribute for the Time [%s]", sname.Data ()));
        if ( ! _tds->isAttribute (sname) )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form("TSensitivity::setIterator method:\n The Time attribute [%s] does not exist in the TDS", sname.Data ()));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    _sTimeAttribute = sname;
}

void TSensitivity::fillIndex(const char * sinputname, const char * sorder,
			     Double_t dval, const char* algo, Double_t dvalCILower, Double_t dvalCIUpper)
{
    sprintf(_sMethod, "%s", _sMethodName.Data());

    sprintf(_sAlgorithm, "%s", algo);
    sprintf(_sInputAttribute, "%s", sinputname);
    sprintf(_sOrder, "%s", sorder);
    _valSobolCrt = dval;
    _valSobolCILower = dvalCILower;
    _valSobolCIUpper = dvalCIUpper;

    _ntresult->Fill();
}

#ifdef ROOT5
void TSensitivity::setFunction(void *fcn, Int_t nx, Int_t ny)
#else
void TSensitivity::setFunction(void (*fcn)(Double_t*,Double_t*), Int_t nx, Int_t ny)
#endif
{
    if (!fcn)
        return;
#ifdef ROOT5
    char *funcname = G__p2f2funcname(fcn);
    _sMethodName = TString(funcname);

    if (funcname)
    {
        fMethodCall = new TMethodCall();
        fMethodCall->InitWithPrototype(funcname, "Double_t*,Double_t*");
    }
    gSensitivity = this;
    fMethod = fcn;
#else
    _pFunction = fcn;
#endif
    // La dimension du vecteur d'entr�e. Elle peut �tre d�j� d�finie
    // via les m�thodes addAttribute(), dans ce cas on met la valeur � -1
    // On peut bien su utiliser GetNAttributes() pour conna�tre cette dimension
    if (nx != -1)
    { // nX est d�fini avec une valeur non nulle
        if ((_nX != 0) && (_nX != nx))
        { // Test si nx est concordant
            cout
		<< "**** Error TSensitivity::setFunction dans la dimension des entr�es "
		<< endl;
            cout << " The dimension _nX[" << _nX << "] used" << endl;
        }
        else
            _nX = nx;
    }

    // La dimension du vecteur de sortie
    _nY = ny;

}


void TSensitivity::checkCanvasCreation(bool newcan)
{
        
    if (gPad && !newcan)
    {
	_canvas = (TCanvas *) (gPad->GetCanvas());
	if( TString(_canvas->GetName()).BeginsWith("__sensitivitycan__"))
	    _canvas->Clear();
	else
	    _canvas = (TCanvas*)gPad;
    }
    else
    {
	stringstream sstr; sstr << "__sensitivitycan__" << _drawingGarbageCollector->GetEntries() ;
	_canvas = new TCanvas(sstr.str().c_str(),sstr.str().c_str(),1);
	_drawingGarbageCollector->Add(_canvas);
    }

    //XYZ<<"[91m The pointer to the canvas created or not is "<<_canvas<<"[0m"<<endl;
    
}


void TSensitivity::drawIndexes(TString sTitre, const char *select, Option_t * option)
{
    TString soption = option;
    soption.ToLower();
    if ( _blog )
        cout << " ** TSensitivity::drawIndexes for method[" << _sMethodName << "] option[" << option << "]" << endl;

    if (_ntresult->GetEntries() == 0) {
        cout << "TSensitivity::drawIndexes Error : the TNtupleD is empty" << endl;
        return;
    }

    TString opt = option;
    opt.ToLower();
    if (opt.Contains("nocanv") || opt.Length() == 0)
        return;

    bool newcan=false;
    if (!opt.Contains("nonewcanv"))
	newcan=true;
	
    checkCanvasCreation(newcan);

    if ( _sMethodName.CompareTo("Morris") == 0 ) {
        cout << " ** case of Morris with option[" << option << "]" << endl;
    }

    if ( _sMethodName.BeginsWith ("SRC") || _sMethodName.BeginsWith ("SRRC") ||
	 _sMethodName.BeginsWith ("PCC") || _sMethodName.BeginsWith ("PRCC") )
    {
        if ( _blog )
	    cout << " ** case of SRC|SRRC|PCC|PRCC with option[" << option << "]" << endl;
        if ( opt.Contains("total") )
	    opt.ReplaceAll ("total","");
        if ( opt.Contains("first") )
	    opt.ReplaceAll ("first","");
        if ( _blog )
	    cout << "End of TSensitivity::drawIndexes  _sMethodName[" << _sMethodName << "] option[" << option << "]" << endl;
    }

    if (opt.Contains("hist"))
	drawIndexesHistogram(sTitre, ((strcmp(select,"")==0||strcmp(select," ")==0) ? "1>0" : select), opt);
    if (opt.Contains("pie"))
	drawIndexesPie(sTitre, ((strcmp(select,"")==0||strcmp(select," ")==0) ? "1>0" : select), opt);
}

void TSensitivity::drawIndexesPie(TString sTitre, const char *select, Option_t * option)
{
    (void)(sTitre);
    Int_t nval = Int_t(_ntresult->GetEntries());
    if (_blog)
    {
        cout << "  **TSensitivity::drawIndexesPie option[" << option << "]" << endl;
        cout << "nval[" << nval << "]" << endl;
        cout << "_sMethodName[" << _sMethodName << "]" << endl;
    }
    TString opt = option;
    opt.ToLower();

    if (_blog)
        cout << "nval[" << nval << "]" << endl;

    stringstream sstr; sstr << "pie_"<<_drawingGarbageCollector->GetEntries();    
    TPie *piepie = new TPie(sstr.str().c_str(), "First Sensitivity Index ", _nX);
    sstr.str(""); sstr << "pietotal_"<<_drawingGarbageCollector->GetEntries(); 
    TPie *piepieTotal = new TPie(sstr.str().c_str(), "Total Sensitivity Index ", _nX);

    _drawingGarbageCollector->Add(piepie);
    _drawingGarbageCollector->Add(piepieTotal);
    _canvas->Divide(2);

    TString method="";
    if(  _sMethodName.BeginsWith ("SRC") || _sMethodName.BeginsWith ("SRRC") || _sMethodName.BeginsWith ("PCC") || _sMethodName.BeginsWith ("PRCC") )
    {
	if( !opt.Contains("src") && !opt.Contains("srrc") && !opt.Contains("pcc") && !opt.Contains("prcc") )
	{
	    if( _sMethodName.Contains("SRC") )       method = "&&(Method==\"SRC^2\")"; 
	    else if( _sMethodName.Contains("SRRC") ) method = "&&(Method==\"SRRC^2\")";
	    else if( _sMethodName.Contains("PCC") )  method = "&&(Method==\"PCC^2\")"; 
	    else if( _sMethodName.Contains("PRCC") ) method = "&&(Method==\"PRCC^2\")";
	}
	else if( opt.Contains("src") )  method = "&&(Method==\"SRC^2\")"; 
	else if( opt.Contains("srrc") ) method = "&&(Method==\"SRRC^2\")";
	else if( opt.Contains("pcc") )  method = "&&(Method==\"PCC^2\")"; 
	else if( opt.Contains("prcc") ) method = "&&(Method==\"PRCC^2\")";

    }
    
    URANIE::DataServer::TAttribute *att;
    for (Int_t i = 0; i < _nX; i++)
    {
        att = (URANIE::DataServer::TAttribute *) (_listOfInputAttributes->At(i));

        Double_t dvalfirst = 0.0;
        TString sSelectForula = Form("((Inp==\"%s\")&&(Algo==\"--first--\")%s&&(%s))", att->GetName(),method.Data(),select);

        Long64_t nDraw = _ntresult->Draw("Sobol", sSelectForula, "goff");
        try {
            if ( nDraw < 1 )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("ERROR : Any First Order Indexes founded in the TDS [%s] with the method [%s]", _tds->GetName (), _sMethod));

        }   catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            exit(-1);
        }
        Double_t *vv = _ntresult->GetV1();
        dvalfirst = vv[0];

        piepie->SetEntryVal(i, dvalfirst);   // Les valeurs
        piepie->SetEntryFillColor(i, i + 3);           // La couleur
        piepie->SetEntryLabel(i, Form("%s", att->GetName())); // Le label

        Double_t dvaltot = 0.0;

        nDraw = _ntresult->Draw("Sobol",
                                Form("((Inp==\"%s\")&&(Algo==\"--total--\")%s&&(%s))", att->GetName(),method.Data(),select),
                                "goff");
        try {
            if ( nDraw < 1 )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("ERROR : Any Total Order Indexes founded in the TDS [%s] with the method [%s]", _tds->GetName (), _sMethod));

        }   catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            exit(-1);
        }
        Double_t *vvtot = _ntresult->GetV1();
        dvaltot = vvtot[0];

        piepieTotal->SetEntryVal(i, dvaltot);   // Les valeurs
        piepieTotal->SetEntryFillColor(i, i + 3);           // La couleur
        piepieTotal->SetEntryLabel(i, Form("%s", att->GetName())); // Le label
        if (_blog)
            cout << " ** attribute " << i << "/" << att->GetName() << " first[" << dvalfirst << "] tot[" << dvaltot << "]" << endl;
    }

    piepie->SetAngularOffset(30.);
    //     piepie->SetEntryRadiusOffset( 4, 0.1);
    piepie->SetRadius(.35);
    piepie->SetLabelFormat("#splitline{%perc}{%txt}");
    piepie->SetTextSize(0.04);
    piepie->SetLabelsOffset(.025);

    piepieTotal->SetAngularOffset(30.);
    //     piepieTotal->SetEntryRadiusOffset( 4, 0.1);
    piepieTotal->SetRadius(.35);
    piepieTotal->SetLabelFormat("#splitline{%perc}{%txt}");
    piepieTotal->SetTextSize(0.04);
    piepieTotal->SetLabelsOffset(.025);

    //     piepie->Draw("nol");
    _canvas->cd(1);
    piepie->Draw("nol");
    _canvas->cd(2);
    piepieTotal->Draw("nol");
}

void TSensitivity::drawIndexesHistogram(TString sTitre, const char *select, Option_t * option)
{
    
    Int_t nval = (Int_t) _ntresult->GetEntries();
    if (_blog)
    {
        cout << "  **TSensitivity::drawIndexesHistogram option[" << option << "]" << endl;
        cout << "nval[" << nval << "]" << endl;
        cout << "_sMethodName[" << _sMethodName << "]" << endl;
    }

    
    // Input Of attributes
    if (_blog)
	_listOfInputAttributes->Print();
    
    TString opt = option;
    opt.ToLower();
    TString sFirstLegend = "First";
    Int_t nfirst = -1, ntotal = -1;

    stringstream sstr; 
    TH1F *hDiv = NULL;
    TH1F *hDivTotal = NULL;
    if (opt.Contains("first") || opt.Contains("all"))
    {
        if (_blog)
            cout << " ** nDraw[(opt.Contains(\"first\") || opt.Contains(\"all\"))]" << endl;
	
	sstr.str(""); sstr << "hDiv_"<<_drawingGarbageCollector->GetEntries();    
        Long64_t nDraw = _ntresult->Draw(TString("Inp>>"+sstr.str()), Form("(Algo==\"--first--\")&&(Inp!=\"__sum__\")&&(Inp!=\"__R2__\")&&(Inp!=\"__R2A__\")&&(%s)",select), "goff");
	hDiv = (TH1F*)gDirectory->Get(sstr.str().c_str());
	hDiv->SetDirectory(0);
	hDiv->SetStats(0);
	
        if (_blog)
            cout << " ** nDraw[" << nDraw << "]" << endl;
        nfirst = Int_t(_ntresult->GetSelectedRows());
        if (_blog)
            cout << " ** nfirst[" << nfirst << "]" << endl;
        try {
            if ( nDraw < 1 )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("ERROR : Any First Order Indexes founded in the TDS [%s] with the method [%s]", _tds->GetName (), _sMethod));

        }   catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            exit(-1);
        }


        if (_blog)
            cout << "  ** First N[" << nfirst << "]" << endl;
    }
    if (opt.Contains("total") || opt.Contains("all"))
    {

	sstr.str(""); sstr << "hDivTotal_"<<_drawingGarbageCollector->GetEntries();    
	Long64_t nDraw = _ntresult->Draw(TString("Inp>>"+sstr.str()), Form("(Algo==\"--total--\")&&(Inp!=\"__sum__\")&&(Inp!=\"__R2__\")&&(Inp!=\"__R2A__\")&&(%s)",select), "goff");
        ntotal = Int_t(_ntresult->GetSelectedRows());
	hDivTotal = (TH1F*)gDirectory->Get(sstr.str().c_str());
	hDivTotal->SetDirectory(0);
	hDivTotal->SetStats(0);
	
        try {
            if ( nDraw < 1 )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("ERROR : Any Total Order Indexes founded in the TDS [%s] with the method [%s]", _tds->GetName (), _sMethod));

        }   catch (URANIE::Exceptions::UErrorExceptions& ue)
        {
            ue.printMessage();
            exit(-1);
        }


        if (_blog)
            cout << "  ** Total N[" << ntotal << "]" << endl;
    }
  

    //////////////////////////////
    // **************************
    //TLeaf *leafx = _ntresult->GetLeaf("Sobol");

//     if (nfirst > 0)
//     {
// // //         _ntresult->Draw("Inp>>hDiv", "Order==\"First\"", "goff");
// // 	_ntresult->Draw("Inp>>hDiv", Form("(Algo==\"--first--\")&&(Inp!=\"__sum__\")&&(Inp!=\"__R2__\")&&(Inp!=\"__R2A__\")&&(%s)",select), "goff");
// //         hDiv = (TH1F*) gDirectory->Get("hDiv");
//         hDiv->SetStats(0);
//     }

//     if (ntotal > 0)
//     {
//         // hDivTotal = (TH1F*) hDiv->Clone("hDivTotal");
//         // _ntresult->Draw("Inp>>hDivTotal", Form("(Algo==\"--total--\")&&(Inp!=\"__sum__\")&&(Inp!=\"__R2__\")&&(Inp!=\"__R2A__\")&&(%s)",select), "goff");
//     }

    TString method="";
    if(  _sMethodName.BeginsWith ("SRC") || _sMethodName.BeginsWith ("SRRC") || _sMethodName.BeginsWith ("PCC") || _sMethodName.BeginsWith ("PRCC") )
    {
	if( !opt.Contains("src") && !opt.Contains("srrc") && !opt.Contains("pcc") && !opt.Contains("prcc") )
	{
	    if( _sMethodName.Contains("SRC") )       {method = "&&(Method==\"SRC^2\")"; sFirstLegend ="SRC^{2}";}
	    else if( _sMethodName.Contains("SRRC") ) {method = "&&(Method==\"SRRC^2\")"; sFirstLegend ="SRRC^{2}";}
	    else if( _sMethodName.Contains("PCC") )  {method = "&&(Method==\"PCC^2\")"; sFirstLegend ="PCC^{2}";}
	    else if( _sMethodName.Contains("PRCC") ) {method = "&&(Method==\"PRCC^2\")"; sFirstLegend ="PRCC^{2}";}
	}
	else if( opt.Contains("src") )  {method = "&&(Method==\"SRC^2\")"; sFirstLegend ="SRC^{2}";}
	else if( opt.Contains("srrc") ) {method = "&&(Method==\"SRRC^2\")"; sFirstLegend ="SRRC^{2}";}
	else if( opt.Contains("pcc") )  {method = "&&(Method==\"PCC^2\")"; sFirstLegend ="PCC^{2}";}
	else if( opt.Contains("prcc") ) {method = "&&(Method==\"PRCC^2\")"; sFirstLegend ="PRCC^{2}";}

	delete hDiv;
	_ntresult->Draw("Inp>>hDiv", Form("(Algo==\"--first--\")&&(Inp!=\"__sum__\")&&(Inp!=\"__R2__\")&&(Inp!=\"__R2A__\")%s&&(%s)",method.Data(),select), "goff");
	hDiv = (TH1F*) gDirectory->Get("hDiv");
	hDiv->SetDirectory(0);
        hDiv->SetStats(0);
    }
      
    // Common objects
    Double_t ymin = 0.80;
    TLegend *legend = new TLegend(0.75, ymin, 0.88, ymin + 0.065);
    _drawingGarbageCollector->Add(legend);
    //    legend->SetHeader("");
    URANIE::DataServer::TAttribute *att;

    if (hDiv)
    {
        hDiv->Reset();
        hDiv->SetBarWidth(0.45);
        hDiv->SetBarOffset(0.1);
        hDiv->SetFillColor(49);
	hDiv->GetYaxis()->SetRangeUser(0,1);
        hDiv->SetTitle(sTitre);
        for (Int_t i = 0; i < _nX; i++)
        {
            att = (URANIE::DataServer::TAttribute *) (_listOfInputAttributes->At(i));
            _ntresult->Draw("Sobol",
			    Form("((Inp==\"%s\")&&(Algo==\"--first--\"))%s&&(%s)", att->GetName(),method.Data(),select),
			    "goff");
            Double_t *vv = _ntresult->GetV1();
            hDiv->SetBinContent(i + 1, vv[0]);
        }

        if ((opt.Contains("first>")) || !(opt.Contains("total<"))
	    || !(opt.Contains("total>")) || (opt.Contains("all")))
            hDiv->LabelsOption(">u");
        if (opt.Contains("first<"))
            hDiv->LabelsOption("<u");
        hDiv->SetMinimum(0.0);
        hDiv->SetMaximum(1.2);
        hDiv->Draw("bar2");
        legend->AddEntry(hDiv, sFirstLegend.Data(), "f");
	_drawingGarbageCollector->Add(hDiv);
    }
    if (hDivTotal)
    {
        hDivTotal->Reset();
	hDiv->GetYaxis()->SetRangeUser(0,1);
        hDivTotal->SetTitle(sTitre);
        hDivTotal->SetBarWidth(0.4);
        hDivTotal->SetBarOffset(0.55);
        hDivTotal->SetFillColor(50);
        if (opt.Contains("total>"))
            hDiv->LabelsOption(">u");
        if (opt.Contains("total<"))
            hDiv->LabelsOption("<u");
        for (Int_t i = 0; i < _nX; i++)
        {
            if (hDiv)
            {
                _ntresult->Draw("Sobol",
				Form("((Inp==\"%s\")&&(Algo==\"--total--\"))&&(%s)",
				     hDiv->GetXaxis()->GetBinLabel(i + 1),select), "goff");
            }
            else
            {
                att = (URANIE::DataServer::TAttribute *) (_listOfInputAttributes->At(i));
                _ntresult->Draw("Sobol",
				Form("((Inp==\"%s\")&&(Algo==\"--total--\"))&&(%s)",
				     att->GetName(),select), "goff");
            }
            Double_t *vv = _ntresult->GetV1();
            hDivTotal->SetBinContent(i + 1, vv[0]);
        }

        if (hDiv)
            hDivTotal->DrawCopy("bar2,same");
        else
            hDivTotal->DrawCopy("bar2");
        legend->AddEntry(hDivTotal, "Total", "f");
	_drawingGarbageCollector->Add(hDivTotal);
    }

    legend->Draw();
    gPad->SetGridy(1);
    gPad->Modified();
    gPad->Update();
}

void TSensitivity::setInputCorrelationMatrix(TMatrixD corrMat)
{
    
    try
    {
        if ((corrMat.GetNcols() != _nX)
            || (corrMat.GetNrows() != corrMat.GetNcols()))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form(
                                                           "TSensitivity::setCorrelationMatrix: invalid correlation matrix: it must be a symetrical square matrix of size %d x %d.\n\
 The one provided is %d x %d.",_nX,_nX,corrMat.GetNrows(),corrMat.GetNcols()));

        _minputCorr.ResizeTo(corrMat);
        _minputCorr = corrMat;
        
        for(int icol=0; icol<corrMat.GetNcols(); icol++)
	{
	    _minputCorr(icol,icol)=1.0;
	    for(int ilig=0; ilig<corrMat.GetNrows(); ilig++)
	    {
		if( fabs(_minputCorr(ilig,icol) - _minputCorr(icol,ilig)) > 1e-14 )
		     throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TSensitivity::setCorrelationMatrix: invalid correlation matrix: it must be a symetrical square matrix.\n\
This is not the case as M(%d,%d) != M(%d,%d). Their difference is %g",ilig,icol,icol,ilig, _minputCorr(ilig,icol) - _minputCorr(icol,ilig)));
	    }
	}
        
        Double_t max = TMath::MaxElement(_minputCorr.GetNoElements(),
                                         _minputCorr.GetMatrixArray());
        Double_t min = TMath::MinElement(_minputCorr.GetNoElements(),
                                         _minputCorr.GetMatrixArray());
        if (max > 1.0 || min < -1.0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TSensitivity::setCorrelationMatrix: correlation values must belong to the interval [-1.0, 1.0] (min and max found are [%g,%g])",
                                                            min,max));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    _bisInputCorrelated = kTRUE;
}


void TSensitivity::printLog(Option_t *option)
{
    cout << " ** TSensitivity::printLog[" << option << "]" << endl;
    cout << " ** Name[" << GetName() << "] Title[" << GetTitle() << "]" << endl;
    cout << " **   ID[" << getID() << "]" << endl;
    cout << " ** _nS[" << _nS << "]" << endl;
    cout << " ** _nSeed[" << _nSeed << "]" << endl;
    cout << " **  Log of DataServer (unactive)" << endl;
    if ( 0 ) {
        _tds->printLog();
        cout << " **   End of Log of DataServer" << endl;
    }
    cout << " ** End of TSensitivity::printLog[" << option << "]" << endl;
}

} // End of namespace Sensitivity
} // End of namespace URANIE

