//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
// TSensitivity jeu sep 14 09:01:14 CEST 2006
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TSensitivity.h
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   thu sep 14 09:01:14 CEST 2006

 \brief Interface of class URANIE::Sensitivity::TSensitivity

 */
/**

 * @class URANIE::Sensitivity::TSensitivity
 *
 * @brief Description of the class TSensitivity.
 *
 \section use_sensitivity Using the TSensitivity class

 This section describes how to implement a new method of sensitivity
 analysis with this \b abstract class \b TSensitivity.
 The \b pure abstract methods are:
 \li \ref computeIndexes Compute the Sensitivity Indexes
 \li \ref generateSample Generate the Design of Experiments to be evaluated


 Abstract methods must be redefined to mostly specify  a generic method inside the new class.

 \deprecated
 \li runSample : Evaluation of the numerical experimental design created by \ref generateSample
 with an analytical function defined by the construction of an object of type \ref TSensitivity(URANIE::DataServer::TDataServer *tds, void *fcn, Int_t ns, const char *varexpinput="", const char *varexpoutput="")

 \section use_sensitivity_script Using inside a  script
 The following methods must be succeeded one another with \em tsensitivity whatever
 the daughter class fille of the abstract class \b TSensitivity.
 \verbatim
 TSensitivity * scmo = new TSensitivity("method1", 7);



 scmo->computeIndice();
 TNtupleD* ntsampl = scmo->getTupleSample();
 ntsampl->Print();
 \endverbatim

 The method \ref computeIndexes launches successively :
 \dot
 digraph tsensitivity_computeIndice {
 node [shape=record, fontname=Helvetica, fontsize=10];
 a [ label="generateSample" URL="\ref generateSample"];
 b [ label="runSample C" URL=" runSample"];
 a -> b [ arrowhead="open", style="dashed" ];
 }
 \enddot


 *
 */
#ifndef TSENSITIVITY_H
#define TSENSITIVITY_H

////// ROOT
#include "TMethodCall.h"
#include "TTree.h"
#include "TMatrixD.h"
#include "TRandom3.h"
#include "TNamed.h"
#include "TNtupleD.h"
#include "TMath.h"

#include <iostream>
#include <sstream>
using std::cout;
using std::endl;

// Uranie
#include "TDataServer.h"
#include "TCode.h"
#include "TRun.h"
#include "UExceptions.h"
#include "Sensitivity.h"
#ifdef WITH_LIBSSH
#include "TCluster.h"
#endif

#define MAXLENGTHSTRING 50

namespace URANIE {
namespace Sensitivity {
class TSensitivity: public TNamed
{

    // Attributes
    //---------------------------------------------
    /** @name Informations store in the TTree result
     *
     * The values of these attributes are stored in the result TTree object.
     * We access to the result object with the \b getResultTuple method
     * \see getResultTuple
     **/
    //@{
protected:
    Char_t _sOutputAttribute[MAXLENGTHSTRING];  ///< The name of the output attribute
    Char_t _sInputAttribute[MAXLENGTHSTRING];   ///< The name of the input attribute
    Char_t _sOrder[MAXLENGTHSTRING];            ///< The order of sensitivity indexes
    Char_t _sMethod[MAXLENGTHSTRING];           ///< The name of the method
    Char_t _sAlgorithm[MAXLENGTHSTRING];        ///< The name of the algorithm to compute the index
    Double_t _valSobolCrt;                      ///< The value of sensitivity indexes
    Double_t _valSobolCILower;                  ///< The value of lower Condidence Interval (95)
    Double_t _valSobolCIUpper;                  ///< The value of upper Condidence Interval (95)
    TMatrixD _minputCorr;                       ///< Input correlation matrix if sample needs to be correlated
    Bool_t _bisInputCorrelated;                 ///< State whether the input correlation matrix is set
    Bool_t _bgoingThroughError;                 ///< State whether the error must not block the computation

    //@}

    //---------------------------------------------
    /** @name Public attributes
     *
     **/
    //@{
public:
    enum ELauncher
    {
        kCode,kCodeRemote,kFunction, kRun, kUnknown
    };
    
    URANIE::DataServer::TDataServer *_tds;          ///< Pointeur vers un TDS
    TList *_listOfInputAttributes;               ///< List of the input branches
    TList *_listOfOutputAttributes;              ///< List of the input branches
    TString _sTimeAttribute;            ///< The name of the Time attribute

#ifdef WITH_LIBSSH
    URANIE::Launcher::TCluster* _cluster;
    URANIE::Launcher::TCluster* getCluster(){
        return _cluster;
    };
    
    void addCluster(URANIE::Launcher::TCluster* cluster){
        _cluster =  cluster;
        _cluster->setDataServer(_tds);
    };
#endif

    Int_t _nS; ///< The number of simulation or other information depend on the used method
    Int_t _nX;               ///< Dimension of the input
    UInt_t _nY;               ///< Dimension of the target
    UInt_t _nElY;               ///< Number of element for one selected output
    Int_t _nbOut;             ///< Total number of Output to be considered
    Int_t _iOut;              ///< counter for output
    unsigned int _iy;          ///< iterator over number of element
    unsigned int _iely;      ///< iterator over number of element

    ELauncher _nLauncher;                           ///< The type of launcher
    TString _sFunctionName;         ///< The Name of the evaluatuor
    URANIE::Launcher::TCode *_code;                 ///< The tcode
    URANIE::Relauncher::TRun *_run;

    TObjArray *_drawingGarbageCollector;         ///< Garbage collector for prints.
    Int_t _nSeed;                          ///< The seed of the random generator

    Bool_t _bChosenOutputs;                    ///<Fact that the input list is provided or not
    Bool_t _blog;                                  ///< Boolean for edit the log
    Bool_t _bdrawProgressBar; ///< Boolean to know if the progress bar has to be drawn
    Bool_t _bnoIntermediateSaved; ///< Boolean to know if the progress bar has to be drawn

    TString _sIteratorName;        ///< The specific iterator attribute for the method
    TString _sMethodName;         ///< The method name

    TString _sSelectedOutput;           ///< The output
    TString _sSelectedInput;           ///< The input

    map<string, unsigned int> _mAttributeSize;  ///< Map of size of element for attribute;
    map<string, vector<int> > _mAttributeElements;  ///< Map of Elements number to run (if vector subselection is requested)
    vector<string>            _vOutputNames; ///< Name of the output
    
    void setNLauncher(ELauncher codeLauncher){
        _nLauncher = codeLauncher;
    };

    TCanvas *_canvas;                             ///< Canvas object to deal with
    
#ifdef ROOT5
    TMethodCall *fMethodCall; //Pointer to MethodCall in case of interpreted function
    void (*fMethod)(Double_t *x, Double_t *y);
#else
    void (*_pFunction)(double *, double *);
#endif
    TTree* _ntresult;         ///< The TTree of results
    //@}
    
    // Operations
private:
    /** @name Constructor helper
     **/
    void initInputs();
    void initResult();

protected:
    //---------------------------------------------
    /** @name visualization
     **/
    //@{
    //!Create a canvas if needed
    void checkCanvasCreation(bool newcan);
    //! Draws indexes with an histogram
    void drawIndexesHistogram(TString sTitre, const char *select = "", Option_t * option = "");
    //! Draws indexes with an pie chart
    void drawIndexesPie(TString sTitre, const char *select = "", Option_t * option = "");
    //@}

    //!PreTreatment for every output
    /*!
      Doing a pre-treatment when running over outputs.
    */
    virtual void preTreatment(){}

    //!PostTreatment for every output
    /*!
      Doing a post-treatment when running over outputs.
    */
    virtual void postTreatment(){}

    //! Set the "only final file" flag
    /*!
      The output file is not created at every iteration, but only at the end. To be used only for (very) quick
      code when re-running it is not an issue
      The method is put in protected so that this option is changed only by option "noIntermediateSaved" in computeIndexes()
    */
    void setNoIntermediateSaved(Bool_t bbool = kTRUE){
        _bnoIntermediateSaved = bbool;
    }

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    TSensitivity();
    //! Default constructor with the name of a function
    TSensitivity(URANIE::DataServer::TDataServer *tds, const char *fcn, Int_t ns,
                 const char *varexpinput = "", const char *varexpoutput = "",
                 Option_t * option = "");
#ifdef ROOT5
    //! Default constructor with a function
    TSensitivity(URANIE::DataServer::TDataServer *tds, void *fcn, Int_t ns,
                 const char *varexpinput = "", const char *varexpoutput = "",
                 Option_t * option = "");
#else
    TSensitivity(URANIE::DataServer::TDataServer *tds,
                 void (*fcn)(Double_t*,Double_t*),
                 const char *varexpinput, const char *varexpoutput,
                 Int_t ns, Option_t * option = "");
#endif
    //! Default constructor
    TSensitivity(URANIE::DataServer::TDataServer *tds, const char *varexpinput,
                 const char *varexpoutput, Option_t * option = "");
    //! Default constructor with TCode arg
    /*!
     */
    TSensitivity(URANIE::DataServer::TDataServer *tds,
                 URANIE::Launcher::TCode *code, Int_t ns, Option_t * option = "");
    //! Default constructor with TRun arg
    /*!
     */
    TSensitivity(URANIE::DataServer::TDataServer *tds,
                 URANIE::Relauncher::TRun *run, Int_t ns, Option_t * option = "");
    //! Default destructor
    virtual ~TSensitivity();
    //@}

    //---------------------------------------------
    /** @name Setting and Getting attributs
     **/
    //@{
    Int_t getID(){
        return ID_SENSITIVITY;
    }
    //!Get the result ntuple (default parameter unused but for Morris method)
    virtual TTree * getResultTuple(bool commonresulttuple=true){
        UNUSED(commonresulttuple);
        return _ntresult;
    }

    vector<int>* getAttributeElements(string str){
        return ((_mAttributeElements.count(str)==0)?NULL:&_mAttributeElements[str]);
    }

#ifdef ROOT5
    TMethodCall *GetMethodCall() const{
        return fMethodCall;
    }
    void setFunction(void *fcn, Int_t nx = -1, Int_t ny = 1);
#else
    void setFunction(void (*fct)(Double_t*, Double_t*), Int_t nx = -1, Int_t ny = 1);
#endif
    TString getFunctionName(){
        return _sFunctionName;
    }
    void setSeed(UInt_t nval){
        _nSeed = nval;
    }
    UInt_t getSeed(){
        return _nSeed;
    }
    //! Set the Method name
    void setMethodName(const char *str){
        _sMethodName = TString(str);
    }
    //! Get the method name
    const char * getMethodName(){
        return _sMethodName.Data();
    }

    //! Get the \em noIntermediateSaved flag
    /*!
      \return (Bool_t) the \em  noIntermediateSaved  flag (boolean)
    */
    Bool_t getNoIntermediateSaved(){
        return _bnoIntermediateSaved;
    }

    //! Get the name of the iterator attribut of the method
    const char * getIteratorName(){
        return _sIteratorName.Data();
    }
    //! Set the iterator name devoted to compute sensitivity indexes
    void setSensitivityIteratorName(const char *str){
        _sIteratorName = TString(str);
    }
    /*!
     * \brief Set the name of the \b time attribute (only one)
     * \param sname (TString)
     */
    void setTimeName(TString sname);
    /*!
     * \brief Get the name of the \b time attribute
     * \return The name of the time attribute
     */
    TString getTimeName() { return _sTimeAttribute;}
    //! Set the "draw progress bar" flag
    /*!
      The progress bar will be drawn if and only if the "draw progress bar"
      flag is true.
    */
    void setDrawProgressBar(Bool_t bbool = kTRUE){
        _bdrawProgressBar = bbool;
    }

    //! Set the "using error results anyway" option
    /*!
      CAREFULL : this is really not recommended, but can be set to further understand weird behaviour
    */
    void setUsingErrors(bool thebool){ _bgoingThroughError=thebool;}
    
    //! Get the \em clean flag
    /*!
      \return (Bool_t) the \em clean flag (boolean)
    */
    Bool_t getDrawProgressBar(){
        return _bdrawProgressBar;
    }
    /*! Get the flag is the input attributes are correlated
     *
     * \return  (Bool_t) the \em _bisInputCorrelated flag (boolean)
     */
    Bool_t isInputCorrelated() {
        return _bisInputCorrelated;
    }
    /*! Return the matrix of the input correlation
     *
     * \return (TMatrixD) The input correlation matrix
     */
    TMatrixD getMatrixInputCorrelation() {
        return _minputCorr;
    }

    /*!
     * \brief Get the number of input attributes
     * \return The number of input attributes ( _nX )
     */
    Int_t getNInput() {
        return _nX;
    }

    /*!
     * \brief Get the number of output attributes
     * \return The number of output attributes ( _nY )
     */
    Int_t getNOutput() {
        return _nY;
    }
    //@}

    //---------------------------------------------
    /** @name Generation of the sample
     **/
    //@{
    virtual void generateSample(Option_t * option = "") = 0;
    void setInputCorrelationMatrix(TMatrixD Corr);
    //@}

    //---------------------------------------------
    /** @name Parse the option
     **/
    //@{
    //! Read the possible options
    /*!
      Possible options are (common to all methods)
      \li output="": restain the list of output to be analysed using the usual format "a:b:c"
      \li noIntermediateSaved: don't save intermediate file if TLauncher is used at some point
    */
    virtual void parseOption(Option_t *option = "");
    //@}

    //---------------------------------------------
    /** @name Check the output list
     **/
    //@{
    //! Check the output list requested by the user
    /*!
      Function introduced in order to be able to run over few elements of a vector y.
      It is called before going over the loop of its elements and check if one or more output contains "[" "]"
      (for example  when the option contains "output=y[2]:y[4]")
    */
    void checkOutputRequested(string attlist, bool fromoption=false);
    //@}

    //---------------------------------------------
    /** @name Compute the indexes
     **/
    //@{
    //! Compute the Sensitivity Indexes.
    /*!
      Possible options are (common to all methods)
      \li output="": restain the list of output to be analysed using the usual format "a:b:c"
      \li noIntermediateSaved: don't save intermediate file if TLauncher is used at some point
    */
    void computeIndexes(Option_t * option = "");

    /*!
     * \brief Evaluates the index from a Specific TDataServer
     *
     * The TDS must contain an attribute Name + "n__iter__sobol" with values
     * \li            -1 : for the M matrix
     * \li             0 : for the N matrix
     * \li          ix+1 : for the matrix Mi
     * \li 100*_nX+ ix+1 : for the matrix Ni
     *
     * \param option (Option_t *) option to pass [""]
     */
    virtual void evaluateIndexes(Option_t * option = "") = 0;

    /*!
     * \brief Method to fill in the tree the value of Sensitivity indexes for an input attribute
     *
     * \param sinputname (const char *) : the name of the input attribute
     * \param sorder (const char *) : the order ["First|Total"] of sensitivity index
     * \param dval (Double_t) : the value of the sensitivity index
     * \param algo (const char *) : the name of the algorithm used
     * \param dvalCILower (Double_t) : the value of the lower Condidence Interval (-1.0)
     * \param dvalCIUpper (Double_t) : the value of the upper Condidence Interval (-1.0)
     */
    void fillIndex(const char * sinputname, const char * sorder,
                   Double_t dval, const char* algo = "",
                   Double_t dvalCILower = -1.0, Double_t dvalCIUpper = -1.0);
    //@}

    //---------------------------------------------
    /** @name Visualization and Scan the data
     **/
    //@{
    virtual void createTuple(Option_t *option = "");
    //! Draws the indexes
    virtual void drawIndexes(TString sTitre, const char *select = "",
                             Option_t * option = "");
    //@}

    //---------------------------------------------
    /** @name Printing Log
     **/
    //@{
    void setLog(){
        _blog = kTRUE;
    }
    void unsetLog(){
        _blog = kFALSE;
    }
    void changeLog(){
        _blog = _blog ? kFALSE : kTRUE;
    }
    Bool_t getLog(){
        return _blog;
    }
    virtual void printLog(Option_t *option = "");
    //@}

    ClassDef(URANIE::Sensitivity::TSensitivity, ID_SENSITIVITY)

};

} // Fin du namespace Sensitivity
} // Fin du namespace URANIE

#ifdef ROOT5
R__EXTERN URANIE::Sensitivity::TSensitivity *gSensitivity;
#endif

#endif
