//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
/*!
 \file   TShapley.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   2020/09/29

 \brief Implement the URANIE::Sensitivity::TShapley class

*/

/// ROOT
#include "TLeaf.h"
#include "TTree.h"

/// Uranie
#include "TShapley.h"
#include "TSampling.h"
#include "TPermutation.h"
#include "TNormalDistribution.h"
#include "TAttributeFormula.h"
#include "UTimer.h"

#define STRNAMEOFATTNI "__xNi__%d"
#define STRNAMEOFATTNORMAL "_n_%s_" /// "__normal__%s__"
#define DTRANSLATION 1000

using std::cout;
using std::endl;
using namespace URANIE::DataServer;
using namespace URANIE::Sampler;

ClassImp (URANIE::Sensitivity::TShapley)

namespace URANIE
{
namespace Sensitivity
{
TShapley::TShapley():
    TSensitivity()
{
    SetName("Shapley");
    setMethodName("Shapley");
    SetTitle("my title of the Shapley values");
    cout << endl << " *** Default constructor TShapley (empty) ***" << endl;
}


TShapley::TShapley(URANIE::DataServer::TDataServer *tds, const char *varexpinput,
                   const char *varexpoutput, Option_t* option) :
    TSensitivity(tds, varexpinput, varexpoutput, option)
{
    SetName("Shapley");
    setMethodName("Shapley");
    SetTitle("my title of the Shapley indexes");
    _sIteratorName = Form("%s__Nv__%s", getMethodName(), _tds->GetName());
    _sIteratorNameNo = Form("%s__No__%s", getMethodName(), _tds->GetName());
    _sIteratorNameNi = Form("%s__Ni__%s", getMethodName(), _tds->GetName());

    /// Looking for the iterator attributes (Nv, No and Ni) for the Shapley Method
    try
      {

        if ( _sSelectedOutput.Length() == 0 )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TShapley constructor \nThe output attribute \"%s\" is empty", _sSelectedOutput.Data()));
        if ( !_tds->isAttribute(_sIteratorName) )
          throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                     Form("TShapley constructor ERROR : the TDataServer \"%s\" do not contains the Nv iterator attribute \"%s\".",
                                                          _tds->GetName(), _sIteratorName.Data()));
        if ( !_tds->isAttribute(_sIteratorNameNo) )
          throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                     Form("TShapley constructor ERROR : the TDataServer \"%s\" do not contains the No iterator attribute \"%s\".",
                                                          _tds->GetName(), _sIteratorNameNo.Data()));
        if ( !_tds->isAttribute(_sIteratorNameNi) )
          throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                     Form("TShapley constructor ERROR : the TDataServer \"%s\" do not contains the Ni iterator attribute \"%s\".",
                                                          _tds->GetName(), _sIteratorNameNi.Data()));

        // case of _nM
        _tds->computeStatistic(_sIteratorName);
        _nM = (Int_t)(_tds->getAttribute(_sIteratorName)->getMaximum());
        cout << " ** _nX[" << _nX << "]  _nM[" << _nM << "]" << endl;

        if ( _nM < 1 )
          throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                     Form("TShapley constructor ERROR : in the TDataServer \"%s\",  the loop in _nM is zero[%d].",
                                                          _tds->GetName(), _nM));

        // case of _No
        TString sSelect =  Form("(%s==1)&&(%s>%d)&&(%s<%d)", _sIteratorName.Data(), _sIteratorNameNo.Data(), DTRANSLATION-10, _sIteratorNameNo.Data(), DTRANSLATION+10+_nX);
        if ( _blog ) cout<< "    *** sSelect[" << sSelect << "]" << endl;
        _tds->setSelect(sSelect);
        Int_t npatnino = _tds->getNPatterns();

        _tds->computeStatistic(_sIteratorNameNi);

        _nO = (Int_t)(_tds->getAttribute(_sIteratorNameNi)->getMaximum());
        if ( _blog ) cout << " ** _nX[" << _nX << "]  _nM[" << _nM << "] _nO[" << _nO << "]" << endl;



        if ( _nO < 1 )
          throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                     Form("TShapley constructor ERROR : in the TDataServer \"%s\",  the loop in _nO is zero[%d].",
                                                          _tds->GetName(), _nO));

        _nI = npatnino / _nO;

        if ( _nI < 3 )
          throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                     Form("TShapley constructor ERROR : in the TDataServer \"%s\",  the loop in _nI [%d] is leaser than 3.",
                                                          _tds->GetName(), _nI));

        _tds->clearSelect();
        // case of _nM
        _tds->setSelect(Form("%s==0", _sIteratorName.Data()));
        _nV = _tds->getNPatterns();

        if ( _nV < 3 )
          throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                     Form("TShapley constructor ERROR : in the TDataServer \"%s\",  the loop in _nV is [%d] leaser than 3.",
                                                          _tds->GetName(), _nV));

        _tds->clearSelect();
        if ( _blog ) {
            cout << " *** _nV[" << _nV << "] _nM[" << _nM << "] _nO[" << _nO << "] _nI[" << _nI << "]" << endl;
            cout << " *** input[" << varexpinput << "] out[" << varexpoutput << "] npat[" << _tds->getNPatterns() << "]" << endl;
        }

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    } catch (URANIE::Exceptions::UExceptions& ue)
    {
        ue.printMessage();
        if (ue.getMessageType() > URANIE::Exceptions::UExceptions::kWARNING)
            throw ue;
    }

    parseOption(option);
}

TShapley::TShapley(TDataServer *tds, const char *fcn, Int_t ns, const char *varexpinput,
               const char *varexpoutput, Option_t* option) :
    TSensitivity(tds, fcn, ns, varexpinput, varexpoutput, option),
    _nV(ns), _nO(1), _nI(3), _nM(0),
    _matNormalCovariance(0,0), _nCondSampler(kNorta)
{
    init(option);
}

TShapley::TShapley(TDataServer *tds, void (*fcn)(Double_t*,Double_t*),
               Int_t ns, const char *varexpinput, const char *varexpoutput,
               Option_t * option) :
    TSensitivity(tds, fcn, varexpinput, varexpoutput, ns),
    _nV(ns), _nO(1), _nI(3), _nM(0),
    _matNormalCovariance(0,0), _nCondSampler(kNorta)
{
    init(option);
}

TShapley::TShapley(URANIE::DataServer::TDataServer *tds,
        URANIE::Launcher::TCode *code, Int_t ns, Option_t * option) :
        TSensitivity(tds, code, ns, option),
        _nV(ns), _nO(1), _nI(3), _nM(0),
        _matNormalCovariance(0,0), _nCondSampler(kNorta)
{
    init(option);
}

TShapley::TShapley(URANIE::DataServer::TDataServer *tds,
        URANIE::Relauncher::TRun *run, Int_t ns, Option_t * option) :
    TSensitivity(tds, run, ns, option),
    _nV(ns), _nO(1), _nI(3), _nM(0),
    _matNormalCovariance(0,0), _nCondSampler(kNorta)
{
    init(option);
}


TShapley::~TShapley()
{
    if ( _blog )
        cout << " *** destroy TShapley(" << GetName() << ") Title[" << GetTitle() << "] *** " << endl;
}


void TShapley::init(Option_t* option) {
    SetName("Shapley");
    setMethodName("Shapley");
    SetTitle("my title of the Shapley indexes");
    _sIteratorName = Form("%s__Nv__%s", getMethodName(), _tds->GetName());
    _sIteratorNameNo = Form("%s__No__%s", getMethodName(), _tds->GetName());
    _sIteratorNameNi = Form("%s__Ni__%s", getMethodName(), _tds->GetName());

    /// not empty TDS not allowed
    try
      {
        if ( ( _tds->hasTuple() ) && ( _tds->getNPatterns() > 0) )
          throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                     Form("TShapley constructor ERROR : the TDataServer \"%s\" is not empty: %d patterns.",
                                                          _tds->GetName(), _tds->getNPatterns()));
      } catch (URANIE::Exceptions::UExceptions& ue)
      {
        ue.printMessage();
        if (ue.getMessageType() > URANIE::Exceptions::UExceptions::kWARNING)
          throw ue;
      }
    // Init the correlation matrix to unit
    _minputCorr.ResizeTo(_nX, _nX);
    _minputCorr.UnitMatrix();

    // Add the auxilliary attributes
    _sSelectedInput += Form(":%s", _sIteratorName.Data());
    _sSelectedInput += Form(":%s", _sIteratorNameNo.Data());
    _sSelectedInput += Form(":%s", _sIteratorNameNi.Data());

    parseOption(option);
}

void TShapley::setNv(Int_t nv) {
    try
      {
        if ( nv < 1 )
          throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                     Form("TShapley::setNv ERROR : the number Nv[%d] must be greater than 1.",
                                                          nv));
        _nV = nv;
      } catch (URANIE::Exceptions::UExceptions& ue)
      {
        ue.printMessage();
        if (ue.getMessageType() > URANIE::Exceptions::UExceptions::kWARNING)
          throw ue;
      }
}

void TShapley::setNm(Int_t nm) {
    try
      {
        if ( nm < 1 )
          throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                     Form("TShapley::setNm ERROR : the number Nm[%d] must be greater than 1.",
                                                          nm));
        _nM = nm;
      } catch (URANIE::Exceptions::UExceptions& ue)
      {
        ue.printMessage();
        if (ue.getMessageType() > URANIE::Exceptions::UExceptions::kWARNING)
          throw ue;
      }
}

void TShapley::setNo(Int_t no) {
    try
      {
        if ( no < 1 )
          throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                     Form("TShapley::setNo ERROR : the number No[%d] must be greater than 1.",
                                                          no));
        _nO = no;
      } catch (URANIE::Exceptions::UExceptions& ue)
      {
        ue.printMessage();
        if (ue.getMessageType() > URANIE::Exceptions::UExceptions::kWARNING)
          throw ue;
      }
}

void TShapley::setNi(Int_t ni) {
    try
      {
        if ( ni < 1 )
          throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                     Form("TShapley::setNi ERROR : the number Ni[%d] must be greater than 1.",
                                                          ni));
        _nI = ni;
      } catch (URANIE::Exceptions::UExceptions& ue)
      {
        ue.printMessage();
        if (ue.getMessageType() > URANIE::Exceptions::UExceptions::kWARNING)
          throw ue;
      }
}

void TShapley::setSamplerMethod(EConditionalSampling nMethod)
{
    try {
        if ( nMethod == kUnknown )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TShapley::setSamplerMethod ERROR :\n The method \"kUnknown\" not authorized.\n Only \"kNorta\" method was implemented.");
        if ( nMethod == kCopulas )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TShapley::setSamplerMethod ERROR :\n Method \"kCopulas\" not yet implemented.\n Only \"kNorta\" method was implemented.");
        if ( nMethod == kVines )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TShapley::setSamplerMethod ERROR :\n Method \"kVines\" not yet implemented.\n Only \"kNorta\" method was implemented.");
        _nCondSampler = nMethod;
    } catch (URANIE::Exceptions::UErrorExceptions& ue) {
        ue.printMessage();
        throw ue;
    }
}

void TShapley::generateSample(Option_t * option)
{
    if (_blog) {
        cout << endl;
        cout << "*********************************" << endl;
        cout << " ***** TShapley::" << __func__ << endl;
        cout << " ***** _nSeed[" << _nSeed << "]" << endl;
        cout << " ***** _nCondSampler[" << _nCondSampler << "]" << endl;
        cout << " ***** _nV[" << _nV << "] _nM[" << _nM << "] _nO[" << _nO << "] _nI[" << _nI << "]" << endl;
        cout << " ***** option[" << option << "]" << endl;
    }

    TString soption = TString(option);
    soption.ToLower();

    // not empty TDS not allowed
    try
    {
        if ( ( _tds->hasTuple() ) && ( _tds->getNPatterns() > 0 ) )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TShapley::generateSample ERROR :\n the TDataServer \"%s\" is not empty: %d patterns.",
                                                            _tds->GetName(), _tds->getNPatterns()));
        if ( _nCondSampler == kCopulas )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TShapley::generateSample ERROR :\n for TDataServer \"%s\", the sampler method \"kCopulas\" not yet implementd.",
                                                            _tds->GetName()));
        if ( _nCondSampler == kVines )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TShapley::generateSample ERROR :\n for TDataServer \"%s\", the sampler method \"kVines\" not yet implementd.",
                                                            _tds->GetName()));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    } catch (URANIE::Exceptions::UExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }


    Int_t nAtt = _tds->getNAttributes();

    // Generate the first sample of size _nV
    {
        TString sMethodSampling = "lhs";
        if (soption.Contains("srs"))
            sMethodSampling = "srs";

        if ( _blog ) {
            cout << " *** step nV : generate the nV DoE with method[" << sMethodSampling << "]" << endl;
        }
        TSampling *tmpSampling = new TSampling(_tds, sMethodSampling, _nV);
        tmpSampling->setSeed(_nSeed);
        tmpSampling->setCorrelationMatrix(_minputCorr);
        tmpSampling->generateSample();

        if ( _blog )
            _tds->computeCorrelationMatrix().Print();

        if ( _blog ) cout<< " *** npat[" << _tds->getNPatterns() << "] natt[" << _tds->getNAttributes() << "]" << endl;
        _tds->addAttribute(_sIteratorName,   "0");
	_tds->getAttribute(_sIteratorName)->delShare();
        _tds->addAttribute(_sIteratorNameNo, "0");
	_tds->getAttribute(_sIteratorNameNo)->delShare();
        _tds->addAttribute(_sIteratorNameNi, "0");
	_tds->getAttribute(_sIteratorNameNi)->delShare();
        _tds->getTuple()->SetScanField(0);
        if ( _blog ) {
            cout << " scan of DoE" << endl;
            _tds->scan("*");
        }
        // Clean
        delete tmpSampling; tmpSampling = NULL;
    }

    // Add the internal attributes
    _tds->getAttribute(_sIteratorName)->setOrigin(TAttribute::kAttribute); // kInternal / kAttribute
    _tds->getAttribute(_sIteratorNameNo)->setOrigin(TAttribute::kAttribute); // kInternal / kAttribute
    _tds->getAttribute(_sIteratorNameNi)->setOrigin(TAttribute::kAttribute); // kInternal / kAttribute

    //
    Bool_t bexact = kFALSE; //  kTRUE kFALSE
    if ( _blog ) cout<< "  ** _nX[" << _nX << "] _nSeed["<< _nSeed  << "]" << endl;
    TPermutation myPermut(Form("_Song_Algorithm_%s", GetName()),  "Permutations in the Song Algorithm for Shapley", _nX, bexact);
    myPermut.setSeed(_nSeed);

    //myPermut.setLog();

    if ( _blog ) myPermut.printLog();
    Int_t *vIndexes = new Int_t[_nX];
    memset(vIndexes, 0, sizeof(Int_t) * (_nX));
    if ( _blog ) {
        cout << " *** original vector (Zero) [";
        for(Int_t i=0; i<_nX; i++)
            cout << vIndexes[i] <<", ";
        cout << "]"<<endl;
    }

    Int_t nL0, nL1;
    Int_t ilno;
    TString sAttlno ="";
    TString sSelect ="";

    if ( _blog ) _listOfInputAttributes->Print();

    // Compute the Variance matrix
    TMatrixD matVariance(_nX, _nX);
    matVariance.UnitMatrix();

    TStochasticAttribute *stoatt;
    // diagonal terms
    for(Int_t ix=0; ix<_nX; ix++) {
        stoatt = dynamic_cast<TStochasticAttribute *>(_listOfInputAttributes->At(ix));
        if ( _blog ) cout << " *** ix(" << ix << "/" << _nX << ") Name[" << stoatt->GetName() << "] TheoreticalStdDev[" << stoatt->getTheoreticalStdDev() << "]" << endl;
        matVariance(ix, ix) = stoatt->getTheoreticalStdDev() * stoatt->getTheoreticalStdDev();
    }

    // Triangular terms
    for(Int_t ix=0; ix<_nX; ix++) {
        for(Int_t jx=0; jx<ix; jx++) {
            matVariance(ix, jx) = matVariance(jx, ix) = _minputCorr(ix, jx) * TMath::Sqrt( matVariance(ix, ix) * matVariance(jx, jx));
        }
    }

    if ( _blog ) {
        cout << " ********  matVariance - matVariance - matVariance - matVariance *****";
        matVariance.Print("");
        cout << " *************************" << endl;
    }

    // Compute the Normal correlation matrix
    computeNormalCorrelationMatrix(matVariance);

    TMatrixD matVarCov0(_nX, _nX), matVarCov1(_nX, _nX), matVarCov01(_nX, _nX);
    matVarCov0.UnitMatrix();
    matVarCov1.UnitMatrix();
    matVarCov01.Zero();
    TVectorD  vecOf01(_nX);


    TMatrixD matCorrTmp(_nX, _nX);
    matCorrTmp.UnitMatrix();

    // Create the return TDS
    TDataServer *theTDSres = new TDataServer("tds_res", "the tds of results");
    // Copy the TStochastic attributes
    TDataServer *copyTDS0 = new TDataServer( Form("tds0_%s_copy_%d", _tds->GetName(), _nM),
                                             Form("the tds of zero title[%s]", _tds->GetTitle()));
    TString sname = "";

    for (Int_t ix = 0; ix < _nX; ix++) {
        sname = _listOfInputAttributes->At(ix)->GetName();
        if ( _blog )
            cout << " --- init TGaussianSampling :: case of ix(" << ix << "/" << _nX << ") att : sname[" << sname << "]" << endl;
        copyTDS0->addAttribute( new TNormalDistribution(Form(STRNAMEOFATTNORMAL, sname.Data()), 0.0, TMath::Sqrt( matVariance(ix, ix)) ));
	copyTDS0->getAttribute(Form(STRNAMEOFATTNORMAL, sname.Data()))->delShare();

        theTDSres->addAttribute( new TAttribute(sname));
	theTDSres->getAttribute(sname)->delShare();
    }
    theTDSres->addAttribute( new TAttribute(_sIteratorName));
    theTDSres->getAttribute(_sIteratorName)->delShare();
    theTDSres->addAttribute( new TAttribute(_sIteratorNameNo));
    theTDSres->getAttribute(_sIteratorNameNo)->delShare();
    theTDSres->addAttribute( new TAttribute(_sIteratorNameNi));
    theTDSres->getAttribute(_sIteratorNameNi)->delShare();

    // Create the TGaussianSampling
    Int_t ndefaultSample = 5;
    TGaussianSampling *tga = new TGaussianSampling(copyTDS0, "srs", ndefaultSample);
    tga->setSeed(_nSeed);
    tga->generateSample();
    if ( _blog )  copyTDS0->scan("*");

//    tga->setIteratorNameNo(_sIteratorNameNo);
//    tga->setIteratorNameNi(_sIteratorNameNi);

//    cout<< " **** _sIteratorName[" << _sIteratorName << "]" << endl;
//    copyTDS0->addAttribute(_sIteratorName,   "0");
    if ( _blog ) cout<< " **** _sIteratorNameNo[" << tga->getIteratorNameNo() << "]" << endl;
    copyTDS0->addAttribute(tga->getIteratorNameNo(), "0");
    copyTDS0->getAttribute(tga->getIteratorNameNo())->delShare();

    if ( _blog ) cout<< " **** _sIteratorNameNi[" << tga->getIteratorNameNi() << "]" << endl;
    copyTDS0->addAttribute(tga->getIteratorNameNi(), "0");
    copyTDS0->getAttribute(tga->getIteratorNameNi())->delShare();
    
    if ( _blog ) copyTDS0->scan("*");

    // Pass the VarianceCovariance Matrix
    for(Int_t ix=0; ix<_nX; ix++) {
        for(Int_t jx=0; jx<ix; jx++) {
            //tga->setUserVarianceCovarianceMatrix(ix, jx, matVarCovar(ix, jx));
            tga->setUserCorrelation(ix, jx, _minputCorr(ix, jx));
        }
    }

    if ( _blog )  {
        cout << " **** Correlation Matrix for TGaussianSampling" << endl;
        tga->getCorrelationMatrix().Print();
        cout << " **** VarianceCovriance Matrix for TGaussianSampling" << endl;
        tga->getCovarianceMatrix().Print();
    }


    // Init the TList of leaves to Store and Transfert
    TList *listOfLeavesStored = new TList();
    for (Int_t ix = 0; ix < _nX; ix++) {
        sname = _listOfInputAttributes->At(ix)->GetName();
        if ( _blog )
            cout << " --- init listOfLeavesStored :: case of ix(" << ix << "/" << _nX << ") att : sname[" << sname << "]" << endl;
        listOfLeavesStored->Add(copyTDS0->getTuple()->GetLeaf(Form(STRNAMEOFATTNORMAL, sname.Data())));
    }

    listOfLeavesStored->Add(copyTDS0->getTuple()->GetLeaf(tga->getIteratorNameNo()));
    listOfLeavesStored->Add(copyTDS0->getTuple()->GetLeaf(tga->getIteratorNameNi()));

    if ( _blog ) listOfLeavesStored->Print("all");

    UTimer timer(_nM, "Uranie Shapley : the generateSample loop");
    if (_bdrawProgressBar) {
        cout << "The ProgressBar for TShapley::" << __func__ <<" loop(" << _nM << ")" << endl;
      }

    // loop on the permutations (m)
    for(Int_t im=0; im<_nM; im++) {
        if ( _blog ) cout << " *** loop im[" << im << "/" << _nM<<"]"<<endl;
        // Generate a permutation
        myPermut.generatePermutation(vIndexes, Form("Shapley--step im[%d/%d]", im, _nM));

        if ( _blog ) {
            cout << " *** Current permutation [";
            for(Int_t i=0; i<_nX; i++)
                cout << vIndexes[i] <<", ";
            cout << "]"<<endl;
        }

        vecOf01.Zero();

//        // Copy the TStochastic attributes
//        TDataServer *copyTDS1 = new TDataServer( Form("tds1_%s_copy_%d", _tds->GetName(), im),
//                                                 Form("the tds of one title[%s]", _tds->GetTitle()));

//        if ( _blog ) cout << " *** copyTDS0 patterns[" << copyTDS0->getNPatterns() << "] att[" << copyTDS0->getNAttributes() << "]" << endl;
//        if ( _blog ) cout << " *** copyTDS1 patterns[" << copyTDS1->getNPatterns() << "] att[" << copyTDS1->getNAttributes() << "]" << endl;

        // loop on nX
        for(Int_t ix=0; ix<_nX-1; ix++) {
            ilno = vIndexes[ix]-1;
            sAttlno = _listOfInputAttributes->At(ilno)->GetName();
            if ( _blog ) cout << " --- case of ix(" << ix << "/" << _nX << ") att : ilno[" << ilno << "] Name[" << ilno+1 << "][" << sAttlno << "]" << endl;
//            copyTDS0->delAttribute(sAttlno);
//            copyTDS1->addAttribute((_tds->getAttribute(_listOfInputAttributes->At(ilno)->GetName())));
//            nL0 = copyTDS0->getNAttributes();
//            nL1 = copyTDS1->getNAttributes();
            vecOf01(ilno) = 10.0+im;

            if ( _blog ) {
                vecOf01.Print("the vector of 01");
//                cout << " ********  _minputCorr before Splitting *****";
//                _minputCorr.Print();
            }

            unsetLog();
            generateConditionnalSampleByNorta(tga, theTDSres, listOfLeavesStored, matVariance, vecOf01);
            if ( _blog ) {
                vecOf01.Print("the vector of 01 After generateConditionnalSampleByNorta");
            }

            if ( _blog )
                theTDSres->scan("*");

            // transfert the data from the samplerTDS into the original TDS
            transfertDataTDS(theTDSres);
            if ( _blog ) cout<< " *** end of transfertDataTDS ix(" << ix << "/" << _nX << ")" << endl;

            //
            unsetLog();
            // Clean
            //delete tdsres; tdsres =NULL;


//            // Erase the values
//            copyTDS0->deleteTuple();
        } // End of ix/_nX

#ifndef WIN32
      if (_bdrawProgressBar) {
          timer.DrawProgressBar(im);
        }
#endif
    } // End Of loop on the indexes _nM (permutations)

    if (_bdrawProgressBar)
      cout << endl << " ** time elapsed " << timer.GetElapsedTime() << endl;

    if ( _blog ) cout<< "  ** _sSelectedInput[" << _sSelectedInput << "]" << endl;
    _tds->exportData("_shapley_sampling_.dat", _sSelectedInput);

    // Clean
    delete listOfLeavesStored; listOfLeavesStored = NULL;
    delete tga; tga = NULL;
    delete copyTDS0; copyTDS0 = NULL;
    delete theTDSres; theTDSres = NULL;
//        delete copyTDS1; copyTDS1 =NULL;
    delete[] vIndexes;
    vIndexes=NULL;
    if ( _blog ) cout << " ** End Of TShapley::" << __func__ << endl;

}

void TShapley::evaluateIndexes(Option_t * option)
{
    if ( _blog )
    {
        cout << "***********************************" << endl;
        cout << " ** TShapley::[" << __func__  << "]" <<     option << "]" << endl;
        cout << " ** npat[" << _tds->getNPatterns() << "]" << endl;
        cout << " ** iter of Y [" << _iy << "]" << endl;
        cout << " ** option[" << option << "]" << endl;
        cout << " ** _nV[" << _nV << "] _nM[" << _nM << "] _nO[" << _nO << "] _nI[" << _nI << "]" << endl;
        cout << " ** _nX[" << _nX << "]" << endl;
    }

    TAttribute *atty = (TAttribute *)(_listOfOutputAttributes->At(_iy));
    TString soutputy = atty->GetName();
    if ( _blog ) cout<< " *** soutputy[" << soutputy << "]" << endl;

    // Compute the Mean and Variance for the nV sample
    TString sSelectV = Form("%s==0", _sIteratorName.Data());
    if ( _blog ) cout<< " *** sSelectV[" << sSelectV << "]" << endl;
    TString sSelect1 =  "";
    TString sSelectX =  "";
    Double_t dvarall;

    _tds->setSelect(sSelectV);

    _tds->computeStatistic(soutputy); // , sSelectV);
    Long64_t npattall=_tds->getNPatterns(); // sSelect.Data());
    Double_t dmeanall = _tds->getAttribute(soutputy)->getMean();
    Double_t dstdall  = _tds->getAttribute(soutputy)->getStd();
    dvarall = dstdall * dstdall;
    if ( _blog ) cout<< " *** soutputy[" << soutputy << "] sSelectV[" << sSelectV << "] N[" << npattall << "] Mean[" << dmeanall << "] Sigma[" << dstdall << "] Variance[" << dvarall << "]" << endl;
    _tds->clearSelect();



    Double_t dprecC=0.0;
    TVectorD vecShapley(_nX);
    vecShapley.Zero();

    TVectorD vecMeanCond(_nX);
    TVectorD  vecOf01(_nX);
    Double_t dstdcond, dvarcond;

    Int_t nxcrt;
    TString sxcrt = "";
    Double_t d1OverM = 1.0 /( dvarall *_nM );

    UTimer timer(_nM, Form("Uranie Shapley : the evaluateIndexes loop for y[%s]", soutputy.Data()));
    if (_bdrawProgressBar) {
        cout << "The ProgressBar for TShapley::" << __func__ <<" for y["<< soutputy << "] loop(" << _nM << ")" << endl;
      }
    // loop on the permutations (m)
    for(Int_t im=0; im<_nM; im++) {
        if ( _blog ) cout << " --- loop m : im(" << im << "/" << _nM << ")" << endl;
        sSelectV = Form("%s==%d", _sIteratorName.Data(), im+1);
        if ( _blog ) cout<< "    *** sSelectV[" << sSelectV << "]" << endl;
//        _tds->getTuple()->SetScanField(0);
//        _tds->scan("*", sSelectV);

        vecMeanCond.Zero();
        vecOf01.Zero();
        dprecC=0.0;

        for(Int_t ix=0; ix<_nX-1; ix++) {
            if ( _blog ) cout << "    --- case of ix(" << ix << "/" << _nX << ") im(" << im << "/" << _nM << ") dprec[" << dprecC << "]" << endl;
            sSelectX =  Form("(%s)&&(%s>%d)&&(%s<%d)", sSelectV.Data(), _sIteratorNameNo.Data(), (ix+1)*DTRANSLATION-10, _sIteratorNameNo.Data(), (ix+1)*DTRANSLATION+10+_nX);
            if ( _blog ) cout<< "    *** sSelectX[" << sSelectX << "]" << endl;
            //_tds->scan("*", sSelectX);
            _tds->computeStatistic(_sIteratorNameNo, sSelectX);
            Double_t dmeannO = _tds->getAttribute(_sIteratorNameNo)->getMean();
            Double_t dstdnO = _tds->getAttribute(_sIteratorNameNo)->getStd();
            Double_t dminnO = _tds->getAttribute(_sIteratorNameNo)->getMinimum();
            Double_t dmaxnO = _tds->getAttribute(_sIteratorNameNo)->getMaximum();
            nxcrt = (Int_t)(dminnO)-(ix+1)*DTRANSLATION;
            sxcrt = _listOfInputAttributes->At(nxcrt-1)->GetName();
            if ( _blog ) cout << "       --- _sIteratorNameNo [" << nxcrt << "] Name[" << sxcrt << "] :: Mean[" << dmeannO << "] Range[" << dminnO << ", " << dmaxnO << "]" << endl;
            vecOf01(nxcrt-1)=1;

            // loop on No
            for(Int_t io=0; io<_nO; io++) {
                sSelect1 = Form("%s&&(%s==%d)", sSelectX.Data(), _sIteratorNameNi.Data(), io+1);
                if ( _blog ) {
                    cout<< "    *** sSelect1[" << sSelect1 << "]" << endl;
                    _tds->scan(Form("%s:%s",_sSelectedInput.Data(), soutputy.Data()), sSelect1);
                }
                _tds->computeStatistic(soutputy, sSelect1);
                dstdcond = _tds->getAttribute(soutputy)->getStd();
                vecMeanCond(nxcrt-1) = ( 1.0 * io * vecMeanCond[nxcrt-1] + dstdcond * dstdcond ) / ( 1.0 * (io+1));
            } // End of Loop on io/_nO
            if ( _blog ) vecMeanCond.Print(Form("Vec Mean in case[%d/%d] nxcft[%d] Name[%s] dprecC[%e] ", ix+1, _nX, nxcrt, sxcrt.Data(), dprecC));
            vecShapley(nxcrt-1) += vecMeanCond(nxcrt-1) - dprecC;
            dprecC = vecMeanCond(nxcrt-1);
            if ( _blog ) vecShapley.Print(Form("Shapley in case[%d/%d] nxcrt[%d] Name[%s]", ix+1, _nX, nxcrt, sxcrt.Data()));
        } // End of Loop on ix/_nX
        // Case of last ix
        if ( _blog ) {
            cout << "    ------- Last case of ix(" << _nX << "/" << _nX << ") im(" << im << "/" << _nM << ") dprec[" << dprecC << "]" << endl;
            vecOf01.Print("Last tour : search the one at zero");
        }
        nxcrt = -1;
        for(Int_t ii=0; ii<_nX; ii++) {
            if ( vecOf01(ii) == 0 )
                nxcrt=ii;
        }

        sxcrt = _listOfInputAttributes->At(nxcrt)->GetName();
        vecShapley(nxcrt) += dvarall - dprecC;
        if ( _blog )
            vecShapley.Print(Form("Shapley in last case ix nxcrt[%d] Name[%s] dvarall[%f] dprec[%f]", nxcrt, sxcrt.Data(), dvarall, dprecC));


#ifndef WIN32
      if (_bdrawProgressBar) {
          timer.DrawProgressBar(im);
        }
#endif

    } // End Of loop on the indexes im/_nM (permutations)
    if (_bdrawProgressBar)
      cout << endl << " ** time elapsed " << timer.GetElapsedTime() << endl;

    vecShapley *=d1OverM;
    vecShapley.Print(Form("Shapley Values for MC[%d]", _nM));
    cout<< "    *** Sum of vecShapley[" << vecShapley.Sum() << "]" << endl;


    /// Fill the values
    ///
    for(Int_t ii=0; ii<_nX; ii++) {
      fillIndex(_listOfInputAttributes->At(ii)->GetName(), getMethodName(), vecShapley(ii), "--first--");
    }
    cout << " end of method[TShpaley::" << __func__<< "]" << endl;
}

void TShapley::transfertDataTDS(TDataServer *tdsCrt) {
    if ( _blog ) cout << " **** function[TShapley::" << __func__ << "] tdsCrt[" << tdsCrt->GetName() << "]" << endl;

    Long64_t npatall = _tds->getNPatterns();
    Int_t nattall = _tds->getNAttributes();
    if ( _blog ) cout << " *** npatall[" << npatall << "] nattall[" << nattall << "]" << endl;

    Long64_t npatcrt = tdsCrt->getNPatterns();
    if ( _blog ) cout << " *** npatcrt[" << npatcrt << "] ___natt___[" << tdsCrt->getNAttributes() << "]" << endl;

    try {
        if ( !tdsCrt->isAttribute(_sIteratorName) )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TShapley::transfertDataTDS ERROR : the TDataServer \"%s\" do not contains the Nv iterator attribute \"%s\".",
                                                            tdsCrt->GetName(), _sIteratorName.Data()));
        if ( !tdsCrt->isAttribute(_sIteratorNameNo) )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TShapley::transfertDataTDS ERROR : the TDataServer \"%s\" do not contains the No iterator attribute \"%s\".",
                                                            tdsCrt->GetName(), _sIteratorNameNo.Data()));
        if ( !tdsCrt->isAttribute(_sIteratorNameNi) )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TShapley::transfertDataTDS ERROR : the TDataServer \"%s\" do not contains the Ni iterator attribute \"%s\".",
                                                            tdsCrt->GetName(), _sIteratorNameNi.Data()));

        TList *listOfLeaves = new TList();
        const Int_t nattstore = nattall+1;
        Double_t *thevalues = new Double_t[nattstore];
        memset(thevalues, 0, sizeof(Double_t) * (nattstore));

        for (Int_t i = 0; i < nattall; i++)
        {
            if ( _blog ) cout << " *** att(" << i << "/" << nattall <<") -- ";
            TAttribute *att_tds = _tds->getAttribute(i);
            if ( _blog ) cout << " Name [" << att_tds->GetName() << "]" << endl;
            if ( tdsCrt->isAttribute(att_tds->GetName()) ) {
                listOfLeaves->Add(tdsCrt->getTuple()->GetLeaf(att_tds->GetName()));
            } else {
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TShapley::transfertDataTDS ERROR : the current TDataServer \"%s\" do not contains the attribute \"%s\".",
                                                                tdsCrt->GetName(), att_tds->GetName()));
            }
        }
        if ( _blog ) listOfLeaves->Print("all");

        for(Long64_t ipat=0; ipat<npatcrt; ipat++) {
            tdsCrt->getTuple()->GetEntry(ipat);
            // the index attribute
            thevalues[0] = npatall+ipat;
            // the original attributes
            for(Int_t iatt=0; iatt<nattall; iatt++) {
                thevalues[1+iatt] = ((TLeaf*) listOfLeaves->At(iatt))->GetValue();
            }
            if ( _blog ) {
                cout << " *********************************************** ipat(" << ipat << "/" << npatcrt << ") - iter[" << thevalues[0] << "] -- ";
                for(Int_t iatt=1; iatt<nattstore; iatt++) {
                    cout<< "(" << thevalues[iatt] << ") ";
                }
                cout << endl;
            }
            _tds->getTuple()->Fill(thevalues);
        }

        // Clean
        delete listOfLeaves; listOfLeaves=NULL;
        delete[] thevalues; thevalues=NULL;


    } catch (URANIE::Exceptions::UExceptions& ue)
    {
        ue.printMessage();
        tdsCrt->printLog();
        if (ue.getMessageType() > URANIE::Exceptions::UExceptions::kWARNING)
            throw ue;
    }




//    // Erase the values
//    tdsCrt->deleteTuple();
    if ( _blog ) cout << " **** End of function[TShapley::" << __func__ << "] tdsCrt[" << tdsCrt->GetName() << "]" << endl;
}

void TShapley::splitMatrixVarCoVar(TMatrixD matOrigin, TVectorD vecOfUnit,
                                   TMatrixD &matVarCovar0, TMatrixD &matVarCovar1, TMatrixD &matVarCovar01,
                                   TMatrixD matCorr, TMatrixD &matCorrTmp) {


    if ( _blog ) {
        cout << " **** function[TShapley::" << __func__ << "]" << endl;
        vecOfUnit.Print("the vector of 01");
    }

    cout << " ********  matOrigin *****";
    matOrigin.Print("");
    matCorr.Print("");
    cout << " *************************";

    Int_t nP = matOrigin.GetNrows();
    Int_t nP0 = matVarCovar0.GetNrows();
    Int_t nP1 = matVarCovar1.GetNrows();
    Int_t nR01 = matVarCovar01.GetNrows();
    Int_t nC01 = matVarCovar01.GetNcols();
    Int_t nSum1 = vecOfUnit.Sum();

    if ( _blog ) {
        cout << " *** nP[" << nP << "] --- nP0[" << nP0 << "] ---";
        cout << " nP1[" << nP1 << "] --- matVarCovar01[" << nR01 << "," << nC01 << "] --- nSum1[" << nSum1 << "]" << endl;
    }

    Int_t nL1 = vecOfUnit.NonZeros ();
    Int_t nL0 = vecOfUnit.GetNrows()-nL1;
    if ( _blog )
        cout << " *** nP[" << nP << "] --- nL0[" << nL0 << "] --- nL1[" << nL1 << "] --- nSum1[" << nSum1 << "]" << endl;



    matVarCovar0.ResizeTo(nL0,nL0);
    matVarCovar0.Zero();
    matCorrTmp.ResizeTo(nL0,nL0);
    matCorrTmp.Zero();

    matVarCovar1.ResizeTo(nL1, nL1);
    matVarCovar1.Zero();

    matVarCovar01.ResizeTo(nL1, nL0);
    matVarCovar01.Zero();


    Int_t icrt11 = 0;
    Int_t jcrt11 = 0;

    Int_t icrt00 = 0;
    Int_t jcrt00 = 0;

    Int_t icrt10 = 0;
    Int_t jcrt10 = 0;

    for(Int_t irow=0; irow<nP; irow++) {
        // Case of One
        if ( vecOfUnit[irow] > 0.5 ) {
            jcrt11=0;
            jcrt10=0;
            for(Int_t jcol=0; jcol<nP; jcol++) {
                if ( vecOfUnit[jcol] > 0.5  ) {
                    if ( _blog )
                        cout << "    +++++ irow[" << irow << "] jcol[" << jcol << "] - icrt11[" << icrt11 << "] jcrt11[" << jcrt11 << "] matOrigin[" << matOrigin(irow,jcol) << "]" << endl;
                    matVarCovar1(icrt11,jcrt11) = matOrigin(irow,jcol);
                    jcrt11++;
                } else {
                    if ( _blog )
                        cout << "    ///// irow[" << irow << "] jcol[" << jcol << "] - icrt10[" << icrt10 << "] jcrt10[" << jcrt10 << "] matOrigin[" << matOrigin[irow][jcol] << "]" << endl;
                    matVarCovar01(icrt10, jcrt10) = matOrigin(irow,jcol);
                    jcrt10++;
                }
            }
            icrt11++;
            icrt10++;
        } else {
            // Case of Zero
            jcrt00=0;
            for(Int_t jcol=0; jcol<nP; jcol++) {
                if ( vecOfUnit[jcol] < 0.5 ) {
                    if ( _blog )
                        cout << "    ----- irow[" << irow << "] jcol[" << jcol << "] matOrigin[" << matOrigin(irow,jcol) << "]" << endl;
                    matVarCovar0(icrt00,jcrt00) = matOrigin(irow,jcol);
                    matCorrTmp(icrt00,jcrt00) = matCorr(irow,jcol);
                    jcrt00++;
                }
            }
            icrt00++;
        }
    }
    if ( _blog ) cout << " ***** End Of function[TShapley::" << __func__ << "]" << endl;
}

void TShapley::computeNormalCorrelationMatrix(TMatrixD matVarCovar) {
    if ( _blog ) cout << " ***** function[TShapley::" << __func__ << "]" << endl;

    Int_t nrows = matVarCovar.GetNrows();
    Int_t ncols = matVarCovar.GetNcols();
    if ( _blog ) cout << " *** _nX[" << _nX << "] nrows[" << nrows << "] --- ncols[" << ncols << "]" << endl;
    try
    {
        if ( nrows != ncols )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TShapley::%s ERROR : the numbers of rows[%d] and columns[%d] are not equal.",
                                                            __func__, nrows, ncols));

        _matNormalCovariance.ResizeTo(nrows, ncols);
        _matNormalCovariance.UnitMatrix();
        Double_t dPiOver6 = TMath::Pi()/6.0;

        for(Int_t irow=0; irow<nrows; irow++) {
            for(Int_t jcol=0; jcol<=irow; jcol++) {
                _matNormalCovariance(irow, jcol) = _matNormalCovariance(jcol, irow) = 2.0* TMath::Sin( dPiOver6 * matVarCovar(irow, jcol));
                _matNormalCovariance(irow, jcol) = _matNormalCovariance(jcol, irow) = matVarCovar(irow, jcol);
            }
        }
        if ( _blog ) {
            cout<< " *** Rank Variance-Covariance Matrix";
            matVarCovar.Print();

            cout<< " *** Correlation Matrice for the Equivalent Normal Distribution";
            _matNormalCovariance.Print();
        }
    } catch (URANIE::Exceptions::UExceptions& ue)
    {
        ue.printMessage();
        if (ue.getMessageType() > URANIE::Exceptions::UExceptions::kWARNING)
            throw ue;
    }
    if ( _blog ) cout << " ***** End Of function[TShapley::" << __func__ << "]" << endl;
}


void TShapley::generateConditionnalSampleByNorta(TGaussianSampling *tsanormal, TDataServer *theTDSres, TList *listOfLeaves, TMatrixD matVarCovar, TVectorD &vecOf01)
{
    if ( _blog ) {
        cout << " **** function[TShapley::" << __func__ << "] _nO[" << _nO << "]" << endl;
        matVarCovar.Print();
        vecOf01.Print("the vector of 01");
    }

    // Empty the TDS of Results
    theTDSres->deleteTuple();
    // Create the Tuple of TDS for results
    theTDSres->createTuple();
    theTDSres->getTuple()->SetScanField(0);

    TDataServer *resTDS = tsanormal->getTDS();
    Int_t ndefault = resTDS->getNPatterns();

    resTDS->getTuple()->SetScanField(0);
    if ( _blog ) {
        cout<< " **** ndefault[" << ndefault << "]" << endl;
        resTDS->Scan("*");
    }

    TString sname = "";
    // loop on nX
    Int_t nxcard=0;
    Int_t nxcrt=-1;
    Int_t nmcrt =-1;
    for(Int_t ix=0; ix<_nX; ix++) {
        if ( vecOf01(ix) > 5. ) {
            nxcrt = ix+1;
            nmcrt = vecOf01(ix)-9;
            vecOf01(ix) = 1.0;
        }
        if ( vecOf01(ix) > 0.5 )
            nxcard++;

        sname = _listOfInputAttributes->At(ix)->GetName();
        if ( _blog ) cout << " --- case of ix(" << ix << "/" << _nX << ") att : sname[" << sname << "]" << endl;
    } // End of ix/_nX


    if ( _blog ) tsanormal->getCovarianceMatrix().Print();
    tsanormal->generateConditionalSample(vecOf01, _nO, _nI);

    if ( _blog ) {
        resTDS->Scan("*");
    }

    const Int_t nattstore = _nX+4;
    if ( _blog ) cout<< " **** nattstore[" << nattstore << "]" << endl;
    Double_t *thevalues = new Double_t[nattstore];
    memset(thevalues, 0, sizeof(Double_t) * (nattstore));


    TStochasticAttribute *stoatt;
    Double_t dvalue, dvalueCDF;

    for (Int_t ipat = ndefault; ipat < resTDS->getNPatterns(); ipat++) {
        resTDS->getTuple()->GetEntry(ipat);

        thevalues[0] = 1.0*(ipat-ndefault);
        for(Int_t iatt=0; iatt<_nX; iatt++) {
            dvalue = ((TLeaf*) listOfLeaves->At(iatt))->GetValue();
            stoatt = dynamic_cast<TStochasticAttribute *>(resTDS->getAttribute(iatt));
            dvalueCDF = stoatt->getCDF(dvalue);
            sname = _listOfInputAttributes->At(iatt)->GetName();
            stoatt = dynamic_cast<TStochasticAttribute *>(_tds->getAttribute(sname));
            thevalues[iatt+1] = stoatt->getInvCDF(dvalueCDF);
        }
        thevalues[_nX+1] = 1.0*nmcrt;
        thevalues[_nX+2] = 1000.*nxcard+nxcrt;
        dvalue = ((TLeaf*) listOfLeaves->At(_nX))->GetValue();
        thevalues[_nX+3] = dvalue;
        theTDSres->getTuple()->Fill(thevalues);
    }

    if ( _blog ) theTDSres->scan("*");


    // Clean
    delete[] thevalues; thevalues = NULL;

    if ( _blog ) cout << " **** End Of function[TShapley::" << __func__ << "]" << endl;
}

TDataServer *TShapley::generateConditionnalSampleByNortaOriginal(TDataServer *tds0, TDataServer *tds1, TMatrixD matVarCovar0, TMatrixD matVarCovar1, TMatrixD matVarCovar01)
{
    if ( _blog ) cout << " **** function[TShapley::" << __func__ << "] _nO[" << _nO << "]" << endl;

    Int_t natt = tds0->getNAttributes();
    Int_t natt1 = tds1->getNAttributes();

    Int_t nR0 = matVarCovar0.GetNrows();
    Int_t nC0 = matVarCovar0.GetNcols();
    if ( _blog ) cout << "  --- matVarCovar0[" << nR0 << "," << nC0 << "]" << endl;
    if ( _blog ) matVarCovar0.Print("");

    TMatrixD matXInv(TMatrixD::kInverted, matVarCovar0);
    if ( _blog ) cout << " --- matXInv := matVarCovar0 inverse" << endl;
    if ( _blog ) matXInv.Print("");

    Int_t nR1 = matVarCovar1.GetNrows();
    Int_t nC1 = matVarCovar1.GetNcols();
    if ( _blog ) cout << "  --- matVarCovar1[" << nR1 << "," << nC1 << "]" << endl;
    if ( _blog ) matVarCovar1.Print("");

    Int_t nR01 = matVarCovar01.GetNrows();
    Int_t nC01 = matVarCovar01.GetNcols();
    if ( _blog ) cout << "  --- matVarCovar01[" << nR01 << "," << nC01 << "]" << endl;
    if ( _blog ) matVarCovar01.Print("");


    TMatrixD matSigma0XInv(matVarCovar01,TMatrixD::kMult,matXInv);
    if ( _blog ) cout << "   ** Matrice matSigma0XInv : Sigma0 * SigmaCInv " << endl;
    if ( _blog ) matSigma0XInv.Print("");

    TMatrixD matSigma0XInvSigma0(matSigma0XInv,TMatrixD::kMultTranspose,matVarCovar01);
    if ( _blog ) cout << "   ** Matrice matSigma0XInv : Sigma0 * SigmaCInv * Sigma0.T" << endl;
    if ( _blog ) matSigma0XInvSigma0.Print("");

    TMatrixD matVarYX(matVarCovar1,TMatrixD::kMinus,matSigma0XInvSigma0);
    if ( _blog ) cout << "   ** Matrice : VarCovariance of Y/X " << endl;
    if ( _blog ) matVarYX.Print("");

    TMatrixD matValues(natt, 1);
    matValues.Zero();
    Double_t dvalue, dvalueCDF;

    // Create the return TDS
    TDataServer *resTDS = new TDataServer("tds_res", "the tds of results");

    TList *listOfLeaves = new TList();
    for (Int_t iatt=0; iatt<natt; iatt++) {
        TAttribute *att_tds = tds0->getAttribute(iatt);
        resTDS->addAttribute(att_tds);
        listOfLeaves->Add(tds0->getTuple()->GetLeaf(att_tds->GetName()));
    }
    if ( _blog ) listOfLeaves->Print("all");

    TDataServer *tdsNi = new TDataServer("_tdsni_", "The TDS for nI");
    for(Int_t iatt=0; iatt<natt1; iatt++) {
        resTDS->addAttribute(tds1->getAttribute(iatt));
        cout << " **** iatt[" << iatt << "/" << natt1 << "] Name[" << Form(STRNAMEOFATTNI, iatt+1) << "] sigma cste(1.0)" << endl;
        tdsNi->addAttribute( new TNormalDistribution(Form(STRNAMEOFATTNI, iatt+1), 0.0, TMath::Sqrt(matVarYX(iatt,iatt))));
        //tdsNi->addAttribute( new TNormalDistribution(Form(STRNAMEOFATTNI, iatt+1), 0.0, 1.0));
        tdsNi->getAttribute(Form(STRNAMEOFATTNI, iatt+1))->delShare();
    }
    if ( _blog ) cout<< "  ** tdsNi  :: N attributes[" << tdsNi->getNAttributes() << "] matVarYX[" << matVarYX.GetNrows() << ", " << matVarYX.GetNcols() << "]" << endl;
    resTDS->addAttribute( new TAttribute(_sIteratorNameNi));
    resTDS->getAttribute(_sIteratorNameNi)->delShare();
    if ( _blog ) cout<< "  ** resTDS :: N attributes[" << resTDS->getNAttributes() << "]" << endl;
    resTDS->createTuple();
    resTDS->getTuple()->SetScanField(0);


    const Int_t nattstore = resTDS->getNAttributes()+1;
    Double_t *thevalues = new Double_t[nattstore];
    memset(thevalues, 0, sizeof(Double_t) * (nattstore));

    // Define the sampler for nI
    Double_t dPiOver6 = TMath::Pi()/6.0;
    Int_t nIcrt = TMath::Max(_nI, 2*natt1+1);
    TGaussianSampling *tgas = new TGaussianSampling(tdsNi, "lhs", nIcrt);
    if ( _nSeed != 0 )
        tgas->setSeed(_nSeed);
    for(Int_t iatt=0; iatt<natt1; iatt++) {
        Double_t dsigmaiatt = TMath::Sqrt(matVarYX(iatt,iatt));
        cout << "     **** iatt[" << iatt << "/" << natt1 << "] matVarYX(" << matVarYX(iatt,iatt) << ") dsigmaiatt(" << dsigmaiatt << ")" << endl;
        for(Int_t j=0; j<iatt; j++) {
            //tgas->setUserCorrelation(iatt, j, matVarYX(iatt, j));
            Double_t dsigmaj = TMath::Sqrt(matVarYX(j,j));
            cout << "        **** j[" << j << "/" << iatt << "] matVarYX(" << matVarYX(j,j) << ") dsigmaiatt(" << dsigmaiatt << ")" << endl;
            //tgas->setUserCorrelation(iatt, j, TMath::Sqrt( 2.0* TMath::Sin( dPiOver6 * matVarYX(iatt, j) ) ) / ( dsigmaiatt * dsigmaj) );
            tgas->setUserCorrelation(iatt, j, matVarYX(iatt, j) / ( dsigmaiatt * dsigmaj) );
        }
    }
    cout << " *** the Correlation matrix for the TGaussianSampling" << endl;
    tgas->getCorrelationMatrix().Print();
    cout << " *** the VarianceCovariance matrix for the TGaussianSampling" << endl;
    tgas->getCovarianceMatrix().Print();
    cout << " *** End Of the Correlation matrix for the TGaussianSampling" << endl;

    TString sSelect ="";
    TStochasticAttribute *stoatt;
    TNormalDistribution uNormal(Form("guranie_unormal_shapley_%s_", GetName()), 0.0, 1.0);

    if ( _blog ) {
        cout << " *** loop nO : to sample the nI DoE od size _nO[" << _nO << "]" <<endl;
        tds0->scan("*", Form("%s<%d", tds0->getIteratorName(), _nO+1));
    }

    Long64_t ipatcrt=1;

    for(Long64_t ipatno=0; ipatno<_nO; ipatno++) {
        if ( _blog ) cout << endl << " *** ipat nO(" << ipatno+1 << "/" << _nO << ")" << endl;
        tds0->getTuple()->GetEntry(ipatno);

        for(Int_t iatt=0; iatt<natt; iatt++) {
            dvalue = ((TLeaf*) listOfLeaves->At(iatt))->GetValue();
            thevalues[iatt+1] = dvalue;
            stoatt = dynamic_cast<TStochasticAttribute *>(tds0->getAttribute(iatt));
            dvalueCDF = stoatt->getCDF(dvalue);
            dvalue = uNormal.getInvCDF(dvalueCDF);
            matValues(iatt, 0) = dvalue;
        }
        if ( _blog ) matValues.Print();

        TMatrixD matMean(matSigma0XInv,TMatrixD::kMult,matValues);
        if ( _blog ) {
            cout << " *** Vector of Mean" << endl;
            matMean.Print();
        }

        // Update the mean values
        TNormalDistribution *stonoratt;
        for(Int_t iatt=0; iatt<0; iatt++) { // natt1
            stonoratt = dynamic_cast<TNormalDistribution *>(tdsNi->getAttribute(Form(STRNAMEOFATTNI, iatt+1)));
            stonoratt->setParameterMu(matMean(iatt, 0));
        }
            // Generate the sample Ni
        tgas->generateSample();
        tdsNi->getTuple()->SetScanField(0);
        tdsNi->scan("*");

        TList *listOfLeavesNi = new TList();
        for(Int_t iatt=0; iatt<natt1; iatt++) {
            listOfLeavesNi->Add(tdsNi->getTuple()->GetLeaf(Form(STRNAMEOFATTNI, iatt+1)));
        }
        if ( _blog ) listOfLeavesNi->Print("all");

        for(Long64_t ipatni=0; ipatni<_nI; ipatni++) {
            if ( _blog ) cout << " *** ipatni(" << ipatni << "/" << _nI << ")" << endl;
            tdsNi->getTuple()->GetEntry(ipatni);
            for(Int_t iatt=0; iatt<natt1; iatt++) {
                cout << "    *** iatt(" << iatt+1<<"/" << natt1 <<") :: ";
                dvalue = ((TLeaf*) listOfLeavesNi->At(iatt))->GetValue();
                cout << " dvalue[" << dvalue << "], ";
                dvalueCDF = uNormal.getCDF(dvalue);
                cout << " dvalueCDF[" << dvalueCDF << "]";
                stoatt = dynamic_cast<TStochasticAttribute *>(tds1->getAttribute(iatt));
                thevalues[natt+iatt+1] = stoatt->getInvCDF(dvalueCDF);
                cout << " thevalues[" << thevalues[natt+iatt+1] << "]" << endl;
            }
            thevalues[0]=ipatcrt;
            thevalues[nattstore-1]=ipatno+1;     // ipatno ipatni
            resTDS->getTuple()->Fill(thevalues);
            ipatcrt++;
        }
        // Clean
        delete listOfLeavesNi; listOfLeavesNi = NULL;
        // Erase the values
        tdsNi->deleteTuple();
    } // end of Loop ipatno/_nO
    if ( 1 ) cout << " *** End of Loop ipatno(" << _nO << ")" << endl;


    // fill other attributes
    for(Int_t iatt=0; iatt<_tds->getNAttributes(); iatt++) {
        if ( _blog ) cout << " *** att(" << iatt+1 << "/" << _tds->getNAttributes() <<") -- ";
        TAttribute *att_tds = _tds->getAttribute(iatt);
        TString sname = att_tds->GetName();
        if ( _blog ) cout << sname;
        if ( ( !sname.BeginsWith(Form("%s__N", getMethodName())) ) &&
             ( att_tds->InheritsFrom(TAttributeFormula::Class()) ) ) {
            if ( _blog ) cout << " XXX";
            resTDS->addAttribute(att_tds);
            resTDS->fillAttribute(sname,"",kFALSE);
        }
        if ( _blog ) cout << endl;
    }

    if ( _blog ) {
        cout << " *** TDS of results to merge with the _tds"<<endl;
        resTDS->scan("*");
    }

    // Clean
    delete listOfLeaves; listOfLeaves = NULL;
    delete tgas; tgas = NULL;
    delete tdsNi; tdsNi = NULL;
    //delete  resTDS; resTDS = NULL;
    delete[] thevalues; thevalues = NULL;

    if ( _blog ) cout << " **** End Of function[TShapley::" << __func__ << "]" << endl;
    return resTDS;
}

void TShapley::printLog(Option_t *option)
{
    cout << " ** TShapley::printLog[" << option << "]" << endl;
    cout << " **   _sIteratorName[" << _sIteratorName << "] _nV[" << _nV << "] _nM[" << _nM << "]" << endl;
    cout << " ** _sIteratorNameNo[" << _sIteratorNameNo << "] _nO[" << _nO << "]" << endl;
    cout << " ** _sIteratorNameNi[" << _sIteratorNameNi << "] _nI[" << _nI << "]" << endl;
    cout << " **    _nCondSampler[" << _nCondSampler << "]" << endl;
    cout << " **           _nSeed[" << _nSeed << "]" << endl;
    cout << " ** End of TShapley::printLog[" << option << "]" << endl;
}

void TShapley::printSummary(Option_t *option)
{
    cout << " *** TShapley::printSummary option[" << option << "]" << endl;
    cout << " *** Number of : Patterns[" << _nS << "] Inputs[" << _nX << "] Outputs[" << _nY << "]" << endl;
    cout << " *** End Of TShapley::printSummary[" << option << "]" << endl;
}

} // End Of namespace Sensitivity
} // End Of namespace URANIE
// End Of File
