//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTFast.cxx
 \author maud Salvoldi (OCC/Areva group)
 \date   $Date$
 
 \brief  Fichier de test de la classe URANIE::Sensitivity::TFast
 
 Cree le : jeu dec 22 12:30:15 CET 2011

 
 */

#define EPSILON 0.01
#define LVLE_EPSILON 0.001 
#define DBL_PRECISION_COMPARE 1.e-12

// Uranie
#include "TestTSensitivity.h"
#include "TestTFast.h"
#include "TFast.h"
#include "TUniformDistribution.h"
#include "TLogUniformDistribution.h"
#include "TOutputFileKey.h"
#include "TOutputFileRow.h"
#include "TOutputFileDataServer.h"
#include "TCode.h"
#include "TLauncher.h"
#include "TLauncherFunction.h"
#include "TCJitEval.h"
#include "TSequentialRun.h"
#include "TThreadedRun.h"
#include "UDisableTest.h"
#include "TestUserFunctions.h"
#include "code_paths.h"
#include "uranie_commands.h"

// ROOT
#include "TString.h"
#include "TMath.h"
#include "TTreePlayer.h"
#include "TROOT.h"
#ifdef ROOT5
#include "Api.h"
#endif

#include <iostream>
#include <iomanip>
using namespace std;

TFastTest::TFastTest() :
        _tds(NULL), _tds2(NULL), _tds3(NULL), _tfast(NULL), ns(0), _mycode(NULL),
	    _p2fIshigami(NULL), _p2fFlowratemodel(NULL), _fout(NULL)
{
#ifdef ROOT5
//   gROOT->LoadMacro("UserFunctions.C");
    G__ClassInfo globalscope;
    G__MethodInfo method;
    long dummy;
    method = globalscope.GetMethod("Ishigami", "double*,double*", &dummy);
    _p2fIshigami = method.PointerToFunc();

    G__ClassInfo globalscope2;
    G__MethodInfo method2;
    long dummy2;
    method2 = globalscope2.GetMethod("flowrateModel", "double*,double*", &dummy2);
    _p2fFlowratemodel = method2.PointerToFunc();
#endif
}

TFastTest::~TFastTest()
{
    UranieTests::removeFile(gSystem->pwd(),"testFastIshigami.out");
    UranieTests::removeFile(gSystem->pwd(),"_fast_launching_.dat");
    UranieTests::removeFile(gSystem->pwd(),"outpuT.dat");
    UranieTests::removeFile(gSystem->pwd(),"outpuS.dat");
    UranieTests::removeFile(gSystem->pwd(),"testFast.out");
}

void TFastTest::setUp()
{
    _tds = new URANIE::DataServer::TDataServer("tdsflowreate",
            "DataBase flowreate");
    _tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("rw", 0.05, 0.15));
    _tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("r", 100.0, 50000.0));
    _tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("tu", 63070.0,
                    115600.0));
    _tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("tl", 63.1, 116.0));
    _tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("hu", 990.0, 1110.0));
    _tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("hl", 700.0, 820.0));
    _tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("l", 1120.0, 1680.0));
    _tds->addAttribute(
            new URANIE::DataServer::TUniformDistribution("kw", 9855.0,
                    12045.0));

    _tds2 = new URANIE::DataServer::TDataServer("tds",
            "TDS for Ishigami function");
    _tds2->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x_{1}",
                    -1.0 * TMath::Pi(), TMath::Pi()));
    _tds2->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x_{2}",
                    -1.0 * TMath::Pi(), TMath::Pi()));
    _tds2->addAttribute(
            new URANIE::DataServer::TUniformDistribution("x_{3}",
                    -1.0 * TMath::Pi(), TMath::Pi()));

    _tds3 = NULL;
    _fout = NULL;
    
    ns = 1;
}

void TFastTest::tearDown()
{
  if (_fout)
    {
      delete _fout;
      _fout = NULL;
    }

  if (_mycode != NULL)
    {
      delete _mycode;
      _mycode = NULL;
    }

  delete _tfast;
  _tfast = NULL;
    
  _tds2->getAttribute("x_{1}")->delShare();
  _tds2->getAttribute("x_{2}")->delShare();
  _tds2->getAttribute("x_{3}")->delShare();
  delete _tds2;
  _tds2 = NULL;
    
  _tds->getAttribute("rw")->delShare();
  _tds->getAttribute("r")->delShare();
  _tds->getAttribute("tu")->delShare();
  _tds->getAttribute("tl")->delShare();
  _tds->getAttribute("hu")->delShare();
  _tds->getAttribute("hl")->delShare();
  _tds->getAttribute("l")->delShare();
  _tds->getAttribute("kw")->delShare();
  delete _tds;
  _tds = NULL;

  if (_tds3)
    {
      delete _tds3;
      _tds3 = NULL;
      
    }

}

////////////////////////////////
///                          ///                            
///         Tests            ///
///                          ///
////////////////////////////////
void TFastTest::setUp1()
{
    ns = 500;
#ifdef ROOT5
    _tfast = new URANIE::Sensitivity::TFast(_tds2, _p2fIshigami, ns);
#else
    _tfast = new URANIE::Sensitivity::TFast(_tds2, &Ishigami, "", "Ishigami", ns);
#endif
    _tfast->setDrawProgressBar(kFALSE);
    _tfast->setSeed(123);
}

void TFastTest::setUp10()
{
    ns = 500;
#ifndef ROOT5
    gROOT->LoadMacro("testMorrisIshigami.C");
#endif
    _tfast = new URANIE::Sensitivity::TFast(_tds2, "Ishigami" , ns);
    _tfast->unsetLog();
    _tfast->setDrawProgressBar(kFALSE);
    _tfast->setSeed(123);
}
void TFastTest::setUp2()
{
    TString sJDDReference = gSystem->pwd();
    sJDDReference += TString("/flowrate_input_with_values_rows.in");
    _tds->getAttribute("rw")->setFileKey(sJDDReference, "Rw", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("r")->setFileKey(sJDDReference, "R", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("tu")->setFileKey(sJDDReference, "Tu", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("tl")->setFileKey(sJDDReference, "Tl", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("hu")->setFileKey(sJDDReference, "Hu", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("hl")->setFileKey(sJDDReference, "Hl", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("l")->setFileKey(sJDDReference, "L", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);
    _tds->getAttribute("kw")->setFileKey(sJDDReference, "Kw", "",
            URANIE::DataServer::TAttributeFileKey::kNewRow);

    // The output file of the code
    _fout = new URANIE::Launcher::TOutputFileRow(
            "_output_flowrate_withRow_.dat");
    // The attribute in the output file
    _fout->addAttribute("yhat");

    // Instanciation de mon code
    _mycode = new URANIE::Launcher::TCode(_tds, UranieTests::code_path("flowrate") + " -s -r");
    _mycode->addOutputFile(_fout);
    ns = 1;
    _tfast = new URANIE::Sensitivity::TFast(_tds, _mycode, ns);
    _tfast->setDrawProgressBar(kFALSE);
    _tfast->setSeed(123);
}

void TFastTest::testIshigamiFAST()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
  cout<<endl<<"To be suppressed ?"<<endl;
    URANIE_IGNORED_TEST() ;

    gROOT->Macro("testFastIshigami.C");

    Double_t dvalx1, dvalx2, dvalx3;
    ifstream outFile("testFastIshigami.out", ios::in);

    outFile >> dvalx1 >> dvalx2 >> dvalx3;

    //Close the file
    outFile.close();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx1, 0.3116684561957,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx2, 0.5038611721187,
            DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dvalx3, 0.000292700925272,
            DBL_PRECISION_COMPARE);

}

#ifdef WITH_LEVELE
// ====================================
//            LevelE
// ====================================
void TFastTest::setUpLevelE()
{
  _tds3 = new URANIE::DataServer::TDataServer("tdsLevelE","levele");
  _tds3->addAttribute( new URANIE::DataServer::TUniformDistribution("t", 100, 1000) );
  _tds3->addAttribute( new URANIE::DataServer::TLogUniformDistribution("kl", 0.001, 0.01) );
  _tds3->addAttribute( new URANIE::DataServer::TLogUniformDistribution("kc", 0.000001, 0.00001) );
  _tds3->addAttribute( new URANIE::DataServer::TLogUniformDistribution("v1", 0.001, 0.1) );
  _tds3->addAttribute( new URANIE::DataServer::TUniformDistribution("l1", 100, 500) );
  _tds3->addAttribute( new URANIE::DataServer::TUniformDistribution("r1", 1, 5) );
  _tds3->addAttribute( new URANIE::DataServer::TUniformDistribution("rc1", 3, 30) );
  _tds3->addAttribute( new URANIE::DataServer::TLogUniformDistribution("v2", 0.01, 0.1) );
  _tds3->addAttribute( new URANIE::DataServer::TUniformDistribution("l2", 50, 200) );
  _tds3->addAttribute( new URANIE::DataServer::TUniformDistribution("r2", 1, 5) );
  _tds3->addAttribute( new URANIE::DataServer::TUniformDistribution("rc2", 3, 30) );
  _tds3->addAttribute( new URANIE::DataServer::TLogUniformDistribution("w", 100000, 10000000) );

  TString name="levelE_input_with_values_rows.in";
  _tds3->getAttribute("t")->setFileKey(name, "t", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds3->getAttribute("kl")->setFileKey(name, "kl", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds3->getAttribute("kc")->setFileKey(name, "kc", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds3->getAttribute("v1")->setFileKey(name, "v1", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds3->getAttribute("l1")->setFileKey(name, "l1", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds3->getAttribute("r1")->setFileKey(name, "r1", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds3->getAttribute("rc1")->setFileKey(name, "rc1", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds3->getAttribute("v2")->setFileKey(name, "v2", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds3->getAttribute("l2")->setFileKey(name, "l2", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds3->getAttribute("r2")->setFileKey(name, "r2", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds3->getAttribute("rc2")->setFileKey(name, "rc2", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);
  _tds3->getAttribute("w")->setFileKey(name, "w", "%e", URANIE::DataServer::TAttributeFileKey::kNewRow);

  _tds3->getAttribute( "t" )->delShare();
  _tds3->getAttribute( "kl" )->delShare();
  _tds3->getAttribute( "kc" )->delShare();
  _tds3->getAttribute( "v1" )->delShare();
  _tds3->getAttribute( "l1" )->delShare();
  _tds3->getAttribute( "r1" )->delShare();
  _tds3->getAttribute( "rc1" )->delShare();
  _tds3->getAttribute( "v2" )->delShare();
  _tds3->getAttribute( "l2" )->delShare();
  _tds3->getAttribute( "r2" )->delShare();
  _tds3->getAttribute( "rc2" )->delShare();
  _tds3->getAttribute( "w" )->delShare();

  _fout = new URANIE::Launcher::TOutputFileRow("_output_levelE_withRow_.dat");
  _fout->addAttribute(new URANIE::DataServer::TAttribute("y", URANIE::DataServer::TAttribute::kVector), 2 );
  _fout->getAttribute(0)->delShare();

#ifdef WIN32
  _mycode = new URANIE::Launcher::TCode(_tds3, "levele.exe");
#else
  _mycode = new URANIE::Launcher::TCode(_tds3, Form("%s/%s",gSystem->pwd(),"../../../../launCHER/teST/coDE/levelE/levele 2> /dev/null "));
#endif
  _mycode->addOutputFile(_fout);  
  
}

void TFastTest::checkLevelEResults(TTree *ntresu, vector<int> *elmts)
{

  int npts = ntresu->Draw("Sobol"," Inp==\"w\" && Algo==\"--first--\" ","goff");
  double *first=ntresu->GetV1();

  double firref[26]={0.0686166588153,0.122312031785,0.157047471924,0.135607482909,0.132410281082,0.132938010498,0.130217761422,0.118467761073,0.116167932564,
		     0.0997544412901,0.0252083265881,0.0164390155121,0.0194102393815,0.023384735752,0.0271113922384,0.0304195826823,0.0330652459783,0.0349530501164,
		     0.0366432546962,0.0358258614813,0.0389126741728,0.0490222006676,0.0627350469347,0.0737428254048,0.0814249828588,0.0890376252055};

  for(int ival=0; ival<npts; ival++)
    CPPUNIT_ASSERT_DOUBLES_EQUAL(firref[ ((elmts==NULL) ? ival : elmts->at(ival)) ], first[ival], LVLE_EPSILON);
}

void TFastTest::testLevelEFast()
{

  setUpLevelE();

  ns = 5;
  _tfast = new URANIE::Sensitivity::TFast(_tds3, _mycode, ns);
  _tfast->setM(2);
  _tfast->setSeed(1234);
  _tfast->setDrawProgressBar(false);
  _tfast->computeIndexes("nointermed");

  TTree *ntresu = (TTree*)_tfast->getResultTuple();
  checkLevelEResults(ntresu);
  
}

void TFastTest::testLevelEFastOutputRestrOption()
{
  setUpLevelE();
  ns = 5;
  _tfast = new URANIE::Sensitivity::TFast(_tds3, _mycode, ns);
  _tfast->setM(2);
  _tfast->setSeed(1234);
  _tfast->setDrawProgressBar(false);
  _tfast->computeIndexes("nointermed,output=y[4]:y[12]:y[21]");
  
  TTree *ntresu = (TTree*)_tfast->getResultTuple();
  checkLevelEResults(ntresu, _tfast->getAttributeElements(string("y")));
  
}
#endif

void TFastTest::testTFastsetTransformFunction()
{

  ns = 500;
#ifdef ROOT5
  _tfast = new URANIE::Sensitivity::TFast(_tds, _p2fFlowratemodel, ns);
#else
  _tfast = new URANIE::Sensitivity::TFast(_tds, &flowrateModel, "", "FlowrateModel", ns);
#endif
  CPPUNIT_ASSERT_EQUAL(URANIE::Sensitivity::TFast::kSaltelliA, _tfast->getTransformFunction());
  _tfast->setTransformFunction(URANIE::Sensitivity::TFast::kCukier);
  CPPUNIT_ASSERT_EQUAL(URANIE::Sensitivity::TFast::kCukier, _tfast->getTransformFunction());
  _tfast->setTransformFunction(URANIE::Sensitivity::TFast::kSaltelliA);
  CPPUNIT_ASSERT_EQUAL(URANIE::Sensitivity::TFast::kSaltelliA, _tfast->getTransformFunction());
  _tfast->setTransformFunction(URANIE::Sensitivity::TFast::kSaltelliB);
  CPPUNIT_ASSERT_EQUAL(URANIE::Sensitivity::TFast::kSaltelliB, _tfast->getTransformFunction());

}
void TFastTest::testTFastsetM()
{
  ns = 500;
#ifdef ROOT5
  _tfast = new URANIE::Sensitivity::TFast(_tds, _p2fFlowratemodel, ns);
#else
  _tfast = new URANIE::Sensitivity::TFast(_tds, &flowrateModel, "", "FlowrateModel", ns);
#endif
  CPPUNIT_ASSERT_EQUAL(6, _tfast->getM());
  _tfast->setM(8);
  CPPUNIT_ASSERT_EQUAL(8, _tfast->getM());
}

void TFastTest::testTFastwithCode()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
  cout<<endl<<"To be suppressed ?"<<endl;
  URANIE_IGNORED_TEST()  ;

    TString sJDDReference = gSystem->pwd();
    sJDDReference += TString("/flowrate_input_with_keys.in");
    _tds->getAttribute("rw")->setFileKey(sJDDReference, "Rw");
    _tds->getAttribute("r")->setFileKey(sJDDReference, "R");
    _tds->getAttribute("tu")->setFileKey(sJDDReference, "Tu");
    _tds->getAttribute("tl")->setFileKey(sJDDReference, "Tl");
    _tds->getAttribute("hu")->setFileKey(sJDDReference, "Hu");
    _tds->getAttribute("hl")->setFileKey(sJDDReference, "Hl");
    _tds->getAttribute("l")->setFileKey(sJDDReference, "L");
    _tds->getAttribute("kw")->setFileKey(sJDDReference, "Kw");

    // The output file of the code
    _fout = new URANIE::Launcher::TOutputFileRow(
            "_output_flowrate_withRow_.dat");
    // The attribute in the output file
    _fout->addAttribute("yhat");

    // Instanciation de mon code  
    TString sRepertoireReference = gSystem->pwd();
#ifdef WIN32
    _mycode = new URANIE::Launcher::TCode(_tds, "flowrate.exe -k 1>NUL");
#else
    _mycode =
            new URANIE::Launcher::TCode(_tds,
                    Form(
                            "%s/../../../../launCHER/teST/coDE/flowREATE/flowrate -s -k>>/dev/null",
                            sRepertoireReference.Data()));
#endif
    _mycode->addOutputFile(_fout);
    _tfast = new URANIE::Sensitivity::TFast(_tds, _mycode, 1);
    _tfast->setDrawProgressBar(kFALSE);
    _tfast->computeIndexes("nointermed");

    Double_t sobol;
    TTree *tree = _tfast->getResultTuple();
    tree->SetBranchAddress("Sobol", &sobol);

    CPPUNIT_ASSERT_EQUAL(16, (Int_t)(tree->GetEntries()));
}

void TFastTest::throwExceptionConstructorTooLongOutputFunction() {
    vector<URANIE::DataServer::TUniformDistribution> inputs;
    TestTSensitivity::add_flowrate_attributes(inputs);

    URANIE::DataServer::TDataServer tds("tds", "");
    for (auto& input: inputs)
        tds.addAttribute(&input);

    const string outputs_1 = "voila_un_nom_beaucoup_trop_long_pour_un_seul_attribut:encore_un_p*tain_de_nom_beaucoup_trop_long_pour_un_seul_attribut";
    TestTSensitivity::constructor_assert_throw<URANIE::Sensitivity::TFast>(tds, &flowrateModel,
        outputs_1);

    const string outputs_2 = "voila_un_nom_beaucoup_trop_long_pour_un_attribut:encore_un_p*tain_de_nom_beaucoup_trop_long_pour_un_seul_attribut";
    TestTSensitivity::constructor_assert_throw<URANIE::Sensitivity::TFast>(tds, &flowrateModel,
        outputs_2);
}

void TFastTest::throwExceptionConstructorTooLongOutputCode() {
    vector<URANIE::DataServer::TUniformDistribution> inputs;
    TestTSensitivity::add_flowrate_attributes(inputs);

    URANIE::DataServer::TDataServer tds("tds", "");
    for (auto& input: inputs)
        tds.addAttribute(&input);

    TString input_file = "flowrate_input_with_keys.in";
    TestTSensitivity::set_file_key_to_upper_first(tds, input_file);

    URANIE::Launcher::TOutputFileKey output_file("_output_flowrate_withKey_.dat");
    output_file.addAttribute("voila_un_nom_beaucoup_trop_long_pour_un_seul_attribut");

    URANIE::Launcher::TCode code(&tds, "");
    code.addOutputFile(&output_file);
    const size_t n_simulations = 50;
    TestTSensitivity::constructor_assert_throw<URANIE::Sensitivity::TFast>(tds, code,
        n_simulations);
}

void TFastTest::throwExceptionConstructorTooLongOutputRun() {
    vector<URANIE::DataServer::TUniformDistribution> inputs;
    TestTSensitivity::add_flowrate_attributes(inputs);
    URANIE::DataServer::TAttribute output("voila_un_nom_beaucoup_trop_long_pour_un_seul_attribut_de_runner");

    URANIE::Relauncher::TCJitEval eval(&flowrateModel);
    for(auto& input: inputs)
        eval.addInput(&input);
    eval.addOutput(&output);

    URANIE::Relauncher::TSequentialRun run(&eval);
    run.startSlave();
    if (run.onMaster()) {
        URANIE::DataServer::TDataServer tds("tds", "");
        for (auto& input: inputs)
            tds.addAttribute(&input);

        const size_t n_simulations = 50;
        TestTSensitivity::constructor_assert_throw<URANIE::Sensitivity::TFast>(tds, run,
            n_simulations);
        run.stopSlave();
    }
}

void TFastTest::throwExceptionTFastConstructorTdsNull()
{
    // The output file of the code
    URANIE::Launcher::TOutputFileRow *fout =
            new URANIE::Launcher::TOutputFileRow(
                    "_output_flowrate_withRow_.dat");
    // The attribute in the output file
    fout->addAttribute("yhat");
  
    // Instanciation de mon code  
    TString sRepertoireReference = gSystem->pwd();
#ifdef WIN32
    URANIE::Launcher::TCode *mycode = new URANIE::Launcher::TCode(_tds, "flowrate.exe -k 1>NUL");
#else
    URANIE::Launcher::TCode *mycode =
            new URANIE::Launcher::TCode(_tds,
                    Form(
                            "%s/../../../../launCHER/teST/coDE/flowREATE/flowrate  -k>>/dev/null",
                            sRepertoireReference.Data()));
#endif

    mycode->addOutputFile(fout);

    URANIE::DataServer::TDataServer *tdsnull =
            new URANIE::DataServer::TDataServer();

    CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TFast(tdsnull, mycode, ns),
            URANIE::Exceptions::UErrorExceptions);

    CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TFast(NULL, mycode, ns),
			 URANIE::Exceptions::UErrorExceptions);
    
    delete fout;
    delete mycode;
    delete tdsnull;
}

void TFastTest::throwExceptionTFastConstructorFalseFnct()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_IGNORED_TEST();

//  CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TFast(_tds, (void*) &toto, ns),URANIE::Exceptions::UErrorExceptions);

}

void TFastTest::throwExceptionTFastConstructorFalseCode()
{
    // The output file of the code
    URANIE::Launcher::TOutputFileRow *fout =
            new URANIE::Launcher::TOutputFileRow(
                    "_output_flowrate_withRow_.dat");
    // The attribute in the output file
    fout->addAttribute("yhat");

    // Instanciation de mon code  
    URANIE::Launcher::TCode *code = new URANIE::Launcher::TCode();

    CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TFast(_tds, code, ns),
			 URANIE::Exceptions::UErrorExceptions);

    CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TFast(_tds,(URANIE::Launcher::TCode *) NULL, ns),
			 URANIE::Exceptions::UErrorExceptions);

    delete fout;
    delete code;
}

void TFastTest::throwExceptionTFastConstructorFalseInput()
{

#ifdef ROOT5
	CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TFast(_tds, _p2fFlowratemodel, ns,
		"rw:r:tu:tl:hulh:toto:kw", "yhat"), URANIE::Exceptions::UErrorExceptions);
#else
	CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TFast(_tds, &flowrateModel,
		"rw:r:tu:tl:hulh:toto:kw", "yhat", ns), URANIE::Exceptions::UErrorExceptions);
#endif

}
void TFastTest::throwExceptionTFastConstructorToMoreInput()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
  URANIE_IGNORED_TEST()   ;

    URANIE::DataServer::TDataServer * tds50 =
            new URANIE::DataServer::TDataServer();
    for (Int_t ik = 0; ik < 50; ik++)
      {
        tds50->addAttribute(new URANIE::DataServer::TUniformDistribution(Form("x_{%d}", ik + 1), 0.0, 1.0));
	tds50->getAttribute(Form("x_{%d}", ik + 1))->delShare();
      }

#ifdef ROOT5
	CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TFast(tds50, _p2fFlowratemodel), URANIE::Exceptions::UErrorExceptions);
#else
	CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TFast(tds50, &flowrateModel, "", "FlowrateModel"), URANIE::Exceptions::UErrorExceptions);
#endif

    delete tds50;
}

void TFastTest::throwExceptionTFastConstructorFalseSize()
{

    TString sJDDReference = gSystem->pwd();
    sJDDReference += TString("/flowrate_input_with_keys.in");
    _tds->getAttribute("rw")->setFileKey(sJDDReference, "Rw");
    _tds->getAttribute("r")->setFileKey(sJDDReference, "R");
    _tds->getAttribute("tu")->setFileKey(sJDDReference, "Tu");
    _tds->getAttribute("tl")->setFileKey(sJDDReference, "Tl");
    _tds->getAttribute("hu")->setFileKey(sJDDReference, "Hu");
    _tds->getAttribute("hl")->setFileKey(sJDDReference, "Hl");
    _tds->getAttribute("l")->setFileKey(sJDDReference, "L");
    _tds->getAttribute("kw")->setFileKey(sJDDReference, "Kw");

    // The output file of the code
    URANIE::Launcher::TOutputFileRow *fout =
            new URANIE::Launcher::TOutputFileRow(
                    "_output_flowrate_withRow_.dat");
    // The attribute in the output file
    fout->addAttribute("yhat");

    // Instanciation de mon code  
    TString sRepertoireReference = gSystem->pwd();
#ifdef WIN32
    URANIE::Launcher::TCode *mycode = new URANIE::Launcher::TCode(_tds, "flowrate.exe -k 1>NUL");
#else
    URANIE::Launcher::TCode *mycode =
            new URANIE::Launcher::TCode(_tds,
                    Form(
                            "%s/../../../../launCHER/teST/coDE/flowREATE/flowrate  -k>>/dev/null",
                            sRepertoireReference.Data()));
#endif
    mycode->addOutputFile(fout);

    ns = 0;

    CPPUNIT_ASSERT_THROW(new URANIE::Sensitivity::TFast(_tds, mycode, ns),
            URANIE::Exceptions::UErrorExceptions);

    delete fout;
    delete mycode;
}

void TFastTest::testTFastsetFrequencies()
{
    TString sJDDReference = gSystem->pwd();
    sJDDReference += TString("/flowrate_input_with_keys.in");
    _tds->getAttribute("rw")->setFileKey(sJDDReference, "Rw");
    _tds->getAttribute("r")->setFileKey(sJDDReference, "R");
    _tds->getAttribute("tu")->setFileKey(sJDDReference, "Tu");
    _tds->getAttribute("tl")->setFileKey(sJDDReference, "Tl");
    _tds->getAttribute("hu")->setFileKey(sJDDReference, "Hu");
    _tds->getAttribute("hl")->setFileKey(sJDDReference, "Hl");
    _tds->getAttribute("l")->setFileKey(sJDDReference, "L");
    _tds->getAttribute("kw")->setFileKey(sJDDReference, "Kw");

    // The output file of the code
    _fout = new URANIE::Launcher::TOutputFileRow(
            "_output_flowrate_withRow_.dat");
    // The attribute in the output file
    _fout->addAttribute("yhat");

    // Instanciation de mon code
    TString sRepertoireReference = gSystem->pwd();
#ifdef WIN32
    _mycode = new URANIE::Launcher::TCode(_tds, "flowrate.exe -k 1>NUL");
#else
    _mycode =
            new URANIE::Launcher::TCode(_tds,
                    Form(
                            "%s/../../../../launCHER/teST/coDE/flowREATE/flowrate  -k>>/dev/null",
                            sRepertoireReference.Data()));
#endif
    _mycode->addOutputFile(_fout);

    ns = 500;
    _tfast = new URANIE::Sensitivity::TFast(_tds, _mycode, ns);
    _tfast->setDrawProgressBar(kFALSE);
    
    Int_t *nff = new Int_t[8];
    nff[0] = 1;
    nff[1] = 2;
    nff[2] = 3;
    nff[4] = 4;
    nff[5] = 5;
    nff[6] = 6;
    nff[7] = 7;
    CPPUNIT_ASSERT_NO_THROW(_tfast->setFrequencies(8, nff));
    delete[] nff;
    nff = NULL;
}

void TFastTest::throwExceptionTFastsetFrequencies()
{

    TString sJDDReference = gSystem->pwd();
    sJDDReference += TString("/flowrate_input_with_keys.in");
    _tds->getAttribute("rw")->setFileKey(sJDDReference, "Rw");
    _tds->getAttribute("r")->setFileKey(sJDDReference, "R");
    _tds->getAttribute("tu")->setFileKey(sJDDReference, "Tu");
    _tds->getAttribute("tl")->setFileKey(sJDDReference, "Tl");
    _tds->getAttribute("hu")->setFileKey(sJDDReference, "Hu");
    _tds->getAttribute("hl")->setFileKey(sJDDReference, "Hl");
    _tds->getAttribute("l")->setFileKey(sJDDReference, "L");
    _tds->getAttribute("kw")->setFileKey(sJDDReference, "Kw");

    // The output file of the code
    URANIE::Launcher::TOutputFileRow *fout =
            new URANIE::Launcher::TOutputFileRow(
                    "_output_flowrate_withRow_.dat");
    // The attribute in the output file
    fout->addAttribute("yhat");

    // Instanciation de mon code
    TString sRepertoireReference = gSystem->pwd();
#ifdef WIN32
    URANIE::Launcher::TCode *mycode = new URANIE::Launcher::TCode(_tds, "flowrate.exe -k 1>NUL");
#else
    URANIE::Launcher::TCode *mycode =
            new URANIE::Launcher::TCode(_tds,
                    Form(
                            "%s/../../../../launCHER/teST/coDE/flowREATE/flowrate  -k>>/dev/null",
                            sRepertoireReference.Data()));
#endif
    mycode->addOutputFile(fout);

    ns = 500;
    _tfast = new URANIE::Sensitivity::TFast(_tds, mycode, ns);
    _tfast->setDrawProgressBar(kFALSE);
    
    Int_t *nff = new Int_t[3];
    nff[0] = 11;
    nff[1] = 21;
    nff[2] = 385;

    CPPUNIT_ASSERT_THROW(_tfast->setFrequencies(3, nff), URANIE::Exceptions::UErrorExceptions);

    delete[] nff;
    delete fout;
    delete mycode;
}

void TFastTest::validateValuesIshigami()
{
    Double_t sobol;
    TTree *tree = _tfast->getResultTuple();
    tree->SetBranchAddress("Sobol", &sobol);

    // Reference values
    Double_t dval[3] =
    { 0.311616, 0.504057, 0.000233093 };

    for (Int_t i = 0; i < 3; i++)
    {
        tree->GetEntry(2 * i);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(dval[i], sobol, 0.00001);
    }
}

void TFastTest::testIshigamiFastComputeIndexesWithTLauncherFunctionName()
{
    setUp10();
    _tfast->computeIndexes();
    validateValuesIshigami();
}

void TFastTest::testIshigamiFastComputeIndexesWithTLauncherFunction()
{
    setUp1();
    _tfast->computeIndexes();
    validateValuesIshigami();
}

void TFastTest::testIshigamiFastComputeIndexesGenerateSampleWithTLauncherFunction()
{
    setUp1();
    _tfast->generateSample();
    _tfast->computeIndexes();
    validateValuesIshigami();
}

void TFastTest::testIshigamiFastComputeIndexesGenerateSampleCreateLauncherWithTLauncherFunction()
{
    setUp1();
    _tfast->generateSample();
    URANIE::Launcher::TLauncherFunction *tlf =
#ifdef ROOT5
            new URANIE::Launcher::TLauncherFunction(_tds2, _p2fIshigami,
#else
            new URANIE::Launcher::TLauncherFunction(_tds2, &Ishigami,
#endif
                    "", "p2fIshigami");
    tlf->setDrawProgressBar(kFALSE);
    tlf->run();
    _tfast->computeIndexes();
    validateValuesIshigami();
    delete tlf;
}


void TFastTest::testIshigamiFastComputeIndexesWithThreadedandNot()
{
    URANIE::DataServer::TAttribute *out = new URANIE::DataServer::TAttribute("TestIshigami");
    out->delShare();
    
    URANIE::Relauncher::TCJitEval eval(&Ishigami);
    eval.addInput(_tds2->getAttribute("x_{1}"));
    eval.addInput(_tds2->getAttribute("x_{2}"));
    eval.addInput(_tds2->getAttribute("x_{3}"));
    eval.addOutput(out);
    
    URANIE::Relauncher::TSequentialRun run(&eval);
    run.startSlave();
    if( run.onMaster() )
    {
        _tfast = new URANIE::Sensitivity::TFast(_tds2, &run, 50);
        _tfast->setDrawProgressBar(kFALSE);
        _tfast->setSeed(123);
        _tfast->computeIndexes();
    
        run.stopSlave();
    }

    TTree *T = _tfast->getResultTuple();    
    ((TTreePlayer*)(T->GetPlayer()))->SetScanRedirect(kTRUE);
    ((TTreePlayer*)(T->GetPlayer()))->SetScanFileName("outpuS.txt");
    T->Scan("*");
    
    _tds2->deleteTuple();
    _tds2->delAttribute("TestIshigami");

    URANIE::DataServer::TAttribute *outT = new URANIE::DataServer::TAttribute("TestIshigamiT");
    outT->delShare();
    
    URANIE::Relauncher::TCJitEval eval2(&Ishigami);
    eval2.addInput(_tds2->getAttribute("x_{1}"));
    eval2.addInput(_tds2->getAttribute("x_{2}"));
    eval2.addInput(_tds2->getAttribute("x_{3}"));
    eval2.addOutput(outT);
    
    URANIE::Relauncher::TThreadedRun runT(&eval2,4);
    runT.startSlave();
    if( runT.onMaster() )
    {
        delete _tfast;

        _tfast = new URANIE::Sensitivity::TFast(_tds2, &runT, 50);
        _tfast->setDrawProgressBar(kFALSE);
        _tfast->setSeed(123);
        _tfast->computeIndexes();

        runT.stopSlave();
    }

    T = _tfast->getResultTuple();    
    ((TTreePlayer*)(T->GetPlayer()))->SetScanRedirect(kTRUE);
    ((TTreePlayer*)(T->GetPlayer()))->SetScanFileName("outpuT.txt");
    T->Scan("*");

    Int_t nret = UranieTests::areFilesEqual(TString("outpuT.dat"),
                                            TString("outpuS.dat"));

    CPPUNIT_ASSERT_EQUAL(nret, 0);        
}

        
void TFastTest::testIshigamiFastComputeIndexesWithTLauncher2()
{
    URANIE::DataServer::TAttribute *out = new URANIE::DataServer::TAttribute("Ishigami");
    out->delShare();
    
    URANIE::Relauncher::TCJitEval eval(&Ishigami);
    eval.addInput(_tds2->getAttribute("x_{1}"));
    eval.addInput(_tds2->getAttribute("x_{2}"));
    eval.addInput(_tds2->getAttribute("x_{3}"));
    eval.addOutput(out);
    
    URANIE::Relauncher::TSequentialRun run(&eval);
    run.startSlave();
    if( run.onMaster() )
    {
        ns = 500;
        _tfast = new URANIE::Sensitivity::TFast(_tds2, &run, ns);
        _tfast->setDrawProgressBar(kFALSE);
        _tfast->setSeed(123);
        _tfast->computeIndexes();
        validateValuesIshigami();

        run.stopSlave();
    }

}


void TFastTest::validateValuesFlowrate()
{
    Double_t sobol;
    TTree *tree = _tfast->getResultTuple();
    tree->SetBranchAddress("Sobol", &sobol);

    // Reference values
    Double_t dval[8] =
    { 0.827924, 2.09805e-06, 2.37838e-06, 3.220e-05, 0.0414499, 0.0413399,  0.0392563, 0.0095494 };

    for (Int_t i = 0; i < 8; i++)
    {
        tree->GetEntry(2 * i);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(dval[i], sobol, 0.00001);
    }
}

void TFastTest::testFlowrateFastComputeIndexesWithTLauncherCode()
{
    setUp2();
    _tfast->computeIndexes("nointermed");
    validateValuesFlowrate();
}

void TFastTest::testFlowrateFastComputeIndexesGenerateSampleWithTLauncherCode()
{
    setUp2();
    _tfast->generateSample();
    _tfast->computeIndexes("nointermed");
    validateValuesFlowrate();
}

void TFastTest::testFlowrateFastComputeIndexesGenerateSampleCreateLauncherWithTLauncherCode()
{
    setUp2();
    _tfast->generateSample();
    URANIE::Launcher::TLauncher *tl = new URANIE::Launcher::TLauncher(_tds,
            _mycode);
    tl->setDrawProgressBar(kFALSE);
    tl->run("nointermed");
    _tfast->computeIndexes("nointermed");
    validateValuesFlowrate();
    delete tl;
}
