//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
#ifndef CPPUNITEST_TFINITEDIFFERENCES_H
#define CPPUNITEST_TFINITEDIFFERENCES_H

#include <cppunit/extensions/HelperMacros.h>
//#include <cppunit/portability/CppUnitDeque.h>
#include <cppunit/Test.h>
#include <cppunit/TestFailure.h>
#include <cppunit/TestResultCollector.h>

// ROOT
#include "TMatrixD.h"

// Uranie
#include "TFiniteDifferences.h"
#include "UExceptions.h"
#include "TUniformDistribution.h"
#include "TNormalDistribution.h"
#include "TOutputFileDataServer.h"

class TFiniteDifferencesTest: public CPPUNIT_NS::TestFixture
{
    CPPUNIT_TEST_SUITE (TFiniteDifferencesTest);

    CPPUNIT_TEST (testFiniteDifferencesWithCode);
    // CPPUNIT_TEST( testFiniteDifferencesWithFunction );
    CPPUNIT_TEST (testFiniteDifferencesWithDataFile);
#ifdef WITH_LEVELE
    CPPUNIT_TEST (testLevelEFiniteDifferences);
    CPPUNIT_TEST (testLevelEFiniteDifferencesOutputRestrVarexp);
    CPPUNIT_TEST (testLevelEFiniteDifferencesOutputRestrOption);
    CPPUNIT_TEST (testLevelEFiniteDifferencesOutputRestrBoth);
#endif
    CPPUNIT_TEST (testIshigamiFiniteDifferencesComputeIndexesWithTLauncherFunction);
    CPPUNIT_TEST (testIshigamiFiniteDifferencesComputeIndexesGenerateSampleWithTLauncherFunction);
    CPPUNIT_TEST (testIshigamiFiniteDifferencesComputeIndexesGenerateSampleCreateLauncherWithTLauncherFunction);
    CPPUNIT_TEST (testIshigamiFiniteDifferencesComputeIndexesWithTLauncher2);
     CPPUNIT_TEST (testIshigamiFiniteDifferencesComputeIndexesThreaded);
    CPPUNIT_TEST (testFlowrateFiniteDifferencesComputeIndexesWithTLauncherCode);
    CPPUNIT_TEST (testFlowrateFiniteDifferencesComputeIndexesGenerateSampleWithTLauncherCode);
    CPPUNIT_TEST (testFlowrateFiniteDifferencesComputeIndexesGenerateSampleCreateLauncherWithTLauncherCode);

    CPPUNIT_TEST (throwExceptionConstructorTooLongOutputFunction);
    CPPUNIT_TEST (throwExceptionConstructorTooLongOutputCode);
    CPPUNIT_TEST (throwExceptionConstructorTooLongOutputRun);

    CPPUNIT_TEST_SUITE_END();

public:
    //---------------------------------------------
    /** @name Constructor and Destructor
     **/
    //@{
    //! Default constructor
    TFiniteDifferencesTest();

    //! Default destructor
    virtual ~TFiniteDifferencesTest();

    void setUp();
    void setUp1();
    void setUp2();
    void tearDown();

    //! the clean tempory directories
    void cleanWorkingDirectories(const char *option = "without");
    //@}

    //---------------------------------------------

    /** @name Validation of the sample values for Ishigami
     **/
    //@{
    void validateValuesIshigami();
    //@}

    /** @name Validation of the sample values for Flowrate
     **/
    //@{
    void validateValuesFlowrate();
    //@}

    //---------------------------------------------
    /** @name Unit Tests
     The list of unit test
     **/
    //@{ 
    //! Compute finite differences of an external code (flowrate)
    void testFiniteDifferencesWithCode();

//  //! Compute finite differences of an analytical function (flowrate)
//   void testFiniteDifferencesWithFunction();

    //! Compute finite differences using an existing dataset
    void testFiniteDifferencesWithDataFile();
#ifdef WITH_LEVELE
    //! Test of the FiniteDifferences for both SRC and SRRC using LevelE
    void testLevelEFiniteDifferences();
    //! Test the Sobol method for the Levele function restraining output vector in constructor
    void testLevelEFiniteDifferencesOutputRestrVarexp();
    //! Test the Sobol method for the Levele function restraining output vector in option
    void testLevelEFiniteDifferencesOutputRestrOption();
    //! Test the Sobol method for the Levele function restraining output vector in both
    void testLevelEFiniteDifferencesOutputRestrBoth();
    //! Check the results of the levelE sobol estrimation
    void checkLevelEResults(TMatrixD mat, vector<int> *elmts = NULL);
#endif
    void testIshigamiFiniteDifferencesComputeIndexesWithTLauncherFunction();
    void testIshigamiFiniteDifferencesComputeIndexesGenerateSampleWithTLauncherFunction();
    void testIshigamiFiniteDifferencesComputeIndexesGenerateSampleCreateLauncherWithTLauncherFunction();
    void testIshigamiFiniteDifferencesComputeIndexesWithTLauncher2();
    void testIshigamiFiniteDifferencesComputeIndexesThreaded();
    void testFlowrateFiniteDifferencesComputeIndexesWithTLauncherCode();
    void testFlowrateFiniteDifferencesComputeIndexesGenerateSampleWithTLauncherCode();
    void testFlowrateFiniteDifferencesComputeIndexesGenerateSampleCreateLauncherWithTLauncherCode();

    void throwExceptionConstructorTooLongOutputFunction();
    void throwExceptionConstructorTooLongOutputCode();
    void throwExceptionConstructorTooLongOutputRun();
    //@}

    //---------------------------------------------
    /** @name Throw exceptions
     **/
    //@{ 
    //! Throw an exception when attributes are not properly set.
    /*!
     */
    //@}

private:
    TString _workingTestPath;

    URANIE::DataServer::TUniformDistribution *rw;
    URANIE::DataServer::TUniformDistribution *r;
    URANIE::DataServer::TUniformDistribution *tu;
    URANIE::DataServer::TUniformDistribution *tl;
    URANIE::DataServer::TUniformDistribution *hu;
    URANIE::DataServer::TUniformDistribution *hl;
    URANIE::DataServer::TUniformDistribution *l;
    URANIE::DataServer::TUniformDistribution *kw;

    URANIE::DataServer::TUniformDistribution *x1;
    URANIE::DataServer::TUniformDistribution *x2;
    URANIE::DataServer::TUniformDistribution *x3;

    TMatrixD _refMat;

    URANIE::DataServer::TDataServer *_tds, *_tds2;
    URANIE::Sensitivity::TFiniteDifferences * _tfd;
    URANIE::Launcher::TCode *_mycode;
    void * _p2fIshigami;
    URANIE::Launcher::TOutputFile *_fout;
};

#endif  // CPPUNITEST_TDATASERVER_H
