//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////

/*!
 @file   TCirce.cxx
 @author Fabrice Gaudier  <fabrice.gaudier@cea.fr>
 @date 2011-08-04 16:14:29

 @brief Implement the TCirce wrapper

 */

#include <ctime>
#include <iostream>
#include <fstream>

using std::cout;
using std::endl;
using std::ostringstream;

// ROOT
#include "TDecompSVD.h"
#include "TRandom3.h"
#include "TLeaf.h"
#include "TMath.h"
#include "TCanvas.h"
#include "TRegexp.h"
#include "TMatrixTUtils.h"

// URANIE
#include "TCirce.h"
#include "UExceptions.h"

ClassImp (URANIE::UncertModeler::TCirce);

namespace URANIE
{
namespace UncertModeler
{

////////////////////////////////////////////////////////////
//  Default constructor
////////////////////////////////////////////////////////////
TCirce::TCirce(URANIE::DataServer::TDataServer *tds, const char *systar,
        const char *syhat, const char *ssensi, const char *sexpuncert,
        Option_t * option) :
        TUncertModeler(tds, "", "", option), _dTol(1.0e-5), _dLikelihood(0), _vecBiais(0), 
                _nPatterns(0), _nParameters(0), _sCirceInputFile(""), 
                _nCirceMethod(kUnknown), _nInitCAlea(1), _listOfTDS(NULL), 
                _listOfInformations(NULL), _listOfBloc(NULL),  _listOfSensitivity(NULL), 
                _vecDiff(0), _vecYStarVariance(0), _matSensibility(0, 0),
                _matInitialC(0,0), _vecInitialB(0),
                _matVarianceMu(0, 0), _matVarianceSigma(0, 0)               
{
    // Add the Data
    addData(tds, systar, syhat, ssensi, sexpuncert);

    // Set the default seed
    setSeed();
}

////////////////////////////////////////////////////////////
//  Default destructor
////////////////////////////////////////////////////////////
TCirce::~TCirce()  
{
  if( _listOfTDS)
    delete _listOfTDS;
  _listOfTDS=NULL;

  if( _listOfInformations)
    {
      _listOfInformations->Delete();
      delete _listOfInformations;
    }
  _listOfInformations=NULL;

  if( _listOfBloc)
    {
      _listOfBloc->Delete();      
      delete _listOfBloc;
    }
  _listOfBloc=NULL;

  if( _listOfSensitivity)
    delete _listOfSensitivity;
  _listOfSensitivity=NULL;
  
}

////////////////////////////////////////////////////////////
//  Add DATA from an another TDS
////////////////////////////////////////////////////////////
void TCirce::addData(URANIE::DataServer::TDataServer *tds, const char *systar,
        const char *syhat, const char *ssensi, const char *sexpuncert)
{
    TString sYStar = TString(systar);
    TString sYHat = TString(syhat);
    TString sYStarSigma = TString(sexpuncert);
    TString sOfSensibility = TString(ssensi);
    sOfSensibility.ReplaceAll(" ", "");
    sOfSensibility.ReplaceAll(",", " ");

    cout << " *********" << endl;
    cout << " ** addData from an another TDS [" << tds->GetName() << "]"
            << endl;
    cout << " ** YStar[" << sYStar << "]  YStarSigma[" << sYStarSigma
            << "]YHat[" << sYHat << "]" << endl;
    cout << " **  Sensitivity Attributes[" << sOfSensibility << "]" << endl;

    try
    {
        if (tds->isAttribute(sYStar) == kFALSE)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TCirce addData method\n the YStar attribute \"%s\" is not in the tds",
                            sYStar.Data()));
        if (tds->isAttribute(sYHat) == kFALSE)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TCirce addData method\n the YHat attribute \"%s\" is not in the tds",
                            sYHat.Data()));
        if (sYStarSigma.Length() > 0)
        {
            if (tds->isAttribute(sYStarSigma) == kFALSE)
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        Form(
                                "TCirce addData method\n the YStarSigma attribute \"%s\" is not in the tds",
                                sYStarSigma.Data()));
        }

        if (sOfSensibility.Length() == 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TCirce addData method\n the sensitivity attributes \"%s\" is empty",
                            sOfSensibility.Data()));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    // Treat the sensibility string to compute the number of parameters
    if (_listOfSensitivity)
        delete _listOfSensitivity;
    _listOfSensitivity = NULL;
    
    try
    {
        istringstream inStream((char *) sOfSensibility.Data());
        string inName;
        while (inStream >> inName)
        {
            TString svar(inName.c_str());
            if (_blog)
                cout << " ** svar[" << svar << "]" << endl;
            try
            {
                addBloc(tds, inName.c_str());
            } catch (URANIE::Exceptions::UExceptions& ue)
            {
                ue.printMessage();
                throw ue;
            }
        }

        cout << " ** nparameter [" << sOfSensibility << "] size["
                << _listOfSensitivity->GetEntries() << "]" << endl;
        if (_nParameters == 0)
        {
            _nParameters = _listOfSensitivity->GetEntries();
        }
        else
        {
            if (_nParameters != _listOfSensitivity->GetEntries())
            {
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        Form(
                                "TCirce addData method\n the number of parameters %d is not the same before %d",
                                _listOfSensitivity->GetEntries(),
                                _nParameters));
            }
        }
    } catch (URANIE::Exceptions::UExceptions& ue)
    {
        ue.printMessage();
        _listOfSensitivity->Print();
        throw ue;
    }

    if (!_listOfTDS)
        _listOfTDS = new TList();

    if (!_listOfInformations)
        _listOfInformations = new TList();

    _listOfTDS->Add(tds);
    Int_t nsize = _listOfTDS->GetEntries();

    _listOfInformations->Add(
            new TNamed(Form("__Circe_YStar_%s_%d__", tds->GetName(), nsize),
                    Form("%s", systar)));
    _listOfInformations->Add(
            new TNamed(Form("__Circe_YHat_%s_%d__", tds->GetName(), nsize),
                    Form("%s", syhat)));
    _listOfInformations->Add(
            new TNamed(
                    Form("__Circe_Sensitivity_%s_%d__", tds->GetName(), nsize),
                    Form("%s", ssensi)));
    if (sYStarSigma.Length() > 0)
        _listOfInformations->Add(
                new TNamed(
                        Form("__Circe_Experimental_Uncertainty_%s_%d__",
                                tds->GetName(), nsize), sexpuncert));

    URANIE::DataServer::TAttribute *tatt = new URANIE::DataServer::TAttribute(
            Form("%s", CIRCE_RESIDU_ATTRIBUTE));
//       tatt->printLog();
    tds->addAttribute(tatt);
    tatt->delShare();

    cout << " ** List Of TDS size[" << _listOfTDS->GetEntries() << "]" << endl;
    _listOfTDS->Print();
    cout << " ** List Of Informations size["
            << _listOfInformations->GetEntries() << "]" << endl;
    _listOfInformations->Print();
    cout << " ** End Of addData from an another TDS [" << tds->GetName() << "]"
            << endl;
    cout << " *********" << endl;
}

////////////////////////////////////////////////////////////
// Get the number of TDS in the List
////////////////////////////////////////////////////////////
Int_t TCirce::getNTDS()
{
    Int_t n = 0;
    if (_listOfTDS)
        n = _listOfTDS->GetEntries();

    return n;
}

////////////////////////////////////////////////////////////
//  Set the tolerance parameter (positive value)
////////////////////////////////////////////////////////////
void TCirce::setTolerance(Double_t dtol)
{
    if (_blog)
        cout << endl << " ** TCirce::setTolerance Tolerance[" << dtol << "]"
                << endl;

    try
    {
        if (dtol < 0.)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TCirce::setTolerance method\n the tolerance is negative \"%e\". Positive value is mandatory.",
                            dtol));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    _dTol = dtol;

}

////////////////////////////////////////////////////////////
//  Set the Circe Input File
////////////////////////////////////////////////////////////
void TCirce::setInputFile(const char * sfile)
{
    _sCirceInputFile = TString(sfile);
    try
    {
        const char * sfindfile = gSystem->FindFile(gSystem->pwd(),
                _sCirceInputFile, kFileExists);
        if (sfindfile == 0)
        {
            _sCirceInputFile = "";
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TCirce::setInputFile Method\n The file \"%s\" does not exist.",
                            sfile));
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

}

////////////////////////////////////////////////////////////
//  Set and Read the Circe Input File
////////////////////////////////////////////////////////////
void TCirce::readInputFile(const char * sfile, ECirceMethod nCirceMethod)
{
    UNUSED(nCirceMethod);
    try
    {
        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                Form(
                        "TCirce::readInputFile Method from \"%s\" \n Not Yet implementd .",
                        sfile));

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

}

////////////////////////////////////////////////////////////
//  setInitialCMatrix method
////////////////////////////////////////////////////////////
void TCirce::setCMatrixInitial(TMatrixD mat)
{
    if (_blog)
    {
        cout << endl << " ******************" << endl
                << " ** TCirce::setInitialCMatrix rows[" << mat.GetNrows()
                << "] col[" << mat.GetNcols() << "]" << endl;
        mat.Print();
    }

    _matInitialC.ResizeTo(mat.GetNrows(), mat.GetNcols());
    _matInitialC = mat;

}
////////////////////////////////////////////////////////////
//  setBVectorInitial method
////////////////////////////////////////////////////////////
void TCirce::setBVectorInitial(TVectorD vec)
{
    if (_blog)
    {
        cout << endl << " ******************" << endl
                << " ** TCirce::setBVectorInitial rows[" << vec.GetNrows()
                << "]" << endl;
        vec.Print();
    }

    _vecInitialB.ResizeTo(vec.GetNrows());
    _vecInitialB = vec;

}

////////////////////////////////////////////////////////////
//  initCMatrix method
////////////////////////////////////////////////////////////
TMatrixD TCirce::initCMatrix(Int_t nrows, Double_t dMulti)
{
    if (_blog)
    {
        cout << endl << " ******************" << endl
                << " ** TCirce::initCMatrix nrows[" << nrows << "] dMulti ["
                << dMulti << "]" << endl;
    }

    TMatrixD matD(nrows, nrows);
    matD.Zero();

    for (Int_t i = 0; i < nrows; i++)
    {
        matD(i, i) = dMulti * gRandom->Rndm();
    }

    if (_blog)
    {
        cout << " ** new Matrix C" << endl;
        matD.Print();
        cout << " ** End Of initCMatrix" << endl << " ******************"
                << endl;
    }
    return matD;
}

////////////////////////////////////////////////////////////
//  init method
////////////////////////////////////////////////////////////
void TCirce::init(Option_t * option)
{
    if (_blog)
        cout << endl << " ** TCirce::init option[" << option << "]" << endl;

    Int_t ntds = _listOfTDS->GetEntries();
    if (_blog)
        cout << " ** size of the list of TDS [" << ntds << "]" << endl;

    ////////////////////////////////////////////
    // Compute the size : sum of tds patterns //
    ////////////////////////////////////////////
    Int_t npat = 0;
    URANIE::DataServer::TDataServer *tdscrt;
    TIter nextTDS(_listOfTDS);
    while ((tdscrt = (URANIE::DataServer::TDataServer*) nextTDS()))
    {
        npat += tdscrt->getNPatterns();
    }
    _nPatterns = npat;
    if (_blog)
        cout << " ** nPatterns[" << _nPatterns << "]" << endl;
    //TCanvas *cdiff = new TCanvas("cdiffcirece");
    //cdiff->Divide(ntds);

    //////////////////////////////////////
    // The vector of diff Yexp - Ycode  //
    //////////////////////////////////////
    _vecDiff.ResizeTo(_nPatterns);
    nextTDS.Reset();
    Int_t itds = 1;
    Int_t ipat = 0;
    while ((tdscrt = (URANIE::DataServer::TDataServer*) nextTDS()))
    {
        if (_blog)
            cout << " ** TDS name[" << tdscrt->GetName()
                    << "] -- Vector of diff Yexp - Ycode" << endl;

        TNamed *circeInfoYStar =
                dynamic_cast<TNamed *>(_listOfInformations->FindObject(
                        Form("__Circe_YStar_%s_%d__", tdscrt->GetName(), itds)));
        if (_blog)
            circeInfoYStar->Print();

        TNamed *circeInfoYHat =
                dynamic_cast<TNamed *>(_listOfInformations->FindObject(
                        Form("__Circe_YHat_%s_%d__", tdscrt->GetName(), itds)));
        if (_blog)
            circeInfoYHat->Print();

        TString sYStar = TString(circeInfoYStar->GetTitle());
        TString sYHat = TString(circeInfoYHat->GetTitle());
        if (_blog)
            cout << " ** sYStar[" << sYStar << "] sYHat[" << sYHat << "]"
                    << endl;

        //cdiff->cd(itds);
        //tdscrt->draw(Form("%s-%s", sYStar.Data(), sYHat.Data()));

        TLeaf * tlStar = tdscrt->getTuple()->GetLeaf(sYStar);
        TLeaf * tlHat = tdscrt->getTuple()->GetLeaf(sYHat);
        for (Int_t i = 0; i < tdscrt->getNPatterns(); i++)
        {
            tdscrt->getTuple()->GetEntry(i);
            _vecDiff(ipat++) = tlStar->GetValue(0) - tlHat->GetValue(0);
        }

        itds++;
    }

    if (_blog)
    {
        cout << endl << " ** vecDiff" << endl;
        _vecDiff.Print();
    }

    //////////////////////////////////////////
    // The vector of Experimental Variance  //
    //////////////////////////////////////////
    nextTDS.Reset();
    itds = 1;
    ipat = 0;
    _vecYStarVariance.ResizeTo(_nPatterns);
    _vecYStarVariance.Zero();
    while ((tdscrt = (URANIE::DataServer::TDataServer*) nextTDS()))
    {
        if (_blog)
            cout << " ** TDS name[" << tdscrt->GetName()
                    << "] -- vector of Experimental Variance" << endl;

        TNamed *circeInfo =
                dynamic_cast<TNamed *>(_listOfInformations->FindObject(
                        Form("__Circe_Experimental_Uncertainty_%s_%d__",
                                tdscrt->GetName(), itds)));

        if (circeInfo)
        {
            if (_blog)
            {
                cout << " ** circeInfo Named[" << circeInfo->GetName() << "]"
                        << endl;
                circeInfo->Print();
            }

            TString sOfExpUncer = TString(circeInfo->GetTitle());
            if (_blog)
                cout << " ** sOfExpUncer[" << sOfExpUncer << "]" << endl;

            Double_t dsigma = 0.0;
            TLeaf * tlystarsigma = tdscrt->getTuple()->GetLeaf(sOfExpUncer);

            for (Int_t i = 0; i < tdscrt->getNPatterns(); i++)
            {
                tdscrt->getTuple()->GetEntry(i);
                dsigma = tlystarsigma->GetValue(0);
                _vecYStarVariance(ipat++) = dsigma * dsigma;
                if (_blog)
                    cout << " ** (" << i << "/" << tdscrt->getNPatterns()
                            << ") value[" << tlystarsigma->GetValue(0)
                            << "] dsigma[" << dsigma << "] Variance["
                            << _vecYStarVariance(ipat - 1) << "]" << endl;
            }
        }
        else
        {
            ipat += tdscrt->getNPatterns();
        }
        itds++;
    }

    if (_blog)
    {
        cout << endl << " ** vecYStarVariance" << endl;
        _vecYStarVariance.Print();
    }

    ////////////////////////////////
    // The matrix of sensitivity  //
    ////////////////////////////////
    nextTDS.Reset();
    itds = 1;
    ipat = 0;
    while ((tdscrt = (URANIE::DataServer::TDataServer*) nextTDS()))
    {
        if (_blog)
            cout << " ** TDS name[" << tdscrt->GetName()
                    << "] -- matrix of sensitivity" << endl;
        TNamed *circeInfo =
                dynamic_cast<TNamed *>(_listOfInformations->FindObject(
                        Form("__Circe_Sensitivity_%s_%d__", tdscrt->GetName(),
                                itds)));

        if (_blog)
            circeInfo->Print();

        if (_listOfSensitivity)
            delete _listOfSensitivity;
        _listOfSensitivity = NULL;

        // Treat the sensibility string if exist
        TString sOfSensibility = TString(circeInfo->GetTitle());
        sOfSensibility.ReplaceAll(" ", "");
        sOfSensibility.ReplaceAll(",", " ");
        if (sOfSensibility.Length() > 0)
        {
            if (_blog)
                cout << " ** Treat the sensibility string [" << sOfSensibility
                        << "]" << endl;
            istringstream inStream((char *) sOfSensibility.Data());
            string inName;
            while (inStream >> inName)
            {
                TString svar(inName.c_str());
                if (_blog)
                    cout << " ** svar[" << svar << "]" << endl;
                addBloc(tdscrt, inName.c_str());
            }
        }

        if (ipat == 0)
        {
            _matSensibility.ResizeTo(_nPatterns,
                    _listOfSensitivity->GetEntries());
            _matSensibility.Zero();
        }

        Int_t icol = 0;
        URANIE::DataServer::TAttribute *tatt;
        TIter next(_listOfSensitivity);
        while ((tatt = (URANIE::DataServer::TAttribute*) next()))
        {
            Int_t ipatcol = ipat;
            if (_blog)
                cout << " ** icol[" << icol << "] att[" << tatt->GetName()
                        << "]" << endl;
            for (Int_t i = 0; i < tdscrt->getNPatterns(); i++)
                _matSensibility(ipatcol++, icol) = tdscrt->getValue(
                        tatt->GetName(), i);
            icol++;
        }

        ipat += tdscrt->getNPatterns();
        itds++;
    }
    if (_blog)
    {
        cout << endl << " ** matSensitivity" << endl;
        _matSensibility.Print();
    }

    // Pass to the Transpose matrix
    _matSensibility.T();
}
////////////////////////////////////////////////////////////
//  addBloc method
////////////////////////////////////////////////////////////
void TCirce::addBloc(URANIE::DataServer::TDataServer *tds, const char *sbloc,
        const char* sname)
{
    if (_blog)
        cout << endl << " ** TCirce::addBloc tds[" << tds->GetName()
                << "] sbloc[" << sbloc << "] name[" << sname << "]" << endl;

    try
    {
        TString svar = TString(sbloc);
        if (svar.CountChar(':') > 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TCirce addBloc method\n the multi-parameter in the same bloc \"%s\" is not yet implemented",
                            svar.Data()));

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    Int_t nbloc = 0;
    if (_listOfBloc == NULL)
        _listOfBloc = new TList();
    else
        nbloc = _listOfBloc->GetEntries();

    if (_listOfSensitivity == NULL)
        _listOfSensitivity = new TList();

    nbloc++;
    TString sBlocName = TString(sname);
    if (sBlocName.Length() == 0)
        sBlocName = Form("%s%d__", CIRCE_BLOC_NAME, nbloc);

    if (_blog)
        cout << " ** Bloc Name[" << sBlocName << "]" << endl;

    _listOfBloc->Add(new TNamed(sBlocName.Data(), sbloc));

    TList * tl = tds->getListOfAttributes(sbloc);
    if (_blog)
        tl->Print();

    // Looking for if the parameter is not already in the list (no double parameter)
    try
    {
        URANIE::DataServer::TAttribute *tatt;
        TIter next(tl);
        while ((tatt = (URANIE::DataServer::TAttribute*) next()))
        {
            URANIE::DataServer::TAttribute *tattList =
                    dynamic_cast<URANIE::DataServer::TAttribute *>(_listOfSensitivity->FindObject(
                            tatt->GetName()));
            if (!tattList)
            {
                if (_blog)
                    cout << " ** tattList name[" << tatt->GetName()
                            << "] not found" << endl;
                _listOfSensitivity->Add(tatt);
            }
            else
            {
                if (_blog)
                    cout << endl << " ** tattList name[" << tatt->GetName()
                            << "] was found" << endl;
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        Form(
                                "TCirce::addBloc method\nThe sensitivity attribute[%s] for the bloc[%s] is already present in the list",
                                tatt->GetName(), sBlocName.Data()));
            }
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
	delete tl;
	tl=NULL;
        _listOfSensitivity->Print();
        throw ue;
    }

    if (_blog)
    {
        cout << " ** List Of bloc [" << _listOfBloc->GetEntries() << "]"
                << endl;
        _listOfBloc->Print();
        cout << " ** List Of Sensitivity Parameter ["
                << _listOfSensitivity->GetEntries() << "]" << endl;
        _listOfSensitivity->Print();
    }

    delete tl;
    tl=NULL;
}
////////////////////////////////////////////////////////////
//  estimate method
////////////////////////////////////////////////////////////
void TCirce::estimate(Option_t * option)
{
    if (_blog)
        cout << " TCirce::estimate option[" << option << "]" << endl;

    // init the matrix
    init(option);

    Int_t npat = _nPatterns;
    Int_t nparam = _listOfSensitivity->GetEntries();

    TMatrixD C1(nparam, nparam);
    TMatrixD C2(nparam, nparam);
    TMatrixD matDeltaC(nparam, nparam);

    // Vector
    TVectorD vecBM(nparam);

    // Matrix
    TMatrixD matAM(nparam, nparam);
    matAM.Zero();

    TMatrixD matB(nparam, nparam);

//    TVectorD vecXDEN(npat);

    _vecBiais.ResizeTo(nparam);
    _matC.ResizeTo(nparam, nparam);
    _matCorrelation.ResizeTo(nparam, nparam);

    TVectorD vecXM1(nparam);
    TVectorD vecXM2(nparam);

    Double_t dY = 0.0;
    Double_t dXDEN = 0.0;
    Double_t dA3 = 0.0;
    Double_t dLikelihood;
    Int_t nSelected = 0;

    // loop for the estimation
    for (Int_t iCAlea = 0; iCAlea < _nInitCAlea; iCAlea++)
    {
        if (1)
            cout << endl << " **********************************" << endl
                 << " ** Begin Of Initial Matrix C [" << iCAlea + 1 << "/"
                 << _nInitCAlea << "]" << endl;

        // Init
        C1.Zero();

        if (iCAlea == 0)
        {
            // Initial C matrix
            if ((_matInitialC.GetNrows() != 0)
                    && (_matInitialC.GetNcols() != 0))
                C2 = _matInitialC;
            else
                C2.UnitMatrix();

            // Initial B vector
            if (_vecInitialB.GetNrows() != 0)
                vecXM2 = _vecInitialB;
            else
                vecXM2.Zero();
        }
        else
        {
            Double_t dMulti = 2.0;
            if (_blog)
                cout << " ** Generate a random C matrix (positive definite)"
                        << endl;
            C2 = initCMatrix(C2.GetNrows(), dMulti);
            for (Int_t irnd = 0; irnd < C2.GetNrows(); irnd++)
                vecXM2(irnd) = dMulti * (2.0 * gRandom->Rndm() - 1.0);
        }

        if (_blog) {
            cout << endl << " **********************************" << endl;
            cout << " ** Initial Matrix C [" << iCAlea + 1 << "/" << _nInitCAlea
                    << "]" << endl;
            C2.Print();
            cout << " ** Initial Vector B [" << iCAlea + 1 << "/" << _nInitCAlea
                    << "]" << endl;
            vecXM2.Print();
            cout << " **********************************" << endl << endl;
        }

        Int_t niter = 0;
        Bool_t bOKcriteria = kFALSE;

        while (bOKcriteria == kFALSE)
        {

            // init to zero
            vecBM.Zero();
            matAM.Zero();
            matDeltaC.Zero();
            dLikelihood = 0.0;

            // Compute the C(t+1) matrix
            for (Int_t ipat = 0; ipat < npat; ipat++) {
                dY = _vecDiff(ipat);

                // Compute XDEN
                TVectorD vecS = TMatrixDColumn_const(_matSensibility, ipat);
                //       if ( _blog ) vecS.Print();
                TVectorD vecXDEN = C2 * vecS;

                if (_blog) {
                    cout << " ** vector vecXDEN dim(" << vecXDEN.GetNrows()
                            << ")" << endl;
                    vecXDEN.Print();
                }

                // Take into account the variance of the Ystar
                dXDEN = vecS * vecXDEN + _vecYStarVariance(ipat);
                dA3 = vecS * vecXM2;

                if (_blog)
                    cout << " ipat[" << ipat << "] dXDEN[" << dXDEN << "]"
                            << endl;

                // Compute XNUM
                matB.Zero();
                for (Int_t jparam = 0; jparam < nparam; jparam++)
                    // Prise en compte des blocs FGA::2011-12-08 15:50:29
                    //for(Int_t kparam=0; kparam<nparam; kparam++)
                    matB(jparam, jparam) = vecXDEN(jparam) * vecXDEN(jparam);

                //        if ( _blog ) matB.Print();

                //  Compute the C Matrix
                for (Int_t jparam = 0; jparam < nparam; jparam++)
                {
                    for (Int_t kparam = 0; kparam < nparam; kparam++)
                    {
                        Double_t dDeltaC = matB(jparam, kparam)
                                * ((dY - dA3) * (dY - dA3) / dXDEN - 1.0);
                        dDeltaC /= dXDEN * npat;
                        matDeltaC(jparam, kparam) += dDeltaC;
                    }
                }

                // Compute BM
                for (Int_t jparam = 0; jparam < nparam; jparam++)
                {
                    vecBM(jparam) += vecS(jparam) * dY / dXDEN;
                    //         cout << " * CRT BM[" << ipat << "," << jparam <<"] :: " << vecS(jparam) << " " << dY << " " << dXDEN << " = " << vecBM(jparam) << endl;
                }

                // Compute AM
                for (Int_t jparam = 0; jparam < nparam; jparam++)
                {
                    for (Int_t kparam = 0; kparam <= jparam; kparam++)
                    {
                        matAM(kparam, jparam) += _matSensibility(jparam, ipat)
                                * _matSensibility(kparam, ipat) / dXDEN;
                        if (kparam != jparam)
                            matAM(jparam, kparam) = matAM(kparam, jparam);
                    }
                }

                // uptodate the LikeLihood
                dLikelihood += TMath::Log(
                        TMath::Exp(-1.0 * (dY - dA3) * (dY - dA3) / (2 * dXDEN))
                                / TMath::Sqrt(2 * TMath::Pi() * dXDEN));

            } // End Of Loop ipat from 0 to npat

            // compute the LikeLihood
            dLikelihood /= 1.0 * npat;

            if (_blog) {
                cout << endl << " ** iter[" << niter << "]" << " ** Likelihood[" << dLikelihood << "]"
                        << endl;

                cout << "matDeltaC" << endl;
                matDeltaC.Print();

                cout << endl << "C2" << endl;
                C2.Print();
            }

            C1 = C2 + matDeltaC;

            if (_blog) {
                cout << endl << "C1" << endl;
                C1.Print();
                cout << endl << "vecBM" << endl;
                vecBM.Print();
                cout << endl << "matAM" << endl;
                matAM.Print();
            }

            TDecompSVD svd(matAM);
            Bool_t bOK = kFALSE;
            vecXM1 = svd.Solve(vecBM, bOK);
            niter++;

            if (_blog) {
                cout << endl << "vecXM1 iter[" << niter << "]" << endl;
                vecXM1.Print();
            }

            // Test of Convergence |C1-C2|/C2 < dTol && |XM1 -XM2|/XM2 < dTol
            bOKcriteria = kTRUE;
            for (Int_t jparam = 0; jparam < nparam; jparam++) {
                for (Int_t kparam = 0; kparam < nparam; kparam++) {
                    if (TMath::Abs(C1(jparam, jparam) - C2(jparam, jparam))
                            / C2(jparam, jparam) > _dTol)
                        bOKcriteria = kFALSE;
                }
            }

            if (bOKcriteria) {
                for (Int_t jparam = 0; jparam < nparam; jparam++) {
                    if (TMath::Abs(vecXM1(jparam) - vecXM2(jparam))
                            / vecXM2(jparam) > _dTol)
                        bOKcriteria = kFALSE;
                }
            }

            // Pass to the new iteration
            C2 = C1;
            vecXM2 = vecXM1;

        }  // Enf Of While

        if (iCAlea == 0) {
            _dLikelihood = dLikelihood;
            nSelected = iCAlea;
            _vecBiais = vecXM1;
            _matC = C1;
        } else {
            if (dLikelihood > _dLikelihood) {
                _dLikelihood = dLikelihood;
                nSelected = iCAlea;
                _vecBiais = vecXM1;
                _matC = C1;
            }
        }
        cout.precision(16);
        cout << endl << " ** CIRCE HAS CONVERGED";
        cout << endl << " ** iter[" << niter << "]" << " ** Likelihood[" << dLikelihood << "]" << endl;
        cout << " ***** Selected :: iter[" << nSelected << "] Likelihood["       << _dLikelihood << "]" << endl;
        cout << " ** matrix C1 " << endl;
        C1.Print();
        cout << endl << " ** vector XM1" << endl;
        vecXM1.Print();
        cout << " ** End Of Initial Matrix C [" << iCAlea + 1 << "/" << _nInitCAlea << "]" << endl;
        cout << " **********************************" << endl;

    } // End Of Loop iCAlea

    // Compute the residu
    TVectorD vecResiu(_nPatterns);
    Double_t dResiu = 0.0;
    Int_t ipat = 0;

    //    TCanvas *cdiff = new TCanvas("cdiffcirece_residu");
    // cdiff->Divide(_listOfTDS->GetEntries());
    Int_t itds = 1;

    URANIE::DataServer::TDataServer *tdscrt;
    TIter nextTDS(_listOfTDS);
    while ((tdscrt = (URANIE::DataServer::TDataServer*) nextTDS()))
    {
        if (_blog)
            cout << " ** TDS name[" << tdscrt->GetName()
                    << "] -- Compute the residu" << endl;
        //TLeaf * tldiff = tdscrt->getTuple()->GetLeaf(CIRCE_DIFF_ATTRIBUTE);
        TBranch* branchAttribut = tdscrt->getTuple()->Branch(
                Form("%s", CIRCE_RESIDU_ATTRIBUTE), &dResiu,
                Form("%s/D", CIRCE_RESIDU_ATTRIBUTE));

        for (Int_t i = 0; i < tdscrt->getNPatterns(); i++)
        {
            tdscrt->getTuple()->GetEntry(i);
            dY = _vecDiff(ipat);
            // Compute XDEN
            TVectorD vecS = TMatrixDColumn_const(_matSensibility, ipat);
            //       if ( _blog ) vecS.Print();
            TVectorD vecXDEN = _matC * vecS;
            //       if ( _blog ) vecXDEN.Print();

            dXDEN = vecS * vecXDEN;
            dA3 = vecS * _vecBiais;

            dResiu = (dY - dA3) / TMath::Sqrt(dXDEN);
            vecResiu(ipat) = dResiu;
            if (_blog)
                cout << " ** Residu ipat[" << ipat << "/" << npat << "] Residu["
                        << dResiu << "]" << endl;
            branchAttribut->Fill();

            ipat++;
        }
	_tds->getTuple()->ResetBranchAddress(branchAttribut);
        //cdiff->cd(itds++);
        //tdscrt->draw(CIRCE_RESIDU_ATTRIBUTE);

    }

    // Mean and Standard-Deviation for the residu
    Double_t dmean = TMath::Mean(_nPatterns, vecResiu.GetMatrixArray());
    Double_t dsigma = TMath::RMS(_nPatterns, vecResiu.GetMatrixArray());
    cout << " ** Residual :: Mean [" << dmean << "] Std[" << dsigma << "]" << endl;

    // Compute the correlation matrix
    for (Int_t i = 0; i < nparam; i++) {
        _matCorrelation(i, i) = TMath::Sqrt(_matC(i, i));
    }
    for (Int_t i = 0; i < nparam; i++) {
        for (Int_t j = 0; j < i; j++) {
            _matCorrelation(i, j) = _matC(i, j) / (_matC(i, i) * _matC(j, j));
        }
    }

    if ( _blog ) {
        cout << " ** matCorrelation  rows[" << _matCorrelation.GetNrows()
             << "] col [" << _matCorrelation.GetNcols() << "]" << endl;
        _matCorrelation.Print();
    }

    // terminate the procedure
    terminate(option);
}

////////////////////////////////////////////////////////////
//  terminate method
////////////////////////////////////////////////////////////
void TCirce::terminate(Option_t * option)
{
    if ( _blog )
        cout << endl << " ** TCirce::terminate option[" << option << "]" << endl;

    if ( _blog )
        cout << " ** TCirce :: Compute Confidence Interval by the Fisher Matrix" << endl;

    TVectorD vecAlpha(_nPatterns);
    vecAlpha.Zero();

    for(Int_t k=0; k<_nPatterns; k++) {
        for(Int_t i=0; i<_nParameters; i++) {
            vecAlpha(k) += _matSensibility(i, k) * _matSensibility(i, k) * _matC(i, i) + _vecYStarVariance(k);
        }
    }

    // Allocate the matrix and init to Zero
    _matVarianceMu.ResizeTo(_nParameters, _nParameters);
    _matVarianceMu.Zero();

    _matVarianceSigma.ResizeTo(_nParameters, _nParameters);
    _matVarianceSigma.Zero();

    // Loop
    Double_t dValCrt;
    for(Int_t i=0; i<_nParameters; i++) {
        for(Int_t j=0; j<_nParameters; j++) {
            for(Int_t k=0; k<_nPatterns; k++) {
                dValCrt = _matSensibility(i, k) * _matSensibility(j, k) / vecAlpha(k);
                _matVarianceMu(i, j) += dValCrt;
                _matVarianceSigma(i,j) += 0.5 * dValCrt * dValCrt;
            }
        }
    }

    // Compute the Inverse
    Double_t dDeterminant;
    _matVarianceMu.Invert(&dDeterminant);
    if ( _blog ) {
        cout << " ** _matVarianceMu Inverse Matrix with Determinant [" << dDeterminant << "]" << endl;
        _matVarianceMu.Print();
    }

    _matVarianceSigma.Invert(&dDeterminant);
    if ( _blog ) {
        cout << " ** _matVarianceSigma Inverse Matrix with Determinant [" << dDeterminant << "]" << endl;
        _matVarianceSigma.Print();
    }

    if ( _blog )
        cout << " ** End Of TCirce::terminate option[" << option << "]" << endl;
}

////////////////////////////////////////////////////////////
//  set the Seed
////////////////////////////////////////////////////////////
void TCirce::setSeed(Int_t ind)
{
    if (ind == 0)
    {
        time_t secondes;
        time(&secondes);
        _nSeed = (Int_t)(secondes);
    }
    else
    {
        _nSeed = ind;
    }
    gRandom->SetSeed(_nSeed);
}


////////////////////////////////////////////////////////////
//  print the results
////////////////////////////////////////////////////////////
void TCirce::printResults(Option_t *option)
{
    cout << " ***************************************" << endl;
    cout << " ** TCirce::printResults[" << option << "]" << endl;
    cout << " ** _sCirceInputFile[" << _sCirceInputFile << "]" << endl;
    cout << " ** _nSeed[" << _nSeed << "]" << endl;
    cout << " ** _nPatterns[" << _nPatterns << "] _nParameters[" << _nParameters << "]" << endl;
    cout << " ** _dTol[" << _dTol << "]" << endl;
    cout << " ** _dLikelihood[" << _dLikelihood << "]" << endl;
    cout << " ** DEFINE CIRCE_DIFF_ATTRIBUTE[" << CIRCE_DIFF_ATTRIBUTE << "]" << endl;


    switch (_nCirceMethod)
    {
    case kBiais:
    {
        cout << " ** CirceMEthod[kBiais]" << endl;
        break;
    }
    case kBiaisBloc:
    {
        cout << " ** CirceMEthod[kBiaisBloc]" << endl;
        break;
    }
    case kBiaisBlocCInitAlea:
    {
        cout << " ** CirceMEthod[kBiaisBlocCInitAlea]" << endl;
        break;
    }
    case kBiaisBlocBootstrap:
    {
        cout << " ** CirceMEthod[kBiaisBlocBootstrap]" << endl;
        break;
    }
    default:
    {
        cout << " ** CirceMEthod[kUnknown]" << endl;
    }
    }

    cout << endl;
    cout << " Complete Analysis"<<endl;
    cout << endl;
    cout << " ***** Resultats for the [" << _nParameters << "] parameters ***" << endl;
    Double_t dMeanBias, dVarBias;
    Double_t dMeanSigma2, dVarSigma2;

    for(Int_t iparam=0; iparam<_nParameters; iparam++) {
        cout << "  **** Parameter[" << iparam << "/" << _nParameters << "] Name[" << _listOfSensitivity->At(iparam)->GetName() << "]" << endl;
        cout << "    **  Bias for parameter [" << _listOfSensitivity->At(iparam)->GetName() << "]" << endl;
        dMeanBias = _vecBiais(iparam);
        dVarBias  = _matVarianceMu(iparam,iparam);
        cout << "         Mean = " << dMeanBias << " , Variance = " << dVarBias << " , Standard-Deviation =" << sqrt(dVarBias) << endl;
        cout << "         95% Confidence Interval = [" << dMeanBias-1.96* sqrt(dVarBias) << " , " << dMeanBias+1.96* sqrt(dVarBias) << "]" << endl;

        cout << "    **  Variance for parameter [" << _listOfSensitivity->At(iparam)->GetName() << "]" << endl;
        dMeanSigma2 = _matC(iparam,iparam);
        dVarSigma2 = _matVarianceSigma(iparam,iparam);
        cout << "         Mean      = " <<  dMeanSigma2 << " , Variance = " << dVarSigma2 << " , Standard-Deviation = " << sqrt(dVarSigma2) << endl;
        cout << "         95% Confidence Interval = [" << dMeanSigma2 - 1.96 * sqrt(dVarSigma2) << " , " << dMeanSigma2 + 1.96 * sqrt(dVarSigma2) << "]" << endl;

        cout << "    **  Standard-Deviation = sqrt(Variance) for parameter [" << _listOfSensitivity->At(iparam)->GetName() << "]" << endl;
        cout << "         Mean    = " << sqrt(dMeanSigma2) << endl;
        cout << "         95% Confidence Interval = [ " << sqrt(dMeanSigma2 - 1.96 * sqrt(dVarSigma2)) << " , " << sqrt(dMeanSigma2 + 1.96 * sqrt(dVarSigma2)) << "]" << endl;
        if ( iparam < _nParameters-1)
            cout << "  ****" << endl;
    }
    cout << " **** End Of Resultats for the [" << _nParameters << "] parameters ***" << endl;
}


////////////////////////////////////////////////////////////
//  print the Log
////////////////////////////////////////////////////////////
void TCirce::printLog(Option_t *option)
{
    cout << " ***************************************" << endl;
    cout << " ** TCirce::printLog[" << option << "]" << endl;
    cout << " ** _sCirceInputFile[" << _sCirceInputFile << "]" << endl;
    cout << " ** _nSeed[" << _nSeed << "]" << endl;
    cout << " ** _nPatterns[" << _nPatterns << "] _nParameters[" << _nParameters
            << "]" << endl;
    cout << " ** _dTol[" << _dTol << "]" << endl;
    cout << " ** _dLikelihood[" << _dLikelihood << "]" << endl;
    cout << " ** DEFINE CIRCE_DIFF_ATTRIBUTE[" << CIRCE_DIFF_ATTRIBUTE << "]"
            << endl;
    if (_listOfBloc)
    {
        cout << " ** List Of bloc [" << _listOfBloc->GetEntries() << "]"
                << endl;
        _listOfBloc->Print();
    }
    if (_listOfSensitivity)
    {
        cout << " ** List Of Sensitivity Parameter ["
                << _listOfSensitivity->GetEntries() << "]" << endl;
        _listOfSensitivity->Print();
    }

    cout << " ** vector _vecDiff Size[" << _vecDiff.GetNrows() << "]" << endl;
    _vecDiff.Print();

    cout << " ** vector Experimental Uncertainty YStarSize["
            << _vecYStarVariance.GetNrows() << "]" << endl;
    _vecYStarVariance.Print();

    cout << " ** matrix _matSensibility Size[" << _matSensibility.GetNrows()
            << "," << _matSensibility.GetNcols() << "]" << endl;
    _matSensibility.Print();

    switch (_nCirceMethod)
    {
    case kBiais:
    {
        cout << " ** CirceMEthod[kBiais]" << endl;
        break;
    }
    case kBiaisBloc:
    {
        cout << " ** CirceMEthod[kBiaisBloc]" << endl;
        break;
    }
    case kBiaisBlocCInitAlea:
    {
        cout << " ** CirceMEthod[kBiaisBlocCInitAlea]" << endl;
        break;
    }
    case kBiaisBlocBootstrap:
    {
        cout << " ** CirceMEthod[kBiaisBlocBootstrap]" << endl;
        break;
    }
    default:
    {
        cout << " ** CirceMEthod[kUnknown]" << endl;
    }
    }
    TUncertModeler::printLog();
}

} // End Of namespace UncertModeler
} // End Of namespace URANIE

//  End Of File TCirce.cxx
