//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 @file   TTestAndersonDarling.cxx
 @author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 @date   Tue Jan  6 11:56:31 2009

 @brief
 \par Modifications
 \li Cedric Forgues (C.S.) 2010/01/22
 Change the prototype TMath:Sort() to adapt to ROOT 5.22.00f
 */

#include <iomanip>

// ROOT
#include "TF1.h"
#include "TH1.h"
#include "TLeaf.h"
#include "TPad.h"
#include "TMath.h"
#include "TCanvas.h"
#include "TRegexp.h"

// Uranie
#include "TTestAndersonDarling.h"
#include "UExceptions.h"
#include "TStochAttribut.h"

ClassImp (URANIE::UncertModeler::TTestAndersonDarling)

namespace URANIE
{
namespace UncertModeler
{

// -----------------------------------------------------------------------
//  Constructors
// -----------------------------------------------------------------------
TTestAndersonDarling::TTestAndersonDarling(URANIE::DataServer::TDataServer *tds,
        const char *varexpinput, Option_t * option) :
        TTester(tds, TTester::kAndersonDarling, varexpinput, option),
        _dEpsilon(1.0e-08)
{
}

// -----------------------------------------------------------------------
//  Destructor
// -----------------------------------------------------------------------
TTestAndersonDarling::~TTestAndersonDarling()
{
    if (_blog)
        cout << " ** Destroy TTestAndersonDarling Test" << endl;
}

// -----------------------------------------------------------------------
//  computeTheCurrentScore
// -----------------------------------------------------------------------
Double_t TTestAndersonDarling::computeTheCurrentScore(Int_t i, Double_t dzi)
{
    Double_t di = 1.0 * i;
    Double_t dn = 1.0 * _nS;

    Double_t dvalcrt = 0.0;

    if ( ( dzi > _dEpsilon ) && ( 1.0 - dzi > _dEpsilon ) )
        dvalcrt = (2.0 * di - 1.0) * TMath::Log(dzi) + (2.0 * dn + 1.0 - 2.0 * di) * TMath::Log(1.0 - dzi);

    if (_blog)
    {
        cout << endl << " ** AD :: di[" << di << "] zi[" << std::setw(1) << dzi << "]";
        cout << " ** dvalcrt[" << dvalcrt << "]";
    }

    return (dvalcrt);
}

// -----------------------------------------------------------------------
//  computeModifiedScore
// -----------------------------------------------------------------------
void TTestAndersonDarling::computeModifiedScore(
        URANIE::DataServer::TStochasticAttribute::ELawType tlaw)
{
    Double_t d1onN = 1.0 / (1.0 * _nS);
    Double_t d1onSqrtN = 1.0 / TMath::Sqrt(1.0 * _nS);

    try
    {
        switch (_nCase)
        {
        case kCase0:    // Case 0
        {
            if (_nS < 5)
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                        Form("TTestAndersonDarling::computeModifiedScore   ERROR \n The number of patterns [%lld] is lesser than 5", _nS));
            _dModifiedScore = _dScore;
        }
            break;
        case kCase3:    // Case 3
        {
            // Set the Modified AndersonDarling value of the statistic
            switch (tlaw)
            {
            case URANIE::DataServer::TStochasticAttribute::kUniform:
            {
                _dModifiedScore = _dScore;               
            }
            break;
            case URANIE::DataServer::TStochasticAttribute::kNormal:
            {
                _dModifiedScore = _dScore * (1.0 + 0.75 * d1onN + 2.25 * d1onN * d1onN);
            }
            break;
            case URANIE::DataServer::TStochasticAttribute::kLogNormal:
            {
                _dModifiedScore = _dScore * (1.0 + 0.75 * d1onN + 2.25 * d1onN * d1onN);
            }
            break;
            case URANIE::DataServer::TStochasticAttribute::kExponential:
            {
                _dModifiedScore = _dScore * (1.0 + 5.4 * d1onN - 11.0 * d1onN * d1onN);
            }
            break;
            case URANIE::DataServer::TStochasticAttribute::kWeibull:
            {
                _dModifiedScore = _dScore * (1.0 + 0.2 * d1onSqrtN);
            }
            break;
            case URANIE::DataServer::TStochasticAttribute::kGamma:
            {
                _dModifiedScore = _dScore;
            }
            break;
            default:
            {
                try
                {
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                               Form("TTestAndersonDarling::computeModifiedScore ** Law not treated [%d] **", tlaw ));
                } catch (URANIE::Exceptions::UErrorExceptions& ue)
                {
                    ue.printMessage();
                    //throw ue;
                }
            }
            break;
            }
        }
            break;
       default:
           break;
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        exit(-1);
    }
}

} // Fin du namespace UncertModeler
} // Fin du namespace URANIE

