//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTTestAndersonDarling.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   $Date$

 \brief  Test File of the class URANIE::DataServer::TTestAndersonDarling

 Created: wen nov 21 07:59:24 2007
 \par Modifications
 \li M. Salvoldi 18/01/2012
 - add 8 news unit tests
 */

#define DBL_PRECISION_COMPARE 1.e-12
#define DBL_PRECISION_COMPARE2 1.e-2
#include <iostream>
using namespace std;

//ROOT
#include "TString.h"

// URANIE
#include "TestTTestAndersonDarling.h"
#include "TUniformDistribution.h"
#include "TSampling.h"

TTestAndersonDarlingTest::TTestAndersonDarlingTest() :
        _tdsBliss(NULL), _tad(NULL)
{
}

TTestAndersonDarlingTest::~TTestAndersonDarlingTest()
{
}

void TTestAndersonDarlingTest::setUp()
{
    TString svar = "w";
    _tdsBliss = new URANIE::DataServer::TDataServer();
    _tdsBliss->fileDataRead("bliss.dat");
    _tdsBliss->computeStatistic(svar);

    _tdsBrisbane = new URANIE::DataServer::TDataServer();
    _tdsBrisbane->fileDataRead("brisbane.dat");
    svar = "w";
    _tdsBrisbane->computeStatistic(svar);

    _tdsForearms = new URANIE::DataServer::TDataServer();
    _tdsForearms->fileDataRead("forearms.dat");
    svar = "w";
    _tdsForearms->computeStatistic(svar);
}

void TTestAndersonDarlingTest::tearDown()
{
    delete _tad;
    _tad = NULL;
    delete _tdsBliss;
    _tdsBliss = NULL;
    delete _tdsBrisbane;
    _tdsBrisbane = NULL;
    delete _tdsForearms;
    _tdsForearms = NULL;
}

void TTestAndersonDarlingTest::testA2blissCase3()
{
    _tad = new URANIE::UncertModeler::TTestAndersonDarling(_tdsBliss, "w");
    _tad->unsetLog();
    Double_t dA2 = _tad->computeScore("normal");
    Double_t dA2m = _tad->getModifiedScore();

    //ORIGIN: Goodness-of-fit techniques D'Agostino et Stephens (page 125)
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.234, dA2, 1.e-3);  // 0.214 without MLE
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.244, dA2m, 1.e-3); // 0.223
}

void TTestAndersonDarlingTest::testA2BrisbaneCase3()
{
    CPPUNIT_ASSERT_EQUAL(44, _tdsBrisbane->getNPatterns());
    _tad = new URANIE::UncertModeler::TTestAndersonDarling(_tdsBrisbane, "w");
    _tad->unsetLog();
    Double_t dA2 = _tad->computeScore("normal");
    Double_t dA2m = _tad->getModifiedScore();
    Double_t dpvalue = _tad->getPValue();

    /// https://www.spcforexcel.com/knowledge/basic-statistics/anderson-darling-test-for-normality#twodatasets
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.717, dA2,  1.e-2);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.7481, dA2m, 1.0e-2);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.000179, dpvalue, 1.e-5);
}

void TTestAndersonDarlingTest::testA2ForearmsCase3()
{
    CPPUNIT_ASSERT_EQUAL(140, _tdsForearms->getNPatterns());
    _tad = new URANIE::UncertModeler::TTestAndersonDarling(_tdsForearms, "w");
    _tad->unsetLog();
    Double_t dA2 = _tad->computeScore("normal");
    Double_t dA2m = _tad->getModifiedScore();
    Double_t dpvalue = _tad->getPValue();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.237, dA2,  1.e-2);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.2452, dA2m, 1.0e-4); // 0.2383
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.760126, dpvalue, 1.e-5); // 0.782045
}

void TTestAndersonDarlingTest::testA2blissCase0()
{
    _tad = new URANIE::UncertModeler::TTestAndersonDarling(_tdsBliss, "w");
    _tad->unsetLog();
    Double_t dA2 = _tad->computeScore("normal(200,35)");
    Double_t dA2m = _tad->getModifiedScore();
    Double_t dpvalue = _tad->getPValue();

    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.019, dA2, 5.e-3);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1.019, dA2m, 5.e-3);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0.347, dpvalue, 2.e-3);
}

void TTestAndersonDarlingTest::testUniformValues()
{
    URANIE::DataServer::TDataServer *tds = new URANIE::DataServer::TDataServer();
    tds->addAttribute(new URANIE::DataServer::TUniformDistribution("u", -5.0, 2.0));
    tds->getAttribute("u")->delShare();
    URANIE::Sampler::TSampling *fsamp = new URANIE::Sampler::TSampling(tds, "lhs", 10000);
    fsamp->generateSample();

    _tad = new URANIE::UncertModeler::TTestAndersonDarling(tds, "u");
    _tad->unsetLog();

    Double_t dA2 = _tad->computeScore("uniform");
    Double_t dA2m = _tad->computeScore("uniform(-5.00000,2.00000)");
    //Double_t dA2ms = _tad->computeScore("uniform(-5.00000,-2.00000)");

    CPPUNIT_ASSERT_DOUBLES_EQUAL(dA2, 0.000000, DBL_PRECISION_COMPARE2);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(dA2m, 0.000000, DBL_PRECISION_COMPARE2);

    delete fsamp;
    delete tds;
    tds = NULL;
}

void TTestAndersonDarlingTest::throwExceptionTestAndersonDarlingConstructorTdsNullEmpty()
{

    CPPUNIT_ASSERT_THROW(
            new URANIE::UncertModeler::TTestAndersonDarling(NULL, "w"),
            URANIE::Exceptions::UErrorExceptions);

    URANIE::DataServer::TDataServer* tds = new URANIE::DataServer::TDataServer();
    CPPUNIT_ASSERT_THROW(
       new URANIE::UncertModeler::TTestAndersonDarling(tds, "w"),
       URANIE::Exceptions::UWarningExceptions);

    delete tds;
}

void TTestAndersonDarlingTest::throwExceptionTestAndersonDarlingConstructorMoreAttribut()
{
    _tdsBliss->addAttribute("E", "5.0*w*w");

    CPPUNIT_ASSERT_THROW(
            new URANIE::UncertModeler::TTestAndersonDarling(_tdsBliss, "w:E"),
            URANIE::Exceptions::UErrorExceptions);
}

void TTestAndersonDarlingTest::throwExceptionTestAndersonDarlingConstructorFalseAttribut()
{

    CPPUNIT_ASSERT_THROW(
            new URANIE::UncertModeler::TTestAndersonDarling(_tdsBliss, "toto"),
            URANIE::Exceptions::UErrorExceptions);
}

void TTestAndersonDarlingTest::throwExceptionTestAndersonDarlingcomputeScoreNoLaw()
{
    _tad = new URANIE::UncertModeler::TTestAndersonDarling(_tdsBliss, "w");
    _tad->unsetLog();

    CPPUNIT_ASSERT_THROW(_tad->computeScore(),
            URANIE::Exceptions::UWarningExceptions);
}

void TTestAndersonDarlingTest::throwExceptionTestAndersonDarlingcomputeScoreBadLaw()
{
    _tad = new URANIE::UncertModeler::TTestAndersonDarling(_tdsBliss, "w");
    _tad->unsetLog();
    CPPUNIT_ASSERT_THROW(_tad->computeScore("azerty"), URANIE::Exceptions::UWarningExceptions);
    CPPUNIT_ASSERT_THROW(_tad->computeScore("invgamma"), URANIE::Exceptions::UWarningExceptions);
}

void TTestAndersonDarlingTest::throwExceptionTestAndersonDarlingcomputeScoreNormalBadDefined()
{
    _tad = new URANIE::UncertModeler::TTestAndersonDarling(_tdsBliss, "w");
    _tad->unsetLog();

    CPPUNIT_ASSERT_THROW(_tad->computeScore("normal()"),
            URANIE::Exceptions::UErrorExceptions);
}

void TTestAndersonDarlingTest::throwExceptionTestAndersonDarlingcomputeScoreUniformBadDefined()
{
    _tad = new URANIE::UncertModeler::TTestAndersonDarling(_tdsBliss, "w");
    _tad->unsetLog();

    CPPUNIT_ASSERT_THROW(_tad->computeScore("uniform()"),
            URANIE::Exceptions::UErrorExceptions);
}

void TTestAndersonDarlingTest::throwExceptionTestAndersonDarlingcomputeScoreLogNormalBadDefined()
{
    _tad = new URANIE::UncertModeler::TTestAndersonDarling(_tdsBliss, "w");
    _tad->unsetLog();

    CPPUNIT_ASSERT_THROW(_tad->computeScore("lognormal()"),
            URANIE::Exceptions::UErrorExceptions);
}
