//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.4 $
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TBiMpiRun.cxx
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date    2011-08-23.10.19.06

 \brief   version mpi a deux niveaux de la distribution
 */

#include "TBiMpiRun.h"

#define HEADER 1
#define DATAOUT 2
#define DATAGRAD 4
#define SIZEOUT 8
#define SIZEIN 16
#define FIN 1024

ClassImp (URANIE::MpiRelauncher::TBiMpiRun);

namespace URANIE
{
namespace MpiRelauncher
{

MPI_Comm TBiMpiRun::__intra;
bool TBiMpiRun::__isBiMpi=false;

// class method
MPI_Comm TBiMpiRun::getCalculMpiComm()
{
    if (!__isBiMpi) throw NotInAMpiContext();
    return __intra;
}

long TBiMpiRun::getCalculMpiCommPy()
{
    if (!__isBiMpi) throw NotInAMpiContext();
    return (long) __intra;
}

bool TBiMpiRun::isBiMpi()
{
    return __isBiMpi;
}

// constructeur
TBiMpiRun::TBiMpiRun(URANIE::Relauncher::TEval *func, int sz, Bool_t init) :
        TMpiRun(func, init)
{
    calculSize(sz);
}

TBiMpiRun::~TBiMpiRun(void)
{
    //stopSlave();
}

void TBiMpiRun::calculSize(int siz)
{
    int id;

    if (siz < 2 || _ncpus % siz != 1)
        throw UncorrectSize();

    _lowsz = siz;
    _lowid = _id % siz;
    _topsz = _ncpus / siz + 1;
    _topid = (_lowid) ? MPI_UNDEFINED : (_ncpus - _id - 1) / siz;
    // creation des communicators
    // low level
    MPI_Comm_split(MPI_COMM_WORLD, _id / siz, _lowid, &__intra);
    __isBiMpi = true;
    // hight level
    id = (_lowid) ? MPI_UNDEFINED : 0;
    MPI_Comm_split(MPI_COMM_WORLD, id, _topid, &_comm);
    _ncpus = _topsz;
}

// utilitaires
void TBiMpiRun::forward_header(int *bufi, int size)
{
    MPI_Bcast(bufi, size, MPI_INT, 0, __intra);
}

void TBiMpiRun::forward_value(double *buff, int size)
{
    MPI_Bcast(buff, size, MPI_DOUBLE, 0, __intra);
}

// public
void TBiMpiRun::leader_loop(int grp)
{
    vector<int> headbuf(_sizein + 2);
    int ret, si, sd, next = 1, stop = 0;
    int i, size;
    double *inbuf;

    URANIE::Relauncher::TEval *mpifun = _fun->rebuild(grp);
    MPI_Status status;

    TIter nexti(_fun->getInputList());
    std::vector<URANIE::DataServer::UEntry*> invec;
    empty_uentry(nexti, invec);
    TIter nexto(_fun->getOutputList());
    std::vector<URANIE::DataServer::UEntry*> outvec;
    empty_uentry(nexto, outvec);
    
    si = receive_header(0, headbuf.data(), _sizein+2);
    while (headbuf[1] != 0) {
        forward_header(headbuf.data(), _sizein+2);
        // read input
        size = 0;
        for (i=0; i<_sizein; i++) size += headbuf[i+2];
        inbuf = new double[size];
        sd = receive_value(0, inbuf, size);
        forward_value(inbuf, size);
        // convert to uentry
        to_uentry(headbuf.data(), inbuf, invec);
        // eval
        ret = mpifun->eval(&invec, &outvec, headbuf[1]);
        // send output
        if (ret) {
            send_uentry(0, ret, _sizeout, &outvec);
        }
        else {
            int *nores = new int[_sizeout+2];
            for (i=0; i< _sizeout+2; i++) nores[i] = 0;
            sd = send_header(0, nores, _sizeout+2);
            delete[] nores;
        }
        // cleanup
        for (auto it : invec) it->clearEntry();
        for (auto it : outvec) it->clearEntry();
        delete[] inbuf;
        si = receive_header(0, headbuf.data(), _sizein+2);
        if (si == -1) {
            std::cout << "Separate Breaked\n";
            break;
        }
    }
    /* sortie */
    forward_header(headbuf.data(), _sizein+2);
    if (mpifun->unbuild()) delete mpifun;
}

void TBiMpiRun::calcul_loop(int grp)
{
    vector<int> headbuf(_sizein + 2);
    int i, size;
    double *inbuf;

    TIter nexti(_fun->getInputList());
    std::vector<URANIE::DataServer::UEntry*> invec;
    empty_uentry(nexti, invec);
    TIter nexto(_fun->getOutputList());
    std::vector<URANIE::DataServer::UEntry*> outvec;
    empty_uentry(nexto, outvec);

    URANIE::Relauncher::TEval *mpifun = _fun->rebuild(grp, kFALSE);

    forward_header(headbuf.data(), _sizein+2);
    while (headbuf[1] != 0)
    {
        // read input
        size = 0;
        for (i=0; i<_sizein; i++) size += headbuf[i+2];
        inbuf = new double[size];
        forward_value(inbuf, size);
        // convert to uentry
        to_uentry(headbuf.data(), inbuf, invec);
        // eval
        mpifun->eval(&invec, &outvec, headbuf[1]);
        // cleanup
        for (auto it : invec) it->clearEntry();
        for (auto it : outvec) it->clearEntry();
        delete[] inbuf;
        // loop
        forward_header(headbuf.data(), _sizein+2);
    }
}

void TBiMpiRun::startSlave(void)
{
    int i;
    _sizein = _fun->getInputList()->GetEntries();
    _sizeout = _fun->getOutputList()->GetEntries();

    if (_topid == 0)
    { //optim
        for (i = 1; i < _ncpus; i++)
            _sleep.push(i);
    }
    else if (_lowid == 0)
    { //leader
        leader_loop(_topid);
    }
    else
    { // calcul
        calcul_loop(_topid);
    }
    
}

Bool_t TBiMpiRun::onMaster(void)
{
    return (_topid == 0) ? kTRUE : kFALSE;
}

void TBiMpiRun::stopSlave(void)
{
    if (!_sizein)
        return;
    //if (!_id)
    //    _id = _topid;
    //else if (!_topid)
    //    _id = 0;
    TMpiRun::stopSlave();
}

} } //namespace
