//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.7 $
// $State$
//////////////////////////////////////////////////////////////
/*!
  \file   TCodeEval.cxx
  \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
  \date    2011-04-04.15.35.52

  \brief   encapsulation d'un TCode
*/

#include "TCodeEval.h"

#ifndef WIN32
#include <sys/wait.h>
#else
#include <io.h>
#endif

//ROOT
#include "TSystem.h"

//URANIE
#include "UExceptions.h"

ClassImp (URANIE::Relauncher::TCodeEval);

namespace URANIE
{
namespace Relauncher
{

/* Constructeur Destructeur */
TCodeBase::TCodeBase(const char *scmd) :
#ifdef WIN32
    _cmd(TString::Format("cd %%D && %s", scmd)),
#else
    _cmd(TString::Format("cd %%D ; %s", scmd)),
#endif	
    _initialdir(gSystem->pwd()),
    _localdir(""), _isnumered(kFALSE),
    _keepAllFolders(false), 
    _workdir(""), _tmpdir("")
{
    _scripts = new TList();
    _result = NULL;
    _origcmd=_cmd;
}

TCodeBase::~TCodeBase(void)
{
    if (!_tmpdir.Length())
        delete _scripts;
}

/* Complete Definition */
void TCodeBase::addInputFile(TCodeScript *f)
{
    // if (_freezed) throw FreezedEval();
    _scripts->Add((TObject *) f);
}

void TCodeBase::addOutputFile(TCodeResult *f)
{
    //if (_freezed) throw FreezedEval();
    if (_result)
        std::cout << "Multi output file: unimplemented feature\n";
    else
        _result = f;

}

void TCodeBase::setReference(TString dir)
{
    //if (_freezed) throw FreezedEval();
    _localdir = dir;
}

void TCodeBase::setWorkingDir(TString dir)
{
    //if (_freezed) throw FreezedEval();
    _workdir = dir;
}

void TCodeBase::setOldTmpDir()
{
    _isnumered = kTRUE;
}

void TCodeBase::keepAllFolders()
{
    _keepAllFolders = kTRUE;
}

/* Close definition */
void TCodeBase::buildDir(void)
{
    /* build file */
    if (_localdir.Length() == 0)
        _localdir = _initialdir;

    /* working directory */
    if (_workdir.Length() == 0)
        _workdir = _initialdir;

#ifdef IMPLICIT
    void *dir = NULL;
    // bug mpi conflict access
    try
    {
        // Create this directory if it doesn't exist
        if (!(dir = gSystem->OpenDirectory(_workdir)))
	    if (gSystem->mkdir(_workdir, kTRUE) == -1)
		throw URANIE::Exceptions::UExceptions(__FILE__, __LINE__, Form("TCodeBase::buildDir ERROR : the directory \"%s\" can not be created", _workdir.Data()));
    }
    catch (URANIE::Exceptions::UExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    if (dir) gSystem->FreeDirectory(dir);
#endif

}

/* execution environment */
void TCodeBase::setTmpDir(int id)
{
    void *dir = NULL;

    if (_isnumered || _keepAllFolders) {
        try {
#ifdef WIN32
            _tmpdir = TString::Format("%s\\URANIE%d", _workdir.Data(), id);
#else		
            _tmpdir = TString::Format("%s/URANIE%d", _workdir.Data(), id);
#endif
            // Create this directory if it doesn't exist
            if (!(dir = gSystem->OpenDirectory(_tmpdir))
                && (gSystem->mkdir(_tmpdir.Data(), kFALSE) == -1)
                ){
                throw URANIE::Exceptions::UExceptions(__FILE__, __LINE__,
						      Form("TCodeBase::setTmpDir ERROR : the directory \"%s\" can not be created", _tmpdir.Data())
                    );
            }
        }
        catch (URANIE::Exceptions::UExceptions& ue) {
            ue.printMessage();
            throw ue;
        }
        if (dir) gSystem->FreeDirectory(dir);
    }
    else {
        try {
            char* tmp = new char[_workdir.Length()+12];
            // tmpdir
            strcpy(tmp, _workdir.Data());
            
#ifndef WIN32	
            strcat(tmp, "/URA_XXXXXX" );
            if (!mkdtemp(tmp)) {
                throw URANIE::Exceptions::UExceptions(__FILE__, __LINE__, 
						      Form("TCodeBase::setTmpDir ERROR : the directory \"%s\" can not be created", tmp)
                    );
            }
#else	
            strcat(tmp, "\\URA_XXXXXX" );
            if (_mktemp_s(tmp,_workdir.Length()+12) || (gSystem->mkdir(tmp, kFALSE) == -1) ) {	
               	
		throw URANIE::Exceptions::UExceptions(__FILE__, __LINE__, 
						      Form("TCodeBase::setTmpDir ERROR : the directory \"%s\" can not be created", tmp)
                    );
	
            }
#endif	 		
            _tmpdir = TString(tmp);
            delete[] tmp;
        }
        catch (URANIE::Exceptions::UExceptions& ue) {
            ue.printMessage();
            throw ue;
        }
        
    }

    /* comande */
    if( _keepAllFolders )
        _cmd = _origcmd;

    _cmd.ReplaceAll("%D", _tmpdir);

}

/* execution environment */
int TCodeBase::putAndRun(Double_t *in, int ncalc)
{
    int ret;
    TCodeScript *fs;
    TIter next(_scripts);

    if( _keepAllFolders )
        setTmpDir( ncalc );

    /* input file */
    while ((fs = (TCodeScript *) next()))
        fs->put(_localdir, _tmpdir, in);

    if(_keepAllFolders!=true)
        _result->clearFile(_tmpdir);

    /* execute */	
    // ret = gSystem->Exec(_cmd);
    ret = system(_cmd.Data());

#ifndef WIN32	
     if (WEXITSTATUS(ret))
        cout  << "Command " << _cmd.Data() << " has returned non-zero exit code ("<<WEXITSTATUS(ret)<<").\n  If any different from 127 (usually for unknown command) and 139 (usually for SIGSEV), the exit code meaning is \"command\" dependent."<<endl;
#endif
    return ret;
}

int TCodeBase::putAndRun(vector<URANIE::DataServer::UEntry*> *in, int ncalc)
{
    int ret;
    TCodeScript *fs;
    TIter next(_scripts);
    
    if( _keepAllFolders )
        setTmpDir( ncalc );

    /* input file */
    while ((fs = (TCodeScript *) next()))
	fs->put(_localdir, _tmpdir, in);

    if(_keepAllFolders!=true)
        _result->clearFile(_tmpdir);

    /* execute */
    // ret = gSystem->Exec(_cmd);
    ret = system(_cmd.Data());
      
#ifndef WIN32	
     if (WEXITSTATUS(ret))
         cout  << "Command " << _cmd.Data() << " has returned non-zero exit code ("<<WEXITSTATUS(ret)<<").\n  If any different from 127 (usually for unknown command) and 139 (usually for SIGSEV), the exit code meaning is \"command\" dependent."<<endl;
    // cout<<"ret is "<<ret<<" => "<<"WIFEXITED(ret)::"<<WIFEXITED(ret)<<"; "<<"WEXITSTATUS(ret)::"<<WEXITSTATUS(ret)<<"; "
    //     <<"WIFSIGNALED(ret)::"<<WIFSIGNALED(ret)<<"; "<<"WTERMSIG(ret)::"<<WTERMSIG(ret)<<"; "<<"WCOREDUMP(ret)::"<<WCOREDUMP(ret)<<"; "
    //     <<"WIFSTOPPED(ret)::"<<WIFSTOPPED(ret)<<"; "<<"WSTOPSIG(ret)::"<<WSTOPSIG(ret)<<"; "<<endl;
#endif
    return ret;
}

// TCodeEval
TCodeEval::TCodeEval(const char *scmd) :
    TShareEval(),
    TCodeBase(scmd)
{
}

TCodeEval::~TCodeEval(void)
{
}

void TCodeEval::freeze(void)
{
    TCodeScript *fs;
    TIter nexts(_scripts);

    if (_freezed)
        return;
    TBaseEval::freeze();

    buildDir();
    /* build file */
    while ((fs = (TCodeScript *) nexts()))
        fs->build(_localdir, &_input);

    try
    {
        if(_result==NULL)
            throw URANIE::Exceptions::UExceptions(__FILE__, __LINE__, "TCodeEval::freeze ERROR : there is no output file defined in this code");
    }
    catch (URANIE::Exceptions::UExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    _result->build(&_output);
}

TEval * TCodeEval::rebuild(int id, Bool_t chef)
{
    if (chef)
    {
        TCodeEval *ret = new TCodeEval(*this);
        if( ret->_keepAllFolders == false)
            ret->setTmpDir(id);
        return (TEval *) ret;
    }
    else
        return (TEval *) this;
}

Bool_t TCodeEval::unbuild()
{
    return kTRUE;
    //return (!_tmpdir.IsNull())? kTRUE : kFALSE ;
    //if (!_tmpdir.EqualTo("")) { printf("TCodeBase unbuild\n"); delete this;}
}

/* TRun use */
int TCodeEval::eval(vector<URANIE::DataServer::UEntry*> *in, vector<URANIE::DataServer::UEntry*> *out, int ncalc)
{
    putAndRun(in, ncalc);

    /* output file */
    int ret = _result->get(_tmpdir, out);
    return ((ret!=0) ? getOutDimension(out) : ret);
}
}
} //namespace

