//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.3 $
// $State$
//////////////////////////////////////////////////////////////
/*!
  \file   TFlatResult.cxx
  \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
  \date    2011-09-09.14.51.09

  \brief   fichier de résultat du TCodeEval
*/

#include "TFlatResult.h"

#include <iostream>
#include <fstream>

//ROOT
//URANIE

ClassImp (URANIE::Relauncher::TFlatResult);

namespace URANIE
{
namespace Relauncher
{

TRegexp TFlatResult::_num("[-+0-9.eE]+");

/* Constructeur destructeur */
TFlatResult::TFlatResult(const char *path) :
    _path(path)
{
    _svecbeg=TString(""); _svecend=TString(""); _svecdel=TString(",");
    _sstrbeg=TString("");  _sstrend=TString(""); _ssepar=TString("");
    _bMultiLine=false;
    _bWarningDone=false;
}

TFlatResult::~TFlatResult()
{
}

void TFlatResult::setVectorProperties( TString beg, TString delim, TString end)
{
    _svecbeg=beg; _svecend=end; _svecdel=delim;
}

void TFlatResult::setStringProperties( TString beg, TString end)
{
    _sstrbeg=beg; _sstrend=end; 
}

  
/* definition */
void TFlatResult::addOutput(URANIE::DataServer::TAttribute *att)
{
    _vars.Add(att);
}

void TFlatResult::setOutputs(int size, URANIE::DataServer::TAttribute *att, ...)
{
    int i;
    va_list ap;
    URANIE::DataServer::TAttribute *attr;

    _vars.Add(att);
    va_start(ap, att);
    for (i = 1; i < size; i++)
    {
        attr = va_arg(ap, URANIE::DataServer::TAttribute *);
        _vars.Add(attr);
    }
    va_end(ap);
}

/* usage */
void TFlatResult::build(TObjArray *out)
{
    TIter next(&_vars);
    URANIE::DataServer::TAttribute *tmp;

    while ((tmp = (URANIE::DataServer::TAttribute *) next()))
        out->Add(tmp);
}

int TFlatResult::get(TString &work, Double_t*res)
{
    int i, deb, len, dec;
    TObject *tmp;
    TIter next(&_vars);
    TString line;
    char* path = new char[work.Length() + _path.Length() + 2];
#ifdef WIN32
    sprintf(path, "%s\\%s", work.Data(), _path.Data());
#else
    sprintf(path, "%s/%s", work.Data(), _path.Data());
#endif	
    std::ifstream fin(path, std::ios::in);
    if (!fin)
    {
        std::cout << "Cannot open :: " << path << std::endl;
        return 0;
    }

    while (line.ReadLine(fin))
    {
        if (line[0] != '#')
            break;
    }
    fin.close();

    deb = i = 0;
    while ((tmp = next()))
    {
        dec = line.Index(_num, &len, deb);
        if (dec == -1)
            return 0;
        TString val(line(_num, deb));
        res[i++] = val.Atof();
        deb = dec + len;
    }

    delete[] path;
    int ret= ((i == _vars.GetEntries()) ? i : 0);
    return ret;
}


int TFlatResult::get(TString &work, vector<URANIE::DataServer::UEntry*> *res)
{
    unsigned int i, isepar;
  
    TIter next(&_vars);
    TString line;
    char* path = new char[work.Length() + _path.Length() + 2];
#ifdef WIN32
    sprintf(path, "%s\\%s", work.Data(), _path.Data());
#else
    sprintf(path, "%s/%s", work.Data(), _path.Data());
#endif	
    ifstream fin(path, std::ios::in);
    if (!fin)
    {
	std::cout << "Cannot open :: " << path << std::endl;
        delete[] path;
	return 0;
    }

    int VarFound=0;
    int ret=0;
    bool FirstLineRead=false;
    if( _bMultiLine ) //If the flat results is shown as multiline
    {
	while (line.ReadLine(fin))
	{
	    if (line[0] == '#')
		continue;

//	    XYZ<<"line is "<<line<<endl;
	    i = 0; isepar=0; string test;
	    istringstream pouet(line.Data());
	    while ( pouet >> test )
	    {
		if( test == _ssepar )
		{
		    isepar++;
		    if(isepar>i) i++;
		    continue;
		}
		if( i >= res->size() )
		{
		    if( !_bWarningDone )
		    {
			cout << "[95m >> WARNING in TFlatResult: "<<endl;
			cout << " >> The number of field in the output file is bigger than the number of output attribute specified ("<< res->size() <<")."<<endl;
			cout << " >> Only the first "<<res->size()<<" columns are read, so check that this is what you expect. "<<endl;
			cout << " << WARNING [0m"<<endl;
			_bWarningDone=true;
		    }
		    break;
		}
		TString val( test );
		//XYZ<<" push "<<val<<" in "<<res->at(i)->name<<",  "<<i<<" component of output vector"<<endl;
		res->at(i)->pushEntry( val.ReplaceAll(_ssepar,""), _svecdel );
		i++;
	    }
            if( !FirstLineRead )
            {
                ret=(((int)i<_vars.GetEntries()) ? 0 : _vars.GetEntries());
                if(ret==0)
                    std::cout << "TFlatResults("<<_path<<") with MultiLine mode: only "<<i<<" values found out of the "<< _vars.GetEntries()<<" expected \n";
                FirstLineRead=true;
            }
	}
        if( !FirstLineRead )
        {
            ret=0;
            std::cout << "TFlatResults("<<_path<<") with MultiLine mode: no lines were found, meaning at least "<< _vars.GetEntries()<<" values are missing\n";
        }
	fin.close();
    }
    else
    {
	while (line.ReadLine(fin))
	{
	    if (line[0] != '#')
		break;
	}
	fin.close();
		
	i = 0; string test;
	istringstream pouet(line.Data());
	while ( pouet >> test )
	{
	    if( i >= res->size() )
	    {
		if( !_bWarningDone )
		{
		    cout << "[95m >> WARNING in TFlatResult: "<<endl;
		    cout << " >> The number of field in the output file is bigger than the number of output attribute specified ("<< res->size() <<")."<<endl;
		    cout << " >> Only the first "<<res->size()<<" columns are read, so check that this is what you expect. "<<endl;
		    cout << " << WARNING [0m"<<endl;
		    _bWarningDone=true;
		}
		break;
	    }
	    TString val( test );
            bool isString = res->at(i)->tag==URANIE::DataServer::UEntry::String;
            //XYZ<<"Calling fillEntry ["<< val <<"]["<< ((isString) ? _ssepar : _svecdel) <<"]["<<((isString) ? _sstrbeg : _svecbeg) <<"]["<< ((isString) ? _sstrend :_svecend )<<"]"<<endl;
            res->at(i)->fillEntry( val, ((isString) ? _ssepar : _svecdel), ((isString) ? _sstrbeg : _svecbeg), ((isString) ? _sstrend :_svecend ));
            //XYZ<<test<<"  =>  "<<res->at(i)->dumpStr()<<endl;
            i++;
	}
        ret=(((int)i<_vars.GetEntries()) ? 0 : _vars.GetEntries());
        if(ret==0)
            std::cout << "TFlatResults("<<_path<<"): only "<<i<<" values found out of the "<< _vars.GetEntries()<<" expected \n";
    }
    delete[] path;
    return ret;
}

void TFlatResult::clearFile(TString tmpdir)
{
#ifdef WIN32
    TString fullPath = tmpdir+"\\"+_path;
    gSystem->Exec( Form("if exist %s del /f %s", fullPath.Data(), fullPath.Data()) ); 
#else
    TString fullPath = tmpdir+"/"+_path;
    gSystem->Exec( Form("rm -f %s", fullPath.Data()) );
#endif
}
  
}
}
