// --- C++ ---
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   TSocketRun.cxx
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2017-11-23
 
 \brief  independant process communicating by socket
*/


#include "TMpiRun.h"

#include "TBiMpiRun.h"
//#include <iostream>
//#include <queue>
//#include <vector>

//ROOT
#include "TROOT.h"
#include "TApplication.h"

#include "loop.h"

#define HEADER 1
#define VALUE 2

ClassImp (URANIE::MpiRelauncher::TMpiRun);

namespace URANIE
{
namespace MpiRelauncher
{

int TMpiRun::_count = 0;

TMpiRun::TMpiRun(URANIE::Relauncher::TEval *func, int *argc, char ***argv) :
        TSeparateRun(func)
{
    if (!_count)
    {
        if( gROOT->GetApplication() != 0 )
        {
            int rootargc = gROOT->GetApplication()->Argc();
            char **rootargv = gROOT->GetApplication()->Argv();
            MPI_Init(&rootargc, &rootargv);
        }
        else if( (argc != NULL) && (argv != NULL) )
        {
            MPI_Init(argc, argv);
        }
    }
    _count++;

    _comm = MPI_COMM_WORLD;
    MPI_Comm_rank(_comm, &_id);
    MPI_Comm_size(_comm, &_ncpus);

    _run.resize(_ncpus);
    _sizein = _sizeout = 0;
}

TMpiRun::TMpiRun(URANIE::Relauncher::TEval *func, Bool_t init) :
        TSeparateRun(func)
{
    if (!_count && init)
    {
        if( gROOT->GetApplication() != 0 )
        {
            int rootargc = gROOT->GetApplication()->Argc();
            char **rootargv = gROOT->GetApplication()->Argv();
            MPI_Init(&rootargc, &rootargv);
        }
        else 
        {
            int flag = -99;
            MPI_Initialized(&flag);
            try{
                if( flag==0 )
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                          "TMpiRun::constructor It seems that you're using this constructor out of a root session.\n\
This implies that you are probably compiling your code on your own but you haven't call MPI_Init. Two possibilities from there:\n\
==> Use the MPI_Init(&argc, &argv); before constructing the TMpiRun objects.\n\
==> Use the TMpiRun constructor with the signature TMpiRun(URANIE::Relauncher::TEval *func, int *argc, char ***argv) like this TMpiRun(eval, &argc, &argv);");
        
            } catch (URANIE::Exceptions::UExceptions& ue){
                ue.printMessage();
                throw ue;
            }
        }
    }
    _count++;

    _comm = MPI_COMM_WORLD;
    MPI_Comm_rank(_comm, &_id);
    MPI_Comm_size(_comm, &_ncpus);

    _run.resize(_ncpus);
    _sizein = _sizeout = 0;
}

TMpiRun::TMpiRun(URANIE::Relauncher::TEval *func, MPI_Comm comm) :
        TSeparateRun(func)
{
    _count++;

    _comm = comm;
    MPI_Comm_rank(_comm, &_id);
    MPI_Comm_size(_comm, &_ncpus);

    _run.resize(_ncpus);
    _sizein = _sizeout = 0;
}

TMpiRun::~TMpiRun()
{
#ifdef IMPLICIT
    stopSlave();
#endif

    _count--;
    if (!_count )
        MPI_Finalize();
}

// utilitaires
int TMpiRun::wait_service()
{
    int id;
    MPI_Status etat;

    MPI_Probe(MPI_ANY_SOURCE, HEADER, _comm, &etat);
    id = etat.MPI_SOURCE;
    return id;
}

int TMpiRun::receive_header(int canal, int *head, int size)
{
    MPI_Status etat;
    return MPI_Recv(head, size, MPI_INT, canal, HEADER, _comm, &etat);
}

int TMpiRun::receive_value(int canal, double *mesg, int size)
{
    MPI_Status etat;
    MPI_Recv(mesg, size, MPI_DOUBLE, canal, VALUE, _comm, &etat);
    return 0;
}

int TMpiRun::send_header(int canal, int *head, int size)
{
    MPI_Send(head, size, MPI_INT, canal, HEADER, _comm);
    return 0;
}

int TMpiRun::send_value(int canal, double *mesg, int size)
{
    MPI_Send(mesg, size, MPI_DOUBLE, canal, VALUE, _comm);
    return 0;
}


// protocole
void TMpiRun::startSlave()
{
    TSeparateRun::startSlave();
    
    if (_id == 0)
    { //optim
        int i;
        for (i = 1; i < _ncpus; i++)
            _sleep.push(i);
    }
    else
    {
        //calcul
        calcul_loop();
    }
}

Bool_t TMpiRun::onMaster()
{
    return (_id) ? kFALSE : kTRUE;
}

Bool_t TMpiRun::readyToGet()
{
    int flag;
    
    MPI_Iprobe(MPI_ANY_SOURCE, HEADER, _comm, &flag, MPI_STATUS_IGNORE);
    return (flag == true)? kTRUE : kFALSE ;
}

}} //namespace
