//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or any 
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision 1.3 $
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TPythonEval.cxx
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date    2012-02-21.15.56.51

 \brief   utilisation dune fonction python dans pyROOT
 */

#include <Python.h>

#include "TPythonEval.h"

ClassImp (URANIE::Relauncher::TPythonEval);

namespace URANIE
{
namespace Relauncher
{

TPythonEval::TPythonEval(PyObject *fun):
    TSimpleEval()
{
    _fun = fun;
}

TPythonEval::~TPythonEval()
{
}

int TPythonEval::eval(vector<URANIE::DataServer::UEntry*> *in, vector<URANIE::DataServer::UEntry*> *out, int ncalc)
{
    UNUSED(ncalc);
    PyGILState_STATE gstate;
    PyObject *pArgs, *pRetour, *pTmp=0;
    const char* ctmp=0;
    int i, sz, osz;

    /* acquire GIL */
    gstate = PyGILState_Ensure();
    /* argument */
    sz = _input.GetEntriesFast();
    pArgs = PyTuple_New(sz);

    //Conversion to python format
    for (i = 0; i < sz; i++)
    {
        switch( in->at(i)->tag)
        {
        case URANIE::DataServer::UEntry::Double :
            pTmp = PyFloat_FromDouble(*in->at(i)->val.dou);
            break;
        case URANIE::DataServer::UEntry::String :
            pTmp = PyUnicode_FromString((*in->at(i)->val.str).c_str());
            break;
        case URANIE::DataServer::UEntry::Vector :
            pTmp = PyList_New( in->at(i)->val.vec->size() );
            for(unsigned int ival=0; ival<in->at(i)->val.vec->size(); ival++)
                PyList_SetItem( pTmp, ival, PyFloat_FromDouble( in->at(i)->val.vec->at(ival) ) );
            break;
        default :
            break;
        }
        PyTuple_SetItem(pArgs, i, pTmp);
    }

    /* appel */
    pRetour = PyObject_CallObject(_fun, pArgs);
    Py_DECREF(pArgs);

    /* résultats */
    if (pRetour == NULL)
    {
        fprintf(stderr, "Call failed\n");
        PyGILState_Release(gstate);
        return 0;
    }
    if (!PyList_Check(pRetour))
    {
        fprintf(stderr, "No return\n");
        PyGILState_Release(gstate);
        return 0;
    }

    osz = _output.GetEntriesFast();
    //Conversion to python format
    for (i = 0; i < osz; i++)
    {
        switch( out->at(i)->tag)
        {
        case URANIE::DataServer::UEntry::Double :
            *out->at(i)->val.dou = PyFloat_AsDouble(PyList_GetItem(pRetour, i));
            break;
        case URANIE::DataServer::UEntry::String :
#ifdef WITH_PYTHON3
            ctmp = PyUnicode_AsUTF8(PyList_GetItem(pRetour, i));
#else
            ctmp = PyString_AsString(PyList_GetItem(pRetour, i));
#endif
            if (ctmp == NULL) std::cout << "toto1" << std::endl;
            *out->at(i)->val.str = string(ctmp);
            break;
        case URANIE::DataServer::UEntry::Vector :
            out->at(i)->val.vec->clear();
            for(unsigned int ival=0; ival<PyList_Size(PyList_GetItem(pRetour, i)); ival++)
                out->at(i)->val.vec->push_back( PyFloat_AsDouble( PyList_GetItem( PyList_GetItem(pRetour, i), ival) ) );
            break;
        default :
            break;
        }
    }
    
    Py_DECREF(pRetour);
    /* release GIL */
    PyGILState_Release(gstate);
    return getOutDimension(out);
}

}
} //namespace

