// --- C++ ---
//////////////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
// 
// This program is free software: you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation, either version 3 of
// the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program. 
// if not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////////////
/*!
 \file   TSocketRun.cxx
 \author Gilles ARNAUD <Gilles.Arnaud@cea.fr>
 \date   2017-11-23
 
 \brief  independant process communicating by socket
*/


#include "TSocketRun.h"

#include <iostream>
#include <queue>
#include <vector>

//ROOT
#include "TROOT.h"

#include "loop.h"

#define TIMEOUT 20000 // 20 sec

ClassImp (URANIE::Relauncher::TSocketRun);

namespace URANIE
{
namespace Relauncher
{

TSocketRun::TSocketRun(URANIE::Relauncher::TEval *func, int port) :
        TSeparateRun(func),
        _port(port)
{
    const char *tmp = gSystem->Getenv("URASOCKHOST");

    if (tmp != NULL && tmp[0]) {
        // client side
        std::cout << "client \n";
        _conn = tmp;
        _sserv = NULL;
        _moni = NULL;
    }
    else {
        // server side
        std::cout << "server\n";
        _sserv = new TServerSocket(port, kTRUE);
        _moni = new TMonitor;
        _moni->Add(_sserv);
    }
}

TSocketRun::~TSocketRun()
{
    int i;
    
    if (_sserv) {
        for EACH(i, _ncpus) {
            _moni->Remove(_sclient[i]);
            _sclient[i]->Close();
            delete _sclient[i];
        }
        delete _moni;
        _sserv->Close();
        delete _sserv;
    }
}

// utilitaires
int TSocketRun::wait_service()
{
    int id;

    TSocket *canal = _moni->Select();
    for EACH(id, _ncpus) if (canal == _sclient[id]) break;
    return id;
}

int TSocketRun::receive_header(int canal, int *head, int size)
{
    return _sclient[canal]->RecvRaw(head, size*sizeof(int));
}

int TSocketRun::receive_value(int canal, double *mesg, int size)
{
    return _sclient[canal]->RecvRaw(mesg, size*sizeof(double));
}

int TSocketRun::send_header(int canal, int *head, int size)
{
    return _sclient[canal]->SendRaw(head, size*sizeof(int));
}

int TSocketRun::send_value(int canal, double *mesg, int size)
{
    return _sclient[canal]->SendRaw(mesg, size*sizeof(double));
}


// protocole
void TSocketRun::startSlave()
{
    TSocket *sock;
    
    TSeparateRun::startSlave();
    
    if (_sserv == NULL) {
        // client side
        sock = new TSocket(_conn, _port);
        if (!sock->IsValid()) {
            std::cout << "can't connect to server\n";
            return;
        }
        _sclient.push_back(sock);    
        calcul_loop();
        sock->Close();
        delete sock;
    }
    else {
        // server side
        TSocket *ss;
        for (ss = _moni->Select(); ss == _sserv; ss = _moni->Select(TIMEOUT)) {
            sock = _sserv->Accept();
            _moni->Add(sock);            
            _sleep.push(_ncpus);
            _sclient.push_back(sock);
            _ncpus++;
        }
        _moni->Remove(_sserv);
        _run.resize(_ncpus);
    }
}

Bool_t TSocketRun::onMaster()
{
    return (_sserv != NULL);
}

Bool_t TSocketRun::readyToGet()
{
    return (_moni->Select(1) == (TSocket *)-1 )? kFALSE : kTRUE;
}

}} //namespace
