//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TestTLauncher.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   $Date$
 
 \brief  Fichier de test de la classe URANIE::Relauncher::TLauncher2
 
 Created on: 2012-11-29

 \par Modifications
 \li S. Pomarede (InckA) 2012-10-29
 Creation of the tests
 ***************************************************************
 
 */

#include "TestTLauncher2.h"
#include "UDisableTest.h"
#include <iostream>

// CPPUNIT
#include <cppunit/extensions/HelperMacros.h>
#include <cppunit/Test.h>

#define DBL_PRECISION_COMPARE 1.e-7

// ROOT
//#include "Api.h"

// Uranie
#include "TLauncher2.h" 
#include "UExceptions.h"
#include "TDataServer.h"
#include "TAttribute.h"
#include "TUniformDistribution.h"

#include "TCJitEval.h"
#include "TCIntEval.h"
#include "TFlatScript.h"
#include "TFlatResult.h"
#include "TLineResult.h"
#include "TLineScript.h"
#include "TKeyResult.h"
#include "TKeyScript.h"
#include "TCodeEval.h"
#include "TComposeEval.h"
//#include "TComposeMulti.h"
#include "TSequentialRun.h"
#include "TThreadedRun.h"
#include "TFlagScript.h"
#include "uranie_commands.h"
#include "code_paths.h"

#ifdef WITH_PYTHON
#include "TPythonEval.h"
#include "python_commands.h"
#endif

#ifdef WITH_MPI 
#include "TMpiRun.h"
#include "mpi_commands.h"
#endif

#ifdef WITH_PCL
#include "pcl.h"
#define StackSIZE 1048576

#include "TROOT.h"
#include "TInterpreter.h"
#include "TRegexp.h"
#include "TMethodCall.h"

#include "TCoroutineRun.h"

struct coarg {
    TMethodCall *pfun;
    URANIE::DataServer::TDataServer *tds;
    URANIE::Relauncher::TCoroutineRun *run;
};
#endif //PCL

// local
#include "TestUserFunctions.h"

//! The dataset Uniform Design with 32 samples for the flowrate external code
#define FLOWRATEASCIIFILEUNIFORMDESIGN "flowrateUniformDesign.dat"

using namespace std;

TLauncher2Test::TLauncher2Test() :
    _bLog(kFALSE), _tds1(NULL), _w1(NULL), _w2(NULL), _v1(NULL), _v2(NULL), _f1(NULL), _thev1(NULL), _thev2(NULL)
{
    _workingTestPath = gSystem->pwd();

#ifdef ROOT5
    // Creation of pointers on functions
    G__ClassInfo globalscope;
    G__MethodInfo method;
    long dummy;

    method = globalscope.GetMethod("precodeComposeScalaire",
				   "Double_t*,Double_t*", &dummy);
    _p2fPrecodeComposeScalaire = method.PointerToFunc();

    method = globalscope.GetMethod("postcodeComposeScalaire",
				   "Double_t*,Double_t*", &dummy);
    _p2fPostcodeComposeScalaire = method.PointerToFunc();

    method = globalscope.GetMethod("precodeComposeMulti", "Double_t*,Double_t*",
				   &dummy);
    _p2fPrecodeComposeMulti = method.PointerToFunc();

    method = globalscope.GetMethod("postcodeComposeMulti",
				   "Double_t*,Double_t*", &dummy);
    _p2fPostcodeComposeMulti = method.PointerToFunc();

    method = globalscope.GetMethod("Ishigami", "Double_t*,Double_t*", &dummy);
    _p2fIshigami = method.PointerToFunc();

    //method = globalscope.GetMethod("Ishigami2", "Double_t*,Double_t*", &dummy);
    //_p2fIshigami2 = method.PointerToFunc();
#endif
}

TLauncher2Test::~TLauncher2Test()
{
}


void TLauncher2Test::cleanWorkingDirectories(const char *option)
{

    if (_bLog)
    {
        cout << endl;
        cout << " ** TLauncher2Test::cleanWorkingDirectories option[" << option
	     << "] _workingTestPath[" << _workingTestPath << "]" << endl;
    }

    UranieTests::removeFile(_workingTestPath.Data(),"bidonUniformDesign.root");
    UranieTests::removeFile(_workingTestPath.Data(),"_dataserver_.root");
    UranieTests::removeFile(_workingTestPath.Data(),"_ThreadedTest_.dat");
    UranieTests::removeFile(_workingTestPath.Data(),"_ThreadedTest_.root");
    UranieTests::removeFile(_workingTestPath.Data(),"flowrateUniformDesign.root");
    UranieTests::removeFile(_workingTestPath.Data(),"Ishigami.root");
    UranieTests::removeFile(_workingTestPath.Data(),"multitype_sampling.root");
    UranieTests::removeFile(_workingTestPath.Data(),"readmultitype_sampling.root");
#ifndef WIN32
    UranieTests::removeFolder(_workingTestPath.Data(), "URANIE*");
#endif

}

void TLauncher2Test::setUp()
{
    if (_bLog)
    {
        cout << endl;
        cout
	    << "*******************************************************************************"
	    << endl;
    }
    gSystem->cd(_workingTestPath);
    cleanWorkingDirectories("setUp");

    // Attributes for Flowrate cases
    _yhat = new URANIE::DataServer::TAttribute("yhat");
    _d = new URANIE::DataServer::TAttribute("d");
    _tdsFlowrate = new URANIE::DataServer::TDataServer("tdsflowrate",
						       "Design of experiments for Flowrate");
    _errFlowrate = NULL;
    
    // Attributes for Ishigami cases
    _ct = new URANIE::DataServer::TAttribute("Ishigami");
    _tdsIshigami = new URANIE::DataServer::TDataServer("tds",
						       "Design of experiments for Ishigami");

    // Attributes for Bidon cases
    _x = new URANIE::DataServer::TAttribute("x", URANIE::DataServer::TAttribute::kVector);
    _diff = new URANIE::DataServer::TAttribute("diff", URANIE::DataServer::TAttribute::kVector);
    _tdsBidon = new URANIE::DataServer::TDataServer("tdsBidon",
						    "Design of experiments for Bidon");
    _errBidon = new URANIE::DataServer::TDataServer("errbidon",
						    "Design of experiments for Bidon not working");
    
    if (_bLog)
        cout
	    << "*******************************************************************************"
	    << endl;
}

void TLauncher2Test::tearDown()
{
    if (_bLog)
    {
        cout << endl;
        cout
	    << "*******************************************************************************"
	    << endl;
    }
    gSystem->cd(_workingTestPath);
    cleanWorkingDirectories("tearDown");

    delete _tdsFlowrate;
    _tdsFlowrate = NULL;
    delete _errFlowrate;
    _errFlowrate = NULL;
    delete _d;
    _d = NULL;
    delete _yhat;
    _yhat = NULL;

    delete _tdsIshigami;
    _tdsIshigami = NULL;
    delete _ct;
    _ct = NULL;

    delete _tdsBidon;
    _tdsBidon = NULL;
    delete _errBidon;
    _errBidon = NULL;
    delete _x;
    _x = NULL;
    delete _diff;
    _diff = NULL;

    if (_bLog)
        cout
	    << "*******************************************************************************"
	    << endl;

    if(_tds1)
    {
	delete _tds1; _tds1=NULL;
	delete _w1; _w1=NULL;
	delete _w2; _w2=NULL;
	delete _v1; _v1=NULL;
	delete _v2; _v2=NULL;
	delete _f1; _f1=NULL;
	delete _thev1; _thev1=NULL;
	delete _thev2; _thev2=NULL;
    }
	
}

////////////////////////////////
///                          ///
///         Tests            ///
///                          ///
////////////////////////////////

// -----------------------------------------------------------------------------

///       COMPOSITION       ///

void TLauncher2Test::testComposeScalaire()
{
    URANIE::DataServer::TAttribute delai("delai"), diff("diff"), yhat("yhat"), d("d");

    // Define the DataServer
    URANIE::DataServer::TDataServer tds("tdsBidon",
					"Design of experiments for Bidon");
    tds.fileDataRead("flowrateUniformDesign.dat");

    URANIE::DataServer::TAttribute *rw = tds.getAttribute("rw");
    URANIE::DataServer::TAttribute *r = tds.getAttribute("r");
    URANIE::DataServer::TAttribute *tu = tds.getAttribute("tu");
    URANIE::DataServer::TAttribute *tl = tds.getAttribute("tl");
    URANIE::DataServer::TAttribute *hu = tds.getAttribute("hu");
    URANIE::DataServer::TAttribute *hl = tds.getAttribute("hl");
    URANIE::DataServer::TAttribute *l = tds.getAttribute("l");
    URANIE::DataServer::TAttribute *kw = tds.getAttribute("kw");

    // pretraitement
#ifdef ROOT5
    URANIE::Relauncher::TCIntEval pre(_p2fPrecodeComposeScalaire);
#else
    URANIE::Relauncher::TCJitEval pre(precodeComposeScalaire);
#endif
    pre.setInputs(2, hu, hl);
    pre.addOutput(&delai);

    // traitement code
    // Set the reference input file and the key for each input attributes
    URANIE::Relauncher::TFlatScript fin("flowrate_input_with_values_rows.in");
    fin.setInputs(8, rw, r, tu, tl, hu, hl, l, kw);

    // The output file of the code
    URANIE::Relauncher::TFlatResult fout("_output_flowrate_withRow_.dat");
    fout.setOutputs(2, &yhat, &d);

    // Instanciation de mon code
    URANIE::Relauncher::TCodeEval mycode((UranieTests::code_path("flowrate") + " -s -r").data());
    
    mycode.setOldTmpDir();
    mycode.addInputFile(&fin);
    mycode.addOutputFile(&fout);

    // postraitement
#ifdef ROOT5
    URANIE::Relauncher::TCIntEval post(_p2fPostcodeComposeScalaire);
#else
    URANIE::Relauncher::TCJitEval post(postcodeComposeScalaire);
#endif
    post.setInputs(1, &d);
    post.setOutputs(1, &diff);

    // compose code
    URANIE::Relauncher::TComposeEval code;
    code.setEvals(3, &pre, &mycode, &post);

    URANIE::Relauncher::TSequentialRun run(&code);
    run.startSlave();
    if (run.onMaster())
    {
        URANIE::Relauncher::TLauncher2 lanceur(&tds, &run);

        // resolution
        lanceur.solverLoop();
        run.stopSlave();
    }

    Int_t npat = tds.getNPatterns();
    CPPUNIT_ASSERT_EQUAL(32, npat);
    CPPUNIT_ASSERT_EQUAL(13, tds.getNAttributes());
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("delai"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("diff"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("yhat"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds.isAttribute("d"));

    int indice[5] =
	{ 0, 1, int(npat / 2), npat - 2, npat - 1 };
    Double_t expectedYhat[5] =
	{ 26.18019, 14.46346, 74.39203, 139.938, 157.591 };
    Double_t expectedD[5] =
	{ 3602.045, 2916.759, 5561.004, 3944.845, 1348.649 };
    for (int i = 0; i < 5; i++)
    {
        int j = indice[i];
        CPPUNIT_ASSERT_DOUBLES_EQUAL(
	    tds.getValue("hu", j) - tds.getValue("hl", j),
	    tds.getValue("delai", j), DBL_PRECISION_COMPARE);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(expectedYhat[i], tds.getValue("yhat", j),
				     DBL_PRECISION_COMPARE);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(expectedD[i], tds.getValue("d", j),
				     DBL_PRECISION_COMPARE);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(2 * tds.getValue("d", j),
				     tds.getValue("diff", j), DBL_PRECISION_COMPARE);
    }
}

#ifdef WITH_PYTHON
void TLauncher2Test::testComposeMulti(int outType, int readempty)
{
    URANIE::DataServer::TAttribute outAtt("OUTTYPE");
    URANIE::DataServer::TAttribute empty("EMPTY");
    URANIE::DataServer::TAttribute delai("delai"), x("x", URANIE::DataServer::TAttribute::kVector);
    URANIE::DataServer::TAttribute diff("diff", URANIE::DataServer::TAttribute::kVector), diff2("diff2", URANIE::DataServer::TAttribute::kVector);

    // Define the DataServer
    _tdsBidon->fileDataRead("bidonUniformDesign.dat");

    URANIE::DataServer::TAttribute *coef = _tdsBidon->getAttribute("coef");
    URANIE::DataServer::TAttribute *pente = _tdsBidon->getAttribute("pente");

    // pretraitement
    URANIE::Relauncher::TCJitEval pre(precodeComposeMulti);
    pre.setInputs(2, coef, pente);
    pre.addOutput(&delai);

    // traitement code
    // Set the reference input file and the key for each input attributes
    URANIE::Relauncher::TFlagScript fin("bidon.py");
    fin.setInputs(4, &delai, "START", pente, "PENTE", &outAtt, "OUTTYPE", &empty, "EMPTY");

    // The output file of the code
    URANIE::Relauncher::TCodeResult *fout=NULL;
    switch(outType)
    {
    case 0: fout = new URANIE::Relauncher::TKeyResult("bidon.res");
        ((URANIE::Relauncher::TKeyResult*)fout)->setOutputs(2, &x, "x", &diff, "diff");
        if(readempty==1)
            ((URANIE::Relauncher::TKeyResult*)fout)->setVectorProperties("[",",","]");
        break;
    case 1: fout = new URANIE::Relauncher::TLineResult("bidon.res");
        ((URANIE::Relauncher::TLineResult*)fout)->setOutputs(2, &x, &diff);
        if(readempty==1)
            ((URANIE::Relauncher::TLineResult*)fout)->setVectorProperties("["," ","]");
        break;
    case 2: fout = new URANIE::Relauncher::TFlatResult("bidon.res");
        ((URANIE::Relauncher::TFlatResult*)fout)->setOutputs(2, &x, &diff);
        if(readempty==1)
            ((URANIE::Relauncher::TFlatResult*)fout)->setVectorProperties("[",",","]");
        break;
    case 3: fout = new URANIE::Relauncher::TFlatResult("bidon.res");
        ((URANIE::Relauncher::TFlatResult*)fout)->setOutputs(2, &x, &diff);
        ((URANIE::Relauncher::TFlatResult*)fout)->isMultiLine();
        break;
    }
    // Instanciation de mon code
    string python_command;
    UranieTests::set_python_executable(python_command);
    python_command += " bidon.py > bidon.res";
    URANIE::Relauncher::TCodeEval mycode(python_command.data());

    mycode.setOldTmpDir();
    mycode.addInputFile(&fin);
    mycode.addOutputFile(fout);

    // postraitement
    URANIE::Relauncher::TCJitEval post(postcodeComposeMulti_hard);
    post.addInput(&diff);
    post.addOutput(&diff2);

    // composition
    URANIE::Relauncher::TComposeEval code;//(&mycode);
    code.addEval(&pre);
    code.addEval(&mycode);
    code.addEval(&post);

    URANIE::Relauncher::TSequentialRun run(&code);
    run.startSlave();
    if (run.onMaster())
    {
	URANIE::Relauncher::TLauncher2 lanceur(_tdsBidon, &run);
        lanceur.addConstantValue(&outAtt, outType);
        lanceur.addConstantValue(&empty, readempty);
        lanceur.setSaveError(_errBidon);
	// resolution       
	lanceur.solverLoop();

	run.stopSlave();
    }

    Int_t npat = _tdsBidon->getNPatterns();
    if(readempty==0)
        CPPUNIT_ASSERT_EQUAL(25, npat);
    if(readempty==1)
        CPPUNIT_ASSERT_EQUAL(32, npat);
    CPPUNIT_ASSERT_EQUAL(6, _tdsBidon->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(kTRUE, _tdsBidon->isAttribute("delai"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, _tdsBidon->isAttribute("x"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, _tdsBidon->isAttribute("diff"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, _tdsBidon->isAttribute("diff2"));
    if(readempty==0)
    {
        Int_t npaterr = _errBidon->getNPatterns();
        CPPUNIT_ASSERT_EQUAL(7, npaterr);
        CPPUNIT_ASSERT_EQUAL(2, _errBidon->getNAttributes());
        CPPUNIT_ASSERT_EQUAL(kFALSE, _errBidon->isAttribute("delai"));
        CPPUNIT_ASSERT_EQUAL(kFALSE, _errBidon->isAttribute("x"));
        CPPUNIT_ASSERT_EQUAL(kFALSE, _errBidon->isAttribute("diff"));
        CPPUNIT_ASSERT_EQUAL(kFALSE, _errBidon->isAttribute("diff2"));
    }

    int event[5] =
	{ 0, 0, npat / 2, npat - 2, npat - 1 };
    int indice[5] =
	{ 0, 1, 1, 0, 0 };
    Double_t expectedX[5] =
	{ 1., 1.414213562, 1.414213562, 1., 1. };
    Double_t expectedDiff[5] =
	{ 0.107569383013, 0.0182558935312, 0.849006234676, 0.132255295665,
	  0.11720701888 };

    for (int i = 0; i < 5; i++)
    {
	int j = event[i];
	int ind=indice[i];
	CPPUNIT_ASSERT_DOUBLES_EQUAL(
	    _tdsBidon->getValue("coef", j) / _tdsBidon->getValue("pente", j),
	    _tdsBidon->getValue("delai", j), DBL_PRECISION_COMPARE);
	CPPUNIT_ASSERT_DOUBLES_EQUAL(
	    _tdsBidon->getValue("diff", j, ind) * _tdsBidon->getValue("diff", j, ind),
	    _tdsBidon->getValue("diff2", j, ind), DBL_PRECISION_COMPARE);
	CPPUNIT_ASSERT_DOUBLES_EQUAL(expectedX[i], _tdsBidon->getValue("x", j, ind),
				     DBL_PRECISION_COMPARE);
	CPPUNIT_ASSERT_DOUBLES_EQUAL(expectedDiff[i], _tdsBidon->getValue("diff", j, ind),
				     DBL_PRECISION_COMPARE);
    }

    delete fout;
    delete _tdsBidon; _tdsBidon=NULL;
    delete _errBidon; _errBidon=NULL;
        
}
void TLauncher2Test::testComposeMultiKey()
{
    testComposeMulti(0);
}
void TLauncher2Test::testComposeMultiKeyWithEmpty()
{
    testComposeMulti(0,1);
}
void TLauncher2Test::testComposeMultiLine()
{
    testComposeMulti(1);
}
void TLauncher2Test::testComposeMultiLineWithEmpty()
{
    testComposeMulti(1,1);
}
void TLauncher2Test::testComposeMultiFlat()
{
    testComposeMulti(2);
}
void TLauncher2Test::testComposeMultiFlatWithEmpty()
{
    testComposeMulti(2,1);
}
void TLauncher2Test::testComposeMultiFlatMultiLine()
{
    testComposeMulti(3);
}
#endif

// -----------------------------------------------------------------------------

///       FLOWRATE       ///

void TLauncher2Test::validateValuesFlowrate()
{
    Int_t npat = _tdsFlowrate->getNPatterns();
    CPPUNIT_ASSERT_EQUAL(32, npat);
    CPPUNIT_ASSERT_EQUAL(11, _tdsFlowrate->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(kTRUE, _tdsFlowrate->isAttribute("yhat"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, _tdsFlowrate->isAttribute("d"));

    int indice[5] =
    { 0, 1, int(npat / 2), npat - 2, npat - 1 };
    Double_t expectedYhat[5] =
    { 26.18019, 14.46346, 74.39203, 139.938, 157.591 };
    Double_t expectedD[5] =
    { 3602.045, 2916.759, 5561.004, 3944.845, 1348.649 };

    // Note: as no order can be expected in a TDS computed with multi-threads or MPI,
    // it is impossible to use
    //    _tdsFlowrate->getValue("name", index)
    // as it would return the index-th line, but not the index-th value of the TDS.
    double *yhatOutput, *dOutput;
    long nbData;
    URANIE::DataServer::TDSNtupleD *tdsTuple = _tdsFlowrate->GetTuple();
    for (int i = 0; i < 5; i++)
    {
        int j = indice[i];
        nbData = (long) tdsTuple->Draw("yhat:d", Form("tdsflowrate__n__iter__==%d", j),"goff");
	CPPUNIT_ASSERT( (nbData >= 0) );
        yhatOutput = tdsTuple->GetV1();
        CPPUNIT_ASSERT_DOUBLES_EQUAL(expectedYhat[i], yhatOutput[0],
                DBL_PRECISION_COMPARE);
        dOutput = tdsTuple->GetV2();
        CPPUNIT_ASSERT_DOUBLES_EQUAL(expectedD[i], dOutput[0],
                DBL_PRECISION_COMPARE);
    }
}

void TLauncher2Test::validateValuesFlowrateFailure()
{
    Int_t npat = _tdsFlowrate->getNPatterns();
    CPPUNIT_ASSERT_EQUAL(27, npat);
    CPPUNIT_ASSERT_EQUAL(11, _tdsFlowrate->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(kTRUE, _tdsFlowrate->isAttribute("yhat"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, _tdsFlowrate->isAttribute("d"));

    Int_t npat2 = _errFlowrate->getNPatterns();
    CPPUNIT_ASSERT_EQUAL(5, npat2);
    CPPUNIT_ASSERT_EQUAL(9, _errFlowrate->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(kFALSE, _errFlowrate->isAttribute("yhat"));
    CPPUNIT_ASSERT_EQUAL(kFALSE, _errFlowrate->isAttribute("d"));

    int indice[5] =
    { 0, 1, 16, 30, 31 };
    Double_t expectedYhat[5] =
    { 26.18019, 14.46346, 74.39203, 139.938, 157.591 };
    Double_t expectedD[5] =
    { 3602.045, 2916.759, 5561.004, 3944.845, 1348.649 };

    // Note: as no order can be expected in a TDS computed with multi-threads or MPI,
    // it is impossible to use
    //    _tdsFlowrate->getValue("name", index)
    // as it would return the index-th line, but not the index-th value of the TDS.
    double *yhatOutput, *dOutput;
    long nbData;
    URANIE::DataServer::TDSNtupleD *tdsTuple = _tdsFlowrate->GetTuple();
    for (int i = 0; i < 5; i++)
    {
        int j = indice[i];
        nbData = (long) tdsTuple->Draw("yhat:d", Form("tdsflowrate__n__iter__==%d", j),"goff");
	CPPUNIT_ASSERT( (nbData >= 0) );
        yhatOutput = tdsTuple->GetV1();
        CPPUNIT_ASSERT_DOUBLES_EQUAL(expectedYhat[i], yhatOutput[0],
                DBL_PRECISION_COMPARE);
        dOutput = tdsTuple->GetV2();
        CPPUNIT_ASSERT_DOUBLES_EQUAL(expectedD[i], dOutput[0],
                DBL_PRECISION_COMPARE);
    }
}

void TLauncher2Test::testFlowrateSequential()
{
    _tdsFlowrate->fileDataRead("flowrateUniformDesign.dat");

    URANIE::DataServer::TAttribute *rw = _tdsFlowrate->getAttribute("rw");
    URANIE::DataServer::TAttribute *r = _tdsFlowrate->getAttribute("r");
    URANIE::DataServer::TAttribute *tu = _tdsFlowrate->getAttribute("tu");
    URANIE::DataServer::TAttribute *tl = _tdsFlowrate->getAttribute("tl");
    URANIE::DataServer::TAttribute *hu = _tdsFlowrate->getAttribute("hu");
    URANIE::DataServer::TAttribute *hl = _tdsFlowrate->getAttribute("hl");
    URANIE::DataServer::TAttribute *l = _tdsFlowrate->getAttribute("l");
    URANIE::DataServer::TAttribute *kw = _tdsFlowrate->getAttribute("kw");

    // Set the reference input file and the key for each input attributes
    URANIE::Relauncher::TFlatScript fin("flowrate_input_with_values_rows.in");
    fin.setInputs(8, rw, r, tu, tl, hu, hl, l, kw);

    // The output file of the code
    URANIE::Relauncher::TFlatResult fout("_output_flowrate_withRow_.dat");
    fout.setOutputs(2, _yhat, _d);

    // Instanciation de mon code
    URANIE::Relauncher::TCodeEval mycode((UranieTests::code_path("flowrate") + " -s -r").data());	
    
    mycode.setOldTmpDir();
    mycode.addInputFile(&fin);
    mycode.addOutputFile(&fout);

    URANIE::Relauncher::TSequentialRun run(&mycode);
    run.startSlave();
    if (run.onMaster())
    {
        URANIE::Relauncher::TLauncher2 lanceur(_tdsFlowrate, &run);

        // resolution
        lanceur.solverLoop();
        run.stopSlave();
    }

    validateValuesFlowrate();
}

void TLauncher2Test::testFlowrateSequentialFailure()
{
    _tdsFlowrate->fileDataRead("flowrateUniformDesign.dat");

    URANIE::DataServer::TAttribute *rw = _tdsFlowrate->getAttribute("rw");
    URANIE::DataServer::TAttribute *r = _tdsFlowrate->getAttribute("r");
    URANIE::DataServer::TAttribute *tu = _tdsFlowrate->getAttribute("tu");
    URANIE::DataServer::TAttribute *tl = _tdsFlowrate->getAttribute("tl");
    URANIE::DataServer::TAttribute *hu = _tdsFlowrate->getAttribute("hu");
    URANIE::DataServer::TAttribute *hl = _tdsFlowrate->getAttribute("hl");
    URANIE::DataServer::TAttribute *l = _tdsFlowrate->getAttribute("l");
    URANIE::DataServer::TAttribute *kw = _tdsFlowrate->getAttribute("kw");

    // Set the reference input file and the key for each input attributes
    URANIE::Relauncher::TFlatScript fin("flowrate_input_with_values_rows.in");
    fin.setInputs(8, rw, r, tu, tl, hu, hl, l, kw);

    // The output file of the code
    URANIE::Relauncher::TFlatResult fout("_output_flowrate_withRow_.dat");
    fout.setOutputs(2, _yhat, _d);

    // Instanciation de mon code
    URANIE::Relauncher::TCodeEval mycode((UranieTests::code_path("flowrate") + " -s -rf").data());	
    
    mycode.setOldTmpDir();
    mycode.addInputFile(&fin);
    mycode.addOutputFile(&fout);

    URANIE::Relauncher::TSequentialRun run(&mycode);
    run.startSlave();

    if (run.onMaster())
    {
        URANIE::Relauncher::TLauncher2 lanceur(_tdsFlowrate, &run);

        _errFlowrate=new URANIE::DataServer::TDataServer("errFlowrate","poeut");
        lanceur.setSaveError(_errFlowrate);
        // resolution
        lanceur.solverLoop();
        run.stopSlave();
    }

    validateValuesFlowrateFailure();
}

void TLauncher2Test::testFlowrateThreaded()
{
    _tdsFlowrate->fileDataRead("flowrateUniformDesign.dat");

    URANIE::DataServer::TAttribute *rw = _tdsFlowrate->getAttribute("rw");
    URANIE::DataServer::TAttribute *r = _tdsFlowrate->getAttribute("r");
    URANIE::DataServer::TAttribute *tu = _tdsFlowrate->getAttribute("tu");
    URANIE::DataServer::TAttribute *tl = _tdsFlowrate->getAttribute("tl");
    URANIE::DataServer::TAttribute *hu = _tdsFlowrate->getAttribute("hu");
    URANIE::DataServer::TAttribute *hl = _tdsFlowrate->getAttribute("hl");
    URANIE::DataServer::TAttribute *l = _tdsFlowrate->getAttribute("l");
    URANIE::DataServer::TAttribute *kw = _tdsFlowrate->getAttribute("kw");

    // Set the reference input file and the key for each input attributes
    URANIE::Relauncher::TFlatScript fin("flowrate_input_with_values_rows.in");
    fin.setInputs(8, rw, r, tu, tl, hu, hl, l, kw);

    // The output file of the code
    URANIE::Relauncher::TFlatResult fout("_output_flowrate_withRow_.dat");
    fout.setOutputs(2, _yhat, _d);

    // Instanciation de mon code
    URANIE::Relauncher::TCodeEval mycode((UranieTests::code_path("flowrate") + " -s -r").data());
    
    mycode.setOldTmpDir();
    mycode.addInputFile(&fin);
    mycode.addOutputFile(&fout);

    Int_t nThread = 10;
    URANIE::Relauncher::TThreadedRun run(&mycode, nThread);
    run.startSlave();
    if (run.onMaster())
    {
        URANIE::Relauncher::TLauncher2 lanceur(_tdsFlowrate, &run);

        // resolution
        lanceur.solverLoop();
        run.stopSlave();
    }

    validateValuesFlowrate();
}

void TLauncher2Test::testFlowrateThreadedFailure()
{
    _tdsFlowrate->fileDataRead("flowrateUniformDesign.dat");

    URANIE::DataServer::TAttribute *rw = _tdsFlowrate->getAttribute("rw");
    URANIE::DataServer::TAttribute *r = _tdsFlowrate->getAttribute("r");
    URANIE::DataServer::TAttribute *tu = _tdsFlowrate->getAttribute("tu");
    URANIE::DataServer::TAttribute *tl = _tdsFlowrate->getAttribute("tl");
    URANIE::DataServer::TAttribute *hu = _tdsFlowrate->getAttribute("hu");
    URANIE::DataServer::TAttribute *hl = _tdsFlowrate->getAttribute("hl");
    URANIE::DataServer::TAttribute *l = _tdsFlowrate->getAttribute("l");
    URANIE::DataServer::TAttribute *kw = _tdsFlowrate->getAttribute("kw");

    // Set the reference input file and the key for each input attributes
    URANIE::Relauncher::TFlatScript fin("flowrate_input_with_values_rows.in");
    fin.setInputs(8, rw, r, tu, tl, hu, hl, l, kw);

    // The output file of the code
    URANIE::Relauncher::TFlatResult fout("_output_flowrate_withRow_.dat");
    fout.setOutputs(2, _yhat, _d);

    // Instanciation de mon code
    URANIE::Relauncher::TCodeEval mycode((UranieTests::code_path("flowrate") + " -s -rf").data());
    
    mycode.setOldTmpDir();
    mycode.addInputFile(&fin);
    mycode.addOutputFile(&fout);

    Int_t nThread = 10;
    URANIE::Relauncher::TThreadedRun run(&mycode, nThread);
    run.startSlave();
    if (run.onMaster())
    {
        URANIE::Relauncher::TLauncher2 lanceur(_tdsFlowrate, &run);

        _errFlowrate=new URANIE::DataServer::TDataServer("errFlowrate","poeut");
        lanceur.setSaveError(_errFlowrate);
        // resolution
        lanceur.solverLoop();
        run.stopSlave();
    }

    validateValuesFlowrateFailure();
}

void TLauncher2Test::testFlowrateSequentialAllFolders()
{
    _tdsFlowrate->fileDataRead("flowrateUniformDesign.dat");

    URANIE::DataServer::TAttribute *rw = _tdsFlowrate->getAttribute("rw");
    URANIE::DataServer::TAttribute *r = _tdsFlowrate->getAttribute("r");
    URANIE::DataServer::TAttribute *tu = _tdsFlowrate->getAttribute("tu");
    URANIE::DataServer::TAttribute *tl = _tdsFlowrate->getAttribute("tl");
    URANIE::DataServer::TAttribute *hu = _tdsFlowrate->getAttribute("hu");
    URANIE::DataServer::TAttribute *hl = _tdsFlowrate->getAttribute("hl");
    URANIE::DataServer::TAttribute *l = _tdsFlowrate->getAttribute("l");
    URANIE::DataServer::TAttribute *kw = _tdsFlowrate->getAttribute("kw");

    // Set the reference input file and the key for each input attributes
    URANIE::Relauncher::TFlatScript fin("flowrate_input_with_values_rows.in");
    fin.setInputs(8, rw, r, tu, tl, hu, hl, l, kw);

    // The output file of the code
    URANIE::Relauncher::TFlatResult fout("_output_flowrate_withRow_.dat");
    fout.setOutputs(2, _yhat, _d);

    // Instanciation de mon code
    URANIE::Relauncher::TCodeEval mycode((UranieTests::code_path("flowrate") + " -s -r").data());
    
    mycode.addInputFile(&fin);
    mycode.addOutputFile(&fout);
    mycode.keepAllFolders();

    URANIE::Relauncher::TSequentialRun run(&mycode);
    run.startSlave();
    if (run.onMaster())
    {
        URANIE::Relauncher::TLauncher2 lanceur(_tdsFlowrate, &run);

        // resolution
        lanceur.solverLoop();
        run.stopSlave();
    }

    validateValuesFlowrate();

    Int_t nbReps = 0;
    void *dir = gSystem->OpenDirectory(".");
    const char* subdir;
    TString name;
    while ((subdir = gSystem->GetDirEntry(dir)))
    {
        name=TString(subdir);
        if(name.Contains("URANIE"))
            nbReps++;
    }
    gSystem->FreeDirectory(dir);
    
    CPPUNIT_ASSERT_EQUAL(nbReps , 32);

}

void TLauncher2Test::testFlowrateThreadedAllFolders()
{
    _tdsFlowrate->fileDataRead("flowrateUniformDesign.dat");

    URANIE::DataServer::TAttribute *rw = _tdsFlowrate->getAttribute("rw");
    URANIE::DataServer::TAttribute *r = _tdsFlowrate->getAttribute("r");
    URANIE::DataServer::TAttribute *tu = _tdsFlowrate->getAttribute("tu");
    URANIE::DataServer::TAttribute *tl = _tdsFlowrate->getAttribute("tl");
    URANIE::DataServer::TAttribute *hu = _tdsFlowrate->getAttribute("hu");
    URANIE::DataServer::TAttribute *hl = _tdsFlowrate->getAttribute("hl");
    URANIE::DataServer::TAttribute *l = _tdsFlowrate->getAttribute("l");
    URANIE::DataServer::TAttribute *kw = _tdsFlowrate->getAttribute("kw");

    // Set the reference input file and the key for each input attributes
    URANIE::Relauncher::TFlatScript fin("flowrate_input_with_values_rows.in");
    fin.setInputs(8, rw, r, tu, tl, hu, hl, l, kw);

    // The output file of the code
    URANIE::Relauncher::TFlatResult fout("_output_flowrate_withRow_.dat");
    fout.setOutputs(2, _yhat, _d);

    // Instanciation de mon code
    URANIE::Relauncher::TCodeEval mycode((UranieTests::code_path("flowrate") + " -s -r").data());
    
    mycode.addInputFile(&fin);
    mycode.addOutputFile(&fout);
    mycode.keepAllFolders();

    Int_t nThread = 10;
    URANIE::Relauncher::TThreadedRun run(&mycode, nThread);
    run.startSlave();
    if (run.onMaster())
    {
        URANIE::Relauncher::TLauncher2 lanceur(_tdsFlowrate, &run);

        // resolution
        lanceur.solverLoop();
        run.stopSlave();
    }

    validateValuesFlowrate();

    Int_t nbReps = 0;
    void *dir = gSystem->OpenDirectory(".");
    const char* subdir;
    TString name;
    while ((subdir = gSystem->GetDirEntry(dir)))
    {
        name=TString(subdir);
        if(name.Contains("URANIE"))
            nbReps++;
    }
    gSystem->FreeDirectory(dir);
    CPPUNIT_ASSERT_EQUAL(nbReps , 32);
}

#ifdef WITH_MPI 
void TLauncher2Test::testFlowrateMPI()
{
    Int_t nProc = 4;

    // Run the macro with MPI
    gSystem->Exec(UranieTests::mpi_command(nProc, "root -l -q -b testFlowrateMPI.C").data());

    // Get the outputs into the TDS
    _tdsFlowrate->fileDataRead( Form("%s/_output_testFlowrateMPI_.dat", gSystem->pwd()) );

    validateValuesFlowrate();
}

void TLauncher2Test::testFlowrateMPIFailure()
{
    Int_t nProc = 4;

    // Run the macro with MPI
    gSystem->Exec(UranieTests::mpi_command(nProc, "root -l -q -b \"testFlowrateMPI.C(1)\"").data());

    // Get the outputs into the TDS
    _tdsFlowrate->fileDataRead( Form("%s/_output_testFlowrateMPI_.dat", gSystem->pwd()) );
    _errFlowrate=new URANIE::DataServer::TDataServer("errFlowrate","poeut");
    _errFlowrate->fileDataRead( Form("%s/_output_testFlowrateMPI_err_.dat", gSystem->pwd()) );

    validateValuesFlowrateFailure();
}

void TLauncher2Test::testFlowrateMPIAllFolders()
{
    Int_t nProc = 4;

    // Run the macro with MPI
    gSystem->Exec(UranieTests::mpi_command(nProc, "root -l -q -b testFlowrateMPIallfolders.C")
        .data());

    // Get the outputs into the TDS
    _tdsFlowrate->fileDataRead( Form("%s/_output_testFlowrateMPI_.dat", gSystem->pwd()) );

    validateValuesFlowrate();

    Int_t nbReps = 0;
    void *dir = gSystem->OpenDirectory(".");
    const char* subdir;
    TString name;
    while ((subdir = gSystem->GetDirEntry(dir)))
    {
        name=TString(subdir);
        if(name.Contains("URANIE"))
            nbReps++;
    }
    gSystem->FreeDirectory(dir);
    CPPUNIT_ASSERT_EQUAL(nbReps , 32);
}
#endif

// -----------------------------------------------------------------------------

///       ISHIGAMI       ///

void TLauncher2Test::validateValuesIshigami(URANIE::DataServer::TDataServer *tds)
{
    if (!tds) tds = _tdsIshigami;
    Int_t npat = tds->getNPatterns();
    CPPUNIT_ASSERT_EQUAL(50, npat);
    CPPUNIT_ASSERT_EQUAL(9, tds->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds->isAttribute("Ishigami"));

    int indice[5] =
    { 0, 1, int(npat / 2), npat - 2, npat - 1 };
    Double_t expectedIshigami[5] =
    { 3.874759183, 1.813092353, -1.994251257, 1.889860022, 1.939837315 };

    // Note: as no order can be expected in a TDS computed with multi-threads or MPI,
    // it is impossible to use
    //    _tdsIshigami->getValue("name", index)
    // as it would return the index-th line, but not the index-th value of the TDS.
    double *ishigamiOutput;
    long nbData;
    URANIE::DataServer::TDSNtupleD *tdsTuple = tds->GetTuple();
    for (int i = 0; i < 5; i++)
    {
        int j = indice[i];
        nbData = (long) tdsTuple->Draw("Ishigami", Form("tds__n__iter__==%d", j),"goff");
	CPPUNIT_ASSERT( (nbData >= 0) );
        ishigamiOutput = tdsTuple->GetV1();
        CPPUNIT_ASSERT_DOUBLES_EQUAL(expectedIshigami[i], ishigamiOutput[0],
                DBL_PRECISION_COMPARE);
    }
}

void TLauncher2Test::validateValuesIshigami3(URANIE::DataServer::TDataServer *tds)
{
    if (!tds) tds = _tdsIshigami;
    Int_t npat = tds->getNPatterns();
    CPPUNIT_ASSERT_EQUAL(26, npat);
    CPPUNIT_ASSERT_EQUAL(9, tds->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds->isAttribute("Ishigami"));

    int indice[5] =
    { 1, 4, 11, 28, 48 };
    Double_t expectedIshigami[5] =
    { 1.813092353, 3.331671643, 1.965493043, 5.060157077, 1.889860022 };

    // Note: as no order can be expected in a TDS computed with multi-threads or MPI,
    // it is impossible to use
    //    _tdsIshigami->getValue("name", index)
    // as it would return the index-th line, but not the index-th value of the TDS.
    double *ishigamiOutput;
    long nbData;
    URANIE::DataServer::TDSNtupleD *tdsTuple = tds->GetTuple();
    for (int i = 0; i < 5; i++)
    {
        int j = indice[i];
        nbData = (long) tdsTuple->Draw("Ishigami", Form("tds__n__iter__==%d", j),"goff");
        CPPUNIT_ASSERT( (nbData >= 0) );
        ishigamiOutput = tdsTuple->GetV1();
        CPPUNIT_ASSERT_DOUBLES_EQUAL(expectedIshigami[i], ishigamiOutput[0],
                DBL_PRECISION_COMPARE);
    }
}

void TLauncher2Test::validateValuesIshigamiErr(URANIE::DataServer::TDataServer *tds)
{
    if (!tds) tds = _tdsIshigami;
    Int_t npat = tds->getNPatterns();
    CPPUNIT_ASSERT_EQUAL(24, npat);
    CPPUNIT_ASSERT_EQUAL(8, tds->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(kFALSE, tds->isAttribute("Ishigami"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds->isAttribute("x_{1}"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, tds->isAttribute("IshigamiStar"));

    int indice[3] =
    { 0, 25, 49 };
    Double_t expectedIshigami[3] =
    { 3.874759183, -1.994251257, 1.939837315 };

    // Note: as no order can be expected in a TDS computed with multi-threads or MPI,
    // it is impossible to use
    //    _tdsIshigami->getValue("name", index)
    // as it would return the index-th line, but not the index-th value of the TDS.
    double *ishigamiOutput;
    long nbData;
    URANIE::DataServer::TDSNtupleD *tdsTuple = tds->GetTuple();
    for (int i = 0; i < 3; i++)
    {
        int j = indice[i];
        nbData = (long) tdsTuple->Draw("IshigamiStar", Form("tdserr__n__iter__==%d", j),"goff");
        CPPUNIT_ASSERT( (nbData >= 0) );
        ishigamiOutput = tdsTuple->GetV1();
        CPPUNIT_ASSERT_DOUBLES_EQUAL(expectedIshigami[i], ishigamiOutput[0],
                DBL_PRECISION_COMPARE);
    }
}

void TLauncher2Test::validateValuesReadmt(URANIE::DataServer::TDataServer *tds, bool nov2)
{
    Int_t npat = tds->getNPatterns();
    CPPUNIT_ASSERT_EQUAL(10, npat);

    //    tds->scan("*");
    int reference = 7 + ((tds->isAttribute("seed"))? 1 : 0 );
    
    CPPUNIT_ASSERT_EQUAL(reference, tds->getNAttributes());

    double expv1[10] = {0.281904968,-0.692320006,-123456789,-0.819208969,-1.56101283,0.884364824,-123456789,1.63357931,-123456789,-0.239765048};
    double expv2[10] = {11.4244901,15.9665363,12.6290655,11.0862816,-123456789,10.3407749,7.1500716,14.5373578,-123456789,9.32100627};
    double injv2[10] = {1.,3.,6.,3.,-123456789,6.,6.,5.,-123456789,-123456789};
    
    // Note: as no order can be expected in a TDS computed with multi-threads or MPI,
    // it is impossible to use
    //    tds->getValue("name", index)
    // as it would return the index-th line, but not the index-th value of the TDS.
    double *thev1, *thev2;
    long nbData;
    URANIE::DataServer::TDSNtupleD *tdsTuple = tds->GetTuple();
    for (int i = 0; i < 10; i++)
    {
        nbData = (long) tdsTuple->Draw("thev1:thev2", Form("foo__n__iter__==%d", i),"goff");
	CPPUNIT_ASSERT( (nbData >= 0) );
        thev1 = tdsTuple->GetV1();
	thev2 = tdsTuple->GetV2();
        CPPUNIT_ASSERT_DOUBLES_EQUAL(expv1[i], thev1[0], DBL_PRECISION_COMPARE);
	if(nov2)
            CPPUNIT_ASSERT_DOUBLES_EQUAL(injv2[i], thev2[0], DBL_PRECISION_COMPARE);
        else
            CPPUNIT_ASSERT_DOUBLES_EQUAL(expv2[i], thev2[0], DBL_PRECISION_COMPARE);
    }
}

void TLauncher2Test::testIshigamiSequential()
{
    _tdsIshigami->fileDataRead("Ishigami.dat");

    URANIE::DataServer::TAttribute *x1 = _tdsIshigami->getAttribute("x_{1}");
    URANIE::DataServer::TAttribute *x2 = _tdsIshigami->getAttribute("x_{2}");
    URANIE::DataServer::TAttribute *x3 = _tdsIshigami->getAttribute("x_{3}");

#ifdef ROOT5
    URANIE::Relauncher::TCIntEval code(_p2fIshigami);
#else
    URANIE::Relauncher::TCJitEval code(Ishigami);
#endif
    code.addInput(x1);
    code.addInput(x2);
    code.addInput(x3);
    code.addOutput(_ct);

    URANIE::Relauncher::TLauncher2 launch(_tdsIshigami, &code);

    // resolution
    launch.solverLoop();

    validateValuesIshigami();
}

void TLauncher2Test::testIshigamiShort()
{
    _tdsIshigami->fileDataRead("Ishigami.dat");

#ifdef ROOT5
    URANIE::Relauncher::TCIntEval code(_p2fIshigami);
#else
    URANIE::Relauncher::TCJitEval code(Ishigami);
#endif
    URANIE::Relauncher::TLauncher2 launch(_tdsIshigami, 
               &code, "x_{1}:x_{2}:x_{3}", "Ishigami");

    // resolution
    launch.solverLoop();

    validateValuesIshigami();
}

void TLauncher2Test::testIshigamiSo()
{
    _tdsIshigami->fileDataRead("Ishigami.dat");

    URANIE::DataServer::TAttribute *x1 = _tdsIshigami->getAttribute("x_{1}");
    URANIE::DataServer::TAttribute *x2 = _tdsIshigami->getAttribute("x_{2}");
    URANIE::DataServer::TAttribute *x3 = _tdsIshigami->getAttribute("x_{3}");

    // Note: we have to create the TCJitEval object with the Ishigami function object,
    // and not with the pointer to function as we do in other tests
    URANIE::Relauncher::TCJitEval code(&Ishigami);
    code.addInput(x1);
    code.addInput(x2);
    code.addInput(x3);
    code.addOutput(_ct);

    Int_t nThread = 10;
    URANIE::Relauncher::TThreadedRun run(&code, nThread);
    run.startSlave();
    if (run.onMaster())
    {
        URANIE::Relauncher::TLauncher2 launch(_tdsIshigami, &run);

        // resolution
        launch.solverLoop();

        run.stopSlave();
    }

    validateValuesIshigami();
}

void TLauncher2Test::testIshigami2So()
{
    _tdsIshigami->fileDataRead("Ishigami.dat");

    URANIE::DataServer::TAttribute *x1 = _tdsIshigami->getAttribute("x_{1}");
    URANIE::DataServer::TAttribute *x2 = _tdsIshigami->getAttribute("x_{2}");
    URANIE::DataServer::TAttribute *x3 = _tdsIshigami->getAttribute("x_{3}");

    // Note: we have to create the TCJitEval object with the Ishigami function object,
    // and not with the pointer to function as we do in other tests
    double extra[2] = {7.0, 0.1};
    URANIE::Relauncher::TCJitEval code(&Ishigami2, (void *) extra);
    code.addInput(x1);
    code.addInput(x2);
    code.addInput(x3);
    code.addOutput(_ct);

    Int_t nThread = 10;
    URANIE::Relauncher::TThreadedRun run(&code, nThread);
    run.startSlave();
    if (run.onMaster())
    {
        URANIE::Relauncher::TLauncher2 launch(_tdsIshigami, &run);

        // resolution
        launch.solverLoop();

        run.stopSlave();
    }

    validateValuesIshigami();
}

void TLauncher2Test::testIshigami3So()
{
    _tdsIshigami->fileDataRead("Ishigami.dat");

    URANIE::DataServer::TAttribute *x1 = _tdsIshigami->getAttribute("x_{1}");
    URANIE::DataServer::TAttribute *x2 = _tdsIshigami->getAttribute("x_{2}");
    URANIE::DataServer::TAttribute *x3 = _tdsIshigami->getAttribute("x_{3}");

    URANIE::DataServer::TDataServer error("tdserr", "tds des erreurs");
    // Note: we have to create the TCJitEval object with the Ishigami function object,
    // and not with the pointer to function as we do in other tests
    URANIE::Relauncher::TCJitEval code(&Ishigami3);
    code.addInput(x1);
    code.addInput(x2);
    code.addInput(x3);
    code.addOutput(_ct);

    Int_t nThread = 10;
    URANIE::Relauncher::TThreadedRun run(&code, nThread);
    run.startSlave();
    if (run.onMaster())
    {
        URANIE::Relauncher::TLauncher2 launch(_tdsIshigami, &run);
        launch.setSaveError(&error);

        // resolution
        launch.solverLoop();

        run.stopSlave();
    }
    error.exportData("caca.dat");

    validateValuesIshigami3();
    validateValuesIshigamiErr(&error);
}


void TLauncher2Test::testIshigami4lambda()
{
    _tdsIshigami->fileDataRead("Ishigami.dat");

    URANIE::DataServer::TAttribute *x1 = _tdsIshigami->getAttribute("x_{1}");
    URANIE::DataServer::TAttribute *x2 = _tdsIshigami->getAttribute("x_{2}");
    URANIE::DataServer::TAttribute *x3 = _tdsIshigami->getAttribute("x_{3}");

    // Note: we have to create the TCJitEval object with the Ishigami function object,
    // and not with the pointer to function as we do in other tests
    double a = 7.0, b = 0.1 ;
    std::function<int(double*,double*)> MyIshigami = [=](double *paramt, double *res)
    {
        res[0] = sin(paramt[0]);
        res[0] += a * sin(paramt[1]) * sin(paramt[1]);
        res[0] += b * paramt[2] * paramt[2] * paramt[2] * paramt[2]
                * sin(paramt[0]);
        //res[1] = 5.0 * paramt[0];
        return  1;
    };
    URANIE::Relauncher::TCJitEval code(&MyIshigami);
    code.addInput(x1);
    code.addInput(x2);
    code.addInput(x3);
    code.addOutput(_ct);

    Int_t nThread = 10;
    URANIE::Relauncher::TThreadedRun run(&code, nThread);
    run.startSlave();
    if (run.onMaster())
    {
        URANIE::Relauncher::TLauncher2 launch(_tdsIshigami, &run);

        // resolution
        launch.solverLoop();

        run.stopSlave();
    }

    validateValuesIshigami();
}



#ifdef WITH_MPI 
void TLauncher2Test::testIshigamiMPI()
{
    Int_t nProc = 4;

    // Run the macro with MPI
    gSystem->Exec(UranieTests::mpi_command(nProc, "root -l -q -b testIshigamiMPI.C").data());

    // Get the outputs into the TDS
    _tdsIshigami->fileDataRead(Form("%s/_output_testIshigamiMPI_.dat", gSystem->pwd()) );

    // Assertions
    validateValuesIshigami();

}
#endif

#ifdef WITH_PYTHON
void TLauncher2Test::testIshigamiSequentialPython()
{
    gSystem->Exec(UranieTests::python_command("testIshigamiSequential.py").data());

    // Get the outputs into the TDS
    _tdsIshigami->fileDataRead(
            Form("%s/_output_testIshigamiSequentialPython_.dat",
                    gSystem->pwd()));

    // Assertions
    validateValuesIshigami();
}
#endif

#ifdef WITH_PYTHON
void TLauncher2Test::testReadmtSequentialPython()
{
    gSystem->Exec(UranieTests::python_command("testReadmtSequential.py").data());

    // Get the outputs into the TDS
    _tdsIshigami->fileDataRead(
            Form("%s/_output_testReadmtSequentialPython_.dat",
                    gSystem->pwd()));

    // Assertions
    validateValuesReadmt(_tdsIshigami);
}
#endif

#ifdef WITH_PYTHON
void TLauncher2Test::testmtSequentialPython()
{
    gSystem->Exec(UranieTests::python_command("testmtSequential.py").data());
    
    // Get the outputs into the TDS
    _tdsIshigami->fileDataRead(
	Form("%s/_output_testmtSequentialPython_.dat",
	     gSystem->pwd()));
    
    // Assertions
    validateOutputMultiTypeResults(_tdsIshigami);
}

#ifdef WITH_MPI 
void TLauncher2Test::testIshigamiMPIPython()
{
    Int_t nProc = 4;

    // Run the macro with MPI
    string python_command;
    UranieTests::set_python_executable(python_command);
    gSystem->Exec(UranieTests::mpi_command(nProc, python_command + " testIshigamiMPI.py").data());
    
    // Get the outputs into the TDS
    _tdsIshigami->fileDataRead( Form("%s/_output_testIshigamiMPIPython_.dat", gSystem->pwd()) );

    // Assertions
    validateValuesIshigami();
}
#endif
#endif

#ifdef WITH_PCL
void callmacro(void *info)
{
    struct coarg *dat = (struct coarg *) info;
    int ress=3;
    void *args[3] = {&dat->tds, &dat->run, &ress};
    
    dat->pfun->SetParamPtrs(args, 3);
    dat->pfun->Execute();
}

void TLauncher2Test::testIshigamiPcl(TString &macro)
{
    char *stack;
    coroutine_t fibre;
    TMethodCall pfun;
    URANIE::Relauncher::TCoroutineRun *trun;
    struct coarg co_arg;
    struct URANIE::Relauncher::element calc;
    int i, ret, ok;
    Ssiz_t start, len;
    double in[3], out[1];
    
    /* pre */
    TRegexp post(".C\\+*$");
    
    TString fun(macro);
    start = fun.Index(post, &len, 0);
    fun = fun.Replace(start, len, "");
    std::cout << std::endl << macro << " with macro " << fun << std::endl;
    
    /* compilation macro */
    do { // bloc for break;
        ok = 1;
        if (gROOT->LoadMacro(macro)) {
            std::cout << "Probleme loading macro [" << macro << "]\n";
            ok = 0;
            break ;
        }
        pfun.InitWithPrototype(fun,
                "URANIE::DataServer::TDataServer**,URANIE::Relauncher::TCoroutineRun**,int*"
            ); 
        if (!pfun.IsValid()) {
            std::cout << "Probleme fun in macro [" << macro << "]\n";
            ok = 0;
            break;
        }
        co_arg.pfun = &pfun;

        /* lancement fibre */
        stack = new char[StackSIZE];
        fibre = co_create(callmacro, (void*) &co_arg, stack, StackSIZE);
        //std::cout << fibre << std::endl;
        //printf("go fibre\n");
        co_call(fibre);
        //printf("come back\n");
        trun = co_arg.run;

        calc.id = -1;
        calc = trun->serviceGet();
        //printf("%g %g\n", calc.in[0], calc.in[1]);
        while (calc.id >= 0) {
            if (calc.inv) {
	            for (i=0; i<3; i++) in[i] = *calc.inv->at(i)->val.dou;
                Ishigami(in, out);
                calc.outv->at(0)->assignEntry((void*) out);
                ret = 1;
                std::cout << ".";
            }
            else ret = 0;
            trun->servicePut(calc.id, ret);
            co_call(fibre);
            calc = trun->serviceGet();
        }
        std::cout << std::endl;
        // clean
        printf("loop back\n");
    } while(0);
    
    // assertion
    //_tdsIshigami->exportData("titi.dat");
    validateValuesIshigami(co_arg.tds);
    
    if (ok) co_call(fibre);
}

void TLauncher2Test::testIshigamiPclInt()
{
#ifdef WIN32
    URANIE_IGNORED_TEST();
#endif
    TString macro("testIshigamiPcl.C");
    testIshigamiPcl(macro);
}

void TLauncher2Test::testIshigamiPclJit()
{
    TString macro("testIshigamiPcl.C++");
    testIshigamiPcl(macro);
}

#endif //PCL

// -----------------------------------------------------------------------------

///       BIDON       ///

void TLauncher2Test::validateValuesBidon(int readempty)
{
    // _tdsBidon->getTuple()->SetScanField(-1);    
    // _tdsBidon->scan("*");
    // _errBidon->getTuple()->SetScanField(-1);
    // _errBidon->scan("*");

    if(readempty==0)
        CPPUNIT_ASSERT_EQUAL(25, _tdsBidon->getNPatterns());
    else
        CPPUNIT_ASSERT_EQUAL(32, _tdsBidon->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(4, _tdsBidon->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(kTRUE, _tdsBidon->isAttribute("diff"));
    CPPUNIT_ASSERT_EQUAL(kTRUE, _tdsBidon->isAttribute("x"));

    int n = 31;
    int event[5]  = { 0, 1, int(n/2), n - 2, n - 1 };
    Double_t expectedX[5] =  { 1., 1., 1., 1., 1. };
    Double_t expectedDiff[5] =
	{ 0.2548256292, 0.22574273581, 0.57265619073, 0.712703705787,
	  0.12001077382 };
    long expectedNbData[5] ={ 3, 2, 5, 6, 1 };
    
    // Note: as no order can be expected in a TDS computed with multi-threads or MPI,
    // it is impossible to use
    //    _tdsBidon->getValue("name", index)
    // as it would return the index-th line, but not the index-th value of the TDS.
    double *diffOutput, *xOutput;
    long nbData;
    URANIE::DataServer::TDSNtupleD *tdsTuple = _tdsBidon->GetTuple();
    for (int i = 0; i < 5; i++)
    {
        int j = event[i];
        nbData = (long) tdsTuple->Draw("x:diff", Form("tdsBidon__n__iter__==%d", j),"goff");
        xOutput = tdsTuple->GetV1();
        diffOutput = tdsTuple->GetV2();
        CPPUNIT_ASSERT_DOUBLES_EQUAL(expectedX[i], xOutput[0],
                DBL_PRECISION_COMPARE);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(expectedDiff[i], diffOutput[0],
                DBL_PRECISION_COMPARE);
        CPPUNIT_ASSERT_EQUAL(expectedNbData[i], nbData);
    }

    // For tdsbidon__n__iter__==5, bidon does not return any value, we test it
    nbData = (long) tdsTuple->Draw("x:diff", "tdsBidon__n__iter__==5","goff");
    CPPUNIT_ASSERT_EQUAL(0L, nbData);

    if(readempty==0)
    {
        CPPUNIT_ASSERT_EQUAL(7, _errBidon->getNPatterns());
        CPPUNIT_ASSERT_EQUAL(2, _errBidon->getNAttributes());
        CPPUNIT_ASSERT_EQUAL(kFALSE, _errBidon->isAttribute("diff"));
        CPPUNIT_ASSERT_EQUAL(kFALSE, _errBidon->isAttribute("x"));
    }
}

#ifdef WITH_PYTHON
void TLauncher2Test::testBidonSequential(int outType, int readempty)
{
    _tdsBidon->fileDataRead("bidonUniformDesign.dat");
    URANIE::DataServer::TAttribute outAtt("OUTTYPE");
    URANIE::DataServer::TAttribute empty("EMPTY");
    URANIE::DataServer::TAttribute *delai = _tdsBidon->getAttribute("coef");
    URANIE::DataServer::TAttribute *pente = _tdsBidon->getAttribute("pente");

    // Set the reference input file and the key for each input attributes
    URANIE::Relauncher::TFlagScript fin("bidon.py");
    fin.addInput(delai, "START");
    fin.addInput(pente, "PENTE");
    fin.addInput(&outAtt, "OUTTYPE");
    fin.addInput(&empty, "EMPTY");

    // The output file of the code
    URANIE::Relauncher::TCodeResult *fout=NULL;
    switch(outType)
    {
    case 0: fout = new URANIE::Relauncher::TKeyResult("bidon.res");
        ((URANIE::Relauncher::TKeyResult*)fout)->setOutputs(2, _x, "x", _diff, "diff");
        if(readempty==1)
            ((URANIE::Relauncher::TKeyResult*)fout)->setVectorProperties("[",",","]");
        break;
    case 1: fout = new URANIE::Relauncher::TLineResult("bidon.res");
        ((URANIE::Relauncher::TLineResult*)fout)->setOutputs(2, _x, _diff);
        if(readempty==1)
            ((URANIE::Relauncher::TLineResult*)fout)->setVectorProperties("["," ","]");
        break;
    case 2: fout = new URANIE::Relauncher::TFlatResult("bidon.res");
        ((URANIE::Relauncher::TFlatResult*)fout)->setOutputs(2, _x, _diff);
        if(readempty==1)
            ((URANIE::Relauncher::TFlatResult*)fout)->setVectorProperties("[",",","]");
        break;
    case 3: fout = new URANIE::Relauncher::TFlatResult("bidon.res");
        ((URANIE::Relauncher::TFlatResult*)fout)->setOutputs(2, _x, _diff);
        ((URANIE::Relauncher::TFlatResult*)fout)->isMultiLine();
        break;
    }

    // Instanciation de mon code
    string python_command;
    UranieTests::set_python_executable(python_command);
    python_command += " bidon.py > bidon.res";
    URANIE::Relauncher::TCodeEval mycode(python_command.data());
      
    mycode.keepAllFolders();
    //    mycode.setOldTmpDir();
    mycode.addInputFile(&fin);
    mycode.addOutputFile(fout);

    URANIE::Relauncher::TSequentialRun run(&mycode);
    run.startSlave();
    if (run.onMaster())
    {
        URANIE::Relauncher::TLauncher2 lanceur(_tdsBidon, &run);
        lanceur.addConstantValue(&outAtt, outType);
        lanceur.addConstantValue(&empty, readempty);
        lanceur.setSaveError(_errBidon);
        // resolution
        lanceur.solverLoop();

        run.stopSlave();
    }

    // Assertions
    validateValuesBidon(readempty);
    delete fout;
}
void TLauncher2Test::testBidonSequentialKey()
{
    testBidonSequential(0);
}
void TLauncher2Test::testBidonSequentialKeyWithEmpty()
{
    testBidonSequential(0,1);
}
void TLauncher2Test::testBidonSequentialLine()
{
    testBidonSequential(1);
}
void TLauncher2Test::testBidonSequentialLineWithEmpty()
{
    testBidonSequential(1,1);
}
void TLauncher2Test::testBidonSequentialFlat()
{
    testBidonSequential(2);
}
void TLauncher2Test::testBidonSequentialFlatWithEmpty()
{
    testBidonSequential(2,1);
}
void TLauncher2Test::testBidonSequentialFlatMultiLine()
{
    testBidonSequential(3);
}
#endif

#ifdef WITH_PYTHON
void TLauncher2Test::testBidonThreaded(int outType, int readempty)
{
    _tdsBidon->fileDataRead("bidonUniformDesign.dat");
    URANIE::DataServer::TAttribute outAtt("OUTTYPE");
    URANIE::DataServer::TAttribute empty("EMPTY");        
    URANIE::DataServer::TAttribute *delai = _tdsBidon->getAttribute("coef");
    URANIE::DataServer::TAttribute *pente = _tdsBidon->getAttribute("pente");

    // Set the reference input file and the key for each input attributes
    URANIE::Relauncher::TFlagScript fin("bidon.py");
    fin.addInput(delai, "START");
    fin.addInput(pente, "PENTE");
    fin.addInput(&outAtt, "OUTTYPE");
    fin.addInput(&empty, "EMPTY");
    
    // The output file of the code
    URANIE::Relauncher::TCodeResult *fout=NULL;
    switch(outType)
    {
    case 0: fout = new URANIE::Relauncher::TKeyResult("bidon.res");
        ((URANIE::Relauncher::TKeyResult*)fout)->setOutputs(2, _x, "x", _diff, "diff");
        if(readempty==1)
            ((URANIE::Relauncher::TKeyResult*)fout)->setVectorProperties("[",",","]");
        break;
    case 1: fout = new URANIE::Relauncher::TLineResult("bidon.res");
        ((URANIE::Relauncher::TLineResult*)fout)->setOutputs(2, _x, _diff);
        if(readempty==1)
            ((URANIE::Relauncher::TLineResult*)fout)->setVectorProperties("["," ","]");        
        break;
    case 2: fout = new URANIE::Relauncher::TFlatResult("bidon.res");
        ((URANIE::Relauncher::TFlatResult*)fout)->setOutputs(2, _x, _diff);
        if(readempty==1)
            ((URANIE::Relauncher::TFlatResult*)fout)->setVectorProperties("[",",","]");        
        break;
    case 3: fout = new URANIE::Relauncher::TFlatResult("bidon.res");
        ((URANIE::Relauncher::TFlatResult*)fout)->setOutputs(2, _x, _diff);
        ((URANIE::Relauncher::TFlatResult*)fout)->isMultiLine();
        break;
    }
    
    // Instanciation de mon code
    string python_command;
    UranieTests::set_python_executable(python_command);
    python_command += " bidon.py > bidon.res";
    URANIE::Relauncher::TCodeEval mycode(python_command.data());
        
    mycode.setOldTmpDir();
    mycode.addInputFile(&fin);
    mycode.addOutputFile(fout);

    Int_t nThread = 10;
    URANIE::Relauncher::TThreadedRun run(&mycode, nThread);
    run.startSlave();
    if (run.onMaster())
    {
        URANIE::Relauncher::TLauncher2 lanceur(_tdsBidon, &run);
        lanceur.setSaveError(_errBidon);
        lanceur.addConstantValue(&outAtt, outType);
        lanceur.addConstantValue(&empty, readempty);
        // resolution
        lanceur.solverLoop();

        run.stopSlave();
    }

//    Assertions
    validateValuesBidon(readempty);
    delete fout;
}
void TLauncher2Test::testBidonThreadedKey()
{
    testBidonThreaded(0);
}
void TLauncher2Test::testBidonThreadedKeyWithEmpty()
{
    testBidonThreaded(0,1);
}
void TLauncher2Test::testBidonThreadedLine()
{
    testBidonThreaded(1);
}
void TLauncher2Test::testBidonThreadedLineWithEmpty()
{
    testBidonThreaded(1,1);
}
void TLauncher2Test::testBidonThreadedFlat()
{
    testBidonThreaded(2);
}
void TLauncher2Test::testBidonThreadedFlatWithEmpty()
{
    testBidonThreaded(2,1);
}
void TLauncher2Test::testBidonThreadedFlatMultiLine()
{
    testBidonThreaded(3);
}
#endif

#ifdef WITH_MPI 
#ifdef WITH_PYTHON
void TLauncher2Test::testBidonMPI(string outType, string readempty)
{
    Int_t nProc = 4;

    // Run the macro with MPI
    string root_command;
    if(string(gSystem->GetBuildArch()) == "win32") {
        root_command = "root -l -q -b \"testBidonMPI.C("+outType+", "+readempty+")\"";
    } else {
        string python_command;
        UranieTests::set_python_executable(python_command, true);
        std::cout << "python_command: [" << python_command << "]" << std::endl;
        root_command = "root -l -q -b \"testBidonMPI.C("+outType+", "+readempty+",\""+python_command+"\")\"";
    }
    gSystem->Exec(UranieTests::mpi_command(nProc, root_command).data());

    // Get the outputs into the TDS
    _tdsBidon->fileDataRead("_output_testBidonMPI_.dat");
    if(readempty!="1")
        _errBidon->fileDataRead("_output_testBidonMPI_err_.dat");
    // Assertions
    validateValuesBidon(atoi(readempty.c_str()));

    UranieTests::removeFile(_workingTestPath.Data(),"_output_*.dat");
}
void TLauncher2Test::testBidonMPIKey()
{
    testBidonMPI("0");
}
void TLauncher2Test::testBidonMPIKeyWithEmpty()
{
    testBidonMPI("0","1");
}
void TLauncher2Test::testBidonMPILine()
{
    testBidonMPI("1");
}
void TLauncher2Test::testBidonMPILineWithEmpty()
{
    testBidonMPI("1","1");
}
void TLauncher2Test::testBidonMPIFlat()
{
    testBidonMPI("2");
}
void TLauncher2Test::testBidonMPIFlatWithEmpty()
{
    testBidonMPI("2","1");
}
void TLauncher2Test::testBidonMPIFlatMultiLine()
{
    testBidonMPI("3");
}
#endif
#endif


void TLauncher2Test::testThreadedWithDataServer()
{
    // Run the macro
    gSystem->Exec(UranieTests::uranie_command("root -l -q -b \"testThreadSafe.C(true)\"").data());

    // Assertions
    validateValuesThreadable(true);
}

void TLauncher2Test::testThreadedWithDataServerFailing()
{
    // Check whether commenting this improves the speed of test (because it implies a REAL crash)
    
    // // Run the macro with MPI
    // gSystem->Exec(Form("bash -c \"( . %s/../../../uranie.bashrc && root -l -q -b testThreadSafe.C )\"",
    //                    gSystem->pwd()));

    // // Assertions
    // validateValuesThreadable(false);    
}

void TLauncher2Test::validateValuesThreadable(bool hasfile)
{

    URANIE::DataServer::TDataServer tds("pouet","pouet");
    
    if( !hasfile)
        CPPUNIT_ASSERT_THROW(tds.fileDataRead("_ThreadedTest_.dat"), URANIE::Exceptions::UErrorExceptions);
    else
    {
        tds.fileDataRead("_ThreadedTest_.dat");
        CPPUNIT_ASSERT_EQUAL(24, tds.getNPatterns());
        CPPUNIT_ASSERT_EQUAL(2,  tds.getNAttributes());

        tds.computeStatistic();

        // cout<<std::setprecision(12)<<tds.getAttribute("multiplier")->getMean()<<" "<<tds.getAttribute("multiplier")->getStd()<<endl;
        // cout<<std::setprecision(12)<<tds.getAttribute("MultMean")->getMean()<<" "<<tds.getAttribute("MultMean")->getStd()<<endl;

        CPPUNIT_ASSERT_DOUBLES_EQUAL(5.504756858,tds.getAttribute("multiplier")->getMean() , DBL_PRECISION_COMPARE);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(2.62712516199,tds.getAttribute("multiplier")->getStd() , DBL_PRECISION_COMPARE);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(907.073835075,tds.getAttribute("MultMean")->getMean() , DBL_PRECISION_COMPARE);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(432.897684156,tds.getAttribute("MultMean")->getStd() , DBL_PRECISION_COMPARE);
        
    }        
}

////////////////////////////////
///                          ///
///      multitype              ///
///                          ///
////////////////////////////////

void TLauncher2Test::setUpMultiType()
{
    _tds1 = new URANIE::DataServer::TDataServer("foo", "multitype usecase");
    _tds1->fileDataRead("multitype_sampling.dat");

    //outputs
    _w1 = new URANIE::DataServer::TAttribute("w1", URANIE::DataServer::TAttribute::kString);
    _w2 = new URANIE::DataServer::TAttribute("w2", URANIE::DataServer::TAttribute::kString);
    _v1 = new URANIE::DataServer::TAttribute("v1", URANIE::DataServer::TAttribute::kVector);
    _v2 = new URANIE::DataServer::TAttribute("v2", URANIE::DataServer::TAttribute::kVector);
    _f1 = new URANIE::DataServer::TAttribute("f1");
    
}

void TLauncher2Test::testMultiTypeOutputFileFlat()
{

    setUpMultiType();

    // Create the input files
    URANIE::Relauncher::TFlatScript inputFile("multitype_input.dat");
    inputFile.setInputs(1, _tds1->getAttribute("seed"), "seed");    

    // // Create the output files
    URANIE::Relauncher::TFlatResult outputFile("_output_multitype_mt_DataServer_.dat");
    outputFile.addOutput(_w1);
    outputFile.addOutput(_v1);
    outputFile.addOutput(_v2);
    outputFile.addOutput(_f1);
    outputFile.addOutput(_w2);

    URANIE::Relauncher::TCodeEval eval((UranieTests::code_path("multitype") + " -mtDS").data());
    // Create the user's evaluation function
    eval.addInputFile(&inputFile);
    eval.addOutputFile(&outputFile);    

    //TSequentialRun runner(&eval);
    URANIE::Relauncher::TSequentialRun runner(&eval);
    
    runner.startSlave();                                           
    if (runner.onMaster())
    {
        
	// Create the launcher
	URANIE::Relauncher::TLauncher2 lanceur(_tds1, &runner);
	lanceur.solverLoop();
        
	// Stop the slave processes
	runner.stopSlave();

    }        
    validateOutputMultiTypeResults(_tds1);  
    
}

void TLauncher2Test::testMultiTypeOutputFileFlat2()
{

    setUpMultiType();

    // Create the input files
    URANIE::Relauncher::TFlatScript inputFile("multitype_input.dat");
    inputFile.setInputs(1, _tds1->getAttribute("seed"), "seed");    

    // // Create the output files
    URANIE::Relauncher::TFlatResult outputFile("_output_multitype_mt_Flat_.dat");
    outputFile.addOutput(_w1);
    outputFile.addOutput(_v1);
    outputFile.addOutput(_v2);
    outputFile.addOutput(_f1);
    outputFile.addOutput(_w2);
    outputFile.setVectorProperties("[",",","]");

    URANIE::Relauncher::TCodeEval eval((UranieTests::code_path("multitype") + " -mtFlat").data());
    // Create the user's evaluation function
    eval.addInputFile(&inputFile);
    eval.addOutputFile(&outputFile);    

    //TSequentialRun runner(&eval);
    URANIE::Relauncher::TSequentialRun runner(&eval);
    
    runner.startSlave();                                           
    if (runner.onMaster())
    {
        
	// Create the launcher
	URANIE::Relauncher::TLauncher2 lanceur(_tds1, &runner);
	lanceur.solverLoop();
        
	// Stop the slave processes
	runner.stopSlave();

    }        
    validateOutputMultiTypeResults(_tds1);  
    
}

void TLauncher2Test::testMultiTypeOutputFileFlatMultiLine()
{

    setUpMultiType();

    // Create the input files
    URANIE::Relauncher::TFlatScript inputFile("multitype_input.dat");
    inputFile.setInputs(1, _tds1->getAttribute("seed"), "seed");    

    // // Create the output files
    URANIE::Relauncher::TFlatResult outputFile("_output_multitype_mt_Row_.dat");
    outputFile.addOutput(_w1);
    outputFile.addOutput(_v1);
    outputFile.addOutput(_v2);
    outputFile.addOutput(_f1);
    outputFile.addOutput(_w2);
    outputFile.isMultiLine(";");

    URANIE::Relauncher::TCodeEval eval((UranieTests::code_path("multitype") + " -mtRow").data());
    // Create the user's evaluation function
    eval.addInputFile(&inputFile);
    eval.addOutputFile(&outputFile);    

    //TSequentialRun runner(&eval);
    URANIE::Relauncher::TSequentialRun runner(&eval);
    
    runner.startSlave();                                           
    if (runner.onMaster())
    {
        
	// Create the launcher
	URANIE::Relauncher::TLauncher2 lanceur(_tds1, &runner);
	lanceur.solverLoop();
        
	// Stop the slave processes
	runner.stopSlave();

    }        
    validateOutputMultiTypeResults(_tds1);  
    
}

void TLauncher2Test::testMultiTypeOutputFileKey()
{

    setUpMultiType();

    // Create the input files
    URANIE::Relauncher::TFlatScript inputFile("multitype_input.dat");
    inputFile.setInputs(1, _tds1->getAttribute("seed"), "seed");    

    // // Create the output files
    URANIE::Relauncher::TKeyResult outputFile("_output_multitype_mt_Key_.dat");
    outputFile.addOutput(_w1, "w1");
    outputFile.addOutput(_v1, "v1");
    outputFile.addOutput(_v2, "v2");
    outputFile.addOutput(_f1, "f1");
    outputFile.addOutput(_w2, "w2");
  
    // Create the user's evaluation function
    URANIE::Relauncher::TCodeEval eval((UranieTests::code_path("multitype") + " -mtKey").data());
    eval.addInputFile(&inputFile);
    eval.addOutputFile(&outputFile);    

    //TSequentialRun runner(&eval);
    URANIE::Relauncher::TSequentialRun runner(&eval);
    
    runner.startSlave();                                           
    if (runner.onMaster())
    {
        
	// Create the launcher
	URANIE::Relauncher::TLauncher2 lanceur(_tds1, &runner);
	lanceur.solverLoop();
        
	// Stop the slave processes
	runner.stopSlave();

    }        
    validateOutputMultiTypeResults(_tds1);  
    
}

void TLauncher2Test::testMultiTypeOutputFileKeyCondensate()
{

    setUpMultiType();

    // Create the input files
    URANIE::Relauncher::TFlatScript inputFile("multitype_input.dat");
    inputFile.setInputs(1, _tds1->getAttribute("seed"), "seed");    

    // // Create the output files
    URANIE::Relauncher::TKeyResult outputFile("_output_multitype_mt_Key_condensate_.dat");
    outputFile.addOutput(_w1, "w1");
    outputFile.addOutput(_v1, "v1");
    outputFile.addOutput(_v2, "v2");
    outputFile.addOutput(_f1, "f1");
    outputFile.addOutput(_w2, "w2");
    outputFile.setVectorProperties("[",",","]");
  
    // Create the user's evaluation function
    URANIE::Relauncher::TCodeEval eval((UranieTests::code_path("multitype") + " -mtKey").data());
    eval.addInputFile(&inputFile);
    eval.addOutputFile(&outputFile);    

    //TSequentialRun runner(&eval);
    URANIE::Relauncher::TSequentialRun runner(&eval);
    
    runner.startSlave();                                           
    if (runner.onMaster())
    {
        
	// Create the launcher
	URANIE::Relauncher::TLauncher2 lanceur(_tds1, &runner);
	lanceur.solverLoop();
        
	// Stop the slave processes
	runner.stopSlave();

    }        
    validateOutputMultiTypeResults(_tds1);  
    
}

void TLauncher2Test::testMultiTypeOutputFileLine()
{

    setUpMultiType();

    // Create the input files
    URANIE::Relauncher::TFlatScript inputFile("multitype_input.dat");
    inputFile.setInputs(1, _tds1->getAttribute("seed"), "seed");    

    // // Create the output files
    URANIE::Relauncher::TLineResult outputFile("_output_multitype_mt_Column_.dat");
    outputFile.addOutput(_w1);
    outputFile.addOutput(_v1);
    outputFile.addOutput(_v2);
    outputFile.addOutput(_f1);
    outputFile.addOutput(_w2);
  
    // Create the user's evaluation function
    URANIE::Relauncher::TCodeEval eval((UranieTests::code_path("multitype") + " -mtCol").data());
    eval.addInputFile(&inputFile);
    eval.addOutputFile(&outputFile);    

    //TSequentialRun runner(&eval);
    URANIE::Relauncher::TSequentialRun runner(&eval);
    
    runner.startSlave();                                           
    if (runner.onMaster())
    {
        
	// Create the launcher
	URANIE::Relauncher::TLauncher2 lanceur(_tds1, &runner);
	lanceur.solverLoop();
        
	// Stop the slave processes
	runner.stopSlave();

    }        
    validateOutputMultiTypeResults(_tds1);  
    
}

void TLauncher2Test::testMultiTypeOutputFileLineThreaded()
{

    setUpMultiType();

    // Create the input files
    URANIE::Relauncher::TFlatScript inputFile("multitype_input.dat");
    inputFile.setInputs(1, _tds1->getAttribute("seed"), "seed");    

    // // Create the output files
    URANIE::Relauncher::TLineResult outputFile("_output_multitype_mt_Column_.dat");
    outputFile.addOutput(_w1);
    outputFile.addOutput(_v1);
    outputFile.addOutput(_v2);
    outputFile.addOutput(_f1);
    outputFile.addOutput(_w2);
  
    // Create the user's evaluation function
    URANIE::Relauncher::TCodeEval eval((UranieTests::code_path("multitype") + " -mtCol").data());
    eval.addInputFile(&inputFile);
    eval.addOutputFile(&outputFile);    

    //TSequentialRun runner(&eval);
    URANIE::Relauncher::TThreadedRun runner(&eval,4);
    
    runner.startSlave();                                           
    if (runner.onMaster())
    {
        
	// Create the launcher
	URANIE::Relauncher::TLauncher2 lanceur(_tds1, &runner);
	lanceur.solverLoop();
        
	// Stop the slave processes
	runner.stopSlave();

    }        
    validateOutputMultiTypeResults(_tds1);  
    
}

void TLauncher2Test::validateOutputMultiTypeResults(URANIE::DataServer::TDataServer *tds)
{
    //Attribute are there  
    CPPUNIT_ASSERT(tds->isAttribute("seed"));
    CPPUNIT_ASSERT(tds->isAttribute("w1"));
    CPPUNIT_ASSERT(tds->isAttribute("v1"));
    CPPUNIT_ASSERT(tds->isAttribute("v2"));
    CPPUNIT_ASSERT(tds->isAttribute("f1"));
    CPPUNIT_ASSERT(tds->isAttribute("w2"));	

    CPPUNIT_ASSERT_EQUAL(10,tds->getNPatterns());

    // tds->getTuple()->SetScanField(-1);
    // tds->scan("*");
    //Few entries
    //v1
    int nentries = tds->getTuple()->Draw("v1","1>0","goff"); double *entries = tds->getTuple()->GetV1();
    CPPUNIT_ASSERT_EQUAL(63,nentries);
    nentries = tds->getTuple()->Draw("v1",Form("foo__n__iter__==%i",0),"goff"); CPPUNIT_ASSERT_EQUAL(12, nentries);
    entries = tds->getTuple()->GetV1(); CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.06901933299, entries[0], DBL_PRECISION_COMPARE);
    nentries = tds->getTuple()->Draw("v1",Form("foo__n__iter__==%i",5),"goff"); CPPUNIT_ASSERT_EQUAL(8, nentries);
    nentries = tds->getTuple()->Draw("v1",Form("foo__n__iter__==%i",9),"goff"); CPPUNIT_ASSERT_EQUAL(5, nentries);
    entries = tds->getTuple()->GetV1(); CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.2397650483, entries[nentries-1], DBL_PRECISION_COMPARE);
  
    //v2
    nentries = tds->getTuple()->Draw("v2","1>0","goff"); entries = tds->getTuple()->GetV1();
    CPPUNIT_ASSERT_EQUAL(78,nentries);
    nentries = tds->getTuple()->Draw("v2",Form("foo__n__iter__==%i",0),"goff"); CPPUNIT_ASSERT_EQUAL(6, nentries);
    entries = tds->getTuple()->GetV1(); CPPUNIT_ASSERT_DOUBLES_EQUAL(11.42449011, entries[0], DBL_PRECISION_COMPARE);
    nentries = tds->getTuple()->Draw("v2",Form("foo__n__iter__==%i",5),"goff"); CPPUNIT_ASSERT_EQUAL(12, nentries);
    nentries = tds->getTuple()->Draw("v2",Form("foo__n__iter__==%i",9),"goff"); CPPUNIT_ASSERT_EQUAL(9, nentries);
    entries = tds->getTuple()->GetV1(); CPPUNIT_ASSERT_DOUBLES_EQUAL(8.552570119, entries[nentries-1], DBL_PRECISION_COMPARE);
  
    //f1
    tds->computeStatistic("f1");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(12.3091863421, tds->getAttribute("f1")->getStd(), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(47.9894841045, tds->getAttribute("f1")->getMean(), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(69.8054975231, tds->getAttribute("f1")->getMaximum(), DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(34.2563231678, tds->getAttribute("f1")->getMinimum(), DBL_PRECISION_COMPARE);

    //w1&w2
    tds->getTuple()->Draw("Entry$",Form("foo__n__iter__==%i",0),"goff"); entries = tds->getTuple()->GetV1();
    CPPUNIT_ASSERT("six"==tds->getString("w1",entries[0]));
    CPPUNIT_ASSERT("zero"==tds->getString("w2",entries[0]));
    tds->getTuple()->Draw("Entry$",Form("foo__n__iter__==%i",5),"goff"); entries = tds->getTuple()->GetV1();
    CPPUNIT_ASSERT("two"==tds->getString("w1",entries[0]));
    CPPUNIT_ASSERT("five"==tds->getString("w2",entries[0]));
    tds->getTuple()->Draw("Entry$",Form("foo__n__iter__==%i",9),"goff"); entries = tds->getTuple()->GetV1();
    CPPUNIT_ASSERT("four"==tds->getString("w1",entries[0]));
    CPPUNIT_ASSERT("six"==tds->getString("w2",entries[0]));
}

// // ===========================================================
// // ================ MultiType in Input =======================
// // ===========================================================
void TLauncher2Test::setUpMultiTypeInput()
{
    _tds1 = new URANIE::DataServer::TDataServer("foo", "multitype usecase");
    _tds1->fileDataRead("readmultitype_sampling.dat");

    _thev1 = new URANIE::DataServer::TAttribute("thev1");
    _thev2 = new URANIE::DataServer::TAttribute("thev2");
}

void TLauncher2Test::testMultiTypeAsInputKey()
{

    setUpMultiTypeInput();
    
    //Get example of input file
    gSystem->Exec((UranieTests::code_path("multitype") + " -mtKey").data());
    // Create the output files
    URANIE::Relauncher::TKeyScript inputFile("_output_multitype_mt_Key_condensate_.dat");
    inputFile.setInputs(5, _tds1->getAttribute("w1"), "w1", _tds1->getAttribute("v1"), "v1",
			_tds1->getAttribute("v2"), "v2", _tds1->getAttribute("f1"), "f1", _tds1->getAttribute("w2"), "w2");
    //  inputFile.setVectorProperties("[",",","]");
	
    // Create the output files
    URANIE::Relauncher::TKeyResult outputFile("_output_multitype_readmt_Key_.dat");
    outputFile.addOutput(_thev1, "thev1");
    outputFile.addOutput(_thev2, "thev2");
    
    // Create the user's evaluation function
    URANIE::Relauncher::TCodeEval eval((UranieTests::code_path("multitype") + " -ReadmtKey").data());
    eval.addInputFile(&inputFile);
    eval.addOutputFile(&outputFile);
    
    URANIE::Relauncher::TSequentialRun runner(&eval);

    runner.startSlave();                                           
    if (runner.onMaster())
    {
	URANIE::Relauncher::TLauncher2 lanceur(_tds1, &runner);
	lanceur.solverLoop();

	// Stop the slave processes
	runner.stopSlave();
    }        
  
    validateValuesReadmt( _tds1);
    
    
}

void TLauncher2Test::testMultiTypeAsInputFlat()
{

    setUpMultiTypeInput();

    // Create the output files
    URANIE::Relauncher::TFlatScript inputFile("_output_multitype_mt_DataServer_.dat");
    inputFile.setInputs(5, _tds1->getAttribute("w1"), _tds1->getAttribute("v1"), _tds1->getAttribute("v2"),
			_tds1->getAttribute("f1"), _tds1->getAttribute("w2"));

    // Create the output files
    URANIE::Relauncher::TFlatResult outputFile("_output_multitype_readmt_DataServer_.dat");
    outputFile.addOutput(_thev1);
    outputFile.addOutput(_thev2);

    // Create the user's evaluation function
    URANIE::Relauncher::TCodeEval eval((UranieTests::code_path("multitype") + " -ReadmtDS").data());
    eval.addInputFile(&inputFile);
    eval.addOutputFile(&outputFile);

    URANIE::Relauncher::TSequentialRun runner(&eval);

    runner.startSlave();                                           
    if (runner.onMaster())
    {
	URANIE::Relauncher::TLauncher2 lanceur(_tds1, &runner);
	lanceur.solverLoop();

	// Stop the slave processes
	runner.stopSlave();
    }        
  
    validateValuesReadmt( _tds1);
    
    
}

void TLauncher2Test::testMultiTypeAsInputFlat2()
{

    setUpMultiTypeInput();

    // Create the output files
    URANIE::Relauncher::TFlatScript inputFile("_output_multitype_mt_Flat_.dat");
    inputFile.setInputs(5, _tds1->getAttribute("w1"), _tds1->getAttribute("v1"), _tds1->getAttribute("v2"),
			_tds1->getAttribute("f1"), _tds1->getAttribute("w2"));

    // Create the output files
    URANIE::Relauncher::TFlatResult outputFile("_output_multitype_readmt_Flat_.dat");
    outputFile.addOutput(_thev1);
    outputFile.addOutput(_thev2);

    // Create the user's evaluation function
    URANIE::Relauncher::TCodeEval eval((UranieTests::code_path("multitype") + " -ReadmtFlat").data());
    eval.addInputFile(&inputFile);
    eval.addOutputFile(&outputFile);

    URANIE::Relauncher::TSequentialRun runner(&eval);

    runner.startSlave();                                           
    if (runner.onMaster())
    {
	URANIE::Relauncher::TLauncher2 lanceur(_tds1, &runner);
	lanceur.solverLoop();

	// Stop the slave processes
	runner.stopSlave();
    }        
  
    validateValuesReadmt( _tds1);
    
    
}

void TLauncher2Test::testMultiTypeAsInputLine()
{

    setUpMultiTypeInput();

    // Create the output files
    URANIE::Relauncher::TLineScript inputFile("_output_multitype_mt_Column_.dat");
    inputFile.setInputs(5, _tds1->getAttribute("w1"), _tds1->getAttribute("v1"), _tds1->getAttribute("v2"),
			_tds1->getAttribute("f1"), _tds1->getAttribute("w2"));

    // Create the output files
    URANIE::Relauncher::TLineResult outputFile("_output_multitype_readmt_Column_.dat");
    outputFile.addOutput(_thev1);
    outputFile.addOutput(_thev2);

    // Create the user's evaluation function
    URANIE::Relauncher::TCodeEval eval((UranieTests::code_path("multitype") + " -ReadmtCol").data());
    eval.addInputFile(&inputFile);
    eval.addOutputFile(&outputFile);

    URANIE::Relauncher::TSequentialRun runner(&eval);

    runner.startSlave();                                           
    if (runner.onMaster())
    {
	URANIE::Relauncher::TLauncher2 lanceur(_tds1, &runner);
	lanceur.solverLoop();

	// Stop the slave processes
	runner.stopSlave();
    }        
  
    validateValuesReadmt( _tds1);
      
}

void TLauncher2Test::testMultiTypeAsInputLineThreaded()
{

    setUpMultiTypeInput();

    // Create the output files
    URANIE::Relauncher::TLineScript inputFile("_output_multitype_mt_Column_.dat");
    inputFile.setInputs(5, _tds1->getAttribute("w1"), _tds1->getAttribute("v1"), _tds1->getAttribute("v2"),
			_tds1->getAttribute("f1"), _tds1->getAttribute("w2"));

    // Create the output files
    URANIE::Relauncher::TLineResult outputFile("_output_multitype_readmt_Column_.dat");
    outputFile.addOutput(_thev1);
    outputFile.addOutput(_thev2);

    // Create the user's evaluation function
    URANIE::Relauncher::TCodeEval eval((UranieTests::code_path("multitype") + " -ReadmtCol").data());
    eval.addInputFile(&inputFile);
    eval.addOutputFile(&outputFile);

    URANIE::Relauncher::TThreadedRun runner(&eval,4);

    runner.startSlave();                                           
    if (runner.onMaster())
    {
	URANIE::Relauncher::TLauncher2 lanceur(_tds1, &runner);
	lanceur.solverLoop();

	// Stop the slave processes
	runner.stopSlave();
    }        
  
    validateValuesReadmt( _tds1);
    
    
}

void TLauncher2Test::testMultiTypeAsInputCJit()
{

    setUpMultiTypeInput();

    // Create the user's evaluation function
    URANIE::Relauncher::TCJitEval eval(FlowrateReatmt_hard);
    eval.setInputs(4, _tds1->getAttribute("v1"), _tds1->getAttribute("w1"), _tds1->getAttribute("v2"), _tds1->getAttribute("w2"));
    eval.addOutput(_thev1);
    eval.addOutput(_thev2);
    
    URANIE::Relauncher::TSequentialRun runner(&eval);

    runner.startSlave();                                           
    if (runner.onMaster())
    {
	URANIE::Relauncher::TLauncher2 lanceur(_tds1, &runner);
	lanceur.solverLoop();

	// Stop the slave processes
	runner.stopSlave();
    }        
  
    validateValuesReadmt( _tds1);        
}

void TLauncher2Test::testLambdaMultiTypeAsInputCJit()
{

    setUpMultiTypeInput();

    double thev1=987654321, thev2=987654321;
    string words[10]={"zero","one","two","three","four","five","six","seven","eight","nine"};

    std::function<int(vector<URANIE::DataServer::UEntry*>*,vector<URANIE::DataServer::UEntry*>*) >
        MyFlowrateMt = [&](vector<URANIE::DataServer::UEntry*> *in,vector<URANIE::DataServer::UEntry*> *out)
    {
        vector<double> *v1 = in->at(0)->val.vec, *v2 = in->at(2)->val.vec;
        string *w1=in->at(1)->val.str, *w2=in->at(3)->val.str;

        for(unsigned int iday=0; iday<10; iday++) {
            if( *w1 == words[iday] )
                thev1 = ( (iday<v1->size()) ? v1->at(iday) : -123456789 );
            if( *w2 == words[iday] )
                thev2 = ( (iday<v2->size()) ? v2->at(iday) : -123456789 );
        }
        out->at(0)->pushEntry(thev1);
        out->at(1)->pushEntry(thev2);

        return 1;
    };

    // Create the user's evaluation function
    URANIE::Relauncher::TCJitEval eval(&MyFlowrateMt);
    eval.setInputs(4, _tds1->getAttribute("v1"), _tds1->getAttribute("w1"), _tds1->getAttribute("v2"), _tds1->getAttribute("w2"));
    eval.addOutput(_thev1);
    eval.addOutput(_thev2);

    URANIE::Relauncher::TSequentialRun runner(&eval);

    runner.startSlave();
    if (runner.onMaster())
    {
        URANIE::Relauncher::TLauncher2 lanceur(_tds1, &runner);
        lanceur.solverLoop();

        // Stop the slave processes
        runner.stopSlave();
    }

    validateValuesReadmt( _tds1);
}

void TLauncher2Test::testMultiTypeAsInputCJitWithConstantVec()
{

    setUpMultiTypeInput();

    delete _tds1;
    _tds1 = new URANIE::DataServer::TDataServer("foo", "multitype usecase");
    _tds1->fileDataRead("readmultitype_sampling_nov2.dat");

    _v2 = new URANIE::DataServer::TAttribute("v2", URANIE::DataServer::TAttribute::kVector);
        
    // Create the user's evaluation function
    URANIE::Relauncher::TCJitEval eval(FlowrateReatmt_hard);
    eval.setInputs(4, _tds1->getAttribute("v1"), _tds1->getAttribute("w1"), _v2, _tds1->getAttribute("w2"));
    eval.addOutput(_thev1);
    eval.addOutput(_thev2);

    URANIE::DataServer::UEntry *tmpConst = new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Vector, "tmpv2");
    for(unsigned int ii=1; ii<=6; ii++)
        tmpConst->pushEntry(double(ii));
    
    URANIE::Relauncher::TSequentialRun runner(&eval);
    runner.startSlave();                                           
    if (runner.onMaster())
    {
	URANIE::Relauncher::TLauncher2 lanceur(_tds1, &runner);
        lanceur.addConstantValue(_v2, tmpConst, true);
	lanceur.solverLoop();

	// Stop the slave processes
	runner.stopSlave();
    }        
  
    validateValuesReadmt( _tds1, true);

    delete tmpConst ;
}

void TLauncher2Test::testMultiTypeAsInputCJitThreaded()
{

    setUpMultiTypeInput();

    // Create the user's evaluation function
    URANIE::Relauncher::TCJitEval eval(FlowrateReatmt_hard);
    eval.setInputs(4, _tds1->getAttribute("v1"), _tds1->getAttribute("w1"), _tds1->getAttribute("v2"), _tds1->getAttribute("w2"));
    eval.addOutput(_thev1);
    eval.addOutput(_thev2);
    
    URANIE::Relauncher::TThreadedRun runner(&eval,4);

    runner.startSlave();                                           
    if (runner.onMaster())
    {
	URANIE::Relauncher::TLauncher2 lanceur(_tds1, &runner);
	lanceur.solverLoop();

	// Stop the slave processes
	runner.stopSlave();
    }        
  
    validateValuesReadmt( _tds1);        
}


void TLauncher2Test::testFillMultiTypeFormat()
{
    
#ifdef WIN32
    URANIE_IGNORED_TEST() ; // No equivalent of sed in windows... Removed for this specific case
#endif
    
    std::string filename("_output_readmultitype_sampling.dat");
    std::ofstream fstr(filename.c_str());
    fstr <<"#NAME: foo\n#TITLE: TDS for flowrate\n#DATE: Fri Sep 23 18:18:22 2016\n#COLUMN_NAMES: v1| w1| v2| w2| f1| foo__n__iter__\n#COLUMN_TYPES: V|S|V|S|D|D\n\n";
    fstr <<"-6.901933299e-02,-1.292435960e-01,4.558876683e-01,5.638486369e-01,-4.767582767e-02,7.102109543e-03,2.819049678e-01,-2.019788082e+00,-2.604401029e+00,-1.617682380e+00,2.894560950e-02,-3.493905850e-01 six 1.142449011e+01,7.318948216e+00,1.502260859e+01,6.041193793e+00,6.729445146e+00,1.128096969e+01 zero 3.425632317e+01 0.000000000e+00";
    fstr.close();
    
    std::string filename2("_Stupid_Flag_.dat");
    std::ofstream fstr2(filename2.c_str());
    fstr2 <<"ldkvhj azdejfaevhdv lkcdnlkdelkvaerkjvn\n\nmljs flkandvjandfvlajev adkjvb @W1@ ldk nlkvnmkdj  of\n\nvkjnla dvalkjevn @V1@  sdlkfj alvfdlknldf nlfn qdsklnv \nlkj dv @F1@\n\n lsdkvh dlvadjnlkf\nlkfhe jhe @W2@ \n skhv akjdvh akjvha kjdvajkdvhadvlh elkjv aldkvh \n lkfhmlakzefhmalzekhjmlkah @V2@\n";
    fstr2.close();
    
    std::string filename3("_Stupid_Key_.dat");
    std::ofstream fstr3(filename3.c_str());
    fstr3 <<"ldkvhj azdejfaevhdv lkcdnlkdelkvaerkjvn\n\nmljs flkandvjandfvlajev adkjvb w1=@Toto) ldk nlkvnmkdj  of\n\nvkjnla dvalkjevn v1=BBBB1.2AAAA  sdlkfj alvfdlknldf nlfn qdsklnv \nlkj dv f1=4.2\n\n lsdkvh dlvadjnlkf\nlkfhe jhe w2=@Tutu) \n skhv akjdvh akjvha kjdvajkdvhadvlh elkjv aldkvh \n lkfhmlakzefhmalzekhjmlkah v2=BBBB1.2AAAA\n";
    fstr3.close();

    _tds1 = new URANIE::DataServer::TDataServer("foo", "multitype usecase");
    _tds1->fileDataRead(filename.c_str());

    _thev1 = new URANIE::DataServer::TAttribute("thev1");
    _thev2 = new URANIE::DataServer::TAttribute("thev2");

    // Create the output files
    URANIE::Relauncher::TLineScript StupidFile1("_Stupid_Line_.dat");
    StupidFile1.setInputs(5, _tds1->getAttribute("w1"), _tds1->getAttribute("v1"), _tds1->getAttribute("v2"),
                          _tds1->getAttribute("f1"), _tds1->getAttribute("w2"));
    StupidFile1.setVectorProperties("BBBB","|","AAAA");
    StupidFile1.setStringProperties("@",")");
    //    StupidFile1.setFormat("%e");
    URANIE::Relauncher::TFlatScript StupidFile2("_Stupid_Flat_.dat");
    StupidFile2.setInputs(5, _tds1->getAttribute("w1"), _tds1->getAttribute("v1"), _tds1->getAttribute("v2"),
                          _tds1->getAttribute("f1"), _tds1->getAttribute("w2"));
    StupidFile2.setVectorProperties("BBBB","|","AAAA");
    StupidFile2.setStringProperties("@",")");
    //    StupidFile2.setFormat("%e");
    URANIE::Relauncher::TKeyScript StupidFile3("_Stupid_Key_.dat");
    StupidFile3.setInputs(5, _tds1->getAttribute("w1"), "w1", _tds1->getAttribute("v1"), "v1", _tds1->getAttribute("v2"), "v2",
                          _tds1->getAttribute("f1"), "f1", _tds1->getAttribute("w2"), "w2");
    StupidFile3.setVectorProperties("BBBB","|","AAAA");
    StupidFile3.setStringProperties("@",")");
    //    StupidFile3.setFormat("%e");
    URANIE::Relauncher::TFlagScript StupidFile4("_Stupid_Flag_.dat");
    StupidFile4.setInputs(5, _tds1->getAttribute("w1"), "@W1@", _tds1->getAttribute("v1"), "@V1@", _tds1->getAttribute("v2"), "@V2@",
                          _tds1->getAttribute("f1"), "@F1@", _tds1->getAttribute("w2"), "@W2@");
    StupidFile4.setVectorProperties("BBBB","|","AAAA");
    StupidFile4.setStringProperties("@",")");
    //    StupidFile4.setFormat("%e");
    // Create the output files
    URANIE::Relauncher::TLineResult outputFile("_output_multitype_readmt_Column_.dat");
    outputFile.addOutput(_thev1);
    outputFile.addOutput(_thev2);

    // Create the user's evaluation function

    URANIE::Relauncher::TCodeEval eval(
	Form("%s/../../../launCHER/teST/coDE/multiTYPE/multitype -ReadmtCol && mv _Stupid*.dat ../",
	     gSystem->pwd()));

    eval.addInputFile(&StupidFile1);
    eval.addInputFile(&StupidFile2);
    eval.addInputFile(&StupidFile3);
    eval.addInputFile(&StupidFile4);
    eval.addOutputFile(&outputFile);

    URANIE::Relauncher::TSequentialRun runner(&eval);
    runner.startSlave();                                           
    if (runner.onMaster())
    {
        URANIE::Relauncher::TLauncher2 lanceur(_tds1, &runner);
        lanceur.solverLoop();

        // Stop the slave processes
        runner.stopSlave();
    }        
    
    std::ofstream ref1("_Stupid_Flag_.dat.ref");
    ref1<<"ldkvhj azdejfaevhdv lkcdnlkdelkvaerkjvn\n\nmljs flkandvjandfvlajev adkjvb @six) ldk nlkvnmkdj  of\n\nvkjnla dvalkjevn BBBB-0.06901933299|-0.129243596|0.4558876683|0.5638486369|-0.04767582767|0.007102109543|0.2819049678|-2.019788082|-2.604401029|-1.61768238|0.0289456095|-0.349390585AAAA  sdlkfj alvfdlknldf nlfn qdsklnv \nlkj dv 34.25632317\n\n lsdkvh dlvadjnlkf\nlkfhe jhe @zero) \n skhv akjdvh akjvha kjdvajkdvhadvlh elkjv aldkvh \n lkfhmlakzefhmalzekhjmlkah BBBB11.42449011|7.318948216|15.02260859|6.041193793|6.729445146|11.28096969AAAA\n";
    ref1.close();
    
    std::ofstream ref2("_Stupid_Flat_.dat.ref");
    ref2<<" @six) BBBB-0.06901933299|-0.129243596|0.4558876683|0.5638486369|-0.04767582767|0.007102109543|0.2819049678|-2.019788082|-2.604401029|-1.61768238|0.0289456095|-0.349390585AAAA BBBB11.42449011|7.318948216|15.02260859|6.041193793|6.729445146|11.28096969AAAA 34.25632317 @zero)\n";
    ref2.close();

    
    std::ofstream ref3("_Stupid_Key_.dat.ref");
    ref3<<"ldkvhj azdejfaevhdv lkcdnlkdelkvaerkjvn\n\nmljs flkandvjandfvlajev adkjvb w1=@six) ldk nlkvnmkdj  of\n\nvkjnla dvalkjevn v1=BBBB-0.06901933299|-0.129243596|0.4558876683|0.5638486369|-0.04767582767|0.007102109543|0.2819049678|-2.019788082|-2.604401029|-1.61768238|0.0289456095|-0.349390585AAAA  sdlkfj alvfdlknldf nlfn qdsklnv \nlkj dv f1=34.25632317\n\n lsdkvh dlvadjnlkf\nlkfhe jhe w2=@zero) \n skhv akjdvh akjvha kjdvajkdvhadvlh elkjv aldkvh \n lkfhmlakzefhmalzekhjmlkah v2=BBBB11.42449011|7.318948216|15.02260859|6.041193793|6.729445146|11.28096969AAAA\n";
    ref3.close();
    
    std::ofstream ref4("_Stupid_Line_.dat.ref");
    ref4<<"@six)\nBBBB-0.06901933299|-0.129243596|0.4558876683|0.5638486369|-0.04767582767|0.007102109543|0.2819049678|-2.019788082|-2.604401029|-1.61768238|0.0289456095|-0.349390585AAAA\nBBBB11.42449011|7.318948216|15.02260859|6.041193793|6.729445146|11.28096969AAAA\n34.25632317\n@zero)\n\n";
    ref4.close();


    Int_t nret = UranieTests::areFilesEqual(TString("_Stupid_Flat_.dat"), TString("_Stupid_Flat_.dat.ref"));
    CPPUNIT_ASSERT_EQUAL(nret, 0);

    nret = UranieTests::areFilesEqual(TString("_Stupid_Flag_.dat"), TString("_Stupid_Flag_.dat.ref"));
    CPPUNIT_ASSERT_EQUAL(nret, 0);

    nret = UranieTests::areFilesEqual(TString("_Stupid_Line_.dat"), TString("_Stupid_Line_.dat.ref"));
    CPPUNIT_ASSERT_EQUAL(nret, 0);

    nret = UranieTests::areFilesEqual(TString("_Stupid_Key_.dat"), TString("_Stupid_Key_.dat.ref"));
    CPPUNIT_ASSERT_EQUAL(nret, 0);


    // UranieTests::removeFile(_workingTestPath.Data(),"_Stupid_*");

    
}

void TLauncher2Test::testFillMultiTypeFormatScientific()
{
    
#ifdef WIN32
    URANIE_IGNORED_TEST() ; // No equivalent of sed in windows... Removed for this specific case
#endif
    
    std::string filename("_output_readmultitype_sampling.dat");
    std::ofstream fstr(filename.c_str());
    fstr <<"#NAME: foo\n#TITLE: TDS for flowrate\n#DATE: Fri Sep 23 18:18:22 2016\n#COLUMN_NAMES: v1| w1| v2| w2| f1| foo__n__iter__\n#COLUMN_TYPES: V|S|V|S|D|D\n\n";
    fstr <<"-6.901933299e-02,-1.292435960e-01,4.558876683e-01,5.638486369e-01,-4.767582767e-02,7.102109543e-03,2.819049678e-01,-2.019788082e+00,-2.604401029e+00,-1.617682380e+00,2.894560950e-02,-3.493905850e-01 six 1.142449011e+01,7.318948216e+00,1.502260859e+01,6.041193793e+00,6.729445146e+00,1.128096969e+01 zero 3.425632317e+01 0.000000000e+00";
    fstr.close();
    
    std::string filename2("_Stupid_Flag_.dat");
    std::ofstream fstr2(filename2.c_str());
    fstr2 <<"ldkvhj azdejfaevhdv lkcdnlkdelkvaerkjvn\n\nmljs flkandvjandfvlajev adkjvb @W1@ ldk nlkvnmkdj  of\n\nvkjnla dvalkjevn @V1@  sdlkfj alvfdlknldf nlfn qdsklnv \nlkj dv @F1@\n\n lsdkvh dlvadjnlkf\nlkfhe jhe @W2@ \n skhv akjdvh akjvha kjdvajkdvhadvlh elkjv aldkvh \n lkfhmlakzefhmalzekhjmlkah @V2@\n";
    fstr2.close();
    
    std::string filename3("_Stupid_Key_.dat");
    std::ofstream fstr3(filename3.c_str());
    fstr3 <<"ldkvhj azdejfaevhdv lkcdnlkdelkvaerkjvn\n\nmljs flkandvjandfvlajev adkjvb w1=@Toto) ldk nlkvnmkdj  of\n\nvkjnla dvalkjevn v1=BBBB1.2AAAA  sdlkfj alvfdlknldf nlfn qdsklnv \nlkj dv f1=4.2\n\n lsdkvh dlvadjnlkf\nlkfhe jhe w2=@Tutu) \n skhv akjdvh akjvha kjdvajkdvhadvlh elkjv aldkvh \n lkfhmlakzefhmalzekhjmlkah v2=BBBB1.2AAAA\n";
    fstr3.close();

    _tds1 = new URANIE::DataServer::TDataServer("foo", "multitype usecase");
    _tds1->fileDataRead(filename.c_str());

    _thev1 = new URANIE::DataServer::TAttribute("thev1");
    _thev2 = new URANIE::DataServer::TAttribute("thev2");

    // Create the output files
    URANIE::Relauncher::TLineScript StupidFile1("_Stupid_Line_.dat");
    StupidFile1.setInputs(5, _tds1->getAttribute("w1"), _tds1->getAttribute("v1"), _tds1->getAttribute("v2"),
                          _tds1->getAttribute("f1"), _tds1->getAttribute("w2"));
    StupidFile1.setVectorProperties("BBBB","|","AAAA");
    StupidFile1.setStringProperties("@",")");
    StupidFile1.setFormat("%e");
    URANIE::Relauncher::TFlatScript StupidFile2("_Stupid_Flat_.dat");
    StupidFile2.setInputs(5, _tds1->getAttribute("w1"), _tds1->getAttribute("v1"), _tds1->getAttribute("v2"),
                          _tds1->getAttribute("f1"), _tds1->getAttribute("w2"));
    StupidFile2.setVectorProperties("BBBB","|","AAAA");
    StupidFile2.setStringProperties("@",")");
    StupidFile2.setFormat("%e");
    URANIE::Relauncher::TKeyScript StupidFile3("_Stupid_Key_.dat");
    StupidFile3.setInputs(5, _tds1->getAttribute("w1"), "w1", _tds1->getAttribute("v1"), "v1", _tds1->getAttribute("v2"), "v2",
                          _tds1->getAttribute("f1"), "f1", _tds1->getAttribute("w2"), "w2");
    StupidFile3.setVectorProperties("BBBB","|","AAAA");
    StupidFile3.setStringProperties("@",")");
    StupidFile3.setFormat("%e");
    URANIE::Relauncher::TFlagScript StupidFile4("_Stupid_Flag_.dat");
    StupidFile4.setInputs(5, _tds1->getAttribute("w1"), "@W1@", _tds1->getAttribute("v1"), "@V1@", _tds1->getAttribute("v2"), "@V2@",
                          _tds1->getAttribute("f1"), "@F1@", _tds1->getAttribute("w2"), "@W2@");
    StupidFile4.setVectorProperties("BBBB","|","AAAA");
    StupidFile4.setStringProperties("@",")");
    StupidFile4.setFormat("%e");
    // Create the output files
    URANIE::Relauncher::TLineResult outputFile("_output_multitype_readmt_Column_.dat");
    outputFile.addOutput(_thev1);
    outputFile.addOutput(_thev2);

    // Create the user's evaluation function

    URANIE::Relauncher::TCodeEval eval(
	Form("%s/../../../launCHER/teST/coDE/multiTYPE/multitype -ReadmtCol && mv _Stupid*.dat ../",
	     gSystem->pwd()));

    eval.addInputFile(&StupidFile1);
    eval.addInputFile(&StupidFile2);
    eval.addInputFile(&StupidFile3);
    eval.addInputFile(&StupidFile4);
    eval.addOutputFile(&outputFile);

    URANIE::Relauncher::TSequentialRun runner(&eval);
    runner.startSlave();                                           
    if (runner.onMaster())
    {
        URANIE::Relauncher::TLauncher2 lanceur(_tds1, &runner);
        lanceur.solverLoop();

        // Stop the slave processes
        runner.stopSlave();
    }        
    
    std::ofstream ref1("_Stupid_Flag_.dat.ref");
    ref1<<"ldkvhj azdejfaevhdv lkcdnlkdelkvaerkjvn\n\nmljs flkandvjandfvlajev adkjvb @six) ldk nlkvnmkdj  of\n\nvkjnla dvalkjevn BBBB-6.901933299000e-02|-1.292435960000e-01|4.558876683000e-01|5.638486369000e-01|-4.767582767000e-02|7.102109543000e-03|2.819049678000e-01|-2.019788082000e+00|-2.604401029000e+00|-1.617682380000e+00|2.894560950000e-02|-3.493905850000e-01AAAA  sdlkfj alvfdlknldf nlfn qdsklnv \nlkj dv 3.425632317000e+01\n\n lsdkvh dlvadjnlkf\nlkfhe jhe @zero) \n skhv akjdvh akjvha kjdvajkdvhadvlh elkjv aldkvh \n lkfhmlakzefhmalzekhjmlkah BBBB1.142449011000e+01|7.318948216000e+00|1.502260859000e+01|6.041193793000e+00|6.729445146000e+00|1.128096969000e+01AAAA\n";
    ref1.close();
    
    std::ofstream ref2("_Stupid_Flat_.dat.ref");
    ref2<<" @six) BBBB-6.901933299000e-02|-1.292435960000e-01|4.558876683000e-01|5.638486369000e-01|-4.767582767000e-02|7.102109543000e-03|2.819049678000e-01|-2.019788082000e+00|-2.604401029000e+00|-1.617682380000e+00|2.894560950000e-02|-3.493905850000e-01AAAA BBBB1.142449011000e+01|7.318948216000e+00|1.502260859000e+01|6.041193793000e+00|6.729445146000e+00|1.128096969000e+01AAAA 3.425632317000e+01 @zero)\n";
    ref2.close();

    
    std::ofstream ref3("_Stupid_Key_.dat.ref");
    ref3<<"ldkvhj azdejfaevhdv lkcdnlkdelkvaerkjvn\n\nmljs flkandvjandfvlajev adkjvb w1=@six) ldk nlkvnmkdj  of\n\nvkjnla dvalkjevn v1=BBBB-6.901933299000e-02|-1.292435960000e-01|4.558876683000e-01|5.638486369000e-01|-4.767582767000e-02|7.102109543000e-03|2.819049678000e-01|-2.019788082000e+00|-2.604401029000e+00|-1.617682380000e+00|2.894560950000e-02|-3.493905850000e-01AAAA  sdlkfj alvfdlknldf nlfn qdsklnv \nlkj dv f1=3.425632317000e+01\n\n lsdkvh dlvadjnlkf\nlkfhe jhe w2=@zero) \n skhv akjdvh akjvha kjdvajkdvhadvlh elkjv aldkvh \n lkfhmlakzefhmalzekhjmlkah v2=BBBB1.142449011000e+01|7.318948216000e+00|1.502260859000e+01|6.041193793000e+00|6.729445146000e+00|1.128096969000e+01AAAA\n";
    ref3.close();
    
    std::ofstream ref4("_Stupid_Line_.dat.ref");
    ref4<<"@six)\nBBBB-6.901933299000e-02|-1.292435960000e-01|4.558876683000e-01|5.638486369000e-01|-4.767582767000e-02|7.102109543000e-03|2.819049678000e-01|-2.019788082000e+00|-2.604401029000e+00|-1.617682380000e+00|2.894560950000e-02|-3.493905850000e-01AAAA\nBBBB1.142449011000e+01|7.318948216000e+00|1.502260859000e+01|6.041193793000e+00|6.729445146000e+00|1.128096969000e+01AAAA\n3.425632317000e+01\n@zero)\n\n";
    ref4.close();


    Int_t nret = UranieTests::areFilesEqual(TString("_Stupid_Flat_.dat"), TString("_Stupid_Flat_.dat.ref"));
    CPPUNIT_ASSERT_EQUAL(nret, 0);

    nret = UranieTests::areFilesEqual(TString("_Stupid_Flag_.dat"), TString("_Stupid_Flag_.dat.ref"));
    CPPUNIT_ASSERT_EQUAL(nret, 0);

    nret = UranieTests::areFilesEqual(TString("_Stupid_Line_.dat"), TString("_Stupid_Line_.dat.ref"));
    CPPUNIT_ASSERT_EQUAL(nret, 0);

    nret = UranieTests::areFilesEqual(TString("_Stupid_Key_.dat"), TString("_Stupid_Key_.dat.ref"));
    CPPUNIT_ASSERT_EQUAL(nret, 0);
    

    UranieTests::removeFile(_workingTestPath.Data(),"_Stupid_*");
}



// // ===========================================================
// // ================= MultiType Composition ===================
// // ===========================================================
void TLauncher2Test::setUpMultiTypeCompo()
{
    setUpMultiType();
    _thev1 = new URANIE::DataServer::TAttribute("thev1");
    _thev2 = new URANIE::DataServer::TAttribute("thev2");
    
}

void TLauncher2Test::testMultiTypeOutputFileKeyCompoKey()
{
    setUpMultiTypeCompo();

    //Get example of input file
    gSystem->Exec((UranieTests::code_path("multitype") + " -mtKey").data());
    
    // Create the input files
    URANIE::Relauncher::TFlatScript inputFile1("multitype_input.dat");
    inputFile1.setInputs(1, _tds1->getAttribute("seed"), "seed");

    // Create the output files
    URANIE::Relauncher::TKeyResult outputFile1("_output_multitype_mt_Key_.dat");
    outputFile1.addOutput(_w1, "w1");
    outputFile1.addOutput(_v1, "v1");
    outputFile1.addOutput(_v2, "v2");
    outputFile1.addOutput(_f1, "f1");
    outputFile1.addOutput(_w2, "w2");

    // Create the user's evaluation function
    URANIE::Relauncher::TCodeEval eval1((UranieTests::code_path("multitype") + " -mtKey").data());
    eval1.addInputFile(&inputFile1);
    eval1.addOutputFile(&outputFile1);

    // Create the output files
    URANIE::Relauncher::TKeyScript inputFile2("_output_multitype_mt_Key_condensate_.dat");
    inputFile2.setInputs(5, _w1, "w1", _v1, "v1", _v2, "v2", _f1, "f1", _w2, "w2");
    //inputFile2.setVectorProperties("[",",","]");
    
    // Create the output files
    URANIE::Relauncher::TKeyResult outputFile2("_output_multitype_readmt_Key_.dat");
    outputFile2.addOutput(_thev1, "thev1");
    outputFile2.addOutput(_thev2, "thev2");

    // Create the user's evaluation function
    URANIE::Relauncher::TCodeEval eval2((UranieTests::code_path("multitype") + " -ReadmtKey").data());
    eval2.addInputFile(&inputFile2);
    eval2.addOutputFile(&outputFile2);

    URANIE::Relauncher::TComposeEval eval;
    eval.addEval(&eval1);
    eval.addEval(&eval2);  

    URANIE::Relauncher::TSequentialRun runner(&eval);
    
    runner.startSlave();                                           
    if (runner.onMaster())
    {      
	URANIE::Relauncher::TLauncher2 lanceur(_tds1, &runner);
	lanceur.solverLoop();
        
	// Stop the slave processes
	runner.stopSlave();
    }        
  
    validateOutputMultiTypeResults(_tds1);
    validateValuesReadmt( _tds1);
}

void TLauncher2Test::testMultiTypeOutputFileFlatCompoFlat()
{
    setUpMultiTypeCompo();

    // Create the input files
    URANIE::Relauncher::TFlatScript inputFile1("multitype_input.dat");
    inputFile1.setInputs(1, _tds1->getAttribute("seed"), "seed");

    // Create the output files
    URANIE::Relauncher::TFlatResult outputFile1("_output_multitype_mt_DataServer_.dat");
    outputFile1.addOutput(_w1);
    outputFile1.addOutput(_v1);
    outputFile1.addOutput(_v2);
    outputFile1.addOutput(_f1);
    outputFile1.addOutput(_w2);

    // Create the user's evaluation function
    URANIE::Relauncher::TCodeEval eval1((UranieTests::code_path("multitype") + " -mtDS").data());
    eval1.addInputFile(&inputFile1);
    eval1.addOutputFile(&outputFile1);

    // Create the output files
    URANIE::Relauncher::TFlatScript inputFile2("_output_multitype_mt_DataServer_.dat");
    inputFile2.setInputs(5, _w1, _v1,_v2, _f1, _w2);
    
    // Create the output files
    URANIE::Relauncher::TFlatResult outputFile2("_output_multitype_readmt_DataServer_.dat");
    outputFile2.addOutput(_thev1);
    outputFile2.addOutput(_thev2);

    // Create the user's evaluation function
    URANIE::Relauncher::TCodeEval eval2((UranieTests::code_path("multitype") + " -ReadmtDS").data());
    eval2.addInputFile(&inputFile2);
    eval2.addOutputFile(&outputFile2);

    URANIE::Relauncher::TComposeEval eval;
    eval.addEval(&eval1);
    eval.addEval(&eval2);  

    URANIE::Relauncher::TSequentialRun runner(&eval);
    
    runner.startSlave();                                           
    if (runner.onMaster())
    {      
	URANIE::Relauncher::TLauncher2 lanceur(_tds1, &runner);
	lanceur.solverLoop();
        
	// Stop the slave processes
	runner.stopSlave();
    }        
  
    validateOutputMultiTypeResults(_tds1);
    validateValuesReadmt( _tds1);
}

void TLauncher2Test::testMultiTypeOutputFileFlat2CompoFlat2()
{
    setUpMultiTypeCompo();

    // Create the input files
    URANIE::Relauncher::TFlatScript inputFile1("multitype_input.dat");
    inputFile1.setInputs(1, _tds1->getAttribute("seed"), "seed");

    // Create the output files
    URANIE::Relauncher::TFlatResult outputFile1("_output_multitype_mt_Flat_.dat");
    outputFile1.addOutput(_w1);
    outputFile1.addOutput(_v1);
    outputFile1.addOutput(_v2);
    outputFile1.addOutput(_f1);
    outputFile1.addOutput(_w2);
    outputFile1.setVectorProperties("[",",","]");

    // Create the user's evaluation function
    URANIE::Relauncher::TCodeEval eval1((UranieTests::code_path("multitype") + " -mtFlat").data());
    eval1.addInputFile(&inputFile1);
    eval1.addOutputFile(&outputFile1);

    // Create the output files
    URANIE::Relauncher::TFlatScript inputFile2("_output_multitype_mt_Flat_.dat");
    inputFile2.setInputs(5, _w1, _v1,_v2, _f1, _w2);
    inputFile2.setVectorProperties("[",",","]");

    // Create the output files
    URANIE::Relauncher::TFlatResult outputFile2("_output_multitype_readmt_Flat_.dat");
    outputFile2.addOutput(_thev1);
    outputFile2.addOutput(_thev2);

    // Create the user's evaluation function
    URANIE::Relauncher::TCodeEval eval2((UranieTests::code_path("multitype") + " -ReadmtFlat").data());
    eval2.addInputFile(&inputFile2);
    eval2.addOutputFile(&outputFile2);

    URANIE::Relauncher::TComposeEval eval;
    eval.addEval(&eval1);
    eval.addEval(&eval2);  

    URANIE::Relauncher::TSequentialRun runner(&eval);
    
    runner.startSlave();                                           
    if (runner.onMaster())
    {      
	URANIE::Relauncher::TLauncher2 lanceur(_tds1, &runner);
	lanceur.solverLoop();
        
	// Stop the slave processes
	runner.stopSlave();
    }        
  
    validateOutputMultiTypeResults(_tds1);
    validateValuesReadmt( _tds1);
}

void TLauncher2Test::testMultiTypeOutputFileFlatMultiLineCompoFlat()
{
    setUpMultiTypeCompo();

    // Create the input files
    URANIE::Relauncher::TFlatScript inputFile1("multitype_input.dat");
    inputFile1.setInputs(1, _tds1->getAttribute("seed"), "seed");

    // Create the output files
    URANIE::Relauncher::TFlatResult outputFile1("_output_multitype_mt_Row_.dat");
    outputFile1.addOutput(_w1);
    outputFile1.addOutput(_v1);
    outputFile1.addOutput(_v2);
    outputFile1.addOutput(_f1);
    outputFile1.addOutput(_w2);
    outputFile1.isMultiLine(";");

    // Create the user's evaluation function
    URANIE::Relauncher::TCodeEval eval1((UranieTests::code_path("multitype") + " -mtRow").data());
    eval1.addInputFile(&inputFile1);
    eval1.addOutputFile(&outputFile1);

    // Create the output files
    URANIE::Relauncher::TFlatScript inputFile2("_output_multitype_mt_DataServer_.dat");
    inputFile2.setInputs(5, _w1, _v1,_v2, _f1, _w2);
    
    // Create the output files
    URANIE::Relauncher::TFlatResult outputFile2("_output_multitype_readmt_DataServer_.dat");
    outputFile2.addOutput(_thev1);
    outputFile2.addOutput(_thev2);

    // Create the user's evaluation function
    URANIE::Relauncher::TCodeEval eval2((UranieTests::code_path("multitype") + " -ReadmtDS").data());
    eval2.addInputFile(&inputFile2);
    eval2.addOutputFile(&outputFile2);

    URANIE::Relauncher::TComposeEval eval;
    eval.addEval(&eval1);
    eval.addEval(&eval2);  

    URANIE::Relauncher::TSequentialRun runner(&eval);
    
    runner.startSlave();                                           
    if (runner.onMaster())
    {      
	URANIE::Relauncher::TLauncher2 lanceur(_tds1, &runner);
	lanceur.solverLoop();
        
	// Stop the slave processes
	runner.stopSlave();
    }        
  
    validateOutputMultiTypeResults(_tds1);
    validateValuesReadmt( _tds1);
}

void TLauncher2Test::testMultiTypeOutputFileLineCompoLine()
{
    setUpMultiTypeCompo();

    // Create the input files
    URANIE::Relauncher::TFlatScript inputFile1("multitype_input.dat");
    inputFile1.setInputs(1, _tds1->getAttribute("seed"), "seed");

    // Create the output files
    URANIE::Relauncher::TLineResult outputFile1("_output_multitype_mt_Column_.dat");
    outputFile1.addOutput(_w1);
    outputFile1.addOutput(_v1);
    outputFile1.addOutput(_v2);
    outputFile1.addOutput(_f1);
    outputFile1.addOutput(_w2);

    // Create the user's evaluation function
    URANIE::Relauncher::TCodeEval eval1((UranieTests::code_path("multitype") + " -mtCol").data());
    eval1.addInputFile(&inputFile1);
    eval1.addOutputFile(&outputFile1);

    // Create the output files
    URANIE::Relauncher::TLineScript inputFile2("_output_multitype_mt_Column_.dat");
    inputFile2.setInputs(5, _w1, _v1,_v2, _f1, _w2);
    
    // Create the output files
    URANIE::Relauncher::TLineResult outputFile2("_output_multitype_readmt_Column_.dat");
    outputFile2.addOutput(_thev1);
    outputFile2.addOutput(_thev2);

    // Create the user's evaluation function
    URANIE::Relauncher::TCodeEval eval2((UranieTests::code_path("multitype") + " -ReadmtCol").data());
    eval2.addInputFile(&inputFile2);
    eval2.addOutputFile(&outputFile2);

    URANIE::Relauncher::TComposeEval eval;
    eval.addEval(&eval1);
    eval.addEval(&eval2);  

    URANIE::Relauncher::TSequentialRun runner(&eval);
    
    runner.startSlave();                                           
    if (runner.onMaster())
    {      
	URANIE::Relauncher::TLauncher2 lanceur(_tds1, &runner);
	lanceur.solverLoop();
        
	// Stop the slave processes
	runner.stopSlave();
    }        
  
    validateOutputMultiTypeResults(_tds1);
    validateValuesReadmt( _tds1);
}

void TLauncher2Test::testMultiTypeOutputFileKeyCompoKeyThreaded()
{
    setUpMultiTypeCompo();

    //Get example of input file
    gSystem->Exec((UranieTests::code_path("multitype") + " -mtKey").data());
    
    // Create the input files
    URANIE::Relauncher::TFlatScript inputFile1("multitype_input.dat");
    inputFile1.setInputs(1, _tds1->getAttribute("seed"), "seed");

    // Create the output files
    URANIE::Relauncher::TKeyResult outputFile1("_output_multitype_mt_Key_.dat");
    outputFile1.addOutput(_w1, "w1");
    outputFile1.addOutput(_v1, "v1");
    outputFile1.addOutput(_v2, "v2");
    outputFile1.addOutput(_f1, "f1");
    outputFile1.addOutput(_w2, "w2");

    // Create the user's evaluation function
    URANIE::Relauncher::TCodeEval eval1((UranieTests::code_path("multitype") + " -mtKey").data());
    eval1.addInputFile(&inputFile1);
    eval1.addOutputFile(&outputFile1);

    // Create the output files
    URANIE::Relauncher::TKeyScript inputFile2("_output_multitype_mt_Key_condensate_.dat");
    inputFile2.setInputs(5, _w1, "w1", _v1, "v1", _v2, "v2", _f1, "f1", _w2, "w2");
    //inputFile2.setVectorProperties("[",",","]");
    
    // Create the output files
    URANIE::Relauncher::TKeyResult outputFile2("_output_multitype_readmt_Key_.dat");
    outputFile2.addOutput(_thev1, "thev1");
    outputFile2.addOutput(_thev2, "thev2");

    // Create the user's evaluation function
    URANIE::Relauncher::TCodeEval eval2((UranieTests::code_path("multitype") + " -ReadmtKey").data());
    eval2.addInputFile(&inputFile2);
    eval2.addOutputFile(&outputFile2);

    URANIE::Relauncher::TComposeEval eval;
    eval.addEval(&eval1);
    eval.addEval(&eval2);  

    URANIE::Relauncher::TThreadedRun runner(&eval,3);
    
    runner.startSlave();                                           
    if (runner.onMaster())
    {      
	URANIE::Relauncher::TLauncher2 lanceur(_tds1, &runner);
	lanceur.solverLoop();
        
	// Stop the slave processes
	runner.stopSlave();
    }        
  
    validateOutputMultiTypeResults(_tds1);
    validateValuesReadmt( _tds1);
}

// // ===========================================================
// // ==================== Type Exceptions ======================
// // ===========================================================
void TLauncher2Test::testThrowExceptionTypeForTCIntEval()
{
    URANIE::DataServer::TAttribute v("v", URANIE::DataServer::TAttribute::kVector);
    URANIE::DataServer::TAttribute s("s", URANIE::DataServer::TAttribute::kString);
    
#ifdef ROOT5
    URANIE::Relauncher::TCIntEval eval("pouet");

    CPPUNIT_ASSERT_THROW(eval.addInput(&v), URANIE::Relauncher::TEval::CanNotDealWithNotDouble);
    CPPUNIT_ASSERT_THROW(eval.addInput(&s), URANIE::Relauncher::TEval::CanNotDealWithNotDouble);
    //Output
    CPPUNIT_ASSERT_THROW(eval.addOutput(&v), URANIE::Relauncher::TEval::CanNotDealWithNotDouble);
    CPPUNIT_ASSERT_THROW(eval.addOutput(&s), URANIE::Relauncher::TEval::CanNotDealWithNotDouble);

    //Input
    CPPUNIT_ASSERT_THROW(eval.setInputs(1,&v), URANIE::Relauncher::TEval::CanNotDealWithNotDouble);
    CPPUNIT_ASSERT_THROW(eval.setInputs(1,&s), URANIE::Relauncher::TEval::CanNotDealWithNotDouble);
    //Output
    CPPUNIT_ASSERT_THROW(eval.setOutputs(1,&v), URANIE::Relauncher::TEval::CanNotDealWithNotDouble);
    CPPUNIT_ASSERT_THROW(eval.setOutputs(1,&s), URANIE::Relauncher::TEval::CanNotDealWithNotDouble);
#else
    CPPUNIT_ASSERT_THROW(URANIE::Relauncher::TCIntEval eval("pouet"), 
            URANIE::Exceptions::UErrorExceptions);
#endif    
}


void TLauncher2Test::testThrowExceptionPythonEvalThreaded()
{

#ifdef ROOT5
    URANIE::Relauncher::TCIntEval cint("pouet");
#endif
    URANIE::Relauncher::TCodeEval code("multitype -mtKey");
    URANIE::Relauncher::TCJitEval cjit(FlowrateReatmt_hard);

    CPPUNIT_ASSERT( cjit.isThreadSafe() == true);
    CPPUNIT_ASSERT( code.isThreadSafe() == true);
#ifdef ROOT5
    CPPUNIT_ASSERT( cint.isThreadSafe() == false);

    CPPUNIT_ASSERT_THROW(URANIE::Relauncher::TThreadedRun run(&cint, 3), URANIE::Relauncher::TEval::EvalIsNotThreadable);

#ifdef WITH_PYTHON
    URANIE::Relauncher::TPythonEval pyth(NULL);
    CPPUNIT_ASSERT( pyth.isThreadSafe() == false);    
    CPPUNIT_ASSERT_THROW(URANIE::Relauncher::TThreadedRun run(&pyth, 3), URANIE::Relauncher::TEval::EvalIsNotThreadable);
#endif
#endif
}

void TLauncher2Test::testThrowExceptionStringInMpi()
{

    URANIE::DataServer::TAttribute v("v", URANIE::DataServer::TAttribute::kVector);
    URANIE::DataServer::TAttribute s("s", URANIE::DataServer::TAttribute::kString);
    URANIE::DataServer::TAttribute d("d");
    URANIE::DataServer::TAttribute o("o");
    
#ifdef ROOT5
    URANIE::Relauncher::TCIntEval cint("pouet");
    CPPUNIT_ASSERT( cint.containsStringType() == false);
    cint.addInput(&d); cint.addOutput(&o);
    CPPUNIT_ASSERT( cint.containsStringType() == false);
#endif
    URANIE::Relauncher::TCJitEval cjit(FlowrateReatmt_hard);

    CPPUNIT_ASSERT( cjit.containsStringType() == false);
    
    cjit.addInput(&o); cjit.addInput(&s); cjit.addInput(&v);
    
#ifdef ROOT5
    cint.freeze();
    CPPUNIT_ASSERT( cint.containsStringType() == false);
#endif
    cjit.freeze();
    CPPUNIT_ASSERT( cjit.containsStringType() == true);
    
#ifdef ROOT5
    URANIE::Relauncher::TComposeEval comp;
    comp.addEval(&cint);
    comp.addEval(&cjit);
    
    URANIE::Relauncher::TCodeEval code("multitype -mtDS");
    CPPUNIT_ASSERT( code.containsStringType() == false);   
    URANIE::Relauncher::TFlatScript input("toto.in");
    input.addInput(&d);
    URANIE::Relauncher::TFlatResult output("toto.out");
    output.addOutput(&s);
    code.addInputFile(&input);
    code.addOutputFile(&output);
    code.freeze();
    CPPUNIT_ASSERT( code.containsStringType() == true);   
    
#ifdef WITH_MPI
    CPPUNIT_ASSERT_THROW(URANIE::MpiRelauncher::TMpiRun run(&cjit), URANIE::MpiRelauncher::TMpiRun::StringInMpi);
    CPPUNIT_ASSERT_THROW(URANIE::MpiRelauncher::TMpiRun run(&comp), URANIE::MpiRelauncher::TMpiRun::StringInMpi);
    CPPUNIT_ASSERT_THROW(URANIE::MpiRelauncher::TMpiRun run(&code), URANIE::MpiRelauncher::TMpiRun::StringInMpi);
#endif
#endif
    
}

void TLauncher2Test::testThrowExceptionConstantIssues()
{

    setUpMultiTypeInput();
        
    // Create the user's evaluation function
    URANIE::Relauncher::TCJitEval eval(FlowrateReatmt_hard);
    eval.setInputs(4, _tds1->getAttribute("v1"), _tds1->getAttribute("w1"), _tds1->getAttribute("v2"), _tds1->getAttribute("w2"));
    eval.addOutput(_thev1);
    eval.addOutput(_thev2);

    URANIE::DataServer::UEntry *tmpConst = new URANIE::DataServer::UEntry(URANIE::DataServer::UEntry::Vector, "tmpv2");
    for(unsigned int ii=1; ii<=6; ii++)
        tmpConst->pushEntry(double(ii));
    
    URANIE::Relauncher::TSequentialRun runner(&eval);
    runner.startSlave();                                           
    if (runner.onMaster())
    {
	URANIE::Relauncher::TLauncher2 lanceur(_tds1, &runner);
        CPPUNIT_ASSERT_THROW(lanceur.addConstantValue(_tds1->getAttribute("v2"), tmpConst, true), URANIE::Exceptions::UErrorExceptions);

	// Stop the slave processes
	runner.stopSlave();
    }

    delete tmpConst;
}

//! End Of File
