//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision
// $State$
//////////////////////////////////////////////////////////////
/*!
 \file   TDataServer.cxx
 \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
 \date   thur apr 26 14:14:11 CEST 2007

 $Revision$
 $Date$

 \brief Implementation of the class URANIE::DataServer::TDataServer
 This is the main object in order to perform incertainties in library mode.

 \par Modifications
 \li JB. Blanchard (CEA) 2016-02-19: Very large modification for the new TDSNtupleD type based on tree directly.
 Modified many methods to deal with string and vector<double> 
 \li S. Pomarede (2012-10-19): the Windows "\r\n" end-of-lines are allowed.
 \li M. Salvoldi (28-11-2011) :
 - suppress the IFDEF WIN32 instruction in all the class
 - add the closure of the file identified by f in ntupleDataRead()
 - replace the Form(%s/%s...) by gSystem->ConcatFileName(..) in fileDataRead()
 \li C. Forgues (11-01-2011 15:12 ): corrects ntupleDataRead(), taking account of inversion of the iterator attribute from createTuple() method.
 \li C. Forgues (2010-12-22 14:17:21)
 - merge() method takes care of hand made TDataServers.
 - TFile in addAttribute() method is now correctly linked to the current TDirectory.
 \li F. Gaudier (2010-12-13 09:03:01)
 - add the \em optstat option in the \b drawTufte method;
 - store and re-apply the \em optstat flag for the \em gStyle  in the \b drawTufte method;
 \li Cedric Forgues (2010-10-27)
 - Variable MAXNBCHARACTERS is now a defined global variable.
 \li F. Gaudier (2010-09-01 15:27:44)
 - add the normalisation method with 3 methods (kCR, kMinusOneOne, kZeroOne) 
 \li F. Gaudier (2010-08-05 14:48:49)
 - put the option argument in the two draw method : drawIter and drawHistogramm
 - put the good legend for the drawIter method
 \li Cedric Forgues (C.S.) 2010/07/23
 - getArchiveFileName is modified to return correct path for the archive name with WIN32.
 \li Cedric Forgues (C.S.) 2010/07/07
 - Save rootfile in fileDataRead() is disabled with WIN32.
 \li Cedric Forgues (C.S.) 2010/06/30
 - A new global variable is added, shortDataTreeName, initialized to GetName() if the numbers of caracters of GetName() is inferior to 15, else to a restriction to the 15 first caracters.
 \li Cedric Forgues (C.S.) 2010/06/29
 - Modifications in createTuple(), if the name of the tds is too long (more than 15 caracters) it is shortened.
 \li Cedric Forgues (C.S.) 2010/06/18
 - Modifications in exportData(), variables names weren't readable.
 - Adding exception in computeInverseNormalCDF() method when giving a non valuable value for probability.
 - Adding exceptions in drawBoxPlot(), drawTufte() methods when giving a non valuable value for probability.
 \li Cedric Forgues (C.S.) 2010/06/18
 - Changing Exception in Draw method with order. 
 \li Cedric Forgues (C.S.) 2010/06/16
 - Changing warning in error for getNPatterns() method. 
 \li Cedric Forgues (C.S.) 2010/04/23
 - Adding of the isWhiteLine(TString) private method. Tests if a TString is a white Line. Used for fileDataRead() method. 
 \li Fabrice Gaudier (2010-04-13 11:24:40)
 - Change the location of the archive file (gSystem->pwd()) instead of the directory of the data source
 - Active the save of the archive file (TFile)
 - Correct the exportData method
 - Correct the bug when added more than one attributes
 \li ven fev 19 12:00:50 2010 FGA
 Correction du bug du calcul de la matrice de correlations uniquement sur les valeurs malgre l'option rang.
 \li Cedric Forgues (C.S.) 2010/01/12
 filedataRead; ostringstream cast in a TString (with the str() method) gives a bad allocation with WIN32.
 \li Cedric Forgues (C.S.) 2009/12/10
 createTuple();
 filedataRead;
 \li Cedric Forgues (C.S.) 2009/12/08
 computeCorrelationMatrix();
 *********************************************************************  
 */
#include <stdlib.h>

#include <iostream>
#include <iomanip>
using std::ostringstream;
//using std::ifstream;
//using std::ofstream;
#include <fstream>
using namespace std;

#include "TROOT.h"
#ifndef ROOT_TTreeFormula
#include "TTreeFormula.h"
#endif
#ifndef ROOT_TSystem
#include "TSystem.h"
#endif
#ifndef ROOT_TFile
#include "TFile.h"
#endif
#include "TThread.h"
#include "TRegexp.h"
#include "TRandom3.h"
#include "TH1.h"
#include "TMultiGraph.h"
#include "TCanvas.h"
#include "TGaxis.h"
#include "TPaveText.h"
#include "TArrow.h"
#include "TLegend.h"
#include "TProfile.h"
#include "TMarker.h"
#include "THistPainter.h"
#include "TEnv.h"
#include "TMath.h"
#include "TPaveLabel.h"
#include "TStyle.h"
#include "TVectorD.h"
//#include "Api.h"
#include "TMatrixTBase.h"
#include "TParallelCoord.h"
#include "TParallelCoordVar.h"


#ifdef WITH_JSONCPP
#include <json/json.h>
#endif

// Uranie
#include "UVersion.h"
#include "TDataServer.h"
#include "TTokenizer.h"
#include "TAttributeFormula.h"
#ifdef WITH_MATHMORE
#include "Math/ProbFuncMathCore.h"
#include "Math/QuantFuncMathMore.h"
#include "Math/QuantFuncMathCore.h"
#endif
#include "UExceptions.h"

#define SINTERNALRAFU "__rafu__dataset__"
#define MAXNBCHARACTERS 55
#define JSON_READ_ERROR() std::cerr << "Error reading in JSON with[]" << std::endl; \
    exit(-1)

//static const char *rcsid = "$Id$";

// Int_t doBanner()
// {
//   cout << endl << "\033[1m--- Uranie v" << URANIE_RELEASE
//        << " --- Developed with ROOT (" << gROOT->GetVersion()
//        << ") by Fabrice Gaudier \033[0m " << endl
//        << "                       Copyright (C) 2013-2022 CEA/DES " << endl
//        << "                       Version : v" << URANIE_RELEASE
//        << " -  Date : " << URANIE_RELEASE_DATE << endl
//        << "                       All rights reserved, please read http://root.cern.ch/"
//        << endl << endl;
//   return 0;
// }

static Int_t dummy = URANIE::DataServer::TDataServer::doBanner();

ClassImp (URANIE::DataServer::TDataServer)

namespace URANIE
{
namespace DataServer
{

  TDataServer::TDataServer() :
    TNamed("_tds_", "_title_"), _dataSpecification(NULL), _datatree(NULL), _canvas(NULL), _friend(NULL),  _blog(kFALSE), 
    _bQuantVectPrinted(kFALSE), _biteratorNameChanged(kFALSE), _bsaveTuple(kTRUE), _sdatafile(""), _sarchivefile("_dataserver_.root"),
    _siteratorChangedName(""), _ccut(0), _norigin(kUnknown), _pelBootstrap(NULL)
  {
    _fLogger = new URANIE::DataServer::UMessageLogger("TDataServer");
    _drawingGarbageCollector = new TObjArray();
    _drawingGarbageCollector->SetOwner(kTRUE);
    checkName();
  }

TDataServer::TDataServer(const char *name, const char* title) :
    TNamed(name, title),
    //  _sMethodName("_unknown_method_"),
    _dataSpecification(NULL), _datatree(NULL), _canvas(NULL), _friend(NULL),  _blog(kFALSE), 
    _bQuantVectPrinted(kFALSE), _biteratorNameChanged(kFALSE), _bsaveTuple(kTRUE), _sdatafile(""), _sarchivefile("_dataserver_.root"),
    _siteratorChangedName(""), _ccut(0), _norigin(kUnknown), _pelBootstrap(NULL)

  {
    _fLogger = new URANIE::DataServer::UMessageLogger("TDataServer");
    _drawingGarbageCollector = new TObjArray();
    _drawingGarbageCollector->SetOwner(kTRUE);

    TString stitle=TString(title);
    stitle.ToLower();

    checkName();
  }

  //! Copy constructor
  TDataServer::TDataServer(const TDataServer &tds) :
    TNamed(Form("Copy_Of_%s", tds.GetName()), tds.GetTitle()), _dataSpecification(tds._dataSpecification),
    _datatree(tds._datatree), _canvas(NULL), _friend(tds._friend), _blog(tds._blog), 
    _bQuantVectPrinted(kFALSE), _biteratorNameChanged(kFALSE), _bsaveTuple(kTRUE), _sdatafile(tds._sdatafile), _sarchivefile(tds._sarchivefile),  
    _shortDataTreeName(tds._shortDataTreeName), _siteratorChangedName(tds._siteratorChangedName),  _ccut(tds._ccut), 
    _norigin(tds._norigin), _pelBootstrap(tds._pelBootstrap)
  
  {
    cout << "Constructeur de copie URANIE::DataServer::TDataServer[" << this
	 << "] name[" << tds.GetName() << "] to new name[" << GetName()
	 << "]" << endl;
    
    _fLogger = new URANIE::DataServer::UMessageLogger("TDataServer");
    _drawingGarbageCollector = new TObjArray();
    _drawingGarbageCollector->SetOwner(kTRUE);
    checkName();
  }

void TDataServer::SetName(const char *name)
{
    //    cout<<__LINE__<<"  Changing from "<<GetName();
    if( !TThread::IsInitialized())
    {
        gROOT->Remove(this);
    
        stringstream sstr; int ival=0; 
        sstr << name;
        while(  gROOT->Get( sstr.str().c_str() ) != 0  )
        {
            sstr.str("");
            sstr << name << "_" << ++ival;
        }
        fName = sstr.str().c_str();
        if( _datatree != NULL )
            _datatree->SetName( getDataTreeName() );

        gROOT->Append(this);
    }
    //    cout<<" to "<<GetName()<<endl;
    //    gROOT->ls();
    
}
  
void TDataServer::checkName()
{
//    TH1::AddDirectory(kFALSE);
    
    TString name = TString(GetName());
    try{
	if( gROOT->Get( name.Data() ) != 0)
	    throw URANIE::Exceptions::UWarningExceptions( __FILE__,
							  __LINE__,
							  Form(
							      "TDataServer construction: \"%s\" already exist... the name is changed by incrementing a index number at the end \"%s_\"",
							      name.Data(), name.Data()));
      

        if( !TThread::IsInitialized())
            gROOT->Append( this);
      
    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
	//ue.printMessage();
	SetName( name.Data() );
    }
}
  

TDataServer::~TDataServer()
{

    if (gROOT->Get(GetName())) {
        if (_blog)
        {
            cout << " ** Destroy TDataServer Name[" << GetName() << "] Title["
	                << GetTitle() << "]" << endl;
            _dataSpecification->setLog();
        }

        if(_drawingGarbageCollector)
            delete _drawingGarbageCollector;
        _drawingGarbageCollector=NULL;
    
        if (_dataSpecification)
        {
            delete _dataSpecification;
            _dataSpecification = NULL;
        }

        if (_fLogger)
        {
            delete _fLogger;
            _fLogger = NULL;
        }

        if (_friend)
        {
            if(_datatree!=NULL)
	        _datatree->RemoveFriend(_friend);
            delete _friend;
            _friend=NULL;
        }

        if (_datatree)
        {
            // Save the TTree in a TFile
            if (_blog)
            {
                cout << "**************************" << endl;
                cout
                    << " ** Before delete attribute, save the data in the sarchivefile["
                    << _sarchivefile << "]" << endl;
            }

            if( _bsaveTuple )
            {
                TFile * tf = TFile::Open(getArchiveFileName(), "UPDATE");
                _datatree->SetDirectory(0);
                _datatree->Write();
                tf->Close();
                if( !TThread::IsInitialized())
                    _datatree->SetDirectory(gROOT);
            }
            if (_blog)
            {
                cout << " ** Close TFile[" << _sarchivefile << "]" << endl;
                cout << "**************************" << endl;

                cout<<"Looking at ntuple "<<_datatree<<endl;
                cout<<" whose name is "<<_datatree->GetName()<<endl;
            }
            delete _datatree;
            _datatree = NULL;
        }

        gROOT->Remove(this);

        if (_pelBootstrap)
            delete _pelBootstrap;
        _pelBootstrap = NULL;
    }

    if (_blog)
        cout << " ** End of destructor TDataServer Name[" << GetName()
             << "] Title[" << GetTitle() << "]" << endl;
}

  ///////////////////////////////////////////////////////////
  //  Lecture des donnees
  ///////////////////////////////////////////////////////////
Bool_t TDataServer::fileDataRead(const char* filename, Bool_t bSave, Bool_t bDontAddAttribute, Option_t *opt)
  {

      TString soption=opt;
      soption.ToLower();

      bool quiet = false;
      if (soption.Contains("quiet"))
          quiet=true;
      
    //Mod by JB Blanchard 2016-02-19: Change the way the file is written in the ntuple to cope with the possibility of reading
    //strings and vector<double>. The latter are written as one block whose component are split by commas.
    // *Add COLUMN_TYPES in the header (prerequisite to used the two added types)
    // *Use the Fill(string) method when dealing with the file
      if (_blog)
          cout << "TDataServe::FileDataRead[" << filename << "]" <<  endl;

    bool preadd=false;
    try
      {
	if (_dataSpecification && !bDontAddAttribute)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form(
							    "TDataServer::fileDataRead ERROR : the TDataServer [%s] is not empty",
							    filename));
	else if (_dataSpecification && bDontAddAttribute && _datatree!= NULL && !quiet)
	    throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 Form(
							     "TDataServer::fileDataRead Warning : You choose to reload the TDataserver [%s]. Previous data will be removed and replaced.\n\
If you know what you are doing and want to get rid of this line, please add \"quiet\" as fourth argument to fileDataRead.", filename));
	
	else if (_dataSpecification && bDontAddAttribute && _datatree == NULL && !quiet)
	{
	    preadd=true;
	    throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 Form(
							     "TDataServer::fileDataRead Warning : You choose to pre-add few attributes. All other attributes found in [%s]  will be added on top.\n\
If you know what you are doing and want to get rid of this line, please add \"quiet\" as fourth argument to fileDataRead.", filename));
	}
	
      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
	ue.printMessage();
	throw ue;
      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
	ue.printMessage();
      }

    ifstream file(filename, ios::in);

    try
      {
	if (!file)
	  {
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form(
							    "TDataServer::fileDataRead ERROR : can't open file [%s]",
							    filename));
	    return kFALSE;
	  }
      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
	ue.printMessage();
	throw ue;
      }

    // Reading of  headers: all line beginning with a "#" character
    Bool_t bentete = kTRUE;
    map < TString, TString > header_file;

   
    // Read the whole line each character by character until End Of Line is reached
    do
      {
	ostringstream sLine;
	char ch;
	while (file.get(ch) && (ch != '\n') && (ch != '\r'))
	  {
	    sLine.put(ch);
	  }
	if (ch == '\r')
	  file.get(ch);

	if (_blog)
            cout << " ** sLine [" << sLine.str() << ']' << endl;
	// Reads the whole line

	string stringSLine = sLine.str();
	TString logLine = TString(stringSLine.c_str());
	if (_blog)
	  cout << " ** logLine[" << logLine << "]" << endl;

	// If the line is a header's one, continue, else breaks the loop
	if (logLine.BeginsWith("#"))
	  bentete = kTRUE;
	else
	  {
	    bentete = kFALSE;
            try {
                if( !UEntry::isWhiteLine(string(logLine)) )
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                               Form(
                                                                   "TDataServer::fileDataRead ERROR : file [%s] must contain an empty line after headers",
                                                                   filename));
            } catch (URANIE::Exceptions::UErrorExceptions& ue)
            {
                ue.printMessage();
                throw ue;
            }
	    break;
	  }

	// Search of notion [name, unity, type, label]
	Ssiz_t npoint = logLine.Index(":");
	if (_blog)
	  cout << "npoint[" << npoint << "]" << endl;
	TString sLabel;
	if (npoint > 0)
	  sLabel = logLine(1, npoint - 1);
	//    else {
	//      npoint = 0;
	//      sLabel = TString("COLUMN_NAMES");
	//    }
	if (_blog)
	  cout << "sLabel[" << sLabel << "]" << endl;
	header_file[sLabel] = logLine.Remove(0, npoint + 1);

      } while (bentete);

    // Racine du nom pour l'arbre
    TString sRacFile;
    TString sTmp = gSystem->BaseName(filename);

    Int_t nTmp = sTmp.Index(".");
    if (nTmp != -1)
      sRacFile = sTmp(0, nTmp);
    else
      sRacFile = sTmp;

    _sdatafile = gSystem->DirName(filename);
    if (_sdatafile.CompareTo(".", TString::kExact) == 0)
      _sdatafile = gSystem->pwd();

    TString sRacFileRoot = Form("%s.root", sRacFile.Data());
    char *tmp = gSystem->ConcatFileName(gSystem->pwd(), sRacFileRoot.Data());
    _sarchivefile = tmp;
    delete[] tmp;
    tmp = gSystem->ConcatFileName(_sdatafile.Data(),
				  gSystem->BaseName(filename));
    _sdatafile = tmp;
    delete[] tmp;

    if (_blog)
      cout << " ** _sarchivefile[" << _sarchivefile << "]" << endl;

    // Name field
    TString strtds = TString(GetName());
    if ( strtds.Contains("_tds_") )
      {
	if (!header_file["NAME"].IsNull())
	  {
	    SetName(header_file["NAME"].Remove(TString::kBoth, ' '));
	  }
	else 
	  {
	    SetName(sRacFile.Data());
	  }
      }	

    
    //Initialize the _shortDataTreeName
    _shortDataTreeName = getDataTreeName();

    //  Create the TDataSpecification object to store the header information
    TDataSpecification *tdsp_bkup=NULL; //pointer to store official address  : if reload file, then it bkup this one and redo a new one
    TDataSpecification *thenewspec=NULL;
    if( !bDontAddAttribute)
    {
	if (_dataSpecification != NULL )
	    delete _dataSpecification;
	
	_dataSpecification = new TDataSpecification(Form("uheader__%s", GetName()),
						    Form("Header of %s", GetTitle()));
    }
    else 
    {
	tdsp_bkup = _dataSpecification;
	_dataSpecification = new TDataSpecification(Form("uheader__cpy__%s", GetName()),
						    Form("Header of %s", GetTitle()));
    }
    
    // Title Field
    strtds = TString(GetTitle());
    if (strtds == "_title_")
      {
	if (!header_file["TITLE"].IsNull())
	  {
	    _dataSpecification->setRelationName(header_file["TITLE"].Data());
	    SetTitle(header_file["TITLE"].Remove(TString::kBoth, ' '));
	  }
	else
	  _dataSpecification->setRelationName(
					      Form("FileDataRead[%s]", filename));
      }

    // creation of the vector of attributes
    TString headerTag = TString("headertag");
    try
      {
	if (header_file["COLUMN_NAMES"].IsNull())
	  {
	    if (_blog)
	      cout
		<< Form(
			"TDataServer::fileDataRead: No header line COLUMN_NAMES specified in the file \"%s\".",
			filename) << endl;
	    if (header_file["COLUMN_TITLES"].IsNull())
	      throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							 Form(
							      "TDataServer::fileDataRead ERROR : neither header line COLUMN_NAMES nor header line COLUMN_TITLES specified in the file \"%s\".",
							      filename));
	    else
	      headerTag = TString("COLUMN_TITLES");
	  }
	else
	  headerTag = TString("COLUMN_NAMES");
      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
	ue.printMessage();
	throw ue;
      }

    header_file[headerTag].ReplaceAll(" ", "").ReplaceAll("\"", "");
    if (_blog)
      cout << "header_file[" << headerTag << "].Data()["
	   << header_file[headerTag].Data() << "]" << endl;

    vector < string > tokens;
    UEntry::StringParser(string(header_file[headerTag]),"|",tokens);
    Int_t nAttribute = (Int_t)(tokens.size());
    vector < string > tokens_types;
    // Ajout du type
    // Added by JBB (2016-02) to cope with strings/vector<double>
    if (!header_file["COLUMN_TYPES"].IsNull())
      {
	Int_t nTokens_Types;
	header_file["COLUMN_TYPES"].ReplaceAll(" ", "").ReplaceAll("\"", "");
	UEntry::StringParser(string(header_file["COLUMN_TYPES"]),"|",tokens_types);
	nTokens_Types = (Int_t)(tokens_types.size());
	try
	  {
	    if (nTokens_Types != nAttribute)
	      throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							 Form(
							      "TDataServer::fileDataRead: COLUMN_TYPES is precised, but number of information (%d) is different from COLUMN_NAMES (%d)",
							      nTokens_Types, nAttribute));
	  } catch (URANIE::Exceptions::UErrorExceptions& ue)
	  {
	    ue.printMessage();
	    throw ue;
	  }		
      }

    //Check whether there is a vector size attribute in case of vector
    try{
    for (UInt_t indx = 0; indx < tokens_types.size(); indx++)
      if ( tokens_types[indx] == "V" )
	  if ( find( tokens.begin(), tokens.end(), SIZENAME+"__"+tokens_types.at(indx) ) != tokens.end() )
	      throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							 Form(
							     "TDataServer::fileDataRead: %s attribute should not be stored in a file, as it is an internal attribute made on the flight to check consistency of a vector. Rename this attribute if it has nothing to do with a vector size or get rid of it.", (SIZENAME+"__"+tokens_types.at(indx)).c_str()));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
	ue.printMessage();
	throw ue;
    }
    
    map<int, int> shiftForV; int nVec=0;// Possible shift for vector
    vector<TString> foundIndex; foundIndex.clear();
    //Looping over the attribute
    for (Int_t indx = 0; indx < nAttribute; indx++)
      {
	TString sname(tokens[indx].c_str());
	try
	  {

	      //It if already exists
	      if (_dataSpecification->isAttribute(sname) && bDontAddAttribute)
		  continue;
	      else if (_dataSpecification->isAttribute(sname) && !bDontAddAttribute )
	      {
		Int_t ncrt = 1;
		TString soldname = sname;
		while (_dataSpecification->isAttribute(Form("%s_%d", sname.Data(), ncrt)))
		  ncrt++;
		//Change the name to name+_incrementtation
		sname += TString(Form("_%d", ncrt));
		_dataSpecification->addAttribute(new TAttribute(sname));
		_dataSpecification->getAttribute(sname)->delShare();
		
		//Add the size attribute in case this is a vector
		if( tokens_types.size() == tokens.size() )
		  {
		    _dataSpecification->getAttribute(sname)->setDataType( tokens_types[indx] );
		    if ( tokens_types[indx] == "V" )
		      {
			_dataSpecification->addAttribute( new TAttribute( (SIZENAME+"__"+string(sname) ).c_str() ) );
			_dataSpecification->getAttribute( (SIZENAME+"__"+string(sname) ).c_str() )->delShare();
			_dataSpecification->getAttribute((SIZENAME+"__"+string(sname) ).c_str() )->setOrigin(TAttribute::kInternal) ;
			nVec++;
		      }
		  }
		throw URANIE::Exceptions::UWarningExceptions( __FILE__,
							      __LINE__,
							      Form(
								   "TDataServer Name: \"%s\" Title\"%s\"\nFile \"%s\"\nAttribute \"%s\" already exists : New name \"%s\"",
								   GetName(), GetTitle(), filename,
								   soldname.Data(), sname.Data()));
	      }
		
	  } catch (URANIE::Exceptions::UWarningExceptions& ue)
	  {
	    ue.printMessage();
	    continue;
	  }

	//regular case where not found yet
	_dataSpecification->addAttribute(new TAttribute(sname));
	_dataSpecification->getAttribute(sname)->delShare();
	//If this matches the iterator ending
	if( sname.EndsWith("__n__iter__") )
	  foundIndex.push_back(sname);
	
	if( tokens_types.size() == tokens.size() )
	  {
	    _dataSpecification->getAttribute(sname)->setDataType( tokens_types[indx] );
	    if ( tokens_types[indx] == "V" )
	      {
		_dataSpecification->addAttribute( new TAttribute( (SIZENAME+"__"+string(sname) ).c_str()) );
		_dataSpecification->getAttribute((SIZENAME+"__"+string(sname) ).c_str() )->setOrigin(TAttribute::kInternal) ;
		_dataSpecification->getAttribute((SIZENAME+"__"+string(sname) ).c_str() )->delShare() ;
		nVec++;
	      }
	  }
	shiftForV[indx+1]=nVec;
      } // End of loop For(indx)


    try
      {
	if ( foundIndex.size() > 1 )
	  throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
						     Form(
							  "TDataServer::fileDataRead: found %lu possible iterator in this file. No one will be chosen", foundIndex.size()));
      }catch(URANIE::Exceptions::UWarningExceptions &ue)
      {
	ue.printMessage();
      }

    Int_t nTokens;
    // Ajout des labels
    if (!header_file["COLUMN_TITLES"].IsNull())
      {
	UEntry::StringParser(string(header_file["COLUMN_TITLES"]),"|",tokens,true);
	nTokens = (Int_t)(tokens.size());
	if (nTokens > nAttribute)
	  nTokens = nAttribute;
	for (Int_t indx = 0; indx < nTokens; indx++)
	  _dataSpecification->getAttribute(indx+shiftForV[indx])->SetTitle(tokens[indx].c_str());
      }

    // Ajout des unites
    if (!header_file["COLUMN_UNITS"].IsNull())
      {
	UEntry::StringParser(string(header_file["COLUMN_UNITS"]),"|",tokens,true);
	nTokens = (Int_t)(tokens.size());
	if (nTokens > nAttribute)
	  nTokens = nAttribute;
	for (Int_t indx = 0; indx < nTokens; indx++)
	  _dataSpecification->getAttribute(indx+shiftForV[indx])->setUnity( TString(tokens[indx].c_str()).Remove(TString::kBoth, ' '));
      }

    // Ajout de l'iterateur Name+"__n__iter__" s'il nexiste pas
    TString siteratorname = TString(getIteratorName());
    int iterindex = -1;
    Bool_t haveIndex = _dataSpecification->isAttribute(siteratorname);

    TAttribute* index = NULL;
    
    if (haveIndex)      
      {
	index = _dataSpecification->getAttribute(siteratorname);
	iterindex = getAttributeIndex( index, true );
	index->setOrigin(TAttribute::kIterator);
      }
    else
      {
	if (foundIndex.size()==1)
	  {
	    try
	      {
		throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							     Form("TDataServer::fileDataRead: Expected iterator %s not found but %s looks like an URANIE iterator => Will be used as so.",
								  siteratorname.Data(),foundIndex.at(0).Data()));
	      }catch(URANIE::Exceptions::UWarningExceptions &ue)
	      {
		ue.printMessage();
	      }
	    haveIndex=true;
	    index = _dataSpecification->getAttribute( foundIndex.at(0).Data() );
	    iterindex = getAttributeIndex( index, true );
	    index->setOrigin(TAttribute::kIterator);
	    changeIteratorName( foundIndex.at(0).Data() );
	  }
	else
	  {
	    index = new TAttribute(siteratorname);
	    index->setOrigin(TAttribute::kIterator);
	    _dataSpecification->addAttribute(index);
	    index->delShare();
	  }
      }

    index->setFormatToSubstitute("%d");
    if (_blog)
      _dataSpecification->printLog();

    //Get back to the proper _dataSpecification if bDontAddAttribute has been used for preadd or reload purpose
    if( bDontAddAttribute )
    {
	try
	{
	    if( !preadd && !tdsp_bkup->isComparableToTDSP(_dataSpecification) )
	    {
		TString newname = _dataSpecification->getVariablesNamesTypes(":", true), oldname = tdsp_bkup->getVariablesNamesTypes(":", true);
		throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							   Form("TDataServer::fileDataRead: The provided file does not match the format of the previously used data.\n \
Attribute names(types) were [%s] while in the file, they are [%s].\n Reload option not used properly", oldname.Data(), newname.Data()) );
	    }
	    if( preadd && !tdsp_bkup->isComparableToTDSP(_dataSpecification, true) )
	    {
		TString newname = _dataSpecification->getVariablesNamesTypes(":", true), oldname = tdsp_bkup->getVariablesNamesTypes(":", true);
		throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							   Form("TDataServer::fileDataRead: The provided file [%s] does not contain some attribute pre-added to the dataserver [%s]. Preadd option not used properly", newname.Data(), oldname.Data()) );
	    }
	}catch(URANIE::Exceptions::UErrorExceptions &ue)
	{
	    ue.printMessage();
	    delete tdsp_bkup; tdsp_bkup=NULL;
	    throw ue;
	}

	//Do the combination so that the tdsp contains the attribute in good order while still using the correct pointers
	thenewspec = new TDataSpecification("tutu","toto");
	TString name = tdsp_bkup->GetName(), title = tdsp_bkup->GetTitle();
	TList *listOfAttributes = _dataSpecification->getListOfAttributes();
	TIter next(listOfAttributes);
	TAttribute *att;
	while( (att = (TAttribute*)next()) )
	{

	    if( att->getOrigin() == TAttribute::kInternal)
		continue;
	    
	    if( tdsp_bkup->isAttribute( att->GetName() ) )
		thenewspec->addAttribute( tdsp_bkup->getAttribute( att->GetName() ) );
	    else
		thenewspec->addAttribute( att );
	}

	delete tdsp_bkup; tdsp_bkup=NULL;
	delete listOfAttributes;
	delete _dataSpecification; _dataSpecification=NULL;

	_dataSpecification = thenewspec;
	_dataSpecification->SetName( name );
	_dataSpecification->SetTitle( title );

    }
	
    // Creation du tuple sur les donnees
    createTuple();

    // Lecture des donnees. Par defaut, les donnees sont mises dans le
    // vecteur des entrees.
    Double_t fCrt = 1.0;
    string line;

    // Changed to used the fill (string) method of TDSNtupleD, JBB (2016-02) to cope with strings/vector<double>
    while ( !UEntry::safeGetline(file, line).eof() )
      {
	if (!(line.length() > 0) || UEntry::isWhiteLine(line))
            break;

	stringstream sstr; sstr << fCrt << " ";

	if(_blog)
	  cout<<" *** Reading line "<<sstr.str()+line<<endl;
	
	if(haveIndex)
	  _datatree->Fill( line, iterindex);
	else
	  _datatree->Fill( sstr.str()+line );
      
	fCrt += 1.0;
      }
    
    //Close the file
    file.close();

    if (_blog)
      {
	cout << " ** TDataServer::fileDataRead read " << getNPatterns()
	     << " points" << endl;
	_datatree->Print();
	_dataSpecification->printLog();
      }

    // Save the TTree in a TFile
    if(bSave)
      {
	TFile *tf = TFile::Open(getArchiveFileName(), "RECREATE");
	_datatree->LoadBaskets();
	_datatree->SetDirectory(0);
	_datatree->Write();
	tf->Close();
        if( !TThread::IsInitialized())        
            _datatree->SetDirectory(gROOT);
      }
                                 
    _norigin = kASCIIFile;
    return kTRUE;
  }

#ifdef WITH_JSONCPP
Bool_t TDataServer::fileDataReadJSon(const char* filename, Bool_t bSave)
{
    if (_blog)
        cout << "TDataServe::fileDataReadJSon[" << filename << "]" << endl;

    ifstream file(filename, ios::in);

    Json::Value root, val;
    Json::Reader reader;
    Json::Value jSONMetaData;
    Json::Value jSONItems;

    try {
        if (_dataSpecification)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							 Form(
							     "TDataServer::fileDataReadJSon ERROR : the TDataServer [%s] is not empty",
							     filename));

        if (!file)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form(
							   "TDataServer::fileDataReadJSon ERROR : can't open file [%s]",
							   filename));

        // Parser the JSON file
        bool parsingSuccessful = reader.parse(file, root);
        if (!parsingSuccessful) {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form(
							   "TDataServer::fileDataReadJSon ERROR : when parsing the JSON file [%s] with JSONCPP library",
							   filename));
        } else {
            if (_blog)
                printf(" ** Parse JSON File [%s] OK \n", filename);
        }

        // Read the _metadata tag
        if (!root.isMember("_metadata")) {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form(
							   "TDataServer::fileDataReadJSon ERROR : The JSON file [%s] does not contain the \"_metadata\" tag",
							   filename));
        } else {
            if (_blog)
                printf(" ** The JSON File [%s] contains _metadata : OK \n", filename);
        }

        // Read the metadata tag
        jSONMetaData = root.get("_metadata",Json::Value::null);

        if (!jSONMetaData.isMember("short_names")) {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form(
							   "TDataServer::fileDataReadJSon ERROR : The MetaData in the JSON file [%s] does not contain the \"short_names\" tag",
							   filename));
        } else {
            if (_blog)
                printf(" ** The JSON File [%s] contains _metadata/short_names : OK \n", filename);
        }


        // Read the items tag
        if (!root.isMember("items")) {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form(
							   "TDataServer::fileDataReadJSon ERROR : The JSON file [%s] does not contain the \"items\" tag",
							   filename));
        } else {
            if (_blog)
                printf(" ** The JSON File [%s] contains items : OK \n", filename);
        }

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    // Racine name of the JSON File
    TString sRacFile;
    TString sTmp = gSystem->BaseName(filename);

    Int_t nTmp = sTmp.Index(".");
    if (nTmp != -1)
        sRacFile = sTmp(0, nTmp);
    else
        sRacFile = sTmp;
    if (_blog)
        cout << " ** sRacFile[" << sRacFile << "] sTmp[" << sTmp << "] nTmp[" << nTmp << "]" << endl;

    // Name field
    TString strtds = TString(GetName());
    sTmp = jSONMetaData.get("table_name",Json::Value::null).asString().c_str();
    sTmp.ReplaceAll(" ","");
    if (strtds == "_tds_") {
        if (_blog)
            cout << "\t ** sTmp[" << sTmp << "]" << endl;
        if ( sTmp.Length() > 0 )
            SetName(sTmp.Data());
        else
            SetName(sRacFile.Data());
    }

    //  Create the TDataSpecification object to store the header information
    _dataSpecification = new TDataSpecification(Form("uheader__%s", GetName()),
                                                Form("Header of %s", GetTitle()));

    // Title field
    strtds = TString(GetTitle());
    sTmp = jSONMetaData.get("table_description", Json::Value::null).asString().c_str();
    if (strtds == "_title_") {
	if (_blog)
	    cout << "\t ** sTmp[" << sTmp << "]" << endl;
        if ( sTmp.Length() > 0 ) {
            _dataSpecification->setRelationName(sTmp.Data());
            SetTitle(sTmp.Data());
        } else {
            _dataSpecification->setRelationName(Form("fileDataReadJSon[%s]", filename));
        }
    }

    // Read the metadata tag
    Json::Value jSONListOfAttributes = jSONMetaData.get("short_names",Json::Value::null);
    Int_t nAttribute = jSONListOfAttributes.size();

    if (_blog) {
        cout << "\t ** TDS Name[" << root["_metadata"]["table_name"] << "] GetName[" << GetName() << "]" << endl;
        cout << "\t ** TDS Title[" << root["_metadata"]["table_description"] << "] GetTitle[" << GetTitle() << "]" << endl;
        cout << "\t ** nAttribute : [" << nAttribute << "]" << endl;
    }
    

    bool hasLongName = false;
    Json::Value jSONListOfLongName = Json::Value::null;
    if ( jSONMetaData.isMember("long_names")) {
        jSONListOfLongName = jSONMetaData.get("long_names",Json::Value::null);
        if ( (Int_t)(jSONListOfLongName.size()) == nAttribute ) {
            hasLongName = true;
            cout << " ** Long Name : Ok " << endl;
        }
    }

    bool hasUnits = false;
    Json::Value jSONListOfUnits;
    if ( jSONMetaData.isMember("units")) {
        jSONListOfUnits = jSONMetaData.get("units",Json::Value::null);
        if ( (Int_t)(jSONListOfUnits.size()) == nAttribute ) {
            hasUnits = true;
            cout << " ** Units : Ok " << endl;
        }
    }

    bool hasTypes = false;
    Json::Value jSONListOfTypes = Json::Value::null;
    if ( jSONMetaData.isMember("types")) {
        jSONListOfTypes = jSONMetaData.get("types",Json::Value::null);
        if ( (Int_t)(jSONListOfTypes.size()) == nAttribute ) {
            hasTypes = true;
            cout << " ** Types : Ok " << endl;
        }
    }
    
    vector<TString> foundIndex; foundIndex.clear();    
    for (Int_t indx = 0; indx < nAttribute; indx++)
    {

        TString sname(jSONListOfAttributes[indx].asString().c_str());
        if ( _blog )
            cout << "\t Att(" <<  indx+1 << "/" << nAttribute << ") Name[" << sname << "]" <<endl;
        try {
            if (_dataSpecification->isAttribute(sname)) {
                Int_t ncrt = 1;
                TString soldname = sname;
                while (_dataSpecification->isAttribute(
                           Form("%s_%d", sname.Data(), ncrt)))
                    ncrt++;
                sname += TString(Form("_%d", ncrt));
                _dataSpecification->addAttribute(new TAttribute(sname));

                throw URANIE::Exceptions::UWarningExceptions( __FILE__,
                                                              __LINE__,
                                                              Form(
                                                                  "TDataServer Name: \"%s\" Title\"%s\"\nFile \"%s\"\nAttribute \"%s\" already exists : New name \"%s\"",
                                                                  GetName(), GetTitle(), filename,
                                                                  soldname.Data(), sname.Data()));
            }
        } catch (URANIE::Exceptions::UWarningExceptions& ue)
        {
            ue.printMessage();
            continue;
        }

        TAttribute *att=NULL;
	if( hasTypes )
	{

	    if( jSONListOfTypes[indx].asString() == "V" )
		att = new TAttribute(sname, URANIE::DataServer::TAttribute::kVector);
	    else if( jSONListOfTypes[indx].asString() == "S" )
		att = new TAttribute(sname, URANIE::DataServer::TAttribute::kString);
	    else
		att = new TAttribute(sname);
	}
	else
	    att = new TAttribute(sname);

        if ( hasLongName )
            att->SetTitle(jSONListOfLongName[indx].asString().c_str());
        if ( hasUnits )
            att->setUnity(jSONListOfUnits[indx].asString().c_str());

        if ( _blog )
            cout << "\t\t Att(" <<  indx+1 << "/" << nAttribute << ") Name[" << att->GetName() << "] Title[" << att->GetTitle() << "] Unit[" << att->getUnity() << "]" <<endl;

	if( sname.EndsWith("__n__iter__") )
	    foundIndex.push_back(sname);
	
        _dataSpecification->addAttribute(att);
        att->delShare();
    } // End of loop For(indx)

    try
    {
	if ( foundIndex.size() > 1 )
	    throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 Form(
							     "TDataServer::fileDataRead: found %lu possible iterator in this file. No one will be chosen", foundIndex.size()));
    }catch(URANIE::Exceptions::UWarningExceptions &ue)
    {
	ue.printMessage();
    }

    // Add the iterator attribute GetName+"__n__iter__" if not exists
    TString siteratorname = TString(getIteratorName());
    Bool_t haveIndex = _dataSpecification->isAttribute(siteratorname);
    Int_t nindexIterator = -1; 
    TAttribute* attIterator = NULL;

    if (haveIndex) {
        attIterator = _dataSpecification->getAttribute(siteratorname);
        attIterator->setOrigin(TAttribute::kIterator);
        nindexIterator = _dataSpecification->getAttributeIndex(attIterator);
    } else {
	if (foundIndex.size()==1)
	{
	    try
	    {
		throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							     Form("TDataServer::fileDataRead: Expected iterator %s not found but %s looks like an URANIE iterator => Will be used as so.",
								  siteratorname.Data(),foundIndex.at(0).Data()));
	    }catch(URANIE::Exceptions::UWarningExceptions &ue)
	    {
		ue.printMessage();
	    }
	    haveIndex=true;
	    attIterator = _dataSpecification->getAttribute( foundIndex.at(0).Data() );
	    nindexIterator = getAttributeIndex( attIterator, true );
	    attIterator->setOrigin(TAttribute::kIterator);
	    changeIteratorName( foundIndex.at(0).Data() );
	    siteratorname = TString(getIteratorName());
	}
	else
	{
	    attIterator = new TAttribute(siteratorname);
	    attIterator->setOrigin(TAttribute::kIterator);
	    _dataSpecification->addAttribute(attIterator);
	    attIterator->delShare();
	}	
    }

    attIterator->setFormatToSubstitute("%d");
    if (_blog) {
        _dataSpecification->printLog();
        cout << " ** nindexIterator(" << nindexIterator << ")" << endl;
    }
    
    // Create the Tuple of data
    createTuple();

    // Read the items tag
    jSONItems = root.get("items", Json::Value::null);

    Int_t nS = (Int_t)(jSONItems.size());
    if (_blog) {
        cout << "\t ** nS : [" << nS << "]" << endl;
    }

    // Lecture des donnees. Par defaut, les donnees sont mises dans le
    // vecteur des entrees.
    Double_t fCrt = 1.0;
    string line;
    int iatt=0;
    vector<string> listofattributes;
    URANIE::DataServer::UEntry::StringParser(string(_dataSpecification->getVariablesNames()),string(":"),listofattributes);

    bool bAlreadyPrintedWarning=false;
    Json::FastWriter fastWriter;
    std::string output;
    if(_blog)
	cout<<"Reading file, with "<<_dataSpecification->getVariablesNamesTypes()<<"  with iterator ? "<<haveIndex<<" name "<<siteratorname<<"  index  "<<nindexIterator<<endl;
    for ( Int_t index = 0; index < (Int_t)(jSONItems.size()); ++index ) 
    {

	stringstream values, it;
	if( !haveIndex )
	    it << fCrt << " ";

	val = jSONItems[index];
	for(vector<string>::iterator itatt=listofattributes.begin(); itatt!=listofattributes.end(); itatt++)
	{

	    if( !haveIndex && (*itatt) == siteratorname ) continue;
	    if( TString(*itatt).BeginsWith("size__") ) continue;

	    iatt=0;
	    while( jSONListOfAttributes[iatt].asString() != (*itatt) )
		iatt++;

	    try{
		if( val[jSONListOfAttributes[iatt].asString()].isNull() )
		{
		    if ( (getAttribute( *itatt )->getDataTypeStr() == "V" ||  getAttribute( *itatt )->getDataTypeStr() == "S") && !bAlreadyPrintedWarning )
		    {
			bAlreadyPrintedWarning=true;
			throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
								   Form("TDataServer::fileDataReadJSON WARNING : the field [%s] is missing for one item. Default is empty string and empty vector",
									jSONListOfAttributes[iatt].asString().c_str()));
		    }
		    
		    if ( getAttribute( *itatt )->getDataTypeStr() == "D" )
			throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
								   Form("TDataServer::fileDataReadJSON ERROR : the field [%s] is missing. As this is a double-type item, method is stopped.",
									jSONListOfAttributes[iatt].asString().c_str()));
		}
	    }catch(URANIE::Exceptions::UErrorExceptions &ue)
	    {
		ue.printMessage();
		throw ue;
	    }
	    catch(URANIE::Exceptions::UWarningExceptions &ue)
	    {
		ue.printMessage();       
	    }

	    output=fastWriter.write(val[jSONListOfAttributes[iatt].asString()]); output=output.substr(0,output.size()-1);

	    if( _blog )
		XYZ<<"Read ["<<jSONListOfAttributes[iatt].asString()<<"] ["<<output<<"]"<<endl;
	    
	    if( (*itatt) == siteratorname )
		it << output << " ";
	    else 
		values << ((output!="null") ? output : "[]") << " ";
	}

	it << values.str();       	
	if (_blog)
	    cout<<"Read "<<it.str()<<endl;
	
	_datatree->Fill( it.str() );
      
	fCrt += 1.0;
    }
    
    //Close the file
    file.close();
    
    if (_blog)
    {
        cout << " ** TDataServer::fileDataReadJSon read " << getNPatterns()
	     << " points" << endl;
        _datatree->Print();
    }

    // Save the TTree in a TFile
    if(bSave)
    {
        TFile *tf = TFile::Open(getArchiveFileName(), "RECREATE");
        _datatree->LoadBaskets();
        _datatree->SetDirectory(0);
        _datatree->Write();
        tf->Close();
    }
    if( !TThread::IsInitialized())
        _datatree->SetDirectory(gDirectory);

    _norigin = kASCIIFile;
    return kTRUE;
}
#endif


Bool_t TDataServer::ntupleDataRead(const char* file, const char *tree,
				     const char* svar, const char *cut)
  {


    try
      {
        if (gSystem->AccessPathName(file, kFileExists))
	  throw URANIE::Exceptions::UExceptions(__FILE__, __LINE__,
						Form(
						     "TDataServer::ntupleDataRead ERROR : can't open file [%s]",
						     file));

        TFile *f = TFile::Open(file);

        TTree *t = (TTree*) f->Get(tree);

        if (t == 0)
	  {
            throw URANIE::Exceptions::UExceptions(__FILE__, __LINE__,
						  Form(
						       "TDataServer::ntupleDataRead ERROR : TDSNtupleD [%s] not found in TFile [%s]",
						       tree, file));
            return kFALSE;
	  }

        if (_dataSpecification)
	  throw URANIE::Exceptions::UExceptions(__FILE__, __LINE__,
						Form(
						     "TDataServer::ntupleDataRead ERROR : the TDataServer [%s] is not empty",
						     _dataSpecification->getRelationName().Data()));

        _dataSpecification = new TDataSpecification(
						    Form("uheader__%s", GetName()),
						    Form("TDataServer::ntupleDataRead[%s][%s][%s]", file,
							 t->GetName(), t->GetTitle()));
        _dataSpecification->setRelationName(
					    Form("TDataServer::ntupleDataRead[%s][%s][%s]", file,
						 t->GetName(), t->GetTitle()));

        //SetNameTitle( TString(t->GetName()).ReplaceAll("__data__tree__",""), t->GetTitle());
	SetName( TString(t->GetName()).ReplaceAll("__data__tree__","")); SetTitle( t->GetTitle());

        // Determine data branches
        TString input(svar);

        if ((input == "*") || (input.BeginsWith("*:", TString::kExact))
	    || (input.EndsWith(":*", TString::kExact))
	    || (input.Contains(":*:", TString::kExact)))
	  {

            TObjArray* listOfLeaves = t->GetListOfLeaves();

            if (_blog)
	      listOfLeaves->Print();

            Int_t nLeaves = listOfLeaves->GetEntries();
            TString slabel;
            if (nLeaves > 0)
	      {
                TLeaf* leaf = (TLeaf *) listOfLeaves->UncheckedAt(0);
                slabel = leaf->GetName();
                for (Int_t i = 1; i < nLeaves; i++)
		  {
		      leaf = (TLeaf *) listOfLeaves->UncheckedAt(i);
                    slabel += Form(":%s", leaf->GetName());
                    //        cout << "TLeaf[" << slabel<<"]"<<endl;
		  }
	      }
            else
	      {
                throw URANIE::Exceptions::UExceptions(__FILE__, __LINE__,
						      Form(
							   "TDataServer::ntupleDataRead ERROR : the ttree [%s] is empty",
							   t->GetName()));
	      }
            //    cout << "*********TLeaf slabel *******[" << slabel<<"]"<<endl;

            if (input == "*")
	      input = slabel;
            //    cout << "*********TLeaf slabel *******[" << input<<"]"<<endl;

            if (input.BeginsWith("*:", TString::kExact))
	      input.Replace(0, 1, slabel);

            //    cout << "*********TLeaf slabel *******[" << input<<"]"<<endl;

            if (input.EndsWith(":*", TString::kExact))
	      input.Replace(input.Length() - 1, 1, slabel);

            //    cout << "*********TLeaf slabel *******[" << input<<"]"<<endl;

            if (input.Contains(":*:", TString::kExact))
	      input.ReplaceAll(":*:", Form(":%s:", slabel.Data()));

            //    cout << "FINALLY *********TLeaf slabel *******[" << input<<"]"<<endl;

	  }

	TString Inputs = "";
	vector<TString> AttributeToAdd;
	
        Int_t nvar = input.CountChar(':') + 1;
        input.ReplaceAll(":", " ");
        input += " ";

        istringstream inStream((char *) input.Data());
        string inName;
        UInt_t nInputs = 0;
	vector<TString> foundIndex; foundIndex.clear();
        while (inStream >> inName)
	  {
            // Recherche si c'est une feuille ou une formule
            TLeaf *leaf = t->GetLeaf(inName.c_str());
            if (leaf != NULL)
	      {
		if( TString( leaf->GetName() ).EndsWith("__n__iter__") )
		  foundIndex.push_back( TString( leaf->GetName() ) );
		
		if( isAttribute( leaf->GetName() ) )
		    continue;

		Inputs+=TString(leaf->GetName())+":";
		
		_dataSpecification->addAttribute( new TAttribute(leaf->GetName()));
		_dataSpecification->getAttribute( leaf->GetName())->delShare();
		TBranch *buffer_br = leaf->GetBranch();
		if( buffer_br )
		  {
		    if ( string(buffer_br->GetClassName())=="vector<double>" )
		      {
			_dataSpecification->getAttribute( leaf->GetName() )->setDataType("V");
			if ( !isAttribute( (SIZENAME+"__"+string(leaf->GetName())).c_str() ) )
			  _dataSpecification->addAttribute( new TAttribute( (SIZENAME+"__"+string(leaf->GetName())).c_str() ) );
			_dataSpecification->getAttribute( (SIZENAME+"__"+string(leaf->GetName())).c_str() )->setOrigin(TAttribute::kInternal);
			_dataSpecification->getAttribute( (SIZENAME+"__"+string(leaf->GetName())).c_str() )->delShare();
		      }
		    else if ( string(buffer_br->GetClassName())=="string" )
		      _dataSpecification->getAttribute( leaf->GetName() )->setDataType("S");
		  }
	      }
            else
	      {
		AttributeToAdd.push_back( Form("%s", inName.c_str()) );
		// _dataSpecification->addAttribute( new TAttributeFormula(Form("var%d", nInputs), Form("%s", inName.c_str())));
		// _dataSpecification->getAttribute( Form("var%d", nInputs) )->delShare();
	      }
	    nInputs++;
	  }
	
	Inputs.Replace(Inputs.Length() - 1, 1, "");
	
	try
	  {
	    if ( foundIndex.size() > 1 )
	      throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 Form(
							      "TDataServer::ntupleDataRead: found %lu possible iterator in this file. No one will be chosen", foundIndex.size()));
	  }catch(URANIE::Exceptions::UWarningExceptions &ue)
	  {
	    ue.printMessage();
	  }

	
	// Ajout de l'iterateur Name+"__n__iter__" s'il nexiste pas
        TString siteratorname = TString(getIteratorName());
        Bool_t haveIndex = _dataSpecification->isAttribute(siteratorname);
        TAttribute* index = NULL;
		
        if (haveIndex)
	  {
            index = _dataSpecification->getAttribute(siteratorname);
            index->setOrigin(TAttribute::kIterator);
	  }
        else
	  {
	    if (foundIndex.size()==1)
	      {
		try
		  {
		    throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
								 Form("TDataServer::ntupleDataRead: Expected iterator %s not found but %s looks like an URANIE iterator => Will be used as so.",
								      siteratorname.Data(),foundIndex.at(0).Data()));
		  }catch(URANIE::Exceptions::UWarningExceptions &ue)
		  {
		    ue.printMessage();
		  }
		haveIndex=true;
		index = _dataSpecification->getAttribute( foundIndex.at(0).Data() );
		index->setOrigin(TAttribute::kIterator);
		changeIteratorName( foundIndex.at(0).Data() );
	      }
	    else
	      {
		index = new TAttribute(siteratorname);
		index->setOrigin(TAttribute::kIterator);
		_dataSpecification->addAttribute(index);
		index->delShare();
		nvar++;
	      }
	  }
	
	if(_datatree) 
	  {
	    delete _datatree;
	    _datatree=NULL;
	  }

	gROOT->cd();
	_datatree = new TDSNtupleD(t, Inputs, cut, true);
	_datatree->SetDirectory(0);
	f->Close();
	
	for(vector<TString>::iterator itAtt = AttributeToAdd.begin(); itAtt!=AttributeToAdd.end(); itAtt++)
	  addAttribute(*itAtt, *itAtt);
	_norigin = kTree;
	
        if (_blog)
	  {
            cout << " ** TDataServer::ntupleDataRead read " << getNPatterns()
		 << " points with svar " << Inputs << endl;
            _datatree->Print();
            _dataSpecification->printLog();
	  }
	
	if (_blog)
	  cout << " ** TDataServer::ntupleDataRead saving in " << getArchiveFileName() << " points" << endl;

        // Save the TTree in a TFile
        TFile *tf = TFile::Open(getArchiveFileName(), "RECREATE");
	//_datatree->SetDirectory(0);
        _datatree->Write();
        tf->Close();
	//_datatree->SetDirectory(gDirectory);
	
        if (_blog)
	  cout << " ** TDataServer::ntupleDataRead saved in " << getArchiveFileName() << " points" << endl;

      } catch (URANIE::Exceptions::UExceptions& ue)
      {
        ue.printMessage();
        if (ue.getMessageType() > URANIE::Exceptions::UExceptions::kWARNING)
	  {
            throw ue;
            exit(-1);
	  }
      }
    
    // Racine du nom pour l'arbre
    _sdatafile = gSystem->DirName(file);
    if (_sdatafile.CompareTo(".", TString::kExact) == 0)
      _sdatafile = gSystem->pwd();

    char *tmp =  gSystem->ConcatFileName(_sdatafile.Data(), gSystem->BaseName(file));
    _sdatafile = tmp;
    delete[] tmp;
    
    return kTRUE;
  }

  Bool_t TDataServer::loadTree(TTree* tt)
  {
    if (_blog)
      cout << "TDataServer::ReadTree[" << tt->GetName() << "]" << endl;

    //SetNameTitle(tt->GetName(), tt->GetTitle());
    SetName(tt->GetName()); SetTitle( tt->GetTitle());
    
    //Initialize the _shortDataTreeName
    _shortDataTreeName = getDataTreeName();

    //Prevent from piling up new TDataSpecification
    if(_dataSpecification) delete _dataSpecification;
    
    _dataSpecification = new URANIE::DataServer::TDataSpecification();
    _dataSpecification->setRelationName(
					Form("from TTree[%s]title[%s]", tt->GetName(), tt->GetTitle()));

    Int_t nvar = 0;
    // Parcours la liste des variables contenues dans le TTree tt
    TObjArray *listOfLeaves = tt->GetListOfLeaves();

    if (_blog)
      listOfLeaves->Print();

    TLeaf *leaf;
    TIter next(listOfLeaves);
    vector<TString> foundIndex; foundIndex.clear();
    while ((leaf = (TLeaf*) next()))
      {
	if( isAttribute( leaf->GetName() ) )
	  continue;
	
        _dataSpecification->addAttribute(new TAttribute(leaf->GetName()));
	_dataSpecification->getAttribute( leaf->GetName())->delShare();
	TBranch *buffer_br = leaf->GetBranch();
	if( buffer_br )
	  {
	    if ( string(buffer_br->GetClassName())=="vector<double>" )
	      {
		_dataSpecification->getAttribute( leaf->GetName() )->setDataType("V");
		if ( !isAttribute( (SIZENAME+"__"+string(leaf->GetName())).c_str() ) )
		  _dataSpecification->addAttribute( new TAttribute( (SIZENAME+"__"+string(leaf->GetName())).c_str() ) );
		_dataSpecification->getAttribute( (SIZENAME+"__"+string(leaf->GetName())).c_str() )->setOrigin(TAttribute::kInternal);
		_dataSpecification->getAttribute( (SIZENAME+"__"+string(leaf->GetName())).c_str() )->delShare();
	      }
	    else if ( string(buffer_br->GetClassName())=="string" )
	      _dataSpecification->getAttribute( leaf->GetName() )->setDataType("S");
	  }
        nvar++;
      }
    
    try
      {
	if ( foundIndex.size() > 1 )
	  throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
						     Form(
							  "TDataServer::loadTree: found %lu possible iterator in this file. No one will be chosen", foundIndex.size()));
      }catch(URANIE::Exceptions::UWarningExceptions &ue)
      {
	ue.printMessage();
      }
    
    // Ajout de l'iterateur Name+"__n__iter__" s'il nexiste pas
    TString siteratorname = TString(getIteratorName());
    if (_blog)
      cout << "siteratorname[" << siteratorname << "]" << endl;

    Bool_t haveIndex = _dataSpecification->isAttribute(siteratorname);
    TAttribute* index = NULL;

    if (haveIndex)
      {
        index = _dataSpecification->getAttribute(siteratorname);
        index->setOrigin(TAttribute::kIterator);
      }
    else
      {
	if (foundIndex.size()==1)
	  {
	    try
	      {
		throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							     Form("TDataServer::loadTree: Expected iterator %s not found but %s looks like an URANIE iterator => Will be used as so.",
								  siteratorname.Data(),foundIndex.at(0).Data()));
	      }catch(URANIE::Exceptions::UWarningExceptions &ue)
	      {
		ue.printMessage();
	      }
	    haveIndex=true;
	    index = _dataSpecification->getAttribute( foundIndex.at(0).Data() );
	    index->setOrigin(TAttribute::kIterator);
	    changeIteratorName( foundIndex.at(0).Data() );
	  }
	else
	  {
	    index = new TAttribute(siteratorname);
	    index->setOrigin(TAttribute::kIterator);
	    _dataSpecification->addAttribute(index);
	    index->delShare();
	    nvar++;
	  }
      }
    // Remplissages des donnees
    TString sVariableNames = _dataSpecification->getVariablesNames();

    if(_datatree != NULL)
      {
        delete _datatree;
        _datatree = NULL;
      } 

    gROOT->cd();
    //method written to construct the tdsntupleD from a tree
    _datatree = new TDSNtupleD(tt);
    _datatree->SetMarkerColor(2);
    _datatree->SetMarkerSize(4);
    _datatree->SetMarkerStyle(7);
    
    return kTRUE;
  }

  ////////////////////////////////////////////////////////////
  //  merge a TDS
  ////////////////////////////////////////////////////////////
  void TDataServer::merge(TDataServer *tds, const char* varexpinput, bool bwithFriend)
  {
    //Mod by JB Blanchard 2016-02-19: Change the way the merger work: propose to get rid of friend way (change default)
    // to properly merge TDSNtupleD instead (method is in TDSNtupleD). In this case, can select which attribute to merge.
    try
      {
        Int_t nEntries = getNPatterns();
        Int_t ncrt = tds->getNPatterns();
        if (nEntries != ncrt)
	  {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form(
							    "TDataServer::merge ERROR : the 2 TDS have not the same number of patterns\n First  : %d - Name = \"%s\" Title = \"%s\"\n Second : %d - Name = \"%s\" Title = \"%s\"",
							    getNPatterns(), GetName(), GetTitle(),
							    tds->getNPatterns(), tds->GetName(),
							    tds->GetTitle()));
	  }

        if (_blog)
	  cout << "TDataServe::Merge Patterns[" << nEntries << "] -- "
	       << " Other TDS[" << ncrt << "]" << endl;

        // Les 2 dataservers sont de meme taille
        if (bwithFriend == kTRUE)
	  {
	    try
	      {
		if (strcmp(varexpinput,"*")!=0)
		  throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							       Form(
								    "TDataServer::merge WARNING\n Merge asked with ROOT friend method and a restricted number of attribute. Not possible, so all attribute will be used"));
	      } catch (URANIE::Exceptions::UWarningExceptions& ue)
	      {
		ue.printMessage();
	      }

	    TFile *f = TFile::Open(tds->getArchiveFileName());
	    if (!f)
	      {
		TFile *ttf = TFile::Open(tds->getArchiveFileName(), "RECREATE");
		//tds->getTuple()->SetDirectory(0);
		tds->getTuple()->Write(tds->getDataTreeName(),
				       TObject::kOverwrite);
		ttf->Close();
		//tds->getTuple()->SetDirectory(gDirectory);
		f = TFile::Open(tds->getArchiveFileName());
	      }

	    _friend = (TTree*) f->Get(tds->getDataTreeName());
	    if (_friend == NULL)
	      throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							 Form(
							      "TDataServer::merge[%s]  Error:: The root file [%s] does not contain [%s] TTree",
							      GetName(), tds->getArchiveFileName(),
							      tds->getDataTreeName()));
            if( !TThread::IsInitialized())
                _friend->SetDirectory(gROOT);
	    f->Close();
	    //	    _friend->SetDirectory(gDirectory);
	    _datatree->AddFriend(_friend);
	  } //End withfriend
	
	//Merge _dataSpecification
	Int_t nattr = tds->getNAttributes(false);
	string attr_names="", attr_oldnames="";
	vector<string> needed;
	if ( strcmp(varexpinput,"*")!=0 && !bwithFriend  )
	  UEntry::StringParser( string(varexpinput), ":", needed);

        for (Int_t ivar = 0; ivar < nattr; ivar++)
	  { // 0 / nattr

            try
	      {
                TAttribute *att = tds->getAttribute(ivar,false);
		if(att->getOrigin()==TAttribute::kIterator) continue; //JBB to be removed at some point
		
                TString sname = att->GetName();
                TString soldname = sname;

		if( strcmp(varexpinput,"*")!=0 &&
		    find( needed.begin(), needed.end(), string(soldname) ) == needed.end() )
		  continue;
		
                if (_blog)
		  cout << ivar << "/" << nattr << " : sname[" << sname << "]";
                if (_dataSpecification->isAttribute(sname))
		  {
                    sname = TString(
				    Form("%s.%s", tds->GetName(), soldname.Data()));

		    TLeaf *leaf = tds->GetTuple()->GetLeaf(soldname);
                    if (leaf)
		      {
                        leaf->SetName(sname);
		      }
                    att->SetNameTitle(sname,
				      Form("%s.%s", tds->GetName(), att->GetTitle()));
		  }

                if (_blog)
		  {
                    cout << " ********** att name[" << att->GetName()
			 << "] Title[" << att->GetTitle() << "]" << endl;
		  }

		//Add to _dataspecification if all variables are requested or only a few
		if( strcmp(varexpinput,"*")==0 || ( bwithFriend && strcmp(varexpinput,"*")!=0)
		    ||  find( needed.begin(), needed.end(),  string(soldname) )!= needed.end() )
		  {
		      _dataSpecification->addAttribute(att, true);		    
		    //concatenate the old and new name for the TDSNTupleD::Merger
		    if ( !bwithFriend  )
		      {
			attr_names+=string(sname)+":";
			attr_oldnames+=string(soldname)+":";
		      }
		    //In case a vector is requested
		    if(att->getDataTypeStr().EqualTo("V"))
		      {
			//If its internal size__ branch is not requested, add it by force !!
			if( strcmp(varexpinput,"*")!=0 &&
			    find( needed.begin(), needed.end(), SIZENAME+"__"+string(soldname) ) == needed.end() )
			  {
			    string newname=((sname!=soldname) ? string( tds->GetName())+"." : "")+SIZENAME+"__"+string(soldname);
			    string oldname=SIZENAME+"__"+string(soldname);
			    TLeaf *leaf = tds->GetTuple()->GetLeaf(oldname.c_str());
			    if (leaf)
			      leaf->SetName(newname.c_str());

			    attr_names+= newname+":";
			    attr_oldnames+=oldname+":";
			    _dataSpecification->addAttribute( new TAttribute( newname.c_str() ) );
			    _dataSpecification->getAttribute( newname.c_str() )->setOrigin(TAttribute::kInternal);
			    _dataSpecification->getAttribute( newname.c_str() )->delShare();
			  }			
		      }
		  }

	      } catch (URANIE::Exceptions::UWarningExceptions& ue)
	      {
                ue.printMessage();
                continue;
	      }
	  }

	// if not with friend, run it
	if (bwithFriend == kFALSE)
	  {	    
	    attr_names = attr_names.substr(0, attr_names.size()-1);
	    attr_oldnames = attr_oldnames.substr(0, attr_oldnames.size()-1);
	    getTuple()->Merger(tds->getTuple(), attr_names, attr_oldnames, true, getIteratorName(),tds->getIteratorName());
	  }
	
      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
        throw ue;
        exit(-1);
      }
    if (_blog)
      cout << " *************** End of TDataServe::Merge Patterns" << endl;

  }
  
  ////////////////////////////////////////////////////////////
  //  normalize attributes
  ////////////////////////////////////////////////////////////
  void TDataServer::normalize(const char* varexp, const char* suffix, ENormalisation method, bool global)
  {
    if (_blog)
      {
        cout << "********************************" << endl;
        cout << " ** TDataServer::normalize" << endl;
        cout << " **  varexp[" << varexp << "]" << endl;
        cout << " **  suffix[" << suffix << "]" << endl;
        cout << " **  method[" << method << "]" << endl;
      }

    // Compute the statistics for the input and the output attributes
    computeStatistic(varexp);

    // Get the current list of attributes
    TList *listOfAttributes = getListOfAttributes(varexp, false);
    TIter next(listOfAttributes);
    TDSNtupleD *TempNtuple=0; //To deal with vectors
    // Loop on each attribute on the list to normalize
    while (TAttribute *att = (TAttribute *) next())
      {
	bool isVector = att->getDataTypeStr().EqualTo("V");
	//Don't run on variable like string and non-constant size vector (but not redoing all warnings)
	if ( att->getMean() == (-1.0 / DLE_MAX) )
	  continue;
	
        TString svarnormalize = Form("%s%s", att->GetName(), suffix);
	unsigned int VectSize=0;
	if( att->getDataTypeStr().EqualTo("V") )
	  VectSize = TMath::Nint(getAttribute( (SIZENAME+"__"+string(att->GetName())).c_str() )->getMean());

        if (_blog)
	  cout << " *** att[" << att->GetName() << "] att normalize["
	       << svarnormalize << "]" << endl;
		
	Double_t dmean = getAttribute(att->GetName())->getMean( ((global) ? VectSize : 0) );
	Double_t dstd = getAttribute(att->GetName())->getStd( ((global) ? VectSize : 0) );
	Double_t dmin = getAttribute(att->GetName())->getMinimum( ((global) ? VectSize : 0) );
	Double_t dmax = getAttribute(att->GetName())->getMaximum( ((global) ? VectSize : 0) );

	//If the normalisation is not global, then 
	if(isVector && !global)
	  {
	    string suffixes[4]={"min","max","mean","std"};
	    string theline="", thebranch=""; vector<double> buffer;
	    for(unsigned int ipat=0; ipat<4; ipat++)
	      {
		_dataSpecification->addAttribute( new TAttribute( Form("%s_%s",suffixes[ipat].c_str(),att->GetName()) ) );
		_dataSpecification->getAttribute( Form("%s_%s",suffixes[ipat].c_str(),att->GetName()) )->setDataType("V");
		_dataSpecification->getAttribute( Form("%s_%s",suffixes[ipat].c_str(),att->GetName()) )->delShare();
		att->getStatisticalVector( suffixes[ipat].c_str(), buffer);
		theline+=UEntry::dumpVectorStr(&buffer)+((ipat==3) ? "" :" ");
		thebranch+=suffixes[ipat]+"_"+string(att->GetName())+((ipat==3) ? "": ":");
	      }

	    TempNtuple = new TDSNtupleD("temporaryNtuple","the temp ntuple", thebranch.c_str(), "V:V:V:V"); 
	    for(int ipat=0; ipat<getNPatterns(); ipat++)
	      TempNtuple->Fill( theline );

	    _datatree->AddFriend( TempNtuple );
	    //_datatree->Merger( TempNtuple, thebranch.c_str(), thebranch.c_str());
	  }
	
	switch (method)
	  {
	  case kCR:
	    { // Centered and Reducted (Standard Deviation equal to 1.0)
	      if (_blog)
		cout << "   *** kCR :: mean[" << dmean << "] std(" << dstd
		     << ")" << endl;
	      if( isVector && !global )
		addAttribute(Form("%s", svarnormalize.Data()), Form("(%s-mean_%s)/std_%s", att->GetName(), att->GetName(), att->GetName()));
	      else
		addAttribute(Form("%s", svarnormalize.Data()), Form("(%s-%e)/%e", att->GetName(), dmean, dstd));
	      break;
	    }
	  case kCentered:
	    { // Centered
	      if (_blog)
		cout << "   *** kCentered :: mean[" << dmean << "]" << endl;
	      if( isVector && !global )
		addAttribute(Form("%s", svarnormalize.Data()), Form("(%s-mean_%s)", att->GetName(), att->GetName()));
	      else
		addAttribute(Form("%s", svarnormalize.Data()), Form("%s-%e", att->GetName(), dmean));
	      break;
	    }
	  case kCenteredSum0fSquareOne:
	    { // Centered and Sum Of Squared equal to 1.0 (different to kCR)
	      if (_blog)
		cout << "   *** kCenteredSum0fSquare1 :: mean[" << dmean << "] std(" << dstd << ") npatterns(" << getNPatterns() << ")" << endl;
	      addAttribute(Form("%s", svarnormalize.Data()), Form("(%s-%e)/%e", att->GetName(), dmean, dstd*TMath::Sqrt(1.0*getNPatterns()-1.0) ));
	      break;
	    }
	  case kMinusOneOne:
	    { // on [-1.0, 1.0]
	      if (_blog)
		cout << "   *** kMinusOneOne :: min[" << dmin << "] max("
		     << dmax << ")" << endl;
	      if( isVector && !global )
		addAttribute(Form("%s", svarnormalize.Data()), Form("2.0*(%s-min_%s)/(max_%s-min_%s) - 1.0", att->GetName(), att->GetName(), att->GetName(), att->GetName()));
	      else
		addAttribute(Form("%s", svarnormalize.Data()), Form("2.0*(%s-%e)/(%e-%e) - 1.0", att->GetName(), dmin, dmax, dmin));
	      break;
	    }
	  case kZeroOne:
	    { // on [0.0, 1.0]
	      if (_blog)
		cout << "   *** kZeroOne :: min[" << dmin << "] max(" << dmax
		     << ")" << endl;
	      if( isVector && !global )
		addAttribute(Form("%s", svarnormalize.Data()), Form("(%s-min_%s)/(max_%s-min_%s)", att->GetName(), att->GetName(), att->GetName(), att->GetName()));
	      else
		addAttribute(Form("%s", svarnormalize.Data()), Form("(%s-%e)/(%e-%e)", att->GetName(), dmin, dmax, dmin));
	      break;
	    }
	  default:
	    {
	      cout << " *** method not known" << endl;
	    }
	  }//End of switch method

	if(isVector && !global)
	  {
	    _datatree->RemoveFriend(TempNtuple);
	    string suffixes[4]={"min","max","mean","std"};
	    for(unsigned int ipat=0; ipat<4; ipat++)
	      _dataSpecification->delAttribute( Form("%s_%s",suffixes[ipat].c_str(), att->GetName() ) );
	    delete TempNtuple;
	  }
	
      }
    
    if(listOfAttributes!=NULL)
      delete listOfAttributes;
    listOfAttributes=NULL;
    
    if (_blog)
      {
        cout << " *** End of TDataServe::normalize" << endl;
        cout << "********************************" << endl;
      }
  }
  
  ///////////////////////////////////////////////////////////
  //  Mix the patterns
  ///////////////////////////////////////////////////////////
  void TDataServer::mix()
  {
    cout << " ** TDataServer::mix" << endl;
    //Clean
    TString stel = Form("uranie_pel_%s", GetName());
    cout << " ** stel[" << stel << "]" << endl;

    if( gROOT->FindObject(Form("%s", stel.Data())) ) gROOT->Delete(Form("%s", stel.Data()));

    //  TEventList * pelmix = _datatree->GetEventList();
    TPatternsEventList * pelmix =
      dynamic_cast<TPatternsEventList *>(_datatree->GetEventList());
    if (!pelmix)
      {
        cout << "no eventlist" << endl;
        //    pelmix = new TEventList(stel.Data(), "all patterns", getNPatterns());
        pelmix = new TPatternsEventList(stel.Data(), "all patterns",
					getNPatterns());
      }
    pelmix->Print("all");
    pelmix->Mix();
    pelmix->Print("all");
    _datatree->SetEventList(pelmix);

  }

  ///////////////////////////////////////////////////////////
  //  Manipulation des patterns
  ///////////////////////////////////////////////////////////
  Int_t TDataServer::getNPatterns()
  {
    Int_t npat = 0;
    try
      {
	if (_datatree == NULL)
	  throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
						       Form(
							    "TDataServer::getNPatterns[%s] WARNING\n The TTree is NULL",
							    GetName()));
	// if(strcmp(_ccut.GetTitle(),"")==0)
	//   npat = (Int_t)(_datatree->GetEntries());
	// else
	  npat = (Int_t)(_datatree->GetEntries( _ccut.GetTitle() ) );

      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
	ue.printMessage();
      }
    return npat;
  }

  ///////////////////////////////////////////////////////////
  //  Manipulation des attributs
  ///////////////////////////////////////////////////////////
  void TDataServer::setIterator(TString svar)
  {
    TAttribute *att = _dataSpecification->getAttribute(svar);
    if (att != NULL)
      {
        att->setOrigin(TAttribute::kIterator);
      }
    else
      throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						 Form(
						      "TDataServer::setIterator [%s]  Variable[%s] is not a variable of the tree",
						      GetName(), svar.Data()));
  }

void TDataServer::addAttribute(TString name, Double_t dmin, Double_t dmax)
{
    TAttribute *att = new TAttribute(name, dmin, dmax);
    try
    {
	addAttribute( att );
    }catch(URANIE::Exceptions::UErrorExceptions& ue)
    {
	delete att;
	throw ue;
    }    
    getAttribute(name)->delShare();
}

void TDataServer::addAttributeUsingData(TString name, double *arr, int size)
{

    clearSelect();
    int npat=((_datatree==NULL)?0:getNPatterns());
    
    try{

        if( npat == 0  && _dataSpecification != NULL )
        {
            if( _dataSpecification->getNAttributes() > 1)
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                           Form("TDataServer::addAttributeUsingData[%s]: ERROR: there's no tree but there are attributes in this TDataServer.\n\
This method should be used when no data and no other attributes are there OR when there are data already.",
                                                               name.Data()));
            
        }

        if(size != npat && npat != 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form(
                                                           "TDataServer::addAttributeUsingData[%s]: ERROR: the size of the array provided is not correct\n\
It is stated as [%d] while it should be equal to the size of dataserver's content [%d]",
                                                           name.Data(), size, npat));

        if(arr==NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       "TDataServer::addAttributeUsingData[%s]:  ERROR: the array provided is empty (null)");
        
    }catch(URANIE::Exceptions::UErrorExceptions& ue)
    {
	ue.printMessage();
        throw ue;
    }

    TAttribute *att = new TAttribute(name);
    try
    {
        addAttribute( att );
    }catch(URANIE::Exceptions::UErrorExceptions& ue)
    {
        delete att;
        throw ue;
    }    
    getAttribute(name)->delShare();

    if (_datatree == NULL)
    {
        (*_fLogger) << "This dataserver does not contain data so far. A new tree will be created, if this is what you meant." << Endl;
        createTuple();

        double values[2];
        for (Int_t ipat = 0; ipat < size; ipat++)
        {
            values[0]=ipat+1;
            values[1]=arr[ipat];
            _datatree->Fill(values);
        }
        
    }
    else
    {    
        //memory buffer for branches
        Double_t* tabValue=NULL;

        //Pointers for branches (_size is in case of vectors) and their memory buffer
        TBranch *brank=0;
        tabValue = new Double_t[1];
        brank = _datatree->Branch(name.Data(), tabValue, Form("%s/D", name.Data()));

        //Add them in the tree
        for (Int_t ipat = 0; ipat < size; ipat++)
        {            
            tabValue[0]=arr[ipat];
            brank->Fill();
        }
        _datatree->ResetBranchAddress(brank);

        delete[] tabValue; tabValue=NULL;
    }
    if (_blog)
      cout << "****** End Of TDataServer::computeRank" << endl;
    
}


  //! Adds an attribute with the name and title
  /*!
    \param name: the name of the attribute
    \param title: the title of the attribute
   */
  void TDataServer::addAttributeByNameTitle(TString name, TString title)
  {
    TAttribute *att = new TAttribute(name.Data(), title.Data());
    try
      {
	addAttribute( att );
      }catch(URANIE::Exceptions::UErrorExceptions& ue)
      {
	delete att;
	throw ue;
      }    
    getAttribute(name)->delShare();
  }

  //! Adds an attribute with the name only
  /*!
    \param name: the name of the attribute
    \param title: the title of the attribute
   */
  void TDataServer::addAttribute(const char *name)
  {
    TAttribute *att = new TAttribute(TString(name));
    try
      {
	addAttribute( att );
      }catch(URANIE::Exceptions::UErrorExceptions& ue)
      {
	delete att;
	throw ue;
      }    
    getAttribute(name)->delShare();
  }

  
void TDataServer::addAttribute(TString name, TString formula, TString title,  TString unit, bool update)
  {
    // modified by ngi: 2013-04-19
    // add exception for the case of string branch and vector one as well

    if (_blog)
      cout << " ** TDataServer::AddAttribut name[" << name << "] title["
	   << title << "] unit[" << unit << "] formula[" << formula << "]"
	   << endl;
    
    try
      {
	// The attribute
	TAttribute *att = 0;

	if ((_dataSpecification != NULL) && (_dataSpecification->isAttribute(name)))
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form(
							  "TDataServer::addAttribute[%s]  ERROR: Attribute %s already exists",
							  name.Data(), name.Data()));
	else
	  {
	    if (formula.Strip().IsNull())
	      throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							 Form(
							      "TDataServer::addAttribute[%s]  ERROR: an empty formula was given",
							      name.Data()));
	    else
	      {
		att = new TAttributeFormula(name, formula);
		att->delShare();
		addAttribute(att);
	      }
	  }

	// Title
	if (title.Length() != 0)
	  att->SetTitle(title);

	// Unit
	if (unit.Length() != 0)
	  att->setUnity(unit);

	// If the tree exists, fill the attribute
	if (_datatree)
            fillAttribute(name,"",update);
      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
	ue.printMessage();
	throw ue;
      }
    if (_blog)
      cout << " ** End Of TDataServer::AddAttribut name[" << name
	   << "] formula[" << formula << "]" << endl;
  }

  void TDataServer::fillAttribute(TString name, TString formula, Bool_t update)
  {
    // ngi: 2013-04-19
    //Mod by JB Blanchard 2016-02-19: Method get exception when using a branch with string. Add detection of vectors
    try
      {
	 
        if (!_datatree)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form(
							  "TDataServer::fillAttribute[%s]  ERROR: the TTree is NULL",
							  name.Data()));
      }catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
	ue.printMessage();
	throw ue;
      }

    TString localFormula = "";
    
    // The call to TDataServer::getAttribute will raise an exception 
    // if the attribute does not exist, so we use it as an existence 
    // check.
    TAttributeFormula* att =
      dynamic_cast<TAttributeFormula *>(getAttribute( name));

    try
      {

	// Check if the attribute is a TAttributeFormula or not, 
	// and retrieve the formula accordingly.
	if (att != NULL)
	  {
	    if (formula != "")
	      throw URANIE::Exceptions::UErrorExceptions(__FILE__,
							 __LINE__,
							 Form(
							      "TDataServer::fillAttribute[%s]  ERROR: when %s is a TAttributeFormula, no other formula should be given.",
							      name.Data(), name.Data()));
	    else
	      localFormula = att->getFormula();
	  }
	else
	  {
	    if (formula == "")
	      throw URANIE::Exceptions::UErrorExceptions(__FILE__,
							 __LINE__,
							 Form(
							      "TDataServer::fillAttribute[%s]  ERROR: no formula was given.",
							      name.Data()));
	    else
	      localFormula = formula;
	  }
	att = 0;

      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
	ue.printMessage();
	throw ue;
      }

    TTreeFormula *inForm=NULL;
    try
      {
        // Create the formula                                
        inForm = new TTreeFormula(name.Data(),
						localFormula.Data(), _datatree);
        // Verify that the formula is correct
        if (inForm->GetNdim()== 0)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form(
							  "TDataServer::fillAttribute[%s]  ERROR: the formula [%s] of the attribute [%s] is not correct",
							  name.Data(), localFormula.Data(), name.Data()));

        // Verify that the branch does not exist
        if (_datatree->GetBranch(name.Data()) != NULL)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form(
							  "TDataServer::fillAttribute[%s]  ERROR: the variable [%s] is already filled",
							  name.Data(), name.Data()));

        // Create the branch
	bool isVector=false; //detect whether branch is a vector
	// while( )
	if( !inForm->GetExpFormula().Contains("$") )
	  for(int ileaf=0; ileaf<inForm->GetNcodes(); ileaf++)
	    if( inForm->GetLeafInfo(ileaf) != 0 )
	      if( _dataSpecification->getAttribute( inForm->GetLeaf(ileaf)->GetName() )->getDataTypeStr().EqualTo("V") )	    
		isVector=true;

	TBranch *bAttribute=NULL, *bAttribute_size=NULL;
	if( !isVector )
	  {
	    Double_t* tabValue = new Double_t[1];
	    bAttribute = _datatree->Branch(name.Data(), tabValue, Form("%s/D", name.Data()));

	    // Fill the branch                
	    for (Int_t n = 0; n < _datatree->GetEntries(); n++)
	      {
		_datatree->GetEntry(n);
		tabValue[0] = inForm->EvalInstance();
		bAttribute->Fill();
	      }

	    delete[] tabValue; tabValue = 0;
	  }
	else // new vector case
	  {
	    TAttribute *att_size = new TAttribute((SIZENAME+"__"+string(name)).c_str());
	    att_size->setOrigin(TAttribute::kInternal);
	    att_size->delShare();
	    _dataSpecification->addAttribute(att_size);
	    _dataSpecification->getAttribute( name )->setDataType("V");

	    vector<Double_t>  *tabValue = new vector<Double_t>();
	    Double_t* tabValue_size = new Double_t[1];
	    bAttribute = _datatree->Branch(name.Data(), &tabValue, "vector<double>");
	    bAttribute_size = _datatree->Branch((SIZENAME+"__"+string(name)).c_str(), tabValue_size, Form("%s/D", (SIZENAME+"__"+string(name)).c_str()) );

	    // Fill the branch                
	    for (Int_t n = 0; n < _datatree->GetEntries(); n++)
	      {
		_datatree->GetEntry(n);
		tabValue->clear();	      
		for (Int_t ncase = 0; ncase < inForm->GetNdata(); ncase++)
		  tabValue->push_back(inForm->EvalInstance(ncase));
		tabValue_size[0] = inForm->GetNdata();
		bAttribute->Fill();
		bAttribute_size->Fill();
	      }
	    delete tabValue;
	    delete[] tabValue_size;
	  }

	if (_blog)
	  cout << " ** Info : create the branch [" << bAttribute->GetName()
	       << "]" << endl;

        // Clean
	if( bAttribute )
	  _datatree->ResetBranchAddress(bAttribute);
	if( bAttribute_size )
	  _datatree->ResetBranchAddress(bAttribute_size);

	delete inForm;

        if (update)
	  {
            // Save the data in a TFile
            TFile *ttf = TFile::Open(getArchiveFileName(), "UPDATE");
            //TFile *ttf = TFile::Open(getArchiveFileName(), "RECREATE");
            _datatree->SetDirectory(0);
            _datatree->Write(); // (getDataTreeName(), TObject::kOverwrite);
            ttf->Close();
            if( !TThread::IsInitialized())
                _datatree->SetDirectory(gROOT);
	  }

      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
	if(inForm!=NULL)
	  delete inForm;
	inForm=NULL;	
        throw ue;
      }
  }


void TDataServer::fillConstantAttribute(TString name, UEntry *ConstUEntry)
{
    // ngi: 2013-04-19
    //Created by JB Blanchard 2019-06-25
    try
    {
	 
        if (!_datatree)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form(
                                                           "TDataServer::fillConstantAttribute[%s]  ERROR: the TTree is NULL",
                                                           name.Data()));
    }catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
	ue.printMessage();
	throw ue;
    }

    try
    {
        // Verify that the branch does not exist
        if (_datatree->GetBranch(name.Data()) != NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form(
                                                           "TDataServer::fillConstantAttribute[%s]  ERROR: the variable [%s] is already filled",
                                                           name.Data(), name.Data()));
        if ( ConstUEntry == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form(
                                                           "TDataServer::fillConstantAttribute[%s]  ERROR: the UEntry pointer is NULL",
                                                           name.Data()));
        
        else if ( ConstUEntry->tag==URANIE::DataServer::UEntry::Empty)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form(
                                                           "TDataServer::fillConstantAttribute[%s]  ERROR: the UEntry pointer has no type defined",
                                                           name.Data()));

        addAttribute(name.Data());

	TBranch *bAttribute=NULL, *bAttribute_size=NULL;
        switch(ConstUEntry->tag)
        {
        case URANIE::DataServer::UEntry::Double:
        {
	    Double_t* douValue = new Double_t[1];
	    bAttribute = _datatree->Branch(name.Data(), douValue, Form("%s/D", name.Data()));
            douValue[0] = *ConstUEntry->val.dou;
            
	    // Fill the branch                
	    for (Int_t n = 0; n < _datatree->GetEntries(); n++)
            {
		_datatree->GetEntry(n);
		bAttribute->Fill();
            }
            
	    delete[] douValue; douValue = 0;
            break;
        }   
	case URANIE::DataServer::UEntry::Vector:
        {   
	    TAttribute *att_size = new TAttribute((SIZENAME+"__"+string(name)).c_str());
	    att_size->setOrigin(TAttribute::kInternal);
	    att_size->delShare();
	    _dataSpecification->addAttribute(att_size);
	    _dataSpecification->getAttribute( name )->setDataType("V");
            
	    vector<Double_t>  *vecValue = new vector<Double_t>();
	    Double_t* vecValue_size = new Double_t[1];
	    bAttribute = _datatree->Branch(name.Data(), &vecValue, "vector<double>");
	    bAttribute_size = _datatree->Branch((SIZENAME+"__"+string(name)).c_str(), vecValue_size, Form("%s/D", (SIZENAME+"__"+string(name)).c_str()) );

            vecValue->clear();	      
            for (unsigned int ncase = 0; ncase < ConstUEntry->val.vec->size(); ncase++)
                vecValue->push_back( ConstUEntry->val.vec->at(ncase) );
            vecValue_size[0] = ConstUEntry->val.vec->size();
            
	    // Fill the branch                
	    for (Int_t n = 0; n < _datatree->GetEntries(); n++)
            {
		_datatree->GetEntry(n);
		bAttribute->Fill();
		bAttribute_size->Fill();
            }
	    delete vecValue;
	    delete[] vecValue_size;
            break;
        }
        case URANIE::DataServer::UEntry::String:
        {
            string *tempo;
            tempo=ConstUEntry->val.str;
            bAttribute = _datatree->Branch(name.Data(), &tempo, "string");
            // Fill the branch                
	    for (Int_t n = 0; n < _datatree->GetEntries(); n++)
            {
		_datatree->GetEntry(n);
		bAttribute->Fill();
            }
            break;
        }
        default:
            break;
        }

	if (_blog)
            cout << " ** Info : create the branch [" << bAttribute->GetName()
                 << "]" << endl;

        // Clean
	if( bAttribute )
            _datatree->ResetBranchAddress(bAttribute);
	if( bAttribute_size )
            _datatree->ResetBranchAddress(bAttribute_size);


    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
}

  void TDataServer::fillOthersAttributes(bool update)
  {

    TAttribute *att = 0;
    TAttributeFormula *formulatt = 0;
    TList *attList = getListOfAttributes("*",false);
    string realattr="";
    vector<TString> formName, formForm;
    for (Int_t iatt = 0; iatt < attList->GetEntries(); iatt++)
      {
        att = (TAttribute *) attList->At(iatt);
        formulatt = dynamic_cast<TAttributeFormula *>(att);
        if (formulatt != NULL)
	  {
	    formName.push_back(TString(formulatt->GetName()));
	    formForm.push_back(TString(formulatt->getFormula()));
	    //	    delAttribute( formulatt->GetName() );
	  }
	else
	  realattr+=string(att->GetName())+":";
      }
    realattr=realattr.substr(0, realattr.size()-1);
    delete attList;

    TDSNtupleD *newtds = new TDSNtupleD(_datatree,realattr.c_str(),"",true);
    delete _datatree;
    _datatree=newtds;
    
    for(unsigned int iform=0; iform<formName.size(); iform++)
        fillAttribute(formName.at(iform),"",update);
      //      addAttribute(formName.at(iform),formForm.at(iform),"","",true);
    
  }

  void TDataServer::addAttribute(TAttribute *att)
  {

    if (_dataSpecification == NULL)
      {
        _dataSpecification = new TDataSpecification(
						    Form("uheader__%s", GetName()),
						    Form("Header of %s", GetTitle()));
        TAttribute* index = new TAttribute(TString(getIteratorName()));
        index->setOrigin(TAttribute::kIterator);
        _dataSpecification->addAttribute(index);
        index->delShare();
      }
    if (att != NULL)
      {
        _dataSpecification->addAttribute(att);
      }
  }

  void TDataServer::delAttribute(TString name)
  {
    if (_blog)
      cout << "TDataServer::delAttribute name[" << name << "]" << endl;

    int index = _dataSpecification->getAttributeIndex(name);
    try
      {
        if (index == -1)
	  {
            throw URANIE::Exceptions::UErrorExceptions( __FILE__, __LINE__,
							Form(
							     "TDataServer Name: \"%s\" Title\"%s\"\nAttribute \"%s\" to delete does not exist",
							     GetName(), GetTitle(), name.Data()));
	  }
        else
	  {
            _dataSpecification->delAttribute(name);
	  }

      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
        throw ue;
        exit(-1);
      }
        
  }

  Int_t TDataServer::getNInputAttributes()
  {
    Int_t natt = 0;
    if (_blog)
      cout << endl << " ** TDataServer::getNInputAttributes" << endl;

    try
      {
        if (_dataSpecification == NULL)
	  throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
						       Form(
							    "TDataServer::getNInputAttributes [%s] ERROR\n the TDataSpecification  is empty",
							    GetName()));
        natt = _dataSpecification->getNInputAttributes();


      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }

    if (_blog)
      cout << " ** End of TDataServer::getNInputAttributes[" << natt << "]"
	   << endl;
    return natt;
  }

  Int_t TDataServer::getNOutputAttributes()
  {
    Int_t natt = 0;
    if (_blog)
      cout << endl << " ** TDataServer::getNOutputAttributes" << endl;

    try
      {
        if (_dataSpecification == NULL)
	  throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
						       Form(
							    "TDataServer::getNOutputAttributes [%s] ERROR\n the TDataSpecification  is empty",
							    GetName()));
        natt = _dataSpecification->getNOutputAttributes();
      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
        ue.printMessage();
        throw ue;
      }

    if (_blog)
      cout << " ** End of TDataServer::getNOutputAttributes[" << natt << "]"
	   << endl;
    return natt;

  }
  
  Bool_t TDataServer::isAttribute(TString name) const
  {

    try
      {
	if(! _dataSpecification )
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     "TDataServer::isAttribute ERROR: the tds is empty");
      }catch(URANIE::Exceptions::UErrorExceptions &ue)
      {
	ue.printMessage();
	throw ue;
	//exit(-1);
      }
    
    return _dataSpecification->isAttribute(name);
  }

  int TDataServer::getAttributeIndex(TAttribute * att, bool onlyAttribute) const
  {

   try
      {
	if(! _dataSpecification )
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     "TDataServer::getAttributeIndex ERROR: the tds is empty");
      }catch(URANIE::Exceptions::UErrorExceptions &ue)
      {
	ue.printMessage();
	throw ue;
	//exit(-1);
      }
    
    return _dataSpecification->getAttributeIndex(att, onlyAttribute);
  }

  int TDataServer::getAttributeIndex(TString name, bool onlyAttribute) const
  {
    try
      {
	if(! _dataSpecification )
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     "TDataServer::getAttributeIndex ERROR: the tds is empty");
      }catch(URANIE::Exceptions::UErrorExceptions &ue)
      {
	ue.printMessage();
	throw ue;
	//exit(-1);
      }
   
    return _dataSpecification->getAttributeIndex(name, onlyAttribute);
  }

  TAttribute *TDataServer::getAttribute(TString str)
  {
    try
      {
	if(! _dataSpecification )
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     "TDataServer::getAttribute ERROR: the tds is empty");
      }catch(URANIE::Exceptions::UErrorExceptions &ue)
      {
	ue.printMessage();
	throw ue;
	//exit(-1);
      }
   
    return _dataSpecification->getAttribute(str);
  }
  
  Int_t TDataServer::getNAttributes(bool onlyAttribute)
  {
    Int_t natt = 0;
    if (_blog)
      cout << endl << "****** TDataServer::getNAttribut" << endl;

    if (_dataSpecification != NULL)
      {
	natt = _dataSpecification->getNAttributes(onlyAttribute);
      }
    
    return natt;
  }


  TList * TDataServer::getListOfAttributes(const char* sListOfAttributes, bool onlyAttribute)
  {

    try
      {
	if(! _dataSpecification )
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     "TDataServer::getListOfAttributes ERROR: the tds is empty");
      }catch(URANIE::Exceptions::UErrorExceptions &ue)
      {
	ue.printMessage();
	throw ue;
	//exit(-1);
      }

    return  _dataSpecification->getListOfAttributes(sListOfAttributes, onlyAttribute);
  }

string TDataServer::getStringListOfAttributes(bool onlyNumerical,bool onlyAttribute)
  {

    try
      {
	if(! _dataSpecification )
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     "TDataServer::getStringListOfAttributes ERROR: the tds is empty");
      }catch(URANIE::Exceptions::UErrorExceptions &ue)
    {
	ue.printMessage();
	throw ue;
	//exit(-1);
    }

    string res="";
    
    TList *list=_dataSpecification->getListOfAttributes("*", onlyAttribute);
    TIter next(list);
    while (TAttribute *att = (TAttribute *) next()) 
    {
        if( onlyNumerical==true && att->getDataTypeStr() == "S")
            continue;
        res+=((res=="")?string(att->GetName()):":"+string(att->GetName()));
    }
    delete list;
    return res;
  }
  
  TAttribute *TDataServer::getAttribute(Int_t index, bool onlyAttribute)
  {
    try
      {
        if (_dataSpecification == NULL)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form("TDataServer::getAttribute[%d]  the header is empty",
							  index));

      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
        throw ue;
        exit(-1);
      }

    return _dataSpecification->getAttribute(index, onlyAttribute);
  }

  //Modified by JBB 2016 -2-23 to get easily vector's value and too complains in few cases:
  // for string, for vector if ielement > size() or ==-1, if the attribute does not exist
Double_t TDataServer::getValue(TString attName, Int_t entry, Int_t ielement)
{
    try
    {
	if (_datatree)
	{
	    if (isAttribute(attName) )
	    {
		if(  getAttribute(attName)->getDataTypeStr().EqualTo("D") )
		{
		    TLeaf* att = _datatree->GetLeaf(attName);
		    if(!att)
			throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
								   Form(
								       "TDataServer::getValue method\n The attribute [%s] exit but not the leaf in the tree",
								       attName.Data()));
		    if(entry < getNPatterns() && entry >= 0)
		    {
			_datatree->GetEntry(entry);
			return att->GetValue();
		    }
		    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							       Form(
								   "TDataServer::getValue method\n The entry [%d] does not exist in the TDS with the attribute [%s]. The entry must be between [0] and [%d].",
								   entry, attName.Data(), getNPatterns()-1));
		}
		else if ( getAttribute(attName)->getDataTypeStr().EqualTo("V") )
		{
		    TLeaf* att = _datatree->GetLeaf(attName);
		    if(!att)
			throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
								   Form(
								       "TDataServer::getValue method\n The attribute [%s] exit but not the leaf in the tree",
								       attName.Data()));
		    _datatree->GetEntry(entry);
		    
		    if(ielement<0 || UInt_t(ielement) >= ((vector<double>*)att->GetValuePointer())->size() )
			throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
								   Form(
								       "TDataServer::getValue method\n The attribute [%s] has a size %lu and cannot contain the index %d",
								       attName.Data(),((vector<double>*)att->GetValuePointer())->size(),ielement));
		    return ((vector<double>*)att->GetValuePointer())->at(ielement);
		}
		else if ( getAttribute(attName)->getDataTypeStr().EqualTo("S") )
		    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							       Form(
								   "TDataServer::getValue method\n The attribute [%s] is a string attribute. In order to read this information, use TDataServer::getString() method",
								   attName.Data()));
	    }
	    else
		throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							   Form(
							       "TDataServer::getValue method\n The attribute [%s] does not exist in the TDS with entry [%d]",
							       attName.Data(), entry));
	}
	else
	    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						       Form(
							   "TDataServer::getValue method\n The tree is not existing yet"));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
	ue.printMessage();
	throw ue;
	exit(-1); // Mod by JBB Not convinced it is a good idea
    }

    return -1.;
}

  vector<double> TDataServer::getVector(TString attName, Int_t entry)
  {
    try
      {
        if (_datatree)
	  {
	    if (!isAttribute(attName) )
	      throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							 Form(
							      "TDataServer::getVector method\n The attribute [%s] does not exist",
							      attName.Data()));

	    else if (!getAttribute(attName)->getDataTypeStr().EqualTo("V") )
	      throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							 Form(
							      "TDataServer::getVector method\n The attribute [%s] is a not a vector attribute, it's type is %s",
							      attName.Data(),getAttribute(attName)->getDataTypeStr().Data()));
	    
	    vector<double> thevec; thevec.clear();
	    TLeaf* att = _datatree->GetLeaf(attName);
	    if(!att)
	      throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							 Form(
							      "TDataServer::getVector method\n The attribute [%s] exit but not the leaf in the tree",
							      attName.Data()));	    
	    _datatree->GetEntry(entry);
	    vector<double> *buffer = ((vector<double>*)att->GetValuePointer());
	    for(unsigned int iel=0; iel<buffer->size(); iel++)
	      thevec.push_back( buffer->at(iel) );

	    return thevec;
	  }
	else
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form(
							  "TDataServer::getVector method\n The tree is not existing yet"));
      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
	throw ue;
	exit(-1); // Mod by JBB Not convinced it is a good idea
      }
  }

  string TDataServer::getString(TString attName, Int_t entry)
  {
    try
      {
        if (_datatree)
	  {
	    if (!isAttribute(attName) )
	      throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							 Form(
							      "TDataServer::getString method\n The attribute [%s] does not exist",
							      attName.Data()));
	    
	    else if (!getAttribute(attName)->getDataTypeStr().EqualTo("S") )
	      throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							 Form(
							      "TDataServer::getString method\n The attribute [%s] is a not a vector attribute, it's type is %s",
							      attName.Data(),getAttribute(attName)->getDataTypeStr().Data()));

	    TLeaf *att = (TLeaf*)_datatree->GetLeaf(attName);
	    if(!att)
	      throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							 Form(
							      "TDataServer::getString method\n The attribute [%s] exit but not the leaf in the tree",
							      attName.Data()));
	    _datatree->GetEntry(entry);
	    return *(string*)att->GetValuePointer();
	  }
	else
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form(
							  "TDataServer::getString method\n The tree is not existing yet"));
      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
	throw ue;
	exit(-1); // Mod by JBB Not convinced it is a good idea
      }
  }

  TMatrixD TDataServer::getMatrix(const char* varexp, const char* selection,
				  Option_t* option)
  {
    // fga::information 2012-03-30 08:58:33
    // suppress scan in the log
    //  if ( _blog) {
    //    _datatree->Scan(varexp, _ccut + selection);
    //  }
    // end of fga::information 2012-03-30 08:58:33
    TString soption = option;
    soption.ToLower();
    TString sListOfAttributes(varexp);

    if (sListOfAttributes == "*")
      {
        TList *listOfAttributes = getListOfAttributes(varexp);
        TIter next(listOfAttributes);
        TAttribute *att = (TAttribute *) next();
        sListOfAttributes = att->GetName();
        while ( (att = (TAttribute *) next()) )
	  {

	    try
	      {
		if( !att->getDataTypeStr().EqualTo("D") )
		  throw URANIE::Exceptions::UWarningExceptions( __FILE__, __LINE__,
								Form(
								     "TDataServer::getMatrix Warning :  The attribute %s is not a double precision one. It is not included in the matrix !!",
								     att->GetName()));
	      }catch(URANIE::Exceptions::UWarningExceptions &ue)
	      {
		ue.printMessage();
		continue;
	      }
	    
            sListOfAttributes += ":";
            sListOfAttributes += att->GetName();
	  }
        delete listOfAttributes;
      }

    Int_t nvar = sListOfAttributes.CountChar(':') + 1;
    //cout << " sy[" << sListOfAttributes << "] -- nvar[" << nvar << "]" << endl;

    sListOfAttributes.ReplaceAll(":", " ");
    sListOfAttributes += " ";

    istringstream inStream((char *) sListOfAttributes.Data());
    string inName;
    UInt_t nInputs = 0;
    TMatrixD vec;

    while (inStream >> inName)
      {
        //cout << " *** inName[" << inName << "] nInputs[" << nInputs<< "]" << endl;
        _datatree->SetEstimate(_datatree->GetEntries());
        _datatree->Draw(inName.c_str(), _ccut + selection, "goff");
        if (nInputs == 0)
	  vec.ResizeTo(nvar, (Int_t) _datatree->GetSelectedRows());

        vec.InsertRow(nInputs, 0, _datatree->GetV1());

        nInputs++;
      }
    vec.T();

    //vec.Print();

    return vec;

  }

  void TDataServer::createTuple(Option_t *option)
  {

    if (_dataSpecification == NULL)
      {
	_dataSpecification = new TDataSpecification(
						    Form("uheader__%s", GetName()),
						    Form("Header of %s", GetTitle()));
	TAttribute* index = new TAttribute(TString(getIteratorName()));
	index->setOrigin(TAttribute::kIterator);
	index->delShare();
	_dataSpecification->addAttribute(index);
      }
    TString sVariableNames = _dataSpecification->getVariablesNames();
    TString sVariableTypes = _dataSpecification->getVariablesTypes();
    TString siteratorname = TString(getIteratorName());
    TString soption(option);    
    if (sVariableNames.Contains(siteratorname))
      {
	int iterindex = getAttributeIndex( _dataSpecification->getAttribute(siteratorname) );
	vector<string> bufname, buftype; string  buflinename="", buflinetype="";
	UEntry::StringParser(string(sVariableNames), string(":"), bufname);
	UEntry::StringParser(string(sVariableTypes), string(":"), buftype);
	for(unsigned int ibuf=0; ibuf<bufname.size(); ibuf++)
	  {
	    if(UInt_t(iterindex)==ibuf) continue;
	    buflinename+=":"+bufname.at(ibuf);
	    buflinetype+=":"+buftype.at(ibuf);
	  }
	sVariableNames=TString(siteratorname+buflinename);
	sVariableTypes=TString("D"+buflinetype);
      }

    _shortDataTreeName = getDataTreeName();
    
    if (_blog)
      {
	cout << " ** TDataServer::createTuple Option[" << option << "]" << endl;
	cout << " ** sVariableNames[" << sVariableNames << "]" << endl;
	cout << " ** sVariableTypes[" << sVariableTypes << "]" << endl;
      }

    if (_datatree != NULL)
      {
	delete _datatree;
	_datatree = NULL;
      }

//    _datatree = new TDSNtupleD(Form("%s__data__tree__", _shortDataTreeName.Data()),
    _datatree = new TDSNtupleD(_shortDataTreeName.Data(),
			       _dataSpecification->getRelationName(), sVariableNames, sVariableTypes);

    _datatree->SetMarkerColor(2);
    _datatree->SetMarkerSize(4);
    _datatree->SetMarkerStyle(7);
    
    if( TThread::IsInitialized())
        _datatree->SetDirectory(0);
  }

  ////////////////////////////////////////////////////////////
  //  Exportation des donnees 
  ///////////////////////////////////////////////////////////
  void TDataServer::exportData(const char *filename, const char *varexp,
			       const char *select, Option_t *option)
  {
    if (_blog)
      {
	cout << endl << "****** TDataServer::exportData [" << GetName()
	     << "] Title[" << GetTitle() << "]" << endl;
	cout << " **     Size[" << getNPatterns() << "]" << endl;
	cout << " ** filename[" << filename << "]" << endl;
	cout << " **   varexp[" << varexp << "]" << endl;
	cout << " **   select[" << select << "]" << endl;
      }

    TList *listOfAttributes=NULL;
    TPatternsEventList * pelmix = NULL;
    bool pelmix_is_new=false;
    TEventList *telapp;
    Int_t npat;
    
    Int_t nPrecision = 9;
    TString sOption = option;
    bool onlyAttribute=true;
    bool bIntOutput=false;
    TString sFormat="%e";
    if(sOption.BeginsWith("precision="))
      {
	Int_t nIndE = sOption.First('=');
	Int_t nLengthO = sOption.Length();
	TString sNum = sOption(nIndE +1, nLengthO-nIndE);
	if(sNum.IsDigit())
	  nPrecision = sNum.Atoi();
      }

    sOption.ToLower();
    if( sOption.Contains("internal") )
	onlyAttribute=false;
    if( sOption.Contains("%i") )
    {
	bIntOutput=true; // Cancel the scientific writting
	sFormat="%i";
    }
    try
      {
	if (_datatree == NULL)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form(
							  "TDataServer::exportData[%s] in file[%s]\n The TTree is NULL",
							  GetName(), filename));
	if (_dataSpecification == NULL)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form(
							  "TDataServer::exportData ERROR : the TDataServer \"%s\" to export is empty",
							  GetName()));
      }catch(URANIE::Exceptions::UErrorExceptions& ue)
      {
	ue.printMessage();
	throw ue;
      }
    // Get the current list of attributes
    try
      {
	listOfAttributes = getListOfAttributes(varexp, onlyAttribute);
      }catch(URANIE::Exceptions::UErrorExceptions& ue)
      {
	if(listOfAttributes!=NULL)
	  delete listOfAttributes;
	throw ue;
      }
    
    if (_blog)
      {
	cout << " ** Iterator Name[" << getIteratorName() << "]" << endl;
	getAttribute(getIteratorName())->printLog();
      }
    if( !listOfAttributes->Contains( getAttribute(getIteratorName())  ) && !sOption.Contains("noiter"))
      listOfAttributes->Add(getAttribute(getIteratorName()));
    
    // Treat the selection
    TCut allCut; allCut += _ccut; allCut += TCut(select);
    TString scut = TString(_ccut);
    if (_blog)
      {
	cout << " ** _ccut[" << _ccut << "]" << endl;
	cout << " ** select[" << select << "]" << endl;
	cout << " ** allCut[" << allCut << "]" << endl;
	cout << " ** scut[" << scut << "]" << endl;
      }
    TString spel = Form("uranie_pel_%s", GetName());
    pelmix = dynamic_cast<TPatternsEventList *>(_datatree->GetEventList());
    if (!pelmix)
      {
	if (_blog)
	  cout << "no eventlist" << endl;
	//    pelmix = new TEventList(stel.Data(), "all patterns", getNPatterns());
	pelmix = new TPatternsEventList(spel.Data(), "all patterns",  getNPatterns());
	pelmix_is_new=true;
      }
    if (_blog)
      pelmix->Print("all");
    
    // Clean the TEventList
    if( gDirectory != gROOT ) gDirectory=gROOT;
    if( gROOT->FindObject("elistUranie") ) gROOT->Delete("elistUranie");
    _datatree->Draw(">>elistUranie", ((strcmp(allCut.GetTitle(),"")==0) ? "1>0": allCut.GetTitle()),"goff");
    telapp = (TEventList*) gROOT->Get("elistUranie");
    npat = telapp->GetN();
    if (_blog)
      {
	telapp->Print("all");
	cout << " ** elistUranie npat[" << npat << "]" << endl;
      }
    
    try
      {
	if (npat == 0)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form(
							  "TDataServer::exportData ERROR : the TDataServer \"%s\" to export have no pattern",
							  GetName()));
      }catch(URANIE::Exceptions::UErrorExceptions& ue)
      {
	ue.printMessage();
	if(pelmix_is_new==true)
	  delete pelmix;
	if(listOfAttributes!=NULL)
	  delete listOfAttributes;
	if( gROOT->FindObject("elistUranie") ) gROOT->Delete("elistUranie");
	throw ue;
      }

    try{
      // Open the output file
      ofstream sortie;
      sortie.open(filename, ios::out);
      sortie.precision(nPrecision);
      if( !bIntOutput ) sortie.setf(ios::scientific);
	
      // Header
      TDatime theTime;
      sortie << "#NAME: " << GetName() << endl;
      sortie << "#TITLE: " << GetTitle() << endl;
      sortie << "#DATE: " << theTime.AsString() << endl;
      sortie << "#COLUMN_NAMES: ";
	
      // Line of attributes names
      Int_t nleaves = listOfAttributes->GetEntries();
      TIter next(listOfAttributes);
      Int_t ivar = 0;
	
      TList *listOfLeaves = new TList();
      string StrType="", StrUnit="", StrTitle=""; bool AtLeastOneUnit=false;
      while (TAttribute *att = (TAttribute *) next())
	{
	  sortie << att->GetName();
	  StrType+=att->getDataTypeStr();
	  StrUnit+=string(att->getUnit());
	  if( string(att->getUnit()) != " " && string(att->getUnit()) != "" && !AtLeastOneUnit)
	    AtLeastOneUnit=true;
	  StrTitle+=string(att->getTitle());
	  if (_blog)
	    cout << " ** ivar[" << ivar << "] name[" << att->GetName() << "]" << endl;
	  if (ivar++ < nleaves - 1)
	    {
	      sortie << "| ";
	      StrType += "|";
	      StrUnit += "|";
	      StrTitle += "|";
	    }
	  if (_datatree->GetLeaf(att->GetName()))
	    listOfLeaves->Add(_datatree->GetLeaf(att->GetName()));
	  else
	    {
	      try
		{
		  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							     Form("TDataServer::exportData ERROR : the Leaf \"%s\" does not exist in the TTree object",
								  att->GetName()));
		}catch(URANIE::Exceptions::UErrorExceptions& ue)
		{
		  ue.printMessage();
		  delete listOfLeaves;
		  listOfLeaves = NULL;
		  throw ue;	  
		}
	    }
	}
      sortie << endl;
      sortie << "#COLUMN_TYPES: " << StrType << endl;
      if( AtLeastOneUnit )
	sortie << "#COLUMN_UNITS: " << StrUnit << endl;
      sortie << endl;
      if (_blog)
	{
	  cout << endl << " **  TLeaf listing  ["
	       << listOfLeaves->GetEntries() << "]" << endl;
	  listOfLeaves->Print();
	}

      //Write the data
      for (Int_t ipat = 0; ipat < npat; ipat++)
	{
	  _datatree->GetEntry(telapp->GetEntry(ipat));
	  if (_blog)
	    {
	      _datatree->Show();
	      cout << endl << " **  Entry  [" << telapp->GetEntry(ipat) << "/"
		   << npat << "]" << endl;
	    }
	  //        TIter nextLeaf(listOfLeaves);
	  //        while ( TLeaf *leaf = (TLeaf *)nextLeaf()) {
	  //      cout << " **** Attribute[" << leaf->GetName() << "] value[" << leaf->GetValue() << "]" << endl;
	  //      sortie << leaf->GetValue() << " ";
	  //        }
	  for (Int_t i = 0; i < nleaves; i++)
	    {
	      TLeaf *leaf = (TLeaf *) (listOfLeaves->At(i));
	      if (_blog)
		cout << " **** Attribute[" << leaf->GetName() << "] value["
		     << leaf->GetValue() << "]" << endl;
		
	      TString theclass = getAttribute( leaf->GetName() )->getDataTypeStr();
	      if( TString(leaf->GetName()).Contains("__n__iter__") )
	      {
		  int iterval = int(leaf->GetValue());
		  sortie << std::fixed << iterval << " "; //sortie.precision(nPrecision); sortie.setf(ios::scientific);
	      }
	      else if(theclass.EqualTo("D"))
	      {
		  if ( sFormat=="%i" )
		  {
		      int theval = int(leaf->GetValue());
		      sortie << std::fixed << theval << " ";
		  }
		  else
		      sortie << std::scientific << leaf->GetValue() << " ";
	      }
	      else if(theclass.EqualTo("S")) sortie << "\"" << *((string*)leaf->GetValuePointer()) << "\" ";
	      else if(theclass.EqualTo("V")) sortie << "[" << UEntry::dumpVectorStr((vector<double>*)(leaf->GetValuePointer()), nPrecision, string(","), sFormat) << "] ";
	    }
	  sortie << endl;
	}

      // Close the output file
      sortie.close();

      //Clean
      if(pelmix_is_new==true)
	delete pelmix;
      delete listOfLeaves;
      delete listOfAttributes;
      // Clean the TEventList
      if( gROOT->FindObject("elistUranie") ) gROOT->Delete("elistUranie");
      if (_blog)
	cout << "****** End of TDataServer::exportData" << endl;
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
	ue.printMessage();
	if( gROOT->FindObject("elistUranie") ) gROOT->Delete("elistUranie");
	if(pelmix_is_new==true)
	  delete pelmix;
	if (listOfAttributes != NULL)
	  delete listOfAttributes;
	throw ue;
      }

  }

  void TDataServer::exportDataHeader(const char *filename, const char *varexp,
             const char *select, Option_t *option)
  {

    if (_blog)
      {
        cout << endl << "****** TDataServer::exportDataHeader [" << GetName()
	     << "] Title[" << GetTitle() << "]" << endl;
        cout << " ** filename[" << filename << "]" << endl;
        cout << " **   varexp[" << varexp << "]" << endl;
        cout << " **   select[" << select << "]" << endl;
        cout << " **   option[" << option << "]" << endl;
      }

    TList *listOfAttributes=NULL;
    TEventList *telapp;
    Int_t npat;

    Int_t nPrecision = SSTR_PREC;
    TString sOption = option;
    sOption.ToLower();
    sOption.ReplaceAll(" ", "");
    sOption.ReplaceAll(",", " ");
    if(sOption.BeginsWith("precision=")) {
        TRegexp reg("precision=[0-9]*");
        TString out = sOption(reg);
        TString snclass = out.Remove(0, 10);
        if(snclass.Length() > 0 )
            nPrecision = snclass.Atoi();
    }
    if ( _blog )
        cout << " ** nPrecision[" << nPrecision << "]"<< endl;

    try
      {
        if (_dataSpecification == NULL)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form(
							  "TDataServer::exportDataHeader ERROR : the TDataServer \"%s\" to export is empty",
							  GetName()));

        // Get the current list of attributes
        listOfAttributes = getListOfAttributes(varexp);

        // Treat the selection and the option
        TCut allCut; allCut += _ccut; allCut += TCut(select);
        // CFO::WARNING
        //Version v2_1_10 CFO, raison inconnue.
        npat = getNPatterns();
        if (npat == 0)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form(
							  "TDataServer::exportDataHeader ERROR : the TDataServer \"%s\" to export have no pattern",
							  GetName()));

	//Check folder in which we're dumping list
	if( gDirectory != gROOT ) gDirectory=gROOT;
	if( gROOT->FindObject("elist1") ) gROOT->Delete("elist1");
        _datatree->Draw(">>elist1", ((strcmp(allCut.GetTitle(),"")==0)?"1>0":allCut.GetTitle()),"goff");
        telapp = (TEventList*) gROOT->Get("elist1");
        npat = telapp->GetN();

      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
        delete listOfAttributes;
        throw ue;
        exit(-1);
      }

    // Ouverture du fichier de sortie
    ofstream sortie;
    sortie.open(filename, ios::out);
    sortie.precision(nPrecision);
    sortie.setf(ios::scientific);

    // entete
    TDatime theTime;
    sortie
      << Form("// File \"%s\" generated by ROOT v%s", filename,
	      gROOT->GetVersion()) << endl;
    sortie << Form("// DateTime %s", theTime.AsString()) << endl;
    sortie << "// DataServer : Name=\"" << GetName() << "\" Title=\""
	   << GetTitle() << "\"";
    if (_ccut != 0)
      sortie << " Global Select=\"" << _ccut << "\"";
    TString sselect = TString(select);
    if (sselect.Length() > 0)
      sortie << " Local Select=\"" << sselect << "\"";
    sortie << endl << endl;

    sortie << "#define " << GetName() << "_nPattern " << npat << endl << endl;

    TList *listOfLeaves = new TList();
    TIter next(listOfAttributes);
    while (TAttribute *att = (TAttribute *) next())
      {
        listOfLeaves->Add(_datatree->GetLeaf(att->GetName()));
      }

    if (_blog)
      {
        Int_t nleaves = listOfAttributes->GetEntries();
        cout << endl << " **  TLeaf listing  [" << nleaves << "]" << endl;
        listOfLeaves->Print();
      }

    TIter nextLeaf(listOfLeaves);
    while (TLeaf *leaf = (TLeaf *) nextLeaf())
      {
	TString svar = leaf->GetName();
        if (_blog)
	  cout << " ** svar[" << svar << "]" << endl;
	
        sortie << "//  Attribute Name=\"" << svar << "\"" << endl;
	TString theclass = getAttribute( svar )->getDataTypeStr();
	if( theclass.EqualTo("D"))
	  {
	    sortie << "Double_t " << svar << "[" << GetName() << "_nPattern] = {" << endl;
	    for (Int_t ipat = 0; ipat < npat; ipat++)
	      {
		_datatree->GetEntry(telapp->GetEntry(ipat));
		sortie << " " << leaf->GetValue() << ( (ipat==npat-1) ? "" : ",") << endl;
	      }	
	  }
	else if( theclass.EqualTo("S"))
	  {
	    sortie << "std::string " << svar << "[" << GetName() << "_nPattern] = {"<< endl;
	    for (Int_t ipat = 0; ipat < npat; ipat++)
	      {
		_datatree->GetEntry(telapp->GetEntry(ipat));
		sortie << " \"" << *((string*)leaf->GetValuePointer()) << ( (ipat==npat-1) ? "\"" : "\",") << endl;
	      }	
	  }
	else if( theclass.EqualTo("V"))
	  {
	    sortie << "// vector inline initialisation only available for C++11... This is not yet tested! Comment it in case of problem" << endl;
	    sortie << "std::vector<Double_t> " << svar << "[" << GetName() << "_nPattern] = {"<< endl;
	    for (Int_t ipat = 0; ipat < npat; ipat++)
	      {
		_datatree->GetEntry(telapp->GetEntry(ipat));
		sortie << " std::vector<Double_t> {" << UEntry::dumpVectorStr(((vector<double>*)leaf->GetValuePointer()), 9) << ( (ipat==npat-1) ? "}" : "},") << endl;
	      }	
	  }

        sortie << "};" << endl;
        sortie << "// End of attribute " << svar << endl;
        // fin
        sortie << endl;
      }
    sortie << "// End of File " << filename << endl;

    // Close the output file
    sortie.close();
        
    //Clean
    if( gROOT->FindObject("elist1") ) gROOT->Delete("elist1");
    delete listOfLeaves;
    listOfLeaves = NULL;
    delete listOfAttributes;
    listOfAttributes = NULL;

    if (_blog)
      cout << "****** End of TDataServer::exportHeader" << endl;
  }

  void TDataServer::exportDataNeMo(const char *filename, const char *varexpinput,
				   const char *varexpoutput, const char *select)
  {

    if (_blog)
      {
        cout << endl << "****** TDataServer::exportDataNeMo [" << GetName()
	     << "] Title[" << GetTitle() << "]" << endl;
        cout << " ** filename[" << filename << "]" << endl;
        cout << " **    input[" << varexpinput << "]" << endl;
        cout << " **   output[" << varexpoutput << "]" << endl;
        cout << " **   select[" << select << "]" << endl;
      }

    TList *listOfAttributes=NULL;
    TEventList *telapp;
    Int_t npat;

    TString sListOfInputAttributes(varexpinput);
    TString sListOfOutputAttributes(varexpoutput);

    try
      {
        if (_dataSpecification == NULL)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form(
							  "TDataServer::exportDataNeMo ERROR : the TDataServer \"%s\" to export is empty",
							  GetName()));

        // Get the current list of attributes
        TString svars = sListOfInputAttributes;
        svars += TString(":");
        svars += sListOfOutputAttributes;
        svars.ReplaceAll(":", " ");
        svars += TString(" ");
        if (_blog)
	  cout << " ** svars[" << svars << "]" << endl;
        listOfAttributes = getListOfAttributes(svars.Data());

        // Treat the selection and the option
        TCut allCut; allCut += _ccut; allCut += TCut(select);
        // CFO::WARNING
        //Version v2_1_10 CFO, raison inconnue.
        npat = getNPatterns();

        if (npat == 0)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form(
							  "TDataServer::exportDataNeMo ERROR : the TDataServer \"%s\" to export have no pattern",
							  GetName()));

	
	if( gDirectory != gROOT ) gDirectory=gROOT;
	if( gROOT->FindObject("elist1") ) gROOT->Delete("elist1");
        _datatree->Draw(">>elist1",  ((strcmp(allCut.GetTitle(),"")==0)?"1>0":allCut.GetTitle()),"goff");
        telapp = (TEventList*) gROOT->Get("elist1");
        npat = telapp->GetN();

        try
	  {
            if (npat < 5)
	      throw URANIE::Exceptions::UWarningExceptions( __FILE__,
							    __LINE__,
							    Form(
								 "TDataServer::exportDataNeMo Warning :  Name: \"%s\"  the number of patterns is low ( %d < 5)",
								 GetName(), npat));
	  } catch (URANIE::Exceptions::UWarningExceptions& ue)
	  {
            ue.printMessage();
	  }

      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
	if ( listOfAttributes!=NULL )
	  delete listOfAttributes;
        throw ue;
        exit(-1);
      }

    // Ouverture du fichier de sortie
    ofstream sortie;
    sortie.open(filename, ios::out);
    sortie.precision(SSTR_PREC);
    sortie.setf(ios::scientific);

    // The number of input/output attributes
    Int_t nvarx = sListOfInputAttributes.CountChar(':') + 1;
    Int_t nvary = sListOfOutputAttributes.CountChar(':') + 1;

    // entete
    sortie << "#NombreExemples " << npat << endl;
    sortie << "#NombreEntrees " << nvarx << endl;
    sortie << "#NombreSorties " << nvary << endl << endl;

    TList *listOfLeaves = new TList();
    TIter next(listOfAttributes);
    while (TAttribute *att = (TAttribute *) next())
      {
        listOfLeaves->Add(_datatree->GetLeaf(att->GetName()));
      }

    if (_blog)
      {
        cout << endl << " **  TLeaf listing  ["
	     << listOfAttributes->GetEntries() << "]" << endl;
        listOfLeaves->Print();
      }

    for (Int_t ipat = 0; ipat < npat; ipat++)
      {
        _datatree->GetEntry(telapp->GetEntry(ipat));
        TIter nextLeaf(listOfLeaves);
        while (TLeaf *leaf = (TLeaf *) nextLeaf())
	  {
	    TString theclass = getAttribute( leaf->GetName() )->getDataTypeStr();
	    if(theclass.EqualTo("D")) sortie << leaf->GetValue() << " ";
	    if(theclass.EqualTo("S")) sortie << *((string*)leaf->GetValuePointer()) << " ";
	    if(theclass.EqualTo("V")) sortie << UEntry::dumpVectorStr((vector<double>*)(leaf->GetValuePointer())) << " ";
	  }
        sortie << endl;
      }
    // Close the output file
    sortie.close();

    //Clean
    if( gROOT->FindObject("elist1") ) gROOT->Delete("elist1");
    delete listOfLeaves;
    listOfLeaves = NULL;
    delete listOfAttributes;
    listOfAttributes = NULL;

    if (_blog)
      cout << "****** End of TDataServer::exportDataNeMo" << endl;
  }

  void TDataServer::exportDataContext(const char *filename, const char *varexp,
				      const char *select)
  {

    if (_blog)
      {
        cout << endl << "****** TDataServer::exportDataContext [" << GetName()
	     << "] Title[" << GetTitle() << "]" << endl;
        cout << " ** filename[" << filename << "]" << endl;
        cout << " **   varexp[" << varexp << "]" << endl;
        cout << " **   select[" << select << "]" << endl;
      }

    TList *listOfAttributes=NULL;
    TEventList *telapp;
    Int_t npat;

    try
      {
        if (_dataSpecification == NULL)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form(
							  "TDataServer::exportDataContext ERROR\n the TDataServer \"%s\" to export is empty",
							  GetName()));

        // Get the current list of attributes
        listOfAttributes = getListOfAttributes(varexp);
        listOfAttributes->Add(getAttribute(getIteratorName()));

        // Treat the selection and the option
        TCut allCut; allCut += _ccut; allCut += TCut(select);

        // Delete the TEventList elist1 if exists
	if( gDirectory != gROOT ) gDirectory=gROOT;
        if( gROOT->FindObject("elist1") ) gROOT->Delete("elist1");
        _datatree->Draw(">>elist1",  ((strcmp(allCut.GetTitle(),"")==0)?"1>0":allCut.GetTitle()),"goff");
        telapp = (TEventList*) gROOT->Get("elist1");

        if (telapp == NULL)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form(
							  "TDataServer::exportDataContext ERROR\n the TDataServer \"%s\" with selection[%s] to export have no pattern",
							  GetName(), allCut.GetTitle()));

        npat = telapp->GetN();
      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
	if(listOfAttributes!=NULL)
	  delete listOfAttributes;
        throw ue;
        exit(-1);
      }

    // Open the output file
    TString stexfile(filename);
    //out.Remove(0,8);
    ofstream sortie;
    sortie.open(filename, ios::out);

    // Header
    TDatime theTime;
    sortie
      << Form("%%%% File \"%s\" generated by ROOT v%s", stexfile.Data(),
	      gROOT->GetVersion()) << endl;
    sortie << Form("%%%% DateTime %s", theTime.AsString()) << endl;
    sortie << "%% DataServer : Name=\"" << GetName() << "\" Title=\""
	   << GetTitle() << "\"";
    if (_ccut != 0)
      sortie << " Global Select=\"" << _ccut << "\"";
    TString sselect = TString(select);
    if (sselect.Length() > 0)
      sortie << " Local Select=\"" << sselect << "\"";
    sortie << endl;

    // Line of attributes names
    Int_t nleaves = listOfAttributes->GetEntries();
    TIter next(listOfAttributes);
    Int_t ivar = 0;

    TList *listOfLeaves = new TList();

    sortie << "\\starttable[|";
    for (Int_t i = 0; i < nleaves - 1; i++)
      sortie << "c|";
    sortie << "l|]" << endl << "\\HL" << endl;

    sortie << "\\VL ";
    while (TAttribute *att = (TAttribute *) next())
      {
        TString svarname = TString(att->GetName());
        svarname.ReplaceAll("_", "\\type{_}");
        if (att->getOrigin() == TAttribute::kIterator)
	  sortie << "Iter";
        else
	  sortie << svarname;
        if (ivar++ < nleaves - 1)
	  sortie << " \\VL ";
        listOfLeaves->Add(_datatree->GetLeaf(att->GetName()));
      }
    sortie << " \\VL \\SR" << endl << "\\HL" << endl;

    if (_blog)
      {
        cout << endl << " **  TLeaf listing  [" << nleaves << "]" << endl;
        listOfLeaves->Print();
      }

    Double_t dval;
    TString sval;
    //Write the data
    for (Int_t ipat = 0; ipat < npat; ipat++)
      {
        _datatree->GetEntry(telapp->GetEntry(ipat));

        TIter nextLeaf(listOfLeaves);
        while (TLeaf *leaf = (TLeaf *) nextLeaf())
	  {
            TAttribute *att = getAttribute(leaf->GetName());
            if (_blog)
	      att->printLog();
            TString sformattosubstitute = TString(att->getFormatToSubstitute());
            dval = leaf->GetValue();
            if (sformattosubstitute.Length() > 0)
	      {
                if (sformattosubstitute.EndsWith("d"))
		  {
                    Int_t nval = (Int_t)(TMath::Ceil(dval));
                    sval = Form(sformattosubstitute, nval);
		  }
                else
		  sval = Form(sformattosubstitute, dval);
                sortie << "\\VL  " << sval << " ";
	      }
            else
	      {
                sortie << "\\VL  " << leaf->GetValue() << " ";
	      }

            //          sortie << "\\VL  " << leaf->GetValue() << " ";
            //          sortie << "\\VL  " << sval << " ";
	  }
        if (ipat == 0)
	  {
            sortie << "\\VL \\FR";
	  }
        else
	  {
            if (ipat == npat - 1)
	      {
                sortie << "\\VL \\LR";
	      }
            else
	      {
                sortie << "\\VL \\MR";
	      }
	  }
        sortie << endl;
      }
    sortie << "\\HL" << endl << "\\stoptable" << endl;
    ;
    // Close the output file
    sortie.close();

    //Clean
    if( gROOT->FindObject("elist1") ) gROOT->Delete("elist1");
    delete listOfLeaves;
    listOfLeaves = NULL;
    delete listOfAttributes;
    listOfAttributes = NULL;
    if (_blog)
      cout << "****** End of TDataServer::exportDataContext" << endl;
  }


#ifdef WITH_JSONCPP
void TDataServer::exportDataJSon(const char *filename, const char *varexp, const char *select, Option_t *option)
{
  if (1) // _blog)
    {
        cout << endl << "****** TDataServer::exportDataJSon [" << GetName()
                << "] Title[" << GetTitle() << "]" << endl;
        cout << " **     Size[" << getNPatterns() << "]" << endl;
        cout << " ** filename[" << filename << "]" << endl;
        cout << " **   varexp[" << varexp << "]" << endl;
        cout << " **   select[" << select << "]" << endl;
    }

    
    TList *listOfAttributes=NULL;
    TEventList *telapp;
    Int_t npat;
    TPatternsEventList * pelmix = NULL;
    bool pelmix_is_new=false;
    
    Int_t nPrecision = 9;
    TString sOption = option;
    sOption.ToLower();
    if(sOption.BeginsWith("precision="))
    {
        Int_t nIndE = sOption.First('=');
        Int_t nLengthO = sOption.Length();
        TString sNum = sOption(nIndE +1, nLengthO-nIndE);
        if(sNum.IsDigit())
            nPrecision = sNum.Atoi();
    }

    try
    {
        if (_datatree == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TDataServer::exportData[%s] in file[%s]\n The TTree is NULL",
                            GetName(), filename));
        if (_dataSpecification == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TDataServer::exportData ERROR : the TDataServer \"%s\" to export is empty",
                            GetName()));

        // Get the current list of attributes
        listOfAttributes = getListOfAttributes(varexp);
        if (_blog)
        {
            cout << " ** Iterator Name[" << getIteratorName() << "]" << endl;
            getAttribute(getIteratorName())->printLog();
        }
        listOfAttributes->Add(getAttribute(getIteratorName()));

        // Treat the selection
        TCut allCut; allCut += _ccut; allCut += TCut(select);
        TString scut = TString(_ccut);
        if (_blog)
        {
            cout << " ** _ccut[" << _ccut << "]" << endl;
            cout << " ** select[" << select << "]" << endl;
            cout << " ** allCut[" << allCut << "]" << endl;
            cout << " ** scut[" << scut << "]" << endl;
        }
        TString spel = Form("uranie_pel_%s", GetName());
        pelmix = dynamic_cast<TPatternsEventList *>(_datatree->GetEventList());
        if (!pelmix)
        {
            if (_blog)
                cout << "no eventlist" << endl;
            //    pelmix = new TEventList(stel.Data(), "all patterns", getNPatterns());
            pelmix = new TPatternsEventList(spel.Data(), "all patterns", getNPatterns());
	    pelmix_is_new=true;
        }
        if (_blog)
            pelmix->Print("all");

	// Clean the TEventList
	if( gDirectory != gROOT ) gDirectory=gROOT;
	if( gROOT->FindObject("elistUranie") ) gROOT->Delete("elistUranie");
	_datatree->Draw(">>elistUranie", ((strcmp(allCut.GetTitle(),"")==0) ? "1>0": allCut.GetTitle()),"goff");
	telapp = (TEventList*) gROOT->Get("elistUranie");
	npat = telapp->GetN();
	if (_blog)
	{
	    telapp->Print("all");
	    cout << " ** elistUranie npat[" << npat << "]" << endl;
	}

        if (npat == 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TDataServer::exportData ERROR : the TDataServer \"%s\" to export have no pattern",
                            GetName()));


        // Header
        TDatime theTime;
	Json::Value tdsFromStratch;
	Json::Value tdsMetaData;
	Json::Value tdsItems;
	tdsMetaData["_comment"] = "CurrentComment";
	tdsMetaData["date"] = theTime.AsString();
	tdsMetaData["table_name"] = GetName();
	tdsMetaData["table_description"] = GetTitle();
	

        // Line of attributes names
        Int_t nleaves = listOfAttributes->GetEntries();
        TIter next(listOfAttributes);
        Int_t ivar = 0;

        TList *listOfLeaves = new TList();

        while (TAttribute *att = (TAttribute *) next())
	  {
	    tdsMetaData["short_names"][ivar] = att->GetName();
	    tdsMetaData["types"][ivar] = att->getDataTypeStr().Data();
	    tdsMetaData["units"][ivar] = att->getUnit().Data();
            if (_blog)
	      cout << " ** ivar[" << ivar << "] name[" << att->GetName()
		   << "]" << endl;
            if (_datatree->GetLeaf(att->GetName()))
	      listOfLeaves->Add(_datatree->GetLeaf(att->GetName()));
            else
	      {
                delete listOfLeaves;
                listOfLeaves = NULL;
                delete listOfAttributes;
                listOfAttributes = NULL;
		if(pelmix_is_new==true)
		{
		    delete pelmix;
		    pelmix=NULL;
		}
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							   Form(
								"TDataServer::exportData ERROR : the Leaf \"%s\" does not exist in the TTree object",
								att->GetName()));
	      }
	    ivar++;
	  }

        if (_blog)
        {
            cout << endl << " **  TLeaf listing  ["
                    << listOfLeaves->GetEntries() << "]" << endl;
            listOfLeaves->Print();
        }

  	tdsFromStratch["_metadata"] = tdsMetaData;
	vector<double> buffer;
	//Write the data
        for (Int_t ipat = 0; ipat < npat; ipat++)
        {
            _datatree->GetEntry(telapp->GetEntry(ipat));
            if (_blog)
            {
                _datatree->Show();
                cout << endl << " **  Entry  [" << telapp->GetEntry(ipat) << "/"
                        << npat << "]" << endl;
            }

            for (Int_t i = 0; i < nleaves; i++)
            {
                TLeaf *leaf = (TLeaf *) (listOfLeaves->At(i));
                if (_blog)
                    cout << " **** Attribute[" << leaf->GetName() << "] value["
                            << leaf->GetValue() << "]" << endl;
		if( getAttribute( leaf->GetName() )->getDataTypeStr() == "V" )
		{
		    buffer = *(vector<double>*)(leaf->GetValuePointer());
		    for(unsigned int ival=0; ival<buffer.size(); ival++)
			tdsItems[ipat][tdsMetaData["short_names"][i].asString()].append( buffer.at(ival) );
		}
		else if( getAttribute( leaf->GetName() )->getDataTypeStr() == "S" )
		    tdsItems[ipat][tdsMetaData["short_names"][i].asString()] = *(string*)(leaf->GetValuePointer());
		else if( getAttribute( leaf->GetName() )->getDataTypeStr() == "D" )
		    tdsItems[ipat][tdsMetaData["short_names"][i].asString()] = leaf->GetValue();
            }
        }
        // Close the output file
	tdsFromStratch["items"] = tdsItems;
	// Write the JSON object
	std::ofstream file_id;
	file_id.open(filename, ios::out);
        file_id.precision(nPrecision);
        file_id.setf(ios::scientific);
	Json::StyledStreamWriter writer("  ");
	writer.write(file_id, tdsFromStratch);
    //std::cout << tdsFromStratch << std::endl;
	file_id.close();

        //Clean
	if(pelmix_is_new==true)
	    delete pelmix;
        delete listOfLeaves;
        delete listOfAttributes;
        // Clean the TEventList
        gROOT->Delete("elistUranie;*");
        if (_blog)
            cout << "****** End of TDataServer::exportData" << endl;
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
     	if( gROOT->FindObject("elistUranie") ) gROOT->Delete("elistUranie");
	if(pelmix_is_new==true)
	    delete pelmix;
	if (listOfAttributes != NULL)
	    delete listOfAttributes;
        throw ue;
    }

}
#endif


void TDataServer::saveTuple(Option_t* option)
{
    TFile * rootfile = TFile::Open(getArchiveFileName(), option);
    _datatree->SetDirectory(0);
    _datatree->Write();
    rootfile->Close();
    if( !TThread::IsInitialized())
        _datatree->SetDirectory(gROOT);
  }


////////////////////////////////////////////////////////
////////////////////////////////////////////////////////
//  Statistical Method
////////////////////////////////////////////////////////
////////////////////////////////////////////////////////

////////////////////////////////////////////////////////
//  computeCAH : Statistical Method
////////////////////////////////////////////////////////
  TMatrixD TDataServer::computeCAH(const char *varexp, ECAHMethod CAHmethod,
				   const char *selection, Option_t *option)
  {
    if (_blog)
      {
        cout << endl << "****** TDataServer::CAH with tds[" << GetName()
	     << "] npat[" << getNPatterns() << "]" << endl;
        cout << " **  varexp[" << varexp << "]" << endl;
        switch (CAHmethod)
	  {
	  case kWard:
            cout << " ** Method: kWard" << endl;
            break;
	  case kWardExact:
            cout << " ** Method: kWardExact" << endl;
            break;
	  case kCentroid:
            cout << " ** Method: kCentroid" << endl;
            break;
	  case kAverage:
            cout << " ** Method: kAverage not yet implemented" << endl;
            break;
	  case kSingle:
            cout << " ** Method: kSingle not yet implemented" << endl;
            break;
	  case kComplete:
            cout << " ** Method: kComplete not yet implemented" << endl;
            break;
	  default:
            cout << "  ** Method unknown [" << CAHmethod << "]" << endl;
	  }
        cout << " **  select[" << selection << "]" << endl;
        cout << " **  option[" << option << "]" << endl;
      }

    TList *listOfAttributes = NULL;
    try
      {

        switch (CAHmethod)
	  {
	  case kSingle:
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 "TDataServer::CAH method\n The Algorithm TDataServer::kSingle not yet implemented\n Return matrix of size 0-0");
            break;
	  case kComplete:
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 "TDataServer::CAH method\n The Algorithm TDataServer::kComplete not yet implemented\n Return matrix of size 0-0");
            break;
	  default:
	      break;
	  }

        TString svar = TString(varexp);
        svar.ReplaceAll(" ", "");
        TString sweight = "__URANIE__Weight__";

        TLeaf *leafWeight = NULL;
        Int_t ncpt = svar.CountChar(',');
        if (ncpt > 0)
	  {
            svar.ReplaceAll(",", " ");
            istringstream inStream((char *) svar.Data());
            inStream >> svar >> sweight;
            ncpt = sweight.CountChar(':');

            if (ncpt != 0)
	      throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							 Form(
							      "TDataServer::CAH method\n The mass attribute contains more than one attribute \"%s\"",
							      sweight.Data()));
            if (isAttribute(sweight) == kFALSE)
	      throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
							 Form(
							      "TDataServer::CAH method\n The TDS \"%s\" does not contain the Weight attribute \"%s\"",
							      GetName(), sweight.Data()));
            leafWeight = _datatree->GetLeaf(sweight.Data());
	  }
        if (_blog)
	  {
            cout << " **  svar attributes [" << svar << "]" << endl;
            cout << " ** Weight attribute [" << sweight << "]" << endl;
	  }
        Int_t npat = getNPatterns();
        Int_t nx = svar.CountChar(':') + 1;
        if (_blog)
	  cout << " ** npat(" << npat << ") nX(" << nx << ")" << endl;

        // Compute the statistic to store normalized data (kCR). Not use the the normalize method
        computeStatistic(varexp, selection);

        //////////////////////////////////////////////////////////////
        // matrix to store all the nodes (original data and nodes of the cluster)
        // 0->nx-1 : First columns are the original attributes centre-reduit
        // nx      : the weight
        // nx+1    : the score of the node (-1 for an original patterns and > 0 for a node)
        // nx+2    : the cumuled score (0 for an original patterns and \sum{j=0){j=i}Score_{j} > 0 for a node)
        // nx+3    : the index of "ainé" pattern
        // nx+4    : the index of "benjamin" pattern
        // The number of rows is 2 * npat -1 : npat first row for the data, and npat-1 clusters
        TMatrixD matCAH(2 * npat - 1, nx + 5);
        matCAH.Zero();

	try
	  {
	    listOfAttributes = getListOfAttributes(svar.Data());
	  }catch(URANIE::Exceptions::UErrorExceptions& ue)
	  {
	    if (listOfAttributes!=NULL)
	      delete listOfAttributes;
	    throw ue;
	  }
        TIter nextatt(listOfAttributes);

        Int_t iatt = 0;
        // Default value for the weight
        Double_t dWeight = 1.0;
        // Fill the normalized data and weight attributes
        cout.setf(ios::fixed);
        cout.precision(SSTR_PREC);
        for (Int_t ipat = 0; ipat < npat; ipat++)
	  {
            _datatree->GetEntry(ipat);
            // data
            iatt = 0;
            while (TAttribute *att = (TAttribute *) nextatt())
	      {
                matCAH(ipat, iatt) =
		  (_datatree->GetLeaf(att->GetName())->GetValue()
		   - att->getMean()) / att->getStd();
                iatt++;
	      }
            nextatt.Reset();

            // The value of the weight is defined
            if (leafWeight)
	      dWeight = leafWeight->GetValue();
            matCAH(ipat, nx) = dWeight;
            matCAH(ipat, nx + 1) = -1.0;
            matCAH(ipat, nx + 2) = ipat;
	  }

        Int_t ipatcrt = npat;
        Int_t nIMinIndex, nJMinIndex, nIIndex;
        TMatrixDSym matD(npat);
        matD.Zero();
        Double_t dlambda;
        Double_t dmin=1e15;
        Int_t imin = -1, jmin = -1;

        // loop to clustering of size=npat-1
        for (Int_t nloop = 0; nloop < npat - 1; nloop++)
	  {
            if (_blog)
	      cout << endl << endl << " ** Begin Of Loop [" << nloop << "]"
		   << endl;
            imin = -1;
            jmin = -1;
            for (Int_t i = 0; i < npat; i++)
	      {
                for (Int_t j = i + 1; j < npat; j++)
		  {

                    // Compute the distance in the first loop
                    if (nloop == 0)
		      {
                        dlambda = 0.0;
                        for (Int_t k = 0; k < nx; k++)
			  {
                            dlambda += (matCAH(i, k) - matCAH(j, k))
			      * (matCAH(i, k) - matCAH(j, k));
			  }
                        switch (CAHmethod)
			  {
			  case kWard:
			  case kWardExact:
                            matD(i, j) = matD(j, i) = (dlambda * matCAH(i, nx)
						       * matCAH(j, nx))
			      / (matCAH(i, nx) + matCAH(j, nx));
                            break;
			  case kAverage:
                            matD(i, j) = matD(j, i) = dlambda;
                            break;
			  case kCentroid:
                            matD(i, j) = matD(j, i) = dlambda;
                            break;
			  default:
			      break;
			  }
		      }

                    // Looking for the min value of the matrix of distance with the index values (imin, jmin)
                    // only if the two patterns are active (i.e. matCAH(i, nx+2) != -1)
                    if ((matCAH(i, nx + 2) > -0.5)
			&& (matCAH(j, nx + 2) > -0.5))
		      {
                        if (imin < 0)
			  {
                            dmin = matD(i, j);
                            imin = i;
                            jmin = j;
                            if (_blog)
			      cout << "  ** BEGIN OF min dmin(" << dmin
				   << ") (" << imin << "," << jmin << ")"
				   << endl;
			  }
                        else
			  {
                            if (matD(i, j) < dmin)
			      {
                                dmin = matD(i, j);
                                imin = i;
                                jmin = j;
                                if (_blog)
				  cout << "  ** new MIN dmin(" << dmin
				       << ") (" << imin << "," << jmin
				       << ")" << endl;
			      }
			  }
		      }
		  }
	      }
            if (_blog)
	      cout << " ** END OF min dmin(" << dmin << ") (" << imin << ","
		   << jmin << ")" << endl;

            // update the line ipatcrt of the CAH matrix
            nIMinIndex = (Int_t)(matCAH(imin, nx + 2));
            nJMinIndex = (Int_t)(matCAH(jmin, nx + 2));

            matCAH(ipatcrt, nx + 1) = TMath::Sqrt(dmin);
            if (nloop == 0)
	      matCAH(ipatcrt, nx + 2) = TMath::Sqrt(dmin);
            else
	      matCAH(ipatcrt, nx + 2) = matCAH(ipatcrt - 1, nx + 2)
		+ matCAH(ipatcrt, nx + 1);
            matCAH(ipatcrt, nx + 3) = imin;
            matCAH(ipatcrt, nx + 4) = jmin;

            // Set the weight of the barycentre (i.e. second equation from book "Analyse de Donn�es" de Volle page 289)
            matCAH(ipatcrt, nx) = matCAH(nIMinIndex, nx)
	      + matCAH(nJMinIndex, nx);
            // Set the barycentre of the two partitions imin and jmin (i.e. first equation from book "Analyse de Donn�es" de Volle page 289)
            for (Int_t k = 0; k < nx; k++)
	      matCAH(ipatcrt, k) = (matCAH(nIMinIndex, nx)
				    * matCAH(nIMinIndex, k)
				    + matCAH(nJMinIndex, nx) * matCAH(nJMinIndex, k))
		/ matCAH(ipatcrt, nx);

            Double_t ddmax = matD.Max();
            // Update matrix of distance
            for (Int_t i = 0; i < npat; i++)
	      {
                nIIndex = (Int_t)(matCAH(i, nx + 2));
                if (nIIndex > -0.5)
		  {
                    if ((i != imin) || (i != jmin))
		      {
                        switch (CAHmethod)
			  {
			  case kWard:
                            ////////////////////////////////
                            //   Ward
                            // Update the matrix of distance for the kWard distance (i.e. equation from book "Analyse de Donn�es" de Volle page 288)
                            ////////////////////////////////
                            matD(imin, i) = matD(i, imin) =
			      ((matCAH(nIMinIndex, nx)
				+ matCAH(nIIndex, nx))
			       * matD(imin, i)
			       + (matCAH(nJMinIndex, nx)
				  + matCAH(nIIndex, nx))
			       * matD(jmin, i)
			       - matCAH(nIIndex, nx)
			       * matD(imin, jmin))
			      / (matCAH(ipatcrt, nx)
				 + matCAH(nIIndex, nx));
                            break;
			  case kWardExact:
                            ////////////////////////////////
                            //   Ward Exact
                            // Recompute the matrix of distance from the coordinates of the new cluster and each points
                            ////////////////////////////////
                            dlambda = 0.0;
                            for (Int_t k = 0; k < nx; k++)
			      {
                                dlambda += (matCAH(ipatcrt, k)
					    - matCAH(nIIndex, k))
				  * (matCAH(ipatcrt, k)
				     - matCAH(nIIndex, k));
			      }
                            dlambda *=
			      matCAH(ipatcrt, nx) * matCAH(nIIndex, nx)
			      / (matCAH(ipatcrt, nx)
				 + matCAH(nIIndex, nx));
                            matD(imin, i) = matD(i, imin) = dlambda;
                            break;
                            ////////////////////////////////
                            //  Average
                            ////////////////////////////////
			  case kAverage:
                            matD(imin, i) = matD(i, imin) = (matCAH(nIMinIndex,
								    nx) * matD(imin, i)
							     + matCAH(nJMinIndex, nx) * matD(jmin, i))
			      / matCAH(ipatcrt, nx);
                            break;
                            ////////////////////////////////
                            //  Centroid
                            ////////////////////////////////
			  case kCentroid:
                            matD(imin, i) = matD(i, imin) = (matCAH(nIMinIndex,
								    nx) * matD(imin, i)
							     + matCAH(nJMinIndex, nx) * matD(jmin, i))
			      / matCAH(ipatcrt, nx)
			      - matCAH(nIMinIndex, nx)
			      * matCAH(nJMinIndex, nx)
			      * matD(imin, jmin)
			      / (matCAH(ipatcrt, nx)
				 * matCAH(ipatcrt, nx));
                            break;
			  default:
			      break;
			  }
		      }
		  }
	      }
            for (Int_t i = 0; i < npat; i++)
	      {
                matD(jmin, i) = ddmax;
                matD(i, jmin) = ddmax;
	      }
            matCAH(imin, nx + 2) = ipatcrt;
            matCAH(jmin, nx + 2) = -1;

            //      matCAH(imin, nx) = matCAH(ipatcrt, nx);
            //      matCAH(jmin, nx) = matCAH(ipatcrt, nx);

            Double_t ddmin = matD.Min();

            if (_blog)
	      {
                matCAH.Print();
                cout << " ** End Of Loop [" << nloop << "] with ddmin(" << ddmin
		     << ") ddmax(" << ddmax << ")" << endl;
	      }
            ipatcrt++;
	  }

	delete listOfAttributes;
        return matCAH;

      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
        ue.printMessage();
	if(listOfAttributes!=NULL)
	  delete listOfAttributes;
        throw ue;
      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
        cout << " ***************************** JUST A  WARNING" << endl;
        ue.printMessage();
        TMatrixD matCAH(0, 0);
        matCAH.Zero();
        return matCAH;
      }

  }

  //Small structure to replace the ranking procedure
  struct CompareIndicesByAnotherVectorValues
  {
    std::vector<double>* _values;
  public:
    CompareIndicesByAnotherVectorValues(std::vector<double>* values) : _values(values) {}
    bool operator() (const int& a, const int& b) const { return _values->at(a) < _values->at(b); }
  };
  
  ////////////////////////////////////////////////////////
  //  computeRank : Statistical Method
  ////////////////////////////////////////////////////////
  void TDataServer::computeRank(const char *varexp, Option_t *option)
  {
    //Mod by JBB 2016-2-24 to take the vectors into account (and get rid of strings)
    if (_blog)
      {
        cout << endl << "****** TDataServer::computeRank" << endl;
        cout << " ** varexp[" << varexp << "]" << endl;
        cout << " **  option[" << option << "]" << endl;
      }

    // Get the current list of attributes
    TList *listOfAttributes = getListOfAttributes(varexp);

    Int_t nvars = listOfAttributes->GetEntries();
    if (_blog)
      {
        cout << endl << " **  Attributes listing  [" << nvars << "] npat["
	     << getNPatterns() << "]" << endl;
        listOfAttributes->Print();
      }


    //memory buffer for branches
    vector<double> *v_dnrank=NULL;
    Double_t* tabValue_size=NULL;
    //Number of pattern for the considered tree
    Int_t npat = getNPatterns();
    TIter next(listOfAttributes);
    while (TAttribute *att = (TAttribute *) next())
      {
	unsigned int VectSize=1;//Default for simple double
        if (isAttribute(Form("Rk_%s", att->GetName())))
	  {
            if (_blog)
	      cout << " *** Attribute [" << att->GetName()
		   << "] already exists" << endl;
            continue;
	  }

        TLeaf *leaf = _datatree->GetLeaf(att->GetName());
	TString theclass = att->getDataTypeStr();
	try{
	  if(theclass.EqualTo("S"))
	    throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 Form(
							      "TDataServer::ComputeRank is only able to deal with double. Branch %s has type %s, so nothing will be done!",
							      att->GetName(), theclass.Data()));
	  //Test the vector case
	  else if(theclass.EqualTo("V"))
	    {
	      string sizebranch=SIZENAME+"__"+string(att->GetName());
	      if(!isAttribute( sizebranch.c_str() ))
		throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							     Form("TDataServer::ComputeRank: vectors are usable when size is an attribute. %s not found so nothing will be done for it!",
								  sizebranch.c_str()));
	      computeStatistic( sizebranch.c_str() );
	      if( getAttribute( sizebranch.c_str() )->getMinimum() != getAttribute( sizebranch.c_str() )->getMaximum() )
		throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							     Form("TDataServer::ComputeRank: vectors are usable when size is not changing. %s Std is not 0 (%g) so nothing will be done for it!",
								  sizebranch.c_str(), getAttribute( sizebranch.c_str() )->getStd() ));
	      VectSize = TMath::Nint(getAttribute( sizebranch.c_str() )->getMean());
	    }

        } catch (URANIE::Exceptions::UWarningExceptions& ue)
	  {
            ue.printMessage();
	    continue;
	  }	
      
        if (_blog)
	  cout << endl << " ** Sort att[" << att->GetName() << "]" << endl;

	//vector to store the values, original index and valuues' indexes
	vector<vector<double> > v_dval; v_dval.resize(VectSize);
	vector<vector<Long64_t> > v_index; v_index.resize(VectSize);
	vector<vector<double> > v_valindex; v_valindex.resize(VectSize);

	vector<double> *v=NULL;
	//Reading the branch
        for (Int_t ipat = 0; ipat < npat; ipat++)
	  {
            _datatree->GetEntry(ipat);

	    //Get the data (either a ector or one double 	
	    if( theclass.EqualTo("D") )
	      v = new vector<double>(1,leaf->GetValue());
	    else if( theclass.EqualTo("V") )
	      v = ((vector<double>*)leaf->GetValuePointer());

	    // Fill dval elements
	    for(unsigned int iel=0; iel<v->size(); iel++)
	       {
		 v_dval[iel].push_back( v->at(iel) );
		 v_valindex[iel].push_back( 0.0 );
		 v_index[iel].push_back( ipat ); //Was 0.0 in the old procedure
	       }
	    
	    if( theclass.EqualTo("D") )
	      delete v;
	  }
	
	//Sort the values
	for(unsigned int iel=0; iel<VectSize; iel++)
	  {
	    //TMath::Sort(Long64_t(npat), &v_dval[iel].at(0), &v_index[iel].at(0), kFALSE);
	    stable_sort(v_index[iel].begin(), v_index[iel].end(), CompareIndicesByAnotherVectorValues(&v_dval[iel]) );
	  }

	//Pointers for branches (_size is in case of vectors) and their memory buffer
	TBranch *brank=0, *brank_size=0;
	v_dnrank = new vector<double>(VectSize, 0);
	tabValue_size = new Double_t[1];
	if( theclass.EqualTo("D") )
	  brank = _datatree->Branch(Form("Rk_%s", att->GetName()), &v_dnrank->at(0), Form("Rk_%s/D", att->GetName()));
	else if( theclass.EqualTo("V") )
	  {
	    //Create both vector and size branch
	    brank = _datatree->Branch(Form("Rk_%s", att->GetName()), &v_dnrank, "vector<double>");
	    brank_size = _datatree->Branch((SIZENAME+"__Rk_"+string(att->GetName())).c_str(), tabValue_size, Form("%s/D", (SIZENAME+"__Rk_"+string(att->GetName())).c_str()) );
	  }

	//Fill the final indexes values
        for (Int_t ipat = 0; ipat < npat; ipat++)
	  for(unsigned int iel=0; iel<VectSize; iel++)
	    v_valindex[iel].at(v_index[iel].at(ipat)) = (Double_t)(ipat + 1);

	//Add them in the tree
        for (Int_t ipat = 0; ipat < npat; ipat++)
	  {
            _datatree->GetEntry(ipat);
	    for(unsigned int iel=0; iel<VectSize; iel++)
	      v_dnrank->at(iel) = v_valindex[iel].at(ipat);
            brank->Fill();
	    if( theclass.EqualTo("V") )
	      {
		tabValue_size[0] = VectSize;
		brank_size->Fill();
	      }
	  }
        brank->ResetAddress();
        addAttribute(new TAttribute(Form("Rk_%s", att->GetName())));
	getAttribute(Form("Rk_%s", att->GetName()))->delShare();
	if( theclass.EqualTo("V") )
	  {
	    brank_size->ResetAddress();
	    getAttribute( Form("Rk_%s", att->GetName()) )->setDataType("V");
	    addAttribute(new TAttribute( (SIZENAME+"__Rk_"+string(att->GetName())).c_str() ));
	    getAttribute( (SIZENAME+"__Rk_"+string(att->GetName())).c_str() )->setOrigin(TAttribute::kInternal);
	    getAttribute( (SIZENAME+"__Rk_"+string(att->GetName())).c_str() )->delShare();
	  }
	delete v_dnrank; v_dnrank=NULL;
	delete[] tabValue_size; tabValue_size=NULL;
      }//end of while {attribute}

    //Clean
    delete listOfAttributes;
    listOfAttributes=NULL;
    if (_blog)
      cout << "****** End Of TDataServer::computeRank" << endl;
  }

  ////////////////////////////////////////////////////////
  //  computeCorrelationMatrix : Statistical Method
  ////////////////////////////////////////////////////////
  TMatrixD TDataServer::computeCorrelationMatrix(const char *varexp,
						 const char *selection, Option_t *option)
  {
    //Mod by JBB 2016-2-25 to take the vectors into account (and get rid of strings)
    if (_blog)
      {
        cout << endl << "****** TDataServer::computeCorrelationMatrix" << endl;
        cout << " ** varexp[" << varexp << "]" << endl;
        cout << " ** select[" << selection << "]" << endl;
        cout << " ** option[" << option << "]" << endl;
      }

    Bool_t btype = kFALSE;

    TString soption = TString(option);
    soption.ToLower();


    bool internalCall=false;
    if( soption.Contains("internalcall"))
        internalCall=true;

    map<string, vector<int> > mAttributeElements;
    string reducedlist=((internalCall) ? string(varexp) : TDataServer::checkAttributes(this,varexp,mAttributeElements));

    // for(map<string, vector<int> >::const_iterator it = mAttributeElements.begin(); it != mAttributeElements.end(); ++it)
    // {
    //     XYZ << it->first << " [";
    //     for(unsigned int iv=0; iv < it->second.size(); iv++)
    //         cout << it->second.at(iv)<<((iv==( it->second.size()-1))? "]" : ", ") ;
    //     cout << endl;
    // }
    
    // Get the current list of attributes
    TList *listOfAttributes = getListOfAttributes(reducedlist.c_str());

    TAttribute *att;
    TString svarx = "";
    TIter nextinit(listOfAttributes);
    while ((att = (TAttribute*) nextinit()))
      svarx += Form(":%s", att->GetName());

    //Case of rank option
    if (soption.Contains("rank"))
      {
        computeRank(varexp);
        svarx.ReplaceAll(":", ":Rk_");
        btype = kTRUE;
      }
    svarx.Replace(0, 1, "");

    // Uptodate list of attributes
    delete listOfAttributes;
    listOfAttributes = getListOfAttributes(svarx.Data());

    // Treat the selection and the option
    TCut allCut; allCut += _ccut; allCut += TCut(selection);
    if( gDirectory != gROOT ) gDirectory=gROOT;
    if( gROOT->FindObject("elist1") ) gROOT->Delete("elist1");
    _datatree->Draw(">>elist1",  ((strcmp(allCut.GetTitle(),"")==0)?"1>0":allCut.GetTitle()), "goff");
    TEventList *telapp = (TEventList*) gROOT->Get("elist1");
    Int_t npat = telapp->GetN();
    if (_blog)
      cout << "   npat[" << npat << "]" << endl;

    try
      {
	if (npat < 3)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form(
							  "TDataServer::computeCorrelationMatrix ERROR : the TDataServer \"%s\" to compute the correlation matrix have low pattern (%d < 3)",
							  GetName(), npat));
      }catch(URANIE::Exceptions::UErrorExceptions &ue)
      {
	ue.printMessage();
	delete listOfAttributes;
	throw ue;
      }
    
    TList *listOfLeaves = new TList();
    Int_t icrt = 0; int nvars = 0;
    TIter next(listOfAttributes);
    map<const char*,unsigned int> VectSize;
    //Check the content of the request in terms of string and vectors
    //increment the nvars int to get the matrix size
    while ( (att = (TAttribute *) next()))
      {
	  TLeaf *leaf = (TLeaf*)_datatree->GetLeaf(att->GetName());
	
	try
	  {
	    TString theclass = att->getDataTypeStr();
	    if(theclass.EqualTo("S"))
	      throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							   Form("TDataServer::ComputeCorrelationMatrix is only able to deal with double and vectors. Branch %s has type %s, so nothing will be done for it!",
								att->GetName(), theclass.Data()));
	    else if(theclass.EqualTo("V"))
	      {
		string sizebranch=SIZENAME+"__"+string(att->GetName());
		if(!isAttribute( sizebranch.c_str() ))
		  throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							       Form("TDataServer::ComputeCorrelationMatrix: vectors are usable when size is an attribute. %s not found so nothing will be done for it!",
								    sizebranch.c_str()));
		computeStatistic( sizebranch.c_str(), selection, option );
		if( getAttribute( sizebranch.c_str() )->getMinimum() != getAttribute( sizebranch.c_str() )->getMaximum() )
		  throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							       Form("TDataServer::ComputeCorrelationMatrix: vectors are usable when size is not changing. %s Std is not 0 (%g) so nothing will be done for it!",
								    sizebranch.c_str(), getAttribute( sizebranch.c_str() )->getStd() ));

                // XYZ<<string(att->GetName())<<"  ::  "<<mAttributeElements.count(string(att->GetName()))<<endl;
                if(mAttributeElements.count( string(att->GetName()) ) != 0)
                {
                    for(unsigned int iiel=0; iiel< mAttributeElements[string(att->GetName())].size(); iiel++)
                        if( mAttributeElements[string(att->GetName())].at(iiel) >= TMath::Nint(getAttribute( sizebranch.c_str() )->getMean()))
                            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                                         Form("TDataServer::ComputeCorrelationMatrix: Vectors element [%d] requested for attributev [%s] while the size is [%g].",
                                                                              mAttributeElements[string(att->GetName())].at(iiel), att->GetName(), getAttribute( sizebranch.c_str() )->getMean() ));
                    
                    VectSize[leaf->GetName()]=mAttributeElements[ string(att->GetName()) ].size();
                }
                else
                    VectSize[leaf->GetName()]=TMath::Nint(getAttribute( sizebranch.c_str() )->getMean());
	      }
	    else // always considering Double as a default"
	      VectSize[leaf->GetName()]=1;
	    
	  } catch (URANIE::Exceptions::UWarningExceptions& ue)
	  {
            ue.printMessage();
	    continue;
	  }catch (URANIE::Exceptions::UErrorExceptions& ue)
	  {
            ue.printMessage();
            throw ue;
	  }
	
        listOfLeaves->Add( leaf );
	nvars+=VectSize[leaf->GetName()];
        icrt++;
      }

    if (_blog)
      {
        cout << endl << " **  Attributes listing  [" << nvars << "] npat["
	     << npat << "]" << endl;
        listOfAttributes->Print();
      }

    //Create the ouptut matrix, defined as NumberOfEntry times NumberOfVariables
    TMatrixD dval(npat, nvars);
    vector<double> vecMean(nvars,0);
    vector<double> vecSigma(nvars,0);
    
    Double_t drapcrt;
    vector<double> *v; //pointer to read the leaf
    for (Int_t ipat = 0; ipat < npat; ipat++)
      {
        _datatree->GetEntry(telapp->GetEntry(ipat));
        TIter nextLeaf(listOfLeaves);
        icrt = 0;
        while (TLeaf * leaf = (TLeaf *) nextLeaf())
	  {
	    if (getAttribute(leaf->GetName())->getDataTypeStr().EqualTo("D"))
	      v = new vector<double>(1, leaf->GetValue() );
	    else
	      v = (vector<double>*)leaf->GetValuePointer();	    

	    for(unsigned int iel=0; iel<VectSize[leaf->GetName()]; iel++)
            {
                double localvalue= v->at( ((mAttributeElements.count(leaf->GetName())==0) ? iel : mAttributeElements[leaf->GetName()].at(iel)) );
                dval(ipat, icrt)=localvalue; 

		// Mean case
		vecMean[icrt] *= (Double_t)(ipat) / (Double_t)(ipat + 1);
		vecMean[icrt] += localvalue / (Double_t)(ipat + 1);

		// Variance=sigma**2 case
		if (ipat > 0)
		  {
		    vecSigma[icrt] *= (Double_t)(ipat - 1) / (Double_t)(ipat);
		    drapcrt = (Double_t)(ipat + 1) / (Double_t)(ipat);
		    vecSigma[icrt] += drapcrt * (localvalue - vecMean[icrt])
		      * (localvalue - vecMean[icrt]) / (Double_t)(ipat);
		  }
		
		if (_blog)
		  cout << "icrt[" << icrt << "] var[" << leaf->GetName() << "] val[" << v->at(iel) << "] mean["
		       << vecMean[icrt] << "] Sigma[" << vecSigma[icrt] << "]" << endl;
		icrt++;

		if (getAttribute(leaf->GetName())->getDataTypeStr().EqualTo("D"))
		  delete v;
	      }
	  }
      }

    // ToBeDone : Take care of the case where vecSigma[ivar] == 0 !!!! JBB
    for (Int_t ipat = 0; ipat < npat; ++ipat)
      {
        // Second Step : Center and reduce the current column
        for (Int_t ivar = 0; ivar < nvars; ivar++)
	  {
            dval(ipat, ivar) -= vecMean[ivar];
            dval(ipat, ivar) /= TMath::Sqrt( (Double_t)(npat - 1) * vecSigma[ivar] );
	  }
      }

    TMatrixD matCorr(dval, TMatrixD::kTransposeMult, dval);
    if (_blog)
      {
        cout << endl << " **  Correlation Matrix (calculated)" << endl;
        matCorr.Print("");
      }

    //Not touched by JBB when coping for vectors. ToBeDone
    if (soption.Contains("context="))
      {
        TRegexp reg("context=[._/@a-zA-Z]*.tex");
        TString out = soption(reg);
        if (_blog)
	  cout << "out[" << out << "]" << endl;
        TString stexfile = out.Remove(0, 8);
        if (_blog)
	  cout << "stexfile[" << stexfile << "]" << endl;
        // Open the context file
        ofstream sortie;
        sortie.open(stexfile, ios::out);
        sortie.precision(3);
        //      sortie.setf(ios::scientific);

        // Header
        TDatime theTime;
        sortie
	  << Form("%%%% File \"%s\" generated by ROOT v%s",
		  stexfile.Data(), gROOT->GetVersion()) << endl;
        sortie << Form("%%%% DateTime %s", theTime.AsString()) << endl;
        sortie << "%% DataServer : Name=\"" << GetName() << "\" Title=\""
	       << GetTitle() << "\"";
        sortie << endl;

        sortie << "\\starttable[|l|";
        for (Int_t i = 0; i < nvars; i++)
	  sortie << "c|";
        sortie << "]" << endl << "\\HL" << endl;
        if (btype)
	  sortie << "\\VL {\\darkred Rank} ";
        else
	  sortie << "\\VL Variables ";
        for (Int_t i = 0; i < nvars; i++)
	  {
	      TLeaf *leaf = (TLeaf *) (listOfLeaves->At(i));
            TString sname = leaf->GetName();
            leaf->Print();
            if (btype)
	      sortie << "\\VL  " << sname.Remove(0, 3) << " ";
            else
	      sortie << "\\VL  " << sname << " ";
	  }
        sortie << "\\VL \\SR" << endl << "\\HL" << endl;
        for (Int_t i = 0; i < nvars; i++)
	  {
            TLeaf *leaf = (TLeaf *) (listOfLeaves->At(i));
            TString sname = leaf->GetName();
            leaf->Print();
            if (btype)
	      sortie << "\\VL  " << sname.Remove(0, 3) << " ";
            else
	      sortie << "\\VL  " << sname << " ";

            for (Int_t j = 0; j < nvars; j++)
	      {
                sortie << "\\VL  ";
                if (j <= i)
		  sortie << matCorr(i, j) << " ";
	      }
            if (i == 0)
	      {
                sortie << "\\VL \\FR";
	      }
            else
	      {
                if (i == nvars - 1)
		  {
                    sortie << "\\VL \\LR";
		  }
                else
		  {
                    sortie << "\\VL \\FR";
		  }
	      }
            sortie << endl;
	  }
        sortie << "\\HL" << endl << "\\stoptable" << endl;
        ;

        // Close the output file
        sortie.close();

      }
    //Clean
    if( gROOT->FindObject("elist1") ) gROOT->Delete("elist1");
    delete listOfLeaves;
    listOfLeaves = NULL;
    delete listOfAttributes;
    listOfAttributes = NULL;

    if (_blog)
      cout << "****** End of TDataServer::computeCorrelationMatrix" << endl
	   << endl;
    return matCorr;
  }


string TDataServer::checkAttributes(TDataServer *tds, string attlist, map<string, vector<int> > &mAttributeElements)
{
    vector<string> buffer; 
    map<string, unsigned int> mAttributeSize;  ///< Map of size of element for attribute;
     ///< Map of Elements number to run (if vector subselection is requested)
    vector<string>            vNames; ///< Name of the output
    
    URANIE::DataServer::UEntry::StringParser( attlist, ":", buffer);
    
    // XYZ<<" checkAttributes "<<attlist<<" buffer size "<<buffer.size()<<" fromopt "<<fromoption<<endl;
    
    try
    {
        for(unsigned int ibuf=0; ibuf<buffer.size(); ibuf++)
        {
            int ipos = ( (buffer.at(ibuf).find_first_of("[")==string::npos) ? buffer.at(ibuf).size() : buffer.at(ibuf).find_first_of("[") );
            int ilen = ( (buffer.at(ibuf).find_first_of("[")==string::npos) ? 0 : buffer.at(ibuf).find_first_of("]")-1-buffer.at(ibuf).find_first_of("["));            
            string restrBuf = buffer.at(ibuf).substr( 0, ipos);
            string restrInd = ((ilen==0)?"":buffer.at(ibuf).substr( ipos+1, ilen));

            // XYZ<<" ipos/ilen "<<ipos<<"/"<<ilen<<"; restInd/Buf "<<restrInd<<"/"<<restrBuf<<" count "<<mAttributeElements.count(restrBuf)<<endl;
            
            if( find(vNames.begin(), vNames.end(), restrBuf) ==  vNames.end() )
                vNames.push_back( restrBuf );
	    
            if( restrInd != "" && mAttributeElements.count(restrBuf)==0)
                mAttributeElements.insert( std::pair<string, vector<int> >(restrBuf, vector<int> (1, atoi(restrInd.c_str()) ) ) );
            else if( restrInd != "" && mAttributeElements.count(restrBuf)!=0 )
                if( find(mAttributeElements[restrBuf].begin(),mAttributeElements[restrBuf].end(), atoi(restrInd.c_str())) == mAttributeElements[restrBuf].end() )
                    mAttributeElements[restrBuf].push_back( atoi(restrInd.c_str()) );
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
	ue.printMessage();
   	throw ue;
    }
        
    string reducedlist="";
    for(unsigned int ival=0; ival<vNames.size(); ival++)
        reducedlist+=vNames.at(ival)+((ival<(vNames.size()-1))?":":"");

    //    XYZ<<"Reduced list is "<<reducedlist<<endl;

    TList *test=NULL;
    try
    {
        test=tds->getListOfAttributes(reducedlist.c_str(),false);
    } catch( URANIE::Exceptions::UErrorExceptions& ue)
    {
	ue.printMessage();
        delete test;
   	throw ue;
    }

    delete test; test=NULL;
    return reducedlist;
}

  ////////////////////////////////////////////////////////
  //  computeStatistic : Statistical Method
  ////////////////////////////////////////////////////////
  void TDataServer::computeStatistic(const char *varexp, const char *selection,
				     Option_t *option)
  {
    //Mod by JBB 2016-2-24 to take the vectors into account (and get rid of strings)
    if (_blog)
      {
        cout << endl << "****** TDataServer::computeStatistic" << endl;
        cout << " **  varexp[" << varexp << "]" << endl;
        cout << " **  select[" << selection << "]" << endl;
        cout << " **  option[" << option << "]" << endl;
      }
        
    // Treat the selection and the option
    TEventList *telapp = _datatree->GetEventList();
    if (!telapp)
    {
        TCut allCut;
        allCut += _ccut;
        allCut += TCut(selection);
        if( gDirectory != gROOT ) gDirectory = gROOT ;      
        if( gROOT->FindObject("elistStat") ) gROOT->Delete("elistStat");
        _datatree->Draw(">>elistStat",  ((strcmp(allCut.GetTitle(),"")==0) ? "1>0" : allCut.GetTitle()), "goff");
        telapp = (TEventList*) gROOT->Get("elistStat");
    }

    Int_t npat = telapp->GetN();
    if (_blog)
      {
        cout << " **  npat[" << npat << "]" << endl;
        cout << " **  TEventList[" << telapp->GetName() << "]" << endl;
        telapp->Print("all");
      }

    try
      {
	if (npat < 3)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form(
							  "TDataServer::computeStatistic ERROR : the TDataServer \"%s\" to compute the correlation matrix have low pattern (%d < 3)",
							  GetName(), npat));
      }catch( URANIE::Exceptions::UErrorExceptions &ue)
      {
	ue.printMessage();
	if( gROOT->FindObject("elistStat") ) gROOT->Delete("elistStat");
	throw ue;
      }

    // -- ngi -- disable unused branches
    _datatree->SetBranchStatus("*", 0);    
    // Get the current list of attributes
    TList *listOfAttributes = NULL;
    
    try{
        listOfAttributes = getListOfAttributes(varexp,false); //mod by JBB to run on all attributes (not only kAttribute and kConstant)
    }catch(URANIE::Exceptions::UErrorExceptions &ue)
    {
	if( gROOT->FindObject("elistStat") ) gROOT->Delete("elistStat");
	if(listOfAttributes!=NULL)
            delete listOfAttributes;
	throw ue;
    }
    
    //Check that the branch in selection are switched ON as well
    TList *fullListOfAttributes=getListOfAttributes("*",false);
    TIter next2(fullListOfAttributes);
    TString cutstr = TString (_ccut.GetTitle())+"_"+TString(selection);
    while (TAttribute *att = (TAttribute*) next2())
      if( cutstr.Contains(att->GetName()) )
	_datatree->SetBranchStatus( att->GetName(), 1);
    if( fullListOfAttributes != NULL)
      delete fullListOfAttributes;
    fullListOfAttributes=NULL;
    
    //Check that the considered branches are double or vectors.
    TIter next(listOfAttributes);
    UInt_t nvars = 0;
    map<const char*, int> Iterator; // Given the name of the attribute it returns the number of either Double or Vector to be studied
    map<int, int> VectorSize; // Given the number of vector that is under study, it returns the size of it
    while (TAttribute *att = (TAttribute*) next())
      {

        TLeaf *leaf = _datatree->GetLeaf(att->GetName());
        if (leaf != NULL)
	  _datatree->SetBranchStatus(leaf->GetBranch()->GetName(), 1);
	else
	  continue;
	
	Iterator[leaf->GetName()]=-1;

	try
	  {

	    TString theclass = att->getDataTypeStr();
	    if(theclass.EqualTo("S"))
	      throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							   Form("TDataServer::ComputeStatistic is only able to deal with double and vectors. Branch %s has type %s, so nothing will be done for it!",
								att->GetName(), theclass.Data()));
	    else if(theclass.EqualTo("V"))
	      {

		string sizebranch=SIZENAME+"__"+string(att->GetName());
		if(!isAttribute( sizebranch.c_str() ))
		  throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							       Form("TDataServer::ComputeStatistic: vectors are usable when size is an attribute. %s not found so nothing will be done for it!",
								    sizebranch.c_str()));	
		_datatree->SetBranchStatus(sizebranch.c_str(),1);
		Long64_t nPat = _datatree->Draw(sizebranch.c_str(), selection, "goff");
        if (nPat > 1000000)
        {
             _datatree->SetEstimate(_datatree->GetEntries());
             nPat = _datatree->Draw(sizebranch.c_str(), selection, "goff");
        }
		Double_t* values = _datatree->GetV1();

		getAttribute( sizebranch.c_str() )->setMinimum(values[TMath::LocMin(nPat, values)]);
		getAttribute( sizebranch.c_str() )->setMaximum(values[TMath::LocMax(nPat, values)]);
		Double_t mean = TMath::Mean(nPat, values);
		Double_t sumVar = 0.0;
		for (Int_t i = 0; i < npat; i++)
		  {
		    sumVar += (values[i] - mean) * (values[i] - mean);
		  }

		getAttribute( sizebranch.c_str() )->setStd(TMath::Sqrt(sumVar / (nPat - 1)));
		getAttribute( sizebranch.c_str() )->setMean(mean);
		_datatree->SetBranchStatus(sizebranch.c_str(),0);

		if( getAttribute( sizebranch.c_str() )->getMinimum() != getAttribute( sizebranch.c_str() )->getMaximum() )
		  throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							       Form("TDataServer::ComputeStatistic: vectors are usable when size is not changing. %s Std is not 0 (%g) so nothing will be done for it!",
								    sizebranch.c_str(), getAttribute( sizebranch.c_str() )->getStd() ));

		VectorSize[nvars]=TMath::Nint(getAttribute( sizebranch.c_str() )->getMean()+1);
		Iterator[leaf->GetName()]=nvars++;
	      }
	    else // always considering Double as a default"
	      {
		VectorSize[nvars]=1;
		Iterator[leaf->GetName()]=nvars++;
	      }
	    
	  } catch (URANIE::Exceptions::UWarningExceptions& ue)
	  {
            ue.printMessage();
	    continue;
	  }
      }

    if (_blog)
      {
        cout << endl << " **  Attributes listing  [" << nvars << "] npat["
	     << npat << "]" << endl;
        listOfAttributes->Print();
      }

    // For vectors && double
    vector<vector<double>* > v_Min, v_Max, v_Mean, v_Sigma;
    for(unsigned int ivect=0; ivect<nvars; ivect++)
      {
	v_Min.push_back( new vector<double>( VectorSize[ivect], DLE_MAX ) );
	v_Max.push_back( new vector<double>( VectorSize[ivect], -DLE_MAX ) );
	v_Mean.push_back( new vector<double>( VectorSize[ivect], 0 ) );
	v_Sigma.push_back( new vector<double>( VectorSize[ivect], 0 ) );
      }


    TList *listOfLeaves = new TList();
    next.Reset();
    //    _datatree->GetEntry(telapp->GetEntry(0));
    while (TAttribute *att = (TAttribute *) next())
      {

        TLeaf *leaf = _datatree->GetLeaf(att->GetName());
        //    draw(att->GetName());
        if (leaf == NULL)
	  {
	     cout << "The leaf [" << att->GetName() << "] is empty" << endl;
            //  _datatree->Print();
	     //_datatree->GetListOfLeaves()->ls();
	    continue;
	  }

	if( Iterator[leaf->GetName()] == -1) continue;
	
        listOfLeaves->Add(leaf);
      }

    Double_t drapcrt;
    Long64_t ncrtpat;
    for (Int_t ipat = 0; ipat < npat; ipat++)
      {
        ncrtpat = telapp->GetEntry(ipat);	
        //    _datatree->GetEntry(telapp->GetEntry(ipat));
        _datatree->GetEntry(ncrtpat);
        TIter nextLeaf(listOfLeaves);

        while (TLeaf *leaf = (TLeaf *) nextLeaf())
	  {
	    //Get rid of information for string and not usable vectors	    
	    if( Iterator[leaf->GetName()] == -1) continue;

	    vector<double> *v=NULL;
	    if( getAttribute( leaf->GetName() )->getDataTypeStr().EqualTo("D") )
	      v = new vector<double>(1,leaf->GetValue());
	    else if( getAttribute( leaf->GetName() )->getDataTypeStr().EqualTo("V") )
	      v = ((vector<double>*)leaf->GetValuePointer());
	    
	    for(unsigned int iel=0; iel<v->size(); iel++)
	      {
		// Min case
		if (v->at(iel) < v_Min[Iterator[leaf->GetName()]]->at(iel))
		  v_Min[Iterator[leaf->GetName()]]->at(iel) = v->at(iel);

		// Max case
		if (v->at(iel) > v_Max[Iterator[leaf->GetName()]]->at(iel))
		  v_Max[Iterator[leaf->GetName()]]->at(iel) = v->at(iel);

		// Mean case
		v_Mean[Iterator[leaf->GetName()]]->at(iel) *= (Double_t)(ipat) / (Double_t)(ipat + 1);
		v_Mean[Iterator[leaf->GetName()]]->at(iel) += v->at(iel) / (Double_t)(ipat + 1);

		// Variance=sigma**2 case
		if (ipat > 0)
		  {
		    v_Sigma[Iterator[leaf->GetName()]]->at(iel) *= (Double_t)(ipat - 1) / (Double_t)(ipat);
		    drapcrt = (Double_t)(ipat + 1) / (Double_t)(ipat);
		    v_Sigma[Iterator[leaf->GetName()]]->at(iel) += drapcrt * (v->at(iel) - v_Mean[Iterator[leaf->GetName()]]->at(iel))
		      * (v->at(iel) - v_Mean[Iterator[leaf->GetName()]]->at(iel)) / (Double_t)(ipat);
		  }

		//Case to have total min, max, mean and std (total==over all element of a vector)
		if( getAttribute( leaf->GetName() )->getDataTypeStr().EqualTo("V") )
		  {
		    // Min case
		    if (v->at(iel) < v_Min[Iterator[leaf->GetName()]]->at(v->size()))
		      v_Min[Iterator[leaf->GetName()]]->at(v->size()) = v->at(iel);

		    // Max case
		    if (v->at(iel) > v_Max[Iterator[leaf->GetName()]]->at(v->size()))
		      v_Max[Iterator[leaf->GetName()]]->at(v->size()) = v->at(iel);

		    // Mean case
		    v_Mean[Iterator[leaf->GetName()]]->at(v->size()) *= (Double_t)(iel+v->size()*ipat) / (Double_t)(iel+v->size()*ipat + 1);
		    v_Mean[Iterator[leaf->GetName()]]->at(v->size()) += v->at(iel) / (Double_t)(iel+v->size()*ipat + 1);

		    // Variance=sigma**2 case
		    if ( (iel+v->size()*ipat) > 0)
		      {
			v_Sigma[Iterator[leaf->GetName()]]->at(v->size()) *= (Double_t)(iel+v->size()*ipat - 1) / (Double_t)(iel+v->size()*ipat);
			drapcrt = (Double_t)(iel+v->size()*ipat + 1) / (Double_t)(iel+v->size()*ipat);
			v_Sigma[Iterator[leaf->GetName()]]->at(v->size()) += drapcrt * (v->at(iel) - v_Mean[Iterator[leaf->GetName()]]->at(v->size()))
			  * (v->at(iel) - v_Mean[Iterator[leaf->GetName()]]->at(v->size())) / (Double_t)(iel+v->size()*ipat);
		      }
		  }
	      }

	    if ((_blog) & ((ipat < 10) || (ipat > npat - 10)))
	      cout << " ** icrt [" << Iterator[leaf->GetName()] << "]; ipat [" << ipat << "/" << ncrtpat << "/" << npat << "]; val ["
		   << UEntry::dumpVectorStr(v) << "]; mean [" << UEntry::dumpVectorStr(v_Mean[Iterator[leaf->GetName()]])
		   << "]; Variance [" << UEntry::dumpVectorStr(v_Sigma[Iterator[leaf->GetName()]]) << "]" << endl;
	    if ((_blog) & (ipat == 10))
	      cout << " ** icrt [" << Iterator[leaf->GetName()] << "]; ipat [" << ipat << "/" << ncrtpat << "/" << npat << "]; ... " << endl;
	    
	    
	    if( getAttribute( leaf->GetName() )->getDataTypeStr().EqualTo("D") )
	      delete v;
	  }
      }

    TIter nextLeaf(listOfLeaves);
    while (TLeaf *leaf = (TLeaf *) nextLeaf())
      {

	//Get rid of information for string and not usable vectors	    
	if( Iterator[leaf->GetName()] == -1) continue;

	TAttribute *att = getAttribute( leaf->GetName()  );
	if (_blog)
	  cout << " **  End att [" << att->GetName() << "] mean["
	       << UEntry::dumpVectorStr(v_Mean[Iterator[leaf->GetName()]]) << "] Sigma[" << UEntry::dumpVectorStr(v_Sigma[Iterator[leaf->GetName()]]) << "] Min["
	       << UEntry::dumpVectorStr(v_Min[Iterator[leaf->GetName()]]) << "] Max[" << UEntry::dumpVectorStr(v_Max[Iterator[leaf->GetName()]]) << "]"
	       << endl;

	for(unsigned int iel=0; iel<v_Min[Iterator[leaf->GetName()]]->size(); iel++)
	  {
	    att->setMinimum(v_Min[Iterator[leaf->GetName()]]->at(iel), iel);
	    att->setMaximum(v_Max[Iterator[leaf->GetName()]]->at(iel), iel);
	    att->setStd(TMath::Sqrt(v_Sigma[Iterator[leaf->GetName()]]->at(iel)), iel);
	    att->setMean(v_Mean[Iterator[leaf->GetName()]]->at(iel), iel);
	  }

      }

    // We delete the data after use
    for(vector<vector<double>* >::iterator it=v_Min.begin(); it!=v_Min.end(); it++)
      delete (*it);
    v_Min.clear();
    for(vector<vector<double>* >::iterator it=v_Max.begin(); it!=v_Max.end(); it++)
      delete (*it);
    v_Max.clear();
    for(vector<vector<double>* >::iterator it=v_Mean.begin(); it!=v_Mean.end(); it++)
      delete (*it);
    v_Mean.clear();
    for(vector<vector<double>* >::iterator it=v_Sigma.begin(); it!=v_Sigma.end(); it++)
      delete (*it);
    v_Sigma.clear();
    
    // -- ngi -- re-activate the branches
    _datatree->SetBranchStatus("*", 1);

    //Clean
    if( gROOT->FindObject("elistStat") ) gROOT->Delete("elistStat");
    delete listOfLeaves;
    listOfLeaves = NULL;
    delete listOfAttributes;
    listOfAttributes = NULL;
    if (_blog)
      {
        cout << endl << "****** End of TDataServer::computeStatistic" << endl;
      }

  }
  
  ////////////////////////////////////////////////////////
  //  Quantiles : Statistical Method
  ////////////////////////////////////////////////////////
  void TDataServer::estimateQuantile(const char *attName, Double_t proba,
				     Double_t &quantile, Double_t confidence, Option_t *option)
  // Nicolas Gilardi 2009.03.18
  // fga::modification 16/12/2009 11:09
  // Make an option to use the interpolation method
  // End Of fga::modification 16/12/2009 11:09
  {
 //Test the requested attribute
    TString theclass = getAttribute( attName )->getDataTypeStr();
    unsigned int VectSize=1;
    try{
      if(theclass.EqualTo("S"))
	throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
						     Form(
							  "TDataServer::EstimateQuantile is only able to deal with double. Branch %s has type %s, so nothing will be done!",
							  attName, theclass.Data()));
      //Test the vector case
      else if(theclass.EqualTo("V"))
	{
	  string sizebranch=SIZENAME+"__"+string(attName);
	  if(!isAttribute( sizebranch.c_str() ))
	    throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 Form("TDataServer::EstimateQuantile: vectors are usable when size is an attribute. %s not found so nothing will be done for it!",
							      sizebranch.c_str()));
	  computeStatistic( sizebranch.c_str() );
	  if( getAttribute( sizebranch.c_str() )->getMinimum() != getAttribute( sizebranch.c_str() )->getMaximum() )
	    throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 Form("TDataServer::EstimateQuantile: vectors are usable when size is not changing. %s Std is not 0 (%g) so nothing will be done for it!",
							      sizebranch.c_str(), getAttribute( sizebranch.c_str() )->getStd() ));
	  VectSize=TMath::Nint(getAttribute( sizebranch.c_str() )->getMean());
	}

    } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
	ue.printMessage();
	return;
      }
    
    try
      {
	if(theclass.EqualTo("V") && !_bQuantVectPrinted)
	  {
	    _bQuantVectPrinted=!_bQuantVectPrinted;
	    throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 Form("TDataServer::estimateQuantile: The requested attribute (%s) is a vector. \n ==>    The double value \"returned\" by the method correspond to the last column of the vector !! \n    ==> Results are available thanks through the dedicated methods of the concerned attribute \n          getQuantile, getQuantiles, getQuantileVector...",attName));
	  }
      }catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
	ue.printMessage();
      }

    Int_t nbData = getNPatterns();
    if (VectSize*(_datatree->GetEntries()) > 1000000)
    {
        _datatree->SetEstimate(VectSize*(_datatree->GetEntries()));
    }
    _datatree->Draw(attName, _ccut, "goff");
    
    Double_t* data = _datatree->GetV1();
    double *col=0;
    Int_t *sortIndex=0;
    for(unsigned int iel=0; iel<VectSize; iel++)
      {
	if(VectSize!=1)
	  {
	    col=new double[nbData];
	    int icol=0;
	    for(unsigned int ipat=0; ipat<VectSize*nbData; ipat++)      
	      if( (ipat % VectSize) == iel )
		col[icol++]=data[ipat];
	  }	    

	sortIndex = new Int_t[nbData];

	TMath::Sort(nbData, ((VectSize==1) ? data : col), sortIndex, 0);
	Int_t index = computeIndexQuantileWilks(proba, confidence, nbData);
	if (_blog)
	  {
	    cout<<endl;
	    cout << " ** Number of Data: [" << nbData << "]; ";
	    cout << "Requested quantile: [" << proba << "]; ";
	    cout << "Confidence: [" << confidence << "]; " << endl;
	    cout << " ** Index Returned [" << index << "]" << endl;
	  }

	// Case of the interpolation
	TString soption = TString(option);
	soption.ToLower();
	if ((nbData < 1000) && (soption.Contains("interpolate")))
	  {
	    if (_blog)
	      cout << " **** Case of interpolate nbData [" << nbData << "]"
		   << endl;
	    Double_t s = 0.0, sold = 0.0, sdiff = 0.0;
	    Int_t j = 0;
	    while (sold < 1. - confidence)
	      {
		s = sold;
		sdiff = TMath::Binomial(nbData, j) * pow(1. - proba, (Double_t) j)
		  * pow(proba, (Double_t)(nbData - j));
		sold += sdiff;
		if (_blog)
		  cout << " **** ** iter j[" << j << "] s[" << s << "] sold["
		       << sold << "] sdiff[" << sdiff << "]" << endl;
		j++;
	      }
	    // decrment the index as we are greater than the thresold
	    //  j--;
	    Int_t ncIndex = TMath::FloorNint(nbData * (1. - proba));
	    Int_t rIndex = j;
	    Int_t nPos = TMath::CeilNint(nbData * proba) + rIndex;
	    if (_blog)
	      cout << " **** Final s[" << s << "] sdiff[" << sdiff << "] sold["
		   << sold << "] -- nc[" << ncIndex << "] r[" << rIndex
		   << "]  nPos[" << nPos << "]" << endl;
	    Double_t fa2 = (1.0 - confidence - s) / sdiff;
	    Double_t fa1 = (sold + confidence - 1.0) / sdiff;
	    if (_blog)
	      cout << " **** Coefficients a1[" << fa1 << "] a2[" << fa2 << "]"
		   << endl;
	    Double_t dlast = ((VectSize==1) ? data[sortIndex[index - 1]] : col[sortIndex[index - 1]]);
	    Double_t dwilks = ((VectSize==1) ? data[sortIndex[index]] : col[sortIndex[index]]);
	    if (_blog)
	      cout << " **** Previous[" << dlast << "] wilks[" << dwilks << "]"
		   << endl;
	    Double_t dquantileinterpole = fa1 * dwilks + fa2 * dlast;
	    if (_blog)
	      cout << " ** Quantile with interpolation [" << dquantileinterpole
		   << "]" << endl;
	    quantile = dquantileinterpole;

	  }
	else
	  {
	    if (index != -1)
	      quantile = ((VectSize==1) ? data[sortIndex[index]] : col[sortIndex[index]]);
	  }

	 getAttribute( attName )->setQuantile(proba, quantile, iel);
	
	// Clean
	if(col) delete[] col;
	delete[] sortIndex;
	sortIndex = NULL;
      }
  }

  void TDataServer::estimateQuantile(const char *attName, Int_t nProba,
				     Double_t* proba, Double_t* quantile, Double_t confidence,
				     Option_t *option)
  // Nicolas Gilardi 2009.03.18    
  // fga::modification 16/12/2009 11:09
  // Make an option to use the interpolation method
  // and use the estimateQuantile method for one quantile for each quantile in the vector
  // End Of fga::modification 16/12/2009 11:09
  {
    Int_t nbData = getNPatterns();

    if (_blog)
      {
        TString probaString("");
        for (Int_t i = 0; i < nProba - 1; i++)
	  {
            probaString += Form("%f, ", proba[i]);
	  }
        probaString += Form("%f", proba[nProba - 1]);

        cout << "** Number of Data: [" << nbData;
        cout << "]; Requested quantile: [" << probaString;
        cout << "]; Confidence: [" << confidence << "]" << endl;
      }

    for (Int_t i = 0; i < nProba; i++)
      {
        //         Int_t index = computeIndexQuantileWilks(proba[i], confidence, nbData);
        //         quantile[i] = data[sortIndex[index]];
        estimateQuantile(attName, proba[i], quantile[i], confidence, option);
      }
  }

  void TDataServer::computeQuantile(const char *attName, Double_t proba,
				    Double_t &quantile, Int_t type)
  {
    //Test the requested attribute
    TString theclass = getAttribute( attName )->getDataTypeStr();
    unsigned int VectSize=1;
    try{
      if(theclass.EqualTo("S"))
	throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
						     Form(
							  "TDataServer::ComputeQuantile is only able to deal with double. Branch %s has type %s, so nothing will be done!",
							  attName, theclass.Data()));
      //Test the vector case
      else if(theclass.EqualTo("V"))
	{
	  string sizebranch=SIZENAME+"__"+string(attName);
	  if(!isAttribute( sizebranch.c_str() ))
	    throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 Form("TDataServer::ComputeQuantile: vectors are usable when size is an attribute. %s not found so nothing will be done for it!",
							      sizebranch.c_str()));
	  computeStatistic( sizebranch.c_str() );
	  if( getAttribute( sizebranch.c_str() )->getMinimum() != getAttribute( sizebranch.c_str() )->getMaximum() )
	    throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 Form("TDataServer::ComputeQuantile: vectors are usable when size is not changing. %s Std is not 0 (%g) so nothing will be done for it!",
							      sizebranch.c_str(), getAttribute( sizebranch.c_str() )->getStd() ));
	  VectSize=TMath::Nint(getAttribute( sizebranch.c_str() )->getMean());
	}

    } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
	ue.printMessage();
	return;
      }

    // try
    //   {
    if(theclass.EqualTo("V") && !_bQuantVectPrinted)
      {
	_bQuantVectPrinted=!_bQuantVectPrinted;
	// 	    throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
	// 							 Form("TDataServer::computeQuantile: The requested attribute (%s) is a vector. \n ==>    The double value \"returned\" by the method correspond to the last column of the vector !! \n    ==> Results are available thanks through the dedicated methods of the concerned attribute \n          getQuantile, getQuantiles, getQuantileVector...",attName));
	// 	  }
	// }catch (URANIE::Exceptions::UWarningExceptions& ue)
	// {
	// 	ue.printMessage();
	(*_fLogger) << "TDataServer::computeQuantile: The requested attribute (" << attName  << ") is a vector " << Endl;
	(*_fLogger) <<" ==> The double value \"returned\" by the method correspond to the last column of the vector !!" << Endl;
	(*_fLogger) <<" ==> Results are available thanks through the dedicated methods of the concerned attribute " << Endl;
	(*_fLogger) <<"          getQuantile, getQuantiles, getQuantileVector..." << Endl;
      }

    // Should we use ROOT TMath::Quantiles function ?
    Bool_t useROOTQuantiles = kFALSE;

    const Int_t nbData = getNPatterns();
    if (VectSize*(_datatree->GetEntries()) >1000000)
    {
        _datatree->SetEstimate(VectSize*(_datatree->GetEntries()));    
    }
    _datatree->Draw(attName, _ccut, "goff");
    
    if (_blog)
      {
        cout << "** Number of Data: [" << nbData << "]; Requested quantile: ["
	     << proba << "]" << endl;
      }

    if (useROOTQuantiles)
      {
        if (_blog)
	  {
            cout << "** using TMath::Quantiles function" << endl;
	  }
        TMath::Quantiles(nbData, 1, _datatree->GetV1(), &quantile, &proba, 0,
			 NULL, type);
      }
    else
      {
        if (_blog)
	  {
            cout << "** using TDataServer::tdsQuantiles function" << endl;
	  }

	double *data=_datatree->GetV1();
	double *col=0;
	for(unsigned int iel=0; iel<VectSize; iel++)
	  {
	    if(VectSize!=1)
	      {
		col=new double[nbData];
		int icol=0;
		for(unsigned int ipat=0; ipat<VectSize*nbData; ipat++)      
		  if( (ipat % VectSize) == iel )
		    col[icol++]=data[ipat];
	      }	    
	    internalQuantilesComputation(nbData, 1, ((VectSize==1) ? data : col), &quantile, &proba, 0, NULL, type);
	    getAttribute( attName )->setQuantile(proba, quantile, iel);
	    if(col) delete[] col;
	  }
	
      }
  }

  void TDataServer::computeQuantile(const char *attName, Int_t nProba,
				    Double_t* proba, Double_t* quantiles, Int_t type)
  {
    //Test the requested attribute
    TString theclass = getAttribute( attName )->getDataTypeStr();
   unsigned int VectSize=1;
    try{
      if(theclass.EqualTo("S"))
	throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
						     Form(
							  "TDataServer::ComputeQuantile is only able to deal with double. Branch %s has type %s, so nothing will be done!",
							  attName, theclass.Data()));
      //Test the vector case
      else if(theclass.EqualTo("V"))
	{
	  string sizebranch=SIZENAME+"__"+string(attName);
	  if(!isAttribute( sizebranch.c_str() ))
	    throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 Form("TDataServer::ComputeQuantile: vectors are usable when size is an attribute. %s not found so nothing will be done for it!",
							      sizebranch.c_str()));
	  computeStatistic( sizebranch.c_str() );
	  if( getAttribute( sizebranch.c_str() )->getMinimum() != getAttribute( sizebranch.c_str() )->getMaximum() )
	    throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
							 Form("TDataServer::ComputeQuantile: vectors are usable when size is not changing. %s Std is not 0 (%g) so nothing will be done for it!",
							      sizebranch.c_str(), getAttribute( sizebranch.c_str() )->getStd() ));
	  VectSize=TMath::Nint(getAttribute( sizebranch.c_str() )->getMean());
	}

    } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
	ue.printMessage();
	return;
      }	

    // try
    //   {
    if(theclass.EqualTo("V") && !_bQuantVectPrinted)
      {
	_bQuantVectPrinted=!_bQuantVectPrinted;
	// 	    throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
	// 							 Form("TDataServer::computeQuantile: The requested attribute (%s) is a vector. \n ==>    The double value \"returned\" by the method correspond to the last column of the vector !! \n    ==> Results are available thanks through the dedicated methods of the concerned attribute \n          getQuantile, getQuantiles, getQuantileVector...",attName));
	// 	  }
	// }catch (URANIE::Exceptions::UWarningExceptions& ue)
	// {
	// 	ue.printMessage();
	(*_fLogger) << "TDataServer::computeQuantile: The requested attribute (" << attName  << ") is a vector " << Endl;
	(*_fLogger) <<" ==> The double value \"returned\" by the method correspond to the last column of the vector !!" << Endl;
	(*_fLogger) <<" ==> Results are available thanks through the dedicated methods of the concerned attribute " << Endl;
	(*_fLogger) <<"          getQuantile, getQuantiles, getQuantileVector..." << Endl;
      } 

    
    // Should we use ROOT TMath::Quantiles function ?
    Bool_t useROOTQuantiles = kFALSE;

    const Int_t nbData = getNPatterns();
    if (VectSize*(_datatree->GetEntries()) > 1000000)
    {
        _datatree->SetEstimate(VectSize*(_datatree->GetEntries()));
    }
    _datatree->Draw(attName, _ccut, "goff");

    if (_blog)
      {
        TString probaString("");
        for (Int_t i = 0; i < nProba - 1; i++)
	  {
            probaString += Form("%f, ", proba[i]);
	  }
        probaString += Form("%f", proba[nProba - 1]);
        cout << "** Number of Data: [" << nbData << "]; Requested quantile: ["
	     << probaString << "]" << endl;
      }

    if (useROOTQuantiles)
      {
        if (_blog)
	  {
            cout << "** using TMath::Quantiles function" << endl;
	  }
        TMath::Quantiles(nbData, nProba, _datatree->GetV1(), quantiles, proba,
			 0, NULL, type);
      }
    else
      {
        if (_blog)
	  {
            cout << "** using TDataServer::tdsQuantiles function" << endl;
	  }
	
	double *data=_datatree->GetV1();
	double *col=0;
	for(unsigned int iel=0; iel<VectSize; iel++)
	  {
	    if(VectSize!=1)
	      {
		col=new double[nbData];
		int icol=0;
		for(unsigned int ipat=0; ipat<VectSize*nbData; ipat++)      
		  if( (ipat % VectSize) == iel )
		    col[icol++]=data[ipat];
	      }
	    
	    internalQuantilesComputation(nbData, nProba, ((VectSize==1) ? data : col), quantiles, proba, 0, NULL, type);

	    for(int iq=0; iq<nProba; iq++)
	      getAttribute( attName )->setQuantile(proba[iq], quantiles[iq], iel);

	    if(col) delete[] col;
	  }
	
      }
  }
  ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  ///                                                    Add by MSA
  ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  void TDataServer::computeQuantileCV(TString yname, Double_t alpha,
				      TString zname, Double_t zalpha, Double_t &quantile, Double_t &rho)
  {

    try
      {
	//verify the entrance data
	if ((alpha <= 0.0) || (alpha >= 1.0))
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form(
							  "TDataServer::computeQuantileCV - ERROR : Invalid value for alpha: [%f]; range of probability parameter alpha is ]0,1[",
							  alpha));
	if (!isAttribute(yname))
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form(
							  "TDataServer::computeQuantileCV - ERROR: this attribut [%s] is unknown",
							  yname.Data()));
	if (!isAttribute(zname))
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form(
							  "TDataServer::computeQuantileCV - ERROR: this attribut [%s] is unknown",
							  zname.Data()));
	Int_t nS, N1, N0, n, K=0;
	N0 = 0;
	N1 = 0;
	n = 0;
	quantile = 0.0;
	nS = getNPatterns();
	Double_t *w = new Double_t[nS];
	if (_blog)
	  cout << "Number of patterns: " << nS << endl;
	for (Int_t j = 0; j < nS; j++)
	  {
	    if (getValue(zname, j) <= zalpha)
	      N0 = N0 + 1;
	    else
	      N1 = N1 + 1;
	  }
	if (_blog)
	  {
	    cout << "Number of NO: " << N0 << endl;
	    cout << "Number of N1:" << N1 << endl;
	  }
	if (N0 <= 0)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form(
							  "TDataServer::computeQuantileCV - ERROR : Invalid N0 (number of z<zalpha) [%d]",
							  N0));
	if (N1 <= 0)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form(
							  "TDataServer::computeQuantileCV - ERROR : Invalid N0 (number of z>zalpha) [%d]",
							  N1));
	if (nS != (N0 + N1))
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     Form(
							  "TDataServer::computeQuantileCV - ERROR : Invalid N0 and N1 (N0+N1 is not equal to nS) [%d+%d # %d]",
							  N0, N1, nS));
	// compute the Wj et sum of the Wj                         
	Int_t *index = new Int_t[nS];
	getTuple()->Draw(yname, "", "goff");
	TMath::Sort(nS, getTuple()->GetV1(), index, kFALSE);
	for (Int_t j = 0; j < nS; j++)
	  {
	    if (getValue(zname, j) <= zalpha)
	      w[j] = alpha / N0;
	    else
	      w[j] = (1 - alpha) / N1;
	    ;
	  }
	if (_blog)
	  {
	    for (Int_t i = 0; i < nS; i++)
	      cout << " W: " << w[index[i]] << "  Z: "
		   << getValue(zname, index[i]) << endl;
	  }
	// compute the sum of ordered Wj and find the index K
	Double_t *sumW = new Double_t[nS];
	Double_t sum;
	sum = 0;
	for (Int_t i = 0; i < nS; i++)
	  {
	    n = index[i];
	    sum += w[n];
	    sumW[n] = sum;
	    if (sum < alpha)
	      K = i;
	  }
	K = K + 1;
	if (_blog)
	  {
	    cout << "I : Y        Z        W         sumW" << endl;
	    for (Int_t i = 0; i < nS; i++)
	      cout << i << ": " << getValue(yname, index[i]) << "  "
		   << getValue(zname, index[i]) << "  " << w[index[i]]
		   << "  " << sumW[index[i]] << endl;
	    cout << "Indice K: " << K << endl;
	  }
	//    if ( !TMath::AreEqualAbs(sumW[index[nS-1]],1.0,1.e-9))
	if (TMath::Abs(sumW[index[nS - 1]] - 1.0) > 1.e-9)
	  throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
						     "TDataServer::computeQuantileCV - ERROR : the sum of Wj is not equal to 1!!! ");
	//seach the empirique quantile
	quantile = getValue(yname, index[K]);
	if (_blog)
	  cout << "Aplha quantile is : " << quantile << endl;
	//compute rho
	Int_t *indZ = new Int_t[nS];
	Int_t *indY = new Int_t[nS];
	for (Int_t i = 0; i < nS; i++)
	  {
	    indZ[i] = 0;
	    indY[i] = 0;
	    if (getValue(zname, i) <= zalpha)
	      indZ[i] = 1;
	    if (getValue(yname, i) <= quantile)
	      indY[i] = 1;
	  }
	if (_blog)
	  {
	    cout << "Value of the indicator function of Y ans Z variables"
		 << endl;
	    for (Int_t i = 0; i < nS; i++)
	      cout << "Indicator Y: " << indY[i] << " Indicator Z: "
		   << indZ[i] << endl;
	    //(Form("%s_zapha",GetName()
	  }
	addAttribute("IndZ", 0, 1);
	addAttribute("IndY", 0, 1);
	Int_t nx;
	TString att;
	nx = getNAttributes();
	TMatrixD x(nx + 1, nS);
	for (Int_t k = 0; k < nS; k++)
	  {
	    x(0, k) = k;
	    for (Int_t i = 0; i < nx - 2; i++)
	      {
		att = getAttribute(i)->GetName();
		x(i + 1, k) = getValue(att, k);
	      }
	    x(nx - 1, k) = indZ[k];
	    x(nx, k) = indY[k];
	  }
	getTuple()->~TDSNtupleD();
	createTuple();
	Double_t *xx = new Double_t[nx + 1];
	for (Int_t k = 0; k < nS; k++)
	  {
	    for (Int_t i = 0; i < nx + 1; i++)
	      xx[i] = x(i, k);
	    getTuple()->Fill(xx);
	  }
	TMatrixD matcorr = computeCorrelationMatrix("IndZ:IndY");
	Double_t alpha_carre = alpha * alpha;
	Double_t nom = matcorr(0, 1) - alpha_carre;
	Double_t dem = alpha - alpha_carre;
	rho = nom / dem;
	rho = TMath::Abs(rho);
	if (_blog)
	  {
	    cout << "Coefficent de correlation entre IndZ et IndY: "
		 << matcorr(0, 1) << endl;
	    cout << " Value of the variance reduction:" << rho << endl;
	  }
	delAttribute("IndZ");
	delAttribute("IndY");

	// Clean
	delete[] w;
	delete[] index;
	delete[] sumW;
	delete[] indY;
	delete[] indZ;
	delete[] xx;
      } catch (URANIE::Exceptions::UWarningExceptions& ue)
      {
	ue.printMessage();
	throw ue;
      } catch (URANIE::Exceptions::UErrorExceptions& ue)
      {
	ue.printMessage();
	throw ue;
	exit(-1);
      }

  }

  ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  ///                                                   End of add by MSA
  ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  void TDataServer::internalQuantilesComputation(Int_t nData, Int_t nProba,
						 Double_t* data, Double_t* quantiles, Double_t* proba, Bool_t isSorted,
						 Int_t* index, Int_t type)
  {
    if (type < 1 || type > 9)
      {
        cout << "*** ERROR: illegal value of type: [" << type << "]" << endl;
        return;
      }

    Int_t *ind = 0;
    Bool_t indAllocated = kFALSE;

    Double_t *sortedData = 0;
    Bool_t sortAllocated = kFALSE;

    if (!isSorted)
      {
        sortedData = new Double_t[nData];
        sortAllocated = kTRUE;
        if (index)
	  ind = index;
        else
	  {
            ind = new Int_t[nData];
            indAllocated = kTRUE;
	  }
        TMath::Sort(nData, data, ind, 0);
        for (Int_t i = 0; i < nData; i++)
	  sortedData[i] = data[ind[i]];
      }
    else
      sortedData = data;

    //------------------------------------------------------------------------//
    // The chosen algorithm is based on the one from Mathematica, given       //
    // on http://mathworld.wolfram.com/Quantile.html                          //
    //------------------------------------------------------------------------//

    // Coefficients            a     b     c     d
    Double_t abcd[9][4] =
      {
	{ 0.0, 0.0, 1.0, 0.0 },   // Method 1
	{ 0.0, 0.0, 0.0, 0.0 },  // This algorithm doesn't work for method 2
	{ 0.5, 0.0, 0.0, 0.0 },   // Method 3
	{ 0.0, 0.0, 0.0, 1.0 },   // Method 4
	{ 0.5, 0.0, 0.0, 1.0 },   // Method 5
	{ 0.0, 1.0, 0.0, 1.0 },   // Method 6
	{ 1.0, -1.0, 0.0, 1.0 },   // Method 7
	{ 1 / 3., 1 / 3., 0.0, 1.0 },   // Method 8
	{ 3 / 8., 1 / 4., 0.0, 1.0 } }; // Method 9
    Double_t x, xFrac;
    Int_t xFloor, xCeil;

    if (type == 2)
      {   // Special case for method 2 
        for (Int_t i = 0; i < nProba; i++)
	  {
            x = proba[i] * nData;
            xFloor = TMath::FloorNint(x);
            xCeil = TMath::CeilNint(x);
            if (xFloor < 1)
	      quantiles[i] = sortedData[0];
            else if (xCeil > nData)
	      quantiles[i] = sortedData[nData - 1];
            else if (xCeil == xFloor && xCeil < nData)
	      quantiles[i] = 0.5
		* (sortedData[xFloor - 1] + sortedData[xFloor]);
            else
	      quantiles[i] = sortedData[xCeil - 1];
	  }
      }
    else
      {   // All the other methods
        for (Int_t i = 0; i < nProba; i++)
	  {
            x = abcd[type - 1][0] + proba[i] * (nData + abcd[type - 1][1]);
            xFloor = TMath::FloorNint(x);
            xCeil = TMath::CeilNint(x);
            xFrac = x - xFloor;
            if (xFloor < 1)
	      quantiles[i] = sortedData[0];
            else if (xCeil > nData)
	      quantiles[i] = sortedData[nData - 1];
            else
	      quantiles[i] =
		sortedData[xFloor - 1]
		+ (sortedData[xCeil - 1]
		   - sortedData[xFloor - 1])
		* (abcd[type - 1][2]
		   + abcd[type - 1][3] * xFrac);
	  }
      }

    if (indAllocated)
      delete[] ind;
    ind = NULL;
    if (sortAllocated)
      delete[] sortedData;
    sortedData = NULL;
  }

Int_t TDataServer::computeIndexQuantileWilks(Double_t alpha, Double_t beta,
        Int_t n)
// Jean-Marc Martinez 2009.03.17
// Nicolas   Gilardi  2009.03.18
{
    try
    {
        // Validate input parameters
        if ((alpha <= 0.0) || (alpha >= 1.0))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "Invalid value for alpha: [%f]; range of probability parameter alpha is ]0,1[",
                            alpha));
        if ((beta > 0.999999999999999) || (beta < 0.0))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "Invalid value for beta: [%f]; range of confidence parameter beta is [0,1[",
                            beta));

        // Compute the minimal size "ns" of the sample requiered for the given 
        // confidence level.
        // The equation is:
        //      $ns = [\log(1-beta)/\log(alpha)] + 1$
        // where [xxx] is the integer part of xxx
        Int_t ns = TMath::FloorNint(TMath::Log(1. - beta) / (TMath::Log(alpha)))
                + 1;
        if (n < ns)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "*** Invalid value for n: [%d]; the sample size is too small for the given confidence [%5.3f] and probability [%5.3f].\n*** In this case, the minimum sample size is: %d",
                            n, beta, alpha, ns));

        // Computing the quantile index (supposed to be between 0 and n-1)
        if (n <= 1000)
        {
            // If n is "small", we use the exact Wilks quantile formula
            Double_t s = 0.0;
            Int_t j;
            Int_t nc = TMath::FloorNint(n * (1. - alpha));
            for (j = 0; j <= nc; j++)
            {
                s += TMath::Binomial(n, j) * pow(1. - alpha, (Double_t) j)
                        * pow(alpha, (Double_t)(n - j));
                if (s > 1. - beta)
                    break;
            }
            return n - j;
        }
        else
        {
            // If n is "large", we use an approximation of the Wilks formula to 
            // reduce the numerical errors        
            Int_t r = TMath::CeilNint(
                    computeInverseNormalCDF(beta)
                            * TMath::Sqrt(alpha * (1.0 - alpha) * n));
            Int_t k = TMath::FloorNint(alpha * n) + r;
            if (_blog)
            {
                cout << " *** r [" << r << "]" << endl;
                cout << " *** alpha * n[" << alpha * n
                        << "] TMath::FloorNint(alpha * n) ["
                        << TMath::FloorNint(alpha * n) << "]" << endl;
                cout << " *** approximation ["
                        << alpha * n
                                + computeInverseNormalCDF(beta)
                                        * TMath::Sqrt(alpha * (1.0 - alpha) * n)
                        << "]" << endl;
            }

            // This should never happen as alpha < 1.0... but who knows ? 
            if (k == n)
            {
                k = n - 1;
            }

            return k;
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;

    }
}

Double_t TDataServer::computeInverseNormalCDF(Double_t p, Double_t mean,
        Double_t stddev)
// This function is a copy of TNormalDistribution::getInvCDF. It should be 
// moved to a utility library.
{

    try
    {
        if (p < 0. || p > 1.)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TDataServer::computeInverseNormalCDF  This is not a valuable given probability");
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    }
    static Double_t borne = 1.e-13;

    const Double_t a[6] =
	{ -3.969683028665376e+01, 2.209460984245205e+02,
	  -2.759285104469687e+02, 1.383577518672690e+02,
	  -3.066479806614716e+01, 2.506628277459239e+00 };
    const Double_t b[5] =
	{ -5.447609879822406e+01, 1.615858368580409e+02, -1.556989798598866e+02,
	  6.680131188771972e+01, -1.328068155288572e+01 };
    const Double_t c[6] =
	{ -7.784894002430293e-03, -3.223964580411365e-01,
	  -2.400758277161838e+00, -2.549732539343734e+00,
	  4.374664141464968e+00, 2.938163982698783e+00 };
    const Double_t d[4] =
	{ 7.784695709041462e-03, 3.224671290700398e-01, 2.445134137142996e+00,
	  3.754408661907416e+00 };
    
    if (p < borne)
        p = borne;
    else if (p > 1. - borne)
        p = 1. - borne;

    if (_blog)
        cout << "p [" << p << "]" << endl;

    Double_t q = p < 1 - p ? p : 1.0 - p;
    Double_t x = mean, t, u;

    if (q > 0.02425)
    {
        /* Rational approximation for central region. */
        u = q - 0.5;
        t = u * u;
        u = u
                * (((((a[0] * t + a[1]) * t + a[2]) * t + a[3]) * t + a[4]) * t
                        + a[5])
                / (((((b[0] * t + b[1]) * t + b[2]) * t + b[3]) * t + b[4]) * t
                        + 1);
    }
    else
    {
        /* Rational approximation for tail region. */
        t = sqrt(-2 * log(q));
        u =
                (((((c[0] * t + c[1]) * t + c[2]) * t + c[3]) * t + c[4]) * t
                        + c[5])
                        / ((((d[0] * t + d[1]) * t + d[2]) * t + d[3]) * t + 1);
    }

    t = TMath::Erfc(-u / sqrt(2.)) / 2. - q;

    t = t * TMath::Sqrt(TMath::TwoPi()) * exp(u * u / 2); /* f(u)/df(u) */
    u = u - t / (1 + u * t / 2); /* Halley's method */

    x = p > 0.5 ? -u : u;
    return x * stddev + mean;
}

///////////////////////////////////////////////////////////
//  Visualisation
///////////////////////////////////////////////////////////

void TDataServer::checkCanvasCreation(Option_t *option)
{
    
    TString soption = TString(option);
    soption.ToLower();
    
    if (gPad)
    {
	_canvas = (TCanvas *) (gPad->GetCanvas());
	if( TString(_canvas->GetName()).BeginsWith("__tdscan__") && !soption.Contains("same") )
	    _canvas->Clear();
	else
	    _canvas = (TCanvas*)gPad;
    }
    else
    {
	stringstream sstr; sstr << "__tdscan__" << _drawingGarbageCollector->GetEntries() ;
	_canvas = new TCanvas(sstr.str().c_str(),sstr.str().c_str(),1);
	_drawingGarbageCollector->Add(_canvas);
    }

//    XYZ<<"[91m The pointer to the canvas created or not is "<<_canvas<<"[0m"<<endl;
    
}

void TDataServer::draw(const char* varexp, const char* selection,
		       Option_t* option, bool internalCall)
{
    TTreeFormula *intest=0;
    try
    {
	if (_datatree)
        {

	    bool ToBeDrawn = true;
	    if( TString(option).Contains("goff") )
		ToBeDrawn = false;

	    if( ToBeDrawn )
		if(!internalCall)
		    checkCanvasCreation(option);
	    stringstream sstr;

	    TString svar = TString(varexp);
	    Int_t ncpt = svar.CountChar(':');
            
            if( !svar.Contains(">>") )
            {
                TString svarcheck = svar;
                svarcheck.ReplaceAll(":", " ");
                istringstream inStream((char *) svarcheck.Data());
                string inName;
                Int_t ncrtpart = 0;
                while (inStream >> inName)
                {
                    TString satt(inName.c_str());
                    intest=new TTreeFormula("test",inName.c_str(), _datatree);
                    if( intest->GetNdim() == 0 )
                        throw URANIE::Exceptions::UErrorExceptions(__FILE__,
                                                                   __LINE__,
                                                                   Form("TDataServer::draw method : the variable [%s] is not part of this dataserver, so it could not be drawn", satt.Data()));
                    delete intest;
                    intest=NULL;
                }
            }

	    TString soption = TString(option);
	    soption.ToLower();
	    soption.ReplaceAll(" ", "");

	    // Keep the current default number of bins for ROOT histograms
	    Int_t old_nBins = gEnv->GetValue("Hist.Binning.1D.x", 4);
	    switch (ncpt)
            {
            case 0: // case of 1D
	    {
                Int_t type_nclass = 0; // NBINX = 100 as .rootrc field NBINX
                // \todo Le 2008-01-16 20:51:57 Mettre la valeur par defaut dans une variable gUranie qui sera toujours remise
                //    avec gEnv->GetValue("Hist.Binning.1D.x", 4);
                Int_t nBins = old_nBins;

                if (soption.Contains("nclass="))
		{
                    TRegexp reg("nclass=[a-z0-9]*");
                    TString out = soption(reg);
                    TString snclass = out.Remove(0, 7);
                    soption.ReplaceAll(Form("nclass=%s", out.Data()), "");

                    if (snclass == "root")
			type_nclass = 0;
                    else if (snclass == "sturges")
			type_nclass = 1;
                    else if (snclass == "fd")
			type_nclass = 2;
                    else if (snclass == "scott")
			type_nclass = 3;
                    else if (snclass.IsDigit())
			type_nclass = 4;
                    else
		    {
                        cout << "Type[" << snclass
			     << "] no recognized in TDataServer::draw method"
			     << endl;
                        return;
		    }

                    //     Int_t npat = (Int_t)(
                    //_datatree->GetEntries(_ccut + selection));
                    TCut allCut; allCut += _ccut; allCut += TCut(selection);
                    Long64_t npat = _datatree->Draw(varexp, allCut, "goff");
                    if (npat > 1000000)
                    {
                        _datatree->SetEstimate( npat );
                        npat = _datatree->Draw(varexp, allCut, "goff");
                    }
                    Double_t dMin = TMath::MinElement(npat, _datatree->GetV1());
                    Double_t dMax = TMath::MaxElement(npat, _datatree->GetV1());
                    if (_blog)
		    {
                        cout << "dMin[" << dMin << "]" << endl;
                        cout << "dMax[" << dMax << "]" << endl;
		    }

                    switch (type_nclass)
		    {
		    case 1:
		    { // Sturges (default en R)
			nBins = (Int_t)(TMath::Log2((Double_t)(npat))) + 1;
			break;
		    }
		    case 2:
		    { // Freedman & Diaconis
			// Calcul des Quantiles
			const Int_t nq = 2;
			Double_t *xq = new Double_t[nq]; // position where to compute the 3 quantiles in 0.25, 0.50, 0.75
			Double_t *yq = new Double_t[nq]; // array to contain the quantiles
			xq[0] = 0.25;
			xq[1] = 0.75;
			TMath::Quantiles(npat, nq, _datatree->GetV1(), yq, xq,
					 kFALSE);
			Double_t RQ = yq[1] - yq[0];
			if (_blog)
			{
			    cout << "yq[0][" << yq[0] << "]" << endl;
			    cout << "yq[1][" << yq[1] << "]" << endl;
			    cout << "dRQ[" << RQ << "]" << endl;
			}
			nBins = (Int_t)(
			    (dMax - dMin) * pow((Double_t)(npat), 1. / 3.)
			    / (2. * RQ));
			delete[] xq;
			delete[] yq;
			break;
		    }
		    case 3:
		    { // Scoot
			Double_t dRMS = TMath::RMS(npat, _datatree->GetV1());
			if (_blog)
                            cout << "dRMS[" << dRMS << "]" << endl;
			nBins = (Int_t)(
			    (dMax - dMin) * pow((Double_t)(npat), 1. / 3.)
			    / (3.5 * dRMS));
			break;
		    }
		    case 4:
		    { // Fixed number
			nBins = snclass.Atoi();
		    }
		    }
		}
                gEnv->SetValue("Hist.Binning.1D.x", nBins);
	    }
            }

	    // the selection
	    TCut allCut; allCut += _ccut; allCut += TCut(selection);

	    // Case of Cobweb :: Add "para" in the soption if the number of attributes is greater than 4 and "para" does not exist
	    if ((ncpt > 4) && !soption.Contains("para"))
            {
		soption.Append(" para");
            }

	    // The ROOT visualisation
	    _datatree->Draw(svar.Data(), ((strcmp(allCut.GetTitle(),"")==0)?"1>0":allCut.GetTitle()), soption.Data());
	    // if(soption.Contains("para"))
	    // {
	    // 	XYZ<<"  svar["<<svar<<"]  allCut["<<allCut<<"]  soption["<<soption<<"] "<<endl;
	    // 	gDirectory->GetList()->ls();
	    // 	cout<<endl;
	    // 	_canvas->GetListOfPrimitives()->ls();
	    // 	cout<<endl;
	    // }

	    ////////////////////////////////////
	    // The Legend of the axis
	    if (!soption.Contains("goff") && !soption.Contains("para") && !soption.Contains("same"))
		drawLegend(varexp, selection);
	    
	    TH1 *htemp = ( (ToBeDrawn) ? (TH1*)gPad->GetPrimitive("htemp") : NULL);
	    if(htemp!=NULL && !internalCall)
	    {
		sstr << "__tdshisto__" << _drawingGarbageCollector->GetEntries() ;
		htemp->SetName(sstr.str().c_str());
	    }
	  
	    // Draw Legend for the cobweb case
	    if (soption.Contains("para") && ToBeDrawn)
            {
		if (ncpt > 0)
                {
		    TParallelCoord* para = (TParallelCoord*) gPad->GetListOfPrimitives()->FindObject("ParaCoord");
		    if (para)
                    {
			TList *tl = para->GetVarList();
			if (tl)
                        {
			    for (Int_t i = 0; i < tl->GetEntries(); i++)
                            {
				TParallelCoordVar* axis = (TParallelCoordVar*) tl->FindObject( tl->At(i)->GetName());
				axis->SetName( _dataSpecification->getAttribute( tl->At(i)->GetName())->getLegend());
				axis->SetBit(kCanDelete);
                            }
                        }
			else
                        {
			    throw URANIE::Exceptions::UErrorExceptions(__FILE__,
								       __LINE__,
								       "TDataServer::draw method with \"para\" option\n The list of variable in the \"ParaCoord\" object is empty");
                        }
                    }
		    else
                    {
			throw URANIE::Exceptions::UErrorExceptions(__FILE__,
								   __LINE__,
								   "TDataServer::draw method with \"para\" option\n No found \"ParaCoord\" object");
                    }

                }
		else
                {
		    throw URANIE::Exceptions::UErrorExceptions(__FILE__,
							       __LINE__,
							       Form(
								   "TDataServer::draw method with \"para\" option\n The \"para\" option is available at least for 2 attributes, and we use [%s].",
								   varexp));
                }
            }

	    // Restore the default number of bins for 1D histograms
	    gEnv->SetValue("Hist.Binning.1D.x", old_nBins);
	    
        }
	else
        {
	    throw URANIE::Exceptions::UExceptions(__FILE__, __LINE__,
						  Form("TDataServer::draw[%s]  ERROR The TTree is NULL",
						       GetName()));
        }

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
	ue.printMessage();
        delete intest;
        intest=NULL;
	throw ue;
	exit(-1);
    }
}

void TDataServer::draw(UInt_t nOrder, const char* selection, Option_t* option)
{
    try
    {
        if (!_dataSpecification)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TDataServer::draw  No dataspecification!");
        UInt_t fInvecLen = (UInt_t)(_dataSpecification->getNAttributes());

        if (!_datatree)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TDataServer::draw  No data Tree!");

	if (nOrder > fInvecLen)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TDataServer::draw  indice superieur a la dimension");
	
        draw(_dataSpecification->getAttribute(nOrder)->GetName(), selection,   option, kTRUE);
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    }
}

void TDataServer::drawQQPlot(const char* varexp, const char* laws, int nq,
                             Option_t* option )
{
    TString soption = TString(option)+",qqplot";
#ifdef WITH_MATHMORE
    drawQQorPPPlot(varexp, laws, nq, soption.Data());
#else
    UNUSED(varexp, laws, nq, option);
    cout << " ** DataServer::drawQQPlot::Problem " << endl;
    cout << "** Nothing can be done as ROOT has not been installed with the Mathmore library" << endl;
#endif
}

void TDataServer::drawPPPlot(const char* varexp, const char* laws, int nq,
                             Option_t* option )
{
    TString soption = TString(option)+",ppplot";
#ifdef WITH_MATHMORE
    drawQQorPPPlot(varexp, laws, nq, soption.Data());
#else
    UNUSED(varexp, laws, nq, option);
    cout << " ** DataServer::drawPPPlot::Problem " << endl;
    cout << "** Nothing can be done as ROOT has not been installed with the Mathmore library" << endl;
#endif
}

void TDataServer::drawQQorPPPlot(const char* varexp, const char* laws, int nq,
                             Option_t* option )
{
#ifdef WITH_MATHMORE

    TString svar = TString(varexp);
    try
    {
        if (!_datatree)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TDataServer::drawQQPlot Error : There is no tree."));
        
        if (svar.CountChar(':') > 0 || svar=="*")
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TDataServer::drawQQPlot Error : MultiVariate variable QQ plot is not yet available"));
        else if ( !_dataSpecification->isAttribute(varexp) )
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TDataServer::drawQQPlot Error : [%s] is not an existing attribute",varexp));
      
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
        
    // Gestion des options
    TString opt = option;
    opt.ToLower();    
    
    bool DrawPP = opt.Contains("ppplot");

    vector<TString> Laws;
    TString svarcheck = laws;
    svarcheck.ToLower();
    
    svarcheck.ReplaceAll(":", " ");
    istringstream inStream((char *) svarcheck.Data());
    string inName;
    Int_t ncrtpart = 0;
    while (inStream >> inName)
        Laws.push_back( TString(inName) );
    
    int Col[5] = {2,3,4,5,6};
    
    // Proba and quantile
    Double_t *dvalp = new Double_t[nq];
    Double_t *dvalq = new Double_t[nq];
    Double_t *dvalT = new Double_t[nq];           
    
    computeStatistic(svar);
    // Empirical
    for(Int_t i=0; i<nq; i++)
        dvalp[i] = ( 1.0 + i)/(1.0+nq);
    computeQuantile(svar, nq, dvalp, dvalq);
    
    checkCanvasCreation(option);
    stringstream sstr;
    int reallaw=0;
    
    vector<double> _paramValues;
    TLegend *leg=NULL;
    vector<TGraph*> gr;
    
    for(unsigned int ilaw=0; ilaw<Laws.size(); ilaw++)
    {
        double dmin=getAttribute(svar)->getMinimum();
        double dmax=getAttribute(svar)->getMaximum();
        double dmean=getAttribute(svar)->getMean();
        double dstd=getAttribute(svar)->getStd();

        TString slaw= Laws.at(ilaw);
        TString parlaw= Laws.at(ilaw);

        if( slaw.BeginsWith("normal") )
        {                    
            if ( slaw.Length()!=6 )
            {
                TRegexp reg("normal([+/-]?[0-9.e+-]*,[+]?[0-9.e+-]*)");
                TString out = slaw(reg);

                if(out.Length() == 0)
                {
                    cout << "You must precise only mean and standard deviation for gaussian law. Dropping this law! ["<<slaw<<"]"<<endl;
                    continue;
                }
                else
                {
                    URANIE::DataServer::TDataServer::findParametersLaw(parlaw.Remove(0, 7), _paramValues, _blog);
                    dmean=_paramValues.at(0);
                    dstd=_paramValues.at(1);                            
                }
            }
            for(Int_t i=0; i<nq; i++)
                dvalT[i] = ( (DrawPP) ? (ROOT::Math::normal_cdf(dvalq[i],dstd,dmean)) : (ROOT::Math::normal_quantile(dvalp[i],dstd) + dmean) );
        }
        else if( slaw.BeginsWith("uniform") )
        {
            if ( slaw.Length()!=7 )
            {
                TRegexp reg("uniform([+/-]?[0-9.e+-]*,[+/-]?[0-9.e+-]*)");
                TString out = slaw(reg);

                if(out.Length() == 0)
                {
                    cout << "You must precise only min and max for uniform law. Dropping this law!"<<endl;
                    continue;
                }
                else
                {
                    URANIE::DataServer::TDataServer::findParametersLaw(parlaw.Remove(0, 8), _paramValues, _blog);
                    dmin=_paramValues.at(0);
                    dmax=_paramValues.at(1);                            
                }
            }
                    
            for(Int_t i=0; i<nq; i++) 
                dvalT[i] = ( (DrawPP) ? ROOT::Math::uniform_cdf(dvalq[i],dmin,dmax) : ROOT::Math::uniform_quantile(dvalp[i],dmin, dmax) );
        }
        else if( slaw.BeginsWith("beta") )
        {
            double alpha=1.0, beta=1.0;
            if ( slaw.Length()!=4 )
            {
                TRegexp reg("beta([+]?[0-9.e+-]*,[+]?[0-9.e+-]*,[+/-]?[0-9.e+-]*,[+/-]?[0-9.e+-]*)");
                TString out = slaw(reg);

                if (out.Length() == 0)
                {

                    TRegexp reg2("beta([+]?[0-9.e+-]*,[+]?[0-9.e+-]*)");
                    TString out2 = slaw(reg2);

                    if (out2.Length() == 0)
                    {
                        cout << "You must precise at least alpha and beta (possibly min and max as well) for beta law. Dropping this law!"<<endl;
                        continue;
                    }
                }

                URANIE::DataServer::TDataServer::findParametersLaw(parlaw.Remove(0, 5), _paramValues, _blog);
                alpha=_paramValues.at(0);
                beta=_paramValues.at(1);
                if( _paramValues.size()==4 )
                {
                    dmin = _paramValues.at(2);
                    dmax = _paramValues.at(3);
                }

            }
                    
            for(Int_t i=0; i<nq; i++)
                dvalT[i] = ( (DrawPP) ? ROOT::Math::beta_cdf( (dvalq[i]-dmin)/(dmax-dmin),alpha,beta) : ROOT::Math::beta_quantile(dvalp[i], alpha, beta)*(dmax-dmin) + dmin );
        }
        else if (slaw.BeginsWith("exponential"))
        {
            if (slaw.Length() != 11)
            {
                TRegexp reg("exponential([+]?[0-9.e+-]*,[+]?[0-9.e+-]*)");
                TString out = slaw(reg);
                try
                {
                    if (out.Length() == 0)
                        throw URANIE::Exceptions::UErrorExceptions(__FILE__,__LINE__,
                                                                   Form("TDataServer::drawQQorPPPlot ERROR \n The current law is not well defined [%s]",slaw.Data()));
                } catch (URANIE::Exceptions::UErrorExceptions& ue)
                {
                    ue.printMessage();
                    delete[] dvalp;
                    delete[] dvalq;
                    delete[] dvalT;   
                    throw ue;
                }

                URANIE::DataServer::TDataServer::findParametersLaw(parlaw.Remove(0, 12), _paramValues, _blog);
                dmean = _paramValues.at(0);
                dstd = _paramValues.at(1);

            }
            for(Int_t i=0; i<nq; i++) 
                dvalT[i] = ( (DrawPP) ? (ROOT::Math::exponential_cdf((dvalq[i]-dstd),dmean)) : (ROOT::Math::exponential_quantile(dvalp[i],dmean)+dstd ) );
        }
        //////////////////////////
        // Case of lognormal law
        else if (slaw.BeginsWith("lognormal"))
        {
            if (slaw.Length() != 9)
            {
                TRegexp reg("lognormal([+/-]?[0-9.e+-]*,[+]?[0-9.e+-]*,[+/-]?[0-9.e+-]*)");
                TString out = slaw(reg);

                if (out.Length() == 0)
                {
                    TRegexp reg2("lognormal([+/-]?[0-9.e+-]*,[+]?[0-9.e+-]*)");
                    TString out2 = slaw(reg2);

                    try
                    {
                        if (out2.Length() == 0)                        
                            throw URANIE::Exceptions::UErrorExceptions(__FILE__,__LINE__,
                                                                       Form("TDataServer::drawQQorPPPlot ERROR \n The current law is not well defined [%s]",slaw.Data()));
                    }
                    catch (URANIE::Exceptions::UErrorExceptions& ue)
                    {
                        ue.printMessage();
                        delete[] dvalp;
                        delete[] dvalq;
                        delete[] dvalT;   
                        throw ue;
                    }
                }

                URANIE::DataServer::TDataServer::findParametersLaw(parlaw.Remove(0, 10), _paramValues, _blog);
                dmean = _paramValues.at(0);
                dstd = _paramValues.at(1);
                if( _paramValues.size()==3 )               
                    dmin = _paramValues.at(2);
                else
                    dmin=0.0;

                Double_t sigmacrt = TMath::Log(dstd) / 1.645;
                dstd=sigmacrt;                                                                         
                Double_t meancrt = TMath::Log(dmean-dmin) - 0.5 * dstd*dstd;
                dmean=meancrt;
            }
            for(Int_t i=0; i<nq; i++)
                dvalT[i] = ( (DrawPP) ?  0.5 * (1 + TMath::Erf( (TMath::Log(dvalq[i]-dmin) - dmean) / TMath::Sqrt( 2 * dstd * dstd))) :  dmin + TMath::Exp(ROOT::Math::normal_quantile(dvalp[i],dstd) + dmean));
        }
        //////////////////////////
        // Case of Weibull law
        else if (slaw.BeginsWith("weibull"))
        {
            if (slaw.Length() != 7)
            {
                TRegexp reg("weibull([+]?[0-9.e+-]*,[+]?[0-9.e+-]*,[+]?[0-9.e+-]*)");
                TString out = slaw(reg);
                try {
                    if (out.Length() == 0)
                        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                                   Form("TDataServer::drawQQorPPPlot ERROR \n The current law is not well defined [%s]",slaw.Data()));
                } catch (URANIE::Exceptions::UErrorExceptions& ue) {
                    ue.printMessage();
                    delete[] dvalp;
                    delete[] dvalq;
                    delete[] dvalT;   
                    throw ue;
                }

                URANIE::DataServer::TDataServer::findParametersLaw(parlaw.Remove(0, 8), _paramValues, _blog);
                dmean = _paramValues.at(0); //lambda
                dstd = _paramValues.at(1); //k 
                dmin = _paramValues.at(2); //
            }
            for(Int_t i=0; i<nq; i++) 
                dvalT[i] = ( (DrawPP) ?  1-exp(-pow((dvalq[i]-dmin)/dmean,dstd)) : dmin + dmean* TMath::Power(-1.0 * TMath::Log(1.0 - dvalp[i]), 1.0 / dstd) );
            //;1-exp(-pow(dvalp[i]/dmean,dstd) ) );
        }
        //////////////////////////
        // Case of Gamma law
        else if (slaw.BeginsWith("gamma"))
        {
         
            if (slaw.Length() != 5)
            {
                TRegexp reg("gamma([+]?[0-9.e+-]*,[+]?[0-9.e+-]*,[+]?[0-9.e+-]*)");
                TString out = slaw(reg);
                try {
                    if (out.Length() == 0)
                        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                                   Form("TDataServer::drawQQorPPPlot ERROR \n The current law is not well defined [%s]",slaw.Data()));
                } catch (URANIE::Exceptions::UErrorExceptions& ue) {
                    ue.printMessage();
                    delete[] dvalp;
                    delete[] dvalq;
                    delete[] dvalT;   
                    throw ue;
                }

                URANIE::DataServer::TDataServer::findParametersLaw(parlaw.Remove(0, 6), _paramValues, _blog);
                dmean = _paramValues.at(0);
                dstd = _paramValues.at(1);
                dmin = _paramValues.at(2);
            }
            for(Int_t i=0; i<nq; i++) 
                dvalT[i] = ( (DrawPP) ? (ROOT::Math::gamma_cdf((dvalq[i]-dmin),dmean,dstd)) : (ROOT::Math::gamma_quantile(dvalp[i],dmean,dstd)+dmin ) );
        }
        //////////////////////////
        // Case of Gumbelmax law
        else if (slaw.BeginsWith("gumbelmax"))
        {
         
            if (slaw.Length() != 9)
            {
                TRegexp reg("gumbelmax([+]?[0-9.e+-]*,[+]?[0-9.e+-]*)");
                TString out = slaw(reg);
                try {
                    if (out.Length() == 0)
                        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                                   Form("TDataServer::drawQQorPPPlot ERROR \n The current law is not well defined [%s]",slaw.Data()));
                } catch (URANIE::Exceptions::UErrorExceptions& ue) {
                    ue.printMessage();
                    delete[] dvalp;
                    delete[] dvalq;
                    delete[] dvalT;   
                    throw ue;
                }

                URANIE::DataServer::TDataServer::findParametersLaw(parlaw.Remove(0, 10), _paramValues, _blog);
                dmean = _paramValues.at(0);
                dstd = _paramValues.at(1);
            }
            for(Int_t i=0; i<nq; i++) 
                dvalT[i] = ( (DrawPP) ?  exp(- exp(- (dvalq[i] - dmean) / dstd)) : dmean - dstd * TMath::Log(-TMath::Log(dvalp[i])) );
        }
        else
        {
            try{
                throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                             Form("TDataServer::drawQQPlot Warning : The law [%s] is not yet available !",slaw.Data()));
            }
            catch(URANIE::Exceptions::UErrorExceptions& ue)
            {
                ue.printMessage();
                delete[] dvalp;
                delete[] dvalq;
                delete[] dvalT;   
                throw ue;
            }
            continue;
        }
        
        //Quantile-Quantile Normal law
        gr.push_back( new TGraph(nq, dvalT, ((DrawPP) ? dvalp : dvalq)) );
        gr.at(reallaw)->SetMarkerStyle(20);
        gr.at(reallaw)->SetMarkerColor(Col[reallaw]);
        gr.at(reallaw)->SetMarkerSize(.9);
        gr.at(reallaw)->Draw( ((reallaw == 0) ? "AP" : "Psame") );

        if( reallaw==0)
        {
            gr.at(reallaw)->GetXaxis()->SetTitle(((DrawPP) ? "Theoretical Probability" : "Theoretical Quantile") );
            gr.at(reallaw)->GetYaxis()->SetTitle(((DrawPP) ? "Empirical Probability" : "Empirical Quantile"));

            // line y=x
            TGraph *tbisq = new TGraph(2);
            tbisq->SetPoint(0, ((DrawPP) ? 0 : dmin), ((DrawPP) ? 0 : dmin));
            tbisq->SetPoint(1, ((DrawPP) ? 1 : dmax), ((DrawPP) ? 1 : dmax));
            tbisq->SetLineStyle(2);
            tbisq->SetLineWidth(4);
            tbisq->Draw("Lsame");
                    
            leg = new TLegend(0.2, 0.6, 0.45,0.85);
            leg->AddEntry(tbisq,"Reference", "l");
            leg->AddEntry(gr.at(reallaw), (svar+TString(" vs ")+slaw.Data()), "p");

        }
        else
            leg->AddEntry(gr.at(reallaw), (svar+TString(" vs ")+slaw.Data()), "p"); 
        
        reallaw++;        
               
    }
    leg->Draw();

    delete[] dvalp;
    delete[] dvalq;
    delete[] dvalT;    
            
#else
    UNUSED(varexp, laws, nq, option);
    cout << " ** DataServer::drawQQorPPPlot::Problem " << endl;
    cout << "** Nothing can be done as ROOT has not been installed with the Mathmore library" << endl;
#endif
    
}

void TDataServer::drawECDF(const char* varexp, const char* selection,
        Option_t* option)
{
    try
    {
        TString svar = TString(varexp);
        if (svar.CountChar(':') > 0)
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TDataServer::drawLegend Error : MultiVariate variable CDF is not yet available"));
        }
        if (_datatree)
        {
            TString sOption = TString(option);
            // Select the data in the TTree and retrieve their number
            TCut allCut; allCut += _ccut; allCut += TCut(selection);
            Int_t nbSelected = (Int_t) _datatree->Draw(varexp, allCut, sOption + TString("goff"));
            if (nbSelected>1000000)
            {
                   _datatree->SetEstimate( nbSelected );
                   nbSelected = _datatree->Draw(varexp, allCut, sOption + TString("goff"));
            }
            // Array storing the data to display
            Double_t *dataSelection = _datatree->GetV1(); // unsorted data
            Int_t *sortIndex = new Int_t[nbSelected];  // indices of sorted data
            Double_t * values = new Double_t[nbSelected];         // sorted data
            Double_t * proba = new Double_t[nbSelected];   // normalised indices

            TMath::Sort(nbSelected, dataSelection, sortIndex, kFALSE);

            // Construction of the empirical CDF        
            for (Int_t i = 0; i < nbSelected; i++)
            {
                values[i] = dataSelection[sortIndex[i]];
                proba[i] = (Double_t)(i + 1) / nbSelected;
            }

	    checkCanvasCreation(option);
	    
            // Drawing the empirical CDF
            TGraph *g = new TGraph(nbSelected, values, proba);
	    _drawingGarbageCollector->Add(g);
            g->Draw(TString("LAP") + sOption);

            // Graph details
            Width_t nwidth = 4;
            g->SetLineWidth(nwidth);
            g->SetLineColor(8);
            g->SetTitle(Form("Experimental CDF : %s", svar.Data()));

	    delete[] sortIndex;
	    delete[] values;
	    delete[] proba;
        }
        else
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form("TDataServer::drawLegend Error : There is no tree."));

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    }
}
void TDataServer::drawCDF(const char* varexp, const char* selection,  Option_t* option)
{
    TString svar = TString(varexp);
    try
    {
        if (!_dataSpecification)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TDataServer::drawCDF Error :\n the TDS[%s] has no dataspecification!!", GetName()));
        if (!_datatree)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TDataServer::drawCDF Error :\n the TDS[%s] has no datatree!!", GetName()));

        if (svar.CountChar(':') > 0)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TDataServer::drawCDF Error :\n MultiVariates drawCDF [%s] not yet implemented in TDS[%s]!", svar.Data(), GetName()));

        if (!_dataSpecification->isAttribute(varexp))
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                       Form("TDataServer::drawCDF Error :\n The attribute [%s] is not present in TDS[%s]!", varexp, GetName()));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }

    TCut sselect = _ccut + TCut(selection);
    TString svartitle = _dataSpecification->getAttribute(varexp)->getLegend();
    if ( _blog )
        cout << " *** svar[" << svar << "] Title[" << svartitle << "]" << endl;


    // Treat the options
    TString sOption = TString(option);
    sOption.ToLower();
    if (_blog)
        cout << " ** sOption[" << sOption << "]" << endl;

    sOption += "goff";

    Bool_t bDrawLeg = kFALSE;
    if (sOption.Contains("legend"))
        bDrawLeg = kTRUE;

    if (sOption.Contains("rafu")) {
        drawCDFRaFu(varexp, selection, option);
    } else {
	    checkCanvasCreation(option);
	    stringstream sstr;

	    _datatree->Draw(Form("%s>>hplot%s", varexp, varexp), sselect, sOption);
	    
	  TH1F *htemp = (TH1F*) gDirectory->Get(Form("hplot%s", varexp));
	  htemp->SetDirectory(0);
	  sstr.str(""); sstr << "__tdsCDFplot__" <<  _drawingGarbageCollector->GetEntries() ;	  
	  htemp->SetName( sstr.str().c_str() );
	  _drawingGarbageCollector->Add(htemp);
	  Int_t nBin = htemp->GetNbinsX();
	  Double_t dSum = htemp->GetSumOfWeights();


        TString stitle = Form("Attribute : %s", svartitle.Data());
        TString ssss = TString(sselect.GetTitle());
        if (ssss.Length() != 0)
            stitle += Form(" ( %s ) ", sselect.GetTitle());

	  TH1* hcdf =(TH1F*) htemp->Clone(Form("hcdf%s",varexp));// htemp->DrawCopy();
        hcdf->SetTitle(stitle);
	  hcdf->GetXaxis()->SetTitle(svartitle);
        hcdf->GetYaxis()->SetTitle(Form("CDF (%s)", svartitle.Data()));
	  hcdf->SetStats(kFALSE);
	  hcdf->SetFillColor(0);
	  hcdf->Reset();

	  TH1F* hccdf = (TH1F*) htemp->Clone(Form("hccdf%s",varexp));
	  hccdf->SetStats(kFALSE);
	  hccdf->Reset();

	  _drawingGarbageCollector->Add(hcdf);
	  _drawingGarbageCollector->Add(hccdf);
	  
	  Double_t sum = 0;
	  Double_t sumccdf = 1.0;
	  for (Int_t i = 0; i < nBin; i++)
            {
	      sum += htemp->GetBinContent(i) / dSum;
	      sumccdf -= htemp->GetBinContent(i) / dSum;
	      hcdf->SetBinContent(i, sum);
	      hccdf->SetBinContent(i, sumccdf);
            }
	  hcdf->SetBinContent(nBin, 1.0);
	  hccdf->SetBinContent(nBin, 0.0);
        Width_t nwidth = 3;
	  hcdf->SetLineWidth(nwidth);
	  hcdf->SetLineColor(kBlack);
	  hcdf->Draw();
	  
        TLegend *leg = new TLegend(0.11, 0.65, 0.30, 0.89);
	  _drawingGarbageCollector->Add(leg);
	  leg->SetBorderSize(0);
	  leg->SetTextAlign(12);
        leg->SetTextSize(0.03);
	  leg->SetFillColor(0);
	  leg->AddEntry(hcdf, "  CDF", "L");

        // Treat again the options
	  sOption = TString(option);
	  sOption += "same";
	  if (_blog)
	    cout << " ** sOption[" << sOption << "]" << endl;

        // Case of CCDF
        if (sOption.Contains(TString("ccdf"))) {
	      hccdf->SetLineColor(kRed);
	      hccdf->SetLineWidth(nwidth);
	      hccdf->SetFillColor(0);
            hccdf->Draw(sOption);

            hcdf->GetYaxis()->SetTitle("");
	      leg->AddEntry(hccdf, "  CCDF", "L");
            bDrawLeg = kTRUE;
            }

        // Case of Histogram
	  // \todo Do like for the macro in tutorial hist/twoscales 
        if (sOption.Contains(TString("hist"))) {
            _datatree->Draw(Form("%s>>htemp", varexp), (const char *) (_ccut + TCut(selection)), sOption);
		TH1F* htemp2 = (TH1F*)gDirectory->Get("htemp");
		htemp2->Scale(1.0 / htemp2->GetMaximum());
		htemp2->Draw(sOption);

            hcdf->GetYaxis()->SetTitle("");
		leg->AddEntry(htemp2, "Histogram", "L");
            bDrawLeg = kTRUE;

		sstr.str(""); sstr << "__tdsCDFhisto__" <<  _drawingGarbageCollector->GetEntries() ;	  
		htemp2->SetName( sstr.str().c_str() );
		_drawingGarbageCollector->Add(htemp2);
            }

        if ( bDrawLeg )
	  leg->Draw();
        }
    }

void TDataServer::drawCDFRaFu(const char* varexp, const char* selection,
        Option_t* option)
{
    TCut sselect = _ccut + TCut(selection);
    TString svar = TString(varexp);
    TString svartitle = TString(varexp);
    if (_dataSpecification->isAttribute(varexp))
        svartitle = _dataSpecification->getAttribute(varexp)->getLegend();

    if (svar.CountChar(':') > 0)
    {
        cout << "Error in TDataServer::drawCDFRaFu MultiVariates [" << svar
                << "] is not yet possible" << endl;
        return;
    }

    if (_datatree)
    {
        TString sOption = TString(option);
        sOption += "goff";

	checkCanvasCreation(option);
		
        TMultiGraph *mg = new TMultiGraph();
	_drawingGarbageCollector->Add(mg);
        mg->SetTitle(Form("Attribute %s : CDF via RaFu ", svartitle.Data()));

        // case des Lower values :
        gDirectory->Delete("hcdfLower");
        _datatree->Draw(Form("%s>>hcdfLower", varexp),
                Form("%s%s<0.0", GetName(), SINTERNALRAFU), sOption);
        // case des Upper values :
        gDirectory->Delete("hcdfUpper");
        _datatree->Draw(Form("%s>>hcdfUpper", varexp),
                Form("%s%s>0.0", GetName(), SINTERNALRAFU), sOption);

        TH1F *hcdfLower = (TH1F*) gDirectory->Get("hcdfLower");
        TH1F *hcdfUpper = (TH1F*) gDirectory->Get("hcdfUpper");
        Int_t nBin = hcdfLower->GetNbinsX();
        Double_t dSumInf = hcdfLower->GetSumOfWeights();
        Double_t dSumSup = hcdfUpper->GetSumOfWeights();

        TGraph *grLower = new TGraph(nBin);
	_drawingGarbageCollector->Add(grLower);
        TGraph *grUpper = new TGraph(nBin);
	_drawingGarbageCollector->Add(grUpper);
        Double_t sumLower = 0;
        Double_t sumUpper = 0;
        for (Int_t i = 0; i < nBin; i++)
        {
            sumLower += hcdfLower->GetBinContent(i) / dSumInf;
            sumUpper += hcdfUpper->GetBinContent(i) / dSumSup;
            grLower->SetPoint(i, hcdfLower->GetBinCenter(i), sumLower);
            grUpper->SetPoint(i, hcdfUpper->GetBinCenter(i), sumUpper);
        }

        // Parameters Lower case
        grLower->SetLineColor(8);
        grLower->SetLineWidth(1504);
        grLower->SetFillStyle(3002);
        grLower->SetFillColor(2);
        // Parameters Upper case
        grUpper->SetLineColor(8);
        grUpper->SetLineWidth(-802);
        grUpper->SetFillStyle(3002);
        grUpper->SetFillColor(2);

        mg->Add(grLower);
        mg->Add(grUpper);
        mg->Draw("AC");
        mg->GetXaxis()->SetTitle(svartitle);
        mg->GetYaxis()->SetTitle("Proba");

        gPad->Modified();
        gPad->Update();
        // Clean
        gDirectory->Delete("hcdfLower");
        gDirectory->Delete("hcdfUpper");

    }
    else
    {
        cout << "TDataServer::drawCDFRaFu: Error pas de tree" << endl;
    }

}

void TDataServer::drawBoxPlot(const char* varexp, const char* selection,
        Option_t* option)
{
    try
    {
        if (_datatree == NULL)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TDataServer::drawBoxPlot Error : pas de tree!");

	checkCanvasCreation(option);
	stringstream sstr; 	
        TString sOption = TString(option);
        sOption += "goff";

            TString svar = TString(varexp);
            try
            {
                if (svar.CountChar(':') > 0)
                    throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                                                               Form(
                                                                   "TDataServer::drawBoxPlot Error : drawBoxPlot MultiVariates [%s] is not yet possible!",
                                                                   svar.Data()));
            } catch (URANIE::Exceptions::UErrorExceptions& ue)
            {
                ue.printMessage();
                throw ue;
            }

        _datatree->Draw(Form("%s>>htemp", varexp),
			(const char *) (_ccut + TCut(selection)), sOption.Data());

	TH1F *htemp = (TH1F*)gDirectory->Get("htemp"); htemp->SetDirectory(0);
	sstr.str(""); sstr << "__tdsboxplot__" <<  _drawingGarbageCollector->GetEntries() ;	  
	htemp->SetName( sstr.str().c_str() );
	_drawingGarbageCollector->Add(htemp);
        THistPainter* histpainter = (THistPainter*) (htemp->GetPainter());
        histpainter->PaintTitle();
	
        TAxis* xaxis = htemp->GetXaxis();

        Double_t xmin = xaxis->GetXmin();
        Double_t xmax = xaxis->GetXmax();

            computeStatistic(varexp);
            Double_t Realmin = getAttribute(varexp)->getMinimum();
            Double_t Realmax = getAttribute(varexp)->getMaximum();
        //Double_t deltax = xmax - xmin;

            Double_t ymin = htemp->GetMinimum();
            Double_t ymax = htemp->GetMaximum();
            if( !sOption.Contains("same") )
            {
        TGaxis* gxaxis = new TGaxis(xmin, 0.1, xmax, 0.1, xmin, xmax);
	_drawingGarbageCollector->Add(gxaxis);
        gxaxis->ImportAxisAttributes(xaxis);
        gxaxis->SetName(varexp);
        gxaxis->Draw();
            }

        // Quantiles
        const Int_t nq = 3;
        Double_t * xq = new Double_t[nq]; // position where to compute the 3 quantiles in 0.25, 0.50, 0.75
        Double_t * yq = new Double_t[nq];  // array to contain the quantiles
        xq[0] = 0.25; yq[0] = 0.0;
        xq[1] = 0.50; yq[1] = 0.0;
        xq[2] = 0.75; yq[2] = 0.0;
        htemp->GetQuantiles(nq, yq, xq);

        Double_t mean = htemp->GetMean();
        Double_t rms = htemp->GetRMS();

        if (_blog)
        {
            cout << "Min[" << xmin << "] MAX[" << xmax << "]" << endl;
            cout << "Mean[" << mean << "] RMS[" << rms << "]" << endl;
            cout << xq[0] << " : " << yq[0] << endl;
            cout << xq[1] << " : " << yq[1] << endl;
            cout << xq[2] << " : " << yq[2] << endl;
        }

            if( !sOption.Contains("same") )
        gPad->Range(xmin, 0, xmax, 1);

            Double_t box_lower = ((sOption.Contains("same")) ?  (ymax-ymin)*0.2+ymin : 0.4);
            Double_t box_upper = ((sOption.Contains("same")) ?  (ymax-ymin)*0.01+ymin : 0.2);
        Double_t box_middle = 0.5 * (box_lower + box_upper);

        // Box
        TPave *boxquartile = new TPave(yq[0], box_lower, yq[2], box_upper, 1);
	_drawingGarbageCollector->Add(boxquartile);
        boxquartile->SetFillColor(10);
        boxquartile->SetLineWidth(2);
        boxquartile->Draw();
        // Mean
        TLine *line_mean = new TLine(mean, box_lower, mean, box_upper);
	_drawingGarbageCollector->Add(line_mean);
        line_mean->SetLineColor(kGreen);
        line_mean->SetLineWidth(3);
        line_mean->Draw();
        // Median
        TLine *line_mediane = new TLine(yq[1], box_lower, yq[1], box_upper);
	_drawingGarbageCollector->Add(line_mediane);
        line_mediane->SetLineColor(1);
        line_mediane->SetLineWidth(3);
        line_mediane->Draw();

        if (rms > 0.00001)
        {
            Double_t dalpha = 3.0;
            //Arrow inf
                Double_t min_arrow = TMath::Max(mean - dalpha * rms, Realmin);
            TArrow *left_arrow = new TArrow(min_arrow, box_middle, yq[0],
                    box_middle, 0.05, "<");
	    _drawingGarbageCollector->Add(left_arrow);
            left_arrow->SetLineWidth(2);
            left_arrow->Draw();
            //Arrow sup
                Double_t max_arrow = TMath::Min(mean + dalpha * rms, Realmax);
            TArrow *right_arrow = new TArrow(yq[2], box_middle, max_arrow,
                    box_middle, 0.05, ">");
	    _drawingGarbageCollector->Add(right_arrow);
            right_arrow->SetLineWidth(2);
            right_arrow->Draw();

            // Lower than Minimal Value
            _datatree->Draw(Form("%s>>hmin", varexp),
                    Form("%s<%g", varexp, min_arrow), "goff");
	    TH1F *hmin = (TH1F*)gDirectory->Get("hmin"); hmin->SetDirectory(0);
            const Int_t nPatInf = (Int_t)(hmin->GetSumOfWeights());
            if (nPatInf > 0)
            {
                Double_t *xminimum = new Double_t[nPatInf];
                xminimum = _datatree->GetV1();
                for (Int_t i = 0; i < nPatInf; i++)
                {
                    TMarker *mark = new TMarker(xminimum[i], box_middle, 7);
		    _drawingGarbageCollector->Add(mark);
                    mark->Draw();
                }
                delete[] xminimum;
            }
	    delete hmin;

            // Upper than Maximal Value
            _datatree->Draw(Form("%s>>hmax", varexp),
                    Form("%s>%g", varexp, max_arrow), "goff");
	    TH1F *hmax = (TH1F*)gDirectory->Get("hmax"); hmax->SetDirectory(0);
            const Int_t nPatSup = (Int_t)(hmax->GetSumOfWeights());
            if (nPatSup > 0)
            {
                Double_t *xmaximum = new Double_t[nPatSup];
                xmaximum = _datatree->GetV1();
                for (Int_t i = 0; i < nPatSup; i++)
                {
                    TMarker *mark = new TMarker(xmaximum[i], box_middle, 7);
		    _drawingGarbageCollector->Add(mark);
                    mark->Draw();
                }
                delete[] xmaximum;
            }
            // Clean
	    delete hmax;

        }

	delete[] xq;
	delete[] yq;
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    }
}

void TDataServer::drawTufte(const char *varexp, const char* selection,
        Option_t* option)
{
    try
    {
        if (!_datatree)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TDataServer::drawTufte Error : drawTufte pas de tree!");
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    }

    TString svar = TString(varexp);
    if (svar.CountChar(':') != 1)
    {
        cout
                << "Error in TDataServer::drawTufte exist if two attributs are defined ["
                << svar << "]" << endl;
        return;
    }

    Ssiz_t npoint = svar.Index(":");

    TString svary = svar(0, npoint);
    TString svarx = svar(npoint + 1, svar.Length());

    checkCanvasCreation(option);

    TPad *pad1 = new TPad("pad1", "pad1", 0.25, 0.025, 0.99, 0.24);
    pad1->SetBottomMargin(0.2);
    TPad *pad2 = new TPad("pad2", "pad2", 0.01, 0.24, 0.255, 0.93);
    pad2->SetLeftMargin(0.2);
    TPad *pad3 = new TPad("pad3", "pad3", 0.25, 0.25, 0.99, 0.93);
    pad3->Draw();
    pad1->Draw();
    pad2->Draw();
    
    TPaveLabel *title = new TPaveLabel(0.40, 0.94, 0.60, 0.98,
            Form("Tufte graph %s", varexp));
    _drawingGarbageCollector->Add(title);
    title->SetFillColor(0);
    title->SetTextFont(42);
    title->SetBorderSize(0);
    title->Draw();

    // Store the boolean of GetOptStat from gStyle
    Bool_t bOptStat = gStyle->GetOptStat();

    // Gestion des options
    TString soption = TString(option);
    soption.ToLower();

    if (soption.Contains("optstat"))
        gStyle->SetOptStat(1);
    else
        gStyle->SetOptStat(0);

    Bool_t bscatt = kFALSE;
    if (soption.Contains("scatter"))
        bscatt = kTRUE;

    // Histogram of x
    pad1->cd();
    draw(Form("%s", svarx.Data()), selection, "bar", kTRUE);
    TH1F *htemp = (TH1F*) gPad->GetPrimitive("htemp");
    Int_t nxbins = htemp->GetXaxis()->GetNbins();
    htemp->SetTitle("");
    htemp->SetLabelSize(0.00, "Y");
    htemp->SetLabelSize(0.085, "X"); htemp->SetTitleSize(0.095, "X"); htemp->SetTitleOffset(1.05, "X");
    htemp->SetLineColor(getTuple()->GetMarkerColor());
    htemp->SetFillColor(getTuple()->GetMarkerColor());
    htemp->SetFillStyle(1001);
    htemp->DrawNormalized("bar", 1.0);

    // Histogram of y
    pad2->cd();
    draw(Form("%s", svary.Data()), selection, "hbar", kTRUE);
    htemp = (TH1F*) gPad->GetPrimitive("htemp");
    Int_t nybins = htemp->GetXaxis()->GetNbins();
    htemp->SetTitle("");
    htemp->SetLabelSize(0.00, "Y");
    htemp->SetLabelSize(0.055, "X"); htemp->SetTitleSize(0.06, "X"); htemp->SetTitleOffset(1.55 , "X");
    htemp->SetLineColor(getTuple()->GetMarkerColor());
    htemp->SetFillColor(getTuple()->GetMarkerColor());
    htemp->SetFillStyle(1001);
    htemp->DrawNormalized("hbar", 1.0);
    

    // Nuage de point
    pad3->cd();
    draw(Form("%s:%s>>htemp(%d,,,%d)", svary.Data(), svarx.Data(), nxbins,
	      nybins), selection,"", kTRUE);

    if (bscatt)
    {
        draw(Form("%s:%s", svary.Data(), svarx.Data()), selection, "contzsame",kTRUE);
        draw(Form("%s:%s", svary.Data(), svarx.Data()), selection, "psame",kTRUE);
    }

    drawLegend(varexp, selection);

    htemp = (TH1F*) gPad->GetPrimitive("htemp");
    htemp->SetTitle("");

    pad3->Modified("");
    pad3->Update();
    gStyle->SetOptStat(bOptStat);

}

void TDataServer::drawPairs(const char *varexp, const char* selection,
        Option_t* option)
{
    try
    {
        if (!_datatree)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TDataServer::drawPairs[%s]  Error :\n The TTree is NULL",
                            GetName()));
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    }

    TList *listOfAttributes = getListOfAttributes(varexp);
    Int_t nX = listOfAttributes->GetSize();
    
    checkCanvasCreation(option);
    _canvas->Divide(nX, nX);

    Int_t npad = 1;
    for (Int_t i = 0; i < nX; i++)
    {
        for (Int_t j = 0; j < nX; j++)
        {
            _canvas->cd(npad++);
            if (i == j)
                draw(Form("%s", listOfAttributes->At(i)->GetName()), selection,
                        option, kTRUE);
            else
                draw(Form("%s:%s", listOfAttributes->At(i)->GetName(),
			  listOfAttributes->At(j)->GetName()), selection,
		     option, kTRUE);
        }
    }
    _canvas->Modified("");
    _canvas->Update();

    delete listOfAttributes;
}

void TDataServer::drawScatterplot(const char *varexp, const char* selection,
        Option_t* option)
{

    try
    {
        if (!_datatree)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TDataServer::drawScatterplot Error : drawScatterplot pas de tree!");
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    }
    
    checkCanvasCreation(option);
    
    TString svar = TString(varexp);
    if (svar.CountChar(':') != 1)
    {
        cout
                << "Error in TDataServer::drawScatterplot exist if 2 attributs must be defined ["
                << svar << "]" << endl;
        return;
    }

    Ssiz_t npoint = svar.Index(":");

    TString svary = svar(0, npoint);
    TString svarx = svar(npoint + 1, svar.Length());

    gStyle->SetPalette(1);
    draw(Form("%s:%s", svary.Data(), svarx.Data()), selection,
	 TString("contz"), kTRUE);
    draw(Form("%s:%s", svary.Data(), svarx.Data()), selection,
	 TString("psame"), kTRUE);

}
void TDataServer::drawCobWeb(const char *varexp, const char* selection, Option_t* option)
{
    checkCanvasCreation(option);
    
    draw(varexp, selection, Form("%s,para", option), kTRUE);
    drawLegend(varexp, selection, "ParaCoord");

    TParallelCoord* para = (TParallelCoord*) gPad->GetListOfPrimitives()->FindObject("ParaCoord");
    stringstream sstr; sstr << "__tdspara__" << _drawingGarbageCollector->GetEntries() ;
    if (para)
    	para->SetName(sstr.str().c_str());
}

void TDataServer::drawProfile(const char *varexp, const char* selection,
        Option_t* option)
{
    TString svar = TString(varexp);
    try {
    if (svar.CountChar(':') != 1)
    {
        throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                Form(
                        "TDataServer::drawProfile [%s]  only two attributes must be specified [%s]",
                        GetName(), varexp));
    }
    }catch(URANIE::Exceptions::UErrorExceptions &ue)
    {
        ue.printMessage();
        throw ue;
    }
    TString sprofname = TString("gURANIE_drawProfile");
    sprofname += TString("__") + GetName() + TString("__");

    Ssiz_t npoint = svar.Index(":");

    TString svary = svar(0, npoint);
    TString svarx = svar(npoint + 1, svar.Length());
    sprofname += svary + TString("__Versus__") + svarx;

    checkCanvasCreation(option);
    
    // Gestion des options
    TString soption = TString(option);
    soption.ToLower();
    if (soption.Contains("same"))
    {
        draw(Form("%s:%s", svary.Data(), svarx.Data()), ((strcmp(selection,"")==0)?"1>0":selection),"", kTRUE);
    }

    Int_t nBins = 100;
    if (soption.Contains("nclass="))
    {
        TRegexp reg("nclass=[0-9]*");
        TString out = soption(reg);
        TString snclass = out.Remove(0, 7);
        nBins = atoi(snclass.Data());
    }
    nBins += 2;

    // Compute the statistics for the input and the output attributes
    computeStatistic(varexp);

    Double_t dxmin = getAttribute(svarx)->getMinimum();
    Double_t dxmax = getAttribute(svarx)->getMaximum();
    Double_t dxcoeff = 0.01;
    Double_t dxrange = dxmax - dxmin;
    Double_t dymin = getAttribute(svary)->getMinimum();
    Double_t dymax = getAttribute(svary)->getMaximum();

    if (_blog)
        cout << " ** dxmin[" << dxmin << "] dxmax[" << dxmax << "] dxrange["
                << dxrange << "] dxcoeff[" << dxcoeff << "] *** dymin[" << dymin
                << "] dymax[" << dymax << "]" << endl;
    dxmin -= dxcoeff * dxrange;
    dxmax += dxcoeff * dxrange;

    draw( Form("%s:%s>>%s(%d,%e,%e)", svary.Data(), svarx.Data(),
	       sprofname.Data(), nBins, dxmin, dxmax), ((strcmp(selection,"")==0)?"1>0":selection),
	  "goffprofs",kTRUE);
    
    TProfile *hprofs = (TProfile*) gDirectory->Get(sprofname);
    hprofs->SetDirectory(0);
    hprofs->SetMarkerColor(kBlue);
    hprofs->SetMarkerSize(0.7);
    hprofs->SetMarkerStyle(21);
    if (soption.Contains("same"))
    {
        hprofs->Draw("same");
        drawLegend(varexp, selection);
	stringstream sstr;
	TH1 *htemp = (TH1*)gPad->GetPrimitive("htemp");
	if(htemp!=NULL)
	{
	    sstr << "__tdshisto__" << _drawingGarbageCollector->GetEntries() ;
	    htemp->SetName(sstr.str().c_str());
	}
    }
    else
    {
	hprofs->Draw();
        TString stitle = Form("Profile %s (Bin = %d )", varexp, nBins);
        hprofs->SetName(stitle);
        stitle += TString(";");
        if (_dataSpecification->isAttribute(svarx))
            stitle += _dataSpecification->getAttribute(svarx)->getLegend();
        else
            stitle += svarx;
        stitle += TString(";");
        if (_dataSpecification->isAttribute(svary))
            stitle += _dataSpecification->getAttribute(svary)->getLegend();
        else
            stitle += svary;

        hprofs->SetTitle(stitle);
    }
    _drawingGarbageCollector->Add(hprofs);
    
    if (_blog)
    {
        hprofs->ls();
        hprofs->Print();
    }

}

void TDataServer::bootstrap(Int_t nsize, Option_t *option)
{
    TString soption = TString(option);
    soption.ToLower();
    soption.ReplaceAll(" ", "");
    if (_blog)
    {
        cout << "********************************" << endl;
        cout << " *** TDataServer::Bootstrap tds[" << GetName() << "] nBoot["
                << nsize << "] soption[" << soption << "]" << endl;
    }

    Int_t ns = nsize;

    if (ns == 0)
    {
        _datatree->SetEventList(0);
        if (_pelBootstrap)
            delete _pelBootstrap;
        _pelBootstrap = NULL;
    }
    else
    {
        if (_pelBootstrap == NULL)
        {
            if (ns < 0)
                ns = getNPatterns();
            if (_blog)
                cout << " ** create TPatternsEventList :"
                        << Form("bootstrap [%d, %d]", ns, getNPatterns())
                        << endl;
            UInt_t nseed = 0;
            if (soption.Contains("seed="))
            {
                TRegexp reg("seed=[0-9]*");
                TString out = soption(reg);
                TString snclass = out.Remove(0, 5);
                if (_blog)
                    cout << " ** soption[" << soption << "] snclass[" << snclass
                            << "]" << endl;
                nseed = atoi(snclass.Data());
            }
            gRandom->SetSeed(nseed);
            _pelBootstrap = new TPatternsEventList("pelbootstrap",
                    Form("bootstrap [%d, %d]", ns, getNPatterns()), ns);
        }
        else
        {
            _pelBootstrap->Reset();
        }
        ns = _pelBootstrap->GetSize();
        if (_blog)
            cout << " ** TDataServer::Bootstrap ns[" << ns
                    << "] getNPatterns()[" << getNPatterns() << "]" << endl;

        // Loop on each cell of the array
        for (Int_t i = 0; i < ns; i++)
        {
            // Tirage aleatoire d'un nombre entier entre 0 et ns
            UInt_t ther = gRandom->Integer(ns);
            if (_blog)
                cout << " ** index i[" << i << "/" << ns << "] value [" << ther
                        << "] " << endl;
            _pelBootstrap->Enter(ther);
        }
        if (soption.Contains("sort"))
            _pelBootstrap->Sort();
        if (_blog)
            _pelBootstrap->Print("all");

        _datatree->SetEventList(_pelBootstrap);
    }
    if (_blog)
    {
        cout << " *** End Of TDataServer::Bootstrap tds[" << GetName()
                << "] nBoot[" << nsize << "]" << endl;
        cout << "********************************" << endl;
    }
}

Int_t TDataServer::scan(const char *varexp, const char *selection,
        Option_t *option)
{
    try
    {
        if (!_datatree)
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form("TDataServer::scan[%s]  ERROR The TTree is NULL",
                            GetName()));
        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    }

    TCut sselect = _ccut + TCut(selection);
    return Int_t(_datatree->Scan(varexp, (const char *) sselect, option));
}

void TDataServer::startViewer()
{
    try
    {
        if (!_datatree)
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    "TDataServer::startViewer Error : startViewer pas de tree!");
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    }

    if (_datatree != NULL)
    {
        _datatree->StartViewer();
    }
    else
    {
        cout << "The TNtuple is NULL" << endl;
    }
}

void TDataServer::drawIter(const char* sVar, const char* selection,
        Option_t* option)
{
    TCut sselect = _ccut + TCut(selection);
    try
    {
        if (_datatree)
        {
            _datatree->Draw(Form("%s:%s", sVar, getIteratorName()), sselect,
                    option);
            drawLegend(Form("%s:%s", sVar, getIteratorName()), sselect);
        }
        else
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form("TDataServer::drawLegend Error : There is no tree."));

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    }
}

void TDataServer::drawLegend(const char* varexp, const char* selection, const char * histname)
{
    try
    {
        if (!gPad)
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form("TDataServer::drawLegend Error : the gPad is empty."));
        }
        TH1D *htemp = (TH1D*) gPad->GetPrimitive(histname);

        if (htemp)
        {

            TString sLabel = TString(varexp);

            // Titre du graphique
            TString stitle;
            Int_t ncpt = sLabel.CountChar(':');

            switch (ncpt)
            {
            case 0: // case of 1D
            {
                stitle = Form("Histogram %s", sLabel.Data());
                break;
            }
            case 1: // case of y
            {
                stitle = Form("Scatterplot %s          ", sLabel.Data());
                break;
            }
            case 2: // case of z
            {
                stitle = Form("Scatterplot 3D %s", sLabel.Data());
                break;
            }
            default:
            {
                stitle = Form("more than 3 variables ( %s )", sLabel.Data());
                break;
            }
            }

            // La selection de pattern
            TCut sselect = _ccut + TCut(selection);
            TString sSelection = TString((const char*) sselect);
            if (sSelection.Length() > 0)
                stitle += Form(" { %s } ", sSelection.Data());

            // Affichage des titres des axes
            Int_t nend = sLabel.Length();
            do
            {
                ncpt = sLabel.Last(':');

                TString svar = sLabel(ncpt + 1, nend);
                if (_dataSpecification->isAttribute(svar))
                    svar = _dataSpecification->getAttribute(svar)->getLegend();

                stitle += Form(";%s", svar.Data());

                nend = ncpt;
                if (ncpt != -1)
                    sLabel.Remove(ncpt);

            } while (nend != -1);

            htemp->SetTitle(stitle);

            gPad->Modified();
            gPad->Update();
        }
        else
        {
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form(
                            "TDataServer::drawLegend Error : The histogram htemp is empty."));

        }
    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);

    }
}

void TDataServer::drawHistogram(const char* sVar, const char* selection,
        Option_t* option)
{
    try
    {
        if (_datatree)
        {
            _datatree->Draw(sVar, (const char *) (_ccut + TCut(selection)),
                    option);
            drawLegend(sVar, selection);
        }
        else
            throw URANIE::Exceptions::UErrorExceptions(__FILE__, __LINE__,
                    Form("TDataServer::drawLegend Error : There is no tree."));

    } catch (URANIE::Exceptions::UErrorExceptions& ue)
    {
        ue.printMessage();
        throw ue;
        exit(-1);
    }
}

///////////////////////////////////////////////////////////
//  Printing Log
///////////////////////////////////////////////////////////
void TDataServer::printLog(Option_t *option)
{
    cout << "TDataServer::printLog[" << option << "]" << endl;
    cout << "Name[" << GetName() << "] Title[" << GetTitle() << "]" << endl;
    cout << "Origin[";
    switch (_norigin)
    {
    case kASCIIFile:
    {
        cout << "ASCII File";
        break;
    }
    case kTree:
    {
        cout << "TTree";
        break;
    }
    default:
        cout << "Unknown";
    }
    cout << "]" << endl;
    //  cout << "_sMethodName[" <<_sMethodName << "]" << endl; 
    cout << " _sdatafile[" << _sdatafile << "] " << endl << " _sarchivefile["
            << _sarchivefile << "]" << endl;
    if (_dataSpecification)
        _dataSpecification->printLog();

    if (_datatree)
        _datatree->Print();
    cout << "fin de TDataServer::printLog[" << option << "]" << endl;
}

//////////////////////////////////////////////////////////////////////////////////////////////
//  Recuperation de donnees basiques du dataTree ou de fichiers de sauvegarde root du dataTree        
//////////////////////////////////////////////////////////////////////////////////////////////
char * TDataServer::getDataFileName()
{

    return Form("%s", _sdatafile.Data());
}
char * TDataServer::getArchiveFileName()
{

    return Form("%s", _sarchivefile.Data());
}

char * TDataServer::getDataTreeName()
{
    //Initialize the _shortDataTreeName
    _shortDataTreeName = GetName();
    if (_shortDataTreeName.Length() >= MAXNBCHARACTERS)
        _shortDataTreeName = _shortDataTreeName(0, MAXNBCHARACTERS);
    return Form("%s__data__tree__", _shortDataTreeName.Data());
}

TDSNtupleD *TDataServer::getTuple(Option_t * option)
{
    TString soption = option;
    soption.ToLower();
    try
    {
        if (!_datatree)
            throw URANIE::Exceptions::UWarningExceptions(__FILE__, __LINE__,
                    "TDataServer::getTuple Error : There is no tree!");
    } catch (URANIE::Exceptions::UWarningExceptions& ue)
    {
        ue.printMessage();
        throw ue;
    }
    return _datatree;
}

 void TDataServer::deleteTuple()
 {
     if(_datatree != NULL)
     {
         delete _datatree;
         _datatree = NULL;
     }
     
     clearSelect();
     
 }

// -----------------------------------------------------------------------
//  Return the parameters of the law contained in the string on the vector
// -----------------------------------------------------------------------
void TDataServer::findParametersLaw(TString soption, vector<double> &vec, bool blog)
{
    vec.clear();
    if (blog)
        cout << endl << " ***********************" << endl << " *** soption["
             << soption << "]" << endl << " ***********************" << endl;
    
    Int_t nparams = soption.CountChar(',') + 1;
    vec.resize(nparams);
    soption.ReplaceAll(")", ",").ReplaceAll("(", "");

    Ssiz_t nbegin = 0;
    for (Int_t i = 0; i < nparams; i++)
    {
        Ssiz_t nend = soption.Index(',', nbegin);
        TString scrt = soption(nbegin, nend);
        vec.at(i) = atof(scrt.Data());
	  
        nbegin = nend + 1;
    }

}

  
} // Fin du namespace DataServer
} // Fin du namespace URANIE

