//////////////////////////////////////////////////////////////
// Copyright (C) 2013-2022 CEA/DES
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation, either version 3 of the License, or any
// later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
// $Id$
// $Author$
// $Date$
// $Revision$
// $State$
//////////////////////////////////////////////////////////////
/*!
  \file   TestTLauncher.cxx
  \author Fabrice Gaudier <fabrice.gaudier@cea.fr>
  \date   $Date$
 
  \brief  Fichier de test de la classe URANIE::Launcher::TLauncher
 
  Cree le : mer nov 21 07:59:24 2007

  \par Modifications
  \li JB Blanchard (CEA) 2016-02-11
  put the "nointermed" option in run command to speed up tests
  \li S. Pomarede (InckA) 2013-07-03
  Creation of 'areFilesEqual' function to test if files are equal whatever the platform and encoding are.
  \li S. Pomarede (InckA) 2012-11-15
  Addition of tests for TOutputFileXML class
  \li S. Pomarede (InckA) 2012-08-01
  Add two functions to test flowrate ability to use XML input files
  \li S. Pomarede (InckA) 2012-05-04
  Add setDrawProgressBar(kFALSE) to the TLauncher not to print the progress bars during tests
  \li M. Salvoldi (Areva/OCC) 2011/09/28
  - add,in cleanWorkingDirectories(), testNewKey(), testNewLine() and testNewColumn(), IFDEF WIN32 to replace rm and diff by del and fc on Windows platform. 
  \li F. Gaudier (2010-12-14 11:26:42)
  add the assert in the testLauncherAttributeWithSeveralInputFiles TU to verify the number of leaves in the TNtupleD database (DI111)
  \li CFO (2010-12-07)
  Correction of the test throwExceptiontestLauncherOutputFileKey(), reference file was not correct.  
  \li F. Gaudier (2010-08-06 09:13:26)
  Adds the 6 tests of the TOuputFileKey classe with lanching the flowrate code on the uniform design (32 patterns)
  Change the name of the datasets using in the tests
  Change the setKey by the setFileKey method
  \li Cedric Forgues (C.S.) 2010/04/12
  Addings of unitary tests testLauncherWithNoSave(), testLauncherWithCurrentDirectorySpecifiedAsWorkingDirectory(), testLauncherWithNoSpecifiedWorkingDirectory(), testLauncherWithNoSpecifiedWorkingDirectory()
  \li Cedric Forgues (C.S.) 2010/03/29
  TEMPORARY: adds relative path for flowrate executable.  
  \li Cedric Forgues (C.S.) 2010/03/18
  Adds methods testLauncherHighNbSamplesWithoutSave(), testLauncherHighNbSamplesWithSave1(), testLauncherHighNbSamplesWithSave2(), testLauncherHighNbSamplesWithSave3().
  \li Cedric Forgues (C.S.) 2010/03/02
  Adds methods testLauncherFunction3()  
  \li Cedric Forgues (C.S.) 2009/12/15
  Adds methods testWorkingdirectory()  
  \li Cedric Forgues (C.S.) 2009/12/04
  Adds methods testSearchIndexAttributeRecreate
  \li Cedric Forgues (C.S.) 2009/12/03
  Adds methods TestNewKey, TestNewLine, TestNewColumn  
  // ***************************************************************
 
  */
//CPPUNIT
//! A true double precision to compare (FGA)
#define DBL_DBL_PRECISION_COMPARE 1.e-10

// Uranie
#include "TestTLauncher.h"
#include "TAttribute.h"
#include "TUniformDistribution.h"
#include "TBasicSampling.h"
#include "TLauncherFunction.h"
#include "TestUserFunctions.h"
#include "TUniformDistribution.h"
#include "UDisableTest.h"
#include "code_paths.h"
#include "uranie_commands.h"

//! The dataset Uniform Design with 32 samples for the flowrate external code
#define FLOWRATEASCIIFILEUNIFORMDESIGN "flowrateUniformDesign.dat"
#define RESFLOWRATEASCIIFILEUNIFORMDESIGN "_run_flowrateUniformDesign.dat.ref"
//! A dataset  with 1200 samples (LHS) for the flowrate external code
#define FLOWRATEASCIIFILE1200SAMPLES "flowrate1200Samples.dat"
//! the referebce dataset for levelE usecase
#define LEVELEDATASET "levelE_database_reference.dat"
// ROOT
#include "TString.h"
#include "TMath.h"
#include "TStopwatch.h"

#include <iostream>
using namespace std;

TLauncherTest::TLauncherTest() :
    tds1(NULL), tc1(NULL), tl1(NULL), tds2(NULL), tc2(NULL), tl2(NULL), tds3(NULL),
    tc3(NULL), tl3(NULL), tds4(NULL), tc4(NULL), tl4(NULL), tlbs(NULL),
    _foutCol(NULL), _foutRow(NULL), _foutKey(NULL), _foutDataServer(NULL),
#ifdef WITH_FMI
    _foutFMI1(nullptr), _foutFMI2(nullptr),
    _tcFMI1(nullptr), _tcFMI2(nullptr),
#endif

#ifdef WITH_LIBXML2
    _foutXML(NULL),
#endif
    ptrHat(NULL), ptrD(NULL), tifr(NULL), _bLog(kFALSE)
{
    _workingTestPath = gSystem->pwd();
}

TLauncherTest::~TLauncherTest()
{
}

void TLauncherTest::cleanWorkingDirectories(const char *option)
{
    if (_bLog)
    {
        cout << endl;
        cout << " ** TLauncherTest::cleanWorkingDirectories option[" << option
	     << "] _workingTestPath[" << _workingTestPath << "]" << endl;
    }
    UranieTests::removeFolder(_workingTestPath.Data(),"URANIE");
    UranieTests::removeFolder(_workingTestPath.Data(),"tmpLanceurUranie");
    UranieTests::removeFile(_workingTestPath.Data(),"_local_run_flowrateUniforDesign.dat");
    UranieTests::removeFile(_workingTestPath.Data(),"__Uranie_Tmp__.bsub");
    UranieTests::removeFile(_workingTestPath.Data(),"__Uranie_Tmp__.log");
    UranieTests::removeFile(_workingTestPath.Data(), "_outputFileKey_*.dat");
    UranieTests::removeFile(_workingTestPath.Data(),"TestFileColumn.in");
    UranieTests::removeFile(_workingTestPath.Data(),"TestFileKey.in");
    UranieTests::removeFile(_workingTestPath.Data(),"TestFileLine.in");
    UranieTests::removeFile(_workingTestPath.Data(),"uranie_substitute_xmlfile_Test_flowrate_input_with_xml.in.xslt");
    UranieTests::removeFile(_workingTestPath.Data(),"_outputFile_levelE_cols_bystep_.dat");
    UranieTests::removeFile(_workingTestPath.Data(),"_outputFile_levelE_cols_.dat");
    UranieTests::removeFile(_workingTestPath.Data(),"_outputFile_levelE_keys_.dat");
    UranieTests::removeFile(_workingTestPath.Data(),"_outputFile_levelE_rows_.dat");
    UranieTests::removeFile(_workingTestPath.Data(),"_outputFile_levelE_TDS_.dat");
    UranieTests::removeFile(_workingTestPath.Data(),"_outputFile_levelE_XML_.dat");
    UranieTests::removeFile(_workingTestPath.Data(),"_output_flowrate_withKey_.dat");
    UranieTests::removeFile(_workingTestPath.Data(),"_output_flowrate_withRow_.dat");
    UranieTests::removeFile(_workingTestPath.Data(),"_output_flowrate_withXML_.dat");
    UranieTests::removeFile(_workingTestPath.Data(),"_output_multitype_mt_Column_.dat");
    UranieTests::removeFile(_workingTestPath.Data(),"_output_multitype_mt_Key_.dat");
    UranieTests::removeFile(_workingTestPath.Data(),"_output_multitype_mt_Key_condensate_.dat");
    UranieTests::removeFile(_workingTestPath.Data(),"_output_multitype_mt_Row_.dat");
    UranieTests::removeFile(_workingTestPath.Data(),"_output_multitype_mt_DataServer_.dat");
    UranieTests::removeFile(_workingTestPath.Data(),"_output_multitype_mt_.xml");
    UranieTests::removeFile(_workingTestPath.Data(),"_output_multitype_readmt_Column_.dat");
    UranieTests::removeFile(_workingTestPath.Data(),"_output_multitype_readmt_Key_.dat");
    UranieTests::removeFile(_workingTestPath.Data(),"_output_multitype_readmt_Row_.dat");
    UranieTests::removeFile(_workingTestPath.Data(),"_output_multitype_readmt_DataServer_.dat");
    UranieTests::removeFile(_workingTestPath.Data(),"_output_multitype_readmt_.xml");
    UranieTests::removeFile(_workingTestPath.Data(),"readmultitype_sampling.root");
    UranieTests::removeFile(_workingTestPath.Data(),"levelE_database_reference.root");
    UranieTests::removeFile(_workingTestPath.Data(),"multitype_sampling.root");
    UranieTests::removeFile(_workingTestPath.Data(),"flowrate1200Samples.root");
    UranieTests::removeFile(_workingTestPath.Data(),"Exemples_For_Blank.root");
    UranieTests::removeFile(_workingTestPath.Data(),"flowrate200Samples.root");
    UranieTests::removeFile(_workingTestPath.Data(),"_dataserver_.root");
    UranieTests::removeFile(_workingTestPath.Data(),"readmultitype_sampling_1.root");
}

void TLauncherTest::cleanTestDirectories()
{
    UranieTests::removeFolder(_workingTestPath.Data(), TString("URANIE"));
    UranieTests::removeFolder(_workingTestPath.Data(), TString("tmpLanceurUranie"));
    UranieTests::removeFile(_workingTestPath.Data(), TString("UranieTmp.bsub"));
}

void TLauncherTest::setUp()
{
    if (_bLog)
    {
        cout << endl;
        cout
	    << "*******************************************************************************"
	    << endl;
    }
    gSystem->cd(_workingTestPath);
    cleanWorkingDirectories("setUp");

    tc1 = NULL;
    tc2 = NULL;
    tc3 = NULL;
    tc4 = NULL;
    tl1 = NULL;
    tl2 = NULL;
    tlbs = NULL;
    tl3 = NULL;
    tl4 = NULL;
    _foutCol = NULL;
    _foutRow = NULL;
    _foutKey = NULL;
    _foutDataServer = NULL;
#ifdef WITH_LIBXML2
    _foutXML = NULL;
#endif
    tds1 = NULL;
    tds2 = NULL;
    tds3 = NULL;
    tds4 = NULL;
    ptrHat = new URANIE::DataServer::TAttribute("yhat");
    ptrD = new URANIE::DataServer::TAttribute("d");
    tifr = NULL;

    if (_bLog)
        cout
	    << "*******************************************************************************"
	    << endl;
}

void TLauncherTest::tearDown()
{
    if (_bLog)
    {
        cout << endl;
        cout
	    << "*******************************************************************************"
	    << endl;
    }
    gSystem->cd(_workingTestPath);
    cleanWorkingDirectories("tearDown");

    // Clean
    if (tifr != NULL)
    {
        delete tifr;
        tifr = NULL;
    }

    if (tc1 != NULL)
    {
        delete tc1;
        tc1 = NULL;
    }
    if (tc2 != NULL)
    {
        delete tc2;
        tc2 = NULL;
    }
    if (tc3 != NULL)
    {
        delete tc3;
        tc3 = NULL;
    }
    if (tc4 != NULL)
    {
        delete tc4;
        tc4 = NULL;
    }

    if (tl1 != NULL)
    {
        delete tl1;
        tl1 = NULL;
    }
    if (tl2 != NULL)
    {
        delete tl2;
        tl2 = NULL;
    }
    if (tl3 != NULL)
    {
        delete tl3;
        tl3 = NULL;
    }
    if (tl4 != NULL)
    {
        delete tl4;
        tl4 = NULL;
    }
    if (tlbs != NULL)
    {
        delete tlbs;
        tlbs = NULL;
    }

    
    if (_foutCol != NULL)
    {
        delete _foutCol;
        _foutCol = NULL;
    }

    if (_foutRow != NULL)
    {
        delete _foutRow;
        _foutRow = NULL;
    }

    if (_foutKey != NULL)
    {
        delete _foutKey;
        _foutKey = NULL;
    }

    if (_foutDataServer != NULL)
    {
        delete _foutDataServer;
        _foutDataServer = NULL;
    }

#ifdef WITH_LIBXML2
    if( _foutXML != NULL )
    {
        delete _foutXML; _foutXML = NULL;
    }
#endif

    if (tds1 != NULL)
    {
        delete tds1;
        tds1 = NULL;
    }
    if (tds2 != NULL)
    {
        delete tds2;
        tds2 = NULL;
    }
    if (tds3 != NULL)
    {
        delete tds3;
        tds3 = NULL;
    }
    if (tds4 != NULL)
    {
        delete tds4;
        tds4 = NULL;
    }

    if (ptrHat != NULL)
    {
        delete ptrHat;
        ptrHat = NULL;
    }
    if (ptrD != NULL)
    {
        delete ptrD;
        ptrD = NULL;
    }

    if (_bLog)
        cout
	    << "*******************************************************************************"
	    << endl;
}

////////////////////////////////
///                          ///
///      Set up Methods      ///
///                          ///
////////////////////////////////

void TLauncherTest::setUp1(bool nodata)
{
    cout << endl << " ********************" << endl;
    cout << " ****" << endl;
    cout << " **** SetUP1" << endl;
    cout << " ****" << endl;
    tds1 = new URANIE::DataServer::TDataServer("tds1", "my title 1");
    if(!nodata)
        tds1->fileDataRead(FLOWRATEASCIIFILEUNIFORMDESIGN);
    else
    {
        // Add the study attributes ( min, max and nominal values)
        tds1->addAttribute( new URANIE::DataServer::TUniformDistribution("rw", 0.05, 0.15)); tds1->getAttribute("rw")->delShare();
        tds1->addAttribute( new URANIE::DataServer::TUniformDistribution("r",  100.0, 50000.0)); tds1->getAttribute("r")->delShare();
        tds1->addAttribute( new URANIE::DataServer::TUniformDistribution("tu", 63070.0, 115600.0)); tds1->getAttribute("tu")->delShare();
        tds1->addAttribute( new URANIE::DataServer::TUniformDistribution("tl", 63.1, 116.0)); tds1->getAttribute("tl")->delShare();
        tds1->addAttribute( new URANIE::DataServer::TUniformDistribution("hu", 990.0, 1110.0)); tds1->getAttribute("hu")->delShare();
        tds1->addAttribute( new URANIE::DataServer::TUniformDistribution("hl", 700.0, 820.0)); tds1->getAttribute("hl")->delShare();
        tds1->addAttribute( new URANIE::DataServer::TUniformDistribution("l",  1120.0, 1680.0)); tds1->getAttribute("l")->delShare();
        tds1->addAttribute( new URANIE::DataServer::TUniformDistribution("kw", 9855.0, 12045.0)); tds1->getAttribute("kw")->delShare();
    }
    // The reference input file
    TString sJDD1 = gSystem->pwd();
#ifdef WIN32
    sJDD1 += TString("\\flowrate_input_with_keys.in");
#else
    sJDD1 += TString("/flowrate_input_with_keys.in");
#endif

    tds1->getAttribute("rw")->setFileKey(sJDD1, "Rw", "%f");
    tds1->getAttribute("r")->setFileKey(sJDD1, "R", "%f");
    tds1->getAttribute("tu")->setFileKey(sJDD1, "Tu", "%f");
    tds1->getAttribute("tl")->setFileKey(sJDD1, "Tl", "%f");
    tds1->getAttribute("hu")->setFileKey(sJDD1, "Hu", "%f");
    tds1->getAttribute("hl")->setFileKey(sJDD1, "Hl", "%f");
    tds1->getAttribute("l")->setFileKey(sJDD1, "L", "%f");
    tds1->getAttribute("kw")->setFileKey(sJDD1, "Kw", "%f");

    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("flowrate") + " -s -k");
    cout << " ****" << endl;
    cout << " **** End Of SetUp1" << endl;
    cout << " ****" << endl;
    cout << " ********************" << endl;
}

void TLauncherTest::setUp2()
{
    cout << endl << " ********************" << endl;
    cout << " ****" << endl;
    cout << " **** SetUP2" << endl;
    cout << " ****" << endl;
    // The reference
    tds2 = new URANIE::DataServer::TDataServer();
    tds2->fileDataRead(FLOWRATEASCIIFILEUNIFORMDESIGN);

    //The reference input file for TInputFileRecreate
    TString sJDD2 = gSystem->pwd();
    //Give an empty inputfile
#ifdef WIN32
    sJDD2 += TString("\\flowrate_input_with_keys.in");
#else
    sJDD2 += TString("/flowrate_input_with_keys.in");
#endif

    tds2->getAttribute("rw")->setFileKey(sJDD2, "Rw", "%e",
					 URANIE::DataServer::TAttributeFileKey::kNewKey);
    tds2->getAttribute("r")->setFileKey(sJDD2, "R", "%e",
					URANIE::DataServer::TAttributeFileKey::kNewKey);
    tds2->getAttribute("tu")->setFileKey(sJDD2, "Tu", "%e",
					 URANIE::DataServer::TAttributeFileKey::kNewKey);
    tds2->getAttribute("tl")->setFileKey(sJDD2, "Tl", "%e",
					 URANIE::DataServer::TAttributeFileKey::kNewKey);
    tds2->getAttribute("hu")->setFileKey(sJDD2, "Hu", "%e",
					 URANIE::DataServer::TAttributeFileKey::kNewKey);
    tds2->getAttribute("hl")->setFileKey(sJDD2, "Hl", "%e",
					 URANIE::DataServer::TAttributeFileKey::kNewKey);
    tds2->getAttribute("l")->setFileKey(sJDD2, "L", "%e",
					URANIE::DataServer::TAttributeFileKey::kNewKey);
    tds2->getAttribute("kw")->setFileKey(sJDD2, "Kw", "%e",
					 URANIE::DataServer::TAttributeFileKey::kNewKey);

    _foutRow = new URANIE::Launcher::TOutputFileRow(
	"_output_flowrate_withRow_.dat");
    _foutRow->addAttribute(ptrHat);
#ifdef WIN32
    tc2 = new URANIE::Launcher::TCode(tds2, "flowrate.exe -s -k");
#else
    tc2 = new URANIE::Launcher::TCode(tds2,
				      Form("%s/%s", _workingTestPath.Data(),
					   "../../teST/coDE/flowREATE/flowrate -s -k"));
#endif
//   tc2->addInputFile(tifr);
    tc2->addOutputFile(_foutRow);
    tl2 = new URANIE::Launcher::TLauncher(tds2, tc2);
    tl2->setDrawProgressBar(kFALSE);
    cout << " ****" << endl;
    cout << " **** End Of SetUp2" << endl;
    cout << " ****" << endl;
    cout << " ********************" << endl;
}

void TLauncherTest::setUp3()
{
    // The reference
    tds3 = new URANIE::DataServer::TDataServer("tdsFlowrate",
					       "Design of experiments for Flowrate");
    tds3->fileDataRead(FLOWRATEASCIIFILE1200SAMPLES);
    TString sJDD3 = gSystem->pwd();
#ifdef WIN32
    sJDD3 += TString("\\flowrate_input_with_keys.in");
#else
    sJDD3 += TString("/flowrate_input_with_keys.in");
#endif

    // Sets the reference input file and the key for each input attributes
    tds3->getAttribute("rw")->setFileKey(sJDD3, "Rw");
    tds3->getAttribute("r")->setFileKey(sJDD3, "R");
    tds3->getAttribute("tu")->setFileKey(sJDD3, "Tu");
    tds3->getAttribute("tl")->setFileKey(sJDD3, "Tl");
    tds3->getAttribute("hu")->setFileKey(sJDD3, "Hu");
    tds3->getAttribute("hl")->setFileKey(sJDD3, "Hl");
    tds3->getAttribute("l")->setFileKey(sJDD3, "L");
    tds3->getAttribute("kw")->setFileKey(sJDD3, "Kw");

    _foutRow = new URANIE::Launcher::TOutputFileRow(
	"_output_flowrate_withRow_.dat");
    _foutRow->addAttribute(ptrHat);

#ifdef WIN32
    tc3 = new URANIE::Launcher::TCode(tds3, "flowrate.exe -s -k");
#else
    tc3 = new URANIE::Launcher::TCode(tds3,
				      Form("%s/%s", _workingTestPath.Data(),
					   "../../teST/coDE/flowREATE/flowrate -s -k"));
#endif
    tc3->addOutputFile(_foutRow);

    tl3 = new URANIE::Launcher::TLauncher(tds3, tc3);
    tl3->setDrawProgressBar(kFALSE);
#ifdef WIN32
    tl3->setWorkingDirectory(gSystem->pwd() + TString("\\tmpLanceurUranie\\flowrate"));
#else
    tl3->setWorkingDirectory(
	gSystem->pwd() + TString("/tmpLanceurUranie/flowrate"));
#endif
}

#ifdef WITH_LIBXML2
void TLauncherTest::setUp4()
{
    tds4 = new URANIE::DataServer::TDataServer("tds4", "my title 1");
    tds4->fileDataRead(FLOWRATEASCIIFILEUNIFORMDESIGN);

    TString sJDD = gSystem->pwd();
#ifdef WIN32
    sJDD += TString("\\flowrate_input_with_xml.in");
#else
    sJDD += TString("/flowrate_input_with_xml.in");
#endif
    tds4->getAttribute("rw")->setFileXMLAttribute(sJDD, "wall_friction/@rw");
    tds4->getAttribute("r")->setFileXMLAttribute(sJDD, "wall_friction/@r");
    tds4->getAttribute("tu")->setFileKey(sJDD, "parameter[@name='Tu']/value/double", "%e", URANIE::DataServer::TAttributeFileKey::kXMLField);
    tds4->getAttribute("tl")->setFileKey(sJDD, "parameter[@name='Tl']/value/double", "%e", URANIE::DataServer::TAttributeFileKey::kXMLField);
    tds4->getAttribute("hu")->setFileKey(sJDD, "parameter[tonode='mesher' and toport='dt_hu']/value/double", "%e", URANIE::DataServer::TAttributeFileKey::kXMLField);
    tds4->getAttribute("hl")->setFileKey(sJDD, "parameter[tonode='mesher' and toport='dt_hl']/value/double", "%e", URANIE::DataServer::TAttributeFileKey::kXMLField);
    tds4->getAttribute("l")->setFileXMLAttribute(sJDD, "parameter[@name='L']/@precision", "%e");
    tds4->getAttribute("kw")->setFileXMLAttribute(sJDD, "kW/@value");

    tc4 = new URANIE::Launcher::TCode(tds4, UranieTests::code_path("flowrate") + " -s -x");

    _foutXML = new URANIE::Launcher::TOutputFileXML("_output_flowrate_withXML_.dat");
    
}
#endif

void TLauncherTest::setUpMultiInput(TString type)
{
    cout << endl << " ********************" << endl;
    cout << " ****" << endl;
    cout << " **** SetUPMulti" << endl;
    cout << " ****" << endl;

    tds1 = new URANIE::DataServer::TDataServer("tds1", "my title 1");
    tds1->fileDataRead(FLOWRATEASCIIFILEUNIFORMDESIGN);
    // The reference input file
    TString sJDD1 = gSystem->pwd();
    TString sJDD2 = gSystem->pwd();

    
#ifdef WIN32
    sJDD1 += TString("\\");
    sJDD2 += TString("\\");
#else
    sJDD1 += TString("/");
    sJDD2 += TString("/");
#endif

    if(type=="mkk")
    {
        sJDD1 += "flowrate_input_with_keys.in";
        sJDD2 += "flowrate_input_with_keys2.in";

        tds1->getAttribute("rw")->setFileKey(sJDD1, "Rw", "%f");
        tds1->getAttribute("tu")->setFileKey(sJDD1, "Tu", "%f");
        tds1->getAttribute("hu")->setFileKey(sJDD1, "Hu", "%f");
        tds1->getAttribute("hl")->setFileKey(sJDD1, "Hl", "%f");
        tds1->getAttribute("kw")->setFileKey(sJDD1, "Kw", "%f");

        tds1->getAttribute("r")->setFileKey(sJDD2, "R", "%f");
        tds1->getAttribute("tu")->setFileKey(sJDD2, "Tu", "%f");
        tds1->getAttribute("tl")->setFileKey(sJDD2, "Tl", "%f");
        tds1->getAttribute("l")->setFileKey(sJDD2, "L", "%f");
        tds1->getAttribute("kw")->setFileKey(sJDD2, "Kw", "%f");
                
    }
    else if(type=="mff")
    {
        sJDD1 += "flowrate_input_with_flags.in";
        sJDD2 += "flowrate_input_with_flags2.in";
        
        tds1->getAttribute("rw")->setFileFlag(sJDD1, "@rw@", "%f");
        tds1->getAttribute("tu")->setFileFlag(sJDD1, "@tu@", "%f");
        tds1->getAttribute("hu")->setFileFlag(sJDD1, "@hu@", "%f");
        tds1->getAttribute("hl")->setFileFlag(sJDD1, "@hl@", "%f");
        tds1->getAttribute("kw")->setFileFlag(sJDD1, "@kw@", "%f");

        tds1->getAttribute("r")->setFileFlag(sJDD2, "@r@", "%f");
        tds1->getAttribute("tu")->setFileFlag(sJDD2, "@tu@", "%f");
        tds1->getAttribute("tl")->setFileFlag(sJDD2, "@tl@", "%f");
        tds1->getAttribute("l")->setFileFlag(sJDD2, "@l@", "%f");
        tds1->getAttribute("kw")->setFileFlag(sJDD2, "@kw@", "%f");
    }
    else if(type=="mkf")
    {
        sJDD1 += "flowrate_input_with_keys.in";
        sJDD2 += "flowrate_input_with_flags.in";

        tds1->getAttribute("rw")->setFileKey(sJDD1, "Rw", "%f");
        tds1->getAttribute("tu")->setFileKey(sJDD1, "Tu", "%f");
        tds1->getAttribute("hu")->setFileKey(sJDD1, "Hu", "%f");
        tds1->getAttribute("hl")->setFileKey(sJDD1, "Hl", "%f");
        tds1->getAttribute("kw")->setFileKey(sJDD1, "Kw", "%f");
        
        tds1->getAttribute("r")->setFileFlag(sJDD2, "@r@", "%f");
        tds1->getAttribute("tu")->setFileFlag(sJDD2, "@tu@", "%f");
        tds1->getAttribute("tl")->setFileFlag(sJDD2, "@tl@", "%f");
        tds1->getAttribute("l")->setFileFlag(sJDD2, "@l@", "%f");
        tds1->getAttribute("kw")->setFileFlag(sJDD2, "@kw@", "%f");
    }
    else if(type=="Folkk")
    {
        sJDD1 += "flowrate_input_with_keys.in";
        sJDD2 += "flowrate_input_with_keys2.in";

        tds1->getAttribute("rw")->setFileKey(sJDD1, "Rw", "%f");
        tds1->getAttribute("r")->setFileKey(sJDD1, "R", "%f");
        tds1->getAttribute("tu")->setFileKey(sJDD1, "Tu", "%f");
        tds1->getAttribute("tl")->setFileKey(sJDD1, "Tl", "%f");
        
        tds1->getAttribute("hu")->setFileKey(sJDD2, "Hu", "%f");
        tds1->getAttribute("hl")->setFileKey(sJDD2, "Hl", "%f");
        tds1->getAttribute("l")->setFileKey(sJDD2, "L", "%f");
        tds1->getAttribute("kw")->setFileKey(sJDD2, "Kw", "%f");
                
    }
    else if(type=="Folff")
    {
        sJDD1 += "flowrate_input_with_flags.in";
        sJDD2 += "flowrate_input_with_flags2.in";
        
        tds1->getAttribute("rw")->setFileFlag(sJDD1, "@rw@", "%f");
        tds1->getAttribute("r")->setFileFlag(sJDD1, "@r@", "%f");
        tds1->getAttribute("tu")->setFileFlag(sJDD1, "@tu@", "%f");
        tds1->getAttribute("tl")->setFileFlag(sJDD1, "@tl@", "%f");

        tds1->getAttribute("hu")->setFileFlag(sJDD2, "@hu@", "%f");
        tds1->getAttribute("hl")->setFileFlag(sJDD2, "@hl@", "%f");
        tds1->getAttribute("l")->setFileFlag(sJDD2, "@l@", "%f");
        tds1->getAttribute("kw")->setFileFlag(sJDD2, "@kw@", "%f");
    }
    else if(type=="Folkf")
    {
        sJDD1 += "flowrate_input_with_keys.in";
        sJDD2 += "flowrate_input_with_flags.in";

        tds1->getAttribute("rw")->setFileKey(sJDD1, "Rw", "%f");
        tds1->getAttribute("r")->setFileKey(sJDD1, "R", "%f");
        tds1->getAttribute("tu")->setFileKey(sJDD1, "Tu", "%f");
        tds1->getAttribute("tl")->setFileKey(sJDD1, "Tl", "%f");

	tds1->getAttribute("hu")->setFileFlag(sJDD2, "@hu@", "%f");
        tds1->getAttribute("hl")->setFileFlag(sJDD2, "@hl@", "%f");
        tds1->getAttribute("l")->setFileFlag(sJDD2, "@l@", "%f");
        tds1->getAttribute("kw")->setFileFlag(sJDD2, "@kw@", "%f");
    }
    
    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("flowrate") + " -s -" +
        string(type));

    cout << " ****" <<endl;
    cout << " **** End Of SetUp1" << endl;
    cout << " ****" << endl;
    cout << " ********************" << endl;
}

////////////////////////////////
///                          ///
///         Tests            ///
///                          ///
////////////////////////////////

void TLauncherTest::testNames()
{
    TString sName = "sTDS1";
    tds1 = new URANIE::DataServer::TDataServer(sName, "my title 1");
    tds1->addAttribute(ptrHat);

    _foutDataServer = new URANIE::Launcher::TOutputFileDataServer(
	"_output_flowrate_withRow_.dat");
    _foutDataServer->addAttribute(ptrD);//    _foutDataServer->addAttribute(ptrHat);

#ifdef WIN32
    tc1 = new URANIE::Launcher::TCode(tds1, "flowrate.exe -s -k");
#else
    tc1 = new URANIE::Launcher::TCode(tds1,
				      Form("%s/%s", _workingTestPath.Data(),
					   "../../teST/coDE/flowREATE/flowrate -s -k"));
#endif
    tc1->addOutputFile(_foutDataServer);

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);
    
    TString sLauncherExpected(
	Form("LauncherTDS%sCODEcode%s", sName.Data(), sName.Data()));
    CPPUNIT_ASSERT(!strcmp(tl1->GetName(), sLauncherExpected.Data()));

    TString sCodeExpected(Form("code%s", sName.Data()));
    CPPUNIT_ASSERT(!strcmp(tc1->GetName(), sCodeExpected.Data()));

    // PrePro and PostPro commands
    TString sPreProCommand = tl1->getPreProCommand();
    CPPUNIT_ASSERT_EQUAL(0, sPreProCommand.Length());

    TString sPostProCommand = tl1->getPostProCommand();
    CPPUNIT_ASSERT_EQUAL(0, sPostProCommand.Length());

    sPreProCommand ="ls -al";
    tl1->setPreProCommand(sPreProCommand);
    CPPUNIT_ASSERT_EQUAL(6, tl1->getPreProCommand().Length());
    CPPUNIT_ASSERT(strcmp(sPreProCommand.Data(), tl1->getPreProCommand().Data()) == kFALSE);

    sPostProCommand =Form("%sR", sPreProCommand.Data());
    tl1->setPostProCommand(sPostProCommand);
    CPPUNIT_ASSERT_EQUAL(7, tl1->getPostProCommand().Length());
    CPPUNIT_ASSERT(strcmp(sPostProCommand.Data(), tl1->getPostProCommand().Data()) == kFALSE);

}

void TLauncherTest::testTitles()
{

    TString sName = "sTDS1";

    tds1 = new URANIE::DataServer::TDataServer(sName, "my title 1");
    tds1->addAttribute(ptrHat);

    _foutDataServer = new URANIE::Launcher::TOutputFileDataServer(
	"_output_flowrate_withRow_.dat");
    _foutDataServer->addAttribute(ptrD);//    _foutDataServer->addAttribute(ptrHat);

#ifdef WIN32
    tc1 = new URANIE::Launcher::TCode(tds1, TString("flowrate.exe -s -m"));
#else
    tc1 = new URANIE::Launcher::TCode(tds1,
				      Form("%s/%s", _workingTestPath.Data(),
					   "../../teST/coDE/flowREATE/flowrate -s -k"));
#endif

    tc1->addOutputFile(_foutDataServer);

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);
    
    TString sLauncherExpected(
	Form("Launcher of code%s with %s", sName.Data(), sName.Data()));

    CPPUNIT_ASSERT(!strcmp(tl1->GetTitle(), sLauncherExpected));

    CPPUNIT_ASSERT(!strcmp(tc1->GetTitle(), tds1->GetTitle()));
}

void TLauncherTest::testWorkingdirectory()
{
    setUp1();
    _foutRow = new URANIE::Launcher::TOutputFileRow(
	"_output_flowrate_withRow_.dat");
    _foutRow->addAttribute(ptrHat);
    tc1->addOutputFile(_foutRow);

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);
    
    TString sCurrentDirectory = gSystem->pwd();

#ifdef WIN32
    tl1->setWorkingDirectory(gSystem->pwd() + TString("\\tmpLanceurUranie\\flowrate"));
#else
    tl1->setWorkingDirectory(
	gSystem->pwd() + TString("/tmpLanceurUranie/flowrate"));
#endif
    Int_t sortie = 0;
    tl1->run("nointermed");

    sortie = ((!gSystem->OpenDirectory("UranieLauncher_1")) ?  0 : 1);

    CPPUNIT_ASSERT_EQUAL(0, sortie);
}


void TLauncherTest::testRunDefault()
{
    setUp1();
    // The output file of the code
    _foutRow = new URANIE::Launcher::TOutputFileRow(
	"_output_flowrate_withRow_.dat");
    //The attribute in the output file
    _foutRow->addAttribute(ptrHat);
    tc1->addOutputFile(_foutRow);
    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);
 
    Int_t nvar = tds1->getNAttributes();
    tds1->getAttribute("ystar")->setOrigin(
	URANIE::DataServer::TAttribute::kConstant);
    tl1->run("nointermed");

    //
    CPPUNIT_ASSERT_EQUAL(32, tds1->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(tds1->getNAttributes(), nvar+1);

    Int_t FirstFieldExpected = 30;
    for (Int_t i = 0; i < tds1->getNAttributes() - 2; i++) //-1 cause "yhat" is not checked
    {
        list<URANIE::DataServer::TAttributeFileKey*> KeyList =
	    tds1->getAttribute(i)->getKeyList();
        list<URANIE::DataServer::TAttributeFileKey*>::iterator first =
	    KeyList.begin();
        CPPUNIT_ASSERT_EQUAL(Int_t(FirstFieldExpected + i * 2),
			     Int_t((*first)->_sEventList->GetEntry(0)));
    }

    tds1->exportData("_local_run_flowrateUniforDesign.dat");
    Int_t nret = UranieTests::areFilesEqual(TString(RESFLOWRATEASCIIFILEUNIFORMDESIGN),
                                            TString("_local_run_flowrateUniforDesign.dat"),"#DATE:");
    CPPUNIT_ASSERT_EQUAL(nret, 0);
    
}

void TLauncherTest::testRunDefaultMultiInputKK()
{
    TString type="mkk";
    setUpMultiInput(type);
    // The output file of the code
    _foutRow = new URANIE::Launcher::TOutputFileRow(
	"_output_flowrate_withRow_.dat");
    //The attribute in the output file
    _foutRow->addAttribute(ptrHat);
    tc1->addOutputFile(_foutRow);
    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);
 
    Int_t nvar = tds1->getNAttributes();
    tds1->getAttribute("ystar")->setOrigin(
	URANIE::DataServer::TAttribute::kConstant);
    tl1->run("nointermed");

    //
    CPPUNIT_ASSERT_EQUAL(32, tds1->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(tds1->getNAttributes(), nvar+1);


    tds1->exportData("_local_run_flowrateUniforDesign.dat");
    Int_t nret = UranieTests::areFilesEqual(TString(RESFLOWRATEASCIIFILEUNIFORMDESIGN),
                                            TString("_local_run_flowrateUniforDesign.dat"),"#DATE:");
    CPPUNIT_ASSERT_EQUAL(nret, 0);
	
}

void TLauncherTest::testRunDefaultMultiInputKF()
{
    TString type="mkf";
    setUpMultiInput(type);
    // The output file of the code
    _foutRow = new URANIE::Launcher::TOutputFileRow(
	"_output_flowrate_withRow_.dat");
    //The attribute in the output file
    _foutRow->addAttribute(ptrHat);
    tc1->addOutputFile(_foutRow);
    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);
 
    Int_t nvar = tds1->getNAttributes();
    tds1->getAttribute("ystar")->setOrigin(
	URANIE::DataServer::TAttribute::kConstant);
    tl1->run("nointermed");

    //
    CPPUNIT_ASSERT_EQUAL(32, tds1->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(tds1->getNAttributes(), nvar+1);

 
    tds1->exportData("_local_run_flowrateUniforDesign.dat");
    Int_t nret = UranieTests::areFilesEqual(TString(RESFLOWRATEASCIIFILEUNIFORMDESIGN),
                                            TString("_local_run_flowrateUniforDesign.dat"),"#DATE:");
    CPPUNIT_ASSERT_EQUAL(nret, 0);	
}

void TLauncherTest::testRunDefaultMultiInputFF()
{
    TString type="mff";
    setUpMultiInput(type);
    // The output file of the code
    _foutRow = new URANIE::Launcher::TOutputFileRow(
	"_output_flowrate_withRow_.dat");
    //The attribute in the output file
    _foutRow->addAttribute(ptrHat);
    tc1->addOutputFile(_foutRow);
    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);
 
    Int_t nvar = tds1->getNAttributes();
    tds1->getAttribute("ystar")->setOrigin(
	URANIE::DataServer::TAttribute::kConstant);
    tl1->run("nointermed");

    //
    CPPUNIT_ASSERT_EQUAL(32, tds1->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(tds1->getNAttributes(), nvar+1);

    tds1->exportData("_local_run_flowrateUniforDesign.dat");
    Int_t nret = UranieTests::areFilesEqual(TString(RESFLOWRATEASCIIFILEUNIFORMDESIGN),
                                            TString("_local_run_flowrateUniforDesign.dat"),"#DATE:");
    CPPUNIT_ASSERT_EQUAL(nret, 0);
	
}

void TLauncherTest::testRunDefaultMultiInputFollowingKK()
{
    TString type="Folkk";
    setUpMultiInput(type);
    // The output file of the code
    _foutRow = new URANIE::Launcher::TOutputFileRow(
	"_output_flowrate_withRow_.dat");
    //The attribute in the output file
    _foutRow->addAttribute(ptrHat);
    tc1->addOutputFile(_foutRow);
    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);
 
    Int_t nvar = tds1->getNAttributes();
    tds1->getAttribute("ystar")->setOrigin(
	URANIE::DataServer::TAttribute::kConstant);
    tl1->run("nointermed");

    //
    CPPUNIT_ASSERT_EQUAL(32, tds1->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(tds1->getNAttributes(), nvar+1);


    tds1->exportData("_local_run_flowrateUniforDesign.dat");
    Int_t nret = UranieTests::areFilesEqual(TString(RESFLOWRATEASCIIFILEUNIFORMDESIGN),
                                            TString("_local_run_flowrateUniforDesign.dat"),"#DATE:");
    CPPUNIT_ASSERT_EQUAL(nret, 0);
	
}

void TLauncherTest::testRunDefaultMultiInputFollowingKF()
{
    TString type="Folkf";
    setUpMultiInput(type);
    // The output file of the code
    _foutRow = new URANIE::Launcher::TOutputFileRow(
	"_output_flowrate_withRow_.dat");
    //The attribute in the output file
    _foutRow->addAttribute(ptrHat);
    tc1->addOutputFile(_foutRow);
    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);
 
    Int_t nvar = tds1->getNAttributes();
    tds1->getAttribute("ystar")->setOrigin(
	URANIE::DataServer::TAttribute::kConstant);
    tl1->run("nointermed");

    //
    CPPUNIT_ASSERT_EQUAL(32, tds1->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(tds1->getNAttributes(), nvar+1);

 
    tds1->exportData("_local_run_flowrateUniforDesign.dat");
    Int_t nret = UranieTests::areFilesEqual(TString(RESFLOWRATEASCIIFILEUNIFORMDESIGN),
                                            TString("_local_run_flowrateUniforDesign.dat"),"#DATE:");
    CPPUNIT_ASSERT_EQUAL(nret, 0);	
}

void TLauncherTest::testRunDefaultMultiInputFollowingFF()
{
    TString type="Folff";
    setUpMultiInput(type);
    // The output file of the code
    _foutRow = new URANIE::Launcher::TOutputFileRow(
	"_output_flowrate_withRow_.dat");
    //The attribute in the output file
    _foutRow->addAttribute(ptrHat);
    tc1->addOutputFile(_foutRow);
    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);
 
    Int_t nvar = tds1->getNAttributes();
    tds1->getAttribute("ystar")->setOrigin(
	URANIE::DataServer::TAttribute::kConstant);
    tl1->run("nointermed");

    //
    CPPUNIT_ASSERT_EQUAL(32, tds1->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(tds1->getNAttributes(), nvar+1);

    tds1->exportData("_local_run_flowrateUniforDesign.dat");
    Int_t nret = UranieTests::areFilesEqual(TString(RESFLOWRATEASCIIFILEUNIFORMDESIGN),
                                            TString("_local_run_flowrateUniforDesign.dat"),"#DATE:");
    CPPUNIT_ASSERT_EQUAL(nret, 0);
	
}

void TLauncherTest::testRunDefaultWithDrawings()
{
    setUp1();

    // The output file of the     
    _foutRow = new URANIE::Launcher::TOutputFileRow(
	"_output_flowrate_withRow_.dat");

    //The attribute in the output file
    _foutRow->addAttribute(ptrHat);
    tc1->addOutputFile(_foutRow);
    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);
    tl1->setVarDraw("rw");
 
    Int_t nvar = tds1->getNAttributes();
    tds1->getAttribute("ystar")->setOrigin(
	URANIE::DataServer::TAttribute::kConstant);
    tl1->run("nointermed");

    //
    CPPUNIT_ASSERT_EQUAL(32, tds1->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(tds1->getNAttributes(), nvar+1);

     tds1->exportData("_local_run_flowrateUniforDesign.dat");
    Int_t nret = UranieTests::areFilesEqual(TString(RESFLOWRATEASCIIFILEUNIFORMDESIGN),
                                            TString("_local_run_flowrateUniforDesign.dat"),"#DATE:");
    CPPUNIT_ASSERT_EQUAL(nret, 0);
}

void TLauncherTest::testRunFlowrateMultithreadAbove4000()
{
    setUp1(true);
    int ns=4000;
    URANIE::Sampler::TBasicSampling *bsam=new URANIE::Sampler::TBasicSampling(tds1,"lhs",ns);
    bsam->generateSample();

    delete bsam;
    
    // The output file of the     
    _foutRow = new URANIE::Launcher::TOutputFileRow(
	"_output_flowrate_withRow_.dat");

    //The attribute in the output file
    _foutRow->addAttribute(ptrHat);
    tc1->addOutputFile(_foutRow);
    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1, "multiquiet");
    tl1->setDrawProgressBar(kFALSE);
 
    Int_t nvar = tds1->getNAttributes();
    tl1->run("localhost=3");

    // pouet
    CPPUNIT_ASSERT_EQUAL(ns, tds1->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(tds1->getNAttributes(), nvar+1);

    URANIE::Launcher::TLauncherFunction *tlf = new URANIE::Launcher::TLauncherFunction(tds1, flowrateModel, "", "flowrateModel");
    tlf->run();
    delete tlf;

    tds1->addAttribute("Diff","flowrateModel-yhat");
    tds1->computeStatistic("Diff");

    //XYZ << "[" << tds1->getAttribute("Diff")->getMean() << "]" << "  [" << tds1->getAttribute("Diff")->getStd() << "]" << endl;
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0, tds1->getAttribute("Diff")->getMean(), 1e-4);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(0, tds1->getAttribute("Diff")->getStd(), 1e-3);    
    
}

void TLauncherTest::testRunDefaultMultithread()
{
    setUp1();
    // The output file of the code
    _foutRow = new URANIE::Launcher::TOutputFileRow(
	"_output_flowrate_withRow_.dat");
    //The attribute in the output file
    _foutRow->addAttribute(ptrHat);
    tc1->addOutputFile(_foutRow);
    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1, "multiquiet");
    tl1->setDrawProgressBar(kFALSE);
 
    Int_t nvar = tds1->getNAttributes();
    tds1->getAttribute("ystar")->setOrigin( URANIE::DataServer::TAttribute::kConstant);
    tl1->run("localhost=2");

    //
    CPPUNIT_ASSERT_EQUAL(32, tds1->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(tds1->getNAttributes(), nvar+1);

    Int_t FirstFieldExpected = 30;
    for (Int_t i = 0; i < tds1->getNAttributes() - 2; i++) //-1 cause "yhat" is not checked
    {
        list<URANIE::DataServer::TAttributeFileKey*> KeyList =
	    tds1->getAttribute(i)->getKeyList();
        list<URANIE::DataServer::TAttributeFileKey*>::iterator first =
	    KeyList.begin();
        CPPUNIT_ASSERT_EQUAL(Int_t(FirstFieldExpected + i * 2),
			     Int_t((*first)->_sEventList->GetEntry(0)));
    }

     tds1->exportData("_local_run_flowrateUniforDesign.dat");
    Int_t nret = UranieTests::areFilesEqual(TString(RESFLOWRATEASCIIFILEUNIFORMDESIGN),
                                            TString("_local_run_flowrateUniforDesign.dat"),"#DATE:");
    CPPUNIT_ASSERT_EQUAL(nret, 0);
	
}

//! Check the Number of Core
void TLauncherTest::testNumberOfCore()
{
    setUp1();
    tlbs = new URANIE::Launcher::TLauncherByStep(tds1, tc1); 
    CPPUNIT_ASSERT(tlbs->getNumberOfCores() == 0 );
    tlbs->setNumberOfCores(5);

    CPPUNIT_ASSERT(tlbs->getNumberOfCores() == 5);
    CPPUNIT_ASSERT(tlbs->getNumberOfCores() != 0);

    tlbs->setNumberOfCores(0);
    CPPUNIT_ASSERT(tlbs->getNumberOfCores() == 0);

//    delete tlbs;
}

//! Check the number of second allocated
void TLauncherTest::testAllocatedSeconds()
{
    setUp1();
    tlbs = new URANIE::Launcher::TLauncherByStep(tds1, tc1); 
    CPPUNIT_ASSERT(tlbs->getAllocatedTime() == 1800);

    tlbs->setAllocatedTime(2000);
    CPPUNIT_ASSERT(tlbs->getAllocatedTime() == 2000);
    CPPUNIT_ASSERT(tlbs->getAllocatedTime() != 1800);

    tlbs->setAllocatedTime(1800);
    CPPUNIT_ASSERT(tlbs->getAllocatedTime() == 1800);

//    delete tlbs;
}


//! Check the default value of ClusterName then when it's set one and two times
void TLauncherTest::testClusterName()
{
    setUp1();
    tlbs = new URANIE::Launcher::TLauncherByStep(tds1, tc1);
    CPPUNIT_ASSERT(tlbs->getClusterName() == "");

    tlbs->setClusterName("toto");
    CPPUNIT_ASSERT(tlbs->getClusterName() == "toto");
    CPPUNIT_ASSERT(tlbs->getClusterName() != "");
    CPPUNIT_ASSERT(tlbs->getClusterName() != "irene");

    tlbs->setClusterName("irene");
    CPPUNIT_ASSERT(tlbs->getClusterName() == "irene");

    //  delete tlbs;
}

//! Check the default value of ProjectName then when it's set one and two times
void TLauncherTest::testProjectName()
{
    setUp1();
    tlbs = new URANIE::Launcher::TLauncherByStep(tds1, tc1);
    CPPUNIT_ASSERT(tlbs->getClusterName() == "");
    
    tlbs->setProjectName("test");
    CPPUNIT_ASSERT(tlbs->getProjectName() == "test");
    CPPUNIT_ASSERT(tlbs->getProjectName() != "");
    CPPUNIT_ASSERT(tlbs->getProjectName() != "Myproject");

    tlbs->setProjectName("Myproject");
    CPPUNIT_ASSERT(tlbs->getProjectName() == "Myproject");
    
//    delete tlbs;
}
//!checks TLauncherByStep with argument run("irene");
//! for batch execution on the irene TGCC machine

void TLauncherTest::testRunByStepCurie()
{
    setUp1();
    tds1->computeStatistic("ystar");

    // The output file of the code
    _foutRow = new URANIE::Launcher::TOutputFileRow(
	"_output_flowrate_withRow_.dat");
    //The attribute in the output file
    _foutRow->addAttribute(ptrHat);
	
    tc1->addOutputFile(_foutRow);
    tlbs = new URANIE::Launcher::TLauncherByStep(tds1, tc1);
    tlbs->setDrawProgressBar(kFALSE);
 
    Int_t nvar = tds1->getNAttributes();
    tds1->getAttribute("ystar")->setOrigin(
	URANIE::DataServer::TAttribute::kConstant);
    tlbs->preTreatment();
    tlbs->setClusterName("irene");
    //check that NumberOfCores and AllocatedTime are positive values
    CPPUNIT_ASSERT(tlbs->getAllocatedTime() > 0);
    CPPUNIT_ASSERT(tlbs->getNumberOfCores() >= 0);
    if (strncmp(gSystem->HostName(),"irene",5)==0) {  // if on irene, check the expected behavior
    	// run the test
        tlbs->run("irene");

    	// check UranieTmp.bsub exists
    	TString bsubFile = "__Uranie_Tmp__.bsub";
    	CPPUNIT_ASSERT(gSystem->AccessPathName(bsubFile, kFileExists) == 0);

    	// check some content of the file
        ifstream ifs(bsubFile.Data(), ios::in);
        TString str;
        str.ReadFile(ifs);
        ifs.close();

        CPPUNIT_ASSERT(str.Index("#MSUB -n 32")>0);
        CPPUNIT_ASSERT(str.Index("#MSUB -t 1800")>0);
        Int_t start = str.Index("ccc_mprun");
        Int_t count = 0;
        while(start>=0){
            start = str.Index("ccc_mprun", start+1);
            count++;
        }
        CPPUNIT_ASSERT_EQUAL(count, 32);

    	// Wait for the the computation to complete
        TString batchid = gSystem->GetFromPipe("awk '{print $NF}' __Uranie_Tmp__.log");
        TString check_command = Form("ccc_mpp | grep %s | wc -l", batchid.Data());

        while(1){
            TString output = gSystem->GetFromPipe(check_command);
            if(output.CompareTo("1") != 0){
                break;
            }
            gSystem->Sleep(500);
        }

        // Only when computation completed, post-treatment and values checking
        tlbs->postTreatment();
        tds1->computeStatistic("yhat");

        CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getAttribute("ystar")->getMean(),
				     tds1->getAttribute("yhat")->getMean(),1e-2);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getAttribute("ystar")->getStd(),
				     tds1->getAttribute("yhat")->getStd(),1e-2);

        CPPUNIT_ASSERT_EQUAL(32, tds1->getNPatterns());
        CPPUNIT_ASSERT_EQUAL(tds1->getNAttributes(), nvar+1);

        Int_t FirstFieldExpected = 30;
        for (Int_t i = 0; i < tds1->getNAttributes() - 2; i++) //-1 cause "yhat" is not checked
        {
            list<URANIE::DataServer::TAttributeFileKey*> KeyList =
		tds1->getAttribute(i)->getKeyList();
            list<URANIE::DataServer::TAttributeFileKey*>::iterator first =
		KeyList.begin();
            CPPUNIT_ASSERT_EQUAL(Int_t(FirstFieldExpected + i * 2),
				 Int_t((*first)->_sEventList->GetEntry(0)));
        }
    }
    else {  // if not on irene, check the exception is thrown
        CPPUNIT_ASSERT_THROW(tlbs->run("irene"), URANIE::Exceptions::UErrorExceptions);
    }

    
}

void TLauncherTest::testInputFileKeyRecreate()
{
    setUp2();

    URANIE::Launcher::TInputFileRecreate *theCrtTIFR =
	dynamic_cast<URANIE::Launcher::TInputFileRecreate *>(tc2->getInputFile(
								 0));
    theCrtTIFR->setHeader("Exemple Header");
    theCrtTIFR->setFooter("Exemple Footer");

    // GetHeader() of the file TInputFileRecreate
    CPPUNIT_ASSERT(!strcmp(theCrtTIFR->getHeader(), "Exemple Header"));

    //GetFooter() of the file TInputFileRecreate
    CPPUNIT_ASSERT(!strcmp(theCrtTIFR->getFooter(), "Exemple Footer"));

    theCrtTIFR->resetHeader();
    theCrtTIFR->resetFooter();
    theCrtTIFR->setHeader(
	"#\n#\n# INPUT FILE with KEYS for the \"FLOWREATE\" code\n# \\date   2008-04-22 12:53:35\n#  parameters : 8\n#\n\n\ndate = 123456 ;\n");
    theCrtTIFR->setFooter("\n\n\n\n\n\nend =6;");

    Int_t nvar = tds2->getNAttributes();
    Int_t npat = tds2->getNPatterns();

    //Test run
    tl2->setSave();
    tl2->run("nointermed");

    CPPUNIT_ASSERT_EQUAL(npat, tds2->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(nvar + 1, tds2->getNAttributes());

	
}

void TLauncherTest::testInputFileFlag()
{
    tds1 = new URANIE::DataServer::TDataServer("tds1", "my title 1");
    tds1->fileDataRead(FLOWRATEASCIIFILEUNIFORMDESIGN);

#ifdef WIN32
    TString sJDD = gSystem->pwd()+TString("\\flowrate_input_with_flags.in");
#else
    TString sJDD = gSystem->pwd() + TString("/flowrate_input_with_flags.in");
#endif
    tds1->getAttribute("rw")->setFileFlag(sJDD, "@rw@", "%f");
    tds1->getAttribute("r")->setFileFlag(sJDD, "@r@", "%f");
    tds1->getAttribute("tu")->setFileFlag(sJDD, "@tu@", "%f");
    tds1->getAttribute("tl")->setFileFlag(sJDD, "@tl@", "%f");
    tds1->getAttribute("hu")->setFileFlag(sJDD, "@hu@", "%f");
    tds1->getAttribute("hl")->setFileFlag(sJDD, "@hl@", "%f");
    tds1->getAttribute("l")->setFileFlag(sJDD, "@l@", "%f");
    tds1->getAttribute("kw")->setFileFlag(sJDD, "@kw@", "%f");

    _foutRow = new URANIE::Launcher::TOutputFileRow(
	"_output_flowrate_withRow_.dat");
    _foutRow->addAttribute(ptrHat);

    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("flowrate") + " -s -f");
    tc1->addOutputFile(_foutRow);

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);
    tl1->run("nointermed");

    CPPUNIT_ASSERT(tds1->isAttribute("yhat"));
    CPPUNIT_ASSERT(!tds1->isAttribute("d"));

    Int_t npat = tds1->getNPatterns();
    CPPUNIT_ASSERT_DOUBLES_EQUAL(26.18019, tds1->getValue("yhat", 0),
				 DBL_DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(157.591, tds1->getValue("yhat", npat - 1),
				 DBL_DBL_PRECISION_COMPARE);
}

#ifdef WITH_LIBXSLT
void TLauncherTest::testInputFileXML_WithoutLIBXML2()
{
    tds1 = new URANIE::DataServer::TDataServer("tds1", "my title 1");
    tds1->fileDataRead(FLOWRATEASCIIFILEUNIFORMDESIGN);

#ifdef WIN32
    TString sJDD = gSystem->pwd()+TString("\\flowrate_input_with_xml.in");
#else
    TString sJDD = gSystem->pwd()+TString("/flowrate_input_with_xml.in");
#endif
    tds1->getAttribute("rw")->setFileKey(sJDD, "wall_friction/@rw", "%e", URANIE::DataServer::TAttributeFileKey::kXMLAttribute);
    tds1->getAttribute("r")->setFileKey(sJDD, "wall_friction/@r", "%e", URANIE::DataServer::TAttributeFileKey::kXMLAttribute);
    tds1->getAttribute("tu")->setFileKey(sJDD, "parameter[@name='Tu']/value/double", "%e", URANIE::DataServer::TAttributeFileKey::kXMLField);
    tds1->getAttribute("tl")->setFileKey(sJDD, "parameter[@name='Tl']/value/double", "%e", URANIE::DataServer::TAttributeFileKey::kXMLField);
    tds1->getAttribute("hu")->setFileKey(sJDD, "parameter[tonode='mesher' and toport='dt_hu']/value/double", "%e", URANIE::DataServer::TAttributeFileKey::kXMLField);
    tds1->getAttribute("hl")->setFileKey(sJDD, "parameter[tonode='mesher' and toport='dt_hl']/value/double", "%e", URANIE::DataServer::TAttributeFileKey::kXMLField);
    tds1->getAttribute("l")->setFileKey(sJDD, "parameter[@name='L']/@precision", "%e", URANIE::DataServer::TAttributeFileKey::kXMLAttribute);
    tds1->getAttribute("kw")->setFileKey(sJDD, "kW/@value", "%e", URANIE::DataServer::TAttributeFileKey::kXMLAttribute);

    Double_t * val = new Double_t[5];
    val[0] = 0.15;
    val[1] = 0.3;
    val[2] = 0.45;
    val[3] = 0.6;
    val[4] = 0.75;
    Int_t n = 0;

    URANIE::Launcher::TInputFileXML* tifx = new URANIE::Launcher::TInputFileXML("Test_flowrate_input_with_xml.in");
    tifx->addAttribute(tds1->getAttribute("rw"));
    tifx->addAttribute(tds1->getAttribute("r"));
    tifx->addAttribute(tds1->getAttribute("tu"));
    tifx->addAttribute(tds1->getAttribute("tl"));
    tifx->addAttribute(tds1->getAttribute("hu"));
    tifx->substitue(n, val);

    Int_t nret = UranieTests::areFilesEqual(TString("Test_flowrate_input_with_xml.ref"),
                                            TString("Test_flowrate_input_with_xml.in"));

    CPPUNIT_ASSERT_EQUAL(nret, 0);
    
    gSystem->CopyFile("flowrate_input_with_xml.in","Test_flowrate_input_with_xml.in",kTRUE);
    
    // Clean
    delete[] val;
    val = NULL;
    delete tifx;
    tifx = NULL;
}
#endif

#ifdef WITH_LIBXML2
#ifdef WITH_LIBXSLT
void TLauncherTest::testInputFileXML()
{
    tds1 = new URANIE::DataServer::TDataServer("tds1", "my title 1");
    tds1->fileDataRead(FLOWRATEASCIIFILEUNIFORMDESIGN);

#ifdef WIN32
    TString sJDD = gSystem->pwd()+TString("\\flowrate_input_with_xml.in");
#else
    TString sJDD = gSystem->pwd()+TString("/flowrate_input_with_xml.in");
#endif
    tds1->getAttribute("rw")->setFileKey(sJDD, "wall_friction/@rw", "%e", URANIE::DataServer::TAttributeFileKey::kXMLAttribute);
    tds1->getAttribute("r")->setFileKey(sJDD, "wall_friction/@r", "%e", URANIE::DataServer::TAttributeFileKey::kXMLAttribute);
    tds1->getAttribute("tu")->setFileKey(sJDD, "parameter[@name='Tu']/value/double", "%e", URANIE::DataServer::TAttributeFileKey::kXMLField);
    tds1->getAttribute("tl")->setFileKey(sJDD, "parameter[@name='Tl']/value/double", "%e", URANIE::DataServer::TAttributeFileKey::kXMLField);
    tds1->getAttribute("hu")->setFileKey(sJDD, "parameter[tonode='mesher' and toport='dt_hu']/value/double", "%e", URANIE::DataServer::TAttributeFileKey::kXMLField);
    tds1->getAttribute("hl")->setFileKey(sJDD, "parameter[tonode='mesher' and toport='dt_hl']/value/double", "%e", URANIE::DataServer::TAttributeFileKey::kXMLField);
    tds1->getAttribute("l")->setFileKey(sJDD, "parameter[@name='L']/@precision", "%e", URANIE::DataServer::TAttributeFileKey::kXMLAttribute);
    tds1->getAttribute("kw")->setFileKey(sJDD, "kW/@value", "%e", URANIE::DataServer::TAttributeFileKey::kXMLAttribute);

    _foutDataServer = new URANIE::Launcher::TOutputFileDataServer("_output_flowrate_withRow_.dat");
    _foutDataServer->addAttribute(ptrHat);

    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("flowrate") + " -s -x");
    tc1->addOutputFile(_foutDataServer);

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);
    // tl1->setWorkingDirectory("InputFileXML");
    // tl1->setSave(-1);
    tl1->run("nointermed");

    CPPUNIT_ASSERT(tds1->isAttribute("yhat"));

    Int_t npat = tds1->getNPatterns();
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 26.18019, tds1->getValue("yhat",0), DBL_DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 157.5910, tds1->getValue("yhat",npat-1), DBL_DBL_PRECISION_COMPARE);
}
#endif
#endif

#ifdef WITH_LIBXML2
#ifdef WITH_LIBXSLT
void TLauncherTest::testInputFileXMLLauncherFailure()
{
    tds1 = new URANIE::DataServer::TDataServer("tds1", "my title 1");
    tds1->fileDataRead("flowrate200Samples.dat");

#ifdef WIN32
    TString sJDD = gSystem->pwd()+TString("\\flowrate_input_with_xml.in");
#else
    TString sJDD = gSystem->pwd()+TString("/flowrate_input_with_xml.in");
#endif
    tds1->getAttribute("rw")->setFileKey(sJDD, "wall_friction/@rw", "%e", URANIE::DataServer::TAttributeFileKey::kXMLAttribute);
    tds1->getAttribute("r")->setFileKey(sJDD, "wall_friction/@r", "%e", URANIE::DataServer::TAttributeFileKey::kXMLAttribute);
    tds1->getAttribute("tu")->setFileKey(sJDD, "parameter[@name='Tu']/value/double", "%d", URANIE::DataServer::TAttributeFileKey::kXMLField);
    tds1->getAttribute("tl")->setFileKey(sJDD, "parameter[@name='Tl']/value/double", "%d", URANIE::DataServer::TAttributeFileKey::kXMLField);
    tds1->getAttribute("hu")->setFileKey(sJDD, "parameter[tonode='mesher' and toport='dt_hu']/value/double", "%e", URANIE::DataServer::TAttributeFileKey::kXMLField);
    tds1->getAttribute("hl")->setFileKey(sJDD, "parameter[tonode='mesher' and toport='dt_hl']/value/double", "%e", URANIE::DataServer::TAttributeFileKey::kXMLField);
    tds1->getAttribute("l")->setFileKey(sJDD, "parameter[@name='L']/@precision", "%e", URANIE::DataServer::TAttributeFileKey::kXMLAttribute);
    tds1->getAttribute("kw")->setFileKey(sJDD, "kW/@value", "%e", URANIE::DataServer::TAttributeFileKey::kXMLAttribute);

    _foutDataServer = new URANIE::Launcher::TOutputFileDataServer("_output_flowrate_withRow_.dat");
    ptrHat->setDefaultValue(-200.0);
    _foutDataServer->addAttribute(ptrHat);

    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("flowrate") + " -s -xf");
    tc1->addOutputFile(_foutDataServer);

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);
    tl1->run("nointermed");

    tds1->setSelect("yhat<=0");

    Int_t nFailure= tds1->getNPatterns();
    CPPUNIT_ASSERT_EQUAL(54,nFailure);
}
#endif
#endif


void TLauncherTest::testNewKey()
{
    Double_t * val = new Double_t[5];
    val[0] = 0.15;
    val[1] = 0.3;
    val[2] = 0.45;
    val[3] = 0.6;
    val[4] = 0.75;
    Int_t n = 0;
    tds2 = new URANIE::DataServer::TDataServer();
    tds2->fileDataRead(FLOWRATEASCIIFILEUNIFORMDESIGN);
    tds2->getAttribute("rw")->setFileKey("TestFileKey.in", "Rw", "%.2f",
					 URANIE::DataServer::TAttributeFileKey::kNewKey);
    tds2->getAttribute("r")->setFileKey("TestFileKey.in", "R", "%.2f",
					URANIE::DataServer::TAttributeFileKey::kNewKey);
    tds2->getAttribute("tu")->setFileKey("TestFileKey.in", "Tu", "%.2f",
					 URANIE::DataServer::TAttributeFileKey::kNewKey);
    tds2->getAttribute("tl")->setFileKey("TestFileKey.in", "Tl", "%.2f",
					 URANIE::DataServer::TAttributeFileKey::kNewKey);
    tds2->getAttribute("hu")->setFileKey("TestFileKey.in", "Hu", "%.2f",
					 URANIE::DataServer::TAttributeFileKey::kNewKey);

    tifr = new URANIE::Launcher::TInputFileRecreate("TestFileKey.in");
    tifr->addAttribute(tds2->getAttribute("rw"));
    tifr->addAttribute(tds2->getAttribute("r"));
    tifr->addAttribute(tds2->getAttribute("tu"));
    tifr->addAttribute(tds2->getAttribute("tl"));
    tifr->addAttribute(tds2->getAttribute("hu"));
    tifr->substitue(n, val);

    Int_t nret = UranieTests::areFilesEqual(TString("TestFileKey.ref"),
                                            TString("TestFileKey.in"));
			        

    CPPUNIT_ASSERT_EQUAL(nret, 0);

    // Clean
    delete[] val;
    val = NULL;
}

void TLauncherTest::testNewLine()
{
    Double_t * val = new Double_t[5];
    val[0] = 0.15;
    val[1] = 0.3;
    val[2] = 0.45;
    val[3] = 0.6;
    val[4] = 0.75;
    Int_t n = 0;
    tds2 = new URANIE::DataServer::TDataServer();
    tds2->fileDataRead(FLOWRATEASCIIFILEUNIFORMDESIGN);
    tds2->getAttribute("rw")->setFileKey("TestFileLine.in", "Rw", "%.2f",
					 URANIE::DataServer::TAttributeFileKey::kNewLine);
    tds2->getAttribute("r")->setFileKey("TestFileLine.in", "R", "%.2f",
					URANIE::DataServer::TAttributeFileKey::kNewLine);
    tds2->getAttribute("tu")->setFileKey("TestFileLine.in", "Tu", "%.2f",
					 URANIE::DataServer::TAttributeFileKey::kNewLine);
    tds2->getAttribute("tl")->setFileKey("TestFileLine.in", "Tl", "%.2f",
					 URANIE::DataServer::TAttributeFileKey::kNewLine);
    tds2->getAttribute("hu")->setFileKey("TestFileLine.in", "Hu", "%.2f",
					 URANIE::DataServer::TAttributeFileKey::kNewLine);

    tifr = new URANIE::Launcher::TInputFileRecreate("TestFileLine.in");
    tifr->addAttribute(tds2->getAttribute("rw"));
    tifr->addAttribute(tds2->getAttribute("r"));
    tifr->addAttribute(tds2->getAttribute("tu"));
    tifr->addAttribute(tds2->getAttribute("tl"));
    tifr->addAttribute(tds2->getAttribute("hu"));
    tifr->substitue(n, val);
    Int_t nret = UranieTests::areFilesEqual(TString("TestFileLine.ref"),
                                            TString("TestFileLine.in"));
    CPPUNIT_ASSERT_EQUAL(nret, 0);

    // Clean
    delete[] val;
    val = NULL;
}

void TLauncherTest::testNewColumn()
{
    Double_t * val = new Double_t[5];
    val[0] = 0.15;
    val[1] = 0.3;
    val[2] = 0.45;
    val[3] = 0.6;
    val[4] = 0.75;
    Int_t n = 0;
    tds2 = new URANIE::DataServer::TDataServer();
    tds2->fileDataRead(FLOWRATEASCIIFILEUNIFORMDESIGN);
    tds2->getAttribute("rw")->setFileKey("TestFileColumn.in", "Rw", "%.2f",
					 URANIE::DataServer::TAttributeFileKey::kNewColumn);
    tds2->getAttribute("r")->setFileKey("TestFileColumn.in", "R", "%.2f",
					URANIE::DataServer::TAttributeFileKey::kNewColumn);
    tds2->getAttribute("tu")->setFileKey("TestFileColumn.in", "Tu", "%.2f",
					 URANIE::DataServer::TAttributeFileKey::kNewColumn);
    tds2->getAttribute("tl")->setFileKey("TestFileColumn.in", "Tl", "%.2f",
					 URANIE::DataServer::TAttributeFileKey::kNewColumn);
    tds2->getAttribute("hu")->setFileKey("TestFileColumn.in", "Hu", "%.2f",
					 URANIE::DataServer::TAttributeFileKey::kNewColumn);

    tifr = new URANIE::Launcher::TInputFileRecreate("TestFileColumn.in");
    tifr->addAttribute(tds2->getAttribute("rw"));
    tifr->addAttribute(tds2->getAttribute("r"));
    tifr->addAttribute(tds2->getAttribute("tu"));
    tifr->addAttribute(tds2->getAttribute("tl"));
    tifr->addAttribute(tds2->getAttribute("hu"));
    tifr->substitue(n, val);
    Int_t nret = UranieTests::areFilesEqual(TString("TestFileColumn.ref"),
                                            TString("TestFileColumn.in"));
    CPPUNIT_ASSERT_EQUAL(nret, 0);

    delete[] val;
    val = NULL;
}

void TLauncherTest::testSearchIndexAttributeRecreate()
{
    tds2 = new URANIE::DataServer::TDataServer();
    tds2->fileDataRead(FLOWRATEASCIIFILEUNIFORMDESIGN);
    tds2->getAttribute("rw")->setFileKey("TestFile.in", "Rw");
    tifr = new URANIE::Launcher::TInputFileRecreate("TestFile.in");
    tifr->searchIndexeAttribute(tds2->getAttribute("rw"));
    list<URANIE::DataServer::TAttributeFileKey *> KeyList = tds2->getAttribute(
	"rw")->getKeyList();
    list<URANIE::DataServer::TAttributeFileKey *>::iterator first =
	KeyList.begin();
    CPPUNIT_ASSERT((*first)->_sEventList == NULL);
}

void TLauncherTest::testLauncherInputRowRecreateOutputFileDataServer()
{
    tds1 = new URANIE::DataServer::TDataServer("tds1", "my title 1");
    tds1->fileDataRead(FLOWRATEASCIIFILEUNIFORMDESIGN);
    Int_t npat = tds1->getNPatterns();
    Int_t ninput = tds1->getNAttributes();
    Int_t noutput = 0;
    Int_t natt = ninput;

    // The reference input file
#ifdef WIN32
    TString sJDD1 = gSystem->pwd()+TString("\\flowrate_input_with_values_rows.in");
#else
    TString sJDD1 = gSystem->pwd()
	+ TString("/flowrate_input_with_values_rows.in");
#endif  
    // Set the input file with "key = value" format CREATED ON FLY with row for each input attributes
    tds1->getAttribute("rw")->setFileKey(sJDD1, "Rw", "",
					 URANIE::DataServer::TAttributeFileKey::kNewRow);
    tds1->getAttribute("r")->setFileKey(sJDD1, "R", "",
					URANIE::DataServer::TAttributeFileKey::kNewRow);
    tds1->getAttribute("tu")->setFileKey(sJDD1, "Tu", "",
					 URANIE::DataServer::TAttributeFileKey::kNewRow);
    tds1->getAttribute("tl")->setFileKey(sJDD1, "Tl", "",
					 URANIE::DataServer::TAttributeFileKey::kNewRow);
    tds1->getAttribute("hu")->setFileKey(sJDD1, "Hu", "",
					 URANIE::DataServer::TAttributeFileKey::kNewRow);
    tds1->getAttribute("hl")->setFileKey(sJDD1, "Hl", "",
					 URANIE::DataServer::TAttributeFileKey::kNewRow);
    tds1->getAttribute("l")->setFileKey(sJDD1, "L", "",
					URANIE::DataServer::TAttributeFileKey::kNewRow);
    tds1->getAttribute("kw")->setFileKey(sJDD1, "Kw", "",
					 URANIE::DataServer::TAttributeFileKey::kNewRow);

    _foutDataServer = new URANIE::Launcher::TOutputFileDataServer(
	"_output_flowrate_withRow_.dat");
    _foutDataServer->addAttribute(ptrD);
    _foutDataServer->addAttribute(ptrHat);

    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("flowrate") + " -s -r");
    tc1->addOutputFile(_foutDataServer);

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);
    tl1->run("nointermed");

    natt += 2;
    noutput += 2;
    CPPUNIT_ASSERT_EQUAL(npat, tds1->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(natt, tds1->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(ninput, tds1->getNInputAttributes());
    CPPUNIT_ASSERT_EQUAL(noutput, tds1->getNOutputAttributes());

    TTree* tuple = (TNtupleD*) tds1->getTuple();
    TObjArray * Leaves = tuple->GetListOfLeaves();
    CPPUNIT_ASSERT(tds1->isAttribute("d"));
    CPPUNIT_ASSERT(tds1->isAttribute("yhat"));
    CPPUNIT_ASSERT(Leaves->FindObject("d") != 0);
    CPPUNIT_ASSERT(Leaves->FindObject("yhat") != 0);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(3602.045, tds1->getValue("d", 0),
				 DBL_DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(26.18019, tds1->getValue("yhat", 0),
				 DBL_DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1348.649, tds1->getValue("d", npat - 1),
				 DBL_DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(157.591, tds1->getValue("yhat", npat - 1),
				 DBL_DBL_PRECISION_COMPARE);
}

void TLauncherTest::testLauncherInputOutputKey_WithBlank()
{

    tds1 = new URANIE::DataServer::TDataServer("pouet","foo");
    tds1->fileDataRead("Exemples_For_Blank.dat");

    // The reference input file
#ifdef WIN32
    TString sFileName = gSystem->pwd()+TString("\\withblank_input.dat");
#else
    TString sFileName = gSystem->pwd()
	+ TString("/withblank_input.dat");
#endif  
    tds1->getAttribute("pressure")->setFileKey(sFileName, "water model.brewery pit pressure ");
    tds1->getAttribute("density")->setFileKey(sFileName, "residual brew density model.factor for glass ");

    URANIE::Launcher::TInputFileKey *in = new URANIE::Launcher::TInputFileKey(sFileName);
    // in->setLog();
    in->setFieldSeparatorCharacter("=");
    in->setSeparatorCharacter(" ");
    in->addAttribute( tds1->getAttribute("pressure") );
    in->addAttribute( tds1->getAttribute("density") );

    TString sOutFileName="withblank_output.dat";

    _foutKey = new URANIE::Launcher::TOutputFileKey(sOutFileName);
    _foutKey->setFieldSeparatorCharacter("=");
    _foutKey->setSeparatorCharacter("\n;");
    URANIE::DataServer::TAttribute *outAtt = new URANIE::DataServer::TAttribute("outpressure");
    outAtt->setFileKey(sOutFileName,"Test doubling.water model.brewery pit pressure ");
    URANIE::DataServer::TAttribute *outAtt2 = new URANIE::DataServer::TAttribute("outdensity");
    outAtt2->setFileKey(sOutFileName,"Chocolat doubling.residual brew density model.factor for glass ");
    _foutKey->addAttribute(outAtt);
    _foutKey->addAttribute(outAtt2);
    outAtt->delShare();
    outAtt2->delShare();
    
    // Create a TCode object with the TDS (attribute and input files) and the command to execute
    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("withblank"));

    // Add the output file
    tc1->addInputFile( in );    
    tc1->addOutputFile( _foutKey );
    
    // Launcher of the code on the Design of experiments (DoE) in the TDS
    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setSave(-1);
    tl1->setClean();
    tl1->run("nointermed");

    tds1->addAttribute("diff_pressure","2*pressure - outpressure");
    tds1->addAttribute("diff_density","3*density - outdensity");

    double *press, *densi;
    tds1->getTuple()->Draw("diff_pressure:diff_density","","goff");
    press=tds1->getTuple()->GetV1();
    densi=tds1->getTuple()->GetV2();
    for(unsigned int ival=0; ival<10; ival++)
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(0, press[ival], 1e-14);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(0, densi[ival], 1e-14);
    }

    //    delete in;
}

void TLauncherTest::testLauncherOutputFileKey_yhat_Relative()
{
    tds1 = new URANIE::DataServer::TDataServer("tds1", "my title 1");
    tds1->fileDataRead(FLOWRATEASCIIFILEUNIFORMDESIGN);
    // The reference input file
    TString sJDD1 = "flowrate_input_with_keys.in";

    tds1->getAttribute("rw")->setFileKey(sJDD1, "Rw");
    tds1->getAttribute("r")->setFileKey(sJDD1, "R");
    tds1->getAttribute("tu")->setFileKey(sJDD1, "Tu");
    tds1->getAttribute("tl")->setFileKey(sJDD1, "Tl");
    tds1->getAttribute("hu")->setFileKey(sJDD1, "Hu");
    tds1->getAttribute("hl")->setFileKey(sJDD1, "Hl");
    tds1->getAttribute("l")->setFileKey(sJDD1, "L");
    tds1->getAttribute("kw")->setFileKey(sJDD1, "Kw");

    _foutKey = new URANIE::Launcher::TOutputFileKey(
	"_output_flowrate_withKey_.dat");
    _foutKey->addAttribute(ptrHat);

    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("flowrate") + " -s -k");

    tc1->addOutputFile(_foutKey);

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);
    TString sCurrentDirectory = gSystem->pwd();

#ifdef WIN32
    tl1->setWorkingDirectory(Form("%s\\tmpLanceurUranie\\flowrate", sCurrentDirectory.Data()));
#else
    tl1->setWorkingDirectory(
	Form("%s/tmpLanceurUranie/flowrate", sCurrentDirectory.Data()));
#endif
    tl1->run("nointermed");
    tds1->exportData("_outputFileKey_yhat_.dat");

    Int_t nret = UranieTests::areFilesEqual(TString("_outputFileKey_yhat_.dat.ref"),
                                            TString("_outputFileKey_yhat_.dat"), TString("#DATE:"));
    CPPUNIT_ASSERT_EQUAL(nret, 0);
}

void TLauncherTest::testLauncherOutputFileKey_yhat()
{
    tds1 = new URANIE::DataServer::TDataServer("tds1", "my title 1");
    tds1->fileDataRead(FLOWRATEASCIIFILEUNIFORMDESIGN);
    // The reference input file
#ifdef WIN32
    TString sJDD1 = gSystem->pwd()+TString("\\flowrate_input_with_keys.in");
#else
    TString sJDD1 = gSystem->pwd() + TString("/flowrate_input_with_keys.in");
#endif  

    tds1->getAttribute("rw")->setFileKey(sJDD1, "Rw");
    tds1->getAttribute("r")->setFileKey(sJDD1, "R");
    tds1->getAttribute("tu")->setFileKey(sJDD1, "Tu");
    tds1->getAttribute("tl")->setFileKey(sJDD1, "Tl");
    tds1->getAttribute("hu")->setFileKey(sJDD1, "Hu");
    tds1->getAttribute("hl")->setFileKey(sJDD1, "Hl");
    tds1->getAttribute("l")->setFileKey(sJDD1, "L");
    tds1->getAttribute("kw")->setFileKey(sJDD1, "Kw");

    _foutKey = new URANIE::Launcher::TOutputFileKey(
	"_output_flowrate_withKey_.dat");
    _foutKey->addAttribute(ptrHat);

    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("flowrate") + " -s -k");

    tc1->addOutputFile(_foutKey);

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);
//   tl1->setSave(-1);
    TString sCurrentDirectory = gSystem->pwd();

#ifdef WIN32
    tl1->setWorkingDirectory(Form("%s\\tmpLanceurUranie\\flowrate", sCurrentDirectory.Data()));
#else
    tl1->setWorkingDirectory(
	Form("%s/tmpLanceurUranie/flowrate", sCurrentDirectory.Data()));
#endif
    tl1->run("nointermed");
    tds1->exportData("_outputFileKey_yhat_.dat");

    Int_t nret = UranieTests::areFilesEqual(TString("_outputFileKey_yhat_.dat.ref"),
                                            TString("_outputFileKey_yhat_.dat"), TString("#DATE:"));
    CPPUNIT_ASSERT_EQUAL(nret, 0);
}

void TLauncherTest::testLauncherOutputFileKey_d()
{
    tds1 = new URANIE::DataServer::TDataServer("tds1", "my title 1");
    tds1->fileDataRead(FLOWRATEASCIIFILEUNIFORMDESIGN);
    // The reference input file
#ifdef WIN32
    TString sJDD1 = gSystem->pwd()+TString("\\flowrate_input_with_keys.in");
#else
    TString sJDD1 = gSystem->pwd() + TString("/flowrate_input_with_keys.in");
#endif  

    tds1->getAttribute("rw")->setFileKey(sJDD1, "Rw");
    tds1->getAttribute("r")->setFileKey(sJDD1, "R");
    tds1->getAttribute("tu")->setFileKey(sJDD1, "Tu");
    tds1->getAttribute("tl")->setFileKey(sJDD1, "Tl");
    tds1->getAttribute("hu")->setFileKey(sJDD1, "Hu");
    tds1->getAttribute("hl")->setFileKey(sJDD1, "Hl");
    tds1->getAttribute("l")->setFileKey(sJDD1, "L");
    tds1->getAttribute("kw")->setFileKey(sJDD1, "Kw");

    _foutKey = new URANIE::Launcher::TOutputFileKey(
	"_output_flowrate_withKey_.dat");
    _foutKey->addAttribute(ptrD);

    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("flowrate") + " -s -k");

    tc1->addOutputFile(_foutKey);

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);
    TString sCurrentDirectory = gSystem->pwd();

#ifdef WIN32
    tl1->setWorkingDirectory(Form("%s\\tmpLanceurUranie\\flowrate", sCurrentDirectory.Data()));
#else
    tl1->setWorkingDirectory(
	Form("%s/tmpLanceurUranie/flowrate", sCurrentDirectory.Data()));
#endif
    tl1->run("nointermed");

    tds1->exportData("_outputFileKey_d_.dat");
    Int_t nret = UranieTests::areFilesEqual(TString("_outputFileKey_d_.dat.ref"),
                                            TString("_outputFileKey_d_.dat"), TString("#DATE:"));
    CPPUNIT_ASSERT_EQUAL(nret, 0);

}

void TLauncherTest::testLauncherOutputFileKey_yhat_d()
{
    tds1 = new URANIE::DataServer::TDataServer("tds1", "my title 1");
    tds1->fileDataRead(FLOWRATEASCIIFILEUNIFORMDESIGN);
    // The reference input file
#ifdef WIN32
    TString sJDD1 = gSystem->pwd()+TString("\\flowrate_input_with_keys.in");
#else
    TString sJDD1 = gSystem->pwd() + TString("/flowrate_input_with_keys.in");
#endif  

    tds1->getAttribute("rw")->setFileKey(sJDD1, "Rw");
    tds1->getAttribute("r")->setFileKey(sJDD1, "R");
    tds1->getAttribute("tu")->setFileKey(sJDD1, "Tu");
    tds1->getAttribute("tl")->setFileKey(sJDD1, "Tl");
    tds1->getAttribute("hu")->setFileKey(sJDD1, "Hu");
    tds1->getAttribute("hl")->setFileKey(sJDD1, "Hl");
    tds1->getAttribute("l")->setFileKey(sJDD1, "L");
    tds1->getAttribute("kw")->setFileKey(sJDD1, "Kw");

    _foutKey = new URANIE::Launcher::TOutputFileKey(
	"_output_flowrate_withKey_.dat");
    _foutKey->addAttribute(ptrHat);
    _foutKey->addAttribute(ptrD);

    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("flowrate") + " -s -k");

    tc1->addOutputFile(_foutKey);

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);
    TString sCurrentDirectory = gSystem->pwd();

#ifdef WIN32
    tl1->setWorkingDirectory(Form("%s\\tmpLanceurUranie\\flowrate", sCurrentDirectory.Data()));
#else
    tl1->setWorkingDirectory(
	Form("%s/tmpLanceurUranie/flowrate", sCurrentDirectory.Data()));
#endif
    tl1->run("nointermed");
    tds1->exportData("_outputFileKey_yhat_d_.dat");

    Int_t nret = UranieTests::areFilesEqual(TString("_outputFileKey_yhat_d_.dat.ref"),
                                            TString("_outputFileKey_yhat_d_.dat"), TString("#DATE:"));
    CPPUNIT_ASSERT_EQUAL(nret, 0);

}

void TLauncherTest::testLauncherOutputFileKey_d_yhat()
{
    tds1 = new URANIE::DataServer::TDataServer("tds1", "my title 1");
    tds1->fileDataRead(FLOWRATEASCIIFILEUNIFORMDESIGN);
    // The reference input file
#ifdef WIN32
    TString sJDD1 = gSystem->pwd()+TString("\\flowrate_input_with_keys.in");
#else
    TString sJDD1 = gSystem->pwd() + TString("/flowrate_input_with_keys.in");
#endif  

    tds1->getAttribute("rw")->setFileKey(sJDD1, "Rw");
    tds1->getAttribute("r")->setFileKey(sJDD1, "R");
    tds1->getAttribute("tu")->setFileKey(sJDD1, "Tu");
    tds1->getAttribute("tl")->setFileKey(sJDD1, "Tl");
    tds1->getAttribute("hu")->setFileKey(sJDD1, "Hu");
    tds1->getAttribute("hl")->setFileKey(sJDD1, "Hl");
    tds1->getAttribute("l")->setFileKey(sJDD1, "L");
    tds1->getAttribute("kw")->setFileKey(sJDD1, "Kw");

    _foutKey = new URANIE::Launcher::TOutputFileKey(
	"_output_flowrate_withKey_.dat");
    _foutKey->addAttribute(ptrD);
    _foutKey->addAttribute(ptrHat);

    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("flowrate") + " -s -k");

    tc1->addOutputFile(_foutKey);

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);
    TString sCurrentDirectory = gSystem->pwd();

#ifdef WIN32
    tl1->setWorkingDirectory(Form("%s\\tmpLanceurUranie\\flowrate", sCurrentDirectory.Data()));
#else
    tl1->setWorkingDirectory(
	Form("%s/tmpLanceurUranie/flowrate", sCurrentDirectory.Data()));
#endif
    tl1->run("nointermed");
    tds1->exportData("_outputFileKey_d_yhat_.dat");
    Int_t nret = UranieTests::areFilesEqual(TString("_outputFileKey_d_yhat_.dat.ref"),
                                            TString("_outputFileKey_d_yhat_.dat"), TString("#DATE:"));
    CPPUNIT_ASSERT_EQUAL(nret, 0);

}

void TLauncherTest::testLauncherSeveralOutputFiles()
{
    tds1 = new URANIE::DataServer::TDataServer("tds1", "my title 1");
    tds1->fileDataRead(FLOWRATEASCIIFILEUNIFORMDESIGN);
    Int_t npat = tds1->getNPatterns();
    Int_t ninput = tds1->getNAttributes();
    Int_t noutput = 0;
    Int_t natt = ninput;
    // The reference input file
#ifdef WIN32
    TString sJDD1 = gSystem->pwd()+TString("\\flowrate_input_with_keys.in");
#else
    TString sJDD1 = gSystem->pwd() + TString("/flowrate_input_with_keys.in");
#endif  

    tds1->getAttribute("rw")->setFileKey(sJDD1, "Rw");
    tds1->getAttribute("r")->setFileKey(sJDD1, "R");
    tds1->getAttribute("tu")->setFileKey(sJDD1, "Tu");
    tds1->getAttribute("tl")->setFileKey(sJDD1, "Tl");
    tds1->getAttribute("hu")->setFileKey(sJDD1, "Hu");
    tds1->getAttribute("hl")->setFileKey(sJDD1, "Hl");
    tds1->getAttribute("l")->setFileKey(sJDD1, "L");
    tds1->getAttribute("kw")->setFileKey(sJDD1, "Kw");

    _foutKey = new URANIE::Launcher::TOutputFileKey(
	"_output_flowrate_withKey_.dat");
    _foutKey->addAttribute(ptrD);
    _foutKey->addAttribute(ptrHat);

    _foutRow = new URANIE::Launcher::TOutputFileRow(
	"_output_flowrate_withRow_.dat");
    _foutRow->addAttribute(TString("yhatbis"));

    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("flowrate") + " -s -k");

    tc1->addOutputFile(_foutKey);
    tc1->addOutputFile(_foutRow);

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);
    TString sCurrentDirectory = gSystem->pwd();

#ifdef WIN32
    tl1->setWorkingDirectory(Form("%s\\tmpLanceurUranie\\flowrate", sCurrentDirectory.Data()));
#else
    tl1->setWorkingDirectory(
	Form("%s/tmpLanceurUranie/flowrate", sCurrentDirectory.Data()));
#endif
    tl1->run("nointermed");
    natt += 3;
    noutput += 3;

    CPPUNIT_ASSERT_EQUAL(npat, tds1->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(natt, tds1->getNAttributes());
    CPPUNIT_ASSERT_EQUAL(ninput, tds1->getNInputAttributes());
    CPPUNIT_ASSERT_EQUAL(noutput, tds1->getNOutputAttributes());

    tds1->exportData("_outputFileKey_SeveralOutputFiles_.dat");
    Int_t nret = UranieTests::areFilesEqual(
	TString("_outputFileKey_SeveralOutputFiles_.dat.ref"),
	TString("_outputFileKey_SeveralOutputFiles_.dat"),
	TString("#DATE:"));
    CPPUNIT_ASSERT_EQUAL(nret, 0);

    TTree* tuple = (TNtupleD*) tds1->getTuple();
    TObjArray * Leaves = tuple->GetListOfLeaves();
    CPPUNIT_ASSERT(tds1->isAttribute("d"));
    CPPUNIT_ASSERT(tds1->isAttribute("yhat"));
    CPPUNIT_ASSERT(tds1->isAttribute("yhatbis"));
    CPPUNIT_ASSERT(Leaves->FindObject("d") != 0);
    CPPUNIT_ASSERT(Leaves->FindObject("yhat") != 0);
    CPPUNIT_ASSERT(Leaves->FindObject("yhatbis") != 0);

    CPPUNIT_ASSERT_DOUBLES_EQUAL(3602.045, tds1->getValue("d", 0),
				 DBL_DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(26.18019, tds1->getValue("yhat", 0),
				 DBL_DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getValue("yhat", 0),
				 tds1->getValue("yhatbis", 0), DBL_DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(1348.649, tds1->getValue("d", npat - 1),
				 DBL_DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(157.591, tds1->getValue("yhat", npat - 1),
				 DBL_DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(tds1->getValue("yhat", npat - 1),
				 tds1->getValue("yhatbis", npat - 1), DBL_DBL_PRECISION_COMPARE);

}

void TLauncherTest::testLauncherOutputFileKey_d_yhat_zWithDefaultValue()
{
    tds1 = new URANIE::DataServer::TDataServer("tds1", "my title 1");
    tds1->fileDataRead(FLOWRATEASCIIFILEUNIFORMDESIGN);
    // The reference input file
#ifdef WIN32
    TString sJDD1 = gSystem->pwd()+TString("\\flowrate_input_with_keys.in");
#else
    TString sJDD1 = gSystem->pwd() + TString("/flowrate_input_with_keys.in");
#endif  

    tds1->getAttribute("rw")->setFileKey(sJDD1, "Rw");
    tds1->getAttribute("r")->setFileKey(sJDD1, "R");
    tds1->getAttribute("tu")->setFileKey(sJDD1, "Tu");
    tds1->getAttribute("tl")->setFileKey(sJDD1, "Tl");
    tds1->getAttribute("hu")->setFileKey(sJDD1, "Hu");
    tds1->getAttribute("hl")->setFileKey(sJDD1, "Hl");
    tds1->getAttribute("l")->setFileKey(sJDD1, "L");
    tds1->getAttribute("kw")->setFileKey(sJDD1, "Kw");

    _foutKey = new URANIE::Launcher::TOutputFileKey(
	"_output_flowrate_withKey_.dat");
    _foutKey->addAttribute(ptrD);
    _foutKey->addAttribute(ptrHat);

    URANIE::DataServer::TAttribute *ptrZ = new URANIE::DataServer::TAttribute("z");
    ptrZ->setDefaultValue(-9.7531);
    ptrZ->delShare();
    _foutKey->addAttribute(ptrZ);

    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("flowrate") + " -s -k");

    tc1->addOutputFile(_foutKey);

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);
    tl1->run("nointermed");

    tds1->exportData("_outputFileKey_d_yhat_z_.dat");

    Int_t nret = UranieTests::areFilesEqual(TString("_outputFileKey_d_yhat_z_.dat.ref"),
                                            TString("_outputFileKey_d_yhat_z_.dat"), TString("#DATE:"));
    CPPUNIT_ASSERT_EQUAL(nret, 0);

}

void TLauncherTest::testLauncherOutputFileKey_d_yhat_zWithoutDefaultValue()
{
    tds1 = new URANIE::DataServer::TDataServer("tds1", "my title 1");
    tds1->fileDataRead(FLOWRATEASCIIFILEUNIFORMDESIGN);
    // The reference input file
#ifdef WIN32
    TString sJDD1 = gSystem->pwd()+TString("\\flowrate_input_with_keys.in");
#else
    TString sJDD1 = gSystem->pwd() + TString("/flowrate_input_with_keys.in");
#endif  

    tds1->getAttribute("rw")->setFileKey(sJDD1, "Rw");
    tds1->getAttribute("r")->setFileKey(sJDD1, "R");
    tds1->getAttribute("tu")->setFileKey(sJDD1, "Tu");
    tds1->getAttribute("tl")->setFileKey(sJDD1, "Tl");
    tds1->getAttribute("hu")->setFileKey(sJDD1, "Hu");
    tds1->getAttribute("hl")->setFileKey(sJDD1, "Hl");
    tds1->getAttribute("l")->setFileKey(sJDD1, "L");
    tds1->getAttribute("kw")->setFileKey(sJDD1, "Kw");

    _foutKey = new URANIE::Launcher::TOutputFileKey(
	"_output_flowrate_withKey_.dat");
    _foutKey->addAttribute(ptrD);
    _foutKey->addAttribute(ptrHat);

    URANIE::DataServer::TAttribute *ptrZ = new URANIE::DataServer::TAttribute("z");
    ptrZ->delShare();
    _foutKey->addAttribute(ptrZ);

    cout << " ***** _foutKey YHat Share[" << ptrHat->getShare() << "]" << endl;
    cout << " ***** _foutKey    d Share[" << ptrHat->getShare() << "]" << endl;
    cout << " ***** _foutKey    Z Share[" << ptrZ->getShare() << "]" << endl;

    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("flowrate") + " -s -k");

    tc1->addOutputFile(_foutKey);

    cout << " ***** tc1 YHat Share[" << ptrHat->getShare() << "]" << endl;
    cout << " ***** tc1    d Share[" << ptrHat->getShare() << "]" << endl;
    cout << " ***** tc1    Z Share[" << ptrZ->getShare() << "]" << endl;

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);
    TString sCurrentDirectory = gSystem->pwd();

    cout << " ***** tl1 YHat Share[" << ptrHat->getShare() << "]" << endl;
    cout << " ***** tl1    d Share[" << ptrHat->getShare() << "]" << endl;
    cout << " ***** tl1    Z Share[" << ptrZ->getShare() << "]" << endl;

#ifdef WIN32
    tl1->setWorkingDirectory(Form("%s\\tmpLanceurUranie\\flowrate", sCurrentDirectory.Data()));
#else
    tl1->setWorkingDirectory(
	Form("%s/tmpLanceurUranie/flowrate", sCurrentDirectory.Data()));
#endif

    cout << " ***** tl1 before::run YHat Share[" << ptrHat->getShare() << "]"
	 << endl;
    cout << " ***** tl1 before::run    d Share[" << ptrHat->getShare() << "]"
	 << endl;
    cout << " ***** tl1 before::run   Z Share[" << ptrZ->getShare() << "]"
	 << endl;

    tl1->run("nointermed");

    cout << " ***** tl1 after::run YHat Share[" << ptrHat->getShare() << "]"
	 << endl;
    cout << " ***** tl1 after::run    d Share[" << ptrHat->getShare() << "]"
	 << endl;
    cout << " ***** tl1after:: run   Z Share[" << ptrZ->getShare() << "]"
	 << endl;

    tds1->exportData("_outputFileKey_d_yhat_zWithoutDefaultValue_.dat");
    Int_t nret = UranieTests::areFilesEqual(
	TString("_outputFileKey_d_yhat_zWithoutDefaultValue_.dat.ref"),
	TString("_outputFileKey_d_yhat_zWithoutDefaultValue_.dat"),
	TString("#DATE:"));
    CPPUNIT_ASSERT_EQUAL(0, nret);

}

void TLauncherTest::testLauncherHighNbSamplesWithoutSave()
{
    setUp3();
    tl3->run("nointermed");
    TString workingDirectory = tl3->getWorkingDirectory();
#ifdef WIN32
    void * openDirectoryResult = gSystem->OpenDirectory( Form("%s\\UranieLauncher_1",workingDirectory.Data()) );
#else
    void * openDirectoryResult = gSystem->OpenDirectory(
	Form("%s/UranieLauncher_1", workingDirectory.Data()));
#endif
    CPPUNIT_ASSERT(openDirectoryResult != 0);
    gSystem->FreeDirectory(openDirectoryResult);

    Int_t nbReps = 0;
    void *dir = gSystem->OpenDirectory(workingDirectory);
    const char* subdir;
    while ((subdir = gSystem->GetDirEntry(dir)))
        nbReps++;
    gSystem->FreeDirectory(dir);

    UranieTests::removeFolder(_workingTestPath.Data(),workingDirectory.Data());

    // On retire 3 au nombre de r�pertoires car on ne consid�re pas UranieResults, ni le r�pertoire courant et le r�pertoire parent.
    CPPUNIT_ASSERT_EQUAL(nbReps - 3, 1);
}

void TLauncherTest::testLauncherHighNbSamplesWithSave1()
{
    setUp3();
    tl3->setSave(980);
    tl3->setClean();
    tl3->run("nointermed");
    TString workingDirectory = tl3->getWorkingDirectory();
#ifdef WIN32
    void * openDirectoryResult = gSystem->OpenDirectory( Form("%s\\UranieLauncher_981",workingDirectory.Data()) );
#else
    void * openDirectoryResult = gSystem->OpenDirectory(
	Form("%s/UranieLauncher_981", workingDirectory.Data()));
#endif
    CPPUNIT_ASSERT(openDirectoryResult == 0);
    gSystem->FreeDirectory(openDirectoryResult);

    Int_t nbReps = 0;
    void *dir = gSystem->OpenDirectory(workingDirectory);
    const char* subdir;
    while ((subdir = gSystem->GetDirEntry(dir)))
        nbReps++;
    gSystem->FreeDirectory(dir);

    UranieTests::removeFolder(_workingTestPath.Data(),workingDirectory.Data());

    // On retire 4 au nombre de r�pertoires car on ne consid�re pas UranieResults, UranieLauncher_9999, le r�pertoire courant et le r�pertoire parent.
    CPPUNIT_ASSERT_EQUAL(nbReps - 4, 980);
}

void TLauncherTest::testLauncherHighNbSamplesWithSave2()
{
    setUp3();
    tl3->setSave(100);
    tl3->setClean();
    tl3->run("nointermed");
    TString workingDirectory = tl3->getWorkingDirectory();
#ifdef WIN32
    void * openDirectoryResult = gSystem->OpenDirectory( Form("%s\\UranieLauncher_101",workingDirectory.Data()) );
#else
    void * openDirectoryResult = gSystem->OpenDirectory(
	Form("%s/UranieLauncher_101", workingDirectory.Data()));
#endif
    CPPUNIT_ASSERT(openDirectoryResult == 0);
    gSystem->FreeDirectory(openDirectoryResult);

    Int_t nbReps = 0;
    void *dir = gSystem->OpenDirectory(workingDirectory);
    const char* subdir;
    while ((subdir = gSystem->GetDirEntry(dir)))
        nbReps++;
    gSystem->FreeDirectory(dir);

    UranieTests::removeFolder(_workingTestPath.Data(),workingDirectory.Data());

    CPPUNIT_ASSERT_EQUAL(nbReps - 4, 100);
}

void TLauncherTest::testLauncherHighNbSamplesWithSave3()
{
    setUp3();
    tl3->setSave(-1);
    tl3->setClean();
    tl3->run("nointermed");
    TString workingDirectory = tl3->getWorkingDirectory();
#ifdef WIN32
    void * openDirectoryResult = gSystem->OpenDirectory( Form("%s\\UranieLauncher_1201",workingDirectory.Data()) );
#else
    void * openDirectoryResult = gSystem->OpenDirectory(
	Form("%s/UranieLauncher_1201", workingDirectory.Data()));
#endif
    CPPUNIT_ASSERT(openDirectoryResult == 0);
    gSystem->FreeDirectory(openDirectoryResult);

    Int_t nbReps = 0;
    void *dir = gSystem->OpenDirectory(workingDirectory);
    const char* subdir;
    while ((subdir = gSystem->GetDirEntry(dir)))
        nbReps++;
    gSystem->FreeDirectory(dir);
    
    UranieTests::removeFolder(_workingTestPath.Data(),workingDirectory.Data());
        
// On retire 3 au nombre de r�pertoires car on ne consid�re pas UranieResults, le r�pertoire courant et le r�pertoire parent.
    CPPUNIT_ASSERT_EQUAL(nbReps - 3, 1200);
}

void TLauncherTest::testLauncherWithNotModifiedInputFiles()
{
    tds1 = new URANIE::DataServer::TDataServer("tds1", "my title 1");
    tds1->fileDataRead(FLOWRATEASCIIFILEUNIFORMDESIGN);
    // The reference input file
#ifdef WIN32
    TString sJDD1 = gSystem->pwd()+TString("\\flowrate_input_with_keys.in");
#else
    TString sJDD1 = gSystem->pwd() + TString("/flowrate_input_with_keys.in");
#endif  

    tds1->getAttribute("rw")->setFileKey(sJDD1, "Rw");
    tds1->getAttribute("r")->setFileKey(sJDD1, "R");
    tds1->getAttribute("tu")->setFileKey(sJDD1, "Tu");
    tds1->getAttribute("tl")->setFileKey(sJDD1, "Tl");
    tds1->getAttribute("hu")->setFileKey(sJDD1, "Hu");
    tds1->getAttribute("hl")->setFileKey(sJDD1, "Hl");
    tds1->getAttribute("l")->setFileKey(sJDD1, "L");
    tds1->getAttribute("kw")->setFileKey(sJDD1, "Kw");
    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("flowrate") + " -s -k");

    _foutRow = new URANIE::Launcher::TOutputFileRow(
	"_output_flowrate_withRow_.dat");
    _foutRow->addAttribute(ptrHat);

    tc1->addOutputFile(_foutRow);
    tc1->setReferenceDirectory(gSystem->pwd());
    // relative path
    tc1->addInputFile("flowrate_input_with_flags.in");
    // absolute path
    char *foo = gSystem->ConcatFileName(gSystem->pwd(),"flowrate_input_with_xml.in") ;
    tc1->addInputFile( foo );
    delete[] foo;
    
    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);
    TString sCurrentDirectory = gSystem->pwd();

#ifdef WIN32
    tl1->setWorkingDirectory(Form("%s\\tmpLanceurUranie\\flowrate", sCurrentDirectory.Data()));
#else
    tl1->setWorkingDirectory(
	Form("%s/tmpLanceurUranie/flowrate", sCurrentDirectory.Data()));
#endif
    tl1->run("nointermed");

    TString sinputfilekey = Form(
	"%s/UranieLauncher_1/flowrate_input_with_keys.in",
	tl1->getWorkingDirectory().Data());
    TString sinputfileflag = Form(
	"%s/UranieLauncher_1/flowrate_input_with_flags.in",
	tl1->getWorkingDirectory().Data());
    TString sinputfilexml = Form(
	"%s/UranieLauncher_1/flowrate_input_with_xml.in",
	tl1->getWorkingDirectory().Data());

    TString soutput1 = Form("%s/UranieLauncher_1/_output_flowrate_withKey_.dat",
			    tl1->getWorkingDirectory().Data());
    TString soutput2 = Form("%s/UranieLauncher_1/_output_flowrate_withRow_.dat",
			    tl1->getWorkingDirectory().Data());

    Int_t nfound = ((gSystem->AccessPathName(sinputfilekey, kFileExists)) ? 1 : 0);
    CPPUNIT_ASSERT(nfound == 0);

    nfound = ((gSystem->AccessPathName(sinputfileflag, kFileExists))? 1 : 0);
    CPPUNIT_ASSERT(nfound == 0);

    nfound = ((gSystem->AccessPathName(sinputfilexml, kFileExists)) ? 1 : 0);
    CPPUNIT_ASSERT(nfound == 0);

    nfound = ((gSystem->AccessPathName(soutput1, kFileExists)) ? 1 : 0);
    CPPUNIT_ASSERT(nfound == 0);

    nfound = ((gSystem->AccessPathName(soutput2, kFileExists)) ? 1 : 0);
    CPPUNIT_ASSERT(nfound == 0);
}

void TLauncherTest::testLauncherAttributeWithSeveralInputFiles()
{
    tds1 = new URANIE::DataServer::TDataServer("tds1", "my title 1");
    tds1->fileDataRead(FLOWRATEASCIIFILEUNIFORMDESIGN);
    // The reference input file with keys
#ifdef WIN32
    TString sJDDKey = gSystem->pwd()+TString("\\flowrate_input_with_keys.in");
#else
    TString sJDDKey = gSystem->pwd() + TString("/flowrate_input_with_keys.in");
#endif  

    tds1->getAttribute("rw")->setFileKey(sJDDKey, "Rw");
    tds1->getAttribute("r")->setFileKey(sJDDKey, "R");
    tds1->getAttribute("tu")->setFileKey(sJDDKey, "Tu");
    tds1->getAttribute("tl")->setFileKey(sJDDKey, "Tl");
    tds1->getAttribute("hu")->setFileKey(sJDDKey, "Hu");
    tds1->getAttribute("hl")->setFileKey(sJDDKey, "Hl");
    tds1->getAttribute("l")->setFileKey(sJDDKey, "L");
    tds1->getAttribute("kw")->setFileKey(sJDDKey, "Kw");

    // The reference input file with flags
#ifdef WIN32
    TString sJDDFlag = gSystem->pwd()+TString("\\flowrate_input_with_flags.in");
#else
    TString sJDDFlag = gSystem->pwd()
	+ TString("/flowrate_input_with_flags.in");
#endif  

    tds1->getAttribute("rw")->setFileKey(sJDDFlag, "@rw@", "",
					 URANIE::DataServer::TAttributeFileKey::kFlag);
    tds1->getAttribute("r")->setFileKey(sJDDFlag, "@r@", "",
					URANIE::DataServer::TAttributeFileKey::kFlag);
    tds1->getAttribute("tu")->setFileKey(sJDDFlag, "@tu@", "",
					 URANIE::DataServer::TAttributeFileKey::kFlag);
    tds1->getAttribute("tl")->setFileKey(sJDDFlag, "@tl@", "",
					 URANIE::DataServer::TAttributeFileKey::kFlag);
    tds1->getAttribute("hu")->setFileKey(sJDDFlag, "@hu@", "",
					 URANIE::DataServer::TAttributeFileKey::kFlag);
    tds1->getAttribute("hl")->setFileKey(sJDDFlag, "@hl@", "",
					 URANIE::DataServer::TAttributeFileKey::kFlag);
    tds1->getAttribute("l")->setFileKey(sJDDFlag, "@l@", "",
					URANIE::DataServer::TAttributeFileKey::kFlag);
    tds1->getAttribute("kw")->setFileKey(sJDDFlag, "@kw@", "",
					 URANIE::DataServer::TAttributeFileKey::kFlag);

    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("flowrate") + " -s -k");

    _foutKey = new URANIE::Launcher::TOutputFileKey(
	"_output_flowrate_withKey_.dat");
    _foutKey->addAttribute(ptrHat);

    tc1->addOutputFile(_foutKey);

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);
    TString sCurrentDirectory = gSystem->pwd();

#ifdef WIN32
    tl1->setWorkingDirectory(Form("%s\\tmpLanceurUranie\\flowrate", sCurrentDirectory.Data()));
#else
    tl1->setWorkingDirectory(
	Form("%s/tmpLanceurUranie/flowrate", sCurrentDirectory.Data()));
#endif
    tl1->run("nointermed");
    CPPUNIT_ASSERT_EQUAL(tds1->getTuple()->GetNbranches(), 12);

#ifdef WIN32
    TString sinputfilekey = Form("%s\\UranieLauncher_1\\flowrate_input_with_keys.in", tl1->getWorkingDirectory().Data());
    TString sinputfileflag = Form("%s\\UranieLauncher_1\\flowrate_input_with_flags.in", tl1->getWorkingDirectory().Data());
    TString soutput1 = Form("%s\\UranieLauncher_1\\_output_flowrate_withKey_.dat", tl1->getWorkingDirectory().Data());
    TString soutput2 = Form("%s\\UranieLauncher_1\\_output_flowrate_withRow_.dat", tl1->getWorkingDirectory().Data());
#else
    TString sinputfilekey = Form(
	"%s/UranieLauncher_1/flowrate_input_with_keys.in",
	tl1->getWorkingDirectory().Data());
    TString sinputfileflag = Form(
	"%s/UranieLauncher_1/flowrate_input_with_flags.in",
	tl1->getWorkingDirectory().Data());
    TString soutput1 = Form("%s/UranieLauncher_1/_output_flowrate_withKey_.dat",
			    tl1->getWorkingDirectory().Data());
    TString soutput2 = Form("%s/UranieLauncher_1/_output_flowrate_withRow_.dat",
			    tl1->getWorkingDirectory().Data());
#endif

    Int_t nfound = 0;
    gSystem->AccessPathName(sinputfilekey, kFileExists) ? nfound = 1 : nfound =
	0;
    CPPUNIT_ASSERT(nfound == 0);

    gSystem->AccessPathName(sinputfileflag, kFileExists) ? nfound = 1 : nfound =
	0;
    CPPUNIT_ASSERT(nfound == 0);

    gSystem->AccessPathName(soutput1, kFileExists) ? nfound = 1 : nfound = 0;
    CPPUNIT_ASSERT(nfound == 0);

    gSystem->AccessPathName(soutput2, kFileExists) ? nfound = 1 : nfound = 0;
    CPPUNIT_ASSERT(nfound == 0);
}

void TLauncherTest::testLauncherWithNoSpecifiedWorkingDirectory()
{
    UranieTests::removeFolder(gSystem->pwd(), "URANIE");

    tds1 = new URANIE::DataServer::TDataServer("tds1", "my title 1");
    tds1->fileDataRead(FLOWRATEASCIIFILEUNIFORMDESIGN);
    // The reference input file with keys
#ifdef WIN32
    TString sJDDKey = gSystem->pwd()+TString("\\flowrate_input_with_keys.in");
#else
    TString sJDDKey = gSystem->pwd() + TString("/flowrate_input_with_keys.in");
#endif  

    tds1->getAttribute("rw")->setFileKey(sJDDKey, "Rw");
    tds1->getAttribute("r")->setFileKey(sJDDKey, "R");
    tds1->getAttribute("tu")->setFileKey(sJDDKey, "Tu");
    tds1->getAttribute("tl")->setFileKey(sJDDKey, "Tl");
    tds1->getAttribute("hu")->setFileKey(sJDDKey, "Hu");
    tds1->getAttribute("hl")->setFileKey(sJDDKey, "Hl");
    tds1->getAttribute("l")->setFileKey(sJDDKey, "L");
    tds1->getAttribute("kw")->setFileKey(sJDDKey, "Kw");

#ifdef WIN32
    tc1 = new URANIE::Launcher::TCode(tds1, "flowrate.exe -s -k");
#else
    tc1 = new URANIE::Launcher::TCode(tds1,
				      Form("%s/%s", _workingTestPath.Data(),
					   "../../teST/coDE/flowREATE/flowrate -s -k"));
#endif

    _foutRow = new URANIE::Launcher::TOutputFileRow(
	"_output_flowrate_withRow_.dat");
    _foutRow->addAttribute(ptrHat);

    tc1->addOutputFile(_foutRow);

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);

    tl1->setSave();
    tl1->setClean();
    tl1->run("nointermed");

#ifdef WIN32
    TString workingDirectory = gSystem->pwd() + TString("\\URANIE");
#else
    TString workingDirectory = gSystem->pwd() + TString("/URANIE");
#endif
    void * openDirectoryResult = gSystem->OpenDirectory(
	workingDirectory.Data());
    CPPUNIT_ASSERT(openDirectoryResult != 0);
    gSystem->FreeDirectory(openDirectoryResult);

    Int_t nfound = 0;
#ifdef WIN32
    if (gSystem->AccessPathName( Form("%s\\UranieResults\\uranielauncher.root",workingDirectory.Data()) )?nfound=1:nfound=0)
#else
	if (gSystem->AccessPathName(
		Form("%s/UranieResults/uranielauncher.root",
		     workingDirectory.Data())) ? (nfound = 1) : (nfound = 0))
#endif
	    CPPUNIT_ASSERT(nfound == 0);
}

void TLauncherTest::testLauncherWithCurrentDirectorySpecifiedAsWorkingDirectory()
{
    UranieTests::removeFolder(gSystem->pwd(), "URANIE");

    tds1 = new URANIE::DataServer::TDataServer("tds1", "my title 1");
    tds1->fileDataRead(FLOWRATEASCIIFILEUNIFORMDESIGN);
    // The reference input file with keys
#ifdef WIN32
    TString sJDDKey = gSystem->pwd()+TString("\\flowrate_input_with_keys.in");
#else
    TString sJDDKey = gSystem->pwd() + TString("/flowrate_input_with_keys.in");
#endif  

    tds1->getAttribute("rw")->setFileKey(sJDDKey, "Rw");
    tds1->getAttribute("r")->setFileKey(sJDDKey, "R");
    tds1->getAttribute("tu")->setFileKey(sJDDKey, "Tu");
    tds1->getAttribute("tl")->setFileKey(sJDDKey, "Tl");
    tds1->getAttribute("hu")->setFileKey(sJDDKey, "Hu");
    tds1->getAttribute("hl")->setFileKey(sJDDKey, "Hl");
    tds1->getAttribute("l")->setFileKey(sJDDKey, "L");
    tds1->getAttribute("kw")->setFileKey(sJDDKey, "Kw");

#ifdef WIN32
    tc1 = new URANIE::Launcher::TCode(tds1, "flowrate.exe -s -k");
#else
    tc1 = new URANIE::Launcher::TCode(tds1,
				      Form("%s/%s", _workingTestPath.Data(),
					   "../../teST/coDE/flowREATE/flowrate -s -k"));
#endif

    _foutRow = new URANIE::Launcher::TOutputFileRow(
	"_output_flowrate_withRow_.dat");
    _foutRow->addAttribute(ptrHat);

    tc1->addOutputFile(_foutRow);

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);
    tl1->setWorkingDirectory(gSystem->pwd());

    tl1->setSave();
    tl1->setClean();
    tl1->run("nointermed");

#ifdef WIN32
    TString workingDirectory = gSystem->pwd() + TString("\\URANIE");
#else
    TString workingDirectory = gSystem->pwd() + TString("/URANIE");
#endif
    void * openDirectoryResult = gSystem->OpenDirectory(
	tl1->getWorkingDirectory().Data());
    CPPUNIT_ASSERT(openDirectoryResult != 0);
    gSystem->FreeDirectory(openDirectoryResult);

    Int_t nfound = 0;
    if (gSystem->AccessPathName(
            Form("%s/UranieResults/uranielauncher.root",
		 workingDirectory.Data())) ? (nfound = 1) : (nfound = 0))
        CPPUNIT_ASSERT(nfound == 0);
}

void TLauncherTest::testLauncherFailure()
{

#ifdef WIN32
    TString workingDirectory = Form("%s\\%s", _workingTestPath.Data(), "tmpLanceurUranie");      
#else
    TString workingDirectory = Form("%s/%s", _workingTestPath.Data(), "tmpLanceurUranie");      
#endif
    UranieTests::removeFolder(_workingTestPath.Data(), "tmpLanceurUranie");      

    // Define the DataServer
    tds1 = new URANIE::DataServer::TDataServer("tdsflowrate",
					       "DataBase flowrate");
    tds1->fileDataRead("flowrate200Samples.dat");
    // The reference input file with keys
#ifdef WIN32
    TString sJDDKey = gSystem->pwd()+TString("\\flowrate_input_with_keys.in");
#else
    TString sJDDKey = gSystem->pwd() + TString("/flowrate_input_with_keys.in");
#endif  

    tds1->getAttribute("rw")->setFileKey(sJDDKey, "Rw");
    tds1->getAttribute("r")->setFileKey(sJDDKey, "R");
    tds1->getAttribute("tu")->setFileKey(sJDDKey, "Tu");
    tds1->getAttribute("tl")->setFileKey(sJDDKey, "Tl");
    tds1->getAttribute("hu")->setFileKey(sJDDKey, "Hu");
    tds1->getAttribute("hl")->setFileKey(sJDDKey, "Hl");
    tds1->getAttribute("l")->setFileKey(sJDDKey, "L");
    tds1->getAttribute("kw")->setFileKey(sJDDKey, "Kw");

    // The output file of the code
    _foutRow = new URANIE::Launcher::TOutputFileRow(
	"_output_flowrate_withRow_.dat");
    ptrHat->setDefaultValue(-200.0);
    _foutRow->addAttribute(ptrHat);

    // Create a TCode object
    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("flowrate") + " -s -kf");

    // Add the output file where to find the output attributes
    tc1->addOutputFile(_foutRow);

    // Create a TLauncher object from a TDataServer and a TCode objects
    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);

    tl1->setWorkingDirectory(workingDirectory);
    tl1->run("nointermed");
    tds1->setSelect("yhat<=0");
    Int_t nFailure = tds1->getNPatterns();

    CPPUNIT_ASSERT_EQUAL(nFailure, 54);

    Int_t nfound = 0;
    // The last pattern gives a failure. "_output_flowrate_withRow_.dat" doesn't have to exist.
#ifdef WIN32
    if (gSystem->AccessPathName( Form("%s\\UranieLauncher_1\\_output_flowrate_withRow_.dat",workingDirectory.Data()) )?nfound=1:nfound=0)
#else
	if (gSystem->AccessPathName(
		Form("%s/UranieLauncher_1/_output_flowrate_withRow_.dat",
		     workingDirectory.Data())) ? (nfound = 1) : (nfound = 0))
#endif
	    CPPUNIT_ASSERT(nfound == 1);

    UranieTests::removeFolder(_workingTestPath.Data(),workingDirectory.Data());

}

void TLauncherTest::testlaunchCodeAttributeNameDifferentKey()
{
    tds1 = new URANIE::DataServer::TDataServer("tds1", "my title 1");
    tds1->fileDataRead(FLOWRATEASCIIFILEUNIFORMDESIGN);

    Int_t natt = tds1->getNAttributes();
    Int_t npat = tds1->getNPatterns();

    // The reference input file
#ifdef WIN32
    TString sJDD1 = gSystem->pwd()+TString("\\flowrate_input_with_keys.in");
#else
    TString sJDD1 = gSystem->pwd() + TString("/flowrate_input_with_keys.in");
#endif  
    tds1->getAttribute("rw")->setFileKey(sJDD1, "Rw");
    tds1->getAttribute("r")->setFileKey(sJDD1, "R");
    tds1->getAttribute("hu")->setFileKey(sJDD1, "Hu");
    tds1->getAttribute("hl")->setFileKey(sJDD1, "Hl");

#ifdef WIN32
    tc1 = new URANIE::Launcher::TCode(tds1, "flowrate.exe -s -k");
#else
    tc1 = new URANIE::Launcher::TCode(tds1,
				      Form("%s/%s", _workingTestPath.Data(),
					   "../../teST/coDE/flowREATE/flowrate -s -k"));
#endif

    TString sOutputFileName = TString("_output2_.dat");
    _foutKey = new URANIE::Launcher::TOutputFileKey(sOutputFileName);
    URANIE::DataServer::TAttribute *atty = new URANIE::DataServer::TAttribute("y");
    atty->setFileKey(sOutputFileName, "yhat");
    atty->delShare();
    _foutKey->addAttribute(atty);
    _foutKey->addAttribute(ptrD);

    tc1->addOutputFile(_foutKey);

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);

#ifdef WIN32
    tl1->setWorkingDirectory(gSystem->pwd() + TString("\\tmpLanceurUranie\\flowrate"));
#else
    tl1->setWorkingDirectory(
	gSystem->pwd() + TString("/tmpLanceurUranie/flowrate"));
#endif
    tl1->run("nointermed");

    CPPUNIT_ASSERT_EQUAL(npat, tds1->getNPatterns());
    CPPUNIT_ASSERT_EQUAL(natt + 2, tds1->getNAttributes());

}

////////////////////////////////
///                          ///
///      levelE              ///
///                          ///
////////////////////////////////
#ifdef WITH_LEVELE

void TLauncherTest::setUpLevelE()
{
    tds1 = new URANIE::DataServer::TDataServer("tds1", "levelE usecase");
    tds1->fileDataRead(LEVELEDATASET);

    TString sJDD = "levelE_input_with_values_rows.in";

    tds1->getAttribute("t")->setFileKey(sJDD, "", "%e",URANIE::DataServer::TAttributeFileKey::kNewRow);
    tds1->getAttribute("kl")->setFileKey(sJDD,"", "%e",URANIE::DataServer::TAttributeFileKey::kNewRow);
    tds1->getAttribute("kc")->setFileKey(sJDD, "", "%e",URANIE::DataServer::TAttributeFileKey::kNewRow);
    tds1->getAttribute("v1")->setFileKey(sJDD, "", "%e",URANIE::DataServer::TAttributeFileKey::kNewRow);
    tds1->getAttribute("l1")->setFileKey(sJDD, "", "%e",URANIE::DataServer::TAttributeFileKey::kNewRow);
    tds1->getAttribute("r1")->setFileKey(sJDD, "", "%e",URANIE::DataServer::TAttributeFileKey::kNewRow);
    tds1->getAttribute("rc1")->setFileKey(sJDD, "", "%e",URANIE::DataServer::TAttributeFileKey::kNewRow);
    tds1->getAttribute("v2")->setFileKey(sJDD, "", "%e",URANIE::DataServer::TAttributeFileKey::kNewRow);
    tds1->getAttribute("l2")->setFileKey(sJDD, "", "%e",URANIE::DataServer::TAttributeFileKey::kNewRow);
    tds1->getAttribute("r2")->setFileKey(sJDD, "", "%e",URANIE::DataServer::TAttributeFileKey::kNewRow);
    tds1->getAttribute("rc2")->setFileKey(sJDD, "", "%e",URANIE::DataServer::TAttributeFileKey::kNewRow);
    tds1->getAttribute("w")->setFileKey(sJDD, "", "%e",URANIE::DataServer::TAttributeFileKey::kNewRow);
}

void TLauncherTest::testLauncherLevelEOutputFileDataServer()
{
    setUpLevelE();

    _foutDataServer = new URANIE::Launcher::TOutputFileDataServer("_output_levelE_withDataServer_.dat");
    _foutDataServer->addAttribute( new URANIE::DataServer::TAttribute("z",URANIE::DataServer::TAttribute::kVector) );
    _foutDataServer->addAttribute( new URANIE::DataServer::TAttribute("tps",URANIE::DataServer::TAttribute::kVector) );
    _foutDataServer->addAttribute( new URANIE::DataServer::TAttribute("y",URANIE::DataServer::TAttribute::kVector) );

#ifdef WIN32
    tc1 = new URANIE::Launcher::TCode(tds1, "levele.exe");
#else
    tc1 = new URANIE::Launcher::TCode(tds1, Form("%s/%s",_workingTestPath.Data(),"../../teST/coDE/levelE/levele 2> /dev/null "));
#endif

    tc1->addOutputFile( _foutDataServer );

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);
    tl1->run("nointermed");

    checkLauncherLevelEResults("_outputFile_levelE_TDS_.dat");
    
}

#ifdef WITH_LIBXML2
void TLauncherTest::testLauncherLevelEOutputFileXML()
{
    setUpLevelE();

    _foutXML = new URANIE::Launcher::TOutputFileXML("_output_levelE_.xml");
    _foutXML->addAttribute(new URANIE::DataServer::TAttribute("y",URANIE::DataServer::TAttribute::kVector), "/levele/theDose/values", URANIE::DataServer::TAttributeFileKey::kXMLField);
    _foutXML->addAttribute(new URANIE::DataServer::TAttribute("tps",URANIE::DataServer::TAttribute::kVector), "/levele/times/values", URANIE::DataServer::TAttributeFileKey::kXMLField);
    _foutXML->addAttribute(new URANIE::DataServer::TAttribute("z",URANIE::DataServer::TAttribute::kVector), "/levele/z/values", URANIE::DataServer::TAttributeFileKey::kXMLField);
  
#ifdef WIN32
    tc1 = new URANIE::Launcher::TCode(tds1, "levele.exe");
#else
    tc1 = new URANIE::Launcher::TCode(tds1, Form("%s/%s",_workingTestPath.Data(),"../../teST/coDE/levelE/levele 2> /dev/null "));
#endif

    tc1->addOutputFile( _foutXML );

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);
    tl1->run("nointermed");

    checkLauncherLevelEResults("_outputFile_levelE_XML_.dat");
    
}
#endif

void TLauncherTest::testLauncherLevelEOutputFileColumn()
{
    setUpLevelE();

    _foutCol = new URANIE::Launcher::TOutputFileColumn("_output_levelE_withColumn_.dat");
    _foutCol->addAttribute(new URANIE::DataServer::TAttribute("tps",URANIE::DataServer::TAttribute::kVector));
    _foutCol->addAttribute(new URANIE::DataServer::TAttribute("y",URANIE::DataServer::TAttribute::kVector));
    _foutCol->addAttribute(new URANIE::DataServer::TAttribute("z",URANIE::DataServer::TAttribute::kVector));

#ifdef WIN32
    tc1 = new URANIE::Launcher::TCode(tds1, "levele.exe");
#else
    tc1 = new URANIE::Launcher::TCode(tds1, Form("%s/%s",_workingTestPath.Data(),"../../teST/coDE/levelE/levele 2> /dev/null "));
#endif

    tc1->addOutputFile( _foutCol );

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
//    tl1->setSave(-1);
    tl1->setDrawProgressBar(kFALSE);
    tl1->run("nointermed");

    checkLauncherLevelEResults("_outputFile_levelE_cols_.dat");

}

void TLauncherTest::testLauncherLevelEOutputFileRow()
{
    setUpLevelE();

    _foutRow = new URANIE::Launcher::TOutputFileRow("_output_levelE_withRow_.dat");
    _foutRow->addAttribute(new URANIE::DataServer::TAttribute("tps",URANIE::DataServer::TAttribute::kVector));
    _foutRow->addAttribute(new URANIE::DataServer::TAttribute("y",URANIE::DataServer::TAttribute::kVector));
    _foutRow->addAttribute(new URANIE::DataServer::TAttribute("z",URANIE::DataServer::TAttribute::kVector));

#ifdef WIN32
    tc1 = new URANIE::Launcher::TCode(tds1, "levele.exe");
#else
    tc1 = new URANIE::Launcher::TCode(tds1, Form("%s/%s",_workingTestPath.Data(),"../../teST/coDE/levelE/levele 2> /dev/null "));
#endif

    tc1->addOutputFile( _foutRow );

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);
    tl1->run("nointermed");

    checkLauncherLevelEResults("_outputFile_levelE_rows_.dat");

}

void TLauncherTest::testLauncherLevelEOutputFileRowMultithread()
{
    setUpLevelE();

    _foutRow = new URANIE::Launcher::TOutputFileRow("_output_levelE_withRow_.dat");
    _foutRow->addAttribute(new URANIE::DataServer::TAttribute("tps",URANIE::DataServer::TAttribute::kVector));
    _foutRow->addAttribute(new URANIE::DataServer::TAttribute("y",URANIE::DataServer::TAttribute::kVector));
    _foutRow->addAttribute(new URANIE::DataServer::TAttribute("z",URANIE::DataServer::TAttribute::kVector));

#ifdef WIN32
    tc1 = new URANIE::Launcher::TCode(tds1, "levele.exe");
#else
    tc1 = new URANIE::Launcher::TCode(tds1, Form("%s/%s",_workingTestPath.Data(),"../../teST/coDE/levelE/levele 2> /dev/null "));
#endif

    tc1->addOutputFile( _foutRow );

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1, "multiquiet");
    tl1->setDrawProgressBar(kFALSE);
    tl1->run("localhost=3");

    checkLauncherLevelEResults("_outputFile_levelE_rows_.dat");

}

void TLauncherTest::testLauncherLevelEOutputFileKeys()
{
    setUpLevelE();

    _foutKey = new URANIE::Launcher::TOutputFileKey("_output_levelE_withKey_.dat");
    _foutKey->addAttribute(new URANIE::DataServer::TAttribute("tps",URANIE::DataServer::TAttribute::kVector), "time");
    _foutKey->addAttribute(new URANIE::DataServer::TAttribute("y",URANIE::DataServer::TAttribute::kVector));
    _foutKey->addAttribute(new URANIE::DataServer::TAttribute("z",URANIE::DataServer::TAttribute::kVector));

#ifdef WIN32
    tc1 = new URANIE::Launcher::TCode(tds1, "levele.exe");
#else
    tc1 = new URANIE::Launcher::TCode(tds1, Form("%s/%s",_workingTestPath.Data(),"../../teST/coDE/levelE/levele 2> /dev/null "));
#endif

    tc1->addOutputFile( _foutKey );

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);
    tl1->run("nointermed");

    checkLauncherLevelEResults("_outputFile_levelE_keys_.dat");

}

void TLauncherTest::testLauncherLevelEOutputFileColumnByStep()
{
    setUpLevelE();

    _foutCol = new URANIE::Launcher::TOutputFileColumn("_output_levelE_withColumn_.dat");
    _foutCol->addAttribute(new URANIE::DataServer::TAttribute("tps",URANIE::DataServer::TAttribute::kVector));
    _foutCol->addAttribute(new URANIE::DataServer::TAttribute("y",URANIE::DataServer::TAttribute::kVector));
    _foutCol->addAttribute(new URANIE::DataServer::TAttribute("z",URANIE::DataServer::TAttribute::kVector));

#ifdef WIN32
    tc1 = new URANIE::Launcher::TCode(tds1, "levele.exe");
#else
    tc1 = new URANIE::Launcher::TCode(tds1, Form("%s/%s",_workingTestPath.Data(),"../../teST/coDE/levelE/levele 2> /dev/null "));
#endif

    tc1->addOutputFile( _foutCol );

    tlbs = new URANIE::Launcher::TLauncherByStep(tds1, tc1);
    tlbs->setDrawProgressBar(kFALSE);
    tlbs->preTreatment();
    tlbs->run();
    tlbs->postTreatment("nointermed");

    checkLauncherLevelEResults("_outputFile_levelE_cols_bystep_.dat");

}

void TLauncherTest::checkLauncherLevelEResults(TString sOutputfileTmp)
{

    Int_t nSizeTime = 26;
    int npat = tds1->getNPatterns();
    
    CPPUNIT_ASSERT_EQUAL( 15, tds1->getNAttributes());
    CPPUNIT_ASSERT_EQUAL( 12, tds1->getNInputAttributes());
    CPPUNIT_ASSERT_EQUAL( 3, tds1->getNOutputAttributes());

    tds1->exportData(sOutputfileTmp, "t:kl:kc:v1:l1:r1:rc1:v2:l2:r2:rc2:w:tps:y:z");
    Int_t nret = UranieTests::areFilesEqual( TString("outputFile_levelE.dat.ref"), sOutputfileTmp, TString("#DATE:"));
    CPPUNIT_ASSERT_EQUAL(0,nret);

    TTree* tuple = (TNtupleD*)tds1->getTuple();
    TObjArray * Leaves = tuple->GetListOfLeaves();
    CPPUNIT_ASSERT(tds1->isAttribute("tps"));
    CPPUNIT_ASSERT(tds1->isAttribute("y"));
    CPPUNIT_ASSERT(tds1->isAttribute("z"));
    CPPUNIT_ASSERT(Leaves->FindObject("tps") != 0);
    CPPUNIT_ASSERT(Leaves->FindObject("y") != 0);
    CPPUNIT_ASSERT(Leaves->FindObject("z") != 0);

    // Time Attribute
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 2.000000000e+04, tds1->getValue("tps",0,0), DBL_DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 3.000000000e+04, tds1->getValue("tps",0,1), DBL_DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 9.000000000e+06, tds1->getValue("tps",0,nSizeTime-1), DBL_DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 2.000000000e+04, tds1->getValue("tps",1,0), DBL_DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 9.000000000e+06, tds1->getValue("tps",1,nSizeTime-1), DBL_DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 9.000000000e+06, tds1->getValue("tps",npat-1,nSizeTime-1), DBL_DBL_PRECISION_COMPARE);

    // Dose Attribute
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1.723721000e-06, tds1->getValue("y",0,0), DBL_DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 3.863967000e-07, tds1->getValue("y",0,1), DBL_DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 5.709311000e-07, tds1->getValue("y",0,nSizeTime-1), DBL_DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 2.390295000e-14, tds1->getValue("y",1,0), DBL_DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 8.234901000e-16, tds1->getValue("y",1,nSizeTime-1), DBL_DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 5.575421000e-15, tds1->getValue("y",npat-1,0), DBL_DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 3.445296000e-12, tds1->getValue("y",npat-1,nSizeTime-1), DBL_DBL_PRECISION_COMPARE);

    // Clean
    tds1->getAttribute("tps")->delShare();
    tds1->getAttribute("y")->delShare();
    tds1->getAttribute("z")->delShare();    
}
#endif



////////////////////////////////
///                          ///
///      FMI                 ///
///                          ///
////////////////////////////////
#ifdef WITH_FMI
void TLauncherTest::setUpFMI()
{
    if (_bLog)
        cout << endl <<
                " ********************" << endl <<
                " *** Method::" << __func__ << endl <<
                " *** PWD[" << gSystem->pwd() << "]" << endl;

    TString sFileFMU = "";
    Int_t nout = 0;

    // useCase1 - FMIv1
    {
        // TDS uscase 1 with FMI v1
        tds1 = new URANIE::DataServer::TDataServer("tdsUseCase1_FMIv1", "TDs for the use Case 1 - FMI v1");
        tds1->unsetLog();
        tds1->fileDataRead("useCase1FMIv1.dat");

        sFileFMU = Form("%s/useCase1_fmi1_cs.fmu", gSystem->pwd());
        //sFileFMU = "./useCase1_fmi1_cs.fmu";
        if (_bLog)
            cout << " **** sFileFMU[" << sFileFMU << "]" << endl;

        tds1->getAttribute("mu")->setFileFMU(sFileFMU, "BOUNCE_COF");

        _foutFMI1 = new URANIE::Launcher::TOutputFileFMU(sFileFMU);
        _foutFMI1->unsetLog();
        _foutFMI1->addAttribute(new URANIE::DataServer::TAttribute("yhat0", URANIE::DataServer::TAttribute::kVector), "HIGHT");
        _foutFMI1->addAttribute(new URANIE::DataServer::TAttribute("yhat1", URANIE::DataServer::TAttribute::kVector), "HIGHT_SPEED");
        //    _foutFMU1FMIv1->addAttribute(new TAttribute("yhat0", TAttribute::kVector), "x0");
        //    _foutFMU1FMIv1->addAttribute(new TAttribute("yhat1", TAttribute::kVector), "x1");

        nout =  _foutFMI1->getNAttributes();
        for(Int_t i=0; i<nout; i++)
            _foutFMI1->getAttribute(i)->delShare();

        tc1 = new URANIE::Launcher::TCodeFMU(tds1, sFileFMU, "loglevel=nothing");
        tc1->unsetLog();
    }

    // useCase1 - FMIv2
    {
        // TDS usecase 1 with FMI v2
        tds2 = new URANIE::DataServer::TDataServer("tdsUseCase1_FMIv2", "TDs for the use Case 1 - FMI v2");
        tds2->unsetLog();
        tds2->fileDataRead("useCase1FMIv2.dat");

        sFileFMU = Form("%s/useCase1_fmi2_cs.fmu", gSystem->pwd());
        if (_bLog)
            cout << " **** sFileFMU[" << sFileFMU << "]" << endl;

        //    _tds1FMIv2->getAttribute("mu")->setFileFMU(sFileFMU, "mu");
        tds2->getAttribute("mu")->setFileFMU(sFileFMU, "BOUNCE_COF");

        _foutFMI2 = new URANIE::Launcher::TOutputFileFMU(sFileFMU);
        _foutFMI2->unsetLog();
        _foutFMI2->addAttribute(new URANIE::DataServer::TAttribute("yhat0", URANIE::DataServer::TAttribute::kVector), "HIGHT");
        _foutFMI2->addAttribute(new URANIE::DataServer::TAttribute("yhat1", URANIE::DataServer::TAttribute::kVector), "HIGHT_SPEED");
        _foutFMI2->addAttribute(new URANIE::DataServer::TAttribute("yhat2", URANIE::DataServer::TAttribute::kVector), "HIGHT_ACC");
        //    _foutFMU1FMIv2->addAttribute(new TAttribute("yhat0", TAttribute::kVector), "x0");
        //    _foutFMU1FMIv2->addAttribute(new TAttribute("yhat1", TAttribute::kVector), "x1");

        nout =  _foutFMI2->getNAttributes();
        for(Int_t i=0; i<nout; i++)
            _foutFMI2->getAttribute(i)->delShare();

        _tcFMI2 = new URANIE::Launcher::TCodeFMU(tds2, sFileFMU, "loglevel=info");
        _tcFMI2->unsetLog();
    }

    if (_bLog)
        cout << endl <<
                " *** end of Method::" << __func__ << endl <<
                " ********************" << endl;
}

void TLauncherTest::tearDownFMI()
{
    if (_bLog)
        cout << endl <<
                " ********************" << endl <<
                " *** Method::" << __func__ << endl <<
                " *** PWD[" << gSystem->pwd() << "]" << endl;

    if (_foutFMI1)
        delete _foutFMI1;
    _foutFMI1=nullptr;

    if (_foutFMI2)
        delete _foutFMI2;
    _foutFMI2=nullptr;

    if (_tcFMI1)
        delete _tcFMI1;
    _tcFMI1 = nullptr;
    if (_tcFMI2)
        delete _tcFMI2;
    _tcFMI2 = nullptr;

    if (_bLog)
        cout << endl <<
                " *** end of Method::" << __func__ << endl <<
                " ********************" << endl;
}

void TLauncherTest::checkValuesFMI(URANIE::DataServer::TDataServer *tds, Int_t nSizeYs)
{
    if (_bLog)
        cout << endl <<
                " ********************" << endl <<
                " *** Method::" << __func__ << endl <<
                " *** tds Name::" << tds->GetName() << endl <<
                " *** nSizeYs::" << nSizeYs << endl <<
                " *** PWD[" << gSystem->pwd() << "]" << endl;

    TString satt = "yhat0";

    Int_t npat = tds->getNPatterns();
    Int_t nsizeY = -1;
    for(Int_t i=0; i<npat; i++) {
        nsizeY = (Int_t)(tds->getValue(Form("size__%s", satt.Data()), i));
        CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Method::%s The size of vector [%s] for index(%d) equal to (%d) and not equal to the reference value (%d)", __func__, satt.Data(), i, nsizeY, nSizeYs),
                                     nsizeY, nSizeYs);
    }

    //Int_t nYSize = tds->getAttribute(satt)->

//    //Statistic for a single realisation of a vector, not considering other events
//    tds->addAttribute("Min_x",  Form("Min$(%s)", satt.Data()));
//    tds->addAttribute("Max_x",  Form("Max$(%s)", satt.Data()));
//    tds->addAttribute("Mean_x", Form("Sum$(%s)/Length$(%s)", satt.Data(), satt.Data()));

//    tds->scan("yhat0:Min_x:Max_x:Mean_x","","colsize=5 col=2::::");


    if (_bLog)
        cout << endl <<
                " *** end of Method::" << __func__ << endl <<
                " ********************" << endl;
}

void TLauncherTest::teFMIv1BoundingBall()
{
#ifdef WIN32
    URANIE_IGNORED_TEST()
#endif
    cout << endl;
    if (_bLog)
        cout << " ********************" << endl <<
                " *** Method::" << __func__ << endl <<
                " *** PWD[" << gSystem->pwd() << "]" << endl;

    setUpFMI();

    // References values
    Int_t noatt_ref=2;
    Int_t nipf_ref = 0, nopf_ref = 0;
    Int_t npat_ref = tds1->getNPatterns();
    Int_t natt_ref = tds1->getNAttributes();
    // Current values
    Int_t npat, natt, nipf, nopf, noatt;

    // The output Attributes
    noatt = _foutFMI1->getNAttributes();
    CPPUNIT_ASSERT_EQUAL_MESSAGE(
                Form("Method::%s The number of Output attributes (%d) is not equal to reference(%d)", __func__, noatt, noatt_ref),
                noatt, noatt_ref);
    // the time Output attribute
    TString sattTime = "__time__";
    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", sattTime.Data()), !tds1->isAttribute(sattTime) );

    nipf = tc1->getNInputFiles();
    nopf = tc1->getNOutputFiles();

    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Method::%s The number of Input files (%d) is not equal to reference(%d)", __func__, nipf, nipf_ref),
                                 nipf, nipf_ref);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Method::%s The number of Output files (%d) is not equal to reference(%d)", __func__, nopf, nopf_ref),
                                 nopf, nopf_ref);

    /// Add the output file
    tc1->addOutputFile(_foutFMI1);
    nopf_ref++;

    nipf = tc1->getNInputFiles();
    nopf = tc1->getNOutputFiles();

    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Method::%s The number of Input files (%d) is not equal to reference(%d)", __func__, nipf, nipf_ref),
                                 nipf, nipf_ref);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Method::%s The number of Output files (%d) is not equal to reference(%d)", __func__, nopf, nopf_ref),
                                 nopf, nopf_ref);

    npat = tds1->getNPatterns();
    natt = tds1->getNAttributes();

    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Method::%s The number of patterns (%d) is not equal to reference(%d)",
                                      __func__, npat, npat_ref), npat, npat_ref);

    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Method::%s The number of attributes (%d) is not equal to reference(%d)",
                                      __func__, natt, natt_ref), natt, natt_ref);

    noatt = _foutFMI1->getNAttributes();
    CPPUNIT_ASSERT_EQUAL_MESSAGE(
                Form("Method::%s The number of Output attributes (%d) is not equal to reference(%d)", __func__, noatt, noatt_ref),
                noatt, noatt_ref);

    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", sattTime.Data()), !tds1->isAttribute(sattTime));

    /// Create the TLauncher object
    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    // Add the One input file
    nipf_ref++;

    nipf = tc1->getNInputFiles();
    nopf = tc1->getNOutputFiles();

    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Method::%s The number of Input files (%d) is not equal to reference(%d)", __func__, nipf, nipf_ref),
                                 nipf, nipf_ref);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Method::%s The number of Output files (%d) is not equal to reference(%d)", __func__, nopf, nopf_ref),
                                 nopf, nopf_ref);

    npat = tds1->getNPatterns();
    natt = tds1->getNAttributes();

    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Method::%s The number of patterns (%d) is not equal to reference(%d)",
                                      __func__, npat, npat_ref), npat, npat_ref);

    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Method::%s The number of attributes (%d) is not equal to reference(%d)",
                                      __func__, natt, natt_ref), natt, natt_ref);

    noatt = _foutFMI1->getNAttributes();
    CPPUNIT_ASSERT_EQUAL_MESSAGE(
                Form("Method::%s The number of Output attributes (%d) is not equal to reference(%d)", __func__, noatt, noatt_ref),
                noatt, noatt_ref);

    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", sattTime.Data()), !tds1->isAttribute(sattTime));

    /// Evaluate the DoE
    tl1->setClean();
    tl1->setSave();
    tl1->run();

//    cout << " ***  natt_ref[" << natt_ref << "]" << endl;
//    cout << " ***     noatt[" << noatt << "]" << endl;
    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", sattTime.Data()), tds1->isAttribute(sattTime));
    noatt_ref++;

    TString satt = "HIGHT";
    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", sattTime.Data()), !tds1->isAttribute(satt));
    satt = "yhat0";
    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", sattTime.Data()), tds1->isAttribute(satt));
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Attribute \"%s\" is type (%s) and it must be kVector", satt.Data(), tds1->getAttribute(satt)->getDataTypeStr().Data()),
                                 URANIE::DataServer::TAttribute::kVector, tds1->getAttribute(satt)->getDataType());

    satt = "HIGHT_SPEED";
    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", sattTime.Data()), !tds1->isAttribute(satt));
    satt = "yhat1";
    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", sattTime.Data()), tds1->isAttribute(satt));

    noatt = _foutFMI1->getNAttributes();
    CPPUNIT_ASSERT_EQUAL_MESSAGE(
                Form("Method::%s The number of Output attributes (%d) is not equal to reference(%d)", __func__, noatt, noatt_ref),
                noatt, noatt_ref);

    natt_ref += noatt_ref;

    npat = tds1->getNPatterns();
    natt = tds1->getNAttributes();

    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Method::%s The number of patterns (%d) is not equal to reference(%d)",
                                      __func__, npat, npat_ref), npat, npat_ref);

    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Method::%s The number of attributes (%d) is not equal to reference(%d)",
                                      __func__, natt, natt_ref), natt, natt_ref);

    nipf = tc1->getNInputFiles();
    nopf = tc1->getNOutputFiles();

    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Method::%s The number of Input files (%d) is not equal to reference(%d)", __func__, nipf, nipf_ref),
                                 nipf, nipf_ref);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Method::%s The number of Output files (%d) is not equal to reference(%d)", __func__, nopf, nopf_ref),
                                 nopf, nopf_ref);

    checkValuesFMI(tds1);

    tearDownFMI();
    if (_bLog)
        cout << endl << " *** end of Method::" << __func__ <<
                endl << " ********************" << endl;
}

void TLauncherTest::teFMIv2BoundingBall()
{
#ifdef WIN32
    URANIE_IGNORED_TEST()
#endif
    cout << endl;
    if (_bLog)
        cout << " ********************" << endl <<
                " *** Method::" << __func__ << endl <<
                " *** PWD[" << gSystem->pwd() << "]" << endl;

    setUpFMI();

    // References values
    Int_t noatt_ref=3;
    Int_t nipf_ref = 0, nopf_ref = 0;
    Int_t npat_ref = tds2->getNPatterns();
    Int_t natt_ref = tds2->getNAttributes();
    // Current values
    Int_t npat, natt, nipf, nopf, noatt;

    // The output Attributes
    noatt = _foutFMI2->getNAttributes();
    CPPUNIT_ASSERT_EQUAL_MESSAGE(
                Form("Method::%s The number of Output attributes (%d) is not equal to reference(%d)", __func__, noatt, noatt_ref),
                noatt, noatt_ref);
    // the time Output attribute
    TString sattTime = "__time__";
    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", sattTime.Data()), !tds2->isAttribute(sattTime) );

    nipf = _tcFMI2->getNInputFiles();
    nopf = _tcFMI2->getNOutputFiles();

    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Method::%s The number of Input files (%d) is not equal to reference(%d)", __func__, nipf, nipf_ref),
                                 nipf, nipf_ref);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Method::%s The number of Output files (%d) is not equal to reference(%d)", __func__, nopf, nopf_ref),
                                 nopf, nopf_ref);

    /// Add the output file
    _tcFMI2->addOutputFile(_foutFMI2);
    nopf_ref++;

    nipf = _tcFMI2->getNInputFiles();
    nopf = _tcFMI2->getNOutputFiles();

    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Method::%s The number of Input files (%d) is not equal to reference(%d)", __func__, nipf, nipf_ref),
                                 nipf, nipf_ref);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Method::%s The number of Output files (%d) is not equal to reference(%d)", __func__, nopf, nopf_ref),
                                 nopf, nopf_ref);

    // Set Start, Step and Stop simulation
    Double_t dstart=0.0, dstep=0.1, dstop=2.0;
    Int_t nSizeYs = (Int_t)((dstop-dstart)/dstep)+1;
    cout << " *** nSizeYs[" << nSizeYs << "]" << endl;

    _tcFMI2->setStartStopSimulation(dstart, dstop);
    _tcFMI2->setStepSimulation(dstep);

    npat = tds2->getNPatterns();
    natt = tds2->getNAttributes();

    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Method::%s The number of patterns (%d) is not equal to reference(%d)",
                                      __func__, npat, npat_ref), npat, npat_ref);

    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Method::%s The number of attributes (%d) is not equal to reference(%d)",
                                      __func__, natt, natt_ref), natt, natt_ref);

    noatt = _foutFMI2->getNAttributes();
    CPPUNIT_ASSERT_EQUAL_MESSAGE(
                Form("Method::%s The number of Output attributes (%d) is not equal to reference(%d)", __func__, noatt, noatt_ref),
                noatt, noatt_ref);

    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", sattTime.Data()), !tds2->isAttribute(sattTime));

    /// Create the TLauncher object
    tl2 = new URANIE::Launcher::TLauncher(tds2, _tcFMI2);
    // Add the One input file
    nipf_ref++;

    nipf = _tcFMI2->getNInputFiles();
    nopf = _tcFMI2->getNOutputFiles();

    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Method::%s The number of Input files (%d) is not equal to reference(%d)", __func__, nipf, nipf_ref),
                                 nipf, nipf_ref);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Method::%s The number of Output files (%d) is not equal to reference(%d)", __func__, nopf, nopf_ref),
                                 nopf, nopf_ref);

    npat = tds2->getNPatterns();
    natt = tds2->getNAttributes();

    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Method::%s The number of patterns (%d) is not equal to reference(%d)",
                                      __func__, npat, npat_ref), npat, npat_ref);

    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Method::%s The number of attributes (%d) is not equal to reference(%d)",
                                      __func__, natt, natt_ref), natt, natt_ref);

    noatt = _foutFMI2->getNAttributes();
    CPPUNIT_ASSERT_EQUAL_MESSAGE(
                Form("Method::%s The number of Output attributes (%d) is not equal to reference(%d)", __func__, noatt, noatt_ref),
                noatt, noatt_ref);

    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", sattTime.Data()), !tds2->isAttribute(sattTime));

    /// Evaluate the DoE
    tl2->setClean();
    tl2->setSave();
    tl2->run();

//    cout << " ***  natt_ref[" << natt_ref << "]" << endl;
//    cout << " ***     noatt[" << noatt << "]" << endl;
    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", sattTime.Data()), tds2->isAttribute(sattTime));
    noatt_ref++;

    TString satt = "HIGHT";
    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", sattTime.Data()), !tds2->isAttribute(satt));
    satt = "yhat0";
    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", sattTime.Data()), tds2->isAttribute(satt));
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Attribute \"%s\" is type (%s) and it must be kVector", satt.Data(), tds2->getAttribute(satt)->getDataTypeStr().Data()),
                                 URANIE::DataServer::TAttribute::kVector, tds2->getAttribute(satt)->getDataType());

    satt = "HIGHT_SPEED";
    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", sattTime.Data()), !tds2->isAttribute(satt));
    satt = "yhat1";
    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", sattTime.Data()), tds2->isAttribute(satt));
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Attribute \"%s\" is type (%s) and it must be kVector", satt.Data(), tds2->getAttribute(satt)->getDataTypeStr().Data()),
                                 URANIE::DataServer::TAttribute::kVector, tds2->getAttribute(satt)->getDataType());

    satt = "HIGHT_ACC";
    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", sattTime.Data()), !tds2->isAttribute(satt));
    satt = "yhat2";
    CPPUNIT_ASSERT_MESSAGE(Form("Exist Attribute %s", sattTime.Data()), tds2->isAttribute(satt));
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Attribute \"%s\" is type (%s) and it must be kVector", satt.Data(), tds2->getAttribute(satt)->getDataTypeStr().Data()),
                                 URANIE::DataServer::TAttribute::kVector, tds2->getAttribute(satt)->getDataType());

    noatt = _foutFMI2->getNAttributes();
    CPPUNIT_ASSERT_EQUAL_MESSAGE(
                Form("Method::%s The number of Output attributes (%d) is not equal to reference(%d)", __func__, noatt, noatt_ref),
                noatt, noatt_ref);

    natt_ref += noatt_ref;

    npat = tds2->getNPatterns();
    natt = tds2->getNAttributes();

    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Method::%s The number of patterns (%d) is not equal to reference(%d)",
                                      __func__, npat, npat_ref), npat, npat_ref);

    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Method::%s The number of attributes (%d) is not equal to reference(%d)",
                                      __func__, natt, natt_ref), natt, natt_ref);

    nipf = _tcFMI2->getNInputFiles();
    nopf = _tcFMI2->getNOutputFiles();

    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Method::%s The number of Input files (%d) is not equal to reference(%d)", __func__, nipf, nipf_ref),
                                 nipf, nipf_ref);
    CPPUNIT_ASSERT_EQUAL_MESSAGE(Form("Method::%s The number of Output files (%d) is not equal to reference(%d)", __func__, nopf, nopf_ref),
                                 nopf, nopf_ref);

    checkValuesFMI(tds2, nSizeYs);

    tearDownFMI();
    if (_bLog)
        cout << endl << " *** end of Method::" << __func__ << endl<<
                " ********************" << endl;
}
#endif

////////////////////////////////
///                          ///
///      multitype              ///
///                          ///
////////////////////////////////

void TLauncherTest::setUpMultiType()
{
    tds1 = new URANIE::DataServer::TDataServer("foo", "multitype usecase");
    tds1->fileDataRead("multitype_sampling.dat");
    tds1->getAttribute("seed")->setFileKey("multitype_input.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewRow);
}

void TLauncherTest::testLauncherMultiTypeOutputFileDataServer()
{

    setUpMultiType();

    _foutDataServer = new URANIE::Launcher::TOutputFileDataServer("_output_multitype_mt_DataServer_.dat");
    _foutDataServer->addAttribute(new URANIE::DataServer::TAttribute("w1",URANIE::DataServer::TAttribute::kString));
    _foutDataServer->addAttribute(new URANIE::DataServer::TAttribute("w2",URANIE::DataServer::TAttribute::kString));
    _foutDataServer->addAttribute(new URANIE::DataServer::TAttribute("v1",URANIE::DataServer::TAttribute::kVector));
    _foutDataServer->addAttribute(new URANIE::DataServer::TAttribute("v2",URANIE::DataServer::TAttribute::kVector));
    _foutDataServer->addAttribute(new URANIE::DataServer::TAttribute("f1"));
    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("multitype") + " -mtDS");
    tc1->addOutputFile( _foutDataServer );

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(false);
    tl1->run("nointermed");
  
    checkLauncherMultiTypeResults();
}

#ifdef WITH_LIBXML2
void TLauncherTest::testLauncherMultiTypeOutputFileXML()
{

    setUpMultiType();

    _foutXML = new URANIE::Launcher::TOutputFileXML("_output_multitype_mt_.xml");
    _foutXML->addAttribute(new URANIE::DataServer::TAttribute("w1",URANIE::DataServer::TAttribute::kString), "/multitypeMT/w1/@value", URANIE::DataServer::TAttributeFileKey::kXMLAttribute );
    _foutXML->addAttribute(new URANIE::DataServer::TAttribute("w2",URANIE::DataServer::TAttribute::kString), "/multitypeMT/w2/@value", URANIE::DataServer::TAttributeFileKey::kXMLAttribute );
    _foutXML->addAttribute(new URANIE::DataServer::TAttribute("v1",URANIE::DataServer::TAttribute::kVector), "/multitypeMT/v1/values", URANIE::DataServer::TAttributeFileKey::kXMLField );
    _foutXML->addAttribute(new URANIE::DataServer::TAttribute("v2",URANIE::DataServer::TAttribute::kVector), "/multitypeMT/v2/values", URANIE::DataServer::TAttributeFileKey::kXMLField );
    _foutXML->addAttribute(new URANIE::DataServer::TAttribute("f1"), "/multitypeMT/f1/@value", URANIE::DataServer::TAttributeFileKey::kXMLAttribute );
    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("multitype") + " -mtXML");
    tc1->addOutputFile( _foutXML );

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1); 
    tl1->setDrawProgressBar(false);
    tl1->run("nointermed");
  
    checkLauncherMultiTypeResults();
}
#endif

void TLauncherTest::testLauncherMultiTypeOutputFileKey()
{

    setUpMultiType();

    _foutKey = new URANIE::Launcher::TOutputFileKey("_output_multitype_mt_Key_.dat");
    _foutKey->addAttribute(new URANIE::DataServer::TAttribute("w1",URANIE::DataServer::TAttribute::kString));
    _foutKey->addAttribute(new URANIE::DataServer::TAttribute("w2",URANIE::DataServer::TAttribute::kString));
    _foutKey->addAttribute(new URANIE::DataServer::TAttribute("v1",URANIE::DataServer::TAttribute::kVector));
    _foutKey->addAttribute(new URANIE::DataServer::TAttribute("v2",URANIE::DataServer::TAttribute::kVector));
    _foutKey->addAttribute(new URANIE::DataServer::TAttribute("f1"));
    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("multitype") + " -mtKey");
    tc1->addOutputFile( _foutKey );

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(false);
    tl1->run("nointermed");

    checkLauncherMultiTypeResults();
}

void TLauncherTest::testLauncherMultiTypeOutputFileKeyCondensate()
{

    setUpMultiType();

    _foutKey = new URANIE::Launcher::TOutputFileKey("_output_multitype_mt_Key_condensate_.dat");
    _foutKey->addAttribute(new URANIE::DataServer::TAttribute("w1",URANIE::DataServer::TAttribute::kString));
    _foutKey->addAttribute(new URANIE::DataServer::TAttribute("w2",URANIE::DataServer::TAttribute::kString));
    _foutKey->addAttribute(new URANIE::DataServer::TAttribute("v1",URANIE::DataServer::TAttribute::kVector));
    _foutKey->addAttribute(new URANIE::DataServer::TAttribute("v2",URANIE::DataServer::TAttribute::kVector));
    _foutKey->addAttribute(new URANIE::DataServer::TAttribute("f1"));
    _foutKey->setVectorProperties("[",",","]");
    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("multitype") + " -mtKey");
    tc1->addOutputFile( _foutKey );

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(false);
    tl1->run("nointermed");

    checkLauncherMultiTypeResults();
}

void TLauncherTest::testLauncherMultiTypeOutputFileRow()
{

    setUpMultiType();

    _foutRow = new URANIE::Launcher::TOutputFileRow("_output_multitype_mt_Row_.dat");
    _foutRow->setFieldSeparatorCharacter(";");
    _foutRow->setSeparatorCharacter(" \n");
    _foutRow->addAttribute(new URANIE::DataServer::TAttribute("w1",URANIE::DataServer::TAttribute::kString),1);
    _foutRow->addAttribute(new URANIE::DataServer::TAttribute("w2",URANIE::DataServer::TAttribute::kString),5);
    _foutRow->addAttribute(new URANIE::DataServer::TAttribute("v1",URANIE::DataServer::TAttribute::kVector),2);
    _foutRow->addAttribute(new URANIE::DataServer::TAttribute("v2",URANIE::DataServer::TAttribute::kVector),3);
    _foutRow->addAttribute(new URANIE::DataServer::TAttribute("f1"),4);
    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("multitype") + " -mtRow");
    tc1->addOutputFile( _foutRow );

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(false);
    tl1->run("nointermed");

    checkLauncherMultiTypeResults();
}

void TLauncherTest::testLauncherMultiTypeOutputFileRowMultithread()
{

    setUpMultiType();

    _foutRow = new URANIE::Launcher::TOutputFileRow("_output_multitype_mt_Row_.dat");
    _foutRow->setFieldSeparatorCharacter(";");
    _foutRow->setSeparatorCharacter(" \n");
    _foutRow->addAttribute(new URANIE::DataServer::TAttribute("w1",URANIE::DataServer::TAttribute::kString),1);
    _foutRow->addAttribute(new URANIE::DataServer::TAttribute("w2",URANIE::DataServer::TAttribute::kString),5);
    _foutRow->addAttribute(new URANIE::DataServer::TAttribute("v1",URANIE::DataServer::TAttribute::kVector),2);
    _foutRow->addAttribute(new URANIE::DataServer::TAttribute("v2",URANIE::DataServer::TAttribute::kVector),3);
    _foutRow->addAttribute(new URANIE::DataServer::TAttribute("f1"),4);
    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("multitype") + " -mtRow");
    tc1->addOutputFile( _foutRow );

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1, "multiquiet");
    tl1->setDrawProgressBar(false);
    tl1->run("localhost=3");
  
    checkLauncherMultiTypeResults();
}


void TLauncherTest::testLauncherMultiTypeOutputFileRowBySteps()
{

    setUpMultiType();

    _foutRow = new URANIE::Launcher::TOutputFileRow("_output_multitype_mt_Row_.dat");
    _foutRow->setFieldSeparatorCharacter(";");
    _foutRow->setSeparatorCharacter(" \n");
    _foutRow->addAttribute(new URANIE::DataServer::TAttribute("w1",URANIE::DataServer::TAttribute::kString),1);
    _foutRow->addAttribute(new URANIE::DataServer::TAttribute("w2",URANIE::DataServer::TAttribute::kString),5);
    _foutRow->addAttribute(new URANIE::DataServer::TAttribute("v1",URANIE::DataServer::TAttribute::kVector),2);
    _foutRow->addAttribute(new URANIE::DataServer::TAttribute("v2",URANIE::DataServer::TAttribute::kVector),3);
    _foutRow->addAttribute(new URANIE::DataServer::TAttribute("f1"),4);
    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("multitype") + " -mtRow");
    tc1->addOutputFile( _foutRow );

    tlbs = new URANIE::Launcher::TLauncherByStep(tds1, tc1);
    tlbs->setDrawProgressBar(false);
    tlbs->preTreatment();
    tlbs->run();
    tlbs->postTreatment("nointermed");
    
    checkLauncherMultiTypeResults();
}

void TLauncherTest::testLauncherMultiTypeOutputFileColumn()
{

    setUpMultiType();

    _foutCol = new URANIE::Launcher::TOutputFileColumn("_output_multitype_mt_Column_.dat");
    _foutCol->addAttribute(new URANIE::DataServer::TAttribute("w1",URANIE::DataServer::TAttribute::kString),1);
    _foutCol->addAttribute(new URANIE::DataServer::TAttribute("w2",URANIE::DataServer::TAttribute::kString),5);
    _foutCol->addAttribute(new URANIE::DataServer::TAttribute("v1",URANIE::DataServer::TAttribute::kVector),2);
    _foutCol->addAttribute(new URANIE::DataServer::TAttribute("v2",URANIE::DataServer::TAttribute::kVector),3);
    _foutCol->addAttribute(new URANIE::DataServer::TAttribute("f1"),4);
    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("multitype") + " -mtCol");
    tc1->addOutputFile( _foutCol );

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(false);
    tl1->run("nointermed");

    checkLauncherMultiTypeResults();
}

void TLauncherTest::checkLauncherMultiTypeResults()
{
    //Attribute are there  
    CPPUNIT_ASSERT(tds1->isAttribute("seed"));
    CPPUNIT_ASSERT(tds1->isAttribute("w1"));
    CPPUNIT_ASSERT(tds1->isAttribute("v1"));
    CPPUNIT_ASSERT(tds1->isAttribute("v2"));
    CPPUNIT_ASSERT(tds1->isAttribute("f1"));
    CPPUNIT_ASSERT(tds1->isAttribute("w2"));	

    //tds1->Scan("*");
    
    CPPUNIT_ASSERT_EQUAL(10,tds1->getNPatterns());

    //Few entries
    //v1
    int nentries = tds1->getTuple()->Draw("v1","1>0","goff"); double *entries = tds1->getTuple()->GetV1();
    CPPUNIT_ASSERT_EQUAL(63,nentries);
    nentries = tds1->getTuple()->Draw("v1",Form("foo__n__iter__==%i",1),"goff"); CPPUNIT_ASSERT_EQUAL(12, nentries);
    entries = tds1->getTuple()->GetV1(); CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.06901933299378, entries[0], DBL_DBL_PRECISION_COMPARE);
    nentries = tds1->getTuple()->Draw("v1",Form("foo__n__iter__==%i",6),"goff"); CPPUNIT_ASSERT_EQUAL(8, nentries);
    nentries = tds1->getTuple()->Draw("v1",Form("foo__n__iter__==%i",10),"goff"); CPPUNIT_ASSERT_EQUAL(5, nentries);
    entries = tds1->getTuple()->GetV1(); CPPUNIT_ASSERT_DOUBLES_EQUAL(-0.2397650482929, entries[nentries-1], DBL_DBL_PRECISION_COMPARE);
  
    //v2
    nentries = tds1->getTuple()->Draw("v2","1>0","goff"); entries = tds1->getTuple()->GetV1();
    CPPUNIT_ASSERT_EQUAL(78,nentries);
    nentries = tds1->getTuple()->Draw("v2",Form("foo__n__iter__==%i",1),"goff"); CPPUNIT_ASSERT_EQUAL(6, nentries);
    entries = tds1->getTuple()->GetV1(); CPPUNIT_ASSERT_DOUBLES_EQUAL(11.42449011404, entries[0], DBL_DBL_PRECISION_COMPARE);
    nentries = tds1->getTuple()->Draw("v2",Form("foo__n__iter__==%i",6),"goff"); CPPUNIT_ASSERT_EQUAL(12, nentries);
    nentries = tds1->getTuple()->Draw("v2",Form("foo__n__iter__==%i",10),"goff"); CPPUNIT_ASSERT_EQUAL(9, nentries);
    entries = tds1->getTuple()->GetV1(); CPPUNIT_ASSERT_DOUBLES_EQUAL(8.552570119034, entries[nentries-1], DBL_DBL_PRECISION_COMPARE);
  
    //f1
    tds1->computeStatistic("f1");
    CPPUNIT_ASSERT_DOUBLES_EQUAL(12.3091863421, tds1->getAttribute("f1")->getStd(), DBL_DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(47.9894841045, tds1->getAttribute("f1")->getMean(), DBL_DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(69.8054975231, tds1->getAttribute("f1")->getMaximum(), DBL_DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL(34.2563231678, tds1->getAttribute("f1")->getMinimum(), DBL_DBL_PRECISION_COMPARE);

    //w1&w2
    tds1->getTuple()->Draw("Entry$",Form("foo__n__iter__==%i",1),"goff"); entries = tds1->getTuple()->GetV1();
    CPPUNIT_ASSERT("six"==tds1->getString("w1",entries[0]));
    CPPUNIT_ASSERT("zero"==tds1->getString("w2",entries[0]));
    tds1->getTuple()->Draw("Entry$",Form("foo__n__iter__==%i",6),"goff"); entries = tds1->getTuple()->GetV1();
    CPPUNIT_ASSERT("two"==tds1->getString("w1",entries[0]));
    CPPUNIT_ASSERT("five"==tds1->getString("w2",entries[0]));
    tds1->getTuple()->Draw("Entry$",Form("foo__n__iter__==%i",10),"goff"); entries = tds1->getTuple()->GetV1();
    CPPUNIT_ASSERT("four"==tds1->getString("w1",entries[0]));
    CPPUNIT_ASSERT("six"==tds1->getString("w2",entries[0]));

    //Cleaning
    tds1->getAttribute("w1")->delShare();
    tds1->getAttribute("w2")->delShare();
    tds1->getAttribute("v1")->delShare();
    tds1->getAttribute("v2")->delShare();
    tds1->getAttribute("f1")->delShare();
}

////////////////////////////////
///                          ///
///   multitype as inputs    ///
///                          ///
////////////////////////////////

void TLauncherTest::setUpMultiTypeAsInput()
{
    tds1 = new URANIE::DataServer::TDataServer("foo", "multitype usecase");
    tds1->fileDataRead("readmultitype_sampling.dat");
}

void TLauncherTest::testLauncherMultiTypeInputFileDataServer()
{

    setUpMultiTypeAsInput();

    tds1->getAttribute("w1")->setFileKey("_output_multitype_mt_DataServer_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewTDS);
    tds1->getAttribute("v1")->setFileKey("_output_multitype_mt_DataServer_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewTDS);
    tds1->getAttribute("v2")->setFileKey("_output_multitype_mt_DataServer_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewTDS);
    tds1->getAttribute("f1")->setFileKey("_output_multitype_mt_DataServer_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewTDS);
    tds1->getAttribute("w2")->setFileKey("_output_multitype_mt_DataServer_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewTDS);    

    URANIE::Launcher::TInputFileRecreate *in;
#ifdef WIN32
    in = new URANIE::Launcher::TInputFileRecreate(Form("%s\\_output_multitype_mt_DataServer_.dat",gSystem->pwd()));
#else
    in = new URANIE::Launcher::TInputFileRecreate(Form("%s/_output_multitype_mt_DataServer_.dat",gSystem->pwd()));
#endif
    in->addAttribute( tds1->getAttribute("w1") );
    in->addAttribute( tds1->getAttribute("v1") );
    in->addAttribute( tds1->getAttribute("v2") );
    in->addAttribute( tds1->getAttribute("f1") );
    in->addAttribute( tds1->getAttribute("w2") );

    _foutDataServer = new URANIE::Launcher::TOutputFileDataServer("_output_multitype_readmt_DataServer_.dat");
    _foutDataServer->addAttribute(new URANIE::DataServer::TAttribute("thev1") );
    _foutDataServer->addAttribute(new URANIE::DataServer::TAttribute("thev2") );
    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("multitype") + " -ReadmtDS");
    tc1->addInputFile( in );
    tc1->addOutputFile( _foutDataServer );

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(false);
    tl1->run("nointermed");

    checkLauncherMultiTypeInputResults();
}

#ifdef WITH_LIBXML2
#ifdef WITH_LIBXSLT
void TLauncherTest::testLauncherMultiTypeInputFileXML()
{

    setUpMultiTypeAsInput();

    gSystem->Exec((UranieTests::code_path("multitype") + " -mtXML").data());
    
    tds1->getAttribute("w1")->setFileKey("_output_multitype_mt_.xml","w1/@value", "", URANIE::DataServer::TAttributeFileKey::kXMLAttribute);
    tds1->getAttribute("v1")->setFileKey("_output_multitype_mt_.xml","v1/values", "", URANIE::DataServer::TAttributeFileKey::kXMLField);
    tds1->getAttribute("v2")->setFileKey("_output_multitype_mt_.xml","v2/values", "", URANIE::DataServer::TAttributeFileKey::kXMLField);
    tds1->getAttribute("f1")->setFileKey("_output_multitype_mt_.xml","f1/@value", "", URANIE::DataServer::TAttributeFileKey::kXMLAttribute);
    tds1->getAttribute("w2")->setFileKey("_output_multitype_mt_.xml","w2/@value", "", URANIE::DataServer::TAttributeFileKey::kXMLAttribute);    

    URANIE::Launcher::TInputFileXML *in;
#ifdef WIN32
    in = new URANIE::Launcher::TInputFileXML(Form("%s\\_output_multitype_mt_.xml",gSystem->pwd()));
#else
    in = new URANIE::Launcher::TInputFileXML(Form("%s/_output_multitype_mt_.xml",gSystem->pwd()));
#endif
    in->setVectorProperties(""," ", "");
    in->addAttribute( tds1->getAttribute("w1") );
    in->addAttribute( tds1->getAttribute("v1") );
    in->addAttribute( tds1->getAttribute("v2") );
    in->addAttribute( tds1->getAttribute("f1") );
    in->addAttribute( tds1->getAttribute("w2") );
    
    _foutXML = new URANIE::Launcher::TOutputFileXML("_output_multitype_readmt_.xml");
    _foutXML->addAttribute(new URANIE::DataServer::TAttribute("thev1"), "/multitypeREADMT/thev1/@value",  URANIE::DataServer::TAttributeFileKey::kXMLAttribute);
    _foutXML->addAttribute(new URANIE::DataServer::TAttribute("thev2"), "/multitypeREADMT/thev2/@value",  URANIE::DataServer::TAttributeFileKey::kXMLAttribute );

    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("multitype") + " -ReadmtXML");
    tc1->addInputFile( in );
    tc1->addOutputFile( _foutXML );

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(false);
    tl1->run("nointermed");

    checkLauncherMultiTypeInputResults();
}
#endif
#endif

void TLauncherTest::testLauncherMultiTypeInputFileRow()
{

    setUpMultiTypeAsInput();

    tds1->getAttribute("w1")->setFileKey("_output_multitype_mt_Row_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewRow);
    tds1->getAttribute("v1")->setFileKey("_output_multitype_mt_Row_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewRow);
    tds1->getAttribute("v2")->setFileKey("_output_multitype_mt_Row_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewRow);
    tds1->getAttribute("f1")->setFileKey("_output_multitype_mt_Row_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewRow);
    tds1->getAttribute("w2")->setFileKey("_output_multitype_mt_Row_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewRow);    

    URANIE::Launcher::TInputFileRecreate *in;
#ifdef WIN32
    in = new URANIE::Launcher::TInputFileRecreate(Form("%s\\_output_multitype_mt_Row_.dat",gSystem->pwd()));
#else
    in = new URANIE::Launcher::TInputFileRecreate(Form("%s/_output_multitype_mt_Row_.dat",gSystem->pwd()));
#endif
    in->addAttribute( tds1->getAttribute("w1") );
    in->addAttribute( tds1->getAttribute("v1") );
    in->addAttribute( tds1->getAttribute("v2") );
    in->addAttribute( tds1->getAttribute("f1") );
    in->addAttribute( tds1->getAttribute("w2") );
    in->setFieldSeparatorCharacter("; ");
    
    _foutRow = new URANIE::Launcher::TOutputFileRow("_output_multitype_readmt_Row_.dat");
    _foutRow->addAttribute(new URANIE::DataServer::TAttribute("thev1") );
    _foutRow->addAttribute(new URANIE::DataServer::TAttribute("thev2") );
    _foutRow->setSeparatorCharacter("; ");
    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("multitype") + " -ReadmtRow");
    tc1->addInputFile( in );
    tc1->addOutputFile( _foutRow );

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(false);
    tl1->run("nointermed");

    checkLauncherMultiTypeInputResults();
}

void TLauncherTest::testLauncherMultiTypeInputFileRowMultithread()
{

    setUpMultiTypeAsInput();

    tds1->getAttribute("w1")->setFileKey("_output_multitype_mt_Row_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewRow);
    tds1->getAttribute("v1")->setFileKey("_output_multitype_mt_Row_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewRow);
    tds1->getAttribute("v2")->setFileKey("_output_multitype_mt_Row_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewRow);
    tds1->getAttribute("f1")->setFileKey("_output_multitype_mt_Row_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewRow);
    tds1->getAttribute("w2")->setFileKey("_output_multitype_mt_Row_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewRow);    

    URANIE::Launcher::TInputFileRecreate *in;
#ifdef WIN32
    in = new URANIE::Launcher::TInputFileRecreate(Form("%s\\_output_multitype_mt_Row_.dat",gSystem->pwd()));
#else
    in = new URANIE::Launcher::TInputFileRecreate(Form("%s/_output_multitype_mt_Row_.dat",gSystem->pwd()));
#endif
    in->addAttribute( tds1->getAttribute("w1") );
    in->addAttribute( tds1->getAttribute("v1") );
    in->addAttribute( tds1->getAttribute("v2") );
    in->addAttribute( tds1->getAttribute("f1") );
    in->addAttribute( tds1->getAttribute("w2") );
    in->setFieldSeparatorCharacter("; ");
    
    _foutRow = new URANIE::Launcher::TOutputFileRow("_output_multitype_readmt_Row_.dat");
    _foutRow->addAttribute(new URANIE::DataServer::TAttribute("thev1") );
    _foutRow->addAttribute(new URANIE::DataServer::TAttribute("thev2") );
    _foutRow->setSeparatorCharacter("; ");
    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("multitype") + " -ReadmtRow");
    tc1->addInputFile( in );
    tc1->addOutputFile( _foutRow );

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1, "multiquiet");
    tl1->setDrawProgressBar(false);
    tl1->run("localhost=3");

    checkLauncherMultiTypeInputResults();
}

void TLauncherTest::testLauncherMultiTypeInputFileRowBySteps()
{

    setUpMultiTypeAsInput();

    tds1->getAttribute("w1")->setFileKey("_output_multitype_mt_Row_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewRow);
    tds1->getAttribute("v1")->setFileKey("_output_multitype_mt_Row_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewRow);
    tds1->getAttribute("v2")->setFileKey("_output_multitype_mt_Row_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewRow);
    tds1->getAttribute("f1")->setFileKey("_output_multitype_mt_Row_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewRow);
    tds1->getAttribute("w2")->setFileKey("_output_multitype_mt_Row_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewRow);    

    URANIE::Launcher::TInputFileRecreate *in;
#ifdef WIN32
    in = new URANIE::Launcher::TInputFileRecreate(Form("%s\\_output_multitype_mt_Row_.dat",gSystem->pwd()));
#else
    in = new URANIE::Launcher::TInputFileRecreate(Form("%s/_output_multitype_mt_Row_.dat",gSystem->pwd()));
#endif
    in->addAttribute( tds1->getAttribute("w1") );
    in->addAttribute( tds1->getAttribute("v1") );
    in->addAttribute( tds1->getAttribute("v2") );
    in->addAttribute( tds1->getAttribute("f1") );
    in->addAttribute( tds1->getAttribute("w2") );
    in->setFieldSeparatorCharacter("; ");
    
    _foutRow = new URANIE::Launcher::TOutputFileRow("_output_multitype_readmt_Row_.dat");
    _foutRow->addAttribute(new URANIE::DataServer::TAttribute("thev1") );
    _foutRow->addAttribute(new URANIE::DataServer::TAttribute("thev2") );
    _foutRow->setSeparatorCharacter("; ");
    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("multitype") + " -ReadmtRow");
    tc1->addInputFile( in );
    tc1->addOutputFile( _foutRow );

    tlbs = new URANIE::Launcher::TLauncherByStep(tds1, tc1);
    tlbs->setDrawProgressBar(false);
    tlbs->preTreatment();
    tlbs->run();
    tlbs->postTreatment("nointermed");

    checkLauncherMultiTypeInputResults();
}

void TLauncherTest::testLauncherMultiTypeInputFileColumn()
{

    setUpMultiTypeAsInput();

    tds1->getAttribute("w1")->setFileKey("_output_multitype_mt_Column_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewColumn);
    tds1->getAttribute("v1")->setFileKey("_output_multitype_mt_Column_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewColumn);
    tds1->getAttribute("v2")->setFileKey("_output_multitype_mt_Column_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewColumn);
    tds1->getAttribute("f1")->setFileKey("_output_multitype_mt_Column_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewColumn);
    tds1->getAttribute("w2")->setFileKey("_output_multitype_mt_Column_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewColumn);    

    URANIE::Launcher::TInputFileRecreate *in;
#ifdef WIN32
    in = new URANIE::Launcher::TInputFileRecreate(Form("%s\\_output_multitype_mt_Column_.dat",gSystem->pwd()));
#else
    in = new URANIE::Launcher::TInputFileRecreate(Form("%s/_output_multitype_mt_Column_.dat",gSystem->pwd()));
#endif
    in->addAttribute( tds1->getAttribute("w1") );
    in->addAttribute( tds1->getAttribute("v1") );
    in->addAttribute( tds1->getAttribute("v2") );
    in->addAttribute( tds1->getAttribute("f1") );
    in->addAttribute( tds1->getAttribute("w2") );

   _foutCol = new URANIE::Launcher::TOutputFileColumn("_output_multitype_readmt_Column_.dat");
    _foutCol->addAttribute(new URANIE::DataServer::TAttribute("thev1") );
    _foutCol->addAttribute(new URANIE::DataServer::TAttribute("thev2") );
    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("multitype") + " -ReadmtCol");
    tc1->addInputFile( in );
    tc1->addOutputFile( _foutCol );

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(false);
    tl1->run("nointermed");

    checkLauncherMultiTypeInputResults();
}

void TLauncherTest::testLauncherMultiTypeInputFileKey()
{

    setUpMultiTypeAsInput();

    tds1->getAttribute("w1")->setFileKey("_output_multitype_mt_Key_condensate_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewKey);
    tds1->getAttribute("v1")->setFileKey("_output_multitype_mt_Key_condensate_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewKey);
    tds1->getAttribute("v2")->setFileKey("_output_multitype_mt_Key_condensate_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewKey);
    tds1->getAttribute("f1")->setFileKey("_output_multitype_mt_Key_condensate_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewKey);
    tds1->getAttribute("w2")->setFileKey("_output_multitype_mt_Key_condensate_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewKey);    

    URANIE::Launcher::TInputFileRecreate *in;
#ifdef WIN32
    in = new URANIE::Launcher::TInputFileRecreate(Form("%s\\_output_multitype_mt_Key_condensate_.dat",gSystem->pwd()));
#else
    in = new URANIE::Launcher::TInputFileRecreate(Form("%s/_output_multitype_mt_Key_condensate_.dat",gSystem->pwd()));
#endif
    in->setVectorProperties("[",",","]");
    in->addAttribute( tds1->getAttribute("w1") );
    in->addAttribute( tds1->getAttribute("v1") );
    in->addAttribute( tds1->getAttribute("v2") );
    in->addAttribute( tds1->getAttribute("f1") );
    in->addAttribute( tds1->getAttribute("w2") );

   _foutKey = new URANIE::Launcher::TOutputFileKey("_output_multitype_readmt_Key_.dat");
    _foutKey->addAttribute(new URANIE::DataServer::TAttribute("thev1") );
    _foutKey->addAttribute(new URANIE::DataServer::TAttribute("thev2") );
    tc1 = new URANIE::Launcher::TCode(tds1, UranieTests::code_path("multitype") + " -ReadmtKey");
    tc1->addInputFile( in );
    tc1->addOutputFile( _foutKey );

    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(false);
    tl1->run("nointermed");

    checkLauncherMultiTypeInputResults();
}

void TLauncherTest::checkLauncherMultiTypeInputResults()
{
    Int_t npat = tds1->getNPatterns();
    CPPUNIT_ASSERT_EQUAL(10, npat);
    // tds1->Scan("*");
    // CPPUNIT_ASSERT_EQUAL(6, tds1->getNAttributes());

    double expv1[10] = {0.2819049677902,-0.6923200061823,-123456789,-0.8192089693621,-1.561012830108,0.8843648237689,-123456789,1.633579305804,-123456789,-0.2397650482929};
    double expv2[10] = {11.42449011404,15.96653626442,12.62906549132,11.08628156181,-123456789,10.34077492413,7.150071600452,14.53735777922,-123456789,9.321006265935};
        
    // Note: as no order can be expected in a TDS1 computed with multi-threads or MPI,
    // it is impossible to use
    //    tds1->getValue("name", index)
    // as it would return the index-th line, but not the index-th value of the TDS1.
    double *thev1, *thev2;
    long nbData;
    cout.precision(15);
    URANIE::DataServer::TDSNtupleD *tdsTuple = tds1->GetTuple();
    for (int i = 0; i < 10; i++)
    {
        nbData = (long) tdsTuple->Draw("thev1:thev2", Form("foo__n__iter__==%d", i),"goff");
	CPPUNIT_ASSERT( (nbData>=0) );
        thev1 = tdsTuple->GetV1();
	thev2 = tdsTuple->GetV2();
        CPPUNIT_ASSERT_DOUBLES_EQUAL(expv1[i], thev1[0], DBL_DBL_PRECISION_COMPARE);
	CPPUNIT_ASSERT_DOUBLES_EQUAL(expv2[i], thev2[0], DBL_DBL_PRECISION_COMPARE);
    }

    tds1->getAttribute("thev1")->delShare();
    tds1->getAttribute("thev2")->delShare();
}

void TLauncherTest::testLauncherFormattingMultiTypeInputFile()
{
#ifdef WIN32
    URANIE_IGNORED_TEST() ; // No equivalent of sed in windows... Removed for this specific case
#endif
    
    tds1 = new URANIE::DataServer::TDataServer("foo", "multitype usecase");
    gSystem->Exec("sed '7,15d' readmultitype_sampling.dat > readmultitype_sampling_1.dat");
    tds1->fileDataRead("readmultitype_sampling_1.dat");
    
    tds1->getAttribute("w1")->setFileKey("_output_multitype_mt_DataServer_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewTDS);
    tds1->getAttribute("v1")->setFileKey("_output_multitype_mt_DataServer_.dat","", "Foo%g",URANIE::DataServer::TAttributeFileKey::kNewTDS);
    tds1->getAttribute("v2")->setFileKey("_output_multitype_mt_DataServer_.dat","", "%i",URANIE::DataServer::TAttributeFileKey::kNewTDS);
    tds1->getAttribute("f1")->setFileKey("_output_multitype_mt_DataServer_.dat","", "TurluTutu_%g",URANIE::DataServer::TAttributeFileKey::kNewTDS);
    tds1->getAttribute("w2")->setFileKey("_output_multitype_mt_DataServer_.dat","", "%e",URANIE::DataServer::TAttributeFileKey::kNewTDS);    

    _foutDataServer = new URANIE::Launcher::TOutputFileDataServer("_output_multitype_readmt_DataServer_.dat");
    URANIE::DataServer::TAttribute *thev1 = new URANIE::DataServer::TAttribute("thev1"); thev1->delShare();
    URANIE::DataServer::TAttribute *thev2 = new URANIE::DataServer::TAttribute("thev2"); thev2->delShare();
    _foutDataServer->addAttribute( thev1 );
    _foutDataServer->addAttribute( thev2 );
    
    tc1 = new URANIE::Launcher::TCode(tds1, Form("%s/%s", _workingTestPath.Data(),"../../teST/coDE/multiTYPE/multitype -ReadmtDS"));
    tc1->addOutputFile( _foutDataServer );

    tlbs = new URANIE::Launcher::TLauncherByStep(tds1, tc1);
    tlbs->setDrawProgressBar(false);
    tlbs->preTreatment();
    
    Int_t nret = UranieTests::areFilesEqual(TString("output_multitype_mt_formatted.dat.ref"),
                                            TString("URANIE/UranieLauncher_1/_output_multitype_mt_DataServer_.dat"));
    
    CPPUNIT_ASSERT_EQUAL(nret, 0);
    UranieTests::removeFile(_workingTestPath.Data(),"readmultitype_sampling_1.dat");
    UranieTests::removeFile(_workingTestPath.Data(),"URANIE/UranieLauncher_1/_output_multitype_mt_DataServer_.dat");

}


////////////////////////////////
///                          ///
///       Exception          ///
///                          ///
////////////////////////////////

void TLauncherTest::throwExceptionCodeAttributeWithSeveralTOutputFiles()
{
    TString sOutputFileName = TString("_output_flowrate_withKey_.dat");
    _foutKey = new URANIE::Launcher::TOutputFileKey(sOutputFileName);
    URANIE::DataServer::TAttribute *atty = new URANIE::DataServer::TAttribute(
	"y");
    atty->setFileKey(sOutputFileName, "flowrate");
    atty->setFileKey("tata.out", "titi");

    CPPUNIT_ASSERT_THROW(_foutKey->addAttribute(atty),
			 URANIE::Exceptions::UErrorExceptions);

    // Clean
    delete atty;
    atty = NULL;
}

void TLauncherTest::throwExceptionCodeAttributeWithDifferentNames()
{

    TString sOutputFileName = TString("_output_flowrate_withKey_.dat");
    _foutKey = new URANIE::Launcher::TOutputFileKey(sOutputFileName);
    URANIE::DataServer::TAttribute *atty = new URANIE::DataServer::TAttribute(
	"y");
    atty->setFileKey("tata.out", "flowrate");

    CPPUNIT_ASSERT_THROW(_foutKey->addAttribute(atty),
			 URANIE::Exceptions::UErrorExceptions);

    //  tc1->addOutputFile( _foutKey );

    // Clean
    delete atty;
    atty = NULL;
}

void TLauncherTest::throwExceptionTCodeWithoutOutputFile()
{
    tds1 = new URANIE::DataServer::TDataServer("tdsFlowrate",
					       "Design of experiments for Flowrate");
    tds1->fileDataRead(FLOWRATEASCIIFILEUNIFORMDESIGN);

    _foutDataServer = new URANIE::Launcher::TOutputFileDataServer(
	"_output_flowrate_withRow_.dat");
    _foutDataServer->addAttribute(ptrHat);

#ifdef WIN32
    tc1 = new URANIE::Launcher::TCode(tds1, "flowrate.exe -s -k");
#else
    tc1 = new URANIE::Launcher::TCode(tds1, "flowrate -s -k");
#endif
    tl1 = new URANIE::Launcher::TLauncher(tds1, tc1);
    tl1->setDrawProgressBar(kFALSE);

    CPPUNIT_ASSERT_THROW(tl1->run("nointermed"), URANIE::Exceptions::UErrorExceptions);
}

#ifdef WITH_LIBXML2
void TLauncherTest::validateTDS4ValueYhat()
{
    CPPUNIT_ASSERT(tds4->isAttribute("yhat"));
    Int_t npat = tds4->getNPatterns();
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 26.18019, tds4->getValue("yhat",0), DBL_DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 157.5910, tds4->getValue("yhat",npat-1), DBL_DBL_PRECISION_COMPARE);
}
void TLauncherTest::validateTDS4ValueD()
{
    CPPUNIT_ASSERT(tds4->isAttribute("d"));
    Int_t npat = tds4->getNPatterns();
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 3602.045, tds4->getValue("d",0), DBL_DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1348.649, tds4->getValue("d",npat-1), DBL_DBL_PRECISION_COMPARE);
}
#endif

#ifdef WITH_LIBXML2 
#ifdef WITH_LIBXSLT
void TLauncherTest::testOutputFileXML_yhat()
{
    setUp4();

    _foutXML->addAttribute(ptrHat, "/steady_state[@name='flowrate']/parameter/value/double", URANIE::DataServer::TAttributeFileKey::kXMLField);
    tc4->addOutputFile(_foutXML);

    tl4 = new URANIE::Launcher::TLauncher(tds4, tc4);
    tl4->setDrawProgressBar(kFALSE);
    // tl4->setWorkingDirectory("_yhat");
    // tl4->setSave(-1);
    tl4->run("nointermed");
    validateTDS4ValueYhat();
}
#endif
#endif

#ifdef WITH_LIBXML2
#ifdef WITH_LIBXSLT
void TLauncherTest::testOutputFileXML_d()
{
    setUp4();

    _foutXML->addAttribute(ptrD, "/steady_state[@name='flowrate']/distance/@value", URANIE::DataServer::TAttributeFileKey::kXMLAttribute);
    tc4->addOutputFile(_foutXML);

    tl4 = new URANIE::Launcher::TLauncher(tds4, tc4);
    tl4->setDrawProgressBar(kFALSE);
    tl4->run("nointermed");
    validateTDS4ValueD();
}
#endif
#endif

#ifdef WITH_LIBXML2
#ifdef WITH_LIBXSLT
void TLauncherTest::testOutputFileXML_yhat_d()
{
    setUp4();

    _foutXML->addAttribute(ptrHat, "/steady_state[@name='flowrate']/parameter/value/double", URANIE::DataServer::TAttributeFileKey::kXMLField);
    _foutXML->addAttribute(ptrD, "/steady_state[@name='flowrate']/distance/@value", URANIE::DataServer::TAttributeFileKey::kXMLAttribute);
    tc4->addOutputFile(_foutXML);

    tl4 = new URANIE::Launcher::TLauncher(tds4, tc4);
    tl4->setDrawProgressBar(kFALSE);
    tl4->run("nointermed");
    validateTDS4ValueYhat();
    validateTDS4ValueD();
}
#endif
#endif

#ifdef WITH_LIBXML2
void TLauncherTest::testOutputFileXML_yhat_d_with_keys_input()
{
    tds3 = new URANIE::DataServer::TDataServer("tdsFlowrate",
					       "Design of experiments for Flowrate");
    tds3->fileDataRead(FLOWRATEASCIIFILEUNIFORMDESIGN);
    TString sJDD3 = gSystem->pwd();
#ifdef WIN32
    sJDD3 += TString("\\flowrate_input_with_keys.in");
#else
    sJDD3 += TString("/flowrate_input_with_keys.in");
#endif

    // Sets the reference input file and the key for each input attributes
    tds3->getAttribute("rw")->setFileKey(sJDD3, "Rw");
    tds3->getAttribute("r")->setFileKey(sJDD3, "R");
    tds3->getAttribute("tu")->setFileKey(sJDD3, "Tu");
    tds3->getAttribute("tl")->setFileKey(sJDD3, "Tl");
    tds3->getAttribute("hu")->setFileKey(sJDD3, "Hu");
    tds3->getAttribute("hl")->setFileKey(sJDD3, "Hl");
    tds3->getAttribute("l")->setFileKey(sJDD3, "L");
    tds3->getAttribute("kw")->setFileKey(sJDD3, "Kw");


    _foutXML = new URANIE::Launcher::TOutputFileXML("_output_flowrate_withXML_.dat");
    _foutXML->addAttribute(ptrHat, "/steady_state[@name='flowrate']/parameter/value/double", URANIE::DataServer::TAttributeFileKey::kXMLField);
    _foutXML->addAttribute(ptrD, "/steady_state[@name='flowrate']/distance/@value", URANIE::DataServer::TAttributeFileKey::kXMLAttribute);

    tc3 = new URANIE::Launcher::TCode(tds3, UranieTests::code_path("flowrate") + " -s -k");
    tc3->addOutputFile(_foutXML);

    tl3 = new URANIE::Launcher::TLauncher(tds3, tc3);
    tl3->setDrawProgressBar(kFALSE);
#ifdef WIN32
    tl3->setWorkingDirectory(gSystem->pwd() + TString("\\tmpLanceurUranie\\flowrate"));
#else
    tl3->setWorkingDirectory(
	gSystem->pwd() + TString("/tmpLanceurUranie/flowrate"));
#endif

    tl3->run("nointermed");
   
    CPPUNIT_ASSERT(tds3->isAttribute("yhat"));
    Int_t npat = tds3->getNPatterns();
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 26.18019, tds3->getValue("yhat",0), DBL_DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 157.5910, tds3->getValue("yhat",npat-1), DBL_DBL_PRECISION_COMPARE);

    CPPUNIT_ASSERT(tds3->isAttribute("d"));
    npat = tds3->getNPatterns();
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 3602.045, tds3->getValue("d",0), DBL_DBL_PRECISION_COMPARE);
    CPPUNIT_ASSERT_DOUBLES_EQUAL( 1348.649, tds3->getValue("d",npat-1), DBL_DBL_PRECISION_COMPARE);
}
#endif


#ifdef WITH_LIBXML2
#ifdef WITH_LIBXSLT
void TLauncherTest::testOutputFileXML_d_yhat()
{
    setUp4();

    _foutXML->addAttribute(ptrD, "/steady_state[@name='flowrate']/distance/@value", URANIE::DataServer::TAttributeFileKey::kXMLAttribute);
    _foutXML->addAttribute(ptrHat, "/steady_state[@name='flowrate']/parameter/value/double", URANIE::DataServer::TAttributeFileKey::kXMLField);
    tc4->addOutputFile(_foutXML);

    tl4 = new URANIE::Launcher::TLauncher(tds4, tc4);
    tl4->setDrawProgressBar(kFALSE);
    tl4->run("nointermed");
    validateTDS4ValueD();
    validateTDS4ValueYhat();
}
#endif
#endif

#ifdef WITH_LIBXML2
#ifdef WITH_LIBXSLT
void TLauncherTest::testOutputFileDataServer_yhat_XML_d()
{
    setUp4();

    _foutDataServer = new URANIE::Launcher::TOutputFileDataServer("_output_flowrate_withRow_.dat");
    _foutDataServer->addAttribute(ptrHat);

    _foutXML->addAttribute(ptrD, "/steady_state[@name='flowrate']/distance/@value", URANIE::DataServer::TAttributeFileKey::kXMLAttribute);
    tc4->addOutputFile( _foutXML );
    tc4->addOutputFile( _foutDataServer );

    tl4 = new URANIE::Launcher::TLauncher(tds4, tc4);
    tl4->setDrawProgressBar(kFALSE);
    tl4->run("nointermed");
    validateTDS4ValueYhat();
    validateTDS4ValueD();
}
#endif
#endif

#ifdef WITH_LIBXML2
void TLauncherTest::throwExceptionTwoOutputFilesForOneAttribute()
{
    _foutDataServer = new URANIE::Launcher::TOutputFileDataServer("_output_flowrate_withRow_.dat");
    _foutDataServer->addAttribute(ptrHat);

    _foutXML = new URANIE::Launcher::TOutputFileXML("_output_flowrate_withXML_.dat");
    _foutXML->addAttribute(ptrHat, "/steady_state[@name='flowrate']/parameter/value/double",URANIE::DataServer::TAttributeFileKey::kXMLField);
}
#endif

#ifdef WITH_LIBXML2
void TLauncherTest::throwExceptionOutputFileTypeUnauthorized()
{
    _foutXML = new URANIE::Launcher::TOutputFileXML("_output_flowrate_withXML_.dat");
    _foutXML->addAttribute(ptrHat, "/steady_state[@name='flowrate']/parameter/value/double", URANIE::DataServer::TAttributeFileKey::kNewRow);
}
#endif

#ifdef WITH_LIBXML2
#ifdef WITH_LIBXSLT
void TLauncherTest::throwExceptionOutputFileXMLNotFound()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_IGNORED_TEST();

    setUp4();
    delete _foutXML;
    _foutXML = new URANIE::Launcher::TOutputFileXML("not_found.xml");
    _foutXML->addAttribute(ptrHat, "/steady_state[@name='flowrate']/parameter/value/double", URANIE::DataServer::TAttributeFileKey::kXMLField);
    tc4->addOutputFile(_foutXML);

    tl4 = new URANIE::Launcher::TLauncher(tds4, tc4);
    tl4->setDrawProgressBar(kFALSE);
    CPPUNIT_ASSERT_THROW( tl4->run("nointermed"), URANIE::Exceptions::UErrorExceptions );
}
#endif
#endif

#ifdef WITH_LIBXML2
#ifdef WITH_LIBXSLT
void TLauncherTest::throwExceptionOutputFileXMLXPathNotFound()
{
    // THIS TEST HAS BEEN DISABLED
    // TODO: fix me
    URANIE_IGNORED_TEST();

    setUp4();
    _foutXML->addAttribute(ptrHat, "/steady_state[@name='flowrate']/parameter/double", URANIE::DataServer::TAttributeFileKey::kXMLField);
    tc4->addOutputFile(_foutXML);

    tl4 = new URANIE::Launcher::TLauncher(tds4, tc4);
    tl4->setDrawProgressBar(kFALSE);

    CPPUNIT_ASSERT_THROW( tl4->run("nointermed"), URANIE::Exceptions::UErrorExceptions );

}
#endif
#endif


void TLauncherTest::throwExceptionCodeFileSetPointSeparator()
{
    tds1 = new URANIE::DataServer::TDataServer("tds1", "my title 1");
    tds1->fileDataRead(FLOWRATEASCIIFILEUNIFORMDESIGN);

#ifdef WIN32
    TString sJDD = gSystem->pwd()+TString("\\flowrate_input_with_flags.in");
#else
    TString sJDD = gSystem->pwd() + TString("/flowrate_input_with_flags.in");
#endif
    tds1->getAttribute("rw")->setFileFlag(sJDD, "@rw@", "%f");
    tds1->getAttribute("r")->setFileFlag(sJDD, "@r@", "%f");
    tds1->getAttribute("tu")->setFileFlag(sJDD, "@tu@", "%f");
    tds1->getAttribute("tl")->setFileFlag(sJDD, "@tl@", "%f");
    tds1->getAttribute("hu")->setFileFlag(sJDD, "@hu@", "%f");
    tds1->getAttribute("hl")->setFileFlag(sJDD, "@hl@", "%f");
    tds1->getAttribute("l")->setFileFlag(sJDD, "@l@", "%f");
    tds1->getAttribute("kw")->setFileFlag(sJDD, "@kw@", "%f");

    _foutRow = new URANIE::Launcher::TOutputFileRow(
	"_output_flowrate_withRow_.dat");
    _foutRow->addAttribute(ptrHat);

    URANIE::Launcher::TCodeFile *fin = new URANIE::Launcher::TCodeFile(sJDD);
    CPPUNIT_ASSERT_THROW( fin->setSeparatorCharacter("_'. "), URANIE::Exceptions::UErrorExceptions );

    CPPUNIT_ASSERT(_foutRow->isAttribute(TString("yhat")));
    CPPUNIT_ASSERT(_foutRow->isAttribute(TString("dgrt")) == kFALSE);

    delete fin;

}

void TLauncherTest::throwExceptionInputFileNotFound()
{
    cout << endl << " ********************" << endl;
    cout << " ****" << endl;
    cout << " **** SetUP1" << endl;
    cout << " ****" << endl;
    tds1 = new URANIE::DataServer::TDataServer("tds1", "my title 1");
    tds1->fileDataRead(FLOWRATEASCIIFILEUNIFORMDESIGN);
    // The reference input file
    TString sJDD1 = gSystem->pwd();
#ifdef WIN32
    sJDD1 += TString("\\Notflowrate_input_with_keys.in");
#else
    sJDD1 += TString("/Notflowrate_input_with_keys.in");
#endif

    tds1->getAttribute("rw")->setFileKey(sJDD1, "Rw", "%f");
    tds1->getAttribute("r")->setFileKey(sJDD1, "R", "%f");
    tds1->getAttribute("tu")->setFileKey(sJDD1, "Tu", "%f");
    tds1->getAttribute("tl")->setFileKey(sJDD1, "Tl", "%f");
    tds1->getAttribute("hu")->setFileKey(sJDD1, "Hu", "%f");
    tds1->getAttribute("hl")->setFileKey(sJDD1, "Hl", "%f");
    tds1->getAttribute("l")->setFileKey(sJDD1, "L", "%f");
    tds1->getAttribute("kw")->setFileKey(sJDD1, "Kw", "%f");
#ifdef WIN32
    tc1 = new URANIE::Launcher::TCode(tds1, "flowrate.exe                                                                                                                 -s                                                                                                                 -k");
#else
    tc1 = new URANIE::Launcher::TCode(tds1,
				      Form("%s/%s", _workingTestPath.Data(),
					   "../../teST/coDE/flowREATE/flowrate -s -k"));
#endif
    cout << " ****" << endl;
    cout << " **** End Of SetUp1" << endl;
    cout << " ****" << endl;
    cout << " ********************" << endl;

#ifndef OLDCLUB
    CPPUNIT_ASSERT_THROW(new URANIE::Launcher::TLauncher(tds1, tc1) , TamponTexteError);
#endif
}


void TLauncherTest::throwExceptionInputAndOutputAttributeWithSameName()
{
    
    tds1 = new URANIE::DataServer::TDataServer("tdsFlowrate", "Design of experiments for Flowrate");
  
    // Add the study attributes ( min, max and nominal values)
    tds1->addAttribute( new URANIE::DataServer::TUniformDistribution("rw", 0.05, 0.15));
    tds1->getAttribute("rw")->delShare();
    
    // The reference input file
    TString sJDD1 = gSystem->pwd();
#ifdef WIN32
    sJDD1 += TString("\\flowrate_input_with_keys.in");
#else
    sJDD1 += TString("/flowrate_input_with_keys.in");
#endif
   
    // Set the reference input file and the key for each input attributes
    tds1->getAttribute("rw")->setFileKey(sJDD1, "Rw");
    
    // The output file of the code
    _foutKey = new URANIE::Launcher::TOutputFileKey("_output_flowrate_withKey_.dat");
    // The attribute in the output file (with the same name than the input one
    URANIE::DataServer::TAttribute *badOut = new URANIE::DataServer::TAttribute("rw");
    badOut->delShare();
    _foutKey->addAttribute(badOut); 
    
    
    // define the code
#ifdef WIN32
    tc1 = new URANIE::Launcher::TCode(tds1, "flowrate.exe -s -k");
#else
    tc1 = new URANIE::Launcher::TCode(tds1,
            Form("%s/%s", _workingTestPath.Data(),
                    "../../teST/coDE/flowREATE/flowrate -s -k"));
#endif   
    
    CPPUNIT_ASSERT_THROW(tc1->addOutputFile( _foutKey ) , URANIE::Exceptions::UErrorExceptions);
    
}


//! End Of File
